/*************************************************************************************
 Error.c
 The file deals with Pedigree errors handlings.
 An error is a struct which includes an error string, that is presented to the user.
*************************************************************************************/
#include "pc_Error.h"
#include "pc_Person.h"
#include "../globals.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>


#define UNRESOLVED_ERROR_STR "An unresolved error occured in person. Please contact PedTool staff at the Technion"
#define UNRESOLVED_ERROR_ID "1000"

ckrResult AddError(ErrorLL * aErrorLL, char * aErrorStr);
ckrResult AddErrorInPerson(ErrorLL * aErrorLL, int aMarkerIdx, char aPersonID[MAX_PERSON_ID_LEN], char * aErrorStr);
ckrResult AddErrorInNuclearFam(ErrorLL * aErrorLL, int aMarkerIdx, NuclearFam * aNuclearFam, char * aErrorStr);
ckrResult AddErrorInPersonLine(ErrorLL * aErrorLL, char aPersonID[MAX_PERSON_ID_LEN], char * aErrorStr);

/***************************************************************************************
Function:CreateErrorLL
This function creates a new errorLL and returns it, after being initializing it.
****************************************************************************************/

ErrorLL * CreateErrorLL()
{
	ErrorLL * errorLL = NULL;

	errorLL = (ErrorLL *)malloc(sizeof(ErrorLL));
	if (NULL == errorLL)
	{
		return NULL;
	}

	/* go to function InitErrorLL */
	InitErrorLL(errorLL);
	return errorLL;
}

/*************************************************************************************
Function:InitErrorLL
This function recieves aErrorLL and initializes it.
/************************************************************************************/
void InitErrorLL(ErrorLL * aErrorLL)
{
	aErrorLL->mFirst = NULL;
	aErrorLL->mLast  = NULL;
	aErrorLL->numOfErrors = 0;
}

/*************************************************************************************
Function:FreeErrorLL
This function recieves aErrorLL and frees it.
Author: Anna Tzemach
/************************************************************************************/

void FreeErrorLL(ErrorLL * aErrorLL){
	 Error* next, *current;
	 
	 current = aErrorLL->mFirst;
	 while  (current != NULL){
		 next = current->mNext;
		 free(current);
		 current = next;
	 }
  	
}

/**************************************************************************************
Function:AddError
This function recieves aErrorLL and an error message.
The function returns ckrOK if the new error message was added successfully to aErrorLL.
**************************************************************************************/
ckrResult AddError(ErrorLL * aErrorLL, char * aErrorStr)
{
	Error * error;

	error = NULL;
	error = (Error *)malloc(sizeof(Error));
	if (NULL == error)
	{
		return ckrMemErr;
	}

	/* init the error's fields */
	strcpy(error->errorString, aErrorStr);
	error->mNext = NULL;

	/* increase num of errors */
	aErrorLL->numOfErrors++;

	/* error list is empty */
	if (NULL == aErrorLL->mFirst)
	{
		aErrorLL->mFirst = error;
		aErrorLL->mLast  = error;
		return ckrOK;
	}

	/* list is not empty. Add error to errors list */
	aErrorLL->mLast->mNext = error;
	aErrorLL->mLast = error;
	return ckrOK;
}

/*************************************************************************************************
Function:AddErrorInPerson
This function recieves aErrorLL, aPersonID and error message.
The function returns ckrOK if the new error message was added successfully .
**************************************************************************************************/
ckrResult AddErrorInPerson(ErrorLL * aErrorLL, 
			   int aMarkerIdx, 
			  char aPersonID[MAX_PERSON_ID_LEN], 
			  //int aPersonID;
			   char * aErrorStr)
{
	char error[MAX_ERROR_ID_LEN];
	char markerIdx[MAX_ERROR_ID_LEN];
	memset(error, 0, MAX_ERROR_ID_LEN);
	memset(markerIdx, 0, MAX_ERROR_ID_LEN);

	strcpy(error, "Error in person ");
	strcat(error, aPersonID);
	if (UNDEFINED_MARKER_IDX != aMarkerIdx)
	{
		strcat(error, ", marker ");
		MyDecimalItoa(aMarkerIdx + 1, markerIdx, MAX_ERROR_ID_LEN);
		strcat(error, markerIdx);
	}
	strcat(error, ": ");
	strcat(error, aErrorStr);

	return AddError(aErrorLL, error);
}


/*************************************************************************************************
Function:AddErrorInNuclearFam
This function recieves aErrorLL, aNuclearFam and error message.
The function returns ckrOK if the new error message was added successfully to the NuclearFam.
**************************************************************************************************/

ckrResult AddErrorInNuclearFam(ErrorLL * aErrorLL, 
			       int aMarkerIdx, 
			       NuclearFam * aNuclearFam, 
			       char * aErrorStr)
{
	char error[MAX_ERROR_ID_LEN];
	char markerIdx[MAX_ERROR_ID_LEN];
	char allele[MAX_ALLELE_ID_LEN];
	Person * aChild = NULL;
	int aFirstChildFlag = 1, allele1, allele2;

	memset(error, 0, MAX_ERROR_ID_LEN);
	memset(markerIdx, 0, MAX_ERROR_ID_LEN);
	memset(allele, 0, MAX_ALLELE_ID_LEN);

	

	if (UNDEFINED_MARKER_IDX != aMarkerIdx)
	{
		strcpy(error,LociData[aMarkerIdx].locusName); // The array begin from zero..
		strcat(error, " - Mendelian error ");
		//MyDecimalItoa(aMarkerIdx + 1, markerIdx, MAX_ERROR_ID_LEN);
		//strcat(error, markerIdx);
	}
	else{
		return ckrErr;
	}

	strcat(error, ": ");
	strcat(error, aErrorStr);
	/*strcat(error, "\n");*/ strcat(error, "\t");

	/* father*/
	strcat(error, "father: ");
	strcat(error, "#");
	strcat(error, aNuclearFam->mFather->mID);
	strcat(error, "\t");
	GetMarkers(aNuclearFam->mFather, &allele1, &allele2, aMarkerIdx);
	MyDecimalItoa(allele1, allele, MAX_ALLELE_ID_LEN);
	strcat(error, allele);
	strcat(error, "/");
	MyDecimalItoa(allele2, allele, MAX_ALLELE_ID_LEN);
	strcat(error, allele);

	/*strcat(error, "\n");*/ strcat(error, "\t");
	
	/* mother*/
	strcat(error, "\t");
	strcat(error, "mother: ");
	strcat(error, "#");
	strcat(error, aNuclearFam->mMother->mID);
	strcat(error, "\t");
	GetMarkers(aNuclearFam->mMother, &allele1, &allele2, aMarkerIdx);
	MyDecimalItoa(allele1, allele, MAX_ALLELE_ID_LEN);
	strcat(error, allele);
	strcat(error, "/");
	MyDecimalItoa(allele2, allele, MAX_ALLELE_ID_LEN);
	strcat(error, allele);
	/*strcat(error, "\n");*/ strcat(error, "\t");strcat(error, "\t");
	
	strcat(error, "children: ");

	aChild = aNuclearFam->mChild;
	while (aChild != NULL)
	{
		strcat(error, "#");
		strcat(error, aChild->mID);
		strcat(error, "\t");
		GetMarkers(aChild, &allele1, &allele2, aMarkerIdx);
		MyDecimalItoa(allele1, allele, MAX_ALLELE_ID_LEN);
		strcat(error, allele);
		strcat(error, "/");
		MyDecimalItoa(allele2, allele, MAX_ALLELE_ID_LEN);
		strcat(error, allele);
		
		if (aChild->mProblem == '*'){
			strcat(error, "\t(*)");
			strcat(error, " ");
			aChild->mProblem = 40;
		}
		
		/*strcat(error, "\n");*/ strcat(error, "\t"); strcat(error, "\t");

		aChild = aChild->mSibling;
	}

	return AddError(aErrorLL, error);
}
/*************************************************************************************************
Function:AddErrorInPersonLine
This function recieves aErrorLL, aPersonID and error message.
The function returns ckrOK if the new error message was added successfully .
**************************************************************************************************/
ckrResult AddErrorInPersonLine(ErrorLL * aErrorLL, 
			   char aPersonID[MAX_PERSON_ID_LEN], 
			   char * aErrorStr)
{
	char error[MAX_ERROR_ID_LEN];
	memset(error, 0, MAX_ERROR_ID_LEN);

	strcpy(error, "Error in line of person ");
	strcat(error, aPersonID);
	strcat(error, ": ");
	strcat(error, aErrorStr);

	return AddError(aErrorLL, error);
}




/* handles person errors */

void SetErrorInPerson(ErrorLL * aErrorLL, int index, char aID[MAX_PERSON_ID_LEN], ErrorType error)

{
   switch(error)
   {

   case (NO_LEGAL_GENDER_ERROR):
	  AddErrorInPerson(aErrorLL, index, aID,"No legal input concerning the gender of the person.");
      break;
   
   case (NO_LEGAL_AFF_STATUS_ERROR):
	  AddErrorInPerson(aErrorLL, index, aID,"No legal input concerning the affected status of the person.");
      break;
   
   case (NOT_FULLY_TYPED_ERROR):
	  AddErrorInPerson(aErrorLL, index, aID, "A person is not fully typed.");
      break;
   
   case (DETAILS_TOO_LONG_ERROR):
      AddErrorInPerson(aErrorLL, index, aID, "A person detail is too long.");
	  break;

   case (ODD_ALLELE_NUM_ERROR):
	  AddErrorInPerson(aErrorLL, index, aID, "A person is half typed - in one of the markers, only one allele is found.");
      break;

  case (PARSING_MARKER_ERROR):
	  AddErrorInPerson(aErrorLL, index, aID, "An error while parsing marker.");
	  break;

  case(ONE_ALLELE_UNKNOWN_ERROR):
	  AddErrorInPerson(aErrorLL, index, aID, "One allele of this marker is known and the other allele is not.");
	  break;

  case(INVALID_ALLELE_ERROR):
	  AddErrorInPerson(aErrorLL, index, aID, "Invalid allele identifier. An Allele can be a number only.");
	  break;

  case(CHILD_AND_PARENTS_NOT_COMPATIBLE_ERROR):
      AddErrorInPerson(aErrorLL, index, aID, "Child and parents alleles are not compatible.");
      break;

  case (PARENTS_ORDER_ERROR):
	  AddErrorInPerson(aErrorLL, index, aID, "Parents of the person are typed in a wrong order.");
      break; 
	  
  case (FATHER_OF_HIMSELF_ERROR):
	  AddErrorInPerson(aErrorLL, index, aID, "Id of the person is identical to his / her father\'s.");
      break; 

  case (MOTHER_OF_HIMSELF_ERROR):
	  AddErrorInPerson(aErrorLL, index, aID, "Id of the person is identical to his / her mother\'s.");
      break; 

  case (NO_FATHER_FOUND_ERROR):
	  AddErrorInPerson(aErrorLL, index, aID, "The father of this person does not exist in the pedigree.");
      break; 

  case (NO_MOTHER_FOUND_ERROR):
	  AddErrorInPerson(aErrorLL, index, aID, "The mother of this person does not exist in the pedigree.");
      break; 

  case (MAN_ALLELES_VALUE_ERROR):
      AddErrorInPerson(aErrorLL, index, aID, "In sexlinked format male persons should have only mothers alleles, thus it should be only one value of allele (or both equal).");
      break; 
  
  // we should not get to default
  default:
	  AddErrorInPerson(aErrorLL, UNDEFINED_MARKER_IDX, UNRESOLVED_ERROR_ID, UNRESOLVED_ERROR_STR);
      break; 
   }
  
}

void SetParentErrorInPerson(ErrorLL * aErrorLL, int index, char aID[MAX_PERSON_ID_LEN], char aParentID[MAX_PERSON_ID_LEN], ErrorType error)
{
	char  errorStr[MAX_ERROR_ID_LEN];

	switch(error)
   {
	case (FATHER_IS_FEMALE_ERROR):
		strcpy(errorStr,"Father of the person is declared as a female (father is ");
		strcat(errorStr,aParentID);   
		strcat(errorStr,")");
		AddErrorInPerson(aErrorLL, index, aID, errorStr);
		break;

	case (MOTHER_IS_MALE_ERROR):
		strcpy(errorStr,"Mother of the person is declared as a male (mother is ");
		strcat(errorStr,aParentID);   
		strcat(errorStr,")");
		AddErrorInPerson(aErrorLL, index, aID, errorStr);
		break;

	// we should not get to default
    default:
	  AddErrorInPerson(aErrorLL, UNDEFINED_MARKER_IDX, UNRESOLVED_ERROR_ID, UNRESOLVED_ERROR_STR);
      break; 
	}
}

/* general errors */
void SetError(ErrorLL * aErrorLL, ErrorType error)
{
   switch(error)
   {

   case(EMPTY_FAMILY_ERROR):
	  AddError(aErrorLL, "A Pedigree without members found");
	  break;

   case(NO_FAMILY_ID_ERROR):
	  AddError(aErrorLL, "The pedigree does not have an ID");
	  break;

   case(MARKERS_CNT_DIFF_IN_PERSONS_ERROR):
	  AddError(aErrorLL,"Number of markers is not equal in all persons of pedigree");
	  break;

   // we should not get to default
   default:
	  AddError(aErrorLL, UNRESOLVED_ERROR_STR);
      break; 

   }
}

/* handles nuclearfam errors */
void SetErrorInNuclearFam(ErrorLL * aErrorLL, int index, NuclearFam * aNFam, ErrorType error)
{

   switch(error)
   {
   case (MORE_THAN_2_HOMO_CHILDREN_ERROR):
     AddErrorInNuclearFam(aErrorLL, index, aNFam, "More than 2 Homozigous children with different alleles");
     break;
    
   case(MORE_THAN_4_IN_SIBSHIP_ERROR):
	   AddErrorInNuclearFam(aErrorLL, index, aNFam, "More than a total of 4 different alleles in the family\'s children");
       break;
   
   case(MORE_THAN_3_IN_SIBSHIP_ERROR):
	   AddErrorInNuclearFam(aErrorLL, index, aNFam, "More than a total of 3 different alleles in the family\'s children,  even though there is a homozygous child");
       break;
     
   case(MORE_THAN_2_IN_SIBSHIP_ERROR):
	   AddErrorInNuclearFam(aErrorLL, index, aNFam, "More than a total of 2 different alleles in the family\'s children, even though there are two different homozygous children");
       break;

   case(NO_POSSIBLE_GENOTYPING_ERROR):
	   AddErrorInNuclearFam(aErrorLL, index, aNFam, "No possible genotyping for one or more pedigree members");
	   break;	

   case(CHILD_AND_PARENTS_NOT_COMPATIBLE_ERROR):
	   AddErrorInNuclearFam(aErrorLL, index, aNFam, "Child's and parents' alleles are not compatible");
	   break;	
   
   case (MORE_THAN_3_IN_SEX_SIBSHIP_ERROR):
		AddErrorInNuclearFam(aErrorLL, index, aNFam, "More than a total of 3 different alleles in the family in sexlinked format");
		break; 

   case (MORE_THAN_1_HOMO_CHILDREN_SEX_ERROR):
		AddErrorInNuclearFam(aErrorLL, index, aNFam, "More than one Homozigous children with different alleles in nuclear family in sexlinked format");
		break; 
   case (MORE_THAN_2_IN_SIBSHIP_SEX_ERROR):
		   AddErrorInNuclearFam(aErrorLL, index, aNFam, "More than a total of 2 different alleles in the family\'s in sexlinked format children, even though there are a homozygous children");
		break; 
	// we should not get to default
    default:
	  AddErrorInNuclearFam(aErrorLL, UNDEFINED_MARKER_IDX, aNFam, UNRESOLVED_ERROR_STR);
      break; 
   }
}  

/* handles errors in the file's line */
void SetErrorInPersonLine(ErrorLL * aErrorLL, char aID[MAX_PERSON_ID_LEN], ErrorType error)
{
	switch(error)
	{
	case (NO_LEGAL_REMARK_ERROR):
	  AddErrorInPersonLine(aErrorLL, aID, "Syntax error - character \"<\" found. Did you mean to use the remark string \"<<\"?");
      break;

	// we should not get to default
    default:
	  AddErrorInPersonLine(aErrorLL, UNRESOLVED_ERROR_ID, UNRESOLVED_ERROR_STR);
      break; 
	}
}

