/*************************************************************************************
 LangeGoradia.c
 The file deals with Lange-Goradia algorithm functions.
*************************************************************************************/
#include <stdio.h>
#include <stdlib.h>

#include "pc_LangeGoradia.h"
#include "pc_Person.h"
#include "pc_Error.h"
#include "pc_FreeMemory.h"

ckrResult LangeGoradia(Pedigree * aPedigree, int aMarkerIdx);
void DeleteOptMarkerLLs(Pedigree * aPedigree);
ckrResult InitOptMarkerLLs(Pedigree * aPedigree, int aMarkerIdx);
ckrResult HandleLangGoradiaNuclearFam(NuclearFam * aNuclearFam, int aMarkerIdx, int * aIsFamChanged, Format aFormat);
ckrResult GetPossibleChildMarkers(MarkerLLNode * a,MarkerLLNode * b, MarkerLLNode ** c);
ckrResult GetPossibleMaleChildMarkers(MarkerLLNode * b, MarkerLLNode ** c);
//ckrResult MarkPossibleChildrenMarkers(Person * aFirstChild, MarkerLLNode * aPossibleChildMarkers);
ckrResult MarkPossibleChildrenMarkers (Person * aFirstChild,MarkerLLNode * aPossibleChildMarkers,
									   Format aFormat,MarkerLLNode * aPossibleChildMaleMarkers);
Result IsMarkerInList(Marker aElement, MarkerLLNode * aList);
void ExcludeNonMarkedMarkers(NuclearFam * aNuclearFam, int * aIsFamChanged);
void SetMarkersMarked(NuclearFam * aNuclearFam, int aMarkVal);
void SetPersonMarkersMarked(Person * aPerson, int aMarkVal);
ckrResult IsAllFamMembersHaveOptMarkers(NuclearFam * aNuclearFam);
void SetAllFamUnchecked(Pedigree * aPedigree);
int isAllPersonsUnknown(Pedigree * aPedigree, int aMarkerIdx);


/*************************************************************************************
Lange-Goradia algorithm as explained in class:
----------------------------------------------

A. For each pedigree member, save only ordered genotypes compatible with his/her 
   phenotype.

B. For each nuclear family:
  1. Consider each mother-father genotype pair:
    a. Determine which zygotes can arise from this pair.
	b. If each child in the nuclear family has one or more of these zygote genotypes
	   among his or her current genotype list, then save the parental genotypes and
	   any child genotype matching one of the created zygote genotypes.
	c. If any child has none of these zygote genotypes among his/her genotype list,
	   then don't save any genotypes.
  2. For each person in the nuclear family, exclude any genotypes not saved during 
     step (1).

C. Repeat part (B) until no more genotypes can be excluded.


During this file, functions will be documented by referencing those steps.
*************************************************************************************/


/*************************************************************************************
 Function: LangeGoradia
	The function runs the Lange-Goradia algorithm for a pedigree, for a specified 
	marker index, and returns whether the pedigree, for that marker index, is 
	genetically consistent or not
 input: Pedigree * aPedigree - the pedigree to check
	    int aMarkerIdx       - the checked marker index
 output: ckrOK - if Lange Goradia algorithm shows that the pedigree is consistent
		 ckrErr - if the pedigree is not consistent
		 ckrMemErr - in case of memory error			     
*************************************************************************************/
ckrResult LangeGoradia(Pedigree * aPedigree, int aMarkerIdx)
{
	int i;
	int aRunAllFamFlag;
	NuclearFam * aNuclearFam;
	ckrResult aRes;

	aNuclearFam = NULL;
	aRunAllFamFlag = 1;
	
	if (1 == isAllPersonsUnknown(aPedigree, aMarkerIdx))
	{
		return ckrOK;
	}

	/*step A - "For each pedigree member, save only ordered genotypes 
	  compatible with his/her phenotype.*/
	aRes = InitOptMarkerLLs(aPedigree, aMarkerIdx);
	RETURN_MEM_ERROR(aRes);

	/* step C - "Repeat part (B) until no more genotypes can be excluded" */
	while (1 == aRunAllFamFlag)
	{
		aRunAllFamFlag = 0;
		SetAllFamUnchecked(aPedigree);

		/* step B - "for each nuclear family:" */
		for (i = 0; i < aPedigree->mPersonCnt; ++i)
		{
			aNuclearFam = aPedigree->mPersonArray[i]->mParentsFam;
			if (NULL == aNuclearFam)
			{
				continue;
			}
			/* only if aNuclearFam really exists (not NULL), check if not checked already */
			if (1 == aNuclearFam->mIsChecked)
			{
				continue;
			}
			
			aNuclearFam->mIsChecked = 1;
			aRes = HandleLangGoradiaNuclearFam(aNuclearFam, aMarkerIdx, &aRunAllFamFlag, aPedigree->mFormat);
			RETURN_MEM_ERROR(aRes);

			/* no consistency in one or more family members */
			if (ckrErr == aRes)
			{
				SetErrorInNuclearFam(aPedigree->mErrorLL2, aMarkerIdx, aNuclearFam, NO_POSSIBLE_GENOTYPING_ERROR);
				DeleteOptMarkerLLs(aPedigree);
				return ckrErr;
			}
		}
		SetAllFamUnchecked(aPedigree);
	}
	DeleteOptMarkerLLs(aPedigree);
	return ckrOK;
}

/*************************************************************************************
 Function: CreateFullMarkerList
	The function creates a full marker list - it goes over all alleles that exist in 
	the pedigree, for the relevant marker index (were saved during parsing) and 
	creates a marker list that includes all combinations of those alleles. (With no 
	importance to alleles' order, meaning if alleles 3,4 exist, then only one of the 
	marker 3/4 or 4/3 will be in the list.
**************************************************************************************/
ckrResult CreateFullMarkerList(MarkerLLNode ** aOptMarkerList, 
							   AlleleLLNode * aMarkerPedAlleles)
{
	ckrResult res = ckrOK;
	AlleleLLNode * iAllele1 = NULL;
	AlleleLLNode * iAllele2 = NULL;

	*aOptMarkerList = NULL;
	iAllele1 = aMarkerPedAlleles;

	
	while (NULL != iAllele1)
	{
		iAllele2 = aMarkerPedAlleles;
		
		while ((NULL != iAllele2) && (iAllele2 != iAllele1))
		{
			res = CreateAddMarker(iAllele1->mAllele, iAllele2->mAllele, aOptMarkerList);
			RETURN_MEM_ERROR(res);

			iAllele2 = iAllele2->mNext;
		}

		if (iAllele2 == iAllele1)
		{
			res = CreateAddMarker(iAllele1->mAllele, iAllele2->mAllele, aOptMarkerList);
			RETURN_MEM_ERROR(res);
		}

		iAllele1 = iAllele1->mNext;
	}
	return ckrOK;
}

/*************************************************************************************
 Function: CreateFullMaleMarkerList
	The function creates a full marker list - it goes over all alleles that exist in 
	the pedigree, for the relevant marker index (were saved during parsing) and 
	creates a marker list that includes all combinations of those alleles. (With no 
	importance to alleles' order, meaning if alleles 3,4 exist, then only one of the 
	marker 3/4 or 4/3 will be in the list.
**************************************************************************************/
ckrResult CreateFullMaleMarkerList(MarkerLLNode ** aOptMarkerList, 
							   AlleleLLNode * aMarkerPedAlleles)
{
	ckrResult res = ckrOK;
	AlleleLLNode * iAllele1 = NULL;

	*aOptMarkerList = NULL;
	iAllele1 = aMarkerPedAlleles;

	while (NULL != iAllele1)
	{
		res = CreateAddMarker(iAllele1->mAllele, iAllele1->mAllele, aOptMarkerList);
		RETURN_MEM_ERROR(res);
		iAllele1 = iAllele1->mNext;
	}

	return ckrOK;
}
/*************************************************************************************
 Function: DeleteOptMarkerLLs
	The function deletes the optional markers linked-list that was saved for each 
	person in the pedigree
 Input: Pedigree * aPedigree - the pedigree to delete its member's optional markers
		linked-lists
**************************************************************************************/
void DeleteOptMarkerLLs(Pedigree * aPedigree)
{
	int i;
	Person * person;

	for (i = 0; i < aPedigree->mPersonCnt; ++i)
	{
		person = aPedigree->mPersonArray[i];
		DeleteMarkerLL(&(person->mOptMarkerLL));
	}
}


/*************************************************************************************
 Function: InitOptMarkerLLs
	The function initializes the optional markers lists, for all members of the 
	pedigree. The optional linked list (the field mOptMarkerLL) holds all the optional
	combinations of marker that person might have.
	In initalization, for people whom their markers are known, the optional list will
	include only the marker they own. For people whom their markers are not known, the
	optional linked list will include all the combinations of alleles for this marker,
	which is, all combinations of the alleles for this marker that exist in the 
	pedigree
 Input: Pedigree * aPedigree - the pedigree we work on
        int aMarkerIdx       - the marker index
**************************************************************************************/
ckrResult InitOptMarkerLLs(Pedigree * aPedigree, int aMarkerIdx)
{
	int i;
	Person * person;
	ckrResult res;
	AlleleLL * allAlleles = NULL;

	GetAlleleLLFromPedAlleles(&allAlleles, aPedigree->mPedAlleles, aMarkerIdx);

	
	for (i = 0; i < aPedigree->mPersonCnt; ++i)
	{
		person = aPedigree->mPersonArray[i];
		/* a person's marker array is valid here. Which means that if one allele
		   is 0, then the other allele must also be 0                           */

		/* ==my two alleles are zero*/
		if ((0 == person->mMarkerArr[aMarkerIdx].mAllele1) && (0 == person->mMarkerArr[aMarkerIdx].mAllele2))
		{
			/* ==my  deside which kind of full list to build*/
			if ((SEXLINKED == aPedigree->mFormat) && (person->mGender == male) )
				res = CreateFullMaleMarkerList(&(person->mOptMarkerLL), allAlleles->mFirstLLNode);
			else
				res = CreateFullMarkerList(&(person->mOptMarkerLL), allAlleles->mFirstLLNode);
		}
		else
		{
			res = CreateAddMarker(person->mMarkerArr[aMarkerIdx].mAllele1, 
								  person->mMarkerArr[aMarkerIdx].mAllele2, 
								  &(person->mOptMarkerLL));
		}

	}
	return ckrOK;
}

/*************************************************************************************
 Function: isAllPersonsUnknown
	The function returns 1 if all person in the pedigree are unknown for marker 
	aMarkerIdx.
	0 else
 Input: Pedigree * aPedigree - the pedigree we work on
        int aMarkerIdx       - the marker index
**************************************************************************************/
int isAllPersonsUnknown(Pedigree * aPedigree, int aMarkerIdx)
{
	int i;
	Person * person;

	for (i = 0; i < aPedigree->mPersonCnt; ++i)
	{
		person = aPedigree->mPersonArray[i];
		/* a person's marker array is valid here. Which means that if one allele
		   is 0, then the other allele must also be 0                           */
		/* my changed to 2 markers check*/
		if ((0 != person->mMarkerArr[aMarkerIdx].mAllele1) && (0 != person->mMarkerArr[aMarkerIdx].mAllele2) )
		{
			return 0;
		}
	}
	return 1;
}

/*************************************************************************************
 Function: HandleLangGoradiaNuclearFam
	The function handles the Lange-Goradia algorithm for a single nuclear family. 
	Meaning, consistency is checked on the given nuclear family, in the way specified 
	in the algorithm.
 Input: NuclearFam * aNuclearFam - the nuclear family to check
		int aMarkerIdx           - the checked marker index
		Format aFormat			 - the format of pedigree (AUTOSOMAL or SEXLINKED)
output: ckrOK - if Lange Goradia algorithm shows that this nuclear family is consistent
		ckrErr - if the nuclear family is not consistent
		ckrMemErr - in case of memory error			     
**************************************************************************************/
ckrResult HandleLangGoradiaNuclearFam(NuclearFam * aNuclearFam, 
									  int aMarkerIdx, 
									  int * aIsFamChanged, 
									  Format aFormat)
{
	int res;

	MarkerLLNode * motherMarker;
	MarkerLLNode * fatherMarker;
	MarkerLLNode * possibleChildMarkers = NULL;
	MarkerLLNode * possibleMaleChildMarkers = NULL;

	motherMarker = aNuclearFam->mMother->mOptMarkerLL;
	fatherMarker = aNuclearFam->mFather->mOptMarkerLL;

	/*step B.1 - Consider each mother-father genotype pair:*/
	while (NULL != motherMarker)
	{
		while (NULL != fatherMarker)
		{
			possibleChildMarkers=NULL;

			/*step B.1.a - Determine which zygotes can arise from this pair. */
			res = GetPossibleChildMarkers(fatherMarker, motherMarker, &possibleChildMarkers);
			RETURN_MEM_ERROR(res);
			
		
			/* ==my  take all possibilities for male in sexlinked format*/
			if ( SEXLINKED == aFormat) 
			{
				res = GetPossibleMaleChildMarkers(motherMarker, &possibleMaleChildMarkers);
				RETURN_MEM_ERROR(res);
			}


			/* Step B.1.b - If each child in the nuclear family has one or more of 
			   these zygote genotypes among his or her current genotype list.....*/
			
			if (YES == MarkPossibleChildrenMarkers(aNuclearFam->mChild, possibleChildMarkers, aFormat, possibleMaleChildMarkers))
			{
				/*... then save the parental genotypes....*/
				fatherMarker->mIsMarked = 1;
				motherMarker->mIsMarked = 1;
				/* ... and any child genotype matching one of the created zygote 
			   genotypes. (was done inside function MarkPossibleChildrenMarkers) */
			}
			
			fatherMarker = fatherMarker->mNext;
		}
		//pc_freeMarkerLLNode(possibleChildMarkers);
		//pc_freeMarkerLLNode(possibleMaleChildMarkers);
		fatherMarker = aNuclearFam->mFather->mOptMarkerLL;
		motherMarker = motherMarker->mNext;
	}
	
	/* Step B.2 - For each person in the nuclear family, exclude any 
	   genotypes not saved during step (1).*/
	ExcludeNonMarkedMarkers(aNuclearFam, aIsFamChanged);
	SetMarkersMarked(aNuclearFam, 0);
	
	return IsAllFamMembersHaveOptMarkers(aNuclearFam);
}

/*************************************************************************************
 Function: MarkPossibleChildrenMarkers
	The function checks if all children in siblinghood have at least one marker included
	in the given markers list.
 Input: Person * aFirstChild - first child in siblinghood
		MarkerLLNode * aPossibleChildMarkers - the possible marker list, that each 
		child should have at least one of his possible markers in that list.
 Output:YES - if all children in the family have at least one matching marker.
        NO  - otherwise
**************************************************************************************/
/* old version
ckrResult MarkPossibleChildrenMarkers(Person * aFirstChild, MarkerLLNode * aPossibleChildMarkers)
{
	Person * child;
	MarkerLLNode * childMarker;
	Result isAllChildrenOK;
	Result isChildOK;
	child = aFirstChild;

	/* Mark the tmp flag for the matching markers 
	isAllChildrenOK = YES;
	while (NULL != child)
	{
		childMarker = child->mOptMarkerLL;
		isChildOK = NO;
		while (NULL != childMarker)
		{
			
			if (YES == IsMarkerInList(childMarker->mMarker, aPossibleChildMarkers))
			{
				childMarker->mTmpIsMarked = 1;
				isChildOK = YES;
			}
			
			childMarker = childMarker->mNext;
		}
		if (NO == isChildOK)
		{
			isAllChildrenOK = NO;
			break;
		}
		child = child->mSibling;
	}

	/* if all children are OK, then save the matching markers, and clear the tmp flag.
	   if not all children OK, - just clear the tmp flag 
	child = aFirstChild;
	while (NULL != child)
	{
		childMarker = child->mOptMarkerLL;
		while (NULL != childMarker)
		{
			if ((YES == isAllChildrenOK) && (1 == childMarker->mTmpIsMarked))
			{
				childMarker->mIsMarked = 1;	
			}
			childMarker->mTmpIsMarked = 0;
			childMarker = childMarker->mNext;
		}
		child = child->mSibling;
	}
	return isAllChildrenOK;
}
*/
/*************************************************************************************
 Function: MarkPossibleChildrenMarkers
	The function checks if all children in siblinghood have at least one marker included
	in the given markers list.
 Input: Person * aFirstChild - first child in siblinghood
		MarkerLLNode * aPossibleChildMarkers - the possible marker list, that each 
		child should have at least one of his possible markers in that list.
 Output:YES - if all children in the family have at least one matching marker.
        NO  - otherwise
**************************************************************************************/
ckrResult MarkPossibleChildrenMarkers (Person * aFirstChild, 
									   MarkerLLNode * aPossibleChildMarkers,
									   Format aFormat,
									   MarkerLLNode * aPossibleChildMaleMarkers)
{
	Person * child;
	MarkerLLNode * childMarker;
	Result isAllChildrenOK;
	Result isChildOK;
	MarkerLLNode * aChildMarkers = aPossibleChildMarkers;	
	child = aFirstChild;
	

	/* Mark the tmp flag for the matching markers */
	isAllChildrenOK = YES;
	while (NULL != child)
	{
		childMarker = child->mOptMarkerLL;
		isChildOK = NO;

		if ((SEXLINKED == aFormat) && (male == child->mGender))
			aChildMarkers = aPossibleChildMaleMarkers;

		while (NULL != childMarker)
		{
			if (YES == IsMarkerInList(childMarker->mMarker, aChildMarkers))
			{
				childMarker->mTmpIsMarked = 1;
				isChildOK = YES;
			}
			
			childMarker = childMarker->mNext;
		}
		if (NO == isChildOK)
		{
			isAllChildrenOK = NO;
			break;
		}
		child = child->mSibling;
	}

	/* if all children are OK, then save the matching markers, and clear the tmp flag.
	   if not all children OK, - just clear the tmp flag */
	child = aFirstChild;
	while (NULL != child)
	{
		childMarker = child->mOptMarkerLL;
		while (NULL != childMarker)
		{
			if ((YES == isAllChildrenOK) && (1 == childMarker->mTmpIsMarked))
			{
				childMarker->mIsMarked = 1;	
			}
			childMarker->mTmpIsMarked = 0;
			childMarker = childMarker->mNext;
		}
		child = child->mSibling;
	}
	return isAllChildrenOK;
}

/*************************************************************************************
 Function: GetPossibleChildMarkers
	The function returns all combinations of markers of potential children of the given
	parent's markers.
 Input: MarkerLLNode * fatherMarker - father's marker 
		MarkerLLNode * motherMarker - mother's marker
		MarkerLLNode ** childPossibleMarkers - the possible combinations of the given 
		parents' markers. The function fills this list.
 Output: ckrMemErr - in case of memory error
		 ckrOK     - otherwise
**************************************************************************************/
ckrResult GetPossibleChildMarkers(MarkerLLNode * fatherMarker,
								  MarkerLLNode * motherMarker, 
								  MarkerLLNode ** childPossibleMarkers)
{
	int fatherAllele1;
	int fatherAllele2;
	int motherAllele1;
	int motherAllele2;
	MarkerLLNode * nextC;

	fatherAllele1=fatherMarker->mMarker.mAllele1;
	fatherAllele2=fatherMarker->mMarker.mAllele2;
	motherAllele1=motherMarker->mMarker.mAllele1;
	motherAllele2=motherMarker->mMarker.mAllele2;

	if ((fatherAllele1 == fatherAllele2) && (motherAllele1 == motherAllele2))
	{
		*childPossibleMarkers = CreateMarkerLLNode(fatherAllele1,motherAllele1);
	}
	else if (((fatherAllele1 == motherAllele1) && (fatherAllele2 == motherAllele2)) || 
			 ((fatherAllele1 == motherAllele2) && (fatherAllele2 == motherAllele1)))
	{
		*childPossibleMarkers            = CreateMarkerLLNode(fatherAllele1,fatherAllele1);
		(*childPossibleMarkers)->mNext     = CreateMarkerLLNode(fatherAllele2,fatherAllele2); 
		nextC = (*childPossibleMarkers)->mNext;
		nextC->mNext = CreateMarkerLLNode(fatherAllele1,fatherAllele2);
	}
	else if (fatherAllele1 == fatherAllele2)
	{
		*childPossibleMarkers        = CreateMarkerLLNode(fatherAllele1,motherAllele1);
		(*childPossibleMarkers)->mNext = CreateMarkerLLNode(fatherAllele1,motherAllele2);
	}
	else if (motherAllele1 == motherAllele2)
	{
		*childPossibleMarkers        = CreateMarkerLLNode(fatherAllele1,motherAllele1);
		(*childPossibleMarkers)->mNext = CreateMarkerLLNode(fatherAllele2,motherAllele1);
	}
	else
	{
		*childPossibleMarkers            = CreateMarkerLLNode(fatherAllele1,motherAllele1);
		(*childPossibleMarkers)->mNext     = CreateMarkerLLNode(fatherAllele1,motherAllele2); 
		nextC     = (*childPossibleMarkers)->mNext;
		nextC->mNext = CreateMarkerLLNode(fatherAllele2,motherAllele1); nextC = nextC->mNext;
		nextC->mNext = CreateMarkerLLNode(fatherAllele2,motherAllele2);
	}
	if (NULL == *childPossibleMarkers)
	{
		return ckrMemErr;
	}
	return ckrOK;
}

/*my************************************************************************************
 Function: GetPossibleMaleChildMarkers
	The function returns all combinations of markers of potential children of the given
	parent's markers.
 Input: MarkerLLNode * motherMarker - mother's marker
		MarkerLLNode ** childPossibleMarkers - the possible combinations of the given 
		parents' markers. The function fills this list.
 Output: ckrMemErr - in case of memory error
		 ckrOK     - otherwise
**************************************************************************************/
ckrResult GetPossibleMaleChildMarkers(MarkerLLNode * motherMarker, 
								      MarkerLLNode ** childPossibleMarkers)
{
	int motherAllele1;
	int motherAllele2;

	motherAllele1=motherMarker->mMarker.mAllele1;
	motherAllele2=motherMarker->mMarker.mAllele2;

	/* mother is homozygotic */
	if ((motherAllele1 == motherAllele2)) 
	{
		*childPossibleMarkers = CreateMarkerLLNode(motherAllele1,motherAllele1);
	}
	else if ( motherAllele1 != motherAllele2) 
	{
		*childPossibleMarkers            = CreateMarkerLLNode(motherAllele1,motherAllele1);
		(*childPossibleMarkers)->mNext     = CreateMarkerLLNode(motherAllele2,motherAllele2); 
	}
	
	if (NULL == *childPossibleMarkers)
	{
		return ckrMemErr;
	}
	return ckrOK;
}

/*************************************************************************************
 Function: ExcludePersonNonMarkedMarkers
	The function exludes all of a person's non marked markers (from the person's 
	optional markers list)
 Input: Person * aPerson - the relevant person
        int * aIsMarkersExcluded - turned to 1 if any marker has been excluded
 Output: void
**************************************************************************************/
void ExcludePersonNonMarkedMarkers(Person * aPerson, int * aIsMarkersExcluded)
{
	MarkerLLNode * marker = aPerson->mOptMarkerLL;
	MarkerLLNode * prev = NULL;
	MarkerLLNode * temp = NULL;
	
	while (NULL != marker)
	{
		if (0 == marker->mIsMarked)
		{	
			*aIsMarkersExcluded = 1;

			/* change the reference from the person */
			/* to the next element */
			if (NULL == prev)
			{
				temp = marker;
				marker = marker->mNext;
				aPerson->mOptMarkerLL = marker;
				free (temp);
				continue;
			}
			else
			{
				temp = marker;
				marker = marker->mNext;
				prev->mNext = marker;
				free (temp);
				continue;
			}
		}
		prev = marker;
		marker = marker->mNext;
	}
}

/*************************************************************************************
 Function: IsAllFamMembersHaveOptMarkers
	The function checks whether all family's members have at least one optional marker
	each.
 Input: NuclearFam * aNuclearFam - the relevant nuclear family
 Output: ckrOK - if all members of the family, have at least one optional marker.
         ckrErr - otherwise
**************************************************************************************/
ckrResult IsAllFamMembersHaveOptMarkers(NuclearFam * aNuclearFam)
{
	Person * child;
	
	if ((NULL == aNuclearFam->mFather->mOptMarkerLL) ||
		(NULL == aNuclearFam->mMother->mOptMarkerLL))
	{
		return ckrErr;
	}
	
	child = aNuclearFam->mChild;
	while (NULL != child)
	{
		if (NULL == child->mOptMarkerLL)
		{
			return ckrErr;
		}
		child = child->mSibling;
	}
	return ckrOK;	
}

/*************************************************************************************
 Function: ExcludeNonMarkedMarkers
	The function exludes all of a person's non marked markers (from the person's 
	optional markers list), for all persons in the given nuclear family
 Input: NuclearFam * aNuclearFam - the relevant family
        int * aIsFamChanged - turned to 1 if any of the family's people had a marker 
		excluded
 Output: void
**************************************************************************************/
void ExcludeNonMarkedMarkers(NuclearFam * aNuclearFam, int * aIsFamChanged)
{
	Person * child;

	ExcludePersonNonMarkedMarkers(aNuclearFam->mFather, aIsFamChanged);
	ExcludePersonNonMarkedMarkers(aNuclearFam->mMother, aIsFamChanged);
	
	child = aNuclearFam->mChild;
	while (NULL != child)
	{
		ExcludePersonNonMarkedMarkers(child, aIsFamChanged);
		child = child->mSibling;
	}

}

/*************************************************************************************
 Function: SetMarkersMarked
	The function sets all the family's persons optional markers list, to be marked or 
	not marked (the marked field is set to aMarkVal, which is 0 or 1 only)
 Input: NuclearFam * aNuclearFam - the family to handle its members
		int aMarkVal - the value to be set. 0 meams all markers will be set to "not 
		marked", 1 means all markers will be set to "marked"
 Output: void
**************************************************************************************/
void SetMarkersMarked(NuclearFam * aNuclearFam, int aMarkVal)
{
	Person * child;

	SetPersonMarkersMarked(aNuclearFam->mFather, aMarkVal);
	SetPersonMarkersMarked(aNuclearFam->mMother, aMarkVal);
	
	child = aNuclearFam->mChild;
	while (NULL != child)
	{
		SetPersonMarkersMarked(child, aMarkVal);
		child = child->mSibling;
	}
}

/*************************************************************************************
 Function: SetPersonMarkersMarked
	The function sets all of the given person's optional markers list to be with the 
	value aMarkVal (which is 0 or 1 only)
 Input: Person * aPerson - the person
        int aMarkVal - the new value to set on markers
 Output: void
**************************************************************************************/
void SetPersonMarkersMarked(Person * aPerson, int aMarkVal)
{
	MarkerLLNode * marker = aPerson->mOptMarkerLL;
	
	while (NULL != marker)
	{
		marker->mIsMarked = aMarkVal;
		marker = marker->mNext;
	}
}

/*************************************************************************************
 Function: IsMarkerInList
	The function checks if the given marker is included in the given markers list
 Input: Marker aMarker - the marker to check
		MarkerLLNode * aMarkerList - the markers list, to check if aMarker is included 
		in it.
 Output: YES - if the marker is included
		 NO  - otherwise
**************************************************************************************/
Result IsMarkerInList(Marker aMarker, MarkerLLNode * aMarkerList)
{
	int allele1 = aMarker.mAllele1;
	int allele2 = aMarker.mAllele2;
	int iMarkerAllele1;
	int iMarkerAllele2;
	MarkerLLNode * iMarker;
	
	iMarker = aMarkerList;

	while (NULL != iMarker)
	{
		iMarkerAllele1    = iMarker->mMarker.mAllele1;
		iMarkerAllele2    = iMarker->mMarker.mAllele2;
		if (((iMarkerAllele1 == allele1)  && (iMarkerAllele2 == allele2)) ||
		    ((iMarkerAllele1 == allele2)  && (iMarkerAllele2 == allele1)))
		{
			return YES;
		}
		iMarker = iMarker->mNext;
	}
	return NO;
}

