/*************************************************************************************
 Level3Checker.c
 The file deals with level 3 checkings functions. 
 Level 3 is mainly to check consistency in the pedigree, and if no consistency was 
 found, then to tell the user what are the critical genotypes for the pedigree.
*************************************************************************************/

#include "pc_Level3Checker.h"
#include "pc_NuclearFam.h"
#include "pc_Pedigree.h"
#include "pc_LangeGoradia.h"
#include "pc_Marker.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

ckrResult CriticalGenotypesILevel(Pedigree * aPedigree, int aMarkerIdx, int aI);
Result SetIUnknownPersons(Pedigree * aPedigree, int aCritincGenotypeLevel, int aMarkerIdx, int * aUnknownIndices);
void UnsetIUnknownPersons(Pedigree * aPedigree, int aCritincGenotypeLevel, int aMarkerIdx, int * aUnknownIndices);
Result SetNextUnknownIndices(int aUnknownIndicesCnt, int * aUnknownIndices, int aPedigreePersonCnt);


/*************************************************************************************
 Function: Level3Errors
 	The function checks level 3 checks of the given pedigree. In case Lange-Goradia 
	algorithm showed there is no consistency in the pedigree, level 3 is resposible 
	to find out which are the minimum sets of people, that when their markers are 
	turned to "unknown" (instead of it's value in the input file), then the pedigree
	becomes consistent. (such a set must exist, because it is obvios that when all 
	pedigree members are turned to be "unknown", then the pedigree is consistent.
	This minumum set of people is called "critical genotypes"
 input: Pedigree * aPedigree - the checked pedigree
 output: ckrMemErr - in case of memory error			   
		 ckrOK     - otherwise
*************************************************************************************/
ckrResult Level3Errors(Pedigree * aPedigree)
{
	int iCriticGenotypeLevel = 0;
	int iMarkerIdx = 0;
	int markerCnt = aPedigree->mMarkerCnt;
	ckrResult res;

	/* detecting critical Genotypes */
	for (iMarkerIdx = 0; iMarkerIdx < markerCnt; ++iMarkerIdx)
	{
		/* correct pedigree in this marker */
		if (0 == aPedigree->mCriticalGenotypesLevel[iMarkerIdx])
		{
			continue;
		}

		/* CRITICAL_GENOTYPES_MAX_LEVEL - contains a max number of ZIVUG people
		  2 - till pairs, ... */
		for (iCriticGenotypeLevel = 1; 
		     iCriticGenotypeLevel <= CRITICAL_GENOTYPES_MAX_LEVEL; 
			 ++iCriticGenotypeLevel)
		{
			res = CriticalGenotypesILevel(aPedigree, iMarkerIdx, iCriticGenotypeLevel);
			RETURN_MEM_ERROR(res);
			if (ckrOK == res)
			{
				/* the minimum level of critical genotypes that 
				   solves this marker, is exactly i.
				   A positive value to mCriticalGenotypesLevel[iMarkerIdx] means 
				   that a solution has been found */
				aPedigree->mCriticalGenotypesLevel[iMarkerIdx] = iCriticGenotypeLevel;
				break;
			}
			else
			{
				/* the minimum level of critical genotypes that 
				   solves this marker, is at least i+1 
				   A negative value to mCriticalGenotypesLevel[iMarkerIdx] means 
				   that a solution has not yet been found */
				aPedigree->mCriticalGenotypesLevel[iMarkerIdx] = -(iCriticGenotypeLevel+1);
			}
		}
	}
	
	return ckrOK;
}

/*************************************************************************************
 Function: CriticalGenotypesILevel
 	The function finds the critical genotype for the given pedigree and marker index, 
	when the size the critical genotypes is i. Meaning, it finds a set of exactly i
	people, that when their markers are turned to "unknown", the pedigree becomes 
	consistent.
	The found sets are documented in the recommendation list of the pedigree
 input:  Pedigree * aPedigree - the checked pedigree
		 int aMarkerIdx - the checked marker index
		 int aI - the size of the set of critical genotypes
 output: ckrErr - in case of invalid input, or when a set of critical genotypes of 
		 size i doesn't exist
		 ckrOK  - when a set of critical genotypes, in size i, is found
		 ckrMemErr - in case of memory error
*************************************************************************************/
ckrResult CriticalGenotypesILevel(Pedigree * aPedigree, int aMarkerIdx, int aI)
{
	/* the indices of the i persons who their markers are set to unkown*/
	int * unknownIndices = NULL;
	int j = 0;
	Result res;
	ckrResult ckrRes;
	ckrResult retRes = ckrErr;

	if ((0 >= aI) || (aPedigree->mPersonCnt < aI))
	{
		return ckrErr;
	}
	
	unknownIndices = (int *)malloc(aI*sizeof(int));
	if (NULL == unknownIndices)
	{
		return ckrMemErr;
	}

	for (j = 0; j < aI; ++j)
	{
		unknownIndices[j] = j;
	}

	while (1)
	{
		res = SetIUnknownPersons(aPedigree, aI, aMarkerIdx, unknownIndices);
		/* run Lange Goradia only if all the i persons have known markers */
		if (YES == res)
		{
			ckrRes = LangeGoradia(aPedigree, aMarkerIdx);
			RETURN_MEM_ERROR(ckrRes);
			if (ckrOK == ckrRes)
			{
				retRes = ckrOK;

				SetRecommendation(aPedigree, aMarkerIdx, aI, unknownIndices);
				
				/* we look for only one option of i critcal genotypes when
				   CRITICAL_GENOTYPES_MULT_RESULTS is 0.
				   if we want to find all the sets of critical genotypes of
				   size i in the pedigree, then turn CRITICAL_GENOTYPES_MULT_RESULTS
				   to be 1 */
				if (0 == CRITICAL_GENOTYPES_MULT_RESULTS)
				{
					return retRes;
				}
			}
			UnsetIUnknownPersons(aPedigree, aI, aMarkerIdx, unknownIndices);
		}
		
		res = SetNextUnknownIndices(aI, unknownIndices, aPedigree->mPersonCnt);
		/* no more options of i unknown people out of the pedigree */
		if (NO == res)
		{
			break;
		}
	}

	free(unknownIndices);
	return retRes;	
}

/*************************************************************************************
 Function: SetIUnknownPersons
 	The function gets as an argument an array of people in the pedigree, which should 
	be turned to "unknown" in the given marker. This is for checking if they a critical
	genotype in the pedigree - meaning, to check if turning their marker into "unkown"
	instead of relaying on the input for their marker, will turn the pedigree to be
	consistent.
	The original markers' value of these people are saved as a "backup"
 input:  Pedigree * aPedigree - the pedigree we work on.
		 int aCritincGenotypeLevel - the number of people to be set to "unkown"
		 int aMarkerIdx - the index of the marker to be set to "unkown"
		 int * aUnknownIndices - an array that specify the indices of the people to turn
		 into "unkown"
 output: YES - in case we've turned the asked people to "unknown" in the asked marker
         successfuly
		 NO  - in case the input to the function is wrong, or when one of the people 
		 in the specified array, is allready "unkown" (which means, the original input
		 on this person's marker, was allready "unknown", and therefor, turning this 
		 person to "unknown" will not help us find consistency in the pedigree)
*************************************************************************************/
Result SetIUnknownPersons(Pedigree * aPedigree, int aCritincGenotypeLevel, int aMarkerIdx, int * aUnknownIndices)
{
	int i;
	Person * iPerson = NULL;

	if (aMarkerIdx >= aPedigree->mPersonArray[0]->mMarkersCnt)
	{
		return NO;
	}

	for (i = 0; i < aCritincGenotypeLevel; ++i)
	{
		iPerson = aPedigree->mPersonArray[aUnknownIndices[i]];

		/* came unknown in the input */
		if ((iPerson->mMarkerArr[aMarkerIdx].mAllele1 == 0) ||
			(iPerson->mMarkerArr[aMarkerIdx].mAllele2 == 0))
		{
			/* if we allready set few persons as unknown, undo that */
			if (i > 0)
			{
				UnsetIUnknownPersons(aPedigree, i, aMarkerIdx, aUnknownIndices);
			}
			return NO;
		}

		iPerson->mIsUnknown = 1;
		iPerson->mMarkerBackupForUnknown.mAllele1 = iPerson->mMarkerArr[aMarkerIdx].mAllele1;
		iPerson->mMarkerBackupForUnknown.mAllele2 = iPerson->mMarkerArr[aMarkerIdx].mAllele2;
		iPerson->mMarkerArr[aMarkerIdx].mAllele1 = 0;
		iPerson->mMarkerArr[aMarkerIdx].mAllele2 = 0;
	}

	return YES;
}

/*************************************************************************************
 Function: UnsetIUnknownPersons
 	The function returns the people whom their indices specified in aUnknownIndices
	array, to have their original values of marker number aMarkerIdx.
	(the original value was backed-up, and turned into "unknown" to check if the person
	is in a critical genotype set).
 input:  Pedigree * aPedigree - the checked pedigree
		 int aCritincGenotypeLevel - number of people in the unknown array
	     int aMarkerIdx - the index of the marker
		 int * aUnknownIndices - the array of people to get them back their original 
		 value.
 output: void
*************************************************************************************/
void UnsetIUnknownPersons(Pedigree * aPedigree, int aCritincGenotypeLevel, int aMarkerIdx, int * aUnknownIndices)
{
	int i;
	Person * iPerson = NULL;

	if (aMarkerIdx >= aPedigree->mPersonArray[0]->mMarkersCnt)
	{
		return;
	}

	for (i = 0; i < aCritincGenotypeLevel; ++i)
	{
		iPerson = aPedigree->mPersonArray[aUnknownIndices[i]];

		iPerson->mIsUnknown = 0;
		iPerson->mMarkerArr[aMarkerIdx].mAllele1 = iPerson->mMarkerBackupForUnknown.mAllele1;
		iPerson->mMarkerArr[aMarkerIdx].mAllele2 = iPerson->mMarkerBackupForUnknown.mAllele2;
	}

	return;
}

/*************************************************************************************
 Function: SetNextUnknownIndices
 	The function takes an array of people in the pedigree, and return the next set of
	people in the pedigree. If the array is initialized properly, than a continues call
	to the function gerentees to return all sub-sets of people in the pedigree, of order
	aUnknownIndicesCnt.
 input:  int aUnknownIndicesCnt - number of people in the array
         int * aUnknownIndices - the array that holds the current subset of people, and
		 will be set to the next subset.
		 int aPedigreePersonCnt - number of people in the pedigree.
 output: YES - if the next subset exist
         NO  - if aUnknownIndices allready contains the last subset
*************************************************************************************/
Result SetNextUnknownIndices(int aUnknownIndicesCnt, int * aUnknownIndices, int aPedigreePersonCnt)
{
	int i;
	int j;

	for (i = 1; i <= aUnknownIndicesCnt; ++i)
	{
		if (aUnknownIndices[aUnknownIndicesCnt - i] < aPedigreePersonCnt - i)
		{
			++aUnknownIndices[aUnknownIndicesCnt - i];
			for (j = i - 1; j >= 1; --j)
			{
				aUnknownIndices[aUnknownIndicesCnt - j] = aUnknownIndices[aUnknownIndicesCnt - j - 1] + 1;
			}
			return YES;
		}
	}
	return NO;
}

