/*************************************************************************************
 PedAlleles.c
 This file deals with structures and functions, which "collect", at parsing time, all
 the alleles in the pedigree, for a certain marker.
 This is important for later stages of the application. (for Lange-Goradia algorithm)
*************************************************************************************/
#include <stdlib.h>
#include "pc_PedAlleles.h"


ckrResult GetAlleleLLFromPedAlleles(AlleleLL ** aAlleleLL, PedAlleles * aPedAlleles, int aMarkerIdx);
ckrResult AddAlleleToAlleleLL(AlleleLL * aAlleleList, int aAlleleValue);


/*************************************************************************************/
/*								Allele functions									 */
/*************************************************************************************/

/*************************************************************************************
Function:IsAlleleIn
Function recieves three alleles and checks if the first one is equal to one of the others 
**************************************************************************************/ 

Result IsAlleleIn(int aCheckedAllele,int aAllele1,int aAllele2)
{
	if ((aCheckedAllele == aAllele1) || (aCheckedAllele == aAllele2))
		return YES;
	return NO;	
}

/*************************************************************************************
Function:CreateAlleleLLNode
Creates a new allele node and initializes it. 
*************************************************************************************/
AlleleLLNode * CreateAlleleLLNode(int aAlleleValue)
{
	AlleleLLNode * allele;


	 allele = (AlleleLLNode *)malloc(sizeof(AlleleLLNode));
	if (allele == NULL)
		return NULL;
	
	allele->mAllele = aAlleleValue;
	allele->mNext = NULL;
	
	return allele;
}


/*************************************************************************************/
/*							PedigreeAlleles functions								 */
/*************************************************************************************/



/*************************************************************************************
Function:CreateAlleleLL
Recieves aMarkerIdx,creates a new allele LL and initializes it. 
*************************************************************************************/
AlleleLL * CreateAlleleLL(int aMarkerIdx)
{
	AlleleLL * alleleLL;

	alleleLL = (AlleleLL *)malloc(sizeof(AlleleLL));
	if (NULL == alleleLL)
	{
		return NULL;
	}
	alleleLL->mFirstLLNode = NULL;
	alleleLL->mIndex = aMarkerIdx;
	alleleLL->mNext = NULL;
	return alleleLL;
}

/*************************************************************************************
Function:CreatePedAlleles
Returns new struct of pedAlleles, and initializing it.
*************************************************************************************/

PedAlleles * CreatePedAlleles()
{
	PedAlleles * pedAlleles;

	pedAlleles = (PedAlleles *)malloc(sizeof(PedAlleles));
	if (NULL == pedAlleles)
	{
		return NULL;
	}

	pedAlleles->mAlleleLists = NULL;
	return pedAlleles;
}


/*************************************************************************************
Function:AddAlleleToPedAlleles
Recieves aAlleleValue and aPedAlleles adds it to alleleList.
*************************************************************************************/
ckrResult AddAlleleToPedAlleles(int aAlleleValue, PedAlleles * aPedAlleles, int aMarkerIdx)
{
	ckrResult      res = ckrOK;
	AlleleLL *     alleleList = NULL;
	
	/* wrong argument */
	if (0 > aMarkerIdx)
	{
		return ckrErr;
	}

	/* the function always returns a not null alleleList (other than in a mem error */
	res = GetAlleleLLFromPedAlleles(&alleleList, aPedAlleles, aMarkerIdx);
	RETURN_MEM_ERROR(res);

	res = AddAlleleToAlleleLL(alleleList, aAlleleValue);
	RETURN_MEM_ERROR(res);

	return ckrOK;
}

/*************************************************************************************
Function:GetAlleleLLFromPedAlleles
Recieves aPedAlleles.
Returns ckrOK if aAlleleLL was found , if not creates one.
/*************************************************************************************/
ckrResult GetAlleleLLFromPedAlleles(AlleleLL ** aAlleleLL, PedAlleles * aPedAlleles, int aMarkerIdx)
{
	AlleleLL * iAlleleLL = aPedAlleles->mAlleleLists;

	*aAlleleLL = NULL;

	while (NULL != iAlleleLL)
	{
		/* such an allele Linked List already exist */
		if (aMarkerIdx == iAlleleLL->mIndex)
		{
			*aAlleleLL = iAlleleLL;
			return ckrOK;
		}
		iAlleleLL = iAlleleLL->mNext;
	}

	/* if we got until here, no such allele LL was 
	   found - create a new one and add it          */
	iAlleleLL = CreateAlleleLL(aMarkerIdx);
	if (NULL == iAlleleLL)
	{
		return ckrMemErr;
	}

	iAlleleLL->mNext = aPedAlleles->mAlleleLists;
	aPedAlleles->mAlleleLists = iAlleleLL;
	*aAlleleLL = iAlleleLL;
	return ckrOK;
}

/*************************************************************************************
Function:GetMaleAlleleLLFromPedAlleles
Recieves aPedAlleles.
Returns ckrOK if aAlleleLL was found , if not creates one.
/*************************************************************************************/
ckrResult GetMaleAlleleLLFromPedAlleles(AlleleLL ** aAlleleLL, PedAlleles * aPedAlleles, int aMarkerIdx)
{
	AlleleLL * iAlleleLL = aPedAlleles->mAlleleLists;

	*aAlleleLL = NULL;

	while (NULL != iAlleleLL)
	{
		/* such an allele Linked List already exist */
		if (aMarkerIdx == iAlleleLL->mIndex)
		{
			*aAlleleLL = iAlleleLL;
			return ckrOK;
		}
		iAlleleLL = iAlleleLL->mNext;
	}

	/* if we got until here, no such allele LL was 
	   found - create a new one and add it          */
	iAlleleLL = CreateAlleleLL(aMarkerIdx);
	if (NULL == iAlleleLL)
	{
		return ckrMemErr;
	}

	iAlleleLL->mNext = aPedAlleles->mAlleleLists;
	aPedAlleles->mAlleleLists = iAlleleLL;
	*aAlleleLL = iAlleleLL;
	return ckrOK;
}


/*************************************************************************************
Function:AddAlleleToAlleleLL
Recieves aAlleleList and aAlleleValue.
Adds allele to linked list only if it doesn't already exists in list.
Returns ckrOK after doing it successfully.
**************************************************************************************/
ckrResult AddAlleleToAlleleLL(AlleleLL * aAlleleList, int aAlleleValue)
{
	AlleleLLNode * allele;

	allele = aAlleleList->mFirstLLNode;
	
	while (NULL != allele)
	{
		/* such an allele value already exists */
		if (aAlleleValue == allele->mAllele)
		{
			return ckrOK;
		}
		allele = allele->mNext;
	}

	/* if we got until here, no such allele value 
	   was found - create a new one and add it */
	allele = CreateAlleleLLNode(aAlleleValue);
	if (NULL == allele)
	{
		return ckrMemErr;
	}
	allele->mNext = aAlleleList->mFirstLLNode;
	aAlleleList->mFirstLLNode = allele;
	return ckrOK;
}

