/*************************************************************************************
 Pedigree.c
 Pedigree is a struct which represenets a pedigree. It holds all the information needed
 for the application's run, on the pedigree, and points to its members.
*************************************************************************************/
#include "pc_Pedigree.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


ckrResult	CompleteDataInPedigree(Pedigree * aPedigree,int aNumPerson);
Result IsUnknownParent(char * aParentID);

/*****************************************************************************************
Function:CreatePedigree
The function recieves aLLPerson,aPersonCnt,aPedAlleles and aPedigreeErrorLL.
Returns a new struct of pedigree and sets the correct information in each 
of its fields.
******************************************************************************************/
Pedigree * CreatePedigree(Person * aLLPerson, 
						  int aPersonCnt, 
						  int aMarkerCnt, 
						  PedAlleles * aPedAlleles, 
						  ErrorLL *  aPedigreeErrorLL)
{   
	Pedigree * aPedigree;
	Person * aPerson;
	int i;

/*	if (aPersonCnt <= 0)
	{
		return NULL;
	}*/

	aPedigree = (Pedigree *)malloc(sizeof(Pedigree));
	if (aPedigree == NULL)
	{
		return NULL;
	}
	memset(aPedigree, 0, sizeof(Pedigree));

	aPedigree->mPedAlleles = aPedAlleles;
	aPedigree->mPersonCnt  = aPersonCnt;

	aPedigree->mMarkerCnt = aMarkerCnt;
	aPedigree->mErrorLL1  = aPedigreeErrorLL;

    aPedigree->mErrorLL2 = CreateErrorLL(); 
//    aPedigree->mErrorLL3 = CreateErrorLL(); 

	aPedigree->mRecommendationLL = CreateRecommendationLL();
	
	aPedigree->mPersonArray = (Person **)malloc(sizeof(Person *)*aPersonCnt);
	if (aPedigree->mPersonArray == NULL)
	{
		free(aPedigree);
		return NULL;
	}

	aPerson=aLLPerson;

	for(i=0; i<aPersonCnt; ++i)
	{
		aPedigree->mPersonArray[i] = aPerson;
		aPerson=aPerson->mNext;
	}

	CompleteDataInPedigree(aPedigree, aPersonCnt);

	return aPedigree;
}  


/****************************************************************************************
Function:SetPedigreeName
Recieves aPedigree and a string (aName) which is the pedigree's name.
The function copies the name to it's field.
*****************************************************************************************/
void SetPedigreeName(Pedigree * aPedigree, const char * aName)
{
	strcpy(aPedigree->mName, aName);
}


/*****************************************************************************************
Function:DeletePedigree
The function recieves aPedigree and is responsible to delete it.
******************************************************************************************/
void DeletePedigree(Pedigree * aPedigree)
{
	if (aPedigree == NULL)
	{
		return;
	}

	if (aPedigree->mPersonArray != NULL)
	{
		free(aPedigree->mPersonArray);
	}
	free(aPedigree);
}


/****************************************************************************************
Function:CompareID
Returns YES if id1 and id2 are the same and NO if not.
*****************************************************************************************/
Result CompareID(char * id1,char * id2)
{
	if (strcmp(id1,id2)==0) 
		return YES;
	else 
		return NO;
}


/*************************************************************************************
Function:GetParentsFam
Recieves pointers to aFather and aMother and returns their NuclearFam.
**************************************************************************************/
NuclearFam * GetParentsFam(Person * aFather,Person * aMother)
{

	NuclearFam * aFam;
	NuclearFam * aPrevFam;

	if ((aFather==NULL) || (aMother==NULL))
	{
		return NULL;
	}
	
	/* run on mother's Nuclear fam list, and find (or create) the family */
	aFam=aMother->mFoundedFam;
	if (aFam==NULL)
	{
		aFam = CreateNuclearFam();
		aMother->mFoundedFam = aFam;
	}
	else
	{
		aPrevFam = NULL;
		aFam = aFam;
		while (aFam!=NULL)
		{
			if (aFam->mFather==aFather)  
				return aFam;
			aPrevFam = aFam;
			aFam = aFam->mNextMotherFam;
		}
		aFam = CreateNuclearFam();
		if (NULL != aPrevFam)
		{
			aPrevFam->mNextMotherFam = aFam;
		}
	}	

	/* connect Father to the New created Nuclear Family */
	if (aFather->mFoundedFam == NULL)
	{
		aFather->mFoundedFam = aFam;
	}
	else
	{
		aPrevFam = aFather->mFoundedFam;
		while (aPrevFam->mNextFatherFam != NULL)
			aPrevFam = aPrevFam->mNextFatherFam;
		aPrevFam->mNextFatherFam = aFam;
	}

	/* connect New created Family to father & Mother */
	aFam->mFather = aFather;
	aFam->mMother = aMother;
	return aFam;
}


/*************************************************************************************
Function:CheckIfToSwitchParents
The function checks if a father of the person is declared as a female and if 
mother of the person is declared as a male or both is incorrect.
**************************************************************************************/
ckrResult CheckIfToSwitchParents(ErrorLL * aErrorLL, 
								 Person ** aFather, 
								 Person ** aMother, 
								 int aFatherIsFemaleFlag, 
								 int aMotherIsMaleFlag,
								 char * aPersonID)
{
	/* both father and mother are correct */
	if ((0 == aFatherIsFemaleFlag) && (0 == aMotherIsMaleFlag))
	{
		return ckrOK;
	}

	/* only father is incorrect */
	if ((1 == aFatherIsFemaleFlag) && (0 == aMotherIsMaleFlag))
	{
		SetParentErrorInPerson(aErrorLL, UNDEFINED_MARKER_IDX, aPersonID, (*aFather)->mID, FATHER_IS_FEMALE_ERROR);
		return ckrErr;
	}
	
	/* only mother is incorrect */
	if ((0 == aFatherIsFemaleFlag) && (1 == aMotherIsMaleFlag))
	{
		SetParentErrorInPerson(aErrorLL, UNDEFINED_MARKER_IDX, aPersonID,(*aMother)->mID, MOTHER_IS_MALE_ERROR );
		return ckrErr;
	}

	/* both are incorrect */
	SetErrorInPerson(aErrorLL, UNDEFINED_MARKER_IDX, aPersonID, PARENTS_ORDER_ERROR);
	return ckrErr;
}


/************************************************************************************
Function:FindParents
The function is responsable to find the parents of a person according to aPersonID.
*************************************************************************************/
ckrResult FindParents(Pedigree * aPedigree, 
				 char * aFatherID, 
				 char * aMotherID, 
				 Person ** aFather, 
				 Person ** aMother,
				 char * aPersonID)
{
	int i;
	Person * aPerson;
	int aFatherSrchFlag=1;
	int aMotherSrchFlag=1;
	int aPersonCnt;
	/* the most weird variables I ever declared on... */
	int fatherIsFemaleFlag = 0;
	int motherIsMaleFlag   = 0;
	int errorFlag = 0;

	*aFather = NULL;
	*aMother = NULL;
	aPersonCnt = aPedigree->mPersonCnt;


	if (0 == strcmp(aFatherID, aPersonID))
	{
		SetErrorInPerson(aPedigree->mErrorLL1, UNDEFINED_MARKER_IDX, aPersonID, FATHER_OF_HIMSELF_ERROR);
		errorFlag = 1;
		
	}
	if (0 == strcmp(aMotherID, aPersonID))
	{
		SetErrorInPerson(aPedigree->mErrorLL1, UNDEFINED_MARKER_IDX, aPersonID, MOTHER_OF_HIMSELF_ERROR);
		errorFlag = 1;
	}
	if (1 == errorFlag)
	{
		return ckrErr;
	}

	if (IsUnknownParent(aFatherID)==YES)
	{
		aFatherSrchFlag = 0;
	}

	if (IsUnknownParent(aMotherID)==YES)
	{
		aMotherSrchFlag = 0;
	}

	for (i=0; i<aPersonCnt; ++i)
	{
		if (aFatherSrchFlag==0 && aMotherSrchFlag==0)
		{
			return CheckIfToSwitchParents(aPedigree->mErrorLL1, 
										 aFather, 
										 aMother, 
										 fatherIsFemaleFlag, 
										 motherIsMaleFlag,
										 aPersonID);
		}

		aPerson = aPedigree->mPersonArray[i];

		if (aFatherSrchFlag && (YES == CompareID(aPerson->mID,aFatherID)))
		{
			if (aPerson->mGender != male)
			{
				fatherIsFemaleFlag = 1;
			}
			aFatherSrchFlag = 0;
			*aFather = aPerson;
		}

		if (aMotherSrchFlag && (YES == CompareID(aPerson->mID,aMotherID)))
		{
			if (aPerson->mGender != female)
			{
				motherIsMaleFlag = 1;
			}
			aMotherSrchFlag = 0;
			*aMother = aPerson;
		}
	}

	/* both of the parents have been found */
	if (aFatherSrchFlag==0 && aMotherSrchFlag==0)
	{
		return CheckIfToSwitchParents(aPedigree->mErrorLL1, 
									  aFather, 
									  aMother, 
									  fatherIsFemaleFlag, 
									  motherIsMaleFlag,
									  aPersonID);
	}

	/* if we got until here, then not both of the parents have been found */
	if (aFatherSrchFlag == 1)
	{
		SetErrorInPerson(aPedigree->mErrorLL1, UNDEFINED_MARKER_IDX, aPersonID, NO_FATHER_FOUND_ERROR);
	}
	if (aMotherSrchFlag == 1)
	{
		SetErrorInPerson(aPedigree->mErrorLL1, UNDEFINED_MARKER_IDX, aPersonID, NO_MOTHER_FOUND_ERROR);
	}
	return ckrErr;
}


/*************************************************************************************/
Result IsUnknownParent(char * aParentID)
{
	if ((strcmp(aParentID,NO_PARENT_1)==0) || 
		(strcmp(aParentID,NO_PARENT_2)==0) || 
		(strcmp(aParentID,NO_PARENT_3)==0))
	{
		return YES;
	}
	return NO;
}


/************************************************************************************
Function:CompleteDataInPedigree
Recieves aPedigree and aNumPerson in it.
Set aMotherID and aFatherID of each person in the pedigree, Connect between a child and his family.
*************************************************************************************/
ckrResult	CompleteDataInPedigree(Pedigree * aPedigree,int aNumPerson)
{
	int i,aRes;
    char * aID;
	char * aMotherID;
	char * aFatherID;

	Person * aFather=NULL;
	Person * aMother=NULL;
    Person * aPerson;
	Person * aSibling;
	NuclearFam * aFam;
    
	/*go over the array and make the "connections" ...*/
	for (i=0;i<aNumPerson;i++) 
	{
		aPerson=aPedigree->mPersonArray[i];
		aID=aPerson->mID;
		aMotherID=aPerson->mMotherID;
		aFatherID=aPerson->mFatherID;
	
		/* Find if the person is a child */
		
		
				
        aRes=FindParents(aPedigree, aFatherID, aMotherID, &aFather, &aMother, aID);		
		
/*		if (aRes==ckrErr) 
		{
			AddErrorInPerson(aPedigree->mErrorLL1, UNDEFINED_MARKER_IDX, aID, "Error 3: a person's parents are not in the pedigree");
		}*/
			
		
         
		if ((aFather==NULL) || (aMother==NULL))
		{
			aFather = NULL;
			aMother = NULL;
			continue;
		}

		/* connect the aPerson as a child in the family */
    	aFam = GetParentsFam(aFather,aMother);
		aSibling = aFam->mChild;
		aFam->mChild = aPerson;
		aPerson->mSibling = aSibling;

		/* make the family the parents family of aPerson */
		aPerson->mParentsFam = aFam;
	}

	return ckrOK;
} 


/**************************************************************************************
Function:PrintPedigreeErrors
Prints the errors according to the level number
Return 1 if errors at this level exist, 0 else
**************************************************************************************/
int PrintPedigreeErrors(int aLevelNum, Pedigree * aPedigree,FILE * output)
{
	Error * error = NULL;
    int numError=0;

     /* check level number */
	switch(aLevelNum)
	{
	case 1:	numError = aPedigree->mErrorLL1->numOfErrors;
	        break;
	case 2: numError = aPedigree->mErrorLL2->numOfErrors;
			break;
//	case 3: numError = aPedigree->mErrorLL3->numOfErrors;
 	}
		

	/* no errors were found */
	if (0 == numError)
	{
		if ('\0' != aPedigree->mName)
		{
			switch (aLevelNum)
			{
			case(1):
				/*fprintf(output,"Pedigree %s had no syntax errors.\n\n", aPedigree->mName);*/
				break;
			case(2):
				/*fprintf(output,"Pedigree %s had no Mendelian errors.\n\n", aPedigree->mName);*/
				break;
			}
		}
		else
		{
		/*	fprintf(output,"A pedigree had no errors in level %d (there is no pedigree name).\n\n", aLevelNum);*/
		}
		return 0;
	}
	else /* errors exist */
	{
		if ('\0' != aPedigree->mName)
		{
			switch (aLevelNum)
			{
			case(1):
				//fprintf(output,"Pedigree %s had the following syntax errors:\n\n", aPedigree->mName);
				break;
			case(2):
				//fprintf(output,"Pedigree %s had the following Mendelian errors:\n\n", aPedigree->mName);
				break;
			}
		}
		else
		{
			switch (aLevelNum)
			{
			case(1):
				//fprintf(output,"A pedigree had the following syntax errors: (there is no name for this pedigree)\n\n");
				break;
			case(2):
				//fprintf(output,"A pedigree had the following Mendelian errors: (there is no name for this pedigree)\n\n");
				break;
			}
		}

		
		switch(aLevelNum)
		{
		case 1:	error = aPedigree->mErrorLL1->mFirst;
		        break;
	 	case 2: error = aPedigree->mErrorLL2->mFirst;
		        break;
//		case 3: error = aPedigree->mErrorLL3->mFirst;
 
		}
		
		while (NULL != error)
		{
			fprintf(output,"%s\n", error->errorString);
			error = error->mNext;
		}
		return 1;
	}
}


/******************************************************************************************
Function:PrintPedigreeRecommendations
Print the recommendations of a pedigree - meaning, print all the sets of critical genotypes
found in the pedigree.
******************************************************************************************/
int PrintPedigreeRecommendations(Pedigree * aPedigree,FILE * output)
{
	Recommendation * rec = NULL;
    int numRecs;
	int i = 0;

	numRecs = aPedigree->mRecommendationLL->numOfRecommendation;

	if (0 == numRecs)
	{
		if ('\0' != aPedigree->mName)
		{
			/*fprintf(output,"Pedigree %s had no recommendations.\n\n", aPedigree->mName);*/
		}
		else
		{
			/*fprintf(output,"A pedigree had no recommendations(there is no pedigree name).\n\n");*/
		}
		return 0;
	}
	else
	{
		rec = aPedigree->mRecommendationLL->mFirst;

		while (NULL != rec)
		{
			if (i > 0)
				fprintf(output, "Or, ");

			fprintf(output,"%s\n", rec->mRecommendationString);
			rec = rec->mNext;
			i++;
		}

		fprintf(output,"\n");
		return 1;
	}

}


