/*************************************************************************************
 Recommendation.c
 The file deals with recommendations functions. 
 A recommendation is a struct, which holds a string which is a recommendation to the 
 user of the application to turn a subset of people to "unknown" in a certain marker,
 in order to make the input pedigree be consistent. Meaning, a recommendation is 
 a "critical genotype" of the pedigree, on a certain marker.
*************************************************************************************/
#include "pc_Recommendation.h"
#include "pc_Pedigree.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

ckrResult AddRecommendation(RecommendationLL * aRecommendationLL, char * aRecommendationStr, int aRecommendationLen);

/*************************************************************************************
 Function: CreateRecommendationLL
 	The function create an empty linked-list of recommendations. 
 input:  void
 output: RecommendationLL * - a new and empty linked list of recommendations.
*************************************************************************************/
RecommendationLL * CreateRecommendationLL()
{
	RecommendationLL * recLL = NULL;

	recLL = (RecommendationLL *)malloc(sizeof(RecommendationLL));
	if (NULL == recLL)
	{
		return NULL;
	}

	recLL->mFirst = NULL;
	recLL->mLast  = NULL;
	recLL->numOfRecommendation = 0;
	
	return recLL;
}

/*************************************************************************************
 Function: AddRecommendation
 	The function adds a new recommendation into a recommendations linked list.
 input:  RecommendationLL * aRecommendationLL - the linked list to add a new 
		 recommendation to.
		 char * aRecommendationStr - the string that will appear in the new added 
		 recomendation.
		 int aRecommendationLen - the length of aRecommendationStr.
 output: ckrResult: ckrErr - in case of invalid input
					ckrMemErr - in case of a memory error
					ckrOK - in case of success.
*************************************************************************************/
ckrResult AddRecommendation(RecommendationLL * aRecommendationLL, char * aRecommendationStr, int aRecommendationLen)
{
	Recommendation * rec = NULL;
	char * recStr = NULL;

	if (0 >= aRecommendationLen)
	{
		return ckrErr;
	}


	rec = (Recommendation *)malloc(sizeof(Recommendation));
	if (NULL == rec)
	{
		return ckrMemErr;
	}

	recStr = (char *)malloc((aRecommendationLen+1)*sizeof(char));
	if (NULL == recStr)
	{
		return ckrMemErr;
	}
	memset(recStr, 0, aRecommendationLen+1);

	strcpy(recStr, aRecommendationStr);
	
	rec->mRecommendationLen    = aRecommendationLen;
	rec->mRecommendationString = recStr;
	rec->mNext = NULL;

	/* increase num of recommendations */
	aRecommendationLL->numOfRecommendation++;

	/* rec list is empty */
	if (NULL == aRecommendationLL->mFirst)
	{
		aRecommendationLL->mFirst = rec;
		aRecommendationLL->mLast  = rec;
		return ckrOK;
	}

	/* list is not empty. Add rec to recommendations list */
	aRecommendationLL->mLast->mNext = rec;
	aRecommendationLL->mLast = rec;
	return ckrOK;
}

ckrResult SetRecommendation(struct _Pedigree* aPedigree, int aMarkerIdx, int aI, int * unknownIndices)
{
	int j;
	//int aRecommendationLen = 0;
	int recLen;
	char * rec = NULL;
	char markerIdxStr[MAX_NUM_STR_LEN];

	/* Add a recommendation to the pedigree which specify all persons whom when turned to "unknown",
	   Lange-Goradia algorithm passes in success */
	recLen = MAX_PERSON_ID_LEN*aI + 100;
	rec = malloc(recLen*sizeof(char));
	if (NULL == rec)
	{
		return ckrMemErr;
	}
				
	strcpy(rec, "A possible resolution for the pedigree: turn marker ");
	MyDecimalItoa(aMarkerIdx + 1, markerIdxStr, MAX_NUM_STR_LEN);
	strcat(rec, markerIdxStr);
				
	if (1 == aI)
		strcat(rec, " in person ");
	else
		strcat(rec, " in persons ");
				
	for (j = 0; j < aI - 1; ++j)
	{
		strcat(rec, aPedigree->mPersonArray[unknownIndices[j]]->mID);
		strcat(rec, ", ");
	}
	strcat(rec, aPedigree->mPersonArray[unknownIndices[j]]->mID);
    strcat(rec, " to unknown.");

	AddRecommendation(aPedigree->mRecommendationLL, rec, recLen);

	free(rec);

	return ckrOK;
}
