/*************************************************************************************
SimlifiedParser.h
This file deals with functions that parse the input file, in the format that was
specified in the project requirements.
The main funtion of this file is ParseSimplifiedMultiPedigree which parse a whole 
input file.
*************************************************************************************/
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>



#include "pc_SimplifiedParser.h"
#include "pc_Common.h"
#include "pc_Marker.h"
#include "pc_NuclearFam.h"
#include "pc_Error.h"


#define UNKNOWN_CHAR_1 '0'
#define UNKNOWN_CHAR_2 '?'
#define UNKNOWN_CHAR_3 '-'


Result IsNewPedigreeDeclared(int familyIndex/*FILE * aPedFile*/, char * aPedName, int * nameMissingFlag);
void GoToEndOfPedigree(FILE * aPedFile);



/************************************************************************************
Function:fillPersonDetails
The function sets the fields of a person according to input and checks its correctness.
/************************************************************************************/
Person * fillPersonDetails(Person * aPerson,
						   PersonDataStruct *personData,
						   ErrorLL * aErrorLL)
{
	if (aPerson == NULL)
		return NULL;

	/* check if the input is correct */

	if (personData->sex == Male)
		aPerson->mGender = male;

	else if (personData->sex == Female)
	{
		aPerson->mGender = female;
	}
	else
	{
		SetErrorInPerson(aErrorLL, UNDEFINED_MARKER_IDX, aPerson->mID, NO_LEGAL_GENDER_ERROR);
		free (aPerson);
		return NULL;
	}

	if (personData->status == 0)
	{
		aPerson->mAffStatus = unknown;
	}
	else if (personData->status == 1)
	{
		aPerson->mAffStatus = healthy;
	}
	else if (personData->status == 2)
	{
		aPerson->mAffStatus = affected;
	}
	else
	{ 		
		SetErrorInPerson(aErrorLL, UNDEFINED_MARKER_IDX, aPerson->mID, NO_LEGAL_AFF_STATUS_ERROR);

		free (aPerson);
		return NULL;
	}

	sprintf(aPerson->mID,"%d", personData->id);
	sprintf(aPerson->mFatherID,"%d", personData->father);
	sprintf(aPerson->mMotherID,"%d", personData->mother);

	return aPerson;
}

/*************************************************************************************
Function name :ParsePersonDetails
This function "reads" five items from input file.
The items are:id of person,id of his parents, his gender and affected status.
*************************************************************************************/
ckrResult ParsePersonDetails(PersonDataStruct *personData, Person * aPerson, ErrorLL * aErrorLL)
{



	if (NULL == fillPersonDetails(aPerson, personData, aErrorLL))
	{
		return ckrErr;
	}
	return ckrOK;
}

/*************************************************************************************/
Result IsUnknown(int allele)
{


	if (allele == 0)
	{
		return YES;
	}

	return NO;
}

Result OneKnownOneNot(int * aAllele1, int * aAllele2)
{
	if ((*aAllele1 != 0) && (*aAllele2 == 0))
	{
		*aAllele1 = 0;
		return YES;
	}

	if ((*aAllele1 == 0) && (*aAllele2  != 0))
	{
		*aAllele2 = 0;
		return YES;
	}

	return NO;
}

Result TwoKnown (int * aAllele1, int * aAllele2)
{
	if ( (*aAllele1 != 0) && (*aAllele2 != 0))
	{
		return YES;
	}
	return NO;
}

void updateAlelesToEqualValue(int* aAllele1, int *aAllele2){
	if (*aAllele1 == 0){
		*aAllele1 = *aAllele2;
		return;
	}
	*aAllele2 = *aAllele1;	
}
/*************************************************************************************
Function name:ParseAllele 
This function returns allele value as an integer.
**************************************************************************************/
Result ParseAllele(int allele, int * aAllele, PedAlleles * aPedAlleles, int aMarkerIdx)
{

	double aDblAllele;

	if (YES == IsUnknown(allele))
	{
		*aAllele = 0;
		return YES;
	}


	aDblAllele = (double)allele; 
	/* TODO - what does it return in case of failure? to build our own 
	function? */

	if (aDblAllele - (int)aDblAllele <= 0.5)
		*aAllele = (int)aDblAllele;
	else
		*aAllele = (int)(aDblAllele + 1);
	/* Check or Add this allele to list of existing alleles */
	AddAlleleToPedAlleles(*aAllele, aPedAlleles, aMarkerIdx);
	return YES;
}

/*************************************************************************************
Function:DeleteMarkerList
Deletes a linked list of markers
**************************************************************************************/
void DeleteMarkerList(MarkerLLNode * aMarkerLLLst)
{
	MarkerLLNode * aMarkerLLN1;
	MarkerLLNode * aMarkerLLN2;

	aMarkerLLN1=aMarkerLLLst;
	while (aMarkerLLN1 != NULL)
	{
		aMarkerLLN2 = aMarkerLLN1->mNext;
		free(aMarkerLLN1);
		aMarkerLLN1 = aMarkerLLN2;	
	}
}

/*************************************************************************************
Function:MarkerListToArr
Creates a an array for markers according to the number of nodes in the linked list of markers.
**************************************************************************************/
Result MarkerListToArr(MarkerLLNode * aMarkerLst, Marker * aMarkerArr, int aLstLen)
{		
	int i;
	MarkerLLNode * aMarker;

	aMarker = aMarkerLst;
	for (i=aLstLen-1; i>=0; --i)
	{
		if (aMarker == NULL)
			return ckrErr;

		CopyMarkerMarkerLLN(&aMarkerArr[i], aMarker);
		aMarker = aMarker->mNext;
	}
	return ckrOK;
}


/*************************************************************************************/
/*
void GoToEndOfLine(FILE * aInFile)
{
	char c;
	c = fgetc(aInFile);
	while ((c!='\n') && (c!=EOF) && (c!=13))
		c = fgetc(aInFile);
		
}*/
/********************************************************************************&&****************
Function:SetPersonMarkers
Recieves aPerson and aMarkerLLNLst, returns ckrOK if the list was transformed correctly to an array.
***************************************************************************************************/
ckrResult SetPersonMarkers(Person * aPerson, MarkerLLNode * aMarkerLLNLst, int aMarkerCnt)
{
	int aRes;
	Marker * aMarkerArr;

	if (0 >= aMarkerCnt)
	{
		aPerson->mMarkerArr = NULL;
		return ckrOK;
	}

	aMarkerArr = CreateMarkerArray(aMarkerCnt);
	if (aMarkerArr == NULL)
	{
		return ckrErr;
	}
	aRes = MarkerListToArr(aMarkerLLNLst, aMarkerArr, aMarkerCnt);
	if (aRes == ckrErr)
	{
		return ckrErr;
	}
	aPerson->mMarkerArr = aMarkerArr;
	return ckrOK;
}
/*************************************************************************************/
ckrResult HandleEndOfParsedLine(Person * aPerson, 
								MarkerLLNode * aMarkerLLNLst, 
								int aNumOfMarkers, 
								int * aMarkerCnt,
								int aCoupleFlag, 
								ErrorLL * aErrorLL)
{
	ckrResult      aRes;
	int errFlag = 0;

	*aMarkerCnt = aNumOfMarkers;
	aPerson->mMarkersCnt = aNumOfMarkers;

	/* an allele is missing in a marker - not an even num of alleles*/
	if (1 == aCoupleFlag)
	{
		SetErrorInPerson(aErrorLL, UNDEFINED_MARKER_IDX, aPerson->mID, ODD_ALLELE_NUM_ERROR);
		errFlag = 1;
	}

	aRes = SetPersonMarkers(aPerson, aMarkerLLNLst, aNumOfMarkers);
	DeleteMarkerList(aMarkerLLNLst);

	if ((errFlag == 1) || (aRes != ckrOK))
		return ckrNonCriticErr;
	else
		return aRes;
}
/**************************************************************************************************
Function:ParsePersonMarkers
This function parses the markers of aPerson.
***************************************************************************************************/
ckrResult ParsePersonMarkers(PersonDataStruct *personData, 
							 Format aFileFormat,
							 Person * aPerson, 
							 int * aMarkerCnt, 
							 PedAlleles * aPedAlleles, 
							 ErrorLL * aErrorLL)
{
	int    i,j;
	int    aAllele1;
	int    aAllele2;
	//char   c;
	int    aCoupleFlag;
	//fpos_t         aPos;
	MarkerLLNode * aMarkerLLN;
	MarkerLLNode * aMarkerLLNLst;


	aMarkerLLNLst = NULL;

	aCoupleFlag = 0;
	for ( j = 0, i = -1; j < ProgramData.numberOfLoci - ProgramData.numberOfDiseaseLoci; j++){
		if (LociData[j].type != NUMBERED_LOCUS){
			continue;
		}else{
			i++;
		}

		//while (1)
		//{
		//	fgetpos(aInFile, &aPos);
		//	c = fgetc(aInFile);
		//	if ((c == ' ')||(c == '\t'))
		//	{
		//	continue;
		//	}
		//	if ((c == '\n')||(c == EOF)||(c == 13))
		//	{
		//		return HandleEndOfParsedLine(aPerson, aMarkerLLNLst, i, aMarkerCnt, aCoupleFlag, aErrorLL);
		//	}
		//	if (((c>='0') && (c<='9')) ||
		//		(c == UNKNOWN_CHAR_1)  ||
		//		(c == UNKNOWN_CHAR_2)  ||
		//		(c == UNKNOWN_CHAR_3))
		//	{
		//		fsetpos(aInFile, &aPos);
		//		if (aCoupleFlag == 0)
		//		{
		if (ParseAllele(personData->markersAlleles[j].allele1, &aAllele1, aPedAlleles, i) == NO){	
			DeleteMarkerList(aMarkerLLNLst);
			SetErrorInPerson(aErrorLL, i, aPerson->mID, PARSING_MARKER_ERROR);
			//GoToEndOfLine(aInFile);
			return ckrNonCriticErr;
		}
		aCoupleFlag = 1;

		//else /* aCoupleFlag == 1*/
		//{
		if (ParseAllele(personData->markersAlleles[j].allele2, &aAllele2, aPedAlleles, i) == NO)
		{	
			DeleteMarkerList(aMarkerLLNLst);
			SetErrorInPerson(aErrorLL, UNDEFINED_MARKER_IDX, aPerson->mID, ODD_ALLELE_NUM_ERROR);
			//GoToEndOfLine(aInFile);
			return ckrNonCriticErr;
		}
		/* ==my== perform the check only if autosomal format or not male */
		if ((aFileFormat == AUTOSOMAL) || (aPerson->mGender != male)){

			if (OneKnownOneNot(&aAllele1, &aAllele2) == YES)
			{
				SetErrorInPerson(aErrorLL, i, aPerson->mID, ONE_ALLELE_UNKNOWN_ERROR);
			}
		}
		/* ==my==*/		
		if ( (aFileFormat == SEXLINKED) && (aPerson->mGender == male))
		{
			if ( (TwoKnown(&aAllele1, &aAllele2) ==  YES) && 
				(aAllele1 != aAllele2) ) 
			{
				SetErrorInPerson(aErrorLL, i, aPerson->mID, MAN_ALLELES_VALUE_ERROR);
			}
			updateAlelesToEqualValue(&aAllele1, &aAllele2);
		}


		aCoupleFlag = 0;			
		aMarkerLLN = CreateMarkerLLNode(aAllele1, aAllele2);
		if (aMarkerLLN == NULL)
		{
			DeleteMarkerList(aMarkerLLNLst);
			return ckrErr;
		}
		AddMarkerToList(aMarkerLLN, &aMarkerLLNLst);

	}
	/*}
	else
	{
	SetErrorInPerson(aErrorLL, i, aPerson->mID, INVALID_ALLELE_ERROR);
	DeleteMarkerList(aMarkerLLNLst);
	GoToEndOfLine(aInFile);
	return ckrNonCriticErr;
	}*/

return HandleEndOfParsedLine(aPerson, aMarkerLLNLst, i+1, aMarkerCnt, aCoupleFlag, aErrorLL);
}

/*************************************************************************************
Function:ParseRow
The function handles the parsing of details & markers of a person.
Returns ckrOK if it was done.
**************************************************************************************/
ckrResult ParseRow(PersonDataStruct *personData,
				   Format aFileFormat,
				   Person ** aPerson, 
				   int * aMarkerCnt, 
				   PedAlleles * aPedAlleles, 
				   ErrorLL * aErrorLL)
{	
	ckrResult aRes;
	*aPerson = CreatePerson();

	if (*aPerson == NULL)
		return ckrErr;

	/* handle details of a person*/ 
	if (ParsePersonDetails(personData, *aPerson, aErrorLL) == ckrOK)
	{
		/* handle markers*/
		aRes = ParsePersonMarkers(personData/* aInFile*/, aFileFormat, *aPerson, aMarkerCnt, aPedAlleles, aErrorLL);
		if ((aRes == ckrOK) || (aRes == ckrNonCriticErr))
			//if (aRes == ckrOK)
			return ckrOK;
	}

	return ckrErr;
}


/*************************************************************************************
Function:ParseSimplified
This function handels all actions concerning "reading" the input of a person and 
setting it into the correct place.
**************************************************************************************/
Person * ParseSimplified(/*FILE * input ,*/
						 PedigreeDataStruct *pedigree,
						 Format aFileFormat,
						 int * aNumPerson,
						 int * aMarkerCnt, 
						 PedAlleles * aPedAlleles, 
						 ErrorLL * aErrorLL)
{
	int /*i,*/pId;
	int        aMarkerCnt1 = 0;
	int        aMarkerCnt2 = 0;
//	int		   aMinMarkerCnt = INT_MAX;
	Person *   aLLPerson = NULL;
	Person *   aPerson1  = NULL;
	Person *   aPerson2  = NULL;
//	char  pedigreeName[MAX_PEDIGREE_NAME_LEN];
	//int aKeepOnParsingFlag = 1;
//	int markerNumFlag = 0;
//	int dummyFlag = 0;
	//ckrResult parseRowRes = ckrOK;
	/*
	if (input == NULL)
	{
	return NULL;
	}
	*/
	/* ParseSimplified is called on the line after a Pedigree declaration (if any) */
	//while (YES == IsNewPedigreeDeclared(input, pedigreeName, &dummyFlag))
	//IsNewPedigreeDeclared(pedigree->index, pedigreeName, &dummyFlag);
	if (pedigree->numPeople < 1){
		SetError(aErrorLL, EMPTY_FAMILY_ERROR);
		return NULL;
	}

	for (pId =0; pId < pedigree->numPeople; pId++){ 
		if (ParseRow(/*input*/pedigree->people+pId, aFileFormat, &aPerson2, &aMarkerCnt2, aPedAlleles, aErrorLL)== ckrOK){
			if (aLLPerson == NULL){
				aLLPerson = aPerson2;
				aPerson1 = aPerson2;
				*aNumPerson=1;
			}else{
				aPerson1->mNext = aPerson2;
				aPerson1 = aPerson2;
				aMarkerCnt1 = aMarkerCnt2;
				aPerson2 = NULL;
				(*aNumPerson)++;
			}
			/*
			if (aMarkerCnt1 < aMinMarkerCnt){
				aMinMarkerCnt = aMarkerCnt1;
			}

				if (YES == IsNewPedigreeDeclared(input, pedigreeName, &dummyFlag))
			{
			aKeepOnParsingFlag = 0;
			}
			*/

			/*		while (1)
			{
			// stop conditions
			if (1 != aKeepOnParsingFlag)
			{
			break;
			}
			parseRowRes = ParseRow(input, aFileFormat, &aPerson2, &aMarkerCnt2, aPedAlleles, aErrorLL);
			if (ckrOK != parseRowRes)
			{
			break;
			}

			*/

		/*	if ((markerNumFlag == 0) && (aMarkerCnt2 != aMarkerCnt1)){
				SetError(aErrorLL, MARKERS_CNT_DIFF_IN_PERSONS_ERROR);
				markerNumFlag = 1;
			}

			if (aMarkerCnt2 < aMinMarkerCnt){
				aMinMarkerCnt = aMarkerCnt2;
			}*/




			/*if (YES == IsNewPedigreeDeclared(input, pedigreeName, &dummyFlag))
			{
			aKeepOnParsingFlag = 0;
			break;
			}
			}*/
		}else{

			/*GoToEndOfLine(input);
			*aNumPerson = 1;
			*aMarkerCnt = aMarkerCnt1;
			return aPerson1;
			GoToEndOfLine(input);
			GoToEndOfPedigree(input);
			*/
			*aNumPerson = 0;
			*aMarkerCnt = 0;
			return NULL;
		}
		/*
		if (ckrOK != parseRowRes)
		{
		GoToEndOfPedigree(input);
		}
		*/
	}

 
	/* the function's output aMarkerCnt is assigned to the minimum marker count 
	to assure that there will be no illegal memory access in further function */
	//*aMarkerCnt=aMinMarkerCnt;
   *aMarkerCnt = aMarkerCnt2;
	return aLLPerson;	
}


/************************************************************************************
Function:GoToEndOfPedigree
The function goes over file (without parsing it) until the next pedigree
***************************************************************************************/
/*void GoToEndOfPedigree(FILE * aPedFile)
{
	fpos_t FilePos;
	char c;
	int nameMissingFlag;
	char  pedigreeName[MAX_PEDIGREE_NAME_LEN];

	while (1)
	{
		if (NO == IsNewPedigreeDeclared(aPedFile, pedigreeName, &nameMissingFlag))
		{
			GoToEndOfLine(aPedFile);
			fgetpos(aPedFile, &FilePos);
			c = fgetc(aPedFile);
			if (c == EOF)
			{
				return;
			}
			fsetpos (aPedFile, &FilePos);
		}
		else
		{
			return;
		}
	}
}
*/
/************************************************************************************
Function:ParseSimplifiedMultiPedigree
This function is the first one in a row of functions that parse the input.
The function calls other functions to check the input and place it in the correct place.
It also checks for new pedigrees.  
***************************************************************************************/
int ParseSimplifiedMultiPedigree(PedigreesStruct Pedigrees,/* char* fileFormat,*/ PedigreeList * aPedigreeLL){

	Pedigree * aPedigree;
	Person *  aLLPerson;
//	char  pedigreeName[MAX_PEDIGREE_NAME_LEN];
	int aNumPerson;
	int MarkerNum;
	ErrorLL * aErrorLL = NULL;
	PedAlleles * aPedAlleles = NULL;
//	int nameMissingFlag = 0;
	Format aFileFormat;
//	int format;
int i;
	/*if (NULL == fileFormat)
	{
		aFileFormat = AUTOSOMAL;
	}
	else
	{
		format = atoi(fileFormat);
		switch (format){
			case SEXLINKED: 
				aFileFormat = SEXLINKED;
				break;
			default:
				aFileFormat = AUTOSOMAL;
		}
	}*/

if (ProgramData.sexLinked){
	aFileFormat = SEXLINKED;
}else{
	aFileFormat = AUTOSOMAL;
}

	//while(1){		
	for (i =0; i < Pedigrees.numOfPedigrees; i++){
		/* create a pedigree and inset it to the pedigree linked list */
		aErrorLL = CreateErrorLL();
		if (NULL == aErrorLL)
		{
			return ckrMemErr;
		}
		aPedAlleles = CreatePedAlleles();
		if (NULL == aPedAlleles)
		{
			return ckrMemErr;
		}



		aLLPerson = ParseSimplified(Pedigrees.pedigrees + i, aFileFormat, &aNumPerson, &MarkerNum, aPedAlleles, aErrorLL); 
		if (NULL == aLLPerson)
		{
			continue;
		}

		aPedigree = CreatePedigree(aLLPerson, aNumPerson, MarkerNum, aPedAlleles, aErrorLL);
		sprintf(aPedigree->mName,"%d",(Pedigrees.pedigrees + i)->index);

		aPedigree->mFormat = aFileFormat;

		AddPedigree(aPedigreeLL, aPedigree);


	}
	
	return ckrOK;
}

/************************************************************************************
Function: IsNewPedigreeDeclared
The function return YES if a new pedigree is declared (in a new line), 
and returns NO, if there is no declaration of a new pedigree.
The function does NOT change the file position.
In addition, the function returns the new pedigree's name (if no new pedigree 
declared, then the name is empty)
************************************************************************************/
Result IsNewPedigreeDeclared(/*FILE * aPedFile*/int familyIndex, char * aPedName, int * nameMissingFlag)
{
//	fpos_t FilePos;
//	char c;
//	int i;
	*nameMissingFlag = 0;
	sprintf(aPedName,"%d",familyIndex);
	return YES;
	//aPedName[0] = '\0';

	///* don't consider all white spaces */
	//while (1)
	//{
	//	fgetpos(aPedFile, &FilePos);
	//	c = fgetc(aPedFile);
	//	if ((c == ' ')||(c == '\t')||(c == '\n')||(c==13))
	//	{
	//		continue;
	//	}
	//	else if (c == EOF)
	//	{
	//		fsetpos (aPedFile, &FilePos);
	//		return NO;		
	//	}
	//	else
	//	{
	//		fsetpos(aPedFile, &FilePos);
	//		break;
	//	}
	//}

	//fgetpos(aPedFile, &FilePos);


	///* is it "pedigree" (without case sensitivity) ?? */
	//c = fgetc(aPedFile);
	//if ((c != 'p') && (c != 'P'))
	//{
	//	fsetpos (aPedFile, &FilePos);
	//	return NO;
	//}
	//c = fgetc(aPedFile);
	//if ((c != 'e') && (c != 'E'))
	//{
	//	fsetpos (aPedFile, &FilePos);
	//	return NO;
	//}
	//c = fgetc(aPedFile);
	//if ((c != 'd') && (c != 'D'))
	//{
	//	fsetpos (aPedFile, &FilePos);
	//	return NO;
	//}
	//c = fgetc(aPedFile);
	//if ((c != 'i') && (c != 'I'))
	//{
	//	fsetpos (aPedFile, &FilePos);
	//	return NO;
	//}
	//c = fgetc(aPedFile);
	//if ((c != 'g') && (c != 'G'))
	//{
	//	fsetpos (aPedFile, &FilePos);
	//	return NO;
	//}
	//c = fgetc(aPedFile);
	//if ((c != 'r') && (c != 'R'))
	//{
	//	fsetpos (aPedFile, &FilePos);
	//	return NO;
	//}
	//c = fgetc(aPedFile);
	//if ((c != 'e') && (c != 'E'))
	//{
	//	fsetpos (aPedFile, &FilePos);
	//	return NO;
	//}
	//c = fgetc(aPedFile);
	//if ((c != 'e') && (c != 'E'))
	//{
	//	fsetpos (aPedFile, &FilePos);
	//	return NO;
	//}

	///* ":" is allowed at the end also */
	//c = fgetc(aPedFile);
	//if (c == ':')
	//{
	//	c = fgetc(aPedFile);
	//}

	//while ((c == ' ') || (c == '\t'))
	//{
	//	c = fgetc(aPedFile);
	//}


	///* new line. Error! since no name was specified. 
	//   The error is reported outside the function */
	//if ((c == '\n')||(c == EOF)||(c == 13))
	//{
	//	*nameMissingFlag = 1;
	//	fsetpos (aPedFile, &FilePos);
	//	return YES;		
	//}

	//for (i=0; ((c != '\n') && (c != EOF) && (c != 13)); ++i)
	//{
	//	if (i == MAX_PEDIGREE_NAME_LEN - 1)
	//	{
	//		aPedName[i] = '\0';
	//		fsetpos (aPedFile, &FilePos);
	//		return YES;
	//	}

	//	aPedName[i] = c;
	//	c = fgetc(aPedFile);
	//}

	//if (i < MAX_PEDIGREE_NAME_LEN)
	//	aPedName[i] = '\0';
	//else
	//	aPedName[MAX_PEDIGREE_NAME_LEN - 1] = '\0';


	//fsetpos (aPedFile, &FilePos);
	//return YES;
}

