/*****************************************************************************/
/*  ArgMax.h															     */
/*  Author: Ma'ayan Fishelson												 */
/*****************************************************************************/
/*																			 */
/*  Comment: This file includes functions used for executing the			 */
/*			 MPE-elimination algorithm used in haplotyping analysis.		 */
/*																			 */
/*****************************************************************************/

#ifndef _SL_ARG_MAX_H_
#define _SL_ARG_MAX_H_

#include <string.h>
#include "sl_Auxiliary.h"
#include "sl_Node.h"

/*------------------------------*/
/*		DEFINITIONS				*/
/*------------------------------*/

#define	UNKNOWN					-1

/* The name of the file to which the haplotype information is printed to. */

#define HAPLOTYPE_FILE_NAME		"haplotypes.txt"

/*------------------------------*/
/*		DATA-STRUCTURES			*/
/*------------------------------*/

/* Definition of the type of argMax data. */

typedef char	argMaxType;
#define elimNode	0
#define breakNode	1
#define	dummyNode	2

/* This structure holds information on one of the nodes that the argmax function 
   depends on. */

typedef struct argMaxNodeData {
	int		currNumVals;		/* The number of possible values of the node at
								   the time the argmax function is created. */
	node	*origNode;			/* Pointer to the node that this structure represents. */
} argMaxNodeData;

/* This structure holds information on variables' values that were eliminated after
   the value of some break node was set. */

typedef struct argMaxElimValsVar {
	int		numElimVals;		/* The number of values of the variable that were
								   eliminated, after the value of the break node was set. */
	int		*elimVals;			/* An array that stores the values that were eliminated. */
	node	*origNode;			/* Pointer to the node that this structure represents. */
} argMaxElimValsVar;


typedef struct argMax {
	argMaxType		type;		/* The type of argMax data. */
	node			*mainNode;	/* Pointer to the node in which elimination this 
								   argMax data was created. */
	union {						/* Holds the necessary data according to the specific 
								   type of argMax. */
		struct {
			int				numNodes;		/* The number of nodes this function 
											   depends on. */
			argMaxNodeData	*nodes;			/* An array of pointers to the function's nodes. */
			unsigned long	sizeOfTable; 	/* The size of the list 'argMaxVals'. */
			int				*argMaxVals;	/* The list of values of 'mainNode' which 
											   maximized the original function for 
											   each value assignment. */
		} elimNodeData;

		struct {
			int					numNodes;	/* The number of nodes whose values were
											   eliminated after the value of this break
											   node was set. */
			argMaxElimValsVar	*nodes;		/* An array of pointers to the data on the
											   values that were eliminated for each node. */
			int					val;		/* The value of the break node. */
		} otherNodeData;
	} specificData;

	struct argMax	*next;		/* Pointer to the argMax of the previous variable 
								   in the elimination order. */
} *argMaxData;



/*------------------------------*/
/*		FUNCTIONS				*/
/*------------------------------*/

/* Functions for allocating and creating different types of argMaxData. */

argMaxData createAMData(argMaxType type, node *mainNode, argMaxData next);
argMaxData createAMDataElim(node *mainNode, argMaxData next, int numNodes, 
							argMaxNodeData *nodes, unsigned long sizeOfTable, 
							int *argMaxVals);
argMaxData createAMDataOther(node *mainNode, argMaxData next, int numNodes, 
							 argMaxElimValsVar *nodes, argMaxType type);

/* Functions for freeing the memory allocated for the argMaxData. */

void freeAMData(argMaxData argMax);
void freeAMList(argMaxData head);

/* Function used in the forward execution of the MPE-elimination algorithm to trace
   the values of the nodes. */

void retrieveArgMaxData(argMaxData argMax);
void setMaxValNode(argMaxData argMax);
void calcMaxVal(argMaxData argMax);
void updateMaxValsNodes(argMaxData argMax);


#endif /* _ARG_MAX_H_ */
