/*****************************************************************************/
/*  Auxiliary.h																 */
/*  Author: Ma'ayan Fishelson												 */
/*****************************************************************************/
/*																			 */
/*  Comment: This file includes general definitions and auxiliary functions  */
/*			 used by different modules.										 */
/*																			 */
/*****************************************************************************/

#ifndef _SL_AUXILIARY_H_
#define _SL_AUXILIARY_H_

#include <stdlib.h>
#ifdef WIN32
#include "..\PedData.h"
#include "..\LociData.h"
#include "..\globals.h"
#else
#include "../PedData.h"
#include "../LociData.h"
#include "../globals.h"
#endif
#ifdef WIN32
	#include <crtdbg.h>
	#include <Assert.h>
#endif
#include <stdio.h>
#include <setjmp.h>
#include "sl_Settings.h"

#if MANAGE_MEM
	#include "sl_MemManage.h"
#endif

#include <stdarg.h>

/*--------------------------------------*/
/*		DEFINITIONS & DATA-STRUCTURES	*/
/*--------------------------------------*/

/* Definition of a boolean type. */
#ifndef BOOLEAN__
#define BOOLEAN__
typedef int	boolean;
#define	FALSE	0
#define TRUE	1
#endif

typedef double Probability;

/* Definition of an interval in the genetic map (can be either left, middle or right). */

typedef char	interval;
#define LEFT	0
#define MIDDLE	1
#define	RIGHT	2
 
/* This structure holds the results of a single likelihood evaluation. */

typedef struct resNode {
	double				lnLike;					/* The ln-likelihood obtained in this position. */
	double				lodScore;				/* The lod-score obtained in this position. */
	double				*distFromMarker1Arr;	/* The distances of the trait loci from the first
												   marker in the scan interval (can be a different
												   marker for each trait locus). */ 
	struct resNode		*nextResNode;			/* Pointer to the next results node in the list. */
	struct resNode		*prevResNode;			/* Pointer to the previous results node in the list. */
} resNode;

/* This struct holds the results of all likelihood evaluations performed. */

typedef struct results {
	resNode			*headRes;		/* Pointer to the results node which is head of the list. */
	resNode			*currRes;		/* Pointer to the current results node. */
	resNode			*maxRes;		/* Pointer to the results node with the maximal score. */
	resNode			*unlinkedRes;	/* Pointer to the results node of the 'no-linkage' case. */
	Probability		*pedsLnLikeArr;	/* Holds the ln-likelihoods of each pedigree, given the
									   theta values, for the current likelihood evaluation. */
	boolean			*isZeroLikeArr;	/* An array of flags that indicates for each pedigree whether
									   the likelihood of that pedigree is zero. */
	double			logLike_NoLinkage;		/* The log-likelihood in the case of no-linkage. */
	Probability		*pedsLogLike_NoLinkageArr;	/* Holds the log-likelihood of each pedigree,
												   for the case of no-linkage. */
} results;

/* Definition of the type of maximization function (either the LOD-score or the
   likelihoood). */

typedef double (*retValFuncType)(resNode *resultNode);

double getLnLike(resNode *resultNode);
double getLodScore(resNode *resultNode);

#if MANAGE_MEM
	boolean manageMemory;
	jmp_buf jumper;
#endif


/* The default name for the results file. */

#define RESULTS_FILE_NAME  "results.txt"

/*------------------------------*/
/*		GLOBAL VARIABLES		*/
/*------------------------------*/

FILE	*resFile;
int		progType;  /* 0 general
				      1 construct net and/or order */


/*------------------------------*/
/*		FUNCTIONS				*/
/*------------------------------*/

#define MIN(x,y) (x < y ? x : y)
#define MAX(x,y) (x < y ? y : x)


/* Allocates a results node and initializes its fields. */

#define allocateResNode(pointer) \
	myMalloc(pointer, sizeof(resNode)); \
	myMalloc(pointer->distFromMarker1Arr, sizeof(double) * loci.numTraitLoci); \
	pointer->nextResNode = NULL;

/* Frees the memory allocated to the results node pointed to by 'pointer'. */

#define freeResNode(pointer) \
	myFree(pointer->distFromMarker1Arr); \
	myFree(pointer);

/* Allocates a results node, updates 'currRes' to point to it and updates the
   distance of the trait locus from the leftmost marker in the scan interval. */

#define initResNode(dist) \
	if (totalRes.headRes == NULL) { \
		allocateResNode(totalRes.headRes); \
		totalRes.currRes = totalRes.headRes; \
		(totalRes.currRes)->prevResNode = NULL; \
	} \
	else { \
		allocateResNode((totalRes.currRes)->nextResNode); \
		((totalRes.currRes)->nextResNode)->prevResNode = totalRes.currRes; \
		totalRes.currRes = (totalRes.currRes)->nextResNode; \
	} \
	if (loci.numTraitLoci ){ \
	(totalRes.currRes)->distFromMarker1Arr[0] = dist; \
	}


/* Allocates a results node, updates 'currRes' to point to it and updates the
   distance of the trait loci from the leftmost marker in their scan intervals. */

#define initResNode2Loc(dist1, dist2) \
	initResNode(dist1); \
	(totalRes.currRes)->distFromMarker1Arr[1] = dist2;

/* Updates the theta pointed to by 'thetaPtr' and it complement (1-theta). */

#define updateThetaCompTheta(thetaPtr, val) \
	(thetaPtr)[0] = val; \
	(thetaPtr)[1] = 1.0 - (thetaPtr)[0]; 


/* Different functions for printing error/warning messages, and for handling different errors. */

void mallocErr(char *message);
void reallocErr(char *message);
void inputErr(char *message, ...);
void inputWarning(char *message, ...);
void error(char *message, ...);

/* This function checks for EOF. */

int P_eof(FILE *f);

/* A function used for swaping between two values. */

void swap(int *val1, int *val2);

/* This function is used for printing messages with parameters. */

void myPrintf(char *str,...);

/* Different functions for handling memory allocations. */

/*void myMalloc(void* pointer, size_t size) */

#if MANAGE_MEM
	#define myMalloc(pointer, size) \
		pointer = malloc(size); \
		if (pointer == NULL){ \
			printf("size: %u\n", size); \
			mallocErr("pointer"); \
		} \
		if (manageMemory == TRUE) \
			insertBlock(pointer);
#else
	#define myMalloc(pointer, size) \
		pointer = malloc(size); \
		if (pointer == NULL) { \
			printf("Size: %u\n", size); \
			mallocErr("pointer"); \
		} 
#endif

/*void myRealloc(void* pointer, size_t size) */

#if MANAGE_MEM 
	#define myRealloc(pointer, size) \
		if (manageMemory == TRUE) \
				removeBlock(pointer); \
		pointer = realloc(pointer, size); \
		if (pointer == NULL) \
			reallocErr("pointer"); \
		if (manageMemory == TRUE) \
			insertBlock(pointer);
#else
	#define myRealloc(pointer, size) \
		pointer = realloc(pointer, size); \
		if (pointer == NULL) \
			reallocErr("pointer"); 
#endif


/*void myFree(void* pointer) */

#if MANAGE_MEM
	#define myFree(pointer) \
		if (pointer) { \
			free(pointer); \
				if (manageMemory == TRUE) \
					removeBlock(pointer); \
			pointer = NULL; \
		}
#else
	#define myFree(pointer) \
		if (pointer) { \
			free(pointer); \
			pointer = NULL; \
		}
#endif

#if MANAGE_MEM
	#define SET_THRESHOLDS() \
		manageMemory = TRUE; \
		if(!setjmp(jumper)){ \
			smallBreakThreshold = BREAK_PRODUCT_BOTTOM_THRESHOLD; \
			bigBreakThreshold = totalPhysMem() / 16.0; \
		} \
		else{ \
			bigBreakThreshold /= 3.0; \
			if(bigBreakThreshold < smallBreakThreshold){ \
				smallBreakThreshold /= 3.0; \
				bigBreakThreshold = smallBreakThreshold; \
			} \
		} \
		zeroFirstMemBlockPointers(); \
		breakProductBottomThreshold = smallBreakThreshold;
#else 
	#define SET_THRESHOLDS() \
		breakProductBottomThreshold = smallBreakThreshold = BREAK_PRODUCT_BOTTOM_THRESHOLD; \
		bigBreakThreshold = totalPhysMem() / 32.0;
#endif


#endif /* _AUXILIARY_H_ */




