/*********************************************************************************/
/*  BayesNet.c												     				 */
/*  Author: Ma'ayan Fishelson													 */
/*********************************************************************************/

#include "sl_BayesNet.h"

/******************************************************************************
*
*       Function        createBayesNetNodesPeds
*       Parameters		isFirstTime (a boolean flag that indicates whether this
*						is the first time that the Bayesian networks are created),
*						createBayesNet (pointer to the function used to create
*						the nodes for each pedigree).
*       Output          none
*       Description		Creates the necessary nodes in the Bayesian networks 
*						that represent the input pedigrees. 
*       Comments		none
*
******************************************************************************/
void createBayesNetNodesPeds(boolean isFirstTime, createBayesNetPed createBayesNet)
{
	pedigree			*pedPtr;


	/* 1.	Loop over all pedigrees. */

	for (pedPtr = peds.head; pedPtr != NULL; pedPtr = pedPtr->nextPed) {

		/* 1.1.	For each pedigree, create the nodes that represent it. */

		pedPtr->numNodesBN = createBayesNet(pedPtr, isFirstTime);
	}
}


/******************************************************************************
*
*       Function        createBayesNetNodesAutosomal
*       Parameters		pedPtr (pointer to the pedigree struct for which the
*						Bayesian network is built), isFirstTime (a boolean flag 
*						that indicates whether this is the first time that the
*						Bayesian networks are created).
*       Output          none
*       Description		Creates the necessary nodes for each person in the 
*						pedigree 'pedPtr'. 
*       Comments		Used for autosomal analysis.
*
******************************************************************************/
long createBayesNetNodesAutosomal(pedigree *pedPtr, boolean isFirstTime)
{
	person				*personPtr, *personPtrEnd;
	int					locusIndMap, locusIndDs, nodeWeight, numLoci,
						interference, *mapToData;
	long				countVariables;
	boolean				isKnown;
	locus				*locusArr;
	locusType			currLocType;
	locusPerson			*currPersonLoc;
	node				**paSelecArr, **maSelecArr, **paLocArr, **maLocArr;
	uniteType			uniteT;

#if DISCONNECT_UNLINKED
	boolean				isUnlinked;
#endif

	locusArr = loci.locusArr;
	numLoci = loci.numLoci;

	interference = recombData.interference;
	countVariables = 0;
	mapToData = loci.mapToData;


	/* 1.	Loop over all people in the pedigree. For each person: */

	for (personPtr = pedPtr->people,  personPtrEnd = personPtr + pedPtr->numPeople;
			personPtr < personPtrEnd; personPtr++) {

		/* 1.1.	Update pointers to the arrays of pointers to the nodes of this person. */

		paSelecArr = personPtr->selectorVar[0];
		maSelecArr = personPtr->selectorVar[1];
		paLocArr = personPtr->locusVar[0];
		maLocArr = personPtr->locusVar[1];

		/* 1.2.	Loop over all loci. */

		for (locusIndMap = 0; locusIndMap < numLoci; locusIndMap++) {

			/* 1.2.1.	Update pointers to the general locus structure and to the 
						locus structure of this person. */

			locusIndDs = mapToData[locusIndMap];
			currLocType = locusArr[locusIndDs].type;
			currPersonLoc = &(personPtr->loci[locusIndDs]);

			#if DISCONNECT_UNLINKED
				if ((locusIndMap > 0) && (recombData.maleThetas[locusIndMap - 1][0] == 0.5)) {
					isUnlinked = TRUE;
				}
				else {
					isUnlinked = FALSE;
				}
			#endif

			/* 1.2.2.	Update the type of union between the locus variables of 
						this person at this locus. */

			#if UNITE_VARS
				if (currLocType == ALLELES ) {
					uniteT = currPersonLoc->locusData.allelesL.uniteT;
				}
				else {
			#endif
				uniteT = noUnite;
			#if UNITE_VARS
			}
			#endif

			/* 1.2.3.	Create a paternal selector node, if this person has parents. */

			if (personPtr->pa != NULL) {	

				/* 1.2.3.1.	Determine the number of possible values for this node. 
							If it is one, then the value of this node is known. */

				if ((currLocType != ALLELES) || 
						(currPersonLoc->locusData.allelesL.paSelectorVals[1] != -1)) {
					nodeWeight = 2;
					isKnown = FALSE;
				}
				else {
					nodeWeight = 1;
					isKnown = TRUE;
				}

				/* 1.2.3.2.	Allocate memory and initialize the fields of the selector node. */

				if (locusIndMap == 0) {
					paSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
											   SELECTOR, PATERNAL, isKnown, noUnite);
				}
			#if DISCONNECT_UNLINKED
				else if (isUnlinked) {
					paSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
											   SELECTOR, PATERNAL, isKnown, noUnite);
				}
			#endif
				else if ((!interference) || (interference && locusIndMap == 1)) {
					paSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 1, personPtr->pNum, locusIndMap, 
											   SELECTOR, PATERNAL, isKnown, noUnite);
				}
				else {
					paSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 2, personPtr->pNum, locusIndMap, 
											   SELECTOR, PATERNAL, isKnown, noUnite);
				}

				/* 1.2.3.3.	Initialize the name of the node (for debugging). */

				#if DEBUG_STAT
					sprintf(paSelecArr[locusIndDs]->name, "ps_%d_%d", personPtr->pNum,
							locusIndMap);
				#endif

				/* 1.2.3.4.	Update the count of the number of nodes in the Bayesian network. */

				countVariables++;
			}
			else {
				paSelecArr[locusIndDs] = NULL;
			}

			/* 1.2.4.	Create a paternal locus node. */

			/* 1.2.4.1.	Update the number of possible values for this node and whether
						it is known. */

			if (currLocType == ALLELES) {
				if (isFirstTime) {

					/* 1.2.4.1.1.	The array of possible values needs to be created only
									if this is the first time the Bayesian networks are
									created. */

					#if PERFORM_ALLELE_RECODING
						nodeWeight = createValArrayAR(currPersonLoc, TRUE, &isKnown);
					#else
						nodeWeight = createValArray(currPersonLoc, TRUE, &isKnown);
					#endif
						
					currPersonLoc->locusData.allelesL.isKnownPa = isKnown;
				}
				else {
					nodeWeight = currPersonLoc->locusData.allelesL.numPaAlleleGroups;
					isKnown = currPersonLoc->locusData.allelesL.isKnownPa;
				}
			}
			else {
				nodeWeight = 2;
				isKnown = FALSE;
			}

			/* 1.2.4.2.	Allocate memory and initialize the fields of the node. */
		
			if (personPtr->pa == NULL)	{				/* this is a founder. */
				paLocArr[locusIndDs] = 
					allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
										   LOCUS, PATERNAL, isKnown, uniteT);
			}
			else if ((locusIndMap == 0) && ((personPtr->pa)->pa == NULL)) {
				paLocArr[locusIndDs] = 
					allocateInitializeNode(nodeWeight, 2, personPtr->pNum, locusIndMap, 
										   LOCUS, PATERNAL, isKnown, uniteT);
			}
			else {
				paLocArr[locusIndDs] = 
					allocateInitializeNode(nodeWeight, 3, personPtr->pNum, locusIndMap, 
										   LOCUS, PATERNAL, isKnown, uniteT);
			}

			#if DEBUG_STAT
				sprintf(paLocArr[locusIndDs]->name, "pl_%d_%d", personPtr->pNum,
						locusIndMap);
			#endif

			countVariables++;

			/* 1.2.5.	Create a maternal selector node, if this person has parents. */

			if (personPtr->ma != NULL) {	

				if ((currLocType != ALLELES) || 
						(currPersonLoc->locusData.allelesL.maSelectorVals[1] != -1)) {
					nodeWeight = 2;
					isKnown = FALSE;
				}
				else {
					nodeWeight = 1;
					isKnown = TRUE;
				}

				if (locusIndMap == 0) {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}
			#if DISCONNECT_UNLINKED
				else if (isUnlinked) {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}
			#endif
				else if ((!interference) || (interference && locusIndMap ==1)) {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 1, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}
				else {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 2, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}

				#if DEBUG_STAT
					sprintf(maSelecArr[locusIndDs]->name, "ms_%d_%d", personPtr->pNum,
							locusIndMap);
				#endif

				countVariables++;
			}
			else {
				maSelecArr[locusIndDs] = NULL;
			}

			/* 1.2.6.	Create a maternal locus node. */

			if (currLocType == ALLELES) {		/* This is an ALLELES type locus. */

				if (isFirstTime) {

					#if PERFORM_ALLELE_RECODING
						nodeWeight = createValArrayAR(currPersonLoc, FALSE, &isKnown);	
					#else
						nodeWeight = createValArray(currPersonLoc, FALSE, &isKnown);	
					#endif

					currPersonLoc->locusData.allelesL.isKnownMa = isKnown;
				}
				else {
					nodeWeight = currPersonLoc->locusData.allelesL.numMaAlleleGroups;
					isKnown = currPersonLoc->locusData.allelesL.isKnownMa;
				}

				/* 1.2.6.1	If this isn't the first locus of a founder,
							the person is typed and hetrozygote at this locus, 
							create a phenotype node. */

			if  ((!UNITE_VARS || RUN_CREATE_BAYES) &&
					 (((personPtr->pa != NULL) || (locusIndMap != 0)) && 
						(personPtr->loci[locusIndDs].unknown == FALSE) && (nodeWeight == 2))) {
							personPtr->phenotypeVar[locusIndDs] = 
								allocateInitializeNode(1, 2, personPtr->pNum, locusIndMap, PHENOTYPE, PATERNAL, TRUE, noUnite);
							countVariables++;

						sprintf(personPtr->phenotypeVar[locusIndDs]->name, "ph_%d_%d", 
								personPtr->pNum, locusIndMap); 
					}
					else {
				
					personPtr->phenotypeVar[locusIndDs] = NULL;
			}
				
			}
			else if (currLocType == affection) {	/* This is an affection type locus. */
				nodeWeight = 2;
				isKnown = FALSE;
	
				/* 1.2.6.1	If the phenotype of the person at this locus is unknown,
							create a phenotype node. */

				if (currPersonLoc->unknown == FALSE) {

					/* 1.2.6.1.1.	If this isn't the first locus of a founder, then the 
									parents of the nodes in the Bayesian network are the 2 
									locus nodes of this person at this locus. Otherwise, 
									the parent of this node is the genotype node of this 
									person at this locus. */

					if ((personPtr->pa != NULL) || (locusIndMap != 0)) {
						personPtr->phenotypeVar[locusIndDs] = 
							allocateInitializeNode(1, 2, personPtr->pNum, locusIndMap, 
												   PHENOTYPE, PATERNAL, TRUE, noUnite);
					}
					else {
						personPtr->phenotypeVar[locusIndDs] = 
							allocateInitializeNode(1, 1, personPtr->pNum, locusIndMap, 
												   PHENOTYPE, PATERNAL, TRUE, noUnite);
					}

					countVariables++;

					#if DEBUG_STAT
						sprintf(personPtr->phenotypeVar[locusIndDs]->name, "ph_%d_%d", personPtr->pNum,
								locusIndMap);
					#endif
				}
				else
					personPtr->phenotypeVar[locusIndDs] = NULL;
			}
			else if (((locusIndMap == loci.traitLociPos[0]) && (loci.traitLociPos[0] > loci.traitLociPos[1])) ||
					 ((locusIndMap == loci.traitLociPos[1]) && (loci.traitLociPos[1] > loci.traitLociPos[0]))) {	

				/* This is the last (rightmost) null type locus on the current map. */

				nodeWeight = 2;
				isKnown = FALSE;

				/* 1.2.6.1.	If the phenotype of the person at this locus is unknown,
							create a phenotype node. */

				if (personPtr->loci[loci.lastNullLocusIndDS].unknown == FALSE) {

					/* 1.2.6.1.1.	If neither of the null loci are the first locus of a 
									founder, then the parents of the phenotype node are the 
									2 loci variables in each of the null loci. Otherwise, the
									parents of this node are the 2 loci variables in one null 
									locus and the genotype node at the other null locus. */

					if ((personPtr->pa != NULL) || 
							((loci.traitLociPos[0] != 0) && (loci.traitLociPos[1] != 0))) {
						personPtr->phenotypeVar[locusIndDs] = 
							allocateInitializeNode(1, 4, personPtr->pNum, locusIndMap, 
												   PHENOTYPE, PATERNAL, TRUE, noUnite);
					}
					else {
						personPtr->phenotypeVar[locusIndDs] = 
							allocateInitializeNode(1, 3, personPtr->pNum, locusIndMap, 
												   PHENOTYPE, PATERNAL, TRUE, noUnite);
					}

					countVariables++;

					#if DEBUG_STAT
						sprintf(personPtr->phenotypeVar[locusIndDs]->name, "ph_%d_%d", personPtr->pNum,
								locusIndMap); 
					#endif
				}
				else
					personPtr->phenotypeVar[locusIndDs] = NULL;
			}
			else
				personPtr->phenotypeVar[locusIndDs] = NULL;

			/* 1.2.7. Allocate memory and initialize the fields of the maternal locus node. */

			if (personPtr->ma == NULL) {	/* this is a founder. */
				maLocArr[locusIndDs] = 
					allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
										   LOCUS, MATERNAL, isKnown, uniteT);
			}
			else if ((locusIndMap == 0) && ((personPtr->ma)->pa == NULL)) {
				maLocArr[locusIndDs] = 
					allocateInitializeNode(nodeWeight, 2, personPtr->pNum, locusIndMap, 
										   LOCUS, MATERNAL, isKnown, uniteT);
			}
			else {
				maLocArr[locusIndDs] = 
					allocateInitializeNode(nodeWeight, 3, personPtr->pNum, locusIndMap, 
											   LOCUS, MATERNAL, isKnown, uniteT);
			}

			sprintf(maLocArr[locusIndDs]->name, "ml_%d_%d", personPtr->pNum,
					locusIndMap); 

			/* 1.2.8.	Update unite information if the two locus nodes need to be treated together. */

			#if UNITE_VARS
				if (uniteT == oneAll || uniteT == twoAll) {
					updateUniteInfo(uniteT, paLocArr, maLocArr, locusIndDs, currPersonLoc);
				}
			#endif

			countVariables++;

			/* 1.2.9.	Create a genotype node, if this is the first locus of a founder. */

			if ((locusIndMap == 0) && (personPtr->pa == NULL)) {

				/* 1.2.9.1.	Calculate the number of different possible genotypes. */

				nodeWeight = calcNumValsGenotype(personPtr, locusIndDs, &isKnown);
				personPtr->genotypeVar = 
					allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
										   GENOTYPE, PATERNAL, isKnown, noUnite);

				#if DEBUG_STAT
					sprintf(personPtr->genotypeVar->name, "ge_%d_%d", personPtr->pNum,
							locusIndMap); 
				#endif

				/* 1.2.9.2.	The 2 original locus nodes are replaced by 1 genotype node. */

				countVariables--;
			}
		}
	}

	/* 2.	Update the genetic loci variables that need to be eliminated jointly 
			with other nodes. */

	#if UNITE_VARS
		updateUniteFourVars(pedPtr);	
	#endif

	/* 3.	Return the number of nodes in the Bayesian network created. */

	return countVariables;
}


/******************************************************************************
*
*       Function        updateUniteInfo
*       Parameters		pedPtr (pointer to the pedigree struct for which the
*						Bayesian network is built).
*       Output          none
*       Description		Updates the unite information in case the two locus 
*						nodes of a person need to be treated together in the
*						computations. 
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void updateUniteInfo(uniteType uniteT, node **paLocArr, node **maLocArr, 
					 int locusIndDs, locusPerson *currPersonLoc)
{
	node		**uniteVars, *paNode, *maNode;

	/* 1.	Update pointers to the paternal and maternal locus nodes of the person
			at locus no. 'locusIndDs'. */

	paNode = paLocArr[locusIndDs];
	maNode = maLocArr[locusIndDs];

	/* 2.	If one of the locus nodes is constant, then the nodes are not united in
			the computations. */

	if ((paNode->numValues == 1) || (maNode->numValues == 1) || RUN_CREATE_BAYES) {
		paNode->uniteT = noUnite;
		maNode->uniteT = noUnite;
		return;
	}

	/* 3.	Allocate memory for an array of pointers to the two nodes and initialize it. */

	myMalloc(uniteVars, sizeof(node*) * 2);
	uniteVars[0] = paNode;
	uniteVars[1] = maNode;

	/* 4.	Update the unite information in each one of the nodes. */

	updateUniteVarsInfo(paNode, uniteVars, TRUE, 2);
	updateUniteVarsInfo(maNode, uniteVars, FALSE, 2);

	/* 5.	If the unite is of type 'oneAll': */

	if (uniteT == oneAll) {

		/* 5.1. Allocate memory to store the index of the certain value 
				in the array of possible values for each of the locus nodes. */

		myMalloc(maLocArr[locusIndDs]->certainVal, sizeof(int));
		myMalloc(paLocArr[locusIndDs]->certainVal, sizeof(int));

		/* 5.1. Find the index of the certain value in the array of
				possible values for each of the locus nodes. */

		#if PERFORM_ALLELE_RECODING
			maLocArr[locusIndDs]->certainVal[0] = 
				findCertainValIndAR(currPersonLoc->locusData.allelesL.maAllelsGroups,
									currPersonLoc->locusData.allelesL.numMaAlleleGroups,
									currPersonLoc->locusData.allelesL.certainAllele[0]);
			paLocArr[locusIndDs]->certainVal[0] = 
				findCertainValIndAR(currPersonLoc->locusData.allelesL.paAllelsGroups,
									currPersonLoc->locusData.allelesL.numPaAlleleGroups,
									currPersonLoc->locusData.allelesL.certainAllele[0]);
		#else
			maLocArr[locusIndDs]->certainVal[0] = 
				findCertainValInd(currPersonLoc->locusData.allelesL.maAlleles,
									currPersonLoc->locusData.allelesL.numMaAlleleGroups,
									currPersonLoc->locusData.allelesL.certainAllele[0]);
			paLocArr[locusIndDs]->certainVal[0] = 
				findCertainValInd(currPersonLoc->locusData.allelesL.paAlleles,
									currPersonLoc->locusData.allelesL.numPaAlleleGroups,
									currPersonLoc->locusData.allelesL.certainAllele[0]);

		#endif
	}
}
#endif


/******************************************************************************
*
*       Function        createBayesNetNodesSexLinked
*       Parameters		pedPtr (pointer to the pedigree struct for which the
*						Bayesian network is built), isFirstTime (a boolean flag 
*						that indicates whether this is the first time that the
*						Bayesian networks are created).
*       Output          none
*       Description		Creates the necessary nodes for each person in the 
*						pedigree 'pedPtr'. 
*       Comments		Used for sex-linked analysis.
*
******************************************************************************/
long createBayesNetNodesSexLinked(pedigree *pedPtr, boolean isFirstTime)
{
	person				*personPtr, *personPtrEnd;
	int					locusIndMap, locusIndDs, nodeWeight, numLoci, interference, 
						*mapToData;
	long				countVariables;
	boolean				isKnown;
	locus				*locusArr;
	locusType			currLocType;
	locusPerson			*currPersonLoc;
	node				**maSelecArr, **paLocArr, **maLocArr;
	uniteType			uniteT;

#if DISCONNECT_UNLINKED
	boolean				isUnlinked;
#endif

	countVariables = 0;
	locusArr = loci.locusArr;
	numLoci = loci.numLoci;
	interference = recombData.interference;
	mapToData = loci.mapToData;

	/* 1.	Loop over all people in the pedigree. */

	for (personPtr = pedPtr->people,  personPtrEnd = personPtr + pedPtr->numPeople;
			personPtr < personPtrEnd; personPtr++) {

		/* 1.1.	Update pointers to the arrays of pointers to the nodes of this person. */

		maSelecArr = personPtr->selectorVar[1];
		maLocArr = personPtr->locusVar[1];

		if (personPtr->personSex == Female) {
			paLocArr = personPtr->locusVar[0];
		}

		/* 1.2.	For each person, loop over all loci. */

		for (locusIndMap = 0; locusIndMap < numLoci; locusIndMap++) {
			locusIndDs = mapToData[locusIndMap];
			currLocType = locusArr[locusIndDs].type;
			currPersonLoc = &(personPtr->loci[locusIndDs]);
			uniteT = noUnite;

			#if DISCONNECT_UNLINKED
				if ((locusIndMap > 0) && (recombData.maleThetas[locusIndMap - 1][0] == 0.5)) {
					isUnlinked = TRUE;
				}
				else {
					isUnlinked = FALSE;
				}
			#endif

			/* 1.2.1.	If this is a female, create a paternal locus node. */

			if (personPtr->personSex == Female) {
				if (currLocType == ALLELES) {		
					if (isFirstTime) {

						#if PERFORM_ALLELE_RECODING
							nodeWeight = createValArrayAR(currPersonLoc, TRUE, &isKnown);
						#else
							nodeWeight = createValArray(currPersonLoc, TRUE, &isKnown);
						#endif

						(currPersonLoc->locusData).allelesL.isKnownPa = isKnown;
					}
					else {
						nodeWeight = (currPersonLoc->locusData).allelesL.numPaAlleleGroups;
						isKnown = (currPersonLoc->locusData).allelesL.isKnownPa;
					}

					#if UNITE_VARS
						uniteT = (currPersonLoc->locusData).allelesL.uniteT;
					#endif
				}
				else {
					nodeWeight = 2;
					isKnown = FALSE;
				}
			
				if (personPtr->pa == NULL)	{ /* this is a founder. */
					paLocArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
											   LOCUS, PATERNAL, isKnown, uniteT);
				}
				else {
					paLocArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 1, personPtr->pNum, locusIndMap, 
											   LOCUS, PATERNAL, isKnown, uniteT);
				}

				#if DEBUG_STAT
					sprintf(paLocArr[locusIndDs]->name, "pl_%d_%d", personPtr->pNum,
							locusIndMap); 
				#endif

				countVariables++;
			}

			/* 1.2.2.	Create a maternal selector node, if this person has parents. */

			if (personPtr->ma != NULL) {	

				if ((currLocType != ALLELES) ||
						(currPersonLoc->locusData.allelesL.maSelectorVals[1] != -1)) {
					nodeWeight = 2;
					isKnown = FALSE;
				}
				else {
					nodeWeight = 1;
					isKnown = TRUE;
				}

				if (locusIndMap == 0) {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}
			#if DISCONNECT_UNLINKED
				else if (isUnlinked) {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}
			#endif
				else if ((!interference) || (interference && locusIndMap == 1)) {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 1, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}
				else {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 2, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}

				#if DEBUG_STAT
					sprintf(maSelecArr[locusIndDs]->name, "ms_%d_%d", personPtr->pNum,
							locusIndMap);  
				#endif	

				countVariables++;
			}
			else {
				maSelecArr[locusIndDs] = NULL;
			}

			/* 1.2.3.	Create a maternal locus node. */
           
			if (currLocType == ALLELES) {		/* This is an ALLELES type locus. */
				if (isFirstTime) {

					#if PERFORM_ALLELE_RECODING
						nodeWeight = createValArrayAR(currPersonLoc, FALSE, &isKnown);
					#else
						nodeWeight = createValArray(currPersonLoc, FALSE, &isKnown);
					#endif

					currPersonLoc->locusData.allelesL.isKnownMa = isKnown;
				}
				else {
					nodeWeight = currPersonLoc->locusData.allelesL.numMaAlleleGroups;
					isKnown = currPersonLoc->locusData.allelesL.isKnownMa;
				}

				/* 1.2.3.1	If this person isn't the first locus of a founder, the person is
							typed and heterozygote at this locus, create a phenotype node. */
				
				if (((personPtr->pa != NULL) || (locusIndMap != 0)) && 
						(currPersonLoc->unknown == FALSE) && (nodeWeight == 2)) {
					personPtr->phenotypeVar[locusIndDs] = 
						allocateInitializeNode(1, 2, personPtr->pNum, locusIndMap, 
											   PHENOTYPE, PATERNAL, TRUE, noUnite);
					countVariables++;
					
					#if DEBUG_STAT
						sprintf(personPtr->phenotypeVar[locusIndDs]->name, "ph_%d_%d", personPtr->pNum,
								locusIndMap); 
					#endif
				}
				else
					personPtr->phenotypeVar[locusIndDs] = NULL;
			}
			else if (currLocType == affection) {	/* This is an affection type locus. */
				nodeWeight = 2;
				isKnown = FALSE;
				
				/* 1.2.3.1	If the phenotype of the person at this locus is known,
							create a phenotype node. */
				
				if (currPersonLoc->unknown == FALSE) {
					
					if ((personPtr->personSex == Female) && ((personPtr->pa != NULL) || (locusIndMap != 0))) {
						personPtr->phenotypeVar[locusIndDs] = 
							allocateInitializeNode(1, 2, personPtr->pNum, locusIndMap, 
												   PHENOTYPE, PATERNAL, TRUE, noUnite);
					}
					else {
						personPtr->phenotypeVar[locusIndDs] = 
							allocateInitializeNode(1, 1, personPtr->pNum, locusIndMap, 
												   PHENOTYPE, PATERNAL, TRUE, noUnite);
					}
					
					countVariables++;
					
					#if DEBUG_STAT
						sprintf(personPtr->phenotypeVar[locusIndDs]->name, "ph_%d_%d", personPtr->pNum,
								locusIndMap);  
					#endif
				}
				else
					personPtr->phenotypeVar[locusIndDs] = NULL;
			}
			else if (((locusIndMap == loci.traitLociPos[0]) && (loci.traitLociPos[0] > loci.traitLociPos[1])) ||
					 ((locusIndMap == loci.traitLociPos[1]) && (loci.traitLociPos[1] > loci.traitLociPos[0]))) {	

				 /* This is the last (rightmost) null type locus on the map. */

				nodeWeight = 2;
				isKnown = FALSE;
			
				/* 1.2.3.1	If the phenotype of the person at this locus is known,
							create a phenotype node. */
				
				if (personPtr->loci[loci.lastNullLocusIndDS].unknown == FALSE) {
	
					if (personPtr->personSex == Female) {
						if ((personPtr->pa != NULL) ||
								((loci.traitLociPos[0] != 0) && (loci.traitLociPos[1] != 0))) {

							personPtr->phenotypeVar[locusIndDs] = 
								allocateInitializeNode(1, 4, personPtr->pNum, locusIndMap, 
													   PHENOTYPE, PATERNAL, TRUE, noUnite);
						}
						else {
							personPtr->phenotypeVar[locusIndDs] = 
								allocateInitializeNode(1, 3, personPtr->pNum, locusIndMap, 
													   PHENOTYPE, PATERNAL, TRUE, noUnite);
						}
					}
					else {
						personPtr->phenotypeVar[locusIndDs] = 
							allocateInitializeNode(1, 2, personPtr->pNum, locusIndMap, 
												   PHENOTYPE, PATERNAL, TRUE, noUnite);
					}
					
					countVariables++;
					
					#if DEBUG_STAT
						sprintf(personPtr->phenotypeVar[locusIndDs]->name, "ph_%d_%d", personPtr->pNum,
							locusIndMap); 
					#endif
				}
				else
					personPtr->phenotypeVar[locusIndDs] = NULL;
			}
			else {
				nodeWeight = 2;
				isKnown = FALSE;
				
				personPtr->phenotypeVar[locusIndDs] = NULL;
			}
			
			if (personPtr->ma == NULL)	{  /* This is a founder. */
				maLocArr[locusIndDs] = 
					allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
										   LOCUS, MATERNAL, isKnown, uniteT);
			}
			else if ((locusIndMap == 0) && ((personPtr->ma)->pa == NULL)) {
				maLocArr[locusIndDs] = 
					allocateInitializeNode(nodeWeight, 2, personPtr->pNum, locusIndMap, 
										   LOCUS, MATERNAL, isKnown, uniteT);
			}
			else {
				maLocArr[locusIndDs] = 
					allocateInitializeNode(nodeWeight, 3, personPtr->pNum, locusIndMap, 
										   LOCUS, MATERNAL, isKnown, uniteT);
			}

			#if DEBUG_STAT
				sprintf(maLocArr[locusIndDs]->name, "ml_%d_%d", personPtr->pNum,
						locusIndMap);  
			#endif

			countVariables++;

			/* 1.2.4.	Update unite information if the two locus nodes need to be treated together. */

			#if UNITE_VARS
				if (personPtr->personSex == Female) {
					if (uniteT == oneAll || uniteT == twoAll) {
						updateUniteInfo(uniteT, paLocArr, maLocArr, locusIndDs, currPersonLoc);
					}
				}
			#endif

			/* 1.2.5.	Create a genotype node, if this is the first locus of a female founder. */

			if ((personPtr->personSex == Female) && (locusIndMap == 0) && (personPtr->pa == NULL)) {

				/* 1.2.5.1.	Calculate the number of different possible genotypes. */

				nodeWeight = calcNumValsGenotype(personPtr, locusIndDs, &isKnown);
				personPtr->genotypeVar = allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
																GENOTYPE, PATERNAL, isKnown, noUnite);

				#if DEBUG_STAT
					sprintf(personPtr->genotypeVar->name, "ge_%d_%d", personPtr->pNum,
							locusIndMap);  
				#endif

				/* 1.2.5.2.	The 2 original locus nodes are replaced by 1 genotype node. */

				countVariables--;
			}
		}
	}

	/* 2.	Return the number of nodes in the Bayesian network created. */

	return countVariables;
}



/******************************************************************************
*
*       Function        updateUniteFourVars
*       Parameters		pedPtr (pointer to the pedigree struct for which the
*						nodes which need to be eliminated jointly with 3 other
*						nodes are being updated).
*       Output          none
*       Description		Updates the genetic loci nodes which need to be 
*						eliminated jointly with 3 other nodes, for each person 
*						in the pedigree 'pedPtr'. 
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void updateUniteFourVars(pedigree *pedPtr)
{
	person				*personPtr, *personPtrEnd, *otherPerson;
	locus				*locusArr;
	locusPerson			*currPersonLoc;
	node				**paLocArr, **maLocArr, **uniteNodes;
	uniteType			uniteT;
	int					pId, locusIndMap, locusIndDs, numLoci, *mapToData;

	numLoci = loci.numLoci;
	locusArr = loci.locusArr;
	mapToData = loci.mapToData;

	/* 1.	Loop over all people in the pedigree. */

	for (personPtr = pedPtr->people,  personPtrEnd = personPtr + pedPtr->numPeople;
			personPtr < personPtrEnd; personPtr++) {
			
		/* 1.1.	Update pointers to the arrays of pointers to the nodes of this person. */

		paLocArr = personPtr->locusVar[0];
		maLocArr = personPtr->locusVar[1];

		/* 1.2.	For each person, loop over all loci. */

		for (locusIndMap = 0; locusIndMap < numLoci; locusIndMap++) {

			/* 1.2.1.	Update pointers to the general locus structure and to the 
						locus structure of this person. */

			locusIndDs = mapToData[locusIndMap];
			currPersonLoc = &(personPtr->loci[locusIndDs]);

			/* 1.2.2.	Update the type of union between the locus variables of 
						this person at this locus. */

			if (locusArr[locusIndDs].type == ALLELES) {
				uniteT = currPersonLoc->locusData.allelesL.uniteT;
			}
			else {
				continue;
			}

			/* 1.2.3.	In case the two genetic loci variables of this person at
						this locus need to be eliminated jointly with the two 
						genetic loci variables of another person, allocate memory
						for an array that would store pointers to these nodes, and 
						update them. */
				
			if (uniteT == fourAll) {

				/* 1.2.3.1.	Update a pointer to the other person. */

				pId = ((currPersonLoc->locusData).allelesL).unitePerson - 1;
				otherPerson = &(pedPtr->people[pId]);

				/* 1.2.3.2.	If the ID of the other person is larger than this person's
							ID, then allocate memory for an array that would store 
							pointers to the nodes that need to be united, update it with
							the pointers, and update the paternal locus node of this person
							with information regarding the union. */

				if (pId > personPtr->pNum - 1) {
					myMalloc(uniteNodes, sizeof(node*) * 4);
					uniteNodes[0] = paLocArr[locusIndDs];
					uniteNodes[1] = maLocArr[locusIndDs];
					uniteNodes[2] = otherPerson->locusVar[0][locusIndDs];
					uniteNodes[3] = otherPerson->locusVar[1][locusIndDs];

					updateUniteNodesVals(paLocArr[locusIndDs], uniteNodes, TRUE, 
										 currPersonLoc, PATERNAL); 
				}
				else {
					/* 1.2.3.2.	Otherwise, update the paternal locus node of 
								this person with information regarding the union. */

					uniteNodes = (otherPerson->locusVar[0][locusIndDs])->uniteVars;

					updateUniteNodesVals(paLocArr[locusIndDs], uniteNodes, FALSE, 
										 currPersonLoc, PATERNAL); 
				}

				/* 1.2.3.3.	Update the maternal locus node of this person
							with information regarding the union. */

				updateUniteNodesVals(maLocArr[locusIndDs], uniteNodes, FALSE, 
									 currPersonLoc, MATERNAL);
			}
		}
	}
}
#endif

/******************************************************************************
*
*       Function        updateUniteNodesVals
*       Parameters		mainNode (pointer to the main node which needs to be 
*						updated regarding the union), uniteNodes (an array
*						of pointers to nodes that need to be eliminated jointly
*						with this node), isMyUniteNodes (a flag that indicates 
*						whether 'mainNode' is the node which stores the main copy
*						of the array 'uniteNodes', for freeing-memory purposes),
*						currLoc (pointer to the locus structure of the person
*						that 'mainNode' represents),  sex (the sex of this person).
*       Output          none
*       Description		Updates 'mainNode' with information on the nodes which 
*						need to be eliminated jointly with it, and on the indices
*						of the possible values for this node.
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void updateUniteNodesVals(node *mainNode, node **uniteNodes, boolean isMyUniteNodes, 
						  locusPerson *currLoc, NodeSex sex)
{
	int		*certainValNode, *certainValsTotal, ind;


	#if PERFORM_ALLELE_RECODING
		alleleSet	*alleleGroupsPtr, *alleleGroupsPtrEnd;
	#else
		int			*alleleArrPtr, *alleleArrPtrEnd;
	#endif


	/* 1.	Update the node to point to the array of nodes with which it needs to
			be eliminated jointly, update the number of nodes in the array, and 
			whether the original array belongs to this node (for memory-freeing 
			purposes). */

	mainNode->uniteT = fourAll;
	mainNode->numVarsUnite = 4;
	mainNode->isMyUniteVars = isMyUniteNodes;
	mainNode->uniteVars = uniteNodes;

	/* 2.	Allocate memory for an array that would store the indices in the total 
			certain-val-array of the possible values for this node, and update a pointer
			to its beginning. */

	myMalloc(mainNode->certainVal, sizeof(int) * mainNode->numValues);
	certainValNode = mainNode->certainVal;

	/* 3.	Update a pointer to the array which holds all four certain ALLELES for the
			four locus nodes. */

	certainValsTotal = ((currLoc->locusData).allelesL).certainAllele;

	#if PERFORM_ALLELE_RECODING

		/* 4.	Update a pointer to the beginning of the array which holds the possible 
				values for this node. */

		if (sex == PATERNAL) {
			alleleGroupsPtr = ((currLoc->locusData).allelesL).paAllelsGroups;
			alleleGroupsPtrEnd = alleleGroupsPtr + ((currLoc->locusData).allelesL).numPaAlleleGroups;
		}
		else {
			alleleGroupsPtr = ((currLoc->locusData).allelesL).maAllelsGroups;
			alleleGroupsPtrEnd = alleleGroupsPtr + ((currLoc->locusData).allelesL).numMaAlleleGroups;
		}

		/* 5.	Iterate over the values in the array, and update for each one the index in
				the certain-val-array. */

		for (ind = 0; alleleGroupsPtr < alleleGroupsPtrEnd; certainValsTotal++, ind++) {

			if (alleleGroupsPtr->alleleArr[0] == (*certainValsTotal)) {
				(*certainValNode) = ind;
				alleleGroupsPtr++;
				certainValNode++;
			}
		}
	#else
		/* 4.	Update a pointer to the beginning of the array which holds the possible 
				values for this node. */

		if (sex == PATERNAL) {
			alleleArrPtr = ((currLoc->locusData).allelesL).paAlleles;
			alleleArrPtrEnd = alleleArrPtr + ((currLoc->locusData).allelesL).numPaAlleleGroups;
		}
		else {
			alleleArrPtr = ((currLoc->locusData).allelesL).paAlleles;
			alleleArrPtrEnd = alleleArrPtr + ((currLoc->locusData).allelesL).numPaAlleleGroups;
		}

		/* 5.	Iterate over the values in the array, and update for each one the index in
				the certain-val-array. */

		for (ind = 0;alleleArrPtr < alleleArrPtrEnd; certainValsTotal++) {

			if ((*alleleArrPtr) == (*certainValsTotal)) {
				(*certainValNode) = ind;
				alleleArrPtr++;
				certainValNode++;
			}
		}

	#endif
}
#endif


/******************************************************************************
*
*       Function        createBayesNetNodesAutosomalHaplo
*       Parameters		pedPtr (pointer to the pedigree struct for which the
*						Bayesian network is built), isFirstTime (a boolean flag 
*						that indicates whether this is the first time that the
*						Bayesian networks are created).
*       Output          none
*       Description		Creates the necessary nodes for each person in the 
*						pedigree 'pedPtr'. 
*       Comments		Used for autosomal analysis when haplotyping is performed.
*
******************************************************************************/
long createBayesNetNodesAutosomalHaplo(pedigree *pedPtr, boolean isFirstTime)
{
	person				*personPtr, *personPtrEnd;
	int					locusIndMap, locusIndDs, nodeWeight, *valArr, valArr1[1],
						interference, *mapToData;
	long				countVariables;
	boolean				isKnown;
	locusPerson			*currPersonLoc;
	node				**paSelecArr, **maSelecArr, **paLocArr, **maLocArr;

#if DISCONNECT_UNLINKED
	boolean				isUnlinked;
#endif

	countVariables = 0;
	valArr1[0] = 0;
	interference = recombData.interference;
	mapToData = loci.mapToData;

	/* 1.	Loop over all people in the pedigree. */
 	for (personPtr = pedPtr->people,  personPtrEnd = personPtr + pedPtr->numPeople;
			personPtr < personPtrEnd; personPtr++) {
		
		/* 1.1.	Update pointers to the arrays of pointers to the nodes of this person. */

		paSelecArr = personPtr->selectorVar[0];
		maSelecArr = personPtr->selectorVar[1];
		paLocArr = personPtr->locusVar[0];
		maLocArr = personPtr->locusVar[1];
    /*my	printf("\n");
	/*my	printf("%d\t", personPtr->pNum);*/
	
		/* 1.2.	For each person, loop over all marker (ALLELES type) loci. */
		for (locusIndMap = 0; locusIndMap < numMarkers; locusIndMap++) {

			locusIndDs = mapToData[locusIndMap];
			currPersonLoc = &(personPtr->loci[locusIndDs]);

	
			#if DISCONNECT_UNLINKED
				if ((locusIndMap > 0) && (recombData.maleThetas[locusIndMap - 1][0] == 0.5)) {
					isUnlinked = TRUE;
				}
				else {
					isUnlinked = FALSE;
				}
			#endif


			/* 1.2.1	Create a paternal selector node, if this person has parents. */

			if (personPtr->pa != NULL) {	

				/* 1.2.1.1.	Determine the number of possible values for this node. If it is
							one, then the value of this node is known. */

				valArr = currPersonLoc->locusData.allelesL.paSelectorVals;

				if (valArr[1] != -1) {
					nodeWeight = 2;
					isKnown = FALSE;
				}
				else {
					nodeWeight = 1;
					isKnown = TRUE;
				}

				/* 1.2.1.2.	Allocate memory and initialize the fields of the selector node. */

				if (locusIndMap == 0) {
					paSelecArr[locusIndDs] = 
							allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
												   SELECTOR, PATERNAL, isKnown, noUnite);
				}
			#if DISCONNECT_UNLINKED
				else if (isUnlinked) {
					paSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
											   SELECTOR, PATERNAL, isKnown, noUnite);
				}
			#endif
				else if ((!interference) || (interference && locusIndMap == 1)) {
					paSelecArr[locusIndDs] = 
							allocateInitializeNode(nodeWeight, 1, personPtr->pNum, locusIndMap, 
												   SELECTOR, PATERNAL, isKnown, noUnite);
				}
				else {
					paSelecArr[locusIndDs] = 
							allocateInitializeNode(nodeWeight, 2, personPtr->pNum, locusIndMap, 
												   SELECTOR, PATERNAL, isKnown, noUnite);
				}

				if (isKnown) {
					paSelecArr[locusIndDs]->currVal = 0;
				}

				/* 1.2.1.3.	Initialize the name of the node (for debugging). */

				#if DEBUG_STAT
					sprintf(paSelecArr[locusIndDs]->name, "ps_%d_%d", personPtr->pNum,
							locusIndMap);
				#endif

				/* 1.2.1.4.	Update the count of the number of nodes in the Bayesian network. */

				countVariables++;
			}
			else {
				paSelecArr[locusIndDs] = NULL;
			}

			/* 1.2.2	Create a paternal locus node. */

			/* 1.2.2.1.	Update the number of possible values for this node and whether
						it is known. */

			#if PERFORM_ALLELE_RECODING
				nodeWeight = createValArrayAR(currPersonLoc, TRUE, 
											  &isKnown);

				if (personPtr->pa == NULL) {
					updateNonTransAlleles(currPersonLoc, TRUE, locusIndDs);
				}
			#else
				nodeWeight = createValArray(currPersonLoc, TRUE, &isKnown);
			#endif

			currPersonLoc->locusData.allelesL.isKnownPa = isKnown;

			/* 1.2.2.2.	Allocate memory and initialize the fields of the node. */
		
			if (personPtr->pa == NULL)	{				/* this is a founder. */
				paLocArr[locusIndDs] = 
					allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
										   LOCUS, PATERNAL, isKnown, noUnite);
			}
			else {
				paLocArr[locusIndDs] = 
					allocateInitializeNode(nodeWeight, 3, personPtr->pNum, locusIndMap, 
										   LOCUS, PATERNAL, isKnown, noUnite);
			}

			if (isKnown) {
				#if PERFORM_ALLELE_RECODING
					paLocArr[locusIndDs]->currVal = 
						currPersonLoc->locusData.allelesL.paAllelsGroups[0].alleleArr[0];
				#else		
					paLocArr[locusIndDs]->currVal = 
						currPersonLoc->locusData.allelesL.paAlleles[0];
				#endif		
			}

			#if DEBUG_STAT
				sprintf(paLocArr[locusIndDs]->name, "pl_%d_%d", personPtr->pNum,
						locusIndMap);
			#endif

			countVariables++;

			/* 1.2.3	Create a maternal selector node, if this person has parents. */

			if (personPtr->ma != NULL) {	

				valArr = currPersonLoc->locusData.allelesL.maSelectorVals;

				if (valArr[1] != -1) {
					nodeWeight = 2;
					isKnown = FALSE;
				}
				else {
					nodeWeight = 1;
					isKnown = TRUE;
				}

				if (locusIndMap == 0) {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}
			#if DISCONNECT_UNLINKED
				else if (isUnlinked) {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}
			#endif
				else if ((!interference) || (interference && locusIndMap ==1)) {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 1, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}
				else {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 2, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}

				if (isKnown) {
					maSelecArr[locusIndDs]->currVal = 0;
				}

				#if DEBUG_STAT
					sprintf(maSelecArr[locusIndDs]->name, "ms_%d_%d", personPtr->pNum,
							locusIndMap);
				#endif	

				countVariables++;
			}
			else {
				maSelecArr[locusIndDs] = NULL;
			}

			/* 1.2.4.	Create a maternal locus node. */

			#if PERFORM_ALLELE_RECODING
				nodeWeight = createValArrayAR(currPersonLoc, FALSE, &isKnown);	

				if (personPtr->pa == NULL) {
					updateNonTransAlleles(currPersonLoc, FALSE, locusIndDs);
				}
			#else
				nodeWeight = createValArray(currPersonLoc, FALSE, &isKnown);	
			#endif

			currPersonLoc->locusData.allelesL.isKnownMa = isKnown;

		
			/* 1.2.5.	If the person is typed and heterozygote at this locus, create a phenotype node. */

			if ((currPersonLoc->unknown == FALSE) && (nodeWeight == 2)) {
					personPtr->phenotypeVar[locusIndDs] = 
						allocateInitializeNode(1, 2, personPtr->pNum, locusIndMap, 
											   PHENOTYPE, PATERNAL, TRUE, noUnite);
					countVariables++;

					#if DEBUG_STAT
						sprintf(personPtr->phenotypeVar[locusIndDs]->name, "ph_%d_%d", 
								personPtr->pNum, locusIndMap); 
					#endif

			}
			else
				personPtr->phenotypeVar[locusIndDs] = NULL;

			/* 1.2.6. Allocate memory and initialize the fields of the maternal locus node. */

			if (personPtr->ma == NULL) {	/* this is a founder. */
				maLocArr[locusIndDs] = 
					allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
										   LOCUS, MATERNAL, isKnown, noUnite);
			}
			else {
				maLocArr[locusIndDs] = 
					allocateInitializeNode(nodeWeight, 3, personPtr->pNum, locusIndMap, 
										   LOCUS, MATERNAL, isKnown, noUnite);
			}

			if (isKnown) {
				#if PERFORM_ALLELE_RECODING
					maLocArr[locusIndDs]->currVal = 
						currPersonLoc->locusData.allelesL.maAllelsGroups[0].alleleArr[0];
				#else		
					maLocArr[locusIndDs]->currVal = 
						currPersonLoc->locusData.allelesL.maAlleles[0];
				#endif		
			}

			#if DEBUG_STAT
				sprintf(maLocArr[locusIndDs]->name, "ml_%d_%d", personPtr->pNum,
						locusIndMap); 
			#endif

			countVariables++;
			/*my  printf("%d\t", (personPtr->locusVar[0])[locusIndDs]->currVal);
			/*my	printf("%d\t", (personPtr->locusVar[1])[locusIndDs]->currVal);*/
		}
	}

	/* 2.	Return the number of nodes in the Bayesian network created. */

	return countVariables;
}




/******************************************************************************
*
*       Function        createBayesNetNodesSexLinkedHaplo
*       Parameters		pedPtr (pointer to the pedigree struct for which the
*						Bayesian network is built), isFirstTime (a boolean flag 
*						that indicates whether this is the first time that the
*						Bayesian networks are created).
*       Output          none
*       Description		Creates the necessary nodes for each person in the 
*						pedigree 'pedPtr', in the case of a sex-linked analysis. 
*       Comments		Used for sex-linked analysis when haplotyping is performed.
*
******************************************************************************/
long createBayesNetNodesSexLinkedHaplo(pedigree *pedPtr, boolean isFirstTime)
{
	person				*personPtr, *personPtrEnd;
	int					locusIndMap, locusIndDs, nodeWeight, valArr1[1],
						*mapToData;
	long				countVariables;
	boolean				isKnown;
	locusPerson			*currPersonLoc;
	node				**maSelecArr, **paLocArr, **maLocArr;

#if DISCONNECT_UNLINKED
	boolean				isUnlinked;
#endif

	countVariables = 0;
	valArr1[0] = 0;
	mapToData = loci.mapToData;

	/* 1.	Loop over all people in the pedigree. */

	for (personPtr = pedPtr->people,  personPtrEnd = personPtr + pedPtr->numPeople;
			personPtr < personPtrEnd; personPtr++) {

		/* 1.1.	Update pointers to the arrays of pointers to the nodes of this person. */

		maSelecArr = personPtr->selectorVar[1];
		maLocArr = personPtr->locusVar[1];

		if (personPtr->personSex == Female) {
			paLocArr = personPtr->locusVar[0];
		}

		/* 1.2.	For each person, loop over all loci. */

		for (locusIndMap = 0; locusIndMap < numMarkers; locusIndMap++) {
			locusIndDs = mapToData[locusIndMap];
			currPersonLoc = &(personPtr->loci[locusIndDs]);

			#if DISCONNECT_UNLINKED
				if ((locusIndMap > 0) && (recombData.maleThetas[locusIndMap - 1][0] == 0.5)) {
					isUnlinked = TRUE;
				}
				else {
					isUnlinked = FALSE;
				}
			#endif

			/* 1.2.1.	If this is a female, create a paternal locus node. */

			if (personPtr->personSex == Female) {

				#if	PERFORM_ALLELE_RECODING
					nodeWeight = createValArrayAR(currPersonLoc, TRUE, &isKnown);

					if (personPtr->pa == NULL)	{ /* this is a founder. */
						updateNonTransAlleles(currPersonLoc, TRUE, locusIndDs);
					}
				#else
					nodeWeight = createValArray(currPersonLoc, TRUE, &isKnown);
				#endif
				
				currPersonLoc->locusData.allelesL.isKnownPa = isKnown;
			
				if (personPtr->pa == NULL)	{ /* this is a founder. */
					paLocArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
											   LOCUS, PATERNAL, isKnown, noUnite);
				}
				else {
					paLocArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 1, personPtr->pNum, locusIndMap, 
											   LOCUS, PATERNAL, isKnown, noUnite);
				}

				#if DEBUG_STAT
					sprintf(paLocArr[locusIndDs]->name, "pl_%d_%d", personPtr->pNum,
							locusIndMap); 
				#endif

				countVariables++;
			}

			/* 1.2.2.	Create a maternal selector node, if this person has parents. */

			if (personPtr->ma != NULL) {	

				if (currPersonLoc->locusData.allelesL.maSelectorVals[1] != -1) {
					nodeWeight = 2;
					isKnown = FALSE;
				}
				else {
					nodeWeight = 1;
					isKnown = TRUE;
				}

				if (locusIndMap == 0) {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}
			#if DISCONNECT_UNLINKED
				else if (isUnlinked) {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}
			#endif
				else if ((!recombData.interference) || (recombData.interference && locusIndMap == 1)) {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 1, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}
				else {
					maSelecArr[locusIndDs] = 
						allocateInitializeNode(nodeWeight, 2, personPtr->pNum, locusIndMap, 
											   SELECTOR, MATERNAL, isKnown, noUnite);
				}

				#if DEBUG_STAT
					sprintf(maSelecArr[locusIndDs]->name, "ms_%d_%d", personPtr->pNum,
							locusIndMap);  
				#endif

				countVariables++;
			}
			else {
				maSelecArr[locusIndDs] = NULL;
			}

			/* 1.2.3.	Create a maternal locus node. */

			#if	PERFORM_ALLELE_RECODING
				nodeWeight = createValArrayAR(currPersonLoc, FALSE, &isKnown);

				if (personPtr->pa == NULL)	{ /* this is a founder. */
					updateNonTransAlleles(currPersonLoc, FALSE, locusIndDs);
				}
			#else
				nodeWeight = createValArray(currPersonLoc, FALSE, &isKnown);
			#endif

			currPersonLoc->locusData.allelesL.isKnownMa = isKnown;
	        
			/* 1.2.3.1	If this person is typed and heterozygote at this locus, create a phenotype node. */
			
			if ((currPersonLoc->unknown == FALSE) && (nodeWeight == 2)) {
				personPtr->phenotypeVar[locusIndDs] = 
					allocateInitializeNode(1, 2, personPtr->pNum, locusIndMap, 
										   PHENOTYPE, PATERNAL, TRUE, noUnite);
				countVariables++;
				
				#if DEBUG_STAT
					sprintf(personPtr->phenotypeVar[locusIndDs]->name, "ph_%d_%d", 
							personPtr->pNum, locusIndMap); 
				#endif
			}
			else
				personPtr->phenotypeVar[locusIndDs] = NULL;

			if (personPtr->ma == NULL)	{  /* this is a founder. */
				maLocArr[locusIndDs] = 
					allocateInitializeNode(nodeWeight, 0, personPtr->pNum, locusIndMap, 
										   LOCUS, MATERNAL, isKnown, noUnite);
			}
			else {
				maLocArr[locusIndDs] = 
					allocateInitializeNode(nodeWeight, 3, personPtr->pNum, locusIndMap, 
										   LOCUS, MATERNAL, isKnown, noUnite);
			}

			#if DEBUG_STAT
				sprintf(maLocArr[locusIndDs]->name, "ml_%d_%d", personPtr->pNum,
						locusIndMap);
			#endif

			countVariables++;
		}
	}

	/* 2.	Return the number of nodes in the Bayesian network created. */

	return countVariables;
}


/******************************************************************************
*
*       Function        allocateInitializeNode
*       Parameters		nodeWeight (the number of possible values of the node),
*						numParents (the number of parents of the node in the 
*						Bayesian network), pNum (the index of the person this
*						node represents), lNum (the index of locus this node
*						represents), nodeType (the type of this node),
*						nodeSex (whether this is a paternal or maternal node),
*						isKnown (a flag that indicates whether this node 
*						represents a constant variable), uniteT (indicates the
*						unite type of this node - whether this node needs to
*						be treated with other nodes).
*       Output          none
*       Description		Allocates a 'node' structure, initializes its fields
*						and returns a pointer to it. 
*       Comments		none
*
******************************************************************************/
node* allocateInitializeNode(int nodeWeight, int numParents, int pNum, int lNum,
							 NodeType nodeType, NodeSex nodeSex, boolean isKnown,
							 uniteType uniteT)
{
	node	*Node;

	/* 1.	Allocate memory for the node structure. */

	myMalloc(Node, sizeof(node));
	
	/* 2.	Initialize the node's fields. */

	Node->numValues = nodeWeight;
	Node->numChildren = 0;
	Node->varInProd = NULL;
	Node->numParents = numParents;
	Node->parents = NULL;
	Node->elimInd = LONG_MAX;
	Node->pNum = pNum;
	Node->lNum = lNum;
	Node->nodeType = nodeType;
	Node->nodeSex = nodeSex;
	Node->isKnown = isKnown;
	Node->currVal = 0;

#if UNITE_VARS
	Node->uniteT = uniteT;
	Node->certainVal = NULL;
#endif
	
	/* 3.	If the node has parents, allocate memory for an array of pointers to
			the parents. */

	if (numParents > 0) {
		myMalloc(Node->parents, sizeof(node*) * numParents);
	}

	/* 4.	Initialize the array of nodes that this node needs to be eliminated
			jointly with to NULL. */

	#if UNITE_VARS
		Node->uniteVars = NULL;
		Node->numVarsUnite = 1;
	#endif

	/* 5.	Return a pointer to the node. */ 

	return Node;
}


/******************************************************************************
*
*       Function        createValArray
*       Parameters		locus (the locus for which the list is transformed into
*						an array), isPaternal (is TRUE if this is a paternal 
*						locus), isKnown (whether the genotype at this locus is 
*						known).
*       Output          none
*       Description		Transforms the list of possible ALLELES of some locus
*						into an array, and frees the list.
*       Comments		none
*
******************************************************************************/
int createValArray(locusPerson *locus, boolean isPaternal, boolean *isKnown)
{
	halfLocusPerson		*loc, *temp = NULL;
	int					numVals, ind1, ind2, *valArr;

	ind1 = 0;
	ind2 = 0;

	/* 1.	Initialize a pointer to the list of possible ALLELES of the person in
			this haplotype. */

	if (isPaternal == TRUE) {
		loc = ((locus->locusData).allelesL).paLocus;
	}
	else {
		loc = ((locus->locusData).allelesL).maLocus;
	}

	/* 2.	Calculate the number of possible values for the node, which is the number 
			of possible ALLELES. */

	numVals = calcNodeWeight(loc);

	/* 3.	Allocate the array and initialize a pointer to the first element 
			of the array. */

	if (isPaternal == TRUE) {
		myMalloc((locus->locusData).allelesL.paAlleles, sizeof(int) * numVals);
		valArr = (locus->locusData).allelesL.paAlleles;
		(locus->locusData).allelesL.numPaAlleleGroups = numVals;
	}
	else {
		myMalloc((locus->locusData).allelesL.maAlleles, sizeof(int) * numVals);
		valArr = (locus->locusData).allelesL.maAlleles;
		(locus->locusData).allelesL.numMaAlleleGroups = numVals;
	}

	/* 4.	Initialize the array with the possible values and free the list elements. */

	for (; loc != NULL; temp = loc, loc = loc->next, ind1++) {
		valArr[ind1] = loc->allele;
		myFree(temp);
	}
	myFree(temp);

	/* 5.   Determine whether the genotype is known (with/without phase). */

	*isKnown = ((numVals == 1) || (locus->unknown == FALSE));

	/* 6.	Return the number of possible values. */

	return numVals;
}


/******************************************************************************
*
*       Function        createValArrayAR
*       Parameters		locus (the locus for which the list is transformed into
*						an array), isPaternal (is TRUE if this is a paternal 
*						haplotype), isKnown (whether the genotype at this locus 
*						is known).
*       Output          none
*       Description		Transforms the list of possible ALLELES of some locus
*						into an array, and frees the list.
*       Comments		Used when allele-recoding is performed.
*
******************************************************************************/
#if PERFORM_ALLELE_RECODING
int createValArrayAR(locusPerson *locus, boolean isPaternal, boolean *isKnown)
{
	halfLocusPerson		*loc, *temp = NULL, *transList, *nonTransList;
	int					numSets, ind1, ind2, numNonTrans;
	alleleSet			*setArr;

	ind1 = 0;
	ind2 = 0;

	/* 1.	Calculate the number of possible values for the node. */

	/* 1.1. If there are non-transmitted ALLELES, initialize the number of
			possible values to 1.  Otherwise, initialize it to 0.
			Initialize a pointer to the list of transmitted ALLELES. */

	if (isPaternal == TRUE) {
		loc = ((locus->locusData).allelesL).paLocusTrans;

		if (((locus->locusData).allelesL).paLocus != NULL)
			numSets	= 1;
		else
			numSets	= 0;
	}
	else {
		loc = ((locus->locusData).allelesL).maLocusTrans;

		if (((locus->locusData).allelesL).maLocus != NULL)
			numSets	 = 1;
		else
			numSets	= 0;
	}

	/* 1.2.	Add the number of transmitted ALLELES to the number of possible values. */

	numSets += calcNodeWeight(loc);

	/* 2.	Allocate the array, initialize a pointer to the first element 
			of the list. Initialize pointers to the first element of the transmitted
			ALLELES list and to the first element of the non-transmitted ALLELES list. */

	if (isPaternal == TRUE) {
		myMalloc((locus->locusData).allelesL.paAllelsGroups, sizeof(alleleSet) * numSets);

		(locus->locusData).allelesL.numPaAlleleGroups = numSets;
		setArr = (locus->locusData).allelesL.paAllelsGroups;
		transList = (locus->locusData).allelesL.paLocusTrans;
		nonTransList = (locus->locusData).allelesL.paLocus;
	}
	else {
		myMalloc((locus->locusData).allelesL.maAllelsGroups, sizeof(alleleSet) * numSets);

		(locus->locusData).allelesL.numMaAlleleGroups = numSets;
		setArr = (locus->locusData).allelesL.maAllelsGroups;
		transList = (locus->locusData).allelesL.maLocusTrans;
		nonTransList = (locus->locusData).allelesL.maLocus;
	}

	/* 3.	If there are non-transmitted ALLELES, calculate the number of non-transmitted
			ALLELES and allocate memory for the non-transmitted allele list. */

	if (nonTransList != NULL) {
		numNonTrans = setArr[ind1].numAlleles = calcNodeWeight(nonTransList);

		myMalloc(setArr[ind1].alleleArr, sizeof(int) * numNonTrans);

		/* 3.1.	Add the set of non-transmitted ALLELES to the array and free the list. */

		for (temp = NULL; nonTransList != NULL; temp = nonTransList, 
				nonTransList = nonTransList->next, ind2++) {

			setArr[ind1].alleleArr[ind2] = nonTransList->allele;
			myFree(temp);
		}

		myFree(temp);
		ind1++;
	}
	else
		numNonTrans = 0;

	/* 4.	Initialize the array with all the possible transmitted ALLELES and free 
			the list elements. */

	for (temp = NULL; transList != NULL; temp = transList, transList = transList->next, ind1++) {
		setArr[ind1].numAlleles = 1;

		myMalloc(setArr[ind1].alleleArr, sizeof(int));

		setArr[ind1].alleleArr[0] = transList->allele;
		myFree(temp);
	}
	myFree(temp);

	/* 5.	Determine whether the genotype is known (with/without phase). */

	*isKnown = (((numSets == 1) && (numNonTrans == 0)) || (locus->unknown == FALSE));

	/* 6.	Return the number of possible values. */

	return numSets;
}
#endif


/******************************************************************************
*
*       Function        updateNonTransAlleles
*       Parameters		locus (the locus for which the set of non-transmitted 
*						ALLELES is updated), isPaternal (is TRUE if this is a 
*						paternal haplotype), locusInd (the index of the locus).
*       Output          none
*       Description		updates the set of non-transmitted ALLELES of this 
*						haplotype to hold only the allele with the highest gene
*						frequency (used for founders).
*       Comments		Used when allele-recoding is performed for haplotyping
*						analysis.
*
******************************************************************************/
#if PERFORM_ALLELE_RECODING
void updateNonTransAlleles(locusPerson *locus, boolean isPaternal, int locusInd)
{
	alleleSet		*alleleGroup;
	Probability		*geneFrequencies, maxGeneFreq;
	int				maxFreqAllele, *allPtr, *allPtrEnd;


	/* 1.	Update a pointer to the first allele set of this person. If there
			are non-transmitted ALLELES, they would be in this set. */

	if (isPaternal) {
		alleleGroup = &((locus->locusData).allelesL.paAllelsGroups[0]);
	}
	else {
		alleleGroup = &((locus->locusData).allelesL.maAllelsGroups[0]);
	}

	/* 2.	If there's more than one non-transmitted allele, find the one with
			the highest gene-frequency and keep it only. */

	if (alleleGroup->numAlleles > 1) {

		maxGeneFreq = 0.0;
		geneFrequencies = loci.locusArr[locusInd].geneFrequencies;

		for (allPtr = alleleGroup->alleleArr, 
				allPtrEnd = allPtr + alleleGroup->numAlleles; 
				allPtr < allPtrEnd; allPtr++) {		

			if (geneFrequencies[*allPtr-1] > maxGeneFreq) {
				maxGeneFreq = geneFrequencies[*allPtr-1];
				maxFreqAllele = *allPtr;
			}
		}
		alleleGroup->alleleArr[0] = maxFreqAllele;
		alleleGroup->numAlleles = 1;
	}
}
#endif

/******************************************************************************
*
*       Function        calcNodeWeight
*       Parameters		locus (the first element in a linked list that holds 
*						the possible ALLELES for one single-locus-haplotype  
*						of some person).
*       Output          none
*       Description		Counts the number of possible values (number of ALLELES)
*						of this single-locus-haplotype.
*       Comments		none
*
******************************************************************************/
int calcNodeWeight(halfLocusPerson *locus)
{
	int		numVals = 0;

	/* 1.	Count the number of possible values. */

	for (;locus != NULL; locus = locus->next)
		numVals++;

	/* 2.	Return the number of possible values (ALLELES). */

	return numVals;
}



/******************************************************************************
*
*       Function        calcNumValsGenotype
*       Parameters		personPtr (a pointer to the person for which the number
*						of possible genotypes in locus no. 'locusNum' in the
*						data-structures needs to be calculated).
*       Output          none
*       Description		Counts the number of possible values (number of different
*						genotypes) for person 'personPtr' in locus no. 'locusNum'.
*       Comments		none
*
******************************************************************************/
int calcNumValsGenotype(person *personPtr, int locusNum, boolean *isKnown)
{
	int		n;
	locus	*locusPtr;

	/* 1.	Update a pointer to the locus structure of this person at this locus. */

	locusPtr = &(loci.locusArr[locusNum]);

	/* 2.	If this is a locus of type affection or null, then there 
			are 3 possible genotypes. */

	if ((locusPtr->type == affection) || (locusPtr->type == null)) {
		*isKnown = FALSE;
		return 3;
	}

	/* 2.	If the genotype of the person is known at this locus, then there is 
			only 1 possible genotype. */

	if (personPtr->loci[locusNum].unknown == FALSE) {
		*isKnown = TRUE;
		return 1;
	}

	/* 3.	If this is an ALLELES type locus and the genotype isn't known, 
			calculate the number of different unordered genotypes possible. */

	n = (personPtr->locusVar[0][locusNum])->numValues;
	*isKnown = FALSE;

	return ((n * (n + 1)) / 2);
}



/******************************************************************************
*
*       Function        findCertainValInd
*       Parameters		valArr (an array of possible values), numVals (the number
*						of values in the array), val (the requested value).
*       Output          none
*       Description		Finds the index of the requested value in the value array. 
*       Comments		Used in the case when allele-recoding isn't performed.
*
******************************************************************************/
int findCertainValInd(int *valArr, int numVals, int val) 
{
	int		*valArrPtr, *valArrPtrEnd;

	/* 1.	Iterate over the values in the array. */

	for (valArrPtr = valArr, valArrPtrEnd = valArrPtr + numVals;
			valArrPtr < valArrPtrEnd; valArrPtr++) {

		/* 1.1.	If we reached the requested value, break. */

		if (*valArrPtr == val)
			break;
	}

	/* 2.	Return the index in the array of the requested value. */

	return (valArrPtr - valArr);
}


/******************************************************************************
*
*       Function        findCertainValIndAR
*       Parameters		valArr (an array of possible values), numVals (the number
*						of values in the array), val (the requested value).
*       Output          none
*       Description		finds the index of the requested value in the possible 
*						value array. 
*       Comments		Used in the case when allele-recoding is performed.
*
******************************************************************************/
#if PERFORM_ALLELE_RECODING
int findCertainValIndAR(alleleSet *setArr, int numSets, int val) 
{
	alleleSet	*setArrPtr, *setArrPtrEnd;

	/* 1.	Iterate over the possible sets of values in the array. */

	for (setArrPtr = setArr, setArrPtrEnd = setArrPtr + numSets;
			setArrPtr < setArrPtrEnd; setArrPtr++) {

		/* 1.1.	If the set contains more than one allele, than it holds the 
				non-transmitted ALLELES. Continue to the next set since the 
				requested value is transmitted. */

		if (setArrPtr->numAlleles > 1) 
			continue;

		/* 1.2.	Otherwise, if we reached the requested value, then break. */

		if (setArrPtr->alleleArr[0] == val)
			break;
	}

	/* 2.	Return the index in the array of the requested value. */

	return (setArrPtr - setArr);
}
#endif



/******************************************************************************
*
*       Function        zeroNodeFieldsPeds
*       Parameters		zeroNodeFields (pointer to the function which needs to 
*						be called in order to zero the fields of the nodes of
*						the pedigrees).
*       Output          none
*       Description		Zeroes the 'varInProd' and the 'numChildren' fields of 
*						the node of the Bayesian networks.
*       Comments		none
*
******************************************************************************/
void zeroNodeFieldsPeds(zeroNodeFieldsPed zeroNodeFields)
{
	pedigree	*currPed;


	/* 1.	Loop over all the pedigrees. */

	for (currPed = peds.head; currPed != NULL; currPed = currPed->nextPed) {

		/* 1.1	For each pedigree, call a function which zeroes the 
				'varInProd' and 'numChildren' fields of all the nodes of 
				this pedigree. */

		zeroNodeFields(currPed);
	}
}


/******************************************************************************
*
*       Function        zeroNodeFieldsPedAutosomal
*       Parameters		currPed (the data-structure that holds the information
*						on the input pedigree). 
*       Output          none
*       Description		Zeroes the 'varInProd' and the 'numChildren' fields of 
*						the node in the Bayesian network that represents this
*						pedigree.
*       Comments		Used in case of an autosomal analysis.
*
******************************************************************************/
void zeroNodeFieldsPedAutosomal(pedigree *currPed)
{
	person		*currPerson;
	int			personInd, locusInd, numLoci;

	numLoci = loci.numLoci;

	/* 1.	Iterate over the people in the pedigree. */

	for (personInd = 0; personInd < currPed->numPeople; personInd++) {
		currPerson = &(currPed->people[personInd]);

		/* 1.1.	For each person in the pedigree, loop over all loci. */

		for (locusInd = 0; locusInd < numLoci; locusInd++) {

			/* 1.1.1.	For each node, zero the 'varInProd' and the 'numChildren' fields. */

			zeroNode(currPerson->locusVar[0][locusInd]);
			zeroNode(currPerson->locusVar[1][locusInd]);
			zeroNode(currPerson->selectorVar[0][locusInd]);
			zeroNode(currPerson->selectorVar[1][locusInd]);
			zeroNode(currPerson->phenotypeVar[locusInd]);
		}

		zeroNode(currPerson->genotypeVar);
	}
}


/******************************************************************************
*
*       Function        zeroNodeFieldsPedSexLinked
*       Parameters		currPed (the data-structure that holds the information
*						on the input pedigree). 
*       Output          none
*       Description		Zeroes the 'varInProd' and the 'numChildren' fields of 
*						the node in the Bayesian network that represents this
*						pedigree.
*       Comments		Used in case of sex-linked analysis.
*
******************************************************************************/
void zeroNodeFieldsPedSexLinked(pedigree *currPed) 
{
	person		*currPerson;
	int			personInd, locusInd, numLoci;

	numLoci = loci.numLoci;

	/* 1	Loop over the people in the pedigree. */

	for (personInd = 0; personInd < currPed->numPeople; personInd++) {
		currPerson = &(currPed->people[personInd]);

		/* 1.1.	For each person in the pedigree, loop over the loci. */

		for (locusInd = 0; locusInd < numLoci; locusInd++) {

			/* 1.1.1.	For each node, zero the 'varInProd' and 'numChildren' fields. */

			if (currPerson->personSex == Female) {
				zeroNode(currPerson->locusVar[0][locusInd]);
			}

			zeroNode(currPerson->locusVar[1][locusInd]);
			zeroNode(currPerson->selectorVar[1][locusInd]);
			zeroNode(currPerson->phenotypeVar[locusInd]);
		}

		zeroNode(currPerson->genotypeVar);
	}
}


/******************************************************************************
*
*       Function        zeroNode
*       Parameters		Node (pointer to the node whose fields needs to be zeroed).
*       Output          none
*       Description		zeroes the fields of Node.
*       Comments		none
*
******************************************************************************/
void zeroNode(node *Node) 
{
	if (Node) { 
		Node->varInProd = NULL; 
		Node->numChildren = 0; 
	}
}



/******************************************************************************
*
*       Function        freeNodes
*       Parameters		none
*       Output          none
*       Description		Frees the memory allocated to the nodes of the Bayesian 
*						networks which represent the input pedigrees.
*       Comments		none
*
******************************************************************************/
void freeNodes()
{
	int			locusInd, numLoci; 
	pedigree	*currPed;
	person		*personPtr, *personPtrEnd;

	numLoci = loci.numLoci;

	/* 1.	Iterate over all pedigrees. */

	for (currPed = peds.head; currPed != NULL; currPed = currPed->nextPed) {

		/* 1.1.	For each pedigree, iterate over all people. */

		for (personPtr = currPed->people, personPtrEnd = personPtr + currPed->numPeople;
				personPtr < personPtrEnd; personPtr++) {

			/* 1.1.1.	For each person in each pedigree, iterate over all loci. */

			for (locusInd = 0; locusInd < numLoci; locusInd++) {

				/* 1.1.1.1.	Free the nodes of the Bayesian network allocated for this locus. */

				if (personPtr->locusVar[0]) {
					freeNode(&(personPtr->locusVar[0][locusInd]));
				}

				freeNode(&(personPtr->locusVar[1][locusInd]));

				if (personPtr->selectorVar[0]) {
					freeNode(&(personPtr->selectorVar[0][locusInd]));
				}

				freeNode(&(personPtr->selectorVar[1][locusInd]));
				freeNode(&(personPtr->phenotypeVar[locusInd]));
			}

			myFree(personPtr->genotypeVar);
		}
	}
}


/******************************************************************************
*
*       Function        freeNode
*       Parameters		Node (the node which needs to be freed).
*       Output          none
*       Description		Frees the memory allocated to 'Node'.
*       Comments		none
*
******************************************************************************/
void freeNode(node **Node)
{
	if (*Node) {
		myFree((*Node)->parents);

		#if UNITE_VARS
			if ((*Node)->uniteT) {
				if ((*Node)->isMyUniteVars) {	
					myFree((*Node)->uniteVars);
				}
				myFree((*Node)->certainVal);
			}
		#endif

		myFree((*Node));
	}
}
