/*****************************************************************************/
/*  BayesNet.h																 */
/*  Author: Ma'ayan Fishelson											     */
/*****************************************************************************/
/*																			 */
/*  Comment: This file includes functions used for creating the Bayesian     */
/*			 networks used for representing the input pedigrees.			 */
/*																			 */
/*****************************************************************************/

#ifndef _SL_BAYES_NET_H_
#define _SL_BAYES_NET_H_


#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "sl_Auxiliary.h"
#include "sl_PedData.h"


/*------------------------------*/
/*		DATA-STRUCTURES			*/
/*------------------------------*/

/* A type of a function that creates the Bayesian network's nodes for a single pedigree. */

typedef long (createBayesNetPed)(pedigree *pedPtr, boolean isFirstTime);

/* A type of a function that zeroes the flags of the Bayesian network's nodes for
   a single pedigree. */

typedef void (zeroNodeFieldsPed)(pedigree *pedPtr);


/*------------------------------*/
/*		GLOBAL VARIABLES		*/
/*------------------------------*/

extern pedData				peds;
extern lociData				loci;
extern int					numMarkers;
extern recombinationData	recombData;		

/*------------------------------*/
/*		FUNCTIONS				*/
/*------------------------------*/

/* Different functions used for creating the Bayesian networks for the autosomal and sexlinked 
   cases, for both likelihood computations and haplotyping. */

void createBayesNetNodesPeds(boolean isFirstTime, createBayesNetPed createBayesNet);
long createBayesNetNodesAutosomal(pedigree *pedPtr, boolean isFirstTime);
long createBayesNetNodesSexLinked(pedigree *pedPtr, boolean isFirstTime);
long createBayesNetNodesAutosomalHaplo(pedigree *pedPtr, boolean isFirstTime);
long createBayesNetNodesSexLinkedHaplo(pedigree *pedPtr, boolean isFirstTime);

int findCertainValInd(int *valArr, int numVals, int val) ;

node* allocateInitializeNode(int nodeWeight, int numParents, int pNum, int lNum,
							 NodeType nodeType, NodeSex nodeSex, boolean isKnown,
							 uniteType uniteT);
int calcNodeWeight(halfLocusPerson *locus);
int createValArray(locusPerson *locus, boolean isPaternal, boolean *isKnown);
int calcNumValsGenotype(person *personPtr, int locusNum, boolean *isKnown);

#if PERFORM_ALLELE_RECODING
	int createValArrayAR(locusPerson *locus, boolean isPaternal, boolean *isKnown);
	void updateNonTransAlleles(locusPerson *locus, boolean isPaternal, int locusInd);
	int findCertainValIndAR(alleleSet *setArr, int numSets, int val);
#endif

/* Different functions used for updating information regarding nodes that need 
   to be treated together when performing the computations. */

#if UNITE_VARS
	void updateUniteInfo(uniteType uniteT, node **paLocArr, node **maLocArr, 
						int locusIndDs, locusPerson *currPersonLoc);
	void updateUniteFourVars(pedigree *pedPtr);
	void updateUniteNodesVals(node *mainNode, node **uniteNodes, boolean isMyUniteNodes, 
							locusPerson *currLoc, NodeSex sex);
#endif

/* Updates information on the nodes with which 'currNode' should be united when performing
   the computations. */

#define updateUniteVarsInfo(currNode, uniteVarsArr, isMyUniteVarsArr, numVarsUniteArr) \
		currNode->uniteVars = uniteVarsArr; \
		currNode->isMyUniteVars = isMyUniteVarsArr; \
		currNode->numVarsUnite = numVarsUniteArr;

/* Different functions used for zeroeing the flags of the Bayesian network's nodes. */

void zeroNodeFieldsPeds(zeroNodeFieldsPed zeroNodeFields);
void zeroNodeFieldsPedAutosomal(pedigree *pedPtr);
void zeroNodeFieldsPedSexLinked(pedigree *pedPtr);
void zeroNode(node *Node);

/* Functions used for freeing the memory allocated to the nodes of the Bayesian networks. */

void freeNodes();
void freeNode(node **Node);


#endif		/*	_BAYES_NET_H_	*/

