/*********************************************************************************/
/*  EliminationOrderMF.c											     	     */
/*  Author: Ma'ayan Fishelson					                                 */
/*********************************************************************************/

#include "sl_EliminationOrderMF.h"


double n_w(struct graph *G, int v) {
	double cost = 1.0;
	int i;

	neighbor *e = G->vertexList[v].neighborList;

	for(i = 0; i < G->vertexList[v].numNeighbors; i++){
		cost *= (G->vertexList[e->neighborInd].varInProd)->numValues;
		e = e->next;
	}
	return cost;
}

/******************************************************************************
*
*       Function        determineEO_MF
*       Parameters		elimProd (the elimination-order-product), finalElimOrd
*						(the final elimination-order-array, when the function
*						is called, it holds the best sequence found up to now),
*						tempElimOrd (a temporary elimination-order-array), 
*						firstElimInd (the 1st index in the elimination-order-array 
*						which needs to be filled), varMappingSize (the size of the 
*						elimProd->oldToNewVarMapping), numLoci (the number of loci
*						in the analysis), tableSizes (an array which would store 
*						the size of the probability tables created in each step of
*						the elimination), numIter (the number of iterations that 
*						the Min-Fill algorithm needs to be executed), cFunc (the
*						cost function that is used to calculate the cost of each
*						vertex in order to determine which vertex to eliminate in
*						each step of the algorithm). 
*       Output          none
*       Description		Executes the Min-Fill heuristic algorithm 'numIter' times.
*						If a better elimination sequence is found, it is 
*						returned in 'finalElimOrd'.
*       Comments		none.
*
******************************************************************************/
boolean determineEO_MF(elimOrderProduct *elimProd, pedigree *currPed, 
					   elimOrderPed *finalElimOrd, elimOrderPed *tempElimOrd, 
					   int firstElimInd, long varMappingSize, int numLoci,
					   double *tableSizes, long *elimVarsSegArr, 
					   long *numVarsSegArr, int numIter, costFunc cFunc, 
					   boolean isWeighted)
{
	char				algType[4];
	struct graph		*G;
	int					iter, complexityCoefficient, algInd;
	elimOrderProduct	*tempElimProd;
	boolean				foundBetterOrd;


	foundBetterOrd = FALSE;

	if (numIter > 0) {

		if (isWeighted) {
			sprintf(algType, "WMF");
			algInd = WMF_IND;
		}
		else {
			sprintf(algType, "MF");
			algInd = MF_IND;
		}

		#if (!CALC_MEM_REQ) 

			/* 1.	Create a copy of the elimination-order-product. */

			tempElimProd = copyElimOrderProduct(elimProd, varMappingSize);

		#endif

		/* 2.	Perform 'numIter' iterations of the Min-Fill heuristic algorithm. */

		for (iter = 0; iter < numIter; iter++) {

			/* 2.1.	Create a copy of the elimination-order-product. */

			#if (CALC_MEM_REQ) 
				tempElimProd = copyElimOrderProduct(elimProd, varMappingSize);
			#endif

			/* 2.2.	Create the graph representation of the elimination-order-product. */

			G = createGraph(tempElimProd, TRUE, varMappingSize);

			/* 2.3.	Find an elimination sequence, according to the Min-Fill heuristic. */

			MinFillSearch(tempElimProd, tempElimOrd, currPed, G, numLoci, firstElimInd,
						  tableSizes, elimVarsSegArr, numVarsSegArr, cFunc);

			/* 2.4.	If the elimination sequence found is better than the one found 
					up to now: */

			if (tempElimOrd->sumTableSizes < finalElimOrd->sumTableSizes) {

				updateNewOrd(finalElimOrd, tempElimOrd, &foundBetterOrd, 
							 &complexityCoefficient, iter, algType, algInd);
			}

			/* 2.5.	Free the memory allocated for the graph. */

			destroyGraph(G, TRUE);

			/* 2.6.	Free the memory allocated for the graph. */

			#if	CALC_MEM_REQ
				freeElimOrderProduct(tempElimProd);
			#endif
		}

		#if	!CALC_MEM_REQ

			/* 3.	Free the memory allocated for the elimination order product. */

			freeElimOrderProduct(tempElimProd);
		#endif
	}

	return foundBetterOrd;
}


/******************************************************************************
*
*       Function        determineEO_MF_Auto
*       Parameters		elimProd (the elimination-order-product), finalElimOrd
*						(the final elimination-order-array, when the function
*						is called, it holds the best sequence found up to now),
*						tempElimOrd (a temporary elimination-order-array), 
*						firstElimInd (the 1st index in the elimination-order-array 
*						which needs to be filled), varMappingSize (the size of the 
*						elimProd->oldToNewVarMapping), numLoci (the number of loci
*						in the analysis), tableSizes (an array which would store 
*						the size of the probability tables created in each step of
*						the elimination), cFunc (the cost function that is used 
*						to calculate the cost of each vertex in order to determine
*						which vertex to eliminate in each step of the algorithm),
*						isWeighted (a flag which indicates whether the weighted 
*						version of the min-fill algorithm should be run, or the
*						unweighted version). 
*       Output          none
*       Description		Executes the Min-Fill heuristic algorithm a number of 
*						times according to the complexity of the problem.
*						If a better elimination sequence is found, it is 
*						returned in 'finalElimOrd'.
*       Comments		none.
*
******************************************************************************/
boolean determineEO_MF_Auto(elimOrderProduct *elimProd, pedigree *currPed, 
					        elimOrderPed *finalElimOrd, elimOrderPed *tempElimOrd, 
						    int firstElimInd, long varMappingSize, int numLoci,
						    double *tableSizes, long *elimVarsSegArr, 
						    long *numVarsSegArr, costFunc cFunc, boolean isWeighted)
{
	char				algType[4];
	struct graph		*G;
	int					iter, maxNumIter, complexityCoefficient, algInd; 
	elimOrderProduct	*tempElimProd;
	bestResults			bestResults;
	double				improvementRate;
	boolean				foundBetterOrd;

	
	foundBetterOrd = FALSE;

	if (isWeighted) {
		maxNumIter = (int)(heuristicEO.maxNumIter * PERCENT_WMF);
		sprintf(algType, "WMF");
		algInd = WMF_IND;
	}
	else {
		maxNumIter = (int)(heuristicEO.maxNumIter * PERCENT_MF);
		sprintf(algType, "MF");
		algInd = MF_IND;
	}

	if (maxNumIter == 0) 
		return FALSE;

	/* 1.	Initialize the different parameters. */

	improvementRate = 0;
	bestResults.sumTableImprovement = 0;
	bestResults.lastImprovementIter = 0;

	#if (!CALC_MEM_REQ) 

		/* 2.	Create a copy of the elimination-order-product. */

		tempElimProd = copyElimOrderProduct(elimProd, varMappingSize);
	#endif

	/* 3.	Perform the number of iterations required by the optimization policy. 
			For each one: */

	for (iter = 0; iter < maxNumIter; iter++) {

		#if (CALC_MEM_REQ) 

			/* 3.1.	Create a copy of the elimination-order-product. */

			tempElimProd = copyElimOrderProduct(elimProd, varMappingSize);
		#endif

		/* 3.1.	Create the graph representation of the elimination-order-product. */

		G = createGraph(tempElimProd, TRUE, varMappingSize);
	
		/* 3.2.	Find an elimination sequence, according to the Min Fill heuristic. */

		MinFillSearch(tempElimProd, tempElimOrd, currPed, G, numLoci, firstElimInd,
					  tableSizes, elimVarsSegArr, numVarsSegArr, cFunc);

		/* 3.3.	If the elimination sequence found is better than the one found 
				up to now: */

		if (tempElimOrd->sumTableSizes < finalElimOrd->sumTableSizes && tempElimOrd->maxMemReq < mem_avl) {

			updateNewOrd(finalElimOrd, tempElimOrd, &foundBetterOrd, 
						 &complexityCoefficient, iter, algType, algInd);

			/* 3.3.2.	Calculate the improvement rate in the sumTableSizes. */

/*			bestResults.lastImprovementIter = iter;
			bestResults.sumTableImprovement = 
				tempElimOrd->sumTableSizes/finalElimOrd->sumTableSizes;
*/
			if (complexityCoefficient < 7)
				break;
		}
/*		else {
			/* 3.3.1.	Check every 200 iterations that the improvement rate is 
						satisfying.  If it is too low,stop the seacrh. */
	
/*			if ((iter % 200 == 0) && (iter != 0)) {
				improvementRate = bestResults.sumTableImprovement/
									(iter - bestResults.lastImprovementIter);

				#if PRINT_DEBUG_OPTIMIZATION_ELIMINATION_ORDER
					printf("\n%s Checkpoint: iteration %d, improvementRate %lf.\n",
							algType, iter, improvementRate);
				#endif

				if (improvementRate < MIN_IMPROVEMENT_RATE) {
					break;
				}
			}
		}
*/
		/* 3.4.	Free the memory allocated for the graph. */

		destroyGraph(G, TRUE);

		#if	CALC_MEM_REQ

			/* 3.5.	Free the memory allocated for the graph. */

			freeElimOrderProduct(tempElimProd);

		#endif
	}

	#if	!CALC_MEM_REQ
		/* 4.	Free the memory allocated for the elimination order product. */

		freeElimOrderProduct(tempElimProd);
	#endif

	return foundBetterOrd;
}


/******************************************************************************
*
*       Function        MinFillSearch
*       Parameters		elimProd (the elimination-order-product), elimOrd 
*						(the elimination order array), G (the graph representation
*						of the elimination-order-product), numLoci (the number of
*						loci in the analysis), firstElimInd (the 1st index in the 
*						elimination order array which needs to be filled),
*						tableSizes (an array which would store the size of the 
*						probability tables created in each step of the elimination), 
*						cFunc (the cost function that is used to calculate the 
*						cost of each vertex in order to determine which vertex to 
*						eliminate in each step of the algorithm).
*       Output          none
*       Description		creates the elimination-order-array from the 
*						elimination-order-product.  The elimination order is 
*						determined according to the Min-Fill heuristic.
*       Comments		Determines an elimination sequence according to the 
*						Minimum-Fill Search heuristic.
*
******************************************************************************/
void MinFillSearch(elimOrderProduct *elimProd, elimOrderPed *elimOrd,
				   pedigree *currPed, struct graph *G, int numLoci, int firstElimInd,
				   double *tableSizes, long *elimVarsSegArr, long *numVarsSegArr,
				   costFunc cFunc)
{
	int				cntVertices;
	vertexInPool	*bestVertex, *worstVertex, *currVertex, *tmpVertex;


	#if (PRINT_VARIOUS_STATISTICS_LEVEL >  1) 
		elimOrd->treeWidth = -1; 
	#endif 


	/* 1.	Allocate memory for the linked list that holds the information 
			on the vertices in the vertex pool. */

	myMalloc(bestVertex, sizeof(vertexInPool));
	worstVertex = bestVertex;

	bestVertex->prev = NULL;
	
	for (cntVertices = 1; cntVertices < VAR_POOL_SIZE; cntVertices++) {
		myMalloc(worstVertex->next, sizeof(vertexInPool));
		(worstVertex->next)->prev = worstVertex;
		worstVertex = worstVertex->next;
	}

	worstVertex->next = NULL;

	#if CALC_MEM_REQ	
		elimOrd->maxMemReq = 0;
	#endif

	/* 3.	Determine an elimination order: either an elimination-order style 
			Lander-Green, where the variables are eliminated locus-locus (first
			the genotype and phenotype variables of the locus, and then the selector
			variables of the locus), or a greedy elimination-order. */

	if (elimOrd->isLGOrder) {
		MinFillSearchLG(elimProd, elimOrd, currPed, G, numLoci, 
						firstElimInd, &bestVertex, &worstVertex, tableSizes, 
						elimVarsSegArr, numVarsSegArr, cFunc);
	}
	else {
		MinFillSearchG(elimProd, elimOrd, G, firstElimInd, &bestVertex, 
					   &worstVertex, tableSizes, cFunc);
	}

	/* 5.	Update the sum of the table sizes. */

	updateSumTableSizes(tableSizes, elimOrd, firstElimInd);

	/* 6.	Free the memory allocated dynamically. */

	for (currVertex = bestVertex; currVertex != NULL;) {
		tmpVertex = currVertex;
		currVertex = currVertex->next;
		myFree(tmpVertex);
	}
}


/******************************************************************************
*
*       Function        MinFillSearchG
*       Parameters		elimProd (the elimination-order-product), elimOrd 
*						(the elimination order array), G (the graph representation
*						of the elimination-order-product), firstElimInd (the 1st
*						index in the elimination order array which needs to be
*						filled).
*       Output          none
*       Description		creates the elimination-order-array from the 
*						elimination-order-product.  The elimination order is 
*						determined according to the Min-Fill heuristic.
*       Comments		none.
*
******************************************************************************/
void MinFillSearchG(elimOrderProduct *elimProd, elimOrderPed *elimOrd, 
					struct graph *G, int firstElimInd, vertexInPool **bestVertex, 
					vertexInPool **worstVertex, double *tableSizes, costFunc cFunc)
{
	elimOrderNode	*elimNode;
	vertex			*nextVertex;
	boolean			isBreakVertex;
	long 			elimInd;
	int				numVertices;


	elimInd = firstElimInd;
	elimNode = &(elimOrd->elimOrderArr[elimInd]);

	/* 1.	While there are still vertices to eliminate from the graph. */

	while (G->currNumVertices) {	
		
		/* 1.1.	Choose a vertex to eliminate (or condition on) next, using a 
				probabilistic algorithm. */

		nextVertex = getNextVertex(G, &numVertices, &isBreakVertex, 
								   bestVertex, worstVertex, cFunc);

		if (nextVertex == NULL)
			break;				

		/* 1.2.	Update the elimination-order node and the graph, regarding this
				vertex, whether it is eliminated, or conditioned on. */

		if (isBreakVertex) {
			fixVertices(nextVertex, numVertices, &elimInd, &elimNode, G, elimProd);
		}
		else { 
			eliminateVertices(nextVertex, numVertices, tableSizes, &elimInd, 
							  &elimNode, G, elimProd);
		}

		#if CALC_MEM_REQ
			if (elimProd->currMemReq > elimOrd->maxMemReq) {
				elimOrd->maxMemReq = elimProd->currMemReq;
			}
		#endif
	}
}


/******************************************************************************
*
*       Function        getNextVertex
*       Parameters		G (the graph representation of the elimination-order 
*						product), 
*       Output          none
*       Description		returns a pointer to the vertex chosen according to a 
*						stochastic greedy heuristic which chooses a vertex 
*						according to the number of marked neighbors of
*						a vertex. The more marked neighbors of a vertex, the 
*						more likely that it would be chosen.
*       Comments		none.
*
******************************************************************************/
vertex* getNextVertex(struct graph *G, int *numVertices, boolean *isBreakVar, 
					  vertexInPool **bestVertexPtr, vertexInPool **worstVertexPtr, 
					  costFunc cFunc)
{
	vertex			*vertexPtr, *vertexPtrEnd, *chosenVertex = NULL;
	double			totalProb, coin, lowerBound;
	vertexInPool	*currVertex, *tmpVertex, *bestVertex, *worstVertex;
	node			*origNode;
		

	/* 1.	Initialize. */

	bestVertex = *bestVertexPtr;
	bestVertex->prev = NULL;
	worstVertex = *worstVertexPtr;
	worstVertex->next = NULL;

	for (currVertex = bestVertex; currVertex != NULL; currVertex = currVertex->next) {
		currVertex->cost = DBL_MAX;
		currVertex->v = NULL;
	}

	/* 2.	Iterate over the existing vertices looking for the best vertex 
			to mark. */

	for (vertexPtr = G->vertexList, vertexPtrEnd = vertexPtr + G->numVertices;
			vertexPtr < vertexPtrEnd; vertexPtr++) {

		/* 2.1.	If the vertex hasn't been marked already (it hasn't been chosen yet). */

		if (vertexPtr->isMarked == FALSE) {

			/* 2.1.1.	If the cost of this vertex is invalid, 
						we need to recalculate it. */

			if (vertexPtr->cost == INVALID) {
				cFunc(G, vertexPtr);
			}

			/* 2.1.2.	If the complexity is simple enough, the search will be stopped here. */

			if (vertexPtr->cost == 0.0) {
				chosenVertex = vertexPtr;
				goto useThisVertex;
			}

			/* 2.1.3.	Compare the cost of the current vertex to the costs of the 
						vertices in the pool.  If it is better than that of some of 
						the vertices in the pool and the pool is full, replace the 
						worst vertex in the pool with the current vertex. Iterate 
						over the vertices, starting from the worst one. */

			for (currVertex = worstVertex; currVertex != NULL;) {
				
				/* 2.1.3.1.	If the score of the current vertex is smaller or 
							equal to that of 'currVertex', */

				if (currVertex->cost <= vertexPtr->cost) {
					
					/* 2.1.3.1.1.	If 'currVertex' is the worst vertex in the pool, 
									then break (we don't want to insert it into the 
									vertex pool). */

					if (currVertex->next == NULL)
						break;

					/* 2.1.3.1.2.	Else, insert the current vertex right after 'currVertex'
									and remove the last vertex from the list. */

					tmpVertex = worstVertex;
					worstVertex = worstVertex->prev;
					(tmpVertex->prev)->next = NULL;
					tmpVertex->cost = vertexPtr->cost;
					tmpVertex->v = vertexPtr;
					tmpVertex->next = currVertex->next;
					tmpVertex->prev = currVertex;
					currVertex->next = tmpVertex;

					if (tmpVertex->next != NULL)
						(tmpVertex->next)->prev = tmpVertex;

					if (worstVertex->next != NULL)
						worstVertex = worstVertex->next;

					break;
				}

				/* 2.1.3.2.	If the cost of the current vertex is larger than that of 
							'currVertex' which is the bestVertex, then insert it before 
							the bestVertex and remove the last vertex from the list. */

				if ((currVertex->prev == NULL) && (currVertex->cost > vertexPtr->cost)) {
					tmpVertex = worstVertex;
					worstVertex = worstVertex->prev;
					(tmpVertex->prev)->next = NULL;
					tmpVertex->cost = vertexPtr->cost;
					tmpVertex->v = vertexPtr;
					tmpVertex->next = currVertex;
					tmpVertex->prev = NULL;
					currVertex->prev = tmpVertex;
					bestVertex = tmpVertex;
					break;
				}
				currVertex = currVertex->prev;
			}
		}
	}

	/* 3.	Iterate over the vertices in the pool and calculate their probability 
			to be chosen, according to the specified probability function. */

	totalProb = 0.0;

	for (currVertex = bestVertex; currVertex != NULL; currVertex = currVertex->next) {
		
		/* 3.1.	If the vertex pool isn't full, give the empty cells zero probability. 
				Otherwise, the probability is determined according to the appropriate 
				probability function. */

		if (currVertex->v == NULL) {
			currVertex->relProb = 0.0;
		}
		else {
			if (currVertex->cost == 0)
				currVertex->relProb = 1;
			else
				currVertex->relProb = sqrt(currVertex->cost);
			totalProb += currVertex->relProb;
		}
	}

	/* 4.	Iterate over the vertices in the pool and determine the relative probability
			of each one of them. */

	for (currVertex = bestVertex; currVertex != NULL; currVertex = currVertex->next) {
		
		if (currVertex->v != NULL) {
			currVertex->relProb  /= totalProb;
		}
	}

	/* 5.	Toss a coin to determine which vertex from the pool will be chosen. */

	coin = (double)(rand()) / RAND_MAX;

	/* 6.	Determine which vertex from the pool has been chosen. This is done by dividing the 
			[0,1] interval to subintervals according to the number of vertices in the pool.
			Each subinterval corresponds to one of the vertices and its size is equal to the
			relative probability of this vertex to be chosen. */

	if (coin == 1.0) {
		currVertex = worstVertex;

		while (currVertex->v == NULL)
			currVertex = currVertex->prev;

		chosenVertex = currVertex->v;
		goto useThisVertex;
	}
	
	for (lowerBound = 0, currVertex = bestVertex; currVertex != NULL; currVertex = currVertex->next) {
		
		if (coin <= lowerBound + currVertex->relProb) {
			chosenVertex = currVertex->v;
			break;
		}

		lowerBound += currVertex->relProb;
	}

useThisVertex: 

	(*bestVertexPtr) = bestVertex;
	(*worstVertexPtr) = worstVertex;

	/* 7.	If the elimination complexity of this vertex is too large, 
			find a vertex to perform conditioning on. */

	if ((chosenVertex->varInProd)->tableSize > breakProductBottomThreshold) {
		*isBreakVar = TRUE;
		return getBreakVertex(G, numVertices);
	}

	*isBreakVar = FALSE;

	origNode = (chosenVertex->varInProd)->originalNode;

	#if	UNITE_VARS
		if (origNode->uniteT) {
			*numVertices = origNode->numVarsUnite;
		}
		else {
	#endif
		*numVertices = 1;
	#if	UNITE_VARS
		}
	#endif

	return chosenVertex;
}


/******************************************************************************
*
*       Function        calcWMinFillCost
*       Parameters		G (the graph representation of the elimination-order
*						product, v (the index of the vertex whose cost needs
*						to be calculated).
*       Output          none
*       Description		Calculates the cost (according to the weighted-min-fill
*						heuristic) of the vertex pointed to by 'vertexPtr'.
*       Comments		none.
*
******************************************************************************/
double calcWMinFillCost(struct graph *G, vertex	*vertexPtr)
{
	double		cost, elimCost;
	vertex		*neighorPtr;
	neighbor	*neighbor1, *neighbor2;
	node		*origNode;

	origNode = (vertexPtr->varInProd)->originalNode;

	/* 1.	If the vertex needs to be eliminated with a few other vertices, 
			compute their joint complexity. */

	#if	UNITE_VARS
		if (origNode->uniteT) {
			return calcWMinFillCostJoint(G, vertexPtr->uniteVertices, origNode->numVarsUnite);
		}
	#endif

	vertexPtr->numNeighborsUnite = vertexPtr->numNeighbors;

	/* 2.	If the vertex has no neighbors (or only one neighbor), 
			then it's cost is zero. It should be a preferred vertex for elimination. */

	if (vertexPtr->numNeighbors == 0) {
		vertexPtr->cost = 0.0;
		(vertexPtr->varInProd)->tableSize = 0.0;
		return 0.0;
	}
	else if (vertexPtr->numNeighbors == 1) {
		vertexPtr->cost = 0.0;
		(vertexPtr->varInProd)->tableSize = 
			(G->vertexList[vertexPtr->neighborList->neighborInd]).varInProd->numValues;
		return 0.0;
	}

	/* 3.	Initialize the cost to zero. */
    	
	cost = 0;
	elimCost = 1.0;

	/* 4.	Update a pointer to the first neighbor. */

	neighbor1 = vertexPtr->neighborList;

	/* 5.	Iterate over the neighbors. */

	while (neighbor1 != NULL) {
		neighorPtr = &(G->vertexList[neighbor1->neighborInd]);
		elimCost *= (neighorPtr->varInProd)->numValues;
		neighbor2 = neighbor1->next;

		/* 5.1.	Iterate over all the neighbors of 'neighbor1' which appear after it
				in the neighbor list (this way we only check each pair of neighbors once). */

		while (neighbor2 != NULL) {

			/* 5.1.1.	For each 2 neighbors of 'vertexPtr' that aren't neighbors of
						each other, add the product of their weights (number of possible 
						values) to the cost of the vertex. */

			if (areNeighbors(G, neighbor1->neighborInd, neighbor2->neighborInd) == FALSE) {
				cost += neighorPtr->varInProd->numValues * 
						((G->vertexList[neighbor2->neighborInd]).varInProd)->numValues;
			}
			
			neighbor2 = neighbor2->next;
		}

		neighbor1 = neighbor1->next;
	}

	/* 6.	Update the cost of the vertex and return it. */

	vertexPtr->cost = cost;
	(vertexPtr->varInProd)->tableSize = elimCost;
	
	return cost;
}


/******************************************************************************
*
*       Function        calcWMinFillCostJoint
*       Parameters		G (the graph representation of the elimination-order
*						product, vertexArr (the array of vertices whose joint
*						cost needs to be calculated).
*       Output          none
*       Description		Calculates the joint cost (according to the 
*						weighted-min-fill heuristic) of of the vertices in the 
*						array  'vertexArr'.
*       Comments		none.
*
******************************************************************************/
double calcWMinFillCostJoint(struct graph *G, vertex **vertexArr, int numVertices)
{
	double		cost, tableSize;
	vertex		**vertexPtr, **vertexPtrEnd, **neighborArr, **neighbor1Ptr, 
				**neighbor2Ptr, **neighborPtrEnd;
	int			numNeighbors;

	/* 1.	Find all the neighbors of the vertices in 'vertexArr'. */

	neighborArr = getNeighbors(G, vertexArr, numVertices, &numNeighbors);

	/* 2.	If the vertex has no neighbors (or only one neighbor), 
			then it's cost is zero. It should be a preferred vertex for elimination. */

	if (numNeighbors == 0) {
		cost = 0;
		tableSize = 0;
		goto updateCost;
	}
	else if (numNeighbors == 1) {
		cost = 0;
		tableSize = (neighborArr[0]->varInProd)->numValues;
		goto updateCost;
	}

	/* 3.	Initialize the cost to zero, and the elimination cost to 1. */
    	
	cost = 0;
	tableSize = 1.0;

	/* 4.	Update a pointer to the first neighbor in the array, and update a pointer
			to the end of the neighbor array. */

	neighbor1Ptr = neighborArr;
	neighborPtrEnd = neighborArr + numNeighbors;

	/* 5.	Iterate over the neighbors. */

	while (neighbor1Ptr < neighborPtrEnd) {
		tableSize *= ((*neighbor1Ptr)->varInProd)->numValues;
		neighbor2Ptr = neighbor1Ptr + 1;

		/* 5.1.	Iterate over all the neighbors of 'neighbor1' which appear after it
				in the neighbor array (this way we only check each pair of neighbors once). */

		while (neighbor2Ptr < neighborPtrEnd) {

			/* 5.1.1.	For each 2 neighbors of the vertices in 'vertexArr' that aren't 
						neighbors of each other, add the product of their weights 
						(number of possible values) to the cost of the vertices in 'vertexArr'. */

			if (areNeighbors(G, (*neighbor1Ptr)->vNum, (*neighbor2Ptr)->vNum) == FALSE) {
				cost += (*neighbor1Ptr)->varInProd->numValues * 
						(*neighbor2Ptr)->varInProd->numValues;
			}
			
			neighbor2Ptr++;
		}

		neighbor1Ptr++;
	}

updateCost:

	/* 6.	Iterate over the vertices in the array, and update their cost. */

	for (vertexPtr = vertexArr, vertexPtrEnd = vertexArr + numVertices; 
			vertexPtr < vertexPtrEnd; vertexPtr++) {

		(*vertexPtr)->cost = cost;
		((*vertexPtr)->varInProd)->tableSize = tableSize;
		(*vertexPtr)->numNeighborsUnite = numNeighbors;
	}

	/* 7.	Free the memory allocated dynamically, and return the cost found. */

	myFree(neighborArr);
	
	return cost;
}


/******************************************************************************
*
*       Function        calcMinFillCost
*       Parameters		G (the graph representation of the elimination-order
*						product, v (the index of the vertex whose cost needs
*						to be calculated).
*       Output          none
*       Description		Calculates the cost (according to the min-fill
*						heuristic) of the vertex pointed to by 'vertexPtr'.
*       Comments		none.
*
******************************************************************************/
double calcMinFillCost(struct graph *G, vertex *vertexPtr)
{
	double		cost, tableSize;
	vertex		*neighorPtr;
	neighbor	*neighbor1, *neighbor2;
	node		*origNode;

	origNode = (vertexPtr->varInProd)->originalNode;

	/* 1.	If the vertex needs to be eliminated with a few other vertices, 
			compute their joint complexity. */

	#if	UNITE_VARS
		if (origNode->uniteT) {
			return calcMinFillCostJoint(G, vertexPtr->uniteVertices, 
										origNode->numVarsUnite);
		}
	#endif

	vertexPtr->numNeighborsUnite = vertexPtr->numNeighbors;

	tableSize = 1.0;

	/* 1.	If the vertex has no neighbors (or only one neighbor), 
			then it's cost is zero. It should be a preferred vertex for elimination. */

	if (vertexPtr->numNeighbors == 0) {
		vertexPtr->cost = 0.0;
		(vertexPtr->varInProd)->tableSize = 0.0;
		return 0.0;
	}
	else if (vertexPtr->numNeighbors == 1) {
		vertexPtr->cost = 0.0;
		(vertexPtr->varInProd)->tableSize = 
			(G->vertexList[vertexPtr->neighborList->neighborInd]).varInProd->numValues;
		return 0.0;
	}

	/* 2.	Initialize the cost to zero. */
    	
	cost = 0;

	/* 3.	Update a pointer to the first neighbor. */

	neighbor1 = vertexPtr->neighborList;

	/* 4.	Iterate over the neighbors. */

	while (neighbor1 != NULL) {
		neighorPtr = &(G->vertexList[neighbor1->neighborInd]);
		tableSize *= (neighorPtr->varInProd)->numValues;
		neighbor2 = neighbor1->next;

		/* 4.1.	Iterate over all the neighbors of 'neighbor1' which appear after it
				in the neighbor list (this way we only check each pair of neighbors once). */

		while (neighbor2 != NULL) {

			/* 4.1.1.	For each 2 neighbors of 'vertexPtr' that aren't neighbors of
						each other, add 1.0 to the cost of the vertex. */

			if (areNeighbors(G, neighbor1->neighborInd, neighbor2->neighborInd) == FALSE) {
				cost += 1.0;
			}
			
			neighbor2 = neighbor2->next;
		}

		neighbor1 = neighbor1->next;
	}

	/* 5.	Update the cost of the vertex and return it. */

	vertexPtr->cost = cost;
	(vertexPtr->varInProd)->tableSize = tableSize;

	return cost; 
}


/******************************************************************************
*
*       Function        calcMinFillCostJoint
*       Parameters		G (the graph representation of the elimination-order
*						product, v (the index of the vertex whose cost needs
*						to be calculated).
*       Output          none
*       Description		Calculates the joint cost (according to the min-fill
*						heuristic) of the vertices in the array  'vertexArr'.
*       Comments		none.
*
******************************************************************************/
double calcMinFillCostJoint(struct graph *G, vertex **vertexArr, int numVertices)
{
	double		cost, tableSize;
	vertex		**vertexPtr, **vertexPtrEnd, **neighborArr, **neighbor1Ptr, 
				**neighbor2Ptr, **neighborPtrEnd;
	int			numNeighbors;

	/* 1.	Find all the neighbors of the vertices in 'vertexArr'. */

	neighborArr = getNeighbors(G, vertexArr, numVertices, &numNeighbors);

	/* 2.	If the vertex has no neighbors (or only one neighbor), 
			then it's cost is zero. It should be a preferred vertex for elimination. */

	if (numNeighbors == 0) {
		cost = 0;
		tableSize = 0;
		goto updateCost;
	}
	else if (numNeighbors == 1) {
		cost = 0;
		tableSize = (neighborArr[0]->varInProd)->numValues;
		goto updateCost;
	}

	/* 3.	Initialize the cost to zero, and the elimination cost to 1. */
    	
	cost = 0;
	tableSize = 1.0;

	/* 4.	Update a pointer to the first neighbor in the array, and update a pointer
			to the end of the neighbor array. */

	neighbor1Ptr = neighborArr;
	neighborPtrEnd = neighborArr + numNeighbors;

	/* 5.	Iterate over the neighbors. */

	while (neighbor1Ptr < neighborPtrEnd) {
		tableSize *= ((*neighbor1Ptr)->varInProd)->numValues;
		neighbor2Ptr = neighbor1Ptr + 1;

		/* 5.1.	Iterate over all the neighbors of 'neighbor1' which appear after it
				in the neighbor array (this way we only check each pair of neighbors once). */

		while (neighbor2Ptr < neighborPtrEnd) {

			/* 5.1.1.	For each 2 neighbors of the vertices in 'vertexArr' that aren't 
						neighbors of each other, add the product of their weights 
						(number of possible values) to the cost of the vertices in 'vertexArr'. */

			if (areNeighbors(G, (*neighbor1Ptr)->vNum, (*neighbor2Ptr)->vNum) == FALSE) {
				cost += 1.0;
			}
			
			neighbor2Ptr++;
		}

		neighbor1Ptr++;
	}

updateCost:

	/* 6.	Iterate over the vertices in the array, and update their cost. */

	for (vertexPtr = vertexArr, vertexPtrEnd = vertexArr + numVertices; 
			vertexPtr < vertexPtrEnd; vertexPtr++) {

		(*vertexPtr)->cost = cost;
		((*vertexPtr)->varInProd)->tableSize = tableSize;
		(*vertexPtr)->numNeighborsUnite = numNeighbors;
	}

	/* 7.	Free the memory allocated dynamically, and return the cost found. */

	myFree(neighborArr);
	
	return cost;
}


/******************************************************************************
*
*       Function        getBreakVertex
*       Parameters		G (the graph representation of the elimination-order 
*						product). 
*       Output          none
*       Description		returns a pointer to the vertex chosen for conditioning. 
*       Comments		none.
*
******************************************************************************/
vertex* getBreakVertex(struct graph *G, int *numVertices)
{
	vertex			*vertexPtr, *vertexPtrEnd, *bestVertex;
	double			bestScore, currScore;
	node			*origNode;

	/* 1.	Initialize. */

	bestScore = MIN_DBL;
	bestVertex = NULL;

	/* 2.	Iterate over the existing vertices looking for the best vertex 
			to condition on. */

	for (vertexPtr = G->vertexList, vertexPtrEnd = vertexPtr + G->numVertices;
			vertexPtr < vertexPtrEnd; vertexPtr++) {

		/* 2.1.	If the vertex hasn't been chosen yet: */ 

		if (vertexPtr->isMarked == FALSE)	{

			/* 2.1.1.	Calculate its score. */

			currScore = (vertexPtr->varInProd)->tableSize * 
								sqrt(vertexPtr->numNeighborsUnite);
		
			/* 2.1.2.	Compare the vertex score to the best score found 
						up to now. */

			if (currScore > bestScore) {
				bestScore = currScore;
				bestVertex	= vertexPtr;
			}
		}
	}

	/* 3.	Update the number of vertices that need to be conditioned on jointly. */

	origNode = (bestVertex->varInProd)->originalNode;

	#if	UNITE_VARS
		if (origNode->uniteT) {
			*numVertices = origNode->numVarsUnite;
		}
		else {
	#endif
		*numVertices = 1;
	#if	UNITE_VARS
		}
	#endif

	/* 4.	Return a pointer to the vertex chosen to be conditioned on next. */

	return bestVertex;
}



/******************************************************************************
*
*       Function        MinFillSearchLG
*       Parameters		elimOrd (stores the elimination order), elimProd (the 
*						elimination-order-product), currPed (pointer to the
*						pedigree for which the elimination order is found), 
*						numLoci (the number of loci in the analysis),
*						maxTableSize (the maximum size of a probability table 
*						created during the elimination),
*						sumTableSizes (the sum of the sizes of all the probability
*						tables created during the elimination).
*       Output          none
*       Description		Determines an elimination-order style Lander-Green.  
*						That is, a peeling order which proceeds locus-locus.
*       Comments		none.
*
******************************************************************************/
void MinFillSearchLG(elimOrderProduct *elimProd, elimOrderPed *elimOrd,
					 pedigree *currPed, struct graph *G, int numLoci, 
					 int firstElimInd, vertexInPool **bestVertex,
					 vertexInPool **worstVertex, double *tableSizes, 
					 long *elimVarsSegArr, long *numVarsSegArr, costFunc cFunc)
{
	minFillSearchInSegs(elimProd, elimOrd, G, currPed->numPeople * 3, 
						numLoci * 2, firstElimInd, bestVertex, 
						worstVertex, tableSizes, elimVarsSegArr, numVarsSegArr,
						cFunc);
}


/******************************************************************************
*
*       Function        minFillSearchInSegs
*       Parameters		elimOrd (stores the elimination order), elimProd (the 
*						elimination-order-product), currPed (pointer to the
*						pedigree for which the elimination order is found), 
*						numLoci (the number of loci in the analysis),
*						maxTableSize (the maximum size of a probability table 
*						created during the elimination),
*						sumTableSizes (the sum of the sizes of all the probability
*						tables created during the elimination).
*       Output          none
*       Description		Determines an elimination-order style Lander-Green.  
*						That is, a peeling order which proceeds locus-locus.
*       Comments		none.
*
******************************************************************************/
void minFillSearchInSegs(elimOrderProduct *elimProd, elimOrderPed *elimOrd, 
						 struct graph *G, long maxSegSize, int numSegs, int firstElimInd, 
						 vertexInPool **bestVertex, vertexInPool **worstVertex,
						 double *tableSizes, long *elimVarsSegArr, 
						 long *numVarsSegArr, costFunc cFunc)
{
	int				seg;
	long			*elimVarsSegArrPtr, *numVarsSegArrPtr, elimOrdInd; 
	vertex			**elimVertexArr, **elimVertexArrPtr, **elimVertexArrPtrEnd;


	/* 1.	Allocate memory for an auxiliary array that will store pointers to
			all the variables of some locus. */

	myMalloc(elimVertexArr, sizeof(vertex*) * maxSegSize);

	elimVarsSegArrPtr = elimVarsSegArr;
	numVarsSegArrPtr = numVarsSegArr;

	elimOrdInd = firstElimInd;

	/* 1.	Iterate over the segments. */

	for (seg = 0; seg < numSegs; seg++) {

		/* 1.1.	For each segment, update the 'elimVarsArr' with pointers to all the
				variables of this segment according to their indices in the 'elimVarsSegArr'. */

		for (elimVertexArrPtr = elimVertexArr, 
				elimVertexArrPtrEnd = elimVertexArrPtr + (*numVarsSegArrPtr);
				elimVertexArrPtr < elimVertexArrPtrEnd; elimVertexArrPtr++, elimVarsSegArrPtr++) {
			
			(*elimVertexArrPtr) = &(G->vertexList[(*elimVarsSegArrPtr)]);
		}

		elimOrdInd = minFillLGPart(elimVertexArr, *numVarsSegArrPtr, elimOrdInd,
								   elimOrd, elimProd, bestVertex, worstVertex,
								   tableSizes, G, cFunc);

		numVarsSegArrPtr++;
	}

	myFree(elimVertexArr);
}


/******************************************************************************
*
*       Function        minFillLGPart
*       Parameters      elimVarsArr (an array of elimination-order-vars for 
*						which an elimination order needs to be found)
*						numNodesInArr (the number of nodes in elimVarsArr), 
*						elimOrdInd (the elimination index of the next elimination 
*						node), elimOrd (stores the elimination order), 
*						elimProd (the elimination-order-product).
*       Output          none
*       Description		finds an elimination order for the variables in 
*						'elimVarArr'.
*       Comments		updated 23.10.
*
******************************************************************************/
long minFillLGPart(vertex **elimVertexArr, long numVerticesInArr, 
				   long elimOrdInd, elimOrderPed *elimOrd, 
				   elimOrderProduct *elimProd, vertexInPool **bestVertex, 
				   vertexInPool **worstVertex, double *tableSizes, struct graph *G,
				   costFunc cFunc)
{
	long				elimInd, numNodesLeft;
	elimOrderNode		*elimNode;
	vertex				**nextElimVertex, *nextVertex;
	boolean				isBreakVertex;
	int					numVertices;

	/* 1.	Initialize the elimination order index. */

	elimInd = elimOrdInd;
	elimNode = &((elimOrd->elimOrderArr)[elimInd]);

	numNodesLeft = numVerticesInArr;
	nextElimVertex = &elimVertexArr[0];


	/* 2.	While there are still variables to eliminate. */

	while (numNodesLeft) {		

		/* 2.1.	Choose a vertex to eliminate (or condition on) next, using a 
				probabilistic algorithm. */

		nextVertex = getNextVertexPart(elimVertexArr, numVerticesInArr, 
									   &nextElimVertex, G, &numVertices, &isBreakVertex, 
									   bestVertex, worstVertex, cFunc);

		if (nextVertex == NULL)
			break;		
	
		/* 2.2.	Update the elimination-order node and the graph, regarding this
				vertex, whether it is eliminated, or conditioned on. */

		if (isBreakVertex) {
			fixVertices(nextVertex, numVertices, &elimInd, &elimNode, G, elimProd);
		}
		else {
			eliminateVertices(nextVertex, numVertices, tableSizes, &elimInd, &elimNode,
							  G, elimProd);
		}

		numNodesLeft -= numVertices;

		#if CALC_MEM_REQ
			if (elimProd->currMemReq > elimOrd->maxMemReq) {
				elimOrd->maxMemReq = elimProd->currMemReq;
			}
		#endif
	}
		
	return elimInd;
}


/******************************************************************************
*
*       Function        getNextVertexPart
*       Parameters		elimVarsArr (an array of elimination-order-vars for 
*						which an elimination order needs to be found)
*						numNodesInArr (the number of nodes in elimVarsArr), 
*					    nextElimVar (the next variable that is tested for 
*						elimination).
*						eliminateComplexity (the elimination complexity
*						of the variable chosen to be eliminated next).
*       Output          none
*       Description		returns a pointer to the variable choosen next to be 
*						eliminated (the variable that's chosen is least complex
*						according to the some greedy heuristic).						 
*       Comments		none
*
******************************************************************************/
vertex* getNextVertexPart(vertex **elimVertexArr, long numVerticesInArr, 
						  vertex ***nextElimVertex, struct graph *G, 
						  int *numVertices, boolean *isBreakVar, 
						  vertexInPool **bestVertexPtr, 
						  vertexInPool **worstVertexPtr, costFunc cFunc)
{
	vertex			**vertexPtr, **vertexPtrEnd, *v, *chosenVertex = NULL;
	double			totalProb, coin, lowerBound;
	vertexInPool	*currVertex, *tmpVertex, *bestVertex, *worstVertex;
	node			*origNode;

	
	/* 1.	Initialize. */

	bestVertex = *bestVertexPtr;
	bestVertex->prev = NULL;
	worstVertex = *worstVertexPtr;
	worstVertex->next = NULL;

	for (currVertex = bestVertex; currVertex != NULL; currVertex = currVertex->next) {
		currVertex->cost = DBL_MAX;
		currVertex->v = NULL;
	}


	/* 2.	Iterate over the existing vertices looking for the best vertex 
			to mark. */

	for (vertexPtr = elimVertexArr, vertexPtrEnd = vertexPtr + numVerticesInArr;
			vertexPtr < vertexPtrEnd; vertexPtr++) {

		v = (*vertexPtr);

		/* 2.1.	If the vertex hasn't been marked already (it hasn't been chosen yet). */

		if (v->isMarked == FALSE) {

			/* 2.1.1.	If the cost of this vertex is invalid, 
						we need to recalculate it. */

			if (v->cost == INVALID) {
				cFunc(G, v);
			}	

			/* 2.1.2.	Compare the cost of the current vertex to the costs of the 
						vertices in the pool.  If it is better than that of some of 
						the vertices in the pool and the pool is full, replace the 
						worst vertex in the pool with the current vertex. Iterate 
						over the vertices, starting from the worst one. */

			for (currVertex = worstVertex; currVertex != NULL;) {

				/* 2.3.1.	If the complexity of the current vertex is bigger or equal to 
							that of 'currVertex', */

				if (currVertex->cost <= v->cost) {

					/* 2.3.1.1.	If 'currVertex' is the worst vertex in the pool, then 
								break (we don't want to insert it into the vertex pool). */

					if (currVertex->next == NULL)
						break;

					/* 2.3.1.2.	Else, insert the current vertex right after 'currVertex' and
								remove the last vertex from the list. */

					tmpVertex = worstVertex;
					worstVertex = worstVertex->prev;
					(tmpVertex->prev)->next = NULL;
					tmpVertex->cost = v->cost;
					tmpVertex->v = v;
					tmpVertex->next = currVertex->next;
					tmpVertex->prev = currVertex;
					currVertex->next = tmpVertex;

					if (tmpVertex->next != NULL)
						(tmpVertex->next)->prev = tmpVertex;

					if (worstVertex->next != NULL)
						worstVertex = worstVertex->next;

					break;
				}

				/* 2.1.1.2.	If the cost of the current vertex is larger than that of 
							'currVertex' which is the bestVertex, then insert it before 
							the bestVertex and remove the last vertex from the list. */

				if ((currVertex->prev == NULL) && (currVertex->cost > v->cost)) {
					tmpVertex = worstVertex;
					worstVertex = worstVertex->prev;
					(tmpVertex->prev)->next = NULL;
					tmpVertex->cost = v->cost;
					tmpVertex->v = v;
					tmpVertex->next = currVertex;
					tmpVertex->prev = NULL;
					currVertex->prev = tmpVertex;
					bestVertex = tmpVertex;
					break;
				}
				currVertex = currVertex->prev;
			}
		}
	} 

	/* 3.	Iterate over the vertices in the pool and calculate their probability 
			to be chosen, according to the specified probability function. */

	totalProb = 0.0;

	for (currVertex = bestVertex; currVertex != NULL; currVertex = currVertex->next) {
		
		/* 3.1.	If the vertex pool isn't full, give the empty cells zero probability. 
				Otherwise, the probability is determined according to the appropriate 
				probability function. */

		if (currVertex->v == NULL) {
			currVertex->relProb = 0.0;
		}
		else {
			if (currVertex->cost == 0)
				currVertex->relProb = 1;
			else
				currVertex->relProb = sqrt(currVertex->cost);
			totalProb += currVertex->relProb;
		}
	}

	/* 4.	Iterate over the vertices in the pool and determine the relative probability
			of each one of them. */

	for (currVertex = bestVertex; currVertex != NULL; currVertex = currVertex->next) {
		
		if (currVertex->v != NULL) {
			currVertex->relProb  /= totalProb;
		}
	}

	/* 5.	Toss a coin to determine which vertex from the pool will be chosen. */

	coin = (double)(rand()) / RAND_MAX;

	/* 6.	Determine which vertex from the pool has been chosen. This is done by dividing the 
			[0,1] interval to subintervals according to the number of vertices in the pool.
			Each subinterval corresponds to one of the vertices and its size is equal to the
			relative probability of this vertex to be chosen. */

	if (coin == 1.0) {
		currVertex = worstVertex;

		while (currVertex->v == NULL)
			currVertex = currVertex->prev;

		chosenVertex = currVertex->v;
		goto useThisVertex;
	}
	
	for (lowerBound = 0, currVertex = bestVertex; currVertex != NULL; currVertex = currVertex->next) {
		
		if (coin <= lowerBound + currVertex->relProb) {
			chosenVertex = currVertex->v;
			break;
		}

		lowerBound += currVertex->relProb;
	}

useThisVertex: 

	(*bestVertexPtr) = bestVertex;
	(*worstVertexPtr) = worstVertex;

	if (chosenVertex->varInProd->tableSize > breakProductBottomThreshold) {
		*isBreakVar = TRUE;
		return getBreakVertexPart(elimVertexArr, numVerticesInArr, numVertices);
	}

	*isBreakVar = FALSE;

	origNode = (chosenVertex->varInProd)->originalNode;

	#if	UNITE_VARS
		if (origNode->uniteT) {
			*numVertices = origNode->numVarsUnite;
		}
		else {
	#endif
		*numVertices = 1;
	#if	UNITE_VARS
		}
	#endif

	return chosenVertex;
}



/******************************************************************************
*
*       Function        getBreakVertexPart
*       Parameters		elimVarsArr (an array of elimination-order-vars for 
*						which an elimination order needs to be found)
*						numNodesInArr (the number of nodes in elimVarsArr), 
*       Output          none
*       Description		finds & returns the "best" variable (among the variables
*						that haven't been eliminated yet) to break the product
*						upon, according to the chosen heuristic. 						 
*       Comments		none
*
******************************************************************************/
vertex* getBreakVertexPart(vertex **elimVertexArr, long numVerticesInArr,
						   int *numVertices)
{
	vertex			**vertexPtr, **vertexPtrEnd, *v, *bestVertex;
	double			bestScore, currScore;
	node			*origNode;

	/* 1.	Initialize. */

	bestScore = MIN_DBL;
	bestVertex = NULL;

	/* 2.	Iterate over the variables in the array. */

	for (vertexPtr = &elimVertexArr[0], vertexPtrEnd = vertexPtr + numVerticesInArr;
			vertexPtr < vertexPtrEnd; vertexPtr++) {

		v = *vertexPtr;
		
		/* 2.1.	If the vertex hasn't been chosen yet: */ 

		if (v->isMarked == FALSE)	{

			/* 2.1.1.	Calculate its score. */

			currScore = (v->varInProd)->tableSize * sqrt(v->numNeighborsUnite);

			/* 2.1.2.	Compare the vertex score to the best score found 
						up to now. */

			if (currScore > bestScore) {
				bestScore = currScore;
				bestVertex = v;
			}
		}
	}

	origNode = (bestVertex->varInProd)->originalNode;

	#if	UNITE_VARS
		if (origNode->uniteT) {
			*numVertices = origNode->numVarsUnite;
		}
		else {
	#endif
		*numVertices = 1;	
	#if	UNITE_VARS
		}
	#endif

	/* 3.	Return the vertex chosen for conditioning. */

	return bestVertex;
}



