/*****************************************************************************/
/*  EliminationOrderProduct.h												 */
/*  Author: Ma'ayan Fishelson												 */
/*****************************************************************************/

#ifndef _SL_ELIMINATION_ORDER_PRODUCT_H_
#define _SL_ELIMINATION_ORDER_PRODUCT_H_

#include <stdio.h>
#include <stdlib.h>
#include "sl_Auxiliary.h"
#include "sl_Node.h"
#include "sl_Product.h"

/*------------------------------*/
/*		DATA-STRUCTURES			*/
/*------------------------------*/

typedef struct elimOrderFunc elimOrderFunc;

/* elimOrderVar - Simulates a variable (node) of the Bayesian network, for the 
   elimination-order-finding process. */

typedef struct elimOrderVar {
	boolean				used;				/* Is this space being used. */
	int					functionSpaces;		/* Number of function slots available. */
	int					currNumFunctions;	/* Number of functions currently pointed to by the 
											   struct (= number of functions that the variable
										       appears in currently).*/
	elimOrderFunc		**funcs;			/* An array of pointers to the functions that depend 
											   on this variable. */
	node				*originalNode;		/* Pointer to the original node in the Bayes-net. */
	double				elimComplexity;		/* The elimination complexity of this variable (if 
										       the value isn't updated currently, has a negative 
										       value.) */
	double				tableSize;			/* The size of the probability table that would be 
										       created if this variable is eliminated. */
	int					numValues;			/* The number of possible values of this variable. */
	boolean				alreadySeen;		/* An auxiliary flag used by different functions. */
	long				indInRealProd;		/* The index of the variable in the real product 
										       which this variable simulates. */
	struct elimOrderVar	**uniteVars;		/* An array of pointers to the variables that need to 
											   be eliminated jointly with this variable. */
	int					numFuncsUnite;		/* If this variable needs to be treated jointly with a
											   few additional variables, then this field holds the
											   total number of functions they appear in. Otherwise,
											   it stores only the number of functions that this 
											   variable participates in. */
#if PERFORM_PREPROCESSING
	boolean		  isUpdatedASimp;
#endif

#if (PRINT_VARIOUS_STATISTICS_LEVEL > 1)
	int			  cliqueSize;			/* The size of the clique that would be created if
										   this variable is eliminated. */
#endif
} elimOrderVar;


/* elimOrderFunc - Simulates a function (probability table) of the Bayesisn network, 
   for the elimination-order-finding process. */

struct elimOrderFunc {
	boolean			used;				/* Is this space being used. */
	int				numVariables;		/* The number of variables of the function. */
	elimOrderVar	**variables;		/* An array of pointers to the variables of the 
										   function. */
	unsigned long	sizeOfTable;		/* The size of the probability table. */
	boolean			alreadySeen;		/* An auxiliary flag used by different functions. */
};


/* elimOrderProduct - Simulates the Bayesian network (the Product which represents
   the Bayesian network), for the elimination-order-finding process. */

typedef struct elimOrderProduct {
	long		  numVariables;			/* The number of variable slots available. */
	long		  currNumVars;			/* The number of variables currently in the product. */
	elimOrderVar  *variables;			/* An array of variables. */
	elimOrderVar  *nextVarPlace;		/* Pointer to the next empty slot in the variable array. */
	elimOrderVar  *nextElimVar;			/* Pointer to the next possible candidate for eliminations. */
	elimOrderVar  **oldToNewVarMapping;	/* A mapping that is used when the product needs to be copied.
										   This is a mapping from the variables in the original 
										   product to the variables in the duplicated product. */
	variable	  *firstVar;			/* A pointer to the first slot in the variable array of
										   the real product. */

	long		  numFunctions;			/* The number of function slots available. */
	long		  currNumFuncs;			/* The number of fucntions currently in the product. */
	elimOrderFunc *functions;			/* An array of functions. */
	elimOrderFunc *nextFuncPlace;		/* Pointer to the next empty slot in the function array. */

#if CALC_MEM_REQ
	double		  currMemReq;			/* The current memory requirements of all probability tables. */
	double		  currMemReqCurrProd;	/* The current memory requirements of the last break product. */	
	int			  condDepth;			/* The depth of the conditioning tree. */
#endif

} elimOrderProduct;	


programData			progData;	

/*------------------------------*/
/*		FUNCTIONS				*/
/*------------------------------*/

elimOrderProduct* allocateElimOrderProduct(long numVariables, long numFunctions,
										   long numVariablesMapping);
void freeElimOrderProduct(elimOrderProduct *elimProd);

elimOrderVar* insertVariableToElimOrderProduct(elimOrderProduct *elimProd, int numFunctions, 
											   node *originalNode, int numValues);
void removeVariableFromElimOrderProduct(elimOrderProduct *elimProd, 
										elimOrderVar *elimVar);
void addFunctionToElimOrderVariable(elimOrderVar *elimVar, elimOrderFunc *elimFunc);
void removeFunctionFromElimOrderVariable(elimOrderVar *elimVar, elimOrderFunc *elimFunc);
elimOrderFunc* insertFunctionToElimOrderProduct(elimOrderProduct *elimProd, 
												int numVariables, elimOrderVar **vars,
												unsigned long sizeOfTable);
void removeFunctionFromElimOrderProduct(elimOrderProduct *elimProd, 
										elimOrderFunc *elimFunc);
elimOrderProduct* createElimOrderProduct(product *prod);
elimOrderProduct* copyElimOrderProduct(elimOrderProduct *oldProd, 
									   long varMappingSize);
void zeroSeenFuncsVars(elimOrderFunc **funcArr, int numFuncs);

void getVarsInCombinedElimOrderFunc(elimOrderFunc **funcArr, int numFuncs,
									elimOrderVar **varArr, int *numVars,
									elimOrderVar *elimVar);
void getVarsInCombinedElimOrderFuncJoint(elimOrderVar **elimVarsArr, int numElimVars,
										 elimOrderFunc **funcArr, int *numFuncs,
										 elimOrderVar **varArr, int *numVars,
									 	 unsigned long *sizeOfTable);

void eliminateVarFromFunction(elimOrderProduct *elimProd, elimOrderFunc *elimFunc, 
							  elimOrderVar *brokenVar, int oldNumVals);
void buildElimOrderVariableFunctionTables(elimOrderProduct *elimProd);


double calcElimComplexity(elimOrderVar *var);
double calcElimComplexityJoint(elimOrderVar **varArr, int numVars);

void eliminateElimOrderVars(elimOrderVar *elimVar, int numElimVars, 
						    double *tableSizes, long *elimInd, 
						    elimOrderNode **elimNodePtr, elimOrderProduct *elimProd);
void eliminateElimOrderVarsPre(elimOrderVar *elimVar, int numElimVars, 
							   long *elimInd, elimOrderNode **elimNodePtr, 
							   elimOrderProduct *elimProd);
void eliminateElimOrderVar(elimOrderProduct *elimProd, elimOrderVar *elimVar);
unsigned long eliminateElimOrderVarsJoint(elimOrderProduct *elimProd, 
									      elimOrderVar **elimVars, int numElimVars);

void conditionOnElimOrderVar(elimOrderVar *condVar, int numVars, long *elimInd,
						     elimOrderNode **elimNodePtr, elimOrderProduct *elimProd);
void setBreakValElimOrder(elimOrderProduct *elimProd, elimOrderVar *brokenVar);


#endif		/*	_ELIMINATION_ORDER_PRODUCT_H_	*/
