/*********************************************************************************/
/*  EliminationOrderSG.c											     	     */
/*  Author: Ma'ayan Fishelson					                                 */
/*********************************************************************************/

#include "sl_EliminationOrderSG.h"


/******************************************************************************
*
*       Function        determineEO_MW
*       Parameters		elimProd (the elimination order product for which 
*						variables an elimination order needs to be determined), 
*						currPed (the data-structure which holds information on 
*						the input pedigree which 'elimProd' represents), 
*						finalElimOrd (will be updated with the final elimination
*						order), tempElimOrd (holds the temporary elimination 
*						order found in each iteration), firstElimInd (the index
*						of the first node which will be eliminated), 
*						varMappingSize (the number of spaces that need to be 
*						allocated for a variable mapping), numLoci (the number
*						of loci in the analysis), tableSizes (an array which is
*						used to store the sizes of the probability tables created
*						during the inference), elimVarsSegArr (if the elimination
*						order needs to be determined in segments, then this
*						array store the variables according to these segments),
*						numVarsSegArr (this array stores the number of variables
*						in each segment), numIter (the number of iterations that
*						need to be performed).
*       Output          none
*       Description		determines an eliminationOrder for the Bayes-net which
*						represents the pedigree 'currPed'.
*       Comments		This function is called when the number of iterations 
*						performed is determined by the used (and not automatically
*						by the program).
*
******************************************************************************/
boolean determineEO_MW(elimOrderProduct *elimProd, pedigree *currPed,
					   elimOrderPed *finalElimOrd, elimOrderPed *tempElimOrd, 
					   int firstElimInd, long varMappingSize, 
					   int numLoci, double *tableSizes, long *elimVarsSegArr, 
					   long *numVarsSegArr, int numIter)
{
	int						iterCnt, complexityCoefficient;
	boolean					foundBetterOrd;
	elimOrderProduct		*tempElimProd;


	foundBetterOrd = FALSE;

	/* 1.	Perform the number of iterations required by the optimization policy. 
			For each one: */

	for (iterCnt = 0; iterCnt < numIter; iterCnt++) {
				
		/* 1.1.	Create a copy of the elimination-order-product. */

		tempElimProd = copyElimOrderProduct(elimProd, varMappingSize);

		/* 1.2.	Find an elimination sequence. */

		MinWeightSearch(tempElimProd, tempElimOrd, currPed, numLoci, 
						firstElimInd, TRUE, tableSizes, elimVarsSegArr, 
						numVarsSegArr);

		/* 1.3.	Compare the result of the current iteration to the best result 
				found up to now in general. If there is an improvement, store  
				the new elimination sequence. */

		if (tempElimOrd->sumTableSizes < finalElimOrd->sumTableSizes) {

			updateNewOrd(finalElimOrd, tempElimOrd, &foundBetterOrd, 
						 &complexityCoefficient, iterCnt, "MW", MW_IND);
		}
	}

	return foundBetterOrd;
}


/******************************************************************************
*
*       Function        determineEO_MW_Auto
*       Parameters		elimProd (the elimination order product for which 
*						variables an elimination order needs to be determined), 
*						currPed (the data-structure which holds information on 
*						the input pedigree which 'elimProd' represents), 
*						finalElimOrd (will be updated with the final elimination
*						order), tempElimOrd (holds the temporary elimination 
*						order found in each iteration), firstElimInd (the index
*						of the first node which will be eliminated), 
*						varMappingSize (the number of spaces that need to be 
*						allocated for a variable mapping), numLoci (the number
*						of loci in the analysis), tableSizes (an array which is
*						used to store the sizes of the probability tables created
*						during the inference), elimVarsSegArr (if the elimination
*						order needs to be determined in segments, then this
*						array store the variables according to these segments),
*						numVarsSegArr (this array stores the number of variables
*						in each segment), numIter (the number of iterations that
*						need to be performed).
*       Output          none
*       Description		determines an eliminationOrder for the Bayes-net which
*						represents the pedigree 'currPed'.
*       Comments		This function is called when the number of iterations 
*						performed is determined automatically by the program.
*
******************************************************************************/
boolean determineEO_MW_Auto(elimOrderProduct *elimProd, pedigree *currPed,
	 					    elimOrderPed *finalElimOrd, elimOrderPed *tempElimOrd, 
						    int firstElimInd, long varMappingSize, 
						    int numLoci, double *tableSizes, long *elimVarsSegArr, 
						    long *numVarsSegArr)
{
	elimOrderProduct		*tempElimProd;
	int						maxNumIter, iterCnt, complexityCoefficient;
	boolean					foundBetterOrd;
/*	double					improvementRate;
	bestResults				bestRes;
*/

	foundBetterOrd = FALSE;

/*******************************************************************************

	// 1.	Used for checking if the improvement rate is satisfying:
			Initialize the different parameters. 

	improvementRate = 0;
	bestRes.sumTableImprovement = 0;
	bestRes.lastImprovementIter = 0;

/******************************************************************************/

	maxNumIter = (int)(heuristicEO.maxNumIter * PERCENT_MW);
			
	/* 2.	Perform the number of iterations required by the optimization policy. 
			For each one: */

	for (iterCnt = 0; iterCnt < maxNumIter; iterCnt++) {
			
		/* 2.1.	Create a copy of the elimination-order-product. */

		tempElimProd = copyElimOrderProduct(elimProd, varMappingSize);

		/* 2.2.	Find an elimination sequence. */

		MinWeightSearch(tempElimProd, tempElimOrd, currPed, numLoci, 
						firstElimInd, TRUE, tableSizes, elimVarsSegArr, numVarsSegArr);

		/* 2.3.	Compare the result of the current iteration to the best result found
				up to now in general. If there is an improvement, store the new 
				elimination sequence. */

		if ((tempElimOrd->sumTableSizes < finalElimOrd->sumTableSizes) &&
				(tempElimOrd->maxMemReq < mem_avl)) {

			updateNewOrd(finalElimOrd, tempElimOrd, &foundBetterOrd, 
						 &complexityCoefficient, iterCnt, "MW", MW_IND);

			if (complexityCoefficient < 7)
				break;

/*******************************************************************************

			// 2.3.1.	Calculate the improvement rate in the sumTableSizes. 

			bestResults.lastImprovementIter = iterCnt;
			bestResults.sumTableImprovement = 
				tempElimOrd->sumTableSizes/finalElimOrd->sumTableSizes;

			// 2.3.2.	Revise the optimization policy according to the 
			//			newly found complexity. 

			if (complexityCoefficient < MAX_ELIMINATION_COMPLEXITY) {
				if ((tmpMaxNumIter = optimizationPolicyArr[complexityCoefficient].maxNumIter * PERCENT_MW)
						< (maxNumIter - iterCnt)) {

					maxNumIter = iterCnt + tmpMaxNumIter;
					heuristicEO.maxNumIter = optimizationPolicyArr[complexityCoefficient].maxNumIter;

					#if PRINT_DEBUG_OPTIMIZATION_ELIMINATION_ORDER
						printf("\nMW Decided to perform only %d more iterations.\n", tmpMaxNumIter);
					#endif
				}
			}

		}
		else {
			// 2.4.1.	Check every 200 iterations that the improvement rate is 
			//			satisfying.  If it is too low,stop the search. 

			if ((iterCnt % 200 == 0) && (iterCnt != 0)) {
				improvementRate = bestResults.sumTableImprovement/
									(iterCnt - bestResults.lastImprovementIter);

				#if PRINT_DEBUG_OPTIMIZATION_ELIMINATION_ORDER
					printf("\nMW Checkpoint: iteration %d, improvementRate %lf.\n",
							iterCnt, improvementRate);
				#endif

				if (improvementRate < MIN_IMPROVEMENT_RATE) {
					break;
				}
			}
*******************************************************************************/
		}
	}

	return foundBetterOrd;
}



/******************************************************************************
*
*       Function        MinWeightSearch
*       Parameters		Product (the product which represents the Bayes-net for
*						which an elimination order needs to be found), 
*						elimOrd (will hold the elimination-order found for the 
*						pedigree pointed to by currPed).
*						currPed (the pedigree for which an elimination-order is 
*						being determined, numLoci (the number of loci in the 
*						analysis), elimInd (the first index in the elimination
*						order array which needs to be initialized).
*       Output          none
*       Description		determines an eliminationOrder for pedigree 'currPed'.
*       Comments		none.
*
******************************************************************************/
void MinWeightSearch(elimOrderProduct *elimProd, elimOrderPed *elimOrd, 
					 pedigree *currPed, int numLoci, int firstElimInd,
					 boolean isStochastic, double *tableSizes, long *elimVarsSegArr, 
					 long *numVarsSegArr)
{

	#if (PRINT_VARIOUS_STATISTICS_LEVEL >  1)
		elimOrd->treeWidth = -1;
	#endif

	#if CALC_MEM_REQ	
		elimOrd->maxMemReq = 0;
	#endif

	/* 1.	Determine an elimination order: either an elimination-order style 
			Lander-Green, where the variables are eliminated locus-locus (first
			the genotype and phenotype variables of the locus, and then the selector
			variables of the locus), or a greedy elimination-order. */

	if (elimOrd->isLGOrder) {
		if (isStochastic) {
			MinWeightSearchLG(elimOrd, elimProd, currPed, numLoci, firstElimInd,
							  tableSizes, elimVarsSegArr, numVarsSegArr,
							  updatePedElimOrderPart_Stoch);
		}
		else {
			MinWeightSearchLG(elimOrd, elimProd, currPed, numLoci, firstElimInd,
							  tableSizes, elimVarsSegArr, numVarsSegArr, 
						      updatePedElimOrderPart);
		}
	}
	else {
		if (isStochastic) {
			MinWeightSearchG_Stoch(elimProd, elimOrd, firstElimInd, tableSizes);
		}
		else {
			MinWeightSearchG(elimProd, elimOrd, firstElimInd, tableSizes);
		}
	}

	/* 2.	Update the sum of the table sizes. */

	updateSumTableSizes(tableSizes, elimOrd, firstElimInd);

	/* 3.	Free the memory allocated for the elimination-order-product. */

	freeElimOrderProduct(elimProd);
}


/******************************************************************************
*
*       Function        MinWeightSearchG
*       Parameters		elimProd (the elimination-order-product), elimOrd 
*						(the elimination order array), maxTableSize (the maximum
*						size of a probability table created during the elimination),
*						sumTableSizes (the sum of the sizes of all the probability
*						tables created during the elimination).
*       Output          none
*       Description		creates the elimination-order-array from the 
*						elimination-order-product.  The elimination order is 
*						determined according to a greedy heuristic.
*       Comments		In this function, stochastic-greedy methods are not 
*						applied for finding an elimination order.
*
******************************************************************************/
void MinWeightSearchG(elimOrderProduct *elimProd, elimOrderPed *elimOrd,
					  int firstElimInd, double *tableSizes)
{
	int				numVars;
	long			elimInd;
	elimOrderVar	*nextVar;
	elimOrderNode	*elimNode;
	boolean			isBreakVar;


	/* 1.	Initialize the elimination order index. */ 

	elimInd = firstElimInd;
	elimNode = &(elimOrd->elimOrderArr[elimInd]);

	/* 2.	While there are still variables to eliminate. */

	while (elimProd->currNumVars) {		

		/* 2.1.	Choose a variable to eliminate (or condition on) next. */

		nextVar = getNextVar(elimProd, &numVars, &isBreakVar);

		if (nextVar == NULL)
			break;				

		/* 2.2.	If the elimination complexity of this variable is too 
				large, perform product-breaking on some variable. */

		if (isBreakVar) {
			conditionOnElimOrderVar(nextVar, numVars, &elimInd, &elimNode, elimProd);
		}
		else {
			eliminateElimOrderVars(nextVar, numVars, tableSizes, &elimInd, 
								   &elimNode, elimProd);
		}


/*		#if (PRINT_VARIOUS_STATISTICS_LEVEL > 1)
			if (elimVar->cliqueSize > elimOrd->treeWidth)
				elimOrd->treeWidth = elimVar->cliqueSize;
		#endif
*/

		#if CALC_MEM_REQ
			if (elimProd->currMemReq > elimOrd->maxMemReq) {
				elimOrd->maxMemReq = elimProd->currMemReq;
			}
		#endif
	}
}


/******************************************************************************
*
*       Function        MinWeightSearchG_Stoch
*       Parameters		elimProd (the elimination-order-product), elimOrd 
*						(the elimination order array), maxTableSize (the maximum
*						size of a probability table created during the elimination),
*						sumTableSizes (the sum of the sizes of all the probability
*						tables created during the elimination).
*       Output          none
*       Description		creates the elimination-order-array from the 
*						elimination-order-product.  The elimination order is 
*						determined according to a greedy heuristic.
*       Comments		In this function, stochastic-greedy methods are  
*						applied for finding an elimination order.
*
******************************************************************************/
void MinWeightSearchG_Stoch(elimOrderProduct *elimProd, elimOrderPed *elimOrd,
							int firstElimInd, double *tableSizes)
{
	elimOrderVar	*nextVar;
	elimOrderNode	*elimNode;
	int				numVars, cntVars;
	long			elimInd;
	varInPool		*bestVar, *worstVar, *currVar, *tmpVar;
	boolean			isBreakVar;


	/* 1.	Initialize the elimination order index. */ 

	elimInd = firstElimInd;
	elimNode = &(elimOrd->elimOrderArr[elimInd]);

	/* 2.	Allocate memory for the linked list that holds the information on the
			variables in the variable pool. */

	myMalloc(bestVar, sizeof(varInPool));
	worstVar = bestVar;

	bestVar->prev = NULL;
	
	for (cntVars = 1; cntVars < VAR_POOL_SIZE; cntVars++) {
		myMalloc(worstVar->next, sizeof(varInPool));
		(worstVar->next)->prev = worstVar;
		worstVar = worstVar->next;
	}

	worstVar->next = NULL;

	/* 3.	While there are still variables to eliminate. */

	while (elimProd->currNumVars) {		

		/* 2.1.	Choose a variable to eliminate (or condition on) next,
				using a probabilistic algorithm. */

		nextVar = getNextVarProb(elimProd, &numVars, &isBreakVar, &bestVar, &worstVar);

		if (nextVar == NULL)
			break;				

		/* 2.2.	If the elimination complexity of this variable is too 
				large, perform product-breaking on some variable. */

		if (isBreakVar) {
			conditionOnElimOrderVar(nextVar, numVars, &elimInd, &elimNode, elimProd);
		}
		else {
			eliminateElimOrderVars(nextVar, numVars, tableSizes, &elimInd, 
								   &elimNode, elimProd);
		}

		/* 3.4.	Else, Update the array of table sizes and the maximal clique size. */

/*		#if (PRINT_VARIOUS_STATISTICS_LEVEL > 1)
			if (elimVar->cliqueSize > elimOrd->treeWidth)
				elimOrd->treeWidth = elimVar->cliqueSize;
		#endif
*/
		#if CALC_MEM_REQ
			if (elimProd->currMemReq > elimOrd->maxMemReq) {
				elimOrd->maxMemReq = elimProd->currMemReq;
			}
		#endif
	}

	/* 4.	Free the memory allocated for the linked list. */

	for (currVar = bestVar; currVar != NULL;) {
		tmpVar = currVar;
		currVar = currVar->next;
		myFree(tmpVar);
	}
}



/******************************************************************************
*
*       Function        getNextVar
*       Parameters		elimProd (the elimination-order-product), 
*						eliminateComplexity (the elimination complexity
*						of the variable chosen to be eliminated next).
*						tableSize (the size of the table created during the 
*						elimination of the variable.
*       Output          none
*       Description		returns a pointer to the variable choosen next to be 
*						eliminated (the variable that's chosen is least complex
*						according to the some greedy heuristic).						 
*       Comments		none
*
******************************************************************************/
elimOrderVar* getNextVar(elimOrderProduct *elimProd, int *numVars, 
						 boolean *isBreakVar)
{
	elimOrderVar	*var, *varEnd, *bestVar;
	double			bestScore;

	/* 1.	Initialize. */

	bestScore = DBL_MAX;
	bestVar = NULL;

	/* 2.	Iterate over the existing variables looking for the best variable 
			to eliminate, starting from elimProd->nextElimVar and wrapping around. */

	var = elimProd->nextElimVar;
	varEnd = elimProd->variables + elimProd->numVariables;

	do {
		if ((var->used) && (var->currNumFunctions > 0)) {
			
			/* 2.1.	If the complexity value of this variable is invalid, 
					we need to recalculate it. */

			if (var->elimComplexity == INVALID)
				calcElimComplexity(var);

			/* 2.2.	If the complexity is simple enough, the search will be 
					stopped here. */

			if (var->elimComplexity <= ELIMINATION_BOTTOM_THRESHOLD) {
				bestVar = var;
				goto useThisVar_G;
			}

			/* 2.3.	Else, compare the complexity against the best var up till now. */

			if (var->elimComplexity < bestScore) {
				bestScore = var->elimComplexity;
				bestVar = var;
			}
		}

		/* 2.4.	Continue to check the next variable. */

		var++;

		if (var == varEnd)
			var = elimProd->variables;

	} while (var != elimProd->nextElimVar);

	useThisVar_G:

	/* 3.	If a variable to be eliminated has been found, */

	if (bestVar) {

		/* 3.1.	If the elimination complexity of this variable is too large, 
				find a variable to perform product-breaking on. */

		if (bestVar->tableSize > breakProductBottomThreshold) {
			*isBreakVar = TRUE;
			return getBreakVar(elimProd, numVars);
				
		}

		*isBreakVar = FALSE;

		/* 3.2.	Update the 'nextElimVar' to point to the following variable in
				the product's variable array. */

		elimProd->nextElimVar = bestVar + 1;

		if (elimProd->nextElimVar >= varEnd)
			elimProd->nextElimVar = elimProd->variables;

		#if	UNITE_VARS
			if ((bestVar->originalNode)->uniteT) {
				*numVars = (bestVar->originalNode)->numVarsUnite;
			}
			else {
		#endif
			*numVars = 1;
		#if	UNITE_VARS
			}
		#endif
	}

	return bestVar;
}



/******************************************************************************
*
*       Function        getNextVarProb
*       Parameters		elimProd (the elimination-order-product), 
*						eliminateComplexity (the elimination complexity
*						of the variable chosen to be eliminated next).
*						tableSize (the size of the table created during the 
*						elimination of the variable.
*       Output          none
*       Description		returns a pointer to the variable choosen next to be 
*						eliminated (the variable that's chosen is least complex
*						according to the some greedy heuristic).						 
*       Comments		none
*
******************************************************************************/
elimOrderVar* getNextVarProb(elimOrderProduct *elimProd, int *numVars, 
							 boolean *isBreakVar, varInPool **bestVarPtr, 
							 varInPool **worstVarPtr)
{
	elimOrderVar	*var, *varEnd, *chosenVar = NULL;
	double			totalProb, coin, lowerBound;
	varInPool		*currVar, *tmpVar, *bestVar, *worstVar;

	
	/* 1.	Initialize. */

	bestVar = *bestVarPtr;
	bestVar->prev = NULL;
	worstVar = *worstVarPtr;
	worstVar->next = NULL;

	for (currVar = bestVar; currVar != NULL; currVar = currVar->next) {
		currVar->complexity = DBL_MAX;
		currVar->var = NULL;
	}

	/* 2.	Iterate over the existing variables looking for the best variable 
			to eliminate, starting from elimProd->nextElimVar and wrapping around. */

	var = elimProd->nextElimVar;
	varEnd = elimProd->variables + elimProd->numVariables;

	do {
		if ((var->used) && (var->currNumFunctions > 0)) {
			
			/* 2.1.	If the complexity value of this variable is invalid, 
					we need to recalculate it. */

			if (var->elimComplexity == INVALID)
				calcElimComplexity(var);

			/* 2.2.	If the complexity is simple enough, the search will be stopped here. */

			if (var->elimComplexity <= ELIMINATION_BOTTOM_THRESHOLD) {
				chosenVar = var;

				goto useThisVarProb_G;
			}

			/* 2.3.	Compare the elimination complexity of the current variable to 
					the elimination complexities of the variables in the pool.  If it
					is better than that of some of the variables in the pool and the
					pool is full, replace the worst variable in the pool with the current
					variable. Iterate over the variables, starting from the worst one. */

			for (currVar = worstVar; currVar != NULL;) {
				
				/* 2.3.1.	If the complexity of the current variable is bigger or 
							equal to that of 'currVar', */

				if (currVar->complexity <= var->elimComplexity) {
					
					/* 2.3.1.1.	If 'currVar' is the worst var in the pool, then break (we don't
								want to insert it into the variable pool). */

					if (currVar->next == NULL)
						break;

					/* 2.3.1.2.	Else, insert the current variable right after 'currVar'
								and remove the last variable from the list. */

					tmpVar = worstVar;
					worstVar = worstVar->prev;
					(tmpVar->prev)->next = NULL;
					tmpVar->complexity = var->elimComplexity;
					tmpVar->var = var;
					tmpVar->next = currVar->next;
					tmpVar->prev = currVar;
					currVar->next = tmpVar;

					if (tmpVar->next != NULL)
						(tmpVar->next)->prev = tmpVar;

					if (worstVar->next != NULL)
						worstVar = worstVar->next;

					break;
				}

				/* 2.3.2.	If the complexity of the current variable is smaller than that of 
							'currVar' which is the bestVar, */

				if ((currVar->prev == NULL) && (currVar->complexity > var->elimComplexity)) {
					tmpVar = worstVar;
					worstVar = worstVar->prev;
					(tmpVar->prev)->next = NULL;
					tmpVar->complexity = var->elimComplexity;
					tmpVar->var = var;
					tmpVar->next = currVar;
					tmpVar->prev = NULL;
					currVar->prev = tmpVar;
					bestVar = tmpVar;
					break;
				}

				currVar = currVar->prev;
			}
		}

		/* 2.4.	Continue to check the next variable. */

		var++;

		if (var == varEnd)
			var = elimProd->variables;

	} while (var != elimProd->nextElimVar);


	/* 3.	Iterate over the variables in the pool and calculate their probability 
			to be chosen, according to the specified probability function. */

	totalProb = 0.0;

	for (currVar = bestVar; currVar != NULL; currVar = currVar->next) {
		
		/* 3.1.	If the variable pool isn't full, give the empty cells zero probability. 
				Otherwise, the probability is determined according to the appropriate 
				probability function. */

		if (currVar->var == NULL) {
			currVar->relProb = 0.0;
		}
		else {
			currVar->relProb = 1.0 / sqrt(currVar->complexity);
			totalProb += currVar->relProb;
		}
	}

	/* 4.	Iterate over the variables in the pool and determine the relative probability
			of each one of them. */

	for (currVar = bestVar; currVar != NULL; currVar = currVar->next) {
		
		if (currVar->var != NULL) {
			currVar->relProb  /= totalProb;
		}
	}

	/* 5.	Toss a coin to determine which variable from the pool will be chosen. */

	coin = (double)(rand()) / RAND_MAX;

	/* 6.	Determine which variable from the pool has been chosen. This is done by dividing the 
			[0,1] interval to subintervals according to the number of variables in the pool.
			Each subinterval corresponds to one of the variables and its size is equal to the
			relative probability of this variable to be chosen. */

	if (coin == 1.0) {
		currVar = worstVar;

		while (currVar->var == NULL)
			currVar = currVar->prev;

		chosenVar = currVar->var;
		goto useThisVarProb_G;
	}
	
	for (lowerBound = 0, currVar = bestVar; currVar != NULL; currVar = currVar->next) {
		
		if (coin <= lowerBound + currVar->relProb) {
			chosenVar = currVar->var;
			break;
		}

		lowerBound += currVar->relProb;
	}

	useThisVarProb_G:

	(*bestVarPtr) = bestVar;
	(*worstVarPtr) = worstVar;

	if (chosenVar) {

		/* 3.1.	If the elimination complexity of this variable is too large, 
				find a variable to perform product-breaking on. */

		if (chosenVar->tableSize > breakProductBottomThreshold) {
			*isBreakVar = TRUE;
			return getBreakVar(elimProd, numVars);
				
		}

		*isBreakVar = FALSE;

		/* 3.2.	Update the 'nextElimVar' to point to the following variable in
				the product's variable array. */

		elimProd->nextElimVar = chosenVar + 1;

		if (elimProd->nextElimVar >= varEnd)
			elimProd->nextElimVar = elimProd->variables;

		#if	UNITE_VARS
			if ((chosenVar->originalNode)->uniteT) {
				*numVars = (chosenVar->originalNode)->numVarsUnite;
			}
			else {
		#endif
			*numVars = 1;
		#if	UNITE_VARS
			}
		#endif
	}

	return chosenVar;
}



/******************************************************************************
*
*       Function        updateUniteVars
*       Parameters		elimVar (the variable which needs to be eliminated jointly
*						with some other variables), tableSizes (the array of
*						sizes of probability tables created during the elimination),
*						elimInd (the current elimination index), elimNodePtr 
*						(pointer to the current elimination node), elimProd
*						(the elimination-order-product).
*       Output          none
*       Description		finds the variables that need to be eliminated jointly,
*						and updates the elimination-order-array and the 'tableSizes'
*						array in case the joint elimination creates a table 
*						which is smaller than the upper threshold. Also, 
*						eliminates these variables from the elimination-order
*						product.
*       Comments		returns TRUE if succeeded to eliminate the variables
*						jointly (none of the variables have been eliminated
*						previously), otherwise, returns FALSE.
*
******************************************************************************/
#if	UNITE_VARS
boolean updateUniteVars(elimOrderVar *elimVar, double *tableSizes, long *elimInd, 
						elimOrderNode **elimNodePtr, elimOrderProduct *elimProd)
{
	int				numVarsUnite;
	long			currElimInd;
	elimOrderVar	**uniteVarsPtr, **uniteVarsPtrEnd;
	elimOrderNode   *elimNode;

	/* 1.	Initialization. */

	numVarsUnite = (elimVar->originalNode)->numVarsUnite;
	uniteVarsPtr = elimVar->uniteVars;
	uniteVarsPtrEnd = uniteVarsPtr + numVarsUnite;
	currElimInd = *elimInd;

	/* 2.	Update the elimination order array with all the nodes that need to be 
			eliminated jointly. */

	elimNode = *elimNodePtr;

	for (; uniteVarsPtr < uniteVarsPtrEnd; uniteVarsPtr++, elimNode++, currElimInd++) {
		
		if (((*uniteVarsPtr) == NULL) || ((*uniteVarsPtr)->used == FALSE)) {
			return FALSE;
		}

		updateElimNode(elimNode, FALSE, ((*uniteVarsPtr)->originalNode));
		tableSizes[currElimInd] = 0;
	}

	/* 3.	Eliminate the variables from the elimination-order-product and update
			the complexity of the elimination in the 'tableSizes' array. */

	if (calcElimComplexityJoint(elimVar->uniteVars, numVarsUnite) < breakProductBottomThreshold) {
		tableSizes[*elimInd] = eliminateElimOrderVarsJoint(elimProd, elimVar->uniteVars, 
														   numVarsUnite);
		*elimInd = currElimInd;
		*elimNodePtr = elimNode;
	}
	else {
		return FALSE;		
	}

	return TRUE;
}
#endif

/******************************************************************************
*
*       Function        getBreakVar
*       Parameters		elimProd (the elimination-order-product)
*       Output          none
*       Description		finds & returns the "best" variable (among the variables
*						that haven't been eliminated yet) to break the product
*						upon, according to the chosen heuristic. 						 
*       Comments		none
*
******************************************************************************/
elimOrderVar* getBreakVar(elimOrderProduct *elimProd, int *numVars)
{
	elimOrderVar	*var, *varEnd,*bestVar;
	double			bestScore, currScore;

	/* 1.	Initialize. */

	bestScore = MIN_DBL;
	bestVar = NULL;

	/* 2.	Iterate over the product's variables. */

	for (var = elimProd->variables, varEnd = var + elimProd->numVariables;
			var < varEnd; var++) {
		
		/* 2.1.	If the variable hasn't been eliminated yet: */ 

		if (var->used)	{

			/* 2.1.1.	Calculate its score. */

			currScore = var->elimComplexity * sqrt(var->numFuncsUnite);
		
			/* 2.1.2.	Compare the variable's score to the best score found 
						up to now. */

			if (currScore > bestScore) {
				bestScore = currScore;
				bestVar	= var;
			}
		}
	}

	#if	UNITE_VARS
		if ((bestVar->originalNode)->uniteT) {
			*numVars = (bestVar->originalNode)->numVarsUnite;
		}
		else {
	#endif
		*numVars = 1;
	#if	UNITE_VARS
		}
	#endif

	return bestVar;
}


/******************************************************************************
*
*       Function        MinWeightSearchLG
*       Parameters		elimOrd (stores the elimination order), elimProd (the 
*						elimination-order-product), currPed (pointer to the
*						pedigree for which the elimination order is found), 
*						numLoci (the number of loci in the analysis),
*						maxTableSize (the maximum size of a probability table 
*						created during the elimination),
*						sumTableSizes (the sum of the sizes of all the probability
*						tables created during the elimination).
*       Output          none
*       Description		Determines an elimination-order style Lander-Green.  
*						That is, a peeling order which proceeds locus-locus.
*       Comments		none.
*
******************************************************************************/
void MinWeightSearchLG(elimOrderPed *elimOrd, elimOrderProduct *elimProd,
								 pedigree *currPed, int numLoci, 
								 int firstElimInd, double *tableSizes, 
								 long *elimVarsSegArr, long *numVarsSegArr,
								 updatePartElimOrderFunc updatePartElimOrderPtr)
{
	minWeightSearchInSegs(elimOrd, elimProd, currPed->numPeople * 3, numLoci * 2,
						  firstElimInd, tableSizes, elimVarsSegArr, numVarsSegArr,
						  updatePartElimOrderPtr);
}



/******************************************************************************
*
*       Function        minWeightSearchInSegs
*       Parameters		elimOrd (stores the elimination order), elimProd (the 
*						elimination-order-product), currPed (pointer to the
*						pedigree for which the elimination order is found), 
*						.
*       Output          none
*       Description		Determines an elimination-order style Lander-Green.  
*						That is, a peeling order which proceeds locus-locus.
*       Comments		none.
*
******************************************************************************/
void minWeightSearchInSegs(elimOrderPed *elimOrd, elimOrderProduct *elimProd,
						   long maxSegSize, int numSegs, int firstElimInd, 
						   double *tableSizes, long *elimVarsSegArr, 
						   long *numVarsSegArr,
						   updatePartElimOrderFunc updatePartElimOrderPtr)
{
	int				seg;
	elimOrderVar	**elimVarsArr, **elimVarsArrPtr, **elimVarsArrPtrEnd;
	long			*elimVarsSegArrPtr, *numVarsSegArrPtr, elimOrdInd; 


	myMalloc(elimVarsArr, sizeof(elimOrderVar*) * maxSegSize);

	elimVarsSegArrPtr = elimVarsSegArr;
	numVarsSegArrPtr = numVarsSegArr;

	elimOrdInd = firstElimInd;

	/* 1.	Iterate over the segments. */

	for (seg = 0; seg < numSegs; seg++) {

		/* 1.1.	For each segment, update the 'elimVarsArr' with pointers to all the
				variables of this segment according to their indices in the 'elimVarsSegArr'. */

		for (elimVarsArrPtr = elimVarsArr, 
				elimVarsArrPtrEnd = elimVarsArrPtr + (*numVarsSegArrPtr);
				elimVarsArrPtr < elimVarsArrPtrEnd; elimVarsArrPtr++, elimVarsSegArrPtr++) {
			
			(*elimVarsArrPtr) = &elimProd->variables[(*elimVarsSegArrPtr)];
		}

		elimOrdInd = updatePartElimOrderPtr(elimVarsArr, *numVarsSegArrPtr, elimOrdInd,
											elimOrd, elimProd, tableSizes);

		numVarsSegArrPtr++;
	}

	myFree(elimVarsArr);
}




/******************************************************************************
*
*       Function        updatePedElimOrderPart
*       Parameters      elimVarsArr (an array of elimination-order-vars for 
*						which an elimination order needs to be found)
*						numNodesInArr (the number of nodes in elimVarsArr), 
*						elimOrdInd (the elimination index of the next elimination 
*						node), elimOrd (stores the elimination order), 
*						elimProd (the elimination-order-product).
*       Output          none
*       Description		finds an elimination order for the variables in 
*						'elimVarArr'.
*       Comments		Used for the case where stochastic-greedy methods are 
*						not to be applied for finding an elimination order.
*
******************************************************************************/
long updatePedElimOrderPart(elimOrderVar **elimVarsArr, long numNodesInArr, 
							long elimOrdInd, elimOrderPed *elimOrd, 
							elimOrderProduct *elimProd, double *tableSizes)
{
	elimOrderVar		*elimVar, **nextElimVar;
	long				elimInd, numNodesLeft;
	elimOrderNode		*elimNode;
	int					numVars;
	boolean				isBreakVar;

	/* 1.	Initialize the elimination order index. */

	elimInd = elimOrdInd;
	elimNode = &((elimOrd->elimOrderArr)[elimInd]);

	numNodesLeft = numNodesInArr;
	nextElimVar = &elimVarsArr[0];

	/* 2.	While there are still variables to eliminate. */

	while (numNodesLeft) {		

		/* 2.1.	Choose a variable to eliminate (or condition on) next. */

		elimVar = getNextVarLGPart(elimVarsArr, numNodesInArr, &nextElimVar,
								   &numVars, &isBreakVar);

		/* 2.2.	Either perform variable elimination or conditioning, according to the
				type of variable. */

		if (isBreakVar) {
			conditionOnElimOrderVar(elimVar, numVars, &elimInd, &elimNode, elimProd);
		}
		else {
			eliminateElimOrderVars(elimVar, numVars, tableSizes, &elimInd, &elimNode, elimProd);
		}


/*		#if (PRINT_VARIOUS_STATISTICS_LEVEL > 1)
			if (elimVar->cliqueSize > elimOrd->treeWidth)
				elimOrd->treeWidth = elimVar->cliqueSize;
		#endif
*/
		numNodesLeft -= numVars;

		#if CALC_MEM_REQ
			if (elimProd->currMemReq > elimOrd->maxMemReq) {
				elimOrd->maxMemReq = elimProd->currMemReq;
			}
		#endif
	}

	return elimInd;
}

/******************************************************************************
*
*       Function        updatePedElimOrderPart_Stoch
*       Parameters      elimVarsArr (an array of elimination-order-vars for 
*						which an elimination order needs to be found)
*						numNodesInArr (the number of nodes in elimVarsArr), 
*						elimOrdInd (the elimination index of the next elimination 
*						node), elimOrd (stores the elimination order), 
*						elimProd (the elimination-order-product).
*       Output          none
*       Description		finds an elimination order for the variables in 
*						'elimVarArr'.
*       Comments		updated 23.10.
*
******************************************************************************/
long updatePedElimOrderPart_Stoch(elimOrderVar **elimVarsArr, long numNodesInArr, 
								  long elimOrdInd, elimOrderPed *elimOrd, 
								  elimOrderProduct *elimProd, double *tableSizes)
{
	elimOrderVar		*elimVar, **nextElimVar;
	long				elimInd, numNodesLeft;
	elimOrderNode		*elimNode;
	varInPool			*bestVar, *worstVar, *currVar, *tmpVar;
	int					cntVars, numVars;
	boolean				isBreakVar;

	/* 1.	Initialize the elimination order index. */

	elimInd = elimOrdInd;
	elimNode = &((elimOrd->elimOrderArr)[elimInd]);

	numNodesLeft = numNodesInArr;
	nextElimVar = &elimVarsArr[0];

	/* 2.	Allocate memory for the linked list that holds the information on the
			variables in the variable pool. */

	myMalloc(bestVar, sizeof(varInPool));
	worstVar = bestVar;

	bestVar->prev = NULL;
	
	for (cntVars = 1; cntVars < VAR_POOL_SIZE; cntVars++) {
		myMalloc(worstVar->next, sizeof(varInPool));
		(worstVar->next)->prev = worstVar;
		worstVar = worstVar->next;
	}

	worstVar->next = NULL;

	/* 3.	While there are still variables to eliminate. */

	while (numNodesLeft) {		
		
		/* 3.1.	Choose a variable to eliminate (or condition on) next, using a 
				probabilistic algorithm. */

		elimVar = getNextVarLGPartProb(elimVarsArr, numNodesInArr, &nextElimVar,
									   &bestVar, &worstVar, &numVars, &isBreakVar);

		/* 3.2.	Either perform variable elimination or conditioning, according to the
				type of variable. */


		if (isBreakVar) {
			conditionOnElimOrderVar(elimVar, numVars, &elimInd, &elimNode, elimProd);
		}
		else {
			eliminateElimOrderVars(elimVar, numVars, tableSizes, &elimInd, &elimNode, elimProd);
		}


/*		#if (PRINT_VARIOUS_STATISTICS_LEVEL > 1)
			if (elimVar->cliqueSize > elimOrd->treeWidth)
				elimOrd->treeWidth = elimVar->cliqueSize;
		#endif
*/

		numNodesLeft -= numVars;

		#if CALC_MEM_REQ
			if (elimProd->currMemReq > elimOrd->maxMemReq) {
				elimOrd->maxMemReq = elimProd->currMemReq;
			}
		#endif
	}

	/* 4.	Free the memory allocated for the linked list. */

	for (currVar = bestVar; currVar != NULL;) {
		tmpVar = currVar;
		currVar = currVar->next;
		myFree(tmpVar);
	}

	return elimInd;
}


/******************************************************************************
*
*       Function        getNextVarLGPart
*       Parameters		elimVarsArr (an array of elimination-order-vars for 
*						which an elimination order needs to be found)
*						numNodesInArr (the number of nodes in elimVarsArr), 
*					    nextElimVar (the next variable that is tested for 
*						elimination).
*       Output          none
*       Description		returns a pointer to the variable choosen next to be 
*						eliminated (the variable that's chosen is least complex
*						according to the some greedy heuristic).						 
*       Comments		none
*
******************************************************************************/
elimOrderVar* getNextVarLGPart(elimOrderVar **elimVarsArr, long numNodesInArr, 
							   elimOrderVar ***nextElimVar, int *numVars, 
							   boolean *isBreakVar)
{
	elimOrderVar	**varPtr, **varPtrEnd, *var, **bestVarPtr;
	double			bestScore;


	/* 1.	Initialize. */

	bestScore = DBL_MAX;
	bestVarPtr = NULL;

	/* 2.	Iterate over the existing variables looking for the best variable 
			to eliminate, starting from nextElimVar and wrapping around. */

	varPtr = *nextElimVar;
	varPtrEnd = &elimVarsArr[0] + numNodesInArr;

	do {
		var = *varPtr;

		if ((var->used) && (var->currNumFunctions > 0)) {
			
			/* 2.1.	If the complexity value of this variable is invalid, 
					we need to recalculate it. */

			if (var->elimComplexity == INVALID)
				calcElimComplexity(var);

			/* 2.2.	If the complexity is simple enough, the search will be 
					stopped here. */

			if (var->elimComplexity <= ELIMINATION_BOTTOM_THRESHOLD) {
				bestVarPtr = varPtr;
				goto useThisVar_LG;
			}

			/* 2.3.	Else, compare the complexity against the best var up till now. */

			if (var->elimComplexity < bestScore) {
				bestScore = var->elimComplexity;
				bestVarPtr = varPtr;
			}
		}

		/* 2.4.	Continue to check the next variable. */

		varPtr++;

		if (varPtr == varPtrEnd) 
			varPtr = &elimVarsArr[0];

	} while (varPtr != *nextElimVar);

	useThisVar_LG:

	/* 3.	If the complexity of this variable is too large, find a variable to perform
			product-breaking on. */

	if ((*bestVarPtr)->tableSize > breakProductBottomThreshold) {
		*isBreakVar = TRUE;
		return getBreakVarLGPart(elimVarsArr, numNodesInArr, numVars);
	}


	*isBreakVar = FALSE;

	#if	UNITE_VARS
		if ((*bestVarPtr)->originalNode->uniteT) {
			*numVars = (*bestVarPtr)->originalNode->numVarsUnite;
		}
		else {
	#endif
		*numVars = 1;
	#if	UNITE_VARS
		}
	#endif

	/* 3.	Update the 'nextElimVar' to point to the following variable in
			the variable array. */

	*nextElimVar = bestVarPtr + 1;

	if (*nextElimVar >= varPtrEnd)
		*nextElimVar = &elimVarsArr[0];

	/* 4.	Return a pointer to the variable chosen for elimination. */

	return (*bestVarPtr);
}



/******************************************************************************
*
*       Function        getEliminateNodeLGPartProb
*       Parameters		elimVarsArr (an array of elimination-order-vars for 
*						which an elimination order needs to be found)
*						numNodesInArr (the number of nodes in elimVarsArr), 
*					    nextElimVar (the next variable that is tested for 
*						elimination).
*						eliminateComplexity (the elimination complexity
*						of the variable chosen to be eliminated next).
*       Output          none
*       Description		returns a pointer to the variable choosen next to be 
*						eliminated (the variable that's chosen is least complex
*						according to the some greedy heuristic).						 
*       Comments		none
*
******************************************************************************/
elimOrderVar* getNextVarLGPartProb(elimOrderVar **elimVarsArr, long numNodesInArr, 
								   elimOrderVar ***nextElimVar, 
								   varInPool **bestVarPtr, varInPool **worstVarPtr,
								   int *numVars, boolean *isBreakVar)
{
	elimOrderVar	**varPtr, **varPtrEnd, *var, **chosenVarPtr = NULL;
	double			totalProb, coin, lowerBound;
	varInPool		*currVar, *tmpVar, *bestVar, *worstVar;


	/* 1.	Initialize. */

	bestVar = *bestVarPtr;
	bestVar->prev = NULL;
	worstVar = *worstVarPtr;
	worstVar->next = NULL;

	for (currVar = bestVar; currVar != NULL; currVar = currVar->next) {
		currVar->complexity = DBL_MAX;
		currVar->varPtr = NULL;
	}

	/* 2.	Iterate over the existing variables looking for the best variable 
			to eliminate, starting from nextElimVar and wrapping around. */

	varPtr = *nextElimVar;
	varPtrEnd = &elimVarsArr[0] + numNodesInArr;

	do {
		var = *varPtr;

		if ((var->used) && (var->currNumFunctions > 0)) {
			
			/* 2.1.	If the complexity value of this variable is invalid, 
					we need to recalculate it. */

			if (var->elimComplexity == INVALID)
				calcElimComplexity(var);

			/* 2.2.	If the complexity is simple enough, the search will be 
					stopped here. */

			if (var->elimComplexity <= ELIMINATION_BOTTOM_THRESHOLD) {
				chosenVarPtr = varPtr;
				goto useThisVarProb_LG;
			}

			/* 2.3.	Else, compare the complexity of the current varible to the elimination
					complexities of the variables in the pool. If it is better than that of some
					of the variables in the pool and the pool is full, replace the worst variable
					in the pool with the current variable. Iterate over the variables, starting
					from the worst one. */

			for (currVar = worstVar; currVar != NULL;) {

				/* 2.3.1.	If the complexity of the current variable is bigger or equal to 
							that of 'currVar', */

				if (currVar->complexity <= var->elimComplexity) {

					/* 2.3.1.1.	If 'currVar' is the worst var in the pool, then break (we
								don't want to insert it into the variable pool). */

					if (currVar->next == NULL)
						break;

					/* 2.3.1.2.	Else, insert the current variable right after 'currVar' and
								remove the last variable from the list. */

					tmpVar = worstVar;
					worstVar = worstVar->prev;
					(tmpVar->prev)->next = NULL;
					tmpVar->complexity = var->elimComplexity;
					tmpVar->varPtr = varPtr;
					tmpVar->next = currVar->next;
					tmpVar->prev = currVar;
					currVar->next = tmpVar;

					if (tmpVar->next != NULL)
						(tmpVar->next)->prev = tmpVar;

					if (worstVar->next != NULL)
						worstVar = worstVar->next;

					break;
				}

				/* 2.3.2.	If the complexity of the current variable is smaller than that of 
							'currVar' which is the bestVar, */

				if ((currVar->prev == NULL) && (currVar->complexity > var->elimComplexity)) {
					tmpVar = worstVar;
					worstVar = worstVar->prev;
					(tmpVar->prev)->next = NULL;
					tmpVar->complexity = var->elimComplexity;
					tmpVar->varPtr = varPtr;
					tmpVar->next = currVar;
					tmpVar->prev = NULL;
					currVar->prev = tmpVar;
					bestVar = tmpVar;
					break;
				}

				currVar = currVar->prev;
			}
		}

		/* 2.4.	Continue to check the next variable. */

		varPtr++;

		if (varPtr == varPtrEnd) 
			varPtr = &elimVarsArr[0];

	} while (varPtr != *nextElimVar);

	/* 3.	Iterate over the variables in the pool and calcultae their probability
			to be chosen, according to the specified probability function. */

	totalProb = 0.0;

	for (currVar = bestVar; currVar != NULL; currVar = currVar->next) {
		
		/* 3.1.	If the variable pool isn't full, give the empty cells zero probability.
				Otherwise, the probability is determined according to the appropriate
				probability function. */

		if (currVar->varPtr == NULL) {
			currVar->relProb = 0.0;
		}
		else {
			currVar->relProb = 1.0 / sqrt(currVar->complexity);
			totalProb += currVar->relProb;
		}
	}

	/* 4.	Iterate over the variables in the pool and determine the relative probability
			of each one of them. */

	for (currVar = bestVar; currVar != NULL; currVar = currVar->next) {
		
		if (currVar->varPtr != NULL) {
			currVar->relProb /= totalProb;
		}
	}

	/* 5.	Toss a coin to determine which variable from the pool will be chosen. */

	coin = (double)(rand()) / RAND_MAX;

	/* 6.	Determine which variable from the pool has been chosen. This is done by
			dividing the [0,1] interval to subintervals according to the number of 
			variables in the pool. Each subinterval corresponds to one of the variables
			and its size is equal to the relative probability of this variable to be chosen. */

	if (coin == 1.0) {

		currVar = worstVar;

		while (currVar->varPtr == NULL)
			currVar = currVar->prev;

		chosenVarPtr = currVar->varPtr;
		goto useThisVarProb_LG;
	}

	for (lowerBound = 0, currVar = bestVar; currVar != NULL; currVar = currVar->next) {

		if (coin <= lowerBound + currVar->relProb) {
			chosenVarPtr = currVar->varPtr;
			break;
		}

		lowerBound += currVar->relProb;
	}

	useThisVarProb_LG:

	(*bestVarPtr) = bestVar;
	(*worstVarPtr) = worstVar;

	if ((*chosenVarPtr)->tableSize > breakProductBottomThreshold) {
		*isBreakVar = TRUE;
		return getBreakVarLGPart(elimVarsArr, numNodesInArr, numVars);
	}

	*isBreakVar = FALSE;

	#if	UNITE_VARS
		if ((*chosenVarPtr)->originalNode->uniteT) {
			*numVars = (*chosenVarPtr)->originalNode->numVarsUnite;
		}
		else {
	#endif
		*numVars = 1;
	#if	UNITE_VARS
		}
	#endif

	/* 3.	Update the 'nextElimVar' to point to the following variable in
			the variable array. */

	*nextElimVar = chosenVarPtr + 1;

	if (*nextElimVar >= varPtrEnd)
		*nextElimVar = &elimVarsArr[0];

	/* 4.	Return a pointer to the variable chosen for elimination. */

	return (*chosenVarPtr);
}



/******************************************************************************
*
*       Function        getBreakVarLGPart
*       Parameters		elimVarsArr (an array of elimination-order-vars for 
*						which an elimination order needs to be found)
*						numNodesInArr (the number of nodes in elimVarsArr), 
*       Output          none
*       Description		finds & returns the "best" variable (among the variables
*						that haven't been eliminated yet) to break the product
*						upon, according to the chosen heuristic. 						 
*       Comments		none
*
******************************************************************************/
elimOrderVar* getBreakVarLGPart(elimOrderVar **elimVarsArr, long numNodesInArr,
								int *numVars)
{
	elimOrderVar	**varPtr, **varPtrEnd, *var, **bestVarPtr;
	double			bestScore, currScore;

	/* 1.	Initialize. */

	bestScore = MIN_DBL;
	bestVarPtr = NULL;

	/* 2.	Iterate over the variables in the array. */

	for (varPtr = &elimVarsArr[0], varPtrEnd = varPtr + numNodesInArr;
			varPtr < varPtrEnd; varPtr++) {

		var = *varPtr;
		
		/* 2.1.	If the variable hasn't been eliminated yet: */ 

		if (var->used)	{

			/* 2.1.1.	Calculate its score. */

			currScore = var->elimComplexity * sqrt(var->currNumFunctions);

			/* 2.1.2.	Compare the variable's score to the best score found 
						up to now. */

			if (currScore > bestScore) {
				bestScore = currScore;
				bestVarPtr = varPtr;
			}
		}
	}

	#if	UNITE_VARS
		if ((*bestVarPtr)->originalNode->uniteT) {
			*numVars = (*bestVarPtr)->originalNode->numVarsUnite;
		}
		else {
	#endif
		*numVars = 1;
	#if	UNITE_VARS
		}
	#endif

	/* 3.	Return the variable chosen for conditioning. */

	return (*bestVarPtr);
}



