/*********************************************************************************/
/*  Haplotype.c																     */
/*  Author: Ma'ayan Fishelson													 */
/*********************************************************************************/

#include "sl_Haplotype.h"
Bool clusterMPEHaplotypes();
/******************************************************************************
*
*       Function        updateAllelesLociHaplo
*       Parameters		none
*       Output          none
*       Description		Finds the number of markers (ALLELES type loci) in the 
*						analysis, updates the mapToData array to point only to 
*						the ALLELES type loci and updates the theta array to hold
*						the thetas between each two successive markers.
*       Comments		none
*
******************************************************************************/
void updateAllelesLociHaplo()
{
	int				locusInd, locusIndMarkerMap, numThetas, thetaInd, 
					thetaIndMarkerMap, numLoci;
	int				deseaseStatusLocus = - 1,  firstNull = 0;
	Probability		accumThetaMale, accumThetaFemale;
	boolean			useAccumTheta;
	locus			*locusPtr;
	thetaCompTheta	*maleThetas, *femaleThetas;

	maleThetas = recombData.maleThetas;
	femaleThetas = recombData.femaleThetas;

	/* 1.	Update the number of thetas on the map. */

	numLoci = loci.numLoci;
	numThetas = numLoci - 1;

	useAccumTheta = FALSE;

	/* 2.	Iterate over the loci in the data-structures. */

	for (locusInd = 0, locusIndMarkerMap = 0, thetaInd = -1, thetaIndMarkerMap = 0,
			locusPtr = loci.locusArr; locusInd < numLoci; locusInd++, thetaInd++,
			locusPtr++) {

		if (locusPtr->type == ALLELES) {

			/* 2.1.1.	If this is an ALLELES type locus, and this isn't the first 
						marker on the map, update the theta value between it and the 
						previous marker. */
			
			if (locusIndMarkerMap != 0) {
				if (useAccumTheta) {
					maleThetas[thetaIndMarkerMap][0] = 
									getHaldaneThetaAC(accumThetaMale, maleThetas[thetaInd][0]);
					femaleThetas[thetaIndMarkerMap][0] = 
									getHaldaneThetaAC(accumThetaFemale, femaleThetas[thetaInd][0]);

					useAccumTheta = FALSE;
				}
				else {
					maleThetas[thetaIndMarkerMap][0] = maleThetas[thetaInd][0];
					femaleThetas[thetaIndMarkerMap][0] = femaleThetas[thetaInd][0];
				}
				maleThetas[thetaIndMarkerMap][1] = 1.0 - maleThetas[thetaIndMarkerMap][0];
				femaleThetas[thetaIndMarkerMap][1] = 1.0 - femaleThetas[thetaIndMarkerMap][0];
				
				thetaIndMarkerMap++;
			}

			/* 2.1.2.	Update the pointer to this locus in the mapping from map to 
						data-structures , and advance the index of the ALLELES type loci. */

			loci.mapToData[locusIndMarkerMap++] = locusInd;
		}
		else {

			if (locusPtr->type == affection){
				deseaseStatusLocus = locusInd;
			}
			else{
				firstNull++;

				if (firstNull > 1) {
					deseaseStatusLocus = locusInd;
				}
			}

			/* 2.1.1.	If this isn't an ALLELES type locus, and there are markers
						preceeding it on the map, turn on the 'useAccumTheta' flag 
						and add the current theta values to the accumalated thetas. */

			if (locusIndMarkerMap != 0) {
				if (useAccumTheta) {
					accumThetaMale = getHaldaneThetaAC(accumThetaMale, maleThetas[thetaInd][0]);
					accumThetaFemale = getHaldaneThetaAC(accumThetaFemale, femaleThetas[thetaInd][0]);
				}
				else {
					accumThetaMale = maleThetas[thetaInd][0];
					accumThetaFemale = femaleThetas[thetaInd][0];
					useAccumTheta = TRUE;
				}
			}
		}
	}

	/* 3.	Updates 'numMarkers' to the number of ALLELES type loci in the data. */

	numMarkers = locusIndMarkerMap;

	/* 4.	Save desease locus at last position */

	if (deseaseStatusLocus >= 0){
		loci.mapToData[locusIndMarkerMap] =  deseaseStatusLocus;
	}
}



/******************************************************************************
*
*       Function        findMostLikelyHaplotypes
*       Parameters		prodArr (an array of products which represent the
*						Bayesian networks built from the input pedigrees),
*						elimOrderArrs (an array of arrays which represent the
*						elimination order determined for the variables of the
*						products), isZeroLike (is updated to ZERO if the 
*						likelihood is ZERO).
*       Output          none
*       Description		Finds the most likely haplotypes (or one most likely 
*						assignemt) for the individuals in the input pedigrees 
*						and prints them into a file.
*       Comments		If there isn't any possible assignment to the haplotypes 
*						(the likelihood of the data is zero, then updates the flag 
*						'isZeroLike' to zero).
*
******************************************************************************/
double findMostLikelyHaplotypes(product** prodArr, elimOrder *elimOrderArrs, 
							  argMaxData *argMaxArr, PedigreeDataStruct *origPedigree)
{
	int				pedNum;
	elimOrderPed	*elimOrdPtr;
	boolean			*isZeroLikePed;
	boolean isZeroLike = FALSE;
	/* 1.	Iterate over the pedigrees. */

	for (pedNum = 0, elimOrdPtr = *elimOrderArrs, isZeroLikePed = totalRes.isZeroLikeArr; 
			pedNum < peds.numPeds; pedNum++, elimOrdPtr++, isZeroLikePed++) {

		if (*isZeroLikePed) {
			continue;
		}
		
		elimOrdPtr->currElimNode = 0;

		/* 1.1	Find the most probable haplotypes for the people of each pedigree. */
		
		findMostLikelyHaplotypesPed(prodArr[pedNum], pedNum, elimOrdPtr, argMaxArr[pedNum]);
		isZeroLike = *isZeroLikePed || isZeroLike ;
	}

	/* 2.	Print the most likely haplotypes into a file. */
/*
	if (!loci.sexLink)
		printMPEHaplotypes();
	else
		printMPEHaplotypesSexLink();
		*/

		if (isZeroLike)
			return MIN_DBL;
		clusterMPEHaplotypes(origPedigree);
		return 1;

}


/******************************************************************************
*
*       Function        findMostLikelyHaplotypesPed
*       Parameters		prod (the product which represents the Bayesian network 
*						for which MPE is being calculated), currPed (the pedigree
*						for which people the most probable haplotypes are being
*						determined), elimOrderArr (an array that stores an 
*						elimination order for the variable of this product) .
*       Output          none
*       Description		calculates the MPE query for the pedigree ('currPed') 
*						represented by the product 'prod', according to the 
*						elimination order stored in 'elimOrderArr'.
*       Comments		returns FALSE if the likelihood is zero (i.e., there's
*						no possible assignemnt to the variables of the product), 
*						otherwise returns TRUE.  The log-likelihood calculated 
*						is added to the total log-likelihood calculated up till now.
*
******************************************************************************/
void findMostLikelyHaplotypesPed(product *prod, int pedNum, elimOrderPed *elimOrderArr,
									argMaxData argMax)
{
	Probability				Res;
	argMaxData				head, tail;


	/* 1.	Initializations. */

	head = tail = NULL;

	/* 2.	Perform inference (MPE query) on the product. */

	Res = inferMPEProduct(prod, elimOrderArr, &head, &tail, 1, 0);
	
	/* 3.	Add the result of the inference to the beginning of the list pointed to by
			argMax. */

	if (head == NULL) {
		head = argMax;
	}
	else {
		tail->next = argMax;
	}

	/* 4.	Backtrack along the variables of the product to determine the most probable
			assignment to the variables of the product (the forward phase of the MPE-elimination
			algorithm. */

	traceMPEProduct(head);

	/* 5.	Free the memory allocated to the argMaxData list. */

	freeAMList(head);

	if (Res < MIN_DBL || totalRes.currRes->lnLike < MIN_DBL)	{				/* The likelihood of this pedigree is zero. */
		totalRes.isZeroLikeArr[pedNum] = TRUE;
		//return FALSE;
	}
	else {
		totalRes.pedsLnLikeArr[pedNum] += Res;
		(totalRes.currRes)->lnLike += Res;
		//return TRUE;
	}
}


/******************************************************************************
*
*       Function        inferMPEProduct
*       Parameters		prod (pointer to the product on which the inference
*						is being performed), problemSize (the size of the 
*						problem that needs to be solved by the current call to
*						the function) , solved (how much of the original problem 
*						has already been solved), optValsProd (the optimizingVals
*						product which stores the information needed for the 
*						backtracking part of the inference), currPed (the pedigree
*						which product represents).
*       Output          if requested, prints a progress report (how much of the
*						original problem has been solved already).
*       Description		performs inference according to the MPE query on 'prod'.						 
*       Comments		returns the result of the inference.
*
******************************************************************************/
Probability inferMPEProduct(product *prod, elimOrderPed *elimOrderArr, 
						    argMaxData *head, argMaxData *tail,  
							float problemSize, float solved)
{
	variable			*nextVar;
	Probability			res, tempRes;
	boolean				isBreakVar, isZeroFunc;
	argMaxData			tmpTail;

	/* 1.	Initialize the probability to zero. */

	res = 0.0;

	/* 2.	Perform MPE inference on the product by eliminating the variables one by one. 
			While there are still variables to eliminate: */

	while (prod->currNumVars) {		

		/* 2.1.	Choose a variable to eliminate (or condition on) next. */

		nextVar = getEliminateVar(elimOrderArr, &isBreakVar);

		/* 2.2.	If the next variable is a break variable (a variable that we 
				condition on), perform product-breaking on it. */

		if (isBreakVar) {
			tempRes = inferMPEProductWithVariableBreaking(prod, nextVar, elimOrderArr,
														  head, &tmpTail, 
														  problemSize, solved);

			/* 2.2.1.	If the result of the inference on the break variable isn't zero: */

			if (tempRes != MIN_DBL) {

				/* 2.2.1.1.	Update the current result of the inference. */

				res += tempRes;

				/* 2.2.1.2.	If the tail of the argMaxData list hasn't been updated yet, update it. */

				if (*tail == NULL) {
					*tail = tmpTail;
				}
			}
			else {
				return MIN_DBL;
			}

			return res;
		}

		/* 2.3.	Else, just eliminate the chosen variable from the product. 
				If the elimination resulted in a constant function, add
				its logProbability to the total logProbability. */
		
		if (eliminateVarMax(prod, nextVar, &tempRes, head, &isZeroFunc)) {
			res += tempRes;	
		}

		/* 2.4.	If the tail of the argMaxData list isn't updated yet, update it to point
				to the argMaxData structure created in this iteration. */

		if (*tail == NULL) {
			*tail = *head;
		}
	}

	/* 3.	If requested, print a progress report. */

	#if PRINT_PROGRESS_REPORT
		printf("Solved %f percent of the problem.\n", (solved+problemSize)*100);
	#endif

	/* 4.	Return the result of the inference. */

	return res;
}


/******************************************************************************
*
*       Function        inferMPEProductWithVariableBreaking
*       Parameters		prod (the product that needs to be infered), 
*						breakVar (the variable on which the product is broken)  
*						problemSize (the size of the problem that needs to be
*						solved by the current call to the function),
*						solved (the size of the problem that has already been
*						solved), optValsProd (the optimizingVals product which
*						stores the information needed for the backtracking part 
*						of the inference), currPed (the pedigree which product 
*						represents).
*       Output          none
*       Description		performs inference on 'prod' by breaking the Product						 
*						on 'breakVar' (conditioning on it) and performing 
*						inference for each value of the breakVar seperately 
*						and then combining the results.
*       Comments		returns the result of the inference.
*
******************************************************************************/
Probability inferMPEProductWithVariableBreaking(product *prod, variable *breakVar,
											    elimOrderPed *elimOrd,
											    argMaxData *head, argMaxData *tail, 
												float problemSize, float solved)
{
	int					breakVarVal, optBreakVarVal;
	product				*brokenProd;
	variable			*brokenVar;
	Probability			currRes, maxRes;
	float				newProblemSize, currSolved;
	long				currElimNode;
	elimOrderNode		*currNode, *endNode;
	boolean				*isTransformedArr, *isTransformedArrPtr, isZeroLike, 
						areValsElim;
	argMaxData			maxHead, maxTail, tmpHead, tmpTail, inferHead, inferTail;

	/* 1.	Allocate memory for an array that will hold the temporary values of
			variables in the elimination-order-array. */

	myMalloc(isTransformedArr, sizeof(boolean) * (elimOrd->numNodes - elimOrd->currElimNode));

	/* 2.	Calculate the size of the problem that will be solved by each call
			to the 'inferProduct' function. */

	newProblemSize = problemSize / breakVar->numValues;
	currSolved = solved;

	currElimNode = elimOrd->currElimNode;

	maxRes = MIN_DBL;
	optBreakVarVal = 0;
	
	maxHead = maxTail = NULL;

	/* 6.	Perform inference on the Product for each value of the breakVar separately .*/

	for (breakVarVal = 0; breakVarVal < breakVar->numValues; breakVarVal++) {
		
		/* 6.1.	Duplicate the Product. */

		brokenProd = copyProduct(prod, elimOrd, isTransformedArr, &breakVar, &brokenVar, 1);

		/* 6.2.	Set the value of 'brokenVar' to 'breakVarVal' and change 
				'brokenProd' accordingly. */

		currRes = setBreakVal(brokenProd, brokenVar, breakVarVal);

		/* 6.3.	Eliminate impossible variable values from the Product's functions. */

		currRes += eliminateImpossibleValuesHaplo(brokenProd, &isZeroLike, &areValsElim, &tmpHead);

		/* 6.4.	Update tmpTail to point to the argMaxData structure that was created for this
				break node. If none was created (no values were eliminated), create a dummy node. */

		if (tmpHead) {
			tmpTail = tmpHead;
		}
		else {
			tmpTail = tmpHead = createAMDataOther(NULL, NULL, 0, NULL, dummyNode);
		}

		/* 6.5.	If this value of the break var is possible: */

		if (!isZeroLike) {

			inferHead = inferTail = NULL;

			/* 6.5.1.	Perform inference on 'brokenProd'. */

			currRes += inferMPEProduct(brokenProd, elimOrd, &inferHead, &inferTail, 
									    newProblemSize, currSolved);

			/* 6.5.2.	If the likelihood isn't zero: */
//anna add check "inferTail != NULL " to avoid segmantation fault
			// Have no idea why there is  segmantation fault
			if (currRes != MIN_DBL && inferTail != NULL ) {

				/* 6.5.2.1.	Add the argMaxData structure of the break variable to the end of the 
							argMaxData structure list of all the variables following it in the 
							elimination order. */

				inferTail->next = tmpHead;
				tmpHead = inferHead;

				/* 6.5.2.2.	Update the maximum result and the argMaxData list, if necessary. */

				if (currRes > maxRes) {

					/* 6.5.2.2.1.	If the current result is better than the best one found up
									to now, save this result, save the value of the break var which 
									produced this result, free the memory allocated to the argMaxData  
									list of the previous best var value and switch between the 
									argMaxData lists. */

					maxRes = currRes;
					optBreakVarVal = breakVarVal;

					freeAMList(maxHead);

					maxHead = tmpHead;
					maxTail = tmpTail;

					goto isNotZero;
				}
			}
		}
		else {		/* This var value isn't possible. */
			/* 6.5.1.	If requested, print a progress report. */

			#if PRINT_PROGRESS_REPORT
				printf("Solved %f percent of the problem.\n", (currSolved+newProblemSize)*100);
			#endif
		}

		/* 6.6.	Free the memory allocated for the argMaxData list 
				of the current var value. */

		freeAMList(tmpHead);

isNotZero:
		elimOrd->currElimNode = currElimNode;

		/* 6.6.	Restore the previous values of the nodes in the elimination-order array. */

		for (currNode = &(elimOrd->elimOrderArr)[elimOrd->currElimNode],
				endNode = &(elimOrd->elimOrderArr)[elimOrd->numNodes], 
				isTransformedArrPtr = isTransformedArr; 
				currNode < endNode; currNode++, isTransformedArrPtr++) {

			if (*isTransformedArrPtr == TRUE) {
				(currNode->originalNode)->varInProd = 
							((currNode->originalNode)->varInProd)->varInOldProd;
			}
		}

		/* 6.7.	Free the memory allocated for the duplicated product. */

		freeProduct(brokenProd);

		/* 6.8.	Update the size of the problem solved up to now. */

		currSolved += newProblemSize;
	}

	/* 7.	If the likelihood isn't zero: */

	if (maxTail != NULL) {

		/* 7.1.	Update the argMaxData structure of the break var with the value that produced
				the maximum, and the list of argMaxData structures of the variables 
				following it in the elimination order. */

		maxTail->mainNode = breakVar->originalNode;
		maxTail->type = breakNode;
		((maxTail->specificData).otherNodeData).val = optBreakVarVal;

		/* 7.2.	Updata 'maxTail' to point to the real tail of the list (past the argMaxData
				structures of the variables that were eliminated to constants in this iteration). */

		while (maxTail->next != NULL) {
			maxTail = maxTail->next;
		}
		
		/* 7.3.	Concatenate the argMaxData structures list passed to the function to the end of the
				list created in this function and update 'head' and 'tail' to point to the endpoints
				of the list created in this function only. */

		maxTail->next = *head;
		*head = maxHead;
		*tail = maxTail;
	}
	else {
		/* 7.1.	Update that the likelihood is zero. */

		maxRes = MIN_DBL;
	}

	/* 8.	Free the memory allocated dynamically. */

	myFree(isTransformedArr);

	/* 9.	Return the result of the inference performed. */

	return maxRes;
}


/******************************************************************************
*
*       Function        eliminateVarMax
*       Parameters		prod (the product from which a variable needs to be 
*						eliminated), elimVar (pointer to the variable that needs
*						to be eliminated), tempRes (if the combined function is
*						a constant, holds the value of this constant),
*						optValsProd (pointer to the optimizingVals product in
*						which the optimizingVals function which would be created
*						during the elimination of this variable needs to be
*						stored.
*       Output          none
*       Description		eliminates 'elimVar' from 'prod' by multiplying all
*						the functions that it appears in and creating one new
*						combined function that 'elimVar' doesn't appear in.					 
*       Comments		returns TRUE if the combined function is a constant 
*						and FALSE otherwise.
*
******************************************************************************/
boolean eliminateVarMax(product *prod, variable *elimVar, Probability *tempRes,
						argMaxData *head, boolean *isZeroFunc)
{
	variable			**varArr, **varPtr, **varPtrEnd;
	int					numVars, rescaleTimer, *argMaxList;
	unsigned long		sizeOfTable;
	Probability			*probabList;
	function			*combinedFunc = NULL, **funcPtr, **funcPtrEnd, *func;
	double				logMultFactor;
	argMaxNodeData		*nodes, *nodePtr;
	argMaxData			argMax;

	/* 1.	Allocate memory for the variable array of the new function that
			will be created, as a result of the elimination of the variable. */

	myMalloc(varArr, sizeof(variable*) * prod->currNumVars);

	/* 2.	Find all the variables that appear in the functions that need 
			to be combined. */

	getVarsInCombinedFunc(&elimVar, 1, varArr, elimVar->functions, elimVar->currNumFunctions,
						  &numVars, &sizeOfTable);

	/* 3. Allocate an array to store the data on the nodes that this function depends on, and
		  update it. */

	if (numVars) {
		myMalloc(nodes, sizeof(argMaxNodeData) * numVars);

		for (varPtr = varArr, varPtrEnd = varPtr + numVars, nodePtr = nodes;
				varPtr < varPtrEnd; varPtr++, nodePtr++) {

			nodePtr->origNode = (*varPtr)->originalNode;
			nodePtr->currNumVals = (*varPtr)->numValues;
		}
	}
	else {
		nodes = NULL;
	}

	/* 4.	Calculate the new logMultFactor & rescaleTimer. */

	logMultFactor = 0.0;
	rescaleTimer = MAX_RESCALE_UNITS;

	for (funcPtr = elimVar->functions, funcPtrEnd = funcPtr + elimVar->currNumFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {
		
		logMultFactor += (*funcPtr)->logMultFactor;

		if ((*funcPtr)->rescaleTimer < rescaleTimer)
			rescaleTimer = (*funcPtr)->rescaleTimer;
	}

	/* 5.	Allocate memory for the new combined function and the list of values of the
			eliminated variable that maximize the join probability table for each value 
			assignment for the rest of the variables of the function,  calculate them & 
			insert the new combined function into the product and the argMaxList into the 
			argMaxData structure, in case it isn't a constant. */

	myMalloc(probabList, sizeof(Probability) * sizeOfTable);
	myMalloc(argMaxList, sizeof(int) * sizeOfTable);

	/* 6.	Create an argMaxData structure to store the data on this eliminated variable. */

	argMax = createAMDataElim(elimVar->originalNode, *head, numVars, nodes, sizeOfTable, argMaxList);
	
	if (numVars) {		/* In case the combined function isn't a constant. */

		/* 6.1.	If there's more than one function that needs to be combined,
				decrease the rescaleTimer by one. */

		if (elimVar->currNumFunctions > 1)
			rescaleTimer--;

		/* 6.2.	Insert the combined function into the product. */

		combinedFunc = insertFunctionToProduct(prod, sizeOfTable, probabList, logMultFactor,
											   TRUE, numVars, varArr, TRUE, rescaleTimer);

		/* 6.4.	Calculate the new combined function. */

		if (sizeOfTable * elimVar->currNumFunctions < ELIMINATION_SIMPLE_THRESHOLD)
			(*isZeroFunc) = multiplyAllFuncsMaxSimp(combinedFunc, elimVar, argMaxList);	
		else
			(*isZeroFunc) = multiplyAllFuncsMaxComplex(combinedFunc, elimVar, argMaxList);	

		/* 6.5.	If, according to the rescaleTimer, the combined function 
				needs to be rescaled, rescale it.*/

		if (*isZeroFunc == FALSE) {
			if (combinedFunc->rescaleTimer <= 0) {
				scaleFunctionProbList(combinedFunc);
				combinedFunc->rescaleTimer = MAX_RESCALE_UNITS;
			}
		}
	}
	else {				/* In case the combined function is a constant. */
		*tempRes = multiplyAllOneVarFuncsMax(elimVar, logMultFactor, argMaxList);
		myFree(probabList);
	}

	/* 7.	Remove the functions that have been multiplied from the variables that
			appeared in them and from the Product. */

	for (funcPtr = elimVar->functions + elimVar->currNumFunctions - 1, 
			funcPtrEnd = elimVar->functions; funcPtr >= funcPtrEnd; funcPtr--) {

		func = *funcPtr;

		for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			removeFunctionFromVariable(*varPtr, func);
		}

		removeFunctionFromProduct(prod, func);
	}

	/* 8.	Add the new combined function to the variables that appear in it. */

	if (numVars) {
		for (varPtr = combinedFunc->variables, varPtrEnd = varPtr + combinedFunc->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			addFunctionToVariable(*varPtr, combinedFunc);
		}
	}

	/* 9.	Trim the variable that was eliminated from the product. */

	removeVariableFromProduct(prod, elimVar);

	/* 10.	Free the memory allocated dynamically. */
	
	myFree(varArr);

	/* 11.	Update the head of the list to point to the new argMaxData structure that
			was created. */

	*head = argMax;

	/* 12.	Return TRUE if the new function is a constant. */

	return ((numVars == 0) ? TRUE : FALSE);
}


/******************************************************************************
*
*       Function        multiplyAllFuncsMaxSimp
*       Parameters		combinedFunc (will store the result of the multiplication),
*						elimVar (the variable that needs to be eliminated),
*						optVals (an array that would store the value of 'elimVar'
*						which optimized the combined function for each possible 
*						assignment to the rest of the variables in the function).
*       Output          none
*       Description		performs a multiplication of the probability functions
*						that elimVar is part of and returns the result of the 
*						multiplication in combinedFunc.
*       Comments		none
*
******************************************************************************/
boolean multiplyAllFuncsMaxSimp(function *combinedFunc, variable *elimVar,
								int *optVals)
{
	Probability *probabListPtr,			/* pointer to the entry of the combined 
										   function that is currently being updated. */
				funcsVal, currVal, maxVal;
	function	**funcPtr, **funcPtrEnd;
	variable	**varPtr, **varPtrEnd, *var;
	int			optVarVal, *optValsPtr;
	boolean		isZeroFunc;


	/* 1.	Initialize the values of the variables of the combined function
			to values that correspond to the first entry in the probability table. */

	initFunctionVars(combinedFunc, varPtr, varPtrEnd);

	/* 2.	Initialize pointers to the beginning of the probability list and the beginning
			of the optimizingVals list. */

	probabListPtr = combinedFunc->probabList;	
	optValsPtr = optVals;

	funcPtrEnd = elimVar->functions + elimVar->currNumFunctions;

	/* 3.	Initialize the jumpFactor of each of the functions that are about 
			to be combined. */

	for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {
		initJumpFactor(*funcPtr, elimVar);
	}

	isZeroFunc = TRUE;	

	/* 4.	Iterate over the entries of the new probability table and update them. */

	while (1) {
		
		start_loop_mafs:

		/* 4.1.	Initialize the value of each entry in the new probability table to MIN_DBL. */

		maxVal = MIN_DBL;
		optVarVal = 0;

		elimVar->currValue = 0;

		/* 4.2.	Calculate, for each function, the index in the probability table that
				corresponds to the current values of its variables. */

		for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {
			calcBaseIndex(*funcPtr);
		}

		/* 4.3.	Iterate over all of the eliminated variable's values and find the
				value which maximizes the probability of thr current assigment to the
				joint function's variables. */	

		for (; elimVar->currValue < elimVar->numValues; elimVar->currValue++) {
			
			/* 4.3.1.	Initialize the table entry for this variable value to 1.0 . */

			currVal = 1.0;

			/* 4.3.2.	Iterate over all the functions. For each function: */

			for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {

				/* 4.3.2.1.	Get the function's value. */

				funcsVal = getFuncValue((*funcPtr), elimVar->currValue);

				/* 4.3.2.2.	If the value is zero, we can continue to the next 
							variable value. */

				if (funcsVal == 0.0) {
					currVal = 0.0;
					break;
				}

				/* 4.3.2.3.	Otherwise, multiply the value for this variable value, 
							by the function's value. */

				currVal *= funcsVal;
			}
			
			/* 4.3.3.	If the value for this variable value is greater than the maximum
						value found up to now, update the maximum value and save the value
						of the eliminated variable which produced it. */

			if (currVal > maxVal) {
				maxVal = currVal;
				optVarVal = elimVar->currValue;
			}
		}

		/* 4.4.	Write the value into the new probability table, and write the value of 
				the eliminated var which produced it into the optimizing values array. */

		*probabListPtr = maxVal;
		*optValsPtr = optVarVal;

		if (maxVal != 0)
			isZeroFunc = FALSE;

		/* 4.5.	Advance the pointers to the next entries in the tables. */

		probabListPtr++;
		optValsPtr++;

		/* 4.6.	Advance the values of the function's variables to correspond to
				the next entry in the table. */

		advanceFunctionVars(combinedFunc, varPtr, varPtrEnd, var, start_loop_mafs);
	}

	return isZeroFunc;
}




/******************************************************************************
*
*       Function        multiplyAllFuncsMaxComplex
*       Parameters		combinedFunc (will store the result of the multiplication),
*						elimVar (the variable that needs to be eliminated),
*						optVals (an array that would store the value of 'elimVar'
*						which optimized the combined function for each possible 
*						assignment to the rest of the variables in the function)
*       Output          none
*       Description		performs a multiplication of the probability functions
*						that elimVar is part of and returns the result of the 
*						multiplication.
*       Comments		a more sophisticated function than the previous one.
*
******************************************************************************/
boolean multiplyAllFuncsMaxComplex(function *combinedFunc, variable *elimVar,
								   int *optVals)
{
	Probability		*probabListPtr, funcsVal, currVal, maxVal;
	function		**funcPtr, **funcPtrEnd;
	variable		**varPtr, **varPtrEnd, *var;
	incDec			*incDecArr, *incDecPtr, *incDecPtrEnd;
	int				indInCombinedFunc, varInd, funcInd, optVarVal, *optValsPtr;
	boolean			isZeroFunc;


	/* 1.	Allocate memory for an auxiliary array. */

	myMalloc(incDecArr, sizeof(incDec) * elimVar->currNumFunctions * combinedFunc->numVariables);

	/* 2.	Use the 'currValue' field of each variable to store its index in 
			the combinedFunc's varArray, temporarily. */

	for (indInCombinedFunc = 0, varPtr = combinedFunc->variables; 
			indInCombinedFunc < combinedFunc->numVariables; indInCombinedFunc++, varPtr++) {
		
		(*varPtr)->currValue = indInCombinedFunc;
	}

	/* 3.	Initialize all incDecArr's entries to zero. */

	for (incDecPtr = incDecArr, 
			incDecPtrEnd = incDecPtr + elimVar->currNumFunctions * combinedFunc->numVariables;
			incDecPtr < incDecPtrEnd; incDecPtr++) {

		incDecPtr->inc = 0;
		incDecPtr->dec = 0;
	}

	funcPtrEnd = elimVar->functions + elimVar->currNumFunctions;

	/* 4.	Calculate the necessary increment in the index of the probability table 
			of each function as a result of an increase of 1 in one of the function's 
			variables. Calculate the  necessary deccrement in the index of the probability 
			table of each function as a result of an decrease of (numVals-1) in one of 
			the function's variables. */

	for (funcPtr = elimVar->functions, incDecPtr = incDecArr, funcInd = 0; funcPtr < funcPtrEnd; 
			funcPtr++, incDecPtr += combinedFunc->numVariables, funcInd++) {

		initJumpFactors(*funcPtr, elimVar, incDecPtr);
		(*funcPtr)->baseInd = 0;
	}

	/* 5.	Initialize the values of the variables of the combined function
			to values that correspond to the first entry in the probability table. */

	initFunctionVars(combinedFunc, varPtr, varPtrEnd);

	/* 6.	Initialize pointers to the first entries of the probability list and the 
			optimizingVals list. */

	probabListPtr = combinedFunc->probabList;	
	optValsPtr = optVals;
	
	varPtrEnd = combinedFunc->variables;
	isZeroFunc = TRUE;

	/* 7.	Iterate over the entries of the new probability table and update them. */

	while (1) {	
		start_loop:

		/* 7.1.	Initialize the value of each entry in the new probability table to MIN_DBL. */

		maxVal = MIN_DBL;
		optVarVal = 0;

		/* 7.2.	Iterate over all of the eliminated variable's values. */

		elimVar->currValue = 0;

		/* Handles the case where the eliminated variable has only 2 possible values. */

		if (elimVar->numValues == 2) {

			/* VALUE 0 */
			currVal = 1.0;

			/* 7.2.1.	Iterate over all the functions. */

			for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {

				/* 7.2.1.1.	Get the function's value. */

				funcsVal = getFuncValue((*funcPtr), elimVar->currValue);

				/* 7.2.1.2.	If the value is zero, we can continue to the next variable value. */

				if (funcsVal == 0.0) {
					currVal = 0.0;
					break;
				}

				/* 7.2.1.3.	Otherwise, multiply the value for this variable value, by the 
							function's value. */

				currVal *= funcsVal;
			}
			
			/* 7.2.2.	Update the maximum value and the var value which produced it. */

			maxVal = currVal;

			/* VALUE 1 */
			currVal = 1.0;

			elimVar->currValue = 1;

			/* 7.2.3.	Iterate over all the functions. */

			for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {

				/* 7.2.3.1.	Get the function's value. */

				funcsVal = getFuncValue((*funcPtr), elimVar->currValue);

				/* 7.2.3.2.	If the value is zero, we can continue to the next entry in the table. */

				if (funcsVal == 0.0) {
					goto End;
				}

				/* 7.2.3.3.	Otherwise, multiply the value for this variable value, by the 
							function's value. */

				currVal *= funcsVal;
			}
			
			/* 7.2.4.	If the value for this variable value is greater than the maximum
						value found up to now, update the maximum value and the variable
						value which produced it. */

			if (currVal > maxVal) {
				maxVal = currVal;
				optVarVal = 1;
			}
					
			End:;
		}
		else {	/* Handles the regular case where the eliminated variable has any number of
			       possible values. */
			for (; elimVar->currValue < elimVar->numValues;	elimVar->currValue++) {
				
				currVal = 1.0;

				/* 7.2.1.	Iterate over all the functions. */

				for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {

					/* 7.2.1.1.	Get the function's value. */

					funcsVal = getFuncValue((*funcPtr), elimVar->currValue);
					
					/* 7.2.1.2.	If the value is zero, we can continue to the next variable value. */

					if (funcsVal == 0.0) {
						currVal = 0.0;
						break;
					}

					/* 7.2.1.3.	Otherwise, multiply the value for this variable value, by the 
								function's value. */

					currVal *= funcsVal;
				}
				
				/* 7.2.2.	If the value for this variable value is greater than the maximum
							value found up to now, update the maximum value and the variable
							value which produced it. */

				if (currVal > maxVal) {
					maxVal = currVal;
					optVarVal = elimVar->currValue;
				}
			}
		}

		/* 7.3.	Write the maximum value found into the new probability table, and write the 
				value of the eliminated var which produced it into the optimizing values array. */

		*probabListPtr = maxVal;
		*optValsPtr = optVarVal;

		if (maxVal != 0)
			isZeroFunc = FALSE;

		/* 7.4.	Advance the pointers to the next entries in the tables. */

		probabListPtr++;
		optValsPtr++;

		/* 7.5.	Update the variables' currValues (to match the next entrance in the 
				joint probability table) and update the baseInd of the functions which
				are being joined. */

		varPtr = combinedFunc->variables + combinedFunc->numVariables - 1;
		varInd = combinedFunc->numVariables - 1;

		while (varPtr >= varPtrEnd) {
			var = *varPtr;

			if (var->currValue < (var->numValues - 1)) {
				var->currValue++;

				for (funcPtr = elimVar->functions, funcPtrEnd = funcPtr + elimVar->currNumFunctions,
						funcInd = 0; funcPtr < funcPtrEnd; funcPtr++, 
						funcInd += combinedFunc->numVariables) {

					(*funcPtr)->baseInd += incDecArr[funcInd + varInd].inc;
				}

				goto start_loop;
			}
			else {
				var->currValue = 0;

				for (funcPtr = elimVar->functions, funcPtrEnd = funcPtr + elimVar->currNumFunctions,
						funcInd = 0; funcPtr < funcPtrEnd; funcPtr++, 
						funcInd += combinedFunc->numVariables) {

					(*funcPtr)->baseInd += incDecArr[funcInd + varInd].dec;
				}
			}

			varPtr--;
			varInd--;
		}

		break;
	} 

	/* 8.	Free the memory that was allocated dynamically. */

	myFree(incDecArr);

	return isZeroFunc;
}


/******************************************************************************
*
*       Function        multiplyAllOneVarFuncsMax
*       Parameters		elimVar (the variable that needs to be eliminated), 
*						logMultFactor (the sum of the logMultFactors of all the 
*						functions that take part in the multiplication),
*						optVals (an array that would store the value of 'elimVar'
*						which optimized the combined function for each possible 
*						assignment to the rest of the variables in the function).
*       Output          none
*       Description		performs a multiplication of the probability functions
*						that elimVar appears in, returns the log of the maximum
*						entry.
*       Comments		none
*
******************************************************************************/
Probability multiplyAllOneVarFuncsMax(variable *elimVar, double logMultFactor,
									  int *optVals)
{
	Probability		currVal, maxVal;
	function		**funcPtr, **funcPtrEnd;
	int				val, optVal;


	funcPtrEnd = elimVar->functions + elimVar->currNumFunctions;

	/* 1.	Initialize the maximum value to MIN_DBL. */
	
	maxVal = MIN_DBL;
	optVal = 0;

	/* 2.	Calculate the probability for each value of the eliminated variable
			and find the maximum value. */

	for (val = 0; val < elimVar->numValues; val++) {
		
		currVal = 1.0;

		for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {

			if ((*funcPtr)->probabList[val] == 0.0) {
				currVal = 0.0;
				break;
			}

			currVal *= (*funcPtr)->probabList[val];
		}

		/* 3.1.	If the current value is larger than the maximum found up to now, store
				it and store the value of the eliminate variable which produced it. */

		if (currVal > maxVal) {
			maxVal = currVal;
			optVal = val;
		}
	}

	/* 4.	Update the optimizingVals list with the value of the eliminated 
			variable which produced the maximum probability. */

	*optVals = optVal;

	/* 5.	Return the result of the multiplication. */

	return (log(maxVal) + logMultFactor);
}




/******************************************************************************
*
*       Function        traceMPEProduct
*       Parameters		optValsProd (the optimizingVals product which holds all
*						the optimizingVals functions).
*       Output          none
*       Description		traces the most probable explanation to the data (i.e,
*						the assignment to the variables of the network which
*						has the highest probability.
*       Comments		none
*
******************************************************************************/
void traceMPEProduct(argMaxData head)
{
	for (;head != NULL; head = head->next) {
		retrieveArgMaxData(head);
	}
}




/******************************************************************************
*
*       Function        printMPEHaplotypes
*       Parameters		none
*       Output          none
*       Description		Prints the most probable assignment for the haplotypes
*						into a file.
*       Comments		Used for autosomal analysis.
*
******************************************************************************/
void printMPEHaplotypes()
{
	int						pedNum, *mapToData, numRecombTot, numNF,
							numIRecombHaplo[5], i;
	double					numRecombExp;
	pedigree				*pedPtr;
	person					*personPtr, *personPtrEnd;
	FILE					*haplotypeFile;
	thetaCompTheta			*thetaPtr, *thetaPtrEnd;
	boolean					*isZeroLikePed;
	Probability				lnLikeNoGeneFreq;


	numRecombTot = 0;
	numNF = 0;
	mapToData = loci.mapToData;

	for (i = 0; i < 5; i++) {
		numIRecombHaplo[i] = 0;
	}

	lnLikeNoGeneFreq = totalRes.currRes->lnLike;

	/* 1.	Open the file for writing. */

	haplotypeFile = fopen(HAPLOTYPE_FILE_NAME, "w");
	fprintf(haplotypeFile, "Superlink v1.5 (Original Citation: Fishelson M. and Geiger D. Bioinformatics,  2002:18, S189-S198).\n\n");
	fprintf(haplotypeFile, "Input files: %s, %s.\n", pedfileName, datafileName);

	/* 2.	Loop over all pedigrees. For each pedigree: */

	for (pedPtr = peds.head, pedNum = 1, isZeroLikePed = totalRes.isZeroLikeArr; 
			pedPtr != NULL; pedPtr = pedPtr->nextPed, pedNum++, isZeroLikePed++) {

		/* 2.1.	If the likelihood of this pedigree is zero, continue to the next
				pedigree. */

		if (*isZeroLikePed) {
			continue;
		}

		/* 2.2.	Print titles. */

		fprintf(haplotypeFile, "The results for pedigree no. %d.\n\n", pedNum);
		fprintf(haplotypeFile, "PERSON\tFATHER\tMOTHER\tSEX\tSTATUS\tMARKERS\n");
		fprintf(haplotypeFile, "=============================================\n");

		/* 2.3.	Zero the auxiliary flags of the people in the pedigree. */

		initFlags(pedPtr);	

		/* 2.4.	Traverse the pedigree to retrieve the values of the nodes of the
				bayesian network which represents this pedigree. */

		traverseDownHaplo(pedPtr->founder, updateChLociHaploAutosomal, 
						  updateFounderHaploAutosomal);
 
		/* 2.5.	Loop over all people in the pedigree. For each person: */

		for (personPtr = pedPtr->people, personPtrEnd = personPtr + pedPtr->numPeople;
				personPtr < personPtrEnd; personPtr++) {

			/* 2.5.1.	Print the paternal haplotype. */

			printHaplotype(personPtr, 'P', personPtr->locusVar[0], haplotypeFile);

			/* 2.5.2.	Print the recombination pattern for both haplotypes, if this is a
						non-founder. Otherwise, print a line between the two haplotypes. */

			if (personPtr->pa != NULL) {
				numNF++;
				numRecombTot += printRecomb(personPtr, haplotypeFile, numIRecombHaplo);
			}
			else {
				printRecombFounder(personPtr, haplotypeFile);
			}

			/* 2.5.3.	Print the maternal haplotype. */

			printHaplotype(personPtr, 'M', personPtr->locusVar[1], haplotypeFile);
			fprintf(haplotypeFile, "\n\n");
		}
	}

	/* 3.	Calculate the expected number of recombinations. */

	numRecombExp = 0.0;

	for (thetaPtr = recombData.maleThetas, thetaPtrEnd = thetaPtr + numMarkers - 1;
			thetaPtr < thetaPtrEnd; thetaPtr++) {
		
		numRecombExp += (*thetaPtr)[0] * numNF;
	}

	for (thetaPtr = recombData.femaleThetas, thetaPtrEnd = thetaPtr + numMarkers - 1;
			thetaPtr < thetaPtrEnd; thetaPtr++) {
		
		numRecombExp += (*thetaPtr)[0] * numNF;
	}

	/* 4.	Print statictics on the number of recombinations. */

	fprintf(haplotypeFile, "RECOMBINATIONS	HAPLOTYPES WITH r RECOMB:\n");
	fprintf(haplotypeFile, "EXPECTED  SEEN	r=0	r=1	r=2	r=3	r>3\n");
	fprintf(haplotypeFile, "---------------------------------------------------\n");
	fprintf(haplotypeFile, "%.2lf	   %d	%d	%d	%d	%d	%d\n\n\n", numRecombExp, 
				numRecombTot,numIRecombHaplo[0], numIRecombHaplo[1], numIRecombHaplo[2],
				numIRecombHaplo[3], numIRecombHaplo[4]);

	/* 5.	Print the ln-likelihood of this assignment into the file. */

	fprintf(haplotypeFile, "The Log-likelihood of this assignment (without gene frequencies) is: %lf.\n", 
				(lnLikeNoGeneFreq / LOG_10));
	fprintf(haplotypeFile, "The Log-likelihood of this assignment (with gene frequencies) is: %lf.\n", 
				(totalRes.currRes->lnLike / LOG_10));
}


/******************************************************************************
*
*       Function        printMPEHaplotypesSexLink
*       Parameters		none
*       Output          none
*       Description		Prints the most probable assignment for the haplotypes
*						into a file.
*       Comments		used for sex-linked analysis.
*
******************************************************************************/
void printMPEHaplotypesSexLink()
{
	int						pedNum, *mapToData, numNF, numIRecombHaplo[5], i,
							numRecombTot;
	double					numRecombExp;
	pedigree				*pedPtr;
	person					*personPtr, *personPtrEnd;
	FILE					*haplotypeFile;
	thetaCompTheta			*thetaPtr, *thetaPtrEnd;
	boolean					*isZeroLikePed;
	Probability				lnLikeNoGeneFreq;


	numRecombTot = 0;
	numNF = 0;
	mapToData = loci.mapToData;

	for (i = 0; i < 5; i++) {
		numIRecombHaplo[i] = 0;
	}

	lnLikeNoGeneFreq = totalRes.currRes->lnLike;

	/* 1.	Open the file for writing. */

	haplotypeFile = fopen(HAPLOTYPE_FILE_NAME, "w");
	fprintf(haplotypeFile, "Superlink v1.5 (Original Citation: Fishelson M. and Geiger D. Bioinformatics,  2002:18, S189-S198).\n\n");
	fprintf(haplotypeFile, "Input files: %s, %s.\n", pedfileName, datafileName);

	for (pedPtr = peds.head, pedNum = 1, isZeroLikePed = totalRes.isZeroLikeArr; 
			pedPtr != NULL; pedPtr = pedPtr->nextPed, pedNum++, isZeroLikePed++) {

		/* 2.1.	If the likelihood of this pedigree is zero, continue to the next
				pedigree. */

		if (*isZeroLikePed) {
			continue;
		}

		/* 2.2.	Print titles. */

		fprintf(haplotypeFile, "The results for pedigree no. %d.\n\n", pedNum);
		fprintf(haplotypeFile, "PERSON\tFATHER\tMOTHER\tSEX\tSTATUS\tMARKERS\n");
		fprintf(haplotypeFile, "=============================================\n");

		/* 2.3.	Zero the auxiliary flags of the people in the pedigree. */

		initFlags(pedPtr);	

		/* 2.4.	Traverse the pedigree to retrieve the values of the nodes of the
				bayesian network which represents this pedigree. */

		traverseDownHaplo(pedPtr->founder, updateChLociHaploSexLinked, 
						  updateFounderHaploSexLinked);

 		/* 2.5.	Loop over all people in the pedigree. For each person: */

		for (personPtr = pedPtr->people,  personPtrEnd = personPtr + pedPtr->numPeople;
				personPtr < personPtrEnd; personPtr++) {

			/* 2.5.1.	If this is a female, print the paternal haplotype. */

			if (personPtr->personSex == Female) {

				/* 2.5.1.2.	Print the paternal haplotype. */

				printHaplotype(personPtr, 'P', personPtr->locusVar[0], haplotypeFile);
			}
			else {	/* This is a male. */

				/* 2.5.1.2.	Print a line for the paternal haplotype. */

				printHaplotypePaternal(personPtr, 'P', haplotypeFile);
			}

			/* 2.5.2.	Print the recombination pattern for the maternal
						haplotype, if this is a non-founder. Otherwise, print 
						a line. */

			if (personPtr->pa != NULL) {
				numNF++;
				numRecombTot += printRecombMaternalHaplo(personPtr, haplotypeFile, 
														 numIRecombHaplo);
			}
			else {
				printRecombFounder(personPtr, haplotypeFile);
			}

			/* 2.5.3.	Print the maternal haplotype. */

			printHaplotype(personPtr, 'M', personPtr->locusVar[1], haplotypeFile);
			fprintf(haplotypeFile, "\n\n");
		}
	}

	/* 3.	Calculate the expected number of recombinations. */

	numRecombExp = 0.0;

	for (thetaPtr = recombData.femaleThetas, thetaPtrEnd = thetaPtr + numMarkers - 1;
			thetaPtr < thetaPtrEnd; thetaPtr++) {
		
		numRecombExp += (*thetaPtr)[0] * numNF;
	}

	/* 4.	Print statictics on the number of recombinations. */

	fprintf(haplotypeFile, "RECOMBINATIONS	HAPLOTYPES WITH r RECOMB:\n");
	fprintf(haplotypeFile, "EXPECTED  SEEN	r=0	r=1	r=2	r=3	r>3\n\n");
	fprintf(haplotypeFile, "%lf		  %d	%d	%d	%d	%d	%d\n", numRecombExp, 
				numRecombTot,numIRecombHaplo[0], numIRecombHaplo[1], numIRecombHaplo[2],
				numIRecombHaplo[3], numIRecombHaplo[4]);

	/* 5.	Print the ln-likelihood of this assignment into the file. */

	fprintf(haplotypeFile, "The Log-likelihood of this assignment (without gene frequencies) is: %lf.\n", 
				(lnLikeNoGeneFreq / LOG_10));
	fprintf(haplotypeFile, "The Log-likelihood of this assignment (with gene frequencies) is: %lf.\n", 
				(totalRes.currRes->lnLike / LOG_10));
}



/******************************************************************************
*
*       Function        traverseUpHaplo
*       Parameters      currPerson (the person that is being traversed currently).
*       Output          none
*       Description		If 'curPerson' has no parents, calls traverseDownHaplo.
*						Otherwise:
*							- if one of currPerson's parents has been visited
*							  already, returns.
*							- else, keeps traversing up to currPerson's father.
*						
*       Comments		The idea is to traverse up until a founder has been 
*						reached. 
*
******************************************************************************/
void traverseUpHaplo(person *currPerson, updateChLociHaplo updateChLociHap,
					 updateFounderHaplo updateFounderHap)
{
	/* 1.	If this isn't a founder. */

	if (currPerson->pa != NULL)	{ 

		/* 1.1.	If one of the parents has been 'visited' already, we'll get to 
				the person in another way. */

		if (((currPerson->pa)->visited == TRUE) || ((currPerson->ma)->visited == TRUE)) 
			return;

		/* 1.2	Otherwise, traverse up to this person's father. */

		traverseUpHaplo(currPerson->pa, updateChLociHap, updateFounderHap);
	}
	else					/* If this person is a founder, traverse down to it */
		traverseDownHaplo(currPerson, updateChLociHap, updateFounderHap);
}



/******************************************************************************
*
*       Function        traverseDownHaplo
*       Parameters      currPerson (the person that is being traversed currently).
*       Output          none
*       Description		updates all of the children of 'currPerson' according to 
*						it and calls traverseDown recursively on each child   
*						which has already been updated by the other parent as well.
*       Comments		none
*
******************************************************************************/
void traverseDownHaplo(person *currPerson, updateChLociHaplo updateChLociHap,
					   updateFounderHaplo updateFounderHap)
{
	person		*ch;			
	sex			Sex;


	ch = currPerson->ch;
	Sex = currPerson->personSex;

	/* 1.	Mark this person as visited. */

	currPerson->visited = TRUE;			

	if (currPerson->pa == NULL) {
		updateFounderHap(currPerson);
		currPerson->updatedParents = TRUE;
	}

	if (Sex == Male) {

		/* 2.	Iterate over the person's children. */

		for (; ch != NULL; ch = ch->NeSiPa) {	

			if ((ch->updatedParents == FALSE) &&
				(currPerson->updatedParents == TRUE && (ch->ma)->updatedParents == TRUE)) {
			
				/* 2.1.	Update the child according to the parent (father). */

				updateChHaplo(updateChLociHap, currPerson, ch);
				updateChHaplo(updateChLociHap, ch->ma, ch);
				ch->updatedParents = TRUE;
			}

			/* 2.2. If the child's mother hasn't been visited yet, traverse up to it
					and then traverse down to the child. 
					Otherwise, if we finished updating the child's mother, traverse
					down to the child. */

			if ((ch->ma)->visited == FALSE)	{
				traverseUpHaplo(ch->ma, updateChLociHap, updateFounderHap);
				traverseDownHaplo(ch, updateChLociHap, updateFounderHap);
			}
			else if ((ch->ma)->done == TRUE)
				traverseDownHaplo(ch, updateChLociHap, updateFounderHap);
		}
	}
	else {			/* (Sex == Female) */

		/* 2.	Iterate over the person's children. */

		for (; ch != NULL; ch = ch->NeSiMa) {

			if ((ch->updatedParents == FALSE) &&
				(currPerson->updatedParents == TRUE && (ch->pa)->updatedParents == TRUE)) {
			
				/* 2.1.	Update the child according to the parent (father). */

				updateChHaplo(updateChLociHap, ch->pa, ch);
				updateChHaplo(updateChLociHap, currPerson, ch);
				ch->updatedParents = TRUE;
			}

			if ((ch->pa)->visited == FALSE)	{ /* we haven't visited the father yet. */
				traverseUpHaplo(ch->pa, updateChLociHap, updateFounderHap);
				traverseDownHaplo(ch, updateChLociHap, updateFounderHap);
			}
			else if ((ch->pa)->done == TRUE)
				traverseDownHaplo(ch, updateChLociHap, updateFounderHap);
		}
	}

	/* 3.	Update that we finished updating this person. */

	currPerson->done = TRUE;
}


/******************************************************************************
*
*       Function        updateFounderHaploAutosomal
*       Parameters      currPerson (the person who needs to be updated).
*       Output          none
*       Description		updates 'currPerson'.
*       Comments		used for autosomal analysis.
*
******************************************************************************/
void updateFounderHaploAutosomal(person *currPerson)
{
	int				locusIndMap, locusIndDs, chosenAllele, maxValInd;
	node			**locusNode;
	locusPerson		*locus;
	Probability		*geneFrequencies;

#if	PERFORM_ALLELE_RECODING
	alleleSet		*alleleGroup;
#endif

	tempRes = 0.0;

	/* 1.	Loop over all marker (ALLELES type) loci. */

	for (locusIndMap = 0; locusIndMap < numMarkers; locusIndMap++) {

		locusIndDs = loci.mapToData[locusIndMap];
		geneFrequencies = (loci.locusArr[locusIndDs]).geneFrequencies;

		/* 1.1.	Update pointers to the data structure which holds the information
				on this locus of the founder. */

		locus = &(currPerson->loci[locusIndDs]);

		/* 1.2.	Update a pointer to the paternal locus node of this founder at
				this locus. */

		locusNode = &(currPerson->locusVar[0][locusIndDs]);

		/* 1.3.	Determine the index of the maximizing value of this node. */

		if (locus->locusData.allelesL.numPaAlleleGroups == 1) {
			maxValInd = 0;
		}
		else {
			maxValInd = (currPerson->locusVar[0][locusIndDs])->currVal;
		}

		#if PERFORM_ALLELE_RECODING

			/* 1.4.	Update a pointer to the maximizing set of ALLELES of this  
					person at the paternal locus node. */

			alleleGroup = 
				&(currPerson->loci[locusIndDs].locusData.allelesL.paAllelsGroups[maxValInd]);

			/* 1.5.	Determine the value of the paternal locus node of 'currPerson' at this 
					locus. Either it was determined by the data. Otherwise, it is determined
					by choosing the allele with the maximum gene frequency. */

			chosenAllele = alleleGroup->alleleArr[0];
		#else
			chosenAllele = locus->locusData.allelesL.paAlleles[maxValInd];
		#endif

		/* 1.6.	Update the value of the paternal locus node. */

		(*locusNode)->currVal = chosenAllele;

		/* 1.7.	Update a pointer to the maternal locus node of this founder at
				this locus. */

		locusNode = &(currPerson->locusVar[1][locusIndDs]);

		/* 1.8.	Determine the index of the maximizing value of this node. */

		if (locus->locusData.allelesL.numMaAlleleGroups == 1) {
			maxValInd = 0;
		}
		else {
			maxValInd = (currPerson->locusVar[1][locusIndDs])->currVal;
		}

		#if PERFORM_ALLELE_RECODING

			/* 1.9.	Update a pointer to the maximizing set of ALLELES of this  
					person at the maternal locus node. */

			alleleGroup = 
				&(currPerson->loci[locusIndDs].locusData.allelesL.maAllelsGroups[maxValInd]);

			/* 1.10.	Determine the value of the maternal locus node of 'currPerson' at this 
						locus. Either it was determined by the data. Otherwise, it is determined
						by choosing the allele with the maximum gene frequency. */

			chosenAllele = alleleGroup->alleleArr[0];
		#else
			chosenAllele = locus->locusData.allelesL.maAlleles[maxValInd];
		#endif

		/* 1.6.	Update the value of the maternal locus node. */

		(*locusNode)->currVal = chosenAllele;
	}
}

/******************************************************************************
*
*       Function        updateFounderHaploSexLinked
*       Parameters      currPerson (the person who needs to be updated).
*       Output          none
*       Description		updates 'currPerson'.
*       Comments		used for sex-linked analysis.
*
******************************************************************************/
void updateFounderHaploSexLinked(person *currPerson)
{
	int				locusIndMap, locusIndDs, chosenAllele, maxValInd;
	node			**locusNode;
	locusPerson		*locus;
	Probability		*geneFrequencies;

#if	PERFORM_ALLELE_RECODING
	alleleSet		*alleleGroup;
#endif

	tempRes = 0.0;

	/* 1.	Loop over all marker (ALLELES type) loci. */

	for (locusIndMap = 0; locusIndMap < numMarkers; locusIndMap++) {

		locusIndDs = loci.mapToData[locusIndMap];
		geneFrequencies = (loci.locusArr[locusIndDs]).geneFrequencies;

		/* 1.1.	Update pointers to the data structure which holds the information
				on this locus of the founder. */

		locus = &(currPerson->loci[locusIndDs]);

		/* 1.2.	If this is a female: */

		if (currPerson->personSex == Female) {

			/* 1.2.1.	Update a pointer to the paternal locus node of this founder
						at this locus. */

			locusNode = &(currPerson->locusVar[0][locusIndDs]);

			/* 1.2.2.	Determine the index of the maximizing value of this node. */

			if (locus->locusData.allelesL.numPaAlleleGroups == 1) {
				maxValInd = 0;
			}
			else {
				maxValInd = (currPerson->locusVar[0][locusIndDs])->currVal;
			}

			#if PERFORM_ALLELE_RECODING

				/* 1.2.3	Update a pointer to the maximizing set of ALLELES of this  
							person at the paternal locus node. */

				alleleGroup = 
					&(currPerson->loci[locusIndDs].locusData.allelesL.paAllelsGroups[maxValInd]);

				/* 1.2.4.	Determine the value of the paternal locus node of 'currPerson' at this 
							locus. Either it was determined by the data. Otherwise, 
							it is determined by choosing the allele with the maximum 
							gene frequency. */

				chosenAllele = alleleGroup->alleleArr[0];
			#else
				chosenAllele = locus->locusData.allelesL.paAlleles[maxValInd];
			#endif

			/* 1.2.5.	Update the value of the paternal locus node. */

			(*locusNode)->currVal = chosenAllele;
		}

		/* 1.3.	Update a pointer to the maternal locus node of this founder at
				this locus. */

		locusNode = &(currPerson->locusVar[1][locusIndDs]);

		/* 1.4.	Determine the index of the maximizing value of this node. */

		if (locus->locusData.allelesL.numMaAlleleGroups == 1) {
			maxValInd = 0;
		}
		else {
			maxValInd = (currPerson->locusVar[1][locusIndDs])->currVal;
		}

		#if PERFORM_ALLELE_RECODING

			/* 1.5.	Update a pointer to the maximizing set of ALLELES of this  
					person at the maternal locus node. */

			alleleGroup = &(locus->locusData.allelesL.maAllelsGroups[maxValInd]);

			/* 1.6.	Determine the value of the maternal locus node of 'currPerson' at this 
						locus. Either it was determined by the data. Otherwise, it is determined
						by choosing the allele with the maximum gene frequency. */

			chosenAllele = alleleGroup->alleleArr[0];
		#else
			chosenAllele = locus->locusData.allelesL.maAlleles[maxValInd];
		#endif

		/* 1.7.	Update the value of the maternal locus node. */

		(*locusNode)->currVal = chosenAllele;
	}
}



/******************************************************************************
*
*       Function        updateChHaplo
*       Parameters      parent (the person whose child needs to be updated
*						according to it), child (the person which needs to be 
*						updated).
*       Output          none
*       Description		updates the child according to the parent.
*       Comments		none
*
******************************************************************************/
void updateChHaplo(updateChLociHaplo updateChLociHap, person *parent, person *child)
{
	int		locusIndMap, locusIndDs;

	/* 1.	Loop over all marker (ALLELES type) loci. */

	for (locusIndMap = 0; locusIndMap < numMarkers; locusIndMap++) {

		locusIndDs = loci.mapToData[locusIndMap];

		/* 1.1.	Update the child according to the parent at this locus. */

		updateChLociHap(parent, child, locusIndDs);
	}
}

/******************************************************************************
*
*       Function        updateChLociHaploAutosomal
*       Parameters      parent (the person whose child needs to be updated
*						according to it), child (the person which needs to be 
*						updated), locusInd (the index of the locus being updated).
*       Output          none
*       Description		updates the child according to the parent, at locus 
*						number 'locusInd'.
*       Comments		used for autosomal analysis.
*
******************************************************************************/
void updateChLociHaploAutosomal(person* parent, person* child, int locusInd)
{
	int			selecInd, selecVal;

	/* 1.	Determine the value of the relevant selector node and locus node,
			according to the parent's sex. */

	if (parent->personSex == Male) {

		/* 1.1.	Determine the index of the maximizing value determined for the 
				paternal selector of the child at locus 'locusInd'. */

		selecInd = child->selectorVar[0][locusInd]->currVal;

		/* 1.2.	Determine the maximizing values determined for this selector. */

		selecVal = child->selectorVar[0][locusInd]->currVal = 
			child->loci[locusInd].locusData.allelesL.paSelectorVals[selecInd];

		/* 1.3.	Determine the value of the paternal locus node of the child. */

		child->locusVar[0][locusInd]->currVal = 
			parent->locusVar[1-selecVal][locusInd]->currVal;
	}
	else {
		/* 1.1.	Determine the index of the maximizing value determined for the 
				maternal selector of the child at locus 'locusInd'. */

		selecInd = child->selectorVar[1][locusInd]->currVal;
	
		/* 1.2.	Determine the maximizing values determined for this selector. */

		selecVal = child->selectorVar[1][locusInd]->currVal = 
			child->loci[locusInd].locusData.allelesL.maSelectorVals[selecInd];

		/* 1.3.	Determine the value of the maternal locus node of the child. */

		child->locusVar[1][locusInd]->currVal = 
			parent->locusVar[1-selecVal][locusInd]->currVal;
	}
}


/******************************************************************************
*
*       Function        updateChLociHaploSexLinked
*       Parameters      parent (the person whose child needs to be updated
*						according to it), child (the person which needs to be 
*						updated), locusInd (the index of the locus being updated).
*       Output          none
*       Description		updates the child according to the parent, at locus 
*						number 'locusInd'.
*       Comments		used for sex-linked analysis.
*
******************************************************************************/
void updateChLociHaploSexLinked(person* parent, person* child, int locusInd)
{
	int			selecInd, selecVal;


	/* 1.	Determine the value of the relevant selector node and locus node,
			according to the parent's sex. */

	if (parent->personSex == Male) {

		/* 1.1.	If the child is a female: */

		if (child->personSex == Female) {

			/* 1.1.3.	Determine the value of the paternal locus node of the child. */

			child->locusVar[0][locusInd]->currVal = 
				parent->locusVar[1][locusInd]->currVal;
		}
	}
	else {
		/* 1.1.	Determine the index of the maximizing value determined  
				for the maternal selector of the child at locus 'locusInd'. */

		selecInd = child->selectorVar[1][locusInd]->currVal;

		/* 1.2.	Determine the  maximizing values determined for this selector. */

		selecVal = child->selectorVar[1][locusInd]->currVal = 
					child->loci[locusInd].locusData.allelesL.maSelectorVals[selecInd];

		/* 1.3.	Determine the value of the maternal locus node of the child. */

		child->locusVar[1][locusInd]->currVal = 
			parent->locusVar[1-selecVal][locusInd]->currVal;
	}
}



/******************************************************************************
*
*       Function        printHaplotype
*       Parameters		personPtr (pointer to the person whose haplotype is
*						printed), haploSex (whether the haplotype is paternal
*						or maternal), locusNodes (the person's nodes of this
*						haplotype), haplotypeFile (pointer to the output file 
*						to which the haplotypes are printed).
*       Output          none
*       Description		Prints the assignment to one of the haplotypes of 
*						'personPtr' into the file 'haplotypeFile'.
*       Comments		none.
*
******************************************************************************/
void printHaplotype(person *personPtr, char haploSex, node **locusNodes, 
					FILE *haplotypeFile)
{
	int			locusIndMap, locusIndDs;


	/* 1.	Print general information on this person. */

	if (personPtr->pa != NULL) {
		fprintf(haplotypeFile, "%c %d\t%d\t%d\t", haploSex, personPtr->pNum, 
					(personPtr->pa)->pNum, (personPtr->ma)->pNum);
	}
	else {
		fprintf(haplotypeFile, "%c %d\t--\t--\t", haploSex, personPtr->pNum);
	}

	if (personPtr->personSex == Male) {
		fprintf(haplotypeFile, "M\t");
	}
	else {
		fprintf(haplotypeFile, "F\t");
	}

	/* 2.	Print the affection status of the person. */

	if (numMarkers < loci.numLoci){
		fprintf(haplotypeFile,"%d\t",personPtr->loci[loci.mapToData[numMarkers]].locusData.affectionL.affectionS);
	}
	else { 
		fprintf(haplotypeFile, "0\t");
	}

	/* 3.	Loop over all marker (ALLELES type) loci. */

	for (locusIndMap = 0; locusIndMap < numMarkers; locusIndMap++) {

		locusIndDs = loci.mapToData[locusIndMap];

		/* 3.1.	Print the most probable haplotype of this person (as been determined). */

		fprintf(haplotypeFile, "%d\t", locusNodes[locusIndDs]->currVal);
	}

	fprintf(haplotypeFile,"\n");
}

/******************************************************************************
*
*       Function        printHaplotypePaternal
*       Parameters		personPtr (pointer to the person whose haplotype is
*						printed), haploSex (whether the haplotype is paternal
*						or maternal), haplotypeFile (pointer to the output file 
*						to which the haplotypes are printed).
*       Output          none
*       Description		Prints a line of 'X' for the paternal haplotype of 
*						'personPtr' into the file 'haplotypeFile'.
*       Comments		used for sex-linked analysis, for males.
*
******************************************************************************/
void printHaplotypePaternal(person *personPtr, char haploSex, FILE *haplotypeFile)
{
	int			locusIndMap;


	/* 1.	Print general information on this person. */

	if (personPtr->pa != NULL) {
		fprintf(haplotypeFile, "%c %d\t%d\t%d\t", haploSex, personPtr->pNum, 
					(personPtr->pa)->pNum, (personPtr->ma)->pNum);
	}
	else {
		fprintf(haplotypeFile, "%c %d\t--\t--\t", haploSex, personPtr->pNum);
	}

	if (personPtr->personSex == Male) {
		fprintf(haplotypeFile, "M\t");
	}
	else {
		fprintf(haplotypeFile, "F\t");
	}

	/* 2.	Print the affection status of the person. */

	if (numMarkers < loci.numLoci){
		fprintf(haplotypeFile,"%d\t",personPtr->loci[loci.mapToData[numMarkers]].locusData.affectionL.affectionS);
	}
	else { 
		fprintf(haplotypeFile, "0\t");
	}


	/* 3.	Loop over all marker (ALLELES type) loci. */

	for (locusIndMap = 0; locusIndMap < numMarkers; locusIndMap++) {

		/* 3.1.	Print 'X'. */

		fprintf(haplotypeFile, "X\t");
	}

	fprintf(haplotypeFile,"\n");
}


/******************************************************************************
*
*       Function        printRecomb
*       Parameters		personPtr (pointer to the person whose haplotype is
*						printed), haplotypeFile (pointer to the output file 
*						to which the haplotypes are printed).
*       Output          none
*       Description		Prints the inheritance pattern for both haplotypes of 
*						'personPtr' into the file 'haplotypeFile'.
*       Comments		returns the number of recombination events found.
*
******************************************************************************/
int printRecomb(person *personPtr, FILE *haplotypeFile, int *numIRecombHaplo)
{
	int			locusIndMap, locusIndDs, prevMaternalSelec, prevPaternalSelec,
				currMaternalSelec, currPaternalSelec, numRecombP, numRecombM;
	char		inheritPattern;
	node		**paternalSelectors, **maternalSelectors;
	boolean		recombM, recombP;

	numRecombP = numRecombM = 0;

	paternalSelectors = personPtr->selectorVar[0];
	maternalSelectors = personPtr->selectorVar[1];

	fprintf(haplotypeFile, "+ \t\t\t\t\t");

	locusIndDs = loci.mapToData[0];

	prevPaternalSelec = paternalSelectors[locusIndDs]->currVal;
	prevMaternalSelec = maternalSelectors[locusIndDs]->currVal;

	/* 1.	Loop over all marker (ALLELES type) loci. */

	for (locusIndMap = 1; locusIndMap < numMarkers; locusIndMap++) {

		locusIndDs = loci.mapToData[locusIndMap];

		currPaternalSelec = paternalSelectors[locusIndDs]->currVal;
		currMaternalSelec = maternalSelectors[locusIndDs]->currVal;

		/* 1.1.	Update the flag 'recombP' to TRUE if there was a recombination in the
				paternal haplotype in the current interval. Otherwise, update it to FALSE. */

		if (prevPaternalSelec != currPaternalSelec) {
			recombP = TRUE;
			numRecombP++;
		}
		else {
			recombP = FALSE;
		}

		/* 1.2.	Update the flag 'recombM' to TRUE if there was a recombination in the
				maternal haplotype in the current interval. Otherwise, update it to FALSE. */

		if (prevMaternalSelec != currMaternalSelec) {
			recombM = TRUE;
			numRecombM++;
		}
		else {
			recombM = FALSE;
		}

		/* 1.3.	Update the character that separates between the two haplotypes according to 
				whether there was no recombination, recombination in the paternal haplotype,  
				recombination in the maternal haplotype, or recombination in both. */

		if (recombP) {
			if (recombM) {
				inheritPattern = '|';
			}
			else {
				inheritPattern = '^';
			}
		}
		else {
			if (recombM) {
				inheritPattern = 'v';
			}
			else {
				inheritPattern = '-';
			}
		}
		fprintf(haplotypeFile,"----%c---", inheritPattern);

		prevPaternalSelec = currPaternalSelec;
		prevMaternalSelec = currMaternalSelec;
	}

	fprintf(haplotypeFile,"-\n");

	if (numRecombP < 4) {
		numIRecombHaplo[numRecombP]++;
	}
	else {
		numIRecombHaplo[4]++;
	}
	if (numRecombM < 4) {
		numIRecombHaplo[numRecombM]++;
	}
	else {
		numIRecombHaplo[4]++;
	}

	return (numRecombP + numRecombM);
}


/******************************************************************************
*
*       Function        printRecombMaternalHaplo
*       Parameters		personPtr (pointer to the person whose haplotype is
*						printed), haplotypeFile (pointer to the output file 
*						to which the haplotypes are printed).
*       Output          none
*       Description		Prints the inheritance pattern for the maternal 
*						haplotype of 'personPtr' into the file 'haplotypeFile'.
*       Comments		used for males in sex-linked analysis.
*
******************************************************************************/
int printRecombMaternalHaplo(person *personPtr, FILE *haplotypeFile, int *numIRecombHaplo)
{
	int			locusIndMap, locusIndDs, prevMaternalSelec, currMaternalSelec,
				numRecombM;
	char		inheritPattern;
	node		**maternalSelectors;
	boolean		recombM;

	numRecombM = 0;

	maternalSelectors = personPtr->selectorVar[1];

	fprintf(haplotypeFile, "+ \t\t\t\t\t");

	locusIndDs = loci.mapToData[0];

	prevMaternalSelec = maternalSelectors[locusIndDs]->currVal;

	/* 1.	Loop over all marker (ALLELES type) loci. */

	for (locusIndMap = 1; locusIndMap < numMarkers; locusIndMap++) {

		locusIndDs = loci.mapToData[locusIndMap];

		currMaternalSelec = maternalSelectors[locusIndDs]->currVal;

		/* 1.1.	Update the flag 'recombM' to TRUE if there was a recombination in the
				maternal haplotype in the current interval. Otherwise, update it to FALSE. */

		if (prevMaternalSelec != currMaternalSelec) {
			recombM = TRUE;
			numRecombM++;
		}
		else {
			recombM = FALSE;
		}

		/* 1.3.	Update the character that separates between the two haplotypes according to 
				whether there was no recombination, or recombination in the maternal haplotype. */

		if (recombM) {
			inheritPattern = 'v';
		}
		else {
			inheritPattern = '-';
		}

		fprintf(haplotypeFile,"----%c---", inheritPattern);

		prevMaternalSelec = currMaternalSelec;
	}

	fprintf(haplotypeFile,"-\n");

	if (numRecombM < 4) {
		numIRecombHaplo[numRecombM]++;
	}
	else {
		numIRecombHaplo[4]++;
	}

	return numRecombM;
}


/******************************************************************************
*
*       Function        printRecombFounder
*       Parameters		personPtr (pointer to the person whose haplotype is
*						printed), haplotypeFile (pointer to the output file 
*						to which the haplotypes are printed).
*       Output          none
*       Description		Prints a line between the two haplotypes of 
*						'personPtr' into the file 'haplotypeFile'.
*       Comments		none
*
******************************************************************************/
void printRecombFounder(person *personPtr, FILE *haplotypeFile)
{
	int			locusIndMap;

	fprintf(haplotypeFile, "+ \t\t\t\t\t");

	/* 1.1.2.	Loop over all marker (ALLELES type) loci. */

	for (locusIndMap = 1; locusIndMap < numMarkers; locusIndMap++) {
		fprintf(haplotypeFile,"--------");
	}

	fprintf(haplotypeFile,"-\n");
}


/******************************************************************************
*
*       Function        insertClusterValues
*       Parameters		personPtr (pointer to the person whose haplotype is
*						printed), haploSex (whether the haplotype is paternal
*						or maternal), locusNodes (the person's nodes of this
*						haplotype), haplotypeFile (pointer to the output file 
*						to which the haplotypes are printed).
*       Output          none
*       Description		Prints the assignment to one of the haplotypes of 
*						'personPtr' into the file 'haplotypeFile'.
*       Comments		Author: Anna Tzemach.
*
******************************************************************************/
void insertClusterValues(person *personPtr,PedigreeDataStruct *origPedigree, node **locusNodes1,node **locusNodes2 )
{
	int			locusIndMap, locusIndDs;
	int base;
	PersonDataStruct *origPersonPtr;
	char mapVal1[500];
	char mapVal2[500];
	/* 3.	Loop over all marker (ALLELES type) loci. */
	origPersonPtr = origPedigree->people + (personPtr->pNum - 1);

	myRealloc((origPersonPtr->clustersAlleles), ((ProgramData.numberOfClusters+1)*sizeof(AlleleDataStruct)));

	strcpy (mapVal1,"original val 1: ");
	strcpy (mapVal2,"original val 2: ");

	origPersonPtr->clustersAlleles[ProgramData.numberOfClusters].allele1 = 0;
	origPersonPtr->clustersAlleles[ProgramData.numberOfClusters].allele2 = 0;
	for (locusIndMap = 0; locusIndMap < numMarkers; locusIndMap++) {
		locusIndDs = loci.mapToData[locusIndMap];
		base = pow(loci.locusArr[ locusIndDs].numAlleles,locusIndDs);
		if (personPtr->loci[locusIndDs].unknown){
			origPersonPtr->clustersAlleles[ProgramData.numberOfClusters].allele1 = -1;
			origPersonPtr->clustersAlleles[ProgramData.numberOfClusters].allele2 = -1;
			break;
		}

		sprintf(mapVal1,"%s %d ",mapVal1,locusNodes1[locusIndDs]->currVal);
		sprintf(mapVal2,"%s %d ",mapVal2,locusNodes2[locusIndDs]->currVal);

		origPersonPtr->clustersAlleles[ProgramData.numberOfClusters].allele1 += 
			((locusNodes1[locusIndDs]->currVal-1)*base);
		origPersonPtr->clustersAlleles[ProgramData.numberOfClusters].allele2 += 
			((locusNodes2[locusIndDs]->currVal - 1)*base); 


		/* 3.1.	Print the most probable haplotype of this person (as been determined). */

		//fprintf(haplotypeFile, "%d\t", locusNodes[locusIndDs]->currVal);

	}
	origPersonPtr->clustersAlleles[ProgramData.numberOfClusters].allele1++;
	origPersonPtr->clustersAlleles[ProgramData.numberOfClusters].allele2++;

	if  ((strcmp (mapVal1,"original val 1: ") != 0) && (strcmp (mapVal2,"original val 2: ") != 0) &&
		(origPersonPtr->clustersAlleles[ProgramData.numberOfClusters].allele1 > 0) && (origPersonPtr->clustersAlleles[ProgramData.numberOfClusters].allele2 > 0)){
		printf("%s result allele: %d\n",mapVal1,origPersonPtr->clustersAlleles[ProgramData.numberOfClusters].allele1);
		printf("%s result allele: %d\n",mapVal2,origPersonPtr->clustersAlleles[ProgramData.numberOfClusters].allele2);
	}
}
/******************************************************************************
*
*       Function        printMPEHaplotypes
*       Parameters		none
*       Output          none
*       Description		Prints the most probable assignment for the haplotypes
*						into a file.
*       Comments		Used for autosomal cluster analysis.
*
******************************************************************************/
Bool clusterMPEHaplotypes(PedigreeDataStruct *origPedigree)
{
	int						pedNum, *mapToData, numRecombTot, numNF;

	pedigree				*pedPtr;
	person					*personPtr, *personPtrEnd;

	boolean					*isZeroLikePed;



	numRecombTot = 0;
	numNF = 0;
	mapToData = loci.mapToData;




	/* 2.	Loop over all pedigrees. For each pedigree: */

	for (pedPtr = peds.head, pedNum = 1, isZeroLikePed = totalRes.isZeroLikeArr; 
		pedPtr != NULL; pedPtr = pedPtr->nextPed, pedNum++, isZeroLikePed++) {

			/* 2.1.	If the likelihood of this pedigree is zero, continue to the next
			pedigree. */

			if (*isZeroLikePed) {
				return FALSE;
			}




			/* 2.3.	Zero the auxiliary flags of the people in the pedigree. */

			initFlags(pedPtr);	

			/* 2.4.	Traverse the pedigree to retrieve the values of the nodes of the
			bayesian network which represents this pedigree. */

			traverseDownHaplo(pedPtr->founder, updateChLociHaploAutosomal, 
				updateFounderHaploAutosomal);

			/* 2.5.	Loop over all people in the pedigree. For each person: */

			for (personPtr = pedPtr->people, personPtrEnd = personPtr + pedPtr->numPeople;
				personPtr < personPtrEnd; personPtr++) {

					/* 2.5.1.	Print the paternal haplotype. */

					//printHaplotype(personPtr, 'P', personPtr->locusVar[0], haplotypeFile);
					if (!loci.sexLink || personPtr->personSex == Female)
						insertClusterValues(personPtr,origPedigree, personPtr->locusVar[0],personPtr->locusVar[1]);
					else
						insertClusterValues(personPtr,origPedigree, personPtr->locusVar[1],personPtr->locusVar[1]);

				}
		}
		return TRUE;
}


