/*********************************************************************************/
/*  Inference.c												     				 */
/*  Author: Ma'ayan Fishelson													 */
/*********************************************************************************/

#include "sl_Inference.h"

/******************************************************************************
*
*       function       envEvaluateLike 
*       parameters     x (the positions, in cM, of the trait loci at which 
*					   the likelihood needs to be evaluated. The position is
*					   specified as a distance from the leftmost marker in the
*					   scan interval in which the trait locus is iterated), 
*					   funcData (a struct which holds additional information
*					   needed for the computations).
*       output         none
*       description    A wrapper function for the function which evaluates the 
*					   likelihood of the data at positions x of the trait loci.
*       comments	   none
*
******************************************************************************/
double envEvaluateLike(double *x, void* funcData)
{
	evaluateLikeData	*evalData;

	evalData = (evaluateLikeData*)funcData;

	return evaluateLike(x, evalData->numTraitLoci, evalData->retFunc, 
						evalData->printResults);
}	


/******************************************************************************
*
*       function       evaluateLike 
*       parameters     x (the positions, in cM, of the trait loci at which 
*					   the likelihood needs to be evaluated. The position is
*					   specified as a distance from the leftmost marker in the
*					   scan interval in which the trait locus is iterated. A 
*					   negative value is treated as a distance to the left of
*					   the marker, a positive value is treated as a distance to 
*					   the right of the marker), numTraitLoci (the number of 
*					   trait loci that need to be positioned), retFunc (a 
*					   function that returns the requested value, either 
*					   ln-likelihood or lod-score), printResults (a flag that
*					   indicates whether to print the results of this likelihood
*					   evaluation or not).
*       output         none
*       description    Evaluates the likelihood of the data at positions x of 
*					   the trait loci.
*       comments	   none
*
******************************************************************************/
double evaluateLike(double *x, int numTraitLoci, retValFuncType retFunc, 
					boolean printResults)
{
	double				*xPtr;
	int					*oldLocPos, *oldLocPosPtr, *newLocPos, *newLocPosPtr,
						*mapToData, numThetas;
	boolean				areValsElim, isChangedNet;
	scanIntervalData	*scanData, *scanDataEnd;
	elimOrder			elimOrderArrs;
	Probability			*newThetas, *newThetasPtr;
	thetaCompTheta		*thetaArr;


	/* 1.	Allocate memory for arrays that would store the old and new positions 
			of the trait loci and the new theta values between the trait loci and their 
			adjacent markers. */

	myMalloc(oldLocPos, sizeof(int) * numTraitLoci);
	myMalloc(newLocPos, sizeof(int) * numTraitLoci);
	myMalloc(newThetas, sizeof(Probability) * numTraitLoci);

	/* 2.	Iterate over the trait loci and update arrays with their new 
			and old positions.  */

	for (scanData = progData.scanData, scanDataEnd = scanData + numTraitLoci,
			xPtr = x, newLocPosPtr = newLocPos, newThetasPtr = newThetas,
			oldLocPosPtr = oldLocPos; scanData < scanDataEnd; scanData++, xPtr++, 
			newLocPosPtr++, newThetasPtr++, oldLocPosPtr++) {

		findTraitLocPos(scanData, *xPtr, newLocPosPtr, newThetasPtr);
		(*oldLocPosPtr) = scanData->currPosTraitLoc;
	}

	/* 3.	Allocate memory for arrays that would store the current positions of 
			the loci, and update them. */

	numThetas = loci.numLoci - 1;
	myMalloc(thetaArr, sizeof(thetaCompTheta) * numThetas);
	memcpy(thetaArr, recombData.maleThetas, sizeof(thetaCompTheta) * numThetas);
	myMalloc(mapToData, sizeof(int) * loci.numLoci);
	memcpy(mapToData, loci.mapToData, sizeof(int) * loci.numLoci);

	/* 4.	Update the positions of the trait loci in the data structures according 
			to their new requested positions. */

	moveTraitLoc(numTraitLoci, oldLocPos, newLocPos, newThetas);

	/* 5.	Allocate memory for a result node to store the results of the
			current likelihood evaluation, update the 'currRes'
			field to point to it, and update the distance of each trait locus
			from the leftmost marker in its scan interval. */


	if (numTraitLoci == 1) {
		initResNode(x[0]);
	}
	else {
		initResNode2Loc(x[0], x[1]);
	}

	/* 6.	Update the relevant flags and calculate the likelihood. */

	areValsElim = TRUE;
	isChangedNet = TRUE;
	elimOrderArrs = NULL;

	performOneIter(&elimOrderArrs, &areValsElim, &isChangedNet, TRUE);

	/* 7.	If requested, print the results of the likelihood evaluation. */

	if (printResults) {
		printResults_OneIter(loci.numLoci - 1, peds.numPeds, (boolean)(retFunc == getLodScore));
	}

	/* 8.	Restore the positions of the loci to their positions before the call
			to this function. */

	memcpy(recombData.maleThetas, thetaArr, sizeof(thetaCompTheta) * numThetas);
	memcpy(recombData.femaleThetas, thetaArr, sizeof(thetaCompTheta) * numThetas);
	memcpy(loci.mapToData, mapToData, sizeof(int) * loci.numLoci);

	for (scanData = progData.scanData, scanDataEnd = scanData + numTraitLoci,
			oldLocPosPtr = oldLocPos; scanData < scanDataEnd; scanData++, oldLocPosPtr++) { 

		scanData->currPosTraitLoc = (*oldLocPosPtr);
	}

	/* 9.	Free the memory allocated dynamically. */

	myFree(oldLocPos);
	myFree(newLocPos);
	myFree(newThetas);
	freeEliminationOrderStructs(&elimOrderArrs, peds.numPeds);
	myFree(thetaArr);
	myFree(mapToData);

	/* 10.	Return the lod-score/ln-likelihood calculated, according to the 
			retFunction. */

	return (*retFunc)(totalRes.currRes);
}


/******************************************************************************
*
*       function       updateTraitLocPos 
*       parameters     scanData (the data on the scan interval in which the
*					   trait locus is being iterated), x (the position, in cM, 
*					   of the trait locus at which the likelihood needs to be
*					   evaluated. The position is specified as a distance from 
*					   the leftmost marker in the scan interval in which the 
*					   trait locus is iterated), newLocPos (is updated with the
*					   new position of the trait locus among the markers),
*					   newTheta (the new theta value between the trait locus 
*					   and its adjacent marker).
*       output         none
*       description    Updates the position of the trait locus to a distance
*					   of x from the leftmost marker in the scan interval.
*       comments	   none 
*
*
******************************************************************************/
void findTraitLocPos(scanIntervalData *scanData, double x, int *newLocPos, 
					 Probability *newTheta)
{
	double			*distFromMarker1Ptr, dist;
	int				currMarker, numMarkers;


	/* 1.	If x is negative, then the position of the trait locus is to the left 
			of the leftmost marker in the interval. Calculate the new position and\
			the theta value between the trait locus and the leftmost marker, and
			return. */

	if (x < 0.0) {
		*newLocPos = scanData->startPos - 1;
		*newTheta = getcMToRec(-x, progData.haldane);
		return;
	}

	numMarkers = scanData->numIntervals + 1;

	/* 2.	Iterate over the markers until reaching a marker whose distance from
			the leftmost marker in the scan interval exceeds the distance 'x'. */
			
	for (currMarker = 0, distFromMarker1Ptr = scanData->distFromMarker1; 
			(currMarker < numMarkers) && (*distFromMarker1Ptr <= x); 
			distFromMarker1Ptr++) {

		currMarker++;
	}

	/* 3.	Update the new position on the genetic map of the trait locus with
			the index of this marker, and update the theta value between the trait
			locus and the marker preceeding it. */

	dist = x - *(distFromMarker1Ptr-1);
	*newLocPos = currMarker + scanData->startPos - 1;
	*newTheta = getcMToRec(dist, progData.haldane);
}



/******************************************************************************
*
*       function       moveTraitLoc 
*       parameters     numTraitLoci (the number of trait loci that need to be 
*					   moved), oldLocPos (an array with the old indices of the
*					   positions of the trait loci), newLocPos (an array with 
*					   the new positions of the trait loci), newThetas (the 
*					   new recombination fractions between the trait loci and
*					   the markers preceeding them).
*       output         none
*       description    Updates the positions of the trait loci in the data
*					   structures.
*       comments	   none 
*
******************************************************************************/
void moveTraitLoc(int numTraitLoci, int *oldLocPos, int *newLocPos, 
				  double *newThetas)
{
	scanIntervalData	*scanData;

	scanData = progData.scanData;

	if (numTraitLoci == 1) {
		moveLocus(*oldLocPos, *newLocPos);
		changeThetas(*newLocPos, *newThetas);
		scanData->currPosTraitLoc = *newLocPos;
	}
	else {
		moveLoci(oldLocPos[0], &(newLocPos[0]), newThetas[0], 
				 oldLocPos[1], &(newLocPos[1]), newThetas[1]);

		scanData[0].currPosTraitLoc = newLocPos[0];
		scanData[1].currPosTraitLoc = newLocPos[1];
	}
}


/******************************************************************************
*
*       Function        calcLikelihood
*       Parameters		prodArr (an array of products which represent the 
*						Bayesian networks built from the input pedigrees),
*						elimOrderArrs (an array of arrays which represent the 
*						elimination order determined for each of the Bayesian
*						networks).
*       Output          none
*       Description		Calculates the likelihood of the data (the products in
*						'prodArr') given the recombination values.
*       Comments		none
*
******************************************************************************/
void calcLikelihood(product** prodArr, elimOrder *elimOrderArrs)
{
	int				pedNum;
	boolean			elimSelectorsTogetherGeneral;
	elimOrderPed	*elimOrdPtr;
	boolean			*isZeroLikePed;

	#if ELIMINATE_SELECTORS_TOGETHER
		elimSelectorsTogetherGeneral = ((!readElimOrder) && (recombData.sexDiff == 0)
										 && (loci.numTraitLoci == 1));
		initVars();
	#endif

	/* 1.	Iterate over the pedigrees. */

	for (pedNum = 0, elimOrdPtr = *elimOrderArrs, isZeroLikePed = totalRes.isZeroLikeArr; 
			currPed = peds.head, pedNum < peds.numPeds; pedNum++, elimOrdPtr++,
			isZeroLikePed++, currPed = currPed->nextPed) {

		/* 1.1.	If the likelihood of this pedigree is known to be zero, continue to
				the next pedigree. */

		if (*isZeroLikePed) 
			continue;
		
		/* 1.2.	Determine whether selectors should be eliminated together. */

		#if ELIMINATE_SELECTORS_TOGETHER
			eliminateSelectorsTogether = (elimSelectorsTogetherGeneral && elimOrdPtr->isLGOrder
											&& (elimOrdPtr->numBreakNodes == 0));
		#endif

		elimOrdPtr->currElimNode = 0;

		/* 1.3.	Calculate the likelihood of each pedigree. */

		#if PERFORM_EXACT_COMPUTATIONS
			if (calcLikelihoodPed(prodArr[pedNum], pedNum, elimOrdPtr) == FALSE) {
				*isZeroLikePed = TRUE;
			}
		#else
			if (calcLikelihoodPedApproximate(prodArr[pedNum], pedNum, elimOrdPtr) == FALSE) {
				*isZeroLikePed = TRUE;
			}
		#endif
	}

	/* 2.	Zero the fields of the Bayesian networks and free the memory allocated 
			dynamically. */

	zeroDataFreeMemory(prodArr);
}



/******************************************************************************
*
*       Function        calcLikelihoodPed
*       Parameters		prod (the product which represents the Bayesian network 
*						of which likelihood is being calculated), pedNum (the  
*						index of the pedigree in the data-structures),
*						elimOrderArr (an array that stores an elimination order
*						for the variable of this product).
*       Output          none
*       Description		Calculates the log-likelihood of the pedigree represented
*						by the product 'prod', according to the elimination
*						order stored in 'elimOrderArr'.
*       Comments		Returns FALSE if the likelihood is zero, otherwise
*						returns TRUE. The log-likelihood calculated is added to 
*						the total log-likelihood calculated up till now.
*
******************************************************************************/
boolean calcLikelihoodPed(product *prod, int pedNum, elimOrderPed *elimOrderArr)
{
	Probability		Res;

	/* 1.	Perform inference on the product. */

	Res = inferProd(prod, elimOrderArr, 1, 0);

	/* 2.	If the likelihood of this pedigree is zero, update that it's zero,
			and return FALSE; Otherwise, add the log-likelihood of this pedigree
			to the total log-likelihood calculated up till now of the rest of 
			the pedigrees, and return TRUE. */

	if (Res <= MIN_DBL)	{		/* The likelihood of this pedigree is zero. */
		totalRes.isZeroLikeArr[pedNum] = TRUE;
		totalRes.pedsLnLikeArr[pedNum] = MIN_DBL;
		return FALSE;
	}
	else {

		totalRes.pedsLnLikeArr[pedNum] += Res;
		totalRes.currRes->lnLike += Res;
		return TRUE;
	}
}


/******************************************************************************
*
*       Function        calcLikelihoodPedApproximate
*       Parameters		prod (the product which represents the bayes-net of
*						which likelihood is being calculated), 
*						elimOrderArr (an array that stores the elimination order
*						determined for the likelihood calculation of this pedigree) .
*       Output          none
*       Description		Calculates an approximation to the log-likelihood of the 
*						pedigree represented by the product 'prod', according 
*						to the elimination order stored in 'elimOrderArr'.
*       Comments		Returns FALSE if the likelihood is zero, otherwise
*						returns TRUE.  The log-likelihood calculated is added to 
*						the total log-likelihood calculated up till now.
*
******************************************************************************/
boolean calcLikelihoodPedApproximate(product *prod, int pedNum, 
									 elimOrderPed *elimOrderArr)
{
	return FALSE;
}


/******************************************************************************
*
*       Function        inferProd
*       Parameters		prod (pointer to the product on which the inference
*						is being performed), elimOrderArr (an array that stores 
*						an elimination order for the variable of this product),
*						problemSize (the size of the problem that needs to be 
*						solved by the current call to the function),
*						solved (how much of the original problem has already
*						been solved).
*       Output          If requested, prints a progress report (how much of the
*						original problem has been solved already).
*       Description		Performs inference on 'prod'.						 
*       Comments		Returns the result of the inference.
*
******************************************************************************/
Probability inferProd(product *prod, elimOrderPed *elimOrderArr, 
					  float problemSize, float solved)
{
	variable		*nextVar;
	Probability		res, tempRes;
	boolean			isBreakVar, isZeroLike;

	/* 1.	Initialize the probability to zero. */

	res = 0.0;

	/* 2.	Perform inference on the product by performing variable-elimination
			(combined with conditioning whenever necessary). 
			While there are still variables to eliminate: */

	while (prod->currNumVars) {		

		/* 2.1.	Choose a variable to eliminate (or condition on) next. */

		nextVar = getEliminateVar(elimOrderArr, &isBreakVar);

		/* 2.2.	If the next variable is a break variable (a variable that we 
				condition on), perform product-breaking on it, add the result of 
				the inference on this variable to the total up to now and return it. */

		if (isBreakVar) {
			tempRes = inferProdWithCond(prod, nextVar, elimOrderArr, problemSize,
										solved);
			res += tempRes;
			return res;
		}

		#if ELIMINATE_SELECTORS_TOGETHER

			/* 2.3.	If this is a small pedigree and we've reached a selector
					variable, it means that we have eliminated all the phenotype 
					and genotype variables of this locus and now we need to eliminate
					all the selector variables of this locus. */

			if ((eliminateSelectorsTogether) && ((nextVar->originalNode)->nodeType == SELECTOR)) {
				if ((nextVar->originalNode)->lNum > 0) {
					elimOrderArr->currElimNode--;
					tempRes = eliminateSelectorsLocusI(prod, elimOrderArr, &isZeroLike);
					res += tempRes;

					if (isZeroLike) {
						res = MIN_DBL;
						break;
					}
					continue;
				}
			}
		
		#endif

		/* 2.4.	Else, just eliminate the chosen variable from the product. 
				If the elimination resulted in a constant function, add
				its logProbability to the total logProbability. If the likelihood 
				is zero, break. */
	//	printProduct(prod, res);
		if (eliminateVar(prod, nextVar, &tempRes, &isZeroLike)) {
			res += tempRes;	
		}
		else if (isZeroLike) {
			res = MIN_DBL;
			break;
		}
	}
//	printProduct(prod, res);
	/* 3.	If requested, print a progress report. */

	#if PRINT_PROGRESS_REPORT
		printf("Solved %f percent of the problem.\n", (solved+problemSize)*100);
	#endif

	/* 4.	Return the result of the inference. */

	return res;
}



/******************************************************************************
*
*       Function        zeroDataFreeMemory
*       Parameters		prodArr (an array of products which represent the 
*						Bayesian networks built from the input pedigrees),
*       Output          none
*       Description		Zeroes the 'varInProd' field for all the nodes of the 
*						Bayesian networks, and frees the memory allocated for 
*						the products.
*       Comments		none
*
******************************************************************************/
void zeroDataFreeMemory(product **prodArr)
{
	/* 1.	Zero the 'varInProd' field for all the nodes of the Bayesian networks. */

	if (loci.sexLink)
		zeroNodeFieldsPeds(zeroNodeFieldsPedSexLinked);
	else
		zeroNodeFieldsPeds(zeroNodeFieldsPedAutosomal);
	
	/* 2.	Free the memory allocated for the products. */

	freeProducts(prodArr, peds.numPeds);
}





