/*****************************************************************************/
/*  MapFunction.h															 */
/*  Author: Ma'ayan Fishelson				                                 */
/*****************************************************************************/

#ifndef _SL_MAP_FUNCTION_H_
#define _SL_MAP_FUNCTION_H_

#include <stdio.h>
#include <stdlib.h>


/*------------------------------*/
/*		DEFINITIONS				*/
/*------------------------------*/

/* The distance (in Morgans) used to indicate a recombination fraction of 0.5. */

#define		UNLINKED_DIST		1000

/* We assume that we have three loci A, B, and C which are present on the 
   chromosome in the order ABC.  */

/* Computes the recombination fraction theta(AC) from the known recombination
   fractions theta(AB) and theta(BC). Haldane's map function is used to compute the
   following recombination fraction. */

#define getHaldaneThetaAC(thetaAB, thetaBC) \
	(thetaAB + thetaBC - (2 * thetaAB * thetaBC));

/* Computes the recombination fraction theta(BC) from the map-function-derived
   theta(AC) and a given theta(AB). Haldane's map function is used to compute the
   following recombination fraction. */

#define getHaldaneThetaBC(thetaAC, thetaAB) \
	((thetaAC - thetaAB) / (1 - 2 * thetaAB))>=0?((thetaAC - thetaAB) / (1 - 2 * thetaAB)):0;


/* Computes the recombination fraction theta(AC) from the known recombination
   fractions theta(AB) and theta(BC). Kosambi's map function is used to compute the
   following recombination fraction. */

#define getKosambiThetaAC(thetaAB, thetaBC) \
	(thetaAB + thetaBC) / (1 + 4 * thetaAB * thetaBC);

/* Computes the recombination fraction theta(AC) from the known recombination
   fractions theta(AB) and theta(BC). Kosambi's map function is used to compute the
   following recombination fraction. */

#define getKosambiThetaBC(thetaAC, thetaAB) \
	((thetaAC - thetaAB) / (1 - 4 * thetaAB * thetaAC))>=0?((thetaAC - thetaAB) / (1 - 4 * thetaAB * thetaAC):0;


/* Computes the recombination fraction (probability) from the given distance 
   in Morgans, using Haldane's map function. */

#define getHaldaneMToRec(dist) \
	(0.5 * (1 - exp(-2 * dist)));

/* Computes the distance in Morgans from the given recombination fraction 
   (probability), using Haldane's map function. */

#define getHaldaneRecToM(rec) \
	(rec == 0.5) ? 4.0 : (log(1 - 2 * rec) * (-0.5));

/* Computes the recombination fraction (probability) from the given distance 
   in Morgans, using Kosambi's map function. */

#define getKosambiMToRec(dist) \
	(0.5*(exp(4*dist)-1)/(exp(4*dist)+1));


/* Computes the distance in Morgans from the given recombination fraction 
   (probability), using Kosambi's map function. */

#define getKosambiRecToM(rec) \
	(0.25*log((1+2*rec)/(1-2*rec)));

/* Computes the recombination fraction theta(BC) from the map-function-derived
   theta(AC) and a given theta(AB). Haldane's or Kosambi's map function is used 
   to compute the following recombination fraction. */

#define getThetaBC(thetaAC, thetaAB, haldane) \
	haldane ? (((thetaAC - thetaAB) / (1 - 2 * thetaAB))>0?(thetaAC - thetaAB) / (1 - 2 * thetaAB):0) : (((thetaAC - thetaAB) / (1 - 4 * thetaAB * thetaAC))>0?(thetaAC - thetaAB) / (1 - 4 * thetaAB * thetaAC):0);

/* Computes the distance in Morgans from the given recombination fraction 
   (probability), using either Haldane's or Kosambi's map function. */

#define	getRecToM(rec, haldane) \
	haldane ? (log(1 - 2 * rec) * (-0.5)) : (0.25*log((1+2*rec)/(1-2*rec)));

/* Computes the recombination fraction (probability) from the given distance 
   in Morgans, using Haldane's or Kosambi's map function. */

#define getMToRec(dist, haldane) \
	haldane ? (0.5 * (1 - exp(-2 * dist))) : (0.5*(exp(4*dist)-1)/(exp(4*dist)+1));

/* Computes the distance in cM from the given recombination fraction 
   (probability), using either Haldane's or Kosambi's map function. */

#define	getRecTocM(rec, haldane) \
	(rec > 0.499) ? UNLINKED_DIST : \
	(haldane ? (log(1 - 2 * rec) * (-50)) : (25*log((1+2*rec)/(1-2*rec))));

/* Computes the recombination fraction (probability) from the given distance 
   in Morgans, using Haldane's or Kosambi's map function. */

#define getcMToRec(dist, haldane) \
	haldane ? (0.5 * (1 - exp(-0.02 * dist))) : (0.5*(exp(0.04*dist)-1)/(exp(0.04*dist)+1));


#endif		/*	_MAP_FUNCTION_H_	*/

