/*********************************************************************************/
/*  Mem.c											     				         */
/*  Author: Nikolay Dobgolevsky					                                 */
/*********************************************************************************/
#include "sl_Mem.h"


/******************************************************************************
*
*       Function        totalPhysMem
*       Parameters		none
*       Output          none
*       Description     The function calculates and returns the total amount of 
*						physical memory of the computer at hand.
*       Comments		The function is supposed to work properly in the 
*                       following Operating Systems: WIN32, LINUX, SUN, MAC OSX
*                       Solaris, Cygwin, HPUX11, IRIX6, TRU64 UNIX V4/5, BSD, 
*						DARWIN and AIX.
*                       The function was tested for WIN32, LINUX, SUN and
*                       MAC OSX only! 
*                       In case the function cannot determine the amount of
*                       physical memory available it returns 33% of the
*                       virtual memory available.
*                       WARNING: function may return wrong values for extremly 
*						big amounts of physical memory (over 4 GB RAM).
*
*******************************************************************************/
double totalPhysMem()
{

#if defined _SC_PHYS_PAGES && defined _SC_PAGESIZE
  { /* This works on linux-gnu, solaris2 and cygwin.  */
    double pages = sysconf (_SC_PHYS_PAGES);
    double pagesize = sysconf (_SC_PAGESIZE);
    if (0 < pages && 0 < pagesize)
      return pages * pagesize;
  }
#endif

#if HAVE_PSTAT_GETSTATIC
  { /* This works on hpux11.  */
    struct pst_static pss;
    if (0 <= pstat_getstatic (&pss, sizeof pss, 1, 0))
      {
	double pages = pss.physical_memory;
	double pagesize = pss.page_size;
	if (0 < pages && 0 < pagesize)
	  return pages * pagesize;
      }
  }
#endif

#if HAVE_SYSMP && defined MP_SAGET && defined MPSA_RMINFO && defined _SC_PAGESIZE
  { /* This works on irix6. */
    struct rminfo realmem;
    if (sysmp (MP_SAGET, MPSA_RMINFO, &realmem, sizeof realmem) == 0)
      {
	double pagesize = sysconf (_SC_PAGESIZE);
	double pages = realmem.physmem;
	if (0 < pages && 0 < pagesize)
	  return pages * pagesize;
      }
  }
#endif

#if HAVE_GETSYSINFO && defined GSI_PHYSMEM
  { /* This works on Tru64 UNIX V4/5.  */
    int physmem;

    if (getsysinfo (GSI_PHYSMEM, (caddr_t) &physmem, sizeof (physmem),
		    NULL, NULL, NULL) == 1)
      {
	double kbytes = physmem;

	if (0 < kbytes)
	  return kbytes * 1024.0;
      }
  }
#endif

#if (HAVE_SYSCTL  || defined __APPLE_CC__) && defined HW_USERMEM
  { /* This works on *bsd and darwin.  */
    unsigned int physmem;
    size_t len = sizeof physmem;
    static int mib[2] = { CTL_HW, HW_PHYSMEM };

    if (sysctl (mib, ARRAY_SIZE (mib), &physmem, &len, NULL, 0) == 0
	&& len == sizeof (physmem))
      return (double) physmem;
  }
#endif

#if HAVE__SYSTEM_CONFIGURATION
  /* This works on AIX.  */
  return _system_configuration.physmem;
#endif

#if defined _WIN32
 	MEMORYSTATUS ms;
	GlobalMemoryStatus (&ms);
	return (double) ms.dwTotalPhys;
#endif

  /* Return 33% of virtual memory and hope for best. */
  return availableVirtMem() / 3.0;
}


/******************************************************************************
*
*       Function        availablePhysMem
*       Parameters		none
*       Output          none
*       Description     Calculates and returns the amount of physical memory 
*						available.
*       Comments		The function is supposed to work properly in the 
*                       following Operating Systems: WIN32, LINUX, SUN, MAC OSX,
*                       Solaris, Cygwin, HPUX11, IRIX6, TRU64 UNIX V4/5, BSD, 
*						DARWIN and AIX.
*                       The function was tested for WIN32, LINUX, SUN and
*                       MAC OSX only! 
*                       In case the function cannot determine the amount of
*                       physical memory available it returns 85% of the
*                       value returned by totalPhysMem().
*                       WARNING: the function may return wrong values for extremly 
*						big amounts of physical memory (over 4 GB RAM).
*
*******************************************************************************/
double availablePhysMem()
{
#if defined _SC_PHYS_PAGES && defined _SC_PAGESIZE
  { /* This works on linux-gnu, solaris2 and cygwin.  */
    double pages = sysconf (_SC_AVPHYS_PAGES);
    double pagesize = sysconf (_SC_PAGESIZE);
    if (0 < pages && 0 < pagesize)
      return pages * pagesize;
  }
#endif

#if HAVE_PSTAT_GETSTATIC && HAVE_PSTAT_GETDYNAMIC
  { /* This works on hpux11.  */
    struct pst_static pss;
    struct pst_dynamic psd;
    if (0 <= pstat_getstatic (&pss, sizeof pss, 1, 0)
	&& 0 <= pstat_getdynamic (&psd, sizeof psd, 1, 0))
      {
	double pages = psd.psd_free;
	double pagesize = pss.page_size;
	if (0 < pages && 0 < pagesize)
	  return pages * pagesize;
      }
  }
#endif

#if HAVE_SYSMP && defined MP_SAGET && defined MPSA_RMINFO && defined _SC_PAGESIZE
  { /* This works on irix6. */
    struct rminfo realmem;
    if (sysmp (MP_SAGET, MPSA_RMINFO, &realmem, sizeof realmem) == 0)
      {
	double pagesize = sysconf (_SC_PAGESIZE);
	double pages = realmem.availrmem;
	if (0 < pages && 0 < pagesize)
	  return pages * pagesize;
      }
  }
#endif

#if HAVE_TABLE && defined TBL_VMSTATS
  { /* This works on Tru64 UNIX V4/5.  */
    struct tbl_vmstats vmstats;

    if (table (TBL_VMSTATS, 0, &vmstats, 1, sizeof (vmstats)) == 1)
      {
	double pages = vmstats.free_count;
	double pagesize = vmstats.pagesize;

	if (0 < pages && 0 < pagesize)
	  return pages * pagesize;
      }
  }
#endif

#if (HAVE_SYSCTL || defined __APPLE_CC__) && defined HW_USERMEM
  { /* This works on *bsd and darwin.  */
    unsigned int usermem;
    size_t len = sizeof usermem;
    static int mib[2] = { CTL_HW, HW_USERMEM };

    if (sysctl (mib, ARRAY_SIZE (mib), &usermem, &len, NULL, 0) == 0
	&& len == sizeof (usermem))
      return (double) usermem;
  }
#endif

#if defined _WIN32
	MEMORYSTATUS ms;
	GlobalMemoryStatus (&ms);
	return (double) ms.dwAvailPhys;
#endif

  /* Guess 85% of physical memory.  */
  return totalPhysMem () * 0.85;
}


/******************************************************************************
*
*       Function        availableVirtMem
*       Parameters		none
*       Output          none
*       Description     Computes and returns the amount of virtual memory 
*						available currently (the size, in bytes, of the maximal
*						block the function succeeded to allocate).
*       Comments		The function is system independent.
*
*******************************************************************************/
double availableVirtMem()
{
	double		block_size;
	_byte		*ptr;

	/* 1.	Initial block size. */

	block_size = 64 * 1024 * 1024;

	/* 2.	As long as succeeded in allocating the memory block, free it and 
			multiply the block size by 1.4142. */

	while (ptr = (_byte*)malloc((size_t)block_size)) {
		free(ptr);
		block_size *= 1.4142;
	}

	return ceil(block_size);
}

