/*****************************************************************************/
/*  Optimize.c																 */
/*  Author: Ma'ayan Fishelson				                                 */
/*****************************************************************************/

#include "sl_Optimize.h"


/******************************************************************************
*
*       Function        findMaxInScanInterval
*       Parameters      func (pointer to the function whose maximum needs to be 
*						evaluated), funcData (additional data which needs to be
*						passed to the evaluated function), xMax (is updated with
*					    the abscissa of the maximum).
*       Output          none
*       Description		Finds the position of the trait locus in the scan interval
*						where the function pointed to by 'func' has a maximum. 
*       Comments		none
*
******************************************************************************/
double findMaxInScanInterval(funcType func, void *funcData, double *xMax)
{
	double				ax, bx, cx, maxFuncVal;
	scanIntervalData	*scanData;

	scanData = progData.scanData;

	/* 1.	Iterate the trait locus within the specified scan interval, updating
			the position which generated the maximum probability whenever necessary. 
			Find a triplet of points ax, bx, cx (positions of the trait locus
			with regards to the beginning of the scan interval) that bracket a 
			maximum of the function (the likelihood function).*/

	iterateTraitScanInterval(func, scanData, funcData, 
							 STEP_SIZE, &ax, &bx, &cx);

	/* 2.	Performs a golden section search for the maximum. */

	maxFuncVal = golden(func, ax, bx, cx, TOL, funcData, xMax);

	/* 3.	Return the maximum value of the function in the specified scan
			interval. */

	return maxFuncVal;
}



/******************************************************************************
*
*       Function        iterateTraitScanInterval
*       Parameters      func (pointer to the function which is evaluated in 
*						each position on the grid), scanData (the data on the 
*						scan interval in which the trait locus needs to be 
*						iterated), funcData (additional data which needs to be
*						passed to the evaluated function), stepSize (the size
*						of the step between successive likelihood evaluations),
*						ax, bx, cx (will be updated with 3 positions that will
*						be passed on to the golden search function).
*       Output          none
*       Description		Iterates the trait locus in the scan interval in 
*						fixed steps (performs a grid search).
*       Comments		none
*
******************************************************************************/
void iterateTraitScanInterval(funcType func, scanIntervalData *scanData, 
							  void *funcData, double stepSize, double *ax, 
							  double *bx, double *cx)
{
	double		currPosGen, currPosInterval, *mapDistArrPtr, 
				*distFromMarker1Ptr, *resArr, *resArrPtr, *maxRes, maxVal, 
				currVal, *stepSizeArr, *stepSizeArrPtr;
	int			numStepsTotal, currInterval, currIter, maxResInd, numIntervals;


	/* 1.	Allocate memory to store an array with the step size between successive
			function evaluaions in each of the internal intervals. */

	numIntervals = scanData->numIntervals + (scanData->offEndDistLeft != 0) + 
						(scanData->offEndDistRight != 0);

	myMalloc(stepSizeArr, sizeof(double) * numIntervals);

	/* 2.	Initialize the total number of steps in the whole scan interval to zero. */

	numStepsTotal = 0;
	stepSizeArrPtr = stepSizeArr;

	/* 3.	Calculate the necessary number of steps in the interval off the left 
			end of the scan interval. */

	if (scanData->offEndDistLeft > 0) {

		numStepsTotal += calcNumSteps(scanData->offEndDistLeft, stepSize, stepSizeArrPtr);
		stepSizeArrPtr++;
	}

	/* 4.	Iterate over the internal intervals in the scan interval. */

	for (currInterval = 0, mapDistArrPtr = scanData->mapDistArr + 1;
			currInterval < scanData->numIntervals; 
			currInterval++, mapDistArrPtr++, stepSizeArrPtr++) {

		/* 4.1.	Calculate the number of steps required in the current interval,
				and update the necessary step size in this interval. */

		numStepsTotal += calcNumSteps(*mapDistArrPtr, stepSize, stepSizeArrPtr) ;
	}

	/* 5.	Calculate the necessary number of steps in the interval off the right 
			end of the scan interval. */

	if (scanData->offEndDistRight > 0) {

		numStepsTotal += calcNumSteps(scanData->offEndDistRight, stepSize, stepSizeArrPtr);
	}

	/* 6.	Allocate memory to store an array with the results of all likelihood
			evaluations. */

	myMalloc(resArr, sizeof(double) * numStepsTotal);

	/* 7.	Position the trait locus at the beginning of the scan interval. */

	maxRes = NULL;
	maxVal = MIN_DBL;
	currIter = 0;
	maxResInd = 0;
	stepSizeArrPtr = stepSizeArr;
	resArrPtr = resArr;

	/* 8.	Perform the required number of steps within the left-side interval (if exists). */

 	if (scanData->offEndDistLeft > 0) {

	   for (currPosInterval = -scanData->offEndDistLeft; currPosInterval < 0.0; 
				currPosInterval += (*stepSizeArrPtr), currIter++, resArrPtr++) {
			
			(*resArrPtr) = currPosInterval;

			/* 8.1.	Calculate the value of the function at the current position. */

			currVal = (*func)(&currPosInterval, funcData);

			/* 8.2.	Update a pointer to the maximum result entry, if necessary. */

			if ((maxRes == NULL) || (maxVal < currVal)) {
				maxRes = resArrPtr;
				maxResInd = currIter;
				maxVal = currVal;
			}
		}

		stepSizeArrPtr++;
	}

	/* 9.	Iterate over the internal intervals in the scan interval. */

	for (currInterval = 0, mapDistArrPtr = scanData->mapDistArr + 1,
			distFromMarker1Ptr = scanData->distFromMarker1; 
			currInterval < scanData->numIntervals; currInterval++,  
			mapDistArrPtr++, distFromMarker1Ptr++, stepSizeArrPtr++) {

		/* 9.1.	Initialize 'currPosGen' to the distance of the left end of the current
				internal interval from the left end of the total scan interval. */

		currPosGen = (*distFromMarker1Ptr);

		/* 9.2.	Perform the required number of steps within the internal interval. */

        for (currPosInterval = 0.0;  currPosInterval < (*mapDistArrPtr); 
				currPosInterval += (*stepSizeArrPtr), currIter++, resArrPtr++, 
				currPosGen += (*stepSizeArrPtr)) {
			
			(*resArrPtr) = currPosGen;

			/* 9.2.1.	Calculate the value of the function at the current position. */

			currVal = (*func)(&currPosGen, funcData);

			/* 9.2.2.	Update the pointer to the maximum result entry, if necessary. */

			if ((maxRes == NULL) || (maxVal < currVal)) {
				maxRes = resArrPtr;
				maxResInd = currIter;
				maxVal = currVal;
			}
		}
	}

	/* 10.	Perform the required number of steps within the right-side interval 
			(if exists). */

 	if (scanData->offEndDistRight > 0) {

		currPosGen = (*distFromMarker1Ptr);

		for (currPosInterval = 0; currPosInterval < scanData->offEndDistRight; 
				currPosInterval += (*stepSizeArrPtr), currIter++, resArrPtr++,
				currPosGen += (*stepSizeArrPtr)) {
			
			(*resArrPtr) = currPosInterval;

			/* 10.1.	Calculate the value of the function at the current position. */

			currVal = (*func)(&currPosGen, funcData);

			/* 10.2.	Update a pointer to the maximum result entry, if necessary. */

			if ((maxRes == NULL) || (maxVal < currVal)) {
				maxRes = resArrPtr;
				maxResInd = currIter;
				maxVal = currVal;
			}
		}

		stepSizeArrPtr++;
	}
	else {

		/* 10.	If not, calculate the value of the function at the right end of the 
				scan interval. */

		currVal = (*func)(distFromMarker1Ptr, funcData);
		(*resArrPtr) = (*distFromMarker1Ptr);

		/* 11.	Update a pointer to the maximum result entry, if necessary. */

		if (maxVal < currVal) {
			maxRes = resArrPtr;
			maxResInd = currIter;
			maxVal = currVal;
		}

		currIter++;
	}

	/* 11.	If the maximum wasn't at one of the ends of the interval,		
			update the position that generated the maximum value of the function
			and the positions on both sides of it. Otherwise, update ax, bx, & cx 
			to hold the position that generated the maximum value of the function,
			and the two positions following it (in case that the the maximum value 
			was found in the left end of the interval) or preceeding it (in case 
			that the the maximum value was found in the right end of the interval). */

	if (maxResInd > 0) {
		if  (maxResInd < currIter - 1) {

			/* 11.1.	The maximum was found in the middle of the scan interval. 
						Update bx to hold the position that generated the maximum 
						value of the function, update ax to hold the position preceeding
						this maximum value position, and cx to hold the position 
						following it. */

			*bx = *maxRes;
			*ax = *(maxRes - 1);
			*cx = *(maxRes + 1);
		}
		else {

			/* 11.1.	The maximum was found in the right end of the scan interval. 
						Update cx to hold the position that generated the maximum 
						value of the function, update ax & bx to hold the two positions
						preceeding this maximum value position. */

			*cx = *maxRes;
			*bx = *(maxRes - 1);  
			*ax = *(maxRes - 2);  
		}
	}
	else {

		/* 11.1.	The maximum was found in the left end of the scan interval. 
					Update ax to hold the position that generated the maximum 
					value of the function, update bx & cx to hold the two positions
					following this maximum value position. */

		*ax = *maxRes;
		*bx = *(maxRes + 1);  
		*cx = *(maxRes + 2);  
	}

	/* 12.	Free the memory allocated dynamically. */

	myFree(resArr);
	myFree(stepSizeArr);
}



/******************************************************************************
*
*       Function        calcNumSteps
*       Parameters      dist (the size of the interval for which the necessary
*						number of steps needs to be calculated), stepSize (the size
*						of the step between successive likelihood evaluations),
*						stepSizeArrPtr (needs to be updated with the step size
*						between successive function evaluations in this interval).
*       Output          none
*       Description		Calculates the number of steps required in the current 
*						interval, and updates the necessary step size in this interval. 
*       Comments		none.
*
******************************************************************************/
int calcNumSteps(double	dist, double stepSize, double *stepSizeArrPtr)
{
	int		numSteps;

	/* 1.	Calculate the number of steps required in the current interval,
			and update the necessary step size in this interval. */

	numSteps = (int)(dist / stepSize);

	if (numSteps < MIN_NO_STEPS) {
		numSteps = MIN_NO_STEPS;
	}

	*stepSizeArrPtr = dist / numSteps;

	/* 2.	Return the total number of steps required for this scan interval. */

	return (1 + numSteps);
}


/******************************************************************************
*
*       function       golden 
*       parameters     func (pointer to the function which is evaluated in 
*					   each position), ax, bx, cx (a bracketing triplet of 
*					   abscissas, such that bx is between ax and cx, and f(bx)
*					   is less than both f(ax) and f(cx)), tol (the required
*					   tolerance), funcData (additional data which needs to be
*					   passed to the evaluated function) xmax (is updated with
*					   the abscissa of the maximum).
*       output         none
*       description    Performs a golden section search for the maximum, 
*					   isolating it to a fractional precision of about 'tol'.
*       comments	   Returns the maximum function value (the value of the 
*					   function in 'xmax').
*
******************************************************************************/
double golden(funcType func, double ax, double bx, double cx, double tol, 
			  void *funcData, double *xmax)
{	
	double		f1, f2, x0, x1, x2, x3, searchTol;


	/* 1.	At any given time, the algorithm keeps track of 4 points: 
			x0, x1, x2, & x3. */

	x0 = ax;
	x3 = cx;

	/* 2.	Make x0 to x1 the smaller segment, and fill in the new point to
			be tried. */

	if (fabs(cx - bx) > fabs(bx - ax)) {
		x1 = bx;
		x2 = bx + C * (cx - bx);
	}
	else {
		x2 = bx;
		x1 = bx - C * (bx - ax);
	}

	/* 3.	Evaluate the function at these two points. */

	f1 = (*func)(&x1, funcData);
	f2 = (*func)(&x2, funcData);

	searchTol = ((progData.progSpecific).optlinkData).searchTol;

	/* 4.	Keep going as long as the required precision hasn't been met. */

	while ((fabs(x3 - x0) > tol * (fabs(x1) + fabs(x2))) && (fabs(f1 - f2) > searchTol)) {

		/* 4.1.	Determine which three points to keep for the next step. */

		if (f2 > f1) {

			/* 4.1.1.	Get rid of the first point. */
			
			shift3(x0, x1, x2, R * x1 + C * x3);
			shift2(f1, f2, (*func)(&x2, funcData));
		}
		else {

			/* 4.1.1.	Get rid of the last point. */

			shift3(x3, x2, x1, R * x2 + C * x0);
			shift2(f2, f1, (*func)(&x1, funcData));
		}
	}

	/* 5.	The required precision has been reached. Output the better of the two
			current values. */

	if (f1 > f2) {
		*xmax = x1;
		return f1;
	}
	else {
		*xmax = x2;
		return f2;
	}
}


