/*********************************************************************************/
/*  PedData.c																     */
/*  Author: Ma'ayan Fishelson													 */
/*********************************************************************************/

#include "sl_PedData.h"


/******************************************************************************
*
*       Function        readPedFile
*       Parameters		pedfileName (the name of the pedigre file being read).
*       Output          none
*       Description		Reads the pedigree file, and initializes the 
*						data-structures that hold the information on the 
*						pedigrees being analyzed. 
*       Comments		none
*
******************************************************************************/
void readPedFile(int numMarkers,int* markersUsed, Bool diseaseUsed, PedigreeDataStruct *currPedigree){
	int			pedInd, numPeds;
	pedigree	*ped;


	/* 1.	Allocate memory and initialize the data-structures that hold 
			information on the input pedigrees, according to the number of pedigrees
			in the pedigree file and the number of people in each pedigree. */

	computePedSize(currPedigree);

	ped = peds.head;
	numPeds = peds.numPeds;

	/* 2.	Iterate over the pedigrees. */

	for (pedInd = 0; pedInd < numPeds; pedInd++) {

		/* 2.1.	Update the data-structures according to the information in the 
				pedigree file. */

		readPed(ped, pedInd+1,numMarkers, markersUsed, diseaseUsed,currPedigree);
		ped = ped->nextPed;
	}
}



/******************************************************************************
*
*       Function        computePedSize
*       Parameters		pedfileName (the name of the pedigree file for which 
*						data-structures need to be allocated).
*       Output          none
*       Description		Reads the pedigree file, counts the number of 
*						pedigrees and the number of people in each pedigree
*						and initializes the data-structures accordingly. 
*       Comments		none
*
******************************************************************************/
void computePedSize(PedigreeDataStruct *currPedigree/*char *pedfileName*/)
{
	int			currPed,  numPeople;
	pedigree	*ped, *last;
//	int i;
	last = NULL;

	/* 1.	Initialize the number of pedigrees to zero. */

	peds.numPeds = 0;
	peds.head = NULL;

	currPed = 0;//i;//  nextPed;
	numPeople = currPedigree->numPeople;
		
		/* 2.3.	Increase the count of the number of pedigrees in the file. */

		peds.numPeds++;

		/* 2.4.	Allocate memory for the pedigree struct. */
		
		myMalloc(ped, sizeof(pedigree));
		myMalloc(ped->people, sizeof(person) * numPeople);

		/* 2.5.	Initialize the pedigree struct's fields. */

		ped->numPeople = numPeople;
		ped->pedNum = currPed;
		ped->nextPed = NULL;
		ped->childLess = NULL;
		ped->founder = NULL;
		ped->numNonFounders = 0;
		
		/* 2.6.	Insert the pedigree struct into the list of pedigree structs. */

		if (last == NULL)	/* This is the 1st pedigree read. */
			peds.head = last = ped;
		else {
			last->nextPed = ped;
			last = ped;
	
}

	/* 3.	Close pedFile and reopen it. */

   // fclose(pedFile);
	//printf("Reopening %s\n", pedfileName);
	//pedFile = fopen(pedfileName, "r");
}



/******************************************************************************
*
*       Function        readPed
*       Parameters		ped (the struct which will hold the information on the
*					    current pedigree being read), pedInd (the index of the 
*						pedigree now being read).
*       Output          none
*       Description		Reads one pedigree from the pedigree file. 
*       Comments		none
*
******************************************************************************/
void readPed(pedigree *ped, int pedInd,int numMarkers,int* markersUsed, Bool diseaseUsed,PedigreeDataStruct *currPedigree)
{
	int			personInd, numPeople;
	boolean		sexLink;

	numPeople = ped->numPeople;
	sexLink = loci.sexLink;

	/* 1.	Iterate over the people in the pedigree and read the information on
			each one of them. */

	for (personInd = 0; personInd < numPeople; personInd++) {
		readPerson(ped, pedInd, sexLink, personInd,markersUsed,currPedigree);
	}

	/* 2.	Check that there is a correspondence among all parent-child 
			relationships.*/

	checkInputValidity(ped);
}



/******************************************************************************
*
*       Function        readPerson
*       Parameters		ped (pointer to the pedigree to which the person being 
*					    read belongs), pedInd (the index of the pedigree now 
*						being read), sexLink (a flag that indicates whether 
*						the data is sex-linked or not).
*       Output          none
*       Description		Reads the data on one person from the pedigree file. 
*       Comments		none
*
******************************************************************************/
void readPerson(pedigree *ped, int pedInd, boolean sexLink,  int personInd,int* markersUsed, PedigreeDataStruct *currPedigree)
{
	int		pedNo, indPerson, indPa, indMa, indFirstCh, indNeMa, indNePa, sex, 
			 locusInd, lociPlace, numLoci;
	person *currPerson;
	locus	*locusArr;
	PersonDataStruct *origPerson;
	int markerInd = 0;
	locusArr = loci.locusArr;
	numLoci = loci.numLoci;

	/* 1.	Read general information about the person. */

	/*fscanf(pedFile,"%d%d%d%d%d%d%d%d%d",&pedNo, &indPerson, &indPa, &indMa, &indFirstCh, 
		   &indNePa, &indNeMa, &sex, &proField);
	
	if (proField > 1){
		error("Superlink does not allow any value except 0 as proband status (9th column in pedigee file).\n");
	}*/
	//Anna
	origPerson = currPedigree->people + personInd;
pedNo = pedInd+1;
indPerson = origPerson->id;
indPa = origPerson->father;
indMa = origPerson->mother;
indFirstCh =origPerson->firstChild;
indNePa = origPerson->siblingFather;
indNeMa = origPerson->siblingMother;
sex = origPerson->sex;
//end
currPerson = &(ped->people)[indPerson-1];

	currPerson->pNum = indPerson;

	/* 2.	Check validity of the person's father, and update a pointer to the father. */

	if (indPa != 0) {

		/* 2.1.	Check that both parents are specified in case of a non-founder. */

		if (indMa == 0) {
			inputErr("Only the father of person #%d, pedigree #%d, is specified. \n", 
					 indPerson, pedInd);
		}

		if (indPa <= ped->numPeople) 
			currPerson->pa = &(ped->people)[indPa - 1];
		else {
			inputErr("The father of person #%d, pedigree #%d, doesn't exist. \n", 
					 indPerson, pedInd);
		}
		ped->numNonFounders++;
	}
	else
		currPerson->pa = NULL;

	/* 3.	Check validity of the person's mother, and update a pointer to the mother. */

	if (indMa != 0) {

		/* 3.1.	Check that both parents are specified in case of a non-founder. */

		if (indPa == 0) {
			inputErr("Only the mother of person #%d, pedigree #%d, is specified. \n", 
					 indPerson, pedInd);
		}
		else if (indPa == indMa) {
			inputErr("The father and mother of person #%d, pedigree #%d, are the same person. \n", 
					 indPerson, pedInd);
		}

		if (indMa <= ped->numPeople)
			currPerson->ma = &(ped->people)[indMa - 1];
		else {
			inputErr("The mother of person #%d, pedigree #%d, doesn't exist. \n", 
					 indPerson, pedInd);
		}
	}
	else
		currPerson->ma = NULL;

	/* 4.	Check validity of the person's first child, and update the pointer to it. */

	if (indFirstCh != 0) {
		if (indFirstCh <= ped->numPeople)
			currPerson->ch = &(ped->people)[indFirstCh - 1];
		else {
			inputErr("The first child of person #%d, pedigree #%d, doesn't exist. \n", 
					 indPerson, pedInd);
		}
	}
	else
		currPerson->ch = NULL;

	/* 5.	Check validity of the person's next paternal sibling, and update the pointer to it. */

	if (indNePa != 0) {
		if (indPa == 0) {
			inputErr("Paternal siblings of person #%d, pedigree #%d, are specified but no parents are reported. \n", 
					 indPerson, pedInd);
		}

		if (indNePa <= ped->numPeople)
			currPerson->NeSiPa = &(ped->people)[indNePa - 1];
		else {
			inputErr("The next paternal sibling of person #%d, pedigree #%d, doesn't exist. \n", 
					 indPerson, pedInd);
		}
	}
	else
		currPerson->NeSiPa = NULL;

	/* 6.	Check validity of the person's next maternal sibling, and update the pointer to it. */

	if (indNeMa != 0) {
		if (indPa == 0) {
			inputErr("Maternal siblings of person #%d, pedigree #%d, are specified but no parents are reported. \n", 
					 indPerson, pedInd);
		}

		if (indNeMa <= ped->numPeople)
			currPerson->NeSiMa = &(ped->people)[indNeMa - 1];
		else {
			inputErr("The next maternal sibling of person #%d, pedigree #%d, doesn't exist. \n", 
					 indPerson, pedInd);
		}
	}
	else
		currPerson->NeSiMa = NULL;

	/* 7.	Update the sex field of the person. */

	currPerson->personSex = (sex == 1) ? Male : Female;

	/* 8.	Update the 'childLess' and the 'founder' fields of the pedigree struct. */

	/* 8.1.	If a childless person hasn't been found yet, and this person is childless -->
			update the 'childLess' field of the pedigree struct to point to it. */

	if (ped->childLess == NULL)	{
		if (currPerson->ch == 0) {	
			ped->childLess = currPerson;
		}
	}

	/* 8.2.	If a founder hasn't been found yet, and this person has no parents -->
			update the 'founder' field of the pedigree struct to point to it. */

	if (ped->founder == NULL) {		
		if (currPerson->pa == 0) {  
			ped->founder = currPerson;
		}
	}

	/* 9. Allocate memory for the struct that will hold the genetic information about
		  the person. */

	myMalloc(currPerson->loci, sizeof(locusPerson) * loci.numLoci); 

	/* 10.	Iterate over the loci. */

	for (locusInd = 0; locusInd < numLoci; locusInd++) {
		lociPlace = loci.lociOrder[locusInd];

		/* 10.1.	For each locus, read the genetic information about the person 
					according to the locus type. */

		switch (locusArr[lociPlace].type) {
			case affection :

				readAffection(currPerson, lociPlace, pedInd, indPerson, origPerson);
				break;

			case ALLELES :

				if ((!sexLink) || (currPerson->personSex == Female))
					readAlleles(currPerson, lociPlace, pedInd, indPerson,markersUsed[markerInd],origPerson);
				else
					readAllelesMale(currPerson, lociPlace, pedInd, indPerson, markersUsed[markerInd],origPerson);
				markerInd++;
				break;

			case null :

				if (lociPlace == loci.traitLociPos[1]) 
					readAffection(currPerson, lociPlace, pedInd, indPerson,origPerson);
				break;

		}
	}

	/* 11.	Zero the flags of the person's struct (used when traversing the pedigree). */

	currPerson->visited = FALSE;
	currPerson->done = FALSE;
	currPerson->updatedParents = FALSE;

	/* 12.	Allocate memory for the nodes that represent the person in the Bayesian network. */

	myMalloc(currPerson->locusVar[1], sizeof(node*) * numLoci);
	myMalloc(currPerson->selectorVar[1], sizeof(node*) * numLoci);
	myMalloc(currPerson->phenotypeVar, sizeof(node*) * numLoci);

	if (!sexLink) {
		myMalloc(currPerson->selectorVar[0], sizeof(node*) * numLoci);
		myMalloc(currPerson->locusVar[0], sizeof(node*) * numLoci);
	}
	else {
		currPerson->selectorVar[0] = NULL;

		if (currPerson->personSex == Female) {
			myMalloc(currPerson->locusVar[0], sizeof(node*) * numLoci);
		}
		else {
			currPerson->locusVar[0] = NULL;
		}
	}

	currPerson->genotypeVar = NULL;

	/* 13.	Get the rest of the line. */

	//fscanf(pedFile, "%*[^\n]");
	//fgetc(pedFile);
}


/******************************************************************************
*
*       Function        readAffection
*       Parameters		currPerson (the struct which holds the information 
*						about the person now being read), locusInd (the index of 
*						the locus being read, according to the input order),
*						pedInd (the index of the pedigree now being read),
*						personInd (the id of the person being read).
*       Output          none
*       Description		Reads the genetic data of some person in an affection 
*						type locus, from the pedigree file.
*       Comments		none
*
******************************************************************************/
void readAffection(person *currPerson, int locusInd, int pedInd, int personInd,PersonDataStruct *origPerson) 
{
	int				affectionS, penClass;
	locusPerson		*currLoc;

	currLoc = &(currPerson->loci[locusInd]);
	
	/* 1.	Read the affection status. */

	//fscanf(pedFile, "%d", &affectionS);
	affectionS = origPerson->status;
	if (affectionS == unknownAS) {

		/* 1.1.	The affection status is unknown. */

		currLoc->locusData.affectionL.affectionS = unknownAS;
		currLoc->unknown = TRUE;
	}
	else if (affectionS == affected) {

		/* 1.1.	The person is affected. */

		currLoc->locusData.affectionL.affectionS = affected;
		currLoc->unknown = FALSE;
	}
	else if (affectionS != 1) {

		/* 1.1.	The affection status is inappropriate. */

		inputErr("Inappropriate value (%d) for affection status in locus #%d of person #%d, pedigree #%d\n", 
				 affectionS, locusInd+1, personInd, pedInd);
	}
	else {

		/* 1.1.	The person is unaffected. */

		currLoc->locusData.affectionL.affectionS = unaffected;
		currLoc->unknown = FALSE;
	}

	/* 2.	In case of more than 1 liability class, read the liability class. */

	if (loci.numClasses > 1) {
		//fscanf(pedFile, "%d", &penClass);
		penClass = origPerson->liabClass;
		/* 2.1.	Input tests. */

		if ((penClass <= 0) || (penClass > loci.numClasses)) {
			inputErr("Inappropriate value (%d) for the liability class in locus #%d of person #%d, pedigree #%d\n", 
				     penClass, locusInd+1, personInd, pedInd);
		}
		currLoc->locusData.affectionL.liabilityClass = penClass;
	}
	else {
		currLoc->locusData.affectionL.liabilityClass = 1;
	}
}


/******************************************************************************
*
*       Function        readAlleles
*       Parameters		currPerson (the struct which will hold the data about the
*						person now being read), locusInd (the index of the locus 
*						being read, according to the input order), 
*						pedInd (the index of the pedigree now being read),
*						personInd (the id of the person being read).
*       Output          none
*       Description		Reads from the pedigree file the phenotype at the locus 
*						"locusInd" (which is of type ALLELES) of the person which 
*						is pointed to by "currPerson". 
*       Comments		none
*
******************************************************************************/
void readAlleles(person *currPerson, int locusInd, int pedInd, int personInd,int realMarkerInd,PersonDataStruct *origPerson) 
{
	int					allele1, allele2;
	locusPerson			*currLoc;
	halfLocusPerson		*firstAllele, *secondAllele;
	
	

	/* 1.	Read the 2 allele numbers that represent the person's phenotype at 
			the specific locus. */

	//fscanf(pedFile, "%d%d", &allele1, &allele2);
	allele1  = origPerson->markersAlleles[realMarkerInd].allele1;
	allele2  = origPerson->markersAlleles[realMarkerInd].allele2;
	/* 2.	Input Tests. */

	/* 2.1.	Check that the first allele number read is in the legal range. */

	if ((allele1 < 0) || (allele1 > loci.locusArr[locusInd].numAlleles)) {
		inputErr("There's an error in allele #1, locus #%d of person #%d, pedigree #%d.\n", 
			     locusInd+1, personInd, pedInd);
	}

	/* 2.2.	Check that the second allele number read is in the legal range. */

	if ((allele2 < 0) || (allele2 > loci.locusArr[locusInd].numAlleles)) {
		inputErr("There's an error in allele #2, locus #%d of person #%d, pedigree #%d.\n", 
			     locusInd+1, personInd, pedInd);
	}

	/* 2.3.	Check that either both allele numbers are known or neither one is known. */
	
	if (((allele1 == 0) || (allele2 == 0)) && (allele1 != allele2)) {
		inputErr("There's an error in locus #%d of person #%d, pedigree #%d. (1 allele is known and the other unknown)\n", 
				 locusInd+1, personInd, pedInd);
	}

	currLoc = &((currPerson->loci)[locusInd]);

	#if UNITE_VARS	
		currLoc->locusData.allelesL.uniteT = noUnite;
		currLoc->locusData.allelesL.certainAllele = NULL;
	#endif

	/* 3.	Update the person's data-structures according to the information read. */

	if ((allele1 == 0) && (allele2 == 0)) {		/* The phenotype is unknown at this locus. */

		currLoc->unknown = TRUE;
		currLoc->locusData.allelesL.paLocus = NULL;
		currLoc->locusData.allelesL.maLocus = NULL;
	}
	else {										/* The genotype is known up to phase. */
		if (allele1 > allele2)
			swap(&allele1, &allele2);

		currLoc->unknown = FALSE;

		/* 3.1.	Update the first possible allele of the locus, in the paternal & 
				maternal haplotype. */

		myMalloc(currLoc->locusData.allelesL.paLocus, sizeof(halfLocusPerson));
		firstAllele = currLoc->locusData.allelesL.paLocus;
		
		firstAllele->isCertain = FALSE;
		firstAllele->allele = allele1;

		myMalloc(currLoc->locusData.allelesL.maLocus, sizeof(halfLocusPerson));
		firstAllele = currLoc->locusData.allelesL.maLocus;
		
		firstAllele->isCertain = FALSE;
		firstAllele->allele = allele1;

		/* 3.2.	Update the second possible allele of the locus, in the paternal & 
				maternal haplotype. */

		if (allele1 != allele2) {	/* The person is hetrozygote in this locus. */
			myMalloc(secondAllele, sizeof(halfLocusPerson));

			secondAllele->allele = allele2;
			secondAllele->next = NULL;
			secondAllele->isCertain = FALSE;
			currLoc->locusData.allelesL.paLocus->next = secondAllele;

			myMalloc(secondAllele, sizeof(halfLocusPerson));

			secondAllele->allele = allele2;
			secondAllele->next = NULL;
			secondAllele->isCertain = FALSE;
			currLoc->locusData.allelesL.maLocus->next = secondAllele;
		}
		else {						/* The person is homozygote in this locus. */
			currLoc->locusData.allelesL.paLocus->next = NULL;
			currLoc->locusData.allelesL.maLocus->next = NULL;
			currLoc->locusData.allelesL.paLocus->isCertain = TRUE;
			currLoc->locusData.allelesL.maLocus->isCertain = TRUE;
		}
	}

	/* 4.	Initalize the possible selector values at this locus. */

	currLoc->locusData.allelesL.paSelectorVals[0] = 0;
	currLoc->locusData.allelesL.paSelectorVals[1] = 1;
	currLoc->locusData.allelesL.maSelectorVals[0] = 0;
	currLoc->locusData.allelesL.maSelectorVals[1] = 1;

#if PERFORM_ALLELE_RECODING
	currLoc->locusData.allelesL.maLocusTrans = NULL;
	currLoc->locusData.allelesL.paLocusTrans = NULL;
#endif
}


/******************************************************************************
*
*       Function        readAllelesMale
*       Parameters		currPerson (the struct which will hold the data about the
*						person now being read, locusInd (the index of the locus 
*						being read, according to the input order), 
*						pedInd (the index of the pedigree now being read),
*						personInd (the id of the person being read).
*       Output          none
*       Description		Reads from the pedigree file the phenotype at the locus 
*						"locusInd" (which is of type ALLELES) of the person which 
*						is pointed to by "currPerson". 
*       Comments		none
*
******************************************************************************/
void readAllelesMale(person *currPerson, int locusInd, int pedInd, int personInd, int realMarkerInd,PersonDataStruct *origPerson) 
{
	int					allele1, allele2;
	locusPerson			*currLoc;
	halfLocusPerson		*allele;
	
	

	/* 1.	Read the 2 allele numbers that represent the person's phenotype at 
			the specific locus. */

	//fscanf(pedFile, "%d%d", &allele1, &allele2);
	allele1  = origPerson->markersAlleles[realMarkerInd].allele1;
	allele2  = origPerson->markersAlleles[realMarkerInd].allele2;

	/* 2.	Input Tests. */

	/* 2.1.	Check that the first allele number read is in the legal range. */

	if ((allele1 < 0) || (allele1 > loci.locusArr[locusInd].numAlleles)) {
		inputErr("There's an error in allele #1, locus #%d of person #%d, pedigree #%d.\n", 
			     locusInd+1, personInd, pedInd);
	}

	/* 2.2.	Check that the second allele number read is in the legal range. */

	if ((allele2 < 0) || (allele2 > loci.locusArr[locusInd].numAlleles)) {
		inputErr("There's an error in allele #2, locus #%d of person #%d, pedigree #%d.\n", 
			     locusInd+1, personInd, pedInd);
	}

	/* 2.3.	Check that there's only a single allele known (or none). */
	
	if ((allele2 != 0) && (allele1 != allele2)) {
		inputErr("There's an error in locus #%d of person #%d, pedigree #%d. (males have a single allele)\n", 
			 	 locusInd+1, personInd, pedInd);
	}

	currLoc = &(currPerson->loci[locusInd]);

	#if UNITE_VARS	
		currLoc->locusData.allelesL.uniteT = noUnite;
		currLoc->locusData.allelesL.certainAllele = NULL;
	#endif

	/* 3.	Update the male's paternal locus to NULL since the X-allele is received from
	        his mother. */

	currLoc->locusData.allelesL.paLocus = NULL;

	#if PERFORM_ALLELE_RECODING
		currLoc->locusData.allelesL.paAllelsGroups = NULL;
		currLoc->locusData.allelesL.numPaAlleleGroups = 0;
	#endif

	currLoc->locusData.allelesL.paAlleles = NULL;

	/* 4.	Update the person's data-structures according to the information read. */

	if (allele1 == 0) {		/* The phenotype is unknown at this locus. */

		currLoc->unknown = TRUE;
		currLoc->locusData.allelesL.maLocus = NULL;
	}
	else {					/* The phenotype is known. */
		if (allele1 == 0)
			allele1 = allele2;

		currLoc->unknown = FALSE;

		/* 4.1.	Update the allele of the locus, in the maternal haplotype. */

		myMalloc(currLoc->locusData.allelesL.maLocus, sizeof(halfLocusPerson));
		allele = currLoc->locusData.allelesL.maLocus;
		
		allele->isCertain = TRUE;
		allele->allele = allele1;
		allele->next = NULL;
	}

	/* 5.	Initialize the possible selector values at this locus. */

	currLoc->locusData.allelesL.paSelectorVals[0] = 0;
	currLoc->locusData.allelesL.paSelectorVals[1] = 1;
	currLoc->locusData.allelesL.maSelectorVals[0] = 0;
	currLoc->locusData.allelesL.maSelectorVals[1] = 1;

	#if PERFORM_ALLELE_RECODING
       currLoc->locusData.allelesL.maLocusTrans = NULL;
	   currLoc->locusData.allelesL.paLocusTrans = NULL;
    #endif
}


/******************************************************************************
*
*       Function        initFlags
*       Parameters		ped (pointer to the pedigree whose flags need to be 
*						turned off).
*       Output          none
*       Description		Turns off the 'visited' & 'done' flags of each person in 
*						the pedigree pointed to by 'ped'.
*       Comments		none
*
******************************************************************************/
void initFlags(pedigree *ped)
{
	int		personInd;

	/* 1.	Iterate over the people in the pedigree. */

	for (personInd = 0; personInd < ped->numPeople; personInd++) {

		/* 1.1.	For each person, turn off its flags. */

		ped->people[personInd].visited = FALSE;
		ped->people[personInd].done = FALSE;
	}
}



/******************************************************************************
*
*       Function        updateFlagsLociAutosomal
*       Parameters		ped (pointer to the pedigree that needs to be updated).
*       Output          none
*       Description		Zeroes the 'visited' & 'done' flags of each person in 
*						the pedigree pointed to by 'ped', and updates all the
*						loci that no data exists about.
*       Comments		Used in case of an autosomal analysis.
*
******************************************************************************/
void updateFlagsLociAutosomal(pedigree *ped)
{
	person	*personPtr, *personPtrEnd;

	/* 1.	Iterate over the people in the pedigree. For each person: */

	for (personPtr = ped->people, personPtrEnd = personPtr + ped->numPeople;
			personPtr < personPtrEnd; personPtr++) {

		/* 1.1.	Turn off its flags. */
		
		personPtr->visited = FALSE;
		personPtr->done = FALSE;

		/* 1.2.	Update the loci of which we have no data about. */

		updateUnknownLoci(personPtr);
	}
}



/******************************************************************************
*
*       Function        updateUnknownLoci
*       Parameters		currPerson (the person that is being updated).
*       Output          none
*       Description		For each locus of 'currPerson' of which nothing is known
*						about, all the possible ALLELES of this locus are added
*						to its haplotypes.
*       Comments		none
*
******************************************************************************/
void updateUnknownLoci(person *currPerson) 
{
	int		locusInd, numLoci;
	locus	*locusPtr;

	numLoci = loci.numLoci;

	/* 1.	Iterate over all loci. */

	for (locusInd = 0, locusPtr = loci.locusArr; locusInd < numLoci; locusInd++,
				locusPtr++)

		/* 1.1.	If the locus is of type ALLELES, */

		if (locusPtr->type == ALLELES) {

			/* 1.1.1. If nothing is known about the paternal haplotype, update 
					  it with all possible ALLELES at this locus. */

			if (currPerson->loci[locusInd].locusData.allelesL.paLocus == NULL)
				addAllAlleles(currPerson, locusInd, TRUE);


			/* 1.1.2. If nothing is known about the maternal haplotype, update 
					  it with all possible ALLELES at this locus. */

			if (currPerson->loci[locusInd].locusData.allelesL.maLocus == NULL)
				addAllAlleles(currPerson, locusInd, FALSE);
		}
}


/******************************************************************************
*
*       Function        updateFlagsLociSexLinked
*       Parameters		ped (pointer to the pedigree that needs to be updated).
*       Output          none
*       Description		Zeroes the 'visited' & 'done' flags of each person in 
*						the pedigree pointed to by 'ped', and updates all the
*						loci that no data exists about.
*       Comments		Used in case of a sex-linked analysis.
*
******************************************************************************/
void updateFlagsLociSexLinked(pedigree *ped)
{
	person	*personPtr, *personPtrEnd;

	/* 1.	Iterate over the people in the pedigree. For each person: */

	for (personPtr = ped->people, personPtrEnd = personPtr + ped->numPeople;
			personPtr < personPtrEnd; personPtr++) {

		/* 1.1.	Turn off its flags. */
		
		personPtr->visited = FALSE;
		personPtr->done = FALSE;

		/* 1.2.	Update the loci of which we have no data about. */

		if (personPtr->personSex == Female)
			updateUnknownLoci(personPtr);
		else
			updateUnknownLociMale(personPtr);
	}
}



/******************************************************************************
*
*       Function        updateUnknownLociMale
*       Parameters		currPerson (the person that is being updated).
*       Output          none
*       Description		For each locus of 'currPerson' of which nothing is known
*						about, all the possible ALLELES of this locus are added.
*						to its maternal haplotype.
*       Comments		Used for males in case of a sex-linked analysis.
*
******************************************************************************/
void updateUnknownLociMale(person *currPerson) 
{
	int		locusInd, numLoci;
	locus	*locusPtr;

	numLoci = loci.numLoci;

	/* 1.	Iterate over all loci. */

	for (locusInd = 0, locusPtr = loci.locusArr; locusInd < numLoci; locusInd++,
			locusPtr++)

		/* 1.1.	If the locus is of type ALLELES, */

		if (locusPtr->type == ALLELES) {

			/* 1.1.1. If nothing is known about the maternal haplotype, update 
					  it with all possible ALLELES at this locus. */

			if (currPerson->loci[locusInd].locusData.allelesL.maLocus == NULL)
				addAllAlleles(currPerson, locusInd, FALSE);
		}
}



/******************************************************************************
*
*       Function        addAllAlleles
*       Parameters		currPerson (pointer to the person being updated),  
*						locusInd (the index of the locus being updated), 
*						isPaHaplotype (is TRUE if this is a paternal haplotype).
*       Output          none
*       Description		Adds all possible ALLELES to the paternal or maternal 
*						(according to the flag 'isPaHaplotype') haplotype of
*						locus no. 'locusInd' of person 'currPerson'.
*       Comments		none
*
******************************************************************************/
void addAllAlleles(person *currPerson, int locusInd, boolean isPaHaplotype)
{
	int					alleleInd, numAlleles;
	locus				*locusPtr;
	halfLocusPerson		*currLocus, *prevLocus;

	locusPtr =  &(loci.locusArr[locusInd]);

	/* 1.	Initialize the number of possible ALLELES of this locus. */

	numAlleles = locusPtr->numAlleles;

	/* 2.	Allocate memory for the struct that will store the first possible value
			of this locus. */

	if (isPaHaplotype == TRUE) {
		myMalloc(currPerson->loci[locusInd].locusData.allelesL.paLocus, 
					sizeof(halfLocusPerson));
		currLocus = currPerson->loci[locusInd].locusData.allelesL.paLocus;
	}
	else {
		myMalloc(currPerson->loci[locusInd].locusData.allelesL.maLocus, 
					sizeof(halfLocusPerson));
		currLocus = currPerson->loci[locusInd].locusData.allelesL.maLocus;
	}

	prevLocus = currLocus;

	/* 3.	Iterate over the rest of the possible ALLELES for this locus.  
			For each possible allele: */

	for (alleleInd = 1; alleleInd < numAlleles; alleleInd++) {

		if ((locusPtr->geneFrequencies)[alleleInd-1] == 0.0)
			continue;

		/* 3.1.	Allocate memory for the struct that will store the allele and 
				update its fields. */

		currLocus->allele = alleleInd;
		currLocus->isCertain = FALSE;
		prevLocus = currLocus;
		myMalloc(currLocus->next, sizeof(halfLocusPerson));
		currLocus = currLocus->next;
	}

	if ((locusPtr->geneFrequencies)[alleleInd-1] == 0.0) {
		prevLocus->next = NULL;
		myFree(currLocus);
	}
	else {
		currLocus->allele = alleleInd;
		currLocus->isCertain = FALSE;
		currLocus->next = NULL;
	}
}




/******************************************************************************
*
*       Function        checkInputValidity
*       Parameters		ped (the pedigree which is being checked for validity).
*       Output          none
*       Description		Checks for correspondence among all parent-child 
*						relationships.
*       Comments		none
*
******************************************************************************/
void checkInputValidity(pedigree *ped)
{
	person	*currPerson, *lastPerson;

	/* 1.	Iterate over the people in the pedigree. */

	for (currPerson = ped->people, lastPerson = currPerson + ped->numPeople; 
			currPerson < lastPerson; currPerson++) {
		
		/* 1.1.	If this person isn't a founder check for a correspondence between
				the person and his parents. */

		if (currPerson->pa != NULL) { 

			/* 1.1.1.	Check for correpondence with the father. */

			checkFatherChildMatch(currPerson, ped->pedNum);	

			/* 1.1.2.	Check for correpondence with the mother. */

			checkMotherChildMatch(currPerson, ped->pedNum);
		}
	}
}

/******************************************************************************
*
*       Function        checkFatherChildMatch
*       Parameters		ped (the pedigree which is being checked for validity),
*						currPerson (the person that is being checked currently).
*       Output          none
*       Description		Checks for correspondence among the father-child 
*						relationship of 'currPerson'.
*       Comments		none
*
******************************************************************************/
void checkFatherChildMatch(person *currPerson, int pedNum)
{
	person		*currChild;

	/* 1.	Check if the father is a male. */

	if ((currPerson->pa)->personSex != Male) {
		inputErr("The father of person #%d (%d) in pedigree %d is a female.",  
			 	 currPerson->pNum, (currPerson->pa)->pNum, pedNum);
	}

	/* 2.	Loook for 'currPerson' in it's father's list of children. */

	for (currChild = (currPerson->pa)->ch; currChild != NULL; currChild = currChild->NeSiPa) {
		if (currChild == currPerson)
			return;
	}

	/* 3.	In case there's no correspondence (the father doesn't point to the child), 
			an error message is printed. */

	inputErr("There is no correpondence between the father-child relation of person %d & person %d in pedigree %d.", 
			 currPerson->pNum, (currPerson->pa)->pNum, pedNum);
}



/******************************************************************************
*
*       Function        checkMotherChildMatch
*       Parameters		ped (the pedigree which is being checked for validity),
*						currPerson (the person that is being checked currently).
*       Output          none
*       Description		Checks for correspondence among the mother-child 
*						relationship of 'currPerson'.
*       Comments		none
*
******************************************************************************/
void checkMotherChildMatch(person *currPerson, int pedNum)
{
	person		*currChild;

	/* 1.	Check if the mother is a female. */

	if ((currPerson->ma)->personSex != Female) {
		inputErr("The mother of person #%d (%d) in pedigree %d is a male.",  
				 currPerson->pNum, (currPerson->ma)->pNum, pedNum);
	}

	/* 2.	Loook for 'currPerson' in it's mother's list of children. */

	for (currChild = (currPerson->ma)->ch; currChild != NULL; currChild = currChild->NeSiMa) {
		if (currChild == currPerson)
			return;
	}

	/* 3.	In case there's no correspondence (the mother doesn't point to the child), 
			an error message is printed. */

	inputErr("There is no correpondence between the mother-child relation of person %d & person %d in pedigree %d.", 
			 currPerson->pNum, (currPerson->ma)->pNum, pedNum);
}




/******************************************************************************
*
*       Function        freePedData
*       Parameters		numLoci (the number of loci in the analysis)
*       Output          none
*       Description		Frees the data-structures that were allocated dynamically
*						to hold all the data about the input pedigrees.
*       Comments		none
*
******************************************************************************/
void freePedData(int numLoci)
{
	int			pedInd, locusIndMap, locusIndDS, *mapToData; 
	pedigree	*tempPed, *currPed;
	person		*personPtr, *personPtrEnd;
	locus		*locusArr;

	locusArr = loci.locusArr;
	mapToData = loci.mapToData;
	
	/* 1.	Iterate over all pedigrees. */

	for (currPed = peds.head, pedInd = 0; pedInd < peds.numPeds; pedInd++) {

		/* 1.1.	For each pedigree, iterate over all people. */

		for (personPtr = currPed->people, personPtrEnd = personPtr + currPed->numPeople;
				personPtr < personPtrEnd; personPtr++) {

			/* 1.1.1.	For each person in each pedigree, iterate over all loci. */

			for (locusIndMap = 0; locusIndMap < numLoci; locusIndMap++) {

				locusIndDS = mapToData[locusIndMap];

				/* 1.1.1.1.	If this is an ALLELES type locus, free its arrays of ALLELES. */

				if (locusArr[locusIndDS].type == ALLELES) {
					#if UNITE_VARS	
						if (personPtr->loci[locusIndDS].locusData.allelesL.uniteT) {
							myFree(personPtr->loci[locusIndDS].locusData.allelesL.certainAllele);
						}
					#endif

					myFree(personPtr->loci[locusIndDS].locusData.allelesL.paAlleles);
					myFree(personPtr->loci[locusIndDS].locusData.allelesL.maAlleles);
				}
				
				/* 1.1.1.2.	Free the nodes of the Bayesian network allocated for this locus. */

				if (personPtr->locusVar[0]) {
					freeNode(&(personPtr->locusVar[0][locusIndDS]));
				}

				freeNode(&(personPtr->locusVar[1][locusIndDS]));

				if (personPtr->selectorVar[0]) {
					freeNode(&(personPtr->selectorVar[0][locusIndDS]));
				}

				freeNode(&(personPtr->selectorVar[1][locusIndDS]));
				freeNode(&(personPtr->phenotypeVar[locusIndDS]));
			}

			myFree(personPtr->loci);
			myFree(personPtr->locusVar[0]);
			myFree(personPtr->locusVar[1]);
			myFree(personPtr->selectorVar[0]);
			myFree(personPtr->selectorVar[1]);
			myFree(personPtr->phenotypeVar);
			myFree(personPtr->genotypeVar);
		}

		/* 1.2.	Free the array of people. */

		myFree(currPed->people);

		/* 1.3.	Free the pedigree struct and proceed to the next pedigree. */

		tempPed = currPed;
		currPed = currPed->nextPed;
		myFree(tempPed);
	}
}


/******************************************************************************
*
*       Function        freePedDataAR
*       Parameters		numLoci (the number of loci in the analysis)
*       Output          none
*       Description		Frees the data-structures that were allocated dynamically
*						to hold all the data about the input pedigrees.
*       Comments		Used in case where allele-recoding is performed.
*
******************************************************************************/
#if PERFORM_ALLELE_RECODING
void freePedDataAR(int numLoci)
{
	int			pedInd, locusIndMap, locusIndDS, numGroups, setInd, *mapToData; 
	pedigree	*tempPed, *currPed;
	person		*personPtr, *personPtrEnd;
	locus		*locusArr;

	locusArr = loci.locusArr;
	mapToData = loci.mapToData;

	currPed = peds.head;

	/* 1.	Iterate over all pedigrees. */

	for (pedInd = 0; pedInd < peds.numPeds; pedInd++) {

		/* 1.1.	For each pedigree, iterate over all people. */

		for (personPtr = currPed->people, personPtrEnd = personPtr + currPed->numPeople;
				personPtr < personPtrEnd; personPtr++) {

			/* 1.1.1.	For each person in each pedigree, iterate over all loci. */

			for (locusIndMap = 0; locusIndMap < numLoci; locusIndMap++) {

				locusIndDS = mapToData[locusIndMap];

				/* 1.1.1.1.	If this is an ALLELES type locus, free its arrays of ALLELES. */

				if (locusArr[locusIndDS].type == ALLELES) {

					#if UNITE_VARS	
						if (personPtr->loci[locusIndDS].locusData.allelesL.uniteT) {
							myFree(personPtr->loci[locusIndDS].locusData.allelesL.certainAllele);
						}
					#endif

					numGroups = personPtr->loci[locusIndDS].locusData.allelesL.numPaAlleleGroups;

					for (setInd = 0; setInd < numGroups; setInd++) {

						myFree(personPtr->loci[locusIndDS].locusData.allelesL.paAllelsGroups[setInd].alleleArr);
					}
					myFree(personPtr->loci[locusIndDS].locusData.allelesL.paAllelsGroups);

					numGroups = personPtr->loci[locusIndDS].locusData.allelesL.numMaAlleleGroups;

					for (setInd = 0; setInd < numGroups; setInd++) {

						myFree(personPtr->loci[locusIndDS].locusData.allelesL.maAllelsGroups[setInd].alleleArr);
					}
					myFree(personPtr->loci[locusIndDS].locusData.allelesL.maAllelsGroups);
				}
				
				/* 1.1.1.2.	Free the nodes of the Bayesian network allocated for this locus. */

				if (personPtr->locusVar[0]) {
					freeNode(&(personPtr->locusVar[0][locusIndDS]));
				}

				freeNode(&(personPtr->locusVar[1][locusIndDS]));

				if (personPtr->selectorVar[0]) {
					freeNode(&(personPtr->selectorVar[0][locusIndDS]));
				}

				freeNode(&(personPtr->selectorVar[1][locusIndDS]));
				freeNode(&(personPtr->phenotypeVar[locusIndDS]));
			}

			myFree(personPtr->loci);
			myFree(personPtr->locusVar[0]);
			myFree(personPtr->locusVar[1]);
			myFree(personPtr->selectorVar[0]);
			myFree(personPtr->selectorVar[1]);
			myFree(personPtr->phenotypeVar);
			myFree(personPtr->genotypeVar);
		}

		/* 1.2.	Free the array of people. */

		myFree(currPed->people);

		/* 1.3.	Free the pedigree struct and proceed to the next pedigree. */

		tempPed = currPed;
		currPed = currPed->nextPed;
		myFree(tempPed);
	}
}
#endif


/******************************************************************************
*
*       Function        deleteAlleles
*       Parameters      Locus (pointer to the single-locus haplotype struct of 
*						a person which needs to be freed).
*       Output          none
*       Description		Deletes the list of halfLocusPerson which 'Locus' point
*						to (a list of possible ALLELES).
*       Comments		none
*
******************************************************************************/
void deleteAlleles(halfLocusPerson *Locus)
{
	halfLocusPerson		*locus, *temp;

	locus = Locus;

	/* 1.	Iterate over the elements of the list and free them. */

	while (locus != NULL) {
		temp = locus;
		locus = locus->next;
		myFree(temp);
	}
}


/******************************************************************************
*
*       Function        printPeds
*       Parameters      none
*       Output          none
*       Description		Prints the known information about the people in all
*						pedigrees.
*       Comments		Used for debugging.
*
******************************************************************************/
#if DEBUG_STAT
void printPeds()
{
	FILE			*outFile;
	pedigree		*ped = peds.head;
	person			*currPerson;
	int				personInd, pedNum = 1, lociInd, numLoci;
	halfLocusPerson *locus;

	numLoci = loci.numLoci;

	outFile = fopen("outFile.txt", "w");

	for (; ped != NULL; ped = ped->nextPed, pedNum++) {
		fprintf(outFile,"pedigree no. %d\n", pedNum);

		for (personInd = 0; personInd < ped->numPeople; personInd++) {
			fprintf(outFile,"\n\nperson no. %d.\n", personInd+1);

			currPerson = &(ped->people[personInd]);

			for (lociInd = 0; lociInd < numLoci; lociInd++) 
				if (loci.locusArr[lociInd].type == ALLELES) {
					fprintf(outFile,"\nlocus no. %d:\n", lociInd);
					fprintf(outFile,"possible ALLELES in paternal haplotype:\n");

					locus = currPerson->loci[lociInd].locusData.allelesL.paLocus;
					
					for (;locus != NULL; locus = locus->next) 
						fprintf(outFile,"%d  ", locus->allele);

					fprintf(outFile,"\npossible ALLELES in maternal haplotype:\n");

					locus = currPerson->loci[lociInd].locusData.allelesL.maLocus;
					
					for (;locus != NULL; locus = locus->next) 
						fprintf(outFile,"%d  ", locus->allele);
				}
				else {
					if (currPerson->loci[lociInd].locusData.affectionL.affectionS == unaffected)
						fprintf(outFile,"unaffected\n");
					else if (currPerson->loci[lociInd].locusData.affectionL.affectionS == affected)
						fprintf(outFile,"affected\n");
					else
						fprintf(outFile,"unknown\n");
				}
		}
	}
	fclose(outFile);
}
#endif

#if DEBUG_STAT
void MyprintPeds()
{
	FILE			*outFile;
	pedigree		*ped = peds.head;
	person			*currPerson;
	int				personInd, pedNum = 1, lociInd, numLoci;
	halfLocusPerson *locus;
	
	int *loc, i;

	numLoci = loci.numLoci;

	outFile = fopen("myoutFile.txt", "w");

	for (; ped != NULL; ped = ped->nextPed, pedNum++) {
		fprintf(outFile,"pedigree no. %d\n", pedNum);

		for (personInd = 0; personInd < ped->numPeople; personInd++) {
			fprintf(outFile,"\n\nperson no. %d.\n", personInd+1);

			currPerson = &(ped->people[personInd]);

			for (lociInd = 0; lociInd < numLoci; lociInd++) 
				if (loci.locusArr[lociInd].type == ALLELES) {
					fprintf(outFile,"\nm%d:  ", lociInd);
					fprintf(outFile,"F{");

/*					locus = currPerson->loci[lociInd].locusData.allelesL.maLocusTrans;*/
					loc = currPerson->loci[lociInd].locusData.allelesL.paAlleles;
					for (i = 0; 
						i < currPerson->loci[lociInd].locusData.allelesL.numPaAlleleGroups;
						i++)
						fprintf(outFile,"%d  ", loc[i]);

					fprintf(outFile,"}\t");
					fprintf(outFile,"M{");

					loc = currPerson->loci[lociInd].locusData.allelesL.maAlleles;
					for (i = 0; 
						i < currPerson->loci[lociInd].locusData.allelesL.numMaAlleleGroups;
						i++)
						fprintf(outFile,"%d  ", loc[i]);
					fprintf(outFile,"}\n");	
				}
				else {
					if (currPerson->loci[lociInd].locusData.affectionL.affectionS == unaffected)
						fprintf(outFile,"unaffected\n");
					else if (currPerson->loci[lociInd].locusData.affectionL.affectionS == affected)
						fprintf(outFile,"affected\n");
					else
						fprintf(outFile,"unknown\n");
				}
		}
	}
	fclose(outFile);
}
#endif

