/*********************************************************************************/
/*  PedValueEliminationAutosomal.c											     */
/*  Author: Ma'ayan Fishelson													 */
/*********************************************************************************/

#include "sl_PedValueElimination.h"

/*------------------------------*/
/*		GLOBAL VARIABLES		*/
/*------------------------------*/

extern lociData loci;
int				pedNum;

/******************************************************************************
*
*       Function        ValueEliminate
*       Parameters      updateChild (the function which is used for updating a
*						child according to its parents), updateFlagsLociFunc 
*						(the function used for updating the flags of the people
*						in the pedigree), updateParents (the function used for
*						updating the parents according to their children).
*       Output          none
*       Description		Performs value elimination on all people in all 
*						pedigrees, by performing a downward traversal followed
*						by an upward traversal and another downward
*						traversal on each pedigree. At the end, a downward  
*						traversal in which impossible selector values are 
*						eliminated is performed.
*       Comments		none
*
******************************************************************************/
void valueEliminate(updateCh updateChild, updateFlagsLoci updateFlagsLociFunc, 
					updatePa updateParents)
{
	pedigree	*ped;
	int			numPeople;

	/* 1.	Iterate over the pedigrees. For each pedigree: */

	for (ped = peds.head, pedNum = 1; ped != NULL; ped = ped->nextPed, pedNum++) {

		numPeople = 0;

		/* 1.1.	Perform a downward traversal. */

		traverseDownUpdate(ped->founder, updateChild, &numPeople);
		
		/* 1.2. Check whether the pedigree is connected. If not, print an input warning. */

		if (numPeople < ped->numPeople) {
			inputWarning("Note that pedigree %d isn't connected.\nThe computations would be only performed on part of the pedigree connected to person #1.\n\n", ped->pedNum);
		}

		/* 1.3.	Zero the flags and update the unknown loci. */

		updateFlagsLociFunc(ped);

		/* 1.4.	Perform an upward traversal. */

		traverseUpUpdate(ped->childLess, updateParents);

		/* 1.5.	Zero the flags. */

		initFlags(ped);							

		/* 1.6.	Perform another downward traversal. */

		traverseDownUpdate(ped->founder, updateChild, &numPeople);

		/* 1.7.	Zero the flags. */

		initFlags(ped);	

		/* 1.8. Perform a downward traversal to update the selector values. */

		traverseDownSelector(ped->founder); 

		/* 1.9.	Zero the flags. */

		initFlags(ped);	

		/* 2.	Trace those cases where the four genetic loci variables of 
				two parents at some locus need to be eliminated jointly, and 
				update the relevant variables. */

		#if UNITE_VARS
			if (loci.sexLink == FALSE) {
				updateUniteVarsFourPed(ped);	
			}
		#endif
	}
}


/******************************************************************************
*
*       Function        traverseUp
*       Parameters      currPerson (the person that is being traversed currently),
*						updateChild (the function which is used for updating a
*						child according to its parents), numPeople (is updated 
*						with the number of people reached during the traversal
*						of the pedigree).
*       Output          none
*       Description		If 'curPerson' has no parents, calls traverseDownUpdate.
*						Otherwise:
*							- if one of currPerson's parents has been visited
*							  already, returns.
*							- else, keeps traversing up to currPerson's father.
*						
*       Comments		The idea is to traverse up until a founder has been 
*						reached. 
*
******************************************************************************/
void traverseUp(person *currPerson, updateCh updateChild, int *numPeople)
{
	/* 1.	If this isn't a founder. */

	if (currPerson->pa != NULL)	{ 

		/* 1.1.	If one of the parents has been 'visited' already, we'll get to 
				the person in another way. */

		if (((currPerson->pa)->visited == TRUE) || ((currPerson->ma)->visited == TRUE)) 
			return;

		/* 1.2	Otherwise, traverse up to this person's father. */

		traverseUp(currPerson->pa, updateChild, numPeople);
	}
	else					/* If this person is a founder, traverse down to it */
		traverseDownUpdate(currPerson, updateChild, numPeople);
}



/******************************************************************************
*
*       Function        traverseDownUpdate
*       Parameters      currPerson (the person that is being traversed currently),
*						updateChild (the function which is used for updating a
*						child according to its parents), numPeople (is updated 
*						with the number of people reached during the traversal
*						of the pedigree).
*       Output          none
*       Description		Updates all of the children of 'currPerson' according to 
*						it and calls traverseDown recursively on each child   
*						which has already been updated by the other parent as well.
*       Comments		none
*
******************************************************************************/
void traverseDownUpdate(person *currPerson, updateCh updateChild, int *numPeople)
{
	person		*ch;			
	sex			Sex;


	ch = currPerson->ch;
	Sex = currPerson->personSex;

	/* 1.	Mark this person as visited. */

	currPerson->visited = TRUE;				

	if (Sex == Male) {

		/* 2.	Iterate over the person's children. */

		for (; ch != NULL; ch = ch->NeSiPa) {	
			
			/* 2.1.	Update the child according to the father. */

			updateChild(currPerson, ch);	

			/* 2.2. If the child's mother hasn't been visited yet, traverse up to it
					and then traverse down to the child. 
					Otherwise, if we finished updating the child's mother, traverse
					down to the child. */

			if ((ch->ma)->visited == FALSE)	{
				traverseUp(ch->ma, updateChild, numPeople);
				traverseDownUpdate(ch, updateChild, numPeople);
			}
			else if ((ch->ma)->done == TRUE)
				traverseDownUpdate(ch, updateChild, numPeople);
		}
	}
	else {			/* (Sex == Female) */

		/* 2.	Iterate over the person's children. */

		for (; ch != NULL; ch = ch->NeSiMa) {

			/* 2.1.	Update the child according to the mother. */
			
			updateChild(currPerson, ch);
		
			/* 2.2. If the child's father hasn't been visited yet, traverse up to it
					and then traverse down to the child. 
					Otherwise, if we finished updating the child's father, traverse
					down to the child. */

			if ((ch->pa)->visited == FALSE)	{ 
				traverseUp(ch->pa, updateChild, numPeople);
				traverseDownUpdate(ch, updateChild, numPeople);
			}
			else if ((ch->pa)->done == TRUE)
				traverseDownUpdate(ch, updateChild, numPeople);
		}
	}

	/* 3.	Update that we finished updating this person, and increment by 1 the
			number of people reached during the pedigree traversal. */

	if (currPerson->done == FALSE) {
		(*numPeople)++;
		currPerson->done = TRUE;
	}
}



/******************************************************************************
*
*       Function        updateChildAutosomal
*       Parameters      parent (the person whose child needs to be updated
*						according to it), child (the person which needs to be 
*						updated).
*       Output          none
*       Description		Updates the child according to the parent.
*       Comments		Used in case of an autosomal analysis.
*
******************************************************************************/
void updateChildAutosomal(person* parent, person *child)
{
	int				locusInd;
	locus			*locusPtr;

	/* 1.	Iterate over the loci. */

	for (locusInd = 0, locusPtr = loci.locusArr; locusInd < loci.numLoci; locusInd++,
			locusPtr++) {

		/* 1.1.	If this is an ALLELES type locus, update the child according to
				the parent at this locus. */

		if (locusPtr->type == ALLELES)
			updateChildLociAutosomal(parent, child, locusInd);
	}
}


/******************************************************************************
*
*       Function        updateChildLociAutosomal
*       Parameters      parent (the person whose child needs to be updated
*						according to it), child (the person which needs to be 
*						updated), locusInd (the index of the locus being updated).
*       Output          none
*       Description		Updates the child according to the parent, at locus 
*						number 'locusInd'.
*       Comments		Used in case of an autosomal analysis.
*
******************************************************************************/
void updateChildLociAutosomal(person *parent, person *child, int locusInd)
{
	locusPerson			*parentLocPtr, *childLocPtr;
	halfLocusPerson		**chHaplo, **chHaploOther, *alleleCh, *alleleChOther, 
						*alleleParent, *prevAllele, *tempAllele; 
	boolean				possible;
	sex					parentSex;


	parentSex = parent->personSex;
	parentLocPtr = &(parent->loci[locusInd]);
	childLocPtr = &(child->loci[locusInd]);

	/* 1.	If we don't have any information about the parent, we can't 
			deduce anything about the child. */

	if ((((parentLocPtr->locusData).allelesL).paLocus == NULL) ||
		(((parentLocPtr->locusData).allelesL).maLocus == NULL))
		return;

	/* 2.	Otherwise, update pointers to the haplotypes of the child. */

	if (parentSex == Male) {
		 chHaplo = &((childLocPtr->locusData).allelesL).paLocus;
		 chHaploOther = &((childLocPtr->locusData).allelesL).maLocus;
	}
	else {
		 chHaplo = &((childLocPtr->locusData).allelesL).maLocus;

		 if ((loci.sexLink == FALSE) || 
			 ((loci.sexLink == TRUE) && (child->personSex == Female))) {

			 chHaploOther = &((childLocPtr->locusData).allelesL).paLocus;
		}
		 else {
			chHaploOther = NULL;	
		 }
	}

	/* 2.	Update the child according to the parent. */

	if (*chHaplo == NULL) {	
			
		/* 3.1.	If we don't know anything about the child in this locus, we
				add all the possible ALLELES of the parent to the child's
				relevant haplotype. */

		alleleParent = ((parentLocPtr->locusData).allelesL).paLocus;

		if (alleleParent != NULL)
			addAlleles(childLocPtr, alleleParent, parentSex);

		alleleParent = ((parentLocPtr->locusData).allelesL).maLocus;

		if (alleleParent != NULL) 
			addAlleles(childLocPtr, alleleParent, parentSex);

		/* 3.2.	If the child is left with only one possible allele in this
				haplotype, then he got this allele for certain from this parent. */
		
		if ((*chHaplo)->next == NULL) {
			(*chHaplo)->isCertain = TRUE;
		}
	}
	else {					
		/* 3.1.	If we know something about the child in this locus. */

		prevAllele = NULL;

		alleleCh = (*chHaplo);

		/* 3.2.	Iterate over all of the child's ALLELES.  For each one of them, */

		while (alleleCh != NULL) {		
			possible = FALSE;

			alleleParent = ((parentLocPtr->locusData).allelesL).paLocus;
			
			/* 3.2.1.	Iterate over the parent's possible paternal ALLELES and 
						check if the child's allele is among them. */
			
			for (; alleleParent!= NULL; alleleParent = alleleParent->next) { 
				if (alleleCh->allele == alleleParent->allele) {
					possible = TRUE;
					break;
				}
			}

			/* 3.2.2.	If the child's allele isn't among the parent's possible 
						paternal ALLELES, iterate over the parent's possible maternal 
						ALLELES and check if the child's allele is among them. */

			if (possible == FALSE) {
				alleleParent = ((parentLocPtr->locusData).allelesL).maLocus;

				for (; alleleParent!= NULL; alleleParent = alleleParent->next) { 
					if (alleleCh->allele == alleleParent->allele) {
						possible = TRUE;
						break;
					}
				}
			}

			/* 3.2.3.	If this allele isn't possible according to the parent,
						remove it from the list and preceed to the next allele
						in the list. */

			if (possible == FALSE) {	

				tempAllele = alleleCh;

				if (prevAllele == NULL) {
					(*chHaplo) = alleleCh->next;
				}
				else {
					prevAllele->next = alleleCh->next;
				}

				alleleCh = alleleCh->next;

				myFree(tempAllele);
			}
			else {
				prevAllele = alleleCh;
				alleleCh = alleleCh->next;
			}
		}

		alleleCh = (*chHaplo);

		/* 3.3.	Validity of input check. If the child has no possible ALLELES left
				in this haplotype, print an error message and exit. */

		if (alleleCh == NULL) {
			inputErr("Invalid value in locus #%d of person #%d, pedigree #%d.\n", 
						locusInd+1, child->pNum, pedNum);
		}

		/* 3.4.	If the child is left with only one possible allele in this
				haplotype, then he got this allele for certain from this parent. */

		if (alleleCh->next == NULL) {

			/* 3.4.1.	Mark this allele as certain. */

			alleleCh->isCertain = TRUE;

			/* 3.4.2.	If the child's genotype is known up to phase in this locus,
						we need to delete the other allele from the other parent
						(since we know for certain the allele received from this parent). */

			#if UNITE_VARS	
				if (((childLocPtr->unknown == FALSE) || (childLocPtr->locusData.allelesL.uniteT == twoAll)) 
						&& (chHaploOther != NULL)) {
			#else
				if ((childLocPtr->unknown == FALSE) && (chHaploOther != NULL)) {
			#endif

				alleleChOther = *chHaploOther;

				if (alleleChOther->next != NULL) {

					if (alleleChOther->allele == alleleCh->allele) {
						*chHaploOther = alleleChOther->next;
						myFree(alleleChOther);
					}
					else {
						myFree(alleleChOther->next);
						(*chHaploOther)->next = NULL;
					}
					(*chHaploOther)->isCertain = TRUE;

					#if UNITE_VARS
						(childLocPtr->locusData).allelesL.uniteT = noUnite;
					#endif
				}
			}
		}
	}
}



/******************************************************************************
*
*       Function        updateChildSexLinked
*       Parameters      parent (the person whose child needs to be updated
*						according to it), child (the person which needs to be 
*						updated).
*       Output          none
*       Description		Updates the child according to the parent.
*       Comments		Used in the case of sex-linked analysis.
*
******************************************************************************/
void updateChildSexLinked(person *parent, person *child)
{
	int		locusInd;
	locus	*locusPtr;


	/* 1.	If the child is a male and the parent is his father, then he can't
			be updated by him (relating markers on the X-chromosome). */

	if ((parent->personSex == Male) && (child->personSex == Male))
		return;

	/* 2.	Iterate over the loci. */

	for (locusInd = 0, locusPtr = loci.locusArr; locusInd < loci.numLoci; locusInd++,
			locusPtr++) {

		/* 2.1.	If this is an ALLELES type locus, update the child according to
				the parent. */

		if (locusPtr->type == ALLELES) {

			if (parent->personSex == Male)
				updateChildLociSexLinked(parent, child, locusInd);
			else
				updateChildLociAutosomal(parent, child, locusInd);
		}
	}
}


/******************************************************************************
*
*       Function        updateChildLociSexLinked
*       Parameters		parent (the person whose child needs to be updated
*						according to it), child (the person which needs to be 
*						updated), locusInd (the index of the locus being updated).
*       Output          none
*       Description		Updates the child according to the father, 
*						at locus no. 'locusInd'.
*       Comments		Used in the case of sex-linked analysis.
*
******************************************************************************/
void updateChildLociSexLinked(person *parent, person *child, int locusInd)
{
	halfLocusPerson		*alleleCh, *alleleChOther, *alleleParent, *prevAllele, 
						*tempAllele; 
	boolean				possible;

	/* 1.	If we don't have any information about the father, we can't 
			deduce anything about the child. */

	if ((((parent->loci[locusInd]).locusData).allelesL).maLocus == NULL)
		return;

	/* 2.	Otherwise. */

	alleleCh = (((child->loci[locusInd]).locusData).allelesL).paLocus;

	if (alleleCh == NULL) {

		 /* 2.1.	If we don't know anything about the child in this locus, we
					add all the possible ALLELES of the father to the child's
					paternal haplotype. */

		alleleParent = (((parent->loci[locusInd]).locusData).allelesL).maLocus;

		if (alleleParent != NULL) 
			addAlleles(&(child->loci[locusInd]), alleleParent, Male);

		/* 2.2.	If the child is left with only one possible allele in his paternal
				haplotype, then he got thus allele for certain from his father. */

		if ((((child->loci[locusInd]).locusData).allelesL).paLocus->next == NULL)
			(((child->loci[locusInd]).locusData).allelesL).paLocus->isCertain = TRUE;
	}
	else {
		 /* 2.1.	If we know something about the child in this locus. */

		prevAllele = NULL;

		 /* 2.2.	Iterate over all of the child's ALLELES.  */

		while (alleleCh != NULL) {	
			 possible = FALSE;

			 /* 2.2.1.	Iterate over all the father's possible ALLELES and 
						check if the child's allele is among them. */
	
			 alleleParent = (((parent->loci[locusInd]).locusData).allelesL).maLocus;
			 
			 for (; alleleParent!= NULL; alleleParent = alleleParent->next) 
				 if (alleleCh->allele == alleleParent->allele) {
					 possible = TRUE;
					 break;
				 }

			/* 2.2.1.	If this allele isn't possible according to the father,
						remove it from the list and preceed to the next allele
						in the list. */

			if (possible == FALSE) {

				tempAllele = alleleCh;

				if (prevAllele == NULL) {
					(((child->loci[locusInd]).locusData).allelesL).paLocus = alleleCh->next;
				}
				else {
					prevAllele->next = alleleCh->next;
				}

				alleleCh = alleleCh->next;

				myFree(tempAllele);
			}
			else {
				prevAllele = alleleCh;
				alleleCh = alleleCh->next;
			}
		}

		/* 2.3.	Validity of input check. If the child has no possible ALLELES left
				in his paternal haplotype, print an error message and exit. */

		if ((((child->loci[locusInd]).locusData).allelesL).paLocus == NULL) {
			inputErr("Invalid value in locus #%d of person #%d, pedigree #%d.\n", 
					  locusInd+1, child->pNum, pedNum);
		}

		alleleCh = (((child->loci[locusInd]).locusData).allelesL).paLocus;

		/* 2.4.	If the child is left with only one possible allele in his paternal
				haplotype, then he got this allele for certain from his father. */

		if (alleleCh->next == NULL) {

			/* 2.4.1.	Mark this allele as certain. */

			alleleCh->isCertain = TRUE;

			/* 2.4.2.	If the child's genotype is known up to phase in this locus,
						we need to delete the other allele from the mother (since we 
						know for certain the allele received from the father). */

			#if UNITE_VARS
				if (((child->loci[locusInd]).unknown == FALSE) || 
						((((child->loci[locusInd]).locusData).allelesL).uniteT == twoAll)) {	
			#else
				if ((child->loci[locusInd]).unknown == FALSE) {	
			#endif

				alleleChOther = (((child->loci[locusInd]).locusData).allelesL).maLocus;
						
				if (alleleChOther->next != NULL) {

					if (alleleChOther->allele == alleleCh->allele) {
						(((child->loci[locusInd]).locusData).allelesL).maLocus = alleleChOther->next;
						myFree(alleleChOther);
					}
					else {
						myFree(alleleChOther->next);
						(((child->loci[locusInd]).locusData).allelesL).maLocus->next = NULL;
					}
					(((child->loci[locusInd]).locusData).allelesL).maLocus->isCertain = TRUE;

					#if UNITE_VARS
						(((child->loci[locusInd]).locusData).allelesL).uniteT = noUnite;
					#endif
				}
			}
		}
	}
}


/******************************************************************************
*
*       Function        addAlleles
*       Parameters      childLoci (the child's locus), alleleParent (the 
*						parent's haplotype), Sex (the sex of the parent).
*       Output          none
*       Description		This function adds all the ALLELES in the haplotype
*						of the parent, 'alleleParent', to the child's relevant
*						haplotype (according to 'Sex') in the locus 'childLoci'.
*       Comments		none
*
******************************************************************************/
void addAlleles(locusPerson *childLoci, halfLocusPerson *alleleParent, sex Sex) 
{
	halfLocusPerson		*alleleCh, *prevAlleleCh, *newAllele;

	if (Sex == Male) {

		/* 1.	If the child's list of ALLELES in the paternal haplotype is 
				empty: */

		if (childLoci->locusData.allelesL.paLocus == NULL) {

			/* 1.1.	Allocate memory for a struct that will hold the first
					possible value for this haplotye */
			
			myMalloc((childLoci->locusData).allelesL.paLocus, sizeof(halfLocusPerson));
			alleleCh = childLoci->locusData.allelesL.paLocus;

			/* 1.2.	Add all the father's ALLELES to the paternal haplotype
					of the child. */

			addParentAlleles(alleleParent, alleleCh);
			return;
		}
		else 
			alleleCh = childLoci->locusData.allelesL.paLocus;
	}
	else {			/* (sex == Female) */

		/* 1.	If the child's list of ALLELES in the maternal haplotype is 
				empty: */

		if (childLoci->locusData.allelesL.maLocus == NULL) {

			/* 1.1.	Allocate memory for the struct that will hold the first
					possible value for this haplotye */

			myMalloc((childLoci->locusData).allelesL.maLocus, sizeof(halfLocusPerson));
			alleleCh = childLoci->locusData.allelesL.maLocus;

			/* 1.2.	Add all the mother's ALLELES to the maternal haplotype
					of the child. */

			addParentAlleles(alleleParent, alleleCh);
			return;
		}
		else 
			alleleCh = childLoci->locusData.allelesL.maLocus;
	}

	prevAlleleCh = NULL;

	/* 2.	In case we already added some ALLELES to this locus (from the other 
			haplotype of the parent), iterate over the possible allele values of
			the parent. For each one: */

	for (; alleleParent != NULL; alleleParent = alleleParent->next) {

		/* 2.1.		Iterate over the child's list of possible ALLELES and check if
					the allele already exists in the list. */

		for (;(alleleCh != NULL) && (alleleCh->allele < alleleParent->allele); 
			 prevAlleleCh = alleleCh, alleleCh = alleleCh->next);

		 /* 2.2.	If this allele already appears in the list, don't duplicate it. */

		 if ((alleleCh != NULL) && (alleleCh->allele == alleleParent->allele))
			 continue;

		 /* 2.3.	Otherwise, allocate memory for the struct that will hold the new 
					allele and insert it before alleleCh. */

		 myMalloc(newAllele, sizeof(halfLocusPerson));
		 newAllele->isCertain = FALSE;

		 if (prevAlleleCh == NULL) {	/* the allele needs to be the first. */
			
			newAllele->allele = alleleCh->allele;
			alleleCh->allele = alleleParent->allele;
			newAllele->next = alleleCh->next;			
			alleleCh->next = newAllele;					
		 }
		 else {
			newAllele->allele = alleleParent->allele;
			prevAlleleCh->next = newAllele;				
			newAllele->next = alleleCh;					
			alleleCh = newAllele;						
		 }
	}
}


/******************************************************************************
*
*       Function        addParentAlleles
*       Parameters      alleleParent (the parent's haplotype), alleleCh (the 
*						child's haplotype).
*       Output          none
*       Description		This function adds all the ALLELES from the haplotype
*						of the parent, 'alleleParent', to the child's 
*						haplotype 'alleleCh'.
*       Comments		none
*
******************************************************************************/
void addParentAlleles(halfLocusPerson *alleleParent, halfLocusPerson *alleleCh)
{
	/* 1.1.	Iterate over the possible ALLELES according to the 
			parent's haplotype, allocate structs to store these 
			ALLELES and initialize them. */

	for (;alleleParent->next != NULL; alleleParent = alleleParent->next) {
		alleleCh->allele = alleleParent->allele;
		myMalloc(alleleCh->next, sizeof(halfLocusPerson));
		alleleCh->isCertain = FALSE;
		alleleCh = alleleCh->next;
	}

	/* 1.2.	Update the fields of the struct that stores the last allele. */

	alleleCh->allele = alleleParent->allele;	
	alleleCh->isCertain = FALSE;
	alleleCh->next = NULL;
}



/******************************************************************************
*
*       Function        traverseUpUpdate
*       Parameters      currPerson (the person that is being traversed currently),
*						updateParents (the function used for updating the parents
*						according to their children).
*       Output          none
*       Description		This function traverses all the children of 'currPerson'.
*						If all children have been updated already (according to
*						their children), it updates 'currPerson' according to
*						them and then traverses up to 'currPerson's parents (if
*						they exist).
*       Comments		none
*
******************************************************************************/
void traverseUpUpdate(person *currPerson, updatePa updateParents)
{
	person		*ch;			
	sex			Sex;

	
	ch = currPerson->ch;	
	Sex = currPerson->personSex;

	/* 1.	Mark this person as visited. */

	currPerson->visited = TRUE;				

	if (Sex == Male) {						
		
		/* 2.	Iterate over all children of this person. */

		for (; ch != NULL; ch = ch->NeSiPa) {		
			
			/* 2.1.	If the child hasn't been visited yet, traverse down to it. 
					From this child we'll traverse back up to the parent later on. */

			if (ch->done == FALSE) {				
				if (ch->visited == FALSE) 				
					traverseDown(ch, updateParents);
				return;									
			}
		}
	}
	else {			/* (Sex == Female) */

		/* 2.	Iterate over all children of this person. */

		for (; ch != NULL; ch = ch->NeSiMa) {			
			
			/* 2.1.	If the child hasn't been visited yet, traverse down to it. 
					From this child we'll traverse back up to the parent later on. */

			if (ch->done == FALSE) {
				if (ch->visited == FALSE)			
					traverseDown(ch, updateParents);
				return;
			}
		}
	}

	ch = currPerson->ch;

	/* 3.	All the children have been updated already. Update the parents now.*/

	if (Sex == Male) {

		/* 3.1.	Iterate over all of this person's children. */

		for (; ch != NULL; ch = ch->NeSiPa) {		
			
			/* 3.1.1.	If the parents haven't been updated by this child yet, 
						update them. */

			if ((ch->ma)->done == FALSE)	
				updateParents(currPerson, ch->ma, ch);
		}
	}
	else {		/* (sex == Female) */
		for (; ch != NULL; ch = ch->NeSiMa) {	
			if ((ch->pa)->done == FALSE)
				updateParents(ch->pa, currPerson, ch);	
		}
	}

	/* 4.	Update this person's ALLELES. */

	if (loci.sexLink && (currPerson->personSex == Male)) {
		updateAllelesMale(currPerson);
	}
	else {
		updateAlleles(currPerson);
	}

	/* 5.	Mark this person as done. */

	currPerson->done = TRUE;			

	/* 6.	If currPerson has a father which hasn't been visited yet, go up. */

	if ((currPerson->pa != NULL) && ((currPerson->pa)->done == FALSE))	
		traverseUpUpdate(currPerson->pa, updateParents);

	/* 7.	If currPerson has a mother which hasn't been visited yet, go up. */

	if ((currPerson->ma != NULL) && ((currPerson->ma)->done == FALSE))	
		traverseUpUpdate(currPerson->ma, updateParents);
}



/******************************************************************************
*
*       Function        traverseDown
*       Parameters      currPerson (the person that is being traversed currently),
*						updateParents (the function used for updating the parents
*						according to their children).
*       Output          none
*       Description		If currPerson has no children, we traverse up to him.
*						Otherwise, if its child hasn't been visited yet, we 
*						traverse down to the child. Otherwise, we return.
*       Comments		none
*
******************************************************************************/
void traverseDown(person *currPerson, updatePa updateParents)
{
	person		*ch;			

	ch = currPerson->ch;


	/* 1.	If this person hasn't got any children, we traverse up to him.
			Otherwise, if his child hasn't been visited yet, we traverse
			down to it.  Otherwise we return (we'll reach this person later,
			by another path). */

	if (ch == NULL)					
		traverseUpUpdate(currPerson, updateParents);
	else {
		if (ch->visited == FALSE) 	
			traverseDown(ch, updateParents);
		else						
			return;
	}
}



/******************************************************************************
*
*       Function        updateParentsAutosomal
*       Parameters      pa, ma, ch (father, mother & child)
*       Output          none
*       Description		Updates the parents according to the child.
*       Comments		Used for autosomal analysis.
*
******************************************************************************/
void updateParentsAutosomal(person *pa, person *ma, person *ch)
{
	int					locusInd;
	locus				*locusPtr;
	halfLocusPerson		*alleleCh;

	
	/* 1.	Iterate over all the loci. */

	for (locusInd = 0, locusPtr = loci.locusArr; locusInd < loci.numLoci; locusInd++,
			locusPtr++) {

		/* 1.1.	If the locus is of type 'ALLELES', update the parents according 
				to the child in the specific locus. */

		if (locusPtr->type == ALLELES) {

			/* 1.1.1.	If we know for certain which allele the child got 
						from his father, update the father accordingly. */

			alleleCh = (((ch->loci[locusInd]).locusData).allelesL).paLocus;

			if (alleleCh->next == NULL) {	
				updateParentLoc(alleleCh, &(pa->loci[locusInd]));
			}

			/* 1.1.2.	If we know for certain which allele the child got 
						from his mother, update the mother accordingly. */

			alleleCh = (((ch->loci[locusInd]).locusData).allelesL).maLocus;

			if (alleleCh->next == NULL) {	
				updateParentLoc(alleleCh, &(ma->loci[locusInd]));
			}
		}
	}
}



/******************************************************************************
*
*       Function        updateParentsSexLinked
*       Parameters      pa, ma, ch (father, mother & child)
*       Output          none
*       Description		Updates the parents according to the child.
*       Comments		Used for sex-linked analysis.
*
******************************************************************************/
void updateParentsSexLinked(person *pa, person *ma, person *ch)
{
	int					locusInd;
	locus				*locusPtr;
	halfLocusPerson		*alleleCh, *allelePa;
	
	/* 1.	Iterate over all loci. */

	for (locusInd = 0, locusPtr = loci.locusArr; locusInd < loci.numLoci; locusInd++,
			locusPtr++) {

		/* 1.1.	If the locus is of type ALLELES, update the parents according 
				to the child in the specific locus. */

		if (locusPtr->type == ALLELES) {

			/* 1.1.1.	If the child is a female then update her father 
						according to her. */

			if (ch->personSex == Female) {

				alleleCh = (((ch->loci[locusInd]).locusData).allelesL).paLocus;

				/* 1.1.1.1.	If we know for certain which allele the daughter 
							got from her father: */

				if (alleleCh->next == NULL) {	

					/* 1.1.1.1.1.	Update the maternal haplotype of the father 
									according to his daughter. */

					allelePa = (((pa->loci[locusInd]).locusData).allelesL).maLocus;

					updateHalfParentLocus(alleleCh->allele, allelePa);
				}
			}

			alleleCh = (((ch->loci[locusInd]).locusData).allelesL).maLocus;

			/* 2.	If we know for certain which allele the child got from his 
					mother, update the mother accordingly. */

			if (alleleCh->next == NULL) {	
				updateParentLoc(alleleCh, &(ma->loci[locusInd]));
			}
		}
	}
}


/******************************************************************************
*
*       Function        updateParentLoc
*       Parameters      alleleCh (allele of the child), parentLoc (the locus
*						of the parent which needs to be updated).
*       Output          none
*       Description		Updates the parent according to the child at a certain 
*						locus.
*       Comments		none
*
******************************************************************************/
void updateParentLoc(halfLocusPerson *alleleCh, locusPerson *parentLoc)
{
  	halfLocusPerson		*alleleParent;
  

	/* 1.1.	Update the paternal haplotype of the parent according to the child. */

	alleleParent = ((parentLoc->locusData).allelesL).paLocus;

	updateHalfParentLocus(alleleCh->allele, alleleParent);

	/* 1.2.	Update the maternal haplotype of the parent according to the child. */

	alleleParent = ((parentLoc->locusData).allelesL).maLocus;

	updateHalfParentLocus(alleleCh->allele, alleleParent);
}



/******************************************************************************
*
*       Function        updateHalfParentLocus
*       Parameters      alleleCh (allele of the child), alleleParent (points 
*						to the list of possible ALLELES of the parent).
*       Output          none
*       Description		If the list that 'alleleParent' points to includes the
*						allele 'alleleCh', it is marked as a certain allele.
*       Comments		none
*
******************************************************************************/
void updateHalfParentLocus(int alleleCh, halfLocusPerson *alleleParent)
{
	/* 1.	Iterate over the allele list of the parent. */

	for (; alleleParent != NULL ; alleleParent = alleleParent->next) {

		/* 1.1.	If the allele of the child is found in the list, it is marked 
				as certain. */

		if (alleleParent->allele == alleleCh) {
			alleleParent->isCertain = TRUE;
			return;
		}
	}
}



/******************************************************************************
*
*       Function        updateAlleles
*       Parameters      person (the person which is being updated).
*       Output          none
*       Description		Updates the person according to the certain ALLELES that
*						have been marked previously when this person was 
*						updated by his children. 
*       Comments		none
*
******************************************************************************/
void updateAlleles(person *Person)
{
	locus			*locusPtr;
	locusPerson		*locPersonPtr, *locPersonPtrEnd;

	/* 1.	Iterate over all loci. In case of an ALLELES type locus, update the
			haplotype of the person according to the certain ALLELES marked
			previously. */
	
	for (locPersonPtr = Person->loci, locPersonPtrEnd = locPersonPtr + loci.numLoci, 
			locusPtr = loci.locusArr; locPersonPtr < locPersonPtrEnd; locPersonPtr++, 
			locusPtr++) {

		if (locusPtr->type == ALLELES) {
			if ((((locPersonPtr->locusData).allelesL).paLocus->next != NULL) ||
				(((locPersonPtr->locusData).allelesL).maLocus->next != NULL)) {

				removeAllelesLocus(locPersonPtr);

				#if UNITE_VARS
					if (locPersonPtr->unknown == FALSE && !RUN_CREATE_BAYES) {
						locPersonPtr->locusData.allelesL.uniteT = twoAll;	
					}
				#endif
			}
		}
	}
}



/******************************************************************************
*
*       Function        removeAllelesLocus
*       Parameters      Locus (pointer to the locus being updated currently)
*       Output          none
*       Description		Updates 'Locus' according to the certain ALLELES marked
*						previously (in some cases, it's possible to erase
*						some of the ALLELES from the list).
*       Comments		none
*
******************************************************************************/
void removeAllelesLocus(locusPerson *Locus)
{
	int					certainPa[2], certainMa[2], i, j;
	
	/* 1.	Update the array certainPa to hold the "certain" ALLELES on the 
			paternal haplotype (there can be at most 2 such ALLELES). */

	i = updateCertainAlleles(((Locus->locusData).allelesL).paLocus, certainPa);

	/* 2.	Update the array certainMa to hold the "certain" ALLELES on the 
			maternal haplotype (there can be at most 2 such ALLELES). */

	j = updateCertainAlleles(((Locus->locusData).allelesL).maLocus, certainMa);

	/* 3.	Check how many ALLELES have been marked as "certain" in each haplotype and
			act accordingly. */

	if (i == 1) {			/* 1 allele has been marked as certain in the paternal haplotype.*/
		if (j == 1) {		/* 1 allele has been marked as certain in the maternal haplotype.*/
			if (certainPa[0] == certainMa[0]) {

				#if UNITE_VARS
					if ((((Locus->locusData).allelesL).paLocus->next != NULL) && 
							(((Locus->locusData).allelesL).maLocus->next != NULL)
							&& !RUN_CREATE_BAYES) {

						Locus->locusData.allelesL.uniteT = oneAll;		
						myMalloc(Locus->locusData.allelesL.certainAllele, sizeof(int));
						Locus->locusData.allelesL.certainAllele[0] = certainPa[0];
					}
				#endif

				return;
			}
		}
		else if (j == 2) {	/* 2 ALLELES have been marked as certain in the maternal haplotype.*/
			if (certainPa[0] == certainMa[0]) 
				certainMa[0] = certainMa[1];
		}
	}
	else if (i == 2) {		/* 2 ALLELES have been marked as certain in the paternal haplotype.*/
		if (j == 1) {		/* 1 allele has been marked as certain in the maternal haplotype.*/
			if (certainPa[0] == certainMa[0])
				certainPa[0] = certainPa[1];
		}
		else if (j == 2) {	/* 2 ALLELES have been marked as certain in the maternal haplotype.*/

			/* 3.1.	Remove all ALLELES except for the 2 marked as certain,
					from the paternal haplotype, and if necessary reorder the ALLELES, 
					so that the smaller one is first. */

			updateHaplo(((Locus->locusData).allelesL).paLocus, certainPa);

			/* 3.2.	Remove all ALLELES except for the 2 marked as certain,
					from the maternal haplotype, and if necessary reorder the ALLELES, 
					so that the smaller one is first. */

			updateHaplo(((Locus->locusData).allelesL).maLocus, certainMa);

			#if UNITE_VARS
			if (!RUN_CREATE_BAYES){
				Locus->locusData.allelesL.uniteT = twoAll;	
			}
			#endif

			return;
		}
	}

	/* 5.	Remove all ALLELES except the certain ones from the maternal & 
			paternal haplotypes. */

	if (i != 0) {	
		deleteAlleles(((Locus->locusData).allelesL).paLocus->next);

		((Locus->locusData).allelesL).paLocus->next = NULL;
		((Locus->locusData).allelesL).paLocus->allele = certainPa[0];
	}

	if (j != 0) {
		deleteAlleles(((Locus->locusData).allelesL).maLocus->next);

		((Locus->locusData).allelesL).maLocus->next = NULL;
		((Locus->locusData).allelesL).maLocus->allele = certainMa[0];
	}
}


/******************************************************************************
*
*       Function        updateCertainAlleles
*       Parameters      personAllele (pointer to the allele list at one halpotype
*						of the person being updated), certainArr (this array
*						will be updated with the ALLELES that are certain).
*       Output          none
*       Description		Updates 'certainArr' to hold the certain ALLELES that
*						have been marked previously when this person has been 
*						updated by his children. 
*       Comments		none
*
******************************************************************************/
int	updateCertainAlleles(halfLocusPerson *personAllele, int *certainArr)
{
	int		i;

	i = 0;

	/* 1.	Iterate over the allele list of this person, and update the array
			'certainArr' to hold the "certain" ALLELES on this person's 
			haplotype. */

	for (; personAllele != NULL ; personAllele = personAllele->next) {
		if (personAllele->isCertain == TRUE) {
			certainArr[i] = personAllele->allele;
			i++;

			if (i == 2)
				break;
		}
	}

	return i;
}



/******************************************************************************
*
*       Function        updateCertainAlleles
*       Parameters      haplo (pointer to the allele list at one halpotype
*						of the person being updated), certainArr (this array
*						holds the ALLELES that are certain).
*       Output          none
*       Description		Remove all the ALLELES except for the 2 ALLELES marked 
*						as certain from the relevant haplotype.
*       Comments		none
*
******************************************************************************/
void updateHaplo(halfLocusPerson *haplo, int *certainArr)
{

	/* 1.	Remove all the ALLELES except for the 2 ALLELES marked as certain,
			from the haplotype. */

	deleteAlleles(haplo->next->next);

	haplo->next->next = NULL;

	/* 2.	If necessary reorder the ALLELES, so that the smaller one is first. */

	if (certainArr[0] > certainArr[1]) 
		swap(&certainArr[0], &certainArr[1]);

	haplo->allele = certainArr[0];
	haplo->next->allele = certainArr[1];
}


/******************************************************************************
*
*       Function        updateAllelesMale
*       Parameters      person (the person which is being updated).
*       Output          none
*       Description		Updates the person according to the certain ALLELES that
*						have been marked previously when this person has been 
*						updated by his children. 
*       Comments		Used for males in case of a sex-linked analysis.
*
******************************************************************************/
void updateAllelesMale(person *Person)
{
	int		locusInd;
	locus	*locusPtr;

	/* 1.	Iterate over all loci. In case of an ALLELES type locus in which
			we known for certain the allele in the maternal haplotype, remove 
			the rest of the ALLELES from the haplotype. */
	
	for (locusInd = 0, locusPtr = loci.locusArr; locusInd < loci.numLoci; locusInd++,
			locusPtr++) {
		if (locusPtr->type == ALLELES)
			if (((Person->loci[locusInd].locusData).allelesL).maLocus->next != NULL) {
				removeAllelesLocusMale(&Person->loci[locusInd]);
			}
	}
}


/******************************************************************************
*
*       Function        removeAllelesLocusMale
*       Parameters      Locus (pointer to the locus being updated currently)
*       Output          none
*       Description		Updates 'Locus' according to the certain ALLELES marked
*						previously (in some cases, it's possible to erase
*						some of the ALLELES from the list).
*       Comments		Used for males in case of a sex-linked analysis.
*
******************************************************************************/
void removeAllelesLocusMale(locusPerson *Locus)
{
	halfLocusPerson		*maLocus, *personAllele;  
	int					certainAllele;
	
	/* 1.	Initializations. */

	maLocus = personAllele = ((Locus->locusData).allelesL).maLocus;

	/* 2.	If some allele has been marked as "certain" in the maternal haplotype,
			remove all other ALLELES from this haplotype. */

	for (; personAllele != NULL ; personAllele = personAllele->next) {
		if (personAllele->isCertain == TRUE) {
			certainAllele = personAllele->allele;
			deleteAlleles(maLocus->next);

			maLocus->next = NULL;
			maLocus->allele = certainAllele;
			return;
		}
	}
}


/******************************************************************************
*
*       Function        traverseUpSelector
*       Parameters      currPerson (the person being traversed currently).
*       Output          none
*       Description		If 'curPerson' has no parents, calls traverseDownUpdate.
*						Otherwise:
*							- if one of currPerson's parents has been visited
*							  already, returns.
*							- else, keeps traversing up to currPerson's father.
*						
*       Comments		The idea is to traverse up until a founder has been 
*						reached. 
*
******************************************************************************/
void traverseUpSelector(person *currPerson)
{
	/* 1.	If this person isn't a founder. */

	if (currPerson->pa != NULL)	{	

		/* 1.1.	If one of the parernts is 'visited' we'll get to the person in 
				another way. */

		if (((currPerson->pa)->visited == TRUE) || ((currPerson->ma)->visited == TRUE)) 
			return;

		/* 1.2	Otherwise, traverse up to this person's father. */

		traverseUpSelector(currPerson->pa);
	}
	else				/* If this person is a founder, traverse down to him. */
		traverseDownSelector(currPerson);
}



/******************************************************************************
*
*       Function        traverseDownSelector
*       Parameters      currPerson (the person being traversed currently).
*       Output          none
*       Description		Updates all children of 'currPerson' and calls 
*						traverseDown recursively on each child which has already  
*						been updated by the other parent as well.
*       Comments		none
*
******************************************************************************/
void traverseDownSelector(person *currPerson)
{
	person		*ch;			
	sex			Sex;

	
	ch = currPerson->ch;
	Sex = currPerson->personSex;

	/* 1.	Mark this person as visited. */

	currPerson->visited = TRUE;				

	if (Sex == Male) {

		/*	2.	Iterate over all children of this person.  For each child: */

		for (; ch != NULL; ch = ch->NeSiPa) {		

			/* 2.1.	Update the paternal selector of the child according to the father. */

			if (!loci.sexLink)
				updateChildSelector(currPerson, ch);	
													   
			/* 2.2. If the child's mother hasn't been visited yet, traverse up to it
					and then traverse down to the child. 
					Otherwise, if we finished with the child's mother, traverse
					down to the child. */

			if ((ch->ma)->visited == FALSE)	{ 
				traverseUpSelector(ch->ma);
				traverseDownSelector(ch);
			}
			else if ((ch->ma)->done == TRUE)
				traverseDownSelector(ch);
		}
	}
	else {			/* (Sex == Female) */

		/*	2.	Iterate over all children of this person.  For each child: */

		for (; ch != NULL; ch = ch->NeSiMa) {

			/* 2.1.	Update the maternal selector of the child according to the mother. */

			updateChildSelector(currPerson, ch);	
		
			/* 2.2. If the child's father hasn't been visited yet, traverse up to it
					and then traverse down to the child. 
					Otherwise, if we finished with the child's father, traverse
					down to the child. */

			if ((ch->pa)->visited == FALSE)	{ 
				traverseUpSelector(ch->pa);
				traverseDownSelector(ch);
			}
			else if ((ch->pa)->done == TRUE)
				traverseDownSelector(ch);
		}
	}

	/* 3.	Mark this person as done. */

	currPerson->done = TRUE;
}



/******************************************************************************
*
*       Function        updateChildSelector
*       Parameters      parent (the person whose child needs to be updated
*						according to it), child (the person that needs to be
*						updated).
*       Output          none
*       Description		Updates the child selectors according to the parent.
*       Comments		none
*
******************************************************************************/
void updateChildSelector(person* parent, person *child)
{
	int					locusInd, *selecVals;
	locus				*locusPtr;
	sex					Sex;
	halfLocusPerson		*alleleCh;
	locusPerson			*parentLoc;
	
	/* 1.	Iterate over all loci. In case of an ALLELES type locus, update
			the child's relevant selector variable according to the parent. */

	Sex = parent->personSex;

	if (Sex == Male) {
		for (locusInd = 0, locusPtr = loci.locusArr; locusInd < loci.numLoci; 
				locusInd++, locusPtr++) {

			if (locusPtr->type == ALLELES) {
				alleleCh = (((child->loci[locusInd]).locusData).allelesL).paLocus;
				selecVals = (((child->loci[locusInd]).locusData).allelesL).paSelectorVals;
				parentLoc = &(parent->loci[locusInd]);
				updateChildOneSelec(alleleCh, parentLoc, selecVals);
			}
		}
	}
	else {		/* (sex == Female) */
		for (locusInd = 0, locusPtr = loci.locusArr; locusInd < loci.numLoci; 
				locusInd++, locusPtr++) {

			if (locusPtr->type == ALLELES) {
				alleleCh = (((child->loci[locusInd]).locusData).allelesL).maLocus;
				selecVals = (((child->loci[locusInd]).locusData).allelesL).maSelectorVals;
				parentLoc = &(parent->loci[locusInd]);
				updateChildOneSelec(alleleCh, parentLoc, selecVals);
			}
		}
	}
}



/******************************************************************************
*
*       Function        updateChildOneSelec
*       Parameters      alleleCh (the allele list of the child at one haplotype), 
*						parentLoc (the parent's locus), selecVals (the array of
*						selector values of the child at one haplotype).
*       Output          none
*       Description		Updates the child selector variable in the relevant locus 
*						according to the parent.
*       Comments		none
*
******************************************************************************/
void updateChildOneSelec(halfLocusPerson *alleleCh, locusPerson *parentLoc, 
						 int *selecVals)
{
 	halfLocusPerson		*alleleParentP, *alleleParentM; 

	alleleParentP = ((parentLoc->locusData).allelesL).paLocus;
	alleleParentM = ((parentLoc->locusData).allelesL).maLocus;

	/* 1.	If we know for certain the allele that was transmitted from 
			the parent to the child. */

	if (alleleCh->next == NULL) {	
		if (!inList(alleleCh->allele, alleleParentP)) {
			
			/* 1.1.	If the allele came from the maternal haplotype of the parent,
					update the possible selector values accordingly. */

			selecVals[0] = 0;
			selecVals[1] = -1;
		}
		else if (!inList(alleleCh->allele, alleleParentM)) {
	
			/* 1.1.	If the allele came from the paternal haplotype of the parent,
					update the possible selector values accordingly. */

			selecVals[0] = 1;
			selecVals[1] = -1;
		}
	}
	else if (!intersection(alleleCh, alleleParentP)) {

		/* 1.1.	If there's no intersection between the allele list of the child
				and the paternal haplotype of the parent then the child's allele
				must have come from the maternal haplotype of the parent. */

		selecVals[0] = 0;
		selecVals[1] = -1;
	}
	else if (!intersection(alleleCh, alleleParentM)) {

		/* 1.1.	If there's no intersection between the allele list of the child
				and the maternal haplotype of the parent then the child's allele
				must have come from the paternal haplotype of the parent. */

		selecVals[0] = 1;
		selecVals[1] = -1;
	}
}



/******************************************************************************
*
*       Function        inList
*       Parameters      allele (an allele), allelesList (an allele list)
*       Output          none
*       Description		Checks if 'allele' exists in the list 'allelesList'.
*       Comments		Returns TRUE in case it finds 'allele' in the list and
*						FALSE otherwise.
*
******************************************************************************/
boolean inList(int allele, halfLocusPerson *allelesList)
{
	halfLocusPerson		*currAllele;

	currAllele = allelesList;

	/* 1.	Iterate over the allele list and search for "allele". If it is 
			found in the list, return TRUE. */

	for (;currAllele != NULL; currAllele = currAllele->next) {
		if (currAllele->allele == allele)
			return TRUE;
	}

	/* 2.	If the allele wasn't found in the list, return FALSE. */

	return FALSE;
}



/******************************************************************************
*
*       Function        intersection
*       Parameters      alleleCh (the allele list of the child),  
*						alleleParent (the allele list of the parent).
*       Output          none
*       Description		Checks if the one of the ALLELES of the child exists 
*						in the allele list of the parent.
*       Comments		Returns TRUE in case there is intersection between the 
*						allele lists and FALSE otherwise.
*
******************************************************************************/
boolean intersection(halfLocusPerson *alleleCh, halfLocusPerson *alleleParent)
{
	halfLocusPerson		*currAlleleCh, *currAllelePa;


	currAlleleCh = alleleCh;
	currAllelePa = alleleParent;

	/* 1.	Iterate over the ALLELES in the allele list of the child. For each one: */

	for (;currAlleleCh != NULL; currAlleleCh = currAlleleCh->next) {

		/* 1.1.		Iterate over the parent's list of possible ALLELES and check if
					the allele exists in the list. */

		for (;(currAllelePa != NULL) && (currAllelePa->allele < currAlleleCh->allele); 
			 currAllelePa = currAllelePa->next);

		if (currAllelePa != NULL) {
			if (currAllelePa->allele == currAlleleCh->allele)
				return TRUE;
		}
		else
			return FALSE;
	}

	return FALSE;
}


/******************************************************************************
*
*       Function        updateUniteVarsFourPed
*       Parameters      ped (the pedigree which needs to be updated regarding
*						the cases where the four genetic loci variables of a 
*						couple at some locus need to be eliminated jointly).
*       Output          none
*       Description		Traces and updates the information regarding the cases 
*						where the four genetic loci variables of a couple at
*						some locus need to be eliminated jointly.
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void updateUniteVarsFourPed(pedigree *ped)
{
	person		*personPtr, *personPtrEnd, *ch;


	/* 1.	Iterate over the people in the pedigree. For each person: */

	for (personPtr = ped->people, personPtrEnd = personPtr + ped->numPeople;
			personPtr < personPtrEnd; personPtr++) {

		ch = personPtr->ch;

		/* 1.1.	If this is a non-founder and it hasn't been marked as 'done'
				yet, check for the need to unite four-vars between this person
				and his/her spouse. */
		
		if (ch && (personPtr->done == FALSE)) {

			updateUniteVarsFour(ch->pa, ch->ma);
		}
	}
}
#endif



/******************************************************************************
*
*       Function        updateUniteVarsFour
*       Parameters      pa (the father), ma (the mother)
*       Output          none
*       Description		Traces and updates the information regarding the cases 
*						where the four genetic loci variables of this couple at
*						some locus need to be eliminated jointly.
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void updateUniteVarsFour(person *pa, person *ma)
{
	person		*ch;
	int			numCh, locusInd, numLoci;
	locus		*locusPtr;

	/* 1.	Iterate over the children of the father, and count the number of joint 
			children with this mother. */

	for (numCh = 0, ch = pa->ch; ch != NULL; ch = ch->NeSiPa) {
		if (ch->ma == ma) {
			numCh++;
		}
	}

	/* 2.	If there's only 1 child, then the parents can't be updated according to
			it. */

	if (numCh == 1)
		return;

	numLoci = loci.numLoci;

	/* 3.	Iterate over the loci. */

	for (locusInd = 0, locusPtr = loci.locusArr; locusInd < numLoci; locusInd++,
			locusPtr++) {

		/* 3.1.	If this is an ALLELES type locus, update the parent according to
				the children at this locus. */

		if (locusPtr->type == ALLELES) {
			updateUniteVarsFourLoc(pa, ma, locusInd);
		}
	}

	/* 4.	Update the flags of the parents to DONE. */

	pa->done = TRUE;
	ma->done = TRUE;
}
#endif

/******************************************************************************
*
*       Function        updateUniteVarsFourLoc
*       Parameters      pa (the father), ma (the mother), locusInd (the index of
*						the locus in which the couple needs to be updated).
*       Output          none
*       Description		Checks whether the four genetic loci variables of this 
*						couple at this locus need to be eliminated jointly.
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void updateUniteVarsFourLoc(person *pa, person *ma, int locusInd)
{
	person		*ch;
	int			certainAll[4], numCertainAll, alleleArrCh[2], numAllelesCh;

	numCertainAll = 0;

	/* 1.	If the two genetic loci variables of either one of the couple need 
			to be treated jointly, then return. */

	if ((pa->loci[locusInd].locusData.allelesL.uniteT == twoAll) ||
		(ma->loci[locusInd].locusData.allelesL.uniteT == twoAll)) {
		return;
	}

	/* 2.	Iterate over the children of the father. */

	for (ch = pa->ch; ch != NULL; ch = ch->NeSiPa) {

		/* 2.1.	If this is a joint child of the couple, */

		if (ch->ma == ma) {

			/* 2.1.1.	and his genotype is known, */

			if (isKnownGenotype(&(ch->loci[locusInd]), alleleArrCh, &numAllelesCh)) {

				/* 2.1.1.1.	Update the array of certain ALLELES of the couple. */

				updateCertainAll(alleleArrCh, numAllelesCh, certainAll, &numCertainAll);
			}
		}
	}

	/* 3.	If, according to the couple's children, there are 4 certain ALLELES 
			update that the four genetic loci variables of the couple at this
			locus need to be eliminated jointly. */

	if (numCertainAll == 4) {
		updatePersonFourAllele(&(pa->loci[locusInd]), certainAll, ma->pNum);
		updatePersonFourAllele(&(ma->loci[locusInd]), certainAll, pa->pNum);
	}
}
#endif


/******************************************************************************
*
*       Function        isKnownGenotype
*       Parameters      loc (pointer to the locus structure of some person at
*						some locus), alleleArr (pointer to an array which will
*						be updated with the ALLELES of this person at this 
*						locus, in case the genotype of this person at this locus
*						is known, up to phase), numAlleles (will be updated with
*						the number of ALLELES in the genotype in case the 
*						genotype is known - either 2 if the person is heterozygote
*						or 1 if he is homozygote).
*       Output          none
*       Description		Checks if the genotype of this person at this locus is
*						known. If so, updates 'alleleArr' with the ALLELES of 
*						this person, and 'numAlleles' with the number of ALLELES
*						in the array.
*       Comments		Returns TRUE if the genotype of this person is known up
*						to phase in this locus, and FALSE otherwise.
*
******************************************************************************/
boolean isKnownGenotype(locusPerson *loc, int *alleleArr, int *numAlleles)
{
	halfLocusPerson		*maHaplo, *paHaplo;

	/* 1.	Initialize the number of seen ALLELES to 0. */

	*numAlleles = 0;

	/* 2.	Update pointers to this person's haplotypes. */

	paHaplo = loc->locusData.allelesL.paLocus;
	maHaplo = loc->locusData.allelesL.maLocus;

	/* 3.	If at most 2 ALLELES appear in the haplotypes of this person, than his
			genotype is known --> return TRUE. */

	if (updateAllelesHaplo(paHaplo, alleleArr, numAlleles) &&
			updateAllelesHaplo(maHaplo, alleleArr, numAlleles)) {
		return TRUE;
	}

	/* 4.	The genotype isn't known --> return FALSE. */

	return FALSE;
}


/******************************************************************************
*
*       Function        updateAllelesHaplo
*       Parameters      alleleList (the allele list of some person in one 
*						single-locus haplotype), alleleArr (the array which is  
*						updated with the ALLELES that appear in this person's  
*						haplotypes), numAlleles (is updated with the number of 
*						ALLELES that appear in this person's haplotypes).
*       Output          none
*       Description		Updates the array 'alleleArr' with the ALLELES that 
*						appear in this person's haplotype 'alleleList', and 
*						updates 'numAlleles' accordingly. 
*       Comments		Returns TRUE if the number of ALLELES seen in this 
*						person's haplotypes hasn't reached 2 yet. Otherwise,  
*						returns FALSE.
*
******************************************************************************/
boolean updateAllelesHaplo(halfLocusPerson *alleleList, int *alleleArr,
						   int *numAlleles) 
{
	int					ind;
	boolean				isFound;
	halfLocusPerson		*currAllele;

	/* 1.	Iterate over the allele list of this person. */

	for (currAllele = alleleList; currAllele != NULL; currAllele = currAllele->next) {

		/* 1.1.	Iterate over the ALLELES that have already been seen in this person's
				genotype. If this allele has already been seen, turn on the flag. */

		isFound = FALSE;

		for (ind = 0; ind < *numAlleles; ind++) {
			if (currAllele->allele == alleleArr[ind]) {
				isFound = TRUE;
				break;
			}
		}

		/* 1.2.	If this allele hasn't been seen yet: */

		if (isFound == FALSE){

			/* 1.2.1.	If the number of ALLELES seen so far is smaller than 2,
						update the array with this allele, and increase the 
						count of seen ALLELES by 1. Otherwise, return FALSE. */

			if (*numAlleles < 2) {
				alleleArr[(*numAlleles)++] = currAllele->allele;
			}
			else {
				return FALSE;
			}
		}
	}

	return TRUE;
}



/******************************************************************************
*
*       Function        updateCertainAll
*       Parameters      alleleArrCh (an array which holds the ALLELES of the 
*						child at some locus), numAllelesCh (the number of ALLELES
*						of this child at this locus), certainAll (an array which
*						is updated with the certain ALLELES of the parents), 
*						numCertainAll (is updated to hold the number of certain
*						ALLELES of the parents).
*       Output          none
*       Description		Updates the array 'certainAll' with the ALLELES that 
*						appear in the child's genotype, and updates 'numCertainAll', 
*						accordingly. 
*       Comments		none
*
******************************************************************************/
void updateCertainAll(int *alleleArrCh, int numAllelesCh, int *certainAll, 
					  int *numCertainAll)
{
	int			indCh, *certainAllPtr, *certainAllPtrEnd, ind, ind2, alleleCh;
	boolean		isFound;

	certainAllPtrEnd = certainAll + *numCertainAll;

	/* 1.	Iterate over the ALLELES of the child. */

	for (indCh = 0; indCh < numAllelesCh; indCh++) {
		isFound = FALSE;
		alleleCh = alleleArrCh[indCh];

		/* 1.1.	Iterate over the certain ALLELES of the parents. */

		for (certainAllPtr = certainAll; certainAllPtr < certainAllPtrEnd; certainAllPtr++) {

			/* 1.1.1.	If the allele of the child already appears in the certain
						ALLELES of the parents, turn on the flag. */

			if (alleleCh == *certainAllPtr) {
				isFound = TRUE;
				break;
			}
		}

		/* 1.2.	If the child's allele doesn't appear in the certain
				ALLELES of the parents, add it in the right place (the array 
				is sorted in increasing order), and update the number of certain 
				ALLELES of the parents. */

		if (isFound == FALSE) {

			/* 1.2.1.	Iterate over the array looking for the right place to 
						insert the child's allele. */

			for (ind = 0, certainAllPtr = certainAll; ind < *numCertainAll; 
					ind++, certainAllPtr++) {

				if (*certainAllPtr > alleleCh)
					break;
			}

			if (ind < *numCertainAll) {
				for (ind2 = *numCertainAll; ind2 > ind; ind2--) {
					certainAll[ind2] = certainAll[ind2-1];
				}
				certainAll[ind2] = alleleCh;
				certainAllPtrEnd++;
			}
			else {
				certainAll[*numCertainAll] = alleleCh;
			}

			(*numCertainAll)++;
		}
	}
}


/******************************************************************************
*
*       Function        updatePersonFourAllele
*       Parameters      loc (the locus structure of the person which needs to be
*						updated according to the certain ALLELES in the joint 
*						genotypes of this person and the other parent), 
*						certainAll (an array which holds all four certain ALLELES 
*						of this person and the other parent), pNum (the ID of 
*						the other parent).
*       Output          none
*       Description		Updates the locus structure of the person according to 
*						the array 'certainAll'. 
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void updatePersonFourAllele(locusPerson *loc, int *certainAll, int pNum)
{
	int		*certainAllPtr, *certainAllPtrEnd, *currAllele;

	/* 1.	Update that the 2 genetic loci variables of this person at this
			locus need to be eliminated jointly with the 2 genetic loci variables
			of person no. 'pNum' at this locus. */

	loc->locusData.allelesL.uniteT = fourAll;
	loc->locusData.allelesL.unitePerson = pNum;

	/* 2.	If this person has more than 4 possible allelels in one of his 
			haplotypes, removes the rest and update pointers to these 4 ALLELES
			only. */

	removeOtherAlleles(loc->locusData.allelesL.paLocus, certainAll);
	removeOtherAlleles(loc->locusData.allelesL.maLocus, certainAll);

	/* 3.	Allocate memory to store these 4 ALLELES. */

	myMalloc(loc->locusData.allelesL.certainAllele, sizeof(int) * 4);

	/* 4.	Update the array with the 4 certain ALLELES. */

	for (certainAllPtr = certainAll, certainAllPtrEnd = certainAllPtr + 4,
			currAllele = loc->locusData.allelesL.certainAllele;
			certainAllPtr < certainAllPtrEnd; certainAllPtr++, currAllele++) {

		(*currAllele) = (*certainAllPtr); 
	}
}
#endif

/******************************************************************************
*
*       Function        removeOtherAlleles
*       Parameters      haplo (the haplotype of some person at some locus), 
*						certainAll (an array which holds the certain ALLELES 
*						of this person and the other parent).
*       Output          none
*       Description		Removes the ALLELES that don't appear in the array 
*						'certainAll' from the haplotype of this person. 
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void removeOtherAlleles(halfLocusPerson	*haplo, int *certainAll)
{
	halfLocusPerson		*currAllele;
	int					numAlleles, *certainAllPtr;

	numAlleles = 0;

	/* 1.	Iterate over the ALLELES that appear in this person's haplotype. */

	for (currAllele = haplo; currAllele != NULL; currAllele = currAllele->next) {
		numAlleles++;
		
		/* 1.1.	If there are more than 4, delete the rest of the ALLELES from
				the haplotype and update the person with the 4 certain ALLELES. */

		if ((currAllele->next != NULL) && (numAlleles == 4)) {
			deleteAlleles(currAllele->next);
			currAllele->next = NULL;
			
			for (currAllele = haplo, certainAllPtr = certainAll; currAllele != NULL; 
					currAllele = currAllele->next, certainAllPtr++) {
				currAllele->allele = *certainAllPtr;
			}
			break;
		}
	}
}
#endif

