/*********************************************************************************/
/*  PedValueElimination.h													     */
/*  Author: Ma'ayan Fishelson													 */
/*********************************************************************************/
/*																				 */
/*  Comment: This file includes functions used for eliminating impossible values */
/*			 from the genetic-loci and selector variables representing the		 */
/*			 people of the input pedigrees in the Bayesian networks.			 */
/*																				 */
/*********************************************************************************/

#ifndef _SL_PED_VALUE_ELIMINATION_H_
#define _SL_PED_VALUE_ELIMINATION_H_

#include <stdio.h>
#include <stdlib.h>
#include "sl_PedData.h"
#include "sl_LociData.h"


/*------------------------------*/
/*		DATA-STRUCTURES			*/
/*------------------------------*/

/* A type of a function that updates the child's genotype according to his parent's genotype. */

typedef void (updateCh)(person* parent, person *child);

/* A type of a function that updates the parents' genotype according to the child's genotype. */

typedef void (updatePa)(person *pa, person *ma, person *ch);

/*------------------------------*/
/*		FUNCTIONS				*/
/*------------------------------*/

void valueEliminate(updateCh updateChild, updateFlagsLoci updateFlagsLociFunc,
					 updatePa updateParents);

/* Functions used in the downward traversal part of the value-elimination algorithm, 
   where children are updated according to their parents. */

void traverseUp(person *currPerson, updateCh updateChild, int *numPeople);
void traverseDownUpdate(person *currPerson, updateCh updateChild, int *numPeople);
void updateChildAutosomal(person* parent, person *child);
void updateChildLociAutosomal(person* parent, person* child, int locusInd);
void updateChildSexLinked(person* parent, person *child);
void updateChildLociSexLinked(person* parent, person* child, int locusInd);
void addAlleles(locusPerson *childLoci, halfLocusPerson *alleleParent, sex Sex); 
void addParentAlleles(halfLocusPerson *alleleParent, halfLocusPerson *alleleCh);

/* Functions used in the upward traversal part of the value-elimination algorithm, 
   where parents are updated according to their children. */

void traverseUpUpdate(person *currPerson, updatePa updateParents);
void traverseDown(person *currPerson, updatePa updateParents);
void updateParentsAutosomal(person *pa, person *ma, person *ch);
void updateParentsSexLinked(person *pa, person *ma, person *ch);
void updateParentLoc(halfLocusPerson *alleleCh, locusPerson *parentLoc);
void updateHalfParentLocus(int alleleCh, halfLocusPerson *alleleParent);
void updateAlleles(person *Person);
void removeAllelesLocus(locusPerson *Locus);
int	updateCertainAlleles(halfLocusPerson *personAllele, int *certainArr);
void updateHaplo(halfLocusPerson  *haplo, int *certainArr);
void updateAllelesMale(person *Person);
void removeAllelesLocusMale(locusPerson *Locus);

/* Functions used in the selector-update part of the value-elimination algorithm, 
   where selector values are being eliminated. */

void traverseUpSelector(person *currPerson);
void traverseDownSelector(person *currPerson);
void updateChildSelector(person* parent, person *child);
void updateChildOneSelec(halfLocusPerson *alleleCh, locusPerson *parentLoc, 
						 int *selecVals);
boolean inList(int allele, halfLocusPerson *allelesList);
boolean intersection(halfLocusPerson *alleleCh, halfLocusPerson *alleleParent);

/* Functions used in the part where four locus-variables that need to be 
   treated jointly in the computations are traced and updated. */

#if UNITE_VARS
	void updateUniteVarsFourPed(pedigree *ped);
	void updateUniteVarsFour(person *pa, person *ma);
	void updateUniteVarsFourLoc(person *pa, person *ma, int locusInd);
	void updatePersonFourAllele(locusPerson *loc, int *certainAll, int pNum);
	void removeOtherAlleles(halfLocusPerson	*haplo, int *certainAll);
#endif

boolean isKnownGenotype(locusPerson *loc, int *alleleArr, int *numAlleles);
boolean updateAllelesHaplo(halfLocusPerson	*alleleList, int *alleleArr,
						   int *numAlleles);
void updateCertainAll(int *alleleArrCh, int numAllelesCh, int *certainAll, 
					  int *numCertainAll);


#endif /* _PED_VALUE_ELIMINATION_H_ */

