/*********************************************************************************/
/*  Probability.c																 */
/*  Author: Ma'ayan Fishelson				                                     */
/*********************************************************************************/

#include "sl_Probability.h"



/******************************************************************************
*
*       Function        createTableTrans
*       Parameters		currPerson (the person for which this probability table
*						is created), locusNum (the index in the data-structures
*						of the locus for which this probability table is created), 
*						isPaternal (is TRUE is this is a paternal transmission 
*						probability table), locusNode (the node to which this  
*						probability table is attached), selectorNode (the 
*						relevant selector node), newTable (will point to the 
*						probability table created).
*       Output          none
*       Description		Creates a transmission probability table.
*						Pr(C[i,p] | A[i,p], A[i,m], Sc[i,p]), if this is a 
*						paternal probability table, or
*						Pr(C[i,m] | B[i,p], B[i, m], Sc[i,m]), if this is a 
*						maternal probability table. (C is the child, A is the 
*						father and B is the mother)
*       Comments		Used in the case when no allele-recoding is performed.
*
******************************************************************************/
int createTableTrans(person *currPerson, int locusNum, boolean isPaternal, 
					 node *locusNode, node *selectorNode, Probability **newTable)
{
	int			*locusVals1=NULL, *locusVals2=NULL, *locusVals3=NULL, 
				*selectorVals, ind[4], indProb, numVals[4], tableSize; 
	node		**parents;
	person		*parent;
	locus		*locusPtr;
	locusPerson	*loc;


	/* 1.	Initializations. */

	locusPtr = &(loci.locusArr[locusNum]);
	parents = locusNode->parents;

	/* 2.	Update the parents of this node and their numChildren field. */

	if (isPaternal) {
		parent = currPerson->pa;
	} else {
		parent = currPerson->ma;
	}

	parents[0] = parent->locusVar[0][locusNum];
	parents[1] = parent->locusVar[1][locusNum];
	parents[2] = selectorNode;
	parents[0]->numChildren++;
	parents[1]->numChildren++;
	parents[2]->numChildren++;

	/* 3.	Calculate the size of the table and initialize an auxiliary array which 
			will hold the number of possible values of each variable of the function. */

	numVals[0] = locusNode->numValues;
	tableSize = numVals[0];
	
	for (ind[0] = 1; ind[0] < 4; (ind[0])++, parents++) {
		numVals[ind[0]] = (*parents)->numValues;
		tableSize *= numVals[ind[0]];
	}

	/* 4.	Allocate memory for the new probability table. */
	
	myMalloc((*newTable), sizeof(Probability) * tableSize);

	/* 4.	Update pointers to the arrays of possible values for each variable of
			the probability table. */

	if (locusPtr->type == ALLELES) {
		loc = &(currPerson->loci[locusNum]);
		
		if (isPaternal) {
			locusVals1 = (loc->locusData).allelesL.paAlleles;
			selectorVals = (loc->locusData).allelesL.paSelectorVals;

			loc = &((currPerson->pa)->loci[locusNum]);
		}
		else {
			locusVals1 = (loc->locusData).allelesL.maAlleles;
			selectorVals = (loc->locusData).allelesL.maSelectorVals;

			loc = &((currPerson->ma)->loci[locusNum]);
		}

		locusVals2 = (loc->locusData).allelesL.paAlleles;
		locusVals3 = (loc->locusData).allelesL.maAlleles;
	}
	else {	/* (loci[data->locusNum] == affection) || (loci[data->locusNum] == null) */
		myMalloc(locusVals1, sizeof(int) * 2);
		myMalloc(locusVals2, sizeof(int) * 2);
		myMalloc(locusVals3, sizeof(int) * 2);
		myMalloc(selectorVals, sizeof(int) * 2);

		locusVals1[0] = locusVals2[0] = locusVals3[0] = 1;
		locusVals1[1] = locusVals2[1] = locusVals3[1] = 2;
		selectorVals[0] = 0;
		selectorVals[1] = 1;
	}

	/* 5.	Initialize the probabilities in the new table. */

	indProb = 0;

	for (ind[0] = 0; ind[0] < numVals[0]; (ind[0])++)
		for (ind[1] = 0; ind[1] < numVals[1]; (ind[1])++)
			for (ind[2] = 0; ind[2] < numVals[2]; (ind[2])++)
				for (ind[3] = 0; ind[3] < numVals[3]; (ind[3])++) {
		
					(*newTable)[indProb] = transmission(locusVals1[ind[0]], 
														locusVals2[ind[1]],
														locusVals3[ind[2]], 
														selectorVals[ind[3]]);

					indProb++;
				}

	/* 6.	Free the memory that was allocated dynamically. */

	if (locusPtr->type != ALLELES) {
		myFree(locusVals1);
		myFree(locusVals2);
		myFree(locusVals3);
		myFree(selectorVals);
	}

	/* 7.	Return the size of the probability table created. */

	return tableSize;
}

/******************************************************************************
*
*       Function        createTableTransAR
*       Parameters		currPerson (the person for which this probability table
*						is created), locusNum (the index in the data-structuires
*						of the locus for which this probability table is created), 
*						isPaternal (is TRUE is this is a paternal transmission 
*						probability table), locusNode (the node to which this  
*						probability table is attached), selectorNode (the 
*						relevant selector node), newTable (will point to the 
*						probability table created).
*       Output          none
*       Description		Creates a transmission probability table.
*						Pr(C[i,p] | A[i,p], A[i,m], Sc[i,p]), if this is a 
*						paternal probability table, or
*						Pr(C[i,m] | B[i,p], B[i, m], Sc[i,m]), if this is a 
*						maternal probability table. (C is the child, A is the 
*						father and B is the mother)
*       Comments		Used in the case when allele-recoding is performed.
*
******************************************************************************/
#if PERFORM_ALLELE_RECODING
int createTableTransAR(person *currPerson, int locusNum, boolean isPaternal, 
					   node *locusNode, node *selectorNode, Probability **newTable)
{
	int			*selectorVals, ind[4], indProb, numVals[4], tableSize; 
	node		**parents;
	person		*parent;
	alleleSet	*locusSets1=NULL, *locusSets2=NULL, *locusSets3=NULL;
	locusPerson	*loc;

	/* 1.	Update the parents of this node and their numChildren field. */

	if (isPaternal) 
		parent = currPerson->pa;
	else
		parent = currPerson->ma;

	parents = locusNode->parents;

	parents[0] = parent->locusVar[0][locusNum];
	parents[1] = parent->locusVar[1][locusNum];
	parents[2] = selectorNode;
	parents[0]->numChildren++;
	parents[1]->numChildren++;
	parents[2]->numChildren++;

	/* 2.	Calculate the size of the probability table and initialize an auxiliary array
			which will hold the number of possible values of each variable of the function. */

	numVals[0] = locusNode->numValues;
	tableSize = numVals[0];
	
	for (ind[0] = 1; ind[0] < 4; (ind[0])++, parents++) {
		numVals[ind[0]] = (*parents)->numValues;
		tableSize *= numVals[ind[0]];
	}

	/* 3.	Allocate memory for the new probability table. */
	
	myMalloc((*newTable), sizeof(Probability) * tableSize);

	/* 4.	Update pointers to the arrays of possible values for each variable of
			the probability table. */

	loc = &(currPerson->loci[locusNum]);

	if (isPaternal) {
		locusSets1 = (loc->locusData).allelesL.paAllelsGroups;
		selectorVals = (loc->locusData).allelesL.paSelectorVals;

		loc = &((currPerson->pa)->loci[locusNum]);
	}
	else {
		locusSets1 = (loc->locusData).allelesL.maAllelsGroups;
		selectorVals = (loc->locusData).allelesL.maSelectorVals;

		loc = &((currPerson->ma)->loci[locusNum]);
	}

	locusSets2 = (loc->locusData).allelesL.paAllelsGroups;
	locusSets3 = (loc->locusData).allelesL.maAllelsGroups;

	/* 5.	Initialize the probabilities in the new table. */

	indProb = 0;

	for (ind[0] = 0; ind[0] < numVals[0]; (ind[0])++)
		for (ind[1] = 0; ind[1] < numVals[1]; (ind[1])++)
			for (ind[2] = 0; ind[2] < numVals[2]; (ind[2])++)
				for (ind[3] = 0; ind[3] < numVals[3]; (ind[3])++) {
		
					(*newTable)[indProb] = ((((selectorVals[ind[3]] == 1) &&
											setInclusion(locusSets2[ind[1]], locusSets1[ind[0]])) ||
											((selectorVals[ind[3]] == 0) &&
											setInclusion(locusSets3[ind[2]], locusSets1[ind[0]]))));
					indProb++;
				}

	/* 6.	Return the size of the probability table created. */

	return tableSize;
}
#endif


/******************************************************************************
*
*       Function        createTableTransFounder
*       Parameters		currPerson (the person for which this probability table
*						is created), locusNum (the index in the data-structures
*						of locus for which this probability table is created), 
*						isPaternal (is TRUE is this is a paternal transmission 
*						probability table), locusNode (the node to which this  
*						probability table is attached), selectorNode (the 
*						relevant selector node), newTable (will point to the
*						probability table created).
*       Output          none
*       Description		Creates a transmission probability table for a child of
*						a founder.
*       Comments		none
*
******************************************************************************/
int createTableTransFounder(person *currPerson, int locusNum, boolean isPaternal, 
							node *locusNode, node *selectorNode, 
							Probability **newTable)
{
	int			*locusVals1=NULL, *locusVals2=NULL, *locusVals3=NULL,  
				*selectorVals, ind[4], indProb, numVals[4], tableSize; 
	node		**parents;
	boolean		isKnownAlleleGenotype;
	person		*parent;
	locus		*locusPtr;
	locusPerson	*loc;

#if PERFORM_ALLELE_RECODING
	alleleSet *locusSets1=NULL, *locusSets2=NULL, *locusSets3=NULL;
#endif

	/* 1.	Initializations. */

	locusPtr = &(loci.locusArr[locusNum]);
	parents = locusNode->parents;

	/* 2.	Update a pointer to the parent of this person. */

	if (isPaternal)
		parent = currPerson->pa;
	else
		parent = currPerson->ma;

	/* 3.	Update the parents of this node and their numChildren field. */

	parent->locusVar[0][locusNum]->numChildren++;
	parent->locusVar[1][locusNum]->numChildren++;
	parents[0] = parent->genotypeVar;
	parents[1] = selectorNode;
	parents[0]->numChildren++;
	parents[1]->numChildren++;

	/* 4.	Calculate the size of the table and initialize an auxiliary array which 
			will hold the number of possible values of each variable of the function. */

	numVals[0] = locusNode->numValues;
	tableSize = numVals[0] * parents[0]->numValues;
	numVals[1] = numVals[2] = (parent->locusVar[0][locusNum])->numValues;
	numVals[3] = parents[1]->numValues;
	
	tableSize *= numVals[3];

	/* 5.	Allocate memory for the new probability table. */

	myMalloc((*newTable), sizeof(Probability) * tableSize);

	/* 6.	Update pointers to the arrays of possible values for each variable of
			the probability table. */

	if (locusPtr->type == ALLELES) {
		
		loc = &(currPerson->loci[locusNum]);

		if (isPaternal) {
			selectorVals = (loc->locusData).allelesL.paSelectorVals;

			#if PERFORM_ALLELE_RECODING
				locusSets1 = (loc->locusData).allelesL.paAllelsGroups;
			#else
				locusVals1 = (loc->locusData).allelesL.paAlleles;
			#endif

			loc = &((currPerson->pa)->loci[locusNum]);
		}
		else {
			selectorVals = (loc->locusData).allelesL.maSelectorVals;

			#if PERFORM_ALLELE_RECODING
				locusSets1 = (loc->locusData).allelesL.maAllelsGroups;
			#else
				locusVals1 = (loc->locusData).allelesL.maAlleles;
			#endif

			loc = &((currPerson->ma)->loci[locusNum]);
		}

		#if PERFORM_ALLELE_RECODING
			locusSets2 = (loc->locusData).allelesL.paAllelsGroups;
			locusSets3 = (loc->locusData).allelesL.maAllelsGroups;
		#else
			locusVals2 = (loc->locusData).allelesL.paAlleles;
			locusVals3 = (loc->locusData).allelesL.maAlleles;
		#endif

		isKnownAlleleGenotype = !(loc->unknown);
	}
	else {	/* (loci[data->locusNum] == affection) || (loci[data->locusNum] == null) */
		myMalloc(locusVals1, sizeof(int) * 2);
		myMalloc(locusVals2, sizeof(int) * 2);
		myMalloc(locusVals3, sizeof(int) * 2);
		myMalloc(selectorVals, sizeof(int) * 2);

		locusVals1[0] = locusVals2[0] = locusVals3[0] = 1;
		locusVals1[1] = locusVals2[1] = locusVals3[1] = 2;
		selectorVals[0] = 0;
		selectorVals[1] = 1;

		isKnownAlleleGenotype = FALSE;
	}

	/* 7.	Initialize the probabilities in the new table. */

	indProb = 0;

#if PERFORM_ALLELE_RECODING
	if (locusPtr->type != ALLELES) {
#endif
		if (isKnownAlleleGenotype == FALSE) {	

			/* 7.1.	If the parent's genotype is unknown at this locus. */

			for (ind[0] = 0; ind[0] < numVals[0]; (ind[0])++)
				for (ind[1] = 0; ind[1] < numVals[1]; (ind[1])++)
					for (ind[2] = ind[1]; ind[2] < numVals[2]; (ind[2])++) 
						for (ind[3] = 0; ind[3] < numVals[3]; (ind[3])++) {
				
							(*newTable)[indProb] = transmission(locusVals1[ind[0]], 
																locusVals2[ind[1]],
																locusVals3[ind[2]],
																selectorVals[ind[3]]);
							indProb++;
						}
		}
		else {									

			/* 7.1.	If the parent's genotype is known at this locus. */

			ind[1] = 0;

			if (numVals[1] == 1)
				ind[2] = 0;
			else
				ind[2] = 1;

			for (ind[0] = 0; ind[0] < numVals[0]; (ind[0])++) 
				for (ind[3] = 0; ind[3] < numVals[3]; (ind[3])++) {

					(*newTable)[indProb] = transmission(locusVals1[ind[0]], 
														locusVals2[ind[1]],
														locusVals3[ind[2]],
														selectorVals[ind[3]]);
					indProb++;
				}
		}

#if PERFORM_ALLELE_RECODING
	} else {
	
		if (isKnownAlleleGenotype == FALSE) {	/* The genotype is unknown at this locus. */

			for (ind[0] = 0; ind[0] < numVals[0]; (ind[0])++)
				for (ind[1] = 0; ind[1] < numVals[1]; (ind[1])++)
					for (ind[2] = ind[1]; ind[2] < numVals[2]; (ind[2])++) 
						for (ind[3] = 0; ind[3] < numVals[3]; (ind[3])++) {
					
							(*newTable)[indProb] = ((((selectorVals[ind[3]] == 1) &&
													setInclusion(locusSets2[ind[1]], locusSets1[ind[0]])) ||
													((selectorVals[ind[3]] == 0) &&
													setInclusion(locusSets3[ind[2]], locusSets1[ind[0]]))));
							indProb++;
						}
		}
		else {									/* The genotype is known at this locus. */

			ind[1] = 0;

			if (numVals[1] == 1)
				ind[2] = 0;
			else
				ind[2] = 1;

			for (ind[0] = 0; ind[0] < numVals[0]; (ind[0])++) 
				for (ind[3] = 0; ind[3] < numVals[3]; (ind[3])++) {

					(*newTable)[indProb] = (((selectorVals[ind[3]] == 1) &&
											setInclusion(locusSets2[ind[1]], locusSets1[ind[0]])) ||
											((selectorVals[ind[3]] == 0) &&
											setInclusion(locusSets3[ind[2]], locusSets1[ind[0]])));
					indProb++;
				}
		}
	}
#endif

	/* 8.	Free the memory that was allocated dynamically. */

	if (locusPtr->type != ALLELES) {
		myFree(locusVals1);
		myFree(locusVals2);
		myFree(locusVals3);
		myFree(selectorVals);
	}

	/* 9.	Return the size of the probability table created. */

	return tableSize;
}


/******************************************************************************
*
*       Function        createTableTransMale
*       Parameters		currPerson (the person for which this probability table
*						is created), locusNum (the index in the data-structures
*						of the locus for which this probability table is created),
*						pLocusNode (the node to which this probability table
*						belongs), newTable (will point to the probability table
*						created).
*       Output          none
*       Description		Creates a special transmission probability table of a 
*						female from her father for the case of sex-linked 
*						analysis.
*						Pr(C[i,p] | A[i,m]), where C is the child and A the father.
*       Comments		returns the size of the probability table created.
*						Used in the case of sex-linked analysis, when no allele
*						recoding is performed.
*
******************************************************************************/
int createTableTransMale(person *currPerson, int locusNum, node *pLocusNode,
						 Probability **newTable)
{
	int		*locusVals1=NULL, *locusVals2=NULL, ind[2], indProb, numVals[2],
			tableSize; 
	node	**parents;
	locus	*locusPtr;


	/* 1.	Initializations. */

	locusPtr = &(loci.locusArr[locusNum]);
	parents = pLocusNode->parents;

	/* 2.	Update the numChildren field of this node's parent. */

	(*parents) = (currPerson->pa)->locusVar[1][locusNum];
	(*parents)->numChildren++;

	/* 3.	Calculate the size of the table and initialize an auxiliary array which 
			will hold the number of possible values of each variable of the function. */

	numVals[0] = pLocusNode->numValues;
	numVals[1] = (*parents)->numValues;

	tableSize = numVals[0] * numVals[1];

	/* 4.	Allocate memory for the new probability table. */
	
	myMalloc((*newTable), sizeof(Probability) * tableSize);

	/* 5.	Update pointers to the arrays of possible values for each variable of
			the probability table. */

	if (locusPtr->type == ALLELES) {
		locusVals1 = ((currPerson->loci[locusNum]).locusData).allelesL.paAlleles;
		locusVals2 = (((currPerson->pa)->loci[locusNum]).locusData).allelesL.maAlleles;
	}
	else {	/* (loci[data->locusNum] == affection) || (loci[data->locusNum] == null) */
		myMalloc(locusVals1, sizeof(int) * 2);
		myMalloc(locusVals2, sizeof(int) * 2);

		locusVals1[0] = locusVals2[0] = 1;
		locusVals1[1] = locusVals2[1] = 2;
	}

	/* 6.	Initialize the probabilities in the new table. */

	indProb = 0;

	for (ind[0] = 0; ind[0] < numVals[0]; (ind[0])++) {
		for (ind[1] = 0; ind[1] < numVals[1]; (ind[1])++) {

			(*newTable)[indProb] = transmissionMale(locusVals1[ind[0]], 
													locusVals2[ind[1]]);

			indProb++;
		}
	}

	/* 7.	Free the memory that was allocated dynamically. */

	if (locusPtr->type != ALLELES) {
		myFree(locusVals1);
		myFree(locusVals2);
	}

	/* 8.	Return the size of the probability table created. */

	return tableSize;
}



/******************************************************************************
*
*       Function        createTableTransMaleAR
*       Parameters		currPerson (the person for which this probability table
*						is created), locusNum (the index in the data-structures
*						of the locus for which this probability table is created), 
*						pLocusNode (the node to which this probability table
*						belongs), newTable (will point to the probability table 
*						created).
*       Output          none
*       Description		Creates a special transmission probability table of a 
*						female from her father for the case of sex-linked 
*						analysis.
*						Pr(C[i,p] | A[i,m]), where C is the child and A the father.
*       Comments		Returns the size of the probability table created.
*						Used in the case of sex-linked analysis, when allele
*						recoding is performed.
*
******************************************************************************/
#if PERFORM_ALLELE_RECODING
int createTableTransMaleAR(person *currPerson, int locusNum, node *pLocusNode,
						   Probability **newTable)
{
	int			ind[2], indProb, numVals[2], tableSize; 
	node		**parents;
	alleleSet	*locusSets1=NULL, *locusSets2=NULL;

	/* 1.	Initializations. */

	parents = pLocusNode->parents;

	/* 2.	Update the numChildren field of this node's parent. */

	(*parents) = (currPerson->pa)->locusVar[1][locusNum];
	(*parents)->numChildren++;

	/* 3.	Calculate the size of the table and initialize an auxiliary array which 
			will hold the number of possible values of each variable of the function. */

	numVals[0] = pLocusNode->numValues;
	numVals[1] = (*parents)->numValues;

	tableSize = numVals[0] * numVals[1];

	/* 4.	Allocate memory for the new probability table. */
	
	myMalloc((*newTable), sizeof(Probability) * tableSize);

	/* 5.	Update pointers to the arrays of possible values for each variable of
			the probability table. */

	locusSets1 = ((currPerson->loci[locusNum]).locusData).allelesL.paAllelsGroups;
	locusSets2 = (((currPerson->pa)->loci[locusNum]).locusData).allelesL.maAllelsGroups;

	/* 6.	Initialize the probabilities in the new table. */

	indProb = 0;

	for (ind[0] = 0; ind[0] < numVals[0]; (ind[0])++)
		for (ind[1] = 0; ind[1] < numVals[1]; (ind[1])++) {

			(*newTable)[indProb] = setInclusion(locusSets2[ind[1]], locusSets1[ind[0]]);

			indProb++;
		}

	/* 7.	Return the size of the probability table created. */

	return tableSize;
}
#endif


/******************************************************************************
*
*       Function        setInclusion
*       Parameters		parentSet (an allele set of the parent), childSet (an
*						allele set of the child).
*       Output          none
*       Description		Checks whether the allele set of the parent is included
*						in the allele set of the child. If so return TRUE, 
*						otherwise returns FALSE.
*       Comments		Used in the case where allele-recoding is performed.
*
******************************************************************************/
#if PERFORM_ALLELE_RECODING
boolean setInclusion(alleleSet parentSet, alleleSet childSet)
{
	int			*alleleCh, *alleleChEnd, *allelePa, *allelePaEnd;
	boolean		isFound;

	/* 1.	Iterate over the ALLELES in the parent's set. */

	for (allelePa = parentSet.alleleArr, 
			allelePaEnd = allelePa + parentSet.numAlleles; allelePa < allelePaEnd;
			allelePa++) {

		/* 1.1.	For each one, check if it is found in the allele set of the child. */

		isFound = FALSE;
		
		for (alleleCh = childSet.alleleArr, 
				alleleChEnd = alleleCh + childSet.numAlleles; 
				alleleCh < alleleChEnd; alleleCh++) {

			if (*allelePa == *alleleCh) {
				isFound = TRUE;
				break;
			}
		}

		/* 1.2.	If the allele isn't found, return FALSE. */

		if (isFound == FALSE)
			return FALSE;
	}

	/* 2.	The allele set of the parent is included in the allele set of the child.
			Return TRUE. */

	return TRUE;
}
#endif


/******************************************************************************
*
*       Function        createTableHalf
*       Parameters		numVals (the number of possible values of the selector
*						variable), newTable (will point to the new probability
*						table created).
*       Output          none
*       Description		Creates a probability table of size 'numVals', of which  
*						each entrance has probability of 0.5
*       Comments		This is the probability table of some selector
*						node of the first locus in the chromosome order.
*
******************************************************************************/
int createTableHalf(int numVals, Probability **newTable)
{
	int		ind;

	/* 1.	Allocate memory for the new probability table. */

	myMalloc((*newTable), sizeof(Probability) * numVals);

	/* 2.	Update the probabilities in the new table. */

	for (ind = 0; ind < numVals; ind++) {
		(*newTable)[ind] = 0.5;
	}

	/* 3.	Return the size of the probability table created. */

	return numVals;
}


/******************************************************************************
*
*       Function        createTableRecomb
*       Parameters		currPerson (the person for which this probability table
*						is created), locusIndMap (the map index of the locus for 
*						which this probability table is created), isPaternal 
*						(is TRUE if this is a paternal Recombination probability 
*						table), selectorNode (the selector node to which this 
*						probability table belongs), newTable (will point to the
*						probability table created).
*       Output          none
*       Description		Creates a recombination probability table.
*						Pr(Sc[i,x] | Sc[i-1,x]).
*       Comments		none
*
******************************************************************************/
int createTableRecomb(person *currPerson, int locusIndMap, boolean isPaternal, 
					  node *selectorNode, Probability **newTable)
{
	int		ind[2], indProb, *selectorVals1, *selectorVals2, numVals[2],
			locusIndDS, prevLocusIndDS, tableSize;
	locus	*locusPtr, *prevLocusPtr;
	node	**parents;

	/* 1.	Find the index in the data-structures of locus no. 'locusIndMap' 
			and of the previous locus on the map. */

	locusIndDS = loci.mapToData[locusIndMap];
	prevLocusIndDS = loci.mapToData[locusIndMap-1];

	locusPtr = &(loci.locusArr[locusIndDS]);
	prevLocusPtr = &(loci.locusArr[prevLocusIndDS]);

	/* 2.	Update the node's parent and its number of children field. */

	parents = selectorNode->parents;

	if (isPaternal)
		parents[0] = currPerson->selectorVar[0][prevLocusIndDS];
	else
		parents[0] = currPerson->selectorVar[1][prevLocusIndDS];

	parents[0]->numChildren++;

	/* 3.	Update pointers to the arrays of possible values for each variable of
			the probability table. */

	if (isPaternal)  {
		if (prevLocusPtr->type == ALLELES)
			selectorVals1 = ((currPerson->loci[prevLocusIndDS]).locusData).allelesL.paSelectorVals;
		else {
			myMalloc(selectorVals1, sizeof(int) * 2);
			selectorVals1[0] = 0;
			selectorVals1[1] = 1;
		}

		if (locusPtr->type == ALLELES)
			selectorVals2 = ((currPerson->loci[locusIndDS]).locusData).allelesL.paSelectorVals;
		else {
			myMalloc(selectorVals2, sizeof(int) * 2);
			selectorVals2[0] = 0;
			selectorVals2[1] = 1;
		}
	}
	else {	/* these are maternal selectors. */
		if (prevLocusPtr->type == ALLELES)
			selectorVals1 = ((currPerson->loci[prevLocusIndDS]).locusData).allelesL.maSelectorVals;
		else {
			myMalloc(selectorVals1, sizeof(int) * 2);
			selectorVals1[0] = 0;
			selectorVals1[1] = 1;
		}

		if (locusPtr->type == ALLELES)
			selectorVals2 = ((currPerson->loci[locusIndDS]).locusData).allelesL.maSelectorVals;
		else {
			myMalloc(selectorVals2, sizeof(int) * 2);
			selectorVals2[0] = 0;
			selectorVals2[1] = 1;
		}
	}

	/* 4.	Update the 'numVals' array to hold the number of possible values of 
			each selector variable. */

	numVals[0] = (selectorVals1[1] != -1) ? 2 : 1;
	numVals[1] = (selectorVals2[1] != -1) ? 2 : 1;

	/* 5.	Calculate the size of the new probability table and allocate memory for it. */

	tableSize = numVals[0] * numVals[1];
	myMalloc((*newTable), sizeof(Probability) * tableSize);

	indProb = 0;

	/* 6.	Calculate the probabilities in the new table. */

	if (isPaternal)  {
		for (ind[0] = 0; ind[0] < numVals[0]; ind[0]++)
			for (ind[1] = 0; ind[1] < numVals[1]; ind[1]++) {

				(*newTable)[indProb] = recombTransMale(selectorVals1[ind[0]], selectorVals2[ind[1]], locusIndMap - 1);

				indProb++;
			}
	}
	else {
		for (ind[0] = 0; ind[0] < numVals[0]; ind[0]++)
			for (ind[1] = 0; ind[1] < numVals[1]; ind[1]++) {

				(*newTable)[indProb] = recombTransFemale(selectorVals1[ind[0]], selectorVals2[ind[1]], locusIndMap - 1);

				indProb++;
			}
	}

	/* 7.	Free the memory that was allocated dynamically. */

	if (prevLocusPtr->type != ALLELES) {
		myFree(selectorVals1);
	}

	if (locusPtr->type != ALLELES) {
		myFree(selectorVals2);
	}

	/* 8.	Return the size of the probability table created. */

	return tableSize;
}



/******************************************************************************
*
*       Function        createTableRecombInterference
*       Parameters		currPerson (the person for which this probability table
*						is created), locusIndMap (the index on the map of the 
*						locus for which this probability table is created), 
*						pSelectorNode (the selector node to which this probability
*						table belongs), newTable (will point to the probability 
*						table created).
*       Output          none
*       Description		Creates a recombination probability table for the case
*						where interference is considered.
*						Pr(Sc[i,x] | Sc[i-1,x], Sc[i-2,x]).
*       Comments		Used for paternal recombination.
*
******************************************************************************/
int createTableRecombInterferenceMale(person *currPerson, node *pSelectorNode, 
									  Probability **newTable)
{
	int		ind[3], indProb, *selectorVals1, *selectorVals2, *selectorVals3,
			numVals[3], thetaNum, locusIndDS, prevLocusIndDS, prev2LocusIndDS,
			tableSize;
	locus	*locusPtr, *prevLocusPtr, *prev2LocusPtr;
	node	**parents;


	/* 1.	Find the index in the data-structures of the 3 loci on the map. */

	locusIndDS = loci.mapToData[2];
	prevLocusIndDS = loci.mapToData[1];
	prev2LocusIndDS = loci.mapToData[0];

	locusPtr = &(loci.locusArr[locusIndDS]);
	prevLocusPtr = &(loci.locusArr[prevLocusIndDS]);
	prev2LocusPtr = &(loci.locusArr[prev2LocusIndDS]);

	parents = pSelectorNode->parents;
	parents[0] = currPerson->selectorVar[0][prevLocusIndDS];
	parents[0]->numChildren++;
	parents[1] = currPerson->selectorVar[0][prev2LocusIndDS];
	parents[1]->numChildren++;

	/* 2.	Update pointers to the arrays of possible values for each variable of
			the probability table. */

	if (prev2LocusPtr->type == ALLELES)
		selectorVals3 = ((currPerson->loci[prev2LocusIndDS]).locusData).allelesL.paSelectorVals;
	else {
		myMalloc(selectorVals3, sizeof(int) * 2);
		selectorVals3[0] = 0;
		selectorVals3[1] = 1;
	}

	if (prevLocusPtr->type == ALLELES)
		selectorVals2 = ((currPerson->loci[prevLocusIndDS]).locusData).allelesL.paSelectorVals;
	else {
		myMalloc(selectorVals2, sizeof(int) * 2);
		selectorVals2[0] = 0;
		selectorVals2[1] = 1;
	}

	if (locusPtr->type == ALLELES)
		selectorVals1 = ((currPerson->loci[locusIndDS]).locusData).allelesL.paSelectorVals;
	else {
		myMalloc(selectorVals1, sizeof(int) * 2);
		selectorVals1[0] = 0;
		selectorVals1[1] = 1;
	}

	/* 3.	Update the 'numVals' array to hold the number of possible values of 
			each selector variable. */

	numVals[0] = (selectorVals1[1] != -1) ? 2 : 1;
	numVals[1] = (selectorVals2[1] != -1) ? 2 : 1;
	numVals[2] = (selectorVals3[1] != -1) ? 2 : 1;

	/* 4.	Calculate the size of the new probability table and allocate memory for it. */

	tableSize = numVals[0] * numVals[1] * numVals[2];
	myMalloc((*newTable), sizeof(Probability) * tableSize);

	indProb = 0;
	thetaNum = 0;

	/* 5.	Calculate the probabilities in the new table. */

	for (ind[0] = 0; ind[0] < numVals[0]; ind[0]++)
		for (ind[1] = 0; ind[1] < numVals[1]; ind[1]++) 
			for (ind[2] = 0; ind[2] < numVals[2]; ind[2]++) {

				(*newTable)[indProb] = recombTransInterferenceMale(selectorVals3[ind[2]], selectorVals2[ind[1]], 
																   selectorVals1[ind[0]], thetaNum);

				indProb++;
			}

	/* 6.	Free the memory that was allocated dynamically. */

	if (locusPtr->type != ALLELES) {
		myFree(selectorVals1);
	}

	if (prevLocusPtr->type != ALLELES) {
		myFree(selectorVals2);
	}

	if (prev2LocusPtr->type != ALLELES) {
		myFree(selectorVals3);
	}

	/* 7.	Return the size of the probability table created. */

	return tableSize;
}



/******************************************************************************
*
*       Function        createTableRecombInterferenceFemale
*       Parameters		currPerson (the person for which this probability table
*						is created), locusIndMap (the index on the map of the 
*						locus for which this probability table is created), 
*						mSelectorNode (the selector node to which this probability
*						table belongs), newTable (will point to the probability 
*						table created).
*       Output          none
*       Description		creates a recombination probability table for the case
*						where interference is considered.
*						Pr(Sc[i,x] | Sc[i-1,x], Sc[i-2,x]).
*       Comments		Used for maternal recombination.
*
******************************************************************************/
int createTableRecombInterferenceFemale(person *currPerson, node *mSelectorNode, 
										Probability **newTable)
{
	int		ind[3], indProb, *selectorVals1, *selectorVals2, *selectorVals3,
			numVals[3], thetaNum, locusIndDS, prevLocusIndDS, prev2LocusIndDS,
			tableSize;
	locus	*locusPtr, *prevLocusPtr, *prev2LocusPtr;
	node	**parents;

	/* 1.	Find the index in the data-structures of the 3 loci on the map. */

	locusIndDS = loci.mapToData[2];
	prevLocusIndDS = loci.mapToData[1];
	prev2LocusIndDS = loci.mapToData[0];

	locusPtr = &(loci.locusArr[locusIndDS]);
	prevLocusPtr = &(loci.locusArr[prevLocusIndDS]);
	prev2LocusPtr = &(loci.locusArr[prev2LocusIndDS]);

	/* 2.	Update the node's parents and their numChildren field. */

	parents = mSelectorNode->parents;
	parents[0] = currPerson->selectorVar[1][prevLocusIndDS];
	parents[0]->numChildren++;
	parents[1] = currPerson->selectorVar[1][prev2LocusIndDS];
	parents[1]->numChildren++;

	/* 3.	Update pointers to the arrays of possible values for each variable of
			the probability table. */

	if (prev2LocusPtr->type == ALLELES)
		selectorVals3 = ((currPerson->loci[prev2LocusIndDS]).locusData).allelesL.maSelectorVals;
	else {
		myMalloc(selectorVals3, sizeof(int) * 2);
		selectorVals3[0] = 0;
		selectorVals3[1] = 1;
	}

	if (prevLocusPtr->type == ALLELES)
		selectorVals2 = ((currPerson->loci[prevLocusIndDS]).locusData).allelesL.maSelectorVals;
	else {
		myMalloc(selectorVals2, sizeof(int) * 2);
		selectorVals2[0] = 0;
		selectorVals2[1] = 1;
	}

	if (locusPtr->type == ALLELES)
		selectorVals1 = ((currPerson->loci[locusIndDS]).locusData).allelesL.maSelectorVals;
	else {
		myMalloc(selectorVals1, sizeof(int) * 2);
		selectorVals1[0] = 0;
		selectorVals1[1] = 1;
	}

	/* 4.	Update the 'numVals' array to hold the number of possible values of 
			each selector variable. */

	numVals[0] = (selectorVals1[1] != -1) ? 2 : 1;
	numVals[1] = (selectorVals2[1] != -1) ? 2 : 1;
	numVals[2] = (selectorVals3[1] != -1) ? 2 : 1;

	/* 5.	Calculate the size of the new probability table and allocate memory for it. */

	tableSize = numVals[0] * numVals[1] * numVals[2];
	myMalloc((*newTable), sizeof(Probability) * tableSize);

	indProb = 0;
	thetaNum = 0;

	/* 6.	Calculate the probabilities in the new table. */

	for (ind[0] = 0; ind[0] < numVals[0]; ind[0]++)
		for (ind[1] = 0; ind[1] < numVals[1]; ind[1]++) 
			for (ind[2] = 0; ind[2] < numVals[2]; ind[2]++) {

				(*newTable)[indProb] = recombTransInterferenceFemale(selectorVals3[ind[2]], 
																	 selectorVals2[ind[1]], 
																     selectorVals1[ind[0]], thetaNum);

				indProb++;
			}

	/* 7.	Free the memory that was allocated dynamically. */

	if (locusPtr->type != ALLELES) {
		myFree(selectorVals1);
	}

	if (prevLocusPtr->type != ALLELES) {
		myFree(selectorVals2);
	}

	if (prev2LocusPtr->type != ALLELES) {
		myFree(selectorVals3);
	}

	/* 8.	Return the size of the probability table created. */

	return tableSize;
}




/******************************************************************************
*
*       Function        recombTransInterferenceMale
*       Parameters		aSelectorVal (the value of selector a), bSelectorVal
*						(the value of selector b), cSelectorVal (the value of
*						selector c), thetaNum (the index of the recombination
*						fraction between Sc[i-2,x]) and Sc[i-1,x]).
*       Output          none
*       Description		evaluates Pr(Sc[i,x] | Sc[i-1,x], Sc[i-2,x]). 
*			If (Sc[i-2,x] == Sc[i-1,x] == Sc[i,x]), returns g00/(1 - theta).
*		    If (Sc[i-2,x] == Sc[i-1,x] != Sc[i,x]), returns g01/(1 - theta).
*		    If (Sc[i-2,x] != Sc[i-1,x] != Sc[i,x]), returns g11/theta.
*			If (Sc[i-2,x] != Sc[i-1,x] == Sc[i,x]), returns g10/theta.
*       Comments		none
*
******************************************************************************/
double recombTransInterferenceMale(int aSelectorVal, int bSelectorVal, 
							       int cSelectorVal, int thetaNum)
{
	if (aSelectorVal == bSelectorVal) {		/* Sc[i-2,x] == Sc[i-1,x] */
		if (bSelectorVal == cSelectorVal) {	/* Sc[i-1,x] == Sc[i,x] */

			return recombData.maleJointThetas[0]/recombData.maleThetas[thetaNum][1];
		}
		else {								/* Sc[i-1,x] != Sc[i,x] */
			return recombData.maleJointThetas[1]/recombData.maleThetas[thetaNum][1];
		}
	}
	else {									/* Sc[i-2,x] != Sc[i-1,x] */
		if (bSelectorVal == cSelectorVal) {	/* Sc[i-1,x] == Sc[i,x] */

			return recombData.maleJointThetas[2]/recombData.maleThetas[thetaNum][0];
		}
		else {								/* Sc[i-1,x] != Sc[i,x] */
			return recombData.maleJointThetas[3]/recombData.maleThetas[thetaNum][0];
		}
	}
}


/******************************************************************************
*
*       Function        recombTransInterferenceFemale
*       Parameters		aSelectorVal (the value of selector a), bSelectorVal
*						(the value of selector b), cSelectorVal (the value of
*						selector c), thetaNum (the index of the recombination
*						fraction between Sc[i-2,x]) and Sc[i-1,x]).
*       Output          none
*       Description		evaluates Pr(Sc[i,x] | Sc[i-1,x], Sc[i-2,x]). 
*			If (Sc[i-2,x] == Sc[i-1,x] == Sc[i,x]), returns g00/(1 - theta).
*		    If (Sc[i-2,x] == Sc[i-1,x] != Sc[i,x]), returns g01/(1 - theta).
*		    If (Sc[i-2,x] != Sc[i-1,x] != Sc[i,x]), returns g11/theta.
*			If (Sc[i-2,x] != Sc[i-1,x] == Sc[i,x]), returns g10/theta.
*       Comments		none
*
******************************************************************************/
double recombTransInterferenceFemale(int aSelectorVal, int bSelectorVal, 
									 int cSelectorVal, int thetaNum)
{
	if (aSelectorVal == bSelectorVal) {		/* Sc[i-2,x] == Sc[i-1,x] */
		if (bSelectorVal == cSelectorVal) {	/* Sc[i-1,x] == Sc[i,x] */

			return recombData.femaleJointThetas[0]/recombData.femaleThetas[thetaNum][1];
		}
		else {								/* Sc[i-1,x] != Sc[i,x] */
			return recombData.femaleJointThetas[1]/recombData.femaleThetas[thetaNum][1];
		}
	}
	else {									/* Sc[i-2,x] != Sc[i-1,x] */
		if (bSelectorVal == cSelectorVal) {	/* Sc[i-1,x] == Sc[i,x] */

			return recombData.femaleJointThetas[2]/recombData.femaleThetas[thetaNum][0];
		}
		else {								/* Sc[i-1,x] != Sc[i,x] */
			return recombData.femaleJointThetas[3]/recombData.femaleThetas[thetaNum][0];
		}
	}
}



/******************************************************************************
*
*       Function        createTableGeneFrequency
*       Parameters		locusNum (the index in the data-structures of the locus 
*						for which this probability table is created), alleleArr 
*						(an array of all the possible allele values for some 
*						person at this locus), numVals (the number of possible values), 
*						newTable (will point to the new probability table created).
*       Output          none
*       Description		Creates a probability table of gene frequencies for some
*						founder. 
*       Comments		Used in the case when no allele-recoding is performed.
*
******************************************************************************/
int createTableGeneFrequency(int locusNum, int *alleleArr, int numVals,
							 Probability **newTable)
{
	int			ind;
	Probability	*geneFrequencies;

	geneFrequencies = (loci.locusArr[locusNum]).geneFrequencies;

	/* 1.	Allocate memory for the new probability table. */

	myMalloc((*newTable), sizeof(Probability) * numVals);

	/* 2.	Update the probabilities in the new table. */

	if (alleleArr) {
		for (ind = 0; ind < numVals; ind++) {
			(*newTable)[ind] = geneFrequencies[alleleArr[ind]-1]; 
		}
	}
	else {
		(*newTable)[0] = geneFrequencies[0]; 
		(*newTable)[1] = geneFrequencies[1]; 
	}

	/* 3.	Return the size of the proabability table created. */

	return numVals;
}


/******************************************************************************
*
*       Function        createTableGeneFrequencyAR
*       Parameters		locusNum (the index in the data-structures of the locus 
*						for which this probability table is created),
*						alleleSetArr (an array of sets of possible allele values 
*						for some person at this locus), numSets (the number of 
*						sets in the array), newTable (will point to the new 
*						probability table created).
*       Output          none
*       Description		Creates a probability table of gene frequencies. 
*       Comments		Used in the case when allele-recoding is performed.
*
******************************************************************************/
#if PERFORM_ALLELE_RECODING
int createTableGeneFrequencyAR(int locusNum, alleleSet *alleleSetArr,  
							   int numSets, Probability **newTable)
{
	int			ind, indSet;
	Probability	*geneFrequencies;

	geneFrequencies = (loci.locusArr[locusNum]).geneFrequencies;

	/* 1.	Allocate memory for the new probability table. */

	myMalloc((*newTable), sizeof(Probability) * numSets);

	/* 2.	Update the probabilities in the new table. */

	for (ind = 0; ind < numSets; ind++) {
		(*newTable)[ind] = 0;

		for (indSet = 0; indSet < alleleSetArr[ind].numAlleles; indSet++) {
			(*newTable)[ind] += geneFrequencies[alleleSetArr[ind].alleleArr[indSet]-1]; 
		}
	}

	/* 3.	Return the size of the probability table created. */

	return numSets;
}
#endif



/******************************************************************************
*
*       Function        createTableGenotypeFrequency
*       Parameters		locusNum (the index in the data-structures of the locus 
*						for which this probability table is created), 
*						alleleArr (an array of all the possible allele values 
*						for some person at this locus), numVals (the number of 
*						possible values), newTable (will point to the new 
*						probability table created), isUnknownGenotype (a flag 
*						that indicates whether the genotype of the person is 
*						unknown at this locus).
*       Output          none
*       Description		Creates a probability table of genotype frequencies for
*						some founder, at the first locus accrding to map order. 
*       Comments		Used in the case when no allele-recoding is performed. 
*
******************************************************************************/
int createTableGenotypeFrequency(int locusNum, int *alleleArr, int numVals,
							     Probability **newTable, boolean isUnknownGenotype)
{
	int				ind1, ind2, indTable, tableSize;
	Probability		alleleFreq1, alleleFreq2;
	Probability	*geneFrequencies;

	geneFrequencies = (loci.locusArr[locusNum]).geneFrequencies;

	/* 1.	Calculate the size of the table. */

	if (alleleArr) {

		/* 1.1.	This is an ALLELES type locus. */

		if (isUnknownGenotype)
			tableSize = (numVals * (numVals + 1)) / 2 ;
		else
			tableSize = 1;
	}
	else {

		/* 1.1.	 This is an affection type locus or a NULL type locus. */

		tableSize = 3;
	}

	/* 2.	Allocate memory for the new probability table. */

	myMalloc((*newTable), sizeof(Probability) * tableSize);

	/* 3.	Update the probabilities in the new table. */

	indTable = 0;

	if (alleleArr) {			/* this is an ALLELES type locus. */
		if (isUnknownGenotype) {
			for (ind1 = 0; ind1 < numVals; ind1++) {
				for (ind2 = ind1; ind2 < numVals; ind2++) {

					alleleFreq1 = geneFrequencies[alleleArr[ind1]-1];
					alleleFreq2 = geneFrequencies[alleleArr[ind2]-1];

					(*newTable)[indTable] = alleleFreq1 * alleleFreq2; 

					if (ind1 != ind2)
						(*newTable)[indTable] *= 2;

					indTable++;
				}
			}
		}
		else {
			alleleFreq1 = geneFrequencies[alleleArr[0]-1];

			if (numVals == 1)		/* the person is an homozygote in this locus. */
				(*newTable)[0] = alleleFreq1 * alleleFreq1; 
			else {					/* the person is an heterozygote in this locus. */
				alleleFreq2 = geneFrequencies[alleleArr[1]-1];
			
				(*newTable)[0] = alleleFreq1 * alleleFreq2 * 2; 
			}
		}
	}
	else {						/* this is an affection/null type locus. */
		alleleFreq1 = geneFrequencies[0]; 
		alleleFreq2 = geneFrequencies[1];

		(*newTable)[0] = alleleFreq1 * alleleFreq1; 
		(*newTable)[1] = alleleFreq1 * alleleFreq2 * 2; 
		(*newTable)[2] = alleleFreq2 * alleleFreq2; 
	}

	/* 4.	Return the size of the probability table created. */

	return tableSize;
}



/******************************************************************************
*
*       Function        createTableGenotypeFrequencyAR
*       Parameters		locusNum (the index in the data-structures of the locus 
*						for which this probability table is created), alleleSetArr 
*						(an array of sets of possible allele values for some person  
*						at this locus), numSets (the number of sets in the array), 
*						newTable (will point to the new probability table created),
*						isUnknownGenotype (a flag that indicates whether the 
*						genotype of the person is unknown at this locus).
*       Output          none
*       Description		Creates a probability table of genotype frequencies. 
*       Comments		Used in the case when allele-recoding is performed.
*
******************************************************************************/
#if PERFORM_ALLELE_RECODING
int createTableGenotypeFrequencyAR(int locusNum, alleleSet *alleleSetArr, 
								   int numSets, Probability **newTable, 
								   boolean isUnknownGenotype)
{
	int				ind1, ind2, indTable, tableSize, indSet;
	Probability		alleleFreq1, alleleFreq2;
	Probability	*geneFrequencies;

	geneFrequencies = (loci.locusArr[locusNum]).geneFrequencies;
	
	/* 1.	Calculate the size of the table. */

	if (isUnknownGenotype)
		tableSize = (numSets * (numSets + 1)) / 2 ;
	else
		tableSize = 1;

	/* 2.	Allocate memory for the new probability table. */

	myMalloc((*newTable), sizeof(Probability) * tableSize);

	/* 3.	Update the probabilities in the new table. */

	indTable = 0;

	if (isUnknownGenotype) {

		for (ind1 = 0; ind1 < numSets; ind1++) {
			for (ind2 = ind1; ind2 < numSets; ind2++) {

				alleleFreq1 = 0;

				for (indSet = 0; indSet < alleleSetArr[ind1].numAlleles; indSet++)
					alleleFreq1 += geneFrequencies[alleleSetArr[ind1].alleleArr[indSet]-1];

				alleleFreq2 = 0;

				for (indSet = 0; indSet < alleleSetArr[ind2].numAlleles; indSet++)
					alleleFreq2 += geneFrequencies[alleleSetArr[ind2].alleleArr[indSet]-1];

				(*newTable)[indTable] = alleleFreq1 * alleleFreq2; 

				if (ind1 != ind2)
					(*newTable)[indTable] *= 2;

				indTable++;
			}
		}
	}
	else {
		alleleFreq1 = geneFrequencies[alleleSetArr[0].alleleArr[0]-1];

		if (numSets == 1)		/* the person is an homozygote in this locus. */
			(*newTable)[0] = alleleFreq1 * alleleFreq1; 
		else {					/* the person is an heterozygote in this locus. */
			alleleFreq2 = geneFrequencies[alleleSetArr[1].alleleArr[0]-1];
		
			(*newTable)[0] = alleleFreq1 * alleleFreq2 * 2; 
		}
	}

	/* 4.	Return the size of the probability table created. */

	return tableSize;
}
#endif


/******************************************************************************
*
*       Function        createTablePenetrance
*       Parameters		currLoc (the locus of the person for which the penetrance 
*						probability table is created), locusNum (the index in the 
*						data-structures of the locus for which this probability
*						table is created), phenotypeNode (the relevant phenotype
*						node of the person), pLocusNode (the relevant paternal
*						locus node), mLocusNode (the relevant maternal locus
*						node), newTable (will point to the new table created).
*       Output          none
*       Description		Creates a penetrance probability table Pr(X[i]|G[i]),
*						where X is the phenotype and G the genotype.
*       Comments		none
*
******************************************************************************/
int createTablePenetrance(locusPerson *currLoc, int locusNum, node *phenotypeNode, 
						  node *pLocusNode, node *mLocusNode, Probability **newTable)
{
	int					ind0, ind1, indProb, indAffection, liabilityClass;
	femalePenClass		*penetrances;
	affectionStatus		affectionS;


	/* 1.		Update the numChildren field of the phenotypeNode's parents. */

	phenotypeNode->parents[0] = pLocusNode;
	phenotypeNode->parents[1] = mLocusNode;

	pLocusNode->numChildren++;
	mLocusNode->numChildren++;

	/* 2.	Update the affection status and liability class of this person. */

	affectionS = currLoc->locusData.affectionL.affectionS;
	liabilityClass = currLoc->locusData.affectionL.liabilityClass - 1;

	penetrances = ((loci.diseaseModel).oneLocus).fPenetrances;
	indAffection = (affectionS == affected)  ? 0 : 1;

	/* 3.	Allocate memory for the new probability table. */

	myMalloc((*newTable), sizeof(Probability) * 4);

	indProb = 0;

	/* 4.	Initialize the probabilities in the table. */

	for (ind0 = 0; ind0 < 2; ind0++) {
		for (ind1 = 0; ind1 < 2; ind1++) {
			(*newTable)[indProb] = penetrances[liabilityClass][ind0 + ind1][indAffection];

			indProb++;
		}
	}

	/* 5.	Return the size of the probability table created. */

	return 4;
}



/******************************************************************************
*
*       Function        createTablePenetranceFounder
*       Parameters		currLoc (the locus of the person for which the penetrance 
*						probability table is created), locusNum (the index in the 
*						data-structures of the locus for which this probability
*						table is created), phenotypeNode (the relevant phenotype
*						node of the person), pLocusNode (the relevant paternal
*						locus node), mLocusNode (the relevant maternal locus
*						node), genotypeNode (the relevant genotype node of the
*						founder), newTable (will point to the new table created).
*       Output          none
*       Description		Creates a penetrance probability table Pr(X[i]|G[i])
*						for a founder.
*       Comments		none
*
******************************************************************************/
int createTablePenetranceFounder(locusPerson *currLoc, int locusNum, 
								 node *phenotypeNode, node *pLocusNode,
								 node *mLocusNode, node *genotypeNode, 
								 Probability **newTable)
{
	int					ind0, ind1, indProb, indAffection, liabilityClass;
	femalePenClass		*penetrances;
	affectionStatus		affectionS;
   
	/* 1.		Update the numChildren field of the phenotypeNode's parents. */

	pLocusNode->numChildren++;
	mLocusNode->numChildren++;
	genotypeNode->numChildren++;

	phenotypeNode->parents[0] = genotypeNode;

	/* 2.	Update the affection status and liability class of this person. */

	affectionS = currLoc->locusData.affectionL.affectionS;
	liabilityClass = currLoc->locusData.affectionL.liabilityClass - 1;
	
	penetrances = (loci.diseaseModel).oneLocus.fPenetrances;
	indAffection = (affectionS == affected)  ? 0 : 1;

	/* 3.	Allocate memory for the new probability table. */

	myMalloc((*newTable), sizeof(Probability) * 3);

	indProb = 0;

	/* 4.	Initialize the probabilities in the table. */

	for (ind0 = 0; ind0 < 2; ind0++) {
		for (ind1 = ind0; ind1 < 2; ind1++) {
			(*newTable)[indProb] = penetrances[liabilityClass][ind0 + ind1][indAffection];

			indProb++;
		}
	}

	/* 5.	Return the size of the probability table created. */

	return 3;
}



/******************************************************************************
*
*       Function        createTableMalePenetrance
*       Parameters		currLoc (the locus of the person for which the penetrance 
*						probability table is created), locusNum (the index in the 
*						data-structures of the locus for which this probability
*						table is created), phenotypeNode (the relevant phenotype
*						node of the person), mLocusNode (the relevant maternal 
*						locus node), newTable (will point to the new table created).
*       Output          none
*       Description		Creates a penetrance probability table Pr(X[i]|G[i])
*						for a male in the case of sex-linked analysis.
*       Comments		Used in the case of a sex-linked analysis.
*
******************************************************************************/
int createTableMalePenetrance(locusPerson *currLoc, int locusNum, 
							  node *phenotypeNode, node *mLocusNode,
							  Probability **newTable)
{
	int					indAffection, liabilityClass;
	malePenClass		*penetrances;
	affectionStatus		affectionS;


    /* 1.	Update the numChildren field of this node's parents. */

	mLocusNode->numChildren++;

	phenotypeNode->parents[0] = mLocusNode;

	/* 2.	Update the affection status and liability class of this person. */

	affectionS = currLoc->locusData.affectionL.affectionS;
	liabilityClass = currLoc->locusData.affectionL.liabilityClass - 1;

	penetrances = (loci.diseaseModel).oneLocus.mPenetrances;
	indAffection = (affectionS == affected)  ? 0 : 1;

	/* 3.	Allocate memory for the new probability table. */

	myMalloc((*newTable), sizeof(Probability) * 2);

	/* 4.	Initialize the probabilities in the table. */

	(*newTable)[0] = penetrances[liabilityClass][0][indAffection];
	(*newTable)[1] = penetrances[liabilityClass][1][indAffection];

	/* 5.	Return the size of the probability table created. */

	return 2;
}



/******************************************************************************
*
*       Function        createTablePenetranceTwo
*       Parameters		currPerson (the person for whom this probability table
*						is created), firstNullLociIndDS (the index in the data-
*						structures of the first null type locus), 
*						secondNullLociIndDS (the index in the data-structures 
*						of the second null type locus), phenotypeNode (the 
*						relevant phenotype node of the person), newTable (will  
*						point to the new table created).
*       Output          none
*       Description		Creates a penetrance probability table Pr(X[i]|G[i],G[j]),
*						for the case of a two-locus disease.
*       Comments		Used in the case of a two-locus disease.
*
******************************************************************************/
int createTablePenetranceTwo(person *currPerson, int firstNullLociIndDS, 
							 int secondNullLociIndDS, node *phenotypeNode, 
							 Probability **newTable)
{
	int					ind0, ind1, ind2, ind3, indProb, indAffection, liabilityClass;
	femalePenClassTwo	*penetrances;
	affectionStatus		affectionS;
	locusPerson			*currLoc;
	node				**parents;

	/* 1.	Update the parents of the phenotype node. */

	parents = phenotypeNode->parents;

    parents[0] = currPerson->locusVar[0][firstNullLociIndDS];
	parents[1] = currPerson->locusVar[1][firstNullLociIndDS];
	parents[2] = currPerson->locusVar[0][secondNullLociIndDS];
	parents[3] = currPerson->locusVar[1][secondNullLociIndDS];

	(parents[0])->numChildren++;
	(parents[1])->numChildren++;
	(parents[2])->numChildren++;
	(parents[3])->numChildren++;

	/* 2.	Update the affection status and liability class of this person. */

	currLoc = &(currPerson->loci[secondNullLociIndDS]);
	affectionS = currLoc->locusData.affectionL.affectionS;
	liabilityClass = currLoc->locusData.affectionL.liabilityClass - 1;

	penetrances = (loci.diseaseModel).twoLocus.fPenetrances;
	indAffection = (affectionS == affected)  ? 0 : 1;

	/* 3.	Allocate memory for the new probability table. */

	myMalloc((*newTable), sizeof(Probability) * 16);

	indProb = 0;

	/* 4.	Initialize the probabilities in the table. */

	for (ind0 = 0; ind0 < 2; ind0++) {
		for (ind1 = 0; ind1 < 2; ind1++) {
			for (ind2 = 0; ind2 < 2; ind2++) {
				for (ind3 = 0; ind3 < 2; ind3++) {
					(*newTable)[indProb] = 
							penetrances[liabilityClass][ind0 + ind1][ind2 + ind3][indAffection];

					indProb++;
				}
			}
		}
	}

	/* 5.	Return the size of the probability table created. */

	return 16;
}



/******************************************************************************
*
*       Function        createTablePenetranceFounderTwo
*       Parameters		currPerson (the person for whom this probability table
*						is created), firstNullLociIndDS (the index in the data-
*						structures of the first null type locus), 
*						secondNullLociIndDS (the index in the data-structures 
*						of the second null type locus), phenotypeNode (the 
*						relevant phenotype node of the person), genotypeNode 
*						(the relevant genotype node of the person), 
*						firstNullLociPosMap (the position on the map of the 
*						first null locus), newTable (will point to the new
*						table created).
*       Output          none
*       Description		Creates a penetrance probability table Pr(X[i]|G[i], G[j]),
*						for a founder, for the case of a two-locus disease.
*       Comments		used in the case of a two-locus disease.
*
******************************************************************************/
int createTablePenetranceFounderTwo(person *currPerson, int firstNullLociIndDS, 
									int secondNullLociIndDS, node *phenotypeNode, 
									node *genotypeNode, int firstNullLociPosMap,
									Probability **newTable)
{
	int					ind0, ind1, ind2, ind3, indProb, indAffection, liabilityClass;
	femalePenClassTwo	*penetrances;
	affectionStatus		affectionS;
	locusPerson			*currLoc;
	node				**parents;

	/* 1.	Update the numChildren field of this node's parents. */

	(currPerson->locusVar[0][firstNullLociIndDS])->numChildren++;
	(currPerson->locusVar[1][firstNullLociIndDS])->numChildren++;
	(currPerson->locusVar[0][secondNullLociIndDS])->numChildren++;
	(currPerson->locusVar[1][secondNullLociIndDS])->numChildren++;

	genotypeNode->numChildren++;

	parents = phenotypeNode->parents;
	parents[0] = genotypeNode;

	if (firstNullLociPosMap == 0) {
		parents[1] = currPerson->locusVar[0][secondNullLociIndDS];
		parents[2] = currPerson->locusVar[1][secondNullLociIndDS];
	}
	else {
		parents[1] = currPerson->locusVar[0][firstNullLociIndDS];
		parents[2] = currPerson->locusVar[1][firstNullLociIndDS];
	}

	/* 2.	Update the affection status and liability class of this person. */

	currLoc = &(currPerson->loci[secondNullLociIndDS]);
	affectionS = currLoc->locusData.affectionL.affectionS;
	liabilityClass = currLoc->locusData.affectionL.liabilityClass - 1;

	penetrances = (loci.diseaseModel).twoLocus.fPenetrances;
	indAffection = (affectionS == affected)  ? 0 : 1;

	/* 3.	Allocate memory for the new probability table. */

	myMalloc((*newTable), sizeof(Probability) * 12);

	indProb = 0;

	/* 4.	Initialize the probabilities in the table. */

	for (ind0 = 0; ind0 < 2; ind0++) {
		for (ind1 = ind0; ind1 < 2; ind1++) {
			for (ind2 = 0; ind2 < 2; ind2++) {
				for (ind3 = 0; ind3 < 2; ind3++) {
					(*newTable)[indProb] = 
							penetrances[liabilityClass][ind0 + ind1][ind2 + ind3][indAffection];

					indProb++;
				}
			}
		}
	}

	/* 5.	Return the size of the probability table created. */

	return 12;
}


/******************************************************************************
*
*       Function        createTableMalePenetranceTwo
*       Parameters		currPerson (the person for whom this probability table
*						is created), firstNullLociIndDS (the index in the data-
*						structures of the first null type locus), 
*						secondNullLociIndDS (the index in the data-structures 
*						of the second null type locus), phenotypeNode (the 
*						relevant phenotype node of the person), newTable (will  
*						point to the new table created).
*       Output          none
*       Description		Creates a penetrance probability table Pr(X[i]|G[i], G[j]),
*						for a male, for the case of a two-locus disease in a 
*						sex-linked analysis.
*       Comments		Used in the case of a two-locus disease in a sex-linked
*						analysis.
*
******************************************************************************/
int createTableMalePenetranceTwo(person *currPerson, int firstNullLociIndDS, 
 							     int secondNullLociIndDS, node *phenotypeNode,
								 Probability **newTable)
{
	int					ind0, ind1, indProb, indAffection, liabilityClass;
	malePenClassTwo		*penetrances;
	affectionStatus		affectionS;
	locusPerson			*currLoc;
	node				**parents;

	/* 1.	Update the numChildren field of this node's parents. */

	parents = phenotypeNode->parents;

	parents[0] = currPerson->locusVar[1][firstNullLociIndDS];
	parents[1] = currPerson->locusVar[1][secondNullLociIndDS];

	(parents[0])->numChildren++;
	(parents[1])->numChildren++;

	/* 2.	Update the person's affection status and liability class. */

	currLoc = &(currPerson->loci[secondNullLociIndDS]);
	affectionS = currLoc->locusData.affectionL.affectionS;
	liabilityClass = currLoc->locusData.affectionL.liabilityClass - 1;

	penetrances = (loci.diseaseModel).twoLocus.mPenetrances;
	indAffection = (affectionS == affected)  ? 0 : 1;

	/* 3.	Allocate memory for the new probability table. */

	myMalloc((*newTable), sizeof(Probability) * 4);

	/* 4.	Initialize the probabilities in the table. */

	indProb = 0;

	for (ind0 = 0; ind0 < 2; ind0++) {
		for (ind1 = 0; ind1 < 2; ind1++) {
	
			(*newTable)[indProb] = penetrances[liabilityClass][ind0][ind1][indAffection];

			indProb++;
		}
	}

	/* 5.	Return the size of the probability table created. */

	return 4;
}


/******************************************************************************
*
*       Function        createTablePhenotype
*       Parameters		phenotypeNode (the phenotype node fto which this 
*						probability table is created), pLocusNode (the relevant
*						paternal locus node), mLocusNode (the relevant maternal 
*						locus node), newTable (will point to the new probability  
*						table created).
*       Output          none
*       Description		Creates a phenotype probability table. (Entrances of 
*						impossible genotypes - homozygotes, will be zero, and 
*						the other 2 entrances will get 1 probability.)
*       Comments		none
*
******************************************************************************/
int createTablePhenotype(node *phenotypeNode, node *pLocusNode, node *mLocusNode,
						 Probability **newTable)
{
	/* 1.	Update the numChildren field of this node's parents. */

	pLocusNode->numChildren++;
	mLocusNode->numChildren++;

	phenotypeNode->parents[0] = pLocusNode;
	phenotypeNode->parents[1] = mLocusNode;

	/* 2.	Allocate memory for the new probability table. */

	myMalloc((*newTable), sizeof(Probability) * 4);

	/* 3.	Initialize the probabilities in the table. */

	(*newTable)[0] = (*newTable)[3] = 0.0;
	(*newTable)[1] = (*newTable)[2] = 1.0;

	/* 4.	Return the size of the probability table created. */

	return 4;
}



/******************************************************************************
*
*       Function        multiplyAllFuncsSimp
*       Parameters		combinedFunc (will store the result of the multiplication),
*						elimVar (the variable that needs to be eliminated).
*       Output          none
*       Description		Performs a multiplication of the probability functions
*						that elimVar is part of and returns the result of the 
*						multiplication in combinedFunc.
*       Comments		none
*
******************************************************************************/
void multiplyAllFuncsSimp(function *combinedFunc, variable *elimVar)
{
	Probability *probabListPtr,			/* pointer to the entry of the combined 
										   function that is currently being updated. */
				funcsVal, currVal, totalVal;
	function	**funcPtr, **funcPtrEnd;
	variable	**varPtr, **varPtrEnd, *var;

	/* 1.	Initialize the values of the variables of the combined function
			to values that correspond to the first entry in the probability table. */

	initFunctionVars(combinedFunc, varPtr, varPtrEnd);

	probabListPtr = combinedFunc->probabList;	
	funcPtrEnd = elimVar->functions + elimVar->currNumFunctions;

	/* 2.	Initialize the jumpFactor of each of the functions that are about 
			to be combined. */

	for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {
		initJumpFactor(*funcPtr, elimVar);
	}

	/* 3.	Iterate over the entries of the new probability table and update them. */

	while (1) {
		
		start_loop_mafs:

		/* 3.1.	Initialize the value of each entry in the new probability table to zero. */

		totalVal = 0.0;

		elimVar->currValue = 0;

		/* 3.2.	Calculate, for each function, the index in the probability table that
				corresponds to the current values of its variables. */

		for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {
			calcBaseIndex(*funcPtr);
		}

		/* 3.3.	Iterate over all of the eliminated variable's values and sum up the
				probabilities for each value. */	

		for (; elimVar->currValue < elimVar->numValues; elimVar->currValue++) {
			
			/* 3.3.1.	Initialize the table entry for this variable value to 1.0 . */

			currVal = 1.0;

			/* 3.3.2.	Iterate over all the functions. For each function: */

			for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {

				/* 3.3.2.1.	Get the function's value. */

				funcsVal = getFuncValue((*funcPtr), elimVar->currValue);

				/* 3.3.2.2.	If the value is zero, we can continue to the next 
							variable value. */

				if (funcsVal == 0.0) {
					goto preceedToNextValue;
				}

				/* 3.3.2.3.	Otherwise, multiply the current value for this  
							variable value, by the function's value. */

				currVal *= funcsVal;
			}
			
			/* 3.3.3.	Add the value for this variable value to the total value 
						for this entry. */

			totalVal += currVal;

			preceedToNextValue: ;
		}

		/* 3.4.	Write the value into the new probability table. */

		*probabListPtr = totalVal;
		
		/* 3.5.	Advance the pointer to the next entry in the table. */

		probabListPtr++;

		/* 3.6.	Advance the values of the function's variables to correspond to
				the next entry in the table. */

		advanceFunctionVars(combinedFunc, varPtr, varPtrEnd, var, start_loop_mafs);
	}
}


/******************************************************************************
*
*       Function        multiplyAllFuncsComplex
*       Parameters		combinedFunc (will store the result of the multiplication),
*						elimVar (the variable that needs to be eliminated).
*       Output          none
*       Description		Performs a multiplication of the probability functions
*						that elimVar is part of and returns the result of the 
*						multiplication.
*       Comments		A more sophisticated function than the previous one.
*
******************************************************************************/
void multiplyAllFuncsComplex(function *combinedFunc, variable *elimVar)
{
	Probability		*probabListPtr, funcsVal, currVal, totalVal;
	function		**funcPtr, **funcPtrEnd, *func;
	variable		**varPtr, **varPtrEnd, *var, **lastVar;
	int				indInCombinedFunc, varInd, lastVarInd;


	/* 1.	Use the 'currValue' field of each variable to store its index in 
			the combinedFunc's varArray, temporarily. */

	for (indInCombinedFunc = 0, varPtr = combinedFunc->variables; 
			indInCombinedFunc < combinedFunc->numVariables; indInCombinedFunc++, varPtr++) {
		
		(*varPtr)->currValue = indInCombinedFunc;
	}

	elimVar->currValue = indInCombinedFunc;

	funcPtrEnd = elimVar->functions + elimVar->currNumFunctions;

	/* 2.	Iterate over the functions that need to be combined. For each one: */  

	for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {

		func = *funcPtr;

		/* 2.1.	Calculate the necessary increment in the index of the function's 
				probability table as a result of an increase in one of the 
				combined function's variables. */

		calcAccOffsetElim(func, combinedFunc, elimVar);

		/* 2.2.	Initialize the index of the function's probability table to the first entry. */

		func->baseInd = 0;
	}

	/* 3.	Initialize the values of the variables of the combined function
			to values that correspond to the first entry in the probability table. */

	initFunctionVars(combinedFunc, varPtr, varPtrEnd);

	probabListPtr = combinedFunc->probabList;	
	varPtrEnd = combinedFunc->variables;
	lastVar = varPtrEnd + combinedFunc->numVariables - 1;
	lastVarInd = combinedFunc->numVariables - 1;

	/* 4.	Iterate over the entries of the new probability table and update them. */

	while (1) {	
		start_loop:

		/* 4.1.	Initialize the value of each entry in the new probability table to zero. */

		totalVal = 0.0;

		/* 4.2.	Iterate over all of the eliminated variable's values. */

		elimVar->currValue = 0;

		/* Handles the case where the eliminated variable has only 2 possible values. */
		if (elimVar->numValues == 2) {

			/* VALUE 0 */
			currVal = 1.0;

			/* 4.2.1.	Iterate over all the functions. */

			for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {

				func = *funcPtr;

				/* 4.2.1.1.	Get the function's value. */

				funcsVal = getFuncVal(func);

				/* 4.2.1.2.	If the value is zero, we can continue to the next variable value. */

				if (funcsVal == 0.0) {
					goto valueOne;
				}

				/* 4.2.1.3.	Otherwise, multiply the value for this variable value, by the 
							function's value. */

				currVal *= funcsVal;
			}
			
			/* 4.2.2.	Add the value for this variable value to the total value for this entry. */

			totalVal += currVal;

			/* VALUE 1 */
			valueOne:   currVal = 1.0;

			elimVar->currValue = 1;

			/* 4.2.3.	Iterate over all the functions. */

			for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {

				func = *funcPtr;

				/* 4.2.3.1.	Update the index of this function's probability table. */ 

				func->baseInd += func->accOffset[indInCombinedFunc];

				/* 4.2.3.2. Get the function's value. */

				funcsVal = getFuncVal(func);

				/* 4.2.3.3.	If the value is zero, we can continue to the next variable value. */

				if (funcsVal == 0.0) {

					/* 4.2.3.3.1.	Update the indices of the rest of the functions'
									probability tables. */ 

					for (funcPtr++; funcPtr < funcPtrEnd; funcPtr++) {
						func = *funcPtr;
						func->baseInd += func->accOffset[indInCombinedFunc];
					}
					goto End;
				}

				/* 4.2.3.4.	Otherwise, multiply the value for this variable value, by the 
							function's value. */

				currVal *= funcsVal;
			}
			
			/* 4.2.4.	Add the value for this variable value to the total value for this entry. */

			totalVal += currVal;
					
			End:;
		}
		else {	/* handles the regular case where the eliminated variable has any number of
			       possible values. */
			while (elimVar->currValue < elimVar->numValues) {
				
				currVal = 1.0;
				elimVar->currValue++;

				/* 4.2.1.	Iterate over all the functions. */

				for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {

					func = *funcPtr;

					/* 4.2.1.1.	Get the function's value. */

					funcsVal = getFuncVal(func);

					/* 4.2.1.1.	If this isn't the last value of the eliminated variable, 
								advance the indcies of the functions' probability tables. */

					if (elimVar->currValue < elimVar->numValues) {
						func->baseInd += func->accOffset[indInCombinedFunc];
					}
					
					/* 4.2.1.2.	If the value is zero, we can continue to the next variable value. */

					if (funcsVal == 0.0) {

						/* 4.2.1.2.1.	If this isn't the last value of the eliminated 
										variable, update the indices of the rest of the 
										function's probability tables. */ 

						if (elimVar->currValue < elimVar->numValues) {
							for (funcPtr++; funcPtr < funcPtrEnd; funcPtr++) {
								func = *funcPtr;
								func->baseInd += func->accOffset[indInCombinedFunc];
							}
						}

						goto preceedToNextValue;
					}

					/* 4.2.1.3.	Otherwise, multiply the value for this variable value, 
								by the function's value. */

					currVal *= funcsVal;
				}
				
				/* 4.2.2.	Add the value for this variable value to the total value for 
							this entry. */

				totalVal += currVal;

				preceedToNextValue: ;
			}
		}

		/* 4.3.	Write the value into the new probability table. */

		*probabListPtr = totalVal;

		/* 4.4.	Advance the pointer to the next entry in the table. */

		probabListPtr++;

		/* 4.5.	Update the variables' currValues (to match the next entrance in the 
				joint probability table) and update the index of the functions which
				are being joined. */

		varPtr = lastVar;
		varInd = lastVarInd;

		while (varPtr >= varPtrEnd) {
			var = *varPtr;

			if (var->currValue < (var->numValues - 1)) {
				var->currValue++;

				for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {
					(*funcPtr)->baseInd += (*funcPtr)->accOffset[varInd];
				}

				goto start_loop;
			}
			else {
				var->currValue = 0;
			}

			varPtr--;
			varInd--;
		}

		break;
	} 
}



/******************************************************************************
*
*       Function        multiplyAllOneVarFuncs
*       Parameters		elimVar (the variable that needs to be eliminated), 
*						logMultFactor (the sum of the logMultFactors of all the 
*						functions that take part in the multiplication).
*       Output          none
*       Description		Performs a multiplication of the probability functions
*						that elimVar appears in, sums all the entries and returns 
*						the log-result of the multiplication.
*       Comments		none
*
******************************************************************************/
Probability multiplyAllOneVarFuncs(variable *elimVar, double logMultFactor)
{
	Probability		currVal, totalVal, funcVal;
	function		**funcPtr, **funcPtrEnd;
	int				val;


	funcPtrEnd = elimVar->functions + elimVar->currNumFunctions;
	
	totalVal = 0.0;

	/* 1.	Calculate the probability for each value of the eliminated variable
			and sum them up. */

	for (val = 0; val < elimVar->numValues; val++) {
		
		currVal = 1.0;

		for (funcPtr = elimVar->functions; funcPtr < funcPtrEnd; funcPtr++) {

			funcVal = (*funcPtr)->probabList[val];

			if (funcVal == 0.0) {
				goto valueWasZero;
			}

			currVal *= funcVal;
		}

		totalVal += currVal;

		valueWasZero: ;
	}

	/* 2.	Update the result of the multiplication according to the logMultFactor. */

	return (log(totalVal) + logMultFactor);
}


/******************************************************************************
*
*       Function        multiplyAllFuncs
*       Parameters		numFuncs (the number of functions that need to be 
*						multiplied), funcArr (an array of pointers to the 
*						functions that need to be multiplied),
*						combinedFunc (will hold the resulting function).
*       Output          none
*       Description		Performs a multiplication of the probability functions
*						in 'funcArr' and stores the result of the multiplication
*						in 'combinedFunc'.
*       Comments		none.
*
******************************************************************************/
void multiplyAllFuncs(int numFuncs, function **funcArr, function *combinedFunc)
{
	Probability		*probabListPtr, funcsVal, currVal;
	function		**funcPtr, **funcPtrEnd, *func;
	variable		**varPtr, **varPtrEnd, **varPtrBegin, *var;
	int				indInCombinedFunc, varInd, varIndBegin, rescaleTimer, numVars;
	double			logMultFactor;

	/* 1.	Store the number of variables in the combined function. */
	
	numVars = combinedFunc->numVariables;

	/* 2.	Use the 'currValue' field of each variable to store its index in 
			the combinedFunc's varArray, temporarily. */

	for (indInCombinedFunc = 0, varPtr = combinedFunc->variables; 
			indInCombinedFunc < numVars; indInCombinedFunc++, varPtr++) {
		
		(*varPtr)->currValue = indInCombinedFunc;
	}

	funcPtrEnd = funcArr + numFuncs;
	logMultFactor = 0.0;
	rescaleTimer = MAX_RESCALE_UNITS;

	/* 3.	Iterate over the functions that need to be combined. For each one: */  

	for (funcPtr = funcArr; funcPtr < funcPtrEnd; funcPtr++) {

		func = (*funcPtr);

		/* 3.1.	Calculate the necessary increment in the index of the function's 
				probability table as a result of an increase of 1 in one of the combined 
				function's variables. */

		calcAccOffset(func, combinedFunc);

		/* 3.2.	Initialize the index of the function's probability table to the first entry. */

		func->baseInd = 0;

		/* 3.3.	Update the 'logMultFactor' and the 'rescaleTimer' fields of the combined function. */

		logMultFactor += func->logMultFactor;

		if (func->rescaleTimer < rescaleTimer)
			rescaleTimer = func->rescaleTimer;
	}

	combinedFunc->logMultFactor = logMultFactor;
	combinedFunc->rescaleTimer = rescaleTimer - 1;

	/* 4.	Initialize the values of the variables of the combined function
			to values that correspond to the first entry in the probability table. */

	initFunctionVars(combinedFunc, varPtr, varPtrEnd);

	probabListPtr = combinedFunc->probabList;	
	varPtrEnd = combinedFunc->variables;

	varPtrBegin = combinedFunc->variables + numVars - 1;
	varIndBegin = numVars - 1;

	/* 5.	Iterate over the entries of the new probability table and update them. */

	while (1) {	
		start_loop:

		/* 5.1.	Initialize the value of each entry in the new probability table to zero. */

		currVal = 1.0;

		/* 5.2.	Iterate over all the functions. */

		for (funcPtr = funcArr; funcPtr < funcPtrEnd; funcPtr++) {

			/* 5.2.1.	Get the function's value. */

			funcsVal = getFuncVal((*funcPtr));
					
			/* 5.2.2.	If the value is zero, we can continue to the next entry. */

			if (funcsVal == 0.0) {
				currVal = 0.0;
				goto preceedToNextEntry;
			}

			/* 5.2.3.	Otherwise, multiply the value for this variable value, by the 
						function's value. */

			currVal *= funcsVal;
		}
				
		preceedToNextEntry:

		/* 5.3.	Write the value into the new probability table. */

		*probabListPtr = currVal;

		/* 5.4.	Advance the pointer to the next entry in the table. */

		probabListPtr++;

		/* 5.5.	Update the variables' currValues (to match the next entrance in the 
				joint probability table) and update the baseInd of the functions which
				are being joined. */

		varPtr = varPtrBegin;
		varInd = varIndBegin;

		while (varPtr >= varPtrEnd) {
			var = *varPtr;

			if (var->currValue < (var->numValues - 1)) {
				var->currValue++;

				for (funcPtr = funcArr; funcPtr < funcPtrEnd; funcPtr++) {
					(*funcPtr)->baseInd += (*funcPtr)->accOffset[varInd];
				}

				goto start_loop;
			}
			else {
				var->currValue = 0;
			}

			varPtr--;
			varInd--;
		}

		break;
	} 
}




/******************************************************************************
*
*       Function        multiplyAllFuncsJointComplex
*       Parameters		combinedFunc (will store the result of the multiplication),
*						elimVarsArr (the array of variables that needs to be 
*						eliminated), numElimVars (the number of variables that
*						need to be eliminated), funcArr (the array of pointers 
*						to the functions that need to be multiplied), numFuncs 
*						(the number of functions in the array).
*       Output          none
*       Description		Performs a multiplication of the probability functions
*						that elimVars are part of and returns the result of the 
*						multiplication.
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void multiplyAllFuncsJointComplex(function *combinedFunc, variable **elimVarsArr, 
								  int numElimVars, function **funcArr, int numFuncs)
{
	Probability			*probabListPtr, funcsVal, currVal, totalVal;
	function			**funcPtr, **funcPtrEnd;
	variable			**varPtr, **varPtrEnd, *var, **elimVarsPtr, **elimVarsPtrEnd,
						**varPtrBegin;
	incDec				*incDecArr, *incDecPtr, *incDecPtrEnd;
	int					indInCombinedFunc, varInd, varIndBegin, funcInd;
	validAssignments	validAssignmentsArr;
	assignment			*assignmentPtr, *assignmentPtrEnd;
	int					*incPtr, *val;


	/* 1.	Determine all the valid joint assignments to the variables that need 
			to be eliminated. */

	findValidJointAssignments(elimVarsArr, numElimVars, &validAssignmentsArr, numFuncs);

	/* 2.	Allocate memory for an auxiliary array. */

	myMalloc(incDecArr, sizeof(incDec) * numFuncs * combinedFunc->numVariables);

	/* 3.	Use the 'currValue' field of each variable to store its index in 
			the combinedFunc's variable Array, temporarily. */

	for (indInCombinedFunc = 0, varPtr = combinedFunc->variables; 
			indInCombinedFunc < combinedFunc->numVariables; indInCombinedFunc++, varPtr++) {
		
		(*varPtr)->currValue = indInCombinedFunc;
	}

	for (varPtr = elimVarsArr, varPtrEnd = varPtr + numElimVars; varPtr < varPtrEnd; varPtr++) {
		(*varPtr)->currValue = -1;	
	}

	/* 4.	Initialize all incDecArr's entries to zero. */

	for (incDecPtr = incDecArr, 
			incDecPtrEnd = incDecPtr + numFuncs * combinedFunc->numVariables;
			incDecPtr < incDecPtrEnd; incDecPtr++) {

		incDecPtr->inc = 0;
		incDecPtr->dec = 0;
	}

	funcPtrEnd = funcArr + numFuncs;
	assignmentPtrEnd = validAssignmentsArr.assignmentArr + validAssignmentsArr.numAssignments; 

	/* 5.	Calculate the necessary increment in the index of the probability table 
			of each function as a result of an increase of 1 in one of the function's 
			variables. Calculate the necessary deccrement in the index of the probability 
			table of each function as a result of an decrease of (numVals-1) in one of 
			the function's variables. Also calculate the necessary increment in the index of
			of each function between each 2 successive valid assignments to the eliminated 
			variables. */

	initJumpFactorsJoint(funcArr, numFuncs, elimVarsArr, numElimVars, incDecArr, 
						 combinedFunc->variables, combinedFunc->numVariables, &validAssignmentsArr);

	/* 6.	Iterate over the functions that need to be combined, and initialize the 
			index of each one to zero. */

	for (funcPtr = funcArr, funcPtrEnd = funcPtr + numFuncs; funcPtr < funcPtrEnd; 
			funcPtr++) {
		(*funcPtr)->baseInd = 0;
	}

	/* 7.	Initialize the values of the variables of the combined function
			to values that correspond to the first entry in the probability table. */

	initFunctionVars(combinedFunc, varPtr, varPtrEnd);

	probabListPtr = combinedFunc->probabList;	
	varPtrEnd = combinedFunc->variables;

	varPtrBegin = combinedFunc->variables + combinedFunc->numVariables - 1;
	varIndBegin = combinedFunc->numVariables - 1;

	/* 8.	Iterate over the entries of the new probability table and update them. */

	while (1) {	
		start_loop:

		/* 8.1.	Initialize the value of each entry in the new probability table to zero. */

		totalVal = 0.0;

		/* 8.2.	Iterate over all of the valid assignment for the eliminated variables. */

		assignmentPtr = validAssignmentsArr.assignmentArr;

		for (; assignmentPtr < assignmentPtrEnd; assignmentPtr++) {
			
			currVal = 1.0;

			/* 8.2.1.	Update the values of the eliminated variables according to this assignment. */

			setAssignment(elimVarsArr, numElimVars, assignmentPtr->varVals, val, elimVarsPtr, 
						  elimVarsPtrEnd);

			/* 8.2.2.	Iterate over all the functions. For each function: */

			for (funcPtr = funcArr, incPtr = assignmentPtr->inc; funcPtr < funcPtrEnd;
					 funcPtr++, incPtr++) {

				/* 8.2.2.1.	Get the function's value. */

				funcsVal = (*funcPtr)->probabList[(*funcPtr)->baseInd + (*incPtr)];
			
				/* 8.2.2.2.	If the value is zero, we can continue to the next valid assignment. */

				if (funcsVal == 0.0) {
					goto preceedToNextValue;
				}

				/* 8.2.2.3.	Otherwise, multiply the value for this variable value, by the 
							function's value. */

				currVal *= funcsVal;
			}
			
			/* 8.2.3.	Add the value for this assignment to the total value for 
						this entry. */

			totalVal += currVal;

			preceedToNextValue: ;
		}

		/* 8.3.	Write the value into the new probability table. */

		*probabListPtr = totalVal;

		/* 8.4.	Advance the pointer to the next entry in the table. */

		probabListPtr++;

		/* 8.5.	Update the variables' currValues (to match the next entrance in the 
				joint probability table) and update the baseInd of the functions which
				are being joined. */

		varPtr = varPtrBegin;
		varInd = varIndBegin;

		while (varPtr >= varPtrEnd) {
			var = *varPtr;

			if (var->currValue < (var->numValues - 1)) {
				var->currValue++;

				for (funcPtr = funcArr, funcInd = 0; funcPtr < funcPtrEnd; funcPtr++, 
						funcInd += combinedFunc->numVariables) {

					(*funcPtr)->baseInd += incDecArr[funcInd + varInd].inc;
				}

				goto start_loop;
			}
			else {
				var->currValue = 0;

				for (funcPtr = funcArr,	funcInd = 0; funcPtr < funcPtrEnd; funcPtr++, 
						funcInd += combinedFunc->numVariables) {

					(*funcPtr)->baseInd += incDecArr[funcInd + varInd].dec;
				}
			}

			varPtr--;
			varInd--;
		}

		break;
	} 

	/* 9.	Free the memory that was allocated dynamically. */

	myFree(incDecArr);

	freeAssignmentMem(&validAssignmentsArr);
}
#endif


/******************************************************************************
*
*       Function        multiplyAllFuncsJoint
*       Parameters		elimVarsArr (the array of variables that needs to be 
*						eliminated), numElimVars (the number of variables that
*						need to be eliminated), logMultFactor (the sum of the 
*						logMultFactors of all the functions that take part in the 
*						multiplication), funcArr (the array of pointers 
*						to the functions that need to be multiplied), numFuncs 
*						(the number of functions in the array).
*       Output          none
*       Description		Performs a multiplication of the probability functions
*						that elimVars appear in, sums all the entries and returns 
*						the log-result of the multiplication.
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
Probability multiplyAllFuncsJoint(variable **elimVarsArr, int numElimVars,
								  double logMultFactor, function **funcArr, 
								  int numFuncs)
{
	Probability			currVal, totalVal, funcVal;
	function			**funcPtr, **funcPtrEnd, *func;
	variable			**varPtr, **varPtrEnd;
	validAssignments	validAssignmentsArr;
	assignment			*assignmentPtr, *assignmentPtrEnd;
	int					*val;


	/* 1.	Determine all the valid joint assignments to the variables that need 
			to be eliminated. */

	findValidJointAssignments(elimVarsArr, numElimVars, &validAssignmentsArr, numFuncs);

	funcPtrEnd = funcArr + numFuncs;
	assignmentPtrEnd = validAssignmentsArr.assignmentArr + validAssignmentsArr.numAssignments; 
	
	totalVal = 0.0;

	/* 2.	Calculate the probability for each valid assignment to the eliminated variables,
			and sum them up. */

	for (assignmentPtr = validAssignmentsArr.assignmentArr;
			assignmentPtr < assignmentPtrEnd; assignmentPtr++) {
			
		/* 2.1.	Update the values of the eliminated variables according to this assignment. */

		setAssignment(elimVarsArr, numElimVars, assignmentPtr->varVals, val, varPtr, varPtrEnd);
		
		currVal = 1.0;

		for (funcPtr = funcArr; funcPtr < funcPtrEnd; funcPtr++) {

			func = *funcPtr;
			calcBaseIndex(func);
			funcVal = func->probabList[func->baseInd];

			if (funcVal == 0.0) {
				goto valueWasZero;
			}

			currVal *= funcVal;
		}

		totalVal += currVal;

		valueWasZero: ;
	}

	/* 3.	Free the memory allocated dynamically. */

	freeAssignmentMem(&validAssignmentsArr);

	/* 4.	Update the result of the multiplication according to the logMultFactor. */

	return (log(totalVal) + logMultFactor);
}
#endif


/******************************************************************************
*
*       Function        findValidJointAssignments
*       Parameters		elimVarsArr (the array of variables that needs to be 
*						eliminated), numElimVars (the number of variables that
*						need to be eliminated), validAssignmentsArr (the array
*						which would store all the valid joint assignments to  
*						the variables that needs to be eliminated), numFuncs
*						(the number of functions which need to be combined in
*						the joint elimination of the variables in 'elimVarsArr').
*       Output          none
*       Description		Allocates memory for the array that would store all the
*						valid joint assignments and initializes it.
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void findValidJointAssignments(variable **elimVarsArr, int numElimVars, 
		  					   validAssignments *validAssignmentsArr, int numFuncs)
{
	assignment		*assignmentPtr;

	/* 1.	Find the valid joint assignments according to the unite type. */

	switch ((elimVarsArr[0]->originalNode)->uniteT) {

		case oneAll: findValidJointAssignmentsOne(elimVarsArr, numElimVars,
												  validAssignmentsArr, numFuncs);
					 #if PRINT_VARIOUS_STATISTICS_LEVEL
						numOneUnite++;
					 #endif

					 return;

		case twoAll: validAssignmentsArr->numAssignments = 2;
					 allocAssignmentsMem(validAssignmentsArr, numElimVars, numFuncs);

					 assignmentPtr = validAssignmentsArr->assignmentArr;
					 assignmentPtr->varVals[0] = 0;
					 assignmentPtr->varVals[1] = 1;
					 assignmentPtr++;
					 assignmentPtr->varVals[0] = 1;
					 assignmentPtr->varVals[1] = 0;


					 #if PRINT_VARIOUS_STATISTICS_LEVEL
						numTwoUnite++;
					 #endif

					 return;

		case fourAll: findValidJointAssignmentsFour(elimVarsArr, numElimVars,
												    validAssignmentsArr, numFuncs);
					  #if PRINT_VARIOUS_STATISTICS_LEVEL
            			  numFourUnite++;
					  #endif	

					  return;
	}
}
#endif



/******************************************************************************
*
*       Function        findValidJointAssignmentsOne
*       Parameters		elimVarsArr (the array of variables that needs to be 
*						eliminated), numElimVars (the number of variables that
*						need to be eliminated), validAssignmentsArr (the array
*						which would store all the valid joint assignments to  
*						the variables that needs to be eliminated), numFuncs
*						(the number of functions which need to be combined in
*						the joint elimination of the variables in 'elimVarsArr').
*       Output          none
*       Description		Allocates memory for the array that would store all the
*						valid joint assignments and initializes it.
*       Comments		Used in the case where the unite type is 'oneAll'.
*
******************************************************************************/
#if UNITE_VARS
void findValidJointAssignmentsOne(variable **elimVarsArr, int numElimVars, 
		  						  validAssignments *validAssignmentsArr, int numFuncs)
{
	variable			*var1, *var2;
	int					numAssignments, ind1, ind2, certainInd1, certainInd2;
	assignment			*assignmentPtr;


	/* 1.	Update pointers to the two genetic loci variables that need to be
			eliminated jointly. */

	var1 = elimVarsArr[0];
	var2 = elimVarsArr[1];

	/* 2.	Calculate the number of valid joint assignments. */

	numAssignments = var1->numValues + var2->numValues - 1;
	validAssignmentsArr->numAssignments = numAssignments;

	/* 3.	Allocate memory for the array that would store all the valid assignments. */

	allocAssignmentsMem(validAssignmentsArr, numElimVars, numFuncs);
	assignmentPtr = validAssignmentsArr->assignmentArr;

	/* 4.	Store the index of the certain value in the value array of the first
			eliminated variable. */

	ind1 = certainInd1 = var1->certainVal[0];

	/* 5.	Iterate over all possible values for the second eliminated variable
			and store all the assignments. */

	if (ind1 != -1) {
		for (ind2 = 0; ind2 < var2->numValues; ind2++, assignmentPtr++) {
			assignmentPtr->varVals[0] = ind1;
			assignmentPtr->varVals[1] = ind2;
		}
	}

	/* 6.	Store the index of the certain value in the value array of the second
			eliminated variable. */

	ind2 = certainInd2 = var2->certainVal[0];

	/* 7.	Iterate over all possible values for the first eliminated variable
			and store all the assignments, except for the one where the two
			variables have the certain value (this assignment has already been stored). */

	if (ind2 != -1) {
		for (ind1 = 0; ind1 < var1->numValues; ind1++) {
			if (ind1 == certainInd1) {
				continue;
			}

			assignmentPtr->varVals[0] = ind1;
			assignmentPtr->varVals[1] = ind2;
			assignmentPtr++;
		}
	}

	#if PRINT_VARIOUS_STATISTICS_LEVEL
		if (varsUniteFile) {
			fprintf(varsUniteFile, "%d --> %d\n", var1->numValues * var2->numValues,
						numAssignments);
		}
	#endif
}
#endif

/******************************************************************************
*
*       Function        findValidJointAssignmentsFour
*       Parameters		elimVarsArr (the array of variables that needs to be 
*						eliminated), numElimVars (the number of variables that
*						need to be eliminated), validAssignmentsArr (the array
*						which would store all the valid joint assignments to  
*						the variables that needs to be eliminated), numFuncs
*						(the number of functions which need to be combined in
*						the joint elimination of the variables in 'elimVarsArr').
*       Output          none
*       Description		Allocates memory for the array that would store all the
*						valid joint assignments and initializes it.
*       Comments		Used in the case where the unite type is 'fourAll'.
*
******************************************************************************/
#if UNITE_VARS
void findValidJointAssignmentsFour(variable **elimVarsArr, int numElimVars, 
		  						  validAssignments *validAssignmentsArr, int numFuncs)
{
	int					ind1, ind2, ind3, ind4, numVals[4], *numValsPtr, 
						*certainValArrs[4], **certainValArrsPtr, chosenInd1,
						chosenInd2, chosenInd3, chosenInd4, realInd1, realInd2,
						realInd3, realInd4;
	assignment			*assignmentPtr;
	boolean				chosen[4], *chosenPtr, *chosenPtrEnd;
	variable			**elimVarPtr;

	/* 1.	Allocate memory for the valid assignment array. */

	validAssignmentsArr->numAssignments = 24;

	allocAssignmentsMem(validAssignmentsArr, numElimVars, numFuncs);

	for (chosenPtr = chosen, chosenPtrEnd = chosenPtr + 4, numValsPtr = numVals,
			elimVarPtr = elimVarsArr, certainValArrsPtr = certainValArrs; 
			chosenPtr < chosenPtrEnd; chosenPtr++, numValsPtr++, elimVarPtr++,
			certainValArrsPtr++) {		

		*chosenPtr = FALSE;
		*numValsPtr = (*elimVarPtr)->originalNode->numValues;
		*certainValArrsPtr = (*elimVarPtr)->certainVal;
	}

	assignmentPtr = validAssignmentsArr->assignmentArr;

	/* 2.	Update the array with all the valid assignments. */

	for (ind1 = 0, realInd1 = -1; ind1 < numVals[0]; ind1++) {
		chosenInd1 = certainValArrs[0][ind1];

		if (chosenInd1 == -1) {
			continue;
		}

		realInd1++;

		chosen[chosenInd1] = TRUE;

		for (ind2 = 0, realInd2 = -1; ind2 < numVals[1]; ind2++) {

			chosenInd2 = certainValArrs[1][ind2];

			if (chosenInd2 == -1) 
				continue;

			realInd2++;

			if (chosen[chosenInd2]) 
				continue;

			chosen[chosenInd2] = TRUE;

			for (ind3 = 0, realInd3 = -1; ind3 < numVals[2]; ind3++) {
				
				chosenInd3 = certainValArrs[2][ind3];

				if (chosenInd3 == -1) 
					continue;

				realInd3++;

				if (chosen[chosenInd3])
					continue;

				chosen[chosenInd3] = TRUE;

				for (ind4 = 0, realInd4 = -1; ind4 < numVals[3]; ind4++) {
			
					chosenInd4 = certainValArrs[3][ind4];
					
					if (chosenInd4 == -1) 
						continue;

					realInd4++;

					if (chosen[chosenInd4])
						continue;

					chosen[chosenInd4] = TRUE;

					assignmentPtr->varVals[0] = realInd1;
					assignmentPtr->varVals[1] = realInd2;
					assignmentPtr->varVals[2] = realInd3;
					assignmentPtr->varVals[3] = realInd4;
					assignmentPtr++;
					chosen[chosenInd4] = FALSE;
				}

				chosen[chosenInd3] = FALSE;
			}

			chosen[chosenInd2] = FALSE;
		}
		
		chosen[chosenInd1] = FALSE;
	}

	/* 3.	Update the number of valid assignments. */

	validAssignmentsArr->numAssignments = 
						assignmentPtr - validAssignmentsArr->assignmentArr;

	#if PRINT_VARIOUS_STATISTICS_LEVEL
		if (varsUniteFile) {
			fprintf(varsUniteFile, "%d --> %d\n", 
						numVals[0] * numVals[1] * numVals[2] * numVals[3],
						validAssignmentsArr->numAssignments);
		}
	#endif
}
#endif


/******************************************************************************
*
*       Function        allocAssignmentsMem
*       Parameters		validAssignmentsArr (pointer to the assignment array 
*						for which memory needs to be allocated), numElimVars 
*						(the number of variables that need to be eliminated),
*						numFuncs (the number of functions that need to be 
*						combined during the elimination).
*       Output          none
*       Description		Allocates memory for the assignmemt array.
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void allocAssignmentsMem(validAssignments *validAssignmentsArr, int numElimVars,
						 int numFuncs)
{
	int				numAssignments;
	assignment		*assignmentPtr, *assignmentPtrEnd;

	/* 1.	Update the number of valid assignments. */

	numAssignments = validAssignmentsArr->numAssignments;

	validAssignmentsArr->numAssignmentSpaces = numAssignments;

	/* 2.	Allocate memory for the array of valid assignments. */

	myMalloc(validAssignmentsArr->assignmentArr, sizeof(assignment) * numAssignments);

	/* 3.	Iterate over the entries of the array, and allocate memory to	
			store the values of the variables in the current assignment, and the
			necessary increment in the index of each function to reach the
			entry which corresponds to this assignment. */
			
	for (assignmentPtr = validAssignmentsArr->assignmentArr, 
			assignmentPtrEnd = assignmentPtr + numAssignments; 
			assignmentPtr < assignmentPtrEnd; assignmentPtr++) {
		
		myMalloc(assignmentPtr->varVals, sizeof(int) * numElimVars);

		if (numFuncs) {
			myMalloc(assignmentPtr->inc, sizeof(int) * numFuncs);
		}
		else {
			assignmentPtr->inc = NULL;
		}
	}
}
#endif


/******************************************************************************
*
*       Function        freeAssignmentMem
*       Parameters		validAssignmentsArr (the assignment array whose memory
*						needs to be freed).
*       Output          none
*       Description		Frees the memory allocated for the assignment array.
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void freeAssignmentMem(validAssignments *validAssignmentsArr)
{
	int					numAssignments;
	assignment			*assignmentPtr, *assignmentPtrEnd;

	/* 1.	Store the number of assignment spaces. */

	numAssignments = validAssignmentsArr->numAssignmentSpaces;

	/* 2.	Iterate over the assignments and free the memory allocated to each one. */

	for (assignmentPtr = validAssignmentsArr->assignmentArr, 
			assignmentPtrEnd = assignmentPtr + numAssignments; 
			assignmentPtr < assignmentPtrEnd; assignmentPtr++) {
		
		myFree(assignmentPtr->varVals);
		myFree(assignmentPtr->inc);
	}

	/* 3.	Free the memory allocated to the entire array. */

	myFree(validAssignmentsArr->assignmentArr);
}
#endif


