/*****************************************************************************/
/*  Product.h																 */
/*  Author: Ma'ayan Fishelson												 */
/*****************************************************************************/
/*																			 */
/*  Comment: This file includes functions used for creating the variables	 */
/*			 functions of the Bayesian networks that represent the pedigrees.*/
/*																			 */
/*****************************************************************************/

#ifndef _SL_PRODUCT_H_
#define _SL_PRODUCT_H_

#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include "sl_PedData.h"
#include "sl_EliminationOrderStructures.h"
#include "sl_BayesNet.h"

/*------------------------------*/
/*		DEFINITIONS				*/
/*------------------------------*/

/* updateConstantVars(node *Node) */

#if (PRINT_ELIMINATION_STATISTICS || (PRINT_VARIOUS_STATISTICS_LEVEL > 1))
	#define updateConstantVars(Node) \
		if ((Node)->numValues == 1) \
			countConstantVars++; 
#else 
	#define updateConstantVars(Node) 
#endif
		
/* updateEliminatedVars(node *Node) */

#if (PRINT_ELIMINATION_STATISTICS || (PRINT_VARIOUS_STATISTICS_LEVEL > 1))
	#define updateEliminatedVars(Node) \
			freeNode(&Node); \
			countEliminatedVars++;
#else
	#define updateEliminatedVars(Node) \
			freeNode(&Node); 
#endif

/* isUniteNode(node *Node) - is TRUE if the node exists and needs to be treated jointly 
   with other node(s). */

#if UNITE_VARS
	#define isUniteNode(Node) \
		(Node) && ((Node)->uniteT)
#endif

/* isNotUsedVar(node *Node) - is TRUE if there's no variable associated with this node,
   or if the variable associated with this node has been shrunk into a constant and eliminated. */

#define isNotUsedVar(Node) \
	!((Node)->varInProd) || !((Node)->varInProd->used)

/* isUsedVar(node *Node) - is TRUE if there's a variable associated with this node,
   and this variable hasn't been shrunk into a constant and eliminated. */

#define isUsedVar(Node) \
	((Node)->varInProd) && ((Node)->varInProd->used)

/* getOffset(long *offset, int ind) - returns 0 if ind equals -1; otherwise, returns the 
   value in entry ind. */

#define getOffset(offset, ind) \
	((ind == -1) ? 0 : offset[ind])

/*------------------------------*/
/*		DATA-STRUCTURES			*/
/*------------------------------*/

typedef struct function function;

/* This structure represents a variable associated with a node of the Bayesian network. */

struct variable {
	boolean		used;				/* Is this space being used. */
	int			functionSpaces;		/* Number of function slots available. */
	int			currNumFunctions;	/* Number of functions currently pointed to by the 
									   struct (= number of functions that the variable
									   appears in currently).*/
	function	**functions;		/* An array of pointers to the functions that depend 
									   on this variable. */
	node		*originalNode;		/* Pointer to the original node in the Bayesian network
									   that this variable represents. */
	int			numValues;			/* The number of possible values of this variable. */
	int			currValue;			/* An auxiliary field used by different functions. */
	boolean		alreadySeen;		/* An auxiliary flag used by different functions. */
	variable	*varInOldProd;		/* Pointer to its "correspoindg" variable in the 
									   old product (used when the product is duplicated). */
	double		elimComplexity;		/* The elimination complexity of this variable (if 
									   the value isn't updated currently, has a negative 
									   value.) */
#if UNITE_VARS
	int			*certainVal;		/* In case 'uniteT' is equal to 'oneAll', this field holds the
									   value of the variable (the index of the allele) which is known 
									   for certain in the genotype of this person at this locus. 
									   In case 'uniteT' is equal to 'fourAll', this field holds the 
									   possible values for this variable. */
#endif
};


/* This structure is used to calculate necessary increments and decrements when traversing
   the probability table of some function. */

typedef struct incDec {
	int inc;
	int dec;
} incDec;


/* This structure represents a probability function in the Bayesian network. */

struct function {

	//anna
		char *ownerName;
		variable *owner;               /* Probability table of which variable ti is */
   //end
	boolean			used;				/* Is this space being used. */
	int				numVariables;		/* The number of variables this function
										   depends on. */
	variable		**variables;		/* An array of pointers to the variables
										   of the function. */
	unsigned long	sizeOfTable;		/* The size of the probability table. */
	Probability		*probabList;		/* The list of probabilities of the function. */
	boolean			myProbabList;		/* A flag that indicates whether the probability
										   list belongs only to this function (or it
										   is shared by another function). */
	incDec			jumpFactor;			/* An auxiliary field that is used when performing
										   different operations on the function. */
	long			baseInd;			/* An auxiliary field that is used when performing
										   different operations on the function. */
	double			logMultFactor;		/* A factor which is used to scale the probabilities
										   in the list. */	
	int				rescaleTimer;		/* A timer that indicates when the function needs to
										   be rescaled (when it is zero). */
	boolean			alreadySeen;		/* An auxiliary flag used by different functions. */
	int				*ordering;			/* An auxiliary field used when performing different operations
										   on this function. */
	long			*accOffset;			/* An auxiliary field used when performing different operations
										   on this function. */
};


/* This structure represents the Bayesian network (the variables and probability functions
   of the network). */

typedef struct product {
		int		 prodNum;	
		long	 numVariables;			/* The number of variable slots available. */
		long	 currNumVars;			/* The number of variables currently in the product.
										   (In the beginning, it includes all the variables in
											the Bayesian network.) */
		variable *variables;			/* An array of variables. */
		variable *nextVarPlace;			/* Pointer to the next empty slot in the variable array. */
		variable *nextElimVar;			/* Pointer to the next candidate for elimination. */

		long	 numFunctions;			/* The number of function slots available. */
		long	 currNumFuncs;			/* The number of fucntions currently in the product.
										   (In the beginning, it includes all the functions in
										   the Bayesian network.) */
		function *functions;			/* An array of functions. */
		function *nextFuncPlace;		/* Pointer to the next empty slot in the function array. */

	/* The following are variables used when performing different statistics on the structure
	   of the Bayesian networks constructed. */

	#if (PRINT_VARIOUS_STATISTICS_LEVEL > 1)
		long	 numVarsElim;			/* The number of variables that have been eliminated from
										   the product during its creation (variables with no children
										   on which there's no evidence). */
		long	 numVarsConst;			/* The number of constant variables in the Bayesian network. */
		long	 numVarsTrim;			/* Variables that have been determined to be constants during
										   the product-value-elimination. */
	#endif

	#if (PRINT_VARIOUS_STATISTICS_LEVEL > 2)
		int		minNumValsVar;			/* The minimum number of possible values of a variable
										   of the Bayesian network. */
		int		maxNumValsVar;			/* The maximum number of possible values of a variable
										   of the Bayesian network. */
		double	avgNumValsVar;			/* The average number of possible values of a variable
										   of the Bayesian network. */
	#endif

	#if (PRINT_VARIOUS_STATISTICS_LEVEL > 3)
		long	minNumSelectorsInterval;	/* The minimum number of selector nodes left after
											   elimination between each two consecutive loci. */
		long	maxNumSelectorsInterval;	/* The maximum number of selector nodes left after
											   elimination between each two consecutive loci. */
		double	avgNumSelectorsInterval;	/* The average number of selector nodes left after
											   elimination between each two consecutive loci. */
	#endif

} product;

/* This structure holds information on a valid assignment to a certain group of variables. */

#if UNITE_VARS
	typedef struct assignment {
		int		*varVals;		/* Holds the value of each variable. */
		int		*inc;			/* The necessary increment in the index of each probability
								table from the index of the previous valid assignment. */
	} assignment;
#endif

/* This structure holds information on all valid assignment to a certain group of variables. */

#if UNITE_VARS
	typedef struct validAssignments {
		assignment	*assignmentArr;			/* An array which holds all the valid assignments. */
		int			numAssignmentSpaces;	/* The size of the assignment array. */
		int			numAssignments;			/* The number of valid assignments in the array. */
	} validAssignments;
#endif


/*------------------------------*/
/*		GLOBAL VARIABLES		*/
/*------------------------------*/

product						*Product;
extern lociData				loci;
Probability					tempRes;
extern recombinationData	recombData;		
results						totalRes;		/* Holds the results of all likelihood evaluations. */

FILE *outFile;

#if ELIMINATE_SELECTORS_TOGETHER
	extern boolean	eliminateSelectorsTogether;
#endif

extern int			numMarkers;

#if PRINT_MATRIX_MULT_STATISTICS
	extern FILE		*matrixFile;
#endif


/*------------------------------*/
/*  FUNCTIONS & DEFINITIONS	    */
/*------------------------------*/

typedef void (createPersonVars)(person *currPerson);

product** allocateCreateProducts(createPersonVars createPersonVarsPed);
product* allocateProduct(long numVariables, long numFunctions);
Probability createProduct(product *prod, pedigree *pedPtr, 
						  createPersonVars createPersonVarsPed);

void traverseUpPed(person *currPerson, createPersonVars createPersonVarsPed);
void traverseDownPed(person *currPerson, createPersonVars createPersonVarsPed);
void createPersonVarsAutosomal(person *currPerson);
void createPersonVarsSexLinked(person *currPerson);
void createPersonVarsAutosomalHaplo(person *currPerson);
void createPersonVarsSexLinkedHaplo(person *currPerson);
void handleProbabList(int tableSize, product *Product, node *Node, 
					  Probability *probabList);

product* copyProduct(product *oldProd, elimOrderPed *elimOrd, 
					 boolean *isTransformedArr, variable **condVarsOld, 
					 variable **condVarsNew, int numCondVars);
void freeProducts(product** prodArr, int numProds);
void freeProduct(product *prod);


void zeroProductFields(product **prodArr, boolean sexLink);

/* Functions for creating the different kinds of probability functions. */ 
int createTablePenetrance(locusPerson *currLoc, int locusNum, node *phenotypeNode, 
						  node *pLocusNode, node *mLocusNode, Probability **newTable);
int createTablePenetranceFounder(locusPerson *currLoc, int locusNum, node *phenotypeNode, 
								 node *pLocusNode, node *mLocusNode, node *genotypeNode, 
								 Probability **newTable);
int createTableMalePenetrance(locusPerson *currLoc, int locusNum, node *phenotypeNode, 
							  node *mLocusNode, Probability **newTable);
int createTablePenetranceTwo(person *currPerson, int firstNullLociIndDS, 
							 int secondNullLociIndDS, node *phenotypeNode, Probability **newTable);
int createTablePenetranceFounderTwo(person *currPerson, int firstNullLociIndDS, 
									int secondNullLociIndDS, node *phenotypeNode, 
									node *genotypeNode, int firstNullLociPosMap,
									Probability **newTable);
int createTableMalePenetranceTwo(person *currPerson, int firstNullLociIndDS, 
								  int secondNullLociIndDS, node *phenotypeNode,
								  Probability **newTable);
int createTablePhenotype(node *phenotypeNode, node *pLocusNode, node *mLocusNode,
						 Probability **newTable);
int createTableTrans(person *currPerson, int locusNum, boolean isPaternal, 
					  node *locusNode, node *selectorNode, Probability **newTable);

#if PERFORM_ALLELE_RECODING
	int createTableTransAR(person *currPerson, int locusNum, boolean isPaternal, 
						   node *locusNode, node *selectorNode, Probability **newTable);
#endif

int createTableTransFounder(person *currPerson, int locusNum, boolean isPaternal, 
							node *locusNode, node *selectorNode, 
							Probability **newTable);
int createTableTransMale(person *currPerson, int locusNum, node *pLocusNode,
						 Probability **newTable);

#if PERFORM_ALLELE_RECODING
	int createTableTransMaleAR(person *currPerson, int locusNum, node *pLocusNode,
							Probability **newTable);
#endif

#if PERFORM_ALLELE_RECODING
boolean setInclusion(alleleSet parentSet, alleleSet childSet);
#endif


int createTableGeneFrequency(int locusNum, int *alleleArr, int numVals, Probability **newTable);

int createTableGenotypeFrequency(int locusNum, int *alleleArr, int numVals,
							     Probability **newTable, boolean isUnknownGenotype);

#if PERFORM_ALLELE_RECODING
int createTableGenotypeFrequencyAR(int locusNum, alleleSet *alleleSetArr, 
								   int numSets, Probability **newTable, 
								   boolean isUnknownGenotype);
int createTableGeneFrequencyAR(int locusNum, alleleSet *alleleSetArr,  
							   int numSets, Probability **newTable);
#endif


int createTableRecomb(person *currPerson, int locusNum, boolean isPaternal, node *selectorNode,
					   Probability **newTable);
int createTableRecombInterferenceMale(person *currPerson, node *pSelectorNode, 
									  Probability **newTable);
int createTableRecombInterferenceFemale(person *currPerson, node *mSelectorNode, 
										Probability **newTable);
int createTableHalf(int numVals, Probability **newTable);


/* Variable.h */
/*void removeVariableFromProduct(product *prod, variable *var) */

#if UNITE_VARS
	#define removeVariableFromProduct(prod, var) \
		myFree(var->functions); \
		myFree(var->certainVal); \
		var->used = FALSE; \
		prod->currNumVars--;
#else
	#define removeVariableFromProduct(prod, var) \
		myFree(var->functions); \
		var->used = FALSE; \
		prod->currNumVars--;
#endif


variable* getNodeVariable(node *Node, product *prod);
variable* insertVariableToProduct(product *prod, node *Node, int numValues);

#if UNITE_VARS
	void updateCertainVals(variable *varPtr, uniteType uniteT, int *certainVal);
#endif

void buildVariableFunctionTables(product *prod);
void addFunctionToVariable(variable *Variable, function *Function);
void removeFunctionFromVariable(variable *Variable, function *Function);


/* Function.h */

/* Probability getFuncValue(function *func, int val); */

#define getFuncValue(func, val)	\
	(val) ? func->probabList[func->baseInd + ((func->jumpFactor).inc * val)] :	\
			func->probabList[func->baseInd];


#define getFuncVal(func)	\
	func->probabList[func->baseInd];


/*  Initialize the values of the function's variables to their initial value, zero, 
	(which correspond to the first entrance in the function's probability table). */

#define initFunctionVars(func, varptr, varptrend) \
	for (varptr = func->variables, varptrend = varptr + func->numVariables; \
			varptr < varptrend; varptr++) { \
		(*varptr)->currValue = 0; \
	}


/*  Advance the values of the function's variable to values which correspond to 
    the next entrance in the function's probability table.	*/

#define advanceFunctionVars(func, varptr, varptrend, var, start_loop) \
	for (varptr = func->variables + func->numVariables - 1, varptrend = func->variables; \
			varptr >= varptrend; varptr--) { \
		var = *varptr; \
		if (var->currValue < (var->numValues - 1)) { \
			var->currValue++; \
			goto start_loop; \
		} \
		else \
			var->currValue = 0; \
	} \
	break; 

/* Zeroes the seen flags of the functions in 'funcArr'. */

#define zeroSeenFuncs(funcArr, numFuncs, funcPtr, funcPtrEnd) \
	for (funcPtr = funcArr, funcPtrEnd = funcPtr + numFuncs; \
			funcPtr < funcPtrEnd; funcPtr++) { \
		(*funcPtr)->alreadySeen = FALSE; \
	}

/* Sets an assignment of values in 'varVals' to the variables in array 'elimVars'. */
#if UNITE_VARS
	#define	setAssignment(elimVars, numElimVars, varVals, val, varPtr, varPtrEnd) \
		for (varPtr = elimVars, varPtrEnd = varPtr + numElimVars, \
				val = varVals; varPtr < varPtrEnd; varPtr++, val++) { \
			(*varPtr)->currValue = (*val); \
		}
#endif

function* insertFunctionToProduct(product *prod, unsigned long tableSize, 
								  Probability *probabList, double logScaleFactor,
								  boolean isMyProbTable, int numVariables, 
								  variable **variables, boolean DuplicateVars,
								  int rescaleTimer);
variable** getFunctionVariables(product *prod, node *Node, int *numVariables);

long double getSizeOfCombinedFunc(variable *elimVar);


void removeFunctionFromProduct(product *prod, function *func);
void removeVariableFromFunction(function *func, variable *var);
function* reduceFunctionTable(product *prod, function *func, boolean *varValueUsageFlagsGen, 
							  int maxValuesVar, int *numValsVarsNew, Probability *tempRes);
boolean reduceFunctionTableOnVar(function *oldFunction, variable *var, int numVals, 
								 int valVar, Probability *currRes);
void zeroSeenVars(int numFuncs, function **funcArr);
void getVarsInCombinedFunc(variable **elimVars, int numElimVars, variable **varArr, 
						   function **funcArr, int numFuncs, int *numVars, 
						   unsigned long *sizeOfTable);
void getVarsInCombinedFuncJoint(variable **elimVarsArr, int numElimVars, 
								function **funcArr, int *numFuncs, variable **varArr, 
								int *numVars, unsigned long *sizeOfTable, 
								double *logMultFactor, int *rescaleTimer);
void initJumpFactor(function *func, variable *elimVar);
void initJumpFactors(function *func, variable *elimVar, incDec *incDecPtr);


#if UNITE_VARS
	void initJumpFactorJoint(function **funcArr, int numFuncs, variable **elimVars, 
							int numElimVars, validAssignments *validAssignmentsArr);
	void initJumpFactorsJoint(function **funcArr, int numFuncs, variable **elimVars, 
							int numElimVars, incDec *incDecArr, variable **combinedVars,
							int numCombinedVars, validAssignments *validAssignmentsArr);
#endif

void calcBaseIndex(function *func);
void calcAccOffsetElim(function *funcX, function *funcY, variable *elimVar);
void calcAccOffset(function *funcX, function *funcY);
boolean scaleFunctionProbList(function *func);
void printFunction(function *func);


/* VariableElimination.h */

variable* getEliminateVar(elimOrderPed *elimOrd, boolean *isBreakVar);
boolean eliminateVar(product *prod, variable *elimVar, Probability *tempRes, 
					 boolean *isZeroLike);

/* ProductBreaking.h */

Probability inferProdWithCond(product *prod, variable* breakVar,
							  elimOrderPed *elimOrderArr, float problemSize, 
							  float solved);

#if UNITE_VARS
	Probability inferProdWithCondJoint(product *prod, variable *condVar,
									elimOrderPed *elimOrd, float problemSize, 
									float solved);
#endif

Probability setBreakVal(product *brokenProd, variable *brokenVar, int breakVarVal);


#if DEBUG_STAT
	void printProduct(product *prod, double tmpResLn);
#endif


#endif		/*	_PRODUCT_H_	*/
