/*********************************************************************************/
/*  Prog.c												     				     */
/*  Author: Ma'ayan Fishelson					                                 */
/*********************************************************************************/

#include "sl_Globals.h"
#include "sl_ProductValueElimination.h"
#include "sl_Inference.h"
#include "sl_MapFunction.h"
#include "sl_IterateAff.h" 
#include "sl_Haplotype.h"
#include "sl_Optimize.h"
#include "sl_XMLFormat.h"


/******************************************************************************
*
*       Function        calcLikelihoodOneLocus
*       Parameters      none
*       Output          none
*       Description     Calculates the likelihood of data which consists of only 
*						one locus.
*       Comments        none
*
******************************************************************************/
double calcLikelihoodOneLocus()
{
    boolean			areValsElim, isChangedNet,  isZeroLike;
    elimOrder       elimOrderArrs = NULL;

	/* 1.	Allocate memory for a results node to store the results of the 
			likelihood evaluation, and update the currRes pointer to point to it. */

	initResNode(0.0);

	/* 2.	Calculate the likelihood. */

	isZeroLike =performOneIter(&elimOrderArrs, &areValsElim, &isChangedNet, TRUE);

	/* 3.	Print and save the results. */

	// printResults_OneIterWOLod(peds.numPeds);

	/* 4.	Free the memory allocated dynamcially. */

	if (elimOrderArrs != NULL) {
		freeEliminationOrderStructs(&elimOrderArrs, peds.numPeds);
	}
   if (isZeroLike)
		return MIN_DBL;
	return totalRes.currRes->lnLike;
}



/******************************************************************************
*
*       Function        calcLikelihoodSuperLinkmap
*       Parameters		none
*       Output          none
*       Description		Linkmap compatible program
*       Comments		none
*
******************************************************************************/
void calcLikelihoodSuperLinkmap()
{
	double			intervalSize, stepSize, maleDist, finalVal, intervalSizeFemale;
	int				currIter, numThetaVaried, numOtherThetaVaried=0, numThetas,
					numPeds, ind, numLoci, locusVaried, numEvalPoints, numMarkers;
	boolean			areValsElim, isChangedNet, updateFirstNullLociPos, 
					updateSecondNullLociPos, createBN;
	elimOrder		elimOrderArrs;
	interval		typeInterval;
	Probability		thetaTemp, thetaTempOther, initialMaleVal=0, initialFemaleVal=0;
	thetaCompTheta	*maleThetas, *femaleThetas, *thetaInitialMale, *thetaInitialFemale;	
	xLabel			markerNames[2];
	scanIntervalData	*scanDataPtr;


	/* 1.	Initializations. */

	elimOrderArrs = NULL;
	numLoci = loci.numLoci;
	numThetas = numLoci - 1;
	numPeds = peds.numPeds;
	createBN = TRUE;

	maleThetas = recombData.maleThetas;
	femaleThetas = recombData.femaleThetas;
	locusVaried = ((progData.progSpecific).linkmapData).locusVaried;
	numEvalPoints = ((progData.progSpecific).linkmapData).numEvalPoints;
	finalVal = ((progData.progSpecific).linkmapData).finalVal;

	scanDataPtr = progData.scanData;

	/* 2.	Calculate the size of the interval on which the varied locus needs to be
			moved and determine the type of map interval (left, middle or right) in which
			the locus will be moved. */
	
	if (locusVaried == 0) {
		typeInterval = LEFT;
		numThetaVaried = 0;
		intervalSize = -0.5;
		
		/* 2.1.	Save the initial value(s) of the recombination to be varied. */
		
		initialMaleVal = maleThetas[numThetaVaried][0];
		initialFemaleVal = femaleThetas[numThetaVaried][0];
		
		/* 2.2.	Calculate the likelihood with recombination to be varied at value 0.5 */
		
		updateThetaCompTheta(maleThetas[numThetaVaried], 0.5);
		updateThetaCompTheta(femaleThetas[numThetaVaried], 0.5);

		scanDataPtr->mapDistArr[0] = 0.0;
		scanDataPtr->distFromMarker1[0] = 0.0;
		updateMarkerName(markerNames, &(loci.locusArr[1]), 1);
		numMarkers = 1;
		scanDataPtr->startPos = 1;

	} else if (locusVaried == numLoci - 1) {
		typeInterval = RIGHT;
		numThetaVaried = numThetas - 1;
		intervalSize = 0.5;
		
		/* 2.1.	Save the initial value(s) of the recombination to be varied. */
		
		initialMaleVal = maleThetas[numThetaVaried][0];
		initialFemaleVal = femaleThetas[numThetaVaried][0];
		
		/* 2.2.	Calculate the likelihood with recombination to be varied at value 0.5 */
		
		updateThetaCompTheta(maleThetas[numThetaVaried], 0.5);
		updateThetaCompTheta(femaleThetas[numThetaVaried], 0.5);

		scanDataPtr->mapDistArr[0] = 0.0;
		scanDataPtr->distFromMarker1[0] = 0.0;
		updateMarkerName(markerNames, &(loci.locusArr[locusVaried]), locusVaried);
		numMarkers = 1;
		scanDataPtr->startPos = locusVaried;
	}
	else {
		typeInterval = MIDDLE;
		numThetaVaried = locusVaried - 1;
		numOtherThetaVaried = locusVaried;
		intervalSize = 
			getHaldaneThetaAC(maleThetas[numThetaVaried][0], maleThetas[numOtherThetaVaried][0]);
		
		/* 2.1.	Allocate memory and save the original recombination values 
				(the ones specified in the input file).*/

		myMalloc(thetaInitialMale, sizeof(thetaCompTheta) * numThetas);
		memcpy(thetaInitialMale, maleThetas, sizeof(thetaCompTheta) * numThetas);

		myMalloc(thetaInitialFemale, sizeof(thetaCompTheta) * numThetas);
		memcpy(thetaInitialFemale, femaleThetas, sizeof(thetaCompTheta) * numThetas);

		/* 2.2.	Place the varied locus first on the map (in the leftmost position) and update the 
				recombination fraction between it and the first marker to 0.5, to test the hypothesis
				of no-linkage. */

		loci.mapToData[0] = locusVaried;
		updateThetaCompTheta(maleThetas[0], 0.5);
		updateThetaCompTheta(femaleThetas[0], 0.5);

		/* 2.3.	Update the rest of the mapping from map position to data-structure position 
				and the rest of the theta values. */

		for (ind = 1; ind < locusVaried; ind++) {
			loci.mapToData[ind] = ind - 1;

			maleThetas[ind][0] = thetaInitialMale[ind-1][0];
			maleThetas[ind][1] = thetaInitialMale[ind-1][1];
			femaleThetas[ind][0] = thetaInitialFemale[ind-1][0];
			femaleThetas[ind][1] = thetaInitialFemale[ind-1][1];
		}

		loci.mapToData[ind] = ind - 1;
		updateThetaCompTheta(maleThetas[locusVaried], intervalSize);	

		if (recombData.sexDiff == 1) {
			intervalSizeFemale = 
				getHaldaneThetaAC(femaleThetas[numThetaVaried][0], femaleThetas[numOtherThetaVaried][0]);
		}
		else {
			intervalSizeFemale = intervalSize;
		}

		updateThetaCompTheta(femaleThetas[locusVaried], intervalSizeFemale);	

		/* 2.4.	If this is a two-locus disease, check if the locus to be varied is one of 
				the trait loci. */

		if (loci.numTraitLoci == 2) {
			if (locusVaried == loci.traitLociPos[0]) {
				updateFirstNullLociPos = TRUE;
				updateSecondNullLociPos = FALSE;
				loci.traitLociPos[0] = 0;
			}
			else if (locusVaried == loci.traitLociPos[1]) {
				updateSecondNullLociPos = TRUE;
				updateFirstNullLociPos = FALSE;
				loci.traitLociPos[1] = 0;
			}
			else {
				updateFirstNullLociPos = FALSE;
				updateSecondNullLociPos = FALSE;
			}
		}

		scanDataPtr->mapDistArr[0] = 0.0;
		scanDataPtr->mapDistArr[1] = getRecTocM(intervalSize, progData.haldane);

		scanDataPtr->distFromMarker1[0] = 0.0;
		scanDataPtr->distFromMarker1[1] = scanDataPtr->mapDistArr[1];

		updateMarkerName(markerNames, &(loci.locusArr[locusVaried-1]), locusVaried);
		updateMarkerName(markerNames+1, &(loci.locusArr[locusVaried+1]), locusVaried+1);
		numMarkers = 2;

		scanDataPtr->startPos = locusVaried;
		scanDataPtr->endPos = locusVaried + 1;
	}
	
	/* 3.	Calculate the step size of the move of the varied locus. */
	
	stepSize = intervalSize / numEvalPoints;

	/* 4.	Calculate the likelihood for the case of no-linkage between the trait locus
			and the marker map. */

	myPrintf("\nResults for no linkage:\n");

	calcLikeNoLinkage(&elimOrderArrs, &areValsElim, &isChangedNet);

	isChangedNet = TRUE;

	/* 5.	Restore the value(s) of the 'recombinations to be varied' to their original value. */
	
	if ((locusVaried == 0) || (locusVaried == numLoci - 1)) {
		updateThetaCompTheta(maleThetas[numThetaVaried], initialMaleVal);
		updateThetaCompTheta(femaleThetas[numThetaVaried], initialFemaleVal);
	}
	else {
		memcpy(maleThetas, thetaInitialMale, sizeof(thetaCompTheta) * numThetas);
		memcpy(femaleThetas, thetaInitialFemale, sizeof(thetaCompTheta) * numThetas);

		for (ind = 0; ind < loci.numLoci; ind++) {
			loci.mapToData[ind] = ind;
		}

		/* 5.1.	If this is a 2-locus disease and one of the trait-loci was iterated,
				restore its position. */

		if (loci.numTraitLoci == 2) {
			if (updateFirstNullLociPos) {
				loci.traitLociPos[0] = locusVaried;
			}
			else if (updateSecondNullLociPos) {
				loci.traitLociPos[1] = locusVaried;
			}
		}		
	}
	
	/* 6.	Perform likelihood evaluations until we reach the final value or the maximum 
			number of evaluations requested. */
	
	for (currIter = 0; currIter <= numEvalPoints; currIter++) {
		
		/* 6.1.	Allocate memory for a result node to store the results of the
				current likelihood evaluation, update the 'currRes'
				field to point to it, and update the distance of the trait locus 
				from the marker delimiting the scan interval from the left. */

		initResNode(getRecTocM(maleThetas[numThetaVaried][0], progData.haldane));

		if (typeInterval == LEFT) {
			totalRes.currRes->distFromMarker1Arr[0] *= -1;
		}

		if (isChangedNet) {
			createBN = TRUE;
		}
		else {
			createBN = FALSE;
		}

		/* 6.2.	Calculate the likelihood for the current theta values. */

		performOneIter(&elimOrderArrs, &areValsElim, &isChangedNet, createBN);

		/* 6.3.	Print the results. */
		
		printResults_OneIter(numThetas, numPeds, TRUE);
		
		/* 6.4.	Update the maximum result if necessary. */

		updateMax(getLodScore);

		/* 6.5.	Update the relevant theta values.*/
		
		thetaTemp = maleThetas[numThetaVaried][0];
		incThetas(&(maleThetas[numThetaVaried]), &(femaleThetas[numThetaVaried]), stepSize);
		
		/* 6.6.	Check if we have reached the final value for theta. */
		
		switch (typeInterval) {
			
			case LEFT:  if (maleThetas[numThetaVaried][0] < finalVal)
							goto freeMemory;
				
						break;
				
			case MIDDLE: thetaTempOther = maleThetas[numOtherThetaVaried][0];
						 updateThetaCompTheta(maleThetas[numOtherThetaVaried],
											  getHaldaneThetaBC(intervalSize, maleThetas[numThetaVaried][0]));
				
						if (maleThetas[numOtherThetaVaried][0] < 0.0)
							goto freeMemory;

						if (recombData.sexDiff == 1) {
							maleDist = getHaldaneRecToM(maleThetas[numOtherThetaVaried][0]);
							updateThetaCompTheta(femaleThetas[numOtherThetaVaried],
												 getHaldaneMToRec((maleDist * recombData.geneticDist)));
						}
						else {
							femaleThetas[numOtherThetaVaried][0] = maleThetas[numOtherThetaVaried][0];
							femaleThetas[numOtherThetaVaried][1] = maleThetas[numOtherThetaVaried][1];
						}
				
			case RIGHT:  if (maleThetas[numThetaVaried][0] > finalVal)
							goto freeMemory;
		}
		
		switch (typeInterval) {
			
			case MIDDLE: if ((((thetaTemp != 0.0) && (maleThetas[numThetaVaried][0] != 0.0)) &&
								((thetaTempOther != 0.0) && (maleThetas[numOtherThetaVaried][0] != 0.0))) &&
							((thetaTemp != 0.5) && (maleThetas[numThetaVaried][0] != 0.5)) &&
								((thetaTempOther != 0.5) && (maleThetas[numOtherThetaVaried][0] != 0.5))) {
							isChangedNet = FALSE;
						 }
						 break;
				
			default:	 if (((thetaTemp != 0.0) && (maleThetas[numThetaVaried][0] != 0.0)) &&
							((thetaTemp != 0.5) && (maleThetas[numThetaVaried][0] != 0.5))) {
							isChangedNet = FALSE;
						 }
					     break;
			}
	} 

 	/* 7.	Free the memory allocated dynamically. */
	
	freeMemory: 
    if (elimOrderArrs != NULL)
		freeEliminationOrderStructs(&elimOrderArrs, numPeds);

	if (typeInterval == MIDDLE) {
		myFree(thetaInitialMale);
		myFree(thetaInitialFemale);
	}

	/* 7.	Print the summary of the results. */
 
	printResultsSummary_OneLocus(numMarkers, markerNames);
}


/******************************************************************************
*
*       Function        incThetas
*       Parameters		maleTheta (pointer to the male theta that needs to be
*						incremented), femaleTheta (pointer to the female theta
*						that needs to be update according to the increment in 
*						the male theta), inc (the increment in the male theta).
*       Output          none
*       Description		Updates the male and female thetas according to the
*						specified increment.
*       Comments		The female theta is updated according to whether there 
						is a sex-difference in the recombination fractions or not.
*
******************************************************************************/
void incThetas(thetaCompTheta *maleTheta, thetaCompTheta *femaleTheta, double inc)
{
	double		maleDist;
	
	/* 1.	Increment the male theta value according to the increment specified. */

	updateThetaCompTheta(*maleTheta, (*maleTheta)[0] + inc);
    
	/* 2.	Update the female recombination to be varied, according to whether
			there is a sex-difference in the recombination fractions or not. */

	if (recombData.sexDiff == 1) {
		maleDist = getHaldaneRecToM((*maleTheta)[0]);
		updateThetaCompTheta(*femaleTheta, 
								getHaldaneMToRec((maleDist * recombData.geneticDist)));
	}
	else if (recombData.sexDiff == 0) {
		(*femaleTheta)[0] = (*maleTheta)[0];
		(*femaleTheta)[1] = (*maleTheta)[1];
	}
}


/******************************************************************************
*
*       Function        calcLikelihoodSuperMlink
*       Parameters      none
*       Output          none
*       Description     Mlink compatible program
*       Comments        none
*
******************************************************************************/
void calcLikelihoodSuperMlink()
{
    int                     numThetas, numPeds, recVaried;
    double                  finalVal, inc, dist;
    boolean                 areValsElim, isChangedNet, createBN;
    elimOrder               elimOrderArrs;
	thetaCompTheta			*maleThetaVaried, *femaleThetaVaried;
	xLabel					markerNames;
	struct mlinkData		*mlinkDataPtr;
    Probability             initialMaleVal, initialFemaleVal;


	/* 1.	Initializations. */

	numThetas = loci.numLoci - 1;
	numPeds = peds.numPeds;

	elimOrderArrs = NULL;

	mlinkDataPtr = &(progData.progSpecific).mlinkData;
	recVaried = mlinkDataPtr->recVaried;
	maleThetaVaried = &(recombData.maleThetas[recVaried]);
	femaleThetaVaried = &(recombData.femaleThetas[recVaried]);
	finalVal = mlinkDataPtr->finalVal;
	inc = mlinkDataPtr->inc;

    /* 2.   Save the initial value(s) of the recombination to be varied. */
    
    initialMaleVal = (*maleThetaVaried)[0];
	initialFemaleVal = (*femaleThetaVaried)[0];

    /* 3.   Calculate the likelihood with recombination to be varied at value 0.5 */
    
	changeThetas(0, 0.5);

	calcLikeNoLinkage(&elimOrderArrs, &areValsElim, &isChangedNet);

	isChangedNet = TRUE;

    /* 4.  If the initial value of the (male) recombination to be varied wasn't 0.5 : */
    
    if (initialMaleVal != 0.5) {
            
		/* 4.1.	Restore the value(s) of the 'recombination to be varied' to their 
				original value. */
            
		updateThetaCompTheta(*maleThetaVaried, initialMaleVal);
		updateThetaCompTheta(*femaleThetaVaried, initialFemaleVal);
           

		/* 4.2.	Continue to calculate the likelihood as long as the the designated 
				male recombination is lower than the final value requested. */
            
		while ((*maleThetaVaried)[0] <= finalVal) {

			/* 4.2.1.	Allocate memory for a result node to store the results of the
						current likelihood evaluation, update the 'currRes'
						field to point to it, and update the distance of the trait locus
						from the leftmost marker in the scan interval. */

			dist = getRecTocM((*maleThetaVaried)[0],progData.haldane);
			dist *= -1;

			initResNode(dist);

            /* 4.2.2.	Perform all the necessary computations. */

			if (isChangedNet) {
				createBN = TRUE;
			}
			else {
				createBN = FALSE;
			}

			performOneIter(&elimOrderArrs, &areValsElim, &isChangedNet, createBN);

			/* 4.2.3.	Print the results. */
            
			printResults_OneIter(numThetas, numPeds, TRUE);

			/* 4.2.4.	Update the maximum result if necessary. */

			updateMax(getLodScore);
            
            /* 4.2.5.	Increment the value of the male & female recombinations to
						be varied. */ 
            
			incThetas(maleThetaVaried, femaleThetaVaried, inc);

			if (((*maleThetaVaried)[0] != inc) && ((*maleThetaVaried)[0] != 0.5)) {
                isChangedNet = FALSE;
			}
		} 
    }

    /* 5.  Free the memory allocated dynamically. */
    
	if (elimOrderArrs != NULL) {
		freeEliminationOrderStructs(&elimOrderArrs, numPeds);
	}

	/* 6.	Update the name of the first (leftmost) marker on the map. */

	updateMarkerName(&markerNames, &(loci.locusArr[1]), 1);

	/* 7.	If more than 1 likelihood evaluation has been performed: */

	if (totalRes.headRes != NULL) {

		/* 7.1.	Reverse the order of the results list. */

		reverseResList();

		/* 7.2.	Print the summary of the results. */
	 
		printResultsSummary_OneLocus(1, &markerNames);
	}
}


/******************************************************************************
*
*       Function        calcLikelihoodNoDisease
*       Parameters      none
*       Output          none
*       Description     Computes liklihood of the data
*       Comments        Author: Anna Tzemach
*
******************************************************************************/
double calcLikelihoodNoDisease(){

    boolean                 areValsElim, isChangedNet, isZeroLike;
    elimOrder               elimOrderArrs;


	/* 1.	Initializations. */
	elimOrderArrs = NULL;


	initResNode(0);
   isZeroLike = performOneIter(&elimOrderArrs, &areValsElim, &isChangedNet, TRUE);

	if (exitAfterFindOrder == TRUE) {
		return totalRes.logLike_NoLinkage;
	}

    /* 3.  Free the memory allocated dynamically. */
    
	if (elimOrderArrs != NULL) {
		freeEliminationOrderStructs(&elimOrderArrs, peds.numPeds);
	}
	
	if (isZeroLike)
		return MIN_DBL;
	return totalRes.currRes->lnLike;
}
/******************************************************************************
*
*       Function        calcLikelihoodSuperPos
*       Parameters      none
*       Output          none
*       Description     Calculates the likelihood at some specific positions of 
*						the disease locus (loci).
*       Comments        none
*
******************************************************************************/
void calcLikelihoodSuperPos()
{
	struct posData			*posDataPtr;
	evaluateLikeData		evalData;
	onePosData				*onePosDataPtr, *onePosDataPtrEnd;

	/* 1.	Perform initializations. */

	posDataPtr = &(progData.progSpecific).posData;
	evalData.numTraitLoci = posDataPtr->numIterLoci;
	evalData.retFunc = getLnLike;
	evalData.printResults = TRUE;

	/* 2.	Iterate over the requested positions and calculate the likelihood 
			at each one of these positions. */

	for (onePosDataPtr = posDataPtr->posDataArr, 
			onePosDataPtrEnd = onePosDataPtr + posDataPtr->numPos;
			onePosDataPtr < onePosDataPtrEnd; onePosDataPtr++) {

		envEvaluateLike(onePosDataPtr->locPos, &evalData);
		myFree(onePosDataPtr->locPos);
	}

	/* 3.	Free the memory allocated dynamically. */

	myFree(posDataPtr->posDataArr);
	myFree(thetaMarkers);
}




/******************************************************************************
*
*       Function        calcLikelihoodSuperGH
*       Parameters		none
*       Output          none
*       Description		Moves the disease locus/loci along the genetic map
*						and calculates the likelihood and LOD score at each position.
*       Comments		none
*
******************************************************************************/
void calcLikelihoodSuperGH()
{
	resNode				*maxResRec, *headResRec, *unlinkResRec;
	xLabel				*markerNames = NULL, *currMarker;
	scanIntervalData	*scanData;
	int					locusInd, numMarkersScan = 0;

	if ((progData.progSpecific).sGHData.numIterLoci == 1)  {

		scanData = progData.scanData;
		numMarkersScan = scanData->numIntervals + 1;

		/* 1.	Allocate memory for an array that would store the names of the markers 
				in the scan interval. */

		myMalloc(markerNames, sizeof(xLabel) * numMarkersScan);

		/* 2.	Update the array with the markers' names. */

		for (locusInd = scanData->startPos, currMarker = markerNames; 
				locusInd <= scanData->endPos; locusInd++, currMarker++) {

			updateMarkerName(currMarker, &(loci.locusArr[locusInd]), locusInd);
		}

		if ((progData.progSpecific).sGHData.checkModel == TRUE) {

			/* 3.	If there's one iterated disease locus and the transmission 
					model isn't known for certain: */

			/* 3.1.	Check recessive transmission model with 50% penetrance. */

			updateTransmissionModel(TRUE, HALF_PENETRANCE);
			iterateOneLocus(numMarkersScan, markerNames);

			/* 3.2.	Store the results obtained by the recessive model. */

			maxResRec = totalRes.maxRes;
			headResRec = totalRes.headRes;
			unlinkResRec = totalRes.unlinkedRes;
			totalRes.headRes = totalRes.maxRes = totalRes.currRes = NULL;

			/* 3.3.	Restore the theta values to their initial values. */

			memcpy(recombData.maleThetas + 1, thetaMarkers, 
					sizeof(thetaCompTheta) * (loci.numLoci-2));
			memcpy(recombData.femaleThetas + 1, thetaMarkers, 
					sizeof(thetaCompTheta) * (loci.numLoci-2));

			for (locusInd = 0; locusInd < loci.numLoci; locusInd++) {
				loci.mapToData[locusInd] = locusInd;
			}

			/* 3.4.	Check dominant transmission model with 50% penetrance. */

			updateTransmissionModel(FALSE, HALF_PENETRANCE);
			iterateOneLocus(numMarkersScan, markerNames);

			/* 3.5.	Compare the maximum result obtained by each model, and choose
					the one which produced the maximum result. */

			if (getLodScore(totalRes.maxRes) < getLodScore(maxResRec)) {
				myPrintf("\n\nRecessive transmission model seems more likely.\n");

				freeList(totalRes.currRes);
				freeResNode(totalRes.unlinkedRes);

				totalRes.headRes = headResRec;
				totalRes.maxRes = maxResRec;
				totalRes.unlinkedRes = unlinkResRec;
			}
			else {
				myPrintf("\n\nDominant transmission model seems more likely.\n");

				freeList(headResRec);
				freeResNode(unlinkResRec);
			}
			
			myPrintf("The results (minus 0.3 for the correction for multiple models):\n");

			/* 3.6.	Print the summary of the results. */
			 
			printResultsSummary_OneLocus_MMLS(numMarkersScan, markerNames);
		}
		else {
			iterateOneLocus(numMarkersScan, markerNames);
		}

		/* 4.	Free the memory allocated dynamically. */

		myFree(markerNames);
	}
	else { 
		iterateTwoLoci();
	}

	freeSuperGHDS();
}


/******************************************************************************
*
*       Function        updateMarkerName
*       Parameters		markerNamesEntry (the label entry which needs to be 
*						updated with the label name), loc (the locus structure 
*						whose name needs to be updated in the label), ind (the
*						index of this locus).
*       Output          none
*       Description		Updates the name of the marker in position 'ind' in the 
*						entry pointed to by 'markerNamesEntry'.
*       Comments		none
*
******************************************************************************/
void updateMarkerName(xLabel *markerNamesEntry, locus *loc, int ind) 
{
	if (loc->name != NULL) { 
		strcpy(*markerNamesEntry, loc->name); 
	} 
	else { 
		sprintf(*markerNamesEntry, "loc%d", ind); 
	}
}


/******************************************************************************
*
*       Function        iterateOneLocus
*       Parameters		none
*       Output          none
*       Description		Moves the disease locus along the genetic map
*						and calculates the likelihood and LOD score at each position.
*       Comments		(1)	This function is for the case where there's only ONE
*						iterated disease locus.
*						(2)	The assumption is that the iterated disease locus
*						appears first in the input files.
*
******************************************************************************/
void iterateOneLocus(int numMarkersScan, xLabel *markerNames)
{
	int				locusInd, numIter, oldLocusInd, numThetas, numPeds,	currPos,
					numIntervalsMap;
	double			theta;
	elimOrder		elimOrderArrs;
	boolean			areValsElim, isChangedNet;
	locusPos		*locusPosPtr, *locusPosPtrEnd;
	superGHData		*sGHData;


	/* 1.	Initializations. */

	elimOrderArrs = NULL;
	numThetas = loci.numLoci - 1;
	numPeds = peds.numPeds;
	sGHData = &(progData.progSpecific).sGHData;

	/* 2.	Initialize 'oldLocusInd' to the initial position of the disease 
			locus in the input (it is first). */

	oldLocusInd = 0;

	/* 3.	Calculate the number of iterations required, determine the position of the
			disease locus on the gentic map in each iteration, and perform initializations. */
	
	sGHData->numIter = initPositionData();

	numIter = sGHData->numIter + 1;

	/* 4.	Calculate the likelihood for the case where the disease locus
			is first on the map and there's no linkage between it and the rest of 
			the loci on the map. */
	
	myPrintf("\nIteration #0 (%d): D#1 (1)", numIter);
	myPrintf("\n==========================\n");

	changeThetas(0, 0.5);

	calcLikeNoLinkage(&elimOrderArrs, &areValsElim, &isChangedNet);

	/* 5.	Perform all necessary iterations. */
	
	for (locusPosPtr = (sGHData->sGHDataLocusArr)->locusPosArr,
			locusPosPtrEnd = locusPosPtr + (sGHData->sGHDataLocusArr)->numPos, currPos = 0, 
			currIter = 1; locusPosPtr < locusPosPtrEnd; locusPosPtr++, currPos++, currIter++) {

		/* 5.1.	Update the position of the disease locus in the current iteration
				and the recombination values between the disease locus and the marker
				preceeding it (in case it isn't first. If it is first, then find the 
				recombination fraction between it and the marker following it). */

		locusInd = locusPosPtr->locusIndMap;
		theta = locusPosPtr->rec;

		/* 5.2.	If necessary, move the disease locus to the right. */

		if (oldLocusInd != locusInd) {
			moveLocus(oldLocusInd, locusInd);

			/* 5.2.1.	 Save the current position of the disease locus. */

			oldLocusInd = locusInd;
		}

		/* 5.3.	 Update the recombination values. */

		changeThetas(locusInd,theta);

		/* 5.4.	Allocate memory for a result node to store the results of the
				current likelihood evaluation, and update the 'currRes'
				field to point to it. */

		initResNode(locusPosPtr->distFromMarker1);

		myPrintf("\nIteration #%d (%d): D#1(%d)",currIter,numIter, locusInd+1); 
		myPrintf("\n===========================\n");

		/* 5.5.	Calculate the likelihood. */

		performOneIter(&elimOrderArrs, &areValsElim, &isChangedNet, TRUE);
	
		/* 5.6.	Print and save the results. */

		printResults_OneIter(numThetas, numPeds, TRUE);
	
		/* 5.7.	Update the maximum result if necessary. */

		updateMax(getLodScore);
	}

 	/* 6.	Free the memory allocated dynamically. */
	
    if (elimOrderArrs != NULL)
		freeEliminationOrderStructs(&elimOrderArrs, numPeds);

   /* 7.	Print the summary of the results. */
 
	printResultsSummary_OneLocus(numMarkersScan, markerNames);

	numIntervalsMap = loci.numLoci - ((progData.progSpecific).sGHData).numIterLoci - 1;
	memcpy(recombData.maleThetas + ((progData.progSpecific).sGHData).numIterLoci, 
			thetaMarkers, sizeof(thetaCompTheta) * numIntervalsMap);

}



/******************************************************************************
*
*       Function        iterateTwoLoci
*       Parameters		none
*       Output          none
*       Description		Moves the 2 disease loci along the genetic map
*						and calculates the likelihood and LOD score at each joint
*						position.
*       Comments		(1)	This function is for the case where there are TWO
*						iterated disease loci.
*						(2)	The assumption is that the iterated disease loci
*						appear first in the input files.
*
******************************************************************************/
void iterateTwoLoci()
{
	int				locusInd1, locusInd2, numIter, oldLocusInd1, oldLocusInd2, 
					numThetas, numPeds, currPos1, currPos2;
	double			theta1, theta2;
	elimOrder		elimOrderArrs;
	boolean			areValsElim, isChangedNet;
	locusPos		*locusPosPtr1, *locusPosPtrEnd1, *locusPosPtr2, *locusPosPtrEnd2;
	thetaCompTheta *maleThetas, *femaleThetas;
	superGHData		*sGHData;


	/* 1.	Initializations. */

	elimOrderArrs = NULL;
	numThetas = loci.numLoci - 1;
	numPeds = peds.numPeds;

	maleThetas = recombData.maleThetas;
	femaleThetas = recombData.femaleThetas;
	sGHData = &(progData.progSpecific).sGHData;

	/* 2.	Initialize 'oldLocusInd1' and oldLocusInd2' to the initial positions of 
			the disease loci in the input (they appear first). */

	oldLocusInd1 = 0;
	oldLocusInd2 = 1;

	/* 3.	Calculate the number of iterations required, determine the positions of the
			disease loci on the gentic map in each iteration, and perform initializations. */
	
	sGHData->numIter = initPositionData();
	numIter = sGHData->numIter + 1;

	/* 4.	Calculate the likelihood for the case where the disease loci
			are first on the map and there's no linkage between them, or between them
			and the rest of the loci on the map. */
	
	locusInd1 = 0;
	locusInd2 = 1;
	femaleThetas[0][0] = maleThetas[0][0] = 0.5;
	femaleThetas[0][1] = maleThetas[0][1] = 0.5;
	femaleThetas[1][0] = maleThetas[1][0] = 0.5;
	femaleThetas[1][1] = maleThetas[1][1] = 0.5;

	myPrintf("\nIteration #0 (%d): D#1(1), D#2(2)", numIter);
	myPrintf("\n=================================\n");

	calcLikeNoLinkage(&elimOrderArrs, &areValsElim, &isChangedNet);

	(totalRes.currRes)->distFromMarker1Arr[1] = UNLINKED_DIST;

	/* 5.	Perform all necessary iterations. Iterate over all requested positions for
			the first disease locus. */

	for (locusPosPtr1 = (sGHData->sGHDataLocusArr[0]).locusPosArr,
			locusPosPtrEnd1 = locusPosPtr1 + (sGHData->sGHDataLocusArr[0]).numPos,
			currPos1 = 0, currIter = 1; locusPosPtr1 < locusPosPtrEnd1; locusPosPtr1++, currPos1++) {

		/* 5.1.	Update the position of the first disease locus in the current iteration
				and the recombination values between the disease locus and the marker
				preceeding it (in case it isn't first. If it is first, then find the 
				recombination fraction between it and the marker following it). */

		theta1 = locusPosPtr1->rec;

		/* 5.2.	Iterate over all requested positions for the second disease locus. */

		for (locusPosPtr2 = (sGHData->sGHDataLocusArr[1]).locusPosArr,
			locusPosPtrEnd2 = locusPosPtr2 + (sGHData->sGHDataLocusArr[1]).numPos,
			currPos2 = 0; locusPosPtr2 < locusPosPtrEnd2; locusPosPtr2++, 
			currPos2++, currIter++) {


			locusInd1 = locusPosPtr1->locusIndMap;

			/* 5.2.1.	Update the position of the disease locus in the current iteration
						and the recombination values between the disease locus and the marker
						preceeding it (in case it isn't first. If it is first, then find the 
						recombination fraction between it and the marker following it). */

			locusInd2 = locusPosPtr2->locusIndMap;
			theta2 = locusPosPtr2->rec;

			/* 5.2.2.	If necessary, move the disease loci, and save their new positions. */

			moveLoci(oldLocusInd1, &locusInd1, theta1, oldLocusInd2, &locusInd2, theta2);

			oldLocusInd1 = locusInd1;
			oldLocusInd2 = locusInd2;

			/* 5.2.3.	Allocate memory for a result node to store the results of the
						current likelihood evaluation, update the 'currRes'
						field to point to it, and update the distances of the trait loci
						from the leftmost markers in their scan intervals. */

			initResNode2Loc(locusPosPtr1->distFromMarker1, locusPosPtr2->distFromMarker1);

			myPrintf("\nIteration #%d (%d): D#1(%d), D#2(%d)",currIter,numIter, locusInd1+1,
															locusInd2+1);
			myPrintf("\n====================================\n");

			/* 5.2.4.	Calculate the likelihood. */

			performOneIter(&elimOrderArrs, &areValsElim, &isChangedNet, TRUE);
			
			/* 5.2.5.	Print and save the results. */

			printResults_OneIter(numThetas, numPeds, TRUE);
			
			/* 5.2.6.	Update the maximum result if necessary. */

			updateMax(getLodScore);
		}
	}


	/* 6.	Free the memory allocated dynamically. */
	
    if (elimOrderArrs != NULL)
		freeEliminationOrderStructs(&elimOrderArrs, numPeds);

    /* 7.	Print the summary of the results. */
 
	printResultsSummary_TwoLoci();
}



/******************************************************************************
*
*       Function        calcLikelihoodSuperOptlink
*       Parameters		none
*       Output          none
*       Description		Finds the position of the affection status locus within
*						the specified scan interval which maximizes the 
*						likelihood of the data.
*       Comments		none
*
******************************************************************************/
void calcLikelihoodSuperOptlink()
{
	double				maxLodRec, maxLodDom, maxLod, maxRecPos, maxDomPos, 
						maxPos, maxLnLikeRec, maxLnLikeDom;
	int					locusInd;
	scanIntervalData	*scanData;
	evaluateLikeData	evalData;


	/* 1.	Perform initializations. */

	scanData = progData.scanData;
	evalData.numTraitLoci = 1;
	evalData.retFunc = getLodScore;
	evalData.printResults = TRUE;

	/* 2.	Find the position where the function has a maximum. */

	if ((progData.progSpecific).optlinkData.checkModel == TRUE) {

		/* 2.	If the transmission model isn't known for certain: */

		/* 2.1.	Check recessive transmission model with 50% penetrance. */

		updateTransmissionModel(TRUE, HALF_PENETRANCE);
		maxLodRec = calcLikeSuperOptlinkOneModel(&maxRecPos, &evalData, scanData);

		/* 2.2.	Store the results obtained for this model. */

		maxLnLikeRec = (totalRes.maxRes)->lnLike;

		/* 2.3.	Restore the theta values to their initial values. */

		memcpy(recombData.maleThetas + 1, thetaMarkers, 
				sizeof(thetaCompTheta) * (loci.numLoci-2));
		memcpy(recombData.femaleThetas + 1, thetaMarkers, 
				sizeof(thetaCompTheta) * (loci.numLoci-2));

		for (locusInd = 0; locusInd < loci.numLoci; locusInd++) {
			loci.mapToData[locusInd] = locusInd;
		}

		/* 2.4.	Free the result list. */

		/* 2.4.1.	Free the result node which stores the result assuming no linkage. */

		if (totalRes.unlinkedRes != NULL) {
			freeResNode(totalRes.unlinkedRes);
		}

		/* 2.4.2.	Free the rest of the results nodes. */

		freeList(totalRes.headRes);

		totalRes.headRes = totalRes.maxRes = totalRes.currRes = NULL;

		/* 2.5.	Check dominant transmission model with 50% penetrance. */

		updateTransmissionModel(FALSE, HALF_PENETRANCE);
		maxLodDom = calcLikeSuperOptlinkOneModel(&maxDomPos, &evalData, scanData);

		/* 2.6.	Store the results obtained for this model. */

		maxLnLikeDom = (totalRes.maxRes)->lnLike;

		/* 2.7.	Print the positions which generated the maximum result for each model. */

		myPrintf("\n---------------------------------------------------------------\n");
		myPrintf("      MODEL     | Trait Position (cM) | -2 LN(LIKE) | LOD_SCORE\n");
		myPrintf("                |  (from marker #%d)   | 	    |\n", scanData->startPos);
		myPrintf("---------------------------------------------------------------\n");
		myPrintf(" recessive, 	|     %-14lf  |  %-9lf | %lf\n", maxRecPos, 
				 maxLnLikeRec * (-2), maxLodRec);
		myPrintf(" 50%% penetrance |		      |		    |	\n");
		myPrintf("---------------------------------------------------------------\n");
		myPrintf(" dominant,  	|     %-14lf  |  %-9lf | %lf\n", maxDomPos, 
				maxLnLikeDom * (-2), maxLodDom);
		myPrintf(" 50%% penetrance |		      |		    |	\n");
		myPrintf("---------------------------------------------------------------\n");
		myPrintf("---------------------------------------------------------------\n");

		/* 2.8.	Compare the maximum result obtained by each model, and choose
				the one which produced the maximum result. */

		if (maxLodRec > maxLodDom) {
			myPrintf(" MAX:recessive  |     %-14lf  | 	    | %lf\n", maxRecPos,
					 maxLodRec - 0.3);
		}
		else {
			myPrintf(" MAX:dominant   |     %-14lf  | 	    | %lf\n", maxDomPos,
					 maxLodDom - 0.3);
		}
		myPrintf("---------------------------------------------------------------\n");
	}
	else {
		maxLod = calcLikeSuperOptlinkOneModel(&maxPos, &evalData, scanData);

		myPrintf("\n---------------------------------------------------------------\n");
		myPrintf("      MODEL     | Trait Position (cM) | -2 LN(LIKE) | LOD_SCORE\n");
		myPrintf("                |  (from marker #%d)   | 	    |\n", scanData->startPos);
		myPrintf("---------------------------------------------------------------\n");
		myPrintf(" MAX:specified 	|     %-14lf  |  %-9lf | %lf\n", maxPos, 
				totalRes.maxRes->lnLike * (-2), maxLod);
		myPrintf("---------------------------------------------------------------\n");
	}

	/* 3.	Free the memory allocated dynamically. */

	myFree(thetaMarkers);
}

/******************************************************************************
*
*       Function        calcLikeSuperOptlinkOneModel
*       Parameters		xMax (is updated with the position of the trait locus
*						in which the function has a maximum), evalData (data
*						which is needed by the evaluating functio), scanData
*						data on the scan interval).
*       Output          none
*       Description		Finds (and updates 'xMax' with) the position of the 
*						affection status locus within the specified scan 
*						interval which maximizes the likelihood of the data.  
*       Comments		none
*
******************************************************************************/
double calcLikeSuperOptlinkOneModel(double *xMax, evaluateLikeData *evalData, 
									scanIntervalData *scanData) 
{
	elimOrder			elimOrderArrs = NULL;
	boolean				areValsElim, isChangedNet, isZeroLike;
	double				maxFuncVal;

	/* 1.	Calculate the likelihood for the case where the disease locus
			is first on the map and there's no linkage between it and the rest of 
			the loci on the map. If the likelihood is zero, return MIN_DBL. */
	
	changeThetas(0, 0.5);

	isZeroLike = calcLikeNoLinkage(&elimOrderArrs, &areValsElim, &isChangedNet);
	
	freeEliminationOrderStructs(&elimOrderArrs, peds.numPeds);

	if (isZeroLike) {
		return MIN_DBL;
	}

	/* 2.	Find the maximum value of the function within the scan interval. */

	maxFuncVal = findMaxInScanInterval(envEvaluateLike, evalData, xMax);

	/* 3.	Return the maximum value of the function found. */

	return maxFuncVal;
}



/******************************************************************************
*
*       Function        performOneIter
*       Parameters      elimOrderArrs (holds the elimination order for all the 
*						products), areValsElim (a flag which indicates whether 
*						values were eliminated from the variables of the products 
*						in the previous iteration), isChangedNet (a flag that 
*						indicates whether the Bayesian network has changed from 
*						the previous iteration to the current one), readElimOrder 
*						(a flag which indicates whether the elimination order 
*						needs to be read from a file or determined by the program).
*       Output          none
*       Description     Performs one iteration of likelihood calculation.
*       Comments        none
*
******************************************************************************/
boolean performOneIter(elimOrder *elimOrderArrs, boolean *areValsElim, 
					   boolean *isChangedNet, boolean createBN)
{
	static boolean		isFirstTime = TRUE;
	boolean				*isZeroLikeArrPtr, isZeroLike;
	product				**prodArr;
	int					pedNum;
	double res;

#if UNITE_VARS
	changeUniteTList	*listPtr;
#endif

#if PRINT_VARIOUS_STATISTICS_LEVEL
    clock_t                 start_time, end_time;
    double                  findElimOrderTime;		/* The time needed for finding 
													   the elimination order. */
#endif

	if (progData.programCode == CALCULATE_LIKELIHOOD || progData.programCode == FIND_ORDER_PROG){
		isFirstTime = TRUE;
	}

	/* 1.	If necessary, create the Bayes networks for the pedigrees. */

	if (createBN) {

		/* 1.1.	Free the memory allocated for the nodes of the Bayesian networks. */

		if (!isFirstTime  ) {
			freeNodes();
		}

		/* 1.2.	Construct the nodes of the Bayesian network for each pedigree. */

		if (loci.sexLink)
			createBayesNetNodesPeds(isFirstTime, createBayesNetNodesSexLinked);
		else
			createBayesNetNodesPeds(isFirstTime, createBayesNetNodesAutosomal);

		/* 1.3.	Update the flags. */

		*areValsElim = TRUE;
		*isChangedNet = TRUE;
	}

	/* 2.	If the intereference option is 2, calculate the probabilities for 
			each combination of existence/lack of recombination in the 2 successive 
			intervals along the chromosome. */
            
    if (recombData.interference == 2) {
		updateJointThetas();
    }

	/* 3.	Allocate memory and create the products. */

	if (loci.sexLink) 
		prodArr = allocateCreateProducts(createPersonVarsSexLinked);
	else
		prodArr = allocateCreateProducts(createPersonVarsAutosomal);
	//printProduct(prodArr[0], 0);
   /* 4.	Eliminate impossible values from the variables of the products. */
    
	if (*areValsElim) {
		eliminateImpossibleValsProducts(prodArr, areValsElim);

		#if UNITE_VARS
			if (*areValsElim) {
				if (loci.sexLink) {
					listPtr = updateUniteVarsPeds(updateUniteVarsPedSexLinked);
				}
				else {
					listPtr = updateUniteVarsPeds(updateUniteVarsPedAutosomal);
				}
			}
		#endif
	}

#if (PRINT_VARIOUS_STATISTICS_LEVEL > 2)
	if (isFirstTime)
		calcVariousStatistics(prodArr, peds.numPeds);
#endif

	/* 5.	If the network changed, free the memory allocated to the elimination
			order structures, if necessary, and determine an elimination order. */
            
#if PRINT_VARIOUS_STATISTICS_LEVEL
    start_time = clock();
#endif

    if (readElimOrder) {
		(*elimOrderArrs) = updateEliminationOrder(prodArr, eliminationOrderFileName);
    }
    else {
	
		if (*isChangedNet) {
			if (*elimOrderArrs != NULL)
				freeEliminationOrderStructs(elimOrderArrs, peds.numPeds);

			if (loci.sexLink) {
				(*elimOrderArrs) = 
					determineEliminationOrder(prodArr, loci.numLoci, updateElimVarsLSegSexlink, 
											  determineElimOrderHeuristic);
			}
			else {
				(*elimOrderArrs) = 
					determineEliminationOrder(prodArr, loci.numLoci, updateElimVarsLSegAutosomal,
											  determineElimOrderHeuristic);
			}
			// Anna

			  if (exitAfterFindOrder == TRUE) {
				totalRes.logLike_NoLinkage = log10(elimOrderArrs[0]->sumTableSizes);
   				exitAfterOrder(prodArr, elimOrderArrs);
				return FALSE;
			}

			if ( (int)(log10(elimOrderArrs[0]->sumTableSizes)) > 8 || (elimOrderArrs[0]->elimOrderArr == NULL) ){
				freeAfterOrder(prodArr, elimOrderArrs);
				return TRUE;
			}
			//Anna
		
		}
	}

#if PRINT_VARIOUS_STATISTICS_LEVEL
    end_time = clock();
	
	if (isFirstTime) {
		findElimOrderTime = ((double)(end_time-start_time))/((double)CLOCKS_PER_SEC);
		fprintf(statisticsFile,"%lf,", findElimOrderTime);

		#if UNITE_VARS
			numOneUnite = numTwoUnite = numFourUnite = 0;
			varsUniteFile = fopen("uniteVars.txt", "a");
			fprintf(varsUniteFile, "\n%s %s\n========\n", datafileName, pedfileName);
		#endif
	}
#endif

	/* 6.	Calculate the likelihood of the pedigrees. */
    
	#if PRINT_MATRIX_MULT_STATISTICS
		matrixFile = fopen("matrixMult.txt", "w");
	#endif
	//printProduct(prodArr[0], 0);
    calcLikelihood(prodArr, elimOrderArrs);

	#if PRINT_MATRIX_MULT_STATISTICS
		fclose(matrixFile);
	#endif

#if (PRINT_VARIOUS_STATISTICS_LEVEL && UNITE_VARS)
	if (isFirstTime) {
		fclose(varsUniteFile);
		varsUniteFile = NULL;
	}
#endif

	for (isZeroLike = FALSE, pedNum = 0, isZeroLikeArrPtr = totalRes.isZeroLikeArr; 
			pedNum < peds.numPeds; pedNum++, isZeroLikeArrPtr++) {

		if (*isZeroLikeArrPtr == TRUE) {
			(totalRes.currRes)->lnLike = MIN_DBL;
			isZeroLike = TRUE;
			break;
		}
	}

	#if (PRINT_VARIOUS_STATISTICS_LEVEL && UNITE_VARS)
		if (isFirstTime) {
			fprintf(statisticsFile,"%d,%d,%d,", numOneUnite, numTwoUnite, numFourUnite);

			numOneUnite = numTwoUnite = numFourUnite = 0;
		}
	#endif


	isFirstTime = FALSE;

	#if UNITE_VARS
		if (*areValsElim) {
			restoreUniteVals(listPtr);
		}
	#endif

	/* 7.	Return TRUE if the likelihood of the data is zero, otherwise
			return FALSE. */

	return isZeroLike;
}


/******************************************************************************
*
*       Function        updateJointThetas
*       Parameters		none
*       Output          none
*       Description		Recalculate all the joint theta values. 
*       Comments		none
*
******************************************************************************/
void updateJointThetas() 
{
	Probability		thetaAB, thetaBC, thetaVal2Seg;;
	
		
	/* 1.	Initialize thetaAB and thetaBC with the male recombination fractions. */
	
	thetaAB = recombData.maleThetas[0][0];
	thetaBC = recombData.maleThetas[1][0];
	
	/* 2.	Calculate thetaAC for males. */
	
	thetaVal2Seg = getKosambiThetaAC(thetaAB, thetaBC);
	
	/* 3.	Calculate g00, g01, g10 and g11 for males. */
	
	recombData.maleJointThetas[0] = 1 - 0.5*(thetaAB + thetaBC + thetaVal2Seg); 
	
	recombData.maleJointThetas[1] = 0.5 * (-thetaAB + thetaBC + thetaVal2Seg);
	
	recombData.maleJointThetas[2] = 0.5 * (thetaAB - thetaBC + thetaVal2Seg);
	
	recombData.maleJointThetas[3] = 0.5 * (thetaAB + thetaBC - thetaVal2Seg);

	/* 4.	If there's a sex-difference in the recombination fractions, calculate 
			the female recombination fractions. Otherwise, initialize the female
			joint recombination fractions according to the male joint recombination
			fractions. */

	if (recombData.sexDiff != 0) {

		/* 1.1.	Initialize thetaAB and thetaBC, with the female recombination 
				fractions. */
		
		thetaAB = recombData.femaleThetas[0][0];
		thetaBC = recombData.femaleThetas[1][0];
		
		/* 4.2.	Calculate thetaAC, for females. */
		
		thetaVal2Seg = getKosambiThetaAC(thetaAB, thetaBC);
		
		/* 4.3.	Calculate g00, g01, g10 and g11, for females. */
		
		recombData.femaleJointThetas[0] = 1 - 0.5*(thetaAB + thetaBC + thetaVal2Seg); 
		
		recombData.femaleJointThetas[1] = 0.5 * (-thetaAB + thetaBC + thetaVal2Seg);
		
		recombData.femaleJointThetas[2] = 0.5 * (thetaAB - thetaBC + thetaVal2Seg);
		
		recombData.femaleJointThetas[3] = 0.5 * (thetaAB + thetaBC - thetaVal2Seg);
	}
	else {
		recombData.femaleJointThetas[0] = recombData.maleJointThetas[0]; 
		
		recombData.femaleJointThetas[1] = recombData.maleJointThetas[1];
		
		recombData.femaleJointThetas[2] = recombData.maleJointThetas[2];
		
		recombData.femaleJointThetas[3] = recombData.maleJointThetas[3];
	}
}


/******************************************************************************
*
*       Function        printResults_OneIter
*       Parameters		numThetas (the number of intervals in the marker map),
*						numPeds (the number of pedigrees in the input).
*       Output          none
*       Description		Prints the results (likelihood of data) of the current 
*						iteration.
*       Comments		none
*
******************************************************************************/
void printResults_OneIter(int numThetas, int numPeds, boolean printLodScore)
{
	thetaCompTheta		*thetaPtr, *thetaPtrEnd;
	int locusIndMap;

	/* 1.	Print the recombination fractions for each interval on the map. */

	myPrintf("\n--------------------------------------------------\n");
	myPrintf("--------------------------------------------------");

	myPrintf("\nLoci order: ");
	for (locusIndMap = 0; locusIndMap < loci.numLoci; locusIndMap++) {
			myPrintf("%d ", (loci.mapToData[locusIndMap]+1));
	}

	if (recombData.sexDiff == 0) {
		myPrintf("\nTHETAS  ");
	}
	else {
		myPrintf("\nMALE THETAS    ");
	}
	
	for (thetaPtr = recombData.maleThetas, thetaPtrEnd = thetaPtr + numThetas;
			thetaPtr < thetaPtrEnd; thetaPtr++) {
		myPrintf("%.3lf  ", (*thetaPtr)[0]);
	}
	if (recombData.sexDiff != 0) {
		myPrintf("\nFEMALE THETAS  ");

		for (thetaPtr = recombData.femaleThetas, thetaPtrEnd = thetaPtr + numThetas;
				thetaPtr < thetaPtrEnd; thetaPtr++) {
			myPrintf("%.3lf  ", (*thetaPtr)[0]);
		}
	}

	myPrintf("\n--------------------------------------------------\n");

	/* 2.	Print the results. */

	if (printLodScore) {
		printResults_OneIterLod(numPeds);
	}
	else {
		printResults_OneIterWOLod(numPeds);
	}
}



/******************************************************************************
*
*       Function        printResults_OneIterLod
*       Parameters		numPeds (the number of pedigrees in the input).
*       Output          none
*       Description		Prints the results (likelihood of data) of the current 
*						iteration, including lod-scores.
*       Comments		none
*
******************************************************************************/
void printResults_OneIterLod(int numPeds)
{
	int					pedInd;
	double				pedsLikelihoodLog, lodScoreT, curPedLogLikelihood, lodScoreF;
	resNode				*currRes;
	boolean				*isZeroLikePed, isZeroLike;
	Probability			*pedLnLike;

	currRes = totalRes.currRes;

	/* 1.	Print the titles. */

	myPrintf("PEDIGREE |  LN(LIKE)  |  LOG 10(LIKE) | LOD_SCORE \n");
	myPrintf("--------------------------------------------------\n");

	/* 2.	Print the ln-likelihood, log-likelihood and LOD-score. */
	
	isZeroLike = FALSE;
	lodScoreT = 0.0;

	/* 3.1.	Print the ln-likelihood, log-likelihood and LOD-score
			for each pedigree separately. */

	for (pedInd = 0, isZeroLikePed = totalRes.isZeroLikeArr, 
			pedLnLike = totalRes.pedsLnLikeArr; pedInd < numPeds;
			pedInd++, isZeroLikePed++, pedLnLike++) {

		curPedLogLikelihood = (*pedLnLike) / LOG_10;

		if (*isZeroLikePed) {
			myPrintf("   %-5d  %-13s %-15s %-15s\n", pedInd+1, "-Infinity", "-Infinity", 
						"-Infinity");
			isZeroLike = TRUE;
		}
		else {
			lodScoreF = curPedLogLikelihood - totalRes.pedsLogLike_NoLinkageArr[pedInd];
			myPrintf("   %-5d  %-12lf  %-14lf  %lf\n", pedInd+1, *pedLnLike, 
						curPedLogLikelihood, lodScoreF);
			lodScoreT += lodScoreF;
		}
	}

	if (isZeroLike) {
		myPrintf("--------------------------------------------------\n");
		myPrintf("TOTAL     %-13s %-15s %-13s\n", "-Infinity", "-Infinity", "-Infinity");
		myPrintf("\n-2 LN(LIKE) = -Infinity \n\n");
	
		lodScoreT = MIN_DBL;
	}
	else {
		pedsLikelihoodLog = currRes->lnLike / LOG_10;

		myPrintf("--------------------------------------------------\n");
		myPrintf("TOTAL     %-13lf %-15lf %-13lf\n", currRes->lnLike, 
					pedsLikelihoodLog, lodScoreT);
		myPrintf("\n-2 LN(LIKE) = %lf \n\n", currRes->lnLike * (-2));
	}

	currRes->lodScore = lodScoreT;
}


/******************************************************************************
*
*       Function        printResults_OneIterWOLod
*       Parameters		numPeds (the number of pedigrees in the input).
*       Output          none
*       Description		Prints the results (likelihood of data) of the current 
*						iteration, without lod-scores.
*       Comments		none
*
******************************************************************************/
void printResults_OneIterWOLod(int numPeds)
{
	int					pedInd;
	double				pedsLikelihoodLog, curPedLogLikelihood;
	resNode				*currRes;
	boolean				*isZeroLikePed, isZeroLike;
	Probability			*pedLnLike;

	currRes = totalRes.currRes;

	/* 1.	Print the titles. */

	myPrintf("PEDIGREE |  LN(LIKE)  |  LOG 10(LIKE) \n");
	myPrintf("--------------------------------------\n");

	/* 2.	Print the ln-likelihood, log-likelihood and LOD-score. */
	
	isZeroLike = FALSE;

	/* 3.1.	Print the ln-likelihood, log-likelihood and LOD-score
			for each pedigree separately. */

	for (pedInd = 1, isZeroLikePed = totalRes.isZeroLikeArr, 
			pedLnLike = totalRes.pedsLnLikeArr; pedInd <= numPeds;
			pedInd++, isZeroLikePed++, pedLnLike++) {

		if (*isZeroLikePed) {
			myPrintf("   %-5d  %-13s %-15s\n", pedInd, "-Infinity", "-Infinity");
			isZeroLike = TRUE;
		}
		else {
			curPedLogLikelihood = (*pedLnLike) / LOG_10;
			myPrintf("   %-5d  %-12lf  %-14lf\n", pedInd, *pedLnLike, curPedLogLikelihood);
		}
	}
	myPrintf("--------------------------------------\n");

	if (isZeroLike) {
		myPrintf("TOTAL     %-13s %-15s\n", "-Infinity", "-Infinity");
		myPrintf("\n-2 LN(LIKE) = -Infinity \n\n");
	}
	else {
		pedsLikelihoodLog = currRes->lnLike / LOG_10;
		myPrintf("TOTAL     %-13lf %-15lf\n", currRes->lnLike, pedsLikelihoodLog);
		myPrintf("\n-2 LN(LIKE) = %lf \n\n", currRes->lnLike * (-2));
	}
}

	
/******************************************************************************
*
*       Function        updateMax
*       Parameters		func (the function whose maximum is seacked).
*       Output          none
*       Description		Updates the pointer to the results node with the maximum
*						value of the function.
*       Comments		none
*
******************************************************************************/
void updateMax(retValFuncType func) 
{
	/* 1.	Update the pointer to the results node with the maximum probability. */

	if (totalRes.maxRes != NULL) {
		if ((*func)(totalRes.maxRes) < (*func)(totalRes.currRes)) {
			totalRes.maxRes = totalRes.currRes;
		}
	}
	else {
		totalRes.maxRes = totalRes.currRes;
	}
}


/******************************************************************************
*
*       Function        printResultsSummary_OneLocus
*       Parameters      numMarkers (the number of markers in the scan interval),
*						markerNames (an array with the names of the markers in
*						the scan interval).
*       Output          none
*       Description		prints the results of the likelihood evaluation at each 
*						iteration (Lod-score, ln-likelihood) and the maximum
*						result (Lod-score, ln-likelihood) obtained in all iterations.
*						Also generates a postscript graph with the results if
*						requested.
*       Comments		this function prints the results in the case where only
*						ONE disease locus is iterated.
*
******************************************************************************/
void printResultsSummary_OneLocus(int numMarkers, xLabel *markerNames)
{
	int					numPoints, currMarker;
	resNode				*currRes;
	scanIntervalData	*scanData;
	xLabel				*markerNamePtr;
	double				*distFromMarker1Ptr;
	
#if PRINT_PS_GRAPH
	point			*vals, *valsPtr;
#endif

	scanData = progData.scanData;

	/* 1.	Print the titles of the results table. */

	myPrintf("\n--------------------------------------------------------------\n");
	myPrintf("  Marker | Trait Position (cM) |   LN(LIKE)  | LOD_SCORE\n");
	myPrintf("   Name  |  (from marker #%d)   | 	     |\n", scanData->startPos);
	myPrintf("--------------------------------------------------------------\n");

	/* 2.	Print the result for each iteration. */

	markerNamePtr = markerNames;
	distFromMarker1Ptr = scanData->distFromMarker1;
	currMarker = 0;
	numPoints = 0;

	for (currRes = totalRes.headRes; currRes != NULL; currRes = currRes->nextResNode) {

		if ((currMarker < numMarkers) && 
				(currRes->distFromMarker1Arr[0] >= *distFromMarker1Ptr)) {
			
			myPrintf("  %s\n", *markerNamePtr);
			markerNamePtr++;
			distFromMarker1Ptr++;
			currMarker++;
		}

		myPrintf("\t\t%lf\t ", currRes->distFromMarker1Arr[0]);

		if (currRes->lnLike > MIN_DBL) {
			myPrintf("%lf\t%lf\n", currRes->lnLike, currRes->lodScore); 
		}
		else {
			myPrintf("-Infinity\t-Infinity\n", currRes->lnLike, currRes->lodScore); 
		}
		numPoints++;
	}

	if (currMarker < numMarkers) {
		myPrintf("  %s\n", *markerNamePtr);
	}

	myPrintf("--------------------------------------------------------------\n");

	/* 3.	Print the maximum result obtained and the position of the disease
			locus that produced it. */

	currRes = totalRes.maxRes;

	if (currRes->lnLike > MIN_DBL) {
		if (currRes == totalRes.unlinkedRes) {
			myPrintf("  MAX\t\tUNLINKED\t ");
		}
		else {
			myPrintf("  MAX\t\t%lf\t ", currRes->distFromMarker1Arr[0]);
		}

		myPrintf("%lf\t%lf\n", currRes->lnLike, currRes->lodScore);
	}
	else {
		myPrintf("  MAX\t\t\t\t -Infinity\t -Infinity\n");
	}

	myPrintf("--------------------------------------------------------------\n");

#if	PRINT_PS_GRAPH

	if (numPoints < 2) {
		return;
	}

	/* 4.	If requested print a Postscript graph of the results. */

	/* 4.1.	Allocate memory for an array that would store all the control points of the curve. */

	myMalloc(vals, sizeof(point) * numPoints);

	/* 4.2.	Update the array with all the points (the x-value is the distance of the
			disease locus from the leftmost marker in the scan interval and the y-value
			is the Lod-score at this position). */
	
	currRes = totalRes.headRes;

	if (currRes->distFromMarker1Arr[0] == -UNLINKED_DIST) {
		numPoints--;
		currRes = currRes->nextResNode;
	}

	for (valsPtr = vals; currRes != NULL; currRes = currRes->nextResNode, valsPtr++) {
		
		valsPtr->x = currRes->distFromMarker1Arr[0];
		valsPtr->y = currRes->lodScore;
	}

	if (vals[numPoints-1].x == UNLINKED_DIST) {
		numPoints--;
	}

	/* 4.3.	Create the graph. */

	createPSGraph("lod_plot.ps", "LOD-Score Plot", "LOD", vals, numPoints, 
				  markerNames, numMarkers, scanData->mapDistArr, FALSE);

	/* 4.4.	Free the memory allocated dynamically. */

	myFree(vals);
#endif
}


/******************************************************************************
*
*       Function        printResultsSummary_OneLocus_MMLS
*       Parameters      numMarkers (the number of markers in the scan interval),
*						markerNames (an array with the names of the markers in
*						the scan interval).
*       Output          none
*       Description		prints the results of the likelihood evaluation at each 
*						iteration (Lod-score, ln-likelihood) and the maximum
*						result (Lod-score, ln-likelihood) obtained in all iterations.
*						Also generates a postscript graph with the results if
*						requested.
*       Comments		this function prints the results in the case where only
*						ONE disease locus is iterated.
*
******************************************************************************/
void printResultsSummary_OneLocus_MMLS(int numMarkers, xLabel *markerNames)
{
	int					numPoints;
	resNode				*currRes;
	scanIntervalData	*scanData;
	
#if PRINT_PS_GRAPH
	point			*vals, *valsPtr;
#endif

	scanData = progData.scanData;

	/* 1.	Print the titles of the results table. */

	myPrintf("\n---------------------------------------------\n");
	myPrintf("  Marker | Trait Position (cM) |   LOD_SCORE\n");
	myPrintf("   Name  |  (from marker #%d)   |\n", scanData->startPos);
	myPrintf("---------------------------------------------\n");

	/* 2.	Print the result for each iteration. */

	numPoints = 0;

	for (currRes = totalRes.headRes; currRes != NULL; currRes = currRes->nextResNode) {
		myPrintf("\t\t%lf\t ", currRes->distFromMarker1Arr[0]);

		if (currRes->lnLike > MIN_DBL) {
			myPrintf("%lf\n", currRes->lodScore - 0.3); 
		}
		else {
			myPrintf("-Infinity\n"); 
		}
		numPoints++;
	}

	myPrintf("---------------------------------------------\n");


	/* 3.	Print the maximum result obtained and the position of the disease
			locus that produced it. */

	currRes = totalRes.maxRes;

	if (currRes->lnLike > MIN_DBL) {
		if (currRes == totalRes.unlinkedRes) {
			myPrintf("  MAX\t\tUNLINKED\t ");
		}
		else {
			myPrintf("  MAX\t\t%lf\t ", currRes->distFromMarker1Arr[0]);
		}

		myPrintf("%lf\n", currRes->lodScore - 0.3);
	}
	else {
		myPrintf("   MAX\t -Infinity\n");
	}

	myPrintf("---------------------------------------------\n");

#if	PRINT_PS_GRAPH

	/* 4.	If requested print a Postscript graph of the results. */

	/* 4.1.	Allocate memory for an array that would store all the control points of the curve. */

	myMalloc(vals, sizeof(point) * numPoints);

	/* 4.2.	Update the array with all the points (the x-value is the distance of the
			disease locus from the leftmost marker in the scan interval and the y-value
			is the Lod-score at this position). */
	
	currRes = totalRes.headRes;

	if (currRes->distFromMarker1Arr[0] == -UNLINKED_DIST) {
		numPoints--;
		currRes = currRes->nextResNode;
	}

	for (valsPtr = vals; currRes != NULL; currRes = currRes->nextResNode, valsPtr++) {
		
		valsPtr->x = currRes->distFromMarker1Arr[0];
		valsPtr->y = currRes->lodScore - 0.3;
	}

	if (vals[numPoints-1].x == UNLINKED_DIST) {
		numPoints--;
	}

	/* 4.3.	Create the graph. */

	createPSGraph("lod_plot.ps", "LOD-Score Plot", "LOD", vals, numPoints, 
				  markerNames, numMarkers, scanData->mapDistArr, FALSE);

	/* 4.4.	Free the memory allocated dynamically. */

	myFree(vals);
#endif
}


/******************************************************************************
*
*       Function        printResultsSummary_TwoLoci
*       Parameters      none
*       Output          none
*       Description		prints the LOD score of each iteration and the maximum
*						LOD score obtained in all iterations.
*       Comments		this function prints the results in the case where 
*						TWO disease loci are iterated.
*
******************************************************************************/
void printResultsSummary_TwoLoci()
{
	scanIntervalData	*scanData1, *scanData2;
	resNode				*currRes;

	/* 1.	Initialize pointers to the relevant entries in the arrays. */

	scanData1 = &(progData.scanData[0]);
	scanData2 = &(progData.scanData[1]);

	/* 2.	Print the titles of the results table. */

	myPrintf("\n--------------------------------------------------------------------------\n");
	myPrintf("  Marker | Trait #1 Position | Trait #2 Position |   LN(LIKE)  | LOD_SCORE\n");
	myPrintf("   Name  | (from marker #%d)  | (from marker #%d)  | 	       |\n", 
						scanData1->startPos, scanData2->startPos);
	myPrintf("---------------------------------------------------------------------------\n");

	/* 3.	Print the result for each iteration. */

	for (currRes = totalRes.headRes; currRes != NULL; currRes = currRes->nextResNode) {
		myPrintf("\t\t%lf\t%lf\t ", currRes->distFromMarker1Arr[0], currRes->distFromMarker1Arr[1]);

		if (currRes->lnLike > MIN_DBL) {
			myPrintf("  %lf\t  %lf\n", currRes->lnLike, currRes->lodScore); 
		}
		else {
			myPrintf("-Infinity\t-Infinity\n", currRes->lnLike, currRes->lodScore); 
		}
	}

	myPrintf("---------------------------------------------------------------------------\n");

	/* 4.	Print the maximum result obtained and the positions of the disease
			loci that produced it. */

	currRes = totalRes.maxRes;

	if (currRes->lnLike > MIN_DBL) {
		myPrintf("   MAX\t\t%lf\t%lf\t   %lf\t  %lf \n", currRes->distFromMarker1Arr[0], 
										currRes->distFromMarker1Arr[1],  
										currRes->lnLike, currRes->lodScore);
	}
	else {
		myPrintf("\t\t%lf\t %lf\t -Infinity\t-Infinity --> MAX\n", currRes->distFromMarker1Arr[0],
															currRes->distFromMarker1Arr[1]);
	}

	myPrintf("---------------------------------------------------------------------------\n");
}



/******************************************************************************
*
*       Function        calcVariousStatistics
*       Parameters		prodArr (the data structure that holds the products), 
*						numProds (the number of products in the array).
*       Output          none
*       Description		Calculates various statistics about the graphs created.
*       Comments		none
*
******************************************************************************/
#if (PRINT_VARIOUS_STATISTICS_LEVEL > 2)
void calcVariousStatistics(product **prodArr, int numProds)
{
	product			*prod;
	int				prodNum;
	long double		sumNumValsVars;
	variable		*varPtr, *varPtrEnd;
	
#if (PRINT_VARIOUS_STATISTICS_LEVEL > 3)
	long			*numSelectorsArr, ind;
#endif
	
	
#if (PRINT_VARIOUS_STATISTICS_LEVEL > 3)
	myMalloc(numSelectorsArr, sizeof(long) * numLoci);
	
	for (ind = 0; ind < numLoci; ind++)
		numSelectorsArr[ind] = 0;
#endif
	
	/* 1.	Iterate over the products. For each product: */
	
	for (prodNum = 0; prodNum < numProds; prodNum++) {
		
		prod = prodArr[prodNum];
		
		/* 1.1.	Calculate the minimum, maximum and average number of values of a variable. */
		
		prod->maxNumValsVar = INT_MIN;
		prod->minNumValsVar = INT_MAX;
		sumNumValsVars = 0;
		
		for (varPtr = prod->variables, varPtrEnd = varPtr + prod->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			if (varPtr->used) {
				sumNumValsVars += varPtr->numValues;
				
				if (varPtr->numValues < prod->minNumValsVar)
					prod->minNumValsVar = varPtr->numValues;
				
				if (varPtr->numValues > prod->maxNumValsVar)
					prod->maxNumValsVar = varPtr->numValues;
				
#if (PRINT_VARIOUS_STATISTICS_LEVEL > 3)
				if (varPtr->originalNode->nodeType == SELECTOR)
					numSelectorsArr[varPtr->originalNode->lNum]++;
#endif
			}
		}
		
		prod->avgNumValsVar = sumNumValsVars / prod->currNumVars;
		
		/* 1.2.	Calculate the minimum, maximum and average number of selectors of each locus. */
		
#if (PRINT_VARIOUS_STATISTICS_LEVEL > 3)
		prod->maxNumSelectorsInterval = INT_MIN;
		prod->minNumSelectorsInterval = INT_MAX;
		prod->avgNumSelectorsInterval = 0;
		
		for (ind = 0; ind < numLoci; ind++) {
			prod->avgNumSelectorsInterval += numSelectorsArr[ind];
			
			if (numSelectorsArr[ind] > prod->maxNumSelectorsInterval)
				prod->maxNumSelectorsInterval = numSelectorsArr[ind];
			
			if (numSelectorsArr[ind] < prod->minNumSelectorsInterval)
				prod->minNumSelectorsInterval = numSelectorsArr[ind];
		}
		
		prod->avgNumSelectorsInterval /= numLoci;
#endif
	}
}
#endif


/******************************************************************************
*
*       Function        superHaploProg
*       Parameters		none
*       Output          none
*       Description		Finds the most likely haplotypes for the individuals
*						in the input pedigrees.
*       Comments		none
*
******************************************************************************/
double superHaploProg(PedigreeDataStruct *origPedigree){
	product					**prodArr;
	boolean					areValsElim;
    elimOrder				elimOrderArrs;
	argMaxData				*argMaxArr;     
	double res;

	elimOrderArrs = NULL;
	initResNode(0);

	/* 1.	Traces the locations in the data-structures of all the markers 
			(ALLELES type loci), updates the number of markers, updates the 
			mapToData array to point only to the markers and updates the theta 
			array to hold the thetas between each two consecutive markers. */

	updateAllelesLociHaplo();

            
    if (recombData.interference == 2) {
		updateJointThetas();
    }

	/* 2.	Construct the nodes of the Bayesian network for each pedigree. */

	if (loci.sexLink)
		createBayesNetNodesPeds(TRUE, createBayesNetNodesSexLinkedHaplo);
	else
		createBayesNetNodesPeds(TRUE, createBayesNetNodesAutosomalHaplo);

	/* 3.	Allocate memory and create the products. */

	if (loci.sexLink) {
		prodArr = allocateCreateProducts(createPersonVarsSexLinkedHaplo);
	}
	else {
		prodArr = allocateCreateProducts(createPersonVarsAutosomalHaplo);
	}

	/* 4.	Allocate memory for an array of argMaxData, one for each pedigree. */

	if (totalRes.isZeroLikeArr[0]){
		res = MIN_DBL;
	}else{
		myMalloc(argMaxArr, sizeof(argMaxData) * peds.numPeds);

		/* 5.	Eliminate impossible values from the variables of the products. */

		areValsElim = TRUE;
		eliminateImpossibleValsProductsHaplo(prodArr, &areValsElim, argMaxArr);

		/* 6.	Determine an elimination order for the variables of the product. */

		if (loci.sexLink) {
			elimOrderArrs = determineEliminationOrder(prodArr, numMarkers, updateElimVarsLSegSexlinkHaplo,
				determineElimOrderHeuristicHaplo);
		}
		else {
			elimOrderArrs = determineEliminationOrder(prodArr, numMarkers, updateElimVarsLSegAutosomalHaplo,
				determineElimOrderHeuristicHaplo);
		}


		if ( elimOrderArrs[0].elimOrderArr == NULL ){
			res = MIN_DBL;
			printf ("\nres = MIN_DBL\n");
		}else{
			/* 7.	Find the most likely haplotypes for the individuals in the pedigrees. */

			res = findMostLikelyHaplotypes(prodArr, &elimOrderArrs, argMaxArr,origPedigree);
		}
		/* 8.  Free the memory allocated dynamically. */

		if (elimOrderArrs != NULL)
			freeEliminationOrderStructs(&elimOrderArrs, peds.numPeds);
		myFree(argMaxArr);
	}
	freeProducts(prodArr, peds.numPeds);
	
	return res;
}

/******************************************************************************
*
*       Function        createExportBayesProg
*       Parameters		none
*       Output          none
*       Description		print bayes network to the file
*       Comments		none
*
******************************************************************************/
void createExportBayesProg(){
	product					**prodArr;
	boolean					areValsElim;
    elimOrder				elimOrderArrs;
	//argMaxData				*argMaxArr;     
	//elimOrder		elimOrderArrs;
	char family[5];
	int i;
	elimOrderArrs = NULL;
	initResNode(1);

	

	/* 2.	Construct the nodes of the Bayesian network for each pedigree. */

	if (loci.sexLink)
		createBayesNetNodesPeds(TRUE, createBayesNetNodesSexLinked);
	else
		createBayesNetNodesPeds(TRUE, createBayesNetNodesAutosomal);

	/* 3.	Allocate memory and create the products. */

	if (loci.sexLink) {
		prodArr = allocateCreateProducts(createPersonVarsSexLinked);
	}
	else {
		prodArr = allocateCreateProducts(createPersonVarsAutosomal);
	}

	if ( progData.programCode == BAYES_NET_XBN_BEFORE_ELIM ){
		 convertToXBN(prodArr, peds.numPeds,&peds,loci.numLoci);
	}else if (progData.programCode == BAYES_NET_XML_BEFORE_ELIM ){
		convertToXML(prodArr, peds.numPeds,&peds,loci.numLoci);
	}else{

		/* 5.	Eliminate impossible values from the variables of the products. */

		areValsElim = TRUE;
		eliminateImpossibleValsProducts(prodArr, &areValsElim);
	/*	#if UNITE_VARS
		
				if (loci.sexLink) {
					 updateUniteVarsPeds(updateUniteVarsPedSexLinked);
				}
				else {
					 updateUniteVarsPeds(updateUniteVarsPedAutosomal);
				}
			
		#endif
		*/
	
//		printProduct(prodArr[0]);

		if ( progData.programCode == BAYES_NET_XBN_AFTER_ELIM ){
			convertToXBN(prodArr, peds.numPeds,&peds,loci.numLoci);
		} 
		if (progData.programCode == BAYES_NET_XML_AFTER_ELIM ){
			convertToXML(prodArr, peds.numPeds,&peds,loci.numLoci);
		}
		if (progData.programCode == PRINT_ELIM_ORDER){
			if (loci.sexLink) {
				(elimOrderArrs) = 
					determineEliminationOrder(prodArr, loci.numLoci, updateElimVarsLSegSexlink, 
											  determineElimOrderHeuristic);
			}
			else {
				(elimOrderArrs) = 
					determineEliminationOrder(prodArr, loci.numLoci, updateElimVarsLSegAutosomal,
											  determineElimOrderHeuristic);
			}
			for (i = 0; i < peds.numPeds; i++){
				sprintf(family, "_f%d",i);
				printElimOrder(elimOrderArrs,family); 
			}
			exitAfterOrder(prodArr, &elimOrderArrs);
		}
	}

	freeProducts(prodArr, peds.numPeds);

}

/******************************************************************************
*
*       Function        calcLikelihoodSuperLinkmapApprox
*       Parameters		none
*       Output          none
*       Description		Linkmap compatible program
*       Comments		this program performs approximate computations.
*
******************************************************************************/
void calcLikelihoodSuperLinkmapApprox()
{

}


/******************************************************************************
*
*       Function        initResultList
*       Parameters		none.
*       Output          none
*       Description		Allocates memory and initializes the fields of the 
*						data-structure which stores the results of all the 
*						likelihood evaluations.
*       Comments		none.
*
******************************************************************************/
void initResultList() 
{
	/* 1.	Init the results list and allocate memory. */

	myMalloc(totalRes.pedsLnLikeArr, sizeof(Probability) * peds.numPeds);
	myMalloc(totalRes.isZeroLikeArr, sizeof(boolean) * peds.numPeds);
	myMalloc(totalRes.pedsLogLike_NoLinkageArr, sizeof(Probability) * peds.numPeds);

	totalRes.headRes = totalRes.maxRes = totalRes.currRes = NULL;

	/* 2.	Open the results file. */

	if ((resFile = fopen(RESULTS_FILE_NAME, "w")) == NULL) {
		error("File %s could not be opened.", RESULTS_FILE_NAME);
	}
}


/******************************************************************************
*
*       Function        freeResultList
*       Parameters		none.
*       Output          none
*       Description		Frees the memory allocated to the data-structure which
*						stores the results of all the likelihood evaluations.
*       Comments		none.
*
******************************************************************************/
void freeResultList() 
{
	/* 1.	Free the result node which stores the result assuming no linkage. */

	if (totalRes.unlinkedRes != NULL) {
		freeResNode(totalRes.unlinkedRes);
	}

	/* 2.	Free the rest of the results nodes. */

	freeList(totalRes.headRes);

	/* 3.	Free the arrays that stored the results by pedigree. */

	myFree(totalRes.pedsLnLikeArr);
	myFree(totalRes.isZeroLikeArr);
	myFree(totalRes.pedsLogLike_NoLinkageArr);

	/* 4.	Close the results file. */

	fclose(resFile);
}

/******************************************************************************
*
*       Function        freeList
*       Parameters		none.
*       Output          none
*       Description		Frees the memory allocated to the data-structure which
*						stores the results of all the likelihood evaluations.
*       Comments		none.
*
******************************************************************************/
void freeList(resNode *resPtr) 
{
	resNode		*tmpRes;

	/* 1.	Free all the results nodes in the list. */

	while (resPtr != NULL) {
				
		tmpRes = resPtr;
		resPtr = resPtr->nextResNode;

		freeResNode(tmpRes);
	}
}


/******************************************************************************
*
*       Function        reverseResList
*       Parameters      none
*       Output          none
*       Description		reverses the order of the results list.
*       Comments		none
*
******************************************************************************/
void reverseResList()
{
	resNode		*currRes, *tmpResPtr;

	/* 1.	Iterate over the results nodes and reverse their pointers. */

	for (currRes = totalRes.headRes; currRes->nextResNode != NULL; 
			currRes = currRes->prevResNode) {
		
		tmpResPtr =  currRes->prevResNode;
		currRes->prevResNode = currRes->nextResNode;
		currRes->nextResNode = tmpResPtr;
	}

	/* 2.	Reverse the pointers of the last results node. */

	currRes->nextResNode = currRes->prevResNode;
	currRes->prevResNode = NULL;

	/* 3.	Make the last node in the list the first. */

	totalRes.headRes = currRes;
}

/******************************************************************************
*
*       Function        calcLikeNoLinkage
*       Parameters		none.
*       Output          none
*       Description		calculates the likelihood in the case when there is no
*						linkage.
*       Comments		none.
*
******************************************************************************/
boolean calcLikeNoLinkage(elimOrder	*elimOrderArrs, boolean	*areValsElim,
						  boolean *isChangedNet)
{
	boolean			isZeroLike;

		
	/* 1.	Allocate memory for a results node to store the results of the 
			likelihood evaluation when no-linkage is assumed, and update the 
			currRes pointer to point to it. */

	allocateResNode(totalRes.unlinkedRes) 
	totalRes.currRes = totalRes.unlinkedRes;
	(totalRes.currRes)->distFromMarker1Arr[0] = UNLINKED_DIST;

	/* 2.	Calculate the likelihood. */

	isZeroLike = performOneIter(elimOrderArrs, areValsElim, isChangedNet, TRUE);

	/* 3.	If the results need to be printed for each pedigree separately, 
			initialize the array which stores the log-likelihood of each 
			pedigree for the case of no-linkage. */

	updateLogLikes(peds.numPeds);

	/* 4.	Print the results of the likelihood evaluation. */

	printResults_OneIter(loci.numLoci - 1, peds.numPeds, TRUE);

	/* 5.	Update the maximum result if necessary. */

	updateMax(getLodScore);

	/* 6.	Return a boolean value that indicates whether the likelihood was zero
			or not. */

	return isZeroLike;
}



/******************************************************************************
*
*       Function        updateLogLikes
*       Parameters		numPeds (the number of pedigrees in the input).
*       Output          none
*       Description		updates the total log-likelihood and the log-likelihoods
*						of each pedigree separately, if required.
*       Comments		none.
*
******************************************************************************/
void updateLogLikes(int numPeds)
{
	int					pedInd;

    /* 1.	Initialize the array which stores the log-likelihood of each 
			pedigree for the case of no-linkage. */

	for (pedInd = 0; pedInd < numPeds; pedInd++) {
		totalRes.pedsLogLike_NoLinkageArr[pedInd] =
								totalRes.pedsLnLikeArr[pedInd] / LOG_10; 
	}

	totalRes.logLike_NoLinkage = (totalRes.unlinkedRes)->lnLike / LOG_10;
}


/******************************************************************************
*
*       Function        exitAfterOrder
*       Parameters		none
*       Output          none
*       Description		
*       Comments		none.
*
******************************************************************************/
void freeAfterOrder(product	**prodArr, elimOrder *elimOrderArrs)
{
	int			numLoc;

	#if PRINT_VARIOUS_STATISTICS_LEVEL
		fclose(statisticsFile);
	#endif 

	/* 1.	Free the memory allocated dynamically. */

	freeProducts(prodArr, peds.numPeds);
	
	freeEliminationOrderStructs(elimOrderArrs, peds.numPeds);
}
/******************************************************************************
*
*       Function        exitAfterOrder
*       Parameters		none
*       Output          none
*       Description		
*       Comments		none.
*
******************************************************************************/
void exitAfterOrder(product	**prodArr, elimOrder *elimOrderArrs)
{
	int			numLoc;

	#if PRINT_VARIOUS_STATISTICS_LEVEL
		fclose(statisticsFile);
	#endif 

	/* 1.	Free the memory allocated dynamically. */

	freeProducts(prodArr, peds.numPeds);
	
	freeEliminationOrderStructs(elimOrderArrs, peds.numPeds);

	numLoc = (progData.programCode == SUPER_HAPLO) ? numMarkers : loci.numLoci;

	#if PERFORM_ALLELE_RECODING
		freePedDataAR(numLoc);
	#else
		freePedData(numLoc);
	#endif

	freeLociDataSL();
	freeResultList();
   
	//change by Anna 26/2/07
	//exit(EXIT_SUCCESS);
	//end change
}


/******************************************************************************
*
*       Function        runHMM
*       Parameters		none
*       Output          true if HMM algorithm is suitable
*       Description		
*       Comments		none.
*
******************************************************************************/
boolean runHMM(){
	int numNonFounders = 0, numFounders = 0;
	pedigree *nextPed;
	if (loci.sexLink == 0 && loci.numLoci > 2
		&& progData.programCode == SUPER_GH_PROG &&
		progData.scanData->numIntervals > 1 &&
		loci.locusArr[0].type == affection &&
		loci.numTraitLoci  == 1 &&
		recombData.sexDiff == 0 && 
		recombData.interference == 0){
			for (nextPed = peds.head ; nextPed != NULL; nextPed = nextPed->nextPed){
				if (numNonFounders < nextPed->numNonFounders){
					numNonFounders = nextPed->numNonFounders;
					numFounders = nextPed->numPeople - nextPed->numNonFounders;
				}
			}
			if ( 2*numNonFounders  < 29){
				return TRUE;
			}
		}

		return FALSE;
}
