/*********************************************************************************/
/*  Variable.c																     */
/*  Author: Ma'ayan Fishelson													 */
/*********************************************************************************/

#include "sl_Product.h"


/******************************************************************************
*
*       Function        getNodeVariable
*       Parameters		Node (the node whose related variable is wanted), 
*						prod (the product to which this variable belongs.
*       Output          none
*       Description		Returns a pointer to the variable assigned to 'Node' in
*						'prod', if one exists. Otherwise, creates a variable
*						for 'Node' and returns a pointer to it.
*       Comments		none
*
******************************************************************************/
variable* getNodeVariable(node *Node, product *prod)
{
	variable	*newVariable;

	/* 1.	Check if the node was already assigned a variable. If so, return
			a pointer to it. */

	if (Node->varInProd)
		return Node->varInProd;

	/* 2.	If not, create a new variable for the node. */

	newVariable = insertVariableToProduct(prod, Node, Node->numValues);
	Node->varInProd = newVariable;

	#if UNITE_VARS
		updateCertainVals(newVariable, Node->uniteT, Node->certainVal);
	#endif	

	/* 3.	Return a pointer to the variable assigned to Node. */

	return newVariable;
}


/******************************************************************************
*
*       Function        insertVariableToProduct
*       Parameters		prod (the product to which a variable is inserted), 
*						Node (the node of the Bayesian network which the 
*						variable represents), numValues (the number of possible
*						values for this variable).
*       Output          none
*       Description		Inserts a new variable which represents 'Node' into the 
*						product.						 
*       Comments		none
*
******************************************************************************/
variable* insertVariableToProduct(product *prod, node *Node, int numValues)
{
	variable	*varPtr, *varPtrEnd;

	/* 1.	Search for an empty place to insert the new variable, starting at 
			'nextVarPlace' and wrapping around. */

	for (varPtr = prod->nextVarPlace, varPtrEnd = prod->variables + prod->numVariables;
			varPtr < varPtrEnd; varPtr++) {
		if (!varPtr->used)
			goto enterVarFields;
	}

	for (varPtr = prod->variables, varPtrEnd = prod->nextVarPlace; 
			varPtr < varPtrEnd; varPtr++) {
		if (!varPtr->used)
			goto enterVarFields;
	}

	/* 2.	Initialize the fields of this place. */

	enterVarFields:

	varPtr->used = TRUE;
	varPtr->functions = NULL;
	varPtr->originalNode = Node;
	varPtr->functionSpaces = 0;
	varPtr->numValues = numValues;
	varPtr->elimComplexity = INVALID;

	/* 3.	Set 'nextVarPlace'. */

	prod->nextVarPlace = varPtr + 1;
	prod->currNumVars++;

	/* 4.	Return a pointer to the slot allocated for the new variable. */

	return varPtr;
}




/******************************************************************************
*
*       Function        updateCertainVals
*       Parameters		varPtr (the variable whose certain values need to be 
*						updated), uniteT (the type of unite),
*						certainVal (the array of indices of certain values of
*						the genotype this variable partially represents).
*       Output          none
*       Description		Allocates memory and updates the array which holds the 
*						indices of the certain values of the genotype this
*						variable partially represents).
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void updateCertainVals(variable *varPtr, uniteType uniteT, int *certainVal)
{
	int		*certainValDst, *certainValSrc, *certainValSrcEnd;


	/* 1.	If this variable needs to be treated jointly with a few other variables,
			allocate memory to store the indices of the certain values for this 
			genotype, and initialize the array with these indices. */

	if (uniteT == oneAll) {
		myMalloc(varPtr->certainVal, sizeof(int));
		varPtr->certainVal[0] = certainVal[0];
	}
	else if (uniteT == fourAll) {
		myMalloc(varPtr->certainVal, sizeof(int) * varPtr->originalNode->numValues);

		for (certainValDst = varPtr->certainVal, certainValSrc = certainVal,
				certainValSrcEnd  = certainValSrc + varPtr->originalNode->numValues; 
				certainValSrc < certainValSrcEnd; certainValSrc++, certainValDst++) {
			
			(*certainValDst) = (*certainValSrc);
		}

	}
	else {
		varPtr->certainVal = NULL;
	}
}
#endif

/******************************************************************************
*
*       Function        buildVariableFunctionTables
*       Parameters		prod (the product for which the tables need to be built).
*       Output          none
*       Description		Builds all the variable->function lookup tables from
*						the existing function->variable lookup tables (each 
*						variable will point to all functions it appears in).						 
*       Comments		none
*
******************************************************************************/
void buildVariableFunctionTables(product *prod)
{
	function	*funcPtr, *funcPtrEnd;
	variable	**varPtr, **varPtrEnd, *variablePtr, *variablePtrEnd;

	/* 1.	Count for each variable, the number of functions it appears in. */

	for (funcPtr = prod->functions, funcPtrEnd = funcPtr + prod->numFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {
		
		if (funcPtr->used) {
			for (varPtr = funcPtr->variables, varPtrEnd = varPtr + funcPtr->numVariables;
					varPtr < varPtrEnd; varPtr++) 
				(*varPtr)->functionSpaces++;
		}
	}

	/* 2.	Allocate the inverse tables for each variable. */

	for (variablePtr = prod->variables, variablePtrEnd = variablePtr + prod->numVariables;
			variablePtr < variablePtrEnd; variablePtr++) {
		
		if (variablePtr->used) {
			myMalloc(variablePtr->functions, sizeof(function*) * variablePtr->functionSpaces);
			variablePtr->currNumFunctions = 0;
		}
	}
	
	/* 3.	Fill the inverse tables for each variable. */

	for (funcPtr = prod->functions, funcPtrEnd = funcPtr + prod->numFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {
		
		if (funcPtr->used) {
			for (varPtr = funcPtr->variables, varPtrEnd = varPtr + funcPtr->numVariables;
					varPtr < varPtrEnd; varPtr++) {
				variablePtr = *varPtr;
				variablePtr->functions[variablePtr->currNumFunctions++] = funcPtr;
			}
		}
	}
}


/******************************************************************************
*
*       Function        addFunctionToVariable
*       Parameters		var (the variable to which a function is added), 
*						func (the function added).
*       Output          none
*       Description		Adds 'func' to var's list of functions and 
*						advances the 'currNumFunctions' counter.						 
*       Comments		none
*
******************************************************************************/
void addFunctionToVariable(variable *var, function *func)
{
	/* 1.	Insert a pointer to the function and update the counter.*/

	var->functions[var->currNumFunctions++] = func;

	/* 2.	Invalidate the elimination complexity cache. */

	var->elimComplexity = INVALID;

}


/******************************************************************************
*
*       Function        removeFunctionFromVariable
*       Parameters		var (the variable from which a function is removed), 
*						func (the function removed).
*       Output          none
*       Description		removes 'func' from var's list of functions.						 
*       Comments		none
*
******************************************************************************/
void removeFunctionFromVariable(variable *var, function *func)
{
	function	**funcPlace, **funcPlaceEnd, **tempFuncPlace;

	/* 1.	Search for the specified function in the Variable's list 
			of functions. */

	for (funcPlace = var->functions, 
			 funcPlaceEnd = funcPlace + var->currNumFunctions;
			 funcPlace < funcPlaceEnd; funcPlace++) {
		
		/* 1.1.	When found:  */

		if (*funcPlace == func) {

			/* 1.1.1.	Shift all the following functions one place back. */

			for (tempFuncPlace = funcPlace; 
					tempFuncPlace < funcPlaceEnd - 1; tempFuncPlace++) {

				(*tempFuncPlace) = *(tempFuncPlace + 1);
			}

			/* 1.1.2.	Update the variable's function counter. */

			var->currNumFunctions--;

			/* 1.1.3.	Invalidate the elimination complexity cache. */

			var->elimComplexity = INVALID;

			return;
		}
	}
}

