/*********************************************************************************/
/*  VariableElimination.c													     */
/*  Author: Ma'ayan Fishelson				                                     */
/*********************************************************************************/

#include "sl_VariableElimination.h"


/******************************************************************************
*
*       Function        getEliminateVar
*       Parameters		elimOrd (the elimination order), 
*						isBreakVar (a flag that would be updated to TRUE if
*						this variable should be conditioned on).
*       Output          none
*       Description		Returns a pointer to the variable chosen next to be 
*						processed (the variable that's chosen is least complex
*						according to the some greedy heuristic).						 
*       Comments		none
*
******************************************************************************/
variable* getEliminateVar(elimOrderPed *elimOrd, boolean *isBreakVar)
{
	elimOrderNode	nextElimNode;
	variable		*nextVar;
	node			*origNode;

#if UNITE_VARS
	node			**uniteVarPtr, **uniteVarPtrEnd;
	boolean			unite;
#endif

	/* 1.	Update 'nextVar' to point to the next variable to be processed. */

	nextElimNode = (elimOrd->elimOrderArr)[elimOrd->currElimNode];
	nextVar = (nextElimNode.originalNode)->varInProd;

	/* 2.	While the next var that is supposed to be processed has been 
			trimmed, go to the next var in the elimination order. */

	while (nextVar->used == FALSE) {
		elimOrd->currElimNode++;
		nextElimNode = (elimOrd->elimOrderArr)[elimOrd->currElimNode];
		nextVar = (nextElimNode.originalNode)->varInProd;
	}

	/* 3.	If this variable needs to be treated jointly with a couple of 
			other variables, check if all the other variables haven't been
			eliminated yet, and if they haven't, update the 'currElimNode'
			pointer to point to the elimination-order-node that appears
			after them. */

	origNode = nextVar->originalNode;

	#if UNITE_VARS
		if (origNode->uniteT) {
			unite = TRUE;

			for (uniteVarPtr = origNode->uniteVars,
					uniteVarPtrEnd = uniteVarPtr + origNode->numVarsUnite;
					uniteVarPtr < uniteVarPtrEnd; uniteVarPtr++) {

				if ((*uniteVarPtr)->varInProd->used == FALSE) {
					unite = FALSE;
					break;
				}
			}

			if (unite == TRUE) {
				*isBreakVar = (nextElimNode.isBreakNode);

				elimOrd->currElimNode += origNode->numVarsUnite;
				return nextVar;
			}
		}
	#endif

	/* 4.	Advance the pointer to the next node that needs to be eliminated. */

	elimOrd->currElimNode++;

	/* 5.	Check if the product needs to be broken according to this variable
			(i.e., if this variable needs to be conditioned on). */

/*	*isBreakVar = ((nextElimNode.isBreakNode) && 
					(getSizeOfCombinedFunc(nextVar) >= breakProductBottomThreshold));
*/
	*isBreakVar = ((nextElimNode.isBreakNode) || 
					(getSizeOfCombinedFunc(nextVar) >= breakProductBottomThreshold));

	/* 6.	Return a pointer to the next variable to be processed. */

	return nextVar;
}


/******************************************************************************
*
*       Function        eliminateVar
*       Parameters		prod (the product from which a variable needs to be 
*						eliminated), elimVar (pointer to the variable that needs
*						to be eliminated), tempRes (if the combined function is
*						a constant, holds the value of this constant),
*						isZeroLike (will be updated to TRUE if this elimination
*						resulted in a constant of ZERO probability).
*       Output          none
*       Description		Eliminates 'elimVar' from 'prod' by multiplying all
*						the functions it appears in and creating one new
*						combined function that 'elimVar' doesn't appear in.					 
*       Comments		Returns TRUE if the combined function is a constant,
*						and FALSE otherwise.
*
******************************************************************************/
boolean eliminateVar(product *prod, variable *elimVar, Probability *tempRes, 
					 boolean *isZeroLike)
{
	variable		**varArr, **varPtr, **varPtrEnd;
	int				numVars, rescaleTimer;
	unsigned long	sizeOfTable;
	Probability		*probabList;
	function		*combinedFunc = NULL, **funcPtr, **funcPtrEnd, *func;
	double			logMultFactor;

#if UNITE_VARS
	node			**uniteVar, **uniteVarEnd;
#endif

	/* 1.	If this variable needs to be treated jointly with some other variables - 
			if all these variables haven't been eliminated or shrunk into 
			constants, then call the function used in this case. */

#if UNITE_VARS
	if (elimVar->originalNode->uniteT) {
		for (uniteVar = elimVar->originalNode->uniteVars, 
				uniteVarEnd = uniteVar + elimVar->originalNode->numVarsUnite;
				uniteVar < uniteVarEnd; uniteVar++) {

			if ((*uniteVar)->varInProd->used == FALSE) 
				goto elimVarAlone;
		}
		return eliminateVarsJoint(prod, elimVar, tempRes, isZeroLike);
	}

elimVarAlone:
#endif

	/* 2.	If this variable appears in only one function, then sum this function on the 
			variable. */

	if (elimVar->currNumFunctions == 1) {
		return sumVar(prod, elimVar, tempRes, isZeroLike);
	}

	*isZeroLike = FALSE;

	/* 3.	Allocate memory for the variable array of the new function that
			will be created, as a result of the elimination of the variable. */

	myMalloc(varArr, sizeof(variable*) * prod->currNumVars);

	/* 4.	Find all the variables that appear in the functions that need 
			to be combined. */

	getVarsInCombinedFunc(&elimVar, 1, varArr, elimVar->functions, 
						  elimVar->currNumFunctions, &numVars, &sizeOfTable);

	/* 5.	Calculate the new logMultFactor & rescaleTimer. */

	logMultFactor = 0.0;
	rescaleTimer = MAX_RESCALE_UNITS;

	for (funcPtr = elimVar->functions, funcPtrEnd = funcPtr + elimVar->currNumFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {
		
		logMultFactor += (*funcPtr)->logMultFactor;

		if ((*funcPtr)->rescaleTimer < rescaleTimer)
			rescaleTimer = (*funcPtr)->rescaleTimer;
	}

	/* 6.	Allocate memory for the new combined function, calculate it & 
			insert it into the product, in case it isn't a constant. */

	myMalloc(probabList, sizeof(Probability) * sizeOfTable);

	if (numVars) {		/* in case the combined function isn't a constant. */
		if (elimVar->currNumFunctions > 1)
			rescaleTimer--;

		combinedFunc = insertFunctionToProduct(prod, sizeOfTable, probabList, logMultFactor,
											   TRUE, numVars, varArr, TRUE, rescaleTimer);

		/* 6.1.	Calculate the new combined function. */

		if (sizeOfTable * elimVar->currNumFunctions < ELIMINATION_SIMPLE_THRESHOLD)
			multiplyAllFuncsSimp(combinedFunc, elimVar);	
		else
			multiplyAllFuncsComplex(combinedFunc, elimVar);	

		/* 6.2.	If, according to the rescaleTimer, the combined function 
				needs to be rescaled, rescale it.*/

		if (combinedFunc->rescaleTimer <= 0) {
			(*isZeroLike) = scaleFunctionProbList(combinedFunc);
			combinedFunc->rescaleTimer = MAX_RESCALE_UNITS;
		}
	}
	else {				/* in case the combined function is a constant. */

		*tempRes = multiplyAllOneVarFuncs(elimVar, logMultFactor);
		myFree(probabList);
	}
	
	/* 7.	Remove the functions that have been multiplied from the variables that
			appear in them and from the product. */

	for (funcPtr = elimVar->functions + elimVar->currNumFunctions - 1, 
			funcPtrEnd = elimVar->functions; funcPtr >= funcPtrEnd; funcPtr--) {

		func = *funcPtr;

		for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			removeFunctionFromVariable(*varPtr, func);
		}

		removeFunctionFromProduct(prod, func);
	}

	/* 8.	Add the new combined function to the variables that appear in it. */

	if (numVars) {
		for (varPtr = combinedFunc->variables, varPtrEnd = varPtr + combinedFunc->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			addFunctionToVariable(*varPtr, combinedFunc);
		}
	}

	/* 9.	Trim the variable that was eliminated from the product. */

	removeVariableFromProduct(prod, elimVar);

	/* 10.	Free the memory allocated. */

	myFree(varArr);

	/* 11.	Return TRUE if the new function is a constant. */

	return ((numVars == 0) ? TRUE : FALSE);
}




/******************************************************************************
*
*       Function        eliminateVarsJoint
*       Parameters		prod (the product from which the variables need to be 
*						eliminated), elimVar (pointer to the variable that needs
*						to be eliminated along with the variables that are 
*						treated jointly with it), tempRes (if the combined 
*						function is a constant, holds the value of this constant),
*						isZeroLike (will be updated to TRUE if this elimination
*						resulted in a constant of ZERO probability).
*       Output          none
*       Description		Eliminates 'elimVar' and the variables that are treated
*						jointly with it from 'prod' by multiplying all
*						the functions that they appear in and creating one new
*						combined function that they don't appear in.					 
*       Comments		Returns TRUE if the combined function is a constant,
*						and FALSE otherwise.
*
******************************************************************************/
#if UNITE_VARS
boolean eliminateVarsJoint(product *prod, variable *elimVar, Probability *tempRes, 
						   boolean *isZeroLike)
{
	variable		**varArr, **varPtr, **varPtrEnd, **elimVarsArr;
	node			**uniteVar;
	int				numElimVars, numVars, rescaleTimer, numFuncs;
	unsigned long	sizeOfTable;
	Probability		*probabList;
	function		*combinedFunc = NULL, **funcPtr, **funcPtrEnd, *func, **funcArr;
	double			logMultFactor;

	*isZeroLike = FALSE;

	/* 1.	Allocate memory for an array of pointers to all the variables that need
			to be eliminated, and update it. */

	numElimVars = (elimVar->originalNode)->numVarsUnite;
	myMalloc(elimVarsArr, sizeof(variable*) * numElimVars);

	for (varPtr = elimVarsArr, varPtrEnd = varPtr + numElimVars, 
			uniteVar = (elimVar->originalNode)->uniteVars, numFuncs = 0; 
			varPtr < varPtrEnd; varPtr++, uniteVar++) {
		(*varPtr) = (*uniteVar)->varInProd;
		numFuncs += (*varPtr)->currNumFunctions;
		zeroSeenFuncs((*varPtr)->functions, (*varPtr)->currNumFunctions, funcPtr, funcPtrEnd);
	}

	/* 2.	Allocate memory for an array of pointers to all the functions that need to
			be combined. */

	myMalloc(funcArr, sizeof(function*) * numFuncs);

	/* 3.	Allocate memory for the variable array of the new function that
			will be created, as a result of the elimination of the variables. */

	myMalloc(varArr, sizeof(variable*) * prod->currNumVars);

	/* 4.	Find all the variables that appear in the functions that need 
			to be combined, update an array with pointers to all the functions that
			need to be combined. */

	getVarsInCombinedFuncJoint(elimVarsArr, numElimVars, funcArr, &numFuncs, varArr, &numVars, 
							   &sizeOfTable, &logMultFactor, &rescaleTimer);

	/* 5.	Allocate memory for the new combined function, calculate it & 
			insert it into the product, in case it isn't a constant. */

	myMalloc(probabList, sizeof(Probability) * sizeOfTable);

	if (numVars) {		/* in case the combined function isn't a constant. */
		if (elimVar->currNumFunctions > 1)
			rescaleTimer--;

		combinedFunc = insertFunctionToProduct(prod, sizeOfTable, probabList, logMultFactor,
											   TRUE, numVars, varArr, TRUE, rescaleTimer);

		/* 5.1.	Calculate the new combined function. */

		multiplyAllFuncsJointComplex(combinedFunc, elimVarsArr, numElimVars, funcArr, numFuncs);	

		/* 5.2.	If, according to the rescaleTimer, the combined function 
				needs to be rescaled, rescale it.*/

		if (combinedFunc->rescaleTimer <= 0) {
			(*isZeroLike) = scaleFunctionProbList(combinedFunc);
			combinedFunc->rescaleTimer = MAX_RESCALE_UNITS;
		}
	}
	else {				/* in case the combined function is a constant. */
		*tempRes = multiplyAllFuncsJoint(elimVarsArr, numElimVars, logMultFactor,
											   funcArr, numFuncs);
		myFree(probabList);
	}
		
	/* 6.	Remove the functions that have been multiplied from the variables that
			appear in them and from the product. */

	for (funcPtr = funcArr + numFuncs - 1, funcPtrEnd = funcArr; funcPtr >= funcPtrEnd; 
			funcPtr--) {

		func = *funcPtr;

		for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			removeFunctionFromVariable(*varPtr, func);
		}

		removeFunctionFromProduct(prod, func);
	}

	/* 7.	Add the new combined function to the variables that appear in it. */

	if (numVars) {
		for (varPtr = combinedFunc->variables, varPtrEnd = varPtr + combinedFunc->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			addFunctionToVariable(*varPtr, combinedFunc);
		}
	}

	/* 8.	Trim the variables that were eliminated from the product. */

	for (varPtr = elimVarsArr, varPtrEnd = varPtr + numElimVars; 
			varPtr < varPtrEnd; varPtr++) {
		removeVariableFromProduct(prod, (*varPtr));
	}

	/* 9.	Free the memory allocated dynamically. */

	myFree(varArr);
	myFree(funcArr);
	myFree(elimVarsArr);

	/* 10.	Return TRUE if the new function is a constant. */

	return ((numVars == 0) ? TRUE : FALSE);
}
#endif


/******************************************************************************
*
*       Function        sumVar
*       Parameters		prod (the product from which a variable needs to be 
*						eliminated), elimVar (pointer to the variable that needs
*						to be eliminated), tempRes (if the combined function is
*						a constant, holds the value of this constant),
*						isZeroLike (will be updated to TRUE if this elimination
*						resulted in a constant of ZERO probability).
*       Output          none
*       Description		Eliminates 'elimVar' from 'prod' by summing over the  
*						entries corrseponding to all possible values of 'elimVar'
*						for each possible assignment to the rest of the variables
*						of the function.
*       Comments		Returns TRUE if the combined function is a constant,
*						and FALSE otherwise.
*
******************************************************************************/
boolean sumVar(product *prod, variable *elimVar, Probability *tempRes, 
			   boolean *isZeroLike)
{
	int				numVars, jumpFactor, skipFactor, sucCnt, currVal;
	unsigned long	sizeOfTable;
	Probability		*probabList, *probPtrSrc, *probPtrDst, *probPtrDstEnd, funcVal;
	function		*combinedFunc = NULL, *func;


	*isZeroLike = FALSE;

	/* 1.	Update a pointer to the function from which the eliminated variable 
			needs to be summed out. */

	func = elimVar->functions[0];
	numVars = func->numVariables - 1;

	/* 2.	Calculate the size of the new function and allocate memory for it. */

	sizeOfTable = func->sizeOfTable / elimVar->numValues;
	myMalloc(probabList, sizeof(Probability) * sizeOfTable);

	if (numVars) {		/* In case the resulting function isn't a constant. */

		/* 2.1.	Initialize the jumpFactor and skipFactor of the function. */

		initJumpFactor(func, elimVar);
		jumpFactor = func->jumpFactor.inc;
		skipFactor = jumpFactor * (elimVar->numValues - 1);

		/* 2.2.	Initialize pointers to the beginning of the probability lists. */

		probPtrDst = probabList;
		probPtrDstEnd = probPtrDst + sizeOfTable;
		probPtrSrc = func->probabList;

		/* 2.3.	As long as we haven't updated all the entries in the probability table: */

		while (probPtrDst < probPtrDstEnd) {
			
			/* 2.3.1.	Iterate over all the successive entries in the source probability 
						table (these entries correspond to different assignments for the 
						variables in the destination table). For each assignment: */

			for (sucCnt = 0; sucCnt < jumpFactor; sucCnt++, 
					probPtrSrc++, probPtrDst++) {

				funcVal = 0.0;
				func->baseInd = 0;

				/* 2.3.1.1.	Iterate over all of the eliminated variable's values and 
							sum up the probabilities for each value. */	

				for (currVal = 0; currVal < elimVar->numValues; currVal++) {
					funcVal += 	*(probPtrSrc + func->baseInd);
					func->baseInd += jumpFactor;
				}

				/* 2.3.1.1.	Update the relevant entry in the destination probability table. */

				*probPtrDst = funcVal;
			}

			/* 2.3.2.	Advance the pointer to the source table to the next set of successive
						assignments. */

			probPtrSrc += skipFactor;
		}

		/* 2.4.	Remove the variable from the list of arguments of the function. */

		removeVariableFromFunction(func, elimVar);

		/* 2.5.	Free the old probability list (if necessary), and update the 
				function to point to the new probability list. */

		if (func->myProbabList) {
			myFree(func->probabList);
		}
		else {
			func->myProbabList = TRUE;
		}
		func->probabList = probabList;
		func->sizeOfTable = sizeOfTable;
	}
	else {				/* In case the combined function is a constant. */

		funcVal = 0.0;

		for (probPtrSrc = func->probabList, currVal = 0; currVal < elimVar->numValues; 
				currVal++, probPtrSrc++) {
			funcVal += 	*probPtrSrc;
		}

		*tempRes = log(funcVal) + func->logMultFactor;
		myFree(probabList);

		/* 2.1.	Remove the function from the product. */

		removeFunctionFromProduct(prod, func);
	}

	/* 3.	Trim the variable that was eliminated from the product. */

	removeVariableFromProduct(prod, elimVar);

	/* 4.	Return TRUE if the new function is a constant. */

	return ((numVars == 0) ? TRUE : FALSE);
}

