# dvir 2011/03
package AnalysisTemplateClass;

use strict;
use General;
use Configuration;
use ProcessingBackend;
use DatPedHandler;
use AnalysisSubmissionFile;
use POSIX qw(strftime); # for strftime

# the analysis classes should inherit from this class

# By default the universe is vanlia
sub getCondorJobUniverse() {
	return 'Vanilla';
}

# By default the priority is 5
sub getCondorJobPriority() {
	return 5;
}

sub getCondorJobRank() {
	return 'KFlops';
}

# By default the job should run 1 hour
sub getCondorJobSecondsToBeExecuted() {
	return 60*60;
}

# By default the job doesn't have special requirments
# 1000 is used for example in HMM
sub getCondorJobMemoryRequirments() {
	return 'Memory >= 1000 &&';
}

# does the executeable of the analysis may run on a 32 bit machine
# by deafult it can run.
sub canRunOn32BitMachine() {
	return 1;
}


# returning an array of files names
# those files will be copied to the machine on which the job will run
# it shouldn't contain the dat & ped files, which will be added by the calling method.
sub getCondorJobInputFiles() {
	&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisTemplateClass.pm',`pwd`,
					"getCondorJobInputFiles: this method should be run from sub class\n".
					"there must be input files to the analysis");
}

# returning an array of files which contain the results of the analysis, 
# those files should be copied to the web server
sub getFileListToBeCopiedToWebServer() {
	&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisTemplateClass.pm',`pwd`,
					"getFileListToBeCopiedToExecutionMachine: this method should be run from sub class\n".
					"there must be files to be copied to web server");
}

# returning an array of commands,
# which will be executed on the web server when the task is about to end
sub getCommandsToBeExecutedOnWebServer($) {
	my $class = $_[0]; # the name of this class
	my $analysisDirAtWebServer = $_[1];
	return ();
}

# the program code needed in order that the executeable will run
# the program code will be planted at the first line of the dat file
sub getDatFileProgramCode() {
	return 8;
}

# returning an array of the names of the output files of a single analysis
sub getJobOutputFiles() {
	&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisTemplateClass.pm',`pwd`,
					"getJobOutputFiles: this method should be run from sub class\n".
					"there must be output files for the analysis job");
}

# if a single job failes, than we will expect that the error will be in this file
# as default we return empty string to indicate that there isn't an error file
sub getJobErrorFile() {
	return '';
}

# whether this analysis supports a multi window run
sub doesAnalysisSupportsMultiWindowRun() {
	return 1;
}

# returning an AnalysisSubmissionFileData with the basic configurations
sub getAnalysisSubmissionFileData( $ ) {
	my $class = $_[0]; # the name of this class

	my $analysisSubmissionFileData = AnalysisSubmissionFileData->new(); # AnalysisSubmissionFileData is from AnalysisSubmissionFile		
	
	$analysisSubmissionFileData->jobUniverse ( $class->getCondorJobUniverse() );
	$analysisSubmissionFileData->jobPriority ( $class->getCondorJobPriority() );
	$analysisSubmissionFileData->jobRank ( $class->getCondorJobRank() );
	$analysisSubmissionFileData->jobMemoryRequirments ( $class->getCondorJobMemoryRequirments() );
	$analysisSubmissionFileData->canRunOn32BitMachine ( $class->canRunOn32BitMachine() );
	
	return $analysisSubmissionFileData;
}

# dvir 2011/08
# this method is called from TaskHandler.pl to inform that there is an error in the parameters.
# each analysis has its own way to handle this error, so this method should be implemented in sub class
sub reportOnUserErrorToWebServer() {
	my $class = $_[0]; # the name of this class
	my $errorText = $_[1];
	
	&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisTemplateClass.pm',`pwd`,
					"reportOnUserErrorToWebServer: this method should be run from sub class\n".
					"there must be files to be copied to web server");
}
# dvir 2011/08
# this method is called from "checkTasks.pl"
# and the method should report to the web server that the analysis has failed.
# for example : the type of failure is that the analysis is still marked as running (at the running folder)
# but doesn't exists at the condor.
# each analysis has its own way to handle error, so this method should be implemented in sub class
sub reportOnFailureToWebServer() {
	my $class = $_[0]; # the name of this class
	my $errorText = $_[1];
	
	&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisTemplateClass.pm',`pwd`,
					"reportOnFailureToWebServer: this method should be run from sub class\n".
					"there must be files to be copied to web server");
}

# dvir 2011/11/16
# this method is called from "checkTasks.pl"
# and the method should report to the web server that the analysis is currently running.
# each analysis may have its own way to report about running.
#
# this implementation works for lm_markers, haplotype, drawing, gl_auto.
sub reportOnRunningToWebServer() {
	my $class = $_[0]; # the name of this class
	
	# by default no reporting is done (cause drawing.pm doesn't need reporting)
	# but AnalysisWith_modefile has an implementation for that method.
}

# dvir 2011/11/16
# this method is called from postJob.pl with determine_job_status_of_leaf_job_with_one_exec
# parameters:
#	the folder name of a job
# this function should check if there was a timeout (by looking in the log).
# it is called before doesAnAcceptableErrorOcurredAtJob, so at doesAnAcceptableErrorOcurredAtJob we can check
# &JobStatus::doesJobHasA_TimedOut_Indication .. (see HMM.pm)
# this method is called with a full path of the jobDir and you cannot assume nothing about the cwd
sub does_A_TimedOutOcurredAtJob($) {
	my $class = $_[0]; # the name of this class
	my $jobDir = $_[1]; # full path or relative to the calling scope
	
	
	my $jobLogFileName = "$jobDir\/".$class->getLeafJob_CondorLogFileName();
	&General::makeSureFileExists($jobLogFileName);
		
	my $logFileText = `cat $jobLogFileName`;	
	if ($logFileText =~ /$Configuration::TIMEOUT_MESSAGE_FRAGMENT_AT_JOB_LOG_FILE/) {
		return 1;
	}
	
	# no timeout has ocurred
	return 0;
}

sub getLeafJob_CondorLogFileName() {
	my $class = $_[0]; # the name of this class
	
	return $AnalysisSubmissionFile::JOB_LOG_FILENAME;
}

# dvir 2011/11/16
# this method is called from postJob.pl with determine_job_status_of_leaf_job_with_one_exec
# parameters:
#	the folder name of a job
# this function should check if there was an acceptaable error.
# if the error is acceptable, than no mail should be sent to the developers.
# this method is called with a full path of the jobDir and you cannot assume nothing about the cwd
sub doesAnAcceptableErrorOcurredAtJob($ $) {
	my $class = $_[0]; # the name of this class
	my $jobDir = $_[1]; # full path or relative to the calling scope
	my $jobExitCode = $_[2]; # the exit code of the job
	
	# by default no acceptable error has ocurr
	return 0;
}

1;