# dvir, 2011-09-04
package AnalysisWith_modefile;

use strict;
use AnalysisTemplateClass;
use AnalysisSubmissionFile;
use DatPedHandler;
use DatWrapper;
use PedWrapper;
use ProcessingBackend;

# for mkpath
use File::Path;
use POSIX;

# this class inherits methods from AnalysisTemplateClass
our @ISA = ('AnalysisTemplateClass');

# returning an indication that says if we need to preform an estimation
sub shouldPreformEstimationForWindowRunTime() {
	# by default, no analysis should preform estimation
	return 0;
}

# by default, no uniuqe check is made
sub preformUniqueChecks( \% ) {
}

# creating the initial output files for each job
sub createInitialOutputFiles(\$ \$ \%) {
	my $class = $_[0]; # the name of this class
	
	my $datWrapperRef = $_[1];
	my $pedWrapperRef = $_[2];
	# each family will have a list of jobs, 
	# and each job will have a list of windows to be run in that job.
	# a window is a list of markers.
	my $windowsPerJobPerFamilyHashRef = $_[3];
	
	&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisWith_modefile.pm',`pwd`,
					"createInitialOutputFiles: this method should be run from sub class\n");	
}

# should create a partial html file (a file with html code)
# that will contain the results of the run of the current job.
sub createFinalOutputFileForCurrentJob() {
	&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisWith_modefile.pm',`pwd`,
					"createFinalOutputFileForCurrentJob: this method should be run from sub class\n");	
}

# should join all the partial output files to one big main html file
# the html output file will be used from the calling scope of this method, to be sent to the web server
sub createOutputFiles() {
	&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisWith_modefile.pm',`pwd`,
					"createOutputFiles: this method should be run from sub class\n");	
}


# returning a list of windows
# each window is a list of markers numbers.
sub getWindowsList( $ $ $ ) {
	my $class = $_[0];
	my $startAt = $_[1];
	my $endAt = $_[2];
	my $jump = $_[3];
	
	my @windowList = ();
	# the first marker of the next window, is one the last marker of our window
	for (my $currMarker = $startAt; $currMarker <= $endAt; $currMarker += $jump) {
		my $currWindowLastMarker = $currMarker + $jump - 1;
		if ($currWindowLastMarker > $endAt) {
			$currWindowLastMarker = $endAt;
		}
		# I'm using '[' and not '(' to create array of arrays
		push @windowList, [$currMarker..$currWindowLastMarker];
	}	
	return @windowList;
}

# a limit of the num of windows for the task
# in order not to make too large computations
# if returns undef, than there is no limit
sub maxNumOfWindowsAllowed() {
	my $class = $_[0]; # the name of this class
	return undef;
}

# a limit of the num of markers in window (window size) for the task
# in order not to make too large computations
# if returns undef, than there is no limit
sub maxNumOfMarkersAllowedInWindow() {
	my $class = $_[0]; # the name of this class
	return undef;
}

# dvir,
# creates the file $Configuration::TASK_MAIN_DAG_NAME for the task (for now haplotyping & morgan_lm_markers & gl_auto)
# uses modefile.txt as the analysis param file.
sub initDag( $ ) {
	my $class = $_[0]; # the name of this class
	
	my $binDir = $Configuration::configfile{ binDir };
	
	my $projectTaskDir = $class->getProjectDir();
	my $workDir = $class->getWorkDir();
	my $analysisTypeFileIndication = $class->getType();
	
	chdir $workDir || &General::reportOnErrorReleaseSemaphoresAndDie('AnalysisWith_modefile.pm',`pwd`, "initDag(): \"chdir $workDir\" failed");
			
	# reading the variables from the user
	my %variables = %{ &ProcessingBackend::initVariablesFromPrePoint( $projectTaskDir ) };
	
	# the dat & ped files which were received from web server
	my $datFileName = "$projectTaskDir\/$variables{ datafile }";
	my $pedFileName = "$projectTaskDir\/$variables{ pedfile }";
	
	&General::makeSureFileExists($datFileName);
	&General::makeSureFileExists($pedFileName);
	
	my $datWrapper = new DatWrapper($datFileName);
	my $numOfMarkersInDatFile = $datWrapper->getNumOfMarkers();
	my $pedWrapper = new PedWrapper($pedFileName, $numOfMarkersInDatFile);
	
	$datWrapper->setProgramCode($class->getDatFileProgramCode());
			
	my @families = $pedWrapper->getFamiliesNumbers();
	# creating a families numbers file
	&DatPedHandler::createFamiliesFileByArr(\@families);
	
	# creating a job dir in the $workDir for each family
	foreach my $currFamily (@families) {
		mkdir(&General::getFamilyRunDir($currFamily));
	}	
	
	$class->preformUniqueChecks(\%variables);
	
	# by default, the analysis is a single window run
	my $isAnalysisASingleWindowRun = 1;
	my $maxTimePerJob = $variables{max_time_per_job};
	my $timeToRunSingleWindow = $variables{run_time_of_window};
							
	# $maxTimePerJob can be also $SharedConsts::UNLIMITED_VALUE
	# so I check only that $timeToRunSingleWindow is numeric.
	# the folowing params aren't relevant at a case of estimation (but since no estimation is preformed now, than I (Dvir) allowed myself to check it here)
	if (! (defined $timeToRunSingleWindow &&
		   defined $maxTimePerJob &&
		   isdigit($timeToRunSingleWindow))) {
		&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisWith_modefile.pm',`pwd`,
			"initDag: missing parameters, or param run_time_of_window isn't integer\n".
			"run_time_of_window : $timeToRunSingleWindow\n".
			"max_time_per_job : $maxTimePerJob\n");
	}
				
	# startAt & $endAt are relevant only when performing multi window run.
	my $startAt = 1;
	if ( defined($variables{ from }) && ( $variables{ from } !~ /Unknown/ ) ) {
		$startAt = $variables{ from };
		if ($startAt < 1 || $startAt > $numOfMarkersInDatFile) {
			&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisWith_modefile.pm',`pwd`,
				"initDag: wrong \$variables{ from }: $startAt\n".
							   "\$numOfMarkersInDatFile: $numOfMarkersInDatFile");
		}
	}

	my $endAt = $numOfMarkersInDatFile;
	if ( defined($variables{ to }) && ( $variables{ to } !~ /Unknown/ ) ) {
		$endAt = $variables{ to };
		if ($endAt < $startAt || $endAt > $numOfMarkersInDatFile) {
			&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisWith_modefile.pm',`pwd`,
				"initDag: wrong \$variables{ to }: $endAt\n".
							   "\$startAt: $startAt\n".
							   "\$numOfMarkersInDatFile: $numOfMarkersInDatFile");
		}
	}
	
	# we will relate to a single window run as a multi window run.
	# each family will have a list of jobs, 
	# and each job will have a list of windows to be run in that job.
	# a window is a list of markers.
	my %windowsPerJobPerFamily = ();
	
	# $numOfMarkersInASingleComputation - the maximum num of markers in a single computation
	# dummy value 0. if it stays 0 in the check at the end of this function, than something is wrong,
	# cause it needs to be updated.
	my $numOfMarkersInASingleComputation = 0;
	
	# checking if this is a multi win analysis.
	# jump is another name for the window size
	# and the window size is how many markers will be computed at each sub job of the task
	# jump is relevant only when from & to parmeters are set.
	my $jump = $variables{ jump };
	if ( defined($jump) && ( $jump !~ /Unknown/) ) {			
		
		my $numOfMarkersInAnalysis = &DatPedHandler::getTotalNumOfMarkersByRange($startAt, $endAt);
		if ($jump < 1 || $jump > $numOfMarkersInAnalysis) {
			&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisWith_modefile.pm',`pwd`,
				"initDag: wrong \$jump: $jump\n".
							   "\$startAt: $startAt\n".
							   "\$endAt: $endAt\n".
							   "\$numOfMarkersInAnalysis: $numOfMarkersInAnalysis");
		}
		# checking that there isn't only one window after all
		elsif ($jump != $numOfMarkersInAnalysis) {
			
			# checking if this type of analysis supports multi window
			if (! $class->doesAnalysisSupportsMultiWindowRun()) {
				&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisWith_modefile.pm',`pwd`,
					"initDag: Analysis Doesn\'t support a multi window run\.");
			}
		
			$numOfMarkersInASingleComputation = $jump;
		
			# marking that this isn't a single window run
			$isAnalysisASingleWindowRun = 0;
			
			# marking that the task is multi window task
			open IND_FILE, ">$Configuration::MULTI_WINDOW_TASK_INDICATION_FILENAME";
			close IND_FILE;
			
			# checking if we need to preform estimation for the time that each window takes.
			# for now this part is not in used!!!!
			# it was in use when haplo lars ran as 1 exec and not parallel.
			if ($class->shouldPreformEstimationForWindowRunTime()) {
				
				die "Dvir say: you must delete this err message, if you are sure you want to preform estimation, at present no task type uses estimation";
				
				# here we should preform an estimation
				# the dat file will be created in cwd (the workDir)
				my $estimationDatFileName = &DatPedHandler::getDatByRangeFileName($datFileName);
				my @markersList = (1..$jump);
				
				# a one dat file for an estimation for each family
				open DATFILE_FH, ">$estimationDatFileName";
				$datWrapper->writeSubDatFileByListOfMarker(\*DATFILE_FH, \@markersList);
				close DATFILE_FH;
				
				# creating the estimation dir for each family and a ped file in each dir
				foreach my $currFamily (@families) {
					# creating the estimation dir for curr family
					my $familyRunDir = &General::getFamilyRunDir($currFamily);
					my $estimationDir = "$familyRunDir\/$Configuration::ESTIMATION_DIR_NAME";
					mkdir($estimationDir);
					
					# creating the estimation ped file for curr family
					my @familiesInDestPedFile = ($currFamily);
					open PEDFILE_FH, ">$estimationDir\/$Configuration::PED_FILE_NAME";
					$pedWrapper->writeSubPedFileByListOfMarker(\*PEDFILE_FH, \@markersList, \@familiesInDestPedFile, 1);
					close PEDFILE_FH;										
				}
				open DAG_FILE, ">$Configuration::TASK_MAIN_DAG_NAME";
				$class->createEstimationDag(\*DAG_FILE, $binDir, $workDir, \@families);				
				close DAG_FILE;
								
				#### editing $Configuration::CONDOR_ESTIMATION_JOB_SUBMIT_DEFINITIONS_FILENAME	
				my $jobsDatFileNameAndPath = "$workDir\/$estimationDatFileName";
				my @pmFilesInBinDirArr = &General::getArrayOf_pm_filesNamesInBinDir();
				my @uniqueInputFilesNames = $class->getCondorJobInputFiles();
				my @jobInputFilesList = (
					@pmFilesInBinDirArr,
					$jobsDatFileNameAndPath,
					$Configuration::PED_FILE_NAME,
					$analysisTypeFileIndication,
					@uniqueInputFilesNames);
				
				# the order of the arguments is important (it the order in which they will be written at the submission file)
				my @jobArguments = (
					$estimationDatFileName, 
					$Configuration::PED_FILE_NAME,
					1, # $shouldWeCreateEstimation
					$class->getCondorJobSecondsToBeExecuted());
					
				# AnalysisSubmissionFileData is from AnalysisSubmissionFile
				my $analysisSubmissionFileData = $class->getAnalysisSubmissionFileData();
				
				$analysisSubmissionFileData->executeableName ( "$binDir\/runAnalysis\.pl" );
				@{$analysisSubmissionFileData->executeableArgumentsList} = @jobArguments;
				@{$analysisSubmissionFileData->executeableInputFilesList} = @jobInputFilesList;
				$analysisSubmissionFileData->jobSecondsToBeExecuted ( $class->getCondorJobSecondsToBeExecuted() );
			
				# creating $Configuration::CONDOR_ESTIMATION_JOB_SUBMIT_DEFINITIONS_FILENAME at cwd
				&AnalysisSubmissionFile::createAnalysisSubmissionFileAtCwd($analysisSubmissionFileData, $Configuration::CONDOR_ESTIMATION_JOB_SUBMIT_DEFINITIONS_FILENAME);
				
				# we can return
				return undef;
			}
			# here we don't need to preform an estimation,
			# because the window run time is given as argument
			else {			
				# making sure that all requied parameters appears.
				# among them there should be the : run time for the window (and it is the reason why we havn't preform an estimation)
				my $maxJobsPerTask = $variables{max_jobs_per_task};				
				my $minTimePerJob = $variables{min_time_per_job};				
				
				if (! (defined $maxJobsPerTask &&
					   defined $minTimePerJob)) {
					&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisWith_modefile.pm',`pwd`,
						"initDag: missing parameters.\n".
						"maxJobsPerTask : $maxJobsPerTask\n".
						"min_time_per_job : $minTimePerJob\n");
				}
				
				# a window is a list of markers
				my @windowsLists = $class -> getWindowsList($startAt, $endAt, $jump);
				my $numOfWindows = scalar (@windowsLists);
				
				my $maxNumOfWindowsAllowed = $class->maxNumOfWindowsAllowed();
				if (defined $maxNumOfWindowsAllowed && $maxNumOfWindowsAllowed < $numOfWindows) {
					&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisWith_modefile.pm',`pwd`,
						"too many windows, beyond limit.\n".
						"numOfWindows : $numOfWindows\n".
						"maxNumOfWindowsAllowed : $maxNumOfWindowsAllowed\n");
				}
				
				(my $windowsPerJob, my $errorMsg) = &General::howManyWindowsInASingleJobVer2(
					$maxJobsPerTask,
					$maxTimePerJob,
					$minTimePerJob,
					$numOfWindows,
					$timeToRunSingleWindow);
							
				# returning the error message, in order to report it to the user.
				if (defined $errorMsg) {
					# reporting about the error to the user.
					$class->reportOnUserErrorToWebServer($errorMsg);
					
					return $errorMsg;
				}
				
				# for saftey we  doubled the time of running the entire job.
				if ($maxTimePerJob ne $SharedConsts::UNLIMITED_VALUE) {
					$maxTimePerJob = 2 * $maxTimePerJob;
				}
				else {
					$maxTimePerJob = 2 * $windowsPerJob * $timeToRunSingleWindow;
				}
				
				# all families has the same num of windows per job, 
				# cause the run time of a window is the same for all the families
				# and since that, all the families has the same num of jobs.
				my $numOfJobsPerFamily = &POSIX::ceil($numOfWindows/$windowsPerJob);
				
				# updating %windowsPerJobPerFamily
				foreach my $currFamily (@families) {					
					my $currWindowNumInCurrJob = 0;
					my $currJobNum = 0;
					for my $currWindow (@windowsLists) {
						$windowsPerJobPerFamily{$currFamily}[$currJobNum][$currWindowNumInCurrJob] = $currWindow;
						
						++$currWindowNumInCurrJob;
						# checking if we need to go to the next window
						if ($currWindowNumInCurrJob == $windowsPerJob) {
							$currWindowNumInCurrJob = 0;
							++$currJobNum;
						}
					}
				}
			}
		}
	}
	
	# we reach to here if we havn't decided to preform estimation.
	# and this is a multi win analysis with a declared run time for a window,
	# or this is a single win analysis.
	
	# checking if this is not a multi window run
	if ($isAnalysisASingleWindowRun) {

		# getting a sorted list of markers to use.
		#my @sortedMarkersList = &DatPedHandler::sortMarkersList($variables { usedMarkersId });
		#Omer 15/9/2011 - I restored the original sort, because the other one has bugs that prevented Morgan from running
		#Also, I checked times and the inner sort of Perl is much faster than this sort.
		#
		# Dvir: we are using the markers list even if the user has selected a range since it,
		# casue we expect that the range markers will be listed in @{$variables { usedMarkersId }}.
		my @sortedMarkersList = sort {$a <=> $b} @{$variables { usedMarkersId }};
		
		$numOfMarkersInASingleComputation = scalar (@sortedMarkersList);
		
		# updating %windowsPerJobPerFamily, each family will have only 1 job
		foreach my $currFamily (@families) {
			# first 0 is job no' 0
			# second 0 is that the window list of the job, contains only one window
			$windowsPerJobPerFamily{$currFamily}[0][0] = \@sortedMarkersList;
		}
		
		# for safety, we are doubling $variables{run_time_of_window}
		# the actual run time will be $variables{run_time_of_window}
		$maxTimePerJob = $variables{run_time_of_window} * 2;
		$timeToRunSingleWindow = $variables{run_time_of_window};
	}

	my $maxNumOfMarkersAllowedInWindow = $class->maxNumOfMarkersAllowedInWindow();
	if (defined $maxNumOfMarkersAllowedInWindow && $maxNumOfMarkersAllowedInWindow < $numOfMarkersInASingleComputation) {
		&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisWith_modefile.pm',`pwd`,
			"window size is beyond limit.\n".
			"numOfMarkersInASingleComputation : $numOfMarkersInASingleComputation\n".
			"maxNumOfMarkersAllowedInWindow : $maxNumOfMarkersAllowedInWindow\n");
	}
	
	# numOfDuplications is the num of duplications for each window
	# or the num of duplications for each 'multi window job'
	# by default, we will have only 1 duplication
	my $numOfDuplications = 1;
	if (defined $variables{num_dups_per_job}) {
		$numOfDuplications = $variables{num_dups_per_job};
	}
	
	# checking if we need to create the DAG here
	# will happen in cases that the analysis is :
	# not a multi window per job with estimation and the window size is not all the markers
	if (scalar(keys %windowsPerJobPerFamily) > 0) {
	
		# creating the dag
		$class->createDagForMultiWindowPerJobAnalysis(
			\$datWrapper,
			\$pedWrapper,
			$numOfDuplications,
			\%windowsPerJobPerFamily,
			$maxTimePerJob,
			$timeToRunSingleWindow,
			1); # 1 to mark that all families has the same multi dat file
		
		# creating the initial output files
		$class->createInitialOutputFiles(
			\$datWrapper,
			\$pedWrapper,
			\%windowsPerJobPerFamily);		
	}
	
	# in order to mark that no input error was detected
	return undef;
}

# creating an estimation dag for the multi window run
sub createEstimationDag($ $ $ \@) {
	my $class = $_[0]; # the name of this class
	my $dagFileHandle = $_[1];
	my $binDir = $_[2];
	my $workDir = $_[3];
	my $familiesArrRef = $_[4];
	
	my $jobsNames = '';	
	
	foreach my $currFamily (@$familiesArrRef) {
		my $currFamilyJobName = "Family_${currFamily}_ESTIMATION_JOB";
		$jobsNames .= " $currFamilyJobName ";
		
		my $currFamilyRunDir = &General::getFamilyEstimationDir($currFamily);
		
		print $dagFileHandle 
			# DIR is the place to which the output of the run will be placed, 
			# and from which it will try to locate the job submission file
			"JOB         $currFamilyJobName $workDir\/$Configuration::CONDOR_ESTIMATION_JOB_SUBMIT_DEFINITIONS_FILENAME DIR $currFamilyRunDir\n".
			"SCRIPT POST $currFamilyJobName $binDir\/postJob.pl \$RETURN $workDir 1\n\n";
	}	
			
	print $dagFileHandle
		"JOB         SIMPLE2 $Configuration::SIMPLE_SHORT_CONDOR_JOB\n".
		"SCRIPT POST SIMPLE2 $binDir\/postJob.pl\n".
		"\n".
		"PARENT      $jobsNames CHILD SIMPLE2\n";	
}

# dvir 2011/09/6
# create dag, dirs dat files & ped files for the multi window per job run for all the families.
# each family can have some jobs.
# each job can have some duplications.
sub createDagForMultiWindowPerJobAnalysis(\$ \$ $ \% $ $ ) {
	my $class = $_[0]; # the name of this class
	
	my $datWrapperRef = $_[1];
	my $pedWrapperRef = $_[2];
	# numOfDuplications is the num of duplications for each window
	# or the num of duplications for each 'multi window job'
	my $numOfDuplications = $_[3];
	# each family will have a list of jobs, 
	# and each job will have a list of windows to be run in that job.
	# a window is a list of markers.
	my $windowsPerJobPerFamilyHashRef = $_[4];
	my $maxTimePerJob = $_[5]; # in seconds
	
	# needed only when the runAnalysis method at the SuperlinkWrapper excepts to recive
	# a time limit for the run (=window)
	# for now this is only relevant at morgan
	# at other cases the value of this parameter can be any posibble value (even one parameter less)
	my $timeToRunSingleWindow = $_[6]; # in seconds
	# can we create only one dat file for all the families jobs, and use soft link to this dat file
	# when using estimation, than this should be 0, cause diffrenet families can have a diffrenet partition of windows to jobs.
	my $allFamiliesJobsDatFilesAreTheSame = $_[7];

	my $workDir = $class->getWorkDir();
	
	my @families = $$pedWrapperRef->getFamiliesNumbers();
	
	
	#### creating the job dirs & ped with no allels for each family
	
	# the value of this list isn't important, cause we are only writing the person data
	my @emptyMarkersList = (); 
	foreach my $currFamily (@families) {
		my $familyRunDir = &General::getFamilyRunDir($currFamily);
		
		# creating the pedfile with no allels at the family dir
		my @familiesInDestPedFile = ($currFamily);
		open PEDFILE_FH, ">$familyRunDir\/$Configuration::PED_WITH_NO_ALLELS_FILE_NAME";
		$$pedWrapperRef->writeSubPedFileByListOfMarker(\*PEDFILE_FH, \@emptyMarkersList, \@familiesInDestPedFile, 1, $PedWrapper::WRITE_ONLY_PERSON_DATA);
		close PEDFILE_FH;
		
		# passing over the familiy's jobs & creating them
		for (my $currJobNum=0;$currJobNum<scalar (@{$$windowsPerJobPerFamilyHashRef{$currFamily}}); ++$currJobNum) {
			my $jobDir = &General::getFamilyJobNumRunDir($currFamily, $currJobNum);
			mkdir $jobDir;
		}
	}

	#### creating the multi dat file for each job
	if ($allFamiliesJobsDatFilesAreTheSame) {
		my $datfilesDirName = "$workDir\/datfiles";
		mkdir $datfilesDirName;
			
		my $someFamilyNum = $families[0];
		# passing over the 'some familiy' jobs
		for (my $currJobNum=0;$currJobNum<scalar (@{$$windowsPerJobPerFamilyHashRef{$someFamilyNum}}); ++$currJobNum) {
								
			# creating the datfile
			my $currDatFileName = "$datfilesDirName\/job_$currJobNum.$Configuration::MULTI_DAT_FILE_NAME";
			open DATFILE_FH, ">$currDatFileName";
			
			# passing over the job's windows, and writing them to the multi dat.			
			
			foreach my $currWindow (@{$$windowsPerJobPerFamilyHashRef{$someFamilyNum}[$currJobNum]}) {				
				# $currWindow is a ref to an array with markers nums that belong to the current window
				$$datWrapperRef->writeSubDatFileByListOfMarker(\*DATFILE_FH, $currWindow);
				print DATFILE_FH $Configuration::MULTI_FILES_IN_SINGLE_FILE_SEPERATOR;
			}
			close DATFILE_FH;
			
			# creating a soft link to the dat files for all the families
			foreach my $currFamily (@families) {
				my $jobDir = &General::getFamilyJobNumRunDir($currFamily, $currJobNum);
				
				symlink $currDatFileName, "$jobDir\/$Configuration::MULTI_DAT_FILE_NAME";
			}
		}
	}
	# we need to create a unique multi dat file for each family jobs
	# cause job no' 0 at diffrenet families can be on diffrenet windows.
	else {
		foreach my $currFamily (@families) {
			
			# passing over the familiy's jobs
			for (my $currJobNum=0;$currJobNum<scalar (@{$$windowsPerJobPerFamilyHashRef{$currFamily}}); ++$currJobNum) {
				
				my $jobDir = &General::getFamilyJobNumRunDir($currFamily, $currJobNum);
				
				# creating the datfile
				open DATFILE_FH, ">$jobDir\/$Configuration::MULTI_DAT_FILE_NAME";
				
				# passing over the job's windows
				foreach my $currWindow (@{$$windowsPerJobPerFamilyHashRef{$currFamily}[$currJobNum]}) {
					# $currWindow is a ref to an array with markers nums that belong to the current window
					$$datWrapperRef->writeSubDatFileByListOfMarker(\*DATFILE_FH, $currWindow);
					print DATFILE_FH $Configuration::MULTI_FILES_IN_SINGLE_FILE_SEPERATOR;
				}			
				close DATFILE_FH;			
			}
		}
	}
	
	#### creating the multi ped file for each job
	foreach my $currFamily (@families) {
		
		my @familiesInDestPedFile = ($currFamily);
		
		# passing over the familiy's jobs
		for (my $currJobNum=0;$currJobNum<scalar (@{$$windowsPerJobPerFamilyHashRef{$currFamily}}); ++$currJobNum) {
			
			my $jobDir = &General::getFamilyJobNumRunDir($currFamily, $currJobNum);
			
			# creating the datfile
			open PEDFILE_FH, ">$jobDir\/$Configuration::MULTI_PED_FILE_NAME";
			
			# passing over the job's windows
			foreach my $currWindow (@{$$windowsPerJobPerFamilyHashRef{$currFamily}[$currJobNum]}) {
				# $currWindow is a ref to an array with markers nums that belong to the current window
				$$pedWrapperRef->writeSubPedFileByListOfMarker(\*PEDFILE_FH, $currWindow, \@familiesInDestPedFile, 1, $PedWrapper::WRITE_ONLY_ALLELS);
				print PEDFILE_FH $Configuration::MULTI_FILES_IN_SINGLE_FILE_SEPERATOR;
			}			
			close PEDFILE_FH;
		}
	}	

	my $dagFileName = $Configuration::TASK_MAIN_DAG_NAME;
	#### backuping the previous output&error files of the previous dag:
	# the previous dag exists when this function is called after we preformed estimation
	# and after estimation, than $allFamiliesJobsDatFilesAreTheSame is 0
	if (! $allFamiliesJobsDatFilesAreTheSame) {
		# sanity check that the log of the prev dag exists at the workDir
		if (! -e $Configuration::CONDOR_DAG_SUBMISSION_OF_MAIN_TASK_OUTPUT_FILENAME) {
			&General::reportOnErrorReleaseSemaphoresAndDie('AnalysisWith_modefile.pm',`pwd`,
				"initDag: createDagForMultiWindowPerJobAnalysis.\n".
				"$Configuration::CONDOR_DAG_SUBMISSION_OF_MAIN_TASK_OUTPUT_FILENAME doesn't exists\n");
		}
		
		`cp $Configuration::CONDOR_DAG_SUBMISSION_OF_MAIN_TASK_OUTPUT_FILENAME ${Configuration::CONDOR_DAG_SUBMISSION_OF_MAIN_TASK_OUTPUT_FILENAME}.old`;
		`cp $Configuration::CONDOR_DAG_SUBMISSION_OF_MAIN_TASK_ERROR_FILENAME ${Configuration::CONDOR_DAG_SUBMISSION_OF_MAIN_TASK_ERROR_FILENAME}.old`;	
		$dagFileName = $Configuration::TASK_SECOND_DAG_FILENAME;
	}
	
	#### Creating the DAG	
	open DAG_FILE, ">$dagFileName";
	$class->createDagForMultiWindowPerJobAnalysis_INNER(\*DAG_FILE, \@families, $numOfDuplications, $windowsPerJobPerFamilyHashRef, $maxTimePerJob, $timeToRunSingleWindow);
	close DAG_FILE;
}

# will be called from createDagForMultiWindowPerJobAnalysis
# here we will create the dag (the directories & dat & ped files were allready created at createDagForMultiWindowPerJobAnalysis)
sub createDagForMultiWindowPerJobAnalysis_INNER(\* \@ $ \% $ $ ) {
	my $class = $_[0]; # the name of this class
	
	my $dagFileHandleRef = $_[1];
	
	my $familiesArrRef = $_[2];
	# numOfDuplications is the num of duplications for each window
	# or the num of duplications for each 'multi window job'
	my $numOfDuplications = $_[3];
	# each family will have a list of jobs, 
	# and each job will have a list of windows to be run in that job.
	# a window is a list of markers.
	my $windowsPerJobPerFamilyHashRef = $_[4];
	my $maxTimePerJob = $_[5]; # in seconds
	
	# needed only when the runAnalysis method at the SuperlinkWrapper excepts to recive
	# a time limit for the run (=window)
	# for now this is only relevant at morgan
	# at other cases the value of this parameter can be any posibble value (even one parameter less)
	my $timeToRunSingleWindow = $_[6]; # in seconds	

	# the paths are relative to the place from which the DAG will be submitted
	# this place is the workdir which is now the (cwd)
	my $binDir = $Configuration::configfile{ binDir };
	
	my $workDir = $class->getWorkDir();
	
	my $condorJobParameterName_FamDir = 'FAM_DIR';
	my $condorJobParameterName_FamDir_AsParameter = '$('.$condorJobParameterName_FamDir.')';
	my $condorJobParameterName_FamJobDir = 'JOB_DIR';
	my $condorJobParameterName_FamJobDir_AsParameter = '$('.$condorJobParameterName_FamJobDir.')';
	my $jobsNames = '';
	
	
	foreach my $currFamily (@$familiesArrRef) {
		my $familyRunDir = &General::getFamilyRunDir($currFamily);
		
		# passing over the familiy's jobs
		for (my $currJobNum=0;$currJobNum<scalar (@{$$windowsPerJobPerFamilyHashRef{$currFamily}}); ++$currJobNum) {
			
			my $jobDir = &General::getFamilyJobNumRunDir($currFamily, $currJobNum);
			
			# passing over the duplications
			my $currDuplicationsNames = '';
			for (my $currDuplicateNum=1;$currDuplicateNum<=$numOfDuplications;++$currDuplicateNum) {
				my $familyJobDuplicateRunDir = &General::getFamilyJobNumDuplicateNumRunDir($currFamily, $currJobNum, $currDuplicateNum);
				
				# creating the duplicate dir
				mkdir $familyJobDuplicateRunDir;
				
				my $currFamilyJobName = "Family_${currFamily}_Job_${currJobNum}_Duplicate_${currDuplicateNum}";
				$currDuplicationsNames .= " $currFamilyJobName ";
				
				print $dagFileHandleRef 
					# DIR is the place to which the output of the run will be placed, 
					# and from which it will try to locate the job submission file (if that file is without a absulte path)
					#
					# $condorJobParameterName_FamDir is used for the files to be copied to execution machine.
					"JOB         $currFamilyJobName $workDir\/$Configuration::CONDOR_JOB_SUBMIT_DEFINITIONS_MULTI_WINDOW_PER_JOB_FILENAME DIR $familyJobDuplicateRunDir\n".
					"VARS        $currFamilyJobName $condorJobParameterName_FamDir=\"$workDir\/$familyRunDir\" $condorJobParameterName_FamJobDir=\"$workDir\/$jobDir\" \n".
					"SCRIPT POST $currFamilyJobName $binDir\/postJob.pl $workDir determine_job_status_of_leaf_job_with_possible_more_than_one_exec \$RETURN\n\n";
			}
			
			# a job that will run after all duplications of the current job
			my $currFamilyJobPostDuplicationsJobName = "Family_${currFamily}_Job_${currJobNum}_Post_All_Duplications";
			$jobsNames .= " $currFamilyJobPostDuplicationsJobName ";
			
			print $dagFileHandleRef
				# the logs of this job should appear at $jobDir
				"JOB         $currFamilyJobPostDuplicationsJobName $Configuration::SIMPLE_SHORT_CONDOR_JOB DIR $jobDir\n".
				"SCRIPT POST $currFamilyJobPostDuplicationsJobName $binDir\/postJob.pl $workDir determine_job_status_of_an_inner_job____and____create_html\n\n".
				"PARENT  $currDuplicationsNames  CHILD  $currFamilyJobPostDuplicationsJobName\n\n\n";
		}
	}	

	# the post script of LAST_JOB will preform:
	# 1. compute the job statues for the families folders
	# 2. compute the job status of the task by its families folders
	# 3. sends an indication ($Configuration::HAP_INDICATION_FILE_FOR_ANALYSIS_END) to hap that the task has ended.

	# the dir of the job will be the work dir
	print $dagFileHandleRef
		"JOB         LAST_JOB $Configuration::SIMPLE_SHORT_CONDOR_JOB\n".
		"SCRIPT POST LAST_JOB $binDir\/postJob.pl $workDir determine_job_status_of_families_dirs_by_their_sub_jobs____and____send_indication_to_web_server_to_mark_that_task_is_done____and____determine_job_status_of_the_task_by_families_dirs\n\n".
		"PARENT  $jobsNames CHILD LAST_JOB\n";
		
	#### editing $Configuration::CONDOR_JOB_SUBMIT_DEFINITIONS_FILENAME
	
	# AnalysisSubmissionFileData is from AnalysisSubmissionFile
	my $analysisSubmissionFileData = $class->getAnalysisSubmissionFileData();
	# the actual time for a job will be twice the max time.
	# it's O.K because we each window won't run more than x seconds, so the enitire job
	# won't run more than : numOfWindows * x seconds
	
	my @pmFilesInBinDirArr = &General::getArrayOf_pm_filesNamesInBinDir();
	my @uniqueInputFilesNames = $class->getCondorJobInputFiles();
	my $analysisTypeFileIndication = $class->getType();
	my @jobInputFilesList = (
		@pmFilesInBinDirArr,
		$analysisTypeFileIndication,
		"$condorJobParameterName_FamJobDir_AsParameter\/$Configuration::MULTI_DAT_FILE_NAME",
		"$condorJobParameterName_FamJobDir_AsParameter\/$Configuration::MULTI_PED_FILE_NAME",
		"$condorJobParameterName_FamDir_AsParameter\/$Configuration::PED_WITH_NO_ALLELS_FILE_NAME",
		@uniqueInputFilesNames);
	
	my @jobArguments = ($timeToRunSingleWindow);

	$analysisSubmissionFileData->executeableName ( "$binDir\/runAnalysisMultiWindow\.pl" );
	@{$analysisSubmissionFileData->executeableArgumentsList} = @jobArguments;
	@{$analysisSubmissionFileData->executeableInputFilesList} = @jobInputFilesList;
	$analysisSubmissionFileData->jobSecondsToBeExecuted ( $maxTimePerJob );
	
	# the creating $Configuration::CONDOR_JOB_SUBMIT_DEFINITIONS_MULTI_WINDOW_PER_JOB_FILENAME at cwd
	&AnalysisSubmissionFile::createAnalysisSubmissionFileAtCwd($analysisSubmissionFileData, $Configuration::CONDOR_JOB_SUBMIT_DEFINITIONS_MULTI_WINDOW_PER_JOB_FILENAME);
}	


# dvir 2011/11/17
# creating a html report that will visible to the user
sub createHtmlWithAMessageThatWillBeSentToWebServer() {
	my $class = $_[0]; # the name of this class
	my $headline = $_[1];
	my $text = $_[2];	
	my $shouldMarkToWebServerThatTaskIsDone = $_[3];
	my $reportHtmlPage_FolderNameAtWorkDir = $_[4]; # a relative folder name to the work dir that a copy of the page that
							   				       # is sent to hap will be also save at.
	
	my $projectTaskDir = $class->getProjectDir();
	my $workDir = $class->getWorkDir();

	# checking that the run report dir exists
	my $reportDir = "$workDir\/$reportHtmlPage_FolderNameAtWorkDir";
	if (! -d $reportDir) {
		# creating the run report dir
		mkdir $reportDir || &General::reportOnErrorReleaseSemaphoresAndDie('AnalysisWith_modefile.pm',`pwd`,"createHtmlWithAMessageThatWillBeSentToWebServer: couldn't create $reportDir");
	}
	
	&ProcessingBackend::setLogFiles("$projectTaskDir\/"); # needed in order that the initVariablesFromPrePoint() will work at debug mode
	my %variables = %{ &ProcessingBackend::initVariablesFromPrePoint( $projectTaskDir ) };

	my $projectName = $variables{PROJ};
	my $programName = $variables{program};
	my $dirAtWebServer = $variables{newDir};

	my $htmlTemplate = `cat $Configuration::REPORT_HTML_TEMPLATE_FILENAME`;
	$htmlTemplate =~ s/_PROJECT_NAME_/$projectName/g;
	$htmlTemplate =~ s/_PROGRAM_/$programName/g;
	$htmlTemplate =~ s/_TEXT_/$text/g;
	$htmlTemplate =~ s/_HEADLINE_/$headline/g;	
	
	
	my $reportHtmlFileName = "$reportDir\/$Configuration::REPORT_HTML_FILENAME";
	open REPORT_HTML, ">$reportHtmlFileName";
	print REPORT_HTML $htmlTemplate;
	close REPORT_HTML;	
	
	# when the web server sees the file $errorHtmlFileName at the $dirAtWebServer,
	# then it marks that the analysis has failed.
	&General::sendFilesToWebServer([$reportHtmlFileName], $dirAtWebServer);
	
	if ($shouldMarkToWebServerThatTaskIsDone) {
		# we are creating a $Configuration::HAP_INDICATION_FILE_FOR_ANALYSIS_END at the $dirAtWebServer, in order 
		# to mark the expirement as finished
		# (the expirement is marked as finished when all sub analyses have this file)
		&General::executeSSHCommandsOnWebServer(["touch $dirAtWebServer\/$Configuration::HAP_INDICATION_FILE_FOR_ANALYSIS_END"]);
	}
}

# dvir 2011/08 2011/11
# this method is called from TaskHandler.pl to inform that there is an error in the parameters.
sub reportOnUserErrorToWebServer() {
	my $class = $_[0]; # the name of this class
	my $userErrorText = $_[1];
	
	$class->createHtmlWithAMessageThatWillBeSentToWebServer('analysis input error.',$userErrorText,1,'.');
}

# dvir 2011/08
# this method is called from "checkTasks.pl"
# and the method should report to the web server that the analysis has failed.
# the type of failure is that the analysis is still marked as running (at the running folder)
# but doesn't exists at the condor.
sub reportOnFailureToWebServer() {
	my $class = $_[0]; # the name of this class
	my $errorText = $_[1];
	my $shouldNotContatSysAdmin = $_[2];
	
	my $htmlErrorText = '';
	if (not defined $shouldNotContatSysAdmin) {
		$htmlErrorText .= '<h3>please contact the system administrator</h3></br>';
	}
	$htmlErrorText .= $errorText;

	$class->createHtmlWithAMessageThatWillBeSentToWebServer('analysis failed',$htmlErrorText,1,'.');
}

# dvir 2011/11/16
# this method is called from "checkTasks.pl"
# and the method should report to the web server that the analysis is currently running.
# each analysis may have its own way to report about running.
#
# this implementation works for lm_markers, haplotype, drawing, gl_auto.
sub reportOnRunningToWebServer() {
	my $class = $_[0]; # the name of this class
	
	my $workDir = $class->getWorkDir();
	
	# cause if there is a more meaningful report, 
	# so in order not to override it at hap,
	# than we won't create a simple report with less information report.
	if ($class->doesAReportExists()) {
		return;
	}
	
	my $now_string = strftime "%a %b %e %H:%M:%S %Y", localtime;
	my $text = '<h2>'.$now_string.'</h2>';
		
	my $reportTextFileName = "$workDir\/$Configuration::REPORT_TEXT_FILENAME";
	if (-e $reportTextFileName) {
		$text .= `cat $reportTextFileName`;
	}	
	
	# the will be saved locally at a dir other than cwd (workDir) 
	# in order to avoid overwritting with a may be created report file after a failure
	$class->createHtmlWithAMessageThatWillBeSentToWebServer('analysis running',$text,0,$Configuration::RUN_REPORT_HTML_FOLDERNAME);
}

sub setReportText($) {
	my $class = $_[0]; # the name of this class
	my $text = $_[1];
	
	my $workDir = $class->getWorkDir();
	my $reportTextFileName = "$workDir\/$Configuration::REPORT_TEXT_FILENAME";
	open REP_TEXT, ">$reportTextFileName";
	print REP_TEXT $text;
	close REP_TEXT;
}

# whether a $Configuration::REPORT_HTML_FILENAME exists in the task workDir
# such a file may contain meaningful information.
sub doesAReportExists() {
	my $class = $_[0]; # the name of this class

	my $workDir = $class->getWorkDir();
	
	if (-e "$workDir\/$Configuration::REPORT_HTML_FILENAME") {
		return 1;
	}
	return 0;
}

1;
