package General;

use strict;
use Configuration qw(%configfile);
use Environment;

use Fcntl ':flock'; # import LOCK_* constants
use POSIX qw(strftime); # in order to track the locks time
use Cwd;
use FindBin();
use lib "$FindBin::Bin";
use Tie::File;
use IPC::SysV qw(IPC_PRIVATE S_IRWXU IPC_CREAT);
use SharedConsts;
use JobStatus;

my $COMMANDS_SEPERATOR		= ';';
my $USER_COMPUTER_SEPERATOR = '@';
my $NEW_LINE				= "\n";
my $FILES_LIST_FILE_NAME	= 'fileList.txt';

# SSH_COMMAND is like 'ssh options user_1@comp_a '
my $SSH_COMMAND = "ssh $configfile{ssh_options} $configfile{ssh_user}$USER_COMPUTER_SEPERATOR$configfile{ssh_remote} ";
my $SCP_COMMAND_PREFIX = 'scp ';
my $SCP_COMMAND_SUFFIX = " $configfile{ssh_user}$USER_COMPUTER_SEPERATOR$configfile{ssh_remote}:";

# when the session uses semphore, than the semaphore will be added as key to the hash: %semphoresInUsedBySession
# when the session releases a semphore, than the semaphore will be removed from the hash: %semphoresInUsedBySession
my %semphoresInUsedBySession = ();

#		trim
#	Parameters:
#		$_[0] - string to trim
#	return:
#		Trim function to remove leading whitespace
sub trim( $ ) {
	my $string = $_[0];
	$string =~ s/^\s+//;
	$string =~ s/\s+$//;
	return $string;
}

# removing the path to a file, and returning only the filename
sub removePath( $ ) {
	my $givenFileName = $_[0];
	$givenFileName =~ s/^.*\///;
	return $givenFileName;
}

sub getTaskDirNameByTaskName ( $ ) {
	my $projectName = $_[0];
	return "$Configuration::TASKS_DIR\/$projectName";
}

#	dvir 2011/03/08
sub getTaskNameAndTaskDirFromTaskWorkDir( $ ) {

	my $workDirAbsolutePath = $_[0];
	
	my @workDirAbsolutePathArr = split '/', $workDirAbsolutePath;
	
	# the second cell from the end is the 
	my $taskName = $workDirAbsolutePathArr[$#workDirAbsolutePathArr - 1];
	my $taskDir = join '/', @workDirAbsolutePathArr[0..($#workDirAbsolutePathArr - 1)];
	
	return ($taskName, $taskDir);
}

#	dvir 2010/09/05
#	execute all the given commands on the web server
#	paramters : 
#		$commandsArrRef - each cell is a command to be executed on the web server
sub executeSSHCommandsOnWebServer( \@) {
	my $commandsArrRef = $_[0];

	if ( scalar(@$commandsArrRef) == 0 ) {
		warn ('executeSSHCommandsOnWebServer, commands array is empty');
		return;
	}
	else {
		my $commandsStr = join ($COMMANDS_SEPERATOR, @$commandsArrRef);
		
		my $result = &General::perfomrWithExponentialBackoff("$SSH_COMMAND '$commandsStr'");
		if ( $result != 0 ) {
			&reportOnErrorReleaseSemaphoresAndDie('General.pm','',"failed: executeSSHCommandsOnWebServer, $SSH_COMMAND.'$commandsStr'");
		}
	}
}

#	dvir 2010/11/23
#	execute the given command on the web server and returnes the output of the command
#	paramters : 
#		$commandStr - the command
#		$commandOutputRef - reference to an array with the output lines of the command
sub executeSSHCommandOnWebServerAndCaptureOutput($ \@) {
	my $commandStr = $_[0];
	my $commandOutputRef = $_[1];
	@$commandOutputRef = `$SSH_COMMAND '$commandStr'`;
}

#	dvir 2010/09/26
#	sends all the given files to the web server
#	paramters : 
#		$filesArrRef - each cell is a file name as an absoulte path or releative path to the current path
#						a cell may also be some file names, when the filenames are seperated by spaces
#		$destDirInWebServer - should be in absulote path
sub sendFilesToWebServer( \@) {
	my $fileNamesArrRef = $_[0];
	my $destDirInWebServer = $_[1];

	if ( scalar(@$fileNamesArrRef) == 0 ) {
		warn ('sendFilesToWebServer, filenames array is empty');
		return;
	}
	else {
		my $fileNamesStr = join (' ', @$fileNamesArrRef);
		
		my $result = &General::perfomrWithExponentialBackoff("$SCP_COMMAND_PREFIX $fileNamesStr $SCP_COMMAND_SUFFIX$destDirInWebServer");
		if ( $result != 0 ) {
			&reportOnErrorReleaseSemaphoresAndDie('General.pm','',"failed: sendFilesToWebServer, $SCP_COMMAND_PREFIX $fileNamesStr $SCP_COMMAND_SUFFIX$destDirInWebServer");
		}
	}
}

#	dvir 2010/09/05
#	add all the given files names, each in seperate line, to FILES_LIST_FILE_NAME.
# 	only the files that are in the FILES_LIST_FILE_NAME, will be sent to the web server.
#	should be called from the post scripts
#	paramters : 
#		$filesNamesArrRef - each cell is a file name to be sent to the web server
sub addFilesNamesToSendListToWebServer( \@) {
	my $filesNamesArrRef = $_[0];

	if ( scalar(@$filesNamesArrRef) == 0 ) {
		warn ('addFilesNamesToSendListToWebServer, file names array is empty');
	}
	else {
		my $filesNamesStr = join ($NEW_LINE, @$filesNamesArrRef);
		open FILE,'>>',$FILES_LIST_FILE_NAME || die;
		print FILE $filesNamesStr.$NEW_LINE;
		close FILE;
	}
}

#	dvir 2010/09/06
#	copy the given files from the bin (.pl & .pm) folder to the dest folder.
sub copySrcFilesToFolder(\@ $) {
	&copyFilesToFolder($configfile{binDir}, $_[0], $_[1]);
}

#	dvir 2010/09/06
#	copy the given files from the src folder to the dest folder.
#	should be called from the init scripts
#	parameters : 
#		$srcFolder - the location of the files to be copied
#		$destFolder - the destinations of the files
#		$filesToBeCopiedArrRef - array of the files names, can be relative to $srcFolder
#		$srcFolder & $destFolder should be absulote or relative to the current path
sub copyFilesToFolder($ \@ $) {
	my $srcFolder = $_[0];
	my $filesToBeCopiedArrRef = $_[1];
	my $destFolder = $_[2];
	
	if ( scalar(@$filesToBeCopiedArrRef) == 0 ) {
		warn ('copyFilesToFolder, files array is empty');
	}
	else {
		my $srcFolderHelp = " $srcFolder\/";
		my $filesNamesStr = join ($srcFolderHelp, @$filesToBeCopiedArrRef);

		system("cp $srcFolderHelp$filesNamesStr $destFolder");
	}
}

#	dvir 2010/09/21
#	converts between seconds since 01-01-1970 and a formated date string such as "Tue Sep 21 17:46:21 UTC 2010"
#	parameters : 
#		$seconds - num of seconds since 1970-01-01
sub secondsToFormatedDate($) {
	my $seconds = $_[0];
	return `echo -n \`date -u --date="1970-01-01 $seconds sec GMT"\``;
}

#	dvir 2010/09/30
# returning the current date
sub getCurrentDate() {
	return strftime "%a %b %e %H:%M:%S %Y", localtime;
}

#	dvir 2010/09/12
#	sends a message to the mail address defined here : $configfile{superlinkDeveloperAddress}
#	parameters : 
#		$subject - the subject of the message
#		$message - the message body of the message
sub sendMailToSuperlinkDeveloper($ $) {
	my $subject = $_[0];
	my $message = $_[1];
	
	`echo "$message" | mail -s "$subject" $SharedConsts::PROGRAMMERS_EMAIL_ADDRESS`;
}

#	dvir 2010/09/12
#	sends a message to the mail address defined here : $configfile{superlinkDeveloperAddress}
#	abount an error that occured in the system
#	parameters : 
#		$moudle - the module in which the error occured
#		$task - the task in which the error occured
#		$errorDetails - the details of the error
sub reportOnError($ $ $) {
	my $moudle = $_[0];
	my $task = $_[1];
	my $errorDetails = $_[2];
	
	if ($task eq "") {
		$task = `pwd`;
	}
	my $messageSubject = "system error. Environment:$Environment::ENVIRONMENT_NAME, Moudle: $moudle, Task: $task";

	my $processStartTime = $^T;	
	my $currentTimeInSeconds = `date +%s`;
	my $timeOfExecutionInSeconds = $currentTimeInSeconds - $processStartTime;
	my $timeOfExecutionInMinutes = $timeOfExecutionInSeconds / 60;
	my $timeOfExecutionInHours = $timeOfExecutionInMinutes / 60;
	my $messageBody = 
		"Environment: $Environment::ENVIRONMENT_NAME\n".
		"Task: $task\n".
		"Moudle: $moudle\n".
		"Error Details: $errorDetails\n\n\n".
		
		"Script name: $0\n".
		"Arguments: @ARGV\n".
		"Command: perl $0 @ARGV\n\n\n".
#		"Command at ps: `ps -p $$`".

		"Cwd: `pwd`\n\n".
		"CPU Process number:$$\n\n".
		
		"Semaphores currently used by process: ".join(' ',keys %semphoresInUsedBySession)."\n\n".
		
		"Script Start Time: ".&secondsToFormatedDate($^T).",   (in seconds since 1970: $^T)\n".
		"Error Time: `date`,   (in seconds since 1970: `date +%s`)\n".
		"Time of execution: \n".
			"\tIn Hours: $timeOfExecutionInHours\n".
			"\tIn Minutes:$timeOfExecutionInMinutes\n".
			"\tIn Seconds:$timeOfExecutionInSeconds\n";

	&sendMailToSuperlinkDeveloper($messageSubject, $messageBody);
}

sub reportOnErrorReleaseSemaphoresAndDie($ $ $) {
	my $moudle = $_[0];
	my $task = $_[1];
	my $errorDetails = $_[2];	
	
	&reportOnError($moudle, $task, $errorDetails);
	
	&releaseAllSemaphoresUsedBySession();
	
	die $errorDetails;
}
#dvir 2011/12/04 - no more use in semaphores
# # catching the semaphore with the given id
# sub catchSemaphore($) {
	# my $semid = $_[0];
	
	# # checking if the session allready has the semaphore
	# if (exists $semphoresInUsedBySession{$semid}) {
		# # reporting on an error, without exiting
		# # dvir: I havn't prefrom exit, to see first if there is a place in the system when the session asks for the same 
		# # semphore twice
		# &reportOnError("General.pm" ,"","The session allready has the semaphore : $semid");
	# }
	
	# semop($semid, pack("s3", 0, -1, 0)) || die "catchSemaphore: $semid, $!\n";
	
	# # marking that the session has the semaphore
	# $semphoresInUsedBySession{$semid} = 1;
# }

# # releasing the semaphore with the given id
# sub releaseSemaphore($) {
	# my $semid = $_[0];
	# semop($semid, pack("s3", 0, +1, 0)) || die "releaseSemaphore: $semid, $!\n";
	# delete $semphoresInUsedBySession{$semid};
# }

# dvir
# releaseing all the semphores used by the session (session = process)
sub releaseAllSemaphoresUsedBySession() {
	foreach my $currSemaphoreId ( keys %semphoresInUsedBySession ) {
		&releaseSemaphore($currSemaphoreId);
	}
}

# dvir
# if the file doesn't exists in the cwd we will preform die
sub makeSureFileExists($) {
	my $fileName = $_[0];
	
	# checking that the files exists
	if (! -e $fileName) {
		my $cwd = `pwd`;
		die "General.pm, makeSureFileExists() : File $fileName doesn't exists in cwd: $cwd";
	}
}

# dvir 2010/12/30
# if the directory doesn't exists in the cwd we will preform die
sub makeSureDirExists($) {
	my $dirName = $_[0];
	
	# checking that the files exists
	if (! -d $dirName) {
		my $cwd = `pwd`;
		die "Directory $dirName doesn't exists in cwd: $cwd";
	}
}

# the var should be a number like 38 and not like -38, +38
sub makeSureVarIsNumeric($) {
	my $var = $_[0];
	
	# checking if $string is not a number
	if ($var !~ /^\d+$/) {
		die "$var is not numeric";
	}
}

# num is poisitive like 38.45 and not like -38.45, +38.45
sub getNumLowerValue($) {
	my $num = $_[0];
	
	makeSureVarIsNumeric($num);
	
	return int($num);
}

# num is poisitive like 38.45 and not like -38.45, +38.45
sub getNumUpperValue($) {
	my $num = $_[0];
	
	makeSureVarIsNumeric($num);
	
	return int($num) + 1;
}

sub saveToFileCurrentCommandAndCWD($) {
	my $fileName = $_[0];
	my $cwd = `pwd`;
	`echo "$0 @ARGV\nCWD: $cwd" > $fileName`;
}

sub getFamilyRunDir($) {
	my $family = $_[0];
	return "$family$Configuration::FAMILY_JOB_DIR_SUFFIX";
}

# dvir 2011/03/07
sub getFamilyEstimationDir($) {
	my $family = $_[0];
	return "$family$Configuration::FAMILY_JOB_DIR_SUFFIX\/$Configuration::ESTIMATION_DIR_NAME";
}

sub getFamilyJobNumRunDir($ $) {
	my $family = $_[0];
	my $jobNum = $_[1];
	return &getFamilyRunDir($family)."\/$jobNum$Configuration::JOB_DIR_SUFFIX";
}

# dvir 2011/08/10, used in morgan lm markers
sub getFamilyJobNumDuplicateNumRunDir($ $) {
	my $family = $_[0];
	my $jobNum = $_[1];
	my $duplicateNum = $_[2];
	return &getFamilyJobNumRunDir($family,$jobNum).'/'.&getDuplicateDir($duplicateNum);
}

# dvir 2011/09/07
# realtive to the job dir
sub getDuplicateDir( $ ) {
	my $duplicateNum = $_[0];
	return "$duplicateNum$Configuration::DUPLICATE_DIR_SUFFIX";
}

#	dvir 2010/11/08
# 	sends a mail if the excuteable failed (the exit code is not 0), according to the params
#	specifiying if the output files exsits at the output dir.
#	
#	parameters : 
#		$executionExitCode - the exit code of the execution
#		$outputFilesNamesArrRef - a pointer to an array of all the files names that need to appear (in CWD) when
#								  the execution ends with success
#		$errorFileName - the name of a file that might contain the execution error in its last ~10 lines.
#						 this parameter can be '', and if so, than it is not treated.
sub checkExecutionExitCodeAndExistenceOfOutputFiles($ $ $ \@ $) {
	my $moudle = $_[0];
	my $task = $_[1]; # may be ''
	my $executionExitCode = $_[2];
	my $outputFilesNamesArrRef = $_[3];
	my $errorFileName = $_[4]; # may be ''
	
	my $someFilesAreMissingInd = 0;
	
	if ( scalar(@$outputFilesNamesArrRef) == 0 ) {
		&reportOnErrorReleaseSemaphoresAndDie($moudle, $task, 'failed: checkExecutionExitCodeAndExistenceOfOutputFiles, filenames array is empty');
	}
	else {		
		my $missingFiles = '';
		
		# for each output file, we will check that the file exists in CWD
		foreach my $fileName( @$outputFilesNamesArrRef ) {
			if (! -e $fileName) {
				$someFilesAreMissingInd = 1;
				$missingFiles.= $fileName.', ';
			}			
		}
		
		# checking if some files are missing \ the executable failed
		if ($executionExitCode != 0 || $someFilesAreMissingInd) {
			my $errorDetails = '';
			$errorDetails = "Executeable Failure, exit code: $executionExitCode\.";
			if ($executionExitCode != 0) {
				if ($errorFileName ne '') {				
					if (-e $errorFileName) {
						my $possibleErrorText = `tail $errorFileName`;
						$errorDetails .= "\nExecuteable output last lines(which may contain the error):\n$possibleErrorText\n";
					}
				}
			}
			if ($someFilesAreMissingInd) {
				$errorDetails.= "\nMoreover, some output files are missing. files: $missingFiles";
			}
			&reportOnError($moudle, $task, $errorDetails);
		}		
	}
	return $someFilesAreMissingInd;
}

#	dvir 2010/12/20
# checking the task in the cwd, whether it succeeded
# updating the current running tasks counter,
# updating the total submited tasks counter
sub markThatTaskIsFinished( $ $ $) {

	my $projectName = $_[0];
	# not obligatory, if this param isn't given, than all sub dir of the current dir, will be
	# consdired as the sub jobs.
	# if this param is undef, than the
	my $subJobsArrRef = $_[1];
	
	my $shouldNotNeedToAquireLock = $_[2]; # not obligatory
	my $shouldNotUpdateStatus = $_[3]; # not obligatory, should be 1 (=true) if the status update had allready been preformed
	
	if (! defined $shouldNotNeedToAquireLock) {
		# updating the counters only when having the lock of the task,
		# in order not to decrease the counters twice:
		# so if we decrease them now,
		# than the kill process will see that the task has finished,
		# and therefore won't decrease the counters too.
		# and if the kill process will get the lock before us,
		# than it will kill the condor dag, and this code won't run,
		# and also we succeded not to decrease the counters twice.
		# when we are inside the lock, we can be sure that $Configuration::TASK_KILLED_FILENAME doesn't exists at the current folder.
		open TASK_LOCK_FILE, ">$Configuration::TASK_LOCK_FILENAME";
		# dvir important: 2011/06/01, the lock is exclusive, in order that checkTasks.pl won't be called at the same time
		flock TASK_LOCK_FILE, $Configuration::EXCLUSIVE_LOCK;
	}
	
	# the job that will be marked as succsessful or failed or timed out or no status or a combination of the prev. at the cwd : '.'
	if (! defined $shouldNotUpdateStatus) {
		$shouldNotUpdateStatus = 0;
	}
	if	(! $shouldNotUpdateStatus) {
		&JobStatus::updateJobStatusByItsSubJobs('.',$subJobsArrRef);
	}
	
	my $cwd = `pwd`;
	
	# decreasing the counters of the currently running tasks & currently submitted tasks
	&General::decreaseCurrentRunningCounter($cwd);
	
	# removing the indication that the current task is running
	&General::markThatTaskEndedRunning($projectName);
	
	if (! defined $shouldNotNeedToAquireLock) {
		close TASK_LOCK_FILE;
	}
}

#	dvir 2010/12/30
# marking the task with the given workDir as killed,
# and addding to the file which indicates the kill a message to help understand how the kill
# was done.
# what was the stage of task when the kill was preform
# posible stages: 
# 1. main dag wasn't submitted,
# 2. main dag was submitted and hasn't finished yet
# 3. main dag was submitted and ended (the task is finished).
# Important: this method should be called when holding the lock of the task.
sub markThatTaskIsKilled($ $ $ $ $ $) {
	my $projectName = $_[0];
	my $taskDir = $_[1];
	my $taskWorkDir = $_[2];
	my $killMessage = $_[3];
	my $shouldDecreaseCounters = $_[4];
	my $doesTheTaskWasMarkedAsRunning = $_[5]; # relevant only when $shouldDecreaseCounters is true
											   # (there may be caeses that this method will be called when the task is allready marked as killed)
		
	&makeSureDirExists($taskDir);
	&makeSureDirExists($taskWorkDir);
	
	open TASK_KILL_FILE, ">>$taskWorkDir\/$Configuration::TASK_KILLED_FILENAME";	
	print TASK_KILL_FILE &getCurrentDate().", $0: $killMessage\n";
	close TASK_KILL_FILE;
	
	if ($shouldDecreaseCounters) {		
		&General::decreaseCurrentRunningCounter($taskDir);		
		
		# if the task was marked as running, than we will remove the mark
		if ($doesTheTaskWasMarkedAsRunning) {
			# removing the indication that the current task is running
			&General::markThatTaskEndedRunning($projectName);
		}
		else {
			&General::removeTaskFromSubmittedDir($projectName);
		}
	}
}

#	dvir 2010/12/30
# marking that the task is now running on condor,
# so that every hour we can check with checkTasks.pl if an error occured and the task
# is not running any more at the condor, but we didn't updaed the counters so that new tasks can come
# Important: this method should be called when holding the lock of the task.
sub markThatTaskIsRunning($) {
	my $projectName = $_[0]; # should be the task dir name without slashes
	
	# it ok, that there isn't a print command,
	# notice to the mode of the open : it creates the file.
	open TASK_RUNNING_INDICATION, ">$Configuration::TASKS_BY_RUN_STATUS_DIR\/$projectName";
	close TASK_RUNNING_INDICATION;
	
	# removing the indication that the current task waits to start running
	&removeTaskFromSubmittedDir($projectName);
}

#	dvir 2010/12/30
# marking that the task is no more running on condor.
sub markThatTaskEndedRunning($) {
	my $projectName = $_[0]; # should be the task dir name without slashes
	# removing the indication that the current task is running
	unlink "$Configuration::TASKS_BY_RUN_STATUS_DIR\/$projectName";
}

sub doesTheTaskIsMarkedAsRunning($) {
	my $projectName = $_[0]; # should be the task dir name without slashes
	return -e "$Configuration::TASKS_BY_RUN_STATUS_DIR\/$projectName";
}

#	dvir 2010/01/06
# marking that the task is submitted
sub markThatTaskIsSubmmited($) {
	my $projectName = $_[0];

	my $projectTaskDir = &General::getTaskDirNameByTaskName($projectName);
	
	# it ok, that there isn't a print command,
	# notice to the mode of the open : it creates the file.
	open SUBMITTED_IND , ">$projectTaskDir\/submitted.indication";
	close SUBMITTED_IND;
	
	# the lock is for cases when someone reads the dir at the same time
	open WAITING_TO_RUN_LOCK, ">$Configuration::TASKS_WAITING_TO_RUN_LOCK";
	flock WAITING_TO_RUN_LOCK, $Configuration::EXCLUSIVE_LOCK;
	
	# it ok, that there isn't a print command,
	# notice to the mode of the open : it creates the file.
	open TASK_SUBMITTED_INDICATION, ">$Configuration::TASKS_BY_WAIT_TO_RUN_STATUS_DIR\/$projectName";
	close TASK_SUBMITTED_INDICATION;
	
	close WAITING_TO_RUN_LOCK;
}
# dvir 2011/04/10
sub removeTaskFromSubmittedDir($) {
	my $projectName = $_[0]; # should be the task dir name without slashes
	
	# the lock is for cases when someone reads the dir at the same time
	open WAITING_TO_RUN_LOCK, ">$Configuration::TASKS_WAITING_TO_RUN_LOCK";
	flock WAITING_TO_RUN_LOCK, $Configuration::EXCLUSIVE_LOCK;
	
	unlink "$Configuration::TASKS_BY_WAIT_TO_RUN_STATUS_DIR\/$projectName";
	
	close WAITING_TO_RUN_LOCK;
}

#	dvir 2011/03/07
# marking that the job (usually an estimation job) has stoped its execution, since a time period
# had passed
# the indication will be created at the run machine and will be copied to the local machine,
# after the job is ended
sub markThatJobHasStoppedAfterTimePeriod( $ ) {
	my $logMessage = $_[0]; # a message that will be writteh to the indication file
	# I'm opening the file >>, because there are cases (morgan multi win run for example, that we may stop the process several times)
	# so we don't want that the previous message will be overwritten by the current message
	open JOB_HAS_STOPPED_INDICATION, ">>$Configuration::JOB_STOPPED_FILENAME";
	print JOB_HAS_STOPPED_INDICATION $logMessage;
	close JOB_HAS_STOPPED_INDICATION;
}

#	dvir 2011/03/07
# returning an indication about whether or not the estimation has stopped for the given family
sub doesEstimationJobHasStoppedForGivenFamily( $ ) {
	my $familyNum = $_[0];
	
	my $familyEstimationDir = &General::getFamilyEstimationDir($familyNum);
	
	# making sure that the estimation dir exists (otherwise, there wasn't any estimation)
	if (!-d $familyEstimationDir) {
		&reportOnErrorReleaseSemaphoresAndDie('General.pm', '', 'failed: doesEstimationJobHasStoppedForGivenFamily, estimation folder does\'nt exists');
	}
	return &doesEstimationJobHasStoppedAtGivenDir($familyEstimationDir);
}

# dvir 2011/03/17
# checking if the estimation job at the current dir has stoped
sub doesEstimationJobHasStoppedAtGivenDir( $ ) {
	my $estimationDir = $_[0];
	
	return -e "$estimationDir\/$Configuration::JOB_STOPPED_FILENAME";
}

# dvir 2011/07/11
# checking if the job at the current dir has stoped
sub wasJobStopped( $ ) {
        my $estimationDir = $_[0];               

        return -e "$estimationDir\/$Configuration::JOB_STOPPED_FILENAME";
}


#	dvir 2010/12/30
# if the status of the dag in condor is "hold", than we also return true.
# the purpose of this method is to check if the dag may still run at condor
sub doesTheMainDagOfTheTaskIsRunning($ \$ ) {
	my $projectName = $_[0];
	my $messageStrRef = $_[1];
	
	$$messageStrRef = '';
	
	my $taskMainDagId = &getClusterIdOfMainDagOfTask($projectName);
	
	if (! defined $taskMainDagId) {
		# no main dag
		$$messageStrRef = "No main dag was found for the task, please check the submission file: $configfile{taskDir}\/$projectName\/workDir\/$Configuration::CONDOR_DAG_SUBMISSION_OF_MAIN_TASK_OUTPUT_FILENAME for more details\.";
		return 0;
	}

	my $condorQueryResult = `$Configuration::CONDOR_BIN_DIR/condor_q -format "%s" ClusterId $taskMainDagId`;
	my $exitCode = $? >> 8;
	if ($condorQueryResult eq '') {$$messageStrRef = "exit code : $exitCode"; return 0;}
	else {return 1;}
}

#	dvir 2010/12/30
# getting the cluster id of the main dag of the task as it was written by the condor_dagman to its standard output file
sub getClusterIdOfMainDagOfTask($) {
	my $projectName = $_[0];
	
	my $taskMainDagId = undef;

	open CONDOR_DAGMAN_SUBMIT_OUTPUT, "$configfile{taskDir}\/$projectName\/workDir\/$Configuration::CONDOR_DAG_SUBMISSION_OF_MAIN_TASK_OUTPUT_FILENAME";
	while (my $currSubmitLine  = <CONDOR_DAGMAN_SUBMIT_OUTPUT>)  {
		if ($currSubmitLine =~ /submitted to cluster /) {
			my @currSubmitLineArr = split /submitted to cluster /, $currSubmitLine;
			my @id_line = split /\./, $currSubmitLineArr[1];
			$taskMainDagId = $id_line[0];
			last;
		}
	}
	close CONDOR_DAGMAN_SUBMIT_OUTPUT;	
	return $taskMainDagId;
}

#	dvir 2011/08/14
sub howManyWindowsInASingleJobVer2($ $ $ $ $ $) {
	my $maxJobsPerTask = $_[0];
	my $maxTimePerJob = $_[1]; # in seconds
	my $minTimePerJob = $_[2];
	my $numOfWindowsInTask = $_[3];
	my $timeToRunSingleWindow = $_[4]; # in seconds
	
	# if there is no limitiation over the num of windows of the job.
	if ($maxJobsPerTask eq $SharedConsts::UNLIMITED_VALUE) {
		return 1, undef;
	}
	
	# assert part
	if ($maxJobsPerTask <= 0) {
		&reportOnErrorReleaseSemaphoresAndDie('General.pm', '', "failed: howManyWindowsInASingleJob, maxJobsPerTask: $maxJobsPerTask (Should be positive)");
	}
	if ($numOfWindowsInTask <= 0) {
		&reportOnErrorReleaseSemaphoresAndDie('General.pm', '', "failed: howManyWindowsInASingleJob, numOfWindowsInTask: $numOfWindowsInTask (Should be positive)");
	}	
	if ($maxTimePerJob <= 0) {
		&reportOnErrorReleaseSemaphoresAndDie('General.pm', '', "failed: howManyWindowsInASingleJob, maxTimePerJob: $maxTimePerJob (Should be positive)");
	}
	# at least we should be able to run one window in a single job
	if ($maxTimePerJob  < $timeToRunSingleWindow) {
		&reportOnErrorReleaseSemaphoresAndDie('General.pm', '', "failed: howManyWindowsInASingleJob, maxTimePerJob: $maxTimePerJob, timeToRunSingleWindow: $timeToRunSingleWindow, timeToRunSingleWindow should be less than maxTimePerJob, in order to be able to run a window");
	}
	
	if ($minTimePerJob  > $maxTimePerJob) {
		&reportOnErrorReleaseSemaphoresAndDie('General.pm', '', "failed: howManyWindowsInASingleJob, maxTimePerJob: $maxTimePerJob, minTimePerJob: $minTimePerJob, minTimePerJob should be less than maxTimePerJob");
	}	
	# end assert part
	
	# if we will create jobs with less windows than $minWindowsPerJob
	# then we will create too many jobs and exceed the max num of jobs per task
	my $minWindowsPerJob = &POSIX::floor ( $minTimePerJob / $timeToRunSingleWindow );

	# if we will create jobs with more windows than $maxWindowsPerJob
	# then our job will run for too much time and exceed the max time per job.
	my $maxWindowsPerJob = &POSIX::floor ($maxTimePerJob  / $timeToRunSingleWindow);

	my $suggestedWindowsPerJob = &POSIX::ceil ( $numOfWindowsInTask / $maxJobsPerTask );

	# we can't run a job with more windows than the total num of windows in the task
	#Note: the following ALWAYS holds: $actualMinWindowsPerJob <= $maxWindowsPerJob (because $minWindowsPerJob < $maxWindowsPerJob)
	#my $actualMinWindowsPerJob = min ($minWindowsPerJob, $numOfWindowsInTask);
	my $actualMinWindowsPerJob = $minWindowsPerJob;
	if ($minWindowsPerJob > $numOfWindowsInTask) {
		$actualMinWindowsPerJob = $numOfWindowsInTask;
	}

	# in order to avoid creating too light (=fast) jobs, with too less windows:
	#my $windowsPerJob = max($actualMinWindowsPerJob, $suggestedWindowsPerJob);
	my $windowsPerJob = $suggestedWindowsPerJob;
	if ($suggestedWindowsPerJob < $actualMinWindowsPerJob) {
		$windowsPerJob = $actualMinWindowsPerJob;
	}
	
	#If the number of jobs is too large, there is an error
	if ($windowsPerJob > $maxWindowsPerJob) {
		#If we're here, this means that $suggestedWindowsPerJob > $actualMinWindowsPerJob (due to assert)
		my $timeToRunSingleWindowInMin = &POSIX::ceil($timeToRunSingleWindow/60);

		my $errMsgMaxRunTimePerWindow = &POSIX::floor($maxTimePerJob / $suggestedWindowsPerJob) * 60;
		
		my $maxWindowsPerTask = $maxJobsPerTask * $maxWindowsPerJob;
		# dvir 2011/09/04, because, there are sometimes that the windows intersects
		#my $errMsgMaxWinSize = &POSIX::floor($numOfMarkersInTask / $maxWindowsPerTask);
		#my $errMsgMaxMarkersPerAnalysis = $maxWindowsPerTask * $singleWindowSize;		
		
		my $errorMsg = 
		"The analysis request was rejected because it could take too much time.\n".
		"<br> The analysis has $numOfWindowsInTask windows, and the maximum allocated time for each window is $timeToRunSingleWindowInMin minutes.\n".
		"<br><br>When performing analysis on $numOfWindowsInTask windows, then the computation time per window should be at most ${errMsgMaxRunTimePerWindow} minutes.\n".
		"<br>When the maximum computation time of each window is ${timeToRunSingleWindowInMin} minutes, the maximum number of allowed windows is $maxWindowsPerTask.\n".
		"<br><br>Please reduce either the number of windows in the analysis or place a time restriction on each computation and then try again.\n";
		
		return undef, $errorMsg;
				  
	}
	return $windowsPerJob, undef;
}

#	dvir 2011/01/11
# terminology: task may contain many jobs.
# from the answer to this question: "howManyWindowsInASingleJob", 
# we can derive how many jobs, each family in the task will have.
#
# our goal is :
# 1. not to create more than $Configuration::MAX_NUM_OF_JOBS_PER_FAMILY+1 jobs per family.
#    the "+1" is for cases of ceil and floor.
# 2. not to create too light (=fast) jobs, in order to reduce the overhead in Condor.
#    we will build jobs, so that each one will take approximately no less time than:
#    $Configuration::OPTIMIZE_MIN_EXECUTION_TIME_OF_A_SINGLE_JOB_IN_SECONDS.
#    (I writed "approximately", since there are cases when the job will take little less than that value).

# there may be jobs with less windows than $numOfWindowsInSingleJob
# if $Configuration::MAX_NUM_OF_JOBS_PER_FAMILY = 40 and $totalNumOfWindowsInTask is 41, if we return 2 windows per job, than there will be a job with one window
sub howManyWindowsInASingleJob($ $ $ $) {
	my $numOfMarkerInTask = $_[0];
	my $singleWindowSize = $_[1];
	my $timeToCompleteSingleWindow = $_[2]; # in seconds
	my $maxSingleJobTimeInSeconds = $_[3];
	
	if ($singleWindowSize <= 0) {
		&reportOnErrorReleaseSemaphoresAndDie('General.pm', '', "failed: howManyWindowsInASingleJob, singleWindowSize: $singleWindowSize");
	}
	if ($numOfMarkerInTask <= 0) {
		&reportOnErrorReleaseSemaphoresAndDie('General.pm', '', "failed: howManyWindowsInASingleJob, numOfMarkerInTask: $numOfMarkerInTask");
	}
	
	# there may be cases when the task will be finish too slow, than for safety 
	# we take only 0.85 of the job time into consideration
	my $actualMaxSingleJobTimeInSeconds = 0.85 * $maxSingleJobTimeInSeconds;
	
	# in case that the job had finished very very quickly.
	if ($timeToCompleteSingleWindow == 0) {
		$timeToCompleteSingleWindow = 1;
	}
		
	my $totalNumOfWindowsInTask = &POSIX::ceil($numOfMarkerInTask / $singleWindowSize);
	# my $totalTimeOfExecutionOfTaskInSeconds = $totalNumOfWindowsInTask * $timeToCompleteSingleWindow;
	
	# the task can be executed, since it dosn't take too much time

	my $numOfJobsIsNoMoreThanTheOptimize = 0;
	# we would like to create minimum jobs.
	my $executionTimeOfASingleJob = $Configuration::OPTIMIZE_MIN_EXECUTION_TIME_OF_A_SINGLE_JOB_IN_SECONDS;
		
	# while 
	while ($actualMaxSingleJobTimeInSeconds > $executionTimeOfASingleJob) {
		# num of windows in a single window, when a job takes $executionTimeOfASingleJob
		# print "timeToCompleteSingleWindow=$timeToCompleteSingleWindow, executionTimeOfASingleJob=$executionTimeOfASingleJob, ";
		my $numOfWindowsInSingleJob = &POSIX::floor($executionTimeOfASingleJob / $timeToCompleteSingleWindow);
		# if the job can contain at least one window
		if ($numOfWindowsInSingleJob) {
		
			# calc the minmum numOfJobs to complete the task, 
			# when a single job has $numOfWindowsInSingleJob windows within it
			my $numOfJobs = &POSIX::floor($totalNumOfWindowsInTask / $numOfWindowsInSingleJob);
			
			# checking if the num of jobs doesn't exceed the limit
			if ($numOfJobs <= $Configuration::MAX_NUM_OF_JOBS_PER_FAMILY) {
				my $numOfJobsInPractice;
				# we don't need more than $totalNumOfWindowsInTask jobs ,
				# (we can reach $totalNumOfWindowsInTask jobs when each job contains only one window)
				# so making sure we don't return larger value than $totalNumOfWindowsInTask
				if ($numOfWindowsInSingleJob > $totalNumOfWindowsInTask) {
					$numOfWindowsInSingleJob = $totalNumOfWindowsInTask;
					$numOfJobsInPractice = 1;					
				}
				else {
					$numOfJobsInPractice = &POSIX::ceil($totalNumOfWindowsInTask / $numOfWindowsInSingleJob);
				}
				return ($numOfWindowsInSingleJob, $numOfJobsInPractice);
			}
		}
		# increasing the time for a single job, in order that it may contain more windows
		# so that the total number of jobs for the task will be less than $Configuration::MAX_NUM_OF_JOBS_PER_FAMILY
		# we are increasing the time in $Configuration::ESTIMATION_TIME_IN_SECONDS
		# since that we can be sure that at least one window ends within $Configuration::ESTIMATION_TIME_IN_SECONDS time
		$executionTimeOfASingleJob+= $Configuration::ESTIMATION_TIME_IN_SECONDS;		
	}
	
	# if we get to here, than the job is very heavy,
	# and demands more than $Configuration::MAX_NUM_OF_JOBS_PER_FAMILY jobs of $actualMaxSingleJobTimeInMinutes.
	# so the num of windows in a job will be, one that doesn't exceed $Configuration::MAX_NUM_OF_JOBS_PER_FAMILY.
	# if $Configuration::MAX_NUM_OF_JOBS_PER_FAMILY = 40 and $totalNumOfWindowsInTask is 79, than we return 2, and by that, we don't exceed 40 jobs.
	my $numOfWindowsInSingleJob = &POSIX::ceil($totalNumOfWindowsInTask / $Configuration::MAX_NUM_OF_JOBS_PER_FAMILY);
	my $numOfJobsInPractice;
	if ($totalNumOfWindowsInTask < $Configuration::MAX_NUM_OF_JOBS_PER_FAMILY) {
		$numOfJobsInPractice = $totalNumOfWindowsInTask;
	}
	else {
		$numOfJobsInPractice = $Configuration::MAX_NUM_OF_JOBS_PER_FAMILY;
	}
	return ($numOfWindowsInSingleJob, $numOfJobsInPractice);
}

# dvir
# in case that the estimation havn't been finished
# returning how many windows there will be in a single job
# and how many jobs will be.
sub howManyWindowsInASingleJobWithoutSingleWindowEstimation($ $) {
	my $numOfMarkerInTask = $_[0];
	my $singleWindowSize = $_[1];
	
	if ($singleWindowSize <= 0) {
		&reportOnErrorReleaseSemaphoresAndDie('General.pm', '', "failed: howManyWindowsInASingleJobWithoutSingleWindowEstimation, singleWindowSize: $singleWindowSize");
	}
	if ($numOfMarkerInTask <= 0) {
		&reportOnErrorReleaseSemaphoresAndDie('General.pm', '', "failed: howManyWindowsInASingleJobWithoutSingleWindowEstimation, numOfMarkerInTask: $numOfMarkerInTask");
	}

	my $totalNumOfWindowsInTask = &POSIX::ceil($numOfMarkerInTask / $singleWindowSize);
	my $numOfWindowsInSingleJob = &POSIX::ceil($totalNumOfWindowsInTask / $Configuration::MAX_NUM_OF_JOBS_PER_FAMILY);
	my $numOfJobsInPractice;
	if ($totalNumOfWindowsInTask < $Configuration::MAX_NUM_OF_JOBS_PER_FAMILY) {
		$numOfJobsInPractice = $totalNumOfWindowsInTask;
	}
	else {
		$numOfJobsInPractice = $Configuration::MAX_NUM_OF_JOBS_PER_FAMILY;
	}
	return ($numOfWindowsInSingleJob, $numOfJobsInPractice);
}

# dvir, 31/07/2011
# this method checks if the analysis is on a SNP input or just DAT&PED input
# if the analysis is on snp input, than the "modfile.txt" contains the postions
# if not then the modefile.txt line looks like this : "List of markers and their positions: _(_[_(_positions=_)_]_)_"
# and the "positions" key is ''
sub doesTheAnalysisIsOn_SNP_Input( \% ) {
	my $variablesHashRef = $_[0];
	
	my $doesTheAnalysisIsOn_SNP_Input = $$variablesHashRef{positions} ne '';
	return $doesTheAnalysisIsOn_SNP_Input;
}

# dvir, 10/10/2011
sub howManyTasksAreInWaitingToRunMode() {
	my @tasksArr = getTasksNamesArr_ofTaskswhichWaitToRun_SortedByDateOfSubmission();
	return scalar (@tasksArr);
}

# dvir, 10/10/2011
# the oldest task is first on the list
sub getTasksNamesArr_ofTaskswhichWaitToRun_SortedByDateOfSubmission() {
	open WAITING_TO_RUN_LOCK, ">$Configuration::TASKS_WAITING_TO_RUN_LOCK";
	flock WAITING_TO_RUN_LOCK, $Configuration::EXCLUSIVE_LOCK;
	
	my $tasks = `ls -rt $Configuration::TASKS_BY_WAIT_TO_RUN_STATUS_DIR`;
	chomp($tasks);
	my @tasksArr = split '\n', $tasks;
	
	close WAITING_TO_RUN_LOCK;
	return @tasksArr;
}

# dvir 2011/10/11
sub makeSureThatSystemIsNotOutOfResources() {
	if (-e $Configuration::SYSTEM_OUT_OF_RESOURCES_FILENAME_IND) {
		# dvir: I'm not using a mail alert, in order not to fluid the programmers mail.
		# they have recived a single mail about it, when the file indication was created
		die "System is out of resources\n";
	}
}

# dvir 2011/10/11
# when marking this, than no user will be allowed to submit \ run a task
sub markThatSystemIsOutOfResources() {
	open IND, ">$Configuration::SYSTEM_OUT_OF_RESOURCES_FILENAME_IND";
	close IND;
	
	&reportOnErrorReleaseSemaphoresAndDie('General.pm','',"SYSTEM IS OUT OF RESOURCE!\nNO NEW SUBMISSION OR RUNNING WILL BE DONE!\nplease check top & ps -ef.");
}

# dvir 2011/10/26
# returns 1 if the task runs locally and not at condor
sub doesTheTaskRunsLocally( $ ) {
	my $projectWorkDir = $_[0];
	
	# for now only segregation runs locally
	return -e "$projectWorkDir\/$Configuration::SEGREGATION_ANALYSIS_FILENAME";
}

# dvir 2011/11/10
sub getArrayOf_pm_filesNamesInBinDir() {
	my $binDir = $Configuration::configfile{ binDir };
	
	my $pmFilesInBinDirStr = `ls $binDir/*.pm`;
	my @pmFilesInBinDirArr = split ' ', $pmFilesInBinDirStr;
	return @pmFilesInBinDirArr;
}

#############################
# Global counters functions - subs from Configuration.pm moved to here by dvir 2010-09-21#
#############################

# Running Tasks Counter
#######################
sub increaseCurrentRunningCounter( $ )
{
	my $projectTaskDir = $_[0];
	
	my $counterOfCurrentTasks = $configfile{counterOfTasksLock}; 
	my $counter =  $configfile{counterOfTasksInSys}; 
	open (COUNTERLOCKFILE, ">>$counterOfCurrentTasks") or die "Can't open $counterOfCurrentTasks file\n";
	flock(COUNTERLOCKFILE,2) or die "Can't lock $counterOfCurrentTasks file (cant update task counter)\n";
	my $now_string = strftime "%a %b %e %H:%M:%S %Y", localtime;
	print COUNTERLOCKFILE "locked by PID $$  (in $projectTaskDir) at $now_string to increase\n";
	
	my $newCounter;
	if (! -e $counter) { $newCounter = 0; }
	else {
		open (COUNTER, "$counter") or die "Can't open $counter file\n";
		my $line = <COUNTER>;
		my @tmp = split(/[ \t\n]+/,$line);
		my $counterVal= $tmp[0];
		$newCounter = $counterVal + 1;
		close COUNTER;
	}
	
	open (COUNTER, ">$counter") or die "Can't open $counter file\n";
	print COUNTER "$newCounter";
	close COUNTER;
	close COUNTERLOCKFILE;
}

sub decreaseCurrentRunningCounter( $ )
{
	my $projectTaskDir = $_[0];
	
	my $counterOfCurrentTasks = $configfile{counterOfTasksLock}; 
	my $counter =  $configfile{counterOfTasksInSys}; 
	open (COUNTERLOCKFILE, ">>$counterOfCurrentTasks") or die "Can't open $counterOfCurrentTasks file\n";
	flock(COUNTERLOCKFILE,2) or die "Can't lock $counterOfCurrentTasks file (cant update task counter)\n";
	my $now_string = strftime "%a %b %e %H:%M:%S %Y", localtime;
	print COUNTERLOCKFILE "locked by PID $$ \(TaskHandler\.pl\ in $projectTaskDir) at $now_string to decrease\n";
	
	my $newCounter;
	if (! -e $counter) {
		$newCounter = 0;
	}
	else {
		open (COUNTER, "$counter") or die "Can't open $counter file\n";
		my $line = <COUNTER>;
		my @tmp = split(/[ \t\n]+/,$line);
		my $counterVal= $tmp[0];
		if ($counterVal > 0) {
			$newCounter = $counterVal - 1;			
		}
		else {
			$newCounter = 0;
		}		
		close COUNTER;
	}
	
	open (COUNTER, ">$counter") or die "Can't open $counter file\n";
	print COUNTER "$newCounter";
	close COUNTER;
	close COUNTERLOCKFILE;
}

sub getCurrentRunningCounter( $ )
{
	my $projectTaskDir = $_[0];
	
	my $counterOfCurrentTasks = $configfile{counterOfTasksLock}; 
	my $counter =  $configfile{counterOfTasksInSys};
	open (COUNTERLOCKFILE, ">>$counterOfCurrentTasks") or die "Can't open $counterOfCurrentTasks \n";
	flock(COUNTERLOCKFILE,2) or die "Can't lock $counterOfCurrentTasks file\n";
	my $now_string = strftime "%a %b %e %H:%M:%S %Y", localtime;
	print COUNTERLOCKFILE "locked by PID $$ \(TaskHandler\.pl\) at $now_string\n";

	my $counterVal;
	unless( -e $counter) {
		open (COUNTER, ">$counter") or die "Can't open $counter file\n";
		print COUNTER "0";
		close COUNTER;
		$counterVal = 0;
	}
	else {
		open (COUNTER, "$counter") or die "Can't open $counter file\n";
		my $line = <COUNTER>;
		my @tmp = split(/[ \t\n\r\s\f]+/,$line);
		$counterVal = $tmp[0];
		close COUNTER;
	}
	close COUNTERLOCKFILE;
	
	return $counterVal;
}
######################################
# Global ExponentialBackoff function #
######################################

# dvir 04/12/2011
# executing the given command while holding a rand lock (one of 5 locks),
# and by holding one of 5 locks, than we can be sure that no more than 5 process are executing this command
# at current moment.
# this method should be called only from perfomrWithExponentialBackoff
# and its porpose is to preform command that requires communication with the web server.
# and this is why we aren'y letting more than $NUM_OF_RAND_LOCK_FILES commands to be executed at once.
my $NUM_OF_RAND_LOCK_FILES = 5;
my $RAND_LOCK_FILE_PREFIX = "$Configuration::LOCKS_DIR\/connectionLockNum";
sub executeCommandWhileHoldingRandLock( $ ) {
	my $command = $_[0];
	
	# choosing randomly a file to lock;
	my $currFileNumToLock = int(rand($NUM_OF_RAND_LOCK_FILES));
	my $lockFileName = $RAND_LOCK_FILE_PREFIX.$currFileNumToLock;
	open (RAND_LOCK, ">$lockFileName") || &General::reportOnErrorReleaseSemaphoresAndDie('General.pm',`pwd`, sprintf($Configuration::ERR_MSG_FAILED_OPENING_FILE, $lockFileName ,$!));
	flock(RAND_LOCK, $Configuration::EXCLUSIVE_LOCK)  || &General::reportOnErrorReleaseSemaphoresAndDie('General.pm',`pwd`, sprintf($Configuration::ERR_MSG_FAILED_LOCKING_FILE, $lockFileName ,$!));
	
	system($command);
	my $ret = $? >> 8;
	
	# releasing the lock
	close RAND_LOCK;
	return $ret;
}

sub perfomrWithExponentialBackoff( $ )
{
	my $command = $_[0];
	
	my $sleepTime = 1;
	my $exit_value  = &executeCommandWhileHoldingRandLock($command);
	while (1)
	{
		
		if( $exit_value != 0 )
		{
			sleep $sleepTime;		
			$sleepTime = $sleepTime * 2;
			
			# Try to perform action again
			$exit_value  = &executeCommandWhileHoldingRandLock($command);
				
			# We can/should determin the factor of max sleep in some other place
			if ($sleepTime >= $Configuration::maxSleepTimeOfExpBackoff ) 
			{
				return -1; # Let go..
			}
		}
		else
		{
			return 0;
		}	
	}
}


##############################################
# Global add\remove to\from PIDList function #
##############################################

# Add only if the pid isn't already in the file
# $_[0] - project name
sub addMeToPIDList( $ )
{
	my $pidFileName 	= $configfile{pidFileName};
	my $tasksDir		= $configfile{taskDir};	
	my $pidFile = $tasksDir ."\/" . $_[0] ."\/" . $pidFileName;
	
	if (-e $pidFile) 
	{
		open (PIDFILE , "$pidFile") or die "Can't open $pidFile\n";
		for my $line (<PIDFILE>)
		{
			if ($line eq $$)
			{
				return -1; # PID Already found in the file. no special meaning to -1..
			}
		}
		close PIDFILE;
	}
	
	open (PIDFILE , ">>$pidFile") or die "Can't open $pidFile\n";
	print PIDFILE "$$\n";
	close PIDFILE;
}

# Remove only if the pid found in the file
sub removeMeFromPIDList( $ )
{
	my $pidFileName 	= $configfile{pidFileName};
	my $tasksDir		= $configfile{taskDir};	
	my $pidFile = $tasksDir ."\/" . $_[0] ."\/" . $pidFileName;
	my @newFile =();
	my $i=0;
	my $flagOfPIDInFile = "false";
	
	if (-e $pidFile)
	{
		open (PIDFILE , "$pidFile") or die "Can't open $pidFile\n";
		
		for my $line (<PIDFILE>)
		{
			$newFile[$i]=$line;
			$i++;
			
			if ($line eq $$)
			{
				$flagOfPIDInFile = "true";
			}
		}
		close PIDFILE;
		
		if ($flagOfPIDInFile eq "false")
		{
			return -1; # PID wasn't found in the file. no special meaning to -1..
		}
		
		open (PIDFILE , ">$pidFile") or die "Can't open $pidFile\n";
		
		my $flagOfKeepingFile = "false";
		for $i (0..$#newFile)
		{
			if (!($newFile[$i] eq $$))
			{
				print PIDFILE "$newFile[$i]\n";
				$flagOfKeepingFile="true";
			}
		}	
		close PIDFILE;
		
		if ($flagOfKeepingFile eq "false")
		{
		 system ("rm $pidFile");
		}
	}
}

# For now, return only the first pid found in the file - aka the parent of this project.
sub getPIDNumberByProjectName( $ )
{
	
	my $pidFileName 	= $configfile{pidFileName};
	my $tasksDir		= $configfile{taskDir};	
	my $projDir = $tasksDir ."\/" . $_[0];
	my $pidFile =  $projDir ."\/" . $pidFileName;
	
	unless( -d $projDir)
	{
		return -3;
	}

	
	
	if (!(-e $pidFile))
	{
		my $statFile = $tasksDir ."\/" . $_[0] ."\/" . "status.log";
		if( -e $statFile)
		{
			# meaning that the task has ended - status exists but pid doesn't
			return -1; 
		}
		else
		{ 
			# hasn't started yet.
			return -2;
		}
	}
	
	my @pidlistFile;
	tie @pidlistFile, 'Tie::File', "$pidFile" or die"couldn't open $pidFile for reading";
	my $line = $pidlistFile[0];
	my @tmp = split(/[ \t\n\r\s\f]+/,$line);
	untie @pidlistFile; 
	
	system("rm $pidFile");
	
	return $tmp[0];

}

#############################
# Global counters functions - subs from Configuration.pm moved to here by dvir 2010-09-21#
# end
#############################

1;
