# dvir 2011/11/16
package HMM;

use strict;
use AnalysisWith_modefile;
use Configuration;
use HMM_Html;
use General;
use JobStatus;
use POSIX qw(strftime); # for strftime

# this class inherits methods from AnalysisWith_modefile
our @ISA = ('AnalysisWith_modefile');




# if there are errors, there are written to the 'results.txt' and to the standrad output
our $STANDARD_ERROR_FILE_HMM = 'results.txt';

my $HMM_COMPUTION_REQUIRES_MORE_MEMORY_ERROR_TEXT = 'Memory required for HMM calculation is much too big for current computer';

# methods we override from AnalysisTemplateClass.pm :

# whether this analysis supports a multi window run
sub doesAnalysisSupportsMultiWindowRun() {
	return 0;
}

# returning an array of files names which contain the results of the analysis, 
# those files should be copied to the web server
sub getFileListToBeCopiedToWebServer() {
	# if there was an error with an html, than although it was copied allready, 
	# I'm coping it again in order not to make an error at the General::sendFilesToWebServer command
	# at postJob.pl
	if (-e $Configuration::REPORT_HTML_FILENAME) {
		return ($Configuration::REPORT_HTML_FILENAME);
	}
	
	my $class = $_[0];
	
	my $projectTaskDir = $class->getProjectDir();
	my $workDir = $class->getWorkDir();
	&ProcessingBackend::setLogFiles("$projectTaskDir\/"); # needed in order that the initVariablesFromPrePoint() will work at debug mode
	my %variables = %{ &ProcessingBackend::initVariablesFromPrePoint( $projectTaskDir ) };
	
	my $doesTheAnalysisIsOn_SNP_Input = &General::doesTheAnalysisIsOn_SNP_Input(\%variables);
		
	my @plottingFiles = &Plotter::getPlottingFilesNamesToBeSentToWebServer();
	
	my @fileList = (@HMM_Html::RESULTS_FILE_LIST, @plottingFiles);
		
	# if the input is SNP, then adding the file for the genome browser
	if ($doesTheAnalysisIsOn_SNP_Input) {
		push @fileList, $GenomeBrowser::BED_OUTPUT_FILENAME;
	}
	
	return @fileList;
}

sub getCondorJobRank() {
	return 'LINKAGEATHOME_COMPUTER*1000 + KFlops';
}

# returning an array of the names of the output files of the analysis
sub getJobOutputFiles() {
	return @HMM_Html::HMM_EXECUTEABLE_OUTPUT_FILES;
}

# if a single job failes, than we will expect that the error will be in this file
sub getJobErrorFile() {
	return $STANDARD_ERROR_FILE_HMM;
}

# dvir 2011/11/16
# this method is called from postJob.pl with determine_job_status_of_leaf_job_with_one_exec
# parameters:
#	the folder name of a job
# this function should check if there was an acceptaable error.
# if the error is acceptable, than no mail should be sent to the developers.
# this method is called with a full path of the jobDir and you cannot assume nothing about the cwd
sub doesAnAcceptableErrorOcurredAtJob($ $) {
	my $class = $_[0]; # the name of this class
	my $jobDir = $_[1]; # full path or relative to the calling scope
	my $jobExitCode = $_[2];
	
	if ($jobExitCode == 0) {
		return 0;
	}
	
	# checking if the job needs more memory
	my $doesHMMCalculationNeedsMoreMemoryInOrderTORun = $class->doesHMMCalculationNeedsMoreMemoryInOrderTORun($jobDir);
	if ($doesHMMCalculationNeedsMoreMemoryInOrderTORun) {
		return 1;
	}
	
	# checking if there was timeout error
	# timeout is acceptable for HMM.
	my $doesJobHasA_TimedOut_Indication = &JobStatus::doesJobHasA_TimedOut_Indication($jobDir);
	if ($doesJobHasA_TimedOut_Indication) {
		return 1;
	}
	
	# if we got to here, than no acceptable error has ocurred.
	return 0;
}

# checking if the job needs more memory
sub doesHMMCalculationNeedsMoreMemoryInOrderTORun( $ ) {
	my $class = $_[0]; # the name of this class
	my $jobDir = $_[1]; # full path or relative to the calling scope
	
	# checking if there was a memory error
	my $errFileName = "$jobDir\/$STANDARD_ERROR_FILE_HMM";
	if (-e $errFileName) {
		open ERR_FILE, $errFileName;
		my $firstLineInErrorFile = <ERR_FILE>;
		close ERR_FILE;
		
		if ($firstLineInErrorFile =~ /$HMM_COMPUTION_REQUIRES_MORE_MEMORY_ERROR_TEXT/) {
			return 1;
		}
	}
	return 0;
}

# dvir 2011/09/6
# create dag, dirs dat files & ped files for the multi window per job run for all the families.
# each family can have some jobs.
# each job can have some duplications.
sub createDagForMultiWindowPerJobAnalysis(\$ \$ $ \% $ $ ) {
	my $class = $_[0]; # the name of this class
	
	my $datWrapperRef = $_[1];
	my $pedWrapperRef = $_[2];
	# numOfDuplications is the num of duplications for each window
	# or the num of duplications for each 'multi window job'
	my $numOfDuplications = $_[3];
	# each family will have a list of jobs, 
	# and each job will have a list of windows to be run in that job.
	# a window is a list of markers.
	my $windowsPerJobPerFamilyHashRef = $_[4];
	my $maxTimePerJob = $_[5]; # in seconds, unused

	# time to run HMM for each family
	my $timeToRunSingleWindow = $_[6]; # in seconds
	# can we create only one dat file for all the families jobs, and use soft link to this dat file
	# when using estimation, than this should be 0, cause diffrenet families can have a diffrenet partition of windows to jobs.
	my $allFamiliesJobsDatFilesAreTheSame = $_[7];
	
	# $allFamiliesJobsDatFilesAreTheSame should be cause for now, each job in haplo has only 1 window,
	# therefore jobs #x for family 1 & family 2 will both be on the same window: window #x.
	if (!$allFamiliesJobsDatFilesAreTheSame) {
		&General::reportOnErrorReleaseSemaphoresAndDie('HMM.pm',`pwd`,
					"createDagForMultiWindowPerJobAnalysis: allFamiliesJobsDatFilesAreTheSame should be 1 in HMM");
	}
	
	my $binDir = $Configuration::configfile{ binDir };
	
	my @families = $$pedWrapperRef->getFamiliesNumbers();

	my $workDir = $class->getWorkDir();		
	
	#### making sure that each family has only 1 job, and that job contains only 1 window
	foreach my $currFamily (@families) {
		
		# making sure that the job has only 1 window
		my $numOfJobsForFamily = scalar (@{$$windowsPerJobPerFamilyHashRef{$currFamily}});
		if ($numOfJobsForFamily != 1) {
			&General::reportOnErrorReleaseSemaphoresAndDie('HMM.pm',`pwd`,
				"createDagForMultiWindowPerJobAnalysis: family $currFamily, has $numOfJobsForFamily jobs (and not only 1 job)");
		}
		
		# checking if the single job (job no' 0) has more than 1 window
		my $numOfWindowsInSingleJob = scalar (@{$$windowsPerJobPerFamilyHashRef{$currFamily}[0]});
		if ($numOfWindowsInSingleJob != 1) {
			&General::reportOnErrorReleaseSemaphoresAndDie('HMM.pm',`pwd`,
				"createDagForMultiWindowPerJobAnalysis: family $currFamily singls jobhas $numOfWindowsInSingleJob windows (and not only 1 window)");
		}
	}
	
	
	# here we can be sure that each family has 1 job that contains 1 window
	my $someFamilyNum = $families[0];
	# job 0 and window 0
	my @markersInAnalysis = @{$$windowsPerJobPerFamilyHashRef{$someFamilyNum}[0][0]};
	
	
	#### creating the dat file for all the families
	my $datfilesDirName = "$workDir\/datfiles";
	mkdir $datfilesDirName;
	
	# creating the datfile
	my $datFileName = "$datfilesDirName\/$Configuration::HMM_DAT_FILENAME";

	open DATFILE_FH, ">$datFileName";		
	$$datWrapperRef->writeSubDatFileByListOfMarker(\*DATFILE_FH, \@markersInAnalysis);
	close DATFILE_FH;	

	
	#### creating a ped for each family one with no allels and the other with allels
	#### and creating a symbolic link to the single dat file.
	foreach my $currFamily (@families) {
		my $familyRunDir = &General::getFamilyRunDir($currFamily);
		
		# creating the pedfile
		my @familiesInDestPedFile = ($currFamily);
		open PEDFILE_FH, ">$familyRunDir\/$Configuration::HMM_PED_FILENAME";
		$$pedWrapperRef->writeSubPedFileByListOfMarker(\*PEDFILE_FH, \@markersInAnalysis, \@familiesInDestPedFile, 1, $PedWrapper::WRITE_PERSON_DATA_AND_ALLELS);
		close PEDFILE_FH;
		&General::makeSureFileExists("$familyRunDir\/$Configuration::HMM_PED_FILENAME");
		
		# creating the symbolic link
		symlink $datFileName, "$familyRunDir\/$Configuration::HMM_DAT_FILENAME";		
		&General::makeSureFileExists($datFileName);		
	}	
	
	
	#### editing $Configuration::CONDOR_JOB_SUBMIT_DEFINITIONS_FILENAME
	
	# AnalysisSubmissionFileData is from AnalysisSubmissionFile
	my $submitFile = 'HMM_run.sub';
	my $analysisSubmissionFileData = $class->getAnalysisSubmissionFileData();
	
	my @pmFilesInBinDirArr = &General::getArrayOf_pm_filesNamesInBinDir();
	my $analysisTypeFileIndication = $class->getType();
	my @jobInputFilesList = (
		$Configuration::HMM_DAT_FILENAME,
		$Configuration::HMM_PED_FILENAME);
	
	my @jobArguments = (
		$Configuration::HMM_DAT_FILENAME,
		$Configuration::HMM_PED_FILENAME,
		"-gh -m800",
		# not needed, cause postJob will catch the timeout.
		# "-hmm_timeout ".&General::getNumUpperValue($timeToRunSingleWindow/60)
		);
	
	$analysisSubmissionFileData->executeableName ( "$binDir\/Superlink\/SuperLink\.linux" );
	@{$analysisSubmissionFileData->executeableArgumentsList} = @jobArguments;
	@{$analysisSubmissionFileData->executeableInputFilesList} = @jobInputFilesList;
	$analysisSubmissionFileData->jobSecondsToBeExecuted ( $timeToRunSingleWindow );
	
	# creating $submitFile at cwd
	&AnalysisSubmissionFile::createAnalysisSubmissionFileAtCwd($analysisSubmissionFileData, $submitFile);
	
	
	#### Creating the DAG	
	my $familiesjobsName = '';
	open DAG_FILE, ">$Configuration::TASK_MAIN_DAG_NAME";
	foreach my $currFamily (@families) {
		my $familyRunDir = &General::getFamilyRunDir($currFamily);
		
		my $currFamilyJobName = "Family_${currFamily}_Job";
		$familiesjobsName .= "$currFamilyJobName ";
		print DAG_FILE 
			# DIR is the place to which the output of the run will be placed, 
			# and from which it will try to locate the job submission file (if that file is without a absulte path)
			"JOB         $currFamilyJobName $workDir\/$submitFile DIR $familyRunDir\n".
			"SCRIPT POST $currFamilyJobName $binDir\/postJob.pl $workDir determine_job_status_of_leaf_job_with_one_exec \$RETURN\n\n";		
	}

	# adding a job that will mark that the task is finished.
	# a successful task is when all families are successful.
	print DAG_FILE 
		"\nJOB LAST_JOB $Configuration::SIMPLE_SHORT_CONDOR_JOB\n".
		"SCRIPT POST LAST_JOB $binDir\/postJob.pl $workDir create_html____and____send_indication_to_web_server_to_mark_that_task_is_done____and____determine_job_status_of_the_task_by_families_dirs\n\n".
		"PARENT  $familiesjobsName  CHILD  LAST_JOB\n";
}

# creating the initial output files for each job
# for now it does nothing
sub createInitialOutputFiles(\$ \$ \%) {
	my $class = $_[0]; # the name of this class
	
	my $datWrapperRef = $_[1];
	my $pedWrapperRef = $_[2];
	# each family will have a list of jobs, 
	# and each job will have a list of windows to be run in that job.
	# a window is a list of markers.
	my $windowsPerJobPerFamilyHashRef = $_[3];
	
	my $workDir = $class->getWorkDir();	
	
	# my $outputDataDir = "$workDir\/$OUTPUT_DIRNAME";
	# mkdir "$workDir\/outputData";
	
	# passing over the families	
	# my @families = ();
	# &DatPedHandler::getFamiliesNumbersArrByFamiliesNumbersFile(\@families);
	
	#`touch createInitialOutputFiles.txt`;
}

# do nothing special
sub createFinalOutputFileForCurrentJob() {
	my $class = $_[0]; # the name of this class
	my $numOfDuplications = $_[1]; # unrelevant
	
	#`touch createFinalOutputFileForCurrentJob.txt`;
}

# dvir 2011/11/17
# creating the final output file
# the current implementation that I have wrote is planned to run when all families calculations are done.
sub createOutputFiles( \% ) {
	my $class = $_[0]; # the name of this class
	my $variablesHashRef = $_[1];
	
	my $workDir = $class->getWorkDir();	
	chdir $workDir;
	
	my $projectTaskDir = $class->getProjectDir();
	my %variables = %{ &ProcessingBackend::initVariablesFromPrePoint( $projectTaskDir ) };
		
	# getting the families
	my @families = ();
	&DatPedHandler::getFamiliesNumbersArrByFamiliesNumbersFile(\@families);
	
	my @successfulFamilies = ();
	my @failedFamilies = ();
	my @timedOutFamilies = ();
	my @notEnoughtMemoryFamilies = ();
	
	# passing over the families   
	foreach my $currFamily (@families) {
		my $familyRunDir = &General::getFamilyRunDir($currFamily);
		
		# all families jobs should be over, cause this function should run when all families jobs are over.
		if (! &JobStatus::isJobFinished($familyRunDir)) {
			&General::reportOnErrorReleaseSemaphoresAndDie('HMM.pm',`pwd`,
				"createOutputFiles: the job of family $currFamily isn't mark as finished");
		}
		
		if (&JobStatus::doesJobHasA_Successful_Indication($familyRunDir)) {
			push @successfulFamilies, $currFamily;
		}
		if (&JobStatus::doesJobHasA_Failed_Indication($familyRunDir) ||
		    &JobStatus::doesJobHasA_NoStatus_Indication($familyRunDir)) {
			push @failedFamilies, $currFamily;
		}
		if (&JobStatus::doesJobHasA_TimedOut_Indication($familyRunDir)) {
			push @timedOutFamilies, $currFamily;
		}
		if (&JobStatus::doesJobHasA_FailedWithAcceptableReason_Indication($familyRunDir)) {
			if ($class->doesHMMCalculationNeedsMoreMemoryInOrderTORun($familyRunDir)) {
				push @notEnoughtMemoryFamilies, $currFamily;
			}			 
			elsif (! &JobStatus::doesJobHasA_TimedOut_Indication($familyRunDir)) {
				# I don't know why the job is FailedWithAcceptableReason, so informing the programmers
				&General::reportOnErrorReleaseSemaphoresAndDie('HMM.pm',`pwd`,
					"createOutputFiles: family $currFamily, is marked as FailedWithAcceptableReason but the failur reason is not time out or not enough memory");
			}
		}
	}
	
	
	
	# the '1' is to mark : not to contact sys admin "contact string"
	my $shouldNotContactSysAdmin = 1;
	my $htmlFailureText = "";
	if (scalar(@failedFamilies) > 0) {
		# undef to mark that there is a need to contact sys admin
		$shouldNotContactSysAdmin = undef;
		$htmlFailureText .= '<div>computation faild for families: '.join(' ', @failedFamilies).'</div></br>';
	}
	
	if (scalar(@notEnoughtMemoryFamilies) > 0) {
		$htmlFailureText .= "<div>Memory required for HMM computation exceeds the capabilities of the system.</br>The HMM algorithm can handle only small pedigrees.</div></br>";
	}
	if (scalar(@timedOutFamilies) > 0) {
		$htmlFailureText .= '<div>computation has timed out for families: '.join(' ', @timedOutFamilies).'</div></br>';
	}
	
	
	# checking if we have successful families cacluations
	if (scalar(@successfulFamilies) > 0) {
		if (not defined $shouldNotContactSysAdmin) {
			$htmlFailureText.= '<div>please contact sys admin<div></br>';
		}
		&HMM_Html::createHtmlPages($htmlFailureText, \@successfulFamilies, $projectTaskDir, \%variables);
	}
	# here we don't have a succesful calcuations at all
	else {		
		$class->reportOnFailureToWebServer($htmlFailureText,$shouldNotContactSysAdmin);
	}
}
1;