#!/usr/bin/perl -w
# dvir 2010-01-06
# checking that all the running tasks are indeed running
# so that there isn't a task that is stuck even though it's actually finished, 
# and therefore, not allowing new tasks to enter.
use strict;

use FindBin ();
use lib "$FindBin::Bin";

use Configuration qw(%configfile);
use General;
use Analysis;
use JobStatus;

# this lock, since the script run in crontab (once in _ minutes)
# and we don't want that 2 scripts will run at the same time
open (CHECK_TASKS_LOCK, ">$Configuration::CHECK_TASKS_SCRIPT_LOCK") || &General::reportOnErrorReleaseSemaphoresAndDie('checkTasks.pl',`pwd`, sprintf($Configuration::ERR_MSG_FAILED_OPENING_FILE, $Configuration::CHECK_TASKS_SCRIPT_LOCK ,$!));
# if the lock failes, than another process is holding the lock, this is an error, since no more than 1 process should run checkTasks.pl at a time.
flock(CHECK_TASKS_LOCK, $Configuration::EXCLUSIVE_OR_FAILURE_LOCK)  || &General::reportOnErrorReleaseSemaphoresAndDie('checkTasks.pl',`pwd`, sprintf($Configuration::ERR_MSG_FAILED_LOCKING_FILE, $Configuration::CHECK_TASKS_SCRIPT_LOCK ,$!));

# checking all the currently running tasks, if one of them is finished
opendir(TASKDIR,$Configuration::TASKS_BY_RUN_STATUS_DIR);
for my $taskName ( readdir(TASKDIR) ) {
	# if the name is '.' or '..' than this isn't a task name so we move to the next taskName
	if( $taskName =~ /^\./ ) {
		next;
	}
	
	# here $taskName is really a name of the task

	my $projectWorkDir = "$configfile{taskDir}\/$taskName\/workDir";
	open TASK_LOCK_FILE, ">$projectWorkDir\/$Configuration::TASK_LOCK_FILENAME";
	# because of the lock, than the taskName file won't be deleted
	flock TASK_LOCK_FILE, $Configuration::SHARED_LOCK;
	
	# checking if the task is not to be run at condor.
	# if so, than as long as it runs, it should hold exclusivly the lock
	# and if we got to here than the it doesn't hold the lock.
	# so we should check that it had finish running (not in the running dir)
	if (&General::doesTheTaskRunsLocally($projectWorkDir)) {
		# checking if the task is still marked as running
		if (! &General::doesTheTaskIsMarkedAsRunning($taskName)) {
			# this should have been done by the task, at the end of local run.
			# remove the task from the running dir
			&General::markThatTaskEndedRunning($taskName);
			
			# reporting on error
			&General::reportOnErrorReleaseSemaphoresAndDie($0,'',"local task: $taskName hasn't finish running by itself, please check.");
		}
	}	
	else { # the task doesn't run locally, meaning it runs on the condor
		
		# making sure that when we hold the lock, the task is still marked as running
		if (! &General::doesTheTaskIsMarkedAsRunning($taskName) ) {
			# another process marked that the task ended its run
			# so we won't check nothing for this task
			# it's O.K.
		}
		else {
			my $analysis = new Analysis($projectWorkDir);
			
			# here the task is still marked as running
			my $possibleError = '';
			my $doesTheMainDagOfTheTaskIsRunning = &General::doesTheMainDagOfTheTaskIsRunning($taskName, \$possibleError);
			
			# if the main dag of the task isn't running, than the task is stucked at running status, although it acutally finished.
			# (the task isn't listed at the condor_q (doesn't has a running dag))
			if (! $doesTheMainDagOfTheTaskIsRunning) {
				sleep 5;    #Omer - sleep for 5 seconds and check again, because apparently sometimes condor_q returns false results.
				$doesTheMainDagOfTheTaskIsRunning = &General::doesTheMainDagOfTheTaskIsRunning($taskName, \$possibleError);
				if (! $doesTheMainDagOfTheTaskIsRunning) {			
					# informing the programmer about a condor main dag of a task that had ended its run, but
					# didn't marked it
					&General::reportOnError('checkTasks.pl',$taskName,"The main DAG of the task doesn\'t exists at condor,\nbut the task was still marked as running\.\nThe task was marked as finished and the counters were updated.\nPlease check why the task wasn't mark BY ITSELF as ended running.\nPossible Error in submission of the DAG: $possibleError");
										
					my $errorText = '';
					if (&JobStatus::doesJobHasA_TimedOut_Indication($projectWorkDir)) {
						$errorText = 'Time Out: Analysis had exceeded time limt.';
					}
					
					# only if there isn't a report file with a meaningful information,
					# than we will create a report file of our own.
					if (! $analysis->doesAReportExists()) {
						$analysis->reportOnFailureToWebServer($errorText);
					}
								
					#Remove the task
					&General::decreaseCurrentRunningCounter(`pwd`);
					&General::markThatTaskEndedRunning($taskName);
				}
			} 
			
			# if the task is listed at the condor_q (has a running dag)
			if ($doesTheMainDagOfTheTaskIsRunning) {
				# checking if the analysis is haplotyping
				if (-e "$projectWorkDir\/$Configuration::HAPLOTYPING_ANALYSIS_FILENAME") {
					# checking that the task havn't run too much time.
					my $currWindowEndTime = $analysis->getCurrentWindowEndTime();
					if ($currWindowEndTime < time) {
						my $taskMainDagId = &General::getClusterIdOfMainDagOfTask($taskName);
				
						my $tasksDir = $configfile{taskDir};
						my $command = "$Configuration::CONDOR_BIN_DIR\/condor_rm $taskMainDagId >> $tasksDir\/$taskName\/workDir\/$Configuration::CONDOR_DAG_REMOVING_OF_MAIN_TASK_OUTPUT_FILENAME 2>> $tasksDir\/$taskName\/workDir\/$Configuration::CONDOR_DAG_REMOVING_OF_MAIN_TASK_ERROR_FILENAME";
						system($command);
						my $ret = $? >> 8;
						if ($ret != 0) {
							&General::reportOnErrorReleaseSemaphoresAndDie($0,'',"failed to preform : $command");
						}
						
						# marking the task as timeout
						&JobStatus::markThatJobTimedOut($projectWorkDir);
						
						# updating the user about a timeout
						my $errorText = 'Time Out: Analysis had exceeded time limt.';
						$analysis->reportOnFailureToWebServer($errorText);
									
						#Remove the task
						&General::decreaseCurrentRunningCounter(`pwd`);
						&General::markThatTaskEndedRunning($taskName);						
					}
				}
				
				# dvir 2011/11/16, sending an html update to the web server saying the analysis is running
				$analysis->reportOnRunningToWebServer();
			}
		}
	}
	close TASK_LOCK_FILE;
}
closedir(TASKDIR);

close(CHECK_TASKS_LOCK);