#!/bin/bash

# The script is included in other BOINC-related scripts and, thus, should not contain any 'Main' part

# For debugging purposes only
#set -x

# Script constants
# We do not declare the constants as readable-only in order to enable
# their dynamic reloading during execution of any script that includes them

# Includes
# For $BOINC_SERVER_NAME and ${BOINC_SERVER_ARRAY[@]}
mypath=`dirname $0`

[ -z "$PEDTOOL_ROOT" ] || mypath="$PEDTOOL_ROOT"

[ -e $mypath/CONFIG.SH ]  || { echo "$0 - Failed to find CONFIG.SH in $mypath"; exit -1; } 

. "${mypath}/CONFIG.SH"
PEDTOOL_ROOT="$mypath"
mypath=""
# End of includes

SSH_FLAGS="-x"

# While querying the database, these flags guarantee that the answer is returned in a table format without the table borders,
# the -A flag boosts the starting time of the client, as it instructs the client not to load the interactive information
MYSQL_FLAGS="-A -s"
MYSQL_TABLE_FLAGS=""

MONITORING_SET_TABLE="monitoring_set"
MONITORING_ID_TABLE="monitoring_id"
MONITORING_MAXIMAL_SET_ID_TABLE="monitoring_maximal_set_id"
WORKUNIT_TABLE="workunit"
RESULT_TABLE="result"
HOST_TABLE="host"
USER_TABLE="user"
CLASSAD_TABLE="classad"

# Status code names
FILLING_STATUS="Filling"
MONITORING_STATUS="Monitoring"
PROGRESS_STATUS="Progress"
FINISHED_STATUS="Finished"
FAILED_STATUS="Failed"
CONDOR_STATUS="Condor"
PURGE_STATUS="Purge"
# End of status code names

# Condor job statuses
CONDOR_JOB_REMOVED="3"
CONDOR_JOB_COMPLETED="4"
CONDOR_JOB_HELD="5"
# End of Condor job statuses

# BOINC constants (from $BOINC_ROOT/db/boinc_db.h)
RESULT_SERVER_STATE_UNSENT=2
RESULT_SERVER_STATE_OVER=5
RESULT_OUTCOME_DIDNT_NEED=5
WU_ERROR_CANCELLED=16
WU_FILE_DELETE_DONE=2
# End of BOINC constants

ZERO_RESULTS_ANSWER="Found 0 results"
PROGRESS_LOG_NAME="progress.log"
NON_BOINC_OUTPUTS_FILE_NAME="non_boinc_outputs.txt"
BOINC_CONDOR_NODE_NAME="Parallel_single_boinc_condor.sub"
CONDOR_NODE_NAME="Parallel_single_condor.sub"
NEW_CLASSADS_FILENAME="new_classads"
SUBMIT_EXTENSION="submit"
WU_TEMPLATE_FILE_NAME="superlink_wu"
NO_DYNAMIC_BUNDLE_FILE="no_change_bundle_size"
RUNNING_TIME_ESTIMATE_FILE="single_job_runtime_estimate"

# Classad attributes
RANK="Rank"
SCHEDULER_REQUIREMENTS="SchedulerRequirements"
FEEDER_REQUIREMENTS="FeederRequirements"
CONCURRENCY="Concurrency"
TAIL="Tail"
BOINC_PRIORITY="BoincPriority"
JOBS_LEFT="JobsLeft"
DEADLINE="Deadline"
# Classad attributes
# Classad defaults
DEFAULT_RANK=0
DEFAULT_FEEDER_REQUIREMENTS="false"
DEFAULT_SCHEDULER_REQUIREMENTS="true"
DEFAULT_CONCURRENCY=1
DEFAULT_TAIL="false"
DEFAULT_BOINC_PRIORITY=0
DEFAULT_JOBS_LEFT=0
DEFAULT_DEADLINE=259200
#DEFAULT_CLASSAD='<c><a n=\"Rank\"><i>0</i></a><a n=\"FeederRequirements\"><b v=\"t\"/></a>'\
#'<a n=\"SchedulerRequirements\"><b v=\"t\"/></a><a n=\"Concurrency\"><i>0</i></a></c>'
# End of classad defaults
# File extension for outputs of outliers' recomputations
OUTLIERS_EXTENSION="outliers"
# End of outliers' analysis-related

# End of script constants

# Auxiliary functions

# Creates an instance of the Boinc node submission file template
# Arguments: $1 - extension of the log/error/output files (default: no extension)
# Example  : instantiate_boinc_node_template ".1" will produce log/error/output files with .1 extension
instantiate_boinc_node_template() {
	local extension=$1

    sed_cwd=`pwd | sed 's/\//\\\\\\//g'`

    cat $SUBFILES/$BOINC_CONDOR_NODE_NAME.tmpl | sed "s/_PEDTOOL_SUBSTITUTE_WORKDIR_/$sed_cwd/" | \
												 sed "s/_PEDTOOL_SUBSTITUTE_ERROR_/error\.boinc_node${extension}/" | \
    											 sed "s/_PEDTOOL_SUBSTITUTE_OUTPUT_/output\.boinc_node${extension}/" | \
    											 sed "s/_PEDTOOL_SUBSTITUTE_LOG_/log\.boinc_node${extension}/" > $BOINC_CONDOR_NODE_NAME
}

# Executes SSH command, passed via input, on the BOINC front-end server
# Exit code: 0 - upon the success, 1 - upon a failure
# Remark   : reads a command from the standard input. This is done in order to ease the use of the function
# 			 with complex commands, which include reserved keywords (such as quotes or double-quotes)
# Example  :
# 	execute_ssh_frontend << EOF
#                . \$HOME/.bash_profile;
#                mysql $MYSQL_TABLE_FLAGS -e "insert into $MONITORING_ID_TABLE (set_id, id) values ($set_id, $job_id)";
# EOF
execute_mysql_command_from_file()
{
  local filename=$1
  [ -e $filename ] || return 1

  local frontend=$(boinc_frontend)
  local command="mysql ${MYSQL_TABLE_FLAGS}"
  local exit_code=0

    for attempt in $(seq 1 ${BOINC_ATTEMPTS})
    do
        cat $filename | ssh $SSH_FLAGS $frontend "$command"
        exit_code=$?

        [[ ${exit_code} == 0 ]] && return 0 #Die "SSH command on ${frontend} failed, exit code ${exit_code}" 
        sleep ${BOINC_SLEEP_TIME}
    done

    return ${exit_code}

}

execute_ssh_frontend() {
	local frontend=$(boinc_frontend)
	local command=$(tee)
	local exit_code=0

	for attempt in $(seq 1 ${BOINC_ATTEMPTS})
	do
		ssh $SSH_FLAGS $frontend "$command"
		exit_code=$?

		[[ ${exit_code} == 0 ]] && return 0 #Die "SSH command on ${frontend} failed, exit code ${exit_code}" 
		
		sleep ${BOINC_SLEEP_TIME} 
	done

	return ${exit_code}
}

# Reads jobs number of the depth that is greater or equal to the depth, specified as an argument
# The format of the jobs numbers map file is as follows:
# depth overall_number_of_jobs complexity
read_depth_jobs_number() {
    local jobs_numbers_file_path=$1
	local depth=$2
	local current_depth jobs_number

    while read line
    do
        # Ignoring blank lines
        [[ "x$line" == "x" ]] && continue

		current_depth=$(echo $line | cut -d' ' -f1)

		if [[ $current_depth -ge $depth ]]
		then
			jobs_number=$(echo $line | cut -d' ' -f2)
			echo $jobs_number
			return 0
		fi
    done < "$jobs_numbers_file_path"

	return 1
}

# Returns depth and its corresponding jobs number that is greater or equal to the jobs number, specified as an argument
read_jobs_number_depth() {
	local jobs_numbers_file_path=$1
	local jobs_number=$2
	local current_jobs_number current_depth

	while read line
	do
		# Ignoring blank lines
		[[ "x$line" == "x" ]] && continue

		current_jobs_number=$(echo $line | cut -d' ' -f2)
		current_depth=$(echo $line | cut -d' ' -f1)

		[[ $current_jobs_number -ge $jobs_number ]] && echo $current_jobs_number $current_depth && return 0
	done < "$jobs_numbers_file_path"

	# If the specified jobs number is greater than jobs numbers of all possible depths, return the greatest jobs number
	echo $current_jobs_number $current_depth
}

# The format of the parameters file is as follows:
# set_id overall_number_of_jobs depth working_directory
#
# Remark: all fields are mandatory but valid routes
read_parameters_file() {
	local parameters_file_path=$1
	local set_id_counter=0

	while read line
	do
		# Ignoring blank lines
		if [[ "x$line" == "x" ]]
		then
			continue
		fi

		set_ids[$set_id_counter]=$(echo $line | cut -d' ' -f1)
		jobs_numbers[$set_id_counter]=$(echo $line | cut -d' ' -f2)
		depths[$set_id_counter]=$(echo $line | cut -d' ' -f3)
		working_directories[$set_id_counter]=$(echo $line | cut -d' ' -f4)

		#valid_routes[$set_id_counter]=$(echo $line | cut -d' ' -f5-)

		let set_id_counter++
	done < "$parameters_file_path"
}

write_parameters_file() {
	local parameters_file_path=$1
	local set_id_counter=0
	rm -f "$parameters_file_path.new"
	for set_id in ${set_ids[@]}
	do		
		#echo "$set_id ${jobs_numbers[$set_id_counter]} ${depths[$set_id_counter]} ${working_directories[$set_id_counter]} ${valid_routes[$set_id_counter]}" \
		#		>> "$parameters_file_path.new"
		echo $set_id ${jobs_numbers[$set_id_counter]} ${depths[$set_id_counter]} ${working_directories[$set_id_counter]} >> "$parameters_file_path.new"

		let set_id_counter++
	done

	mv "$parameters_file_path.new" "$parameters_file_path"
}

# Rounds the positive argument up
ceiling()
{
	local number=$1 integer_part floating_part sign

	[[ ! ( "$number" =~ "\." ) ]] && echo $number && return 0

	local integer_part=${number/.*}
	local floating_part=${number/*.}
	local sign=${number:0:1}

	[[ -z "$integer_part" ]] && integer_part=0

	# Removing multiple zeros from the floating part in order to be able to compare it with 0 as a string
	floating_part=$(echo "0.$floating_part + 0" | bc -l)

	# Positive numbers
	if [[ "x$sign" != "x-" ]]
	then
		[[ $floating_part == 0 ]] && echo $integer_part
		[[ $floating_part != 0 ]] && echo $(( $integer_part + 1 ))
	# Negative numbers
	else
		echo "$integer_part"
	fi
}

# Returns the maximum of two integers
max()
{
	local first=$1; 
	local second=$2; 

	[[ $first -gt $second ]] && echo $first && return 0; 

	echo $second;
}

# Returns the minimum of two integers
min()
{
	local first=$1
	local second=$2

	[[ $first -gt $second ]] && echo $second && return 0;

	echo $first;
}

# Evaluates the arithmetic expression, given as the argument
evaluate_arithmetic()
{
	local expression=$1

	echo $(echo "scale=9;$expression" | bc -l)
}

# Returns the name of the BOINC-frontend server
# Remark: in order for the function to work properly, the configuration file (CONFIG.SHARED.SH) has to be included in the caller script
boinc_frontend()
{
	local task_directory=${1:-.}

	[ -f "$task_directory/$BOINC_SERVER_NAME" ] && cat $BOINC_SERVER_NAME && return 0

	echo ${BOINC_SERVER_ARRAY[0]}
}

# Increases the maximal set id, known to the database, by the given number of sets
update_maximal_set_id()
{
#set -x
	local sets_number=$1
	local mysql_execution_file=maximal_set_id.sql

	echo "set autocommit=0; start transaction; \
			create table if not exists $MONITORING_MAXIMAL_SET_ID_TABLE \
			( id int not null comment 'The maximal set id, known to the database' ) \
                              type=InnoDB comment='Contains the maximal set id';\
			lock tables $MONITORING_MAXIMAL_SET_ID_TABLE as $MONITORING_MAXIMAL_SET_ID_TABLE write, $MONITORING_MAXIMAL_SET_ID_TABLE as mmsi read; \
			insert into $MONITORING_MAXIMAL_SET_ID_TABLE select (case when count(id) = 0 then $sets_number when 1 then id + $sets_number end) id \
                                             			 from $MONITORING_MAXIMAL_SET_ID_TABLE as mmsi; \
			set @maximal_set_id = (select max(id) from $MONITORING_MAXIMAL_SET_ID_TABLE as mmsi); \
			delete from $MONITORING_MAXIMAL_SET_ID_TABLE where id <> @maximal_set_id; \
			commit; \
			unlock tables;" > $mysql_execution_file
	local frontend=$(boinc_frontend)

	#ssh $SSH_FLAGS $frontend << EOF 

	[ -e $PWD/$mysql_execution_file ] || return -1;
    execute_mysql_command_from_file $mysql_execution_file

#	 execute_ssh_frontend << EOF
#		 mysql $MYSQL_TABLE_FLAGS 
#EOF
	exit_code=$?
	[[ ${exit_code} != 0 ]] && return ${exit_code}
#	rm -f $mysql_execution_file

	# Returning the first set id to use
	maximal_set_id=$(execute_ssh_frontend << EOF
		mysql $MYSQL_TABLE_FLAGS -e "select id - $sets_number from $MONITORING_MAXIMAL_SET_ID_TABLE" | tail -1
EOF
	)
	exit_code=$?
	[[ ${exit_code} != 0 ]] && return ${exit_code}
	echo ${maximal_set_id}
#set +x
}

find_array_index()
{
	local -a set_ids=( $1 )
	local sought_set_id=$2
	local index=0

	for set_id in ${set_ids[@]}
	do
		[[ $set_id == $sought_set_id ]] && echo $index && return 0
		let index++
	done

	return 1
}

numeric_sort_by_key()
{
	local file_to_sort=$1
	local sorting_key=$2
	
	local temporary_file=$(mktemp)
	sort -n "${file_to_sort}" -k ${sorting_key} | uniq  > "${temporary_file}"
	mv -f "${temporary_file}" "${file_to_sort}"
}

safe_condor_submit()
{
	local submit_file_path=$1

	for attempt in $(seq 1 ${BOINC_ATTEMPTS})
	do
		condor_submit $submit_file_path

		[ $? == 0 ] && return 0
	done
	return 1
}

change_sets_status()
{
	local -a set_ids=( $1 )
	local new_status=$2

	local set_ids_translated=$(echo "${set_ids[@]}" | tr ' ' ',')
	execute_ssh_frontend << EOF
		. \$HOME/.bash_profile; \
		mysql $MYSQL_TABLE_FLAGS -e "update $MONITORING_SET_TABLE set status='$new_status' where id in ($set_ids_translated)";
EOF
	return $?
}

change_jobs_status()
{
	local -a set_ids=( $1 )
	local -a job_ids=( $2 )
	local new_status=$3

	local sql_command=""
	local set_id_counter=0

	for set_id in ${set_ids[@]}
	do
		sql_command="update $MONITORING_ID_TABLE set status = '$new_status', mod_time = unix_timestamp() where id = ${job_ids[$set_id_counter]} and set_id = $set_id; $sql_command"
		let set_id_counter++
	done

	execute_ssh_frontend << EOF
		. \$HOME/.bash_profile; \
		mysql -e "$sql_command";
EOF
	return $?
}
# Updates the number of restarts
restart_boinc_monitoring_job()
{
	local set_id=$1
	local job_id=$2
	local new_status=$3

	execute_ssh_frontend << EOF
		. \$HOME/.bash_profile; \
		mysql $MYSQL_TABLE_FLAGS -e "update $MONITORING_ID_TABLE set restarts = restarts + 1, status = '$new_status', mod_time = unix_timestamp() \
					     where id = $job_id and set_id = $set_id"
EOF
	return $?
}
# Restarts the specified Condor job and updates the number of restarts
restart_condor_job_monitoring()
{
	local set_id=$1
	local directory=$2
	local job_id=$3
	local depth=$4
	local priority=$5
	local restarts=$6

	restart_boinc_monitoring_job ${set_id} ${job_id} "${CONDOR_STATUS}"
	# Do not submit a new BOINC job if the status has not been changed yet
	exit_code=$?
	[[ ${exit_code} != 0 ]] && return ${exit_code}
	# Restarting another job with the same parameters
	report_progress_log "${directory}" ${job_id} "Restart" "restarted ${restarts} time"
    submit_condor_job "${set_id}" "${directory}" "${job_id}" "${depth}" "${priority}"
}

finish_set()
{
	local set_id=$1
	local status=$2

#	ssh $SSH_FLAGS $FRONTEND << EOF
	execute_ssh_frontend << EOF
	. \$HOME/.bash_profile; \
	mysql $MYSQL_TABLE_FLAGS -e "set autocommit=0; start transaction; update $MONITORING_ID_TABLE set status='$FINISHED_STATUS', mod_time = unix_timestamp() \
	     			 where set_id = $set_id; update $MONITORING_SET_TABLE set status='$status' where id = $set_id; commit;"
EOF
	return $?
}

report_progress_log()
{
	local directory=$1
	local job_id=$2
	local message_type=$3
	local message_body=$4

	now_timestamp=`date +'%Y-%m-%d %H:%M:%S.0000'`

	echo "$now_timestamp $message_type: job #$job_id, $message_body" >> "$directory/$PROGRESS_LOG_NAME"
	Log "Job no. $job_id from $directory: $message_type ($message_body)"
}

exclude_failed_jobs_total()
{
	local working_directory=$1

	local gather_total_jobs_file="${working_directory}/${GATHER_TOTAL_JOBS_FILE}"
	local total_jobs_file="${working_directory}/${TOTAL_JOBS_FILE}"

	[ -f "${gather_total_jobs_file}" ]         && local failed_jobs=$(cat "${gather_total_jobs_file}")
	[[ ! ( -f "${gather_total_jobs_file}" ) ]] && local failed_jobs=$(cat "${total_jobs_file}")

	local new_total_jobs=$(( $( cat "${TOTAL_JOBS_FILE}" ) - ${failed_jobs} ))

	echo ${new_total_jobs} > "${TOTAL_JOBS_FILE}"
}

create_monitoring_tables()
{
	#ssh $SSH_FLAGS $FRONTEND << EOF
	execute_ssh_frontend << EOF
		. \$HOME/.bash_profile; \
		mysql $MYSQL_TABLE_FLAGS -e "create table if not exists $MONITORING_SET_TABLE \
		 (id     int not null comment 'Set of monitoring job ids', \
		  primary key (id), \
		  status varchar(10) binary default '$FILLING_STATUS' comment 'Status of the jobs set in the monitoring program', \
		  directory text not null comment 'Working directory of the set, containing datafile, pedfile and order file', \
		  priority int not null comment 'Jobs are started with this priority', \
		  task_id int not null default '-1' comment 'Task id of the monitoring set', \
		  unique index directory_index using btree (directory(256)), \
  		  constraint task_id_fkey foreign key (task_id) references ${CLASSAD_TABLE}(id) \
		  	on delete cascade on update no action) type=InnoDB \
		  comment 'Contains ids of sets of BOINC jobs'; \
	# Monitoring job entry is created only when it's located in some jobs set
		 create table if not exists $MONITORING_ID_TABLE \
		 (id       int not null comment 'Job id', \
		  set_id   int not null comment 'Id of set, to which the job pertains', \
		  restarts int not null comment 'Number of restarts for this job id', \
		  status varchar(10) binary default '$PROGRESS_STATUS' comment 'Status of the jobs id in the monitoring program', \
		  mod_time int comment 'Last time, the monitoring status changed', \
		  index status_hash_index using hash (status), \
		  constraint set_id_fkey foreign key(set_id) references $MONITORING_SET_TABLE(id) \
			on delete cascade on update no action) type=InnoDB \
		  comment 'Contains relation between job ids and set ids'"
EOF
	return $?
}

restart_condor_job()
{
	#local condor_job=$1
	local set_id=$1
	local directory=$2
	local job_id=$3
	local depth=$4
	local priority=$5

	# Trying to remove the old (stuck) Condor job
	# We cannot wait for the purger to wake up, we have to do it right now, otherwise the new Condor job will be removed along with the old one
	condor_qedit -constraint "MonitoringSetId == \"${set_id}\" && MonitoringId == \"${job_id}\"" LeaveJobInQueue false
	condor_rm -constraint "MonitoringSetId == \"${set_id}\" && MonitoringId == \"${job_id}\""
	#echo "MonitoringSetId+MonitoringId string_pair ${set_id} ${job_id}" >> "${PURGE_DIR}/${TASKID}"

	# Restarting another job with the same parameters
	submit_condor_job "${set_id}" "${directory}" "${job_id}" "${depth}" "${priority}"
}
# Return an estimate of the floating point operations per second for the specified directory
# Arguments: $1 - directory of the set, whose floating point operations per second are estimated
floating_point_estimate()
{
	local directory="$1"

	local rsc_fpops_est=0;
	local estimate_file="${directory}/${RUNNING_TIME_ESTIMATE_FILE}"

    rsc_fpops_est=$(execute_ssh_frontend << EOF
        . \$HOME/.bash_profile; echo "<boinc>"\$(cat \$PROJECT_ROOT/templates/${WU_TEMPLATE_FILE_NAME})"</boinc>" | \$PROJECT_ROOT/bin/parse_xml rsc_fpops_est
EOF
    )
    exit_code=$?
    [[ ${exit_code} != 0 ]] && Warn "SSH command failed to parse wu template (exit code ${exit_code}), unable to create BOINC jobs" && return 1

    [[ -s "${estimate_file}" ]] &&
    {
            time_estimate=$(cat "${estimate_file}");
            rsc_fpops_est_override=0;
            rsc_fpops_est_override=$(echo "${time_estimate} * 3000000000" | bc -l);

            [[ -z "$rsc_fpops_est_override" ]] || rsc_fpops_est=${rsc_fpops_est_override};
            Log "Runtime override found - ${rsc_fpops_est}";

    }
	# 'rsc_fpops_est' might be expressed in the '<mantissa> * e<exponent>' format, which is well handled by 'atof'
    rsc_fpops_est=$("${PEDTOOL_ROOT}/atof" ${rsc_fpops_est})
	echo ${rsc_fpops_est}
}

valid_routes_number()
{
	local set_id=$1

	echo $(wc -l "${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}" | cut -f1 -d' ')
}

# Cleans all the information, pertaining to the specified monitoring set ids 
# Arguments: $1 - a space-separated string, containing all the monitoring set ids to purge
#            $2 - unique task id for preventing races between various scripts that call the function
clean_sets()
{
set -x
	local -a set_ids=( $1 )
	local task_id=$2
	#local frontend=$(boinc_frontend)

	local set_ids_list=$(echo ${set_ids[*]} | tr ' ' ',')
	# Appending new records for the purger
	echo "TaskId string ${task_id}" >> "${PURGE_DIR}/${task_id}"
	# Making sure the Condor purger is called
	"${PEDTOOL_ROOT}/Purger.sh"
	
	# Actually it is redundant to update the status of each job id to be "Purge", since the database purger will
	# anyway pass though the monitoring sets' table only and will delete entries from it, which will imply the deletion
	# of entries from the monitoring ids' table by cascade. We are doing it only for the sake of completeness.
	# We are calling the database purger to ensure the deletion of monitoring entries, we do it with a big number of minimal age
	# days to prevent the deletion of data that is not considered as obsolete by the definition of the project, we are connecting to
	execute_ssh_frontend << EOF
		. \$HOME/.bash_profile; \
		mysql $MYSQL_TABLE_FLAGS -e "set autocommit=0; start transaction; \
					     update $MONITORING_ID_TABLE set status='$PURGE_STATUS' where set_id in ($set_ids_list); \
			  		     update $MONITORING_SET_TABLE set status='$PURGE_STATUS' where id in ($set_ids_list); \
						 update result, workunit, $MONITORING_SET_TABLE ms set result.server_state = $RESULT_SERVER_STATE_OVER, \
																			   result.outcome = $RESULT_OUTCOME_DIDNT_NEED \
						 where result.server_state = $RESULT_SERVER_STATE_UNSENT and result.workunitid = workunit.id and \
							   ms.id in (${set_ids_list}) and \
						 	   workunit.xml_doc like CONCAT('%<location>', ms.directory, '</location>%'); \
						 update workunit, $MONITORING_SET_TABLE ms set workunit.error_mask = workunit.error_mask | $WU_ERROR_CANCELLED, \
								workunit.transition_time = UNIX_TIMESTAMP() - 1 where ms.id in (${set_ids_list}) and \
								workunit.xml_doc like CONCAT('%<location>', ms.directory, '</location>%'); \
						 commit;";
EOF
	local ssh_exit_code=$?
 execute_ssh_frontend << EOF
        . \$HOME/.bash_profile; \
		cd \$PROJECT_ROOT/bin; \
		./db_purge -min_age_days ${BIG_MIN_NUM_DAYS} -gzip -one_pass; 
EOF
 execute_ssh_frontend << EOF
		. \$HOME/.bash_profile; \
		touch \$PROJECT_ROOT/${NEW_CLASSADS_FILENAME};
EOF
	# Deleting the gradual submission state file to start the submission anew
	rm -f "${BOINC_SUBMISSION_STATE_FILE}"
	# Deleting the bundle size file to start from the initial bundle size at the (possible) next restart
	[ ! -e $NO_DYNAMIC_BUNDLE_FILE ] && rm -f "${BOINC_BUNDLE_SIZE_FILE}"
	# Deleting the classads file to start from the initial bundle size at the (possible) next restart
#	rm -f "${BOINC_CLASSADS}"
	return ${ssh_exit_code}
}

#find_jobs_in_danger()
#{
#	# Finding all the jobs in progress, which have used a certain percent of the allowed restarts. By sorting them in the ascending order we ensure
#	# that the most outdated are to be handled first
#	execute_ssh_frontend << EOF
#		. \$HOME/.bash_profile; \
#		mysql $MYSQL_TABLE_FLAGS -e \
#		   "\P true; \
#			set @comma=''; \
#			set @id_list := ''; \
#			select @id_list := CONCAT(@id_list, @comma, id), @comma := ','  \
#			from (select w.id, count(r.id) count_result, w.max_error_results \
#			 	 from ${WORKUNIT_TABLE} w, ${RESULT_TABLE} r \
#			 	 where w.file_delete_state <> 2 and r.workunitid = w.id \
#				 group by w.id \
#				 order by w.create_time asc) counts \
#			where max_error_results * ${SCHED_MATCH_SAFE_RESTARTS_PERCENT} <= count_result \
#			limit ${SCHED_MATCH_LIMIT}; \
#			set @command := CONCAT('update ${WORKUNIT_TABLE} set create_time = UNIX_TIMESTAMP() where id in (', @id_list, ')'); \
#			prepare statement from @command; \
#			execute statement; \
#			deallocate prepare statement"
#EOF
#	local ssh_exit_code=$?
#  
#	return ${ssh_exit_code}
#}

default_classad()
{
	local classad_text=$(echo "$1" | tr -d '\n')
	local classad_addon=""

	[[ ! ( "${classad_text}" =~ "[\[ ;]${RANK}[ ]*=[^=]" ) ]]                   && classad_addon="${RANK} = ${DEFAULT_RANK}; ${classad_addon}"
	[[ ! ( "${classad_text}" =~ "[\[ ;]${SCHEDULER_REQUIREMENTS}[ ]*=[^=]" ) ]] && \
		classad_addon="${SCHEDULER_REQUIREMENTS} = ${DEFAULT_SCHEDULER_REQUIREMENTS}; ${classad_addon}"
	[[ ! ( "${classad_text}" =~ "[\[ ;]${FEEDER_REQUIREMENTS}[ ]*=[^=]" ) ]]    && classad_addon="${FEEDER_REQUIREMENTS} = ${DEFAULT_FEEDER_REQUIREMENTS}; ${classad_addon}"
	[[ ! ( "${classad_text}" =~ "[\[ ;]${CONCURRENCY}[ ]*=[^=]" ) ]]            && classad_addon="${CONCURRENCY} = ${DEFAULT_CONCURRENCY}; ${classad_addon}"
	[[ ! ( "${classad_text}" =~ "[\[ ;]${TAIL}[ ]*=[^=]" ) ]]                   && classad_addon="${TAIL} = ${DEFAULT_TAIL}; ${classad_addon}"

	echo "${classad_text}" | sed "s/\[/\[ ${classad_addon}/"
}

add_classad()
{
	local task_id=$1
	local classad_text

	[ ! -s "${BOINC_CLASSADS}" ] && classad_text="[]"
	[ -s "${BOINC_CLASSADS}" ]   && classad_text=$(cat "${BOINC_CLASSADS}")
	local classad=$(default_classad "${classad_text}")
	local temporary_classad=$(mktemp)

	echo "${classad}" > "${temporary_classad}"
	
	local classad_xml

	classad_xml=$("${PEDTOOL_ROOT}/xml_to_classad" "${temporary_classad}" classad)

	local exit_code=$?

	[[ ${exit_code} -eq 0 ]] &&
	{
 		classad_xml=$(echo "${classad_xml}" | sed 's/"/\\"/g' | sed 's/\\\./\\\\\\\./g')
		execute_ssh_frontend << EOF
			. \$HOME/.bash_profile; \
			mysql $MYSQL_TABLE_FLAGS -e \
				"insert into ${CLASSAD_TABLE} (id, text) values (${task_id}, '${classad_xml}')"; \
			touch \$PROJECT_ROOT/${NEW_CLASSADS_FILENAME}
EOF
	}
	mv -f "${temporary_classad}" "${BOINC_CLASSADS}"

	return ${exit_code}
}

update_classad()
{
	local task_id=$1
	local classad_text

	[ ! -s "${BOINC_CLASSADS}" ] && classad_text="[]"
	[ -s "${BOINC_CLASSADS}" ]   && classad_text=$(cat "${BOINC_CLASSADS}")
	local classad=$(default_classad "${classad_text}")
	local temporary_classad=$(mktemp)

	echo "${classad}" > "${temporary_classad}"

	local classad_xml

	classad_xml=$("${PEDTOOL_ROOT}/xml_to_classad" "${temporary_classad}" classad)

	local exit_code=$?

	[[ ${exit_code} -eq 0 ]] &&
	{
		classad_xml=$(echo "${classad_xml}" | sed 's/"/\\"/g' | sed 's/\\\./\\\\\\\./g')
		execute_ssh_frontend << EOF
			. \$HOME/.bash_profile; \
			mysql $MYSQL_TABLE_FLAGS -e \
				"update ${CLASSAD_TABLE} set text = '${classad_xml}' where id = ${task_id}"; \
			touch \$PROJECT_ROOT/${NEW_CLASSADS_FILENAME}
EOF
	}
	mv -f "${temporary_classad}" "${BOINC_CLASSADS}"

	return ${exit_code}
}

bundle_size()
{
	local bundle

	[ -s "${BOINC_BUNDLE_SIZE_FILE}" ] && 
	{
		bundle=$(cat "${BOINC_BUNDLE_SIZE_FILE}")
		echo ${bundle}

		return 0
	}
	bundle=${BOINC_INITIAL_BUNDLE_SIZE}
	echo ${bundle} > "${BOINC_BUNDLE_SIZE_FILE}"
	echo ${bundle}
	#bundle=$(execute_ssh_frontend << EOF
#	. \$HOME/.bash_profile; . \$PROJECT_ROOT/bin/definitions.sh; echo \$BUNDLE 
#EOF
#    )
#	local exit_code=$?
#
#	echo ${bundle}
#	return ${exit_code}
}

active_hosts()
{
	local hosts

	hosts=$(execute_ssh_frontend << EOF
    . \$HOME/.bash_profile; . \$PROJECT_ROOT/bin/definitions.sh; tail -1 \$datafiles_directory/hosts | cut -f3 -d' '
EOF
    )
	local exit_code=$?

	echo ${hosts}
	return ${exit_code}
}

submission_lower_watermark()
{
	[ ! -s "${BOINC_SUBMISSION_LOWER_WATERMARK}" ] && echo ${BOINC_DEFAULT_SUBMISSION_LOWER_WATERMARK} && return 0
	
	local lower_watermark=$(cat "${BOINC_SUBMISSION_LOWER_WATERMARK}")
	echo ${lower_watermark}

	return 0
}

load_submission_state_file()
{
	local length=$1

	local set_id
	local line

	if [ ! -s "${BOINC_SUBMISSION_STATE_FILE}" ]
	then
		set_id=$(head -1 "${BOINC_PARAMETERS_FILE_NAME}"  | awk '{ print $1 }')
		line=0
	else
		set_id=$(head -1 "${BOINC_SUBMISSION_STATE_FILE}" | awk '{ print $1 }')
		line=$(head -1 "${BOINC_SUBMISSION_STATE_FILE}"   | awk '{ print $2 }')
	fi
	# If the corresponding parameters file does not exist, this is the end of our set
	[ ! -f "${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}" ] && return 1

	local routes_number=$(wc -l "${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}" | awk '{ print $1 }')
	local tail_size=$(( ${routes_number} - ${line} ))
	local sequence_file_path="${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}.${SUBMIT_EXTENSION}"
	local bundle
	# The number of workunits depends on the number of routes per bundle, that is why we prepare as much as ${length} * ${bundle} routes 
	bundle=$(bundle_size)
	exit_code=$?
	[[ ${exit_code} != 0 ]] && return 2
	
	tail -${tail_size} "${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}" | head -$(( ${length} * ${bundle} )) > "${sequence_file_path}"
	echo $(pwd)/"${sequence_file_path}"
	echo ${set_id}
	#echo ${line}

	return 0
}

update_submission_state_file()
{
	local length=$1
	local set_id
	local line

	if [ ! -s "${BOINC_SUBMISSION_STATE_FILE}" ]
	then
		set_id=$(head -1 "${BOINC_PARAMETERS_FILE_NAME}"  | awk '{ print $1 }')
		line=0
	else
		set_id=$(head -1 "${BOINC_SUBMISSION_STATE_FILE}" | awk '{ print $1 }')
		line=$(head -1 "${BOINC_SUBMISSION_STATE_FILE}"   | awk '{ print $2 }')
	fi
	# If the corresponding parameters file does not exist, this is the end of our set
	[ ! -f "${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}" ] && return 1

	local bundle=$(bundle_size)
	exit_code=$?
    [[ ${exit_code} != 0 ]] && return 2

	#let line+=$(wc -l "${sequence_file_path}" | awk '{ print $1 }' )
	let line+=${length}*${bundle}
	local routes_number=$(wc -l "${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}" | awk '{ print $1 }')
	# The line number might be equal to the number of lines in the parameters file minus one, hence the non-strict inequality
	[[ ${line} -ge ${routes_number} ]] &&
	{
		local set_id_line=$(grep -n "^${set_id} " "${BOINC_PARAMETERS_FILE_NAME}" | awk 'BEGIN { FS=":" } { print $1 }')
		let set_id_line++

		if [[ ${set_id_line} -gt $(wc -l "${BOINC_PARAMETERS_FILE_NAME}" | awk '{ print $1 }' ) ]]
		then
			set_id=end
		else
			set_id=$(head -${set_id_line} "${BOINC_PARAMETERS_FILE_NAME}" | tail -1 | awk '{ print $1 }')
		fi
		line=0
	}
	echo ${set_id} ${line} > "${BOINC_SUBMISSION_STATE_FILE}"
}

unsubmitted_jobs()
{
	local set_id_line=0
	local jobs=0

	[ -s "${BOINC_SUBMISSION_STATE_FILE}" ] &&
	{
		local set_id=$(head -1 "${BOINC_SUBMISSION_STATE_FILE}"  | awk '{ print $1 }')
		local route=$(head  -1 "${BOINC_SUBMISSION_STATE_FILE}"  | awk '{ print $2 }')
		# If the corresponding parameters file does not exist, this is the end of our set
		[ ! -f "${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}" ] && echo 0 && return 0

		set_id_line=$(grep -n "^${set_id} " "${BOINC_PARAMETERS_FILE_NAME}" | awk 'BEGIN { FS=":" } { print $1 }')

		local routes_number=$(wc -l "${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}" | awk '{ print $1 }')
		local route_line=$(grep -n "^${route}\$" "${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}" | awk 'BEGIN { FS=":" } { print $1 }')
		let jobs+=$(( ${routes_number} - ${route_line} + 1))
	}
	local line_counter=-1

	while read line
	do
		let line_counter++

		[[ ${line_counter} -lt ${set_id_line} ]] && continue

		local set_id=$(echo ${line} | awk '{ print $1}')
		let jobs+=$(wc -l "${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}" | awk '{ print $1 }')
	done < "${BOINC_PARAMETERS_FILE_NAME}"

	echo ${jobs}
}

update_bundle_size()
{

	[ -e "${NO_DYNAMIC_BUNDLE_FILE}" ] && return;
	[ -z "$1" ] || { echo $1 > "${BOINC_BUNDLE_SIZE_FILE}"; return; }
	local jobs_left=$(unsubmitted_jobs)
	local bundle=$(bundle_size)

	[ $bundle -eq 1 ] && return;

	while [[ $(( ${jobs_left} - ${bundle} * ${BOINC_SUBMISSION_FADING_FACTOR} * ${BOINC_DEFAULT_SUBMISSION_LOWER_WATERMARK} )) -lt 0 ]]
	do
		local new_bundle=$(( ${jobs_left} / ${BOINC_SUBMISSION_FADING_FACTOR} / ${BOINC_DEFAULT_SUBMISSION_LOWER_WATERMARK} ))
		new_bundle=$(min $new_bundle $bundle)
		bundle=$(max ${new_bundle} 1)
		# Bundle size has nowhere to go beneath 1
		[[ ${bundle} -eq 1 ]] && break
	done
	
	#[ $bundle -eq 1 ] && touch ${BOINC_BUNDLE_SIZE_FILE}.tail;
	
	echo ${bundle} > "${BOINC_BUNDLE_SIZE_FILE}"
}

check_watermark()
{
	local task_id=$1

	local unfinished_jobs=$(execute_ssh_frontend << EOF
	. \$HOME/.bash_profile; \
	mysql ${MYSQL_FLAGS} -e "select count(*) \
	from ${RESULT_TABLE} r, ${WORKUNIT_TABLE} w where r.workunitid = w.id and w.task_id = ${task_id} and r.server_state = ${RESULT_SERVER_STATE_UNSENT}"
EOF
	)
    local exit_code=$?
	# The output is defined only when the exit code is 0
	echo ${unfinished_jobs}

	return ${exit_code}
}

are_all_finished()
{
	local -a set_ids=( $1 )

	# Preparing a list for SQL WHERE clause
	local set_ids_list=$(echo $1 | tr ' ' ',')
	local all_sets=${#set_ids[*]}
	local finished_sets=$(execute_ssh_frontend << EOF
				. \$HOME/.bash_profile; \
				mysql $MYSQL_FLAGS -e "select count(*) from $MONITORING_SET_TABLE \
						       where status in ('$FINISHED_STATUS', '$FAILED_STATUS', '$PURGE_STATUS') and \
						       id in ($set_ids_list)"
EOF
	)
	local exit_code=$?
	# If there was an error in the SSH or the database connection, we keep functioning
	if [[ ${exit_code} -eq 0 && ${all_sets} == ${finished_sets} ]]
	then
		echo 1
	else
		echo 0
	fi
}

is_tail_close()
{
#	local -a set_ids=( $1 )
#	# Preparing a list for SQL WHERE clause
#    local set_ids_list=$(echo $1 | tr ' ' ',')
#	local unfinished_ids
#
#	unfinished_ids=$(execute_ssh_frontend << EOF
#				. \$HOME/.bash_profile; \
#				mysql $MYSQL_FLAGS -e "select count(*) from $MONITORING_ID_TABLE \
#                               where status not in ('$FINISHED_STATUS', '$PURGE_STATUS') and \
#                               set_id in ($set_ids_list)"
#EOF
#	)
#	local exit_code=$?
#	[[ ${exit_code} -ne 0 ]] && echo 0 && return
#
	[ -e "${NO_DYNAMIC_BUNDLE_FILE}" ] && echo 0 && return 0

#	[ -e "${BOINC_BUNDLE_SIZE_FILE}.tail" ] && echo 1 && return 0
	
	echo 0 && return 0
#	local hosts
#
#	hosts=$(active_hosts)
#	exit_code=$?
#	[[ ${exit_code} -ne 0 ]] && echo 0 && return
#
#	if [[ $(( ${unfinished_ids} - ${bundle} * ${hosts} )) -le 0 ]]
#	then
#		echo 1
#	else
#		echo 0
#	fi
#	return 0
}

find_debug_info()
{
	local monitoring_id=$1
	local set_id=$2
	local location=$3
	
	#echo "Monitoring table entry:"
	echo ${location} ${monitoring_id}
	execute_ssh_frontend << EOF
	. \$HOME/.bash_profile; \
	workunitid_name=( \$(mysql ${MYSQL_FLAGS} -e \
		"select id, name from ${WORKUNIT_TABLE} \
		 where xml_doc like '%<location>${location}</location><jobs>${monitoring_id},%</jobs>%' or \
			   xml_doc like '%<location>${location}</location><jobs>%,${monitoring_id},%</jobs>%' or \
			   xml_doc like '%<location>${location}</location><jobs>${monitoring_id}</jobs>%' or \
			   xml_doc like '%<location>${location}</location><jobs>%,${monitoring_id}</jobs>%'" ) ); \
	echo \${workunitid_name[*]}; \
	workunitid=\${workunitid_name[0]}; \
	name=\${workunitid_name[1]}; \
	[ -z "\$workunitid" ] && { echo "no workunit found for location: ${location} monitoring ID ${monitoring_id} - \${workunitid} "; exit 0; } ;\
	mysql -A -t -e \
		"select 'Monitoring table entry:' as '' from dual; \
		 select * from ${MONITORING_ID_TABLE} where id = ${monitoring_id} and set_id = ${set_id}; \
		 select 'Monitoring set entry:' as '' from dual; \
		 select * from ${MONITORING_SET_TABLE} where id = ${set_id}; \
		 select 'Workunit table entry:' as '' from dual; \
		 select id, FROM_UNIXTIME(create_time) as creation_time, canonical_resultid, \
				FROM_UNIXTIME(transition_time) as transition_time, \
				( case error_mask \
					when 0 then 'Success' \
					when 1 then 'Could not send result' \
					when 2 then 'Too many errors' \
					when 4 then 'Too many successes' \
					when 8 then 'Too many totals' \
					when 16 then 'Cancelled' \
					else 'Combination of errors' end) as error, \
				( case file_delete_state \
					when 0 then 'Init' \
					when 1 then 'Ready' \
					when 2 then 'Done' \
					when 3 then 'Error' end ) as file_delete_state, \
				( case assimilate_state \
					when 0 then 'Init' \
					when 1 then 'Ready' \
					when 2 then 'Done' end ) as assimilate_state, \
				priority,
				mod_time
		 from ${WORKUNIT_TABLE} \
		 where id = \$workunitid; \
		 select 'Result table entries:' as '' from dual; \
		 select FROM_UNIXTIME(r.create_time) as create_time, \
				( case r.server_state \
					when 1 then 'Inactive' \
					when 2 then 'Unsent' \
					when 4 then 'In progress' \
					when 5 then 'Over' end ) as server_state, \
				( case r.outcome \
					when 1 then 'Success' \
					when 2 then 'Could not send' \
					when 3 then 'Client error' \
					when 4 then 'No reply' \
					when 5 then 'Did not need' \
					when 6 then 'Validation error' \
					when 7 then 'Detached' end ) as outcome, \
				( case r.client_state \
					when 0 then 'New' \
					when 1 then 'Downloading' \
					when 2 then 'Computing' \
					when 3 then 'Error' \
					when 4 then 'Uploading' \
 					when 5 then 'Done' \
					when 6 then 'Aborted' end ) as client_state, \
				h.domain_name as host, \
				u.name as user, \
				FROM_UNIXTIME(r.report_deadline) as deadline, \
				FROM_UNIXTIME(r.sent_time) as sent_time, \
				FROM_UNIXTIME(r.received_time) as received_time, \
				r.cpu_time, \
				( case r.file_delete_state \
                    when 0 then 'Init' \
                    when 1 then 'Ready' \
                    when 2 then 'Done' \
                    when 3 then 'Error' end ) as file_delete_state, \
				( case r.validate_state \
					when 0 then 'Init' \
					when 1 then 'Valid' \
					when 2 then 'Invalid' \
					when 3 then 'Skipped' \
					when 4 then 'Inconclusive' \
					when 5 then 'Too late' end ) as validate_state \
		 from ${RESULT_TABLE} r, ${USER_TABLE} u, ${HOST_TABLE} h \
		 where workunitid = \$workunitid and r.userid = u.id and r.hostid = h.id"; \
	cd \$PROJECT_ROOT/upload; \
	echo '+----------------+'; \
	echo '| Found outputs: |'; \
	echo '+----------------+'; \
	find . -name "\${name}_*" -exec cat {} \; -print; \
	echo '+---------------------------+'; \
	echo '| Ln-likelihood calculated: |'; \
	echo '+---------------------------+'; \
	grep -e " \${monitoring_id}\$" "\${location}/output" | cut -f1 -d' '
EOF

}
# Prints out sets with percentage of accomplished jobs
# Arguments   : $1 -- the ids of sets, whose percentages are calculated in the function
# Return value: array of lines, each line containing (1) set id, (2) percentage of accomplished jobs, via standard output
percentage_done()
{
	#local -a set_ids=( $1 )
    local set_ids_list=$(echo "$1" | tr ' ' ',')
	local -a set_ids_percentages

	set_ids_percentages=$(execute_ssh_frontend << EOF
		. \$HOME/.bash_profile; \
		mysql ${MYSQL_FLAGS} -e "select numerator.id, 100 * ifnull(numerator.count_id / denominator.count_id, 0) * (denominator.count_id > ${OUTLIERS_SMALL_SET})
								from 
								(select count(*) count_id, ms.id  
								from ${MONITORING_ID_TABLE} mi , ${MONITORING_SET_TABLE} ms 
								where ms.id = mi.set_id and mi.status in ('${FINISHED_STATUS}', '${FAILED_STATUS}', '${PURGE_STATUS}') and ms.id in (${set_ids_list}) 
								group by ms.id, ms.status) numerator, 
								(select count(*) count_id, ms.id 
								from ${MONITORING_ID_TABLE} mi , ${MONITORING_SET_TABLE} ms 
								where ms.id = mi.set_id and ms.id in (${set_ids_list}) 
								group by ms.id, ms.status) denominator 
								where numerator.id = denominator.id"
EOF
	)
	echo ${set_ids_percentages[@]}
}
# Prints out whether the outliers analysis should be started or not 
# Arguments   : $1 -- a pedfile directory
#               $2 -- a percentage of accomplished jobs
# Return value: success/failure (via standard output)
start_outliers_analysis()
{
	local current_directory="$1"
	local percentage="$2"
	local last_percentage=0

	[ -e ${current_directory}/${OUTLIERS_SUBMITTED} ] && { 
		local last_percentage=$(cat "${current_directory}/${OUTLIERS_SUBMITTED}");
	}
	

	last_percentage=${last_percentage:-0}
	# The analysis has already been started
	#[ -e "${current_directory}/${OUTLIERS_SUBMITTED}" ] && echo 0 && return
	# The percentage is too low
	#[[ $(expr ${percentage} \< ${OUTLIERS_ANALYSIS_PERCENTS} ) == 1 ]] && echo 0 && return
	[[ $(echo "${percentage} < ${OUTLIERS_ANALYSIS_PERCENTS}" | bc -l ) == 1 ]] && echo 0 && return
	# Too little workunits were accomplished since the last analysis
	[[ $(echo "${percentage} - ${last_percentage} < ${OUTLIERS_ANALYSIS_DELTA}" | bc -l ) == 1 ]] && echo 0 && return

	echo 1
}

default_priority()
{
	local task_number="$1"

	echo $(( ${DEFAULT_PRIORITY} - ${task_number} ))
}
# End of auxiliary functions

