#!/bin/bash
basepath=`dirname $0`
. ${basepath}/JobUtils.sh
RedirOutToTaskLog
#set -x

# this script assumes presence of one or several ComplexityEstimate files in the data directories
# it calculates the complexity of each subtask by taking the best order, it removes too heavy jobs if
# this is a special run, and finally it decides if the task should run in frontfill or backfill
# It produces several files:
# 
# this will be the file with the complexity map - an input for the PreParallelDag
# $TASK_COMPLEXITY_MAP_FILE
# this is a file containing only directories in the order which will be used for all Parallelization needs in the following scripts
# ${TASK_COMPLEXITY_MAP_FILE}.order
# this is a file containing directories and total sum of tables, to be further sorted by tablesize ( this file is NOT sorted)
# ${TASK_COMPLEXITY_MAP_FILE}.sort
# this is a file containing directories and split sizes, in the original order
# ${TASK_COMPLEXITY_MAP_FILE}.splits

# It is considered "SECOND_ESTIMATE", so it marks itself in all state updates as Second estimate

# decision is in SECOND_ESTIMATE_RESULT_FILE, or no output if can continue


# this routine selects best order out of all orders in the specified directory
selectBestOrder(){
	local orderDir=$1
	local bestComplexity=1000000000000000000000000000000
	local counter=0
	local bestComplexityIdx=0
	while [ -e $orderDir/ComplexityEstimate${counter}.txt ]; do
		local estimateFile=$orderDir/ComplexityEstimate${counter}.txt

		ParseComplexityEstimateFile $estimateFile
		local complexity=$task_ts
		if [ ! -z "$complexity" ]; then
			LessThen $complexity $bestComplexity  && {
				bestComplexityIdx=$counter
				bestComplexity=$complexity;
			}
		else
			Log "Potential error: Empty complexity file: $estimateFile"
		fi
		let counter++
	done
	# rename the files to meet the convention:
	cp -f $orderDir/ComplexityEstimate${bestComplexityIdx}.txt $orderDir/ComplexityEstimate.txt || Die "Failed to create $orderDir/ComplexityEstimate.txt"
	cp -f $orderDir/NumOfJobs${bestComplexityIdx}.txt $orderDir/${COMPLEXITY_TABLE_FILE} || Die "Failed to create $orderDir/${COMPLEXITY_TABLE_FILE}"
	cp -f $orderDir/order${bestComplexityIdx}.txt $orderDir/order.txt || Die "Failed to create $orderDir/order.txt"

}

# this file is invoked after all SecondEstimation jobs are done. We have to decide where we stand with complexity
cwd=`pwd`

alldirs=`FindAllDataDirs`

[ -z "$alldirs" ] && DieNoLogin "This is a BUG - no DATA dirs found: $alldirs. I should have detected it much earlier!"

# let's first copy all threshold conifugration which will be used here and in other files

cat > ${CHECKPOINT_CONFIG_FILE} <<EOF
LOW_PARALLELIZATION_THRESHOLD=${LOW_PARALLELIZATION_THRESHOLD}
REASONABLE_COMPLEXITY_THRESHOLD=${REASONABLE_COMPLEXITY_THRESHOLD}
COMPLEXITY_TOO_HIGH_THRESHOLD=${COMPLEXITY_TOO_HIGH_THRESHOLD}
MAX_JOBS_PER_TASK=${MAX_JOBS_PER_TASK}
SINGLE_JOB_LINKAGE_THRESHOLD=${SINGLE_JOB_LINKAGE_THRESHOLD}
EOF


Log "Now deciding regarding BACKFILL/TOO_HEAVY"

createTaskComplexityMap(){
	
	TotalComplexity=1
	TooComplex=""
	# we sort according to the complexity to get all simpler positions done earlier
	rm ${TASK_COMPLEXITY_MAP_FILE}.sort
	touch ${TASK_COMPLEXITY_MAP_FILE}.sort

	# sort according to complexity 
	for i in $alldirs; do
		# we want to select the best order out of all created
		selectBestOrder $i
		ParseComplexityEstimateFile $i/ComplexityEstimate.txt
		res=$task_ts
		[ -z "$res" ] && res=0
		echo $res $i >> ${TASK_COMPLEXITY_MAP_FILE}.sort
	done
	cat  ${TASK_COMPLEXITY_MAP_FILE}.sort | sort -k1 -n | awk '{print $2}' >  ${TASK_COMPLEXITY_MAP_FILE}.order

	[ -s  ${TASK_COMPLEXITY_MAP_FILE}.order ] || DieNoLogin "Failed to create inital order ${TASK_COMPLEXITY_MAP_FILE}.order"
	
	# this allows us to keep an order among all invocation of this script. 
	alldirs=`cat ${TASK_COMPLEXITY_MAP_FILE}.order`
	
	rm -f ${TASK_COMPLEXITY_MAP_FILE}.order ${TASK_COMPLEXITY_MAP_FILE}.sort $TASK_COMPLEXITY_MAP_FILE ${TASK_COMPLEXITY_MAP_FILE}.splits > /dev/null 2>&1
# this will be the file with the complexity map - an input for the PreParallelDag
	touch $TASK_COMPLEXITY_MAP_FILE
# this is a file containing only directories in the order which will be used for all Parallelization needs in the following scripts
	touch ${TASK_COMPLEXITY_MAP_FILE}.order
# this is a file containing directories and complexity, sorted by complexity
	touch ${TASK_COMPLEXITY_MAP_FILE}.sort
# this is a file containing directories and split sizes, in the original order
	touch ${TASK_COMPLEXITY_MAP_FILE}.splits

	complexity_array=""	

	for i in $alldirs; do

	
		if [ -e $i/ComplexityEstimate.txt -a -e $i/${COMPLEXITY_TABLE_FILE} ]; then
 			 ParseComplexityEstimateFile $i/ComplexityEstimate.txt
	
			COMPLEXITY=$task_complexity
			[ -z "$complexity_array" ] && complexity_array="$COMPLEXITY" || complexity_array="${complexity_array} $COMPLEXITY"
			TABLESUM=$task_ts
			TotalComplexity=`echo $TotalComplexity + $TABLESUM | bc -l`
			[ $COMPLEXITY -gt 0 ] || { COMPLEXITY=$COMPLEXITY_TOO_HIGH_THRESHOLD; TABLESUM=0; }

			if [ ! -z "$COMPLEXITY" ]; then
			
					
				LessThen $COMPLEXITY_TOO_HIGH_THRESHOLD $COMPLEXITY && {
					TooComplex="${TooComplex} $i $COMPLEXITY"
					# we do the same trick as in FirstEstimate. 
					# This task is too heavy, let's Ignore it
					MarkDoNotCompute  $i
					TotalComplexity=`echo $TotalComplexity-$TABLESUM | bc -l`
					
					continue
				}


				# if below the LOW_PARALLELIZATION_THRESHOLD -> we just put 0 in complexity file
				if [ $COMPLEXITY -le $LOW_PARALLELIZATION_THRESHOLD ]; then
					echo 0 >> ${TASK_COMPLEXITY_MAP_FILE}
					echo $i >> ${TASK_COMPLEXITY_MAP_FILE}.order
					echo $TABLESUM $i >> ${TASK_COMPLEXITY_MAP_FILE}.sort
					echo $i 0  >> ${TASK_COMPLEXITY_MAP_FILE}.splits
					continue
				fi
			
				# we pass all possible splits and their complexities through the filter, which lets through only those 
				# above or equal!! Low parallelization threshold, below or equal to reasonable complexity threshold
				# we also want to take only the first split of LOW_PARALLELIZATION_THRESHOLD. 
				depth=(`cat  ${i}/${COMPLEXITY_TABLE_FILE} | awk "{if ( \\$3 <= ${REASONABLE_COMPLEXITY_THRESHOLD} && counter == 0 && \\$3 > 0 && \\$2 > 0 ) print  \\$2 ; if ( \\$3 > 0 && \\$3 <= ${LOW_PARALLELIZATION_THRESHOLD} || \\$2 < 0 ) counter++; }"`)
				
				if [ -z "${depth[*]}" ]; then
				# because all jobs are above REASONABLE_COMPLEXITY_TH
					TooComplex="${TooComplex} $i $COMPLEXITY"
					# ignore this task for now
					MarkDoNotCompute $i
					TotalComplexity=`echo $TotalComplexity-$TABLESUM | bc -l`
					continue
				else
					# finally write the depth into task complexity map file.
					echo "${#depth[*]} ${depth[*]}" >> ${TASK_COMPLEXITY_MAP_FILE}
					echo $i >> ${TASK_COMPLEXITY_MAP_FILE}.order
					echo $TABLESUM $i >> ${TASK_COMPLEXITY_MAP_FILE}.sort
					echo "$i ${#depth[*]} ${depth[*]}" >> ${TASK_COMPLEXITY_MAP_FILE}.splits
				fi
				
			else
				Log "ComplexityEstimate.txt does not contain anything"
				echo $i >> ${TASK_COMPLEXITY_MAP_FILE}.order
				echo 0 >> ${TASK_COMPLEXITY_MAP_FILE}
				echo 0 $i >> ${TASK_COMPLEXITY_MAP_FILE}.sort
				echo $i 0  >> ${TASK_COMPLEXITY_MAP_FILE}.splits
			fi
		else
			Log "Job is probably trivial - no Complexity file"
			echo 0 >> ${TASK_COMPLEXITY_MAP_FILE}
			echo $i >> ${TASK_COMPLEXITY_MAP_FILE}.order
			echo 0 $i >> ${TASK_COMPLEXITY_MAP_FILE}.sort
			echo $i 0  >> ${TASK_COMPLEXITY_MAP_FILE}.splits
		fi
	done
	
	UpdateTaskState "$SECOND_ESTIMATE_STATE COMPLEXITY_ARRAY=\"${complexity_array}\""
	Log "Checking if there are too complex tasks: TooComplex- ${TooComplex} "

	RunCommand "${BIN}/log10 ${TotalComplexity}" || exit ${GENERIC_ERROR}

	# this is a global variable
	complexity=$command_output

	Log "Here is the total Task complexity: TableSizeSum= ${TotalComplexity} , log10= $command_output"
	UpdateTaskState "${SECOND_ESTIMATE_STATE} TOTAL_TABLE_SIZE=${TotalComplexity}"
	UpdateTaskState "${SECOND_ESTIMATE_STATE} TOTAL_COMPLEXITY=$command_output"

	echo ${command_output} > ${TOTAL_COMPLEXITY_FILE}

	[ ! -z "$command_output" ] || Die "log10 ${TotalComplexity} returned no output. - probably BUG"

	IsNumber $command_output || Die "log10 returned something which is not a number"

}

createTaskComplexityMap

if [ ! -z "${TooComplex}" ]; then
	# let's check if we are not doing BEST_EFFORT or Single point.
	HandleTooHeavyJobs ${SECOND_ESTIMATE_RESULT_FILE} ${SECOND_ESTIMATE_STATE} ${TooComplex}
	# we are still alive -> it's a special run, for sure
	# rerun createTaskComplexityMap 
	rm -f ${TASK_COMPLEXITY_MAP_FILE}.order
	alldirs=`FindAllDataDirs`
	createTaskComplexityMap
fi 

TooComplex=""

### if we are in a BEST EFFORT, we will have to reduce the load to minimum
LessThen  $COMPLEXITY_TOO_HIGH_THRESHOLD $complexity  && { 
	if [ ! -z "$TASK_BEST_EFFORT" ]; then
		Log "Detected BEST_EFFORT run -> try to remove all heavy jobs"
		# sort all the reset in the decreasing complexity order, and remove one after one, 
		# until we get reasonable total compelxity
		alldirs=(`cat ${TASK_COMPLEXITY_MAP_FILE}.sort | sort -n -k1 -r | awk '{ print $2 }'`)
		complexities=(`cat ${TASK_COMPLEXITY_MAP_FILE}.sort | sort -n -k1 -r | awk '{ print $1 }'`)
		can_continue=""
		counter=0
		
		for i in ${alldirs[*]}; do
			TotalComplexity=`echo $TotalComplexity - ${complexities[${counter}]} | bc -l`
			RunCommand "${BIN}/log10 ${TotalComplexity}" || exit ${GENERIC_ERROR}
			Log "Reducing complexity - removing point $i"
			TooComplex="${alldirs[${counter}]}"
			MarkDoNotCompute ${alldirs[${counter}]}

			[ $command_output -ge $COMPLEXITY_TOO_HIGH_THRESHOLD ] || { 
				# we can stop reducing complexity now
				complexity=$command_output
				can_continue="true"; break;
			}
			let counter++
		done
		
		[ -z "$can_continue" ] && {
			echo $JOB_TOO_HEAVY > ${SECOND_ESTIMATE_RESULT_FILE}; 
			exit $JOB_TOO_HEAVY; 
		}
		
		HandleTooHeavyJobs ${SECOND_ESTIMATE_RESULT_FILE} ${SECOND_ESTIMATE_STATE} $TooComplex
		# get alldirs again, and regenerate the complexity map
		rm -f ${TASK_COMPLEXITY_MAP_FILE}.order
		alldirs=`FindAllDataDirs`
		createTaskComplexityMap
		
	else
		echo $JOB_TOO_HEAVY > ${SECOND_ESTIMATE_RESULT_FILE}; 
		exit $JOB_TOO_HEAVY; 
	fi;
}

LessThen $BACKFILL_THRESHOLD $complexity && { 
	echo $BACKFILL_JOB > ${SECOND_ESTIMATE_RESULT_FILE}; 
	exit ${BACKFILL_JOB};
}

# now we want to decide how many jobs to be in each task
TooComplex=""
RunCommand "${BIN}/${SPLIT_DISTRIBUTION_BIN} ${TASK_COMPLEXITY_MAP_FILE} ${MAX_JOBS_PER_TASK}" || {
# we're here only if this command fails
	if [ $ret -ne 6 ]; then
		Die "Error after running split distribution";
	else
		Log "The command returned that there is no solution - task is too heavy"; 
		if [ -z "$TASK_BEST_EFFORT" ]; then
			echo $JOB_TOO_HEAVY > ${SECOND_ESTIMATE_RESULT_FILE}; 
			exit ${JOB_TOO_HEAVY}; 
		else
			Log "Found BEST_EFFORT run -> let's try to reduce the load"
			for i in $alldirs; do
				# create new MAP_FILE by removing the task with the largest number of jobs
				dirname=`cat ${TASK_COMPLEXITY_MAP_FILE}.splits | sort -n -k3 -r | awk '{print $1}' | head -1`
				Log "Removing directory $dirname"
				[ -z "$dirname" ] && Die "Problem while parsing $TASK_COMPLEXITY_MAP_FILE.splits"
				cat ${TASK_COMPLEXITY_MAP_FILE}.splits | grep -v $dirname >  ${TASK_COMPLEXITY_MAP_FILE}.splits.tmp
				mv  ${TASK_COMPLEXITY_MAP_FILE}.splits.tmp  ${TASK_COMPLEXITY_MAP_FILE}.splits
				TooComplex="${TooComplex} $dirname"
				# mark as not-computable
				MarkDoNotCompute $dirname
				dirs_left=`cat ${TASK_COMPLEXITY_MAP_FILE}.splits | cut -f1 -d" "`
				[ -z "$dirs_left" ] && {
					# we cannot do anything -> no dirs left => all dirs are TOO Heavy
					echo $JOB_TOO_HEAVY > ${SECOND_ESTIMATE_RESULT_FILE}; 
					exit ${JOB_TOO_HEAVY}; 
				}
				cat ${TASK_COMPLEXITY_MAP_FILE}.splits | cut -f1 -d" " > ${TASK_COMPLEXITY_MAP_FILE}.order
				cat ${TASK_COMPLEXITY_MAP_FILE}.splits | cut -f2- -d" " > ${TASK_COMPLEXITY_MAP_FILE}
				
				# if the following is successfull -> we'll keep running with the slit inside command_output variable
				RunCommand "${BIN}/${SPLIT_DISTRIBUTION_BIN} ${TASK_COMPLEXITY_MAP_FILE} ${MAX_JOBS_PER_TASK}" && break || {
					# if successful we can continue. Otherwise -> just break out
					[ $ret -eq 6 ] && continue 
					Die "Failed to run ${BIN}/${SPLIT_DISTRIBUTION_BIN}";
				}
			done
		fi;
	fi; 
}
# we output the split result 
echo $command_output > $SPLIT_RESULT_FILE
[ ! -z "$TooComplex" ] && {
        # get alldirs again, and regenerate the complexity map
	 HandleTooHeavyJobs ${SECOND_ESTIMATE_RESULT_FILE} ${SECOND_ESTIMATE_STATE} $TooComplex 
}

exit $SUCCESS
