#!/bin/bash

# For debugging purposes only
set -x

# Script constants
DIRECTORY_NAME=$(dirname $0)
SCRIPT_NAME=$(basename $0)
VALID_ROUTE_LENGTH=1000
ESTIMATE_DIRECTORY_NODE_NAME=estimate_directory_node.sub
EXECUTABLE_WRAPPER_NAME=EstimationWrapper.sh
EXECUTABLE_PREFIX=Superlink_Vanilla

declare -r VALID_ROUTE_LENGTH
declare -r DIRECTORY_NAME
declare -r SCRIPT_NAME
declare -r ESTIMATE_DIRECTORY_NODE_NAME
declare -r EXECUTABLE_WRAPPER_NAME 
declare -r EXECUTABLE_PREFIX
# End of script constants

# Includes
. "$DIRECTORY_NAME/JobUtils.sh"
. "$DIRECTORY_NAME/BoincNodeConstants.sh"
# End of includes

# Auxiliary functions

get_current_date()
{
	local current_date=`date +%s`

	echo $current_date
}

# The format of the jobs numbers map file is as follows:
# depth overall_number_of_jobs complexity
#read_depth_jobs_number()
#{
#    local jobs_numbers_file_path=$1
#	local depth=$2
#	local current_depth jobs_number
#
#    while read line
#    do
#        # Ignoring blank lines
#        [[ "x$line" == "x" ]] && continue
#
#		current_depth=$(echo $line | cut -d' ' -f1)
#
#		if [[ $current_depth == $depth ]]
#		then
#			jobs_number=$(echo $line | cut -d' ' -f2)
#			echo $jobs_number
#			return 0
#		fi
#    done < "$jobs_numbers_file_path"
#
#	return 1
#}

# Arguments   : $1 - overall number of jobs in this set
# Return value: number of sample jobs of this set
calculate_bins_number()
{
	echo $( min $( ceiling $( evaluate_arithmetic "$1 / $ESTIMATION_SAMPLES_PER_BIN" ) ) $ESTIMATION_MAX_SAMPLES )
}

submit_new_set()
{
	local set_id=$1
	local jobs_number=$2
	local depth=$3
	local working_directory=$4
	local valid_routes_number=$( valid_routes_number $set_id )

	# For a randomly chosen job id in each one of the bins run a Superlink instance
	local bins=$(calculate_bins_number $valid_routes_number)

	for bin_index in $(seq 1 $bins)
	do
		local current_bin_samples=$ESTIMATION_SAMPLES_PER_BIN

		[[ $bin_index == $bins ]] && current_bin_samples=$(( $valid_routes_number % $ESTIMATION_SAMPLES_PER_BIN ))

		# The random number is generated out of the range [0, 2**15 - 1]
		local route_index=$(( $(( $bin_index - 1 )) * $ESTIMATION_SAMPLES_PER_BIN + $RANDOM * $current_bin_samples / 2**15))
		local job_id=$( head -$(( $route_index + 1)) "${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}" | tail -1 )

		local total_mem=$(( $MAX_AVAILABLE_MEMORY * 1024 ))
		
		cat > $ESTIMATE_DIRECTORY_NODE_NAME << EOF
#executable = $DIRECTORY_NAME/Superlink/$EXECUTABLE_PREFIX.\$\$(Arch).\$\$(OpSys)
executable = $DIRECTORY_NAME/$EXECUTABLE_WRAPPER_NAME
#arguments = datafile.dat pedfile.dat -e order.txt $job_id $depth
#  $MAX_AVAILABLE_MEMORY MB limit on virtual memory consumption
arguments = $total_mem $EXECUTABLE_PREFIX.\$\$(Arch).\$\$(OpSys) datafile.dat pedfile.dat -e order.txt $job_id $depth

transfer_input_files = $DIRECTORY_NAME/Superlink/$EXECUTABLE_PREFIX.\$\$(Arch).\$\$(OpSys), $working_directory/datafile.dat, $working_directory/pedfile.dat, $working_directory/order.txt
universe = vanilla
should_transfer_files = YES
when_to_transfer_output = ON_EXIT

priority   = $(( ${HIGHEST_PRIORITY} - ${TASKID} - ${bins} ))
output     = $working_directory/output.estimate_node.$depth.$set_id.$job_id
error      = $working_directory/error.estimate_node.$depth.$set_id.$job_id
log        = $working_directory/log.estimate_node.$depth.$set_id.$job_id

Requirements = Memory >=  $MAX_AVAILABLE_MEMORY && \\
                ((Arch == "INTEL" && OpSys == "LINUX")   || \\
                (Arch == "X86_64" && OpSys == "LINUX")   || \\
                (Arch == "ppc64"  && OpSys == "LINUX"))

# EstimationWrapper does not run on windows
# || \\
#                (Arch == "INTEL"  && OpSys == "WINNT51") || \\
#                (Arch == "INTEL"  && OpSys == "WINNT50"))

ImageSize     = 100
Rank          = KFlops 
copy_to_spool = False

getenv = true
notification = never
+EstimationSetId = $set_id
+EstimationJobId = $job_id
+EstimationDepth = $depth
+EstimationDir = "$working_directory"
leave_in_queue = true
queue
EOF

		safe_condor_submit $ESTIMATE_DIRECTORY_NODE_NAME
	done
}

# Arguments: $1 - the id of set that is going to be removed from the pool
#            $2 - (optional) the working directory of the above set to clean temporary files
remove_set()
{
	declare -a cluster_ids
	local set_id=$1
	local working_directory=$2

	Log "Appending set no. $set_id from directory $working_directory to the purger's list"
	cluster_ids=( $(condor_q -format "%s\n" ClusterId -constraint "EstimationSetId == $set_id") ) 

	# condor_rm -forcex causes problems to schedd due to an internal bug, that is why we switched to use condor_qedit instead
	#for cluster_id in ${cluster_ids[@]}
	#do
	#	condor_qedit $cluster_id LeaveJobInQueue false 
	#done
#	[ -n "${cluster_ids[*]}" ] && condor_rm ${cluster_ids[@]}
	echo "ClusterId integer ${cluster_ids[@]}" >> "${PURGE_DIR}/${TASKID}"
	${DIRECTORY_NAME}/Purger.sh
	#condor_rm -forcex ${cluster_ids[@]}

	[[ "x$working_directory" != "x" && -n "$ESTIMATION_REMOVE_TEMPORARY_RESULTS" ]] && \
		find "$working_directory" -name "*\.estimate_node\.*" | xargs rm -f >& /dev/null
}

remove_all_sets()
{
	local set_id_counter=0
	local -a removed_set_ids=( $1 ) removed_working_directories=( $2 )

	for set_id in ${set_ids[@]}
	do
		remove_set $set_id ${removed_working_directories[$set_id_counter]}
		let set_id_counter++
	done
}

constrict_valid_routes()
{
	local set_id=$1
	local previous_level_cardinality=$2
	local current_level_cardinality=$3
	local temporary_file_name=constrict_valid_routes.old

	local constriction_factor=$(( $current_level_cardinality / $previous_level_cardinality ))

	# We do not want to monitor the new routes file preparation, as it is an excessive information
	set +x
	cat "${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}" | while read line; do echo "scale=0;" $line "/$constriction_factor;" ; done | bc -l | uniq > \
		"$temporary_file_name"
	set -x

	mv "$temporary_file_name" "${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}"
}

# Return value: 0 - upon success, 1 - when the number of routes exceeds the allowed maximum
expand_valid_routes()
{
	local set_id=$1
	local current_level_cardinality=$2
	local next_level_cardinality=$3

	local routes_number=$( valid_routes_number $set_id )
	local expansion_factor=$(( $next_level_cardinality / $current_level_cardinality ))

	[[ $(( $expansion_factor * $routes_number )) -gt $ESTIMATION_MAX_ROUTES ]] && {
		Log "The prospective number of valid routes ($routes_number) would exceed $ESTIMATION_MAX_ROUTES, cannot expand routes"

		return 1
	}

	Log "Expanding valid routes by $expansion_factor"
	local valid_route_counter=0
	local temporary_file_name=expand_valid_routes.old

	# We do not want to monitor the routes file preparation, as it is an excessive information
	set +x
	while read valid_route 
	do
		local start=$(( $valid_route * $expansion_factor ))
		local end=$(( $start + $expansion_factor - 1 ))

		seq --format=%.0f $start $end >> $temporary_file_name

		[[ $(( $valid_route_counter % $VALID_ROUTE_LENGTH )) == 0 ]] && Log "Expanding valid routes counting ($valid_route_counter)"
		let valid_route_counter++
	done < "${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}" 
	set -x

	mv "$temporary_file_name" "${ESTIMATION_PARAMETERS_FILE_NAME}.${set_id}"
}

decrease_depth()
{
	local set_id=$1
	local set_id_counter=$2

	remove_set $set_id ${working_directories[$set_id_counter]}
	new_jobs_number=`read_depth_jobs_number "${working_directories[$set_id_counter]}/NumOfJobs.txt" \
											 $((${depths[$set_id_counter]} - 1))` 
	[[ $? != 0 ]] && remove_all_sets "${set_ids[*]}" "${working_directories[*]}" && exit $ESTIMATION_INVALID_DEPTH 

	# To constrict the number of valid routes to the previous depth, one has to divide all the routes by the cardinality of the
	# previous level.
	constrict_valid_routes $set_id $new_jobs_number ${jobs_numbers[$set_id_counter]}
	jobs_numbers[$set_id_counter]=$new_jobs_number
	let depths[$set_id_counter]--
}

write_estimation_state_file()
{
	local state_file_path=$1
	local set_id_counter=0

	# Emptying the state file to start everything over
	[ -f "$state_file_path" ] && mv "$state_file_path" "${state_file_path}.old"

	for set_id in ${set_ids[@]}
	do
		local jobs_number=${jobs_numbers[$set_id_counter]}
		local depth=${depths[$set_id_counter]}
		local working_directory=${working_directories[$set_id_counter]}
		local accomplished_set=${accomplished_sets[$set_id_counter]}
		local orignal_depth=${original_depths[$set_id_counter]}
		
		echo $set_id $jobs_number $depth $working_directory $accomplished_set $orignal_depth >> $state_file_path

		let set_id_counter++
	done

	[ -f "${state_file_path}.old" ] && rm ${state_file_path}.old
}

read_estimation_state_file()
{
	local state_file_path=$1
	local set_id_counter=0

	while read line
	do
		# Ignoring blank lines
		if [[ "x$line" == "x" ]]
		then
			continue
		fi

		set_ids[$set_id_counter]=$(echo $line | cut -d' ' -f1)
		jobs_numbers[$set_id_counter]=$(echo $line | cut -d' ' -f2)
		depths[$set_id_counter]=$(echo $line | cut -d' ' -f3)
		working_directories[$set_id_counter]=$(echo $line | cut -d' ' -f4)
		accomplished_sets[$set_id_counter]=$(echo $line | cut -d' ' -f5)
		original_depths[$set_id_counter]=$(echo $line | cut -d' ' -f6)

		#Log "Valid routes in set no. ${set_ids[$set_id_counter]}: ${#set_valid_routes[@]}"

		let set_id_counter++
	done < "$state_file_path"
}

restart_unaccomplished_sets()
{
	local set_id_counter=-1

	Log "Restarting unaccomplished sets ${accomplished_sets[*]}"

	for set_id in ${set_ids[*]}
	do
		let set_id_counter++
		# Restart only unaccomplished sets
		[[ "${accomplished_sets[$set_id_counter]}" == 1 ]] && continue

		Log "Restarting accomplished set no. $set_id"
		jobs_number=${jobs_numbers[$set_id_counter]}
		depth=${depths[$set_id_counter]}
		working_directory=${working_directories[$set_id_counter]}
	
		# Appending the long rule for Condor status constraint
		submit_new_set $set_id $jobs_number $depth $working_directory 
		set -x
	done
}

average()
{
	local -a running_times=( $1 )
	local average_time

	average_time=$( echo "(${running_times[@]})/${#running_times[@]}" | tr ' ' '+' | bc -l )
	average_time=$(ceiling "${average_time}")

	echo ${average_time}
}
# End of auxiliary functions

# Signals trapping

# Sending SIGUSR2 signal to this script causes it to reload its configuration parameters
trap '. "$DIRECTORY_NAME/JobUtils.sh"; . "$DIRECTORY_NAME/BoincNodeConstants.sh"' USR2
# Removing all Condor jobs upon SIGUSR1 signal
trap 'remove_all_sets "${set_ids[*]}" "${working_directories[*]}"; exit 1' USR1

# End of signals trapping

# Main
Log "Starting $SCRIPT_NAME"

# Script parameters processing
parameters_file_path=$1

# In-memory database
declare -a set_ids depths jobs_numbers working_directories 
# The array of zeros and ones designates sets that have finished successfully and that we should not take care of
declare -a accomplished_sets
# The array of average times for sets that have finished successfully
declare -a accomplished_average_times
# This depth is the in-memory record of the depth, we started from. It is crucial for us lest to get into a loop of increasing/decreasing the depth
declare -a original_depths
# End of in-memory database

# Either there is a state file, then we start over from the checkpoint, or we start everything from the beginning
if [[ ( ! ( -f "$ESTIMATION_STATE_FILE" ) ) && ( ! ( -f "${ESTIMATION_STATE_FILE}.old" ) ) ]]
then
	read_parameters_file "$parameters_file_path"
	
	set_id_counter=0
	
	Log "Submitting new estimation sets: ${set_ids[*]}"
	
	for set_id in ${set_ids[*]}
	do
		jobs_number=${jobs_numbers[$set_id_counter]}
		depth=${depths[$set_id_counter]}
		working_directory=${working_directories[$set_id_counter]}
	
		# Appending the long rule for Condor status constraint
		submit_new_set $set_id $jobs_number $depth $working_directory
	
		original_depths[$set_id_counter]=$depth
		accomplished_sets[$set_id_counter]=0
		let set_id_counter++
	done
else
	# The script was killed when it was recording its state
	[ -f "${ESTIMATION_STATE_FILE}.old" ] && mv "${ESTIMATION_STATE_FILE}.old" "$ESTIMATION_STATE_FILE"
	read_estimation_state_file "$ESTIMATION_STATE_FILE"
	# Cleaning leftovers of previous reincarnation
	remove_all_sets "${set_ids[*]}" "${working_directories[*]}"

	restart_unaccomplished_sets
fi

# End of script parameters processing

Log "Starting the main estimation cycle"

# Waiting until all of the directory estimation nodes are over
while [[ 1 == 1 ]]
do
	# Checking if all Condor jobs are accomplished 
	declare -a accomplished_jobs running_jobs failed_set_ids

	set_id_counter=-1
 	failed_set_ids=( $(condor_q -format "%s\n" EstimationSetId \
			     -constraint "regexp(\".*$EXECUTABLE_WRAPPER_NAME.*\",Cmd)" -constraint "ExitCode != 0" | sort | uniq) )

	# If there are jobs that finished with non-zero exit code, we terminate the entire script
	if [[ $? == 0 && ${#failed_set_ids[@]} != 0 ]]
	then
		Log "There were Condor jobs in set(s) ${failed_set_ids[*]} that finished with non-zero exit code, finalizing everything in these directories"
		
		for failed_set_id in ${failed_set_ids[@]}
		do
			failed_set_id_counter=$( find_array_index "${set_ids[*]}" $failed_set_id )
			[[ $? != 0 ]] && continue
			remove_set $failed_set_id ${working_directories[$failed_set_id_counter]}

			# Creating the error file for tracking down the progress of the directory
            #touch "${working_directories[$failed_set_id_counter]}/$FRONTEND_ERROR_FILE"
			MarkDoNotCompute "${working_directories[$failed_set_id_counter]}"
			unset accomplished_average_times[$failed_set_id_counter]
			accomplished_sets[$failed_set_id_counter]=1
			let failed_set_id_counter++
		done
		#remove_all_sets "${set_ids[*]}" "${working_directories[*]}"
		#exit $ESTIMATION_EXECUTABLE_FAILURE_EXIT_CODE 
	fi

	for set_id in ${set_ids[@]}
	do
		let set_id_counter++

		[[ ${accomplished_sets[$set_id_counter]} == 1 ]] && continue

		directory_constraint="EstimationDir==\"${working_directories[$set_id_counter]}\""
   
#		deleted_jobs=( $(condor_q -format "%d\n" RemoteWallClockTime\
#                                    -constraint "$directory_constraint" -constraint "JobStatus == 5" \
#									 -constraint "EstimationSetId == $set_id" \
#                                     -constraint "regexp(\".*$EXECUTABLE_WRAPPER_NAME.*\",Cmd)" ) )
#  
		accomplished_jobs=( $(condor_q -format "%d\n" RemoteWallClockTime \
                                     -constraint "$directory_constraint" -constraint "JobStatus > 2" \
									 -constraint "EstimationSetId == $set_id" \
                                     -constraint "regexp(\".*$EXECUTABLE_WRAPPER_NAME.*\",Cmd)" ) )
		# Sleeping till the schedd is back to work properly
		[[ $? != 0 ]] && break
		running_jobs=( $(condor_q -format "%d\n" $(get_current_date)-JobCurrentStartDate \
									 -constraint "$directory_constraint" -constraint "JobStatus == 2" \
									 -constraint "EstimationSetId == $set_id" \
									 -constraint "regexp(\".*$EXECUTABLE_WRAPPER_NAME.*\",Cmd)" ) )
		[[ $? != 0 ]] && break

		actual_execution_time=( $(condor_q -format "%d\n" $(get_current_date)-JobStartDate \
									 -constraint "$directory_constraint" -constraint "JobStatus == 2" \
									 -constraint "EstimationSetId == $set_id" \
									 -constraint "regexp(\".*$EXECUTABLE_WRAPPER_NAME.*\",Cmd)" ) )
		[[ $? != 0 ]] && break
		max_run_time=$( echo ${accomplished_jobs[@]} ${running_jobs[@]} | tr ' ' '\n' | uniq | sort -n    | tail -1)
		max_actual_execution_time=$( echo ${accomplished_jobs[@]} ${actual_execution_time[@]} | tr ' ' '\n' | uniq | sort -n    | tail -1)
		min_run_time=$( echo ${accomplished_jobs[@]} ${running_jobs[@]} | tr ' ' '\n' | uniq | sort -n -r | tail -1)
		max_allowed_execution_time=`echo $ESTIMATION_UPPER_BOUND*2 | bc -l`

		routes_number=$( valid_routes_number $set_id )
		bins=$( calculate_bins_number $routes_number )

		let done_jobs="${#accomplished_jobs[@]}"
		# if we hit the pathalogical case when all jobs had to be removed since it took too long to wait in the system
		# so we assume the worst case
		if [[ ${#accomplished_jobs[@]} -eq 0 && ${max_actual_execution_time} -ge $max_allowed_execution_time ]] 
		then
			let max_run_time="$ESTIMATION_UPPER_BOUND"-"1" 
			min_run_time=$ESTIMATION_UPPER_BOUND
			accomplished_jobs[0]=$ESTIMATION_UPPER_BOUND
			Warn " All jobs have to be deleted since Estimate have been stuck for too long: exiting and assuming runtime = $ESTIMATION_UPPER_BOUND "
		fi

		[[ ${done_jobs} -gt $bins ]] && \
			Warn "The number of accomplished jobs of set no. ${set_ids[$set_id_counter]} is strictly greater than the number of submitted jobs" 

		# Checking that all the accomplished sampled jobs ran longer than the lower bound or took too long in the batch queue
		if [[ ${max_actual_execution_time} -ge $max_allowed_execution_time || ${done_jobs} -ge $bins && \
			( $min_run_time -ge $ESTIMATION_LOWER_BOUND || ${depths[$set_id_counter]} == 1 ) ]]
		then
			Log "Set no. ${set_ids[$set_id_counter]} finished just within lower and upper bounds of the estimation"
			accomplished_average_times[$set_id_counter]=$(average "${accomplished_jobs[*]}")
			Log "The average time of set no. ${set_ids[$set_id_counter]} is ${accomplished_average_times[$set_id_counter]}"
			echo "${accomplished_average_times[$set_id_counter]}" > "${working_directories[$set_id_counter]}/${RUNNING_TIME_ESTIMATE_FILE}"
			accomplished_sets[$set_id_counter]=1
		# The jobs are accomplished too early, so we try to decrease the depth
		elif [[ ${done_jobs} -ge $bins && ${original_depths[$set_id_counter]} -ge ${depths[$set_id_counter]} ]]
		then
			decrease_depth $set_id $set_id_counter

			Log "Set no. ${set_ids[$set_id_counter]} finished too early, decreasing the depth to ${depths[$set_id_counter]}"
			# We calculate the average in case, the jobs with the decreased depth will not be able to make it within lower and upper bounds. In this
			# case we stay at the previous (not decreased) depth, so the calculation, performed here will be valid
			accomplished_average_times[$set_id_counter]=$(average "${accomplished_jobs[*]}")
			echo "${accomplished_average_times[$set_id_counter]}" > "${working_directories[$set_id_counter]}/${RUNNING_TIME_ESTIMATE_FILE}"
			submit_new_set $set_id ${jobs_numbers[$set_id_counter]} ${depths[$set_id_counter]} ${working_directories[$set_id_counter]}
		# The jobs are accomplished too early, but the depth was once increased, so lest stay in the loop forever, we end the set
		elif [[ ${done_jobs} -ge $bins ]]
		then
			Log "Set no. ${set_ids[$set_id_counter]} finished too early, but its' depth has once been increased, so we end it"
			accomplished_average_times[$set_id_counter]=$(average "${accomplished_jobs[*]}")
			Log "The average time of set no. ${set_ids[$set_id_counter]} is ${accomplished_average_times[$set_id_counter]}"
			echo "${accomplished_average_times[$set_id_counter]}" > "${working_directories[$set_id_counter]}/${RUNNING_TIME_ESTIMATE_FILE}"
			accomplished_sets[$set_id_counter]=1
		# The jobs have been running for too long, so we try to increase the depth
		elif [[ $max_run_time -ge $ESTIMATION_UPPER_BOUND && ${original_depths[$set_id_counter]} -le ${depths[$set_id_counter]} ]]
		then
			remove_set $set_id ${working_directories[$set_id_counter]}

			# When the running time of one of the samples has exceeded the upper bound, we gradually increase the depth as a tradeoff
			# between the number of jobs and their computation time 
			new_jobs_number=`read_depth_jobs_number "${working_directories[$set_id_counter]}/NumOfJobs.txt" \
													 $((${depths[$set_id_counter]} + 1))`
			[[ $? != 0 ]] && remove_all_sets "${set_ids[*]}" "${working_directories[*]}" && Die $ESTIMATION_INVALID_DEPTH

			# To expand the number of valid routes to the next depth, one has to multiply each route by the cardinality of the
			# current level and expand all the possible offsprings 
			expand_valid_routes $set_id ${jobs_numbers[$set_id_counter]} $new_jobs_number

			[[ $? != 0 ]] && {
				Log "Unable to increase the depth, finalizing ${working_directories[$set_id_counter]}"
				# Creating the error file for tracking down the progress of the directory
				unset accomplished_average_times[$set_id_counter]
				accomplished_sets[$set_id_counter]=1
				MarkDoNotCompute  ${working_directories[$set_id_counter]}
	    		`HandleTooHeavyJobs ${SECOND_ESTIMATE_RESULT_FILE} ${SECOND_ESTIMATE_STATE} ${working_directories[$set_id_counter]}`
				[[ $? != 0 ]] && {
					# we remove everything since the job is too heavy
					Log "Task is too heavy - exiting from EstimationNode.sh"
					remove_all_sets "${set_ids[*]}" "${working_directories[*]}"
					exit $JOB_TOO_HEAVY;
				}
				continue;
			}
			jobs_numbers[$set_id_counter]=$new_jobs_number
			let depths[$set_id_counter]++

			Log "Set no. ${set_ids[$set_id_counter]} has been running for too long, increasing the depth to ${depths[$set_id_counter]}"
			submit_new_set $set_id ${jobs_numbers[$set_id_counter]} ${depths[$set_id_counter]} ${working_directories[$set_id_counter]}
		# The jobs have been running for too long, but the depth of the set has once been decreased, so we end the set and assign the previous depth
		elif [[ $max_run_time -ge $ESTIMATION_UPPER_BOUND ]]
		then
			decrease_depth $set_id $set_id_counter
			Log "Set no. ${set_ids[$set_id_counter]} has been running for too long, but its' depth has once been decreased, so we end it with the previous depth"
			# The average of running times has been calculated in the previous iteration
			accomplished_sets[$set_id_counter]=1
		fi
	done

	# All the sets are accomplished successfully
	if [[ ! ( "x ${accomplished_sets[@]} x" =~ "x.* 0 .*x" ) ]]
	then
		write_estimation_state_file $ESTIMATION_STATE_FILE
		remove_all_sets "${set_ids[*]}" "${working_directories[*]}"
		write_parameters_file $BOINC_PARAMETERS_FILE_NAME
		#sets_average_time=$( average "${accomplished_average_times[*]}" )
		#Log "The average running time of all the sets is ${sets_average_time}"
		#echo "${sets_average_time}" > "${RUNNING_TIME_ESTIMATE_FILE}"
		Log "All sets are accomplished successfully, ending $SCRIPT_NAME"
		
		exit 0
	fi

	write_estimation_state_file $ESTIMATION_STATE_FILE
	sleep $ESTIMATION_ROUND_TIME
done

# End of main
