#!/bin/bash
#this script handles any failure/success of single Pedtool taks
basepath=`dirname $0`
. ${basepath}/JobUtils.sh

[ $# -ne 1 ] && { echo "Params <ReturnStatus>  "; exit -1; }
#RedirOutToTaskLog

set -x

Log "Pedtool DAG completed with status $1. Making cleanup and reports"

moveToBackfill(){
	msg="$1"
# we are here due to backfill. So we run MoveToBackfill script. if it fails -> we can't move job
# if we don't have BACKFILL_ACCOUNT - just fail
	UpdateTaskState "PEDTOOL.BACKFILL"
	if [ ! -z ${BACKFILL_ACCOUNT} ]; then
		${BIN}/Resubmit.sh ${BACKFILL_ACCOUNT} ${BACKFILL_ASKPASS} ${BACKFILL_ACCEPT_SCRIPT_PATH}|| {
  			ReportWebUpdateToUser "$msg `$BIN/CreateComplexityTable.sh`"
			ReportUpdateToUser "Rejected $msg";
			UpdateTaskState "PEDTOOL.BACKFILL.${FAILED}";
			UpdateFileForUser `basename $TASK_OUTPUT_FILE`;
			UpdateFileForUser $FRONTEND_ARCHIVE_FILE
		}
	else
		UpdateFileForUser $FRONTEND_ARCHIVE_FILE
		UpdateFileForUser `basename $TASK_OUTPUT_FILE`
		ReportWebUpdateToUser "$msg";
		ReportUpdateToUser "Rejected $msg";
	fi
}

migrateTask(){

	mymsg="$1"
	target=`cat ${MIGRATION_MARKER_FILE}`
	real_target=`echo $target | awk '{ print $1}'`
	if [ "$real_target" = "${BACKFILL_ACCOUNT}" ]; then
		Log "Detected backfill migration"
		moveToBackfill "$mymsg"
		return
	fi
		
	Log "Attempting to migrate to $target"
	${BIN}/Resubmit.sh $target && return
	Log "Failed to migrate.- will now attempt to restart"
	
	# we can resubmit only DAG with another name, otherwise DAGman is confused
	# importantly, we first check that PedtoolSubmit.dag exists. If it does, we recreate PedtoolSubmit.dag_restart
	# This part is robust if is invoked in multiple sites
	if [ -e PedtoolSubmit.dag ]; then
		mv  PedtoolSubmit.dag PedtoolSubmit.dag_restart
		dag=PedtoolSubmit.dag_restart
	else
		mv  PedtoolSubmit.dag_restart PedtoolSubmit.dag
		dag=PedtoolSubmit.dag
	fi
	
		
	msg="Your task failed due to major system crash. Please try again later."
	
	Log "Fixing GlobalDag.dag"
	MergeRescueDag "GlobalDag.dag"
	
	CondorSubmitDag "$dag" || {
		ReportUpdateToUser "Failed $msg"
		ReportWebUpdateToUser "$msg"
		DieNoLogin "Failed to execute condor_submit_dag $dag";
	}

        RunCommand "condor_submit ${dag}.condor.sub" || {
		ReportUpdateToUser "Failed $msg"
		ReportWebUpdateToUser "$msg"
		DieNoLogin "Failed to execute condor_submit_dag $dag";
	}
	condor_reschedule;

	Log "Restart successfull"
	exit 0	
	
}


# first - cleanup:
# remove JOB_COMPLEXITY_ESTIMATE_LOCK if exists
rm -f ${JOB_COMPLEXITY_ESTIMATE_LOCK} > /dev/null 2>&1


status=$1

# we detect termination according to the presence of  file TASK_COMPLETE_FILE

[ -e ${TIMEOUT_MARKER_FILE} ] && {
	status=${TIMEOUT_ERROR}
	Log "Task timeout detected";
}

[ -e ${MIGRATION_MARKER_FILE} ] && {
	status=${MIGRATION_REQUEST};
	Log "Migration request detected";
}

[ -e ${TASK_COMPLETE_FILE} ] && status=0
old_status=$status
if [ $status -ne ${CONDOR_RM} -a $status -ne 0 -a $status -ne ${TIMEOUT_ERROR} -a $status -ne ${MIGRATION_REQUEST} ]; then 
	[ -e ${ACCEPT_RESULT_FILE} ] && status=`cat ${ACCEPT_RESULT_FILE}`
	[ -e ${LANDER_GREEN_RUN_FILE} ] && status=`cat ${LANDER_GREEN_RUN_FILE}`
	[ -e ${FIRST_ESTIMATE_RESULT_FILE} ] && status=`cat  ${FIRST_ESTIMATE_RESULT_FILE}`
	[ -e ${SECOND_ESTIMATE_RESULT_FILE} ] && status=`cat  ${SECOND_ESTIMATE_RESULT_FILE}`
	[ -e ${PARALLEL_RESULT_FILE} ] && status=`cat  ${PARALLEL_RESULT_FILE}`
	[ -e ${UBNORMAL_TERMINATION_FILE} ] && status=`cat ${UBNORMAL_TERMINATION_FILE}`
fi

# if the status was not 0, but became zero as a result of these files -> clearly we didn't mean that
[ $status -eq 0 -a $old_status -ne 0 ] && status=$old_status

[ -e TOTAL_COMPUTATION_TIME.txt ] && UpdateTaskState "PEDTOOL TOTAL_TIME=`cat TOTAL_COMPUTATION_TIME.txt`"

case "$status" in
0 )

	UpdateTaskState "PEDTOOL.${DONE}"
	UpdateFileForUser $FRONTEND_ARCHIVE_FILE
	UpdateFileForUser `basename $TASK_OUTPUT_FILE`
	ReportFinalResultToUser 
	ReportWebUpdateToUser
	# make this interactive
	${BIN}/RemoteNotificationHandler.sh
;;
$SUPERLINK_INPUT_ERROR )
	UpdateTaskState "PEDTOOL.${FAILED} INPUT_ERROR"
	LogReject "INPUT_ERROR"
	msg="Task failed: `cat $SUPERLINK_INPUT_ERROR_FILE`"
	ReportWebUpdateToUser "$msg"
	ReportUpdateToUser "Error $msg"
	;;
$TASK_SPECIFIC_POINT_ERROR )
	UpdateTaskState "PEDTOOL.${FAILED} TASK_SPECIFIC_POINT_ERROR"
	LogReject "TASK_SPECIFIC_POINT_ERROR"
	msg="Task failed: The specified point ${TASK_SPECIFIC_POINT} in pedigre ${TASK_SPECIFIC_PED} is not found."
	ReportWebUpdateToUser "$msg"
	ReportUpdateToUser "Error $msg"
	;;
$LOAD_TOO_HIGH ) 
	LogReject "LOAD_TOO_HIGH"
	UpdateTaskState "PEDTOOL.${FAILED} LOAD_TOO_HIGH"
	UpdateFileForUser $FRONTEND_ARCHIVE_FILE
	UpdateFileForUser `basename $TASK_OUTPUT_FILE`
	msg="Our appologies. The system is under heavy load and cannot accept new tasks. Please try again later."
	ReportWebUpdateToUser  "$msg"
	ReportUpdateToUser "Failure $msg"
	;;
${TIMEOUT_ERROR} )
	LogReject "TIMEOUT_ERROR"
	UpdateFileForUser $FRONTEND_ARCHIVE_FILE
	UpdateFileForUser `basename $TASK_OUTPUT_FILE`
	msg="Task failed to complete due to low availability of resources or due to high complexity. If you attempted to evaluate LOD score at multiple locations, it is recommended to request a single location evaluation by using the appropriate option from the Run menu at superlink-online web page."
	ReportWebUpdateToUser  "$msg"
	ReportUpdateToUser "Failure $msg"
	UpdateTaskState "PEDTOOL.TIMEOUT"
	;;
$CONDOR_RM )
	LogReject "REMOVED"
	UpdateTaskState "PEDTOOL.${FAILED} CONDOR_RM"
	UpdateFileForUser $FRONTEND_ARCHIVE_FILE
	UpdateFileForUser `basename $TASK_OUTPUT_FILE`
	msg="Your task was removed from the system."
	ReportWebUpdateToUser "$msg"
	msg="Removed Your task was removed from the system."
	ReportUpdateToUser "$msg"
	;;
$JOB_TOO_HEAVY )
	./${FRONTEND_POST_SCRIPT}
	LogReject "JOB_TOO_HEAVY"
	UpdateTaskState "PEDTOOL.${FAILED} JOB_TOO_HEAVY"
	UpdateFileForUser $FRONTEND_ARCHIVE_FILE
	UpdateFileForUser `basename $TASK_OUTPUT_FILE`
	msg="Task failed: The complexity of the task is too high. It cannot be computed in reasonable time. Please consider reducing the number of markers or simplifying the pedigree.If you attempted to evaluate LOD score at multiple locations, it is recommended to request a single location evaluation by using the appropriate option from the Run menu at superlink-online web page."
	ReportWebUpdateToUser "$msg `$BIN/CreateComplexityTable.sh`" 
	ReportUpdateToUser "TooHard $msg"
	;;
$BACKFILL_JOB )
	LogReject "BACKFILL_JOB"
	msg="Task failed due to low availability of resources. Please try again later. You may also want to consider reducing the number of markers or simplifying the pedigree. If you attempted to evaluate LOD score at multiple locations, it is recommended to request a single location evaluation by using the appropriate option from the Run menu at superlink-online web page."
	moveToBackfill "$msg"
	;;
$MIGRATION_REQUEST )
	Log "Attempting to migrate"
	msg="Task failed due to low availability of resources. Please try again later. You may also want to consider reducing the number of markers or simplifying the pedigree."
	migrateTask "$msg"
	Log "Migration accomplished"
	UpdateTaskState "PEDTOOL.${DONE} PEER_MIGRATION"
	;;
$TOO_MANY_POINTS )
	Log "Failed due to too many points"
	msg="Task rejected as the total amount of positions to compute LOD score is above the system limit. Please reduce the total number of markers M so that P<=$MAX_NUM_MARKERS. This limitation is temporary and will be removed in future version."
	LogReject "TOO_MANY_POINTS"
	ReportWebUpdateToUser  "$msg"
	ReportUpdateToUser "Rejected $msg"
	UpdateTaskState "PEDTOOL.TOO_MANY_POINTS"
;;
*)
	LogReject "UNKNOWN"
	UpdateTaskState "PEDTOOL.${FAILED} $status"
	UpdateFileForUser $FRONTEND_ARCHIVE_FILE
	UpdateFileForUser `basename $TASK_OUTPUT_FILE`
	msg="Our applogies. The system failed to execute your task due to internal failure. You may want to retry to submit it once again. The problem has been automatically reported to the system administrator"
	ReportWebUpdateToUser "$msg"
	ReportUpdateToUser "Failure $msg"
	CreateMailForAdmin "$msg"
;;
esac

# remove the working dir
Log "++++++++++++++++++Finalizing task ${TASKID}++++++++++++++++++"
# we want the log file to travel back to the user dir
UpdateFileForUser $FRONTEND_ARCHIVE_FILE
UpdateFileForUser ${TASK_LOG_FILE}

ENABLE_TASK_LOG=""
softlink=`cat ${JOB_UID_FILE}`


cd ..
# we are here since we can now move the results to DONE_DIR
#RunCommand "tar -z -c -f ${DONE_DIR}/${TASKID}.tar.gz ${TASKID}" && Log "Task saved in ${DONE_DIR}" || Log "Was unable to save the task in ${DONE_DIR}"
#cp -a ${TASKID} ${DONE_DIR}/
touch ${DONE_DIR}/${TASKID}
# mark mail dir for deletion
touch ${MAIL_DIR}/${TASKID}/LAST

#rm -rf ${TASKID}
[ -z "$softlink" ] || rm -f  $softlink

exit 0
