#!/bin/bash
set -x
# The script should be run as a crontab script to dynamically control the level of submitted clients. When there are no jobs left to process, the
# script should fade out the number of BOINC clients. The pre-requirements of the script are 
# 1) the passwordless ssh access to all the submission and the database nodes
# 2) the same interface on all the submission nodes, specifically, the part of it which is in charge of setting the maximal number of clients

# Script constants
DIRECTORY_NAME=$(dirname $0)
SCRIPT_NAME=$(basename $0)

# Exit codes

# End of exit codes
WORKUNITS_LEFT_UNDETERMINED=1
BUNDLE_SIZE_UNDETERMINED=2
# End of script constants

# Includes
. "$DIRECTORY_NAME/BoincNodeConstants.sh"
# End of includes


# Database
# The syntax of the array should be as follows <user name>@<host name>. If the user name is the same, then it can be omitted along with the '@' sign
#declare -a submission_nodes=( linkageathome@cbl-link08 condors@cbl-link08 pedtool@diet.cs.wisc.edu pedtool@decaf.cs.wisc.edu amazon@cbl-link08 )
declare -a submission_nodes=( condors@cbl-link08 pedtool@diet.cs.wisc.edu pedtool@decaf.cs.wisc.edu amazon@cbl-link08 mark@haplo1.ics.uci.edu )

# The syntax of the array is exactly like the array above. The entries in the two arrays should be in sync, that is, the nth entry in submission nodes' array 
# should be the submission node of the BOINC project, whose database account is precisely the nth entry in the database nodes' array
declare -a database_nodes=( linkageathome@cbl-boinc-server2 linkageathome@cbl-boinc-server2 linkageathome@cbl-boinc-server2 linkageathome@cbl-boinc-server2 linkageathome@cbl-boinc-server2 linkageathome@cbl-boinc-server2 )
# The array determines which connections use SSH_ASKPASS feature of ssh connections. Fill in 0 for connections that do not use it, fill in the name of the
# askpass program otherwise. The entries in the array should be in sync with the previous two arrays, that is, the nth entry in 'ask_password' array should 
# be related to the nth entries in the above arrays. The askpass programs must all be present in ${BIN_DIR}
#declare -a ask_password=( 0 0 madison_askpass madison_askpass 0)
declare -a ask_password=( 0 madison_askpass madison_askpass 0 0 )
#declare -a profiles=(.bash_profile .bash_profile .bash_profile .bash_profile .bash_profile)
declare -a profiles=( .bash_profile .bash_profile .bash_profile .bash_profile .bash_profile )
#declare -a classad_clauses=( ${BOINC_CLASSADS}.egee ${BOINC_CLASSADS}.technion ${BOINC_CLASSADS}.madison ${BOINC_CLASSADS}.osg ${BOINC_CLASSADS}.amazon )
declare -a classad_clauses=( ${BOINC_CLASSADS}.technion ${BOINC_CLASSADS}.madison ${BOINC_CLASSADS}.osg ${BOINC_CLASSADS}.amazon ${BOINC_CLASSADS}.ics )
# End of database

# Auxiliary functions
# Arguments   : $1 -- the path to the task's directory
#               $2 -- default shares of various clusters
#               $3 -- classad clauses that will be matched versus the existing classad
# Return value: an array, containing how many jobs to request from each one of the clusters for the task
task_shares()
{
	local task_directory="$1"
	local -a shares=( $2 )
	local -a clauses=( $3 )

	local jobs_left=$(( $(cat "${task_directory}/${TOTAL_JOBS_FILE}") - $(cat "${task_directory}/${TOTAL_DONE_JOBS_FILE}") ))
	local excluded=0
	[ -e $task_directory/$EXCLUDED_NUM_JOBS_FILE ] && excluded=`cat $task_directory/$EXCLUDED_NUM_JOBS_FILE`
	let jobs_left="$jobs_left"-"$excluded"
	local task_classad="${task_directory}/${BOINC_CLASSADS}"
	[ -e $task_classad ] || return
	local local_classad=$(mktemp)
	local clause_counter=-1

	local -a output_shares=()

	[[ ${jobs_left} -lt 0 ]] && jobs_left=0

	for clause in ${clauses[*]}
	do
		let clause_counter++

		echo "[ ${BOINC_PRIORITY} = 0; ${CONCURRENCY} = 1; ${JOBS_LEFT} = ${jobs_left}; " > "${local_classad}"
		cat "${DIRECTORY_NAME}/${clause}" >> "${local_classad}"
		echo " ]" >> "${local_classad}"
		scheduler_requirements=$("${DIRECTORY_NAME}/${CLASSAD_VALIDATION_TOOL}" "${task_directory}/${BOINC_CLASSADS}" "${local_classad}" | \
			grep "${SCHEDULER_REQUIREMENTS}" | awk 'BEGIN{ FS="="}{ print $2 }')

		output_shares=( ${output_shares[@]}  $(( ${scheduler_requirements} * ${shares[${clause_counter}]} )) )
	done

	rm -f "${local_classad}"

	local sum_shares=$(echo ${output_shares[@]} | tr ' ' '+' | bc -l)

#	[[ ${sum_shares} -le 0 ]] && output_shares=( ${shares[@]}) && sum_shares=$(echo ${shares[@]} | tr ' ' '+' | bc -l)

	# Outputting the shares of each cluster in the request for the task's jobs
	for share in ${output_shares[*]}
	do
		echo "scale = 0; ${share} * ${jobs_left} / ${sum_shares}" | bc -l
	done
}
# Arguments   : $1 -- default shares of various clusters
# Return value: an array, containing how many jobs to request from each one of the clusters for all the tasks
tasks_shares()
{
	local -a shares=( $1 )
	# Looking for all the symbolic links which represent tasks in process: the slash at the end of the searched directory is crucial,
	# when the work directory itself is a symbolic link
	local -a tasks=( $(find "${WORK_DIR}/" -mindepth 1 -maxdepth 1 -type l) )
	local -a output_shares=()

	for share in ${shares[*]}
	do
		output_shares=( ${output_shares[@]} 0 )
	done

	for task in ${tasks[*]}
	do
		summand_shares=( $(task_shares "${task}" "${shares[*]}" "${classad_clauses[*]}") ) 

		local share_counter=-1

		for share in ${summand_shares[*]}
		do
			let share_counter++
		
			output_shares[${share_counter}]=$(( ${output_shares[${share_counter}]} + ${share} ))
		done
	done
	echo ${output_shares[@]}
}
# End of auxiliary functions

# Main
declare -a unique_database_nodes=( $(echo ${database_nodes[@]} | tr " " "\n" | sort | uniq) )

for unique_database_node in ${unique_database_nodes[@]}
do
	node_counter=-1
	declare -a shares=()

	for submission_node in ${submission_nodes[@]}
	do
		let node_counter++

		[[ "${database_nodes[$node_counter]}" != "$unique_database_node" ]] && continue	
		profile=${profiles[${node_counter}]}	

		[[ "${ask_password[$node_counter]}" == 0 ]] && { 
			share=$(ssh $SSH_FLAGS $submission_node ". \$HOME/${profile}; . \$PLATFORM_LOCATION/scripts/definitions.sh; echo \$HARD_MAXIMAL_RUNNING_JOBS")
			[[ $? != 0 ]] && echo "Problems with SSH in $submission_node" && share=0 
		}
		[[ "${ask_password[$node_counter]}" != 0 ]] && {
			share=$( "$DIRECTORY_NAME/Passwordless_ssh.sh" $submission_node "$DIRECTORY_NAME/${ask_password[$node_counter]}" \
					 ". \$HOME/${profile}; . \$PLATFORM_LOCATION/scripts/definitions.sh; echo \$HARD_MAXIMAL_RUNNING_JOBS" )
			[[ $? != 0 ]] && echo "Problems with SSH in $submission_node" && share=0
		}

		shares=( ${shares[@]} $share )
	done
	#workunits_left=$(ssh $SSH_FLAGS $unique_database_node mysql $MYSQL_FLAGS -e "'select count(*) from monitoring_id where status = \"Progress\" '") 
#MARK CHANGE - ADD DISALLOW SUBMISSION FLAG
#	workunits_left=0;
#
#    if [ ! -e ${LOG_DIR}/${DISALLOW_GRID_SUBMISSIONS_MARKER} ]; then
#		workunits_left=$(ssh $SSH_FLAGS $unique_database_node mysql $MYSQL_FLAGS -e "'select count(*) from result where server_state = 2'")
#		[[ $? -ne 0 ]] && echo "Unable to retrieve the number of workunits left" && exit ${WORKUNITS_LEFT_UNDETERMINED}
#	else
#		echo " ${LOG_DIR}/${DISALLOW_GRID_SUBMISSIONS_MARKER} found - submitting $workunits_left "
#	fi
#	
#	#bundle=$(bundle_size)
#	#[[ $? -ne 0 ]] && echo "Unable to retrieve the bundle size" && exit ${BUNDLE_SIZE_UNDETERMINED}
#	# Validating the number of workunits
#	[[ ! ( $workunits_left =~ ^[0-9]+$ ) ]] && echo "Workunits number is invalid" && continue 
	
	required_jobs_array=( $(tasks_shares "${shares[*]}") )

	shares_sum=$(( $(echo ${shares[@]} | tr " " "+") ))
	[[ $shares_sum -le 0 ]] && continue

	node_counter=-1
	share_counter=-1

	for submission_node in ${submission_nodes[@]}
	do
		let node_counter++

		[[ "${database_nodes[$node_counter]}" != "$unique_database_node" ]] && continue
		profile=${profiles[${node_counter}]}

		let share_counter++
		#required_jobs=$(ceiling $(echo "$workunits_left * ${shares[$share_counter]} / $shares_sum / ${bundle}" | bc -l) ) 
		#required_jobs=$(ceiling $(echo "$workunits_left * ${shares[$share_counter]} / $shares_sum" | bc -l) )
		required_jobs=${required_jobs_array[${share_counter}]}

		[[ "${ask_password[$node_counter]}" == 0 ]] && \
			ssh $SSH_FLAGS $submission_node ". \$HOME/${profile}; \$PLATFORM_LOCATION/scripts/set_maximal_jobs.sh $required_jobs"
		[[ "${ask_password[$node_counter]}" != 0 ]] && \
			"$DIRECTORY_NAME/Passwordless_ssh.sh" $submission_node "$DIRECTORY_NAME/${ask_password[$node_counter]}" \
					 ". \$HOME/${profile}; \$PLATFORM_LOCATION/scripts/set_maximal_jobs.sh $required_jobs"
	done
done
# End of main
