/* $Id: AlleleRecoding.c,v 1.4 2008/08/03 08:15:45 superlink Exp $ */
/* $Id: AlleleRecoding.c,v 1.4 2008/08/03 08:15:45 superlink Exp $ */
 
/*****************************************************************************/
/*  AlleleRecoding.c														 */
/*  Author: Ma'ayan Fishelson											     */
/*****************************************************************************/

#include "AlleleRecoding.h"

#if PERFORM_ALLELE_RECODING

/******************************************************************************
*
*       Function        performAlleleRecoding
*       Parameters		none
*       Output          none
*       Description		Performs allele-recoding on the pedigrees in "peds". 
*       Comments		none
*
******************************************************************************/
void performAlleleRecoding()
{
	pedigree	*pedPtr;
	person		*currPerson, *currPersonEnd;

	/* 1.	Loop over the pedigrees. For each one: */

	for (pedPtr = peds.head; pedPtr != NULL; pedPtr = pedPtr->nextPed) {

		/* 1.1.	Zero the auxiliary flags of the people in the pedigree. */

		initFlags(pedPtr);	

		/* 1.2.	Iterate over the people in the pedigree. */

		for (currPerson = pedPtr->people, currPersonEnd = currPerson + pedPtr->numPeople;
				currPerson < currPersonEnd; currPerson++) {
			
			/* 1.2.1.	Update the transmitted set of each locus of type 
						'alleles', where the person is typed . */

			updatePersonTransmitted(currPerson);
		}

		/* 1.3.	Traverse the pedigree in a bottom-up manner to update the 
				transmitted and non-transmitted allele sets of each untyped 
				person. */

		traverseUpTransmitted(pedPtr->childLess);
	}
}

/******************************************************************************
*
*       Function        updatePersonTransmitted
*       Parameters      currPerson (the person which is being updated).
*       Output          none
*       Description		updates the transmitted alleles of 'currPerson', at 
*						each 'alleles'-type locus in which this person is typed.
*       Comments		none
*
******************************************************************************/
void updatePersonTransmitted(person *currPerson)
{
	int				locusInd, numLoci;
	locus			*locusPtr;
	
	numLoci = loci.numLoci;

	/* 1.	Iterate over all loci. */

	for (locusInd = 0, locusPtr = loci.locusArr; locusInd < numLoci; locusInd++,
				locusPtr++) {

		/* 1.1.	Update the transmitted set of the person in each locus of type 
				'alleles', where the person is typed. */

		if (locusPtr->type == alleles) {

			/* 1.1.1. If the person is typed at this locus, move the alleles from
					  the non-transmitted sets to the transmitted sets, and update the
					  non-transmitted sets to NULL. */

			if ((currPerson->loci[locusInd]).unknown == FALSE) {	
				(((currPerson->loci[locusInd]).locusData).allelesL).paLocusTrans = 
							(((currPerson->loci[locusInd]).locusData).allelesL).paLocus;

				(((currPerson->loci[locusInd]).locusData).allelesL).paLocus = NULL;

				(((currPerson->loci[locusInd]).locusData).allelesL).maLocusTrans = 
							(((currPerson->loci[locusInd]).locusData).allelesL).maLocus;

				(((currPerson->loci[locusInd]).locusData).allelesL).maLocus = NULL;
			}
		}
	}
}


/******************************************************************************
*
*       Function        traverseUpTransmitted
*       Parameters      currPerson (the person that is being traversed).
*       Output          none
*       Description		This function traverses all the children of 'currPerson'.
*						If all children have been updated already (according to
*						their children), it updates 'currPerson' according to
*						them and then traverses up to 'currPerson's parents (if
*						they exist).
*       Comments		none
*
******************************************************************************/
void traverseUpTransmitted(person *currPerson)
{
	person	*ch;			
	sex		Sex;


	ch = currPerson->ch;
	Sex = currPerson->personSex;

	/* 1.	Mark this person as visited. */

	currPerson->visited = TRUE;			
	
	if (Sex == Male) {						
		
		/* 2.	Iterate over all the children of this person. */

		for (; ch != NULL; ch = ch->NeSiPa) {		
			
			/* 2.1.	If the child hasn't been visited yet, traverse down to it. 
					From this child we'll traverse back up to the parent later on. */

			if (ch->done == FALSE) {
				if (ch->visited == FALSE) 			
					traverseDownTransmitted(ch);
				return;									
			}
		}
	}
	else {			/* (Sex == Female) */

		/* 2.	Iterate over all the children of this person. */

		for (; ch != NULL; ch = ch->NeSiMa) {			
			
			/* 2.1.	If the child hasn't been visited yet, traverse down to it. 
					From this child we'll traverse back up to the parent later on. */

			if (ch->done == FALSE) { 
				if (ch->visited == FALSE) 
					traverseDownTransmitted(ch);
				return;
			}
		}
	}

	ch = currPerson->ch;

	/* 3.	All the children have been updated already. Update the parents now.*/

	if (Sex == Male) {

		/* 3.1.	Iterate over all of this person's children. */

		for (; ch != NULL; ch = ch->NeSiPa) {		
			
			/* 3.1.1.	If the parents haven't been updated by this child yet, 
						update them. */

			if ((ch->ma)->done == FALSE)	
				updateParentsTransmitted(currPerson, ch->ma, ch);
		}
	}
	else {		/* (sex == Female) */
		for (; ch != NULL; ch = ch->NeSiMa) {	
			if ((ch->pa)->done == FALSE)
				updateParentsTransmitted(ch->pa, currPerson, ch);	
		}
	}

	/* 4.	Mark this person as done. */

	currPerson->done = TRUE;			

	/* 6.	If currPerson has a father which hasn't been updated yet, go up. */

	if ((currPerson->pa != NULL) && ((currPerson->pa)->done == FALSE))	
		traverseUpTransmitted(currPerson->pa);

	/* 7.	If currPerson has a mother which hasn't been updated yet, go up. */

	if ((currPerson->ma != NULL) && ((currPerson->ma)->done == FALSE))	
		traverseUpTransmitted(currPerson->ma);
}



/******************************************************************************
*
*       Function        traverseDownTransmitted
*       Parameters      currPerson (the person that is being traversed).
*       Output          none
*       Description		If currPerson has no children, traverse up to him.
*						If its child hasn't been visited yet, traverse down
*						to the child. Otherwise, return.
*       Comments		none
*
******************************************************************************/
void traverseDownTransmitted(person *currPerson)
{
	person		*ch;
	
	ch = currPerson->ch;

	/* 1.	If this person hasn't got any children, than there's nothing to update
			regarding his transmitted alleles (he has no transmitted alleles).
			Otherwise:
				- if his child hasn't been visited yet, we traverse
				  down to it.  
				- Otherwise we return (we'll reach this peron later,
				  by another path). */

	if (ch == NULL)					
		traverseUpTransmitted(currPerson);
	else {
		if (ch->visited == FALSE) 	
			traverseDownTransmitted(ch);
	}
}


/******************************************************************************
*
*       Function        updateParentsTransmitted
*       Parameters      pa, ma, ch (father, mother & child)
*       Output          none
*       Description		Updates the parents according to the child.
*       Comments		none
*
******************************************************************************/
void updateParentsTransmitted(person *pa, person *ma, person *ch)
{
	int					locusInd, numLoci;
	halfLocusPerson		*alleleCh;
	locusPerson			*parentLocus;
	locus				*locusPtr;

	numLoci = loci.numLoci;

	/* 1.	Iterate over all loci. */

	for (locusInd = 0, locusPtr = loci.locusArr; locusInd < numLoci; locusInd++,
			locusPtr++) {

		/* 1.1.	If this is an 'alleles' type locus, update the parents according 
				to the child in this locus. */

		if (locusPtr->type == alleles) {

			/* 1.1.1.	If the father is untyped at this locus, update his transmitted
						alleles according to the child. */

			if ((pa->loci[locusInd]).unknown == TRUE) {
				alleleCh = (((ch->loci[locusInd]).locusData).allelesL).paLocusTrans;
				parentLocus = &(pa->loci[locusInd]);
				addTransmittedAlleles(alleleCh, parentLocus);
			}

			/* 1.1.1.	If the mother is untyped at this locus, update her transmitted
						alleles according to the child. */

			if ((ma->loci[locusInd]).unknown == TRUE) {
				alleleCh = (((ch->loci[locusInd]).locusData).allelesL).maLocusTrans;
				parentLocus = &(ma->loci[locusInd]);
				addTransmittedAlleles(alleleCh, parentLocus);
			}
		}
	}
}




/******************************************************************************
*
*       Function        addTransmittedAlleles
*       Parameters      alleleCh (the transmitted allele list of the child that
*						might have been inherited from the parent), 
*						locusParent (pointer to the locus struct of the parent).
*       Output          none
*       Description		Updates the transmitted alleles of the parent according 
*						to the child.
*       Comments		none
*
******************************************************************************/
void addTransmittedAlleles(halfLocusPerson *alleleCh, locusPerson *locusParent)
{
	halfLocusPerson	*alleleParent, *prevAlleleParent, *alleleParentTrans,
					*prevAlleleParentTrans, *newTransAllele, *tmpAlleleCh;


	/* 1.	Initialize 'alleleParent' to point to the beginning of the paternal 
			allele list of the parent. */

	prevAlleleParent = NULL;
	alleleParent = ((locusParent->locusData).allelesL).paLocus;

	/* 2.	Iterate over the allele list of the child. */

	for (tmpAlleleCh = alleleCh; tmpAlleleCh != NULL ; tmpAlleleCh = tmpAlleleCh->next) {

		/* 2.1.	Iterate over the paternal alleles of the parent (this list will hold 
				the non-transmitted paternal alleles eventually). */

		for (; (alleleParent != NULL) && (alleleParent->allele < tmpAlleleCh->allele); 
				prevAlleleParent = alleleParent, alleleParent = alleleParent->next);

		/* 2.2.	If the allele of the child is found in this list, delete it from 
				this list and add it to the list of transmitted paternal alleles of
				the parent. */

		if ((alleleParent != NULL) && (alleleParent->allele == tmpAlleleCh->allele)) {

			newTransAllele = alleleParent;
			
			/* 2.2.1. Remove the allele from the non-transmitted paternal allele
					  list of the parent, and update 'alleleParent' to point to
					  the next allele in the list. */

			if (prevAlleleParent == NULL) {
				((locusParent->locusData).allelesL).paLocus = alleleParent->next;
				alleleParent = ((locusParent->locusData).allelesL).paLocus;
			}
			else {
				prevAlleleParent->next = alleleParent->next;
				alleleParent = prevAlleleParent->next;
			}

			/* 2.2.2. Insert the allele into the transmitted paternal allele list of the
					  parent (the list is sorted in increasing order). */

			prevAlleleParentTrans = NULL;
			alleleParentTrans = ((locusParent->locusData).allelesL).paLocusTrans;

			for (; (alleleParentTrans != NULL) && (alleleParentTrans->allele < newTransAllele->allele); 
					prevAlleleParentTrans = alleleParentTrans, 
					alleleParentTrans = alleleParentTrans->next);

			if (prevAlleleParentTrans == NULL) {	/* the allele needs to be the first. */

				newTransAllele->next = alleleParentTrans;
				((locusParent->locusData).allelesL).paLocusTrans = newTransAllele;
			}
			else {
				prevAlleleParentTrans->next = newTransAllele;				
				newTransAllele->next = alleleParentTrans;					
			}	
		}
	}

	/* 3.	Initialize 'alleleParent' to point to the beginning of the maternal 
			allele list of the parent. */

	prevAlleleParent = NULL;
	alleleParent = ((locusParent->locusData).allelesL).maLocus;

	/* 4.	Iterate over the allele list of the child. */

	for (tmpAlleleCh = alleleCh; tmpAlleleCh != NULL ; tmpAlleleCh = tmpAlleleCh->next) {

		/* 4.1.	Iterate over the maternal alleles of the parent (this list will hold 
				the non-transmitted maternal alleles eventually). */

		for (; (alleleParent != NULL) && (alleleParent->allele < tmpAlleleCh->allele); 
				prevAlleleParent = alleleParent, alleleParent = alleleParent->next);

		/* 4.2.	If the allele of the child is found in this list, delete it from 
				this list and add it to the list of transmitted maternal alleles. */

		if ((alleleParent != NULL) && (alleleParent->allele == tmpAlleleCh->allele)) {

			newTransAllele = alleleParent;
			
			/* 4.2.1. Remove the allele from the non-transmitted maternal allele
					  list of the parent, and update 'alleleParent' to point to
					  the next allele in the list. */

			if (prevAlleleParent == NULL) {
				((locusParent->locusData).allelesL).maLocus = alleleParent->next;
				alleleParent = ((locusParent->locusData).allelesL).maLocus;
			}
			else {
				prevAlleleParent->next = alleleParent->next;
				alleleParent = prevAlleleParent->next;
			}

			/* 4.2.2. Insert the allele into the transmitted maternal allele list of the
					  parent (the list is sorted in increasing order). */

			prevAlleleParentTrans = NULL;
			alleleParentTrans = ((locusParent->locusData).allelesL).maLocusTrans;

			for (; (alleleParentTrans != NULL) && (alleleParentTrans->allele < newTransAllele->allele); 
					prevAlleleParentTrans = alleleParentTrans, 
					alleleParentTrans = alleleParentTrans->next);

			if (prevAlleleParentTrans == NULL) {	/* the allele needs to be the first. */

				newTransAllele->next = alleleParentTrans;
				((locusParent->locusData).allelesL).maLocusTrans = newTransAllele;
			}
			else {
				prevAlleleParentTrans->next = newTransAllele;				
				newTransAllele->next = alleleParentTrans;					
			}	
		}
	}
}

#endif

