/* $Id: ArgMax.c,v 1.2 2008/08/03 08:15:45 superlink Exp $ */
/*****************************************************************************/
/*  ArgMax.c																 */
/*  Author: Ma'ayan Fishelson											     */
/*****************************************************************************/

#include "ArgMax.h"

/******************************************************************************
*
*       function        createAMData
*       parameters      type (the type of 'argMaxData' created), mainNode (the 
*						node in which elimination this struct is created), 
*						next (pointer to the 'argMaxData' created for the previous
*						variable in the elimination order).
*       output          none
*       description		Allocates and initializes the common fields of an 
*						'argMaxData' structure.
*       comments		none
*
******************************************************************************/
argMaxData createAMData(argMaxType type, node *mainNode, argMaxData next)
{
	argMaxData		argMax;

	/* 1.	Allocate memory for the structure. */

	myMalloc(argMax, sizeof(struct argMax));

	/* 2.	Update its fields. */

	argMax->type = type;
	argMax->mainNode = mainNode;
	argMax->next = next;

	/* 3.	Return a pointer to the structure created. */

	return argMax;
}

/******************************************************************************
*
*       function        createAMDataElim
*       parameters      type (the type of 'argMaxData' created), mainNode (the 
*						node in which elimination this structure is created), 
*						next (pointer to the 'argMaxData' created for the previous
*						variable in the elimination order), numNodes (the number 
*						of nodes this argMax function depends on), nodes (an array
*						with data on the nodes this function depends on), 
*						sizeOfTable (the size of this argMax function), 
*						argMaxVals (the list of values of 'mainNode' which 
*						maximized the original prbability function for each value
*						assignment to the rest of the nodes the function depends on).  
*       output          none
*       description     Allocates and initializes the fields of an 
*						'argMaxData' structure of type 'elimNode'.
*       comments		none
*
******************************************************************************/
argMaxData createAMDataElim(node *mainNode, argMaxData next, int numNodes, 
							argMaxNodeData *nodes, unsigned long sizeOfTable, 
							int *argMaxVals)
{
	argMaxData				argMax;

	/* 1.	Allocate memory for the structure, and update the fields that are
			common to all types of 'argMaxData'. */

	argMax = createAMData(elimNode, mainNode, next);

	/* 2.	Update the fields specific to 'argMaxData' structure 
			of type 'elimNode'. */

	((argMax->specificData).elimNodeData).numNodes = numNodes;
	((argMax->specificData).elimNodeData).nodes = nodes;
	((argMax->specificData).elimNodeData).sizeOfTable = sizeOfTable;
	((argMax->specificData).elimNodeData).argMaxVals = argMaxVals;

	/* 3.	Return a pointer to the structure created. */

	return argMax;
}

/******************************************************************************
*
*       function        createAMDataOther
*       parameters      type (the type of 'argMaxData' created), mainNode (the 
*						node in which elimination this structure is created), 
*						next (pointer to the 'argMaxData' created for the previous
*						variable in the elimination order), numNodes (the number 
*						of nodes whose values were eliminated after the value
						of this break node was set), nodes (an array with data
*						on those nodes and the values that were eliminated).
*       output          none
*       description     Allocates and initializes the fields of an 
*						'argMaxData' structure of type 'breakNode'.
*       comments		none
*
******************************************************************************/
argMaxData createAMDataOther(node *mainNode, argMaxData next, int numNodes, 
							 argMaxElimValsVar *nodes, argMaxType type)
{
	argMaxData				argMax;

	/* 1.	Allocate memory for the structure, and update the fields that are
			common to all types of 'argMaxData'. */

	argMax = createAMData(type, mainNode, next);

	/* 2.	Update the fields specific to 'argMaxData' structure 
			of type 'breakNode'. */

	((argMax->specificData).otherNodeData).numNodes = numNodes;
	((argMax->specificData).otherNodeData).nodes = nodes;

	/* 3.	Return a pointer to the structure created. */

	return argMax;
}


/******************************************************************************
*
*       function        freeAMData
*       parameters      argMax (pointer to an 'argMaxData' structure whose memory 
*						needs to be freed).
*       output          none
*       description     Frees the memory allocated dynamically to 'argMax'
*       comments		none
*
******************************************************************************/
void freeAMData(argMaxData argMax)
{
	argMaxElimValsVar	*nodePtr, *nodePtrEnd;

	/* 1.	If there's nothing to free, return. */

	if (argMax == NULL) {
		return;
	}

	/* 2.	Free the memory allocated to the internal fields according to the type
			of 'argMaxData'. */

	if (argMax->type == elimNode) {
		myFree(((argMax->specificData).elimNodeData).nodes);
		myFree(((argMax->specificData).elimNodeData).argMaxVals);
	}
	else {
		for (nodePtr = ((argMax->specificData).otherNodeData).nodes,
				nodePtrEnd = nodePtr + ((argMax->specificData).otherNodeData).numNodes;
				nodePtr < nodePtrEnd; nodePtr++) {	

			myFree(nodePtr->elimVals);
		}
		myFree(((argMax->specificData).otherNodeData).nodes);
	}

	/* 3.	Free the memory allocated to the 'argMaxData' structure itself. */

	myFree(argMax);
}


/******************************************************************************
*
*       function        freeAMList
*       parameters      head (pointer to the head of a list of 'argMaxData'
*						structures).
*       output          none
*       description     Frees the memory allocated dynamically to the list 
*						pointed to by 'head'.
*       comments		none
*
******************************************************************************/
void freeAMList(argMaxData head)
{
	argMaxData	curr, next;

	/* 1.	Initialize a pointer to the beginning of the list. */

	curr = head;

	/* 2.	As long as there are still elements in the list: */

	while (curr != NULL) {

		/* 2.1.	Update a pointer to the next element. */

		next = curr->next;

		/* 2.2.	Free the current element. */

		freeAMData(curr);

		/* 2.3.	Update the pointer to the current element. */

		curr = next;
	}	
}


/******************************************************************************
*
*       function        retrieveArgMaxData
*       parameters      argMax (pointer to an 'argMaxData' structure).
*       output          none
*       description     Retrieves the data stored during the backward phase of
*						the MPE-elimination algorithm in the 'argMaxData' structure
*       comments		This is performed during the forward stage of the 
*						MPE-elimination algorithm.
*
******************************************************************************/
void retrieveArgMaxData(argMaxData argMax)
{
	argMaxType	type;

	type = argMax->type;
	
	/* 1.	If this isn't a dummy node, update the optimizing index of the node. */
	
	if (type != dummyNode) {
		setMaxValNode(argMax);
	}

	/* 2.	If this is a breakNode or a dummyNode, update the optimizing index of
			nodes whose values were eliminated. */

	if (type != elimNode) {
		updateMaxValsNodes(argMax);
	}
}


/******************************************************************************
*
*       function        setMaxValNode
*       parameters      argMax (pointer to an 'argMaxData' structure).
*       output          none
*       description     Updates the optimizing index of the node that 'argMax'
*						represents. 
*       comments		This is performed during the forward stage of the  
*						MPE-elimination algorithm.
*
******************************************************************************/
void setMaxValNode(argMaxData argMax)
{
	if (argMax->type == breakNode) {
		(argMax->mainNode)->currVal = ((argMax->specificData).otherNodeData).val;
	}	
	else {
		calcMaxVal(argMax);
	}
}


/******************************************************************************
*
*       Function        calcMaxVal
*       Parameters		argMax (pointer to an argMaxData structure).
*       Output          none
*       Description		Finds the optimizing value of the node that 'argMax'
*						represents.					 
*       Comments		none
*
******************************************************************************/
void calcMaxVal(argMaxData argMax)
{
	argMaxNodeData	*nodePtr, *nodePtrEnd;	
	unsigned long	multFactor,ind;
	
	/* 1.	Initialize the index and the multFactor of the function. */

	ind = 0;
	multFactor = ((argMax->specificData).elimNodeData).sizeOfTable;

	/* 2.	Iterate over the function's nodes and add the contribution of each 
			node to the index. */

	for (nodePtr = ((argMax->specificData).elimNodeData).nodes, 
			nodePtrEnd = nodePtr + ((argMax->specificData).elimNodeData).numNodes;
			nodePtr < nodePtrEnd; nodePtr++) {

		multFactor /= nodePtr->currNumVals;
	#ifdef WIN32
		assert((nodePtr->origNode)->currVal >= 0);
	#endif
		ind += (multFactor * (nodePtr->origNode)->currVal);
	}

	/* 3.	Update the optimizing value determined for this node (actually, the index 
			of the optimizing value in the array of possible values for this node) . */

	(argMax->mainNode)->currVal = 
			((argMax->specificData).elimNodeData).argMaxVals[ind];

	#ifdef WIN32
		assert((argMax->mainNode)->currVal >= 0);
	#endif
}


/******************************************************************************
*
*       function        updateMaxValsNodes
*       parameters      argMax (pointer to an 'argMaxData' structure).
*       output          none
*       description     Updates the optimizing values of the nodes according to variable
*						values that have been eliminated during value elimination.
*						
******************************************************************************/
void updateMaxValsNodes(argMaxData argMax)
{
	argMaxElimValsVar	*nodePtr, *nodePtrEnd;
	node				*currNode;
	int					*elimValPtr, *elimValPtrEnd;

	/* 1.	Iterate over all the nodes whose values were eliminated. */

	for (nodePtr = ((argMax->specificData).otherNodeData).nodes,
			nodePtrEnd = nodePtr + ((argMax->specificData).otherNodeData).numNodes;
			nodePtr < nodePtrEnd; nodePtr++) {

		currNode = nodePtr->origNode;

		/* 1.1.	For each node, iterate over all the values that were eliminated. */

		for (elimValPtr = nodePtr->elimVals, 
				elimValPtrEnd = elimValPtr + nodePtr->numElimVals;
				elimValPtr < elimValPtrEnd; elimValPtr++) {
			
			/* 1.1.1.	If the value eliminated is lower or equal to the optimizing
						value, then increase it by 1. Otherwise, break 
						(the value are in increasing order). */

			if (*elimValPtr <= currNode->currVal) {
				currNode->currVal++;
			}
			else {
				break;
			}
		}
	}
}

