/* $Id: Auxiliary.c,v 1.2 2008/08/03 08:15:45 superlink Exp $ */
/*****************************************************************************/
/*  Auxiliary.c																 */
/*  Author: Ma'ayan Fishelson				                                 */
/*****************************************************************************/

#include "Auxiliary.h"


/******************************************************************************
*
*       function        mallocErr
*       parameters      message (the error message printed before exiting).
*       output          none
*       description     Prints an error message when malloc fails to allocate
*						memory, and exits.
*       comments		none
*
******************************************************************************/
void mallocErr(char *message)
{
	fprintf(stderr, "\nProblem with malloc, probably not enough space.\n");
	fprintf(stderr, "Problem occurred when allocating %s.\n", message);

#if MANAGE_MEM
	freeAllBlocks();
	longjmp(jumper, 1);
#endif 
	
	exit(EXIT_FAILURE);
}


/******************************************************************************
*
*       function        reallocErr
*       parameters      message (the error message printed before exiting).
*       output          none
*       description     Prints an error message when reallocErr fails to allocate
*						additional space, and exits.
*       comments		none
*
******************************************************************************/
void reallocErr(char *message)
{
	fprintf(stderr, "\nProblem with realloc, probably not enough space.\n");
	fprintf(stderr, "Problem occurred when allocating %s.\n", message);

#if MANAGE_MEM
	freeAllBlocks();
	longjmp(jumper, 1);
#endif

	exit(EXIT_FAILURE);
}


/******************************************************************************
*
*       function        inputErr
*       parameters      message (the error message printed before exiting).
*       output          none
*       description     Prints an error message when there's an input error,
*						and exits.
*       comments		none
*
******************************************************************************/
void inputErr(char *message, ...) 
{
	va_list		ap;


	/* 1.	Set a pointer to the beginning of the argument list. */

	va_start(ap, message);

	/* 2.	Print the requested string to stderr. */

	fprintf(stderr, "\nAn input error has been found:\n");
	vfprintf(stderr, message, ap);

	/* 4.	Reset the pointer to the beginning of the argument list. */

	va_end(ap);

	/* 5.	Exit. */

	exit(EXIT_FAILURE);
}


/******************************************************************************
*
*       function        inputWarning
*       parameters      message (the warning message printed).
*       output          none
*       description     Prints a warning message when there might be an input 
*						error.
*       comments		none
*
******************************************************************************/
void inputWarning(char *message, ...) 
{
	va_list		ap;

	/* 1.	Set a pointer to the beginning of the argument list. */

	va_start(ap, message);

	/* 2.	Print the requested string to stderr. */

	fprintf(stdout, "\nA possible input error has been found:\n");
	vfprintf(stdout, message, ap);

	/* 3.	Reset the pointer to the beginning of the argument list. */

	va_end(ap);
}


/******************************************************************************
*
*       function        error
*       parameters      message (the error message printed before exiting).
*       output          none
*       description     Prints an error message and exits.
*       comments		none
*
******************************************************************************/
void error(char *message, ...) 
{
	va_list		ap;

	/* 1.	Set a pointer to the beginning of the argument list. */

	va_start(ap, message);

	/* 2.	Print the requested string to stderr. */

	fprintf(stderr, "\nError: ");
	vfprintf(stderr, message, ap);

	/* 3.	Reset the pointer to the beginning of the argument list. */

	va_end(ap);

	/* 4.	Exit. */

	exit(EXIT_FAILURE);
}


/******************************************************************************
*
*       Function        P_eof
*       Parameters		FILE *f
*       Output          none
*       Description		Tests for end of file. 
*       Comments		(this function is also used in "unknown.c" in Fastlink)
*
******************************************************************************/
int P_eof(FILE *f)
{
    register int ch;

    if (feof(f))
        return 1;
    if (f == stdin)
        return 0;    /* not safe to look-ahead on the keyboard! */
    ch = getc(f);
    if (ch == EOF)
        return 1;
    
	ungetc(ch, f);

    return 0;
}

/******************************************************************************
*
*       Function        swap
*       Parameters      val1, val2 (the two arguments whose values need to be 
*						swapped).
*       Output          none
*       Description		Swaps between the values of 'val1' & 'val2'.
*       Comments		none
*
******************************************************************************/
void swap(int *val1, int *val2)
{
	int temp = *val1;

	*val1 = *val2;
	*val2 = temp;
}


/******************************************************************************
*
*       Function        getLnLike
*       Parameters      resultNode (the results node whose lnLikelihood is 
*						requested).
*       Output          none
*       Description		Returns the lnlikelihood of the results node 'resultNode'.
*       Comments		none
*
******************************************************************************/
double getLnLike(resNode *resultNode)
{
	return resultNode->lnLike;
}


/******************************************************************************
*
*       Function        getLodScore
*       Parameters      resultNode (the results node whose lod-score is 
*						requested).
*       Output          none
*       Description		Returns the lod-score of the results node 'resultNode'.
*       Comments		none
*
******************************************************************************/
double getLodScore(resNode *resultNode)
{
	return resultNode->lodScore;
}


/******************************************************************************
*
*       Function        myPrintf
*       Parameters      str (a format specification), ... (the additional arguments).
*       Output          none
*       Description		Prints the string passed to it into stdout and into a 
*						results file.
*       Comments		none
*
******************************************************************************/
void myPrintf(char *str,...)
{
	va_list		ap;


	/* 1.	Set a pointer to the beginning of the argument list. */

	va_start(ap, str);

	/* 2.	Print the requested string to stdout. */

	vfprintf(stdout, str, ap);

	/* 3.	Print the requested string to the results file. */

	vfprintf(resFile, str, ap);

	/* 4.	Reset the pointer to the beginning of the argument list. */

	va_end(ap);
}


