/* $Id: EliminationOrder.h,v 1.2 2008/08/03 08:15:46 superlink Exp $ */
/*****************************************************************************/
/*  EliminationOrder.h														 */
/*  Author: Ma'ayan Fishelson												 */
/*****************************************************************************/

#ifndef _ELIMINATION_ORDER_H_
#define _ELIMINATION_ORDER_H_

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include "PedData.h"
#include "EliminationOrderGraph.h"
#include "Mem.h"


/*------------------------------*/
/*		DEFINITIONS				*/
/*------------------------------*/

/* A file that is used to gather different statistics on the elimination-order finding 
   process (used for development of the software). */

#define STATISTICS_FILE_NAME  "statistics.csv"

#define	MAX_ELIMINATION_COMPLEXITY	16		/* The number of different elimination 
											   complexity classes. */

#define	VAR_POOL_SIZE				3		/* The size of a variable pool from which
											   the next variable is chosen in each iteration. */

#define MIN_IMPROVEMENT_RATE		0.0001	/* The minimum improvement rate required for continuing 
											   with the current probability function. */
#define	NUM_ALGS					3		/* The number of heuristic algorithms executes to 
											   determine an elimination order. */

/* The codes of the different algorithms. */
#define	MW_IND						0
#define	WMF_IND						1
#define	MF_IND						2

/* The percentage of time spent executing each one of the algorithms. */

#define	PERCENT_MW					0.5		/* Stochastic-Minimum-Weight heuristic. */
#define	PERCENT_WMF					0.25	/* Stochastic-Weighted-Minimum-Fill heuristic. */
#define	PERCENT_MF					0.25	/* Stochastic-Minimum-Fill heuristic. */

/* Definitions used for the case where the elimination order is read from a file,
   instead of being determined by the program. */

#define	MAX_LINE					50
#define	DEFAULT_ELIMINATION_ORDER_NAME	"eliminationOrder.txt"

/* Returns the cost of the elimination sequence according to the sum of table sizes with
   penalty according to the number of break nodes (used for comparing different elimination
   sequences). */

#define	elimOrderCost(elimOrderPed)	\
	elimOrderPed.sumTableSizes * pow(2.0, elimOrderPed.numBreakNodes)

/* A type of function for determining the heuristics that will be used for finding an 
   elimination order. */

typedef boolean (determineOptHeuristic)(int *complexityCoefficient, pedigree *currPed, 
										elimOrderProduct *elimProd, elimOrderPed *elimOrderPedPtr, 
										elimOrderPed *tempElimOrderPed, int numVariables, 
										int numLoci, double *tableSizes, long *elimVarsSegArr, 
										long *numVarsSegArr);

/*------------------------------*/
/*		DATA-STRUCTURES			*/
/*------------------------------*/

/* This structure holds information on the elimination order finding process. */

typedef struct bestResults {
	int			lastImprovementIter;		/* The index of the last iteration that improved
											   the elimination order (the sum of table sizes). */
	double		sumTableImprovement;		/* The improvement in the sum of table sizes from 
											   the last improvement. */
} bestResults;


/* This structure holds the parameters of the heuristic currently used and information
   on the best elimination order found. */

typedef struct currHeuristicEO {
#ifdef PARALLEL
  boolean         stopAfterEstimate;                      /*      A flag to indicate not to continue */
    int indexNum;                             /* index to be added to all output files with metadata.if < 0 -> nothing added */
  int totalTime;      /*Total time allowed (max) for order search*/
  int startTime;      /*Time when the program started*/
#endif

	boolean		isAutomatic;				/*	A flag which indicates whether the optimization policy
												would be determined automatically, by the program, 
												according to the problem at hand, or by the user 
												(in the command line). */
	int			maxNumIter;					/* The maximum number of iterations performed. */

	int			numIterAlgs[NUM_ALGS];		/* The maximum number of iterations performed by each algorithm. */
	double		sumTableSizesAlgs[NUM_ALGS];/* Holds the best results (sum of table sizes) 
											   found by the each of the algorithms. */
	double		sumTableSizesDG;			/* The sum of table sizes of the elimination sequence 
											   found by the deterministic-min-weight-greedy heuristic. */
	int			numBreakNodes;				/* The number of break nodes (variables that are conditioned
											   on) in the elimination sequence found. */	
#if (PRINT_VARIOUS_STATISTICS_LEVEL > 1)
	int			treeWidthDG;				/* The treewidth of the elimination sequence found by the
												 deterministic algorithm. */
	int			treeWidthAlgs[NUM_ALGS];	/* The treewidths of the elimination sequences 
											   found by the other stochastic algorithms. */
#endif
} currHeuristicEO;


/* This structure defines the optimization policy applied to a problem in the search for an 
   elimination sequence. */

typedef struct optimizationPolicyDef {
	int			maxNumIter;				/* The maximum number of iterations performed. */
	double		maxOptimizationTime;	/* The maximum time spent on optimizing the 
										   elimination order. */
} optimizationPolicyDef;

/* A type of a function that calculates the cost of a vertex in the graph. */

typedef double (costFunc)(graph *G, vertex *v);

/* A type of a function that updates an array with the variables in the order of
   locus by locus, first the phenotype and genetic-loci variables of the first locus,
   then the selector variables of the first locus, then the phenotype and genetic-loci 
   variables of the second locus, and so on. */

typedef void (updateElimVarsLSeg)(long *elimVarsSegArr, long *numVarsSegArr,
								  elimOrderProduct *elimProd, pedigree *currPed, 
								  int numLoci);

/*------------------------------*/
/*		GLOBAL VARIABLES		*/
/*------------------------------*/

/* File descriptor for statistics file. */

#if PRINT_VARIOUS_STATISTICS_LEVEL
	extern FILE *statisticsFile;
#endif 

double mem_avl;

extern	currHeuristicEO			heuristicEO;
optimizationPolicyDef			*optimizationPolicyArr, *optimizationPolicyDefPtr;

extern FILE		*outFile;
extern boolean	readElimOrder;
extern char		*eliminationOrderFileName;

#ifdef PARALLEL
extern long currentPermuationSid; /*permutation number of conditioning set for parallelization*/
extern int numOfSplitNodes; /* depth of the parallelization */
#endif


extern char	*datafileName, *pedfileName;

extern double	breakProductBottomThreshold;

boolean exitAfterFindOrder;		

/*------------------------------*/
/*		FUNCTIONS				*/
/*------------------------------*/

void setEOOptimizationPolicy(int argc, char* argv[]);

#if PRINT_VARIOUS_STATISTICS_LEVEL
	void openInitEOStatisticsFile();
	void printStatisticsResults(pedigree *ped, int numLoci, 
							double findElimOrderTimeIter, product *prod, 
							double sumTableSizesPre);
#endif

optimizationPolicyDef* configureOptimizationPolicy();
elimOrder determineEliminationOrder(product** prodArr, int numLoci,
									updateElimVarsLSeg updateElimVarsLSegFunc,
									determineOptHeuristic determineOptHeuristicFunc);

boolean determineElimOrderHeuristic(int *complexityCoefficient, pedigree *currPed, 
									elimOrderProduct *elimProd, elimOrderPed *elimOrderPedPtr, 
									elimOrderPed *tempElimOrderPed, int numVariables, 
									int numLoci, double *tableSizes, long *elimVarsSegArr, 
									long *numVarsSegArr);
boolean determineElimOrderHeuristicHaplo(int *complexityCoefficient, pedigree *currPed, 
										 elimOrderProduct *elimProd, 
										 elimOrderPed *elimOrderPedPtr,  
										 elimOrderPed *tempElimOrderPed, int numVariables, 
									     int numLoci, double *tableSizes, 
										 long *elimVarsSegArr, long *numVarsSegArr);

long performAdditionalIterAuto(int *complexityCoefficient, double findElimOrderTimeIter, 
	 						   elimOrderPed *elimOrderPedPtr, elimOrderProduct *elimProd,
							   pedigree *currPed, elimOrderPed *tempElimOrderPed, 
							   long numVariables, int numLoci, double *tableSizes, 
							   long *elimVarsSegArr, long *numVarsSegArr);
long performAdditionalIterManual(int *complexityCoefficient, double findElimOrderTimeIter, 
								 elimOrderPed *elimOrderPedPtr, elimOrderProduct *elimProd,
								 pedigree *currPed, elimOrderPed *tempElimOrderPed, 
								 long numVariables, int numLoci, double *tableSizes, 
								 long *elimVarsSegArr, long *numVarsSegArr);

void updateNewOrd(elimOrderPed *finalElimOrd, elimOrderPed *tempElimOrd, 
				  boolean *foundBetterOrd, int *complexityCoefficient,
				  int iterCnt, char *alg, int algInd);


#if PERFORM_PREPROCESSING
	int preprocessEOPed(elimOrderProduct *elimProd, elimOrderPed *elimOrd,
						long *sumTableSizesPre);
	int simplicialElim(elimOrderProduct *elimProd, elimOrderPed *elimOrd,
					   long *sumTableSizesPre, double *low);
	int almostSimplicialElim(elimOrderProduct *elimProd, elimOrderPed *elimOrd,
							long *sumTableSizesPre, double low);
	boolean checkASimp(elimOrderVar *var, double low);
#endif 

#ifdef PARALLEL
/**compute complexity taking into account split nodes***/
double  parallelSumTableSizes(double *tableSizes, elimOrderPed *elimOrd,
                              long firstElimInd, int splitDepth, 
			      int totalBreakNodes);

#endif

void updateSumTableSizes(double *tableSizes, elimOrderPed *elimOrd, 
						   long firstElimInd);
void updateElimVarsLSegAutosomal(long *elimVarsSegArr, long *numVarsSegArr,
								 elimOrderProduct *elimProd, pedigree *currPed, 
								 int numLoci);
void updateElimVarsLSegSexlink(long *elimVarsSegArr, long *numVarsSegArr,
							   elimOrderProduct *elimProd, pedigree *currPed, 
							   int numLoci);
void updateElimVarsLSegAutosomalHaplo(long *elimVarsSegArr, long *numVarsSegArr,
									  elimOrderProduct *elimProd, pedigree *currPed, 
									  int numLoci);
void updateElimVarsLSegSexlinkHaplo(long *elimVarsSegArr, long *numVarsSegArr,
									elimOrderProduct *elimProd, pedigree *currPed, 
									int numLoci);

void freeEliminationOrderStructs(elimOrder *elimOrderArrs, int numPeds);
void moveSelectorBreakNodesToStart(elimOrderPed *elimOrderPedPtr, int numSeg, 
								   long *numVarsSegArr, int numNonFounders);

elimOrder updateEliminationOrder(product** prodArr, char *elimOrderFileName);
void parseLine(char *line, NodeSex *nodeSex, NodeType *nodeType, int *personNum, 
			   int *locusNum);



void MinWeightSearch(elimOrderProduct *elimProd, elimOrderPed *elimOrd, 
					 pedigree *currPed, int numLoci, int firstElimInd,
					 boolean isStochastic, double *tableSizes, long *elimVarsSegArr, 
					 long *numVarsSegArr, double bestSize);
boolean determineEO_MW(elimOrderProduct *elimProd, pedigree *currPed,
					   elimOrderPed *finalElimOrd, elimOrderPed *tempElimOrd, 
					   int firstElimInd, long varMappingSize, 
					   int numLoci, double *tableSizes, long *elimVarsSegArr, 
					   long *numVarsSegArr, int numIter);
boolean determineEO_MW_Auto(elimOrderProduct *elimProd, pedigree *currPed,
						    elimOrderPed *finalElimOrd, elimOrderPed *tempElimOrd, 
						    int firstElimInd, long varMappingSize, 
						    int numLoci, double *tableSizes, long *elimVarsSegArr, 
						    long *numVarsSegArr, long maxIterations, long* iterationsLeft);


boolean determineEO_MF(elimOrderProduct *elimProd, pedigree *currPed, 
					   elimOrderPed *finalElimOrd, elimOrderPed *tempElimOrd, 
					   int firstElimInd, long varMappingSize, int numLoci,
					   double	*tableSizes, long *elimVarsSegArr, 
					   long *numVarsSegArr, int numIter, costFunc cFunc, 
					   boolean isWeighted);
boolean determineEO_MF_Auto(elimOrderProduct *elimProd, pedigree *currPed, 
						    elimOrderPed *finalElimOrd, elimOrderPed *tempElimOrd, 
						    int firstElimInd, long varMappingSize, int numLoci, 
						    double *tableSizes, long *elimVarsSegArr, 
						    long *numVarsSegArr, costFunc cFunc, boolean isWeighted,long maxIterations, long* iterationsLeft);

double calcWMinFillCost(struct graph *G, vertex	*vertexPtr);
double calcMinFillCost(graph *G, vertex	*vertexPtr);



#define needToElimNode(currNode, prod) \
	((currNode != NULL) && \
	 (currNode->varInProd != NULL) && \
	 ((currNode->varInProd)->used == TRUE) && \
	 (prod->oldToNewVarMapping[currNode->varInProd - prod->firstVar] != NULL) && \
	 (prod->oldToNewVarMapping[currNode->varInProd - prod->firstVar]->used == TRUE))

#ifdef PARALLEL

void printElimOrder(elimOrderPed *elimOrderPedPtr);
#else
void printElimOrder(elimOrderPed *elimOrderPedPtr, char *family);
#endif


#endif		/*	_ELIMINATION_ORDER_H_	*/



