/* $Id: EliminationOrderGraph.c,v 1.2 2008/08/03 08:15:46 superlink Exp $ */
/*********************************************************************************/
/*  EliminationOrderGraph.c											     	     */
/*  Author: Ma'ayan Fishelson					                                 */
/*********************************************************************************/

#include "EliminationOrderGraph.h"


/******************************************************************************
*
*       Function        createGraph
*       Parameters		elimProd (the elimination-order-product),
*						createAdjMatrix (a flag that indicates whether to create
*						an adjacency matrix).
*       Output          none
*       Description		Translates the elimination-order-product into a graph
*						representation and returns it.
*       Comments		none.
*
******************************************************************************/
struct graph* createGraph(elimOrderProduct *elimProd, boolean createAdjMat, 
						  long varMappingSize)
{
	struct graph	*G;
	elimOrderVar	*varPtr, *varPtrEnd, **varArr, **neighPtr, **neighPtrEnd;
	vertex			*vertexPtr;
	int				vNum, numNeighbors, ind;

#if UNITE_VARS
	vertex			**uniteVertices, **uniteVertexPtr;
	node			**uniteNodePtr, **uniteNodePtrEnd, *origNode;
#endif

	/* 1.	Allocate memory for the graph representation. */

	myMalloc(G, sizeof(struct graph));
	myMalloc(G->vertexList, sizeof(vertex) * elimProd->currNumVars);
	myMalloc(G->varToVertexMap, sizeof(vertex*) * varMappingSize);

	for (ind = 0; ind < varMappingSize; ind++) {
		G->varToVertexMap[ind] = NULL;
	}

	G->currNumVertices = G->numVertices = elimProd->currNumVars;

	myMalloc(varArr, sizeof(elimOrderVar*) * elimProd->currNumVars); 

	/* 2.	Iterate over all the variables of the product. */

	for (varPtr = elimProd->variables, varPtrEnd = varPtr + elimProd->currNumVars,
			vertexPtr = G->vertexList, vNum = 0; varPtr < varPtrEnd; varPtr++, 
			vertexPtr++, vNum++) {

		/* 2.1.	Update the pointer to the variable in the elimination-order-product 
				which corresponds to this vertex. */

		vertexPtr->varInProd = varPtr;
		vertexPtr->vNum = vNum;
		vertexPtr->numNeighbors = 0;
		vertexPtr->cost = INVALID;
		(vertexPtr->varInProd)->tableSize = INVALID;
		vertexPtr->isMarked = FALSE;
		vertexPtr->uniteVertices = NULL;

		/* 2.2.	Update the mapping from variables in the real product to vertices
				in the graph. */

		G->varToVertexMap[varPtr->indInRealProd] = vertexPtr;

		/* 2.3.	Allocate memory for the list of neighbors of the vertex and update
				the list-of-neighbors and the number-of-neighbors fields. */

		getVarsInCombinedElimOrderFunc(varPtr->funcs, varPtr->currNumFunctions, varArr, 
									   &numNeighbors, varPtr);

		vertexPtr->neighborList = NULL;

		if (numNeighbors > 0) {

			for (neighPtr = varArr, neighPtrEnd = neighPtr + numNeighbors;
					neighPtr < neighPtrEnd; neighPtr++) {
				addNeighbor(vertexPtr, (*neighPtr) - elimProd->variables);
			}
		}
	}

	myFree(varArr);

	/* 5.	For each group of vertices that need to be treated jointly, allocate an array
			that would point to all the vertices, initialize it, and update the whole group
			to point to it. */

	#if UNITE_VARS
		for (varPtr = elimProd->variables, varPtrEnd = varPtr + elimProd->currNumVars,
				vertexPtr = G->vertexList, vNum = 0; varPtr < varPtrEnd; varPtr++, 
				vertexPtr++, vNum++) {

			origNode = varPtr->originalNode;

			if ((origNode->uniteT) && (vertexPtr->uniteVertices == NULL)) {
				myMalloc(uniteVertices, sizeof(vertex*) * origNode->numVarsUnite);

				for (uniteVertexPtr = uniteVertices, uniteNodePtr = origNode->uniteVars, 
						uniteNodePtrEnd = uniteNodePtr + origNode->numVarsUnite;
						uniteNodePtr < uniteNodePtrEnd; uniteNodePtr++, uniteVertexPtr++) {

					(*uniteVertexPtr) = 
							G->varToVertexMap[(*uniteNodePtr)->varInProd - elimProd->firstVar];
					(*uniteVertexPtr)->uniteVertices = uniteVertices;
				}
			}
		}
	#endif

	if (createAdjMat) {
		createAdjMatrix(G);
	}

	/* 3.	Return a pointer to the graph. */

	return G;
}


/******************************************************************************
*
*       Function        createAdjMatrix
*       Parameters		G (a graph representation of the elimination-order-product).
*       Output          none
*       Description		Creates an adjacency matrix for the graph.
*       Comments		none.
*
******************************************************************************/
void createAdjMatrix(struct graph *G)
{
	int				sizeRow;
	adjMatrixT		**adjMatrixTPtr;
	vertex			*vertexPtr, *vertexPtrEnd;
	neighbor		*neighborPtr;

	/* 1.	Allocate memory for the adjacency matrix data-structure. */

	myMalloc(G->adjMat, sizeof(adjMatrix*) * G->numVertices); 

	/* 2.	Iterate over the rows of the matrix (one row for each vertex of 
			the graph) and allocate memory for them. Only half of the matrix
			(one triangle) is stored. */

	for (adjMatrixTPtr = G->adjMat, sizeRow = G->numVertices;
			sizeRow > 0; adjMatrixTPtr++, sizeRow--) {

		myMalloc((*adjMatrixTPtr), (sizeRow/8 + 1) * sizeof(adjMatrixT));
		memset(*adjMatrixTPtr, 0, (sizeRow/8 + 1) * sizeof(adjMatrixT));
	}

	/* 3.	Iterate over the vertices of the graph and update the adjacency matrix. */

	for (vertexPtr = G->vertexList, vertexPtrEnd = vertexPtr + G->numVertices;
			vertexPtr < vertexPtrEnd; vertexPtr++) {
		
		/* 3.1.	Iterate over all the neighbors of the vertex: */

		for (neighborPtr = vertexPtr->neighborList; neighborPtr != NULL; 
				neighborPtr = neighborPtr->next) {

			/* 3.1.1.	If the number of the vertex is smaller than that of the
						neighbor, add the edge to the adjacency matrix. */

			if (vertexPtr->vNum < neighborPtr->neighborInd) {
				addEdgeToAdjMat(vertexPtr->vNum, neighborPtr->neighborInd, G);
			}
		}
	}
}


/******************************************************************************
*
*       Function        addEdgeToAdjMat
*       Parameters		v1 (the number of the first vertex), v2 (the number 
*						of the second vertex), G (a graph representation of the 
*						elimination-order-product).
*       Output          none
*       Description		Adds an edge between vertices 'v1' & 'v2' to the
*						adjacency matrix of the graph 'G'.
*       Comments		none.
*
******************************************************************************/
void addEdgeToAdjMat(int v1, int v2, struct graph *G)
{
	int				i, mod;
	adjMatrixT		*unit;

	/* 1.	Calculate the unit in the row of 'v1Num' where the requested bit 
			is (specifed by i), and the requested bit (specified by mod). */

	if (v1 < v2) {
		i = (v2 - v1) / 8;
		mod = (v2 - v1) % 8;
		unit = &(G->adjMat[v1][i]);
	}
	else {
		i = (v1 - v2) / 8;
		mod = (v1 - v2) % 8;
		unit = &(G->adjMat[v2][i]);
	}

	switch (mod) {
		case 0:	unit->a = 1;
				break;
		case 1: unit->b = 1;
				break;
		case 2: unit->c = 1;
				break;
		case 3: unit->d = 1;
				break;
		case 4: unit->e = 1;
				break;
		case 5: unit->f = 1;
				break;
		case 6: unit->g = 1;
				break;
		case 7: unit->h = 1;
				break;
	}
}


/******************************************************************************
*
*       Function        removeEdgeFromAdjMat
*       Parameters		v1 (the number of the first vertex), v2 (the number 
*						of the second vertex), G (a graph representation of the 
*						elimination-order-product).
*       Output          none
*       Description		Removes the edge between vertices 'v1' & 'v2' from 
*						the adjacency matrix of the graph 'G'.
*       Comments		none.
*
******************************************************************************/
void removeEdgeFromAdjMat(int v1, int v2, struct graph *G)
{
	int				i, mod;
	adjMatrixT		*unit;

	/* 1.	Calculate the unit in the row of 'v1Num' where the requested bit 
			is (specifed by i), and the requested bit (specified by mod). */

	if (v1 < v2) {
		i = (v2 - v1) / 8;
		mod = (v2 - v1) % 8;
		unit = &(G->adjMat[v1][i]);
	}
	else {
		i = (v1 - v2) / 8;
		mod = (v1 - v2) % 8;
		unit = &(G->adjMat[v2][i]);
	}

	switch (mod) {
		case 0:	unit->a = 0;
				break;
		case 1: unit->b = 0;
				break;
		case 2: unit->c = 0;
				break;
		case 3: unit->d = 0;
				break;
		case 4: unit->e = 0;
				break;
		case 5: unit->f = 0;
				break;
		case 6: unit->g = 0;
				break;
		case 7: unit->h = 0;
				break;
	}
}


/******************************************************************************
*
*       Function        addNeighbor
*       Parameters		vertexPtr (a pointer to the vertex to which a neighbor 
*						needs to be added), neighborInd (the index of the 
*						neighbor which is being added).
*       Output          none
*       Description		Adds a neighbor to the list of neighbors of
*						vertex 'vertexPtr'.
*       Comments		none.
*
******************************************************************************/
void addNeighbor(vertex *vertexPtr, int neighborInd)
{
	neighbor	*newNeighbor;

	/* 1.	Allocate memory for the new neighbor struct. */

	myMalloc(newNeighbor, sizeof(neighbor));

	/* 2.	Update the new neighbor struct with the index of the new neighbor. */

	newNeighbor->neighborInd = neighborInd;

	/* 3.	Insert the new neighbor struct at the beginning of the neighbor list. */

	newNeighbor->next = vertexPtr->neighborList;
	vertexPtr->neighborList = newNeighbor;
	vertexPtr->numNeighbors++;
}



/******************************************************************************
*
*       Function        removeNeighbor
*       Parameters		vertexPtr (a pointer to the vertex from which a neighbor 
*						needs to be removed), neighborInd (the index of the 
*						neighbor which is being removed).
*       Output          none
*       Description		Reomves a neighbor from the list of neighbors of
*						vertex 'vertexPtr'.
*       Comments		none.
*
******************************************************************************/
void removeNeighbor(vertex *vertexPtr, int neighborInd)
{
	neighbor	*currNeighbor, *prevNeighbor;

	vertexPtr->numNeighbors--;

	/* 1.	Initialize 'currNeighbor' to point to the first neighbor in the
			neighbor list of 'vertexPtr'. */

	currNeighbor = vertexPtr->neighborList;

	/* 2.	If the first neighbor on the list is the one that needs to be removed,
			remove it from the list and free the memory allocated for it. */

	if (currNeighbor->neighborInd == neighborInd) {
		vertexPtr->neighborList = currNeighbor->next;
		myFree(currNeighbor);
		return;
	}

	/* 3.	While the requested neighbor hasn't been found yet, continue to iterate 
			over the list of neighbors. */

	while (currNeighbor->neighborInd != neighborInd) {
		prevNeighbor = currNeighbor;
		currNeighbor = currNeighbor->next;
	}
	
	/* 4.	Remove the neighbor from the list and free the memory allocated for it. */

	prevNeighbor->next = currNeighbor->next;
	myFree(currNeighbor);
}



/******************************************************************************
*
*       Function        addEdge
*       Parameters		v1 (the number of the first vertex), v2 (the number 
*						of the second vertex), G (a graph representation of the 
*						elimination-order-product).
*       Output          none
*       Description		adds an edge between 'v1' and 'v2' to the graph 'G'.
*       Comments		none.
*
******************************************************************************/
void addEdge(int v1, int v2, struct graph *G)
{
	/* 1.	If 'v1' and 'v2' aren't neighbors already: */

	if (areNeighbors(G, v1, v2) == FALSE) {
		/* 1.1.	Add 'v2' to the neighbor list of 'v1'. */

		addNeighbor(&(G->vertexList[v1]), v2);

		/* 1.2.	Add 'v1' to the neighbor list of 'v2'. */

		addNeighbor(&(G->vertexList[v2]), v1);

		/* 1.3.	Add the edge between 'v1' and 'v2' to the adjacency matrix of G. */

		addEdgeToAdjMat(v1, v2, G);
	}
}



/******************************************************************************
*
*       Function        updateGraph
*       Parameters		G (the graph that needs to be updated), elimProd (the 
*						elimination-order-product).
*       Output          none
*       Description		Update the pointers of the vertices in the graph to the
*						variables of the product.
*       Comments		none.
*
******************************************************************************/
void updateGraph(struct graph *G, elimOrderProduct *elimProd)
{
	elimOrderVar	*varPtr;
	vertex			*vertexPtr, *vertexPtrEnd;

	/* 1.	Iterate over all the vertices of the graph. */

	for (vertexPtr = G->vertexList, vertexPtrEnd = vertexPtr + G->numVertices,
			varPtr = elimProd->variables; vertexPtr < vertexPtrEnd; varPtr++, 
			vertexPtr++) {

		/* 1.1.	Update the pointer to the variable in the elimination-order-product 
				which corresponds to this vertex. */

		vertexPtr->varInProd = varPtr;
	}
}


/******************************************************************************
*
*       Function        destroyGraph
*       Parameters		G (the graph)
*       Output          none
*       Description		Frees the memory that was allocated for the graph.
*       Comments		none.
*
******************************************************************************/
void destroyGraph(struct graph *G, boolean isAdjMat)
{
	vertex			*vertexPtr, *vertexPtrEnd; 
	adjMatrixT		**adjMatrixTPtr, **adjMatrixTPtrEnd;
	neighbor		*currNeighbor, *nextNeighbor;

	/* 1.	Free the memory allocated for the graph representation. */

	for (vertexPtr = G->vertexList, vertexPtrEnd = vertexPtr + G->numVertices;
			vertexPtr < vertexPtrEnd; vertexPtr++) {

		if (vertexPtr->numNeighbors > 0) {
			currNeighbor = vertexPtr->neighborList;

			while (currNeighbor != NULL) {
				nextNeighbor = currNeighbor->next;
				myFree(currNeighbor);
				currNeighbor = nextNeighbor;
			}
		}
	}

	/* 2.	Free the memory allocated for the 'vertexList' and the 'varToVertexMap'. */

	myFree(G->vertexList);
	myFree(G->varToVertexMap);

	/* 3.	If an adjacency matrix was allocated, free the memory allocated for it. */

	if (isAdjMat) {
		for (adjMatrixTPtr = G->adjMat, 
				adjMatrixTPtrEnd = adjMatrixTPtr + G->numVertices;
				adjMatrixTPtr < adjMatrixTPtrEnd; adjMatrixTPtr++) {

			myFree(*adjMatrixTPtr);
		}
		myFree(G->adjMat);
	}

	/* 4.	Free the memory allocated for the graph structure itself. */

	myFree(G);
}



/******************************************************************************
*
*       Function        areNeighbors
*       Parameters		v1 (the number of the first vertex, v2 (the number 
*						of the second vertex, G (a graph representation of the 
*						elimination-order-product).
*       Output          none
*       Description		Checks if there's an edge between vertices 'v1' & 'v2', 
*						and returns TRUE if there is, and FALSE otherwise.
*       Comments		none.
*
******************************************************************************/
boolean areNeighbors(struct graph *G, int v1, int v2)
{
	int				i, mod;
	adjMatrixT		*unit;

	/* 1.	Calculate the unit in the row of 'v1Num' where the requested bit 
			is (specifed by i), and the requested bit (specified by mod). */

	if (v1 < v2) {
		i = (v2 - v1) / 8;
		mod = (v2 - v1) % 8;
		unit = &(G->adjMat[v1][i]);
	}
	else {
		i = (v1 - v2) / 8;
		mod = (v1 - v2) % 8;
		unit = &(G->adjMat[v2][i]);
	}

	switch (mod) {
		case 0:	return (unit->a);
		case 1: return (unit->b);
		case 2: return (unit->c);
		case 3: return (unit->d);
		case 4: return (unit->e);
		case 5: return (unit->f);
		case 6: return (unit->g);
		case 7: return (unit->h);
		default: return FALSE;
	}
}



/******************************************************************************
*
*       Function        eliminateVertices
*       Parameters		vNum (the number of the vertex that needs to be removed
*						from the graph, G (the graph representation of the 
*						elimination-order-product).
*       Output          none
*       Description		Eliminates vertex number 'vNum' from graph 'G'.
*       Comments		none.
*
******************************************************************************/
void eliminateVertices(vertex *elimVertex, int numElimVertices, double *tableSizes, 
					   long *elimInd, elimOrderNode **elimNodePtr, struct graph *G, 
					   elimOrderProduct *elimProd)
{
	elimOrderNode	*elimNode;
	vertex			**elimVertices, **elimVertexPtr, **elimVertexPtrEnd;
	long			currElimInd;


	elimNode = *elimNodePtr;
	
	if (numElimVertices == 1) {

		/* 1.	If there's only 1 vertex to eliminate in the current iteration,
				update the current elimination-order-node regarding it, the array
				of table-sizes, advance the pointers, and eliminate it from the graph. */

		updateElimNode(elimNode, FALSE, (elimVertex->varInProd)->originalNode);

		tableSizes[*elimInd] = (elimVertex->varInProd)->tableSize;
		(*elimNodePtr)++;
		(*elimInd)++;

		eliminateVertex(elimVertex, G);
	}
	else {
		/* 1.	If there is more than 1 vertex to eliminate in the current iteration: */

		/* 1.1.	Initialization. */

		elimVertices = elimVertex->uniteVertices;
		elimVertexPtr = elimVertices;
		elimVertexPtrEnd = elimVertexPtr + numElimVertices;
		currElimInd = *elimInd;

		/* 2.	Update the elimination order array with all the nodes that need to be 
				eliminated jointly. */

		for (; elimVertexPtr < elimVertexPtrEnd; elimVertexPtr++, elimNode++, 
				currElimInd++) {
			
			updateElimNode(elimNode, FALSE, ((*elimVertexPtr)->varInProd->originalNode));
			tableSizes[currElimInd] = 0;
		}

		tableSizes[*elimInd] = (elimVertex->varInProd)->tableSize;
		*elimInd = currElimInd;
		*elimNodePtr = elimNode;

		eliminateVerticesJoint(elimVertices, numElimVertices, G);

		myFree(elimVertices);
	}
}



/******************************************************************************
*
*       Function        eliminateVertex
*       Parameters		vNum (the number of the vertex that needs to be removed
*						from the graph, G (the graph representation of the 
*						elimination-order-product).
*       Output          none
*       Description		Eliminates vertex number 'vNum' from graph 'G'.
*       Comments		none.
*
******************************************************************************/
void eliminateVertex(vertex *elimVertex, struct graph *G) 
{
	int			vNum;
	neighbor	*currNeighbor, *nextNeighbor;


	vNum = elimVertex->vNum;

	if (elimVertex->numNeighbors > 0) {

		/* 1.1.	Create a clique from the neighbors of vertex number 'vNum'. */

		if (elimVertex->numNeighbors > 1) {

			createClique(elimVertex->neighborList, G);
		}

		/* 1.2.	Iterate over the neighbors of 'vertexPtr'. */

		currNeighbor = elimVertex->neighborList;

		while (currNeighbor != NULL) {

			nextNeighbor = currNeighbor->next;

			/* 1.2.1.	Remove the edge between 'vertexPtr' and 'currNeighbor' 
						from the adjacency matrix. */

			removeEdgeFromAdjMat(vNum, currNeighbor->neighborInd, G);

			/* 1.2.2.	Remove vertex no. 'vNum' from the neighbor list of
						the neighbor 'currNeighbor'. */

			removeNeighbor(&(G->vertexList[currNeighbor->neighborInd]), vNum);

			/* 1.2.3.	Free the struct allocated for the current neighbor,
						and advance the pointers to the following neighbor structs. */

			myFree(currNeighbor);
			currNeighbor = nextNeighbor;
		}
	}
	G->currNumVertices--;
	elimVertex->isMarked = TRUE;
	elimVertex->numNeighbors = 0;
}


/******************************************************************************
*
*       Function        createClique
*       Parameters		neighborList (pointer to the beginning of a list of 
*						vertices in the graph which need to be turned into a clique).
*       Output          none
*       Description		Turns the vertices in 'neighborList' into a clique.
*       Comments		none.
*
******************************************************************************/
void createClique(neighbor *neighborList, struct graph *G)
{

	vertex		*neighborPtr;
	neighbor	*neighbor1, *neighbor2;

	
	/* 1.	Initialize pointers to the first neighbor in the list. */

	neighbor1 = neighborList;

	/* 2.	As long as this isn't the last neighbor. */

	while (neighbor1->next != NULL) {

		/* 2.1.	Invalidate the cost of this neighbor. */

		neighborPtr = &(G->vertexList[neighbor1->neighborInd]);
		invalidateVertex(neighborPtr, G);

		/* 2.2.	Initialize a pointer to the neighbor following 'neighbor1' in
					the neighbor list. */

		neighbor2 = neighbor1->next;

		/* 2.2.	If 'neighbor1' isn't connected to all the vertices in the graph, */

		if (neighborPtr->numNeighbors < G->numVertices - 1) {

			while (neighbor2 != NULL) {

				/* 2.2.1.	Add an edge between 'neighbor1' and 'neighbor2'. */

				addEdge(neighbor1->neighborInd, neighbor2->neighborInd, G);

				/* 2.2.2.	Advance 'neighbor2' to point to the next neighbor in the list. */

				neighbor2 = neighbor2->next;
			}
		}

		/* 2.3.	Advance 'neighbor1' to point to the next neighbor in the list. */

		neighbor1 = neighbor1->next;
	}

	/* 3.	Handle the last neighbor. */

	neighborPtr = &(G->vertexList[neighbor1->neighborInd]);
	invalidateVertex(neighborPtr, G);
}



/******************************************************************************
*
*       Function        connectVertices
*       Parameters		vertexArr (an array of vertices in the graph which need 
*						to be turned into a clique), numVertices (the number of
*						vertices in the array), G (the graph).
*       Output          none
*       Description		Turns the vertices in 'vertexArr' into a clique.
*       Comments		none.
*
******************************************************************************/
void connectVertices(vertex **vertexArr, int numVertices, struct graph *G)
{
	vertex		**vertexPtr1, *vertex1, **vertexPtr2, *vertex2, **vertexPtrEnd;

	
	/* 1.	Iterate over the vertices in the array. */

	for (vertexPtr1 = vertexArr, vertexPtrEnd = vertexArr + numVertices;
			vertexPtr1 < vertexPtrEnd; vertexPtr1++) {

		vertex1 = (*vertexPtr1);

		/* 1.1.	Invalidate the cost of this vertex. */

		invalidateVertex(vertex1, G);

		/* 1.2.	If 'vertex1' isn't connected to all the vertices in the graph, */

		if (vertex1->numNeighbors < G->numVertices - 1) {

			/* 1.2.1.	Iterate over all the vertices following it in the array. */

			for (vertexPtr2 = vertexPtr1 + 1; vertexPtr2 < vertexPtrEnd; vertexPtr2++) {

				vertex2 = *vertexPtr2;

				/* 1.2.1.1.	Add an edge between 'vertex1' and 'vertex2'. */

				addEdge(vertex1->vNum, vertex2->vNum, G);
			}
		}
	}
}



/******************************************************************************
*
*       Function        eliminateVerticesJoint
*       Parameters		elimVertexArr (the array of vertices which need to be
*						eliminated), numElimVertices (the number of vertices in
*						the array), G (the elimOrderGraph from which the vertices 
*						need to be eliminated), cliqueSize (the size of the 
*						clique created by the joint elimination).
*       Output          none
*       Description		Eliminates the vertices in the array 'elimVertexArr' 
*						from graph 'G'.
*       Comments		none.
*
******************************************************************************/
void eliminateVerticesJoint(vertex **elimVertexArr, int numElimVertices, 
							struct graph *G)
{
	vertex		**jointNeighbors, **elimVertexPtr, **elimVertexPtrEnd, *elimVertex;
	int			numNeighbors, elimVNum;
	neighbor	*currNeighbor, *nextNeighbor;


	/* 1.	Get the list of all the neighbors of the vertices in 'elimVertexArr'. */

	jointNeighbors = getNeighbors(G, elimVertexArr, numElimVertices, &numNeighbors);

	if (numNeighbors > 1) {

		/* 2.	Turn the neighbors in 'jointNeighbors' into a clique. */

		connectVertices(jointNeighbors, numNeighbors, G);
	}

	/* 3.	Free the memory allocated dynamically. */

	myFree(jointNeighbors);

	/* 4.	For each one of the eliminated vertices: */

	for (elimVertexPtr = elimVertexArr, elimVertexPtrEnd = elimVertexPtr + numElimVertices;
			elimVertexPtr < elimVertexPtrEnd; elimVertexPtr++) {	

		elimVertex = *elimVertexPtr;
		elimVNum = elimVertex->vNum;

		/* 4.1.	Iterate over its neighbors. */

		currNeighbor = elimVertex->neighborList;

		while (currNeighbor != NULL) {

			nextNeighbor = currNeighbor->next;

			/* 4.1.1.	Remove the edge between 'elimVertex' and 'currNeighbor' 
						from the adjacency matrix. */

			removeEdgeFromAdjMat(elimVNum, currNeighbor->neighborInd, G);

			/* 4.1.2.	Remove vertex no. 'vNum' from the neighbor list of
						the neighbor 'currNeighbor'. */

			removeNeighbor(&(G->vertexList[currNeighbor->neighborInd]), elimVNum);

			/* 4.1.3.	Free the struct allocated for the current neighbor,
						and advance the pointers to the following neighbor structs. */

			myFree(currNeighbor);
			currNeighbor = nextNeighbor;
		}

		/* 4.2.	Update the graph according to the eliminated vertex. */

		G->currNumVertices--;
		elimVertex->isMarked = TRUE;
		elimVertex->numNeighbors = 0;
	}
}




/******************************************************************************
*
*       Function        fixVertices
*       Parameters		vNum (the number of the vertex that needs to be 
*						conditioned on (set) in graph, G (the graph representation  
*						of the elimination-order-product).
*       Output          none
*       Description		Fixes (conditions on) vertex number 'vNum' from graph 'G'.
*       Comments		none.
*
******************************************************************************/
void fixVertices(vertex	*vertexPtr, int numVertices, long *elimInd, 
				 elimOrderNode **elimNodePtr, struct graph *G, elimOrderProduct *elimProd)
{
	vertex			**condVertices, **condVertexPtr, **condVertexPtrEnd;
	elimOrderNode	*elimNode;


	if (numVertices == 1) {
		condVertices = &vertexPtr;
	}
	else {
		condVertices = vertexPtr->uniteVertices; 
	}

	elimNode = *elimNodePtr;

	/* 1.	Iterate over the variables that need to be conditioned on. */

	for (condVertexPtr = condVertices, condVertexPtrEnd = condVertexPtr + numVertices;
			condVertexPtr < condVertexPtrEnd; condVertexPtr++, elimNode++) {

		/* 1.1.	Remove the variable from the product. */
	
		fixVertex(*condVertexPtr, G);

		/* 1.2.	Update the current elimination-order-node. */

		updateElimNode(elimNode, TRUE, (*condVertexPtr)->varInProd->originalNode);
	}

	/* 2.	Update the elimination index and the elimination-order-node. */

	*elimInd += numVertices;
	*elimNodePtr = elimNode;

	/* 3.	Free the memory allocated for the array of pointers. */

	if (numVertices > 1) {
		myFree(condVertices);
	}
}



/******************************************************************************
*
*       Function        fixVertex
*       Parameters		vNum (the number of the vertex that needs to be 
*						conditioned on (set) in graph, G (the graph representation  
*						of the elimination-order-product).
*       Output          none
*       Description		Fixes (conditions on) vertex number 'vNum' from graph 'G'.
*       Comments		none.
*
******************************************************************************/
void fixVertex(vertex *vertexPtr, struct graph *G) 
{
	int			vNum;
	neighbor	*currNeighbor, *nextNeighbor;
	vertex		*neighborVertex;


	vNum = vertexPtr->vNum;
	vertexPtr = &(G->vertexList[vNum]);

	if (vertexPtr->numNeighbors > 0) {

		/* 1.1.	Iterate over the neighbors of 'vertexPtr'. */

		currNeighbor = vertexPtr->neighborList;
		vertexPtr->neighborList = NULL;

		while (currNeighbor != NULL) {
			nextNeighbor = currNeighbor->next;

			neighborVertex = &(G->vertexList[currNeighbor->neighborInd]);

			/* 1.1.1.	Remove vertex no. 'vNum' from the neighbor list of
						the neighbor 'currNeighbor'. */

			removeNeighbor(neighborVertex, vNum);

			/* 1.1.2.	Invalidate the cost of the neighbor. */

			invalidateVertex(neighborVertex, G);

			/* 1.1.3.	Remove the edge between 'vertexPtr' and 'currNeighbor' 
						from the adjacency matrix. */

			removeEdgeFromAdjMat(vNum, currNeighbor->neighborInd, G);

			/* 1.1.4.	Free the struct allocated for the current neighbor,
						and advance the pointers to the following neighbor structs. */

			myFree(currNeighbor);
			currNeighbor = nextNeighbor;
		}
	}

	G->currNumVertices--;
	vertexPtr->isMarked = TRUE;
	vertexPtr->numNeighbors = 0;
}


/******************************************************************************
*
*       Function        calcElimComplexityJointG
*       Parameters		G (the elimOrderGraph from which the vertices need to be
*						eliminated), elimVertexArr (the array of vertices which
*						need to be eliminated), numElimVertices (the number of
*						vertices in the array).
*       Output          none
*       Description		calculates the joint elimination complexity of the
*						vertices in the array.
*       Comments		none
*
******************************************************************************/
double calcElimComplexityJointG(struct graph *G, vertex **elimVertexArr, 
								int numElimVertices)
{
	vertex			**elimVertexPtr, **elimVertexPtrEnd, *elimVertex, *neighorPtr;
	double			elimCost;		
	neighbor		*neighbor;


	elimVertexPtrEnd = elimVertexArr + numElimVertices;

	/* 1.	Iterate over the vertices in the array. */

	for (elimVertexPtr = elimVertexArr; elimVertexPtr < elimVertexPtrEnd; 
			elimVertexPtr++) {

		/* 1.1.	Zero all the 'seen' flags of the neighbors of the vertex. */

		zeroSeenVertices(*elimVertexPtr, G);
	}

	/* 2.	Turn on the flags of the vertices that are eliminated. */

	for (elimVertexPtr = elimVertexArr; elimVertexPtr < elimVertexPtrEnd; 
			elimVertexPtr++) {
		
		(*elimVertexPtr)->alreadySeen = TRUE;
	}

	/* 3.	Initialize the elimination cost to 1. */

	elimCost = 1;

	/* 4.	Iterate over the vertices in the array. */

	for (elimVertexPtr = elimVertexArr; elimVertexPtr < elimVertexPtrEnd; 
			elimVertexPtr++) {

		elimVertex = *elimVertexPtr;

		/* 4.1.	Iterate over the neighbors of the vertex. */

		for (neighbor = elimVertex->neighborList; neighbor != NULL; 
				neighbor = neighbor->next) {

			neighorPtr = &(G->vertexList[neighbor->neighborInd]);

			/* 4.1.1.	If this vertex hasn't been seen yet, update the elimination
						cost according to it. */

			if (neighorPtr->alreadySeen == FALSE) {
				elimCost *= (neighorPtr->varInProd)->numValues;
			}
		}
	}

	/* 5.	Returns the elimination complexity of var. */

	return elimCost;
}


/******************************************************************************
*
*       Function        zeroSeenVertices
*       Parameters		elimVertex (the vertex whose neighbors' flags need to be
*						turned off), G (the graph).
*       Output          none
*       Description		turns off the flags of all the neighbors of 'elimVertex'.
*       Comments		none
*
******************************************************************************/
void zeroSeenVertices(vertex *elimVertex, struct graph *G)
{
	vertex			*neighorPtr;
	neighbor		*neighbor;


	/* 1.	Iterate over the neighbors of the vertex. */

	for (neighbor = elimVertex->neighborList; neighbor != NULL; 
			neighbor = neighbor->next) {

		neighorPtr = &(G->vertexList[neighbor->neighborInd]);

		/* 1.1.	Turn off the flag of this vertex. */

		neighorPtr->alreadySeen = FALSE;
	}
}



/******************************************************************************
*
*       Function        getNeighbors
*       Parameters		G (the graph), vertexArr (an array of pointers to vertices
*						in the graph), numVertices (the number of vertices in the
*						array).
*       Output          none
*       Description		Returns an array with all the neighbors of the vertices
*						in the array 'vertexArr'.
*       Comments		none
*
******************************************************************************/
vertex** getNeighbors(struct graph *G, vertex **vertexArr, int numVertices,
					  int *numNeighbors)
{
	vertex			**vertexPtr, **vertexPtrEnd, *currVertex, *neighorPtr,
					**neighborArr, **neighborArrPtr;
	neighbor		*neighbor;


	vertexPtrEnd = vertexArr + numVertices;
	*numNeighbors = 0;

	/* 1.	Iterate over the vertices in the array. */

	for (vertexPtr = vertexArr; vertexPtr < vertexPtrEnd; vertexPtr++) {

		/* 1.1.	Zero all the 'seen' flags of the neighbors of each vertex. */

		zeroSeenVertices(*vertexPtr, G);
		*numNeighbors += (*vertexPtr)->numNeighbors;
	}

	/* 2.	Turn on the flags of the vertices that are in 'vertexArr'. */

	for (vertexPtr = vertexArr; vertexPtr < vertexPtrEnd; vertexPtr++) {
		
		(*vertexPtr)->alreadySeen = TRUE;
	}

	/* 3.	Allocate memory for the array that would store pointers to all the
			neighbors. */

	myMalloc(neighborArr, sizeof(vertex*) * (*numNeighbors));

	/* 4.	Iterate over the vertices in the array. */

	for (vertexPtr = vertexArr, neighborArrPtr = neighborArr; 
			vertexPtr < vertexPtrEnd; vertexPtr++) {

		currVertex = *vertexPtr;

		/* 4.1.	Iterate over the neighbors of each vertex. */

		for (neighbor = currVertex->neighborList; neighbor != NULL; 
				neighbor = neighbor->next) {

			neighorPtr = &(G->vertexList[neighbor->neighborInd]);

			/* 4.1.1.	If this vertex hasn't been seen yet, update the neighbor
						array with a pointer to it. */

			if (neighorPtr->alreadySeen == FALSE) {
				neighorPtr->alreadySeen = TRUE;
				*neighborArrPtr = neighorPtr;
				neighborArrPtr++;
			}
		}
	}

	/* 5.	Calculate the total number of neighbors, and return the neighbor 
			array. */

	*numNeighbors = neighborArrPtr - neighborArr;

	return neighborArr;
}




/******************************************************************************
*
*       Function        invalidateVertex
*       Parameters		
*       Output          none
*       Description		
*       Comments		none
*
******************************************************************************/
void invalidateVertex(vertex *currVertex, struct graph *G) 
{
	node		*origNode;
	vertex		*neighborVertex;
	neighbor	*neighborItt;

#if UNITE_VARS
	vertex		**unitePtr, **unitePtrEnd;
#endif

	/* 1.	If the cost of this vertex is valid: */

	if (currVertex->cost != INVALID) {

		/* 1.1.	Invalidate it. */

		currVertex->cost =  INVALID; 

		/* 1.2.	If this vertex needs to treated along with a few other vertices,
				invalidate their cost as well. */

		origNode = (currVertex->varInProd)->originalNode; 

		#if UNITE_VARS
			if (origNode->uniteT) { 
				unitePtr = currVertex->uniteVertices; 
				unitePtrEnd = unitePtr + origNode->numVarsUnite; 

				for (; unitePtr < unitePtrEnd; unitePtr++) { 
					invalidateVertex(*unitePtr, G);
				} 
			}
		#endif

		/* 1.3.	Iterate over all the neighbors of the vertex and 
				invalidate their cost too. */

		neighborItt = currVertex->neighborList;

		while (neighborItt) {
			neighborVertex = &(G->vertexList[neighborItt->neighborInd]);
			neighborVertex->cost = INVALID;
			neighborItt = neighborItt->next;
		}
	}
}






