/* $Id: EliminationOrderGraph.h,v 1.3 2008/08/03 08:31:23 marks Exp $ */
/*****************************************************************************/
/*  EliminationOrderGraph.h													test */
/*  Author: Ma'ayan Fishelson												 */
/*****************************************************************************/

#ifndef _ELIMINATION_ORDER_GRAPH_H_
#define _ELIMINATION_ORDER_GRAPH_H_

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "EliminationOrderProduct.h"


/*------------------------------*/
/*		DATA-STRUCTURES			*/
/*------------------------------*/

/* This is a 2-dimensional bit array, which contains information about the graph 
   (adjacency matrix, indicates for every 2 vertices whether they are neighbors). */

typedef struct adjMatrixT {
	unsigned char	a:1, b:1, c:1, d:1, e:1, f:1, g:1, h:1;
} adjMatrixT;

typedef adjMatrixT **adjMatrix;

typedef struct neighbor {
	int				neighborInd;
	struct neighbor	*next;
} neighbor;

/* vertex - Simulates a vertex (node) of the Bayesian network, for the 
   elimination-order-finding process. */

typedef struct vertex {
	int				vNum;				/* The index of the vertex in the vertex array. */
	int				numNeighbors;		/* The number of neighbors of the vertex. */
	boolean			isMarked;			/* A flag that indicates whether the vertex has already
										   been chosen for elimination. */
	neighbor		*neighborList;		/* The neighbor list of the vertex. */
	double			cost;				/* Holds the cost of eliminating this vertex, if it's
										   valid. */
	int				numNeighMarked;		/* Holds the number of neighbors of this vertex which 
										   have already been chosen for elimination. */
	elimOrderVar	*varInProd;			/* Pointer to the variable in the elimination-order-product
										   which corresponds to this vertex. */
	boolean			alreadySeen;		/* An auxiliary flag used by different functions. */
	struct vertex	**uniteVertices;	/* An array of pointers to the vertices that need to 
										   be eliminated jointly with this vertex. */
	int				numNeighborsUnite;	/* If this vertex needs to be treated jointly with a
										   few additional vertices, then this field holds the
										   total number of neighbors they have. Otherwise,
										   it stores only the number of neighbors of this vertex. */ 
} vertex;

/* graph - Simulates the Bayesian network, for the elimination-order-finding process. */

typedef struct graph {	
	long			numVertices;		/* The initial number of vertices in the graph. */
	long			currNumVertices;	/* The current number of vertices in the graph. */
	vertex			*vertexList;		/* The vertices in the graph. */
	adjMatrix		adjMat;				/* The adjacency matrix of the graph. */
	vertex			**varToVertexMap;	/* A mapping from the variables of the product
										   to the vertices of the graph. */
} graph;


/* This structure holds the data on one of the vertices in the vertex pool. */
typedef struct vertexInPool {
	vertex				*v;
	vertex				**vPtr;
	double				cost;		/* the cost of this vertex. */
	double				relProb;	/* the relative probability of this
									   vertex to be chosen. */
	struct vertexInPool	*prev;
	struct vertexInPool	*next;
} vertexInPool;



/*------------------------------*/
/*		FUNCTIONS				*/
/*------------------------------*/


struct graph* createGraph(elimOrderProduct *elimProd, boolean createAdjMat, 
						  long varMappingSize);
void updateGraph(graph *G, elimOrderProduct *elimProd);
void destroyGraph(graph *G, boolean isAdjMat);

void createAdjMatrix(graph *G);
void zeroAdjMat(graph *G);
void addEdgeToAdjMat(int v1, int v2, graph *G);
void removeEdgeFromAdjMat(int v1, int v2, graph *G);
void addNeighbor(vertex *vertexPtr, int neighborInd);
void removeNeighbor(vertex *vertexPtr, int neighborInd);
void addEdge(int v1, int v2, graph *G);
boolean areNeighbors(graph *G, int v1, int v2);

void eliminateVertices(vertex *elimVertex, int numElimVertices, double *tableSizes, 
					   long *elimInd, elimOrderNode **elimNodePtr, struct graph *G, 
					   elimOrderProduct *elimProd);
void eliminateVertex(vertex *elimVertex, struct graph *G);
void createClique(neighbor *neighborList, struct graph *G);
void connectVertices(vertex **vertexArr, int numVertices, struct graph *G);
void eliminateVerticesJoint(vertex **elimVertexArr, int numElimVertices, 
							struct graph *G);

void fixVertices(vertex	*vertexPtr, int numVertices, long *elimInd, 
				 elimOrderNode **elimNodePtr, struct graph *G, elimOrderProduct *elimProd);
void fixVertex(vertex *vertexPtr, struct graph *G);

double calcElimComplexityJointG(struct graph *G, vertex **elimVertexArr, 
								int numElimVertices);
void zeroSeenVertices(vertex *elimVertex, struct graph *G);

vertex** getNeighbors(struct graph *G, vertex **vertexArr, int numVertices,
					  int *numNeighbors);

void invalidateVertex(vertex *currVertex, struct graph *G);

#endif		/*	_ELIMINATION_ORDER_GRAPH_H_	*/
