/* $Id: EliminationOrderProduct.c,v 1.2 2008/08/03 08:15:46 superlink Exp $ */
/*********************************************************************************/
/*  EliminationOrderProduct.c												     */
/*  Author: Ma'ayan Fishelson					                                 */
/*********************************************************************************/

#include "EliminationOrderProduct.h"


/******************************************************************************
*
*       Function        allocateElimOrderProduct
*       Parameters		numVariables (number of slots that need to be allocated
*						for variables), numFunctions (number of slots that need 
*						to be allocated for functions),
*						numVariablesMapping (number of spaces that need to be 
*						allocated for a variable mapping).
*       Output          none
*       Description		allocates memory for the elimination order product
*						for one of the pedigrees.
*       Comments		returns a pointer to the product struct that was allocated.
*
******************************************************************************/
elimOrderProduct* allocateElimOrderProduct(long numVariables, long numFunctions,
										   long numVariablesMapping)
{	
	elimOrderProduct		*tempProd;
	elimOrderVar			*varPtr, *varPtrEnd;
	elimOrderFunc			*funcPtr, *funcPtrEnd;


	/* 1.	Allocate the product data-structure. */

	myMalloc(tempProd, sizeof(elimOrderProduct));

	/* 2.	Initialize the variable related fields of the product structure. */

	tempProd->numVariables = numVariables;
	tempProd->currNumVars = 0;
	myMalloc(tempProd->variables, sizeof(elimOrderVar) * numVariables);
	tempProd->nextVarPlace = &tempProd->variables[0];
	tempProd->nextElimVar = &tempProd->variables[0];
	
	for (varPtr = tempProd->variables, varPtrEnd = varPtr + numVariables;
			varPtr < varPtrEnd; varPtr++) {
		varPtr->used = FALSE;
	}

	myMalloc(tempProd->oldToNewVarMapping, sizeof(elimOrderVar*) * numVariablesMapping);

	/* 3.	Initialize the function related fields of the product structure. */

	tempProd->numFunctions = numFunctions;
	tempProd->currNumFuncs = 0;
	myMalloc(tempProd->functions, sizeof(elimOrderFunc) * numFunctions);
	tempProd->nextFuncPlace = &tempProd->functions[0];
	
	for (funcPtr = tempProd->functions, funcPtrEnd = funcPtr + numFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {
		funcPtr->used = FALSE;
		funcPtr->variables = NULL;
	}

#if CALC_MEM_REQ
	tempProd->currMemReq = 0;
	tempProd->currMemReqCurrProd = 0;
	tempProd->condDepth = 1;
#endif

	/* 4.	Return a pointer to the elimOrderProduct structure that was allocated
			and initialized. */

	return tempProd;
}



/******************************************************************************
*
*       Function        freeElimOrderProduct
*       Parameters      elimProd (the elimination-order product that needs to
*						be freed).
*       Output          none
*       Description		frees the memory that was allocated for the elimination
*						order product, 'Product'.
*       Comments		none
*
******************************************************************************/
void freeElimOrderProduct(elimOrderProduct *elimProd)
{
	elimOrderVar	*varPtr, *varPtrEnd;
	elimOrderFunc	*funcPtr, *funcPtrEnd;

#if	UNITE_VARS
	elimOrderVar	**uniteVars, **uniteVar, **uniteVarEnd;
#endif

	/* 1.	Free the memory allocated for the variables and remove them from 
			the elimOrderProduct. */

	for (varPtr = elimProd->variables, varPtrEnd = varPtr + elimProd->numVariables;
			varPtr < varPtrEnd; varPtr++) {
		if (varPtr->used) {

			/* 1.1.	If this variable was treated jointly with a few other variables,
					update all these variables to point to NULL, and free the memory
					allocated for the array of pointers to these variables. */

			#if	UNITE_VARS
				if ((varPtr->originalNode->uniteT) && (varPtr->uniteVars != NULL)) {
					uniteVars = varPtr->uniteVars;

					for (uniteVar = uniteVars, 
							uniteVarEnd = uniteVar + varPtr->originalNode->numVarsUnite;
							uniteVar < uniteVarEnd; uniteVar++) {
						(*uniteVar)->uniteVars = NULL;
					}
					myFree(uniteVars);
				}
			#endif

			removeVariableFromElimOrderProduct(elimProd, varPtr);
		}
	}

	/* 2.	Free the memory allocated for the functions and remove them from 
			the elimOrderProduct. */

	for (funcPtr = elimProd->functions, funcPtrEnd = funcPtr + elimProd->numFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {
		if (funcPtr->used)
			removeFunctionFromElimOrderProduct(elimProd, funcPtr);
	}

	/* 3.	Free the elimOrderProduct's arrays and structure. */

	myFree(elimProd->variables);
	myFree(elimProd->functions);
	myFree(elimProd->oldToNewVarMapping);
	myFree(elimProd);
}



/******************************************************************************
*
*       Function        insertVariableToElimOrderProduct
*       Parameters		elimProd (the elimination-order-Product to which the 
*						variable needs to be inserted), numFunctions (the 
*						number of functions that depend on this variable),
*						originalNode (the node of the Bayesian network that this
*						variable represents), numValues (the number of possible
*						values of this variable).
*       Output          none
*       Description		inserts a new elimination-order-variable into the 
*						elimination-order-product for the node 'originalNode' 
*						in the Bayesian network.
*       Comments		none
*
******************************************************************************/
elimOrderVar* insertVariableToElimOrderProduct(elimOrderProduct *elimProd, 
											   int numFunctions, 
											   node *originalNode, int numValues)
{
	elimOrderVar	*varPtr, *varPtrEnd;

	/* 1.	Search for an empty place to insert the new variable, starting at 
			nextVarPlace and wrapping around. */

	for (varPtr = elimProd->nextVarPlace, 
			varPtrEnd = elimProd->variables + elimProd->numVariables;
			varPtr < varPtrEnd; varPtr++) {
		if (!varPtr->used)
			goto enterVarFields;
	}

	for (varPtr = elimProd->variables, varPtrEnd = elimProd->nextVarPlace; 
			varPtr < varPtrEnd; varPtr++) {
		if (!varPtr->used)
			goto enterVarFields;
	}

	/* 2.	Initialize the fields of this place. */

	enterVarFields:

	varPtr->used = TRUE;
	varPtr->functionSpaces = numFunctions;

	myMalloc(varPtr->funcs, sizeof(elimOrderFunc*) * numFunctions);

	varPtr->currNumFunctions = 0;
	varPtr->originalNode = originalNode;
	varPtr->elimComplexity = INVALID;
	varPtr->numValues = numValues;
	varPtr->uniteVars = NULL;

#if PERFORM_PREPROCESSING
	varPtr->isUpdatedASimp = FALSE;
#endif

	/* 3.	Set nextVarPlace. */

	elimProd->nextVarPlace = varPtr + 1;
	elimProd->currNumVars++;

	return varPtr;
}



/******************************************************************************
*
*       Function        removeVariableFromElimOrderProduct
*       Parameters		elimProd (the elimination-order-product from which the
*						variable needs to be removed), elimVar (the variable 
*						that needs to be removed).
*       Output          none
*       Description		removes the elimination-order-variable, 'elimVar',
*						from the elimination-order-product, 'elimProd'.						 
*       Comments		none
*
******************************************************************************/
void removeVariableFromElimOrderProduct(elimOrderProduct *elimProd, 
										elimOrderVar *elimVar)
{
	/* 1.	Free the fields of this variable. */

	myFree(elimVar->funcs);

	/* 2.	Mark this field as free and update the elimOrderProduct's variable 
			counter. */

	elimVar->used = FALSE;
	elimProd->currNumVars--;
}



/******************************************************************************
*
*       Function        addFunctionToElimOrderVariable
*       Parameters		elimVar (a function needs to be added to this variable's
*						list of functions), elimFunc (the function that needs 
*						to be added).
*       Output          none
*       Description		adds 'elimFunc' to elimVar's list of functions.						 
*       Comments		none
*
******************************************************************************/
void addFunctionToElimOrderVariable(elimOrderVar *elimVar, elimOrderFunc *elimFunc)
{
#if UNITE_VARS
	elimOrderVar	**uniteVar, **uniteVarEnd;
#endif

	/* 1.	Add the function to the array of functions that depend on the variable. */

	elimVar->funcs[elimVar->currNumFunctions++] = elimFunc;

	/* 2.	Invalidate the elimination-complexity cache of this variable. */

	elimVar->elimComplexity = INVALID;

	/* 3.	If this variable needs to be eliminated jointly with a few other variables,
			invalidate the elimination-complexity cache of these variables as well. */

	#if	UNITE_VARS
		if (elimVar->originalNode->uniteT) {
			for (uniteVar = elimVar->uniteVars, 
					uniteVarEnd = uniteVar + elimVar->originalNode->numVarsUnite;
					uniteVar < uniteVarEnd; uniteVar++) {

				(*uniteVar)->elimComplexity = INVALID;
			}
		}
	#endif
}


/******************************************************************************
*
*       Function        removeFunctionFromElimOrderVariable
*       Parameters		elimVar (a function needs to be removed from this 
*						variable's list of functions), elimFunc (the function
*						that needs to be removed).
*       Output          none
*       Description		removes 'elimFunc' from elimVar's list of functions.						 
*       Comments		none
*
******************************************************************************/
void removeFunctionFromElimOrderVariable(elimOrderVar *elimVar, 
										 elimOrderFunc *elimFunc)
{
	elimOrderFunc	**functionPlace, **functionPlaceEnd, **tempFunctionPlace;

#if UNITE_VARS
	elimOrderVar	**uniteVar, **uniteVarEnd;
#endif

	/* 1.	Search for the specified function in Variable's list of functions. */

	for (functionPlace = elimVar->funcs, 
			 functionPlaceEnd = functionPlace + elimVar->currNumFunctions;
			 functionPlace < functionPlaceEnd; functionPlace++) {
		
		/* 1.1.	When found:  */

		if (*functionPlace == elimFunc) {
			
			/* 1.1.1.	Shift all the following functions one place back. */

			for (tempFunctionPlace = functionPlace; tempFunctionPlace < functionPlaceEnd - 1; 
					tempFunctionPlace++) {
				(*tempFunctionPlace) = *(tempFunctionPlace + 1);
			}

			/* 1.1.2.	Invalidate the elimination-complexity cache of this variable. */

			elimVar->elimComplexity = INVALID;

			/* 1.1.3.	If this variable needs to be eliminated jointly with a few other 
						variables, invalidate the elimination-complexity cache of these 
						variables as well. */

			#if	UNITE_VARS
				if (elimVar->originalNode->uniteT) {
					for (uniteVar = elimVar->uniteVars, 
							uniteVarEnd = uniteVar + elimVar->originalNode->numVarsUnite;
							uniteVar < uniteVarEnd; uniteVar++) {

						(*uniteVar)->elimComplexity = INVALID;
					}
				}
			#endif

			/* 1.1.4.	Update the variable's function counter. */

			elimVar->currNumFunctions--;
			return;
		}
	}
}



/******************************************************************************
*
*       Function        insertFunctionToElimOrderProduct
*       Parameters		elimProd (the product to which the function is inserted),
*						numVariables (the number of variables of the function),
*						vars (an array of pointers to the variables of the 
*						function), bucketInd (indicates the index of the bucket 
*						where this function was generated).
*       Output          none
*       Description		inserts a new function into the product.						 
*       Comments		none.
*
******************************************************************************/
elimOrderFunc* insertFunctionToElimOrderProduct(elimOrderProduct *elimProd, 
												int numVariables, 
												elimOrderVar **vars,
												unsigned long sizeOfTable)
{
	elimOrderFunc	*funcPtr, *funcPtrEnd;
	int				varCount;

#if	CALC_MEM_REQ
	double			memReq;
#endif


	/* 1.	Search for an empty place to insert the new function, starting at 
			nextFuncPlace and wrapping around. */

	for (funcPtr = elimProd->nextFuncPlace, 
			funcPtrEnd = elimProd->functions + elimProd->numFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {
		if (!funcPtr->used)
			goto enterFuncFields;
	}

	for (funcPtr = elimProd->functions, funcPtrEnd = elimProd->nextFuncPlace; 
			funcPtr < funcPtrEnd; funcPtr++) {
		if (!funcPtr->used)
			goto enterFuncFields;
	}

	/* 2.	Initialize the fields of this place. */

	enterFuncFields:

	myMalloc(funcPtr->variables, sizeof(elimOrderVar*) * numVariables);

	for (varCount = 0; varCount < numVariables; varCount++) {
		funcPtr->variables[varCount] = vars[varCount];
	}

	funcPtr->used = TRUE;
	funcPtr->numVariables = numVariables;
	funcPtr->sizeOfTable = sizeOfTable;

	/* 3.	Update the elimOrderProduct's fields. */

	elimProd->nextFuncPlace = funcPtr + 1;
	elimProd->currNumFuncs++;

#if CALC_MEM_REQ
	memReq = sizeOfTable * sizeof(Probability);
	elimProd->currMemReq += memReq;
	elimProd->currMemReqCurrProd += memReq;
#endif

	/* 4.	Return a pointer to the new function. */

	return funcPtr;
}


/******************************************************************************
*
*       Function        removeFunctionFromElimOrderProduct
*       Parameters		elimProd (the product from which a function needs to be
*						removed), elimFunc (the functions that needs to be 
*						removed).
*       Output          none
*       Description		removes 'elimFunc' from 'elimProd'.						 
*       Comments		none
*
******************************************************************************/
void removeFunctionFromElimOrderProduct(elimOrderProduct *elimProd, 
										elimOrderFunc *elimFunc)
{

#if CALC_MEM_REQ
	double		memReq;
#endif

	/* 1.	Free the memory and the slot. */

	myFree(elimFunc->variables);

	/* 2.	Mark this field as free and update the elimOrderProduct's function 
			counter. */

	elimFunc->used = FALSE;
	elimProd->currNumFuncs--;

#if CALC_MEM_REQ
	memReq = elimFunc->sizeOfTable * sizeof(Probability);
	elimProd->currMemReq -= memReq;
	elimProd->currMemReqCurrProd -= memReq;
#endif
}


/******************************************************************************
*
*       Function        createElimOrderProduct
*       Parameters		prod (the "real" product from which the elimOrderProduct
*						is created - the elimOrderProduct simulates this product).
*       Output          none
*       Description		creates the elimination-order-product (elimProd) from
*						the real product (Prod).					 
*       Comments		returns a pointer to the elimination-order product 
*						created.
*
******************************************************************************/
elimOrderProduct* createElimOrderProduct(product *prod)
{
	elimOrderProduct	*elimProd;
	variable			*var, *varEnd, **oldVarPtr, **oldVarPtrEnd;
	elimOrderVar		**newVarPtr, *newVar, **funcNewVars,
						**oldToNewMapping;
	function			*func, *funcEnd;
	int					maxVarsInFunc;		

#if UNITE_VARS
	node				*origNode, **uniteNodePtr, **uniteNodePtrEnd;
	elimOrderVar		**uniteVarPtr, **uniteVars, *newVarEnd;
#endif

	/* 1.	Allocate memory for the new elimOrderProduct. */

	elimProd = allocateElimOrderProduct(prod->currNumVars + 1, prod->currNumFuncs + 1,
										prod->numVariables);

	maxVarsInFunc = 0;

	/* 2.	Find the maximum number of variables in a function. */

	for (func = prod->functions, funcEnd = func + prod->numFunctions; func < funcEnd; func++) {
		if (func->used) {
			if (func->numVariables > maxVarsInFunc)
				maxVarsInFunc = func->numVariables;
		}
	}

	/* 3.	Allocate memory for auxiliary data-structures. */

	myMalloc(funcNewVars, sizeof(elimOrderVar*) * maxVarsInFunc);

	oldToNewMapping = elimProd->oldToNewVarMapping;

	/* 4.	Create new variables.  Each variable in the elimOrderProduct holds the index
			of the variable in the "real" product which it simulates. There is also
			a mapping from the variables in the "real" product to the variables in the
			"elimOrderProduct". */

	for (var = prod->variables, varEnd = var + prod->numVariables; var < varEnd; var++) {
		if (var->used) {
			newVar = insertVariableToElimOrderProduct(elimProd, var->currNumFunctions,
													  var->originalNode, var->numValues);

			oldToNewMapping[var - prod->variables] = newVar;
			newVar->indInRealProd = var - prod->variables;

			#if	UNITE_VARS
				if ((var->originalNode)->uniteT) {
					newVar->uniteVars = NULL;
				}
			#endif
		}
	}

	/* 5.	For each group of variables that need to be treated jointly, allocate an array
			that would point to all the variables, initialize it, and update the whole group
			to point to it. */

	#if	UNITE_VARS
		for (newVar = elimProd->variables, newVarEnd = newVar + elimProd->numVariables; 
				newVar < newVarEnd; newVar++) {

			if (newVar->used) {
			
				origNode = newVar->originalNode;

				if ((origNode->uniteT) && (newVar->uniteVars == NULL)) {
					myMalloc(uniteVars, sizeof(elimOrderVar*) * origNode->numVarsUnite);

					for (uniteVarPtr = uniteVars, uniteNodePtr = origNode->uniteVars, 
							uniteNodePtrEnd = uniteNodePtr + origNode->numVarsUnite;
							uniteNodePtr < uniteNodePtrEnd; uniteNodePtr++, uniteVarPtr++) {

						(*uniteVarPtr) = oldToNewMapping[(*uniteNodePtr)->varInProd - prod->variables];
						(*uniteVarPtr)->uniteVars = uniteVars;
					}
				}
			}
		}
	#endif

	/* 6.	Update the "firstVar" field of the "elimOrderProduct" to point to the first
			variable of the "real" product. */

	elimProd->firstVar = prod->variables;

	/* 7.	Create new functions. */

	for (func = prod->functions, funcEnd = func + prod->numFunctions; func < funcEnd; func++) {
		if (func->used) {
			for (oldVarPtr = func->variables, oldVarPtrEnd = oldVarPtr + func->numVariables,
					newVarPtr = funcNewVars; oldVarPtr < oldVarPtrEnd; oldVarPtr++, newVarPtr++) {
				
				*newVarPtr = oldToNewMapping[*oldVarPtr - prod->variables];
			}

			insertFunctionToElimOrderProduct(elimProd, func->numVariables, funcNewVars,
											 func->sizeOfTable);
		}
	}

	/* 8.	Build variable->function lookup tables. */

	buildElimOrderVariableFunctionTables(elimProd);

	/* 9.	Free memory. */

	myFree(funcNewVars);

	/* 10.	Return a pointer to the elimination-order-product created. */

	return elimProd;
}


/******************************************************************************
*
*       Function        copyElimOrderProduct
*       Parameters      oldProd (pointer to the product that needs to be 
*						duplicated), varMappingSize (number of spaces that need  
*						to be allocated for a variable mapping).
*       Output          none
*       Description		duplicates the product 'oldProd' and returns the new
*						duplicated product.
*       Comments		none
*
******************************************************************************/
elimOrderProduct* copyElimOrderProduct(elimOrderProduct *oldProd, 
									   long varMappingSize)
{
	elimOrderProduct	*newProd;
	elimOrderVar		**funcNewVars, *newVar, **oldToNewVarMapping, *var,
						*varEnd, **oldVarPtr, **oldVarPtrEnd, **newVarPtr;
	elimOrderFunc		*func, *funcEnd;
	int					maxVarsInFunc, ind;	

#if UNITE_VARS
	elimOrderVar		*newVarEnd, **uniteVarPtr, **uniteVars;
	node				*origNode, **uniteNodePtr, **uniteNodePtrEnd;
#endif
	
	/* 1.	Allocate memory for the new product. */

	newProd = allocateElimOrderProduct(oldProd->currNumVars + 1, oldProd->currNumFuncs + 1,
									   varMappingSize);

	maxVarsInFunc = 0;

	/* 2.	Find the maximum number of variables in a function. */

	for (func = oldProd->functions, funcEnd = func + oldProd->numFunctions; 
			func < funcEnd; func++) {
		if (func->used) {
			if (func->numVariables > maxVarsInFunc)
				maxVarsInFunc = func->numVariables;
		}
	}

	/* 3.	Allocate memory for auxiliary data-structures. */

	myMalloc(funcNewVars, sizeof(elimOrderVar*) * maxVarsInFunc);
	myMalloc(oldToNewVarMapping, sizeof(elimOrderVar*) * oldProd->numVariables);

	for (ind = 0; ind < varMappingSize; ind++) {
		newProd->oldToNewVarMapping[ind] = NULL;
	}

	/* 4.	Create new variables. Update the mapping from variables in the "real" product
			to the variables in the elimOrderProduct that simulate them. Also, update a 
			mapping from variables in the old elimOrderProduct to the corresponding
			variables in the new elimOrderProduct. */

	for (var = oldProd->variables, varEnd = var + oldProd->numVariables; var < varEnd; var++) {
		if (var->used) {
			newVar = insertVariableToElimOrderProduct(newProd, var->currNumFunctions,
													  var->originalNode, var->numValues);
			
			newVar->indInRealProd = var->indInRealProd;
			oldToNewVarMapping[var - oldProd->variables] = newVar;
			newProd->oldToNewVarMapping[var->indInRealProd] = newVar;
		}
	}

	/* 5.	For each group of variables that need to be treated jointly, allocate an array
			that would point to all the variables, initialize it, and update the whole group
			to point to it. */

	#if	UNITE_VARS
		for (newVar = newProd->variables, newVarEnd = newVar + newProd->numVariables; 
				newVar < newVarEnd; newVar++) {

			if (newVar->used) {
			
				origNode = newVar->originalNode;

				if ((origNode->uniteT) && (newVar->uniteVars == NULL)) {
					myMalloc(uniteVars, sizeof(elimOrderVar*) * origNode->numVarsUnite);

					for (uniteVarPtr = uniteVars, uniteNodePtr = origNode->uniteVars, 
							uniteNodePtrEnd = uniteNodePtr + origNode->numVarsUnite;
							uniteNodePtr < uniteNodePtrEnd; uniteNodePtr++, uniteVarPtr++) {

						(*uniteVarPtr) = newProd->oldToNewVarMapping[(*uniteNodePtr)->varInProd - oldProd->firstVar];
						(*uniteVarPtr)->uniteVars = uniteVars;
					}
				}
			}
		}
	#endif

	/* 5.	Update the "firstVar" field of the new elimOrderProduct to point to the first
			variable in the "real" product. */

	newProd->firstVar = oldProd->firstVar;

	/* 6.	Create new functions. */

	for (func = oldProd->functions, funcEnd = func + oldProd->numFunctions; func < funcEnd; func++) {
		if (func->used) {
			for (oldVarPtr = func->variables, oldVarPtrEnd = oldVarPtr + func->numVariables,
					newVarPtr = funcNewVars; oldVarPtr < oldVarPtrEnd; oldVarPtr++, newVarPtr++) {
				
				*newVarPtr = oldToNewVarMapping[*oldVarPtr - oldProd->variables];
			}

			insertFunctionToElimOrderProduct(newProd, func->numVariables, funcNewVars, func->sizeOfTable);
		}
	}

	/* 7.	Build new variable->function lookup tables. */

	buildElimOrderVariableFunctionTables(newProd);

	/* 8.	Free memory. */

	myFree(oldToNewVarMapping);
	myFree(funcNewVars);

	/* 9.	Return a pointer to the duplicated product. */

	return newProd;
}



/******************************************************************************
*
*       Function        zeroSeenFuncsVars
*       Parameters		funcArr (an array of functions whose variables 'seen' flags
*						need to be zeroed), numFuncs (the number of functions in
*						the array).
*       Output          none
*       Description		zeros the 'seen' flags of all the variables that appear
*						in some function in 'funcArr'.
*       Comments		none
*
******************************************************************************/
void zeroSeenFuncsVars(elimOrderFunc **funcArr, int numFuncs)
{
	elimOrderFunc	**funcPtr, **funcPtrEnd, *func;
	elimOrderVar	**varPtr, **varPtrEnd;

	/* 1.	Iterate over all the functions the variable appears in. */

	for (funcPtr = funcArr, funcPtrEnd = funcPtr + numFuncs;
			funcPtr < funcPtrEnd; funcPtr++) {

		func = *funcPtr;

		/* 1.1.	For each function, iterate over all of its variables and update 
				their 'seen' flags to FALSE. */

		for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
				varPtr < varPtrEnd; varPtr++) 
			(*varPtr)->alreadySeen = FALSE;
	}
}



/******************************************************************************
*
*       Function        getVarsInCombinedElimOrderFunc
*       Parameters		funcArr (an array of functions whose variables are 
*						requested), numFuncs (the number of functions in the 
*						array), varArr (an array that would be upadted with 
*						pointers to all the variables in 'funcArr'), numVars 
*						(the number of variables in the functions in 'funcArr').
*       Output          none
*       Description		inserts into varArr all the variables that appear in one
*						of the functions in 'funcArr' & updates 'numVars' to the 
*						number of variables.
*       Comments		none
*
******************************************************************************/
void getVarsInCombinedElimOrderFunc(elimOrderFunc **funcArr, int numFuncs,
									elimOrderVar **varArr, int *numVars,
									elimOrderVar *elimVar)
{
	int				varCount;
	elimOrderFunc	**funcPtr, **funcPtrEnd, *func;
	elimOrderVar	**varPtr, **varPtrEnd, *var, **varArrPlace;

	/* 1.	Initialize. */

	varCount = 0;
	varArrPlace = varArr;

	/* 2.	Zero the 'seen' flags of all the variables in the functions of
			'funcArr'. */

	zeroSeenFuncsVars(funcArr, numFuncs);

	if (elimVar != NULL)
		elimVar->alreadySeen = TRUE;

	/* 3.	Iterate over all of the functions in the array. */

	for (funcPtr = funcArr, funcPtrEnd = funcPtr + numFuncs; 
			funcPtr < funcPtrEnd; funcPtr++) {

		func = *funcPtr;
		
		/* 3.1.	For each function, iterate over all of its variables. */

		for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			var = *varPtr;

			/* 3.1.1.	If we haven't seen this variable yet */

			if (var->alreadySeen == FALSE) {

				/* 3.1.1.1.	Update that we've seen it. */

				var->alreadySeen = TRUE;

				/* 3.1.1.2.	Insert it into varArr, and update the varCount. */

				*varArrPlace++ = var;
				varCount++;
			}
		}
	}

	*numVars = varCount;
}


/******************************************************************************
*
*       Function        getVarsInCombinedElimOrderFuncJoint
*       Parameters		elimVar (a pointer to the first variable being 
*						eliminated), varArr (will hold pointers to the all the 
*						variables that appear in the functions in 'funcArr' and
*						aren't among the variables that are being eliminated),
*						numVars (will hold the number of variables in the combined
*						function), sizeOfTable (will hold the size of the 
*						probability table of the combined function).
*       Output          none 
*       Description		inserts into varArr all the variables that will appear in
*						the combined function after eliminating the variable in
*						'elimVar', and all the rest of the variables that need to 
*						be eliminated jointly with it, updates 'numVars' with the 
*						number of variables	and 'sizeOfTable' to the multiplication 
*						of the numVals of the variables. Also updates 'logMultFactor'
*						and 'rescaleTimer' to the required values for the combined
*						function.
*       Comments		none
*
******************************************************************************/
void getVarsInCombinedElimOrderFuncJoint(elimOrderVar **elimVarsArr, int numElimVars,
										 elimOrderFunc **funcArr, int *numFuncs,
										 elimOrderVar **varArr, int *numVars,
									 	 unsigned long *sizeOfTable)
{
	int				varCount, funcCount;
	elimOrderFunc	**funcPtr, **funcPtrEnd, *func, **funcArrPtr;
	elimOrderVar	**varPtr, **varPtrEnd, *var, **varArrPlace;
	unsigned long	tableSize;

	/* 1.	Initializations. */

	varCount = 0;
	varArrPlace = varArr;
	tableSize = 1;
	funcArrPtr = funcArr;
	funcCount = 0;

	/* 2.	Update an array with pointers to all the functions that need to be combined. */

	for (varPtr = elimVarsArr, varPtrEnd = varPtr + numElimVars; varPtr < varPtrEnd; 
			varPtr++) {
		var = (*varPtr);

		for (funcPtr = var->funcs, funcPtrEnd = funcPtr + var->currNumFunctions;
				funcPtr < funcPtrEnd; funcPtr++) {	

			func = (*funcPtr);

			if (func->alreadySeen == FALSE) {
				*funcArrPtr++ = func;
				funcCount++;
				func->alreadySeen = TRUE;
			}
		}
	}

	*numFuncs = funcCount;

	/* 3.	Zero the 'seen' flags of all of the variables that appear in the functions
			which need to be combined except for the variables that need to be eliminated. */

	zeroSeenFuncsVars(funcArr, funcCount);

	for (varPtr = elimVarsArr, varPtrEnd = varPtr + numElimVars; varPtr < varPtrEnd; 
			varPtr++) {
		(*varPtr)->alreadySeen = TRUE;
	}

	/* 4.	Iterate over all of the functions in 'funcArr'. */

	for (funcPtr = funcArr, funcPtrEnd = funcPtr + funcCount; funcPtr < funcPtrEnd; funcPtr++) {

		func = *funcPtr;

		/* 4.1	For each function, iterate over all of its variables. */

		for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			var = *varPtr;

			/* 4.1.1	If we haven't seen this variable yet */

			if (var->alreadySeen == FALSE) {

				/* 4.1.1.1	Update that we've seen it. */

				var->alreadySeen = TRUE;

				/* 4.1.1.2	Insert it into varArr, and update the numVars count 
							and the tableSize calculation. */

				*varArrPlace++ = var;
				varCount++;
				tableSize *= var->numValues;
			}
		}
	}

	*numVars = varCount;
	*sizeOfTable = tableSize;
}




/******************************************************************************
*
*       Function        eliminateVarFromFunction
*       Parameters		elimFunc (the function which needs to be restricted, 
*						the variable 'brokenVar' needs to be removed from the 
*						function), brokenVar (the variable that needs to be 
*						removed from the function), oldNumVals (the number of 
*						possible values of 'brokenVar' originally).
*       Output          none
*       Description		removes the variable 'brokenVar' from the function
*						'elimFunc' (removes it from the variable array of the 
*						function, updates the variable count and the size of 
*						the probability table of the function).  
*       Comments		this function is called when we condition on 
*						some variable (i.e., we set the value of this
*						variable and turn it into a constant).
*
******************************************************************************/
void eliminateVarFromFunction(elimOrderProduct *elimProd, elimOrderFunc *elimFunc, 
							  elimOrderVar *brokenVar, int oldNumVals)
{
	elimOrderVar	**newVariables;
	elimOrderVar	**varPtrOld, **varPtrEnd, **varPtrNew;

#if CALC_MEM_REQ	
	unsigned long	oldSizeOfTable, memReqUpdate;
#endif

	/* 1.	Allocate memory for the new variable array of the function. */

	myMalloc(newVariables, sizeof(elimOrderVar*) * (elimFunc->numVariables - 1));

	varPtrOld = elimFunc->variables;
	varPtrNew = newVariables;

	/* 2.	Iterate over the function's variables. */

	for (varPtrOld = elimFunc->variables, varPtrEnd = varPtrOld + elimFunc->numVariables,
			varPtrNew = newVariables; varPtrOld < varPtrEnd; varPtrOld++) {

		/* 2.1.	Copy each variable which isn't 'brokenVar' to the new variable
				array and invalidate its elimination complexity cache. */

		if (*varPtrOld != brokenVar) {
			*varPtrNew++ = *varPtrOld;
			
			(*varPtrOld)->elimComplexity = INVALID;
		}
	}

	/* 3.	Update the relevant function fields. */

	myFree(elimFunc->variables);
	elimFunc->variables = newVariables;
	elimFunc->numVariables--;

#if CALC_MEM_REQ	
	oldSizeOfTable = elimFunc->sizeOfTable;
#endif

	elimFunc->sizeOfTable /= brokenVar->numValues;

#if CALC_MEM_REQ
	memReqUpdate = (oldSizeOfTable - elimFunc->sizeOfTable) * sizeof(Probability);
	elimProd->currMemReqCurrProd -= memReqUpdate;
#endif
}



/******************************************************************************
*
*       Function        buildElimOrderVariableFunctionTables
*       Parameters		elimProd (the elimOrderProduct which needs to be updated
*						so that each variable will point to the functions that
*						depend on it).
*       Output          none
*       Description		builds all the variable->function lookup tables from
*						the existing function->variable lookup tables (each 
*						variable will point to all the functions it appears in).						 
*       Comments		none
*
******************************************************************************/
void buildElimOrderVariableFunctionTables(elimOrderProduct *elimProd)
{
	elimOrderFunc	*funcPtr, *funcPtrEnd;
	elimOrderVar	**varPtr, **varPtrEnd, *variablePtr;

	/* 1.	Iterate over the elimOrderProduct's functions. */

	for (funcPtr = elimProd->functions, funcPtrEnd = funcPtr + elimProd->numFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {
		
		/* 1.1.	For each used function: */

		if (funcPtr->used) {

			/* 1.1.1.	Iterate over all its variables */

			for (varPtr = funcPtr->variables, varPtrEnd = varPtr + funcPtr->numVariables;
					varPtr < varPtrEnd; varPtr++) {

				/* 1.1.1.1. and update their inverse tables. */

				variablePtr = *varPtr;
				variablePtr->funcs[variablePtr->currNumFunctions++] = funcPtr;
			}
		}
	}
}


/******************************************************************************
*
*       Function        calcElimComplexity
*       Parameters		var (the variable of which elimination complexity 
*						is calculated).
*       Output          none
*       Description		calculates the elimination complexity of 'var'.
*       Comments		none
*
******************************************************************************/
double calcElimComplexity(elimOrderVar *var)
{
	elimOrderFunc	**funcPtr, **funcPtrEnd, *func;
	elimOrderVar	**varPtr, **varPtrEnd, *currVar;
	double			elimComplexity = 1;		/* the multiplication of the number 
											   of possible values of each of var's 
											   neighbors. */
	int				cliqueSize = 1;			/* the number of neighbors of var + 1. */
	int				maxVarsInFunc = 0;		/* will hold the maximum number of 
											   variables in a function, among the 
											   functions in which var appears. */

	#if	UNITE_VARS
		if ((var->originalNode)->uniteT) {
			return calcElimComplexityJoint(var->uniteVars, var->originalNode->numVarsUnite);
		}
	#endif

	/* 1.	Zero all the 'seen' flags of the variable's neighbors. */

	zeroSeenFuncsVars(var->funcs, var->currNumFunctions);

	var->alreadySeen = TRUE;

	/* 2.	Iterate over the functions in which 'var' appears. */

	for (funcPtr = var->funcs, funcPtrEnd = funcPtr + var->currNumFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {
		
		func = *funcPtr;
		
		maxVarsInFunc = (func->numVariables > maxVarsInFunc) ? func->numVariables : 
						maxVarsInFunc;

		/* 2.1.	Iterate over all of the variables in each one of the functions. */

		for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			currVar = *varPtr;

			/* 2.1.1.	If we haven't seen the variable yet, mark that it's been 
						seen, update the 'number of neighbors' count and update the
						elimination complexity calculation. */

			if (currVar->alreadySeen == FALSE) {
				currVar->alreadySeen = TRUE;
				cliqueSize++;

				elimComplexity *= currVar->numValues;
			}
		}
	}

	/* 3.	If some function includes all of var's neighbors, then the elimination complexity
			of the var is zero. Otherwise, it is the multiplication of the number of possible
			values of each of var's neighbors. */

	var->elimComplexity = (cliqueSize == maxVarsInFunc) ? 0 : elimComplexity;
	var->tableSize = elimComplexity;
	var->numFuncsUnite = var->currNumFunctions;

	#if (PRINT_VARIOUS_STATISTICS_LEVEL > 1)
		var->cliqueSize = cliqueSize;
	#endif

	/* 4.	Returns the elimination complexity of var. */

	return elimComplexity;
}


/******************************************************************************
*
*       Function        calcElimComplexityJoint
*       Parameters		varArr (the variables of which joint elimination 
*						complexity is calculated), numVars (the number of
*						variables in the array).
*       Output          none
*       Description		calculates the joint elimination complexity of the
*						variables in the array.
*       Comments		none
*
******************************************************************************/
double calcElimComplexityJoint(elimOrderVar **varArr, int numVars)
{
	elimOrderFunc	**funcPtr, **funcPtrEnd, *func;
	elimOrderVar	**varPtr, **varPtrEnd, *currVar, **elimVarPtr, **elimVarPtrEnd, 
					*elimVar;
	double			elimComplexity = 1;		/* The multiplication of the number 
											   of possible values of each of var's 
											   neighbors. */
	int				cliqueSize = 1;			/* the number of neighbors of var + 1. */
	int				maxVarsInFunc = 0;		/* will hold the maximum number of 
											   variables in a function, among the 
											   functions in which var appears. */
	int				numFuncs = 0;			/* The total number of functions that the
											   eliminated variables appear in. */


	/* 1.	Iterate over the variables in the array. */

	for (elimVarPtr = varArr, elimVarPtrEnd = elimVarPtr + numVars; 
			elimVarPtr < elimVarPtrEnd; elimVarPtr++) {

		elimVar = *elimVarPtr;

		/* 1.1.	Zero the 'seen' flags of the functions in which the variable participates. */

		zeroSeenFuncs(elimVar->funcs, elimVar->currNumFunctions, funcPtr, funcPtrEnd);

		/* 1.2.	Zero all the 'seen' flags of the variable's neighbors. */

		zeroSeenFuncsVars(elimVar->funcs, elimVar->currNumFunctions);
	}

	for (elimVarPtr = varArr, elimVarPtrEnd = elimVarPtr + numVars; 
			elimVarPtr < elimVarPtrEnd; elimVarPtr++) {
		
		elimVar = *elimVarPtr;
		elimVar->alreadySeen = TRUE;
	}

	/* 2.	Iterate over the variables in the array. */

	for (elimVarPtr = varArr, elimVarPtrEnd = elimVarPtr + numVars; 
			elimVarPtr < elimVarPtrEnd; elimVarPtr++) {

		elimVar = *elimVarPtr;

		/* 2.1.	Iterate over the functions in which the variable appears. */

		for (funcPtr = elimVar->funcs, funcPtrEnd = funcPtr + elimVar->currNumFunctions;
				funcPtr < funcPtrEnd; funcPtr++) {

			func = *funcPtr;

			if (func->alreadySeen)
				continue;

			func->alreadySeen = TRUE;
		
			maxVarsInFunc = (func->numVariables > maxVarsInFunc) ? func->numVariables : 
								maxVarsInFunc;
			numFuncs++;

			/* 2.1.1.	Iterate over all of the variables in each one of the functions. */

			for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
					varPtr < varPtrEnd; varPtr++) {

				currVar = *varPtr;

				/* 2.1.1.1.	If we haven't seen the variable yet, mark that it's been 
							seen, update the 'number of neighbors' count and update the
							elimination complexity calculation. */

				if (currVar->alreadySeen == FALSE) {
					currVar->alreadySeen = TRUE;

					elimComplexity *= currVar->numValues;
					cliqueSize++;
				}
			}
		}
	}

	/* 3.	If some function includes all of var's neighbors, then the elimination complexity
			of the var is zero. Otherwise, it is the multiplication of the number of possible
			values of each of var's neighbors. */

	for (elimVarPtr = varArr, elimVarPtrEnd = elimVarPtr + numVars; 
			elimVarPtr < elimVarPtrEnd; elimVarPtr++) {

		elimVar = *elimVarPtr;

		elimVar->elimComplexity = (cliqueSize == maxVarsInFunc) ? 0 : elimComplexity;
		elimVar->tableSize = elimComplexity;
		elimVar->numFuncsUnite = numFuncs;

		#if (PRINT_VARIOUS_STATISTICS_LEVEL > 1)
			elimVar->cliqueSize = cliqueSize;
		#endif
	}

	/* 3.	Returns the elimination complexity of var. */

	return elimComplexity;
}




/******************************************************************************
*
*       Function        eliminateElimOrderVars
*       Parameters		elimProd (the elimination-order-product), 
*						elimVar (pointer to the variable that needs to be 
*						eliminated from the product).
*       Output          none
*       Description		eliminates 'elimVar' from 'elimProd' by multiplying all
*						the functions that it appears in and creating one new
*						combined function that 'elimVar' doesn't appear in.					 
*       Comments		updated 23.10.
*
******************************************************************************/
void eliminateElimOrderVars(elimOrderVar *elimVar, int numElimVars, 
						    double *tableSizes, long *elimInd, 
						    elimOrderNode **elimNodePtr, elimOrderProduct *elimProd)
{
	elimOrderNode	*elimNode;
	elimOrderVar	**elimVars, **elimVarsPtr, **elimVarsPtrEnd;
	long			currElimInd;


	elimNode = *elimNodePtr;
	
	if (numElimVars == 1) {
		updateElimNode(elimNode, FALSE, elimVar->originalNode);

		tableSizes[*elimInd] = elimVar->tableSize;
		(*elimNodePtr)++;
		(*elimInd)++;

		eliminateElimOrderVar(elimProd, elimVar);
	}
	else {
		/* 1.	Initialization. */

		elimVars = elimVar->uniteVars;
		elimVarsPtr = elimVars;
		elimVarsPtrEnd = elimVarsPtr + numElimVars;
		currElimInd = *elimInd;

		/* 2.	Update the elimination order array with all the nodes that need to be 
				eliminated jointly. */

		for (; elimVarsPtr < elimVarsPtrEnd; elimVarsPtr++, elimNode++, currElimInd++) {
			
			updateElimNode(elimNode, FALSE, ((*elimVarsPtr)->originalNode));
			tableSizes[currElimInd] = 0;
		}

		tableSizes[*elimInd] = elimVars[0]->tableSize;
		*elimInd = currElimInd;
		*elimNodePtr = elimNode;

		eliminateElimOrderVarsJoint(elimProd, elimVars, numElimVars);
	}
}


/******************************************************************************
*
*       Function        eliminateElimOrderVars
*       Parameters		elimProd (the elimination-order-product), 
*						elimVar (pointer to the variable that needs to be 
*						eliminated from the product).
*       Output          none
*       Description		eliminates 'elimVar' from 'elimProd' by multiplying all
*						the functions that it appears in and creating one new
*						combined function that 'elimVar' doesn't appear in.					 
*       Comments		updated 23.10.
*
******************************************************************************/
void eliminateElimOrderVarsPre(elimOrderVar *elimVar, int numElimVars, 
							   long *elimInd, elimOrderNode **elimNodePtr, 
							   elimOrderProduct *elimProd)
{
	elimOrderNode	*elimNode;
	elimOrderVar	**elimVars, **elimVarsPtr, **elimVarsPtrEnd;
	long			currElimInd;


	elimNode = *elimNodePtr;
	
	if (numElimVars == 1) {
		updateElimNode(elimNode, FALSE, elimVar->originalNode);

		(*elimNodePtr)++;
		(*elimInd)++;

		eliminateElimOrderVar(elimProd, elimVar);
	}
	else {
		/* 1.	Initialization. */

		elimVars = elimVar->uniteVars;
		elimVarsPtr = elimVars;
		elimVarsPtrEnd = elimVarsPtr + numElimVars;
		currElimInd = *elimInd;

		/* 2.	Update the elimination order array with all the nodes that need to be 
				eliminated jointly. */

		for (; elimVarsPtr < elimVarsPtrEnd; elimVarsPtr++, elimNode++, currElimInd++) {
			
			updateElimNode(elimNode, FALSE, ((*elimVarsPtr)->originalNode));
		}

		*elimInd = currElimInd;
		*elimNodePtr = elimNode;

		eliminateElimOrderVarsJoint(elimProd, elimVars, numElimVars);
	}
}



/******************************************************************************
*
*       Function        eliminateElimOrderVar
*       Parameters		elimProd (the elimination-order-product), 
*						elimVar (pointer to the variable that needs to be 
*						eliminated from the product).
*       Output          none
*       Description		eliminates 'elimVar' from 'elimProd' by multiplying all
*						the functions that it appears in and creating one new
*						combined function that 'elimVar' doesn't appear in.					 
*       Comments		updated 23.10.
*
******************************************************************************/
void eliminateElimOrderVar(elimOrderProduct *elimProd, elimOrderVar *elimVar)
{
	elimOrderVar		**varArr, **varPtr, **varPtrEnd;
	int					numVars;
	elimOrderFunc		*combinedFunc = NULL, **funcPtr, **funcPtrEnd, *func;


	/* 1.	Allocate memory for the variable array of the new combined function. */

	varArr=malloc(sizeof(elimOrderVar*) * elimProd->currNumVars);
	/*	myMalloc(varArr, sizeof(elimOrderVar*) * elimProd->currNumVars);*/

	/* 2.	Find all the variables that appear in the functions that 
			need to be combined. */

	getVarsInCombinedElimOrderFunc(elimVar->funcs, elimVar->currNumFunctions, varArr, 
								   &numVars, elimVar);

	/* 3.	Allocate memory for the new combined function & insert it into 
			the product, in case the combined function isn't a constant. */

	if (numVars) {		
		combinedFunc = insertFunctionToElimOrderProduct(elimProd, numVars, varArr,
													    (unsigned long)elimVar->tableSize);

		#if CALC_MEM_REQ
			if (progData.programCode == SUPER_HAPLO) {
				elimProd->currMemReq += 
						(elimVar->tableSize * sizeof(int) * elimProd->condDepth);
			}
		#endif
	}

	/* 4.	Remove the functions that have been multiplied from the variables 
			that appear in them and from the Product. */

	for (funcPtr = elimVar->funcs + elimVar->currNumFunctions - 1, 
			funcPtrEnd = elimVar->funcs; funcPtr >= funcPtrEnd; funcPtr--) {

		func = *funcPtr;

		for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			removeFunctionFromElimOrderVariable(*varPtr, func);
		}

		removeFunctionFromElimOrderProduct(elimProd, func);
	}

	/* 5.	Add the new combined function to all the variables that appear in it. */

	if (numVars) {
		for (varPtr = combinedFunc->variables, varPtrEnd = varPtr + combinedFunc->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			addFunctionToElimOrderVariable(*varPtr, combinedFunc);
		}
	}

	/* 6.	Trim the variable that was eliminated from the product. */

	removeVariableFromElimOrderProduct(elimProd, elimVar);

	myFree(varArr);
}



/******************************************************************************
*
*       Function        eliminateElimOrderVarsJoint
*       Parameters		elimProd (the elimination-order-product), 
*						elimVar (pointer to the variable that needs to be 
*						eliminated from the product).
*       Output          none
*       Description		eliminates 'elimVar' from 'elimProd' by multiplying all
*						the functions that it appears in and creating one new
*						combined function that 'elimVar' doesn't appear in.					 
*       Comments		none
*
******************************************************************************/
unsigned long eliminateElimOrderVarsJoint(elimOrderProduct *elimProd, 
									      elimOrderVar **elimVars, int numElimVars)
{
	elimOrderVar		**varArr, **varPtr, **varPtrEnd, *var, **uniteVars;
	int					numVars, numFuncs;
	elimOrderFunc		*combinedFunc = NULL, **funcPtr, **funcPtrEnd, *func,
						**funcArr;
	unsigned long		tableSize;


	/* 1.	Iterate over the variables that need to be eliminated jointly.
			Count the number of functions they appear in (an overestimate),
			and initialize the 'alreadySeen' flags of the functions to FALSE. */

	numFuncs = 0;

	for (varPtr = elimVars, varPtrEnd = varPtr + numElimVars; varPtr < varPtrEnd; 
			varPtr++) {
		var = (*varPtr);

		numFuncs += var->currNumFunctions;

		for (funcPtr = var->funcs, funcPtrEnd = funcPtr + var->currNumFunctions;
				funcPtr < funcPtrEnd; funcPtr++) {	

			(*funcPtr)->alreadySeen = FALSE;
		}
	}

	/* 2.	Allocate memory for the variable array of the new combined function. */

	myMalloc(varArr, sizeof(elimOrderVar*) * elimProd->currNumVars);

	/* 3.	Allocate memory for a temporary array that would store pointers to all
			the functions that need to be combined. */

	myMalloc(funcArr, sizeof(elimOrderFunc*) * numFuncs);

	/* 4.	Find all the variables that appear in the functions that 
			need to be combined. */

	getVarsInCombinedElimOrderFuncJoint(elimVars, numElimVars, funcArr, &numFuncs,
									    varArr, &numVars, &tableSize);

	/* 5.	Allocate memory for the new combined function & insert it into 
			the product, in case the combined function isn't a constant. */

	if (numVars) {		
		combinedFunc = insertFunctionToElimOrderProduct(elimProd, numVars, varArr,
													    tableSize);
	}

	/* 5.	Remove the functions that have been multiplied from the variables 
			that appear in them and from the Product. */

	for (funcPtr = funcArr, funcPtrEnd = funcPtr + numFuncs; funcPtr < funcPtrEnd;
			funcPtr++) {

		func = *funcPtr;

		for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			removeFunctionFromElimOrderVariable(*varPtr, func);
		}

		removeFunctionFromElimOrderProduct(elimProd, func);
	}

	/* 6.	Add the new combined function to all the variables that appear in it. */

	if (numVars) {
		for (varPtr = combinedFunc->variables, varPtrEnd = varPtr + combinedFunc->numVariables; 
				varPtr < varPtrEnd; varPtr++) {

			addFunctionToElimOrderVariable(*varPtr, combinedFunc);
		}
	}

	/* 7.	Trim the variables that were eliminated from the product. */

	uniteVars = elimVars;

	for (varPtr = elimVars, varPtrEnd = varPtr + numElimVars;
			varPtr < varPtrEnd; varPtr++) {

		removeVariableFromElimOrderProduct(elimProd, *varPtr);
	}
	
	myFree(uniteVars);

	/* 8.	Free the memory allocated dynamically. */

	myFree(varArr);
	myFree(funcArr);

	/* 9.	Return the size of the probability table created during the joint 
			elimination of the variables. */

	return tableSize;
}



/******************************************************************************
*
*       Function        conditionElimOrderVar
*       Parameters		condVars (an array of pointers to the variables that 
*						need to be conditioned on), numVars (the number of 
*						variables in the array), elimInd (the current elimination
*						index), elimNodePtr (pointer to the current elimination
*						order node),  elimProd (the elimination-order-product).
*       Output          none
*       Description		removes the variables in 'condVars' from the functions 
*						that they appear in and from the product. Updates the 
*						elimination order accordingly.
*       Comments		none.
*
******************************************************************************/
void conditionOnElimOrderVar(elimOrderVar *condVar, int numVars, long *elimInd,
						     elimOrderNode **elimNodePtr, elimOrderProduct *elimProd)
{
	elimOrderVar	**condVars, **condVarPtr, **condVarPtrEnd;
	elimOrderNode	*elimNode;


	if (numVars == 1) {
		condVars = &condVar;
	}
	else {
		condVars = condVar->uniteVars; 
	}

	elimNode = *elimNodePtr;

	/* 1.	Iterate over the variables that need to be conditioned on. */

	for (condVarPtr = condVars, condVarPtrEnd = condVarPtr + numVars;
			condVarPtr < condVarPtrEnd; condVarPtr++, elimNode++) {

		/* 1.1.	Remove the variable from the product. */
	
		setBreakValElimOrder(elimProd, *condVarPtr);

		/* 1.2.	Update the current elimination-order-node. */

		updateElimNode(elimNode, TRUE, (*condVarPtr)->originalNode);
	}

	/* 2.	Update the elimination index and the elimination-order-node. */

	*elimInd += numVars;
	*elimNodePtr = elimNode;

	/* 3.	Free the memory allocated for the array of pointers. */

	if (numVars > 1) {
		myFree(condVars);
	}
}



/******************************************************************************
*
*       Function        setBreakValElimOrder
*       Parameters		elimProd (the elimination-order-product), brokenVar (the
*						variable on which the product is broken).
*       Output          none
*       Description		removes the 'brokenVar' from the functions that it 
*						appears in and from the product.
*       Comments		none.
*
******************************************************************************/
void setBreakValElimOrder(elimOrderProduct *elimProd, elimOrderVar *brokenVar)
{
	elimOrderFunc	**funcPtr, **funcPtrEnd, *func;
	int				oldNumFunctions;

	oldNumFunctions = brokenVar->currNumFunctions;

	/* 1.	Remove brokenVar from each of the functions that it appears in. */

	for (funcPtr = brokenVar->funcs, funcPtrEnd = funcPtr + oldNumFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {

		func = *funcPtr;

		eliminateVarFromFunction(elimProd, func, brokenVar, brokenVar->numValues);
	}

	/* 2.	Remove the broken variable from the product. */

	removeVariableFromElimOrderProduct(elimProd, brokenVar);

#if CALC_MEM_REQ
	elimProd->currMemReq += elimProd->currMemReqCurrProd;
	elimProd->condDepth++;
#endif
}



