/* $Id: Function.c,v 1.2 2008/08/03 08:15:46 superlink Exp $ */
/*********************************************************************************/
/*  Function.c																     */
/*  Author: Ma'ayan Fishelson                                                    */
/*********************************************************************************/

#include "Product.h"

/*------------------------------*/
/*		GLOBAL VARIABLES		*/
/*------------------------------*/

extern FILE *outFile;


/******************************************************************************
*
*       Function        insertFunctionToProduct
*       Parameters		prod (the Product to which the function is inserted),
*						tableSize (the size of the probability table of the 
*						function), probabList (the probability table of the 
*						function), logMultFactor (a factor used to rescale the
*						probabilities in the table to prevent underflow), 
*						isMyProbTable (is TRUE if the probability table of this
*						function isn't shared with another function), 
*						numVariables (the number of variables of the function),
*						variables (an array of pointers to the variables of the
*						function), DuplicateVars (is TRUE if the variable array
*						needs to be copied), rescaleTimer (a timer that indicates
*						when the probability table of the function needs to be 
*						rescaled).
*       Output          none
*       Description		Inserts a new function into the product and updates its
*						fields.						 
*       Comments		none
*
******************************************************************************/
function* insertFunctionToProduct(product *prod, unsigned long tableSize, 
								  Probability *probabList, double logMultFactor,
								  boolean isMyProbTable, int numVariables, 
								  variable **variables, boolean DuplicateVars,
								  int rescaleTimer)
{
	function	*funcPtr, *funcPtrEnd;
	int			varCount;

	/* 1.	Search for an empty place to insert the new function, starting at 
			nextFuncPlace and wrapping around. */

	for (funcPtr = prod->nextFuncPlace, 
			funcPtrEnd = prod->functions + prod->numFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {
		if (!funcPtr->used)
			goto enterFuncFields;
	}

	for (funcPtr = prod->functions, funcPtrEnd = prod->nextFuncPlace; 
			funcPtr < funcPtrEnd; funcPtr++) {
		if (!funcPtr->used)
			goto enterFuncFields;
	}

	/* 2.	Initialize the fields of this place. */

	enterFuncFields:

	funcPtr->used = TRUE;
	funcPtr->numVariables = numVariables;

	if (DuplicateVars) {
		myMalloc(funcPtr->variables, sizeof(variable*) * numVariables);

		for (varCount = 0; varCount < numVariables; varCount++)
			funcPtr->variables[varCount] = variables[varCount];
	}
	else
		funcPtr->variables = variables;

	funcPtr->sizeOfTable = tableSize;
	funcPtr->probabList = probabList;
	funcPtr->logMultFactor = logMultFactor;
	funcPtr->rescaleTimer = rescaleTimer;
	funcPtr->myProbabList = isMyProbTable;
	funcPtr->ordering = NULL;
	funcPtr->accOffset = NULL;

	/* 3.	Update the Product's fields. */

	prod->nextFuncPlace = funcPtr + 1;
	prod->currNumFuncs++;

	/* 4.	Return a pointer to the new function. */

	return funcPtr;
}


/******************************************************************************
*
*       Function        getFunctionVariables
*       Parameters		prod (the product of which this function is part of), 
*						Node (the requested function is the probability function
*					    of this node given its parents in the Bayesian network), 
*						numVariables (will be updated to hold the number of 
*						variables this function depends on).
*       Output          none
*       Description		Returns an array of pointers to the variables of the
*						conditional probability function of 'Node', and updates
*						'numVariables' to hold the number of variables in the
*						function.
*       Comments		none
*
******************************************************************************/
variable** getFunctionVariables(product *prod, node *Node, int *numVariables)
{
	variable		**variables = NULL, **varPtr;
	int				varCount;
	node			**nodePtr, **nodePtrEnd, *parentNode;


	varCount = 0;

	/* 1.	Count one variable for the node if its value isn't fixed. */

	if (Node->numValues > 1)
		varCount++;

	/* 2.	Count another variable for each unfixed parent of the node. */

	for (nodePtr = Node->parents, nodePtrEnd = nodePtr + Node->numParents;
			nodePtr < nodePtrEnd; nodePtr++) {

		parentNode = *nodePtr;

		if (parentNode->numValues > 1)
			varCount++;
	}

	/* 3.	Allocate memory for the list of variables. */

	if (varCount > 0) {
		myMalloc(variables, sizeof(variable*) * varCount);

		varPtr = variables;

		/* 3.1	Add the node to the variable array, if it's not fixed. */

		if (Node->numValues > 1) 
			*varPtr++ = getNodeVariable(Node, prod);

		/* 3.2	Add each unfixed parent to the array. */

		for (nodePtr = Node->parents, nodePtrEnd = nodePtr + Node->numParents;
				nodePtr < nodePtrEnd; nodePtr++) {
			parentNode = *nodePtr;

			if (parentNode->numValues > 1) 
				*varPtr++ = getNodeVariable(parentNode, prod);
		}
	}

	/* 4.	Update the count of the number of variables. */

	*numVariables = varCount;

	/* 5.	Return the array of pointers to the function's variables. */

	return variables;
}



/******************************************************************************
*
*       Function        removeFunctionFromProduct
*       Parameters		prod (a pointer to the product from which the function  
*						is removed), func (the function that is being removed
*						from the product).
*       Output          none
*       Description		Removes 'func' from 'prod'.						 
*       Comments		none
*
******************************************************************************/
void removeFunctionFromProduct(product *prod, function *func)
{
	/* 1.	Free the memory allocated to the function's fields and the slot. */

	myFree(func->variables);

	if (func->myProbabList) {
		myFree(func->probabList);
	}

	myFree(func->ordering);
	myFree(func->accOffset);

	/* 2.	Mark this field as free and update the Product's function counter. */

	func->used = FALSE;
	prod->currNumFuncs--;
}


/******************************************************************************
*
*       Function        removeVariableFromFunction
*       Parameters		func (pointer to the function from which the variable
*						is being removed), var (pointer to the variable 
*						that is being removed from the function).
*       Output          none
*       Description		Removes 'var' from the variable array of 'func'.						 
*       Comments		none
*
******************************************************************************/
void removeVariableFromFunction(function *func, variable *var)
{
	variable	**varPlace, **varPlaceEnd, **tempVarPlace;

	/* 1.	Search for 'var' in the variable array of 'func'. */

	for (varPlace = func->variables, varPlaceEnd = varPlace + func->numVariables;
			varPlace < varPlaceEnd; varPlace++) {
		
		/* 1.1	Once found, shift all the following variables one slot back and
				update the counter of the number of variables of the function. */

		if (*varPlace == var) {
			for (tempVarPlace = varPlace; tempVarPlace < varPlaceEnd - 1; tempVarPlace++)
				(*tempVarPlace) = *(tempVarPlace + 1);

			func->numVariables--;
			return;
		}
	}
}


/******************************************************************************
*
*       Function        reduceFunctionTable
*       Parameters		prod (the product of which the function being reduced is
*						part of), func (the function being reduced),
*						varValueUsageFlagsGen (an array of flags which indicate
*						which variable values are impossible and need to be 
*						eliminated), maxValuesVar (maximum number of possible 
*						values of a variable), numValsVarsNew (an array of the 
*						number of possible values of each variable after the 
*						elimination), tempRes (will hold the logProbability of 
*						the new function, if it is reduced into a constant).
*       Output          none
*       Description		Creates the new function from the old one, by  
*						eliminating the impossible variable values.  Returns a 
*						pointer to the new reduced function, if it's not 
*						a constant.  Otherwise, returns NULL.			 
*       Comments		none
*
******************************************************************************/
function* reduceFunctionTable(product *prod, function *func, 
							  boolean *varValueUsageFlagsGen, int maxValuesVar, 
							  int *numValsVarsNew, Probability *tempRes)
{	
	Probability		*newTable, *newFuncPtr, *oldFuncPtr;
	boolean			**currVarPlace, **currVarPlacePtr;
	variable		**varPtr, **varPtrEnd, *var, **variables = NULL, **newFuncVarPtr,
					**advanceVarPtr, **advanceVarPtrEnd, *advanceVar;
	unsigned long	tableSize = 1;
	int				numVars = 0;
	long			varInd;
	function		*funcPtr;

	/* 1.	Allocate memory for an auxiliary array. */

	myMalloc(currVarPlace, sizeof(boolean*) * func->numVariables);

	/* 2.	Calculate the new table size and the number of variables that 
			appear in it. Update the currVarPlace array (this array holds pointers to the 
			first entry, in the varValueUsageFlagsGen array, that corresponds to each
			of the variables in 'func'). */

	for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables,
			currVarPlacePtr = currVarPlace; varPtr < varPtrEnd; varPtr++, currVarPlacePtr++) {

		varInd = (*varPtr) - prod->variables;
	
		if (numValsVarsNew[varInd] > 1) {
			tableSize *= numValsVarsNew[varInd];
			numVars++;
		}
		
		*currVarPlacePtr = &varValueUsageFlagsGen[varInd * maxValuesVar];
	}

	/* 3.	Allocate memory for the new probability table. */

	myMalloc(newTable, sizeof(Probability) * tableSize);

	/* 4.	Allocate memory for the variable array of the new function, and update it. */

	if (numVars > 0) {
		myMalloc(variables, sizeof(variable*) * numVars);

		newFuncVarPtr = variables;

		for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables; 
				varPtr < varPtrEnd; varPtr++) {
		
			var = *varPtr;

			if (numValsVarsNew[var - prod->variables] > 1) {
				*newFuncVarPtr = var;
				newFuncVarPtr++;
			}
		}
	}

	/* 5.	Initialize the values of the function's variables to correspond 
			to the first entry in the probability table (i.e, to zeros). */

	initFunctionVars(func, varPtr, varPtrEnd);

	varPtrEnd = func->variables + func->numVariables;
	newFuncPtr = newTable;
	oldFuncPtr = func->probabList;

	/* 6.	Iterate over the old function's probability table. */

	while (1) {

		start_loop_rft:

		/* 6.1	Check for each variable if its current value is possible. */

		for (varPtr = func->variables, currVarPlacePtr = currVarPlace; 
				varPtr < varPtrEnd; varPtr++, currVarPlacePtr++) { 
			var = *varPtr;

			/* 6.1.1	If this variable's value is impossible, preceed to the next entry. */

			if ((*currVarPlacePtr)[var->currValue] == FALSE) {
				goto preceedToNextEntry;
			}
		}

		/* 6.2	Otherwise, copy this entry into the new function's probability table and 
				advance the pointer of the new function to the next entry. */

		*newFuncPtr = *oldFuncPtr;          
		newFuncPtr++;

		/* 6.3.	Advance the pointer of the old function to the next entry. */

		preceedToNextEntry:;
		oldFuncPtr++;		

		advanceFunctionVars(func, advanceVarPtr, advanceVarPtrEnd, advanceVar, start_loop_rft);
	} 

	/* 7.	Free the memory allocated dynamically. */

	myFree(currVarPlace);

	/* 8.	If the function didn't shrink into a constant, insert it into the 
			product and return a pointer to it. */

	if (numVars) {
		
		
		funcPtr = insertFunctionToProduct(prod, tableSize, newTable, func->logMultFactor, TRUE, 
									   numVars, variables, FALSE, func->rescaleTimer);
/* anna 28/12/o4 - for export bayes network */
		funcPtr->owner = func->owner;
		funcPtr->ownerName = func->ownerName;
		return funcPtr;
/* end change */
	}


	/* 9.	Otherwise, calculate its logProbability, free the memory allocated 
			to it, and return NULL. */

	*tempRes = log(newTable[0]) + func->logMultFactor;	

	myFree(newTable);

	return NULL;
}


/******************************************************************************
*
*       Function        reduceFunctionTableOnVar
*       Parameters		prod (the product of which the function is part of),
*						oldFunction (the function that needs to be reduced),
*						var (the variable of which values are being reduced),
*						valVar (the value of 'var' which is still possible),
*						currRes (if the function reduces to a costant, 'currRes'
*						will hold the value of that constant).
*       Output          none
*       Description		Creates the new function from the old one by eliminating 
*						all entries of the old table in which 'var' has a value
*						other than 'valVar'. Returns TRUE if the function has
*						shrunk into a constant (in this case, it returns the 
*						logProbability of the reduced funtion in 'currRes'.)
*						Otherwise, returns FALSE.
*       Comments		none
*
******************************************************************************/
boolean reduceFunctionTableOnVar(function *oldFunction, variable *var, 
								 int numVals, int valVar, Probability *currRes)
{
	variable		**varPtr, **varPtrEnd;
	Probability		*newTable, *newTablePtr, *oldTablePtr, *oldTablePtrEndBlock;
	long			sizeBlockCopy, sizeBlockSkip, tableSize;
	int				numTimesCopy;
	

	/* 1.	Calculate the size of the new function's probability table. */

	tableSize = oldFunction->sizeOfTable / numVals;

	/* 2.	Allocate memory for the new function's probability table. */

	myMalloc(newTable, sizeof(Probability) * tableSize);

	/* 3.	Calculate the size of the block that needs to be copied each time, the number
			of entries that need to be skipped between each 2 consecutive blocks that 
			are copied and the number of blocks that need to be copied. */

	sizeBlockCopy = 1; 
	numTimesCopy = 1;

	for (varPtr = oldFunction->variables + oldFunction->numVariables - 1; 
			*varPtr != var; varPtr--) {
		
		sizeBlockCopy *= (*varPtr)->numValues;
	}

	varPtr--;

	for (varPtrEnd = oldFunction->variables; varPtr >= varPtrEnd; varPtr--) {
		numTimesCopy *= (*varPtr)->numValues;
	}

	sizeBlockSkip = sizeBlockCopy * (numVals - 1);
	newTablePtr = newTable;
	oldTablePtr = oldFunction->probabList + sizeBlockCopy * valVar;

	/* 4.	Copy the relevant parts of the probability table. */

	while (numTimesCopy--) {
		for (oldTablePtrEndBlock = oldTablePtr + sizeBlockCopy; 
				oldTablePtr < oldTablePtrEndBlock; oldTablePtr++, newTablePtr++) {
			
			*newTablePtr = *oldTablePtr;
		}
		
		oldTablePtr += sizeBlockSkip;
	}


	/* 5.	If the function didn't shrink into a constant, update its fields. */

	if (oldFunction->numVariables > 1) {
		
		/* 5.1.	Update the sizeOfTable and the probabList fields. */

		oldFunction->sizeOfTable = tableSize;

		if (oldFunction->myProbabList) {
			myFree(oldFunction->probabList);
		}
		else {
			oldFunction->myProbabList = TRUE;
		}

		oldFunction->probabList = newTable;

		/* 5.2.	Remove 'var' from the variable array of the function. */

		removeVariableFromFunction(oldFunction, var);

		return FALSE;
	}
	

	/* 6.	Otherwise, calculate its logProbability, free the allocated memory,
			and return TRUE. */

	*currRes = log(newTable[0]) + oldFunction->logMultFactor;

	myFree(newTable);

	return TRUE;
}



/******************************************************************************
*
*       Function        zeroSeenVars
*       Parameters		numFuncs (the number of functions in 'funcArr'), 
*						funcArr (an array of pointers to functions).
*       Output          none
*       Description		Zeros the 'seen' flags of all the variables that appear
*						in one of the functions in 'funcArr'.
*       Comments		none
*
******************************************************************************/
void zeroSeenVars(int numFuncs, function **funcArr)
{
	function	**funcPtr, **funcPtrEnd, *func;
	variable	**varPtr, **varPtrEnd;

	/* 1.	Iterate over the functions in 'funcArr'. */

	for (funcPtr = funcArr, funcPtrEnd = funcPtr + numFuncs;
			funcPtr < funcPtrEnd; funcPtr++) {

		func = *funcPtr;

		/* 1.1	For each function, iterate over its variables and 
				update their 'seen' flags to FALSE. */

		for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
				varPtr < varPtrEnd; varPtr++) 
			(*varPtr)->alreadySeen = FALSE;
	}
}




/******************************************************************************
*
*       Function        getVarsInCombinedFunc
*       Parameters		elimVars (an array of pointers to the variables being 
*						eliminated), numElimVars (the number of variables being
*						eliminated), varArr (will hold pointers to the all the 
*						variables that appear in the functions in 'funcArr' and
*						aren't among the variables that are being eliminated),
*						funcArr (an array of pointers to the functions being
*						combined), numFuncs (the number of functions being 
*						combined), numVars (will hold the number of variables 
*						in the combined function), sizeOfTable (will hold the  
*						size of the probability table of the combined function).
*       Output          none 
*       Description		Inserts into 'varArr' all the variables that will appear in
*						the combined function after eliminating the variables in
*						'elimVars', updates 'numVars' to the number of variables		 
*						and 'sizeOfTable' to the multiplication of the numVals of
*						the variables.
*       Comments		none
*
******************************************************************************/
void getVarsInCombinedFunc(variable **elimVars, int numElimVars, variable **varArr, 
						   function **funcArr, int numFuncs, int *numVars, 
						   unsigned long *sizeOfTable)
{
	int				varCount;
	function		**funcPtr, **funcPtrEnd, *func;
	variable		**varPtr, **varPtrEnd, *var, **varArrPlace;
	unsigned long	tableSize;


	varCount = 0;
	varArrPlace = varArr;
	tableSize = 1;

	/* 1.	Zero the 'seen' flags of all variables that appear in the functions
			in 'funcArr', except for the variables in 'elimVars'. */

	zeroSeenVars(numFuncs, funcArr);

	for (varPtr = elimVars, varPtrEnd = varPtr + numElimVars; varPtr < varPtrEnd; 
			varPtr++) {
		(*varPtr)->alreadySeen = TRUE;
	}

	#if PRINT_MATRIX_MULT_STATISTICS
		if (elimVars != NULL) {
			fprintf(matrixFile, "\n\nEliminating %s: %d functions.\n==================", 
						(*elimVars)->originalNode->name, numFuncs);
		}
	#endif

	/* 2.	Iterate over the functions in 'funcArr'. */

	for (funcPtr = funcArr, funcPtrEnd = funcPtr + numFuncs; funcPtr < funcPtrEnd; funcPtr++) {

		func = *funcPtr;
		
		#if PRINT_MATRIX_MULT_STATISTICS
			fprintf(matrixFile, "\nFunction %d (%d variables):", funcPtr - funcArr, func->numVariables);
		#endif

		/* 2.1	For each function, iterate over its variables. */

		for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			var = *varPtr;

			#if PRINT_MATRIX_MULT_STATISTICS
				fprintf(matrixFile, "%s (%d), ", var->originalNode->name, var->numValues);
			#endif

			/* 2.1.1	If we haven't seen this variable yet, */

			if (var->alreadySeen == FALSE) {

				/* 2.1.1.1	Update that we've seen it. */

				var->alreadySeen = TRUE;

				/* 2.1.1.2	Insert it into 'varArr', and update the variable count 
							and the calculation of the table size. */

				*varArrPlace++ = var;
				varCount++;
				tableSize *= var->numValues;
			}
		}
	}

	*numVars = varCount;
	*sizeOfTable = tableSize;
}




/******************************************************************************
*
*       Function        getVarsInCombinedFuncJoint
*       Parameters		elimVar (a pointer to the first variable being 
*						eliminated), varArr (will hold pointers to the all the 
*						variables that appear in the functions in 'funcArr' and
*						aren't among the variables that are being eliminated),
*						numVars (will hold the number of variables in the combined
*						function), sizeOfTable (will hold the size of the 
*						probability table of the combined function),
*						logMultFactor (the 'logMultFactor' of the combined 
*						function), rescaleTimer (the required value of 
*						'rescaleTimer' for the combined function).
*       Output          none 
*       Description		Inserts into 'varArr' all the variables that will appear in
*						the combined function after eliminating the variables in
*						'elimVarsArr', updates 'numVars' with the number of 
*						variables and 'sizeOfTable' with the multiplication 
*						of the numVals of these variables. Also updates 'logMultFactor'
*						and 'rescaleTimer' to the required values for the combined
*						function.
*       Comments		Used when several variable need to be treated jointly 
*						during the computations.
*
******************************************************************************/
void getVarsInCombinedFuncJoint(variable **elimVarsArr, int numElimVars, 
								function **funcArr, int *numFuncs, variable **varArr, 
								int *numVars, unsigned long *sizeOfTable, 
								double *logMultFactor, int *rescaleTimer)
{
	int				varCount, funcCount;
	function		**funcPtr, **funcPtrEnd, *func, **funcArrPtr;
	variable		**varPtr, **varPtrEnd, *var, **varArrPlace;
	unsigned long	tableSize;

	/* 1.	Initializations. */

	varCount = 0;
	varArrPlace = varArr;
	tableSize = 1;
	funcArrPtr = funcArr;
	funcCount = 0;

	/* 2.	Update an array with pointers to all the functions that need to be combined. */

	for (varPtr = elimVarsArr, varPtrEnd = varPtr + numElimVars; varPtr < varPtrEnd; 
			varPtr++) {
		var = (*varPtr);

		for (funcPtr = var->functions, funcPtrEnd = funcPtr + var->currNumFunctions;
				funcPtr < funcPtrEnd; funcPtr++) {	

			func = (*funcPtr);

			if (func->alreadySeen == FALSE) {
				*funcArrPtr++ = func;
				funcCount++;
				func->alreadySeen = TRUE;
			}
		}
	}

	*numFuncs = funcCount;

	/* 3.	Zero the 'seen' flags of all of the variables that appear in the functions
			which need to be combined, except for the variables that need to be eliminated. */

	zeroSeenVars(funcCount, funcArr);

	for (varPtr = elimVarsArr, varPtrEnd = varPtr + numElimVars; varPtr < varPtrEnd; 
			varPtr++) {
		(*varPtr)->alreadySeen = TRUE;
	}

	*logMultFactor = 0.0;
	*rescaleTimer = MAX_RESCALE_UNITS;

	/* 4.	Iterate over the functions in 'funcArr'. */

	for (funcPtr = funcArr, funcPtrEnd = funcPtr + funcCount; funcPtr < funcPtrEnd; funcPtr++) {

		func = *funcPtr;

		/* 4.1.	Update the logMultFactor and the rescaleTimer of the combined function. */

		*logMultFactor += func->logMultFactor;

		if (func->rescaleTimer < *rescaleTimer)
			*rescaleTimer = func->rescaleTimer;
		
		/* 4.2.	For each function, iterate over its variables. */

		for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			var = *varPtr;

			/* 4.2.1.	If we haven't seen this variable yet */

			if (var->alreadySeen == FALSE) {

				/* 4.2.1.1.	Update that we've seen it. */

				var->alreadySeen = TRUE;

				/* 4.2.1.2.	Insert it into 'varArr', and update the variable count 
							and the table-size calculation. */

				*varArrPlace++ = var;
				varCount++;
				tableSize *= var->numValues;
			}
		}
	}

	*numVars = varCount;
	*sizeOfTable = tableSize;
}



/******************************************************************************
*
*       Function        getSizeOfCombinedFunc
*       Parameters		elimVar (pointer to the variable being eliminated).
*       Output          none 
*       Description		Calculates the size of the probability table that would
*						be created as a result of eliminating 'elimVar', and 
*						returns it.
*       Comments		none
*
******************************************************************************/
long double getSizeOfCombinedFunc(variable *elimVar)

{
	function		**funcPtr, **funcPtrEnd, *func;
	variable		**varPtr, **varPtrEnd, *var;
	long double		tableSize;


	tableSize = 1;

	/* 1.	Zero the 'seen' flags of all of elimVar's neighbors. */

	zeroSeenVars(elimVar->currNumFunctions, elimVar->functions);
	elimVar->alreadySeen = TRUE;

	/* 2.	Iterate over the functions that 'elimVar' appears in. */

	for (funcPtr = elimVar->functions, funcPtrEnd = funcPtr + elimVar->currNumFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {

		func = *funcPtr;
		
		/* 2.1.	For each function, iterate over its variables. */

		for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			var = *varPtr;

			/* 2.1.1.	If we haven't seen this variable yet */

			if (var->alreadySeen == FALSE) {

				/* 2.1.1.1.	Update that we've seen it. */

				var->alreadySeen = TRUE;

				/* 2.1.1.2.	Update the tableSize calculation. */

				tableSize *= var->numValues;
			}
		}
	}

	/* 3.	Return the size of the probability table. */

	return tableSize;
}




/******************************************************************************
*
*       Function        initJumpFactor
*       Parameters		func (the function whose jumpFactor is being initialized), 
*						elimVar (pointer to the variable for which the jump 
*						factor is being calculated).
*       Output          none
*       Description		Calculates the necessary increment in the index of func's
*						probability table as a result of an increment of 1 in 
*						elimVar's value.
*       Comments		none
*
******************************************************************************/
void initJumpFactor(function *func, variable *elimVar)
{
	variable	**varPtr, *var;

	/* 1.	Initialize 'varPtr' to point to the last variable in the function's
			variable array. */

	varPtr = func->variables + func->numVariables - 1;
	var = *varPtr;

	/* 2.	Initialize the jump factor to 1. */

	(func->jumpFactor).inc = 1;

	/* 3. For each variable which follows 'elimVar' in the function, multiply 
		  the jump factor by the number of possible values of this variable. */

	while (var != elimVar) {
		(func->jumpFactor).inc  *= var->numValues;
		
		varPtr--;
		var = *varPtr;
	}
}



/******************************************************************************
*
*       Function        initJumpFactors
*       Parameters		func (pointer to the function of which jump factors are
*						being calculated), elimVar (the variable who is being 
*						eliminated), incDecPtr (an array that will store the
*						jump factors).
*       Output          none
*       Description		Calculates for each variable of the function:
*						1) the necessary increment in the index of the 
*						   function's probability table as a result of an 
*						   increment of 1 in its value.
*						2) the necessary decrement in the index of the 
*						   function's probability table as a result of a 
*						   decrement of (numVals -1) in its value.
*       Comments		none
*
******************************************************************************/
void initJumpFactors(function *func, variable *elimVar, incDec *incDecPtr)
{
	variable	**varPtr, **varPtrEnd, *var;
	long		jumpFactor;

	/* 1.	Initialize the function's jump factor. */

	jumpFactor = 1;

	/* 2.	Iterate over the functions variables. */

	for (varPtr = func->variables + func->numVariables - 1, varPtrEnd = func->variables;
			varPtr >= varPtrEnd; varPtr--) {

		var = *varPtr;

		/* 2.1	The necessary increment/decrement due to a change in the value 
				of 'elimVar' are stored in the 'jumpFactor' field of the function. 
				The necessary increment/decrement due to a change in the value of
				some other variable are stored in the incDecPtr array. */

		if (var == elimVar) {
			(func->jumpFactor).inc = jumpFactor;
			(func->jumpFactor).dec = jumpFactor * (1 - var->numValues);
		}
		else {
			incDecPtr[var->currValue].inc = jumpFactor;
			incDecPtr[var->currValue].dec = jumpFactor * (1 - var->numValues);
		}

		/* 2.2	Multiply the jump factor by the number of values of the variable. */

		jumpFactor *= var->numValues;
	}
}



/******************************************************************************
*
*       Function        initJumpFactorJoint
*       Parameters		funcArr (array of pointers to functions whose  
*						jump-factors need to be calculated), numFuncs (the number
*						of functions in the array), elimVars (pointers to the 
*						variables for which the jump-factors are being 
*						calculated), numElimVars (the number of variables in the
*						array 'elimVars'), validAssignmentsArr (the array which
*						holds the valid assignments for the variables in the
*						array 'elimVars').
*       Output          none
*       Description		Calculates the necessary increment in the index of each
*						of the functions for each valid assignment to the 
*						variables in the 'elimVars' array.
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void initJumpFactorJoint(function **funcArr, int numFuncs, variable **elimVars, 
						 int numElimVars, validAssignments *validAssignmentsArr)
{
	assignment	*assignmentPtr, *assignmentPtrEnd;
	variable	**varPtr, **varPtrEnd;
	function	**funcPtr, **funcPtrEnd, *func;
	int			*incPtr, *val;


	/* 1.	Iterate over the valid joint assignments to the variables in the array
			'elimVars'. */

	for (assignmentPtr = validAssignmentsArr->assignmentArr, 
			assignmentPtrEnd = assignmentPtr + validAssignmentsArr->numAssignments;
			assignmentPtr < assignmentPtrEnd; assignmentPtr++) {

		/* 1.1.	Update the values of the eliminated variables according to the current
				assignment. */

		setAssignment(elimVars, numElimVars, assignmentPtr->varVals, val, varPtr, varPtrEnd);

		/* 1.2.	For each one, iterate over the functions that need to be combined. */

		for (funcPtr = funcArr, funcPtrEnd = funcPtr + numFuncs, 
				incPtr = assignmentPtr->inc; funcPtr < funcPtrEnd; funcPtr++, incPtr++) {	

			/* 1.2.1.	Calculate the increment in the index of the function which 
						corresponds to the current assignment relative the previous one,
						and store it. */

			func = *funcPtr;
			calcBaseIndex(func);

			(*incPtr) = func->baseInd;
		}
	}
}
#endif


/******************************************************************************
*
*       Function        initJumpFactorsJoint
*       Parameters		funcArr (array of pointers to functions whose  
*						jump-factors need to be calculated), numFuncs (the number
*						of functions in the array), elimVars (pointers to the 
*						variables for which the jump-factors are being 
*						calculated), numElimVars (the number of variables in the
*						array 'elimVars'), incDecArr (an array that will store 
*						the jump-factors), combinedVars (an array of all the 
*						variables that will appear in the combined function), 
*						numCombinedVars (the number of variables in the combined
*						function), validAssignmentsArr (the array which holds
*						the valid assignments for the variables in the array
*						'elimVars').
*       Output          none
*       Description		Calculates the necessary increment in the index of each
*						of the functions for each valid assignment to the 
*						variables in the 'elimVars' array. 
*						Calculates for each variable of the combined function:
*						1) the necessary increment in the index of the 
*						   functions' probability tables as a result of an 
*						   increment of 1 in its value.
*						2) the necessary decrement in the index of the 
*						   functions' probability tables as a result of a 
*						   decrement of (numVals -1) in its value.
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void initJumpFactorsJoint(function **funcArr, int numFuncs, variable **elimVars, 
						  int numElimVars, incDec *incDecArr, variable **combinedVars,
						  int numCombinedVars, validAssignments *validAssignmentsArr)
{
	variable	**varPtr, **varPtrEnd, *var;
	long		jumpFactor;
	function	**funcPtr, **funcPtrEnd, *func;
	incDec		*incDecPtr;


	/* 1.	Iterate over the functions in the array. */

	for (funcPtr = funcArr, funcPtrEnd = funcPtr + numFuncs, incDecPtr = incDecArr; 
			funcPtr < funcPtrEnd; funcPtr++, incDecPtr += numCombinedVars) {

		func = (*funcPtr); 

		/* 1.1.	Initialize the function's jump factor. */

		jumpFactor = 1;

		/* 1.2.	Iterate over the functions variables. */

		for (varPtr = func->variables + func->numVariables - 1, varPtrEnd = func->variables;
				varPtr >= varPtrEnd; varPtr--) {

			var = *varPtr;

			/* 1.2.1	The necessary increment/decrement due to a change in the value of
						any variable that isn't eliminated are stored in the incDecPtr array. */

			if (var->currValue >= 0) {
				incDecPtr[var->currValue].inc = jumpFactor;
				incDecPtr[var->currValue].dec = jumpFactor * (1 - var->numValues);
			}

			/* 1.2	Multiply the jump factor by the number of values of the variable. */

			jumpFactor *= var->numValues;
		}
	}

	/* 2.	Zero the value of all the variables in the 'combinedVars' array. */

	for (varPtr = combinedVars, varPtrEnd = varPtr + numCombinedVars; varPtr < varPtrEnd;
			varPtr++) {	
		(*varPtr)->currValue = 0;
	}

	/* 2.	Calculate the necessary increments in the index of each of the functions 
			for each valid assignment to the variables in the 'elimVars' array. */

	initJumpFactorJoint(funcArr, numFuncs, elimVars, numElimVars, validAssignmentsArr);
}
#endif




/******************************************************************************
*
*       Function        calcBaseIndex
*       Parameters		func (the function whose base index needs to be calculated).
*       Output          none
*       Description		Calculates the index of the entry in the probability   
*						table that corresponds to the current values of the 
*						function's variables.					 
*       Comments		none
*
******************************************************************************/
void calcBaseIndex(function *func)
{
	variable		**varPtr, **varPtrEnd;
	unsigned long	multFactor;

	/* 1.	Initialize the index and the multFactor. */

	func->baseInd = 0;
	multFactor = func->sizeOfTable;

	/* 2.	Iterate over the function's variables and add the contribution of each 
			variable to the index. */

	for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
			varPtr < varPtrEnd; varPtr++) {

		multFactor /= (*varPtr)->numValues;
		func->baseInd += (multFactor * (*varPtr)->currValue);
	}
}




/******************************************************************************
*
*       Function        calcAccOffsetElim
*       Parameters		funcX, funcY, elimVar
*       Output          none
*       Description		Calculates the accumulated offset of the set of variables
*						of the function 'funcX' induced by the set of variables 
*						of function 'funcY' appended with the variable 'elimVar' at
*						the end.
*       Comments		the idea is taken from a paper "Operating with potentials
*						of discrete variables", by M. Arias and F.J. Diez (not 
*						yet published).
*
******************************************************************************/
void calcAccOffsetElim(function *funcX, function *funcY, variable *elimVar)
{
	int				numVars, *orderingPtr, *orderingPtrEnd, indX, numVals;
	long			*offset, *offsetPtr, currOffset, *accOffsetPtr;
	variable		**varPtr, **varPtrEnd;


	numVars = funcY->numVariables + 1;

	/* 1.	Allocate memory for the arrays that would store the relative order of
			funcY compared to funcX, and the accumulated offset for an increase of 1
			in one of funcY's variables. */

	myMalloc(funcX->ordering, sizeof(int) * numVars);
	myMalloc(funcX->accOffset, sizeof(unsigned long) * numVars);
	myMalloc(offset, sizeof(unsigned long) * funcX->numVariables);

	/* 2.	Initialize the entries of the ordering array with -1. */

	for (orderingPtr = funcX->ordering, orderingPtrEnd = orderingPtr + numVars;
			orderingPtr < orderingPtrEnd; orderingPtr++) {
		*orderingPtr = -1;
	}

	/* 3.	Update the ordering array regarding all the variables of funcY that
			appear in funcX as well, and initialize an auxiliary array of offsets. */

	for (varPtr = funcX->variables, varPtrEnd = varPtr + funcX->numVariables, 
			indX = 0, currOffset = funcX->sizeOfTable, offsetPtr = offset; 
			varPtr < varPtrEnd; varPtr++, indX++, offsetPtr++) {
		
		funcX->ordering[(*varPtr)->currValue] = indX;

		currOffset /= (*varPtr)->numValues;
		*offsetPtr = currOffset;
	}

	/* 4.	Calculate the accumulated offset. */

	orderingPtr = &(funcX->ordering[numVars-1]);
	accOffsetPtr = funcX->accOffset + numVars - 1;
	*accOffsetPtr = getOffset(offset, (*orderingPtr));
	accOffsetPtr--;
	numVals = elimVar->numValues;

	for (varPtrEnd = funcY->variables, varPtr = varPtrEnd + funcY->numVariables - 1; 
			varPtr >= varPtrEnd; varPtr--, accOffsetPtr--, orderingPtr--) {

		*accOffsetPtr = getOffset(offset, *(orderingPtr-1)) - 
						(numVals * getOffset(offset, (*orderingPtr))) +
						*(accOffsetPtr+1);

		numVals = (*varPtr)->numValues;
	}

	/* 5.	Free the memory allocated dynamically for the offset array. */

	myFree(offset);
}


/******************************************************************************
*
*       Function        calcAccOffset
*       Parameters		funcX, funcY.
*       Output          none
*       Description		Calculates the accumulated offset of the set of variables
*						of the function 'funcX' induced by the set of variables 
*						of function 'funcY'.
*       Comments		the idea is taken from a paper "Operating with potentials
*						of discrete variables", by M. Arias and F.J. Diez (not 
*						yet published).
*
******************************************************************************/
void calcAccOffset(function *funcX, function *funcY)
{
	int				numVars, *orderingPtr, *orderingPtrEnd, indX;
	long			*offset, *offsetPtr, currOffset, *accOffsetPtr;
	variable		**varPtr, **varPtrEnd;


	numVars = funcY->numVariables;

	/* 1.	Allocate memory for the arrays that would store the relative order of
			funcY compared to funcX, and the accumulated offset for an increase of 1
			in one of funcY's variables. */

	myMalloc(funcX->ordering, sizeof(int) * numVars);
	myMalloc(funcX->accOffset, sizeof(unsigned long) * numVars);
	myMalloc(offset, sizeof(unsigned long) * funcX->numVariables);

	/* 2.	Initialize the entries of the ordering array with -1. */

	for (orderingPtr = funcX->ordering, orderingPtrEnd = orderingPtr + numVars;
			orderingPtr < orderingPtrEnd; orderingPtr++) {
		*orderingPtr = -1;
	}

	/* 3.	Update the ordering array regarding all the variables of funcY that
			appear in funcX as well, and initialize an auxiliary array of offsets. */

	for (varPtr = funcX->variables, varPtrEnd = varPtr + funcX->numVariables, 
			indX = 0, currOffset = funcX->sizeOfTable, offsetPtr = offset; 
			varPtr < varPtrEnd; varPtr++, indX++, offsetPtr++) {
		
		funcX->ordering[(*varPtr)->currValue] = indX;

		currOffset /= (*varPtr)->numValues;
		*offsetPtr = currOffset;
	}

	/* 4.	Calculate the accumulated offset. */

	orderingPtr = &(funcX->ordering[numVars-1]);
	accOffsetPtr = funcX->accOffset + numVars - 1;
	*accOffsetPtr = getOffset(offset, (*orderingPtr));
	accOffsetPtr--;

	for (varPtrEnd = funcY->variables, varPtr = varPtrEnd + funcY->numVariables - 1;
			varPtr > varPtrEnd; varPtr--, accOffsetPtr--, orderingPtr--) {

		*accOffsetPtr = getOffset(offset, *(orderingPtr-1)) - 
						((*varPtr)->numValues * getOffset(offset, (*orderingPtr))) +
						*(accOffsetPtr+1);
	}

	/* 5.	Free the memory allocated dynamically for the offset array. */

	myFree(offset);
}


/******************************************************************************
*
*       Function        scaleFunctionProbList
*       Parameters		func (the function whose probability table needs to be 
*						scaled).
*       Output          none
*       Description		Scales the function's probability table according to 
*						its highest entry. 					 
*       Comments		none
*
******************************************************************************/
boolean scaleFunctionProbList(function *func)
{
	Probability		*probPtr, *probPtrEnd;
	double			maxVal;


	maxVal = MIN_DBL;

	/* 1.	Find the maximum value in the probability table. */

	for (probPtr = func->probabList, probPtrEnd = probPtr + func->sizeOfTable;
			probPtr < probPtrEnd; probPtr++) {

		if ((*probPtr) > maxVal)
			maxVal = (*probPtr);
	}

	if (maxVal == 0.0)
		return TRUE;

	/* 2.	Rescale each value in the probability table according to 'maxVal'. */

	for (probPtr = func->probabList, probPtrEnd = probPtr + func->sizeOfTable;
			probPtr < probPtrEnd; probPtr++) {

		(*probPtr) /= maxVal;
	}

	/* 3.	Adjust the 'logMultFactor' of the function accordingly. */

	func->logMultFactor += log(maxVal);

	return FALSE;
}


/******************************************************************************
*
*       Function        printFunction
*       Parameters		func (the function who's being printed). 
*						scaled).
*       Output          none
*       Description		Prints the function. 					 
*       Comments		Used for debugging purposes.
*
******************************************************************************/
#if DEBUG_STAT
void printFunction(function *func)
{
	int count = 0;
	int var;
	variable **varPtr, **varPtrEnd, **advanceVarPtr, **advanceVarPtrEnd, *advanceVar;

	if (func == NULL)
		return;

	/* 1.	Opens a file for printing. */

	outFile = fopen("out.txt", "a");

	/* 2.	Prints the number of variables of the function. */

	fprintf(outFile,"function has %d variables:", func->numVariables);

	/* 3.	Prints the names of the variables. */

	for (var = 0; var < func->numVariables; var++)
		fprintf(outFile, "%s  ", func->variables[var]->originalNode->name);

	/* 4.	Prints the logMultFactor. */

	fprintf(outFile,"%lf, %lf\n", func->logMultFactor, exp(func->logMultFactor));

	/* 5.	Prints the non-zero entries of the probability function. */

	fprintf(outFile,"\n\nProbabilities (%u):\n", func->sizeOfTable);

	initFunctionVars(func, varPtr, varPtrEnd);

	while (1) {
		start_loop_pf:

		if ((func->probabList)[count] != 0.0) {
			fprintf(outFile,"%d.  ", count);  
			
			for (var = 0; var < func->numVariables; var++) 
				fprintf(outFile,"%d  ", func->variables[var]->currValue); 

			fprintf(outFile,"[%lf, %lf] \n", (func->probabList)[count], log((func->probabList)[count]) + func->logMultFactor);
		}
		count++;

		advanceFunctionVars(func, advanceVarPtr, advanceVarPtrEnd, advanceVar, start_loop_pf);
	}

	fprintf(outFile,"\n\n");

	/* 6.	Closes the output file. */

	fclose(outFile);
}
#endif





