/* $Id: graph.h,v 1.2 2008/08/03 08:15:47 superlink Exp $ */
/////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////            Graph.h           /////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/*                 Contains the graph DB and definitions of graph functions.               */
/////////////////////////////////////////////////////////////////////////////////////////////

#ifndef GRAPH_H
#define GRAPH_H

#include "elim_auxiliary.h"
#include "../../Auxiliary.h"

//////////////////////////////////////////////////////////////////////////////////////////////
/*     type AdjMatrix is designated for 2-dimentional bit array, which contains information */
/*     about graph (adjacency matrix). (see "struct Graph" specifications)                  */
//////////////////////////////////////////////////////////////////////////////////////////////

struct AdjMatrix_t{
	unsigned char b1 : 1, b2 : 1, b3 : 1, b4 : 1, b5 : 1, b6 : 1, b7 : 1, b8 : 1;
};
typedef struct AdjMatrix_t **AdjMatrix;

//////////////////////////////////////////////////////////////////////////////////////////////
/*     struct Edge_t contains information about the neighbour of the given vertex and is    */
/*     used in the linked list of neighbours. (see "struct Vertex_t" specifications)        */
//////////////////////////////////////////////////////////////////////////////////////////////

struct Edge_t{
	int neighbor;
	struct Edge_t *next;
};
typedef struct Edge_t *Edge; 

//////////////////////////////////////////////////////////////////////////////////////////////
/*                    struct Vertex_t contains data about vertex in graph                   */
/*     fields:      edgesNum   - number of neighbours the given vertex has.                 */
/*                  neighbours - linked list of vertices that the given vertex is connected */
/*                               with by a single edge.                                     */
/*     comments:    is used as a cell in array of vertices in "struct Graph" type.          */
//////////////////////////////////////////////////////////////////////////////////////////////

struct Vertex_t{
	int edgesNum;
	Edge neighbors;
	boolean ifMarked;
	boolean ifRemoved;
    int weight;
	char name[100];
};
typedef struct Vertex_t *AdjList;

//////////////////////////////////////////////////////////////////////////////////////////////
/*             struct Graph contains graph data, i. e. its vertices and edges               */
/*     fields:      list - pointer to array of linked lists. Cell num. i in array contains  */
/*                         information about  vertex num. i in graph - linked list of       */
/*                         vertices it is connected with.                                   */
/*                  matrix - contains 2-dimentional bit array in the form of lower          */
/*                           triangular (adjacency matrix DB).                              */
/*                  numOfVertices - Number of vertices in the graph.                        */
/*                  currNumOfVertices - current number of vertices during elimination, that */
/*                                      are connected to at least one other vertex.         */
/*     comments:    specified DB enables to delete vertices from graph in O(n) and to       */
/*                  insert or to check whether vertices i, j are connected in O(1).         */
//////////////////////////////////////////////////////////////////////////////////////////////

struct Graph_t{
	AdjList list;
	AdjMatrix matrix;
	int numOfVertices;
	int edgesNum;
	int numOfRemovedVertices;
	double prepr_cost;
};
typedef struct Graph_t *Graph;

//////////////////////////////////////////////////////////////////////////////////////////////
/*          CreateGraph - creates new graph from data received from the given file.         */
/*     parameters:   filename      - Name of file which contains the data about the graph.  */
/*                   gr            - Double pointer to the new graph.                       */
/*     return value: none.                                                                  */
/*     comments:     the given file should be in Koster format.                             */
//////////////////////////////////////////////////////////////////////////////////////////////

void CreateGraph(char *filename, Graph *gr);

//////////////////////////////////////////////////////////////////////////////////////////////
/*                            copyGraph - copies a given graph.                             */
/*     parameters:   gr - pointer to the graph to be copied.                                */
/*     return value: pointer to the new graph.                                              */                    
/*     comments:     none.                                                                  */
//////////////////////////////////////////////////////////////////////////////////////////////

Graph copyGraph(Graph oldGr);

//////////////////////////////////////////////////////////////////////////////////////////////
/*                        AddEdge - adds edge to the graph.                                 */
/*     parameters:   vertex1Num & vertex2Num - number of vertices which are to be connected */
/*                   oldGr      - pointer to graph.                                         */
/*     return value: FAILURE - if there is already  an edge between specified vertices.     */
/*                   SUCCESS - a new edge was added to the graph.                           */
/*     comments:     "vertex1Num" & "vertex2Num" specify numbers of cells in array of       */
/*                   vertices (see "struct Graph" specifications).                          */
//////////////////////////////////////////////////////////////////////////////////////////////

Result AddEdge(int vertex1Num, int vertex2Num, Graph gr);

//////////////////////////////////////////////////////////////////////////////////////////////
/*                        removeEdge - removes edge from the graph.                         */
/*     parameters:   vertex1Num & vertex2Num - number of vertices which are to be           */
/*                   disconnected.                                                          */
/*                   gr      - pointer to graph.                                            */
/*     return value: FAILURE - if there specified vertices are not connected.               */
/*                   SUCCESS - edge was removed from the graph.                             */
/*     comments:     given edge is removed from both adjacency matrix (0 instead of 1), and */
/*                   from array of vertices (from two linked lists).                        */
//////////////////////////////////////////////////////////////////////////////////////////////

Result removeEdge(int vertex1Num, int vertex2Num, Graph gr);

//////////////////////////////////////////////////////////////////////////////////////////////
/*                        removeVertex - removes vertex from the graph.                     */
/*     parameters:   vertexNum - number of vertex to be removed.                            */
/*                   gr      - pointer to graph.                                            */
/*     return value: FAILURE - if there specified vertices are not connected.               */
/*                   SUCCESS - edge was removed from the graph.                             */
/*     comments:     given vertex is removed from both adjacency matrix (all the line is    */
/*                   set to 0), and from array of vertices (from all linked lists) and his  */
/*                   own list of neighbours is destroyed and pointer set to NULL.           */
//////////////////////////////////////////////////////////////////////////////////////////////

void removeVertex(int vertexNum, Graph gr);

boolean ifConnected(Graph gr, int v1, int v2);

//////////////////////////////////////////////////////////////////////////////////////////////
/*               DestroyGraph - frees the memory allocated by the given graph.              */
/*     parameters:   gr      - pointer to graph.                                            */
/*     return value: none.                                                                  */ 
/*     comments:     none.                                                                  */
//////////////////////////////////////////////////////////////////////////////////////////////

void DestroyGraph(Graph gr);

#endif



