/* $Id: InferEngine.h,v 1.2 2008/08/03 08:15:47 superlink Exp $ */
/****************************************************************************/
/*  InferEngine.h:						   									*/
/*	Current class is used in order to make inference on the input product	*/
/*	which represents Bayes Net. Inference is made using 1 additional array:	*/
/*	1. "notElimVars" - array with specific order of the variables NOT to be	*/
/*	eliminated ("numNotElimVars" - size of the array)						*/
/*	The main method which perform the above inference						*/
/*	is "computePartialProb".												*/
/*	The output is ONE table with "notElimVars" variables in the same order	*/
/*	as they appear in "notElimVars" array									*/
/*																			*/
/*	Assumptions on the input product:										*/
/*	1. The structure of the product is defined in "Product.h"				*/
/*	2. The following fields of the product must be initiated before using	*/
/*	"CInferEngine" class: numVariables, numFunctions, variables				*/
/*	variables[i].originalNode (pNum, name) for each (0 <= i < numVariables),*/
/*	variables[i].numValues, functions, functions[j].variables,				*/
/*	functions[j].numVariables, functions[j].sizeOFTable,					*/
/*	functions[j].probabList for each (0 <= j < numFunctions).				*/
/*	3. The order of variables in "variables" array of the product restricts	*/
/*	the order of the functions in "functions" array							*/
/*	in the following mannner: for each variable, the function of this		*/
/*	variable given it's parents must appear under the same index in			*/
/*	"functions" array as the variable (for example, if "X_5" variable has	*/
/*	index 14 in "variables" array, then function "X_5 | Lf_5,Lm_5" must		*/
/*	appear under index 14 in "functions" array.								*/
/*	4. The order of variables in function: if the function is of type 		*/
/*	"variable V	given its parents Pi, for (1 <= i <= n)" then V must appear	*/
/*	under index 0 in function variables, and then all parents variables,	*/
/*	such that variable V is MSB and the last parent variable is LSB.		*/
/*	The order between parents doesn't matter.								*/
/****************************************************************************/



#ifndef INFERENGINE_IDO
#define INFERENGINE_IDO

#include<stdio.h>
#include<stdlib.h>
#include <iostream>
#include <vector>
#include <fstream>
#include <string>
using namespace std;
#include <string.h>
#include <time.h>

#ifdef  __cplusplus
extern "C" {
#endif

	#include "../Product.h"
	#include "ElimOrder/elimination.h"
	#include "ElimOrder/sumMaxCliques.h"
	#include "../EliminationOrder.h"
#ifdef  __cplusplus
}
#endif



namespace InferEngine_Ido
{
	#include "InferEng_BayesNet.h"
	#include "DiscreteFunc.h"
	#include "DiscreteVar.h"
	#include "Bucket.h"
	#include "BucketsEliminator.h"
	#include "InferEng_Settings.h"
	#include "ValHolder.h"

	/* the parsing files *
	#include "XMLBIFParserWin\Utils.h"
	#include "XMLBIFParserWin\Node.h"
	#include "XMLBIFParserWin\Product.h"
	*/

	// This class is exported from the InferEngine.dll
	class CInferEngine {
	public:
		CInferEngine(void);
		virtual ~CInferEngine();
		_prob*		computePartialProb(product* prod, int* notElimVars, int numNotElimVars, bool& problemOccured);
		
	protected:
		BayesNet*		bn;				/* the original Bayes Net */
		_prob			prob;			/* the probability of the XML file */

		void readOrderFromFile(int* varElimOrder);
		void		initializeDB(product* prod);
		int*		getVarEliminationOrder();
	};

	
	//*********************************************************************************88
	// Implementation
	
	CInferEngine::CInferEngine()
	{ 
		this->bn = NULL;//new BayesNet();
		this->prob = 0;
	}


	CInferEngine::~CInferEngine() {

		if (this->bn) {
			delete this->bn;
		}
	}





	void CInferEngine::initializeDB(product* prod) {
		
		// initialize the BayesNet 
		this->bn = new BayesNet();
		 
		
		this->bn->Init(prod);
		
	//	bn->printBN(BN_FILENAME);
	}





	// does partial inference and returns the vector of probabilities
	// for the not eliminated variables
	//__declspec(dllexport) 
	_prob* CInferEngine::computePartialProb(product* prod, int* notElimVars, int numNotElimVars, bool& problemOccured) {

		_prob* res;
		BucketEliminator* be;
		int place,v;
		int* elimOrder;

		initializeDB(prod);
		
		//	getting elimination order
		elimOrder = getVarEliminationOrder();

		ValHolder* vh = new ValHolder(bn->numVariables);

		be = new BucketEliminator(bn->numVariables, vh);

		for (v=0; v<bn->numVariables ;v++) {
			place=elimOrder[v];
			be->setExactPriority(v, place);	
		}
		for (int ne=0; ne<numNotElimVars ;ne++) {
			be->setExactPriority(notElimVars[ne], DONT_TOUCH_PRIORITY);
		}

		be->Init(bn);

		be->arrangeMatricesInBuckets();

		bool mode = INFERENCE_MODE;


		if (CONDITIONING == 1) {
			be->eliminateAllPossiblesCond(mode);
		}
		else {
			be->eliminateAllPossiblesFull(mode);
		}
					
		FuncMatrix* func = be->combineAllTo1Matrix();
		//	Perform the conversion of function probs from log-scale to probability scale
		
		func->convertFuncToProb();
		//	Check that all variables in the resulted "func" are in "notElimVars" array
		
		for(int i = 0; i < bn->getNumVars(); ++i) {
			bn->getVarAt(i)->setIndex(i);
		}
		
		if (func->markNotPresentVars(notElimVars, numNotElimVars)) {
			problemOccured = true;
			//	Sum over redundant variables (marked as "not present")
			func->sumOverSignedVars(false);
		}
		
		res = new _prob[func->numberOfEntries];
		//	instead of changing notElimVars change func so variables will be 
		//	according to notElimVars
		DiscreteVar** msbVars = new DiscreteVar*[numNotElimVars];
		for (int i=0; i<numNotElimVars; i++) {
			msbVars[i] = bn->variables[notElimVars[i]];
		}
		
		func->changeMSB2(msbVars, numNotElimVars);
		
		
		for (int ent=0; ent<func->numberOfEntries ;ent++) {
			res[ent] = func->probs[ent] + be->multFactor;
		}
		
		delete be;
		delete func;

		return res;
	}


	int* CInferEngine::getVarEliminationOrder() {

		DiscreteFunc* func;
		bool inFunc;

		/* create a graph in file GRAPH_FILE */
		ofstream outFile(GRAPH_FILE);	    
		outFile << "p edge " << bn->numVariables << endl;
	
		// debugging
		int minAlloc = 100;
		if (bn->numVariables > 100) {
			minAlloc = bn->numVariables;
		}
		bool* neighbs = new bool[minAlloc];
		bool* childs = new bool[minAlloc];
		int numChilds;
	    
	    
		/* initialize the neighboring array */
		for (int i=0; i<bn->numVariables ;i++) {
			neighbs[i] = false;
			childs[i] = false;
		}

		for (int v=0; v<bn->numVariables ;v++) {

			numChilds = 0;

			for (int j=0; j<bn->numVariables ;j++) {
				func = bn->functions[j];
				inFunc = false;

				// check if any of the members of func is v
				if (func->getVar()->index == v) {
					inFunc = true;
				}
				else {
					for (int p=0; p<func->numParents ;p++) {
						if (func->parents[p]->index == v) {
							inFunc = true;
							childs[func->getVar()->index] = true;
							++numChilds;
						}
					}
				}

				if (inFunc == true) {
					for (int p=0; p<func->numParents ;p++) {
						if (func->parents[p]->index != v) {
							neighbs[func->parents[p]->index] = true;
						}
					}
					if (func->var->index != v) {
						neighbs[func->var->index] = true;
					}
				}
			}
	        

			/* print to outFile the neighbor variables of v */
			for (int n=v+1; n<bn->numVariables ;n++) {
				if (neighbs[n] == true) {
					outFile << "e " << (v+1) << " " << (n+1) << endl;
					/* keep the array initialized to false for the next variable */
					neighbs[n] = false;
				}
			}
		}
		delete[] neighbs;
		delete[] childs;

		/* print to outFile the number of values of every variable */
		for (int v1=0; v1<bn->numVariables ;v1++) {
			outFile << "v " << (v1+1) << " " << 
			bn->variables[v1]->getNumVals() << endl;
		}
		//	operating algorithms for elimination order optimization
		//	cost mode : ALL_CLQ
		//	algorithms: SG, WSG, MF, WMF
		char it_time[15] = "it";	
		int *varElimOrder = new int[bn->numVariables];
		optimizationPolicyDef* optimizationPolicyArr = configureOptimizationPolicy();
		//	running SG algorithm to determine complexity (1 iteration)
		int complexity = (int)log10((double)elimOrder_main(SG, ALL_CLQ, 1, it_time));
		//	checking the need to perform additional algorithms iterations
		if (optimizationPolicyArr[complexity].maxNumIter > 0)
		{
			//	operate number of algorithms and choosing the order 
			//	with the best complexity
			int bestComplexity = MAX_ELIMINATION_COMPLEXITY + 1;
			int numIterations = optimizationPolicyArr[complexity].maxNumIter;
			for(int alg = SG; alg < NUM_ALGORITHMS; alg++)
			{
				int numIters = numIterations;
				switch (alg)
				{
					case SG:	numIters *=	PERCENT_MW;
								break;
					case WSG:	continue;
					case MF:	numIters *= PERCENT_MF;
								break;
					case WMF:	numIters *= PERCENT_WMF;
								break;
					default:	break;
				}
				if (numIters == 0)
				{
					numIters = 1;
				}
				complexity = (int)log10((double)elimOrder_main((ELIM_ALG)alg, ALL_CLQ, 
					numIters, it_time));
				if (complexity < bestComplexity)
				{
					readOrderFromFile(varElimOrder);
					bestComplexity = complexity;
				}	
			}
		}
		else
		{
			//	no need for additional algorithms iterations, return current order
			readOrderFromFile(varElimOrder);
		}    
		
		outFile.close();
		remove(GRAPH_FILE);
		remove(ELIM_FILE);	
		return varElimOrder;
	}



	void CInferEngine::readOrderFromFile(int* varElimOrder)
	{
		ifstream inFile(ELIM_FILE);
	    
		for (int pos=0, v2; pos<bn->numVariables ;pos++) {
			inFile >> v2;
			varElimOrder[v2-1] = bn->numVariables-pos;    
		}
		inFile.close();
	}
		
};

#endif



