/* $Id: IterateAff.c,v 1.2 2008/08/03 08:15:46 superlink Exp $ */
/*****************************************************************************/
/*  IterateAff.c															 */
/*  Author: Anna Reitman and Ma'ayan Fishelson. Some fixes by Mark Silberstein  */
/*****************************************************************************/

#include "IterateAff.h"

/******************************************************************************
*
*       Function        initPositionData
*       Parameters      none
*       Output          returns the number of iteration necessary.
*       Description		Initializes data needed for performing the scan of the
*						iterated disease loci over the marker map.
*       Comments		none
*
******************************************************************************/


double computeMarkerDistance(int marker1, int marker2, thetaCompTheta* map){
  int i=0;
  double totalDist=0;
  
  if (marker1==marker2) return 0.0;
  
  for (i=marker1-1;i<marker2-1;i++){
	
	if (map[i][0]>=0.499){
	  totalDist=UNLINKED_DIST;
	  break;
	}
	totalDist+=getRecTocM(map[i][0],progData.haldane);
  }
  
  return totalDist;
}



/******************************************************************************
*
*       Function        initPositionData
*       Parameters      none
*       Output          returns the number of iteration necessary.
*       Description		Initializes data needed for performing the scan of the
*						iterated disease loci over the marker map.
*       Comments		none
*
******************************************************************************/

int initPositionData()
{
	int				numIntervalsMap, numIter, i, numMarkers, numIterLoci;

	numIterLoci = ((progData.progSpecific).sGHData).numIterLoci;

	/* 1.	Calculate the number of stationary loci and the number of intervals 
			in the marker map (includes all the stationary loci). */

	numMarkers = loci.numLoci - numIterLoci;
	numIntervalsMap = numMarkers - 1;

	/* 2.	Allocate memory and save the original recombination values 
			(the ones specified in the input file), between the stationary loci.*/

	myMalloc(thetaMarkers, sizeof(thetaCompTheta) * numIntervalsMap);
	memcpy(thetaMarkers, recombData.maleThetas + numIterLoci, 
				sizeof(thetaCompTheta) * numIntervalsMap);

	/* 3.	Initialize the number of iterations to 1. */

	numIter = 1;

	/* 4.	For each iterated disease locus, update the necessary information 
			on all its positions in the scan, and update the requested number
			of iterations. */

	for (i = 0; i < numIterLoci; i++) {
		numIter *= initPositionDataLocus(i, numMarkers);
	}

	/* 5.	Return the number of iterations necessary. */
	
	return numIter;
}


/******************************************************************************
*
*       Function        initPositionDataLocus
*       Parameters      ind (the index of the iterated disease locus, either 0
*						or 1), numMarkers (the number of stationary loci in the
*						marker map), distFromMarker1Arr (an array which would be 
*						updated with the distances in Morgans of each marker 
*						from the maker delimiting the scan interval on the left).
*       Output          returns the number of requested positions along the scan
*						interval of iterated disease locus no. 'ind'.
*       Description		Calculates the number of requested positions for the 
*						iterated disease locus no. 'ind' and finds all the 
*						positions. Also, translates the step size and the distance
*						off the ends of the map to Morgans. 
*       Comments		none
*
******************************************************************************/
int initPositionDataLocus(int ind, int numMarkers)
{
	sGHDataLocus		*sGHDataLocusPtr;
	int					numIntervals;
	double				totalDist;
	scanIntervalData	*scanDataPtr;

	/* 1.	Update a pointer to the entry in the data-structure which corresponds
			to this iterated locus. */

	sGHDataLocusPtr = &(((progData.progSpecific).sGHData).sGHDataLocusArr[ind]);
	scanDataPtr = &(progData.scanData[ind]);

	/* 2.	Update an auxiliary array with the distance of each marker in the scan 
			interval from the marker preceeding it and from the leftmost marker. */

	updateDistArrs(scanDataPtr, thetaMarkers);

	/* 3.	Calculate the requested number of positions for the disease locus. */

	if (sGHDataLocusPtr->isStepIter) {

		/* 3.1.	Calculate the total size of the scan interval (in cM). */

		totalDist = scanDataPtr->offEndDistLeft + scanDataPtr->offEndDistRight +
						scanDataPtr->distFromMarker1[scanDataPtr->numIntervals];

		sGHDataLocusPtr->numPos = (int)floor(totalDist / sGHDataLocusPtr->stepIter) + 1; 
	}
	else {
		numIntervals = scanDataPtr->numIntervals +
					   (scanDataPtr->offEndDistLeft != 0) + (scanDataPtr->offEndDistRight != 0);

		sGHDataLocusPtr->numPos = numIntervals * sGHDataLocusPtr->numIterInterval + 1;
	}

	/* 4.	Allocate memory for an array that would store information on all the
			requested positions of the iterated locus along the genetic map. */
			
	myMalloc(sGHDataLocusPtr->locusPosArr, sizeof(locusPos) * sGHDataLocusPtr->numPos);

	/* 5.	Update the information on all the requested positions of the iterated 
			locus along the gentic map. */

	if (sGHDataLocusPtr->isStepIter)
		findRequestedPosStepIter(sGHDataLocusPtr, scanDataPtr);
	else
		findRequestedPosNumIter(sGHDataLocusPtr, scanDataPtr);

	/* 6.	Return the number of requested positions for the disease locus. */

	return sGHDataLocusPtr->numPos;
}


/******************************************************************************
*
*       Function        findRequestedPosStepIter
*       Parameters      sGHDataLocusPtr (pointer to the entry in the array
*						that stores the information on one of the iterated 
*						disease loci), recValuesMorgan (an array that stores 
*						the distances in cM of each marker in the scan 
*						interval from the leftmost marker in the scan interval).
*       Output          none
*       Description		Finds all the requested positions of the iterated disease 
*						locus in the marker map.
*       Comments		none
*
******************************************************************************/
void findRequestedPosStepIter(sGHDataLocus *sGHDataLocusPtr, 
							  scanIntervalData *scanDataPtr)
{
	double		currPos, distFromMarker, *distFromMarker1Arr, *distFromMarker1ArrPtr;
	int			currLocusIndMap;
	locusPos	*locusPosPtr, *locusPosPtrEnd;

		 
	locusPosPtr = sGHDataLocusPtr->locusPosArr;
	locusPosPtrEnd = locusPosPtr + sGHDataLocusPtr->numPos;

	distFromMarker1Arr = scanDataPtr->distFromMarker1;
	distFromMarker1ArrPtr = distFromMarker1Arr + 1;

	/* 1.	Update the information on the first position of the iterated locus. */

	if (scanDataPtr->offEndDistLeft > 0) {

		/* 1.1.	If the scan starts off the map. */

		currLocusIndMap = scanDataPtr->startPos - 1;
		locusPosPtr->locusIndMap = currLocusIndMap;
		currPos = -scanDataPtr->offEndDistLeft;
		locusPosPtr->rec = getcMToRec(-currPos, progData.haldane);
		locusPosPtr->distFromMarker1 = currPos;
	}
	else {
		/* 1.1.	If the scan starts inside the map. */

		currLocusIndMap = scanDataPtr->startPos;
		locusPosPtr->locusIndMap = currLocusIndMap;
		currPos = 0;
		locusPosPtr->rec = 0;
		locusPosPtr->distFromMarker1 = 0;
	}

	/* 2.	Update the information on the rest of the positions of the iterated locus. */

	for (locusPosPtr++; locusPosPtr < locusPosPtrEnd; locusPosPtr++) {
		
		if (currPos < 0) {
			
			/* 2.1.	The iterated locus is still outside the map, on the left side.
					Advance its position according to the step-size. */	
			
			currPos += sGHDataLocusPtr->stepIter;

			/* 2.2.	If the iterated locus has passed to the following interval (the
					first interval in the map), advance its index. Otherwise, update
					the recombination fraction between the iterated locus and the marker
					following it. */

			if (currPos >= 0) {
				currLocusIndMap++;
			}
			else {
				locusPosPtr->rec = getcMToRec(-currPos, progData.haldane);
				locusPosPtr->locusIndMap = currLocusIndMap;
				locusPosPtr->distFromMarker1 = currPos;
				continue;
			}
		}
		else {

			/* 2.1.	The iterated locus is inside the map.
					Advance its position according to the step-size. */	

			currPos += sGHDataLocusPtr->stepIter;
		}

		/* 2.3.	Find the interval in the genetic map which the iterated locus reached. */

		while ((currLocusIndMap < scanDataPtr->endPos) && 
				(currPos >= *distFromMarker1ArrPtr)) {
			currLocusIndMap++;
			distFromMarker1ArrPtr++;
		}

		/* 2.4.	Update the fields of the current entry in the array with the index
				in the map of the iterated locus and the recombination fraction between
				it and the marker preceeding it. */

		locusPosPtr->locusIndMap = currLocusIndMap;
		distFromMarker = currPos - *(distFromMarker1ArrPtr - 1);
		locusPosPtr->rec = getcMToRec(distFromMarker, progData.haldane);
		locusPosPtr->distFromMarker1 = currPos;
	}
}


/******************************************************************************
*
*       Function        findRequestedPosNumIter
*       Parameters      sGHDataLocusPtr (pointer to the entry in the array
*						that stores the information on one of the iterated 
*						disease loci), recValuesMorgan (an array that stores 
*						the distances in Morgans of each marker in the scan 
*						interval from the leftmost marker in the scan interval).
*       Output          none
*       Description		Finds all the requested positions of the iterated disease 
*						locus in the marker map.
*       Comments		none
*
******************************************************************************/
void findRequestedPosNumIter(sGHDataLocus *sGHDataLocusPtr, 
							 scanIntervalData *scanDataPtr)
{
	double		distFromMarker, sizeInterval, sizeStep, currPos, *distFromMarker1Arr,
				*distFromMarker1ArrPtr;
	int			currLocusIndMap, numInteriorIntervals, numIterInterval, interval, 
				posInterval;
	locusPos	*locusPosPtr;

	
	distFromMarker1Arr = scanDataPtr->distFromMarker1;

	/* 1.	Calculate the number of interior intervals on the map which the 
			scan includes. */

	numInteriorIntervals = scanDataPtr->endPos - scanDataPtr->startPos;
	numIterInterval = sGHDataLocusPtr->numIterInterval;

	currPos = 0.0;
	distFromMarker = 0.0;
	currLocusIndMap = scanDataPtr->startPos;
	
	locusPosPtr = sGHDataLocusPtr->locusPosArr;

	/* 2.	If the scan continues past the left end of the map, calculate the positions
			of the disease locus on the interval to the left of the map. */

	if (scanDataPtr->offEndDistLeft > 0) {
		currLocusIndMap--;
		distFromMarker = -scanDataPtr->offEndDistLeft;
		sizeStep = scanDataPtr->offEndDistLeft / numIterInterval;

		for (posInterval = 0; posInterval < numIterInterval; posInterval++, 
				locusPosPtr++) {
			locusPosPtr->rec = getcMToRec(-distFromMarker, progData.haldane);
			locusPosPtr->locusIndMap = currLocusIndMap;
			locusPosPtr->distFromMarker1 = distFromMarker;
			distFromMarker += sizeStep;
		}
		currLocusIndMap++;
	}

	/* 3.	Update the information on the rest of the positions of the 
			iterated locus. */

	/* 3.1.	Iterate over all the interior intervals in the scan. */

	distFromMarker1ArrPtr = distFromMarker1Arr;

	for (interval = 0; interval < numInteriorIntervals; interval++) {

		/* 3.1.1.	Calculate the step size between 2 successive iterations. */

		currPos = *distFromMarker1ArrPtr;
		sizeInterval = *(distFromMarker1ArrPtr + 1) - currPos;
		sizeStep = sizeInterval / numIterInterval;
		distFromMarker = 0;

		/* 3.1.2.	Iterate over all the positions for the iterated locus in the
					current interval. */

		for (posInterval = 0; posInterval < numIterInterval; posInterval++, 
				locusPosPtr++) {

			/* 3.1.2.1.	Update the entries for this position. */
	
			locusPosPtr->locusIndMap = currLocusIndMap;
			locusPosPtr->rec = getcMToRec(distFromMarker, progData.haldane);
			locusPosPtr->distFromMarker1 = currPos + distFromMarker;

			/* 3.1.2.2.	Advance the distance of the iterated disease locus 
						from the marker preceeding it. */

			distFromMarker += sizeStep;
		}

		/* 3.1.3.	Advance the index of the iterated disease locus. */

		currLocusIndMap++;
		distFromMarker1ArrPtr++;
	}

	/* 3.2.	If the scan continues past the right end of the map, calculate the positions
			of the disease locus on the interval to the right of the map. */

	if (scanDataPtr->offEndDistRight > 0) {

		sizeStep = scanDataPtr->offEndDistRight / numIterInterval;
		distFromMarker = 0;
		currPos = *distFromMarker1ArrPtr;

		for (posInterval = 0; posInterval <= numIterInterval; posInterval++, 
				locusPosPtr++) {
			locusPosPtr->rec = getcMToRec(distFromMarker, progData.haldane);
			locusPosPtr->locusIndMap = currLocusIndMap;
			locusPosPtr->distFromMarker1 = currPos + distFromMarker;
			distFromMarker += sizeStep;
				}
	}
	else if (numInteriorIntervals == 0){ 
		locusPosPtr->locusIndMap = currLocusIndMap;
		locusPosPtr->rec = getcMToRec(distFromMarker, progData.haldane);
		locusPosPtr->distFromMarker1 = currPos + distFromMarker;
	}else{
		locusPosPtr->locusIndMap = currLocusIndMap - 1;
		locusPosPtr->rec = getcMToRec(distFromMarker, progData.haldane);
		locusPosPtr->distFromMarker1 = currPos + distFromMarker;
	}
}


/******************************************************************************
*
*       Function        moveLocus
*       Parameters      oldLocusIndMap (the current index on the genetic map of  
*						the locus which needs to be moved),
*						newLocusIndMap (the future index on the genetic map of  
*						the locus which is moved).
*       Output          none
*       Description		Updates the mapping from the genetic map to the 
*						data-structures regarding the move of the disease locus 
*						from 'oldLocusIndMap' to 'newLocusIndMap'.
*       Comments		none
*
******************************************************************************/
void moveLocus(int oldLocusIndMap, int newLocusIndMap)
{
	int			ind, oldLocusIndDS, *mapToData;

	mapToData = loci.mapToData;
	oldLocusIndDS = mapToData[oldLocusIndMap];

	if (oldLocusIndMap < newLocusIndMap) {
		for (ind = oldLocusIndMap; ind < newLocusIndMap; ind++) {
			mapToData[ind] = mapToData[ind+1];
		}
	}
	else {
		for (ind = oldLocusIndMap; ind > newLocusIndMap; ind--) {
			mapToData[ind] = mapToData[ind-1];
		}
	}

	mapToData[newLocusIndMap] = oldLocusIndDS;

	if (loci.numTraitLoci == 2) {
		loci.traitLociPos[0] = newLocusIndMap;

		if ((loci.traitLociPos[1] > oldLocusIndMap) && 
				(loci.traitLociPos[1] <= newLocusIndMap)) {
			loci.traitLociPos[1]--;
		}
	}
}


/******************************************************************************
*
*       Function        moveLoci
*       Parameters      oldLocusIndMap1 (the current index on the genetic map of  
*						the first locus which needs to be moved),
*						newLocusIndMap1 (the future index on the genetic map of  
*						the first locus which is moved), theta1 (the required 
*						recombination fraction between the first locus which
*						needs to be moved and the preceeding marker).
*						oldLocusIndMap2 (the current index on the genetic map of  
*						the second locus which needs to be moved),
*						newLocusIndMap2 (the future index on the genetic map of  
*						the second locus which is moved), theta2 (the required 
*						recombination fraction between the second locus which
*						needs to be moved and the preceeding marker).
*       Output          none
*       Description		Updates the mapping from the gentic map to the 
*						data-structures regarding the move of the first locus 
*						from 'oldLocusIndMap1' to 'newLocusIndMap1', and the 
*						second locus from 'oldLocusIndMap2' to 'newLocusIndMap2'.
*       Comments		Fixes by MARK to allow analysis over unlinked markers.
*
******************************************************************************/
void moveLoci(int oldLocusIndMap1, int *newLocusIndMap1, double theta1,
			  int oldLocusIndMap2, int *newLocusIndMap2, double theta2)
{
	boolean		move1, move2;
	boolean unlinkedRefMarkers;
	int			markerIndDS, ind, firstNewInd, secondNewInd, numLoci, *mapToData;
	int         refMarker1, refMarker2;

	numLoci = loci.numLoci;
	mapToData = loci.mapToData;
	refMarker1=progData.scanData[0].startPos;
	refMarker2=progData.scanData[1].startPos;


	if ((*newLocusIndMap1) < (*newLocusIndMap2)) {

		/* 1.1.	If the second disease locus is to the right of the first disease 
				locus on the map, then advance its index by 1. */

		(*newLocusIndMap2)++;
	}
	else if ((*newLocusIndMap1) == (*newLocusIndMap2)) {
	  
		/* 1.1.	If both disease loci are in the same interval on the map,
				then check which one is first according to the recombination
				fractions between them and the adjacent marker. */
	  if (refMarker1==refMarker2){
		// if they are referenced fromt he same ref. Marker use thettas
		if ((*newLocusIndMap1) != 0) {
			( theta1 <= theta2 ) ? (*newLocusIndMap2)++ : (*newLocusIndMap1)++;
		}
		else {
			(theta1 <= theta2 ) ? (*newLocusIndMap1)++ : (*newLocusIndMap2)++;
		}
	  }else{
		// if they are referenced from different ref. Markers => use ref. marker.
	   refMarker1<refMarker2 ? (*newLocusIndMap2)++ : (*newLocusIndMap1)++;
	  }
	}
	else {
		/* 1.1.	If the first disease locus is to the right of the second disease 
				locus on the map, then advance its index by 1. */

		(*newLocusIndMap1)++;
	}

	/* 1.2.	Move the disease loci to their required positions. */

	move1 = ((oldLocusIndMap1 != *newLocusIndMap1) ? TRUE: FALSE);
	move2 = ((oldLocusIndMap2 != *newLocusIndMap2) ? TRUE: FALSE);

/* 	if (move1 == TRUE) { */
/* 		if (move2 == TRUE) { */

			/* 1.2.1.	Both disease loci need to be moved. Find out which
						one appears first on the map. */

			if ((*newLocusIndMap1) < (*newLocusIndMap2)) {
				firstNewInd = (*newLocusIndMap1);
				secondNewInd =(*newLocusIndMap2);
			}
			else {
				firstNewInd = (*newLocusIndMap2);
				secondNewInd =(*newLocusIndMap1);
			}

			/* 1.2.2.	Update the mapping of both disease loci. */

			loci.mapToData[(*newLocusIndMap1)] = 0;
			loci.mapToData[(*newLocusIndMap2)] = 1;

			markerIndDS = 2;

			/* 1.2.3.	Update the mapping of all the loci that appear left
						of the first disease locus. */

			for (ind = 0; ind < firstNewInd; ind++, markerIndDS++) {
				mapToData[ind] = markerIndDS;
			}

			/* 1.2.4.	Update the mapping of all the loci that appear between
						the first disease locus and the second disease locus. */

			for (ind++; ind < secondNewInd; ind++, markerIndDS++) {
				mapToData[ind] = markerIndDS;
			}

			/* 1.2.3.	Update the mapping of all the loci that appear right
						of the second disease locus. */

			for (ind++; ind < numLoci; ind++, markerIndDS++) {
				mapToData[ind] = markerIndDS;
			}
/* 		} */
/* 		else { */
/* 			moveLocus(oldLocusIndMap1, *newLocusIndMap1); */
/* 		} */
/* 	} */
/* 	else if (move2 == TRUE) { */
/* 		moveLocus(oldLocusIndMap2, *newLocusIndMap2); */
/* 	} */

	loci.traitLociPos[0] = *newLocusIndMap1;
	loci.traitLociPos[1] = *newLocusIndMap2;

	/* 1.3.	 Update the recombination values. */
	unlinkedRefMarkers=computeMarkerDistance(refMarker1,refMarker2,thetaMarkers)==UNLINKED_DIST;

	if (*newLocusIndMap1 < *newLocusIndMap2)
	  changeThetas2Loci(*newLocusIndMap1, theta1, *newLocusIndMap2, theta2, unlinkedRefMarkers);
	else
	  changeThetas2Loci(*newLocusIndMap2, theta2, *newLocusIndMap1, theta1, unlinkedRefMarkers);

}


/******************************************************************************
*
*       Function        changeThetas
*       Parameters      locusInd (the place on the map where the affection status 
*						locus is placed), rec (the recombination fraction between 
*						the disease locus and the marker preceeding it. If the 
*						disease locus is first, then this is the recombination 
*						fraction between it and the first marker).
*       Output          none
*       Description		Updates the recombination fractions according to the new
*						position of the disease locus.
*       Comments		none
*
******************************************************************************/
void changeThetas(int locusInd, double rec)
{
	int					numThetas;
	thetaCompTheta		*maleThetas, *femaleThetas;

	
	maleThetas = recombData.maleThetas;
	femaleThetas = recombData.femaleThetas;

	/* 1.	Calculate the number of intervals on the map. */

	numThetas = loci.numLoci - 1;

	/* 2.	If there is only one marker on the map, or the requested position of 
			the disease locus is on the left side of the leftmost marker 
			on the map. */

	if ((loci.numLoci == 2) || (locusInd == 0)) {
		
		/* 2.1.	Update the theta value between the disease locus and that marker. */

		femaleThetas[0][0] = maleThetas[0][0] = rec;
		femaleThetas[0][1] = maleThetas[0][1] = 1 - rec;
		return;
	}
	
	/* 3.	If the  requested position of the disease locus is inside the map. */

	if ((locusInd > 0) && (locusInd < loci.numLoci-1)) {

		/* 3.1.	Restore the initial recombination fractions that preceed and follow
				the position of the disease locus. */

		if (locusInd-1 > 0)
			memcpy(maleThetas, thetaMarkers, sizeof(thetaCompTheta) * (locusInd-1));

		if (numThetas-locusInd-1 > 0)
			memcpy(maleThetas+locusInd+1, thetaMarkers + locusInd, sizeof(thetaCompTheta) * (numThetas-locusInd-1));

		/* 3.2.	Update the recombination fraction between the disease locus
				and the markers on both its sides. */

		maleThetas[locusInd-1][0] = rec;
		maleThetas[locusInd-1][1] = 1 - rec;

		maleThetas[locusInd][0] = getThetaBC(thetaMarkers[locusInd-1][0],rec, progData.haldane);
		maleThetas[locusInd][1] = 1 - maleThetas[locusInd][0];

		/* 3.3.	Update the female recombination fractions according to the male
				recombination fractions. */

		memcpy(femaleThetas, maleThetas, sizeof(thetaCompTheta) * numThetas);

		return;
	}
	
	/* 4.	If the  requested position of the disease locus is on the right side 
			of the rightmost marker on the map. */

	if (locusInd == loci.numLoci-1) {

		/* 4.1.	Restore the initial recombination fractions that preceed the position
				of the disease locus. */

		memcpy(maleThetas, thetaMarkers, sizeof(thetaCompTheta) * (numThetas-1));

		/* 4.2.	Update the recombination fraction between the rightmost marker
				and the disease locus. */

		maleThetas[numThetas-1][0] = rec;
		maleThetas[numThetas-1][1] = 1 - rec;

		/* 4.3.	Update the female recombination fractions according to the male
				recombination fractions. */

		memcpy(femaleThetas, maleThetas, sizeof(thetaCompTheta) * numThetas);
	}
}


/******************************************************************************
*
*       Function        changeThetas2Loci
*       Parameters      locusInd1 (the place on the map where the first disease
*						locus is placed), rec1 (the recombination fraction between 
*						the first disease locus and the marker preceeding it. If  
*						the disease locus is first, then this is the recombination 
*						fraction between it and the first marker).
*       Output          none
*       Description		Updates the recombination fractions on the map according 
*						to the new positions of the two disease loci. 
*       Comments		none
*
******************************************************************************/
void changeThetas2Loci(int locusInd1, double rec1, int locusInd2, double rec2, boolean unlinked)
{
	int				numThetas, numIntervalsMap;
	double			tmpRec;
	boolean			sameInterval;
	thetaCompTheta	*maleThetas, *femaleThetas;


	maleThetas = recombData.maleThetas;
	femaleThetas = recombData.femaleThetas;

	/* 1.	Calculate the number of intervals on the map. */

	numThetas = loci.numLoci - 1;
	numIntervalsMap = loci.numLoci - 3;

	/* 2.	Check if the two disease loci are in the same interval on the marker
			map. */

	sameInterval = (abs(locusInd1 - locusInd2) == 1);

	/* 3.	If the first disease locus is on the left side of the leftmost
			marker on the map. */

	if (locusInd1 == 0) {
		
		/* 3.1.	If the second disease locus is in the same interval as the first
				disease locus. */

	  if (sameInterval) {
		  
		if (unlinked) {

			tmpRec=0.5;

		}else{
		
		  /* 3.1.1.	Calculate the recombination fraction between the two disease
			 loci. */
		  
		  tmpRec = getThetaBC(rec1, rec2, progData.haldane);
		} 
		/* 3.1.2.	Update the theta value between the two disease loci. */
		
		maleThetas[0][0] = tmpRec;
		maleThetas[0][1] = 1 - tmpRec;
		
		/* 3.1.3.	Update the theta value between the 2nd disease locus and the
		   first marker on the map. */
		
		maleThetas[1][0] = rec2;
		maleThetas[1][1] = 1 - rec2;
		
		/* 3.1.4.	Update the rest of the theta values according to the initial 
		   recombination fractions between the markers. */
		
		memcpy(maleThetas+2, thetaMarkers, sizeof(thetaCompTheta) * numIntervalsMap);
		
		goto updateFemaleRecs;
		
	  }
	  else {
		/* 3.1.	The second disease locus is in a different interval. */

		/* 3.1.1.	Update the theta value between the first disease locus and 
		   the leftmost marker. */
		
		maleThetas[0][0] = rec1;
		maleThetas[0][1] = 1 - rec1;
	  }
	}
	else if (locusInd1 < loci.numLoci-1) {
		
		/* 3.	If the requested position of the first disease locus is inside the map. */

		/* 3.1.	Restore the initial recombination fractions that preceed 
				the position of the first disease locus. */

		if (locusInd1 > 1)

		  memcpy(maleThetas, thetaMarkers, sizeof(thetaCompTheta) * (locusInd1-1));
		
		/* 3.2.	Update the recombination fraction between the first disease locus
		   and the marker preceeding it. */
		
		maleThetas[locusInd1-1][0] = rec1;
		maleThetas[locusInd1-1][1] = 1 - rec1;

		/* 3.1.	If the second disease locus is in the same interval as the first
				disease locus. */

		if (sameInterval) {	

			/* 3.1.1.	Calculate the recombination fraction between the two disease
						loci. */

		  if (unlinked){

			tmpRec=0.5;

		  }else{

			tmpRec = getThetaBC(rec2, rec1, progData.haldane);

		  }
			
		  maleThetas[locusInd1][0] = tmpRec;
		  maleThetas[locusInd1][1] = 1 - tmpRec;
		  
		  // if second locus is inside the map:

		  if (locusInd2< loci.numLoci-1){
			/* 3.1.2.	Calculate the recombination fraction between the 2nd disease
			   locus and the marker following it. */
			
			// if two reference markers are unlinked and both loci are in the same interval 
			// -> original recomb. fraction is necessarily b/w the next marker and the disease locus
			
			if (unlinked){
			  maleThetas[locusInd2][0]=rec2;
			  maleThetas[locusInd2][1]=1-rec2;
			}else{
			  maleThetas[locusInd2][0] = getThetaBC(thetaMarkers[locusInd1-1][0], rec2, progData.haldane);
			  maleThetas[locusInd2][1] = 1 - maleThetas[locusInd1+1][0];
			}
			
			/* 3.1.3.	Restore the initial recombination fractions that follow
			   the position of the 2nd disease locus. */
			
			
			if (numIntervalsMap > locusInd1)
			  memcpy(maleThetas+locusInd2+1, thetaMarkers + locusInd1, sizeof(thetaCompTheta) * (numIntervalsMap-locusInd1));
		  }
		  
		  goto updateFemaleRecs;
		
		}
		else {

			/* 3.1.1.	Update the recombination fraction between the first disease 
			   locus and the marker following it. */
		  
		  maleThetas[locusInd1][0] = getThetaBC(thetaMarkers[locusInd1-1][0], rec1, progData.haldane);
		  maleThetas[locusInd1][1] = 1 - maleThetas[locusInd1][0];
		}
	}
	else {
	  /* 3.	The requested position of the first disease locus is on the right of the 
		 of the rightmost marker on the map. The second disease locus must be in
		 the same interval as well. */
	  
	  /* 3.1.	Restore the initial recombination fractions that preceed the position
				of the first disease locus. */

	  memcpy(maleThetas, thetaMarkers, sizeof(thetaCompTheta) * (numIntervalsMap));

	  /* 3.2.	Update the recombination fraction between the rightmost marker
		 and the first disease locus. */
	  
	  maleThetas[locusInd1-1][0] = rec1;
	  maleThetas[locusInd1-1][1] = 1 - rec1;
	  
	  /* 3.3.	Update the recombination fraction between the first disease locus
		 and the second disease locus. */
	  tmpRec = getThetaBC(rec2, rec1, progData.haldane);
	  
	  maleThetas[locusInd1][0] = tmpRec;
	  maleThetas[locusInd1][1] = 1 - tmpRec;

	  goto updateFemaleRecs;
	}

	/* 4.	Update the initial recombination fractions that follow the first
			disease locus and preceed the second disease locus. */

	if ((locusInd2-locusInd1-2) > 0)
		memcpy(maleThetas+locusInd1+1, thetaMarkers+locusInd1, sizeof(thetaCompTheta) * (locusInd2-locusInd1-2));

	/* 5.	Update the recombination fraction between the second disease locus
			and the marker preceeding it. */
	
	maleThetas[locusInd2-1][0] = rec2;
	maleThetas[locusInd2-1][1] = 1 - rec2;

	/* 6.	If the  requested position of the 2nd disease locus is inside the map. */
	/* new anna 12/7*/
	if (locusInd2 < loci.numLoci-1) {

		/* 6.1.	Update the recombination fraction between the 2nd disease locus
		and the marker following it. */
		if (thetaMarkers[locusInd2-2][0] != 0.5){
			maleThetas[locusInd2][0] = getThetaBC(thetaMarkers[locusInd2-2][0], rec2, progData.haldane);
			maleThetas[locusInd2][1] = 1 - maleThetas[locusInd2][0];
		}else{
			maleThetas[locusInd2][0] = maleThetas[locusInd2-1][0];
			maleThetas[locusInd2][1] = maleThetas[locusInd2-1][1];

			maleThetas[locusInd2-1][0] = maleThetas[locusInd2-1][1] =0.5;
		}
/* end new*/

		/* 6.2.	Restore the initial recombination fractions that follow
				the position of the second disease locus. */

		if ((numIntervalsMap-locusInd2+1) > 0)
			memcpy(maleThetas+locusInd2+1, thetaMarkers+locusInd2-1, sizeof(thetaCompTheta) * (numIntervalsMap-locusInd2+1));
	}


	/* 7.	Update the female recombination fractions according to the male
			recombination fractions. */

updateFemaleRecs:
	memcpy(femaleThetas, maleThetas, sizeof(thetaCompTheta) * numThetas);
}



/******************************************************************************
*
*       Function        freeSuperGHDS
*       Parameters      none.
*       Output          none.
*       Description		frees the memory allocated for the data-structures that
*						are used specifically in the Super-GH program. 
*       Comments		none
*
******************************************************************************/
void freeSuperGHDS()
{
	superGHData		*sGHData;
	sGHDataLocus	*sGHDataLocusPtr, *sGHDataLocusPtrEnd;

	sGHData = &((progData.progSpecific).sGHData);

	/*	1. Iterate over the data-structure allocated for each iterated locus. */

	for (sGHDataLocusPtr = sGHData->sGHDataLocusArr, 
			sGHDataLocusPtrEnd = sGHDataLocusPtr + sGHData->numIterLoci;
			sGHDataLocusPtr < sGHDataLocusPtrEnd; sGHDataLocusPtr++) {

		/*	1.1. Free the array that stores all the positions for this locus. */

		myFree(sGHDataLocusPtr->locusPosArr);
	}

	/* 2.	Free the arrays that were allocated to store the lod-scores obtained
			at all positions, the positions of the iterated loci that produced the
			maximum lod-score and the array that stored information on each one
			of the iterated loci. */

	myFree(sGHData->sGHDataLocusArr);

	myFree(thetaMarkers);
}

