/* $Id: LociData.h,v 1.2 2008/08/03 08:15:46 superlink Exp $ */
/*********************************************************************************/
/*  LociData.h																     */
/*  Author: Ma'ayan Fishelson													 */
/*********************************************************************************/
/*																				 */
/*  Comment: This file includes functions used for reading all the data from the */
/*			 locus file and initializing the relevant data-structures.			 */
/*																				 */
/*********************************************************************************/

#ifndef _LOCI_DATA_H_
#define _LOCI_DATA_H_

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "Auxiliary.h"
#include "MapFunction.h"

/*------------------------------*/
/*		DEFINITIONS				*/
/*------------------------------*/

/* Program_Codes for the different programs options. */

#define	SUPER_LINKMAP_PROG			4
#define SUPER_MLINK_PROG			5
#define	SUPER_POS_PROG				6
#define	SUPER_GH_PROG				8
#define	SUPER_HAPLO					10
#define	SUPER_LINKMAP_APPROX_PROG	11		/* Not implemented currently. */
#define	SUPER_OPTLINK_PROG			12
#define SUPER_OPTMARGINAL_PROG      13
#define BAYES_NET_XML_BEFORE_ELIM	20
#define BAYES_NET_XML_AFTER_ELIM	21
#define BAYES_NET_XBN_BEFORE_ELIM	30
#define BAYES_NET_XBN_AFTER_ELIM	31
#define PRINT_ELIM_ORDER			40
#define PRINT_ELIM_ORDER_NO_BREAK               41

/* The default name for the locus file. */

#define DEFAULT_DATAFILE_NAME  "datafile.dat"

/* Codes for the different types of loci possible. */

#define AFFECTION_LOCUS	1
#define	ALLELES_LOCUS	3
#define	NULL_LOCUS		4

/* The different types of loci. */

typedef	char		locusType;
#define	affection	0
#define	alleles		1
#define	null		2

#define	HALF_PENETRANCE			0.5

/*------------------------------*/
/*		DATA-STRUCTURES			*/
/*------------------------------*/

/*  The program will just create bayes net and exit */
int RUN_CREATE_BAYES;

/* Holds theta in the first entry and (1 - theta) in the second entry. */
typedef Probability thetaCompTheta[2];

/* Holds the female penetrance values of a one-locus trait. */
typedef thetaCompTheta femalePenClass[3];

/* Holds the male penetrance values of a one-locus trait. */
typedef thetaCompTheta malePenClass[2];

/* Holds the female penetrance values of a two-locus trait. */
typedef thetaCompTheta femalePenClassTwo[3][3];

/* Holds the male penetrance values of a two-locus trait. */
typedef thetaCompTheta malePenClassTwo[2][2];

/* Holds the joint probabilities for a combination of existence/lack of
   recombination in 2 successive intervals along the chromosome (for the case
   where interference is taken into account).
   If we have the following sequence of loci A B C, then:
	- the first entry holds the probability that there's NO recombination
	  between A and B AND NO recombination between B and C (g00). 
	- the second entry holds the probability that there's NO recombination
	  between A and B AND there is a recombination between B and C (g01). 
	- the third entry holds the probability that there is a recombination
	  between A and B AND there is NO recombination between B and C (g10). 
	- the fourth entry holds the probability that there is a recombination
	  between A and B AND there is a recombination between B and C (g11). */

typedef Probability jointProbs[4];

/* Holds the information on one locus. */
typedef struct locus {
	locusType		type;						/* The type of locus */
	int				numAlleles;					/* Number of possible alleles. */
	Probability		*geneFrequencies;			/* The allele frequencies. */
	char			*name;						/* The name of the locus. */
} locus;

/* Holds information on all loci. */
typedef struct lociData {
	boolean		sexLink;				/* A flag that indicates whether the data is sex-linked
										   or autosomal. */
	int			numLoci;				/* The number of loci in the analysis. */
	locus*		locusArr;				/* The array which holds the information on the loci. */
	int			numTraitLoci;			/* The number of trait loci (currently, either 1 or 2). */
	int			*traitLociPos;			/* The positions of the trait loci (in map order). If there
										   are 2 trait loci, the 1st entry holds the position of
										   the first trait locus in the data-structures. */
	int			lastNullLocusIndDS;		/* The place in the data-structures of the last NULL locus. */
	int			*lociOrder;				/* A mapping from input order of the loci to chromosome order. */
	int			*mapToData;				/* A mapping from genetic order of the loci to their order in
										   the data-structures (entry i in the array holds the index in 
										   the data-structures, loci and peds, of the locus that is in 
										   position i in the current genetic map). */
	int			numClasses;				/* The number of liability classes (penetrance classes). */
	union {
		struct {								/* Penetrances in the case of a one-locus disease. */
			femalePenClass	*fPenetrances;		/* The female penetrances. */
			malePenClass    *mPenetrances;		/* The male penetrances. */
		} oneLocus;
		struct {								/* Penetrances in the case of a two-locus disease. */
			femalePenClassTwo	*fPenetrances;	/* The female penetrances. */
			malePenClassTwo     *mPenetrances;	/* The male penetrances. */
		} twoLocus;
	} diseaseModel;
} lociData;

/* Stores the information on one position along the genetic map of one disease locus,
   used for the SuperGH program option. */
typedef struct locusPos {
	int				locusIndMap;		/* The index of the disease locus on the map. */
	Probability		rec;				/* The recombination fraction between the disease
										   locus and the marker preceeding it, or between the
										   disease locus and the marker following it (in case
										   the disease locus is first on the map). */
	double			distFromMarker1;	/* The distance in cM from the marker from which the
										   scan began (the leftmost marker in the scan interval). */
} locusPos;


/* Stores all necessary information on one of the disease loci, for the SuperGH program. */
typedef struct sGHDataLocus {
	locusPos	*locusPosArr;			/* An array that stores the information on all the 
										   requested positions along the genetic map of one 
										   of the disease loci. */
	int			numPos;					/* The size of the array 'locusPosArr'. */
	boolean		isStepIter;				/* A flag which indicates whether the size of the step
										   between each 2 successive positions was specified, or
										   the number of steps in each interval. */
	double		stepIter;				/* The size of the step between successive positions
										   of one of the disease loci.*/
	int			numIterInterval;		/* The number of requested positions of one of the
										   disease loci between each 2 markers on the map. */
} sGHDataLocus;


/* Stores all the necessary information for the SuperGH program. */

typedef struct superGHData {
	int				numIterLoci;		/* The number of disease loci that are to be iterated on. */
	int				numIter;			/* The number of likelihood evaluations performed. */
	sGHDataLocus	*sGHDataLocusArr;	/* An array that stores information on the positions
										   in the scan of each of the disease loci that are to 
										   be iterated on. */
	boolean			checkModel;			/* A flag that indicates uncertainty of the disease
										   model, and therefore two models will be compared. */
} superGHData;

/* A data-structure which holds information on the scan interval. */
typedef struct scanIntervalData {
	int		startPos;			/* The number of the marker which delimits the scan interval
								   from the left. */
	int		endPos;				/* The number of the marker which delimits the scan interval
								   from the right. */
	double	*mapDistArr;		/* The map distances (in cM) between the markers in the scan 
								   interval. */
	double	*distFromMarker1;	/* The distances (in cM) of the markers in the scan interval 
								   from the leftmost marker in the interval. */
	int		numIntervals;		/* The number of intervals in the total scan interval (equals
								   the number of markers in the scan interval minus 1). */
	int		currPosTraitLoc;	/* The current position of the trait locus in the scan interval. */
	double	offEndDistLeft;		/* The distance in cM on the left side of the map where 
								   the scan should continue. */
	double	offEndDistRight;	/* The distance in cM on the right side of the map where 
								   the scan should continue. */
} scanIntervalData;

/* A data-structure which holds information on the position(s) of the iterater disease
   locus (loci) for one likelihood evaluation for the SuperPos program option. */
typedef struct onePosData {
  double		*locPos;		/* The position(s) of the iterated disease locus (loci). */
  int           *refMarker;      /* The marker to refer the distance from */
} onePosData;

/* A data-structure that holds information required by the analysis programs. */
typedef struct programData {
	int					programCode;	/* The code of the analysis program that should be run. */
	int					numScanIntervals;
	scanIntervalData	*scanData;		/* Holds the data on the scan intervals. */
	boolean				haldane;		/* A flag that indicates whether Haldane's or Kosambi's
										   map function are to be used. */

	union progSpecific {				/* Program-specific data. */
		struct linkmapData {
			int		locusVaried;
			int		numEvalPoints; 
			double	finalVal;	
		} linkmapData;

		struct mlinkData {
			int		recVaried;
			double	inc;
			double	finalVal;	
		} mlinkData;

		struct posData {
			int			numIterLoci;	/* The number of disease loci that are to be iterated on. */
			int			numPos;			/* The number of positions in which the likelihood needs
										   to be evaluated (i.e., the requested number of likelihood 
										   evaluations). */
			onePosData	*posDataArr;	/* An array which stores the positions of the iterated 
										   disease locus (loci) in which the likelihood neeeds 
										   to be computed. */
		} posData;

		superGHData	sGHData;

		struct optlinkData {
			boolean	checkModel;			/* A flag that indicates uncertainty of the disease
										   model, and therefore two models will be compared 
										   (the MMLS-C approach). */
			double	searchTol;			/* The search tolerance. */
		} optlinkData;

	} progSpecific;

} programData;

/* A data-structure which holds all the information regarding the recombination fractions. */
typedef struct recombinationData {
	thetaCompTheta *maleThetas;		/* A vector which holds the theta values between each 2
									   consecutive loci on the chromosome, for males. */
    thetaCompTheta *femaleThetas;	/* A vector which holds the theta values between each 2
									   consecutive loci on the chromosome, for females. */
    Probability maleThetas2Seg;		/* Holds the male theta value between the 2 flanking markers in case
									   of a 3-loci analysis, when interference is specified. */
	Probability femaleThetas2Seg;	/* Holds the female theta value between the 2 flanking markers in case
									   of a 3-loci analysis, when interference is specified. */
	jointProbs	maleJointThetas;	/* Holds the probabilities of a combination of existence/lack of 
									   recombination in 2 successive intervals along the chromosome, 
									   for males.  (Currently, interference can only be specified 
									   for a three-loci analysis). */
	jointProbs	femaleJointThetas;	/* Holds the probabilities of a combination of existence/lack of 
									   recombination in 2 successive intervals along the chromosome, 
									   for females.  (Currently, interference can only be specified 
									   for a three-loci analysis). */

	int			sexDiff;			/* A flag that indicates whether there is no sex-difference, 
									   constant sex-difference or variable sex-difference in the 
									   recombination fractions. */
	double		geneticDist;		/* The genetic distance between the male and female recombination
									   fractions in case of a constant sex-difference. */
	int			interference;		/* A flag that indicates whether there's no interefence, user-
									   defined interference or intereference according to Kosambi's
									   mapping function. */
} recombinationData;


/*------------------------------*/
/*		GLOBAL VARIABLES		*/
/*------------------------------*/

extern FILE				*dataFile;
extern boolean			haldane;
extern lociData			loci;
int						numTraitLociRead;

//	Added by Anisenia Andrei for indication of imprinting case
boolean					imprinting;				

/*------------------------------*/
/*		FUNCTIONS				*/
/*------------------------------*/

void readLocusFile();
void readLocus(int locusInd);
void readAffectionLocus(int locusInd);
void readNullLocus(int locusInd);
void readGeneFrequencies(int locusInd);
void freeLociData();

void readSuperLinkmapSpecificData();
void readSuperMlinkSpecificData();
void readSuperPosSpecificData();
void readSuperGHSpecificData();
void readSuperOptLinkSpecificData();

void updateTransmissionModel(boolean isRecessive, double penetrance);
int ifCheckModel();
void readIntervalEnds(scanIntervalData	*scanDataPtr, int numMarkers, 
					  int diseaseLocInd);
void updateOffEndDist(scanIntervalData *scanDataPtr, int numIterLoci, int numMarkers);
void updateDistArrs(scanIntervalData *scanData, thetaCompTheta *thetaMarkers);

#endif /* _LOCI_DATA_H_ */
