/*********************************************************************************/
/*  Main.c												     				     */
/*  Author: Ma'ayan Fishelson					                                 */
/*********************************************************************************/
#define _CRT_SECURE_NO_DEPRECATE

#define MY_DEBUG 0					//	0 - no debug; 1- debug mode!

//	DEBUG prints of infer time 
#define PRINT_INFER_ENGINE_TIME 0
#define INFER_ENGINE_TIMES_FILE "inferTime.txt"

//	Available external memory default parameter = 2Gb
#define DEFAULT_AVAIL_EXT_MEMORY 1024.00*1024*1024*2

//	In case when there is not enough memory for HMM calculation,
//	NOT_ENOUGH_MEMORY is returned by exit(NOT_ENOUGH_MEMORY)
#define NOT_ENOUGH_MEMORY 3

#include "InferEngine/InferEngine.h"
#include "gh/MemoryManager/MemoryManager.h"

#ifdef  __cplusplus
extern "C" {
#endif
	#include "Globals.h"
	#include "Haplotype.h"
		
	void initStructuresForHMM(char *pedfileName,char *datafileName);
	inferenceInput* initInferAndElimOrderArray(int givenMem, int shouldExit);
	inferenceInput* initInferAndElimOrderArrayForSelectors();
	void destroyInferenceInput(inferenceInput* infer);
	int gh_main(double* diseaseProbs, MemoryManager* memManager, double probForCheck);
	void testMemoryManager(MemoryManager* memManager);
														
	boolean runHMM();
	/*defined in Prog.c*/
	void createExportBayesProg();
	product** createExportProduct();

	MemoryManagerCreateInput* getMemoryManagerInput(int givenMem, double extMem);
	void deleteMemoryManagerInput(MemoryManagerCreateInput* managerInput);

	product* testExampleProduct();

#ifdef  __cplusplus
}
#endif



/******************************************************************************
*
*       Function        main
*       Parameters		none
*       Output          none
*       Description		The main function of the program.
*       Comments		none
*
******************************************************************************/
int main(int argc, char* argv[])
{
	int			numLoc;
	clock_t		start_time, end_time;
	double		calcLikelihoodTime;
	
#ifdef BOINCIO
        INIT_BOINC()
#endif
	start_time=clock();
	
#if MANAGE_MEM
	manageMemory = FALSE;
#endif

/*#ifdef NDEBUG
	mem_avl = availableVirtMem();
#else
	mem_avl = 1000000000;
#endif
*/
	SET_THRESHOLDS();
	
	/* 1.	Get the names of the input files. */

	if (argc > 1) 
		datafileName = argv[1];
	else
		datafileName = DEFAULT_DATAFILE_NAME;

	if (argc > 2)
		pedfileName = argv[2];
	else
		pedfileName = DEFAULT_PEDFILE_NAME;

	/* 2.	Determine whether the optimization of the elimination order would be
			determined automatically by the program, or by the user in the command line. */
#ifdef PARALLEL
       heuristicEO.stopAfterEstimate = FALSE;
       heuristicEO.indexNum = -1;
       heuristicEO.totalTime = INT_MAX;
       heuristicEO.startTime = time(NULL);
       fprintf(stdout,"+++++>Memory threshold set to %f<+++++\n",bigBreakThreshold);
       
#endif


	exitAfterFindOrder = FALSE;
	RUN_GH = 0;
	RUN_SL = 0;
	if ((argc > 3) && (argv[3][0] == '-')) 
		setEOOptimizationPolicy(argc, argv);
	else
		heuristicEO.isAutomatic = TRUE;

	/* 3.	Read the input files and initialize the program's data-structures. */

	
	init(datafileName, pedfileName);


	if (argc > 3 && strcmp(argv[3], "-alg") == 0){
		if (runHMM() )
			return 200;
		else 
			return 100;
	}

	int shouldExit = 0;
	int mem = 0;
	double extMem = DEFAULT_AVAIL_EXT_MEMORY;
	//	Check if only HMM should run
	if (argc > 4 && strcmp(argv[3], "-gh") == 0) {
		//	Check if the program should exit after memory check
		if (strcmp(argv[4], "-exit") == 0) {
			shouldExit = 1;
		}
		else if (argc > 5 && strcmp(argv[5], "-exit") == 0) {
			shouldExit = 1;
		}
		//	Check which amount of memory should be used
		if (argv[4][1] == 'm') {
			mem = atoi(argv[4] + 2);
		}
		else if (argc > 5 && argv[5][1] == 'm') {
			mem = atoi(argv[5] + 2);
		}
		//	Check which amount of hard disc space in Gb the program is allowed to use
		if (argv[4][1] == 'g') {
			extMem = atof(argv[4] + 2) * 1024.00*1024*1024;
		}
	}
	
	
	if (RUN_GH || (!RUN_SL && !exitAfterFindOrder && runHMM())) {
		imprinting = TRUE;
		initStructuresForHMM(pedfileName, datafileName);
		inferenceInput* infer = initInferAndElimOrderArray(mem, shouldExit);
		//	Creating memory manager
		MemoryManagerCreateInput* managerInput = getMemoryManagerInput(mem, extMem);
		MemoryManager* manager = NULL;
		try {
		    manager = new MemoryManager(*managerInput);
			if (manager->getRequiredExtMem() > 0) {
				cout<<"The program will use "<<(manager->getRequiredExtMem())/1024/1024/1024<<
					" Gb of disc space in order to run HMM calculation"<<endl;
			}
		}
		catch (MemoryManager::NotEnoughMemoryException e) {
			cout<<"Memory required for HMM calculation is too big for current computer"<<endl;
			cout<<"If "<<(e.requiredExtMemory)/1024/1024/1024<<" Gb are available on hard disc,"<<
				endl<<"rerun the program with parameters 'datafile pedfile -gh -gX' where X"<<endl<<
				"equals to the number of Gb the program is allowed to use (with the minimal value mentioned above)"<<endl;
			destroyInferenceInput(infer);
			infer = NULL;
		}
		catch(MemoryManager::NullFuncPointerException) {
			cout<<"Null function pointer error"<<endl;
		}
		deleteMemoryManagerInput(managerInput);
		if ((infer != NULL) && (!shouldExit)) {
			//	Check that HMM calculation can be performed
			if (manager != NULL) {
				#if MY_DEBUG == 0		//	no debug
					//	using Ido infer engine
					//	IMPORTANT: how to use "InferEngine_Ido::CInferEngine" see 
					//	in "InferEngine.h" (format of input product)

					FILE* inferFile;
					#if (PRINT_INFER_ENGINE_TIME == 1) 
						inferFile = fopen(INFER_ENGINE_TIMES_FILE, "w");
						fprintf(inferFile,  "Start writing = %lf\t\t", ((double)clock())/(double)CLOCKS_PER_SEC);	
					#endif
					
					InferEngine_Ido::CInferEngine* idoEngine = new InferEngine_Ido::CInferEngine();
					bool problemOccured = false;
					double* diseaseTable = idoEngine->computePartialProb(infer->productPtr,
						infer->notElimOrderArray, infer->numNotElimVars, problemOccured);
					double probForCheck = -1;
					if (problemOccured) {
						inferenceInput* inferForCheck = initInferAndElimOrderArrayForSelectors();		
						double* diseaseForCheck = idoEngine->computePartialProb(inferForCheck->productPtr,
							inferForCheck->notElimOrderArray, inferForCheck->numNotElimVars, problemOccured);
						probForCheck = diseaseForCheck[0];
						delete[] diseaseForCheck;
						destroyInferenceInput(inferForCheck);
					}
					delete idoEngine;

					#if (PRINT_INFER_ENGINE_TIME == 1) 
						fprintf(inferFile,  "End writing = %lf\n", ((double)clock())/(double)CLOCKS_PER_SEC);
						fclose(inferFile);
					#endif
					
					int res = gh_main(diseaseTable, manager, probForCheck);
					delete manager;
					return res;
				#else			//	debug mode
					testMemoryManager(manager);
					delete manager;
					return 0;
				#endif
			}
			destroyInferenceInput(infer);
		}
		else if (infer == NULL) {
			//	Not enogh memory for HMM calcualtion
			exit(NOT_ENOUGH_MEMORY);
		}
		else {
			exit(0);
		}
	}	
	
	if (RUN_GH){ 
		return 2;
	}
	else{
		init(datafileName, pedfileName);
	}

	/* 4.	Perform initial value-elimination on the pedigrees. */

	if (loci.sexLink) {
		valueEliminate(updateChildSexLinked, updateFlagsLociSexLinked,
					   updateParentsSexLinked);
	}
	else {
		valueEliminate(updateChildAutosomal, updateFlagsLociAutosomal,
					   updateParentsAutosomal);
	}

	/* 5.	If allele-recoding is requested, perform allele-recoding. */

	#if	PERFORM_ALLELE_RECODING
		performAlleleRecoding();
	#endif	

	/* 6.	Initialize the random seed of the system. */

	srand((unsigned)time(NULL));

	/* 7.	Open the elimination-order statistics file for writing. */

	#if PRINT_VARIOUS_STATISTICS_LEVEL
		openInitEOStatisticsFile();
	#endif

	/* 8.	Initialize the data-structure that would hold the results of all the
			likelihood calculations. */
			initResultList();

	if ( RUN_CREATE_BAYES){
			createExportBayesProg();
	}




	/* 9.	If there's only 1 locus and this isn't a haplotyping analysis, call the 
	function which calculates the likelihood for 1 locus. */

	if ((loci.numLoci == 1) && (progData.programCode != SUPER_HAPLO)
		&& !RUN_CREATE_BAYES) {
			calcLikelihoodOneLocus();
			goto end;
		}

	/* 10.	Calculate the likelihood, according to the requested program. */
	
	switch (progData.programCode) {

		case SUPER_LINKMAP_PROG: calcLikelihoodSuperLinkmap();
								 break;

		case SUPER_MLINK_PROG:	calcLikelihoodSuperMlink();
								break;

		case SUPER_POS_PROG:	calcLikelihoodSuperPos();
								break;

		case SUPER_GH_PROG:		calcLikelihoodSuperGH();
								break;

		case SUPER_HAPLO:		superHaploProg();
								break;

		case SUPER_LINKMAP_APPROX_PROG: calcLikelihoodSuperLinkmapApprox();
									    break;

		case SUPER_OPTLINK_PROG:	calcLikelihoodSuperOptlink();
									break;

		case SUPER_OPTMARGINAL_PROG:   OptimizeMarginalLikelihood();
                                        break;	
	}

end:

	/* 11.	Calculate and print the total computation time. */

	end_time=clock();
	calcLikelihoodTime = ((double)(end_time-start_time))/((double)CLOCKS_PER_SEC);

	myPrintf("\n\nThe time needed for calculating the likelihood of the pedigree is %lf sec.\n",calcLikelihoodTime);
	myPrintf("\nSuperlink v1.7 (Original Citation: Fishelson M. and Geiger D. Bioinformatics,  2002:18, S189-S198).\n");
	myPrintf("(c)2002-2008  Technion - Israel Institute of Technology\n\n");
	/* 12.	Print the total run time to the statistics files and close them. */

	#if PRINT_VARIOUS_STATISTICS_LEVEL
		fprintf(statisticsFile, "%lf\n", calcLikelihoodTime);
		fclose(statisticsFile);
	#endif 

	/* 13.	Free the memory allocated dynamically. */
   #ifdef WIN32
	_ASSERT(_CrtCheckMemory());

#endif
	numLoc = (progData.programCode == SUPER_HAPLO) ? numMarkers : loci.numLoci;

	#if PERFORM_ALLELE_RECODING
		freePedDataAR(numLoc);
	#else
		freePedData(numLoc);
	#endif

	freeLociData();
	freeResultList();

#ifdef PARALLEL
	myPrintf("@@@###$$$\n");
#endif

#ifdef BOINCIO
        END_BOINC(0);
#endif

	return 0;
}



