/* $Id: Marginal.c,v 1.2 2008/08/03 08:15:46 superlink Exp $ */
/*********************************************************************************/
/*  Marginal.c											     				     */
/*  Author: Dmitry Rusakov  					                                 */
/*********************************************************************************/


#include <assert.h>
#include <stdlib.h>
#include <math.h>

#include "Marginal.h"
#include "Optimize.h"


/* data structures for wrapper functions */
#define INITIAL_RESCALE_FACTOR_LN 10
typedef struct likelihood_data_tag {
    /* data for normal likelihood evaluations */
    boolean     isRecessive;
    double      location;
    /* for near zero function computations, in evaluate_at_penetrance_adaptive() */
    double      rescale_factor_ln;  /* logarithm of rescale factor */
    /*double      prior;*/
} likelihood_data;

typedef struct marginal_data_tag {
    double ln_appBaseR;
    double ln_appBaseD;
    double ln_appBase;
} marginal_data;


/* wrapper functions for different evaluations */
double evaluate_at_penetrance_adaptive(double pen, void* func_data);
double evaluate_at_penDD_penHD_scaled(double penDD, double penHD, void *func_data);
double evaluate_at_penDD_penHD(double penDD, double penHD, void *func_data);
double evaluate_full_marginal_lod(double pen, void* func_data);
double evaluate_double_marginal_lod(double pen, void* func_data);
double evaluate_rec_marginal_lod(double pen, void* func_data);
double evaluate_dom_marginal_lod(double pen, void* func_data);
/* outdated code, not used in mainstream computations
double evaluate_at_theta(double theta, void* func_data);
double evaluate_at_penetrance(double pen, void* func_data);
double evaluate_at_theta_and_penetrance(double theta, double pen, void *func_data);
*/

/* debug functions */
double test_fun(double x, void* func_data);
double test_fun2(double x, double y, void* func_data);
void dump_matlab_file(functype2 fun, double xmin, double xmax, int steps, char* name, void* func_data);




/******************************************************************************
*
*           EXTERNAL INTERFACE FUNCTIONS
*
******************************************************************************/


/******************************************************************************
*
*       Functions       OptimizeMarginalLikelihood
*       Parameters      none
*       Output          Print various info in stdout.
*       Description     High-level wrapper function,
*                       calls one of Best/Full/Double Optimize..MarginalLikelihood functions.
*                       Main function of Superlink Program 13.
*
*                       Finds the position of the affection status locus within
*						the specified scan interval which maximizes the 
*						marginal likelihood of the data,
*                       averaged by the penetrance model and penetrance itself,
*                       3 methods:
*                       Best - likelihood of data given disease locus location and dominant or recessive
*                              penetrance model is averaged over panetrance values in given model (PEN_MIN to PEN_MAX)
*                              LOD score versus unlinked disease is computed for both models and best model is seleted.
*                              maximum LOD score is corrected by MAXLOD_CORRECTION and printed on screen.
*                       Double - likelihood of data given disease locus location and dominant or recessive
*                                penetrance model is averaged over panetrance values in given model (PEN_MIN to PEN_MAX),
*                                then two likelihoods (given dominant or recessive) penetrance models are averaged with equal
*                                weights 0.5 for each model.
*                                LOD score versus unlinked disease averaged likelihood (by the same method) is computed and reported.
*                       Full - likelihood of data given disease locus location and unspecified penetrance model is computed
*                              Penetrances are averaged over: PEN_MIN <= pen(HD) <= pen(DD) <= PEN_MAX
*                              LOD score versus unlinked disease is computed and reported.
*
*       Comments        Preliminary tests have indicated that 'Best' method outperforms other methods
*                       in terms power to detect disease gene for complex 2 and more locus disease tested.
*                       Double and Full methods has showed similar performance and are slightly better
*                       than 'Best' on single locus intermediate models.
*
*                       The results for 'Best' method were reported in
*                       Fishelson, Rusakov, Geiger "A Bayesian LOD score for Linkage Analysis of Complex Deseases"
*
******************************************************************************/
void OptimizeMarginalLikelihood()
{
    //OptimizeDoubleAvMarginalLikelihood();
    //printf("\n");
    //OptimizeFullAvMarginalLikelihood();
    //printf("\n");
    OptimizeBestMarginalLikelihood();
}

/******************************************************************************
*
*       Functions       OptimizeBestMarginalLikelihood
*       Parameters      none
*       Output          Print various info in stdout.
*       Description     see description for OptimizeMarginalLikelihood
*       Comments        
*
******************************************************************************/
void OptimizeBestMarginalLikelihood()
{
	scanIntervalData	*scanData;
	double				intervalSize, currPos;
    double              bestPos,bestLOD,theta;
    double              bestPosR,bestLODR;
    double              bestPosD,bestLODD;
    double              ln_appBaseR,ln_appBaseD;
    double              ln_appR,ln_appD,lod;
    double              ftol;
    char*               model_name;

    /* Evaluate likelihood for unlinked disease gene */
    currPos = -UNLINKED_DIST;
    ln_appBaseR = evaluateLnMarginalLikelihood(currPos,TRUE, PEN_MIN,PEN_MAX,PRIOR_UNIFORM);
    ln_appBaseD = evaluateLnMarginalLikelihood(currPos,FALSE,PEN_MIN,PEN_MAX,PRIOR_UNIFORM);
    printf("BaseLine: rec = %.2f,  dom = %.2f\n",ln_appBaseR,ln_appBaseD);

    /* Read scan interval data */
    scanData = progData.scanData;
	intervalSize = *(scanData->distFromMarker1 + scanData->numIntervals);
    if ( scanData->startPos == scanData->endPos ) {
        currPos = 0;
        intervalSize += MAX(scanData->offEndDistRight,scanData->offEndDistLeft);
    }
    else {
        currPos = -scanData->offEndDistLeft;
        intervalSize += scanData->offEndDistRight;
    }

    /* Search for most likely gene position in recessive and dominant models, and select the best model */
    printf("OptimizeBestMarginalLikelihood: model search in interval [%.2f, %.2f]\n",currPos,currPos+intervalSize);
    ftol = ((progData.progSpecific).optlinkData).searchTol;
    bestLODR = search_max(evaluate_rec_marginal_lod,currPos,currPos+intervalSize,ftol,SEARCH_XTOL,&ln_appBaseR,&bestPosR);
    bestLODD = search_max(evaluate_dom_marginal_lod,currPos,currPos+intervalSize,ftol,SEARCH_XTOL,&ln_appBaseD,&bestPosD);
    if (bestLODR < bestLODD ) {
        bestLOD = bestLODD;
        bestPos = bestPosD;
        model_name = "dom";
    }
    else {
        bestLOD = bestLODR;
        bestPos = bestPosR;
        model_name = "rec";
    }

    /* Report results */
    theta = getcMToRec(bestPos,progData.haldane);
    printf("Most likely best_position: %.2f cM, theta = %.3f,  cLOD = %.2f %s\n",bestPos,theta,bestLOD-MAXLOD_CORRECTION,model_name);

    /* Evaluate test position */
    currPos = TEST_POSITION;
    theta = getcMToRec(currPos,progData.haldane);
    ln_appR = evaluateLnMarginalLikelihood(currPos,TRUE, PEN_MIN,PEN_MAX,PRIOR_UNIFORM);
    ln_appD = evaluateLnMarginalLikelihood(currPos,FALSE,PEN_MIN,PEN_MAX,PRIOR_UNIFORM);
    if ( ln_appR - ln_appBaseR < ln_appD - ln_appBaseD ) {
        lod = (ln_appD - ln_appBaseD)/log(10);
        model_name = "dom";
    }
    else {
        lod = (ln_appR - ln_appBaseR)/log(10);
        model_name = "rec";
    }
    printf("test_position: %.2f cM, theta = %.3f, cLOD = %.2f %s\n",currPos,theta,lod-MAXLOD_CORRECTION,model_name);
    
	myFree(thetaMarkers);
    return;
}


/******************************************************************************
*
*       Functions       Optimize{Double,Full}MarginalLikelihood functions
*       Parameters      none
*       Output          Print various info in stdout.
*       Description     see description for OptimizeMarginalLikelihood
*       Comments        
*
******************************************************************************/
void OptimizeDoubleAvMarginalLikelihood()
{
	scanIntervalData	*scanData;
	double				intervalSize, currPos;
    double              bestPos,bestLOD,theta;
    marginal_data       mdata;
    double              ln_appR,ln_appD,ln_app;
    double              ftol;

    currPos = -UNLINKED_DIST;
    mdata.ln_appBaseR = evaluateLnMarginalLikelihood(currPos,TRUE, PEN_MIN,PEN_MAX,PRIOR_UNIFORM);
    mdata.ln_appBaseD = evaluateLnMarginalLikelihood(currPos,FALSE,PEN_MIN,PEN_MAX,PRIOR_UNIFORM);
    /* averaging in ln-space */
    mdata.ln_appBase = mdata.ln_appBaseR + log( (1+exp(mdata.ln_appBaseD-mdata.ln_appBaseR))/2 );

    printf("BaseLine: rec = %.3e,  dom = %.3e, ",mdata.ln_appBaseR,mdata.ln_appBaseD);
    printf("double average = %.3e\n",mdata.ln_appBase);

	scanData = progData.scanData;

	intervalSize = *(scanData->distFromMarker1 + scanData->numIntervals);
    if ( scanData->startPos == scanData->endPos ) {
        currPos = 0;
        intervalSize += MAX(scanData->offEndDistRight,scanData->offEndDistLeft);
    }
    else {
        currPos = -scanData->offEndDistLeft;
        intervalSize += scanData->offEndDistRight;
    }
    printf("OptimizeDoubleAvMarginalLikelihood: search in interval [%.2f, %.2f]\n",currPos,currPos+intervalSize);
    ftol = ((progData.progSpecific).optlinkData).searchTol;
    bestLOD = search_max(evaluate_double_marginal_lod,currPos,currPos+intervalSize,ftol,SEARCH_XTOL,&mdata,&bestPos);

    theta = getcMToRec(bestPos,progData.haldane);
    printf("Most likely double_position: %.2f cM, theta = %.3f  dbl_av_LOD = %.2f\n",bestPos,theta,bestLOD);
    
    currPos = TEST_POSITION;
    theta = getcMToRec(currPos,progData.haldane);
    ln_appR = evaluateLnMarginalLikelihood(currPos,TRUE, PEN_MIN,PEN_MAX,PRIOR_UNIFORM);
    ln_appD = evaluateLnMarginalLikelihood(currPos,FALSE,PEN_MIN,PEN_MAX,PRIOR_UNIFORM);
    ln_app = ln_appR + log( (1+exp(ln_appD-ln_appR))/2 ); /* averaging in ln-space */
    printf("test_position: %.2f cM, theta = %.3f, dbl_av_LOD = %.2f\n",currPos,theta,(ln_app - mdata.ln_appBase)/log(10));

    myFree(thetaMarkers);
    return;
}

void OptimizeFullAvMarginalLikelihood()
{
	scanIntervalData	*scanData;
	double				intervalSize, currPos;
    double              appBase,app,ftol;
    double              bestPos,bestLOD,theta;

    /* TODO(16/02/04):  rewrite this function to support near-zero cases.
                        Similar to integrations in Optimize{Best,Double}... */

    /* Evaluation of base (unlinked) values */
    currPos = -UNLINKED_DIST;
    appBase = dblquad_gaussian(evaluate_at_penDD_penHD_scaled,PEN_MIN,PEN_MAX,PEN_MIN,PEN_MAX,QUAD_TOL,&currPos);
    printf("BaseLine: full average = %.3e\n",appBase);

	scanData = progData.scanData;

	intervalSize = *(scanData->distFromMarker1 + scanData->numIntervals);
    if ( scanData->startPos == scanData->endPos ) {
        currPos = 0;
        intervalSize += MAX(scanData->offEndDistRight,scanData->offEndDistLeft);
    }
    else {
        currPos = -scanData->offEndDistLeft;
        intervalSize += scanData->offEndDistRight;
    }
    printf("OptimizeFullAvMarginalLikelihood:  search in interval [%.2f, %.2f]\n",currPos,currPos+intervalSize);
    ftol = ((progData.progSpecific).optlinkData).searchTol;
    bestLOD = search_max(evaluate_full_marginal_lod,currPos,currPos+intervalSize,ftol,SEARCH_XTOL,&appBase,&bestPos);

    theta = getcMToRec(bestPos,progData.haldane);
    printf("Most likely full_position: %.2f cM, theta = %.3f  full_av_LOD = %.2f\n",bestPos,theta,bestLOD);

    currPos = TEST_POSITION;
    theta = getcMToRec(currPos,progData.haldane);
    app = dblquad_gaussian(evaluate_at_penDD_penHD_scaled,PEN_MIN,PEN_MAX,PEN_MIN,PEN_MAX,QUAD_TOL,&currPos);
    printf("test_position: %.2f cM, theta = %.3f, full_av_LOD = %.2f\n",currPos,theta,log10(app/appBase));

    myFree(thetaMarkers);
    return;
}


/******************************************************************************
*
*       Function        evaluateLnMarginalLikelihood
*       Parameters      loc         -   Location of the disease gene on the map.
*                                       Relative to the scanData->startPos,endPos. 
*                                       in centiMorgans, values >= UNLINKED_DIST are considered unlinked.
*                       IsRecessive -   Penetrance model.
*                                       Recessive (true) or dominant (false).
*                       min_pen,max_pen -   Penetrances range for current model.
*                       prior       -   Penetrance prior: PRIOR_NONE, PRIOR_UNIFORM.
*       Output          Natural logarithm of integral of likelihood function, for penetrance varing
*                       in [min_pen,max_pen].
*       Description     Evaluates the desired integral by function specified in MARGINAL_INTEGRAL
*                       with relative error of REL_ERROR.
*                       Treats near zero cases by rescaling all the likelihood values by
*                       data.rescale_factor_ln. The value of 'rescale_factor_ln' is determined in
*                       the first call to 'evaluate_at_penetrance_adaptive' function.
*       Comments        Use PRIOR_UNIFORM to make averaging.
*                       Use loc = -UNLINKED_DIST to evaluate "base marginal likelihood",
*                       i.e., marg. likelihood for disease gene independent of markers.
*                       For debugging: Note that before 16/02/04 this function returned integral
*                       value itself and not the natural logarithm of integral.
*
******************************************************************************/
double evaluateLnMarginalLikelihood(double loc, boolean isRecessive, double min_pen, double max_pen, int prior)
{
    likelihood_data data;
    double app,ln_app;

    assert(min_pen<=max_pen);
    if (min_pen==max_pen) {
        /* dummy marginal likelihood evaluation. */
        updateTransmissionModel(isRecessive,min_pen);
        ln_app = evaluateLike(&loc,1,getLnLike,FALSE);
        return ln_app;
    }
    
    data.isRecessive = isRecessive;
    data.location = loc;
    /* using the evaluate_at_penetrance_adaptive() function in order to deal with near-zero situations. */
    data.rescale_factor_ln = INITIAL_RESCALE_FACTOR_LN;
    app = MARGINAL_INTEGRAL(evaluate_at_penetrance_adaptive,min_pen,max_pen,REL_ERROR,&data);
    if (prior == PRIOR_UNIFORM)
        app /= (max_pen-min_pen);
    else {
        /* other priors are not supported */
        assert(prior==PRIOR_NONE);
    }

    ln_app = log(app) + data.rescale_factor_ln;
    return ln_app;
}



/******************************************************************************
*
*       Function        updateTransmissionModelV2
*       Parameters		penHH, penHD, penDD - penetrances.
*       Output          none
*       Description		Updates the penetrances of the current model.
*                       Generalization of updateTransmissionModel from LociData.c
*                       allows specification of arbitrary transmission model.
*       Comments		modeled after updateTransmissionModel function from LociData.c
*
******************************************************************************/
void updateTransmissionModelV2(double penHH, double penHD, double penDD)
{
	femalePenClass	*fPenetrances;
	malePenClass    *mPenetrances;
	int				count;

	fPenetrances = loci.diseaseModel.oneLocus.fPenetrances;

	/* 1.	Iterate over the different liability classes and update the penetrances
			according to either a dominant or a recessive disease model, with 
			penetrances of 50%. */

	for (count = 0; count < loci.numClasses; count++) {	
			updateThetaCompTheta(fPenetrances[count][0], penHH);
			updateThetaCompTheta(fPenetrances[count][1], penHD);
			updateThetaCompTheta(fPenetrances[count][2], penDD);
		if (loci.sexLink) {
            assert(0); /* NOT IMPLEMENTED YET, as of 10/02/04 */
			mPenetrances = loci.diseaseModel.oneLocus.mPenetrances;
			updateThetaCompTheta(mPenetrances[count][0], 0.0);
			updateThetaCompTheta(mPenetrances[count][1], 0.5);
		}
	}
}



/******************************************************************************
*
*           PUBLIC INTEGRATION (QUADRATURE) FUNCTIONS
*
******************************************************************************/


/******************************************************************************
*
*       Function        quad_trapezoid
*       Parameters      fun -   the integrand function.
*                               Receives (x, func_data) as input.
*                       xmin, xmax - integration limits.
*                       rtol - relative error tolerance. (ignored)
*                       func_data - additional parameters to 'fun' function.
*       Output          integral value
*       Description     The function approximates the integral of function 'fun'
*                       in interval [xmin,xmax] by the trapezoidal method.
*       Comments        Trapezoidal rule is based on 10 points, rtol value ignored.
*
******************************************************************************/
double quad_trapezoid(functype fun,double xmin,double xmax,double rtol, void* func_data)
{
    double ival,fval,fvalprev,x,xprev,step;

    ival = 0;
    step = (xmax-xmin)/10;
    xprev = xmin;
    fvalprev = (*fun)(xprev,func_data);
#ifdef _DEBUG
    printf("quad_trapezoid on [%f %f], step = %f\nfun(%.2f) = %e\n",xmin,xmax,step,xmin,fvalprev);
#endif
    for( x = xmin+step; x <= xmax+step/2; x += step ) {
        fval = (*fun)(x,func_data);
        ival += (fvalprev+fval)/2;
#ifdef _DEBUG
        printf("fun(%.2f) = %e,  int = %e\n",x,fval,step*(fvalprev+fval)/2);
#endif
        xprev = x;
        fvalprev = fval;
    }
    ival *= step;
#ifdef _DEBUG
    printf("[%.3f %.3f], integral = %.3e\n",xmin,xmax,ival);
#endif

    return ival;
}


/******************************************************************************
*
*       Function        quad_adaptive_simpson_step
*       Parameters      fun -   the integrand function.
*                               Receives (x, func_data) as input.
*                       a,b,c - integration limits and midpoint of integration,
*                               i.e., c = (a+b)/2;
*                       a_val,b_val,c_val - precomputed function values in a,b,c.
*                       tol - maximum absolute error tolerance.
*                       func_data - additional parameters to 'fun' function.
*                       level - recursion level.
*                               Max recursion depth is set by RECURSION_LIMIT.
*       Output          integral value
*       Description     The function performs one step approximation of the integral
*                       of function 'fun' in interval [a,b] by the adaptive Simpson method.
*       Comments        Algorithm adopted from Burden and Faires "Numerical Analysis", p 218.
*
******************************************************************************/
double quad_adative_simpson_step(functype fun,double a,double b, double c, double a_val, double b_val, double c_val, double tol, void* func_data, int level)
{
    double d,e,d_val,e_val,h,s,s1,s2;

    h = c-a;
    d = a+h/2;
    e = a+3*h/2;
    d_val = (*fun)(d,func_data);
    e_val = (*fun)(e,func_data);
    s = h*(a_val+4*c_val+b_val)/3;
    s1 = h*(a_val+4*d_val+c_val)/6;
    s2 = h*(c_val+4*e_val+b_val)/6;
    /*printf("fun(a=%.2f) = %.6e\n",a,a_val);
    printf("fun(d=%.2f) = %.6e\n",d,d_val);
    printf("fun(c=%.2f) = %.6e\n",c,c_val);
    printf("fun(e=%.2f) = %.6e\n",e,e_val);
    printf("fun(b=%.2f) = %.6e\n",b,b_val);
    printf("s1+s2-s = %.3e,  tol = %.3e\n",s1+s2-s,tol);*/
    /*printf("4d+4e-a-6c-b = %.3e,  6*tol*h = %.3e\n",4*d_val+4*e_val-a_val-6*c_val-b_val,6*tol*h);*/

    if (s1+s2-s>tol || s1+s2-s < -tol) {
        if (level >= RECURSIVE_LIMIT) {
            //printf("warning: quad_adaptive_simpson: recursive limit %d reached on [%f,%f].\n",RECURSIVE_LIMIT,a,b);
        }
        else {      
            s1 = quad_adative_simpson_step(fun,a,c,d,a_val,c_val,d_val,tol/2,func_data,level+1);
            s2 = quad_adative_simpson_step(fun,c,b,e,c_val,b_val,e_val,tol/2,func_data,level+1);
        }
    }
    /*
    else
        printf("[%.3f,%.3f], level = %d, int = %.3e, s = %.3e, tol = %.3e\n",a,b,level,s1+s2,s,tol);
        */
    return s1+s2;
}

/******************************************************************************
*
*       Function        quad_adaptive_simpson
*       Parameters      fun -   the integrand function.
*                               Receives (x, func_data) as input.
*                       xmin, xmax - integration limits.
*                       rtol - relative error tolerance.
*                       func_data - additional parameters to 'fun' function.
*       Output          integral value
*       Description     The function approximates the integral of function 'fun'
*                       in interval [xmin,xmax] by the adaptive Simpson method.
*       Comments        Algorithm adopted from Burden and Faires "Numerical Analysis", p 218.
*
******************************************************************************/
double quad_adaptive_simpson(functype fun,double xmin,double xmax,double rtol, void* func_data)
{
    double c,a_val,b_val,c_val,app;

    c = (xmax+xmin)/2;
    a_val = (*fun)(xmin,func_data);
    b_val = (*fun)(xmax,func_data);
    c_val = (*fun)(c,func_data);
    app = (a_val+4*c_val+b_val)*(xmax-xmin)/6;

    /*    printf("quad_adaptive_simpson on [%.3f,%.3f], initial app = %.3e\n",xmin,xmax,app); */
    app = quad_adative_simpson_step(fun,xmin,xmax,c,a_val,b_val,c_val,10*rtol*app,func_data,1);
    /*printf("quad_adaptive_simpson on [%.3f,%.3f], final app = %.3e, rtol = %.3f%%\n",xmin,xmax,app,rtol*100);*/

    return app;
}

/******************************************************************************
*
*       Function        quad_gaussian_n5
*       Parameters      fun -   the integrand function.
*                               Receives (x, func_data) as input.
*                       xmin, xmax - integration limits.
*                       rtol - relative error tolerance. IGNORED.
*                       func_data - additional parameters to 'fun' function.
*       Output          integral value
*       Description     The function approximates the integral of function 'fun'
*                       in interval [xmin,xmax] by Gaussian quadrature method based
*                       on 5 points.
*       Comments        Algorithm adopted from Burden and Faires "Numerical Analysis", pp 222-8.
*
******************************************************************************/
double quad_gaussian_n5(functype fun, double xmin, double xmax, double rtol, void *func_data)
{
    int i;
    double x,val,app;
    double r5[5] = {0.9061798459, 0.5384693101, 0.0, -0.5384693101, -0.9061798459};
    double c5[5] = {0.2369268850, 0.4786286705, 0.5688888889, 0.4786286705, 0.2369268850};

    app = 0.0;
    for (i=0; i<5; i++ ) {
        x = (r5[i]*(xmax-xmin)+xmax+xmin)/2.0;
        val = (*fun)(x,func_data);
        app += val*c5[i]*(xmax-xmin)/2;
    }
    return app;
}



/******************************************************************************
*
*       Double Quadrature functions.
*
******************************************************************************/
typedef struct dblquad_internal_integral_data_tag {
    functype2   fun;
    double      ymin;
    double      ymax;
    double      rtol;
    void        *func_data;
    quadfunc    quad;
    /* data for dblquad_general2() function */
    functype    ymin_fun;
    functype    ymax_fun;
} dblquad_internal_integral_data;

typedef struct dblquad_internal_func_data_tag {
    functype2   fun;
    double      x;
    void        *func_data;
} dblquad_internal_func_data;


/******************************************************************************
*
*       Function        dblquad_internal_func
*       Parameters      y -     the y value.
*                       data -  additional function data,
*                               casted to type dblquad_internal_func_data
*       Output          value of function data->fun at (data->x,y,data->func_data)
*       Description     The function simulates the 1-d function for dblquad_internal_integral
*                       while returning the values of 2-d function data->fun for
*                       a fixed x (data->x).
*       Comments        none
*
******************************************************************************/
double dblquad_internal_func(double y, void* data)
{
    dblquad_internal_func_data *internal_data;

    internal_data = (dblquad_internal_func_data*)data;
    return (*(internal_data->fun))(internal_data->x,y,internal_data->func_data);
}


/******************************************************************************
*
*       Function        dblquad_internal_integral
*       Parameters      x -     the x value.
*                       data -  additional function data,
*                               casted to type dblquad_internal_integral_data
*       Output          Integral value for fixed x and variable y in range
*                       [data->ymin,data->ymax] with tolerance data->rtol.
*                       Integrand function is given by data->fun.
*       Description     The function approximates the integral of function 'data->fun'
*                       for a given fixed x and y varied in [data->ymin,data->ymax].
*                       The intenral is approximated by the data->quad method.
*       Comments        none.
*
******************************************************************************/
double dblquad_internal_integral(double x, void* data)
{
    double ymin,ymax;
    dblquad_internal_integral_data *ii_data;
    dblquad_internal_func_data if_data;

    ii_data = (dblquad_internal_integral_data*)data;
    if_data.fun = ii_data->fun;
    if_data.func_data = ii_data->func_data;
    if_data.x = x;
    if ( ii_data->ymin_fun == NULL ) {
        assert(ii_data->ymax_fun==NULL);
        ymin = ii_data->ymin;
        ymax = ii_data->ymax;
    }
    else {
        ymin = (*(ii_data->ymin_fun))(x,ii_data->func_data);
        ymax = (*(ii_data->ymax_fun))(x,ii_data->func_data);
    }
    return (*(ii_data->quad))(dblquad_internal_func,ymin,ymax,ii_data->rtol,&if_data);
}


/******************************************************************************
*
*       Function        dblquad_general
*       Parameters      fun -   the integrand function.
*                               Receives (x, y, func_data) as input.
*                       xmin, xmax, ymin, ymax - integration limits.
*                       rtol -  relative error tolerance.
*                       func_data - additional parameters to 'fun' function.
*                       quad -  1-d integration function.
*                               If quad==NULL, then the default 'quad_adaptive_simpson' is used.
*       Output          integral value.
*       Description     The function approximates the integral of function 'fun'
*                       in domain [xmin,xmax]\times[ymin,ymax] by recursively
*                       approximating the corresponding 1-d integrals by 'quad' function.
*       Comments        If 'quad' function is not specified. I.e., quad==NULL,
*                       then 'quad_adaptive_simpson' is used.
*
******************************************************************************/
double dblquad_general(functype2 fun, double xmin, double xmax, double ymin, double ymax, double rtol, void* func_data, quadfunc quad)
{
    dblquad_internal_integral_data ii_data;
    double rtol2,app;

    if (quad==NULL) 
        quad = quad_adaptive_simpson;

    rtol2 = sqrt(1+rtol)-1;
    ii_data.fun = fun;
    ii_data.func_data = func_data;
    ii_data.quad = quad;
    ii_data.rtol = rtol2;
    ii_data.ymin = ymin;
    ii_data.ymax = ymax;
    ii_data.ymin_fun = NULL;
    ii_data.ymax_fun = NULL;
    app = (*quad)(dblquad_internal_integral,xmin,xmax,rtol2,&ii_data);

#ifdef _DEBUG
    printf("dblquad_general on [%.3f,%.3f]x[%.3f,%.3f], final app = %.3e, rtol2 = %.3f%%, rtol = %.3f%%\n",xmin,xmax,ymin,ymax,app,rtol2*100,rtol*100);
#endif
    return app;
}


/******************************************************************************
*
*       Function        dblquad_gaussian
*       Parameters      fun -   the integrand function.
*                               Receives (x, y, func_data) as input.
*                       xmin, xmax, ymin, ymax - integration limits.
*                       rtol -  relative error tolerance. (ignored)
*                       func_data - additional parameters to 'fun' function.
*       Output          integral value.
*       Description     The function approximates the integral of function 'fun'
*                       in domain [xmin,xmax]\times[ymin,ymax] by gaussian double
*                       integral method, on grid 5 by 5.
*       Comments        rtol value is ignored.
*                       The algorithm is adopted from Burden and Faires "Numerical Analysis", p 237.
*
******************************************************************************/
double dblquad_gaussian(functype2 fun, double xmin, double xmax, double ymin, double ymax, double rtol, void* func_data)
{
    double h1,h2,k1,k2,app,x,y,fval,app2;
    int i,j;
    double r5[5] = {0.9061798459, 0.5384693101, 0.0, -0.5384693101, -0.9061798459};
    double c5[5] = {0.2369268850, 0.4786286705, 0.5688888889, 0.4786286705, 0.2369268850};

    h1 = (xmax-xmin)/2;
    h2 = (xmax+xmin)/2;
    k1 = (ymax-ymin)/2;
    k2 = (ymax+ymin)/2;
    app = 0;

    for (i=1; i<=5; i++ ) {
        x = h1*r5[i-1]+h2;
        app2 = 0;
        for (j=1; j<=5; j++ ) {
            y = k1*r5[j-1]+k2;
            fval = (*fun)(x,y,func_data);
            app2 += c5[j-1]*fval;
        }
        app += c5[i-1]*k1*app2;
    }
    app *= h1;

/*    printf("dblquad_gaussian on [%.3f,%.3f]x[%.3f,%.3f], final app = %.3e\n",xmin,xmax,ymin,ymax,app);*/

    return app;
}

/******************************************************************************
*
*       Function        dblquad_general2
*       Parameters      fun -   the integrand function.
*                               Receives (x, y, func_data) as input.
*                       xmin, xmax - integration limits on x.
*                       ymin_fun, ymax_fun - integration limits on y, as function of x.
*                       rtol -  relative error tolerance.
*                       func_data - additional parameters to 'fun', 'ymin_fun'
*                                   and 'ymax_fun' functions.
*                       quad -  1-d integration function.
*                               If quad==NULL, then the default 'quad_adaptive_simpson' is used.
*       Output          integral value.
*       Description     The function approximates the integral of function 'fun'
*                       in domain [xmin,xmax] \times [ ymin_fun, ymax_fun] by recursively
*                       approximating the corresponding 1-d integrals by 'quad' function.
*       Comments        If 'quad' function is not specified. I.e., quad==NULL,
*                       then 'quad_adaptive_simpson' is used.
*
******************************************************************************/
double dblquad_general2(functype2 fun, double xmin, double xmax, functype ymin_fun, functype ymax_fun, double rtol, void* func_data, quadfunc quad)
{
    dblquad_internal_integral_data ii_data;
    double rtol2,app;

    if (quad==NULL) 
        quad = quad_adaptive_simpson;

    rtol2 = sqrt(1+rtol)-1;
    ii_data.fun = fun;
    ii_data.func_data = func_data;
    ii_data.quad = quad;
    ii_data.rtol = rtol2;
    ii_data.ymin = 0;
    ii_data.ymax = 0;
    ii_data.ymin_fun = ymin_fun;
    ii_data.ymax_fun = ymax_fun;
    app = (*quad)(dblquad_internal_integral,xmin,xmax,rtol2,&ii_data);

    /*printf("dblquad_general on [%.3f,%.3f]x[%.3f,%.3f], final app = %.3e, rtol2 = %.3f%%, rtol = %.3f%%\n",xmin,xmax,ymin,ymax,app,rtol2*100,rtol*100);*/

    return app;
}
/* common boundary definition functions, for dblquad_general2 */
double return_zero(double x, void* data)
{
    return 0.0;
}
double return_x(double x, void* data)
{
    return x;
}


/******************************************************************************
*
*       Function        search_max
*       Parameters      func -  function to be optimized.
*						xmin, xmax - interval in which to search for maximum.
*                       ftol -  tolerance in function value.
*                               search is stopped in difference in function values
*                               in search for maximum is less than ftol.
*						xtol -  tolerance in x.
*                               \argmax(fun) is located within tol from x_opt.
*						func_data - additional data to be passed to fun.
*						x_opt - returned approximation of \argmax(fun)
*       Output          fun(x_opt). Maximum found value of 'fun'.
*       Description		Searches for maximum of 'fun'.
*                       Method: scan interval first with SCAN_STEPS steps and
*                               then pass the maximum region to 'golden' search
*                               function from Optimize.h
*       Comments		none.
*
******************************************************************************/
double search_max(functype func, double xmin, double xmax, double ftol, double xtol, void* func_data, double* x_opt)
{
    int i,max_i,scan_steps;
    double x,x_step,val;
    double vals[SCAN_STEPS+1];
    double a,b,c;

    assert(xtol>0);
    assert(ftol>=0);
    assert(SCAN_STEPS>=3);
    scan_steps = MIN((int)ceil((xmax-xmin)/xtol),SCAN_STEPS);
    x_step = (xmax-xmin)/scan_steps;
    max_i = 0;
    *x_opt = xmin;
    for ( x=xmin, i=0; i<=scan_steps; i++, x+=x_step ) {
        /* checking upper bound, fearing for numerical instability */
        if ( x > xmax )  x = xmax;
        vals[i] = (*func)(x,func_data);
        if ( vals[i] > vals[max_i] ) {
            max_i = i;
            *x_opt = x;
        }
    }
    if ( x_step<=xtol ) return vals[max_i];
    assert(scan_steps==SCAN_STEPS);
    if ( max_i==0 ) {
        a = *x_opt;
        b = *x_opt + x_step;
        c = *x_opt + 2*x_step;
    }
    else if ( max_i==scan_steps ) {
        a = *x_opt - 2*x_step;
        b = *x_opt - x_step;
        c = *x_opt;
    }
    else {
        a = *x_opt - x_step;
        b = *x_opt;
        c = *x_opt + x_step;
    }
    /*printf("a = %.f2, b = %.2f, c = %.2f\n",a,b,c);*/
    val = goldenV2(func,a,b,c,ftol,xtol,func_data,x_opt);
    return val;
}

/******************************************************************************
*
*       function       goldenV2, a generalized copy of golden() from optimize.c
*       parameters     func - pointer to the function which is evaluated in 
*					          each position on the grid,
*                      ax, bx, cx - a bracketing triplet of abscissas,
*                                   such that bx is between ax and cx,
*					                and f(bx) is greater than both f(ax) and f(cx)),
*                      ftol - the required tolerance in function value,
*                      xtol - the required tolerance in x,
*                      funcData - additional data which needs to be passed to
*                                 the evaluated function,
*                      x_opt - is updated with the abscissa of the maximum.
*       output  	   returns the maximum function value.
*                      (the optimal value of x is returned in 'x_opt').
*       description    performs a golden section search for the maximum, 
*					   isolating it to a fractional precision of about 'xtol'.
*       comments       none
*
******************************************************************************/
double goldenV2(functype func, double ax, double bx, double cx, double ftol, double xtol, void *funcData, double *x_opt)
{	
	double		f1, f2, x0, x1, x2, x3;
    /*printf("input: ax = %.2f, bx = %.2f, cx = %.2f, tol = %.2f\n",ax,bx,cx,tol);*/

	/* 1.	At any given time, the algorithm keeps track of 4 points: 
			x0, x1, x2, & x3. */
	x0 = ax;
	x3 = cx;

	/* 2.	Make x0 to x1 the smaller segment, and fill in the new point to
			be tried. */
	if (fabs(cx - bx) > fabs(bx - ax)) {
		x1 = bx;
		x2 = bx + C * (cx - bx);
	}
	else {
		x2 = bx;
		x1 = bx - C * (bx - ax);
	}

	/* 3.	Evaluate the function at these two points. */
	f1 = (*func)(x1, funcData);
	f2 = (*func)(x2, funcData);

	/* 4.	Keep going as long as the required precision hasn't been met. */
	while ( (fabs(x3 - x0) > xtol) && (fabs(f1 - f2) > ftol) ) {
        /*printf("x0 = %.2f, x1 = %.2f, x2 = %.2f, x3 = %.2f\n",x0,x1,x2,x3);*/

		/* 4.1.	Determine which three points to keep for the next step. */
		if (f2 > f1) {
			/* 4.1.1.	Get rid of the first point. */	
			shift3(x0, x1, x2, R * x1 + C * x3);
			shift2(f1, f2, (*func)(x2, funcData));
		}
		else {
			/* 4.1.1.	Get rid of the last point. */
			shift3(x3, x2, x1, R * x2 + C * x0);
			shift2(f2, f1, (*func)(x1, funcData));
		}
	}

	/* 5.	The required precision has been reached. Output the better of the two
			current values. */
	if (f1 > f2) {
		*x_opt = x1;
		return f1;
	}
	else {
		*x_opt = x2;
		return f2;
	}
}



/******************************************************************************
*
*           INTERNAL FUNCTIONS
*
******************************************************************************/


/******************************************************************************
*
*       Functions       evaluate_at_penetrance_adaptive
*       Parameters      pen - penetrance value.
*                       func_data - pointer to likelihood_data structure.
*       Output          Likelihood of data for given penetrance (pen and func_data->isRecessive)
*                       and assumed disease positions func_data->location
*       Description     The function works in couple with evaluateLnMarginalLikelihood.
*                       initially, before the start of integration 
*                                  the value of func_data->rescale_factor_ln is set
*                                  to INITIAL_RESCALE_FACTOR in evaluateLnMarginalLikelihood
*                       in the first call, this function updates func_data->rescale_factor_ln
*                                  to the current ln_likelihood value, and
*                                  in this way it resets the operational range of integration,
*                                  keeping the system from underflowing.
*                       the subsequent calls, modify the returned function value by substracting
*                                  the func_data->rescale_factor_ln from the ln-likelihood approximation.   
*       Comments        The value of INITIAL_RESCALE_FACTOR should be positive ineteger > 1,
*                       To ensure that no likelihood computation can occasionaly return the same value.
*
******************************************************************************/
double evaluate_at_penetrance_adaptive(double pen,void *func_data)
{
    likelihood_data *data;
    double ln_app,app;

    data = (likelihood_data*)func_data;
    updateTransmissionModel(data->isRecessive,pen);
    ln_app = evaluateLike(&(data->location),1,getLnLike,FALSE);
    if (data->rescale_factor_ln==INITIAL_RESCALE_FACTOR_LN) {
        data->rescale_factor_ln = ln_app;
        app = 1;
    }
    else {
        ln_app -= data->rescale_factor_ln;
        app = exp(ln_app);
    }
    return app;
}


/******************************************************************************
*
*       Functions       evaluate_at_penDD_penHD and evaluate_at_penDD_penHD_scaled
*       Parameters      penDD - penetrance for DD (double decease) gene.
*                       penHD - penetrance for HD (healthy/decease) gene,
*                       func_data - pointer to location (double).
*       Output          Likelihood of data for given penetrances and specified
*                       decease location (in *func_data).
*       Description     The function is adopted to OptimizeMarginalLikelihood
*                       integration of all HD and DD penetrances in for pen(HD)<pen(DD).
*       Comments        The second function is suited for use in double integration
*                       functions with standard range [0,1]^2, while the true integration
*                       range is desired to be 0<penHD<penDD<1.
*
******************************************************************************/
double evaluate_at_penDD_penHD(double penDD, double penHD, void *func_data)
{
    double ln_app;

    /* Controlling the input parameters,
       as of 30/12/03, evaluateLike works incorrectly with 0 and 1 penetrances. */
    /* still not fixed on 10/02/04. */
    if (penDD<0.001) penDD=0.001;
    if (penDD>0.999) penDD=0.999;
    if (penHD<0.001) penHD=0.001;
    if (penHD>0.999) penHD=0.999;
    updateTransmissionModelV2(0.0,penHD,penDD);
    ln_app = evaluateLike((double*)func_data,1,getLnLike,FALSE);
    return exp(ln_app);
}
double evaluate_at_penDD_penHD_scaled(double penDD, double penHD, void *func_data)
{
    /* This function supposes that penHD is given as fraction of penDD */
    /* Used to compute double integral evaluate_at_penDD_penHD in range 0<penHD<penDD<1
       while calling the standard integration for range [0,1]^2 */
    double val;
    penHD *= penDD;
    val = evaluate_at_penDD_penHD(penDD,penHD,func_data);
    /* correcting the output accordingly, to produce correct integration */
    return penDD*val;
}


/******************************************************************************
*
*       Functions       evaluate_*_marginal_lod
*       Parameters      loc - hypothetized location of disease gene.
*                       func_data - additional function data.
*                                   *func_data - marginal likelihood of unlinked disease.
*       Output          Lod score of marginal likelihood of disease being evaluated
*                       at specified location.
*       Description     full:   marginal integrated over all 0<penHD<pedDD<1.
*                       double: marginal is average of two integrals,
*                               recessive model: penHD=0; 0<penDD<1;
*                               and dominant model: 0<penHD=penDD<1;
*                       rec     recessive model: penHD=0, 0<penDD<1;
*                       dom     dominant model: 0<penHD=penDD<1;
*       Comments        none
*
*       TODO(16/02/04): rewrite these function and the branch of code that uses these functions,
*                       to use be able to detect near-zero cases and treat them correctly.
*
******************************************************************************/
double evaluate_full_marginal_lod(double loc, void* func_data)
{
    double app;
    app = dblquad_gaussian(evaluate_at_penDD_penHD_scaled,0,1,0,1,QUAD_TOL,&loc);
    printf( "pos = %.3f, full average = %.2e, LOD = %.2f\n",loc,app,log10(app/(*((double*)func_data))) );
    return log10( app/(*((double*)func_data)) );
}
double evaluate_double_marginal_lod(double loc, void* func_data)
{
    double ln_appR,ln_appD,ln_app,lodD,lodR;
    marginal_data* mdata;

    mdata = (marginal_data*)func_data;
    ln_appR = evaluateLnMarginalLikelihood(loc, TRUE,  PEN_MIN, PEN_MAX, PRIOR_UNIFORM);
	ln_appD = evaluateLnMarginalLikelihood(loc, FALSE, PEN_MIN, PEN_MAX, PRIOR_UNIFORM);
    ln_app = ln_appR + log( (1+exp(ln_appD-ln_appR))/2 ); /* averaging in ln-space */
    lodR = (ln_appR - mdata->ln_appBaseR)/log(10);
    lodD = (ln_appD - mdata->ln_appBaseD)/log(10);
    printf("pos = %.3f, rec_LOD = %.2f, dom_LOD = %.2f, dbl_av_LOD = %.2f\n",loc,lodR,lodD,(ln_app - mdata->ln_appBase)/log(10));
    return (ln_app - mdata->ln_appBase)/log(10);
}
double evaluate_rec_marginal_lod(double loc, void* func_data)
{
    double ln_app,lod;
    ln_app = evaluateLnMarginalLikelihood(loc, TRUE,  PEN_MIN, PEN_MAX, PRIOR_UNIFORM);
    lod = ( ln_app - (*((double*)func_data)) ) / log(10);
    printf( "pos = %.3f, recessive_model = %.2f, LOD = %.2f\n",loc,ln_app,lod);
    return lod;
}
double evaluate_dom_marginal_lod(double loc, void* func_data)
{
    double ln_app,lod;
    ln_app = evaluateLnMarginalLikelihood(loc, FALSE,  PEN_MIN, PEN_MAX, PRIOR_UNIFORM);
    lod = ( ln_app - (*((double*)func_data)) ) / log(10);
    printf( "pos = %.3f, dominant_model = %.2f, LOD = %.2f\n",loc,ln_app,lod);
    return lod;
}


/* OUTDATED CODE (16/02/04) - not used in mainstream computations
double evaluate_at_theta(double theta,void *func_data)
{
    double ln_app,loc;
    
    // I'm forced to use here global variable progData, to transform
    //   recombination theta into centiMorgans.
    loc = getRecTocM(theta,progData.haldane);
    ln_app = evaluateLike(&loc,1,getLnLike,FALSE);
    
    return exp(ln_app);
}


double evaluate_at_penetrance(double pen,void *func_data)
{
    likelihood_data *data;
    double ln_app;

    data = (likelihood_data*)func_data;
    updateTransmissionModel(data->isRecessive,pen);
    ln_app = evaluateLike(&(data->location),1,getLnLike,FALSE);
    return exp(ln_app);
}

double evaluate_at_theta_and_penetrance(double theta, double pen, void *func_data)
{
    double ln_app,loc;
    
    // I'm forced to use here global variable progData, to transform
    //   recombination theta into centiMorgans.
    loc = getRecTocM(theta,progData.haldane);
    updateTransmissionModel(*((boolean*)func_data),pen);
    ln_app = evaluateLike(&loc,1,getLnLike,FALSE);
    return exp(ln_app);
}
*/

/******************************************************************************
*
*    Debug functions.
*
*******************************************************************************/
double test_fun(double x, void* a)
{
    return sin(x);
}

double test_fun2(double x, double y, void* func_data)
{
    y = y*x;
    return (x*x+y*y)*x;
}

void dump_matlab_file(functype2 fun, double xmin, double xmax, int steps, char* name, void* func_data)
{
    FILE *mat_file;
    double step,x,y,val;
    int i,j;

    mat_file = fopen (name,"a");
    assert(mat_file);

    step = (xmax-xmin)/steps;
    for (x=xmin,i=0; i<=steps; x+=step, i++ ) {
        for (y=xmin,j=0; j<=steps; y+=step, j++ ) {
            val = (*fun)(x,y,func_data);
            fprintf(mat_file,"%e ",val);
        }
        fprintf(mat_file,"\n");
    }
    fclose(mat_file);
}
