/* $Id: PSGraph.c,v 1.2 2008/08/03 08:15:46 superlink Exp $ */
/*********************************************************************************/
/*  PSGraph.c												     			     */
/*  Author: Ma'ayan Fishelson					                                 */
/*********************************************************************************/

#include "PSGraph.h"


/******************************************************************************
*
*       Function        createPSGraph
*       Parameters		psFileName (the name of the postscript file where the  
*						graph should be drawn), graphTitle (the title of the 
*						graph), yTitle (the title of the Y axis), vals (the 
*						values of the points in the graph), numPoints (the 
*						number of points in the graph), markerNames (the names 
*						of the markers that appear in the graph - xLabels), 
*						numXLabels (the number of markers that appear in the 
*						graph),  mapDistArr (the distances between the markers   
*						that are to appear in the graph), drawBezCurve (a flag  
*						that indicates whether to draw a Bezier curve between 
*						the control points, or only connect the points with  
*						straight lines).
*       Output          none
*       Description		Creates a graph in a PS file according to the specified 
*						parameters.  
*       Comments		This graph contains the results of the likelihood 
*						computations.
*
******************************************************************************/
void createPSGraph(char *psFileName, char *graphTitle, char *yTitle, point *vals, 
				   int numPoints, xLabel *markerNames, int numXLabels, 
				   double *mapDistArr, boolean	drawBezCurve)
{
	boolean		zeroToOne;
    int			bottomY, topY, numYLabels, prevLabel=0, i;
    char		xTitle[20];
     double		minY, maxY, maxX, xOffset, xScale, yScale, tooClose, distFromPrevLabel, 
				*yAxisVals, *yAxisValsPtr, *newMapPos, currYVal, yInc;
	yLabel		*yAxisLabels, *yAxisLabelsPtr, *yAxisLabelsPtrEnd;
	xLabel		*xAxisLabels;
	point		*valsPtr, *valsPtrEnd;
	FILE		*psFile;


	/* 1.	Open the ps file for writing. */

	psFile = fopen(psFileName, "w");

	if (psFile == NULL) {
		error("File %s was not opened", psFileName);
	}

	/* 2.	Print starting comments and definitions. */

    psFileStart(psFile);

	/* 3.	Print the title of the graph, and the names of the input files. */

	fprintf(psFile,"%% Printing the title of the graph. \n");
    fprintf(psFile,"36 360 moveto\n");
	fprintf(psFile,"gsave 90 rotate 0 0 rmoveto\n");
	fprintf(psFile,"12 /Times-Bold SF (%s) show grestore\n", graphTitle);
	fprintf(psFile,"gsave 90 rotate 0 -14 rmoveto\n");
	fprintf(psFile,"10 /Times-Roman SF (Files: %s, %s) show grestore\n",
			pedfileName, datafileName);

   /* 4.	Calculate appropriate X-scale (according to length of 9 inches). */

	xOffset = (vals->x < 0) ? ((-1.0) * vals->x) : 0;
    maxX = vals[numPoints-1].x + xOffset;

    xScale = 648.0 / maxX;

    /* 5.	Find the minimum and the maximum y-values. */

	maxY = minY = 0.0;

	/* 5.1.	Define a Y-scale: Iterate over the different y values. */

	for (valsPtr = vals, valsPtrEnd = valsPtr + numPoints; 
			valsPtr < valsPtrEnd; valsPtr++) {

		/* 5.1.1.	If the y-value is not -Infinity: */

		if (valsPtr->y > MIN_DBL) { 

			/* 5.1.1.1.	Compare it to the maximum and minimum values found up to now,
						and update them if necessary. */
	
			if (valsPtr->y > maxY) {
				maxY = valsPtr->y;
			}
			else {
				if (valsPtr->y < minY) {
					minY = valsPtr->y;
				}
			}
		}
    }

    /* 6.	Calculate appropriate y-scale. */

    if (maxY <= 1.0 && minY >= 0.0) {

		/* 6.1.	All the y-values are between zero and one. */

 		zeroToOne = TRUE;

 		/* 6.2.	Set the top and bottom y to 1 and 0. */

        topY = 1; 
		bottomY = 0;

		/* 6.3.	The scale is set according to a length of 6 inches. */

		yScale = 432.0;			
    } 
	else {

		/* 6.1.	The y-values are not between zero and one. */

        zeroToOne = FALSE;

		/* 6.2.	Set the top and bottom y according to the y-values. */

		topY = (int)(maxY + 1.0);
		bottomY = (int)(minY - 1.0);

		/* 6.3.	The scale is set according to a length of 6 inches. */

		yScale = 432.0 / (topY - bottomY);	
    }

    /* 7.	Eliminate any off chart points (e.g.,-INFINITY LOD-scores). */

	for (valsPtr = vals, valsPtrEnd = valsPtr + numPoints; 
			valsPtr < valsPtrEnd; valsPtr++) {
		if (valsPtr->y < bottomY) {
			valsPtr->y = bottomY;
		}
	}

    /* 8.	Create Y-axis labels. */

	/* 8.1.	Determine the number of labels necessary, and the increment in the 
			y-value from one label to the next. */

	if (zeroToOne) { 
		numYLabels = 11;
		yInc = 0.1;
	}
	else {
		numYLabels = (topY - bottomY)*2 + 1;
		yInc = 0.5;
	}

	/* 8.2.	Allocate memory for arrays that would store the labels in string 
			format and in number format. */

	myMalloc(yAxisLabels, sizeof(yLabel) * numYLabels);
	myMalloc(yAxisVals, sizeof(double) * numYLabels);

	/* 8.3.	Update the arrays with the labels. */

	for (yAxisLabelsPtr = yAxisLabels, yAxisLabelsPtrEnd = yAxisLabelsPtr + numYLabels,
			yAxisValsPtr = yAxisVals, currYVal = bottomY; yAxisLabelsPtr < yAxisLabelsPtrEnd; 
			yAxisLabelsPtr++, yAxisValsPtr++, currYVal += yInc) {

		*yAxisValsPtr = currYVal;
		sprintf(*yAxisLabelsPtr,"%.1lf",*yAxisValsPtr);
	}

    /* 9.	Create X-axis labels. */

	/* 9.1.	Allocate memory for an array that would store the labels. */

	myMalloc(xAxisLabels, sizeof(xLabel) * numXLabels);

	/* 9.2.	Calculate a value which indicates when 2 labels are too close to each other. */

    tooClose = 6.0 / xScale;

	/* 9.3.	Iterate over the markers and create the labels. */

	distFromPrevLabel = 0.0;

    for (i = 0; i < numXLabels; i++) {

		/* 9.3.1.	If this isn't the first label and it is too close to the previous
					one, add the name of the current marker to the end of the previous label. */

        if ((i > 0) && (mapDistArr[i] + distFromPrevLabel < tooClose)) {
			strcat(xAxisLabels[prevLabel]," ");
			strcat(xAxisLabels[prevLabel], markerNames[i]);
			strcpy(xAxisLabels[i],"");
			distFromPrevLabel += mapDistArr[i];
		}
		else {
			strcpy(xAxisLabels[i], markerNames[i]);
			prevLabel = i;
			distFromPrevLabel = 0.0;
		}
    }

	/* 10.	Allocate memory for an array that would store the new positions of the 
			markers on the map relative to the offset in the x-axis. */

	myMalloc(newMapPos, sizeof(double) * numXLabels);

 	if (xOffset > 0.0) {
		newMapPos[0] = mapDistArr[0] + xOffset;
	}
	else {
		newMapPos[0] = mapDistArr[0];
	}

	for (i = 1; i < numXLabels; i++) {
		newMapPos[i] = mapDistArr[i] + newMapPos[i-1];

		if (newMapPos[i] > maxX) {
			numXLabels = i;
			break;
		}
	}

	sprintf(xTitle,"%.2lf cM", maxX);

	/* 11.	Draw the axes. */

	drawAxes(psFile, numXLabels, newMapPos, xAxisLabels, numYLabels, yAxisVals, 
				yAxisLabels, xTitle, yTitle, xScale, yScale);
	
	/* 12.	Update all the x-values with the offset in this axis. */

	if (xOffset > 0.0) {
		for (valsPtr = vals, valsPtrEnd = valsPtr + numPoints; valsPtr < valsPtrEnd;
				valsPtr++) {
			valsPtr->x += xOffset;
		}
	}

	/* 13.	Draw lines connecting between each 2 consecutive points. */

	drawCurve(psFile, vals, numPoints, SOLID_LINE, xScale, yScale);

	/* 14.	Draw the Bezier curve, if requested. */

	if (drawBezCurve) {
		drawBezier(psFile, vals, numPoints, 0.0, 0.0, DOTTED2, xScale, yScale);
	}

	/* 15.	Restore the x-values to their original value. */

    if (xOffset > 0.0){
		for (valsPtr = vals, valsPtrEnd = valsPtr + numPoints; valsPtr < valsPtrEnd;
				valsPtr++) {
			valsPtr->x -= xOffset;
		}
	}

	/* 16.	Print closing definitions, and close the file. */

    psPageEnd(psFile);
	fclose(psFile);

	/* 17.	Free the memory allocated dynamically. */

	myFree(yAxisLabels);
	myFree(yAxisVals);
	myFree(xAxisLabels);
	myFree(newMapPos);
}



/******************************************************************************
*
*       Function        psFileStart
*       Parameters		psFile (pointer to the ps file)
*       Output          none
*       Description		Prints starting comments and definitions to the ps file.
*       Comments		none.
*
******************************************************************************/
void psFileStart(FILE *psFile)
{
    fprintf(psFile,"%%!\n");
    fprintf(psFile,"%%%%Creator: SUPERLINK\n");
    fprintf(psFile,"%%%%BeginProlog\n");
	fprintf(psFile,"%%%%BeginResource: procset DrawingDict\n");
    fprintf(psFile,"/DrawingDict 200 dict def\n");
    fprintf(psFile,"DrawingDict begin\n");
	fprintf(psFile,"/setLW {setlinewidth} def	%% stack: line_width ==>\n");
	fprintf(psFile,"/SF {findfont exch scalefont setfont} def");	
	fprintf(psFile,"	%% stack: font_size font_type ==>\n");	
    fprintf(psFile,"end\n");
    fprintf(psFile,"%%%%EndResource\n");
    fprintf(psFile,"%%%%EndProlog\n");
    fprintf(psFile,"%%%%BeginSetup\n");
    fprintf(psFile,"DrawingDict begin\n");
    fprintf(psFile,"gsave\n");
    fprintf(psFile,"%%%%EndSetup\n");

	fprintf(psFile,"%%%%Page: ? 1\n");
    fprintf(psFile,"%%%%BeginPageSetup\n");
    fprintf(psFile,"save\n");
    fprintf(psFile,"%%%%EndPageSetup\n");
}


/******************************************************************************
*
*       Function        drawAxes
*       Parameters		psFile (pointer to the postscript file where the axes 
*						of the graph are to be drawn), numXVals (the number of
*						labels on the X-axis), xVals (the x values of the labels
*						on the X-axis), xLabels (the labels on the X-axis), 
*						numYVals (the number of labels on the Y-axis), 
*						yVals (the y values of the labels on the Y-axis), 
*						yLabels (the labels on the Y-axis), yTitle (the title
*						of the Y axis), xTitle (the title of the X axis)
*						xScale (the parameter according to which the graph 
*						needs to be scaled in the X-axis), yScale (the 
*						parameter according to which the graph needs to be 
*						scaled in the Y-axis). 
*       Output          none
*       Description		Draws the axes of the graph according to the parameters 
*						specified.
*       Comments		none.
*
******************************************************************************/
void drawAxes(FILE *psFile, int numXVals, double *xVals, xLabel *xLabels, 
			  int numYVals, double *yVals, yLabel *yLabels, char *xTitle, 
			  char *yTitle, double xScale, double yScale)
{
    double		currVal, smallY, yOffset, *valsPtr, *valsPtrEnd;
    time_t		currTime;
	xLabel		*xLabelsPtr;
	yLabel		*yLabelsPtr;


	/* 1.	Set the postscript origin at the origin of this set of axes. */

    /* 1.1.	If all y values are positive, set the origin at y=0.
			Otherwise, set it at the lowest negative value. */

	smallY = (*yVals < 0.0) ? (*yVals) : 0.0;

	/* 2.	Draw the axes (rotated 90 degrees). 
			Axes length: X=684 (9.5 inches), Y=450 (6.25 inches). */

    /* 2.1.	Draw the X-axis along the long side of the page (left side). */

	fprintf(psFile,"\n%% Drawing the X-axis of the graph.\n\n");
    fprintf(psFile,"504 72 translate\n");
	fprintf(psFile,"90 rotate\n");
	fprintf(psFile,"0 0 moveto\n");
	fprintf(psFile,"gsave %s 684 0 rlineto stroke grestore\n", THIN_LINE);
	fprintf(psFile,"gsave 694 0 rmoveto\n");
	fprintf(psFile,"14 /Times-Roman SF (%s) show grestore\n", xTitle);

    /* 2.2.	Draw the Y-axis along the short side of the page (bottom). */

	fprintf(psFile,"\n%% Drawing the Y-axis of the graph.\n\n");
	fprintf(psFile,"gsave %s 0 450 rlineto stroke grestore\n", THIN_LINE);
	fprintf(psFile,"gsave 0 460 rmoveto\n");
	fprintf(psFile,"14 /Times-Roman SF (%s) show grestore\n", yTitle);

	/* 3.	Draw a dotted line (parallel to X-axis) at y=0. */

	fprintf(psFile,"\n%% Drawing the line y=0.\n\n");
	fprintf(psFile,"0 %.2lf moveto\n", ((0 - smallY) * yScale));
	fprintf(psFile,"gsave %s %s 684 0 rlineto stroke grestore\n",THIN_LINE, EVEN_DASH);

	/* 4.	Print the current time at the bottom of the page. */
	
	fprintf(psFile,"\n%% Printing the current time at the lower right side of the graph.\n\n");
    currTime = time((time_t*) 0);
	fprintf(psFile,"580 -70 moveto\n");
	fprintf(psFile,"10 /Times-Roman SF (%s) show\n", ctime(&currTime));

    /* 5.	Draw the X-axis labels and name. */

	fprintf(psFile,"\n%% Printing the labels and name of the X-axis.\n\n");

    for (valsPtr = xVals, valsPtrEnd = valsPtr + numXVals, xLabelsPtr = xLabels; 
			valsPtr < valsPtrEnd; valsPtr++, xLabelsPtr++) {

        currVal = (*valsPtr) * xScale;
		fprintf(psFile,"%.2lf -4 moveto\n", currVal);
		fprintf(psFile,"gsave %s 0 2 rmoveto 0 4 rlineto stroke grestore\n", THIN_LINE);
		fprintf(psFile,"gsave -90 rotate 0 -3 rmoveto\n");
		fprintf(psFile,"8 /Times-Roman SF (%s) show grestore\n", *xLabelsPtr);
    }

    /* 6.	If the postscript origin isn't at y=0, translate it to y=0 for 
			subsequent drawing. */

	fprintf(psFile,"\n%% Drawing the labels and name of the Y-axis.\n\n");

	if (smallY < 0) {
		yOffset = smallY * yScale * (-1.0);
		fprintf(psFile,"0 %.2lf translate\n", yOffset);
	}

    /* 7.	Draw the labels and name of the Y-axis. */

    for (valsPtr = yVals, valsPtrEnd = valsPtr + numYVals, yLabelsPtr = yLabels; 
			valsPtr < valsPtrEnd; valsPtr++, yLabelsPtr++) {
        
		fprintf(psFile,"-4 %.2lf moveto\n", (double)(*valsPtr) * yScale);
		fprintf(psFile,"gsave %s 9 0 rlineto stroke grestore\n", THIN_LINE);
		fprintf(psFile,"gsave -20 -3 rmoveto\n");
		fprintf(psFile,"9 /Times-Roman SF (%s) show grestore\n", *yLabelsPtr);
    }
}

/******************************************************************************
*
*       Function        drawCurve
*       Parameters		psFile (pointer to the postscript file where the curve
*						is to be drawn), vals (the values of the points in  
*						the graph, the control points), numPoints (the number of 
*						points in the graph), lineType (the type of line that 
*						needs to be used for the curve), xScale (the parameter 
*						according  to which the graph needs to be scaled in  
*						the X-axis), yScale (the parameter according to which 
*						the graph needs to be scaled in the Y-axis). 
*       Output          none
*       Description		Draws a curve which connects between each 2 successive 
*						control points with straight lines.
*       Comments		none.
******************************************************************************/
void drawCurve(FILE *psFile, point *vals, int numPoints, char *lineType, 
			   double xScale, double yScale)
{
	point		*valsPtr, *valsPtrEnd;


	fprintf(psFile,"\n%% Drawing lines connecting the points of the graph.\n\n");

    /* 1.	Update the given control points according to the scales of the axes,
			and draw the lines connecting between them. */
 
	valsPtr = vals;
	valsPtrEnd = valsPtr + numPoints;

	fprintf(psFile,"newpath\n");
	fprintf(psFile,"%s\n", lineType);

	fprintf(psFile,"%.4lf %.4lf moveto\n",(valsPtr->x * xScale), (valsPtr->y * yScale));

	for (valsPtr++; valsPtr < valsPtrEnd; valsPtr++) {
		
		fprintf(psFile,"%.4lf %.4lf lineto\n", (valsPtr->x * xScale), 
					(valsPtr->y * yScale));
	}

	fprintf(psFile,"stroke\n");
}


/******************************************************************************
*
*       Function        drawBezier
*       Parameters		psFile (pointer to the postscript file where the Bezier  
*						curve is to be drawn), vals (the values of the points on  
*						the graph, the control points), numPoints (the number of 
*						points on the graph), s0 (used as the slope of the 
*						tangent to the function at the left end-point), sn (used 
*						as the slope of the tangent to the function at the right  
*						end-point), lineType (the type of line that needs to
*						be used for the curve), xScale (the parameter according 
*						to which the graph needs to be scaled in the X-axis), 
*						yScale (the parameter according to which the graph  
*						needs to be scaled in the Y-axis). 
*       Output          none
*       Description		Draws a Bezier curve made up of several cubic Bezier 
*						curves pieced together, in such a way that it passes 
*						through all the control points.
*						The end points for each cubic Bezier curve are 2 successive
*						control points. Each control point which isn't extreme 
*						appears in 2 successive curves, to achieve continuity.
*						The 2 middle points for each cubic curve are computed to
*						be on the tangent to the end point adjacent to them.
*       Comments		A Bezier curve doesn't pass through any of the control
*						points except the first and last one. Therefore, in order
*						for the curve to pass through all the control points,
*						it is made up of several cubic Bezier curves pieced 
*						together.
*						Draws an appropriate curve between points (x0,y0) & 
*						(x3,y3) based on Bezier cubic control points (x1,y1) & 
*						(x2,y2). (x1,y1) lies on the tangent at (x0,y0) and 
*						(x2,y2) lies on the tangent at (x3,y3) where 
*						x0 < x1 < x2 < x3. 
******************************************************************************/
void drawBezier(FILE *psFile, point *vals, int numPoints, double s0, double sn, 
				char *lineType, double xScale, double yScale)
{
    int					numCurves;
	cubicControlPts		*bezArr, *bezPtr, *bezPtrEnd, *bezPtrPrev;
	point				*valsPtr, *valsPtrEnd;
    double				*slopeArr, *currSlope, s1, s2;


	/* 1.	Allocate memory for arrays that would store all the necessary information
			for computing the cubic Bezier curves. */

	myMalloc(bezArr, sizeof(cubicControlPts) * numPoints);
 	myMalloc(slopeArr, sizeof(double) * numPoints);

    /* 2.	Update the given control points according to the scales of the axes. */
 
	for (valsPtr = vals, valsPtrEnd = valsPtr + numPoints; valsPtr < valsPtrEnd; 
			valsPtr++) {
		
		valsPtr->x *= xScale;
		valsPtr->y *= yScale;
	}

	/* 3.	Update the end points of each cubic Bezier curve according to the control
			points. */

	numCurves = numPoints-1;

    for(bezPtr = bezArr, bezPtrEnd = bezPtr + numCurves, valsPtr = vals; 
			bezPtr < bezPtrEnd; bezPtr++, valsPtr++) {
		bezPtr->p0 = *valsPtr;
		bezPtr->p3 = *(valsPtr+1);
    }

    /* 4.	Estimate the slope of the tangent at each control point (data point).
			Given 3 points on a curve (x1,y1), (x2,y2), & (x3,y3), a good guess 
			for the slope of the tangent at (x2,y2) is the average of the slopes 
			of the line connecting (x1,y1) & (x2,y2) and the line connecting (x2,y2) 
			& (x3,y3). Obviously, this method does not work for the end points on 
			the curve, and therefore those values are passed to the function as s0 
			and sn (usually values of 0.0 will be ok). */

	currSlope = slopeArr;
    (*currSlope) = s0;

    for(bezPtrPrev = bezArr, bezPtr = bezArr+1, 
			bezPtrEnd = bezPtr + numCurves; bezPtr< bezPtrEnd; 
			currSlope++, bezPtr++, bezPtrPrev++) {

		if (bezPtrPrev->p3.x == bezPtrPrev->p0.x) {
			s1 = 0.0;		/* Discontinuity */ 
		}
		else {
			s1 = (bezPtrPrev->p3.y - bezPtrPrev->p0.y) / 
					(bezPtrPrev->p3.x - bezPtrPrev->p0.x);
		}

		if (bezPtr->p3.x == bezPtr->p0.x) {
			s2 = 0.0;		/* Discontinuity */
		}
		else {
			s2 = (bezPtr->p3.y - bezPtr->p0.y) / 
					(bezPtr->p3.x - bezPtr->p0.x);
		}

		(*currSlope) = (s1 + s2) / 2.0;
	
		if (*currSlope > 10.0) {
			*currSlope = 10.0;
		}

		if (*currSlope < -10.0) {
			*currSlope = -10.0;
		}
    }
    *currSlope = sn;


    /* 5.	Calculate the middle 2 points, (x1,y1) & (x2,y2), for each cubic Bezier
			curve according to the approximate tangent lines that were computed
			for the real data points (the control points). */

    for(bezPtr = bezArr, bezPtrEnd = bezPtr + numCurves,
			currSlope = slopeArr; bezPtr< bezPtrEnd; 
			currSlope++, bezPtr++) {

		/* 5.1.	Divide the x-interval between p0 and p3 into 3 equal segments,
				and place p1 & p2 at 1/3 & 2/3 the distance. */

		bezPtr->p1.x = (bezPtr->p3.x - bezPtr->p0.x) / 3.0;
		bezPtr->p1.y = bezPtr->p0.y + bezPtr->p1.x * (*currSlope);
		bezPtr->p2.x = bezPtr->p1.x;
		bezPtr->p2.y = bezPtr->p3.y - bezPtr->p2.x * (*(currSlope+1));
		bezPtr->p1.x += bezPtr->p0.x;
		bezPtr->p2.x += bezPtr->p1.x;
    }

    /* 6.	Draw the cubic Bezier curves, using the curveto command of Postscript which
			draws a Bezier cubic curve from the current position to position (x3,y3) 
			based on the control points (x1,y1) and (x2,y2). */

	fprintf(psFile,"\n%% Drawing the Bezier curve of the graph.\n\n");

	bezPtr = bezArr;

	fprintf(psFile,"%.2lf %.2lf moveto\n",bezPtr->p0.x,bezPtr->p0.y);

	for(bezPtr = bezArr, bezPtrEnd = bezPtr + numCurves; 
			bezPtr< bezPtrEnd; bezPtr++) {
		
		if (bezPtr->p0.x != bezPtr->p3.x) {
			fprintf(psFile, "gsave %s %.4lf %.4lf %.4lf %.4lf %.4lf %.4lf curveto stroke grestore\n",
					lineType, bezPtr->p1.x, bezPtr->p1.y, bezPtr->p2.x, bezPtr->p2.y, 
					bezPtr->p3.x, bezPtr->p3.y);
		}
        fprintf(psFile,"%.4lf %.4lf moveto\n", bezPtr->p3.x, bezPtr->p3.y); 
	}

	fprintf(psFile,"%.4lf %.4lf moveto\n", (bezPtr-1)->p3.x, (bezPtr-1)->p3.y);	
	
	/* 7.	Free the memory allocated dynamically. */

    myFree(bezArr);
    myFree(slopeArr);
}



/******************************************************************************
*
*       Function        psPageEnd
*       Parameters		psFile (pointer to the ps file).
*       Output          none
*       Description		Prints ending comments and commands to the ps file 
*						pointed to by 'psFile'.
*       Comments		none
*
******************************************************************************/
void psPageEnd(FILE *psFile)
{
    fprintf(psFile,"%%%%Trailer\n");
    fprintf(psFile,"grestore\n");
    fprintf(psFile,"end %% DrawingDict\n");
    fprintf(psFile,"%%%%EOF\n");
    fprintf(psFile,"restore showpage\n");
}



