/* $Id: PedData.h,v 1.2 2008/08/03 08:15:46 superlink Exp $ */
/*********************************************************************************/
/*  PedData.h																     */
/*  Author: Ma'ayan Fishelson													 */
/*********************************************************************************/
/*																				 */
/*  Comment: This file includes functions used for reading all the data from the */
/*			 pedigree file and initializing the relevant data-structures.		 */
/*																				 */
/*********************************************************************************/

#include "UseBoincIO.h"


#ifndef _PED_DATA_H_
#define _PED_DATA_H_

#include <stdio.h>
#include <stdlib.h>
#include "LociData.h"
#include "Node.h"


/*--------------------------------------*/
/*		DEFINITIONS & DATA-STRUCTURES	*/
/*--------------------------------------*/

/* The default name for the pedigree file. */

#define DEFAULT_PEDFILE_NAME	"pedfile.dat"	

/* Codes for the different affection status possible. */
typedef	char		affectionStatus;
#define	unknownAS	0
#define	unaffected	1
#define	affected	2

/* Codes for the sex of a person. */
typedef	char	sex;
#define	Male	0
#define	Female	1


/* This structure holds information on one possible allele of a person in one haplotype. */
typedef struct halfLocusPerson {		
	int						allele;		/* The code of the allele. */
	struct halfLocusPerson *next;		/* Pointer to the strcuture that holds information
										   on the next possible allele for this haplotype. */
	boolean					isCertain;	/* A flag that is TRUE if the person has this
										   allele for certain. */
} halfLocusPerson;

/* This structure holds information on an allele set in the case where allele-recoding 
   is performed. */
typedef struct alleleSet {
	int		numAlleles;					/* The number of alleles in the set. */
	int		*alleleArr;					/* The array of alleles. */
} alleleSet;


/* This structure holds information on one locus of a person. */
typedef struct locusPerson {
	boolean unknown;					/* A flag that indicates if the phenotype is known
										   at this locus. */
	union {
		struct {
				halfLocusPerson *maLocus;				/* A list of the possible maternal alleles. */
				boolean isKnownMa;						/* The value calculated at createValArray. */
				int	numMaAlleleGroups;					/* The number of sets of maternal alleles 
														   or the array size. */
			#if PERFORM_ALLELE_RECODING 
				halfLocusPerson	*maLocusTrans;			/* A list of the transmitted maternal alleles. */
				alleleSet		*maAllelsGroups;		/* A list of all the sets of maternal 
														   alleles (both transmitted and non-transmitted). */
			#endif
				int				*maAlleles;				/* An array that would hold all the possible
														   maternal alleles (if no allele-recoding 
														   is performed). */
				int				maSelectorVals[2];		/* The possible maternal selector values. */
				halfLocusPerson *paLocus;				/* A list of the possible paternal alleles. */
				int				numPaAlleleGroups;		/* The number of sets of paternal alleles, 
														   or the array size. */
				boolean			isKnownPa;				/* The value calculated at createValArray. */
			
			#if PERFORM_ALLELE_RECODING 
				halfLocusPerson	*paLocusTrans;			/* A list of the transmitted paternal alleles. */
				alleleSet		*paAllelsGroups;		/* A list of all the sets of paternal 
														   alleles (both transmitted and non-transmitted). */
			#endif
				int				*paAlleles;				/* An array that would hold all the possible
														   paternal alleles (if no allele-recoding 
														   is performed). */
				int				paSelectorVals[2];		/* The possible paternal selector values. */

			#if UNITE_VARS
				uniteType		uniteT;					/* Indicates whether the locus variables that correspond
														   to this locus of this person need to be
														   united with each other or also with other locus 
														   variables during the variable elimination. */
				int				*certainAllele;			/* In case 'uniteT' is equal to 'oneAll', this field 
														   holds the allele which is known for certain in the 
														   genotype of this person at this locus. 
														   In case 'uniteT' is equal to 'fourAll', this field 
														   holds all four alleles which are known for certain
														   for this person and the other parent. */
				int				unitePerson;			/* In case 'uniteT' is equal to 'fourAll', this field 
														   holds the id of the person of which genetic loci
														   variables at this locus need to be eliminated jointly
														   with the genetic loci variables of this person. */
			#endif
		} allelesL;
		struct {
			affectionStatus affectionS;				/* The affection status. */
			int				liabilityClass;			/* The liability class this person belongs to. */
		} affectionL;
	} locusData;
} locusPerson;


/* An array which holds the data about all loci of some person. */
typedef locusPerson* Loci;				


/* This structure holds information about one person. */
typedef struct person {
	int				pNum;			/* The index of this person in the pedigree. */
	Loci			loci;			/* Holds information on the phenotypes in all loci. */
	struct person	*ch;			/* Pointer to children */
	struct person	*ma, *pa;		/* Pointers to mother & father. */
	struct person	*NeSiMa;        /* Pointer to the next sibling from 
									   the same mother */
	struct person	*NeSiPa;		/* Pointer to the next sibling from 
									   the same father */
	sex				personSex;		/* The sex of this person. */
	boolean			visited;		/* A flag that indicates if this person has been
									   visited already. */
	boolean			done;			/* A flag that indicates if this person has been updated
									   according to all of his children already. */
	boolean			updatedParents;	/* A flag that indicates if this person parent's 
									   have been updated according to this person already. */

	node			**locusVar[2];		/* An array of pointers to all the locus nodes that represent
										   this person in the Bayesian network. */
	node			**selectorVar[2];	/* An array of pointers to all the selector nodes that
										   represent this person in the Bayesian network. */
	node			**phenotypeVar;     /* An array of pointers to all the phenotype nodes that
										   represent this person in the Bayesian network. */
	node			*genotypeVar;		/* A pointer to the genotype node that represents this
										   person in the Bayesian network.*/
} person;



/* This structure holds information on one pedigree. */
typedef struct pedigree {
	int				numPeople;		/* The number of people in this pedigree. */
	int				pedNum;			/* The serial number of the pedigree. */
	person			*people;		/* An array which holds information about the
									   people in the pedigree. */
	person			*founder;		/* A pointer to one of the founders in the pedigree. */
	person			*childLess;		/* A pointer to a person without children. */
	struct pedigree *nextPed;		/* A pointer to the next pedigree structure. */
	long			numNodesBN;		/* The number of nodes in the Bayeasian network that
									   represents the pedigree. */
	int				numNonFounders;	/* The number of non-founders in the pedigree. */
} pedigree;


/* This structure holds information on all pedigrees in the input. */
typedef struct pedData {
	int			numPeds;			/* The number of pedigrees. */
	pedigree	*head;				/* Linked list of pedigree information. */
} pedData;


typedef void (updateFlagsLoci)(pedigree *ped);


/*------------------------------*/
/*		GLOBAL VARIABLES		*/
/*------------------------------*/

extern FILE			*pedFile;
extern pedData		peds;
extern lociData		loci;

/*------------------------------*/
/*		FUNCTIONS				*/
/*------------------------------*/

void readPedFile();
void computePedSize(char *pedfileName);
void readPed(pedigree *ped, int pedInd);
void readPerson(pedigree *ped, int pedInd, boolean sexLink);
void readAffection(person *currPerson, int locusInd, int pedInd, int personInd); 
void readAlleles(person *currPerson, int locusInd, int pedInd, int personInd); 
void readAllelesMale(person *currPerson, int locusInd, int pedInd, int personInd); 

void initFlags(pedigree *ped);
void updateFlagsLociAutosomal(pedigree *ped);
void updateUnknownLoci(person *currPerson);
void updateFlagsLociSexLinked(pedigree *ped);
void updateUnknownLociMale(person *currPerson);
void addAllAlleles(person *currPerson, int locusNum, boolean isPaHaplotype);

void checkInputValidity(pedigree *ped);
void checkFatherChildMatch(person *currPerson, int pedNum);
void checkMotherChildMatch(person *currPerson, int pedNum);

void freePedData(int numLoci);

#if PERFORM_ALLELE_RECODING
	void freePedDataAR(int numLoci);
#endif

void deleteAlleles(halfLocusPerson *Locus);
void freeNode(node **Node);

#if DEBUG_STAT
	void printPeds();
#endif


#endif /* _PED_DATA_H_ */

