/* $Id: Probability.h,v 1.2 2008/08/03 08:15:46 superlink Exp $ */
/*********************************************************************************/
/*  Probability.h																 */
/*  Author: Ma'ayan Fishelson													 */
/*********************************************************************************/
/*																				 */
/*  Comment: This file includes functions used for creating, multiplying, and	 */
/*		     marginalizing the probability functions of the Bayesian networks.	 */
/*																				 */
/*********************************************************************************/


#ifndef _PROBABILTY_H_
#define _PROBABILTY_H_

#include <float.h>
#include "Auxiliary.h"
#include "PedData.h"
#include "Product.h"
#include "Globals.h"



/*------------------------------*/
/*		DEFINITIONS				*/
/*------------------------------*/

/*  Evaluates Pr(C[i,p] | A[i,p], A[i,m], Sc[i,p])	*/
#define transmission(childAllele, paternalAllele, maternalAllele, selectorVal)	\
					(double)(((childAllele == paternalAllele) && (selectorVal == 1)) ||	\
						 	 ((childAllele == maternalAllele) && (selectorVal == 0)));

/*  Evaluates Pr(C[i,p] | A[i,p], A[i,m], Sc[i,p])	*/
#define transmissionMale(childAllele, maternalAllele)	\
						(double)(childAllele == maternalAllele);

/* Evaluates Pr(Sc[i,x] | Sc[i-1,x]). If (Sc[i,x] == Sc[i-1,x]), returns (1 - theta).  
   Otherwise, returns theta. */
#define recombTransMale(selectorVal, adjacentSelectorVal, thetaNum)		\
	(selectorVal == adjacentSelectorVal) ? recombData.maleThetas[thetaNum][1] : recombData.maleThetas[thetaNum][0];

#define recombTransFemale(selectorVal, adjacentSelectorVal, thetaNum)		\
	(selectorVal == adjacentSelectorVal) ? recombData.femaleThetas[thetaNum][1] : recombData.femaleThetas[thetaNum][0];

/*------------------------------*/
/*		FUNCTIONS				*/
/*------------------------------*/

/* Functions used for creating the different probability tables. */

int createTableTrans(person *currPerson, int locusNum, boolean isPaternal, 
					  node *locusNode, node *selectorNode, Probability **newTable);

#if PERFORM_ALLELE_RECODING
	int createTableTransAR(person *currPerson, int locusNum, boolean isPaternal, 
						   node *locusNode, node *selectorNode, Probability **newTable);
#endif

int createTableTransFounder(person *currPerson, int locusNum, boolean isPaternal, 
							node *locusNode, node *selectorNode, 
							Probability **newTable);
int createTableTransMale(person *currPerson, int locusNum, node *pLocusNode,
						 Probability **newTable);

#if PERFORM_ALLELE_RECODING
	int createTableTransMaleAR(person *currPerson, int locusNum, node *pLocusNode,
							Probability **newTable);
#endif

int createTablePenetrance(locusPerson *currLoc, int locusNum, node *phenotypeNode, 
						  node *pLocusNode, node *mLocusNode, Probability **newTable);
int createTablePenetranceFounder(locusPerson *currLoc, int locusNum, 
								 node *phenotypeNode, node *pLocusNode,
								 node *mLocusNode, node *genotypeNode, 
								 Probability **newTable);
int createTableMalePenetrance(locusPerson *currLoc, int locusNum, node *phenotypeNode, 
							  node *mLocusNode, Probability **newTable);
int createTablePenetranceTwo(person *currPerson, int firstNullLociIndDS, 
							 int secondNullLociIndDS, node *phenotypeNode, Probability **newTable);
int createTablePenetranceFounderTwo(person *currPerson, int firstNullLociIndDS, 
									int secondNullLociIndDS, node *phenotypeNode, 
									node *genotypeNode, int firstNullLociPosMap,
									Probability **newTable);
int createTableMalePenetranceTwo(person *currPerson, int firstNullLociIndDS, 
								  int secondNullLociIndDS, node *phenotypeNode,
								  Probability **newTable);
int createTablePhenotype(node *phenotypeNode, node *pLocusNode, node *mLocusNode,
						 Probability **newTable);

#if PERFORM_ALLELE_RECODING
boolean setInclusion(alleleSet parentSet, alleleSet childSet);
#endif


int createTableGeneFrequency(int locusNum, int *alleleArr, int numVals, Probability **newTable);
int createTableGenotypeFrequency(int locusNum, int *alleleArr, int numVals,
							     Probability **newTable, boolean isUnknownGenotype);

#if PERFORM_ALLELE_RECODING
int createTableGenotypeFrequencyAR(int locusNum, alleleSet *alleleSetArr, 
								   int numSets, Probability **newTable, 
								   boolean isUnknownGenotype);
int createTableGeneFrequencyAR(int locusNum, alleleSet *alleleSetArr,  
							   int numSets, Probability **newTable);
#endif

int createTableRecomb(person *currPerson, int locusNum, boolean isPaternal, node *selectorNode,
					   Probability **newTable);
int createTableRecombInterferenceMale(person *currPerson, node *pSelectorNode, 
									  Probability **newTable);
int createTableRecombInterferenceFemale(person *currPerson, node *mSelectorNode, 
										Probability **newTable);
double recombTransInterferenceMale(int cSelectorVal, int bSelectorVal, int aSelectorVal, 
								   int thetaNum);
double recombTransInterferenceFemale(int cSelectorVal, int bSelectorVal, int aSelectorVal, 
								   int thetaNum);
int createTableHalf(int numVals, Probability **newTable);

/* Functions used for multiplying the probability tables. */

void multiplyAllFuncsSimp(function *combinedFunc, variable *elimVar);
void multiplyAllFuncsComplex(function *combinedFunc, variable *elimVar);
Probability multiplyAllOneVarFuncs(variable *elimVar, double logMultFactor);
void multiplyAllFuncs(int numFuncs, function **funcArr, function *combinedFunc);

#if UNITE_VARS
	void multiplyAllFuncsJointComplex(function *combinedFunc, variable **elimVarsArr, 
									int numElimVars, function **funcArr, int numFuncs);
	Probability multiplyAllFuncsJoint(variable **elimVarsArr, int numElimVars,
									double logMultFactor, function **funcArr, 
									int numFuncs);
#endif

/* Functions used for finding all valid joint assignemtns to a group of two or four 
   variabels that need to be treated jointly. */

#if UNITE_VARS
	void findValidJointAssignments(variable **elimVarsArr, int numElimVars, 
		  						validAssignments *validAssignmentsArr, int numFuncs);
	void findValidJointAssignmentsOne(variable **elimVarsArr, int numElimVars, 
		  							validAssignments *validAssignmentsArr, int numFuncs);
	void findValidJointAssignmentsFour(variable **elimVarsArr, int numElimVars, 
		  							validAssignments *validAssignmentsArr, int numFuncs);
	void allocAssignmentsMem(validAssignments *validAssignmentsArr, int numElimVars,
							int numFuncs);
	void freeAssignmentMem(validAssignments *validAssignmentsArr);
#endif

#endif /* _PROBABILTY_H_ */

