/* $Id: Product.c,v 1.2 2008/08/03 08:15:46 superlink Exp $ */
/*********************************************************************************/
/*  Product.c												     				 */
/*  Author: Ma'ayan Fishelson				                                     */
/*********************************************************************************/

#include "Product.h"
#include "Globals.h"
#include "XMLFormat.h"
#if (PRINT_ELIMINATION_STATISTICS || (PRINT_VARIOUS_STATISTICS_LEVEL > 1))
	int countEliminatedVars;
	int countConstantVars;
#endif

void andreiPrintProduct(product* Product);



/******************************************************************************
*
*       Function        allocateCreateProducts
*       Parameters		createPersonVarsPed (the function that is used to create
*						the variables and functions of a single individual). 
*       Output          none
*       Description		Allocates memory and creates the products from the 
*						nodes of the Bayesian networks and the data regarding 
*						the loci being analyzed.
*       Comments		none
*
******************************************************************************/
product** allocateCreateProducts(createPersonVars createPersonVarsPed)
{
	pedigree		*pedPtr;
	Probability		Res, *pedLnLike;
	long			numSlotsProd;
	int				prodNum;
	product			**prodArr, **currProd;
	boolean			*isZeroLikePed;
	resNode			*currResNode;

	currResNode = totalRes.currRes;

	/* 1.	Initialize the likelihood of the data to zero. */

	currResNode->lnLike = 0.0;

	/* 2.	Allocate memory for the product array. */

	myMalloc(prodArr, sizeof(product*) * peds.numPeds); 

	/* 3.	Iterate over the pedigrees and for each one:  */

	for (pedPtr = peds.head, prodNum = 0, isZeroLikePed = totalRes.isZeroLikeArr,
			pedLnLike = totalRes.pedsLnLikeArr, currProd = prodArr; pedPtr != NULL;
			pedPtr = pedPtr->nextPed, prodNum++, isZeroLikePed++, pedLnLike++, currProd++) {

		numSlotsProd = pedPtr->numNodesBN + 1;

		/* 3.1.	Allocate & initialize the product's fields. */

		*currProd = allocateProduct(numSlotsProd, numSlotsProd); 

		/* 3.2.	Create the variables and functions of the product from the nodes of 
				the Bayesian network. */

		Res = createProduct(*currProd, pedPtr, createPersonVarsPed);

		/* 3.3.	Update the results data structure, according to the constant functions
				that were created. */

		if (Res < MIN_DBL) {
			*isZeroLikePed = TRUE;
			currResNode->lnLike = MIN_DBL;
			(*pedLnLike) = MIN_DBL;
		}
		else {
			*isZeroLikePed = FALSE;
			currResNode->lnLike += Res;
			(*pedLnLike) = Res;
		}
	}

	/* 4.	Return the product array. */

	return prodArr;
}


/******************************************************************************
*
*       Function        allocateProduct
*       Parameters		numVariables (number of spaces that need to be allocated
*						for variables), numFunctions (number of spaces that need 
*						to be allocated for functions).
*       Output          none
*       Description		Allocates memory for the product & initializes its fields.
*       Comments		Returns a pointer to the product struct that was allocated.
*
******************************************************************************/
product* allocateProduct(long numVariables, long numFunctions)
{
	product		*prod;
	variable	*varPtr, *varPtrEnd;
	function	*funcPtr, *funcPtrEnd;

	/* 1.	Allocate the product data-structure. */

	myMalloc(prod, sizeof(product));

	/* 2.	Initialize the variables related fields of the product structure. */

	prod->numVariables = numVariables;
	prod->currNumVars = 0;
	myMalloc(prod->variables, sizeof(variable) * numVariables);
	prod->nextVarPlace = &prod->variables[0];
	prod->nextElimVar = &prod->variables[0];
	
	for (varPtr = prod->variables, varPtrEnd = varPtr + numVariables;
			varPtr < varPtrEnd; varPtr++) {
		varPtr->used = FALSE;
	}

	/* 2.	Initialize the functions related fields of the product structure. */

	prod->numFunctions = numFunctions;
	prod->currNumFuncs = 0;
	myMalloc(prod->functions, sizeof(function) * numFunctions);
	prod->nextFuncPlace = &prod->functions[0];
	
	for (funcPtr = prod->functions, funcPtrEnd = funcPtr + numFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {
		funcPtr->used = FALSE;
		funcPtr->variables = NULL;
	}

	/* 3.	Return a pointer to the product data structure that was allocated 
			and initialized. */

	return prod;
}



/******************************************************************************
*
*       Function        createProduct
*       Parameters		prod (the product that will be created), pedPtr (the 
*						pedigree struct which the product will be built upon),
*						createPersonVarsPed (the function that is used to create
*						the variables and functions of a single individual).
*       Output          If requested, prints elimination statistics.
*       Description		Generates the variables & the functions of the product.
*       Comments		Returns the joint logProbability of the constant 
*						functions that were supposed to be in the product.
*
******************************************************************************/
Probability createProduct(product *prod, pedigree *pedPtr, 
						  createPersonVars createPersonVarsPed)
{
	Product = prod;
	initFlags(pedPtr);

	tempRes = 0.0;

	#if (PRINT_ELIMINATION_STATISTICS || (PRINT_VARIOUS_STATISTICS_LEVEL > 1))
		countConstantVars = 0;
		countEliminatedVars = 0;
	#endif

	/* 1.	Create the variables and functions of the product. */

	traverseUpPed(pedPtr->childLess, createPersonVarsPed); 

	
	/* 2.	Build, for each variable, a table that points to all the functions
			it appears in. */

	buildVariableFunctionTables(prod);

	/* 3.	Print elimination statistics, if requested. */

	#if PRINT_ELIMINATION_STATISTICS
		printf("\n\nEliminated %d nodes.\n", countEliminatedVars);
		printf("%d nodes are constant.\n", countConstantVars);
		printf("Number of nodes %d.\n", prod->currNumVars);
	#endif

	#if (PRINT_VARIOUS_STATISTICS_LEVEL > 1)
		prod->numVarsElim = countEliminatedVars;
		prod->numVarsConst = countConstantVars;
	#endif

	/* 4.	Return the joint log-probability of the constant functions. */
	return tempRes;
}

void andreiPrintProduct(product* Product) {
	FILE* outFile;
	int i, j;

	outFile = fopen("SLproduct.txt", "w");
	
	//	print variables
	fprintf(outFile, "variables\n");
	for(i = 0; i < Product->numVariables; ++i) {
		variable* var = Product->variables + i;
		fprintf(outFile, "%s\t", var->originalNode->name);
		fprintf(outFile, "%d\n", var->numValues);
	}

	//	print functions
	fprintf(outFile, "\n\nfunctions\n");
	for(i = 0; i < Product->numFunctions; ++i) {
		function* func = Product->functions + i;
		fprintf(outFile, "\n(");
		for(j = 0; j < func->numVariables; ++j) {
			fprintf(outFile, "%s ", func->variables[j]->originalNode->name);
		}
		fprintf(outFile, ")\n");
		for(j = 0; j < func->sizeOfTable; ++j) {
			fprintf(outFile, "%lf\n", func->probabList[j]);
		}
		fprintf(outFile, "fEnd\n");
	}

	fclose(outFile);
}



/******************************************************************************
*
*       Function        traverseUpPed
*       Parameters      currPerson (the person that is being traversed currently),
*						createPersonVarsPed (the function that is used to create
*						the variables and functions of a single individual).
*       Output          none
*       Description		This function traverses all the children of 'currPerson'.
*						If the variables that represent all children have been 
*						generated already, it generates the variables that
*						represent 'currPerson' and then traverses up to 
*						'currPerson's parents (if they exist).
*       Comments		none
*
******************************************************************************/
void traverseUpPed(person *currPerson, createPersonVars createPersonVarsPed)
{
	person	*ch;			
	sex		Sex;

	ch = currPerson->ch;
	Sex = currPerson->personSex;

	/* 1.	Mark this person as visited. */

	currPerson->visited = TRUE;				

	/* 2.	Iterate over all children of 'currPerson' and check if the variables 
			that represent them have been created already. */

	if (Sex == Male) {	

		/* 2.1.	Iterate over all children. */
		
		for (; ch != NULL; ch = ch->NeSiPa) {	
			
			/* 2.1.1.	If we haven't created this child's variables. */

			if (ch->done == FALSE) {		

				/* 2.1.1.1.	If we haven't reached the child yet. */

				if (ch->visited == FALSE) {

					/* 2.1.1.1.1.	Traverse down to the child and from this child we
									will later traverse back up to the parent. */

					traverseDownPed(ch, createPersonVarsPed);
				}
				return;							
			}
		}
	}
	else {			/* (Sex == Female) */

		for (; ch != NULL; ch = ch->NeSiMa) {	
			
			if (ch->done == FALSE) {
				if (ch->visited == FALSE)		
					traverseDownPed(ch, createPersonVarsPed);
				return;
			}
		}
	}

	/* 3.	Create the variables that represent 'currPerson' in the product. */

	createPersonVarsPed(currPerson);

	/* 4.	Mark that we've finished creating the variables that represnt 'currPerson'. */

	currPerson->done = TRUE;			

	/* 5.	If 'currPerson' has a father which hasn't been visited yet,
			travers up to him. */

	if ((currPerson->pa != NULL) && ((currPerson->pa)->done == FALSE))	
		traverseUpPed(currPerson->pa, createPersonVarsPed);

	/* 6.	If 'currPerson' has a mother which hasn't been visited yet, 
			travers up to her. */

	if ((currPerson->ma != NULL) && ((currPerson->ma)->done == FALSE))	
		traverseUpPed(currPerson->ma, createPersonVarsPed);
}



/******************************************************************************
*
*       Function        traverseDownPed
*       Parameters      currPerson (the person that is being traversed currently),
*						createPersonVarsPed (the function that is used to create
*						the variables and functions of a single individual).
*       Output          none
*       Description		If 'currPerson' is childless, we traverse up to him.
*						If his child has been visited already, the function returns. 
*						Otherwise, we traverse down to the child.
*       Comments		none
*
******************************************************************************/
void traverseDownPed(person *currPerson, createPersonVars createPersonVarsPed)
{
	person *ch;

	ch = currPerson->ch;

	if (ch == NULL)					/* This person doesn't have any children. */
		traverseUpPed(currPerson, createPersonVarsPed);
	else {
		if (ch->visited == FALSE) 	/* We haven't visited the child yet. */
			traverseDownPed(ch, createPersonVarsPed);
		else				/* We'll reach this person later (by another path). */
			return;
	}
}



/******************************************************************************
*
*       Function        createPersonVarsAutosomal
*       Parameters		currPerson (the person whose variables need to be 
*						created).
*       Output          none
*       Description		Creates the variables that represent 'currPerson' in the
*						product.
*       Comments		Variables are created only for nodes that have more than 
*						one possible value, namely aren't constants. 
*						Only functions that aren't constant are inserted into
*						the product (the probability of the constant functions 
*						is added to the joint probability).
*						Used in case of an autosomal analysis.
*
******************************************************************************/
void createPersonVarsAutosomal(person *currPerson)
{
	int			locusIndMap, locusIndDS, tableSize=0, firstNullLociIndDS,
				secondNullLociIndDS;
	node		**pLocusNode, **mLocusNode, **pSelectorNode, **mSelectorNode, 
				**phenotypeNode, **genotypeNode;
	Probability *probabList;
	locus		*locusArr;

#if DISCONNECT_UNLINKED
	boolean				isUnlinked;
#endif


	locusArr = loci.locusArr;
	genotypeNode = &(currPerson->genotypeVar);

	/* 1.	Iterate over all loci from the last one (leftmost) on the map to 
			the first one (rightmost). */

	for (locusIndMap = loci.numLoci - 1; locusIndMap >= 0; locusIndMap--) {

		locusIndDS = loci.mapToData[locusIndMap];

		/* 1.1.	Initialize pointers to currPerson's nodes of this locus. */

		pLocusNode = &(currPerson->locusVar[0][locusIndDS]);

		mLocusNode = &(currPerson->locusVar[1][locusIndDS]);
		pSelectorNode = &(currPerson->selectorVar[0][locusIndDS]);
		mSelectorNode = &(currPerson->selectorVar[1][locusIndDS]);
		phenotypeNode = &(currPerson->phenotypeVar[locusIndDS]);

		#if DISCONNECT_UNLINKED
			if ((locusIndMap > 0) && (recombData.maleThetas[locusIndMap - 1][0] == 0.5)) {
				isUnlinked = TRUE;
			}
			else {
				isUnlinked = FALSE;
			}
		#endif

		/* 1.2.	If necessary, create a variable for the phenotype node. */

		if ((*phenotypeNode) != NULL) {

			switch ((locusArr[locusIndDS]).type) {

			case affection :
				if ((currPerson->pa != NULL) || (locusIndMap != 0)) {

					tableSize = createTablePenetrance(&(currPerson->loci[locusIndDS]), locusIndDS, *phenotypeNode, 
													  *pLocusNode, *mLocusNode, &probabList);
				}
				else {		/*	This is the first locus of a founder. */	

					tableSize = createTablePenetranceFounder(&(currPerson->loci[locusIndDS]), locusIndDS, *phenotypeNode, 
															 *pLocusNode, *mLocusNode, *genotypeNode, &probabList);
				}
		
				break;

			case null: 

				firstNullLociIndDS = loci.mapToData[loci.traitLociPos[0]];
				secondNullLociIndDS = loci.mapToData[loci.traitLociPos[1]];

				/* 1.2.1.	If this isn't a founder or this is the second null locus on the map
							of a two-locus disease, and neither one of the null loci is first 
							on the map. (loci.traitLociPos holds the positions on the map of the 
							two null loci). */
				
				if ((currPerson->pa != NULL) || 
						((loci.traitLociPos[0] != 0) && (loci.traitLociPos[1] != 0))) {

					tableSize = createTablePenetranceTwo(currPerson, firstNullLociIndDS, 
														 secondNullLociIndDS, 
														 *phenotypeNode, &probabList);
				}
				else {		/*	One of the null loci is the first locus of a founder. */	

					tableSize = createTablePenetranceFounderTwo(currPerson, firstNullLociIndDS, 
																secondNullLociIndDS, *phenotypeNode, 
																*genotypeNode, loci.traitLociPos[0], &probabList);
				}		
				break;	

			case alleles: 			/* This is an alleles type locus. */

				tableSize = createTablePhenotype(*phenotypeNode, *pLocusNode, *mLocusNode, &probabList);
				break;
			}

			handleProbabList(tableSize, Product, *phenotypeNode, probabList);
			updateConstantVars(*phenotypeNode);
		}

		/* 1.3.	Check if the paternal locus node is necessary. */ 

		if (*pLocusNode) {
			if (((*pLocusNode)->numChildren > 0) || (*pLocusNode)->isKnown) { 

				if (currPerson->pa != NULL) {	/* currPerson isn't a founder. */

					/* 1.3.1.	If this isn't the first locus of a founder's child. */

					if (((currPerson->pa)->pa != NULL) || (locusIndMap != 0)) {
	
						/* 1.3.1.2.		Create this nodes probability function. */

						#if PERFORM_ALLELE_RECODING
							if ((locusArr[locusIndDS]).type == alleles)
								tableSize = createTableTransAR(currPerson, locusIndDS, TRUE, *pLocusNode, 
															   *pSelectorNode, &probabList);
							else {
						#endif
								tableSize = createTableTrans(currPerson, locusIndDS, TRUE, *pLocusNode, 
															*pSelectorNode, &probabList);
						#if PERFORM_ALLELE_RECODING
							}
						#endif
					}
					else {		/* This is the first locus of a founder's child. */
	
						/* 1.3.1.2.		Create this nodes probability function. */

						tableSize = createTableTransFounder(currPerson, locusIndDS, TRUE, *pLocusNode,
															*pSelectorNode, &probabList);
					}

					/* 1.3.2.	If the function isn't a constant, insert it into the product.
								Otherwise, add its log-probability to the total log-probablity. */

					handleProbabList(tableSize, Product, *pLocusNode, probabList);
				}
				else {							/* currPerson is a founder. */

					/* 1.3.1.	If this is the first locus of a founder, create a genotype-frequency
								probability table for the paternal & maternal locus of the person. 
								Otherwise, create a gene-frequency probability table for the paternal 
								locus only. */

					if (locusIndMap == 0) {
						if ((locusArr[locusIndDS]).type == alleles){
							#if PERFORM_ALLELE_RECODING
								tableSize = createTableGenotypeFrequencyAR(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.paAllelsGroups,
																		   (*pLocusNode)->numValues, &probabList, currPerson->loci[locusIndDS].unknown);
							#else
								tableSize = createTableGenotypeFrequency(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.paAlleles,
																		 (*pLocusNode)->numValues, &probabList, currPerson->loci[locusIndDS].unknown);
							#endif
						} else
							tableSize = createTableGenotypeFrequency(locusIndDS, NULL, (*pLocusNode)->numValues, &probabList, TRUE);

						handleProbabList(tableSize, Product, *genotypeNode, probabList);
					}
					else {
						if (locusArr[locusIndDS].type == alleles) {
							#if PERFORM_ALLELE_RECODING	
								tableSize = createTableGeneFrequencyAR(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.paAllelsGroups,
																	   (*pLocusNode)->numValues, &probabList);
							#else
								tableSize = createTableGeneFrequency(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.paAlleles,
														             (*pLocusNode)->numValues, &probabList);
							#endif		
						} else {
							tableSize = createTableGeneFrequency(locusIndDS, NULL, (*pLocusNode)->numValues, &probabList);
						}

						handleProbabList(tableSize, Product, *pLocusNode, probabList);
					}
				}

				updateConstantVars(*pLocusNode);
			}
			else {	/* The paternal locus node isn't necessary.  Free the memory allocated for it. */ 

				updateEliminatedVars(*pLocusNode);
			}
		}

		/* 1.4.	Check if the paternal selector is necessary. */ 

		if (*pSelectorNode) {
			if (((*pSelectorNode)->numChildren > 0) || (*pSelectorNode)->isKnown) {  

				if (locusIndMap == 0) {		/* This is the selector of the first locus. */
					tableSize = createTableHalf((*pSelectorNode)->numValues, &probabList);
				}
			#if DISCONNECT_UNLINKED
				else if (isUnlinked) {
					tableSize = createTableHalf((*pSelectorNode)->numValues, &probabList);
				}
			#endif
				else if ((!recombData.interference) || (recombData.interference && locusIndMap == 1)) {		

					/* If interference isn't taken into consideration, or if this is 
					   the selector of the second locus on the map. */

					tableSize = createTableRecomb(currPerson, locusIndMap, TRUE, *pSelectorNode, 
												  &probabList);
				}
				else {	/* Interference is taken into consideration and this isn't 
						   the selector of the first or second locus. */
	
					tableSize = createTableRecombInterferenceMale(currPerson, *pSelectorNode, &probabList);
				}

				handleProbabList(tableSize, Product, *pSelectorNode, probabList);
				updateConstantVars(*pSelectorNode);
			}
			else {
				updateEliminatedVars(*pSelectorNode);
			}
		}

		/* 1.5.	Check if the maternal locus node is necessary. */ 

		if (*mLocusNode) {
			if (((*mLocusNode)->numChildren > 0) || (*mLocusNode)->isKnown) {  

				if (currPerson->ma != NULL) {	/* If this isn't a founder. */

					if (((currPerson->ma)->pa != NULL) || (locusIndMap != 0)) {

						#if PERFORM_ALLELE_RECODING
							if (locusArr[locusIndDS].type == alleles)
								tableSize = createTableTransAR(currPerson, locusIndDS, FALSE, *mLocusNode, 
																*mSelectorNode, &probabList);
							else {
						#endif
								tableSize = createTableTrans(currPerson, locusIndDS, FALSE, *mLocusNode, 
																*mSelectorNode, &probabList);
						#if PERFORM_ALLELE_RECODING
							}
						#endif
					}
					else {
						tableSize = createTableTransFounder(currPerson, locusIndDS, FALSE, *mLocusNode,
															*mSelectorNode, &probabList);
					}

					handleProbabList(tableSize, Product, *mLocusNode, probabList);
				}
				else {		/* If this is a founder. */
					if (locusIndMap != 0) {
						if (locusArr[locusIndDS].type == alleles) {
							#if PERFORM_ALLELE_RECODING	
								tableSize = createTableGeneFrequencyAR(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.maAllelsGroups,
																	   (*mLocusNode)->numValues, &probabList);
							#else
								tableSize = createTableGeneFrequency(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.maAlleles,
														             (*mLocusNode)->numValues, &probabList);
							#endif
						} else {
							tableSize = createTableGeneFrequency(locusIndDS, NULL, (*mLocusNode)->numValues, &probabList);
						}

						handleProbabList(tableSize, Product, *mLocusNode, probabList);
					}
				}
					
				updateConstantVars(*mLocusNode);
			}
			else {
				updateEliminatedVars(*mLocusNode);
			}
		}

		/* 1.6.	Check if the maternal selector is necessary. */ 

		if (*mSelectorNode) {
			if (((*mSelectorNode)->numChildren > 0) || (*mSelectorNode)->isKnown) {  

				if (locusIndMap == 0) {
					tableSize = createTableHalf((*mSelectorNode)->numValues, &probabList);
				}
			#if DISCONNECT_UNLINKED
				else if (isUnlinked) {
					tableSize = createTableHalf((*mSelectorNode)->numValues, &probabList);
				}
			#endif
				else if ((!recombData.interference) || (recombData.interference && locusIndMap == 1)) {		
					
					/* If interference isn't taken into consideration, or if this is 
					   the selector of the second locus on the map. */

					tableSize = createTableRecomb(currPerson, locusIndMap, FALSE, *mSelectorNode, 
												  &probabList);
				}
				else {	/* Interference is taken into consideration and this isn't 
						   the selector of the first or second locus. */
	
					tableSize = createTableRecombInterferenceFemale(currPerson, *mSelectorNode, &probabList);
				}

				handleProbabList(tableSize, Product, *mSelectorNode, probabList);
				updateConstantVars(*mSelectorNode);
			}
			else {
				updateEliminatedVars(*mSelectorNode);
			}
		}
	}
}



/******************************************************************************
*
*       Function        createPersonVarsSexLinked
*       Parameters		currPerson (the person whose variables need to be
*						created).
*       Output          none
*       Description		Creates the variables that represent currPerson in the
*						product.
*       Comments		Variables are created only for nodes that have more than 
*						one possible value, namely aren't constant. 
*						Only functions that aren't constant are inserted into  
*						the product (the probability of the constant functions  
*						is added to the joint probability).
*						Used in case of a sex-linked analysis.
*
******************************************************************************/
void createPersonVarsSexLinked(person *currPerson)
{
	int			locusIndMap, locusIndDS, tableSize=0, firstNullLociIndDS,
				secondNullLociIndDS;
	node		**pLocusNode=NULL, **mLocusNode, **mSelectorNode, 
				**phenotypeNode, **genotypeNode;
	Probability *probabList;
	locus		*locusArr;

#if DISCONNECT_UNLINKED
	boolean				isUnlinked;
#endif

	locusArr = loci.locusArr;
	genotypeNode = &(currPerson->genotypeVar);

	/* 1.	Iterate over all loci from the last one (leftmost) on the map to 
			the first one (rightmost). */

	for (locusIndMap = loci.numLoci - 1; locusIndMap >= 0; locusIndMap--) {

		locusIndDS = loci.mapToData[locusIndMap];

		/* 1.1.	Initialize pointers to currPerson's nodes of this locus. */

		mLocusNode = &(currPerson->locusVar[1][locusIndDS]);
		mSelectorNode = &(currPerson->selectorVar[1][locusIndDS]);
		phenotypeNode = &(currPerson->phenotypeVar[locusIndDS]);

		if (currPerson->personSex == Female)
			pLocusNode = &(currPerson->locusVar[0][locusIndDS]);

		#if DISCONNECT_UNLINKED
			if ((locusIndMap > 0) && (recombData.maleThetas[locusIndMap - 1][0] == 0.5)) {
				isUnlinked = TRUE;
			}
			else {
				isUnlinked = FALSE;
			}
		#endif

		/* 1.2.	If necessary, create a variable for the phenotype node. */

		if ((*phenotypeNode) != NULL) {

			switch (locusArr[locusIndDS].type) {

			case affection:

				if (currPerson->personSex == Male) {

					/* 2.1.1.	If this a Male. */

					tableSize = createTableMalePenetrance(&(currPerson->loci[locusIndDS]),
														  locusIndDS, *phenotypeNode, *mLocusNode, 
														   &probabList);
				}
				else if ((currPerson->pa != NULL) || (locusIndMap != 0)) {

					/* 2.1.1.	If this isn't the first locus of a female founder. */

					tableSize = createTablePenetrance(&(currPerson->loci[locusIndDS]), locusIndDS, *phenotypeNode, 
													  *pLocusNode, *mLocusNode, &probabList);
				}
				else {		
					/* 2.1.1.	If this is the first locus of a female founder. */

					tableSize = createTablePenetranceFounder(&(currPerson->loci[locusIndDS]), locusIndDS, *phenotypeNode, 
															 *pLocusNode, *mLocusNode, *genotypeNode, &probabList);
				}
				break;

			case null:
				
				firstNullLociIndDS = loci.mapToData[loci.traitLociPos[0]];
				secondNullLociIndDS = loci.mapToData[loci.traitLociPos[1]];

				/* 2.1.1.	If this is a female. */

				if (currPerson->personSex == Female) {

					/* 2.1.1.1.	If this isn't a founder or this is the second NULL locus on the map 
								of a 2-loci-disease, and neither one of the NULL loci is first on the
								map. */

					if ((currPerson->pa != NULL) || 
							((loci.traitLociPos[0] != 0) && (loci.traitLociPos[1] != 0))) {

						tableSize = createTablePenetranceTwo(currPerson, firstNullLociIndDS, 
															secondNullLociIndDS, 
															*phenotypeNode, &probabList);
					}
					else {	/*	One of the null loci is the first locus of a founder. */	

					tableSize = createTablePenetranceFounderTwo(currPerson, firstNullLociIndDS, 
																secondNullLociIndDS, *phenotypeNode, 
																*genotypeNode, loci.traitLociPos[0], &probabList);
					}
				}
				else {		/* 2.1.1.	If this is a male. */

					tableSize = createTableMalePenetranceTwo(currPerson, firstNullLociIndDS, 
															 secondNullLociIndDS, *phenotypeNode,
															 &probabList);
				}

				break;

			case alleles:

				tableSize = createTablePhenotype(*phenotypeNode, *pLocusNode, *mLocusNode, &probabList);
				break;
			}

			handleProbabList(tableSize, Product, *phenotypeNode, probabList);
			updateConstantVars(*phenotypeNode);
		}

		/* 2.2.	Check if the paternal locus node is necessary. */ 

		if ((currPerson->personSex == Female) && (*pLocusNode)) {
			if (((*pLocusNode)->numChildren > 0) ||  (*pLocusNode)->isKnown) {  

				if (currPerson->pa != NULL) {	/* 'currPerson' isn't a founder. */

					#if PERFORM_ALLELE_RECODING
						if (locusArr[locusIndDS].type == alleles)
							tableSize = createTableTransMaleAR(currPerson, locusIndDS, *pLocusNode,
															   &probabList);
						else {
					#endif
							tableSize = createTableTransMale(currPerson, locusIndDS, *pLocusNode,
														&probabList);
					#if PERFORM_ALLELE_RECODING
						}
					#endif
				
					handleProbabList(tableSize, Product, *pLocusNode, probabList);
				}
				else {							/* 'currPerson' is a founder. */
					if (locusIndMap == 0) {
						if (locusArr[locusIndDS].type == alleles) {
							#if PERFORM_ALLELE_RECODING
								tableSize = createTableGenotypeFrequencyAR(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.paAllelsGroups,
																		   (*pLocusNode)->numValues, &probabList, currPerson->loci[locusIndDS].unknown);
							#else
								tableSize = createTableGenotypeFrequency(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.paAlleles,
																		 (*pLocusNode)->numValues, &probabList, currPerson->loci[locusIndDS].unknown);
							#endif
						} else
							tableSize = createTableGenotypeFrequency(locusIndDS, NULL, (*pLocusNode)->numValues, &probabList, TRUE);

						handleProbabList(tableSize, Product, *genotypeNode, probabList);
					}
					else {
						if (locusArr[locusIndDS].type == alleles) {
							#if PERFORM_ALLELE_RECODING
								tableSize = createTableGeneFrequencyAR(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.paAllelsGroups,
																	   (*pLocusNode)->numValues, &probabList);
							#else
								tableSize = createTableGeneFrequency(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.paAlleles,
														             (*pLocusNode)->numValues, &probabList);
							#endif
						} else {
							tableSize = createTableGeneFrequency(locusIndDS, NULL, (*pLocusNode)->numValues, &probabList);
						}

						handleProbabList(tableSize, Product, *pLocusNode, probabList);
					}
				}

				updateConstantVars(*pLocusNode);
			}
			else {	/* The paternal locus node isn't necessary. Free the memory allocated for it. */ 

				updateEliminatedVars(*pLocusNode);
			}
		}

		/* 2.3.	Check if the maternal locus node is necessary. */ 

		if (*mLocusNode) {
			if (((*mLocusNode)->numChildren > 0) || (*mLocusNode)->isKnown) { 

				if (currPerson->ma != NULL) {	/* If this isn't a founder. */

					if (((currPerson->ma)->pa != NULL) || (locusIndMap != 0)) {

						#if PERFORM_ALLELE_RECODING
							if (locusArr[locusIndDS].type == alleles)
								tableSize = createTableTransAR(currPerson, locusIndDS, FALSE, *mLocusNode, 
															*mSelectorNode, &probabList);
							else {
						#endif
								tableSize = createTableTrans(currPerson, locusIndDS, FALSE, *mLocusNode, 
															*mSelectorNode, &probabList);
						#if PERFORM_ALLELE_RECODING
							}
						#endif
					}
					else {
						tableSize = createTableTransFounder(currPerson, locusIndDS, FALSE, *mLocusNode,
															*mSelectorNode, &probabList);
					}

					handleProbabList(tableSize, Product, *mLocusNode, probabList);
				}
				else {		/* If this is a founder. */
					/*	If this is a male, or this isn't the first locus on the map,
						then create a gene-frequency probability table. */

					if ((currPerson->personSex == Male) || (locusIndMap != 0)) {

						if (locusArr[locusIndDS].type == alleles) {
							#if PERFORM_ALLELE_RECODING
								tableSize = createTableGeneFrequencyAR(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.maAllelsGroups,
																	   (*mLocusNode)->numValues, &probabList);
							#else
								tableSize = createTableGeneFrequency(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.maAlleles,
																	 (*mLocusNode)->numValues, &probabList);
							#endif
						} else {
							tableSize = createTableGeneFrequency(locusIndDS, NULL, (*mLocusNode)->numValues, &probabList);
						}

						handleProbabList(tableSize, Product, *mLocusNode, probabList);
					}
				}
	
				updateConstantVars(*mLocusNode);
			}
			else {
				updateEliminatedVars(*mLocusNode);
			}
		}

		/* 2.5.	Check if the maternal selector is necessary. */ 

		if (*mSelectorNode) {
			if (((*mSelectorNode)->numChildren > 0) || (*mSelectorNode)->isKnown) { 

				if (locusIndMap == 0) {			/* This is the selector of the first locus. */
					tableSize = createTableHalf((*mSelectorNode)->numValues, &probabList);
				}
			#if DISCONNECT_UNLINKED
				else if (isUnlinked) {
					tableSize = createTableHalf((*mSelectorNode)->numValues, &probabList);
				}
			#endif
				else if ((!recombData.interference) || (recombData.interference && locusIndMap == 1)) {		
					
					/* If interference isn't taken into consideration, or if this is 
					   the selector of the second locus. */

					tableSize = createTableRecomb(currPerson, locusIndMap, FALSE, *mSelectorNode,
												  &probabList);
				}
				else {	  /* Interference is taken into consideration and this isn't 
							 the selector of the first or second locus. */
	
					tableSize = createTableRecombInterferenceFemale(currPerson, *mSelectorNode, &probabList);
				}

				handleProbabList(tableSize, Product, *mSelectorNode, probabList);
				updateConstantVars(*mSelectorNode);
			}
			else {
				updateEliminatedVars(*mSelectorNode);
			}
		}
	}
}


/******************************************************************************
*
*       Function        createPersonVarsAutosomalHaplo
*       Parameters		currPerson (the person whose variables need to be 
*						created).
*       Output          none
*       Description		Creates the variables that represent 'currPerson' in the
*						product.
*       Comments		Variables are created only for nodes that have more than 
*						one possible value. Only functions that aren't constant
*						are inserted into the product (the probability of the
*						constant functions is added to the joint probability).
*						Used in case of an autosomal haplotyping analysis.
*
******************************************************************************/
void createPersonVarsAutosomalHaplo(person *currPerson)
{
	int			locusIndMap, locusIndDS, tableSize;
	node		**pLocusNode, **mLocusNode, **pSelectorNode, **mSelectorNode, 
				**phenotypeNode;
	Probability *probabList;

#if DISCONNECT_UNLINKED
	boolean				isUnlinked;
#endif

	/* 1.	Iterate over all marker (alleles type) loci from the last one 
			(leftmost) on the map to the first one (rightmost). */

	for (locusIndMap = numMarkers - 1; locusIndMap >= 0; locusIndMap--) {

		locusIndDS = loci.mapToData[locusIndMap];

		/* 1.1.	Initialize pointers to currPerson's nodes of this locus. */

		pLocusNode = &(currPerson->locusVar[0][locusIndDS]);
		mLocusNode = &(currPerson->locusVar[1][locusIndDS]);
		pSelectorNode = &(currPerson->selectorVar[0][locusIndDS]);
		mSelectorNode = &(currPerson->selectorVar[1][locusIndDS]);
		phenotypeNode = &(currPerson->phenotypeVar[locusIndDS]);

		#if DISCONNECT_UNLINKED
			if ((locusIndMap > 0) && (recombData.maleThetas[locusIndMap - 1][0] == 0.5)) {
				isUnlinked = TRUE;
			}
			else {
				isUnlinked = FALSE;
			}
		#endif

		/* 1.2.	If necessary, create a variable for the phenotype node. */

		if ((*phenotypeNode) != NULL) {

			/* 1.2.1.	Create this nodes probability function. */

			tableSize = createTablePhenotype(*phenotypeNode, *pLocusNode, *mLocusNode, &probabList);

			handleProbabList(tableSize, Product, *phenotypeNode, probabList);
			updateConstantVars(*phenotypeNode);
		}

		/* 1.3.	Check if the paternal locus node is necessary. */ 

		if (*pLocusNode) {
			if (currPerson->pa != NULL) {	/* 'currPerson' isn't a founder. */

				/* 1.3.2.	Create this node's probability function. */

				#if PERFORM_ALLELE_RECODING
					tableSize = createTableTransAR(currPerson, locusIndDS, TRUE, *pLocusNode, 
												   *pSelectorNode, &probabList);
				#else
					tableSize = createTableTrans(currPerson, locusIndDS, TRUE, *pLocusNode, 
												*pSelectorNode, &probabList);
				#endif

				/* 1.3.3.	If the function isn't a constant, insert it into the product.
							Otherwise, add its log-probability to the total log-probablity. */

				handleProbabList(tableSize, Product, *pLocusNode, probabList);
			}
			else {							/* 'currPerson' is a founder. */

				/* 1.3.1.	Create a gene-frequency probability table for the paternal locus. */

				#if PERFORM_ALLELE_RECODING
					tableSize = createTableGeneFrequencyAR(locusIndDS, 
														 currPerson->loci[locusIndDS].locusData.allelesL.paAllelsGroups,
										 				 (*pLocusNode)->numValues, &probabList);
				#else				
					tableSize = createTableGeneFrequency(locusIndDS, 
														 currPerson->loci[locusIndDS].locusData.allelesL.paAlleles,
										 				 (*pLocusNode)->numValues, &probabList);
				#endif	

				handleProbabList(tableSize, Product, *pLocusNode, probabList);
			}

			updateConstantVars(*pLocusNode);
		}

		/* 1.4.	Check if the paternal selector is necessary. */ 

		if (*pSelectorNode) {
			if (((*pSelectorNode)->numChildren > 0) || (*pSelectorNode)->isKnown) {  

				if (locusIndMap == 0) {		/* This is the selector of the first locus. */
					tableSize = createTableHalf((*pSelectorNode)->numValues, &probabList);
				}
			#if DISCONNECT_UNLINKED
				else if (isUnlinked) {
					tableSize = createTableHalf((*pSelectorNode)->numValues, &probabList);
				}
			#endif
				else if ((!recombData.interference) || (recombData.interference && locusIndMap == 1)) {		

					/*	If interference isn't taken into consideration, or if this is 
						the selector of the second locus on the map. */

					tableSize = createTableRecomb(currPerson, locusIndMap, TRUE, *pSelectorNode,
													&probabList);
				}
				else {	/* Interference is taken into consideration and this isn't 
						   the selector of the first or second locus. */
	
					tableSize = createTableRecombInterferenceMale(currPerson, *pSelectorNode, &probabList);
				}

				handleProbabList(tableSize, Product, *pSelectorNode, probabList);
				updateConstantVars(*pSelectorNode);
			}
			else {
				updateEliminatedVars(*pSelectorNode);
			}
		}

		/* 1.5.	Check if the maternal locus node is necessary. */ 

		if (*mLocusNode) {
			if (currPerson->ma != NULL) {	/* If this isn't a founder. */

				#if PERFORM_ALLELE_RECODING
					tableSize = createTableTransAR(currPerson, locusIndDS, FALSE, *mLocusNode, 
												  *mSelectorNode, &probabList);
				#else
					tableSize = createTableTrans(currPerson, locusIndDS, FALSE, *mLocusNode, 
												 *mSelectorNode, &probabList);
				#endif

				handleProbabList(tableSize, Product, *mLocusNode, probabList);
			}
			else {		/* If this is a founder. */

				#if PERFORM_ALLELE_RECODING
					tableSize = createTableGeneFrequencyAR(locusIndDS, 
														 currPerson->loci[locusIndDS].locusData.allelesL.maAllelsGroups,
														 (*mLocusNode)->numValues, &probabList);
				#else				
					tableSize = createTableGeneFrequency(locusIndDS, 
														 currPerson->loci[locusIndDS].locusData.allelesL.maAlleles,
														 (*mLocusNode)->numValues, &probabList);
				#endif
					
				handleProbabList(tableSize, Product, *mLocusNode, probabList);
			}


			updateConstantVars(*mLocusNode);
		}

		/* 1.6.	Check if the maternal selector is necessary. */ 

		if (*mSelectorNode) {
			if (((*mSelectorNode)->numChildren > 0) || (*mSelectorNode)->isKnown) {  

				if (locusIndMap == 0) {
					tableSize = createTableHalf((*mSelectorNode)->numValues, &probabList);
				}
			#if DISCONNECT_UNLINKED
				else if (isUnlinked) {
					tableSize = createTableHalf((*mSelectorNode)->numValues, &probabList);
				}
			#endif
				else if ((!recombData.interference) || (recombData.interference && locusIndMap == 1)) {		
					
					/* If interference isn't taken into consideration, or if this is 
					   the selector of the second locus on the map. */

					tableSize = createTableRecomb(currPerson, locusIndMap, FALSE, *mSelectorNode,
													&probabList);
				}
				else {		/* Interference is taken into consideration and this isn't 
							   the selector of the first or second locus. */
	
					tableSize = createTableRecombInterferenceFemale(currPerson, *mSelectorNode, &probabList);
				}

				handleProbabList(tableSize, Product, *mSelectorNode, probabList);
				updateConstantVars(*mSelectorNode);
			}
			else {
				updateEliminatedVars(*mSelectorNode);
			}
		}
	}
}



/******************************************************************************
*
*       Function        createPersonVarsSexLinkedHaplo
*       Parameters		currPerson (the person whose variables need to be
*						created).
*       Output          none
*       Description		Creates the variables that represent 'currPerson' in the
*						product.
*       Comments		Variables are created only for nodes that have more than 
*						one possible value. Only functions that aren't constant
*						are inserted into the product (the probability of the
*						constant functions is added to the joint probability).
*						Used in case of an sex-linked haplotyping analysis.
*
******************************************************************************/
void createPersonVarsSexLinkedHaplo(person *currPerson)
{
	int			locusIndMap, locusIndDS, tableSize;
	node		**pLocusNode=NULL, **mLocusNode, **mSelectorNode, 
				**phenotypeNode;
	Probability *probabList;

#if DISCONNECT_UNLINKED
	boolean				isUnlinked;
#endif

	/* 1.	Iterate over all marker (alleles type) loci from the last one (leftmost) 
			on the map to the first one (rightmost). */

	for (locusIndMap = numMarkers - 1; locusIndMap >= 0; locusIndMap--) {

		locusIndDS = loci.mapToData[locusIndMap];

		/* 1.1.	Initialize pointers to currPerson's nodes of this locus. */

		mLocusNode = &(currPerson->locusVar[1][locusIndDS]);
		mSelectorNode = &(currPerson->selectorVar[1][locusIndDS]);
		phenotypeNode = &(currPerson->phenotypeVar[locusIndDS]);

		if (currPerson->personSex == Female)
			pLocusNode = &(currPerson->locusVar[0][locusIndDS]);

		#if DISCONNECT_UNLINKED
			if ((locusIndMap > 0) && (recombData.maleThetas[locusIndMap - 1][0] == 0.5)) {
				isUnlinked = TRUE;
			}
			else {
				isUnlinked = FALSE;
			}
		#endif

		/* 1.2.	If necessary, create a variable for the phenotype node. */

		if ((*phenotypeNode) != NULL) {

			tableSize = createTablePhenotype(*phenotypeNode, *pLocusNode, *mLocusNode, &probabList);

			handleProbabList(tableSize, Product, *phenotypeNode, probabList);
			updateConstantVars(*phenotypeNode);
		}

		/* 2.2.	Check if the paternal locus node is necessary. */ 

		if ((currPerson->personSex == Female) && (*pLocusNode)) {
			if (currPerson->pa != NULL) {	/* 'currPerson' isn't a founder. */

				#if	PERFORM_ALLELE_RECODING
					tableSize = createTableTransMaleAR(currPerson, locusIndDS, *pLocusNode,
													   &probabList);
				#else
					tableSize = createTableTransMale(currPerson, locusIndDS, *pLocusNode,
													&probabList);
				#endif
				
				handleProbabList(tableSize, Product, *pLocusNode, probabList);
			}
			else {							/* 'currPerson' is a founder. */

				#if	PERFORM_ALLELE_RECODING
				tableSize = createTableGeneFrequencyAR(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.paAllelsGroups,
													 	   (*pLocusNode)->numValues, &probabList);
				#else
					tableSize = createTableGeneFrequency(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.paAlleles,
													 	 (*pLocusNode)->numValues, &probabList);
				#endif

				handleProbabList(tableSize, Product, *pLocusNode, probabList);
			}

			updateConstantVars(*pLocusNode);
		}

		/* 2.3.	Check if the maternal locus node is necessary. */ 

		if (*mLocusNode) {
			if (currPerson->ma != NULL) {	/* If this isn't a founder. */

				#if	PERFORM_ALLELE_RECODING
					tableSize = createTableTransAR(currPerson, locusIndDS, FALSE, *mLocusNode, 
												   *mSelectorNode, &probabList);
				#else
					tableSize = createTableTrans(currPerson, locusIndDS, FALSE, *mLocusNode, 
												  *mSelectorNode, &probabList);
				#endif
			}
			else {		/* If this is a founder create a gene-frequency probability table. */

				#if	PERFORM_ALLELE_RECODING
				tableSize = createTableGeneFrequencyAR(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.maAllelsGroups,
															(*mLocusNode)->numValues, &probabList);
				#else
					tableSize = createTableGeneFrequency(locusIndDS, currPerson->loci[locusIndDS].locusData.allelesL.maAlleles,
															(*mLocusNode)->numValues, &probabList);
				#endif
			}

			handleProbabList(tableSize, Product, *mLocusNode, probabList);
			updateConstantVars(*mLocusNode);
		}

		/* 2.5.	Check if the maternal selector is necessary. */ 

		if (*mSelectorNode) {
			if (((*mSelectorNode)->numChildren > 0) || (*mSelectorNode)->isKnown) { 

				if (locusIndMap == 0) {			/* This is the selector of the first locus. */
					tableSize = createTableHalf((*mSelectorNode)->numValues, &probabList);
				}
			#if DISCONNECT_UNLINKED
				else if (isUnlinked) {
					tableSize = createTableHalf((*mSelectorNode)->numValues, &probabList);
				}
			#endif
				else if ((!recombData.interference) || (recombData.interference && locusIndMap == 1)) {		
					
					/* If interference isn't taken into consideration, or if this is 
					   the selector of the second locus. */

					tableSize = createTableRecomb(currPerson, locusIndMap, FALSE, *mSelectorNode,
													&probabList);
				}
				else {	  /* Interference is taken into consideration and this isn't 
							 the selector of the first or second locus. */
	
					tableSize = createTableRecombInterferenceFemale(currPerson, *mSelectorNode, &probabList);
				}

				handleProbabList(tableSize, Product, *mSelectorNode, probabList);
				updateConstantVars(*mSelectorNode);
			}
			else {
				updateEliminatedVars(*mSelectorNode);
			}
		}
	}
}



/******************************************************************************
*
*       Function        handleProbabList
*       Parameters		tableSize (the size of the probability table), Product
*						(pointer to the product into which the function needs to
*						be inserted), Node (the node to which the probability
*						table is associated), probabList (the probability list).
*       Output          none
*       Description		Inserts a function into the product in case the probability
*						list is longer than 1; otherwise, frees it and adds its 
*						probability to the result.
*       Comments		none.
*
******************************************************************************/
void handleProbabList(int tableSize, product *Product, node *Node, 
					  Probability *probabList)
{
	int			numVariables;
	variable	**variables;
	function*   VarFunction;

	if (tableSize > 1) {
		variables = getFunctionVariables(Product, Node, &numVariables);
		VarFunction = insertFunctionToProduct(Product, tableSize, probabList, 0.0, TRUE, 
			numVariables, variables, FALSE, MAX_RESCALE_UNITS);
		if (progData.programCode == BAYES_NET_XBN_BEFORE_ELIM ||
			progData.programCode == BAYES_NET_XBN_AFTER_ELIM ||
			progData.programCode == BAYES_NET_XML_BEFORE_ELIM ||
			progData.programCode == BAYES_NET_XML_AFTER_ELIM 
#ifdef PARALLEL
	// if we only prepare the elim order
		|| PERFORM_PRODUCT_BREAKING==0
#endif

){

				connectFuncToNode(VarFunction, Node);

		}
	}
	else {		/* the function is a constant. */
		tempRes += log(probabList[0]);
		myFree(probabList);
	}
}

/******************************************************************************
*
*       Function        copyProduct
*       Parameters      oldProd (pointer to the product that needs to be 
*						duplicated), elimOrd (the elimination order of the 
*						variables in the product), isTransformedArr (an array of 
*						flags that indicate for each node in the elimination order
*						whether it's corresponding variable has been copied or not),
*						condVarsOld (an array of pointers to a couple of variables
*						in 'oldProd'), condVarsNew (will hold pointers to the new
*						variables in the duplicated product that have been created 
*						due to the variables in 'condVarsOld'), numCondVars (the
*						number of variables in the array 'condVarsOld').
*       Output          none
*       Description		Duplicates the product 'oldProd' and returns the new
*						duplicated product.
*       Comments		none
*
******************************************************************************/
product* copyProduct(product *oldProd, elimOrderPed *elimOrd, 
					 boolean *isTransformedArr, variable **condVarsOld, 
					 variable **condVarsNew, int numCondVars)
{
	product			*newProd;
	variable		*var, *varEnd, **oldVarPtr, **oldVarPtrEnd, **newVarPtr,
					**oldToNewVarMapping, **funcNewVars, *newVar;
	function		*func, *funcEnd;
	int				maxVarsInFunc;		
	elimOrderNode	*currNode, *endNode;
	boolean			*isTransformedArrPtr;

	
	/* 1.	Allocate memory for the new product. */

	newProd = allocateProduct(oldProd->currNumVars + 1, oldProd->currNumFuncs + 1);

	maxVarsInFunc = 0;

	/* 2.	Find the maximum number of variables in a function. */

	for (func = oldProd->functions, funcEnd = func + oldProd->numFunctions; 
			func < funcEnd; func++) {
		if (func->used) {
			if (func->numVariables > maxVarsInFunc)
				maxVarsInFunc = func->numVariables;
		}
	}

	/* 3.	Allocate memory for auxiliary data-structures. */

	myMalloc(oldToNewVarMapping, sizeof(variable*) * oldProd->numVariables);
	myMalloc(funcNewVars, sizeof(variable*) * maxVarsInFunc);

	/* 4.	Create new variables. */

	for (var = oldProd->variables, varEnd = var + oldProd->numVariables; var < varEnd; var++) {
		if (var->used) {
			newVar = insertVariableToProduct(newProd, var->originalNode, var->numValues);
			
			#if UNITE_VARS
				updateCertainVals(newVar, (var->originalNode)->uniteT, var->certainVal);
			#endif

			oldToNewVarMapping[var - oldProd->variables] = newVar;
		}
	}

	/* 5.	Save the old values of the variables in the elimination-order-array.
			Each variable points to the variable in the original product from
			which it was duplicated. The node of the Bayesian network is updated
			to point to the variable in the new product which is representing it. */

	for (currNode = &(elimOrd->elimOrderArr)[elimOrd->currElimNode], 
			isTransformedArrPtr = isTransformedArr,
			endNode = &(elimOrd->elimOrderArr)[elimOrd->numNodes]; 
			currNode < endNode; currNode++, isTransformedArrPtr++) {

		var = (currNode->originalNode)->varInProd;

		if (var->used == TRUE) {
			newVar = oldToNewVarMapping[var - oldProd->variables];
			newVar->varInOldProd = var;
			(newVar->originalNode)->varInProd = newVar;
			*isTransformedArrPtr = TRUE;
		}
		else
			*isTransformedArrPtr = FALSE;
	}

	/* 6.	Create new functions. */

	for (func = oldProd->functions, funcEnd = func + oldProd->numFunctions; 
			func < funcEnd; func++) {
		if (func->used) {
			for (oldVarPtr = func->variables, oldVarPtrEnd = oldVarPtr + func->numVariables,
					newVarPtr = funcNewVars; oldVarPtr < oldVarPtrEnd; oldVarPtr++, newVarPtr++) {
				
				*newVarPtr = oldToNewVarMapping[*oldVarPtr - oldProd->variables];
			}

			insertFunctionToProduct(newProd, func->sizeOfTable, func->probabList, func->logMultFactor,
									FALSE, func->numVariables, funcNewVars, TRUE, func->rescaleTimer);
		}
	}

	/* 7.	Build new variable->function lookup tables. */

	buildVariableFunctionTables(newProd);

	/* 8.	Iterate over the variables in 'condVarsOld' array and update
			pointers to the replicates of these variables in the new product. */

	for (oldVarPtr = condVarsOld, oldVarPtrEnd = oldVarPtr + numCondVars,
			newVarPtr = condVarsNew; oldVarPtr < oldVarPtrEnd; oldVarPtr++,
			newVarPtr++) {

		*newVarPtr = oldToNewVarMapping[*oldVarPtr - oldProd->variables];
	}

	/* 9.	Free memory allocated dynamcially. */

	myFree(oldToNewVarMapping);
	myFree(funcNewVars);

	/* 10.	Return a pointer to the new product. */

	return newProd;
}



/******************************************************************************
*
*       Function        zeroProductFields
*       Parameters      prodArr (the array of products whose fields need to be
*						reset), sexLink (a flag that indicates whether this is a 
*						sex-linked analysis or not).
*       Output          none
*       Description		Resets the fields of the product and its variables.
*       Comments		none
*
******************************************************************************/
void zeroProductFields(product **prodArr, boolean sexLink)
{
	variable	*varPtr, *varPtrEnd;
	function	*funcPtr, *funcPtrEnd;
	product		**prodPtr, **prodPtrEnd, *currProd;

	
	/* 1.	Iterate over the products. */

	for (prodPtr = prodArr, prodPtrEnd = prodPtr + peds.numPeds; prodPtr < prodPtrEnd; 
			prodPtr++) {

		currProd = *prodPtr;

		/* 1.1.	Free the memory allocated for the variables and remove them from 
				the product. */

		for (varPtr = currProd->variables, varPtrEnd = varPtr + currProd->numVariables;
				varPtr < varPtrEnd; varPtr++) {
			if (varPtr->used) {
				removeVariableFromProduct(currProd, varPtr);
			}
		}

		/* 1.2.	Free the memory allocated for the functions and remove them from 
				the product. */

		for (funcPtr = currProd->functions, funcPtrEnd = funcPtr + currProd->numFunctions;
				funcPtr < funcPtrEnd; funcPtr++) {
			if (funcPtr->used) {
				removeFunctionFromProduct(currProd, funcPtr);
			}
		}

		/* 1.3.	Reset the fields of the product. */

		currProd->currNumFuncs = 0;
		currProd->currNumVars = 0;
		currProd->nextFuncPlace = currProd->functions;
		currProd->nextVarPlace = currProd->variables;
	}

	/* 2.	Zero the 'varInProd' field for all the nodes of the Bayesian networks. */

	if (sexLink)
		zeroNodeFieldsPeds(zeroNodeFieldsPedSexLinked);
	else
		zeroNodeFieldsPeds(zeroNodeFieldsPedAutosomal);
}


/******************************************************************************
*
*       Function        freeProducts
*       Parameters		prodArr (an array of products), numProds (the number of
*						products in the array).
*       Output          none
*       Description		Frees the memory allocated dynamically to the products.
*       Comments		none
*
******************************************************************************/
void freeProducts(product** prodArr, int numProds) 
{
	int		pedNum;

	/* 1.	Iterate over the products. */

	for (pedNum = 0 ; pedNum < numProds; pedNum++) {
	
		/* 1.1	Free the memory allocated to the product. */

		freeProduct(prodArr[pedNum]);
	}

	/* 4.	Free the memory allocated to the product array. */

	myFree(prodArr);
}



/******************************************************************************
*
*       Function        freeProduct
*       Parameters      prod (the product that needs to be freed).
*       Output          none
*       Description		Frees the memory that was allocated for 'prod'.
*       Comments		none
*
******************************************************************************/
void freeProduct(product *prod)
{
	variable	*varPtr, *varPtrEnd;
	function	*funcPtr, *funcPtrEnd;


	/* 1.	Free the memory allocated to the variables and remove them from 
			the product. */

	if (prod->currNumVars) {
		for (varPtr = prod->variables, varPtrEnd = varPtr + prod->numVariables;
				varPtr < varPtrEnd; varPtr++) {
			if (varPtr->used) {
				removeVariableFromProduct(prod, varPtr);
			}
		}
	}

	/* 2.	Free the memory allocated to the functions and remove them from 
			the product. */

	if (prod->currNumFuncs) {
		for (funcPtr = prod->functions, funcPtrEnd = funcPtr + prod->numFunctions;
				((prod->currNumFuncs > 0) && (funcPtr < funcPtrEnd)); funcPtr++) {
			if (funcPtr->used) {
				removeFunctionFromProduct(prod, funcPtr);
			}
		}
	}

	/* 3.	Free the Product's arrays and structure. */

	myFree(prod->variables);
	myFree(prod->functions);
	myFree(prod);
}




/******************************************************************************
*
*       Function        printProduct
*       Parameters      prod (the product which needs to be printed).
*       Output          none
*       Description		Prints the information on the product.
*       Comments		Used for debugging purposes.
*
******************************************************************************/
#if DEBUG_STAT
void printProduct(product *prod, double tmpResLn)
{
	function	*func, *funcEnd;
	int			count = 0;

	/* 1.	Open the file for writing. */

	outFile = fopen("out.txt", "w");

	fprintf(outFile, "lnLike %lf %lf\n", exp(totalRes.currRes->lnLike),totalRes.currRes->lnLike);
	fprintf(outFile, "tmpRes %lf %lf\n", exp(tmpResLn), tmpResLn);

	/* 2.	Print the number of functions in the product. */
    fprintf(outFile, "== Printing product ==\n");
	fprintf(outFile, "Total num of functions %ld\n", prod->numFunctions);
	fclose(outFile);
	/* 3.	Iterate over the function slots. */

	for (func = prod->functions, funcEnd = func + prod->numFunctions;
			func < funcEnd; func++) {

		/* 3.1.	If the function is used, print it. */

		if (func->used) {
			printFunction(func);
			count++;
		}
	}

	/* 4.	Print the number of functions and close the file. */
outFile = fopen("out.txt", "a");
	fprintf(outFile, "%d\n", count);
	fclose(outFile);
}
#endif


