/* $Id: ProductBreaking.c,v 1.2 2008/08/03 08:15:46 superlink Exp $ */
/*********************************************************************************/
/*  ProductBreaking.c														     */
/*  Author: Ma'ayan Fishelson				                                     */
/*********************************************************************************/

#include "Product.h"
#include "ProductValueElimination.h"
#include "Inference.h"


/******************************************************************************
*
*       Function        inferProdWithCond
*       Parameters		prod (the product on which inference is performed), 
*						breakVar (the variable on which the product is broken),
*						elimOrd (the elimination order according to which the 
*						variables of the product are processed),
*						problemSize (the size of the problem that needs to be
*						solved by the current call to the function),
*						solved (the size of the problem that has already been
*						solved).
*       Output          none
*       Description		Performs inference on 'prod' by breaking the Product						 
*						on 'breakVar' (conditioning on it) and performing 
*						inference for each value of the breakVar seperately 
*						and then combining the results.
*       Comments		Returns the result of the inference.
*
******************************************************************************/
Probability inferProdWithCond(product *prod, variable *breakVar,
							  elimOrderPed *elimOrd, float problemSize, 
							  float solved)
{
	int					breakVarVal;
	product				*brokenProd;
	variable			*brokenVar;
	Probability			*resArr, res;
	double				maxLogVal = MIN_DBL;
	float				newProblemSize, currSolved;
	long				currElimNode;
	elimOrderNode		*currNode, *endNode;
	boolean				*isTransformedArr, *isTransformedArrPtr, isZeroLike, 
						areValsElim;
	int	maxBreakNodeValue;
	int minBreakNodeValue;
#if UNITE_VARS
	changeUniteTList	*listPtr=NULL;
#endif

	/* 1.	If the break var needs to be treated jointly with other variables, then
			call the function that handles this case. */

	#if UNITE_VARS
		if (breakVar->originalNode->uniteT) {
			return inferProdWithCondJoint(prod, breakVar, elimOrd, problemSize, 
										solved);
		}
	#endif

	/* 2.	Allocate memory for an array that will hold the results of the 
			inference for each value of the break variable. */

	myMalloc(resArr, sizeof(Probability) * breakVar->numValues);

	/* 3.	Allocate memory for an array that will hold the temporary values of
			variables in the elimination-order-array. */

	myMalloc(isTransformedArr, sizeof(boolean) * (elimOrd->numNodes - elimOrd->currElimNode));

	/* 3.	Calculate the size of the problem that will be solved by each call
			to the 'inferProduct' function. */

	newProblemSize = problemSize / breakVar->numValues;
	currSolved = solved;

	currElimNode = elimOrd->currElimNode;

#ifdef PARALLEL
	if( stackDepth == maxBreakStack[ped_num]){
	  /* we proceed as normal if we already reached the required number of break nodes */
#endif

	  maxBreakNodeValue=breakVar->numValues;
	  minBreakNodeValue=0;
	  
#ifdef PARALLEL

	}else{
	
	  /* get the assignment for the current break node */
	  /* instead of counting likelihood for all assignments*/
	  minBreakNodeValue = breakNodeAssignment[ped_num][stackDepth++];
	  maxBreakNodeValue=minBreakNodeValue+1;
	  
	  /* if some of the previous values of this break node have been trimmed */
	  /* this computational path is superfluous and should be aborted */
	  if(minBreakNodeValue >= breakVar->numValues){ 
		/* print the special symbol - the end of the output*/
		printf("@@@###$$$\n");
		exit(0);
	  }
	}
#endif
	/* 4.	Perform inference on the product for each value of the breakVar seperately .*/

	for (breakVarVal=minBreakNodeValue; breakVarVal < maxBreakNodeValue; breakVarVal++) {
	  
	  /* 4.1.	Duplicate the product. */

		brokenProd = copyProduct(prod, elimOrd, isTransformedArr, &breakVar, 
								 &brokenVar, 1);

		/* 4.2.	Set the value of 'brokenVar' to 'breakVarVal' and change 
				'brokenProd' accordingly. */

		resArr[breakVarVal] = setBreakVal(brokenProd, brokenVar, breakVarVal);

		/* 4.3.	Eliminate impossible variable values from the product's functions. */

		resArr[breakVarVal] += eliminateImpossibleValues(brokenProd, &isZeroLike, &areValsElim);

		/* 4.4.	If some variable values have been eliminated, update the unite status
				of nodes whose values have been eliminated, in case necessary. */

		#if UNITE_VARS
			if (areValsElim) {
				listPtr = updateUniteVarsPed(currPed);
			}
		#endif

		/* 4.5.	If this value of the break var is possible: */

		if (!isZeroLike) {

			/* 4.5.1.	Perform inference on 'brokenProd'. */

			resArr[breakVarVal] += inferProd(brokenProd, elimOrd, newProblemSize, currSolved);

			/* 4.5.2.	Compare the current result to the maximum result up to now. */

			if (resArr[breakVarVal])
				if (resArr[breakVarVal] > maxLogVal) 
					maxLogVal = resArr[breakVarVal];
		}
		else {
			resArr[breakVarVal] = MIN_DBL;

			/* 4.5.1.	If requested, print a progress report. */

			#if PRINT_PROGRESS_REPORT
				printf("Solved %f percent of the problem.\n", (currSolved+newProblemSize)*100);
			#endif
		}

		elimOrd->currElimNode = currElimNode;

		/* 4.6.	Restore the previous values of the nodes in the elimination-order array. */

		for (currNode = &(elimOrd->elimOrderArr)[elimOrd->currElimNode],
				endNode = &(elimOrd->elimOrderArr)[elimOrd->numNodes], 
				isTransformedArrPtr = isTransformedArr; 
				currNode < endNode; currNode++, isTransformedArrPtr++) {

			if (*isTransformedArrPtr == TRUE) {
				(currNode->originalNode)->varInProd = 
							((currNode->originalNode)->varInProd)->varInOldProd;
			}
		}

		/* 4.7.	Restore values of unite nodes, in case of nodes whose values changed after
				the value elimination. */

		#if UNITE_VARS
			if (areValsElim) {
				restoreUniteVals(listPtr);
			}
		#endif

		/* 4.8.	Free the memory allocated for the duplicated product. */

		freeProduct(brokenProd);

		/* 4.9.	Update the size of the problem solved up to now. */

		currSolved += newProblemSize;
	}

	/* 5.	Combine the results of the inference on the different values of breakVar. */

	if (maxLogVal != MIN_DBL) {
		res = 0.0;

		for (breakVarVal = minBreakNodeValue; breakVarVal < maxBreakNodeValue; breakVarVal++) {
			if (resArr[breakVarVal] != MIN_DBL)
				res += exp(resArr[breakVarVal] - maxLogVal);
		}

		res = log(res) + maxLogVal;
	}
	else {
		res = MIN_DBL;
	}

#ifdef PARALLEL	
	/* if we used this break node for splitting -> climb back in the stack */
	/* minBreakNodeValue!=0 only if we use this node for splitting.*/
	if( minBreakNodeValue!=0 ){
	  stackDepth--;
	}
#endif


	/* 6.	Free the memory allocated dynamically. */

	myFree(resArr);
	myFree(isTransformedArr);

	/* 7.	Return the result of the inference performed. */

	return res;
}



/******************************************************************************
*
*       Function        inferProdWithCondJoint
*       Parameters		prod (the product on which inference is performed), 
*						condVar (the variable on which conditioning is performed),
*						elimOrd (the elimination order according to which the 
*						variables of the product are processed),
*						problemSize (the size of the problem that needs to be
*						solved by the current call to the function),
*						solved (the size of the problem that has already been
*						solved).
*       Output          none
*       Description		Performs inference on 'prod' by conditioning						 
*						on 'condVar' and the variable that are processed jointly
*						with it, and performing inference seperately for each 
*						valid assignment to the conditioning variables
*						and then combining the results.
*       Comments		Returns the result of the inference. Used in the case 
*						where a few variables need to be conditioned on jointly.
*
******************************************************************************/
#if UNITE_VARS
Probability inferProdWithCondJoint(product *prod, variable *condVar,
								   elimOrderPed *elimOrd, float problemSize, 
								   float solved)
{
	int					numAssignments, *varValPtr, numCondVars;
	product				*brokenProd;
	variable			**condVarsArr, **varPtr, **varPtrEnd, **condVarsArrNew;
	Probability			*resArr, *resArrPtr, *resArrPtrEnd, res;
	double				maxLogVal = MIN_DBL;
	float				newProblemSize, currSolved;
	long				currElimNode;
	elimOrderNode		*currNode, *endNode;
	boolean				*isTransformedArr, *isTransformedArrPtr, isZeroLike, 
						areValsElim;
	validAssignments	validAssignmentsArr;
	assignment			*assignmentPtr, *assignmentPtrEnd;
	node				**uniteVar;
	changeUniteTList	*listPtr=NULL;


	/* 1.	Allocate memory for an array of pointers to all the variables that 
			need to be conditioned on, and update it with the pointers. */

	numCondVars = (condVar->originalNode)->numVarsUnite;
	myMalloc(condVarsArr, sizeof(variable*) * numCondVars);
	myMalloc(condVarsArrNew, sizeof(variable*) * numCondVars);

	for (varPtr = condVarsArr, varPtrEnd = varPtr + numCondVars, 
			uniteVar = (condVar->originalNode)->uniteVars; 
			varPtr < varPtrEnd; varPtr++, uniteVar++) {

		(*varPtr) = (*uniteVar)->varInProd;
	}

	/* 2.	Determine all the valid joint assignments to the variables that need 
			to be conditioned on. */

	findValidJointAssignments(condVarsArr, numCondVars, &validAssignmentsArr, 0);

	/* 3.	Allocate memory for an array that will hold the results of the 
			inference for each valid assignment to the variables that are
			conditioned on. */

	numAssignments = validAssignmentsArr.numAssignments;

	myMalloc(resArr, sizeof(Probability) * numAssignments);

	/* 4.	Allocate memory for an array that will hold the temporary values of
			variables in the elimination-order-array. */

	myMalloc(isTransformedArr, sizeof(boolean) * (elimOrd->numNodes - elimOrd->currElimNode));

	/* 5.	Calculate the size of the problem that will be solved by each call
			to the 'inferProd' function. */

	newProblemSize = problemSize / numAssignments;
	currSolved = solved;

	currElimNode = elimOrd->currElimNode;

	/* 6.	Perform inference on the Product for each valid assignment to the variables 
			that are conditioned on. */

	for (assignmentPtr = validAssignmentsArr.assignmentArr, 
			assignmentPtrEnd = assignmentPtr + numAssignments, 
			resArrPtr = resArr; assignmentPtr < assignmentPtrEnd; 
			assignmentPtr++, resArrPtr++) {
		
		/* 6.1.	Duplicate the Product. */

		brokenProd = copyProduct(prod, elimOrd, isTransformedArr, condVarsArr,
								 condVarsArrNew, numCondVars);

		/* 6.2.	Iterate over the variables that are conditioned on: */

		*resArrPtr = 0.0;

		for (varPtr = condVarsArrNew, varPtrEnd = varPtr + numCondVars,
				varValPtr = assignmentPtr->varVals;
				varPtr < varPtrEnd; varPtr++, varValPtr++) {

			/* 6.2.1.	Set the value of the variable according to the current
						assignment and change 'brokenProd' accordingly. */

			*resArrPtr += setBreakVal(brokenProd, *varPtr, *varValPtr);
		}

		/* 6.3.	Eliminate impossible variable values from the Product's functions. */

		*resArrPtr += eliminateImpossibleValues(brokenProd, &isZeroLike, 
												&areValsElim);

		/* 6.4.	If some variable values have been eliminated, update the unite status
				of nodes whose values have been eliminated, in case necessary. */

		if (areValsElim) {
			listPtr = updateUniteVarsPed(currPed);
		}

		/* 6.5.	If this assignment of values to the variables that are 
				conditioned on is possible: */

		if (!isZeroLike) {

			/* 6.5.1.	Perform inference on 'brokenProd'. */

			*resArrPtr += inferProd(brokenProd, elimOrd, newProblemSize, currSolved);

			/* 6.5.2.	Compare the current result to the maximum result up to now. */

			if (*resArrPtr) {
				if (*resArrPtr > maxLogVal) {
					maxLogVal = *resArrPtr;
				}
			}
		}
		else {
			*resArrPtr = MIN_DBL;

			/* 6.5.1.	If requested, print a progress report. */

			#if PRINT_PROGRESS_REPORT
				printf("Solved %f percent of the problem.\n", (currSolved+newProblemSize)*100);
			#endif
		}

		elimOrd->currElimNode = currElimNode;

		/* 6.6.	Restore the previous values of the nodes in the elimination-order array. */

		for (currNode = &(elimOrd->elimOrderArr)[elimOrd->currElimNode],
				endNode = &(elimOrd->elimOrderArr)[elimOrd->numNodes], 
				isTransformedArrPtr = isTransformedArr; 
				currNode < endNode; currNode++, isTransformedArrPtr++) {

			if (*isTransformedArrPtr == TRUE) {
				(currNode->originalNode)->varInProd = 
							((currNode->originalNode)->varInProd)->varInOldProd;
			}
		}

		/* 6.7.	Restore values of unite nodes, in case of nodes whose values changed after
				the value elimination. */

		if (areValsElim) {
			restoreUniteVals(listPtr);
		}

		/* 6.8.	Free the memory allocated for the duplicated product. */

		freeProduct(brokenProd);

		/* 6.9.	Update the size of the problem solved up to now. */

		currSolved += newProblemSize;
	}

	/* 7.	Combine the results of the inference on the different values of breakVar. */

	if (maxLogVal != MIN_DBL) {
		res = 0.0;

		for (resArrPtr = resArr, resArrPtrEnd = resArrPtr + numAssignments;
				resArrPtr < resArrPtrEnd; resArrPtr++) {
			
			if (*resArrPtr != MIN_DBL)
				res += exp(*resArrPtr - maxLogVal);
		}

		res = log(res) + maxLogVal;
	}
	else {
		res = MIN_DBL;
	}

	/* 8.	Free the memory allocated dynamically. */

	myFree(resArr);
	myFree(isTransformedArr);
	myFree(condVarsArr);
	myFree(condVarsArrNew);
	freeAssignmentMem(&validAssignmentsArr);

	/* 9.	Return the result of the inference performed. */

	return res;
}
#endif


/******************************************************************************
*
*       Function        setBreakVal
*       Parameters		brokenProd (the copied product), brokenVar (the variable
*						on which the product needs to be conditioned on), 
*						breakVarVal (the current value of 'brokenVar').
*       Output          none
*       Description		Sets the value of 'brokenVar' to 'breakVarVal' and 
*						adjusts 'brokenProd' accordingly (reduces the 
*						functions that brokenVar appears in).
*       Comments		If some functions are reduced to constants, returns
*						their joint probability.
*
******************************************************************************/
Probability setBreakVal(product *brokenProd, variable *brokenVar, int breakVarVal)
{
	function	**funcPtr, **funcPtrEnd;
	int			oldNumVals, oldNumFunctions;
	Probability currRes, res;

	/* 1.	Initialize res to zero. */

	res = 0.0;

	/* 2.	Update the number of values of the broken variable to 1 and save the
			original number of values. */

	oldNumVals = brokenVar->numValues;
	brokenVar->numValues = 1;

	/* 3.	Store the number of functions in which the original variable 
			participated in. */

	oldNumFunctions = brokenVar->currNumFunctions;

	/* 4.	Update each of the functions in which the brokenVar appears in. 
			For each function: */

	for (funcPtr = brokenVar->functions, funcPtrEnd = funcPtr + oldNumFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {

		/* 4.1.	Create a new, reduced function from the old function. If the function
				is reduced to a constant, add its log-probability to the total 
				log-probability and remove it from the product.*/

		if (reduceFunctionTableOnVar(*funcPtr, brokenVar, oldNumVals, breakVarVal, &currRes)) {
			res += currRes;
			removeFunctionFromProduct(brokenProd, (*funcPtr));
		}
	}

	/* 5.	Remove the broken variable from the product. */

	removeVariableFromProduct(brokenProd, brokenVar);

	/* 6.	Return the result of the inference. */

	return res;
}
