/* $Id: ProductValueElimination.c,v 1.2 2008/08/03 08:15:46 superlink Exp $ */
/*****************************************************************************/
/*  ProductValueElimination.c												 */
/*  Author: Ma'ayan Fishelson				                                 */
/*****************************************************************************/

#include "ProductValueElimination.h"


/******************************************************************************
*
*       Function        eliminateImpossibleValsProducts
*       Parameters		prodArr (the array of product on which value elimination 
*						needs to be performed), areValsElim (is updated 
*						to TRUE if some of the variables' values have been 
*						eliminated).
*       Output          none
*       Description		Eliminates impossible values of the variables of the
*						products from the products' functions.
*       Comments		If during the elimination, constant functions are 
*						created, their probability is calculated and added to 
*						total probability. 
*						Used for likelihood computations.
*
******************************************************************************/
void eliminateImpossibleValsProducts(product** prodArr, boolean *areValsElim)
{
	int				prodNum, numProds;
	Probability		res, currRes, *pedLnLike;
	boolean			*isZeroLikePed, areValsEliminatedPed;
	pedigree		*currPed;

	/* 1.	Initializations. */

	numProds = peds.numPeds;
	res = 0.0;
	*areValsElim = FALSE;

	/* 2.	Iterate over the products. For each product: */

	for (prodNum = 0, isZeroLikePed = totalRes.isZeroLikeArr, currPed = peds.head,
			pedLnLike = totalRes.pedsLnLikeArr; prodNum < numProds; prodNum++, 
			isZeroLikePed++, pedLnLike++, currPed++) {

		/* 2.1.	If the likelihood of this pedigree is zero, continue to the
				next one. */

		if (*isZeroLikePed) 
			continue;

		/* 2.2.	Eliminate impossible values from the product's functions. */

		currRes = eliminateImpossibleValues(prodArr[prodNum], isZeroLikePed, 
											&areValsEliminatedPed);

		if (areValsEliminatedPed) {
			*areValsElim = TRUE;
		}

		/* 2.3.	Add the log-probability of the constant functions created
				during the value-elimination to the total result, and to the
				likelihood of this pedigree. */

		res += currRes;
		(*pedLnLike) += currRes;
	}

	/* 3.	Add the ln-probability of the functions that became constant to the
			total ln-probability. */

	(totalRes.currRes)->lnLike += res;
}


/******************************************************************************
*
*       Function        eliminateImpossibleValuesProductsHaplo
*       Parameters		prodArr (the array of product on which value elimination 
*						needs to be performed), areValsElim (is updated 
*						to TRUE if some of the variables' values have been 
*						eliminated), argMaxArr (holds information on the values
*						of the variables that were eliminated for the tracing
*						of the assignment of values to the variables of the 
*						Bayesian networks which maximizes the probabiility of 
*						the data).
*       Output          none
*       Description		Eliminates impossible values of the variables of the
*						products from the products' functions.
*       Comments		If during the elimination, constant functions are 
*						created, their probability is calculated and returned. 
*						Used for haplotyping analysis.
*
******************************************************************************/
void eliminateImpossibleValsProductsHaplo(product** prodArr, boolean *areValsElim, 
										  argMaxData *argMaxArr)
{
	int				prodNum, numProds;
	Probability		res, currRes, *pedLnLike;
	boolean			*isZeroLikePed, areValsEliminatedPed;
	pedigree		*currPed;


	/* 1.	Initializations. */

	numProds = peds.numPeds;
	res = 0.0;
	*areValsElim = FALSE;

	/* 2.	Iterate over the products. For each product: */

	for (prodNum = 0, currPed = peds.head, isZeroLikePed = totalRes.isZeroLikeArr,
			pedLnLike = totalRes.pedsLnLikeArr; prodNum < numProds;
			prodNum++, currPed = currPed->nextPed, isZeroLikePed++, pedLnLike++) {

		/* 2.1.	Eliminate impossible values from the product's functions. */

		currRes = eliminateImpossibleValuesHaplo(prodArr[prodNum], isZeroLikePed, 
												 &areValsEliminatedPed, &argMaxArr[prodNum]);

		if (areValsEliminatedPed)
			*areValsElim = TRUE;

		/* 2.2.	Update the results data structure with the log-probability of 
				the constant functions created during the value-elimination */

		if (*isZeroLikePed) {
			(*pedLnLike) = MIN_DBL;
		}
		else {
			res += currRes;
			(*pedLnLike) += currRes;
		}
	}

	/* 3.	Update the total result with the log-probability of the functions that
			became constant. */

	(totalRes.currRes)->lnLike += res;
}


/******************************************************************************
*
*       Function        eliminateImpossibleValues
*       Parameters		prod (the product on whose variables value elimination
*						is being performed), isZeroLike (if all the values of 
*						some variable of the product have been eliminated, this
*						variable is updated to TRUE. Otherwise, it is updated to
*						FALSE), areValsElim (is updated to TRUE if some
*						variable values have been eliminated).
*       Output          none
*       Description		Eliminates impossible values of the variables of the
*						product, from the product's functions.
*       Comments		If during the elimination, constant functions are 
*						created, their probability is calculated and returned. 
*						Used for likelihood computations.
*
******************************************************************************/
Probability eliminateImpossibleValues(product *prod, boolean *isZeroLike, 
									  boolean *areValsElim)
{
	/* The total number of values eliminated & the number of values eliminated 
	   in one round.*/
	int			numValuesElim, numValuesElimRound;	

	/* The maximum number of possible values of a variable & the maximum 
	   number of variables in a function. */
	int			maxValuesVar, maxVarsInFunc;		
	int			varValue;

	variable	*var, *varEnd, **varPtr, **varPtrEnd;
	function	*func, *funcEnd, *newFunc;

	/* Arrays used to mark which variable values are possible, in general and in 
	   a specific function. */
	boolean		*varValueUsageFlagsGen, *varValueUsageFlagsFunc;  

	/* An array used to mark which functions need to be reduced, and a pointer to the
	   array. */
	boolean		*elimFuncs, *elimFuncsPlace;
	boolean		*nextPlace;

	/* An array which holds the number of possible values of each variable and a 
	   pointer to the array. */
	int			*numValsVarsNew, *numValArrPtr;	
	Probability res, currRes;

#if UNITE_VARS
	int			*certainVal, *certainValEnd; 
#endif

	
	*isZeroLike = FALSE;


	#if (PRINT_ELIMINATION_STATISTICS || (PRINT_VARIOUS_STATISTICS_LEVEL > 1))
		countTrimmedVars = 0;
	#endif


	res = 0.0;

	/* 1.	Allocate memory for an auxiliary array. */

	myMalloc(numValsVarsNew, sizeof(int) * prod->numVariables);

	/* 2.	Find the maximum number of possible values of a variable and 
			initialize the array that holds the number of possible values of each variable. */

	maxValuesVar = 0;
	for (var = prod->variables, varEnd = var + prod->numVariables, numValArrPtr = numValsVarsNew; 
			var < varEnd; var++, numValArrPtr++) {
		if (var->used) {
			if (var->numValues > maxValuesVar)
				maxValuesVar = var->numValues;
			*numValArrPtr = var->numValues;
		}
	}

	/* 3.	Find the maximum number of variables in a function. */

	maxVarsInFunc = 0;
	for (func = prod->functions, funcEnd = func + prod->numFunctions; func < funcEnd; func++) {
		if (func->used)
			if (func->numVariables > maxVarsInFunc)
				maxVarsInFunc = func->numVariables;
	}

	/* 4.	Allocate memory for the various data-structures used. */

	myMalloc(varValueUsageFlagsGen, sizeof(boolean) * prod->numVariables * maxValuesVar);
	myMalloc(varValueUsageFlagsFunc, sizeof(boolean) * maxVarsInFunc * maxValuesVar);
	myMalloc(elimFuncs, sizeof(boolean) * prod->numFunctions);

	/* 5.	Start with the value-elimination, the elimination is carried on for as long as
			something has been found to eliminate in the last iteration. */

	numValuesElim = 0;

	while (1) {
		
		/* 5.1.	Initialize the variable-usage flags to indicate that all the values 
				are possible.*/

		for (var = prod->variables, varEnd = var + prod->numVariables, 
			 nextPlace = varValueUsageFlagsGen; var < varEnd; var++, nextPlace += maxValuesVar) {
			if (var->used) {
				for (varValue = 0; varValue < var->numValues; varValue++)
					nextPlace[varValue] = TRUE;
			}
		}

		/* 5.2.	Initialize the function-shrinking flags to indicate that no function 
				needs to be shrunk.*/

		for (func = prod->functions, funcEnd = func + prod->numFunctions, elimFuncsPlace = elimFuncs; 
				func < funcEnd; func++, elimFuncsPlace++) {
			*elimFuncsPlace = FALSE;
		}

		/* 5.3.	Calculate an update on the usage-flags from all the functions. */

		numValuesElimRound = updateVariableUsageInFuncs(prod, varValueUsageFlagsGen, 
										    		    varValueUsageFlagsFunc, maxValuesVar, 
										    		    elimFuncs, numValsVarsNew);

		/* 5.4.	In case that nothing needs to be eliminated in this iteration, 
				we can stop the elimination process. */

		if (numValuesElimRound == 0)
			break;

		/* 5.5.	If values that can be eliminated have been found, update the total 
				number of values that have been eliminated. */

		numValuesElim += numValuesElimRound;

		/* 5.6.	Rebuild the functions that need to be reduced. For each function: */

		for (func = prod->functions, funcEnd = func + prod->numFunctions, elimFuncsPlace = elimFuncs;
				func < funcEnd;	func++, elimFuncsPlace++) {
		
			if (*elimFuncsPlace == TRUE) {
				
				/* 5.6.1.	Create the new function. */

				newFunc = reduceFunctionTable(prod, func, varValueUsageFlagsGen, maxValuesVar,
											  numValsVarsNew, &currRes);

				/* 5.6.2.	Update the product and the variables regarding the change in the 
							functions. */

				/* 5.6.2.1.	Remove the old function from the variables that appear in it & 
							from the product. */

				for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables; 
						varPtr < varPtrEnd; varPtr++) {

					removeFunctionFromVariable(*varPtr, func);
				}
				
				removeFunctionFromProduct(prod, func);

				/* 5.6.2.2.	If the new function isn't a constant, add it to the variables that
							appear in it.  Otherwise, add the function's logProbability to the
							total logProbability. */

				if (newFunc != NULL) {
					for (varPtr = newFunc->variables, varPtrEnd = varPtr + newFunc->numVariables;
							varPtr < varPtrEnd; varPtr++) {

						addFunctionToVariable(*varPtr, newFunc);
					}
				}
				else {	
					res += currRes;
				}
			}
		}

		/* 5.7. Iterate over the variables of the product. */

		for (var = prod->variables, varEnd = var + prod->numVariables, 
				numValArrPtr = numValsVarsNew; var < varEnd; var++, numValArrPtr++) {
			if (var->used) {

				/* 5.7.1.	If some values of the variable have been eliminated in this
							round: */

				if (var->numValues != *numValArrPtr) {

					/* 5.7.1.1.	Update the current number of values of the variable. */

					var->numValues = *numValArrPtr;

					/* 5.7.1.2.	Remove the variables that have only 1 value left, 
								from the product. 
								If all the values of some variable have been eliminated,
								update that the likelihood of the pedigree is zero, and return. */

					if (var->numValues == 1) {
						removeVariableFromProduct(prod, var);

						#if (PRINT_ELIMINATION_STATISTICS || (PRINT_VARIOUS_STATISTICS_LEVEL > 1))
							countTrimmedVars++;
						#endif
					}
					else if (var->numValues == 0) {
						*isZeroLike = TRUE;
						goto free_data_and_return;
					}
				#if	UNITE_VARS
					else if (var->originalNode->uniteT == fourAll) {
						for (certainVal = var->certainVal,
								certainValEnd = certainVal + var->originalNode->numValues;
								certainVal < certainValEnd; certainVal++) {

							if (*certainVal == INVALID_CURR) {
								*certainVal = INVALID_PRE;
							}
						}
					}
				#endif
				}
			}
		}
	}

	/* 6.	Free the memory allocated dynamically. */

	free_data_and_return:

	myFree(varValueUsageFlagsGen); 
	myFree(varValueUsageFlagsFunc);
	myFree(elimFuncs);
	myFree(numValsVarsNew);

	#if PRINT_ELIMINATION_STATISTICS
		printf("Eliminated %d values from the variables.\n", numValuesElim);
		printf("Trimmed %d variables.\n", countTrimmedVars);
		printf("Num variables in BN is %d.\n", prod->currNumVars);
	#endif

	#if (PRINT_VARIOUS_STATISTICS_LEVEL > 1)
		prod->numVarsTrim = countTrimmedVars;
	#endif

	*areValsElim = (numValuesElim > 0);

	if (res < MIN_DBL) {
		*isZeroLike = TRUE;
	}

	/* 7.	Return the logProbability of the functions that were reduced to constants. */

	return res;
}




/******************************************************************************
*
*       Function        eliminateImpossibleValuesHaplo
*       Parameters		prod (the product on whose variables value elimination
*						is being performed), isZeroLike (if all the values of 
*						some variable of the product have been eliminated, this
*						variable is update to TRUE. Otherwise, it is updated to
*						FALSE), areValsElim (is updated to TRUE if some
*						variable values have been eliminated), argMax (holds 
*						information on the values of the variables that were 
*						eliminated for the purpose of tracing the assignment of
*						values to the variables of the Bayesian networks which
*						maximizes the probabiility of the data).
*       Output          none
*       Description		Eliminates impossible values of the variables of the
*						product, from the product's functions.
*       Comments		If during the elimination, constant functions are 
*						created, their probability is calculated and returned. 
*						Used for haplotyping analysis.
*
******************************************************************************/
Probability eliminateImpossibleValuesHaplo(product *prod, boolean *isZeroLike, 
										   boolean *areValsElim, argMaxData *argMax)
{
	/* The total number of values eliminated & the number of values eliminated 
	   in one round.*/
	int			numValuesElim, numValuesElimRound;	

	/* The maximum number of possible values of a variable & the maximum 
	   number of variables in a function. */
	int			maxValuesVar, maxVarsInFunc;		
	int			varValue;

	variable	*var, *varEnd, **varPtr, **varPtrEnd;
	function	*func, *funcEnd, *newFunc;

	/* Arrays used to mark which variable values are possible, in general and in 
	   a specific function. */
	boolean		*varValueUsageFlagsGen, *varValueUsageFlagsFunc;  

	boolean		*varValueUsageFlagsGenTot, *nextPlaceTot, *possibleVal, *possibleValEnd;
	int			inc, val, varValueOld;

	/* An array used to mark which functions need to be reduced, and a pointer to the
	   array. */
	boolean		*elimFuncs, *elimFuncsPlace;
	boolean		*nextPlace;

	/* An array which holds the number of possible values of each variable and a 
	   pointer to the array. */
	int			*numValsVarsNew, *numValArrPtr;	
	Probability res, currRes;

	/* An array which holds the number of possible values of each variable originally 
	   (when the function was called) and a pointer to the array. */ 
	int			*numValsVarsOrig, *numValArrOrigPtr;

	/* An auxiliary argMaxData structure to store data on variables that are eliminated
	   to constants, and a pointer to the head of the list of argMaxData structures that 
	   store data on variables eliminated to constants. */
	argMaxData	constArgMax, headConstArgMax = NULL;

	/* An array of size 1 that will hold the value of the node eliminated to a constant. */
	int			*constArgMaxTable;

	/* An array to store data on nodes whose values were eliminated, and a pointer to the array. */
	argMaxElimValsVar	*elimNodesArr, *elimNodesArrPtr;

	/* Stores the number of nodes whose values were eliminted. */
	int					numElimNodes, *elimValPtr;
	
	*isZeroLike = FALSE;


	#if (PRINT_ELIMINATION_STATISTICS || (PRINT_VARIOUS_STATISTICS_LEVEL > 1))
		countTrimmedVars = 0;
	#endif


	res = 0.0;

	/* 1.	Allocate memory for auxiliary arrays. */

	myMalloc(numValsVarsNew, sizeof(int) * prod->numVariables);
	myMalloc(numValsVarsOrig, sizeof(int) * prod->numVariables);

	/* 2.	Find the maximum number of possible values of a variable and initialize 
			the array that holds the number of possible values of each variable. */

	maxValuesVar = 0;
	for (var = prod->variables, varEnd = var + prod->numVariables, numValArrPtr = numValsVarsNew,
			numValArrOrigPtr = numValsVarsOrig; var < varEnd; var++, numValArrPtr++, numValArrOrigPtr++) {
		if (var->used) {
			if (var->numValues > maxValuesVar) {
				maxValuesVar = var->numValues;
			}
			*numValArrPtr = var->numValues;
			*numValArrOrigPtr = var->numValues;
		}
	}

	/* 3.	Find the maximum number of variables in a function. */

	maxVarsInFunc = 0;
	for (func = prod->functions, funcEnd = func + prod->numFunctions; func < funcEnd; func++) {
		if (func->used)
			if (func->numVariables > maxVarsInFunc)
				maxVarsInFunc = func->numVariables;
	}

	/* 4.	Allocate memory for the various data-structures used. */

	myMalloc(varValueUsageFlagsGen, sizeof(boolean) * prod->numVariables * maxValuesVar);
	myMalloc(varValueUsageFlagsGenTot, sizeof(boolean) * prod->numVariables * maxValuesVar);
	myMalloc(varValueUsageFlagsFunc, sizeof(boolean) * maxVarsInFunc * maxValuesVar);
	myMalloc(elimFuncs, sizeof(boolean) * prod->numFunctions);

	/* 5.	Initialize the variable-usage flags to indicate that all the values 
			are possible.*/

	for (var = prod->variables, varEnd = var + prod->numVariables, 
			 nextPlaceTot = varValueUsageFlagsGenTot; var < varEnd; var++, 
			 nextPlaceTot += maxValuesVar) {
		if (var->used) {
			for (varValue = 0; varValue < var->numValues; varValue++)
					nextPlaceTot[varValue] = VALID_VAL;
		}
	}

	/* 6.	Start with the value-elimination, the elimination is carried on for as long as
			something has been found to eliminate in the last iteration. */

	numValuesElim = 0;

	while (1) {
		
		/* 6.1.	Initialize the variable-usage flags to indicate that all the values 
				are possible.*/

		for (var = prod->variables, varEnd = var + prod->numVariables, 
			 nextPlace = varValueUsageFlagsGen; var < varEnd; var++, nextPlace += maxValuesVar) {
			if (var->used) {
				for (varValue = 0; varValue < var->numValues; varValue++)
					nextPlace[varValue] = TRUE;
			}
		}

		/* 6.2.	Initialize the function-shrinking flags to indicate that no function 
				needs to be shrunk.*/

		for (func = prod->functions, funcEnd = func + prod->numFunctions, elimFuncsPlace = elimFuncs; 
				func < funcEnd; func++, elimFuncsPlace++) {
			*elimFuncsPlace = FALSE;
		}

		/* 6.3.	Calculate an update on the usage-flags from all the functions. */

		numValuesElimRound = updateVariableUsageInFuncs(prod, varValueUsageFlagsGen, 
										     		    varValueUsageFlagsFunc, maxValuesVar, 
											    	    elimFuncs, numValsVarsNew);

		/* 6.4.	In case that nothing needs to be eliminated in this iteration, 
				we can stop the elimination process. */

		if (numValuesElimRound == 0)
			break;

		/* 6.5.	If values that can be eliminated have been found, update the total 
				number of values that have been eliminated. */

		numValuesElim += numValuesElimRound;

		/* 6.6.	Rebuild the functions that need to be reduced. For each function: */

		for (func = prod->functions, funcEnd = func + prod->numFunctions, elimFuncsPlace = elimFuncs;
				func < funcEnd;	func++, elimFuncsPlace++) {
		
			if (*elimFuncsPlace == TRUE) {
				
				/* 6.6.1.	Create the new function. */

				newFunc = reduceFunctionTable(prod, func, varValueUsageFlagsGen, maxValuesVar,
											  numValsVarsNew, &currRes);

				/* 6.6.2.	Update the Product and the variables regarding the change in the 
							functions. */

				/* 6.6.2.1.	Remove the old function from the variables that appear in it & 
							from the product. */

				for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables; 
						varPtr < varPtrEnd; varPtr++) {

					removeFunctionFromVariable(*varPtr, func);
				}
				
				removeFunctionFromProduct(prod, func);

				/* 6.6.2.2.	If the new function isn't a constant, add it to the 
							variables that appear in it.  Otherwise, add the 
							function's logProbability to the total logProbability. */

				if (newFunc != NULL) {
					for (varPtr = newFunc->variables, varPtrEnd = varPtr + newFunc->numVariables;
							varPtr < varPtrEnd; varPtr++) {

						addFunctionToVariable(*varPtr, newFunc);
					}
				}
				else {	
					res += currRes;
				}
			}
		}

		/* 6.7. Iterate over the variables of the product. */

		for (var = prod->variables, varEnd = var + prod->numVariables, 
				numValArrPtr = numValsVarsNew, nextPlace = varValueUsageFlagsGen,
				nextPlaceTot = varValueUsageFlagsGenTot; var < varEnd; var++, 
				numValArrPtr++, nextPlace += maxValuesVar, nextPlaceTot += maxValuesVar) {

			/* 6.7.1.	If some of the values of the variable have been eliminated in this
						round: */

			if ((var->used) && (var->numValues != *numValArrPtr)) {

				/* 6.7.1.1.	Iterate over the values of the variable and update the array
							which holds all the possible values for this variable. */

				for (varValue = 0, possibleVal = nextPlaceTot, inc = 0, val = 0; 
						varValue < var->numValues; varValue++) {

					/* 6.7.1.1.1.	If this value has been found to be invalid for the variable,
									find the corresponding position in the original array of
									possible values for this variable and mark it as invalid. */

					if (nextPlace[varValue] == FALSE) {
						varValueOld = varValue + inc;

						for (possibleValEnd = possibleVal + (var->originalNode)->numValues; 
								possibleVal < possibleValEnd; possibleVal++, val++) {
							if (*possibleVal == INVALID_VAL) {
								varValueOld++;
								inc++;
							}
							else if (val == varValueOld) {
								(*possibleVal) = INVALID_VAL;
								possibleVal++;
								val++;
								break;
							}
						}
					}
				}

				/* 6.7.1.2.	Update the current number of values of the variable. */

				var->numValues = *numValArrPtr;

				/* 6.7.1.3.	Remove the variables that have only 1 value left from the
							Product. If all values of some variable have been eliminated, 
							update that the likelihood of the pedigree is zero, and return. */

				if (var->numValues == 1) {
					removeVariableFromProduct(prod, var);

					#if (PRINT_ELIMINATION_STATISTICS || (PRINT_VARIOUS_STATISTICS_LEVEL > 1))
						countTrimmedVars++;
					#endif

					/* 6.7.1.3.1.	Create a new argMaxData structure for the variable that was eliminated
									to a constant and connect it to the beginning of the list. */

					myMalloc(constArgMaxTable, sizeof(int));

					*constArgMaxTable = getConstVal(nextPlaceTot, maxValuesVar);
					
					#ifdef WIN32
						assert(*constArgMaxTable >= 0);
					#endif

					constArgMax = createAMDataElim(var->originalNode, headConstArgMax, 0, NULL, 
													1, constArgMaxTable);

					headConstArgMax = constArgMax;
				}
				else if (var->numValues == 0) {
					*isZeroLike = TRUE;
					*argMax = NULL;
					freeAMList(headConstArgMax);
					goto free_data_and_return;
				}
			}
		}
	}

	/* 7.	If the values of some nodes were eliminated, create an argMaxData structure of type
			'dummyNode' to store the data on all these nodes and the eliminated values. */

	if (numValuesElim > 0) {

		/* 7.1.	Count the number of nodes whose values were eliminated. */

		numElimNodes = 0;
		
		for (var = prod->variables, varEnd = var + prod->numVariables, numValArrOrigPtr = numValsVarsOrig;
				var	< varEnd; var++, numValArrOrigPtr++) {	

			if (var->used && (var->numValues != *numValArrOrigPtr)) {
				numElimNodes++;
			}
		}

		/* 7.2.	Allocate memory for an array that would store the data on all the eliminated values. */

		myMalloc(elimNodesArr, sizeof(argMaxElimValsVar) * numElimNodes);

		/* 7.3.	Update the array with the data. */

		for (var = prod->variables, varEnd = var + prod->numVariables, 
				numValArrOrigPtr = numValsVarsOrig, elimNodesArrPtr = elimNodesArr, 
				nextPlaceTot = varValueUsageFlagsGenTot; var < varEnd; var++, 
				numValArrOrigPtr++, nextPlaceTot += maxValuesVar) {	

			if (var->used && (var->numValues != *numValArrOrigPtr)) {
				
				elimNodesArrPtr->origNode = var->originalNode;
				elimNodesArrPtr->numElimVals = *numValArrOrigPtr - var->numValues;

				myMalloc(elimNodesArrPtr->elimVals, sizeof(int) * elimNodesArrPtr->numElimVals);

				for (varValue = 0, elimValPtr = elimNodesArrPtr->elimVals; 
						varValue < *numValArrOrigPtr; varValue++) {

					if (nextPlaceTot[varValue] == INVALID_VAL) {
						*elimValPtr = varValue;
						elimValPtr++;
					}
				}

				elimNodesArrPtr++;
			}
		}

		/* 7.4.	Create an argMaxData of type 'dummyNode' and add it at the beginning of the list
				of argMaxData structures of nodes that turned into constants. */

		*argMax = createAMDataOther(NULL, headConstArgMax, numElimNodes, elimNodesArr, 
									dummyNode);
	}
	else {
		/* 7.1.	No values were eliminated, update argMax to NULL. */

		*argMax = NULL;
	}


	/* 8.	Free the memory allocated. */

	free_data_and_return:

	myFree(varValueUsageFlagsGen); 
	myFree(varValueUsageFlagsGenTot);
	myFree(varValueUsageFlagsFunc);
	myFree(elimFuncs);
	myFree(numValsVarsNew);
	myFree(numValsVarsOrig);

	#if PRINT_ELIMINATION_STATISTICS
		printf("Eliminated %d values from the variables.\n", numValuesElim);
		printf("Trimmed %d variables.\n", countTrimmedVars);
		printf("Num variables in BN is %d.\n", prod->currNumVars);
	#endif

	#if (PRINT_VARIOUS_STATISTICS_LEVEL > 1)
		prod->numVarsTrim = countTrimmedVars;
	#endif

	*areValsElim = (numValuesElim > 0);

	/* 9.	Return the logProbability of the functions that were reduced to constants. */

	return res;
}



/******************************************************************************
*
*       Function        updateVariableUsageInFuncs
*       Parameters		prod (the product on which value elimination is performed), 
*						varValueUsageFlagsGen (an array of flags used to mark the
*						values of each variable which are impossible), 
*						varValueUsageFlagsFunc (an array of flags which is used 
*						to mark the values of each variable of the function that
*						are used), maxValuesVar (the maximum number of values of
*						a variable), elimFuncs (an array of functions that need
*						to be reduced), numValsVarsNew (the number of values of 
*						each variable after the value elimination).
*       Output          none
*       Description		Searches for variable values that are impossible according
*						to some function and marks those values.
*       Comments		none
*
******************************************************************************/
int updateVariableUsageInFuncs(product *prod, boolean *varValueUsageFlagsGen, 
							   boolean *varValueUsageFlagsFunc, int maxValuesVar, 
							   boolean *elimFuncs, int *numValsVarsNew)
{
	function	*func, *funcEnd, **funcPtr, **funcPtrEnd;
	variable	**varPtr, **varPtrEnd, *var, **advanceVarPtr, **advanceVarPtrEnd, 
				*advanceVar;
	int			totalVarValues, varValue, numValsElim; 
	boolean		*nextPlace, *currPlace;
	Probability *probTablePtr;

#if UNITE_VARS
	variable	*uniteVar;
	int			*certainVal, *certainValEnd, varValueOld; 
	boolean		*currPlaceUnite;
#endif

	numValsElim = 0;

	/* 1.	Iterate over the Product's functions and check if there are any impossible
			values. */

	for (func = prod->functions, funcEnd = func + prod->numFunctions; func < funcEnd; func++) {
		
		if (func->used) {
			varPtrEnd = func->variables + func->numVariables;

			/* 1.1.	Initialize the variable-usage flags to indicate that all the 
					variable values are impossible.*/

			totalVarValues = 0;

			for (varPtr = func->variables, nextPlace = varValueUsageFlagsFunc; 
					varPtr < varPtrEnd; varPtr++, nextPlace += maxValuesVar) {

				var = *varPtr;

				totalVarValues += var->numValues;

				for (varValue = 0; varValue < var->numValues; varValue++)
					nextPlace[varValue] = FALSE;
			}

			/* 1.2.	Initialize the values of the function's variables to correspond to the 
					first entrance in the function's probability table. */

			initFunctionVars(func, varPtr, varPtrEnd);
			probTablePtr = func->probabList;

			/* 1.3.	Iterate over the function's probability table and set the value-usage 
					flags of the variables to TRUE for non-zero entries. */

			while (1) {
				start_loop_uvuif:

				if (*probTablePtr != 0.0) {
					for (varPtr = func->variables, nextPlace = varValueUsageFlagsFunc; 
							varPtr < varPtrEnd; varPtr++, nextPlace += maxValuesVar) { 

						currPlace = nextPlace + (*varPtr)->currValue;

						if (*currPlace == FALSE) {
							*currPlace = TRUE;
							totalVarValues--;

							/* 1.3.1. If we find that all of the values of this 
									  function's variables are possible according 
									  to it, we preceed to the next function. */

							if (totalVarValues == 0)
								goto preceedToNexFunction;
						}
					}
				}
				probTablePtr++;

				advanceFunctionVars(func, advanceVarPtr, advanceVarPtrEnd, advanceVar, start_loop_uvuif);
			}


			/* 1.4.	Check which variable values are impossible and update the product 
					general-usage-flags accordingly. */

			for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables, 
					nextPlace = varValueUsageFlagsFunc; varPtr < varPtrEnd; 
					varPtr++, nextPlace += maxValuesVar) {

				var = *varPtr;

				currPlace = &varValueUsageFlagsGen[(var - prod->variables) * maxValuesVar];

				for (varValue = 0; varValue < var->numValues; varValue++, currPlace++) {
					
					if (*currPlace) {
						if (!nextPlace[varValue]) {

							/* 1.4.1.	Update the indices of the certain values according
										to the elimination of the impossible variable values. */

							#if UNITE_VARS
								if (var->originalNode->uniteT == oneAll) {
									if (var->certainVal[0] > varValue) {
										var->certainVal[0]--;
									}
									else if (var->certainVal[0] == varValue) {
										var->certainVal[0] = -1;
									}
								} 
								else if (var->originalNode->uniteT == fourAll) {
									varValueOld = varValue;

									for (certainVal = var->certainVal, 
											certainValEnd = certainVal + var->originalNode->numValues;
											certainVal < certainValEnd; certainVal++) {

										if (*certainVal == INVALID_PRE) {
											varValueOld++;
										}
										else if (*certainVal == varValueOld) {
											(*certainVal) = INVALID_CURR;
											break;
										}
									}
								}
								else if (var->originalNode->uniteT == twoAll) {

									/* 1.4.1.	If this variable needs to be eliminated along with another one,	
												update the possible value left for the other variable as well. */

									uniteVar = ((var->originalNode)->uniteVars[0])->varInProd;

									if (var == uniteVar) {
										uniteVar = ((var->originalNode)->uniteVars[1])->varInProd;
									}

									currPlaceUnite = &varValueUsageFlagsGen[(uniteVar - prod->variables) * maxValuesVar];
									currPlaceUnite += (1-varValue);

									*currPlaceUnite = FALSE;
									numValsElim++;
									numValsVarsNew[uniteVar - prod->variables]--;

									/* 1.4.2.	Update all the functions that this variable appears in, 
												that they need to be reduced. */

									for (funcPtr = uniteVar->functions, funcPtrEnd = funcPtr + uniteVar->currNumFunctions; 
											funcPtr < funcPtrEnd; funcPtr++) {
										
										elimFuncs[(*funcPtr) - prod->functions] = TRUE;

									}
								}
							#endif

							/* 1.4.1.	Some value of var has just been found to be impossible. */

							*currPlace = FALSE;
							numValsElim++;
							numValsVarsNew[var - prod->variables]--;

							/* 1.4.2.	Update all the functions that this variable appears in, 
										that they need to be reduced. */

							for (funcPtr = var->functions, funcPtrEnd = funcPtr + var->currNumFunctions; 
									funcPtr < funcPtrEnd; funcPtr++) {
								
								elimFuncs[(*funcPtr) - prod->functions] = TRUE;
							}
						}
					}
				}
			}
		}
		preceedToNexFunction:;
	}

	/* 2.	Return the number of values were eliminated in this round. */

	return numValsElim;
}



/******************************************************************************
*
*       Function        getConstVal
*       Parameters		arr (an array of boolean type), arrSize (the array size).
*       Output          none
*       Description		Searches for an entry which is TRUE and returns its index
*						if found; otherwise, returns -1.
*       Comments		none
*
******************************************************************************/
int getConstVal(boolean *arr, int arrSize)
{
	int		entry;

	for (entry = 0; entry < arrSize; entry++, arr++) {
		if (*arr == TRUE) 
			return entry;
	}

	return -1;
}


/******************************************************************************
*
*       Function        updateUniteVarsPeds
*       Parameters		updateUniteVars (the function which is used for updating
*						the uniteT field of the nodes of the Bayesian in case
*						of variables that have turned into constants during the
*						value elimination).
*       Output          none
*       Description		Updates the 'uniteT' field of the nodes of the Bayesian
*						networks in case of variables that have turned into
*						constants during the value elimination stage.
*       Comments		none
*
******************************************************************************/
#if	UNITE_VARS
changeUniteTList* updateUniteVarsPeds(updateUniteVarsPedFunc updateUniteVars)
{
	pedigree			*currPed;
	changeUniteTList	*listPtr;

	/* 1.	Allocate memory for a struct that would store the information on all
			the nodes whose unite status changed due to the value elimination 
			which turned some nodes into constants. */

	myMalloc(listPtr, sizeof(changeUniteTList));
	myMalloc(listPtr->head, sizeof(changedUniteTNode));

	/* 2.	Initialize the struct. */

	listPtr->next = listPtr->head;
	listPtr->numNodes = 0;

	/* 3.	Loop over all the pedigrees. */

	for (currPed = peds.head; currPed != NULL; currPed = currPed->nextPed) {

		/* 3.1.	For each pedigree, call a function which updates the unite status
				of the nodes of the Bayesian networks in case of variables that have 
				turned into constants during the value elimination stage. */

		updateUniteVars(currPed, listPtr);
	}

	myFree(listPtr->next);

	/* 4.	Return a pointer to the struct. */

	return listPtr;
}
#endif

/******************************************************************************
*
*       Function        updateUniteVarsPed
*       Parameters		currPed (the pedigree which needs to be updated regarding
*						the unite status of the nodes of the Bayesian network
*						which represent it).
*       Output          none
*       Description		Updates the uniteT field of the nodes of the Bayesian
*						networks in case of variables that have turned into
*						constants during the value elimination stage.
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
changeUniteTList* updateUniteVarsPed(pedigree *currPed)
{
	changeUniteTList	*listPtr;
	

	/* 1.	Allocate memory for a struct that would store the information on all
			the nodes whose unite status changed due to the value elimination 
			which turned some nodes into constants. */

	myMalloc(listPtr, sizeof(changeUniteTList));
	myMalloc(listPtr->head, sizeof(changedUniteTNode));

	/* 2.	Initialize the struct. */

	listPtr->next = listPtr->head;
	listPtr->numNodes = 0;

	/* 3.	Call a function which updates the unite status of the nodes of the 
			Bayesian network . */

	if (loci.sexLink) {
		updateUniteVarsPedSexLinked(currPed, listPtr);
	}
	else {
		updateUniteVarsPedAutosomal(currPed, listPtr);
	}

	myFree(listPtr->next);

	/* 4.	Return a pointer to the struct. */

	return listPtr;
}
#endif



/******************************************************************************
*
*       Function        updateUniteVarsPedAutosomal
*       Parameters		currPed (the data-structure that holds the information
*						on the input pedigree), listPtr (will be updated with 
*						the information on those nodes whose unite status has
*						changed).
*       Output          none
*       Description		Updates the 'uniteT' field of the nodes of the Bayesian
*						network that represents this pedigree, in case of 
*						variables that have turned into constants during the
*						value elimination stage. 
*       Comments		Used in case of an autosomal analysis.
*
******************************************************************************/
#if UNITE_VARS
void updateUniteVarsPedAutosomal(pedigree *currPed, changeUniteTList *listPtr)
{
	person				*currPerson;
	int					personInd, locusInd, numLoci;
	node				**uniteVarPtr, **uniteVarPtrEnd, **paLoc, **maLoc;
	uniteType			uniteT;
	changedUniteTNode	*currNode;

	numLoci = loci.numLoci;

	currNode = listPtr->next;

	/* 1.	Iterate over the people in the pedigree. */

	for (personInd = 0; personInd < currPed->numPeople; personInd++) {
		currPerson = &(currPed->people[personInd]);

		paLoc = currPerson->locusVar[0];
		maLoc = currPerson->locusVar[1];

		/* 1.1.	For each person in the pedigree, loop over all loci. */

		for (locusInd = 0; locusInd < numLoci; locusInd++) {

			/* 1.1.1.	If this node needs to be united with other nodes, and 
						one of the nodes has been shrunk into a constant,
						free the memory allocated to the array of nodes and update
						that it won't be united. */

			if ((isUniteNode(*paLoc)) && (isNotUsedVar(*paLoc) || isNotUsedVar(*maLoc))){
				uniteT = (*paLoc)->uniteT;

				for (uniteVarPtr = (*paLoc)->uniteVars, 
						uniteVarPtrEnd = uniteVarPtr + (*paLoc)->numVarsUnite;
						uniteVarPtr < uniteVarPtrEnd; uniteVarPtr++) {

					if (isUsedVar(*uniteVarPtr)) {
						myFree((*uniteVarPtr)->varInProd->certainVal);
					}

					(*uniteVarPtr)->uniteT = noUnite;

					currNode->Node = *uniteVarPtr;
					currNode->uniteT = uniteT;
					myMalloc(currNode->next, sizeof(changedUniteTNode));
					currNode = currNode->next;
					listPtr->numNodes++;
				}
			}

			paLoc++;
			maLoc++;
		}
	}

	listPtr->next = currNode;
}
#endif

/******************************************************************************
*
*       Function        updateUniteVarsPedSexLinked
*       Parameters		currPed (the data-structure that holds the information
*						on the input pedigree), listPtr (will be updated with 
*						the information on those nodes whose unite status has
*						changed). 
*       Output          none
*       Description		Updates the 'uniteT' field of the nodes of the Bayesian
*						network that represents this pedigree in case of 
*						variables that have turned into constants during the
*						value elimination stage. 
*       Comments		Used in case of a sex-linked analysis.
*
******************************************************************************/
#if UNITE_VARS
void updateUniteVarsPedSexLinked(pedigree *currPed, changeUniteTList *listPtr)
{
	person				*currPerson;
	int					personInd, locusInd, numLoci;
	node				**uniteVarPtr, **uniteVarPtrEnd, **paLoc, **maLoc;
	uniteType			uniteT;
	changedUniteTNode	*currNode;

	numLoci = loci.numLoci;
	currNode = listPtr->next;

	/* 1.	Iterate over the people in the pedigree. */

	for (personInd = 0; personInd < currPed->numPeople; personInd++) {
		currPerson = &(currPed->people[personInd]);

		if (currPerson->personSex == Female) {
			paLoc = currPerson->locusVar[0];
			maLoc = currPerson->locusVar[1];

			/* 1.1.	For each person in the pedigree, loop over all loci. */

			for (locusInd = 0; locusInd < numLoci; locusInd++) {

				/* 1.1.1.	If this node needs to be united with other nodes, and 
							one of the nodes has been shrunk into a constant,
							free the memory allocated to the array of nodes and update
							that it won't be united. */

				if ((isUniteNode(*paLoc)) && (isNotUsedVar(*paLoc) || isNotUsedVar(*maLoc))){

					uniteT = (*paLoc)->uniteT;

					for (uniteVarPtr = (*paLoc)->uniteVars, 
							uniteVarPtrEnd = uniteVarPtr + (*paLoc)->numVarsUnite;
							uniteVarPtr < uniteVarPtrEnd; uniteVarPtr++) {

						if (isUsedVar(*uniteVarPtr)) {
							myFree((*uniteVarPtr)->varInProd->certainVal);
						}

						(*uniteVarPtr)->uniteT = noUnite;

						currNode->Node = *uniteVarPtr;
						currNode->uniteT = uniteT;
						myMalloc(currNode->next, sizeof(changedUniteTNode));
						currNode = currNode->next;
						listPtr->numNodes++;
					}
				}

				paLoc++;
				maLoc++;
			}
		}
	}

	listPtr->next = currNode;
}
#endif


/******************************************************************************
*
*       Function        restoreUniteVals
*       Parameters		listPtr (holds information on all the nodes whose unite
*						status has changed and needs to be restored). 
*       Output          none
*       Description		Restores the unite status of all the nodes in 'listPtr'. 
*       Comments		none
*
******************************************************************************/
#if UNITE_VARS
void restoreUniteVals(changeUniteTList *listPtr)
{
	changedUniteTNode	*currNode, *nextNode;
	int					numNodes;

	currNode = listPtr->head;
	numNodes = 0;

	/* 1.	As long as there are still more nodes in the list, restore their unite 
			status to the original status and free the nodes of the list. */

	while (numNodes < listPtr->numNodes) {
		nextNode = currNode->next;
		(currNode->Node)->uniteT = currNode->uniteT;
		myFree(currNode);

		currNode = nextNode;
		numNodes++;
	}

	/* 2.	Free the struct. */

	myFree(listPtr);
}
#endif
