/* $Id: SpecialMult.c,v 1.2 2008/08/03 08:15:46 superlink Exp $ */
/*********************************************************************************/
/*  SpecialMult.c												     			 */
/*  Author: Ma'ayan Fishelson	 												 */
/*********************************************************************************/

#include "SpecialMult.h"



/******************************************************************************
*
*       Function        initVars
*       Parameters		none
*       Output          none
*       Description		Initialize an array of size 32 with powers of 2 
*						(starting at 2^0 up to 2^32).
*       Comments		none
*
******************************************************************************/
void initVars()
{
	int i;

    power2_sl[0] = 1;

    for (i = 1; i < 32; i++) 
		power2_sl[i] = 2 * power2_sl[i-1];
}




/******************************************************************************
*
*       Function        eliminateSelectorsLocusI
*       Parameters		prod (pointer to the product on which the inference
*						is being performed), elimOrd (the elimination-order-array
*						that holds pointers to the variables in the order in 
*						which they should be eliminated. When this function is
*						called, all the selector variables of the current locus
*						appear one after another in the array at the next spot).
*       Output          none
*       Description		Eliminates all the selector variables of the current 
*						locus (locus #i) from the Product. The selector variables
*						are eliminated in groups, instead of one by one.
*       Comments		none
*
******************************************************************************/
Probability eliminateSelectorsLocusI(product *prod, elimOrderPed *elimOrd, 
									 boolean *isZeroLike)
{
	variable		**selectorArr, **selectorPtr, **selectorPtrEnd, *selector,
					**varArr, **mapCurrToPrevArr, *prevSelector, **varPtr, **varPtrEnd;
	int				numSelectors, locusIndMap, prevLocusIndMap;
	Probability		res, tempRes, *probabList;
	function		*newFunc, *func;


	/* 1.	Initialize the probability to zero. */

	res = 0.0;
	*isZeroLike = FALSE;

	/* 2.	Update 'selectorArr' to point to all the selector variables
			of the current locus. */

	selectorArr = getSelectorVarsLocusI(elimOrd, &numSelectors);

	/* 3.	Initialize 'locusIndMap' with the index on the map of the current locus, 
			and 'prevLocusIndMap' with the index on the map of the previous locus. */

	locusIndMap = (selectorArr[0]->originalNode)->lNum;
	prevLocusIndMap = locusIndMap - 1;

	/* 4.	Allocate memory for an array that maps each selector variable of the 
			current locus to the relevant selector variable of the previous locus. */

	myMalloc(mapCurrToPrevArr, sizeof(variable*) * prod->numVariables);

	/* 5.	Iterate over the selector variables of the current locus. */

	for (selectorPtr = selectorArr, selectorPtrEnd = selectorPtr + numSelectors;
			selectorPtr < selectorPtrEnd; selectorPtr++) {

		/* 5.1.	If the relevant selector variable of the previous locus doesn't
				exist, eliminate this selector variable alone. 
				If it does exist: 
				-	Remove the transmission probability function, that is defined 
				on the selector variable of this locus and that of the previous locus.
				-   if this selector variable appears in more than one function, 
				multiply all these functions to create one function. */

		if (checkPrevSelector(*selectorPtr, prevLocusIndMap, &prevSelector) == FALSE) {
			if (eliminateVar(prod,*selectorPtr, &tempRes, isZeroLike)) {
				res += tempRes;
			}
			else if (*isZeroLike) {
				res = MIN_DBL;
				myFree(mapCurrToPrevArr);
				return res;
			}
		}
		else {
			mapCurrToPrevArr[*selectorPtr - prod->variables] = prevSelector;

			removeTransFunc(*selectorPtr, prod, prevLocusIndMap);

			if ((*selectorPtr)->currNumFunctions > 1) {
				createOneFunc(*selectorPtr, prod);
			}
		}
	}

	/* 6.	Iterate over the selector variables of this locus. */

	for (selectorPtr = selectorArr, selectorPtrEnd = selectorPtr + numSelectors;
			selectorPtr < selectorPtrEnd; selectorPtr++) {

		selector = *selectorPtr;

		/* 6.1.	If the selector variable hasn't been eliminated already: */
		
		if (selector->used) {

			/* 6.1.1.	Initialize a pointer to the function it appears in. */

			func = selector->functions[0];

			/* 6.1.2.	Allocate memory for the new probability function. */

			myMalloc(probabList, sizeof(Probability) * func->sizeOfTable);

			/* 6.1.3.	Create the new function (according to either Yate's algorithm
						or the FFT algorithm). */

		#if !USE_FFT

			Yate(probabList, func->probabList, recombData.maleThetas[prevLocusIndMap][0],
				 func->sizeOfTable);
		#else

			calcFFT(probabList, func->probabList, recombData.maleThetas[prevLocusIndMap][0],
					func->sizeOfTable);
		#endif	

			/* 6.1.4.	Allocate memory for the variable array of the new function. */

			myMalloc(varArr, sizeof(variable*) * func->numVariables); /* check */

			/* 6.1.5.	Initialize the variable array of the new function, with the
						selector variables of the previous locus. */

			updateFuncsVariables(varArr, func->variables, func->numVariables, 
								 prod, mapCurrToPrevArr);

			/* 6.1.6.	Insert the new function into the product. */

			newFunc = insertFunctionToProduct(prod, func->sizeOfTable, 
											  probabList, func->logMultFactor, TRUE, 
											  func->numVariables, varArr, 
											  FALSE, func->rescaleTimer-1);

			/* 6.1.7.	If the function needs to be rescaled, rescale it. */

			if (newFunc->rescaleTimer <= 0) {
				scaleFunctionProbList(newFunc);
				newFunc->rescaleTimer = MAX_RESCALE_UNITS;
			}

			/* 6.1.8.	Add the new function to all the selector variables that 
						appear in it. */

			for (varPtr = newFunc->variables, varPtrEnd = varPtr + newFunc->numVariables; 
					varPtr < varPtrEnd; varPtr++) {
				
				addFunctionToVariable(*varPtr, newFunc);
			}

			/* 6.1.9.	Remove the old function from the selector variables that
						appear in it, and remove them from the product. */

			for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
					varPtr < varPtrEnd; varPtr++) {

				removeFunctionFromVariable(*varPtr, func);
				removeVariableFromProduct(prod, (*varPtr));
			}

			/* 6.1.10.	Remove the old function from the Product. */

			removeFunctionFromProduct(prod, func);
		}
	}

	/* 7.	Free the memory allocated dynamically. */

	myFree(mapCurrToPrevArr);
	myFree(selectorArr);

	/* 8.	Return the likelihood of the constant functions that were created. */

	return res;
}



/******************************************************************************
*
*       Function        getSelectorVarsLocusI
*       Parameters		elimOrd (the elimination-order array), 
*						numSelectors (is updated to hold the number of selector
*						variables of the current locus).
*       Output          none
*       Description		Returns an array of all the selector variables of the
*						current locus, and updates 'elimOrd->currElimNode' to 
*						point to the next node to be eliminated after all these
*						selector variables.
*       Comments		none
*
******************************************************************************/
variable** getSelectorVarsLocusI(elimOrderPed *elimOrd, int *numSelectors)
{
	elimOrderNode	*nextElimNode;
	variable		*nextVar, **selectorArr, **selectorArrPtr;
	int				numSelecs = 0, locusNum = -1;
	long			currElimNode;


	/* 1.	Save the first index in the elimination order array of a selector
			variable of this locus. */

	currElimNode = elimOrd->currElimNode;

	nextElimNode = &((elimOrd->elimOrderArr)[currElimNode]);
	
	/* 2.	While there are still selector nodes of this locus. */

	while ((currElimNode < elimOrd->numNodes) && 
				((nextElimNode->originalNode)->nodeType == SELECTOR)) {

		nextVar = (nextElimNode->originalNode)->varInProd;

		/* 2.1.	While the next var that is supposed to be eliminated has been 
				trimmed, go to the next var in the elimination order. */

		while ((currElimNode < elimOrd->numNodes) && (nextVar->used == FALSE)) {
			currElimNode++;
			nextElimNode++;
			nextVar = (nextElimNode->originalNode)->varInProd;
		}
		
		/* 2.2.	If this is a selector node of this locus, update the counter. 
				Otherwise, break. */

		if ((nextElimNode->originalNode)->nodeType == SELECTOR) {
			if (locusNum != -1) {
				if ((nextElimNode->originalNode)->lNum != locusNum) {
					break;
				}
			}
			else {
				locusNum = (nextElimNode->originalNode)->lNum;
			}
			numSelecs++;
		}

		/* 2.3.	Advance the pointer and the counter to the next node that needs 
			    to be eliminated. */

		currElimNode++;
		nextElimNode++;
	}

	/* 3.	Allocate memory for an array that would hold pointers to all the selector 
			variables at the current locus. */

	myMalloc(selectorArr, sizeof(variable*) * numSelecs);

	selectorArrPtr = selectorArr;

	/* 4.	Update 'nextVar' to point to the next variable to be eliminated. */

	currElimNode = elimOrd->currElimNode;

	nextElimNode = &((elimOrd->elimOrderArr)[currElimNode]);
	
	/* 5.	While there are still selector nodes of this locus. */

	while ((currElimNode < elimOrd->numNodes) && 
				((nextElimNode->originalNode)->nodeType == SELECTOR)) {

		nextVar = (nextElimNode->originalNode)->varInProd;

		/* 5.1.	While the next var that is supposed to be eliminated has been 
				trimmed, go to the next var in the elimination order. */

		while ((currElimNode < elimOrd->numNodes) && (nextVar->used == FALSE)) {
			currElimNode++;
			nextElimNode++;
			nextVar = (nextElimNode->originalNode)->varInProd;
		}
		
		/* 5.2.	If this is a selector node, update the selector array, 
				and advance the pointer and the counter to the next node 
				that needs to be eliminated. */

		if ((nextElimNode->originalNode)->nodeType == SELECTOR) {
			if ((nextElimNode->originalNode)->lNum == locusNum) {
				*(selectorArrPtr++) = nextVar;
				currElimNode++;
				nextElimNode++;
			}
			else
				break;
		}
	}

	/* 6.	Update the "next-node-to-be-eliminated" field. */

	elimOrd->currElimNode = currElimNode;

	/* 7.	Update the number of selector variables at the current locus. */

	*numSelectors = numSelecs;

	/* 8.	Return a pointer to the array of selector variables of the current locus. */

	return selectorArr;
}



/******************************************************************************
*
*       Function        checkPrevSelector
*       Parameters		selectorPtr (pointer to the selector variable), 
*						prevLocusNum (the index of the previous locus to that 
*						of which this selector belongs to), prevSelector (will
*						point to the relevant selector variable of the previous 
*						locus, if this exists).
*       Output          none
*       Description		The function checks if the relevant selector of the 
*						previous locus hasn't been eliminated - if it hasn't, it
*						returns TRUE and updates 'prevSelector' to point to it.
*						Otherwise, it returns FALSE.
*       Comments		none
*
******************************************************************************/
boolean checkPrevSelector(variable *selectorPtr, int prevLocusNum, 
						  variable **prevSelector)
{
	function	**funcPtr, **funcPtrEnd;

	/* 1.	Iterate over the functions that the selector variable appears in. */

	for (funcPtr = selectorPtr->functions, 
			funcPtrEnd = funcPtr + selectorPtr->currNumFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {

		/* 1.1.	For each function, check if the relevant selector variable of 
				the previous locus, appears in the function. If it does, return 
				TRUE and update 'prevSelector' to point to it. */

		if (checkFuncVars(*funcPtr, prevLocusNum, prevSelector) == TRUE) {
			return TRUE;
		}
	}

	/* 2.	If the relevant selector variable of the previous locus doesn't appear
			in any of the functions (i.e., it has been eliminated) - return FALSE. */

	return FALSE;
}


/******************************************************************************
*
*       Function        checkFuncVars
*       Parameters		funcPtr (pointer to the function which needs to be 
*						checked), locusNum (the index of the locus this 
*						selector relates to), prevSelector (will hold a pointer
*						to the selector variable of locus #locusNum, if this
*						exists).
*       Output          none
*       Description		Cheks if a selector of locus #locusNum exists in the  
*						variable array of the function 'funcPtr'. If it does,
*						returns TRUE and updates 'prevSelector' to point to it.
*       Comments		none
*
******************************************************************************/
boolean checkFuncVars(function *funcPtr, int locusNum, variable **prevSelector)
{
	variable	**varPtr, **varPtrEnd;

	/* 1.	Iterate over the variable array of the function: */

	for (varPtr = funcPtr->variables, 
			varPtrEnd = varPtr + funcPtr->numVariables; varPtr < varPtrEnd; varPtr++) {

		/* 1.1.	For each variable, check if it is a selector variable of 
				locus #locusNum. If it is, return TRUE and update 'prevSelector'
				to point to it. */

		if ((((*varPtr)->originalNode)->lNum == locusNum) && 
				(((*varPtr)->originalNode)->nodeType == SELECTOR)) {
			*prevSelector = *varPtr;
			return TRUE;
		}
	}

	/* 2.	If none of the function's variables were selector variables of locus
			#locusNum, return FALSE. */

	return FALSE;
}



/******************************************************************************
*
*       Function        removeTransFunc
*       Parameters		selectorPtr (pointer to the selector variable of the 
*						current locus whose transmission probability function 
*						needs to be removed), prod (the product which is 
*						being processed), prevLocusNum (the index of the previous 
*						locus).
*       Output          none
*       Description		Removes the transmission probability function defined on
*						'selecVarPtr' and the relevant selector variable of the
*						previous locus from both selector variables and from the
*						product.
*       Comments		none
*
******************************************************************************/
void removeTransFunc(variable *selecVarPtr, product *prod, int prevLocusNum)
{
	function	**funcPtr, **funcPtrEnd, *func;
	variable	*prevSelector;

	/* 1.	Iterate over the functions of this selector variable. */

	for (funcPtr = selecVarPtr->functions, 
			funcPtrEnd = funcPtr + selecVarPtr->currNumFunctions; 
			funcPtr < funcPtrEnd; funcPtr++) {

		func = *funcPtr;

		/* 1.1.	If this is a function that includes the relevant selector variable
				of the previous locus, remove this function from the Product. */

		if (checkFuncVars(func, prevLocusNum, &prevSelector) == TRUE) {
			removeFunctionFromVariable(selecVarPtr, func);
			removeFunctionFromVariable(prevSelector, func);
			removeFunctionFromProduct(prod, func);
			return;
		}
	}
}


/******************************************************************************
*
*       Function        createOneFunc
*       Parameters		selecVarPtr (pointer to the selector variable whose 
*						functions need to be multiplied to create one function), 
*						prod (the product which is being processed).
*       Output          none
*       Description		Multiplies all the functions that 'selecVarPtr' appears
*						in to create one function.  Updates the product accordingly.
*       Comments		none
*
******************************************************************************/
void createOneFunc(variable *selecVarPtr, product *prod)
{
	variable		**varPtr, **varPtrEnd, **varArr;
	function		*combinedFunc, **funcPtr, **funcPtrEnd, *func;
	int				numVars;
	unsigned long	sizeOfTable;
	Probability		*probabList;

	/* 1.	Allocate memory for a variable array that would hold the variables
			of the new function that will be created. */

	myMalloc(varArr, sizeof(variable*) * prod->currNumVars);

	/* 2.	Find all the variables that appear in the functions that need 
			to be combined. */

	getVarsInCombinedFunc(NULL, 0, varArr, selecVarPtr->functions, 
						  selecVarPtr->currNumFunctions, &numVars, &sizeOfTable);

	/* 3.	Allocate memory for the new combined function. */

	myMalloc(probabList, sizeof(Probability) * sizeOfTable);

	/* 4.	Allocate memory for the new combined function & insert it into the product. */

	combinedFunc = insertFunctionToProduct(prod, sizeOfTable, probabList, 0,
										   TRUE, numVars, varArr, TRUE, MAX_RESCALE_UNITS);

	myFree(varArr);

	/* 5.	Multiply all the functions that the variable appears in to create
			one function. */

	multiplyAllFuncs(selecVarPtr->currNumFunctions, selecVarPtr->functions, combinedFunc);

	/* 6.	Remove the functions that have been multiplied from the variables that
			appear in them and from the Product. */

	for (funcPtr = selecVarPtr->functions + selecVarPtr->currNumFunctions - 1, 
			funcPtrEnd = selecVarPtr->functions; funcPtr >= funcPtrEnd; funcPtr--) {

		func = *funcPtr;

		for (varPtr = func->variables, varPtrEnd = varPtr + func->numVariables;
				varPtr < varPtrEnd; varPtr++) {

			removeFunctionFromVariable(*varPtr, func);
		}

		removeFunctionFromProduct(prod, func);
	}

	/* 7.	Add the new combined function to the variables that appear in it. */

	for (varPtr = combinedFunc->variables, varPtrEnd = varPtr + combinedFunc->numVariables;
			varPtr < varPtrEnd; varPtr++) {

		addFunctionToVariable(*varPtr, combinedFunc);
	}
}




/******************************************************************************
*
*       Function        findFunc
*       Parameters		selectorVar (the selector variable whose relevant function
*						needs to be found), prevLocusNum (the index of the previous 
*						locus).
*       Output          none
*       Description		Finds the function which includes only selector variables
*						of the current locus and returns a pointer to it.
*       Comments		none
*
******************************************************************************/
function* findFunc(variable *selectorVar, int prevLocusNum)
{
	function	**funcPtr, **funcPtrEnd;
	variable	*prevSelector;

	for (funcPtr = selectorVar->functions,  
			funcPtrEnd = funcPtr + selectorVar->currNumFunctions;
			funcPtr < funcPtrEnd; funcPtr++) {
		if (checkFuncVars(*funcPtr, prevLocusNum, &prevSelector) == FALSE)
			return *funcPtr;
	}

	return NULL;
}



/******************************************************************************
*
*       Function        Yate
*       Parameters		result (a vector used to store the result of the 
*						multiplication), input (the input vector), theta (the
*						recombination value), size (the size of both vectors,
*						the input vector and the result vector).
*       Output          none
*       Description		Multiplies the 'input' vector by the transition 
*						probability matrix and returns the result in 'result'.
*       Comments		The vector 'result' needs to be allocated before the
*						function is called.
*						Written by Anna Reitman.
*
******************************************************************************/
void Yate(double *result, double *input, double theta, int size)
{
	int		i,j,numIter, halfJ, halfSize;
	double	*tmp, compTheta;
	size_t  copySize;

	numIter = (int)ceil(log(size)/log(2.0));
	copySize = sizeof(double) * size;

	myMalloc(tmp, copySize);
	memcpy(tmp, input, copySize); 

	compTheta = 1.0 - theta;
	halfSize = size/2;

	for (i = 0; i < numIter; i++){
		for (j = 0; j < size; j += 2){
			halfJ = j/2;
			result[j] = (tmp[halfJ] * compTheta) + (tmp[halfJ + halfSize] * theta);
		    result[j+1] = (tmp[halfJ] * theta) + (tmp[halfJ + halfSize] * compTheta);
		}
	    memcpy(tmp, result, copySize); 
	}
	myFree(tmp);
}



/******************************************************************************
*
*       Function        updateFuncsVariables
*       Parameters		newVarArr (an array of pointers to the variables of the 
*						new function), oldVarArr (an array of pointers to the 
*						variables of the old function), numVars (the numver of
*						variables in each array), prod (the product this function
*						belongs to), mapCurrToPrevArr (maps the selectors of the
*						current locus to the relevant selectors of the previous
*						locus).
*       Output          none
*       Description		Updates the array 'newVarArr' with pointers to the 
*						selectors of the previous locus which correspond to the
*						selectors of this locus which appear in 'oldVarArr'.
*       Comments		none
*
******************************************************************************/
void updateFuncsVariables(variable **newVarArr, variable **oldVarArr, int numVars,
						product *prod, variable **mapCurrToPrevArr)
{
	variable **varPtr, **varPtrEnd, **newVarPtr;

	/* 1.	Iterate over the selector variables in 'oldVarArr'. For each one, update
			'newVarArr' to hold the relevant selector variable of the previous 
			locus. */

	for (varPtr = oldVarArr,  varPtrEnd = varPtr + numVars, newVarPtr = newVarArr;
			varPtr < varPtrEnd; varPtr++, newVarPtr++) {

		*newVarPtr = mapCurrToPrevArr[*varPtr - prod->variables];
	}
}


/******************************************************************************
*
*       Function        calcFFT
*       Parameters		result (a vector used to store the result of the 
*						multiplication), input (the input vector), theta (the
*						recombination value), size (the size of both vectors,
*						the input vector and the result vector).
*       Output          none
*       Description		Multiplies the 'input' vector by the transition 
*						probability matrix and returns the result in 'result'.
*       Comments		The vector 'result' needs to be allocated before the
*						function is called.					
*
******************************************************************************/
void calcFFT(double *result, double *input, double theta, int size)
{
	int		n;

	/* 1.	Calculate n (= the size of the inheritance vector). */

	n = (int)ceil(log(size)/log(2.0));

    if (n <= 32) {
		calcComplexFFT(result, input, theta, size, n);
	}
	else {
		calcSimpleFFT(result, input, theta, size, n);
	}
}


/******************************************************************************
*
*       Function        calcComplexFFT
*       Parameters		result (a vector used to store the result of the 
*						multiplication), input (the input vector), theta (the
*						recombination value), size (the size of both vectors,
*						the input vector and the result vector), n (the size of
*						the inheritance vector).
*       Output          none
*       Description		Multiplies the 'input' vector by the transition 
*						probability matrix and returns the result in 'result'.
*       Comments		The vector 'result' needs to be allocated before the
*						function is called.					
*
******************************************************************************/
void calcComplexFFT(double *result, double *input, double theta, int size, int n)
{
	int		i, j, *wArr, weight, *wPtr, *wPtrEnd, *wPtrStart;
	double	base, *Tvals;

	/* 1.	Copy the input vector onto the result vector. */

	memcpy(result, input, sizeof(double) * size);

	/* 2.	Calculate the Fourier Transform of the input vector. */

	calcComplexFTrans(result, size, n);

	/* 3.	Calculate the Fourier Transform of the transition matrix. */

	base = 1 - 2 * theta;

	/* 4.	Allocate memory for an array that would hold the "n+1" 
			different values in the transition matrix, which correspond
			to the number of bits that are different. */

	myMalloc(Tvals, sizeof(double) * (n+1));

	/* 5.	Initialize the array. */

	Tvals[0] = 1;
	for (i = 1; i <= n; i++) {
		Tvals[i] = Tvals[i-1] * base;
	}

	/* 6.	Allocate memory for an n-bit array that would be used to
			iterate over all possible values for the inheritance vector. */

	myMalloc(wArr, sizeof(int) * n);

	for (i=0; i < n; i++)
		wArr[i] = 0;

	wPtrStart = wArr + n - 1;
	wPtrEnd = wArr;

	/* 7.	Calculate the Fourier Transform of the result vector. */

	for (j = 0, weight = 0; j < size; j++){
		result[j] *= Tvals[weight];

		for (wPtr = wPtrStart; wPtr >= wPtrEnd; wPtr--) { 
			if (*wPtr == 0) { 
				*wPtr = 1; 
				weight++;
				break;
			} 
			else {
				*wPtr = 0; 
				weight--;
			}
		}
	}

	/* 8.	Calculate the Reverse Fourier Transform of the result vector. */

	calcComplexFTrans(result, size, n);

	for (j = 0; j < size; j++)
		result[j] /= size;

	/* 9.	Free memory allocated dynamically. */

	myFree(wArr);
	myFree(Tvals);
}



/******************************************************************************
*
*       Function        calcComplexFTrans
*       Parameters		input (the input vector), size (the size of both vectors,
*						the input vector and the result vector), n (the size of
*						the inheritance vector).
*       Output          none
*       Description		Calculate the Fourier Transform of inp.
*       Comments		none
*
******************************************************************************/
void calcComplexFTrans(double *inp, int size, int n)
{
	int		i, j, ei, jFlipBitI;
	double	tmp;
	
    for (i = 0; i < n; i++){
		ei = power2_sl[i];

		for (j = 0; j < size; j++){

			if ((j & ei) == 0) {
				jFlipBitI = j ^ ei;

				tmp = inp[j];
				inp[j] = tmp + inp[jFlipBitI];
				inp[jFlipBitI] = tmp - inp[jFlipBitI];
			}
		}
	}
}

/******************************************************************************
*
*       Function        calcSimpleFFT
*       Parameters		result (a vector used to store the result of the 
*						multiplication), input (the input vector), theta (the
*						recombination value), size (the size of both vectors,
*						the input vector and the result vector), n (the size of
*						the inheritance vector).
*       Output          none
*       Description		Multiplies the 'input' vector by the transition 
*						probability matrix and returns the result in 'result'.
*       Comments		The vector 'result' needs to be allocated before the
*						function is called.					
*
******************************************************************************/
void calcSimpleFFT(double *result, double *input, double theta, int size, int n)
{
	int		i, j, *wArr, weight, *wPtr, *wPtrEnd, *wPtrStart, *incArr;
	double	base, *Tvals;
	long	jumpFactor;

	/* 1.	Copy the input vector onto the result vector. */

	memcpy(result, input, sizeof(double) * size);

	/* 2.	Allocate memory for an n-bit array that would be used to
			iterate over all possible values for the inheritance vector. */

	myMalloc(wArr, sizeof(int) * n);

	/* 3.	Allocate memory for an auxiliary array used to store the 
			necessary increment/decrement in the index of the 
			input/result vectors as a result of an increment/decrement
			of 1 in each one of the n bits of the inheritance vector. */
	
	myMalloc(incArr, sizeof(int) * n);

	/* 4.	Initialize the array. */

	/* 4.1.	Initialize the jump factor. */

	jumpFactor = 1;

	/* 4.2.	Iterate over the vector's bits. */

	for (i = n-1; i >= 0; i--) {

		/* 4.2.1.	Initialize the jumFactor of bit i. */

		incArr[i] = jumpFactor;

		/* 4.2.2	Multiply the jumpFactor by the number of values of the bit (2). */

		jumpFactor *= 2;
	}

	/* 5.	Calculate the Fourier Transform of the input vector. */

	calcSimpleFTrans(result, size, wArr, n, incArr);


	/* 6.	Calculate the Fourier Transform of the transition matrix. */

	base = 1 - 2 * theta;

	/* 7.	Allocate memory for an array that would hold the "n+1" 
			different values in the transition matrix, which correspond
			to the number of bits that are different. */

	myMalloc(Tvals, sizeof(double) * (n+1));

	/* 8.	Initialize the array. */

	Tvals[0] = 1;

	for (i = 1; i <= n; i++) {
		Tvals[i] = Tvals[i-1] * base;
	}


	for (i=0; i < n; i++)
		wArr[i] = 0;

	wPtrStart = wArr + n - 1;
	wPtrEnd = wArr;

	/* 4.	Calculate the Fourier Transform of the result vector. */

	for (j = 0, weight = 0; j < size; j++){
		result[j] *= Tvals[weight];

		for (wPtr = wPtrStart; wPtr >= wPtrEnd; wPtr--) { 
			if (*wPtr == 0) { 
				*wPtr = 1; 
				weight++;
				break;
			} 
			else {
				*wPtr = 0; 
				weight--;
			}
		}
	}

	/* 5.	Calculate the Reverse Fourier Transform of the result vector. */

	calcSimpleFTrans(result, size, wArr, n, incArr);

	for (j = 0; j < size; j++)
		result[j] /= size;

	/* 6.	Free memory. */

	myFree(wArr);
	myFree(incArr);
	myFree(Tvals);
}



/******************************************************************************
*
*       Function        calcSimpleFTrans
*       Parameters		inp (the input vector), size (the size of both vectors,
*						the input vector and the result vector), n (the size of
*						the inheritance vector).
*       Output          none
*       Description		Calculates the Fourier Transform of input.
*       Comments		none
*
******************************************************************************/
void calcSimpleFTrans(double *inp, int size, int *wArr, int n, int *incArr)
{
	int		i, j, *wPtr, *wPtrEnd, *wPtrStart, ind;
	double	operand1, operand2;


	/* 1.	Calculate the Fourier Transform of the input vector. */

	for (i=0; i < n; i++)
		wArr[i] = 0;

	wPtrStart = wArr + n - 1;
	wPtrEnd = wArr;

    for (i = 0; i < n; i++){
		for (j = 0; j < size; j++){

			if (wArr[i] == 0) {
				operand1 = inp[j];
				ind = j + incArr[i];
				operand2 = inp[ind];

                inp[j] = operand1 + operand2;
				inp[ind] = operand1 - operand2;
			}

			for (wPtr = wPtrStart; wPtr >= wPtrEnd; wPtr--) { 
				if (*wPtr == 0) { 
					*wPtr = 1; 
					break;
				} 
				else {
					*wPtr = 0; 
				}
			}
		}
	}
}







