/* $Id: AllelesInheritance.c,v 1.2 2008/08/03 08:15:48 superlink Exp $ */
/**************************************************************************************************************************
//	AllelesInheritance.c: 
//	See all the declarations and explanation in AllelesInheritance.h
//	Written by Anisenia Andrei on 16/07/2008
/**************************************************************************************************************************/


#include "AllelesInheritance.h"
#include "MapWithPositiveRange.h"


int getMaxPersonIndex(int pedSize, Pedigree* pedigree);

AlleleNode* initFounderAlleles(PedigreeDataStruct* pedigreeData, 
									Pedigree* pedigree, MapWithPositiveRange* indexesMap);

void initAlleleNode(AlleleNode* alleleNode, int allele, AlleleNode* parentAllele, int step, int founderAllele);

void initNonFounderAlleles(AllelesInheritance* allelesInhObj, PedigreeDataStruct* pedigreeData, 
						   Pedigree* pedigree, MapWithPositiveRange* indexesMap);

int setPersonAlleles(AllelesInheritance* allelesInhObj, Pedigree* currNonFounder, 
					  MapWithPositiveRange*  indexesMap, int currInhBit);

int updateParentAllele(AlleleNode* childAllele, AlleleNode* parentAllele, AlleleData* parentAlleleData, 
										BitToAllele* inheritanceBit);

void updateAllelesAndInheritanceBits(AllelesInheritance* allelesInhObj, int vectorOfChanges);

int getNumInheritanceBits(Pedigree* pedigree, int* nonFounders, int numOfNonFounders, 
						  MapWithPositiveRange* indexesMap);

int checkBit(AlleleData* parentalAllele, int parentIndex);


/**************************************************************************************************************************
*
*       Function        createAllelesInheritance - Constructor (create function) for the AllelesInheritance
*       Parameters	pedigreeData  - pointer to the struct of type PedigreeDataStruct
*								pedigree
*       Output           AllelesInheritance pointer
*       Description	initiates AllelesInheritance structure with vector = 0
*       Comments		none
*
/**************************************************************************************************************************/
AllelesInheritance* createAllelesInheritance(PedigreeDataStruct* pedigreeData, Pedigree* pedigree) {
	
	AllelesInheritance* allelesInhObj;
	int maxIndex;
	MapWithPositiveRange* indexesMap;

	//	1. initiating mapping of persons indexes
	maxIndex = getMaxPersonIndex(pedigreeData->PedigreeSize, pedigree);
	indexesMap = createMapWithPositiveRange(maxIndex);

	//	2. initiating founder alleles and mapping founder indexes
	sl_myMalloc(allelesInhObj, sizeof(AllelesInheritance));
	allelesInhObj->allAlleles = initFounderAlleles(pedigreeData, pedigree, indexesMap);

	//	3. computing non-founder alleles and mapping non-founder indexes, 
	//		and initiating inheritance vector to 0
	allelesInhObj->nonfounderAlleles = allelesInhObj->allAlleles + 2*(pedigreeData->NumOfFounders);
	allelesInhObj->vector = 0;
	allelesInhObj->numInheritanceBits = getNumInheritanceBits(pedigree, pedigreeData->NonFounders, 
																pedigreeData->NumOfNonFounders, indexesMap);
	sl_myMalloc(allelesInhObj->inheritanceBits, (allelesInhObj->numInheritanceBits)*sizeof(BitToAllele));
	initNonFounderAlleles(allelesInhObj, pedigreeData, pedigree, indexesMap);
		
	deleteMapWithPositiveRange(indexesMap);
	return allelesInhObj;
}



/**************************************************************************************************************************
*
*       Function			getAllele
*       Parameters		allelesInhObj
*								vector 
*								allelePosition  - in the range [0..(2*NumNonFounders - 1)]
*       Output				required allele
*       Description		Get the specific inheritance allele (out of 2*NumNonFounders alleles) 
*								for the given inheritance vector
*       Comments		none
*
/**************************************************************************************************************************/
int getAllele(AllelesInheritance* allelesInhObj, int vector, int allelePosition) {
	int allele;
	//	Check if there is a need to update the internal inheritance vector
	if ((vector != allelesInhObj->vector) && 
				(allelesInhObj->nonfounderAlleles[allelePosition].parentAlleleNode != NULL)) {
		//	Find the vector with changed bits set to 1 (all the rest are 0) using bitwize XOR operation
		int vectorOfChanges = vector ^ (allelesInhObj->vector);
		//	Update inheritance vector and alleles (due to the new vector)
		allelesInhObj->vector = vector;
		updateAllelesAndInheritanceBits(allelesInhObj, vectorOfChanges);
	}
	//	Return required allele
	allele = allelesInhObj->nonfounderAlleles[allelePosition].allele;
	return allele;
}



/**************************************************************************************************************************
*
*       Function			deleteAllelesInheritance
*       Parameters		allelesInhObj
*       Output				none
*       Description		deletes allelesInhObj
*       Comments		none
*
/**************************************************************************************************************************/
void deleteAllelesInheritance(AllelesInheritance* allelesInhObj) {
	free(allelesInhObj->inheritanceBits);
	free(allelesInhObj->allAlleles);
	free(allelesInhObj);
}



/**************************************************************************************************************************
*
*       Function			getNumInheritanceBits
*       Parameters		pedigree
*								nonFounders - order on non-founders such that chldren appear after their parents
*								numOfNonFounders
*       Output				the number of bits
*       Description		counts number of regular bits + const bits that represent the const inheritance 
*								only from non-founders (and not the const inheritance from founders)
*       Comments		none
*
/**************************************************************************************************************************/
int getNumInheritanceBits(Pedigree* pedigree, int* nonFounders, int numOfNonFounders, 
						  MapWithPositiveRange* indexesMap) {
	int i;
	int numBits = 0;

	//	Go over all non-founders
	for(i = 0; i < numOfNonFounders; i++) {
		Pedigree* currNonFounder = pedigree + nonFounders[i];
		
		//	Check his paternal bit
		AlleleData* parentalAllele = currNonFounder->paternalAllele;
		int parentIndex = getValue(indexesMap, currNonFounder->father);
		numBits += checkBit(parentalAllele, parentIndex);

		//	Check his maternal bit
		parentalAllele = currNonFounder->maternalAllele;
		parentIndex = getValue(indexesMap, currNonFounder->mother);
		numBits += checkBit(parentalAllele, parentIndex);
	}
	return numBits;
}

//	Help function of "getNumInheritanceBits" function
int checkBit(AlleleData* parentalAllele, int parentIndex) {
	int bit = 0;
	//	Check if it's regular bit or const non-founder bit
	if ((parentalAllele->bitType == RegularBit) || 
			((parentalAllele->bitType == ConstBit) && (parentIndex == -1))) {
		bit = 1;
	}
	return bit;
}



/**************************************************************************************************************************
*
*       Function			updateAllelesAndInheritanceBits
*       Parameters		allelesInhObj
*								vectorOfChanges - vector with changed bits set to 1 (all the rest are 0)
*       Output				none
*       Description		goes over all bits in the inheritance vector and for each updates the alleles 
*								and parentAlleleNode pointers
*       Comments		none
*
/**************************************************************************************************************************/
void updateAllelesAndInheritanceBits(AllelesInheritance* allelesInhObj, int vectorOfChanges) {
	int bit;
	int powerOf2 = 1;

	//	Go over all inheritance bits in the vector 
	for(bit = 0; bit < allelesInhObj->numInheritanceBits; bit++) {
		AlleleNode* relatedAllele =  allelesInhObj->inheritanceBits[bit].relatedAllele;
		//	Check if the current bit is regular bit
		if (allelesInhObj->inheritanceBits[bit].bitType == RegularBit) {
			//	Check if the current bit was changed (using bitwize AND operation)
			if (vectorOfChanges & powerOf2) {
				//	It was changed => update the parentAlleleNode pointer of the allele related to the current bit
				relatedAllele->parentAlleleNode += relatedAllele->step;
				relatedAllele->step *= (-1);
			}
			powerOf2 *= 2;
		}
		//	Update the allele from the parent
		relatedAllele->allele = relatedAllele->parentAlleleNode->allele;
	}
}



/**************************************************************************************************************************
*
*       Function			getMaxPersonIndex
*       Parameters		pedSize
*								pedigree
*       Output				maximal index (integer)
*       Description		goes over all persons and finds max index between them
*       Comments		none
*
/**************************************************************************************************************************/
int getMaxPersonIndex(int pedSize, Pedigree* pedigree) {
	int i;
	int max = 0;

	for(i = 0; i < pedSize; i++) {
		if (pedigree[i].index > max) {
			max = pedigree[i].index;
		}
	}
	return max;
}



/**************************************************************************************************************************
*
*       Function      initFounderAlleles
*       Parameters	founders - array of founders indexes in the pedigree (in some order)
*							numOfFounders
*							pedigree
*							indexesMap - structure with mapping from person indexes to position 
*													of their first allele in the allAlleles array
*       Output          allAlleles array with initiated founder's alleles
*       Description	goes over all founders and initiates it's alleles in the allAlleles array 
*       Comments	none
*
/**************************************************************************************************************************/
AlleleNode* initFounderAlleles(PedigreeDataStruct* pedigreeData, 
						Pedigree* pedigree, MapWithPositiveRange* indexesMap) {
	int i;
	int* Founders;
	AlleleNode* allAlleles;
	sl_myMalloc(allAlleles, 
		2*(pedigreeData->NumOfFounders + pedigreeData->NumOfNonFounders)*sizeof(AlleleNode));
	
	Founders = pedigreeData->Founders;
	//	Go over all founders defined in the Founders array 
	for(i = 0; i < pedigreeData->NumOfFounders; i++) {
		//	Mapping founder's index (to it's location in the allAlleles array)
		int key = pedigree[Founders[i]].index;
		int value = 2*i;
		setValue(indexesMap, key, value); 
		//	Updating founder's alleles
		initAlleleNode(allAlleles + value, value, NULL, 0, 1);
		initAlleleNode(allAlleles + value + 1, value + 1, NULL, 0, 1);
	}
	return allAlleles;
}


//	Help function for initiating allele node in allAlleles array
void initAlleleNode(AlleleNode* alleleNode, int allele, AlleleNode* parentAllele, int step, int founderAllele) {
	alleleNode->allele = allele;
	alleleNode->parentAlleleNode = parentAllele;
	alleleNode->step = step;
	alleleNode->founderAllele = founderAllele;
}



/**************************************************************************************************************************
*
*       Function      initNonFounderAlleles
*       Parameters	allelesInhObj - structure AllelesInheritance to be updated
*							pedigreeData - pedigree specific data
*							pedigree - array of the persons 
*							indexesMap - structure with mapping from person indexes to position 
*													of their first allele in the allAlleles array
*       Output          none
*       Description	goes over all non-founders and initiates their alleles in the allAlleles array 
*							simultaneously initiating inheritanceBits to the appropriate alleles 
*       Comments	uses pedigreeData->NonFounders array and assumes that
*							the order of non-founder in it is such that parents appear before the children 
*
/**************************************************************************************************************************/
void initNonFounderAlleles(AllelesInheritance* allelesInhObj, PedigreeDataStruct* pedigreeData, 
						   Pedigree* pedigree, MapWithPositiveRange* indexesMap) {
	int i;
	int* NonFounders;
	int currInhBit = 0;

	NonFounders = pedigreeData->NonFounders;
	//	Go over all non-founders and for each:
	for(i = 0; i < pedigreeData->NumOfNonFounders; i++) {
		Pedigree* currNonFounder = pedigree + NonFounders[i];
		//	1. Map the non-founder index to its position in the allAlleles array
		int key = currNonFounder->index;
		int value = 2*(pedigreeData->NumOfFounders + i);
		setValue(indexesMap, key, value); 

		//	2. Update current non-founder allele nodes and appropriate inheritance vector bits
		currInhBit += setPersonAlleles(allelesInhObj, currNonFounder, indexesMap, currInhBit);
	}
}



/**************************************************************************************************************************
*
*       Function      setPersonAlleles
*       Parameters	allelesInhObj - structure AllelesInheritance to be updated
*							currNonFounder - pointer to the currently treated non-founder in the pedigree 
*							indexesMap - structure with mapping from person indexes to position 
*													of their first allele in the allAlleles array
*							currInhBit - currently treated inheritance vector bit
*       Output          the number of {"ConstBit" inherited from non-founder, "RegularBit"} bits of the person
*       Description	performs the computation of the person alleles and stores them into the allAllleles array,
*							updates pointers from inheritance bits to the related alleles if needed
*       Comments	none 
*
/**************************************************************************************************************************/
int setPersonAlleles(AllelesInheritance* allelesInhObj, Pedigree* currNonFounder, 
										MapWithPositiveRange*  indexesMap, int currInhBit) {
	AlleleNode* myAllele;
	AlleleNode* parentAllele;
	int myAlleleIndex, parentAlleleIndex;
	int numBits = 0; 

	myAlleleIndex = getValue(indexesMap, currNonFounder->index);
	myAllele = allelesInhObj->allAlleles + myAlleleIndex;
	
	//	1. treating father's allele of the person
	parentAlleleIndex = getValue(indexesMap, currNonFounder->father);
	parentAllele = allelesInhObj->allAlleles + parentAlleleIndex;
	numBits += updateParentAllele(myAllele, parentAllele, currNonFounder->paternalAllele, 
														allelesInhObj->inheritanceBits + currInhBit);

	//	2. treating mother's allele of the person
	myAllele++;
	parentAlleleIndex = getValue(indexesMap, currNonFounder->mother);
	parentAllele = allelesInhObj->allAlleles + parentAlleleIndex;
	numBits += updateParentAllele(myAllele, parentAllele, currNonFounder->maternalAllele, 
														allelesInhObj->inheritanceBits + currInhBit + numBits);
	return numBits;
}



/**************************************************************************************************************************
*
*       Function      updateParentAllele
*       Parameters	childAllele - pointer to the child allele node
*							parentAllele - pointer to the parent allele node
*							parentAlleleData - pointer to the specific inheritance data from the parent to the child
*							inheritanceBits - pointer to the currently treated bit int the array inheritanceBits
*														it will be updated due to parentAlleleData
*       Output         1 - if the current bit is (RegularBit) or (ConstBit inherited from non-founder)
*							0 - else
*       Description	performs the computation of the person alleles and stores them into the allAllleles array,
*							updates pointers from inheritance bits to the related alleles if needed
*       Comments	none 
*
/**************************************************************************************************************************/
int updateParentAllele(AlleleNode* childAllele, AlleleNode* parentAllele, AlleleData* parentAlleleData, 
						BitToAllele* inheritanceBit) {
	int allele;
	int step = 0;
	int inhBit = 0;

	switch (parentAlleleData->bitType) {
		case NotUsedBit:		allele = 0;
										parentAllele = NULL;
										break;
		case ConstBit:			allele = parentAllele[parentAlleleData->value].allele;
										if (parentAllele->founderAllele == 0) {
											//	This bit is inherited from non-founder
											inheritanceBit->relatedAllele = childAllele;
											inheritanceBit->bitType = ConstBit;
											parentAllele = parentAllele + parentAlleleData->value;
											inhBit = 1;
										}
										else {
											//	This bit is inherited from founder
											parentAllele = NULL;
										}
										break;
		case ReductionBit:	allele = parentAllele->allele;
										parentAllele = NULL;
										break;
		case RegularBit:		step = 1;
										allele = parentAllele->allele;
										inheritanceBit->relatedAllele = childAllele;
										inheritanceBit->bitType = RegularBit;
										inhBit = 1;
										break;
		default:						break;
	}
	initAlleleNode(childAllele, allele, parentAllele, step, 0);
	return inhBit;
}