/* $Id: AllelesInheritance.h,v 1.2 2008/08/03 08:15:48 superlink Exp $ */
/************************************************************************************/
//	AllelesInheritance.h:
//	Includes the declaration of AllelesInheritance module (structure + functions):
//	
//	AllelesInheritance is aimed to replace the AlleleInh matrix - matrix where for each inheritance vector
//	the 2*n alleles of non-founders are stored, assuming the inheritance from {0,1,..,2*f-1} founder alleles
//	due to the specific inheritance vector.
//	The need to replace the existing matrix arised from the high memory requirements of matrix storage:
//	AlleleInh matrix occupies O(2^(2n-f)*2n) space  - this requirement can cancel the usage of HMM
//	in the LOD-score calculation for the big families (more than 22 inheritance vector bits).
//
//	The new AllelesInheritance module stores (2n+2f) alleles (founders + non-founders) for the specific
//	inheritance vector, and when the inheritance vector changes, the inheritance alleles must be recalculated,
//	So it occupies (O(2n+2f)+O(2n-f)) = O(4n+f) space
//	(2n+2f) - for storing all alleles of founders + non-founders
//	(2n-f) - for storing the bits of inheritance vector
//
//	In the AlleleInh matrix each operation of retrieving some allele for some vector takes O(1) time,
//	In the AllelesInheritance module: 
//	if the parameter inheritance vector equals to the vector stored in the module then it takes O(1) time
//	to retrieve the required allele,
//	else O(2n-f) time - in order to update the alleles inheritance due to the new inheritance vector
//
//	Created by Anisenia Andrei on 11/07/2008 
/************************************************************************************/
#ifndef ALLELES_INHERITANCE_H
#define ALLELES_INHERITANCE_H

#include "../gdata.h"

/************************************************************************************/
//	AllelesInheritance structure definition:

//	Help structure for storing the required information for 1 inheritance allele
typedef struct AlleleNode{
	int allele;														//	value of the currently stored allele 
	struct AlleleNode* parentAlleleNode;		//	pointer to the parent allele node (due to current inheritance vector)
	int step;					//	the movement of parentAlleleNode pointer if the current allele inheratance would change
									//	it may be +1 or -1 according to the current position (paternal or maternal)
	int founderAllele;		//	1 - if this allele is founder's allele, else 0 
}AlleleNode;

//	Help structure for storing the pointers from inheritance vector's bits to the related inheritance allele
typedef struct BitToAllele{
	BitType bitType;			//	May be "ConstBit" or "RegularBit"; 
										//	"ReductionBit" and "NotUsedBit" don't appear in this structure
	AlleleNode* relatedAllele;
}BitToAllele;


//	The main structure
typedef struct AllelesInheritance{
	AlleleNode* allAlleles;							//	the array of all alleles of size = 2n+2f (founders +nonfounders)
	AlleleNode* nonfounderAlleles;				//	the array of nonfounder alleles of size = 2n (sub-array of allAlleles)
	BitToAllele* inheritanceBits;					//	the array of inheritance bits pointing to the related alleles of size = 2n-f
	int vector;												//	currently presented inheritance vector in the structure
	int numInheritanceBits;							//	the number of inheritance bits in vector
}
AllelesInheritance;



/************************************************************************************/
//	 AllelesInheritance functions declaration:

//	Constructor (create function) for the AllelesInheritance
AllelesInheritance* createAllelesInheritance(PedigreeDataStruct* pedigreeData, Pedigree* pedigree);


//	Get the specific inheritance allele (out of 2*NumNonFounders alleles) for the specific inheritance vector
int getAllele(AllelesInheritance* allelesInhObj, int vector, int allelePosition);  


//	Destructor (delete function) for the AllelesInheritance
void deleteAllelesInheritance(AllelesInheritance* allelesInhObj);









#endif

