/* $Id: initFoundersChildren.c,v 1.2 2008/08/03 08:15:48 superlink Exp $ */
/**************************************************************************************************************************
//	initFoundersChildren.c:
//	Contains functions for initiating the array FoundersChildren defined in gdata.h
//	The above array contains integer number for each founder defined in the PedigreeData.Founders array,
//	that means it's size equals to the size of PedigreeData.Founders array and the order of foundes is also
//	the same.
//	The integer number stored per founder is actually a bit array (in integer's binary representation)
//	where for each bit "i": if this founder has a child with one if his alleles related to the bit number "i" in the 
//	full-inheritance space vector (vector's size = 2n), then it's set to 1, else 0.
//	Written by Anisenia Andrei on 18/07/2008
/**************************************************************************************************************************/

#include "MapWithPositiveRange.h"
#include "../gdata.h"


/**************************************************************************************************************************/
//	Functions declaration:


//	The main function for the external use
int* initFoundersChildren(PedigreeDataStruct* pedigreeData, Pedigree* pedigree);


//	The internal functions
int getMaxIndex(int pedSize, Pedigree* pedigree);

void mapFounderIndexesToChildren(int* founders, int numOfFounders, Pedigree* pedigree, 
								 MapWithPositiveRange* indexesMap);

void setFoundersChildrenBits(int* foundersChildren, int* nonFounders, int numOfNonFounders, 
							 Pedigree* pedigree, MapWithPositiveRange* indexesMap);

void addChildBit(int parentPosition, int* foundersChildren, AlleleData* parentAllele);



/**************************************************************************************************************************/
//	Implementation


/**************************************************************************************************************************
*
*       Function			initFoundersChildren
*       Parameters		pedigreeData  - pointer to the struct of type PedigreeDataStruct
*								pedigree
*       Output				FoundersChildren
*       Description		initiates FoundersChildren
*       Comments		none
*
/**************************************************************************************************************************/
int* initFoundersChildren(PedigreeDataStruct* pedigreeData, Pedigree* pedigree) {
	int* foundersChildren;
	int maxIndex;
	MapWithPositiveRange* indexesMap;

	//	1. get the maximal index in the pedigree 
	maxIndex = getMaxIndex(pedigreeData->PedigreeSize, pedigree);
	//	2. initiate the map
	indexesMap = createMapWithPositiveRange(maxIndex);
	//	3. for each founder map his "index" in the pedigree to his position 
	//		in the Founders array (=position in the foundersChildren array)
	mapFounderIndexesToChildren(pedigreeData->Founders, pedigreeData->NumOfFounders, 
		pedigree, indexesMap);
	//	4. initiate foundersChildren array with 0 in each cell
	sl_myCalloc(foundersChildren, pedigreeData->NumOfFounders, sizeof(int));
	//	5. for all founder's children in the pedigree add their full-space inheritance bits to foundersChildren array
	setFoundersChildrenBits(foundersChildren, pedigreeData->NonFounders, 
		pedigreeData->NumOfNonFounders, pedigree, indexesMap);

	deleteMapWithPositiveRange(indexesMap);
	return foundersChildren;
}



/**************************************************************************************************************************
*
*       Function			getMaxPersonIndex
*       Parameters		pedSize
*								pedigree
*       Output				maximal index (integer)
*       Description		goes over all persons and finds max index between them
*       Comments		none
*
/**************************************************************************************************************************/
int getMaxIndex(int pedSize, Pedigree* pedigree) {
	int i;
	int max = 0;

	for(i = 0; i < pedSize; i++) {
		if (pedigree[i].index > max) {
			max = pedigree[i].index;
		}
	}
	return max;
}



/**************************************************************************************************************************
*
*       Function			mapFounderIndexesToChildren
*       Parameters		founders - order of founders in the pedigree
*								numOfFounders
*								pedigree
*								indexesMap - map to be filled with mappings
*       Output				none
*       Description		goes over all founders defined in the "founders" array and for each founder
*								stores the pair (founder index in the pedigree, founder position in the founders array)
*								into the indexesMap
*       Comments		none
*
/**************************************************************************************************************************/
void mapFounderIndexesToChildren(int* founders, int numOfFounders, Pedigree* pedigree, 
								 MapWithPositiveRange* indexesMap) {
	int i;
	for(i = 0; i < numOfFounders; i++) {
		setValue(indexesMap, pedigree[founders[i]].index, i);
	}
}



/**************************************************************************************************************************
*
*       Function			setFoundersChildrenBits
*       Parameters		foundersChildren
*								nonFounders, numOfNonFounders
*								pedigree
*								indexesMap - map to be used in order to get founders position in the 
*										foundersChildren array using their "index" in the pedigree
*       Output				none
*       Description		goes over all non-founders defined in the "nonFounders" array and for each non-founder
*								if his parent is a founder, then add his full-space inheritance bits to foundersChildren array
*       Comments		none
*
/**************************************************************************************************************************/
void setFoundersChildrenBits(int* foundersChildren, int* nonFounders, int numOfNonFounders, 
							 Pedigree* pedigree, MapWithPositiveRange* indexesMap) {
	int i;
	int parentPosition;
	AlleleData* parentAllele;

	for(i = 0; i < numOfNonFounders; i++) {
		Pedigree* currNonFounder  = pedigree + nonFounders[i];
		
		if (!(ProgramSpecificData.sexLinked)) {
			//	Check if the father is founder and if yes, add to foundersChildren the bit 
			parentPosition = getValue(indexesMap, currNonFounder->father);
			parentAllele = currNonFounder->paternalAllele; 
			addChildBit(parentPosition, foundersChildren, parentAllele);
		}

		//	Check if the mother is founder and if yes, add to foundersChildren the bit 
		parentPosition = getValue(indexesMap, currNonFounder->mother);
		parentAllele = currNonFounder->maternalAllele; 
		addChildBit(parentPosition, foundersChildren, parentAllele);
	}
}


//	Help function for "setFoundersChildrenBits" function
void addChildBit(int parentPosition, int* foundersChildren, AlleleData* parentAllele) {
	if (parentPosition != -1) {
			//	Generate a number with current non-founder full-inheritance space bit set to 1 using
			//	"shift left" bit-operation, all the rest of the bits are 0
			int currNonFounderBit = 1<<(parentAllele->noSpeedUpNumber);
			//	Add current non-founder bit to the number of the parent in the foundersChildren array
			//	using bitwize OR operation
			foundersChildren[parentPosition] |= currNonFounderBit;
		}
}
