/* $Id: InfoBucketStorager.h,v 1.2 2008/08/03 08:15:48 superlink Exp $ */
/************************************************************************************/
//	InfoBucketStorager.h
//	Class InfoBucketStorager<T>: Abstract Base Class
//	Defines the interface for storage of some information nodes - infos into buckets,
//	meaning that it is possible to store 2 infos with the same identifiers in different buckets,
//	while the storage of 2 infos with the same identifier in the same bucket is impossible.
//	Storing/restoring is performed using 2 essential parameters: bucket index and info identifier.
//	The class has template parameter which is the type of info identifier.
//	In the current version the type of info pointer is void*, but it may be changed.
//	The class throws exceptions (lists of exceptions can be found in method declarations)
//	All possible exceptions are defined in the class InfoBucketStoragerException
/************************************************************************************/
#ifndef INFO_BUCKET_STORAGER_H
#define INFO_BUCKET_STORAGER_H

#include "../MemoryManagerInputOutputFormat.h"


class InfoBucketStoragerException {
	
	public:
		//	Definition of possible exception values
		typedef enum {		BucketOutOfRange,		//	when the given bucket index is out of buckets range
										BucketFull,						//	bucket is full and no additional info can be stored in it
										InfoExist,						//	info with the given identifier already exists in the given bucket
										InfoNotExist,					//	info with the given identifier doesn't exist in the given bucket
										InfoCorrupted				//	the info was corrupted due to some problem in the storage
		} ExceptionValue;

		//	The value of current exception
		const ExceptionValue value;

		//	Constructor
		InfoBucketStoragerException(ExceptionValue v) : value(v) {};
		//	Destructor is compiler defined (default destructor)
};



template <typename InfoIdentifier> class InfoBucketStorager {

	public:
		
		//	Operational interface:
		//	Store the info with the given identifier into the given bucket (by its pointer)
		virtual void storeInfoToBucket(InfoPtr infoPtr, InfoIdentifier infoID, int bucketIndex) 
							throw(InfoBucketStoragerException) = 0;
		//	Restore the pointer to info with the given identifier from the given bucket (and move it from the bucket)
		virtual InfoPtr restoreInfoFromBucket(InfoIdentifier infoID, int bucketIndex)
							throw(InfoBucketStoragerException) = 0;
		
		//	Virtual destructor
		virtual ~InfoBucketStorager() {};


	protected:
		//	throwBucketOutOfRangeE: if the bucketIndex is out of [0..numBuckets-1] range =>
		//	throw BucketOutOfRange exception, else do nothing
		inline void throwBucketOutOfRangeE(int bucketIndex, int numBuckets) throw(InfoBucketStoragerException) {
			if ((bucketIndex < 0) || (bucketIndex >= numBuckets)) {
				throw 	InfoBucketStoragerException(InfoBucketStoragerException::BucketOutOfRange);
			}
		};

		//	Size of slot for storage (or the upper limit of this size) in bytes
		//	For the use in derived classes
		int slotInfoSize;

		//	Protected Constructor for the use in derived classes
		InfoBucketStorager(int pSlotInfoSize) : slotInfoSize(pSlotInfoSize) {};
};


#endif

