/* $Id: SingleInfoBucketStoragerHDD.h,v 1.3 2008/08/25 07:52:17 sandanis Exp $ */
/************************************************************************************/
//	SingleInfoBucketStoragerHDD.h
//	Class SingleInfoBucketStoragerHDD<T>: Derieved from InfoBucketStorager<T>
//	Implements the interface defined by InfoBucketStorager<T> class (see description in InfoBucketStorager.h)
//	The class has the following implementation specifications:
//	a. It allows maximum 1 info node in the bucket (trying to store the 2nd info into the same bucket leads
//	to the BucketFullException)
//	b. The infos are stored to the Hard Disc Drive (HDD) 
//	c. 3 external C-style function pointers must be supplied (defined in MemoryManagerInputOutputFormat.h):
//		1.	function for deleting the InfoPtr from RAM
//		2. function for writing the data pointed by InfoPtr to the file on hard disc
//		3. function for reading the data from the file on the disc into the InfoPtr 
//		None of the functions may be NULL!!! Else the NullFuncPointerException is thrown 
//	d. The storager of this type has it's own folder to save files (folder's name is passed as a parameter to constructor)
/************************************************************************************/
#ifndef SINGLE_INFO_BUCKET_STORAGER_HDD_H
#define SINGLE_INFO_BUCKET_STORAGER_HDD_H


#include <string>
#include <stdio.h>
using namespace std;

#if defined _WIN32
	#include <direct.h>
	#include <windows.h>
#else
	#include <sys/stat.h> 
	#include <dirent.h>
#endif

#include "InfoBucketStorager.h"


/************************************************************************************/
//	Class definition:
template <typename InfoIdentifier> 
class SingleInfoBucketStoragerHDD: public virtual InfoBucketStorager<InfoIdentifier> {

	public:
		//	Exception definition (thrown in the case on null function pointer)
		class NullFuncPointerException{};

		//	Constructor
		SingleInfoBucketStoragerHDD(string pFolderPath, size_t pNumBuckets,
																FuncDeleteInfoFromRAM fDeleteInfo,
																FuncWriteInfoToFile fWriteInfoToFile,
																FuncReadInfoFromFile fReadInfoFromFile,
																int pSlotInfoSize)
			throw(NullFuncPointerException);
		
		//	Pure virtual methods implementation:
		//	Method storeInfoToBucket
		virtual void storeInfoToBucket(InfoPtr infoPtr, InfoIdentifier infoID, int bucketIndex) 
			throw(InfoBucketStoragerException);

		//	Method restoreInfoFromBucket
		virtual InfoPtr restoreInfoFromBucket(InfoIdentifier infoID, int bucketIndex)
			throw(InfoBucketStoragerException) ;

		//	Virtual destructor
		virtual ~SingleInfoBucketStoragerHDD();


	private:
		string storageFolderPath;
		int numBuckets;
		//	Internal structure for storing info identifier + info file name
		class InfoNode {
			public:
				InfoIdentifier infoID;
				string infoFileName;
				InfoNode(InfoIdentifier id, string fileName) : infoID(id), infoFileName(fileName) {};
		};
		//	Vector of buckets (each bucket points to InfoNode)
		vector<InfoNode*> buckets;
		//	Pointer to c-style function for deleting the infos from RAM 
		FuncDeleteInfoFromRAM funcDeleteInfoFromRAM;
		//	Pointer to c-style function for writing the given info to file
		FuncWriteInfoToFile funcWriteInfoToFile;
		//	Pointer to c-style function for reading the given info from file
		FuncReadInfoFromFile funcReadInfoFromFile;

		//	Method for removing pre-existing storage folders with all content (for Windows and for Unix)
		void removePreexistingStorageFolder(string storageFolderPath);	
};





//---------------------------------------------------------------------------------------------------------------------------
//	Implementation of SingleInfoBucketStoragerRAM public methods:



//	Constructor
template <typename InfoIdentifier>
SingleInfoBucketStoragerHDD<InfoIdentifier>::SingleInfoBucketStoragerHDD(
		string pFolderPath, 
		size_t pNumBuckets,
		FuncDeleteInfoFromRAM fDeleteInfo,
		FuncWriteInfoToFile fWriteInfoToFile,
		FuncReadInfoFromFile fReadInfoFromFile,
		int pSlotInfoSize)
			throw(NullFuncPointerException):
				numBuckets(pNumBuckets), 
				buckets(pNumBuckets, NULL),
				InfoBucketStorager<InfoIdentifier>(pSlotInfoSize) {

	//	Update function pointers and check them for NULL
	if ((fDeleteInfo == NULL) || (fWriteInfoToFile == NULL) || (fReadInfoFromFile == NULL)) {
		throw NullFuncPointerException();
	}
	else {
		funcDeleteInfoFromRAM = fDeleteInfo;
		funcWriteInfoToFile = fWriteInfoToFile;
		funcReadInfoFromFile = fReadInfoFromFile;
	}
	//	Updating folder path name for the storager
	storageFolderPath = pFolderPath;
	//	Creating the specified folder
	removePreexistingStorageFolder(storageFolderPath);
	#if defined _WIN32
		mkdir(storageFolderPath.c_str());
	#else
		mkdir(storageFolderPath.c_str(), 00777);
	#endif
}



/******************************************************************************
*
*       Function        storeInfoToBucket
*       Parameters	infoPtr - pointer to the stored information
*								infoID - identifier of the stored information
*								bucketIndex - index of the bucket for information storing
*       Output           none
*       Description	Stores the info with the given identifier into the given bucket (by its pointer) to HDD
*								and deletes the info by the pointer
*       Comments		throws 4 types of exceptions:
*								1. BucketOutOfRange - bucket index is not in the range [0..numBuckets-1]
*								2. BucketFull - there is already some info in the specified bucket
*								3. InfoExist - the info with the given identifier already exists in the
*								given bucket
*								    Exception 3 is a sub-case of Exception 2 (and is thrown instead)
*								4. InfoCorrupted exception is thrown in the case when there was a problem 
*									while writing to file 
*
******************************************************************************/
template <typename InfoIdentifier> 
void SingleInfoBucketStoragerHDD<InfoIdentifier>::
storeInfoToBucket(InfoPtr infoPtr, InfoIdentifier infoID, int bucketIndex) 
			throw(InfoBucketStoragerException) {						
	
	//	Check that given bucket index is in the valid range, else throw exception
	InfoBucketStorager<InfoIdentifier>::throwBucketOutOfRangeE(bucketIndex, numBuckets);
	//	Check that the specified bucket is full
	if (buckets[bucketIndex] == NULL) {
		//	Bucket is empty => store the given information:
		//	1. Create file path name 
		char bucketIndexString[10];
		sprintf(bucketIndexString,"%d", bucketIndex);
		string filePath = storageFolderPath + "/" + bucketIndexString;
		//	2. Write to file and check the returned value
		if (funcWriteInfoToFile(infoPtr, filePath.c_str(), InfoBucketStorager<InfoIdentifier>::slotInfoSize) == 0) {
			throw  InfoBucketStoragerException(InfoBucketStoragerException::InfoCorrupted);
		}
		//	3. Writing succeded => create a new info node in the buckets array
		buckets[bucketIndex] = new InfoNode(infoID, filePath);
		//	4. Delete the info from RAM
		funcDeleteInfoFromRAM(infoPtr);
	}
	else {
		//	Bucket is not empty: throw BucketFullException or InfoExistException(if the id is equal)
		if (buckets[bucketIndex]->infoID == infoID) {
			throw InfoBucketStoragerException(InfoBucketStoragerException::InfoExist);
		}
		else {
			throw InfoBucketStoragerException(InfoBucketStoragerException::BucketFull);
		}
	}			
}



/******************************************************************************
*
*       Function        restoreInfoFromBucket
*       Parameters	infoID - identifier of the stored information
*								bucketIndex - index of the bucket for information restoring
*       Output           pointer to the required info
*       Description	Restore the info with the given identifier from the given bucket 
*								by reading it from the file, deletes the file and deletes the bucket cell
*       Comments		throws 3 types of exceptions:
*								1. BucketOutOfRange - bucket index is not in the range [0..numBuckets-1]
*								2. InfoNotExist - if the info with the specified id doesn't exist in the 
*								given bucket
*								3. InfoCorrupted - if the reading operation didn't succeed
*
******************************************************************************/
template <typename InfoIdentifier> 
InfoPtr SingleInfoBucketStoragerHDD<InfoIdentifier>::
restoreInfoFromBucket(InfoIdentifier infoID, int bucketIndex)
			throw(InfoBucketStoragerException) {
	
	//	Check that given bucket index is in the valid range, else throw exception
	InfoBucketStorager<InfoIdentifier>::throwBucketOutOfRangeE(bucketIndex, numBuckets);
	//	Check that info with the specified id exists in the specified bucket
	InfoNode* infoNodePtr = buckets[bucketIndex];
	if ((infoNodePtr == NULL) || (infoNodePtr->infoID != infoID)) {
		throw InfoBucketStoragerException(InfoBucketStoragerException::InfoNotExist);
	}
	//	Info exists => restore the information from the bucket and clean the bucket cell:
	//	1. Read the information from the file of infoNode
	InfoPtr infoPtr = funcReadInfoFromFile((infoNodePtr->infoFileName).c_str(), 
																InfoBucketStorager<InfoIdentifier>::slotInfoSize);
	//	2. Check that didn't return NULL (else throw InfoCorrupted)
	if (infoPtr == NULL) {
		throw InfoBucketStoragerException(InfoBucketStoragerException::InfoCorrupted);
	}
	//	3. Reading succeded => remove the file and delete the bucket cell
	remove((infoNodePtr->infoFileName).c_str());
	delete infoNodePtr;
	buckets[bucketIndex] = NULL;
	return infoPtr;
}



/******************************************************************************
*
*       Function        Destructor of SingleInfoBucketStoragerHDD<InfoIdentifier> object
*       Parameters	none
*       Output           none
*       Description	Removes all files mentioned in the buckets cells, and deletes all cells != NULL 
*								removes the storager directory
*       Comments		none 
*
******************************************************************************/
template <typename InfoIdentifier> 
SingleInfoBucketStoragerHDD<InfoIdentifier>::~SingleInfoBucketStoragerHDD() {
	//	Go through all info nodes and delete them
	for(int i = 0; i < numBuckets; i++) {
		if (buckets[i] != NULL) {
			//	Remove the file by it's name
			remove((buckets[i]->infoFileName).c_str());
			//	Delete info node 
			delete buckets[i];
		}
	}
	//	Remove storager folder
	rmdir(storageFolderPath.c_str());
}



//---------------------------------------------------------------------------------------------------------------------------
//	Implementation of SingleInfoBucketStoragerRAM private methods:



/******************************************************************************
*
*       Function			removePreexistingStorageFolderWindows
*       Parameters		storageFolderPath - path to the folder to be removed
*       Output				none
*       Description		Removes all files in the folder and the folder itself
*       Comments		none
*
******************************************************************************/
template <typename InfoIdentifier>
void SingleInfoBucketStoragerHDD<InfoIdentifier>:: 
removePreexistingStorageFolder(string storageFolderPath) {
	
	#if defined _WIN32	
	//	Windows-specific code -------------------------------------------------------------------
		//	1. Defining windows variables for listing files in the directory   
		WIN32_FIND_DATA fileData;
		HANDLE hFind;
		//	2. Trying to find some file in the directory
		wchar_t* folderPath = new wchar_t[storageFolderPath.size() + 3];
		mbstowcs(folderPath, (storageFolderPath + "\\*").c_str(), storageFolderPath.size() + 3);
		hFind = FindFirstFile(folderPath, &fileData);
		if (hFind != INVALID_HANDLE_VALUE) {
			//	loop over all files in the directory and remove them
			do {
				char fileName[100];
				wcstombs(fileName, fileData.cFileName, 100);
				remove((storageFolderPath + "/" + fileName).c_str());
			} 
			while (FindNextFile(hFind, &fileData) != 0);
		}
		FindClose(hFind);
		delete[] folderPath;
	
	#else
	//	Unux-specific code -------------------------------------------------------------------------
		DIR* dirPtr = opendir(storageFolderPath.c_str());
		if (dirPtr != NULL) {
			for(struct dirent* filePtr = readdir(dirPtr); filePtr != NULL; filePtr = readdir(dirPtr)) { 
				char fileName[100];
				sprintf(fileName, "%s", filePtr->d_name);
				remove((storageFolderPath + "/" + fileName).c_str());
			}
		}
	#endif
	//	Remove preexisting storager folder
	rmdir(storageFolderPath.c_str());
}





#endif