/* $Id: SingleInfoBucketStoragerHDD_RAM.h,v 1.3 2008/09/28 07:03:22 sandanis Exp $ */
/************************************************************************************/
//	SingleInfoBucketStoragerHDD_RAM.h
//	Class SingleInfoBucketStoragerHDD_RAM<T>: 
//	Derieved from SingleInfoBucketStoragerHDD<T> and SingleInfoBucketStoragerRAM<T>
//	Implements the interface defined by InfoBucketStorager<T> class (see description in InfoBucketStorager.h)
//	The class has the following implementation specifications:
//	a. All the specifications of its base classes (their union)
//	b. Stores both to RAM and to HDD according to the following specification:
//		1. Has 2 fields (that are also parameters to the constructor): numBucketsRAM and numBucketsHDD
//		2. (numBucketsRAM + numBucketsHDD) buckets can be filled with info nodes
//		3. When the (bucket index < numBucketsHDD) then SingleInfoBucketStoragerHDD_RAM<T> behaves as
//			SingleInfoBucketStoragerHDD<T> in the [0 .. (numBucketsHDD-1)] buckets range
//		4. When the (bucket index >= numBucketsHDD) then SingleInfoBucketStoragerHDD_RAM<T> behaves as
//			SingleInfoBucketStoragerRAM<T> in the [0 .. (numBucketsRAM-1)] buckets range by using the
//			(bucket index - numBucketsHDD) as the index of SingleInfoBucketStoragerRAM<T> bucket
//		5. Hence all the indexes in range [0 .. (numBucketsHDD-1)] are stored to HDD and
//			all the indexes in range [numBucketsHDD .. (numBucketsRAM+numBucketsHDD-1)] are stored to RAM
/************************************************************************************/
#ifndef SINGLE_INFO_BUCKET_STORAGER_HDD_RAM_H
#define SINGLE_INFO_BUCKET_STORAGER_HDD_RAM_H


#include "SingleInfoBucketStoragerRAM.h" 
#include "SingleInfoBucketStoragerHDD.h" 



/************************************************************************************/
//	Class definition:
template <typename InfoIdentifier> 
class SingleInfoBucketStoragerHDD_RAM:	public SingleInfoBucketStoragerHDD<InfoIdentifier>,
		public SingleInfoBucketStoragerRAM<InfoIdentifier> {

	public:
		//	Constructor: NullFuncPointerException may be thrown by SingleInfoBucketStoragerHDD base class
		SingleInfoBucketStoragerHDD_RAM(size_t pNumBucketsHDD, size_t pNumBucketsRAM, 
																string pFolderPath, 
																FuncDeleteInfoFromRAM fDeleteInfo,
																FuncWriteInfoToFile fWriteInfoToFile,
																FuncReadInfoFromFile fReadInfoFromFile,
																int pSlotInfoSize)
								throw(typename SingleInfoBucketStoragerHDD<InfoIdentifier>::NullFuncPointerException):
			numBucketsHDD(pNumBucketsHDD),
			numBucketsRAM(pNumBucketsRAM),
			InfoBucketStorager<InfoIdentifier>(pSlotInfoSize),
			SingleInfoBucketStoragerHDD<InfoIdentifier>(pFolderPath, pNumBucketsHDD,
																fDeleteInfo, fWriteInfoToFile, fReadInfoFromFile, pSlotInfoSize),
			SingleInfoBucketStoragerRAM<InfoIdentifier>(pNumBucketsRAM, fDeleteInfo, pSlotInfoSize) {};


		//	Pure virtual methods implementation:
		//	Method storeInfoToBucket
		virtual void storeInfoToBucket(InfoPtr infoPtr, InfoIdentifier infoID, int bucketIndex) 
			throw(InfoBucketStoragerException);

		//	Method restoreInfoFromBucket
		virtual InfoPtr restoreInfoFromBucket(InfoIdentifier infoID, int bucketIndex)
			throw(InfoBucketStoragerException) ;

		//	Virtual destructor
		virtual ~SingleInfoBucketStoragerHDD_RAM() {};

	private:
		int numBucketsHDD;
		int numBucketsRAM;
};



//---------------------------------------------------------------------------------------------------------------------------
//	Implementation of SingleInfoBucketStoragerRAM public methods:



/******************************************************************************
*
*       Function        storeInfoToBucket
*       Parameters	infoPtr - pointer to the stored information
*								infoID - identifier of the stored information
*								bucketIndex - index of the bucket for information storing
*       Output           none
*       Description	Choose one of the base storagers due to the bucket index and operates it
*       Comments		InfoBucketStoragerException may be thrown from the method calls to
*								storeInfoToBucket of SingleInfoBucketStoragerHDD or SingleInfoBucketStoragerRAM
*
******************************************************************************/
template <typename InfoIdentifier>
void SingleInfoBucketStoragerHDD_RAM<InfoIdentifier>::
storeInfoToBucket(InfoPtr infoPtr, InfoIdentifier infoID, int bucketIndex) 
			throw(InfoBucketStoragerException) {	
			
	//	Check that given bucket index is in the valid range, else throw exception
	InfoBucketStorager<InfoIdentifier>::throwBucketOutOfRangeE(bucketIndex, 
		numBucketsHDD + numBucketsRAM);
	//	Choose the base storager to operate
	if (bucketIndex < numBucketsHDD) {
		SingleInfoBucketStoragerHDD<InfoIdentifier>::storeInfoToBucket(infoPtr, infoID, bucketIndex);
	}
	else {
		bucketIndex -= numBucketsHDD;
		SingleInfoBucketStoragerRAM<InfoIdentifier>::storeInfoToBucket(infoPtr, infoID, bucketIndex);
	}
}



/******************************************************************************
*
*       Function        restoreInfoFromBucket
*       Parameters	infoID - identifier of the stored information
*								bucketIndex - index of the bucket for information restoring
*       Output           pointer to the required info
*       Description	Choose one of the base storagers due to the bucket index and operates it
*       Comments		InfoBucketStoragerException may be thrown from the method calls to
*								restoreInfoFromBucket of SingleInfoBucketStoragerHDD or SingleInfoBucketStoragerRAM
*
******************************************************************************/
template <typename InfoIdentifier> 
InfoPtr SingleInfoBucketStoragerHDD_RAM<InfoIdentifier>::
restoreInfoFromBucket(InfoIdentifier infoID, int bucketIndex)
			throw(InfoBucketStoragerException) {
			
	//	Check that given bucket index is in the valid range, else throw exception
	InfoBucketStorager<InfoIdentifier>::throwBucketOutOfRangeE(bucketIndex, 
		numBucketsHDD + numBucketsRAM);
	//	Choose the base storager to operate
	InfoPtr infoPtr = NULL;
	if (bucketIndex < numBucketsHDD) {
		infoPtr = SingleInfoBucketStoragerHDD<InfoIdentifier>::restoreInfoFromBucket(infoID, bucketIndex);
	}
	else {
		bucketIndex -= numBucketsHDD;
		infoPtr = SingleInfoBucketStoragerRAM<InfoIdentifier>::restoreInfoFromBucket(infoID, bucketIndex);
	}	
	return infoPtr;
}



#endif



