/* $Id: SingleInfoBucketStoragerRAM.h,v 1.2 2008/08/03 08:15:48 superlink Exp $ */
/************************************************************************************/
//	SingleInfoBucketStoragerRAM.h
//	Class SingleInfoBucketStoragerRAM<T>: Derieved from InfoBucketStorager<T>
//	Implements the interface defined by InfoBucketStorager<T> class (see description in InfoBucketStorager.h)
//	The class has 3 implementation specifications:
//	1. It allows maximum 1 info node in the bucket (trying to store the 2nd info into the same bucket leads
//	to the BucketFullException)
//	2. The infos are stored into the RAM 
//	3. The external C-style function pointer must be supplied in order to delete the InfoPtr from the RAM
//	(the client may supply NULL as the function pointer, then the infos won't be deleted)
/************************************************************************************/
#ifndef SINGLE_INFO_BUCKET_STORAGER_RAM_H
#define SINGLE_INFO_BUCKET_STORAGER_RAM_H

#include <vector> 
using namespace std;

#include "InfoBucketStorager.h"


/************************************************************************************/
//	Class definition:
template <typename InfoIdentifier> 
class SingleInfoBucketStoragerRAM: public virtual InfoBucketStorager<InfoIdentifier> {
	
	public:
		//	Constructor
		SingleInfoBucketStoragerRAM(size_t pNumBuckets, FuncDeleteInfoFromRAM func,
															int pSlotInfoSize) :
			numBuckets(pNumBuckets), buckets(pNumBuckets, NULL), funcDeleteInfoFromRAM(func),
			InfoBucketStorager<InfoIdentifier>(pSlotInfoSize) {};
			
		//	Pure virtual methods implementation:
		//	Method storeInfoToBucket
		virtual void storeInfoToBucket(InfoPtr infoPtr, InfoIdentifier infoID, int bucketIndex) 
			throw(InfoBucketStoragerException);

		//	Method restoreInfoFromBucket
		virtual InfoPtr restoreInfoFromBucket(InfoIdentifier infoID, int bucketIndex)
			throw(InfoBucketStoragerException) ;

		//	Virtual destructor
		virtual ~SingleInfoBucketStoragerRAM();


	private:
		int numBuckets;
		//	Internal structure for storing info identifier + info pointer
		class InfoNode {
			public:
				InfoIdentifier infoID;
				InfoPtr infoPtr;
				InfoNode(InfoIdentifier id, InfoPtr ptr) : infoID(id), infoPtr(ptr) {};
		};
		//	Vector of buckets (each bucket points to InfoNode)
		vector<InfoNode*> buckets;
		//	Pointer to c-style function for deleting the infos from RAM 
		FuncDeleteInfoFromRAM funcDeleteInfoFromRAM;
};





//---------------------------------------------------------------------------------------------------------------------------
//	Implementation of SingleInfoBucketStoragerRAM public methods:




/******************************************************************************
*
*       Function        storeInfoToBucket
*       Parameters	infoPtr - pointer to the stored information
*								infoID - identifier of the stored information
*								bucketIndex - index of the bucket for information storing
*       Output           none
*       Description	Store the info with the given identifier into the given bucket (by its pointer)
*       Comments		throws 3 types of exceptions:
*								1. BucketOutOfRange - bucket index is not in the range [0..numBuckets-1]
*								2. BucketFull - there is already some info in the specified bucket
*								3. InfoExist - the info with the given identifier already exists in the
*								given bucket
*								Exception 3 is a sub-case of Exception 2 (and is thrown instead)
*
******************************************************************************/
template <typename InfoIdentifier> 
void SingleInfoBucketStoragerRAM<InfoIdentifier>::
storeInfoToBucket(InfoPtr infoPtr, InfoIdentifier infoID, int bucketIndex) 
			throw(InfoBucketStoragerException) {						
	
	//	Check that given bucket index is in the valid range, else throw exception
	InfoBucketStorager<InfoIdentifier>::throwBucketOutOfRangeE(bucketIndex, numBuckets);
	//	Check that the specified bucket is full
	if (buckets[bucketIndex] == NULL) {
		//	Bucket is empty => store the given information
		buckets[bucketIndex] = new InfoNode(infoID, infoPtr);
	}
	else {
		//	Bucket is not empty: throw BucketFullException or InfoExistException(if the id is equal)
		if (buckets[bucketIndex]->infoID == infoID) {
			throw InfoBucketStoragerException(InfoBucketStoragerException::InfoExist);
		}
		else {
			throw InfoBucketStoragerException(InfoBucketStoragerException::BucketFull);
		}
	}			
}



/******************************************************************************
*
*       Function        restoreInfoFromBucket
*       Parameters	infoID - identifier of the stored information
*								bucketIndex - index of the bucket for information restoring
*       Output           pointer to the required info
*       Description	Restore the info with the given identifier from the given bucket, cleans the bucket cell
*       Comments		throws 2 types of exceptions:
*								1. BucketOutOfRange - bucket index is not in the range [0..numBuckets-1]
*								2. InfoNotExist - if the info with the specified id doesn't exist in the 
*								given bucket
*
******************************************************************************/
template <typename InfoIdentifier> 
InfoPtr SingleInfoBucketStoragerRAM<InfoIdentifier>::
restoreInfoFromBucket(InfoIdentifier infoID, int bucketIndex)
			throw(InfoBucketStoragerException) {
	
	//	Check that given bucket index is in the valid range, else throw exception
	InfoBucketStorager<InfoIdentifier>::throwBucketOutOfRangeE(bucketIndex, numBuckets);
	//	Check that info with the specified id exists in the specified bucket
	InfoNode* infoNodePtr = buckets[bucketIndex];
	if ((infoNodePtr == NULL) || (infoNodePtr->infoID != infoID)) {
		throw InfoBucketStoragerException(InfoBucketStoragerException::InfoNotExist);
	}
	//	Info exists => restore the information from the bucket and clean the bucket cell
	InfoPtr infoPtr = infoNodePtr->infoPtr;
	delete infoNodePtr;
	buckets[bucketIndex] = NULL;
	return infoPtr;
}



/******************************************************************************
*
*       Function        Destructor of SingleInfoBucketStoragerRAM<InfoIdentifier> object
*       Parameters	none
*       Output           none
*       Description	Deletes all the information pointers that are not NULL (using funcDeleteInfoFromRAM)
*       Comments		none 
*
******************************************************************************/
template <typename InfoIdentifier> 
SingleInfoBucketStoragerRAM<InfoIdentifier>::~SingleInfoBucketStoragerRAM() {
	//	Go through all info nodes and delete them
	for(int i = 0; i < numBuckets; i++) {
		if (buckets[i] != NULL) {
			//	Delete info  
			if ((buckets[i]->infoPtr != NULL) && (funcDeleteInfoFromRAM != NULL)) {
				funcDeleteInfoFromRAM(buckets[i]->infoPtr);
			}
			//	Delete info node 
			delete buckets[i];
		}
	}
}



#endif