/* $Id: MemoryManager.cpp,v 1.3 2008/08/25 07:52:17 sandanis Exp $ */
#include "MemoryManager.h"
#include "InfoBucketStoragers/SingleInfoBucketStoragerRAM.h" 

#include <math.h>
//------------------------------------------------------------------------------
//	Implementation of MemoryManager public methods:


/******************************************************************************
*
*       Function        Constructor of MemoryManager
*       Parameters	input - struct of type MemoryManagerCreateInput (see MemoryManagerInputOutputFormat.h)	
*       Output           none
*       Description	calculates the maximal possible window size (if it is less that 2 => 
*								throw NotEnoughMemoryException), initiates the vector of InfoBucketStoragers
*       Comments		none
*
******************************************************************************/
MemoryManager::MemoryManager(const MemoryManagerCreateInput& input) 
	throw(NotEnoughMemoryException, NullFuncPointerException) :
	numSlots(input.numSlots), firstSlotNumber(UNDEFINED_INDEX), firstSlotStoreIndex(UNDEFINED_INDEX), 
		numInfoTypes(input.numSlotsInfoTypes), availMemForWindowCalc(input.availMemForWindowCalc) {
		
	//	1. Calculate the required memory for storing 1 slot (sum over all info types) + check for null function pointers  
	double sizeStoredSlotInfo = getSizeStoredSlot(input.slotsInfoTypesArray);
	//	2. Calculate window size
	calcWindowSize(input.availMemForWindowCalc, input.sizeOfRunningSlotInfo, sizeStoredSlotInfo);
	if (windowSize < 2) {
		//	Not enough RAM => perform a new calculation of window size using external mem (and minimizing its usage) 
		calcWindowSizeUsingHDD(input.availMemForWindowCalc, input.sizeOfRunningSlotInfo, sizeStoredSlotInfo);
		//	Check that client permitted the usage of required HDD space (else throw exception)
		if (input.availExternalMemForWindowCalc < requiredExtMem) {
			throw NotEnoughMemoryException(requiredExtMem);
		}
	}
	//	3. Check if there is no need for MemoryManager
	if (windowSize >= numSlots) {
		numInfoTypes = 0;
		lastOperationResult = MemoryManagerNotNeeded;
	}
	else {
		//	4. Initiate all memory storagers
		initiateMemoryStoragers(input.slotsInfoTypesArray);
		lastOperationResult = Success;
	}
}


//	MemoryManager Debug Constructor:
//	state: 0 - only RAM; 1 - only HD; 2 - 0.5 RAM and 0.5 HD
MemoryManager::MemoryManager(const MemoryManagerCreateInput& input, int size, int state) 
throw(NotEnoughMemoryException, NullFuncPointerException):
	numSlots(input.numSlots), firstSlotNumber(UNDEFINED_INDEX), firstSlotStoreIndex(UNDEFINED_INDEX), 
		numInfoTypes(input.numSlotsInfoTypes) {
	
	//	1.	
	double sizeStoredSlotInfo = getSizeStoredSlot(input.slotsInfoTypesArray);
	//	2.
	windowSize = size;
	//	3.
	numSlotsToStore = (numSlots % (windowSize - 1) != 0) ? 
			(numSlots/(windowSize - 1) + 1) : (numSlots/(windowSize - 1));
	//	4.
	if (state == 0) {
		numSlotsToStoreRAM = numSlotsToStore;
		numSlotsToStoreExtMem = 0;
	}
	else if (state == 1) {
		numSlotsToStoreRAM = 0;
		numSlotsToStoreExtMem = numSlotsToStore;
	}
	else {
		numSlotsToStoreRAM = numSlotsToStore/2;
		numSlotsToStoreExtMem = numSlotsToStore - numSlotsToStoreRAM;
	}
	//	5.
	requiredExtMem = numSlotsToStoreExtMem*sizeStoredSlotInfo;
	//	6.
	initiateMemoryStoragers(input.slotsInfoTypesArray);			
}



/******************************************************************************
*
*       Function        getWindowSize
*       Parameters	none	
*       Output           returns window size of the window calculation
*       Description	none 
*       Comments		none
*
******************************************************************************/
int MemoryManager::getWindowSize() {
	return windowSize;
}


double MemoryManager::getAvailMemForWindowCalc() {
	return availMemForWindowCalc;
}


/******************************************************************************
*
*       Function        getLastOperationResult
*       Parameters	none	
*       Output           returns lastOperationResult field 
*       Description	none
*       Comments		none
*
******************************************************************************/
MemoryOperationResult MemoryManager::getLastOperationResult() {
	return lastOperationResult;
}



/******************************************************************************
*
*       Function        storeSlotInfo
*       Parameters	slotInfoPtr - pointer to the stored information, 
*								slotInfoType - type of the slot information, 
*								slotNumber - index of the slot in the range [0..(numSlots-1)], 
*       Output           none
*       Description	performs storing of the given info into the MemoryManager, 
*								due to the result of the performed operation updates "lastOperationResult"
*       Comments		none
*
******************************************************************************/
void MemoryManager::storeSlotInfo(InfoPtr slotInfoPtr, int slotInfoType, int slotNumber) {
	try {
		//	1. Check for errors: MemoryManagerNotNeeded, NullInfoPtr, NoSuchInfoType and SlotOutOfRange
		basicErrorsCheck(slotInfoType, slotNumber);
		if (slotInfoPtr == NULL) {
			throw NullInfoPtr;
		}
		//	2. Get bucket index and check for errors:  WindowCorruption
		int slotBucketIndex = getBucketIndexForSlot(slotNumber);
		//	3. Store the information and check for errors: MemoryManagerInternalError and SlotInfoExists
		performInfoStoring(slotInfoPtr, slotInfoType, slotNumber, slotBucketIndex);
		//	4. Initiate the firstSlotNumber and firstSlotStoreIndex if needed
		if (firstSlotNumber == UNDEFINED_INDEX) {
			firstSlotNumber = slotNumber;
			firstSlotStoreIndex = slotBucketIndex;
		}
		lastOperationResult = Success;
	}
	catch (MemoryOperationResult operationError) {
		//	Save the occured error in the lastOperationResult
		lastOperationResult = operationError;
	}
}



/******************************************************************************
*
*       Function        restoreSlotInfo
*       Parameters	slotInfoType - type of the slot information, 
*								slotNumber - index of the slot in the range [0..(numSlots-1)], 
*       Output           pointer to the info of slotInfoType with the specified slotNumber
*								or NULL if there is no such an info in MemoryManager
*       Description	performs restoring of info from the MemoryManager, 
*								due to the result of the performed operation updates "lastOperationResult"
*       Comments		none
*
******************************************************************************/
InfoPtr MemoryManager::restoreSlotInfo(int slotInfoType, int slotNumber) {
	try {
		//	1. Check for errors: MemoryManagerNotNeeded, NoSuchInfoType and SlotOutOfRange
		basicErrorsCheck(slotInfoType, slotNumber);
		//	2. Get bucket index and check for errors:  WindowCorruption
		int slotBucketIndex = getBucketIndexForSlot(slotNumber);
		//	3. Restore the information and check for errors: MemoryManagerInternalError and SlotInfoDoesntExist
		InfoPtr infoPtr = performInfoRestoring(slotInfoType, slotNumber, slotBucketIndex);
		lastOperationResult = Success;
		return infoPtr;
	}
	catch (MemoryOperationResult operationError) {
		//	Save the occured error in the lastOperationResult and return NULL
		lastOperationResult = operationError;
		return NULL;
	}
}




/******************************************************************************
*
*       Function        Destructor of MemoryManager
*       Parameters	none	
*       Output           none
*       Description	deletes storagers
*       Comments		none
*
******************************************************************************/
MemoryManager::~MemoryManager() {
	for(int i = 0; i < numInfoTypes; i++) {
		delete storagers[i];
	}
}




//------------------------------------------------------------------------------
//	Implementation of MemoryManager private methods:

/******************************************************************************
*
*       Function        getSizeStoredSlot
*       Parameters	slotsInfoTypesArray - array with sizes for all types of stored slots	
*       Output           returns sum of the above sizes
*       Description	none
*       Comments		may throw NullFuncPointerException if one of the client provided function pointers is null
*
******************************************************************************/
double MemoryManager::getSizeStoredSlot(SlotInfoType* slotsInfoTypesArray) throw(NullFuncPointerException) {
	double sizeOfStoredSlot = 0;
	for(int i = 0; i < numInfoTypes; i++) {
		//	Update sizeOfStoredSlot
		sizeOfStoredSlot += slotsInfoTypesArray[i].slotInfoSize;
		//	Check for null function pointes
		if (slotsInfoTypesArray[i].funcDeleteSlotInfoFromRAM == NULL || 
			slotsInfoTypesArray[i].funcReadInfoFromFile == NULL ||
			slotsInfoTypesArray[i].funcWriteInfoToFile == NULL) {
				throw NullFuncPointerException();
		}
	}
	return sizeOfStoredSlot;
}



/******************************************************************************
*
*       Function        calcWindowSize
*       Parameters	availMemForWindowCalc - available RAM for algorithm calculation 
*								sizeOfRunningSlotInfo - size for running slot
*								sizeStoredSlotInfo - sum of sizes for all types of stored slots
*       Output           none
*       Description	calculates window size
*       Comments		none
*
******************************************************************************/
void MemoryManager::calcWindowSize(double availMemForWindowCalc, 
								  double sizeOfRunningSlotInfo, double sizeStoredSlotInfo) {
	double size = 0;
	//	Help variables
	double x1 = availMemForWindowCalc/sizeOfRunningSlotInfo;		// MEM/sizeB 
	double x2 = availMemForWindowCalc/numSlots;							// MEM/m
	double x3 = sizeOfRunningSlotInfo/numSlots;								// sizeB/m
	//	Check that there is a solution for the given constrains
	double v1 = 4 / (1 + x1) * (sizeStoredSlotInfo + x2) / (x2 + x3);
	if (1 - v1 >= 0) {
		//	There is a solution => complete the calculation
		double v2 = 0.5 + 0.5 * x1;
		size = v2 * (1 + sqrt(1 - v1));
	}
	windowSize = (int)size;
	if (windowSize >= 2) {
		//	Calculate the numSlotsToStore, numSlotsToStoreRAM, numSlotsToStoreExtMem, requiredExtMem
		calcStorageParams(1, sizeStoredSlotInfo);					  
	}
}



/******************************************************************************
*
*       Function        calcWindowSizeUsingHDD
*       Parameters	availMemForWindowCalc - available RAM for algorithm calculation 
*								sizeOfRunningSlotInfo - size for running slot
*								sizeStoredSlotInfo - sum of sizes for all types of stored slots
*       Output           none
*       Description	minimizes the function y(x) = (x*sizeB - MEM)/sizeS + m/(x - 1)
*								by taking it's first derivation, comparing it to zero
*								and then taking the minimum between the received value and 2 boundaries:
*								the range of x (= windowSize) is between 2 and MEM/sizeB
*       Comments		none
*
******************************************************************************/
void MemoryManager::calcWindowSizeUsingHDD(double availMemForWindowCalc, 
										   double sizeRunningSlotInfo, double sizeStoredSlotInfo) {

	double xMin, yMin;
	//	1. Preprocessing : get the minimum between the boundaries
	double l1 = 2;
	double yl1 = (l1*sizeRunningSlotInfo - availMemForWindowCalc)/sizeStoredSlotInfo + numSlots/(l1 - 1);
	double l2 = availMemForWindowCalc/sizeRunningSlotInfo;		//	MEM/sizeB
	double yl2 = numSlots/(l2 - 1);
	//	Choose the minimum between yl1 and yl2
	if ((yl1 > numSlots/(l1 - 1)) || (yl2 < yl1)) {
		//	(yl1 is out of range) or (yl2 < yl1) => choose yl2
		xMin = l2;
		yMin = yl2;
	}
	else {
		//	(yl1 is in the range) and (yl1 =< yl2) => choose yl1
		xMin = l1;
		yMin = yl1;
	}
	//	2. Calculation of the optimal size in order to minimize the required external memory space 
	double size = 1 + sqrt(numSlots * (sizeStoredSlotInfo/sizeRunningSlotInfo));			//	1 + sqrt(m*sizeS/sizeB)
	if ((size >= l1) && (size <= l2)) {
		//	size is in the permitted range => check that it is the minimum
		double ySize = (size*sizeRunningSlotInfo - availMemForWindowCalc)/sizeStoredSlotInfo + numSlots/(size - 1);
		if (ySize < yMin) {
			yMin = ySize;
			xMin = size;
		}
	}
	//	3. Find percent of RAM storing and window size
	double percentStoringRAM = 1 - yMin * ((xMin - 1)/(numSlots - xMin + 1));
	windowSize = (int)xMin;
	//	4. Calculate the numSlotsToStore, numSlotsToStoreRAM, numSlotsToStoreExtMem, requiredExtMem
	calcStorageParams(percentStoringRAM, sizeStoredSlotInfo);
}



/******************************************************************************
*
*       Function        calcStorageParams
*       Parameters	percentStoringRAM - the relative part of slots to be stored in the RAM
*								sizeStoredSlotInfo - sum of sizes for all types of stored slots
*       Output           none
*       Description	using the given parameters, updates the 
*								numSlotsToStore, numSlotsToStoreRAM, numSlotsToStoreExtMem, requiredExtMem
*								class fields
*       Comments		windowSize field is required
*
******************************************************************************/
void MemoryManager::calcStorageParams(double percentStoringRAM, double sizeStoredSlotInfo) {
	
	//	1. Check if window size is too big => no need to use MemoryManager 
	if (windowSize >= numSlots) {
		requiredExtMem = numSlotsToStore = numSlotsToStoreRAM = numSlotsToStoreExtMem = 0;
	}
	else {
		//	2. Calculate the overall number of slots to store in the MemoryManager
		numSlotsToStore = (numSlots % (windowSize - 1) != 0) ? 
			(numSlots/(windowSize - 1) + 1) : (numSlots/(windowSize - 1));
		//	3. Check if percentStoringRAM is close to 1 => all slots are stored to RAM
		if (percentStoringRAM > 0.999999) {
			numSlotsToStoreRAM = numSlotsToStore;
			requiredExtMem = numSlotsToStoreExtMem = 0;
		}
		//	4. Check if percentStoringRAM is too small => no storage to RAM at all, only to external memory
		else if (percentStoringRAM < 0.1) {
			numSlotsToStoreExtMem = numSlotsToStore;
			numSlotsToStoreRAM = 0;
			requiredExtMem = sizeStoredSlotInfo * (numSlots / (windowSize - 1));
		}
		//	5. Both storage to RAM and to external memory simultaneously
		else {
			numSlotsToStoreRAM = (percentStoringRAM * numSlots) / (windowSize - 1);
			numSlotsToStoreExtMem = numSlotsToStore - numSlotsToStoreRAM;
			requiredExtMem =numSlotsToStoreExtMem * sizeStoredSlotInfo;
		}
	}
}



/******************************************************************************
*
*       Function        initiateMemoryStoragers
*       Parameters	slotsInfoTypesArray - array with function pointers for treating info slots
*								storagerToClone - pointer to the storager to be cloned
*       Output           none
*       Description	allocates and initiates "storagers" vector
*       Comments		none
*
******************************************************************************/
void MemoryManager::initiateMemoryStoragers(SlotInfoType* slotsInfoTypesArray) {
	
	//	Initiate empty "storagers" vector
	InfoBucketStorager<int>* tmp = NULL;
	vector<InfoBucketStorager<int>*> storagersTmp(numInfoTypes, tmp);
	storagers = storagersTmp;
	//	For each info type initiate it's storager
	for(int i = 0; i < numInfoTypes; i++) {
		storagers[i] = getStoragerForInfoType(slotsInfoTypesArray + i, i);
	}
}


/******************************************************************************
*
*       Function        getStoragerForInfoType
*       Parameters	infoTypePtr - pointer to the struct with parameters for constructor of the storager
*								to be initiated
*								storagerIndex - index of the storager to be initiated 
*       Output           pointer to the base class object (InfoBucketStorager)
*       Description	constructs the storager whoose type is chosen dinamically according to
*								the storage parameters of MemoryManager (numSlotsToStoreExtMem and numSlotsToStoreRAM)
*       Comments		none
*
******************************************************************************/
InfoBucketStorager<int>* MemoryManager::
getStoragerForInfoType(SlotInfoType* infoTypePtr, int storagerIndex) {
	
	if (numSlotsToStoreExtMem == 0) {
		//	All the storage to RAM:
		return new SingleInfoBucketStoragerRAM<int>
			(numSlotsToStore, infoTypePtr->funcDeleteSlotInfoFromRAM, infoTypePtr->slotInfoSize);
	}
	else if (numSlotsToStoreRAM == 0) {
		//	All the storage to HDD:
		char storagerIndexStr[10];
		sprintf(storagerIndexStr, "%d", storagerIndex);
		return new SingleInfoBucketStoragerHDD<int>
			(STORAGER_STRING + storagerIndexStr, numSlotsToStore, infoTypePtr->funcDeleteSlotInfoFromRAM, 
				infoTypePtr->funcWriteInfoToFile, infoTypePtr->funcReadInfoFromFile, infoTypePtr->slotInfoSize);
	}
	else {
		//	Storage to HDD (1st) + RAM (2nd)
		char storagerIndexStr[10];
		sprintf(storagerIndexStr, "%d", storagerIndex);
		return new SingleInfoBucketStoragerHDD_RAM<int>
			(numSlotsToStoreExtMem, numSlotsToStoreRAM, STORAGER_STRING + storagerIndexStr, 
				infoTypePtr->funcDeleteSlotInfoFromRAM, infoTypePtr->funcWriteInfoToFile, 
				infoTypePtr->funcReadInfoFromFile, infoTypePtr->slotInfoSize);
	}
}



/******************************************************************************
*
*       Function        basicErrorsCheck
*       Parameters	slotInfoType, slotNumber
*       Output           none
*       Description	checks input parameters for the following errors
*       Comments		throws MemoryManagerNotNeeded, NoSuchInfoType or SlotOutOfRange exceptions
*
******************************************************************************/
void MemoryManager::basicErrorsCheck(int slotInfoType, int slotNumber) throw (MemoryOperationResult){
	//	Check for the MemoryManagerNotNeeded
	if (lastOperationResult == MemoryManagerNotNeeded) {
		throw MemoryManagerNotNeeded;
	}
	//	Check for NoSuchInfoType
	else if ((slotInfoType < 0) || (slotInfoType >= numInfoTypes)) {
		throw NoSuchInfoType;
	}
	//	Check for SlotOutOfRange
	else if ((slotNumber < 0) || (slotNumber >= numSlots)) {
		throw SlotOutOfRange;
	}
}



/******************************************************************************
*
*       Function        getBucketIndexForSlot
*       Parameters	slotNumber - number of slot (integer) 
*       Output           index for storing in bucket storager
*       Description	performs calculation of bucket index (mapping of slotNumber into numSlotsToStore range): 
*								1. using the distance between slotNumber and firstSlotNumber if firstSlotNumber 
*								was already initiated;
*								2. using only slotNumber if firstSlotNumber was not initiated yet
*       Comments		assumes that slotNumber is in the range [0..(numSlots-1)]
*
******************************************************************************/
int MemoryManager::getBucketIndexForSlot(int slotNumber) throw (MemoryOperationResult) {
	int index;
	//	Check if firstSlotNumber and firstSlotStoreIndex were initiated
	if (firstSlotNumber != UNDEFINED_INDEX) {
		//	Check that the given slotNumber doesn't corrupt the existing window frame
		if ( (firstSlotNumber - slotNumber) % (windowSize - 1) == 0) {
			//	Calculate index using firstSlotStoreIndex, firstSlotNumber
			index = firstSlotStoreIndex - (firstSlotNumber - slotNumber) / (windowSize - 1);
		}
		else {
			//	Given slotNumber corrupts the existing window frame => throw exception
			throw WindowCorruption;
		}
	}
	else {
		//	firstSlotNumber was not initiated yet, so don't use it
		index = slotNumber / (windowSize - 1);
	}
	return index;
}



/******************************************************************************
*
*       Function        performInfoStoring
*       Parameters	infoPtr - pointer to the stored information, 
*								slotInfoType - index in the range [0..(numInfoTypes-1)], 
*								slotNumber - index in the range [0..(numSlots-1)], 
*								bucketIndex - index in the range [0..(numSlotsToStore-1)]
*       Output           none
*       Description	invokes the storager of slotInfoType to store the desired info, 
*								catches possible storager's exceptions
*       Comments		throws MemoryManagerInternalError and SlotInfoExists exceptions
*
******************************************************************************/
void MemoryManager::performInfoStoring(InfoPtr infoPtr, int slotInfoType, int slotNumber, int bucketIndex)
		throw (MemoryOperationResult) {
	try {
		//	Invoking the storager
		storagers[slotInfoType]->storeInfoToBucket(infoPtr, slotNumber, bucketIndex);
	}
	catch (InfoBucketStoragerException e) {
		MemoryOperationResult error = translateStoragerToManagerException(e.value);
		throw error;
	}
}



/******************************************************************************
*
*       Function        performInfoRestoring
*       Parameters	slotInfoType - index in the range [0..(numInfoTypes-1)], 
*								slotNumber - index in the range [0..(numSlots-1)], 
*								bucketIndex - index in the range [0..(numSlotsToStore-1)]
*       Output           pointer to the desired information
*       Description	invokes the storager of slotInfoType to restore the desired info, 
*								catches possible storager's exceptions
*       Comments		throws MemoryManagerInternalError and SlotInfoDoesntExist exceptions
*
******************************************************************************/
InfoPtr MemoryManager::performInfoRestoring(int slotInfoType, int slotNumber, int bucketIndex) 
		throw (MemoryOperationResult) {
	try {
		//	Invoking the storager
		InfoPtr infoPtr = storagers[slotInfoType]->restoreInfoFromBucket(slotNumber, bucketIndex);
		return infoPtr;
	}
	catch (InfoBucketStoragerException e) {
		MemoryOperationResult error = translateStoragerToManagerException(e.value);
		throw error;
	}
}



/******************************************************************************
*
*       Function        translateStoragerToManagerException
*       Parameters	eValue - value for exception of type InfoBucketStoragerException
*       Output           translated value of type MemoryOperationResult 
*       Description	for each specific value of InfoBucketStoragerException::ExceptionValue 
*								returns specific exception of type MemoryOperationResult (using switch)
*       Comments		none
*
******************************************************************************/
MemoryOperationResult 
MemoryManager::translateStoragerToManagerException(InfoBucketStoragerException::ExceptionValue eValue) {

	MemoryOperationResult res;
	switch (eValue) {
		case InfoBucketStoragerException::BucketOutOfRange :	res = MemoryManagerInternalError;
																											break;	
		case InfoBucketStoragerException::BucketFull :					res = SlotInfoExists;
																											break;			
		case InfoBucketStoragerException::InfoExist :						res = SlotInfoExists;
																											break;	
		case InfoBucketStoragerException::InfoNotExist :				res = SlotInfoDoesntExist;
																											break;	
		case InfoBucketStoragerException::InfoCorrupted :			res = MemoryManagerInternalError;
																											break;	
		default :																						res = MemoryManagerInternalError;		
																											break;	
	}
	return res;
}



//------------------------------------------------------------------------------
//	Implementation of wrapper functions:

int getWindowSize(MemoryManager* manager) {
	return manager->getWindowSize();
}

double getAvailMemForWindowCalc(MemoryManager* manager) {
	return manager->getAvailMemForWindowCalc();
}

MemoryOperationResult getLastOperationResult(MemoryManager* manager) {
	return manager->getLastOperationResult();
}

void storeSlotInfo(MemoryManager* manager, InfoPtr ptr, int slotInfoType, int slotNumber) {
	manager->storeSlotInfo(ptr, slotInfoType,	slotNumber);
}

InfoPtr restoreSlotInfo(MemoryManager* manager, int slotInfoType, int slotNumber) {
	return manager->restoreSlotInfo(slotInfoType, slotNumber);
}



