/* $Id: MemoryManager.h,v 1.3 2008/08/25 07:52:17 sandanis Exp $ */
/************************************************************************************/
//	MemoryManager.h
//	Class MemoryManager (MM) definition and wrapper C-functions declaration:
//	MM implements the storage of information (for each slot, it may be some probabilities array or value)
//	during the window calculation of some iterative algorithm (iteration on some slots) called "client".
//	MM manages storing and restoring information during the client calculation in intermediate slots.
//	When the client restores information, MM deletes it from its internal structures.
//	When MM is deleted, it deletes all stored information that was not restored by the client.
//	In order to delete the information (its type is unknown for MM), MM uses function supplied by the client.
//	In current version client is Superlink => is written in C code, and it supplies pointer to C-style function.
//	IMPORTANT: if the supplied pointer is NULL => MM doesn't delete anything when it is deleted itself!!!
//	If the client tries to restore information which doesn't exist in MM information pool, it returns NULL.
//	Storing/restoring is performed using slot identifiers that are integer positive numbers (int).
//	The range of identifiers is specified in the construction of MM (MemoryManagerCreateInput.numSlots).
//	In the following 3 cases the slot information cannot be stored in MM:
//	1. Slot identifier is out of range
//	2. The difference between the first stored slot identifier and the currently stored slot identifier cannot be
//	divided by "windowSize - 1" without carry out, that means that the client tries to store slot not due to 
//	the window frame = not on the boundary between 2 adjacent windows, its forbidden by MM !!!
//	3. Slot identifier already exists in MM pool
//	The window frame (case 2) is freezed when the first slot is succeessfully stored in MM.
//	MM is able to store different types of slot information.
//	After each storing/restoring operation the status of the performed operation (MemoryOperationResult)
//	can be retrieved by the getLastOperationResult() invocation
/************************************************************************************/
#ifndef MEMORY_MANAGER_H
#define MEMORY_MANAGER_H

#include "MemoryManagerInputOutputFormat.h"


//----------------------------------------------------------------------------------------------------
//	Class declaration:
#ifdef __cplusplus	
	#include <vector>
	#include <string>
	#include <iostream> 
	using namespace std;
	#include "InfoBucketStoragers/SingleInfoBucketStoragerHDD_RAM.h" 

	const string STORAGER_STRING = "_Superlink_temp_hmm_storager";

	class MemoryManager {
		
		public:
			//	Exception thrown from the constructor: if one of the pointers to the client functions is null
			class NullFuncPointerException{};
			//	Exception thrown from the constructor: if there is not enough RAM memory for calculation
			class NotEnoughMemoryException {
				public:
					const double requiredExtMemory;
					NotEnoughMemoryException(double mem) : requiredExtMemory(mem) {};
			};
			//	MemoryManager Constructor
			MemoryManager(const MemoryManagerCreateInput& input) 
				throw(NotEnoughMemoryException, NullFuncPointerException);

			//	MemoryManager DEBUG Constructor
			MemoryManager(const MemoryManagerCreateInput& input, int size, int state) 
				throw(NotEnoughMemoryException, NullFuncPointerException);

			//	Status retrieving methods (don't affect lastOperationResult):
			int getWindowSize();
			double getAvailMemForWindowCalc();
			MemoryOperationResult getLastOperationResult();
			
			//	Store / Restore memory operations (affect lastOperationResult): 
			void storeSlotInfo(InfoPtr slotInfoPtr, int slotInfoType, int slotNumber);
			InfoPtr restoreSlotInfo(int slotInfoType, int slotNumber);

			inline double getRequiredExtMem() {return requiredExtMem; };

			//	Destructor
			~MemoryManager();


		private:	
			//	Constants:
			static const int UNDEFINED_INDEX = -1;		//	class constant for identifying non-existing index
			
			//	Fields:
			int numSlots;								//	overall number of slots in the whole calculation
			int windowSize;							//	size of the window for window calculation
			int numSlotsToStore;				//	maximal number of slots to be stored in MemoryManager
			int numSlotsToStoreRAM;		//	the number of slots to be stored in RAM
			int numSlotsToStoreExtMem;	//	the number of slots to be stored in the external memory
			double requiredExtMem;			//	the size of the required external memory in order to store 
																//	numSlotsToStoreExtMem slots
			int firstSlotNumber;					//	for the first stored slot info: index in the range [0..numSlots-1]
			int firstSlotStoreIndex;				//	for the first stored slot info: index in the range [0..numSlotsToStore-1]
			int numInfoTypes;						//	number of slots info types 
			vector<InfoBucketStorager<int>*> storagers;							//	vector of slot info storagers
			MemoryOperationResult lastOperationResult;		//	updated after each storing/restoring operation
			double availMemForWindowCalc;		//	available RAM memory for window calculation

			//Private methods declaration:
			double getSizeStoredSlot(SlotInfoType* slotsInfoTypesArray) throw(NullFuncPointerException);
			
			void calcWindowSize(double availMemForWindowCalc, 
				double sizeOfRunningSlotInfo, double sizeStoredSlotInfo);
			void calcWindowSizeUsingHDD(double availMemForWindowCalc, 
				double sizeOfRunningSlotInfo, double sizeStoredSlotInfo);
			void calcStorageParams(double percentStoringRAM, double sizeStoredSlotInfo);

			void initiateMemoryStoragers(SlotInfoType* slotsInfoTypesArray);
			InfoBucketStorager<int>* getStoragerForInfoType(SlotInfoType* infoTypePtr, int storagerIndex);
			void basicErrorsCheck(int slotInfoType, int slotNumber) throw (MemoryOperationResult);
			int getBucketIndexForSlot(int slotNumber) throw (MemoryOperationResult);

			MemoryOperationResult 
				translateStoragerToManagerException(InfoBucketStoragerException::ExceptionValue);
			void performInfoStoring(InfoPtr infoPtr, int slotInfoType, int slotNumber, int bucketIndex) 
				throw (MemoryOperationResult);
			InfoPtr performInfoRestoring(int slotInfoType, int slotNumber, int bucketIndex) 
				throw (MemoryOperationResult);
	};
#else	
	typedef struct MemoryManager MemoryManager;
#endif


//----------------------------------------------------------------------------------------------------
//	Wrapper function declarations: interface for using MemoryManager from C code
#ifdef __cplusplus
extern "C" {
#endif	
		
		extern double getAvailMemForWindowCalc(MemoryManager*);
		extern int getWindowSize(MemoryManager*);	
		extern MemoryOperationResult getLastOperationResult(MemoryManager*);
		extern void storeSlotInfo(MemoryManager*, InfoPtr, int, int);
		extern	InfoPtr restoreSlotInfo(MemoryManager*, int, int);
		
#ifdef __cplusplus
}
#endif
 

#endif	//MEMORY_MANAGER_H