/* $Id: NPLstudents.h,v 1.2 2008/08/03 08:15:47 superlink Exp $ */
/****************************************************************************/
/*  NPLstudents.h										   					*/
/*  Finds NPL score for each testPoint									 	*/
/*  Author: Andrei Anisenia													*/
/****************************************************************************/
#ifndef NPL_STUDENTS_H
#define NPL_STUDENTS_H

#include "gdata.h"



/****************************************************************************/
/*  Data structures which are used											*/
/****************************************************************************/

/* number of founders in the pedigree										*/
int f;

/* number of nonfounders in the pedigree									*/
int n;

/* For each human in pedigree we define his personal information in ped_node*/
typedef struct ped_node_t
{
	struct ped_node_t* next_af;	/* only for affected individuals:
								   pointer to the next affected individual
								   in the pedigree							*/
	int mother;					/* mother index	(starting 0, in founder -1)	*/
	int father;					/* father index	(starting 0, in founder -1)	*/
	Bool af;					/* false - not affected; true - affected	*/
	int allele1;				/* allele1 - from father; allele2 - from mother;		*/
	int allele2;				/* two of 2*f alleles {0,1,...,2*f-1}		*/
	int* existingAllele;		/* one of "allele1" or "allele2"			*/
	Bool usefull;				/* shows if the person is usefull in the computation of NPL		
								   (meaning he is affected or has affected children)	*/
	int PedIndex;				/* index of the person in the main pedigree starting 0	*/
}ped_Node;


/*  the array of pointers to ped_nodes for all the individuals				*/
ped_Node** ped;

/*  the array of pointers to ped_nodes for all affected individuals			*/
ped_Node** affected;

/*  the number of affected individuals (=size of "affected" array)			*/
int af_num;
int af_2allelesNum;

/*  the array of size "af_2allelesNum", for each affected individual shows
	which allele is chosen (in the computation of Spairs score): 
	allele1 or allele2 (and so the value in the appropriate cell is 1 or 2)	*/
int* aff;

/*  the array of size 2*f, for each founder allele (from 2*f founder alleles)
	shows the number of times that this allele appears in some collection	*/
int* f_alleles;
int numRealFAlleles;

/*  definition of inheritance vector (size 2*n-f) with all additional 
	necessery information which is used in the computation of NPL score		*/
typedef struct vector_node
{
	int inheritance;			/*	grandpaternal(0) or grandmaternal(1)	*/
								/*  for the current allele					*/
	Bool usefull;				/*	like in ped_node						*/
	int person;					/*	index of the person in "ped"			*/
	int parent;					/*	current allele inherited from father(0) */ 
								/*	or mother(1)							*/
}vector_node;
/*	Ivector is vector of "Const" + "Regular" bits							*/
vector_node* Ivector;
int sizeOfIvector;

/*	vector with size = "PedigreeData.SpeedUpNumOfMeioses"					*/
/*	this is vector of indexes of "Regular" bits in "Ivector" 				*/
int* mappingVector;

/*  for each inheritance vector v we'll calculate 2 scores:					*/
/*  Spairs(v) and Sall(v)													*/
/*  the functions used: getSpairs() and getSall()							*/
double Spairs;
double Sall;
/*  is used in the computing of Sall score									*/
double SallTmp;

/*	for all 2^(2*n - f) vectors these scores will be stored in 2 arrays:	*/
/*	SpairsArray and SallArray												*/
double* SpairsArray;
double* SallArray; 

/*	pointer to the above arrays indicating the next cell to fill in			*/
int next;

/*	help array for computing the "next", used by function getNext(int i)	*/
/*	of size: 2*n - f														*/
int* powersOf2;


/*  numCollections = 2^af_2allelesNum												*/
double numCollections;


/*  Used functions															*/

/****************************************************************************
*
*       Function		InitFounderAlleles  
*       Parameters	  	none	
*       Output			none    
*       Description		parses the alleles {0,...,2*f-1} between 
*						all the founders in ped, allocates f_alleles
*       Comments		none
*
*****************************************************************************/
void InitFounderAlleles();


/****************************************************************************
*
*       Function		InitAffected  
*       Parameters	  	none	
*       Output			none    
*       Description		creates "affected" array and links it to 
*						affected individuals in ped,
*						allocates "aff" array  
*       Comments		none
*
*****************************************************************************/
void InitAffected();


/****************************************************************************
*
*       Function		InitUsefull  
*       Parameters	  	none	
*       Output			none    
*       Description		for each person in ped: if he is affected 
*						or has affected children then 
*						"usefull" is set to true, else false 
*       Comments		none
*
*****************************************************************************/
void InitUsefull();


/****************************************************************************
*
*       Function		RecursiveUsefullTrue  
*       Parameters	  	ind - index of person in ped	
*       Output			none    
*       Description		recursive function, set "usefull" of the person "ind"
*						and all of his relatives upstream to true
*       Comments		help function of InitUsefull
*
*****************************************************************************/
void RecursiveUsefullTrue(int ind);


/****************************************************************************
*
*       Function		InitVector  
*       Parameters	  	none	
*       Output			none    
*       Description		allocates and initiates "Ivector", 
*						parses 2*f founder alleles between n nonfounders
*						according to initial Ivector "inheritance"(zero vector)
*       Comments		should be called after calling InitUsefull
*
*****************************************************************************/
void InitVector();


/****************************************************************************
*
*       Function		RearrangeAlleles  
*       Parameters	  	vect_ind : index of person's allele in Ivector
*						from which we should compute arrangement of alleles	
*						downstream in Ivector for each person
*       Output			none    
*       Description		none
*       Comments		help function of InitVector (but can be used anywhere)
*
*****************************************************************************/
void RearrangeAlleles(int vect_ind);


/****************************************************************************
*
*       Function		InitPed
*       Parameters	  	none	
*       Output			none    
*       Description		intiates in each node in "ped" the next fields:
*						"af", "mother" and "father" according to the 
*						information in Pedigree
*       Comments		none
*
*****************************************************************************/
void InitPedigree();


/****************************************************************************
*
*       Function		Destroy  
*       Parameters	  	none
*       Output			none    
*       Description		frees all the allocated memory
*       Comments		none
*
*****************************************************************************/
void Destroy();


/****************************************************************************
*
*       Function		getSpairs()
*       Parameters	  	none
*       Output			none    
*       Description		computes Spairs score for the current Ivector
*						saves it in "Spairs" variable and in "SpairsArray"
*						at "next" cell
*       Comments		sets to zero "f_alleles" before computing score
*
*****************************************************************************/
void getSpairs();


/****************************************************************************
*
*       Function		getNewScores  
*       Parameters	  	i - index in mappingVector
*       Output			none    
*       Description		after changing "inheritance" at index "i" 
*						computes Spairs and Sall scores (by using
*						functions getSpairs and getSall)
*       Comments		none
*
*****************************************************************************/
void getNewScores(int index);


/****************************************************************************
*
*       Function		computeScores  
*       Parameters	  	none
*       Output			none    
*       Description		computes Spairs and Sall scores 
*						for all inheritance vectors and saves them
*						in SpairsArray and SallArray
*       Comments		none
*
*****************************************************************************/
void computeScores();


/****************************************************************************
*
*       Function		getSall()
*       Parameters	  	none
*       Output			none    
*       Description		computes Sall score for the current Ivector
*						saves it in "Sall" variable and in "SallArray"
*						at "next" cell
*       Comments		sets to zero "f_alleles" and to 1 "aff"
*						before computing score	
*					
*****************************************************************************/
void getSall();


/****************************************************************************
*
*       Function		getSallTmp
*       Parameters	  	i - index in "aff"
*       Output			none    
*       Description		recomputes the value of SallTmp by using it's
*						previous value and the index of the affected person
*						for whom allele's collection was changed
*       Comments		none
*
*****************************************************************************/
void getSallTmp(int i);


/****************************************************************************
*
*       Function		normalizeScores
*       Parameters	  	none
*       Output			none    
*       Description		brings Sall and Spairs scores to standart normal form
*       Comments		none
*
*****************************************************************************/
void normalizeScores();


/****************************************************************************
*
*       Function		normalizeS
*       Parameters	  	stype - type of the score (Sall or Spairs) 
*						to be normalized
*       Output			none  
*       Description		help function of normalizeScores
*       Comments		none
*
*****************************************************************************/
void normalizeS(SCORE stype);


/****************************************************************************
*
*       Function		getRealFounders
*       Parameters	  	none	
*       Output			array of indexes of founders in PedigreeGraph   
*       Description		these are indexes of founders which have children
*						(not childless founders)
*       Comments		none
*
*****************************************************************************/
int* getRealFounders();

/* the main function, for external use										*/
/******************************************************************************
*
*       Function        computeNPL
*       Parameters		none
*       Output          none
*       Description		compute NPL for all points 
*       Comments		is used in "gh_main.c"
*
*****************************************************************************/
void computeNPL();

#endif














































