#include "WindowInputProvider.h"

#include <iostream>
#include <fstream>
using namespace std;

#include <string.h>


//------------------------------------------------------------------------------
//	Implementation of MemoryManager public methods:

//	Constructor
WindowInputProvider::WindowInputProvider(char *datafileName, char *pedfileName) {
	ifstream datafile(datafileName);
	ifstream pedfile(pedfileName);
	readDatafile(datafile);
	readPedfile(pedfile);
	datafile.close();
	pedfile.close();
	startMarker = 0;
	stopMarker = numLoci - 1;
};



//	Set window for which input data will be provided
//	start - the number of the 1st locus in the window (count starting 0)
//	size - the size of the window
void WindowInputProvider::setInputWindow(int start, int size) {
	startMarker = start;
	stopMarker = startMarker + size - 1;
};



int* WindowInputProvider::getLociOrder() {
	int currNumLoci = stopMarker - startMarker + 1;
	int* lociOrder = new int[currNumLoci + 1];
	for(int i = 0; i <= currNumLoci; ++i) {
		lociOrder[i] = i;
	}
	return lociOrder;
};



//	locusIndex - starting 0, not includes disease locus
void WindowInputProvider::getLocusName(int locusIndex, char* name) {
	strcpy(name, markerNames[startMarker + locusIndex].c_str());
};	



int WindowInputProvider::getLocusNumFreqs(int locusIndex) {
	return numAllelesForLoci[startMarker + locusIndex];	
};



double* WindowInputProvider::getLocusFreqs(int locusIndex) {
	return frequencies[startMarker + locusIndex];
};



//	the size of returned array is 3 for non-sex-linked and 5 for sex-linked (female, then male)
//	liabClass - starting 0
double* WindowInputProvider::getPensForLiabClass(int liabClass) {
	return penetrances[liabClass];
};	



double* WindowInputProvider::getRecombValues() {
	return (recombValues + startMarker);
};



void WindowInputProvider::getProgSpecData(char* specData) {
	int spaceCounter = 0;
	int i = 0;
	while (spaceCounter < 3) {
		char c = progSpecData[i];
		if ((c == ' ') || (c == '\t')) {
			++spaceCounter;
		}
		++i;
	}
	string progSpecData1 = progSpecData.substr(0, i);
	while((progSpecData[i] >= 48) && (progSpecData[i] <= 57)) {
		++i;
	}
	string progSpecData2 = progSpecData.substr(i, progSpecData.size() - i);
	sprintf(specData, "%s%d%s", progSpecData1.c_str(), getNumLoci(), progSpecData2.c_str());
};



//	Destructor
WindowInputProvider::~WindowInputProvider() {
	//	Free penetrances
	for(int i = 0; i < numLiabClasses; ++i) {
		delete[] penetrances[i];
	}
	delete[] penetrances;
	//	Free frequencies and marker names
	for(int i = 0; i < numLoci; ++i) {
		delete[] frequencies[i];
	}
	delete[] frequencies;
	delete[] markerNames;
	//	Free recombination values
	delete[] recombValues;
	//	Free all arrays in "personsData" list
	while (personsData.size() > 0) {
		int* arr = personsData.back();
		delete[] arr;
		personsData.pop_back();
	}
};






//------------------------------------------------------------------------------
//	Implementation of MemoryManager private methods:


void WindowInputProvider::readDatafile(ifstream& datafile) {
	int trash;
	datafile>>numLoci;
	--numLoci;
	datafile>>trash>>sexLinked>>programCode;
	string line;
	getline(datafile, line);	//	the rest of the first line
	getline(datafile, line);	//	2nd line (mut locus...)
	getline(datafile, line);	//	3rd line (numbers of loci)
	getline(datafile, line);	//	4th line (1 2...)
	datafile>>disFrequency;
	getline(datafile, line);
	datafile>>numLiabClasses;
	penetrances = new double*[numLiabClasses];
	getline(datafile, line);
	//	Read all penetrances
	for(int i = 0; i < numLiabClasses; ++i) {
		if (sexLinked == 1) {
			penetrances[i] = new double[5];
		}
		else {
			penetrances[i] = new double[3];
		}
		datafile>>penetrances[i][0]>>penetrances[i][1]>>penetrances[i][2];
		getline(datafile, line);
		if (sexLinked == 1) {
			datafile>>penetrances[i][3]>>penetrances[i][4];
			getline(datafile, line);
		}
	}
	//	Read all loci allele frequencies and marker names
	markerNames = new string[numLoci];
	frequencies = new double*[numLoci];
	numAllelesForLoci = new int[numLoci];
	for(int i = 0; i < numLoci; ++i) {
		datafile>>trash>>numAllelesForLoci[i];
		getMarkerName(datafile, markerNames[i], i);
		frequencies[i] = new double[numAllelesForLoci[i]];
		for(int j = 0; j < numAllelesForLoci[i]; ++j) {
			datafile>>frequencies[i][j];
		}
		getline(datafile, line);
	}
	//	Read recombination values
	getline(datafile, line);	//	SEX DIFFERENCE...
	datafile>>trash;
	recombValues = new double[numLoci - 1];
	for(int i = 0; i < numLoci - 1; ++i) {
		datafile>>recombValues[i];
	}
	getline(datafile, line);
	getline(datafile, line);
	getline(datafile, progSpecData);
	progSpecData = line + " " + progSpecData;
};




void WindowInputProvider::readPedfile(ifstream& pedfile) {
	//	Calculate the number of columns for each person
	int numCols = 8;	//	person, father, mother, child, nesipa, nesima, sex, 0
	if (numLiabClasses == 1) {
		numCols += 1 + 2*numLoci;
		startMarkersData = 9;
	}
	else {
		numCols += 2 + 2*numLoci;
		startMarkersData = 10;
	}
	int family = 0;
	pedfile>>family;
	while(family != 0) {
		int* currPersonData = new int[numCols];
		for(int i = 0; i < numCols; ++i) {
			pedfile>>currPersonData[i];
		}
		while (pedfile.get() != '\n') {}
		personsData.push_back(currPersonData);
		family = 0;
		pedfile>>family;
	}
};



void WindowInputProvider::getMarkerName(ifstream& datafile, string& markerName, int markerNum) {
	char c;
	do {
		datafile.get(c);
	} while ((c != '#') && (c != '\n'));
	if (c == '\n') {
		char name[100];
		sprintf(name, "loc%d", markerNum);
		markerName = name;
	}
	else {
		getline(datafile, markerName, '#');
		while (datafile.get() != '\n') {}
	}
};



//------------------------------------------------------------------------------
//	Implementation of wrapper functions:


void setInputWindow(WindowInputProvider* inputProvider, int start, int size) {
	inputProvider->setInputWindow(start, size);
};

int getNumLoci(WindowInputProvider* inputProvider) {
	return inputProvider->getNumLoci();
};

int getSexLinked(WindowInputProvider* inputProvider) {
	return inputProvider->getSexLinked();
};

int* getLociOrder(WindowInputProvider* inputProvider) {
	return inputProvider->getLociOrder();
};

void getLocusName(WindowInputProvider* inputProvider, int locusIndex, char* name) {
	inputProvider->getLocusName(locusIndex, name);
};

int getLocusNumFreqs(WindowInputProvider* inputProvider, int locusIndex) {
	return inputProvider->getLocusNumFreqs(locusIndex);
};

double* getLocusFreqs(WindowInputProvider* inputProvider, int locusIndex) {
	return inputProvider->getLocusFreqs(locusIndex);
};

int getNumLiabClasses(WindowInputProvider* inputProvider) {
	return inputProvider->getNumLiabClasses();
}

double* getPensForLiabClass(WindowInputProvider* inputProvider, int liabClass) {
	return inputProvider->getPensForLiabClass(liabClass);
};

double* getRecombValues(WindowInputProvider* inputProvider) {
	return inputProvider->getRecombValues();
};


void getProgSpecData(WindowInputProvider* inputProvider, char* specData) {
	inputProvider->getProgSpecData(specData);
};

int getNumPersons(WindowInputProvider* inputProvider) {
	return inputProvider->getNumPersons();
};

int* getPersonData(WindowInputProvider* inputProvider, int personIndex) {
	return inputProvider->getPersonData(personIndex);
};