/************************************************************************************/
//	WindowInputProvider.h
//	Class WindowInputProvider : 
//	Reads datafile and pedfile into internal data-structures and then provides
//	all saved data in fixed window of markers.
//	May be used from C-code due to the defined interface
//
//	Written by Anisenia Andrei
//	On 31/08/2010
/************************************************************************************/

#ifndef WINDOW_INPUT_PROVIDER_H
#define WINDOW_INPUT_PROVIDER_H


//----------------------------------------------------------------------------------------------------
//	Class declaration:
#ifdef __cplusplus	

	#include <string>
	#include <list>
	#include <vector>
	using namespace std;

	class WindowInputProvider {

		public:
			//	Constructor
			WindowInputProvider(char* datafile, char* pedfile);

			//	Set window for which input data will be provided
			//	start - the number of the 1st locus in the window (count starting 0)
			//	size - the size of the window
			void setInputWindow(int start, int size);

			//	Getters
			int getNumLoci() { return (stopMarker - startMarker + 1); };
			int getSexLinked() { return sexLinked; };
			int* getLociOrder();
			void getLocusName(int locusIndex, char* name);	//	locusIndex - starting 0, not includes disease locus
			int getLocusNumFreqs(int locusIndex);
			double* getLocusFreqs(int locusIndex);
			int getNumLiabClasses() { return numLiabClasses; };
			//	the size of returned array is 3 for non-sex-linked and 5 for sex-linked (female, then male)
			double* getPensForLiabClass(int liabClass);	//	liabClass - starting 0
			double* getRecombValues();	//	only relevant values
			void getProgSpecData(char* specData);
			int getNumPersons() { return numPersons; };
			//	personIndex - count starting 0
			int* getPersonData(int personIndex) { return personsData[personIndex]; };

			//	Destructor
			~WindowInputProvider();


		private:
			int numLoci;					//	Without disease locus
			int sexLinked;
			int programCode;
			double disFrequency;
			int numLiabClasses;
			double** penetrances;	//	[numLiabClasses][3] for non-sex-linked, or [numLiabClasses][5] for sex-linked
												//	5 = 3 (first female penetrances) + 2 (then male penetrances)
			double** frequencies;		//	[numLoci][number of alleles at each locus]
			string* markerNames;	
			int* numAllelesForLoci;		//	for each locus it's number of alleles
			double* recombValues;		//	(numLoci-1) number of recomb. values between loci
			string progSpecData;		//	The last line of datafile
			int numPersons;
			vector<int*> personsData;			//	list of arrays of [num columns of data in pedfile for each person]
			int startMarkersData;			//	column in personsData where markers data starts 
														//	(counting of columns starts from 0)
			//	Information that will be available from outside
			int startMarker;		//	counting starts from 0
			int stopMarker;		//	counting starts from 0

			//	private methods
			void readDatafile(ifstream& datafile);
			void readPedfile(ifstream& pedfile);
			void getMarkerName(ifstream& datafile, string& markerName, int markerNum);
	};

#else	
	typedef struct WindowInputProvider WindowInputProvider;
#endif


//----------------------------------------------------------------------------------------------------
//	Wrapper function declarations: interface for using MemoryManager from C code
#ifdef __cplusplus
extern "C" {
#endif	

	extern void setInputWindow(WindowInputProvider*, int , int );
	extern int getNumLoci(WindowInputProvider*);
	extern int getSexLinked(WindowInputProvider*);
	extern int* getLociOrder(WindowInputProvider*);
	extern void getLocusName(WindowInputProvider*, int, char*);
	extern int getLocusNumFreqs(WindowInputProvider*, int);
	extern double* getLocusFreqs(WindowInputProvider*, int);
	extern int getNumLiabClasses(WindowInputProvider*);
	extern double* getPensForLiabClass(WindowInputProvider*, int);
	extern double* getRecombValues(WindowInputProvider*);
	extern void getProgSpecData(WindowInputProvider*, char*);
	extern int getNumPersons(WindowInputProvider*);
	extern int* getPersonData(WindowInputProvider*, int );

#ifdef __cplusplus
}
#endif







#endif	//WINDOW_INPUT_PROVIDER_H