/* $Id: fft.c,v 1.2 2008/08/03 08:15:47 superlink Exp $ */
/*************************************************************************/
/* fft.c												     			 */
/* Implements FFT (Fast Fourier Trasmorm) for multiplication			 */
/* Author: Ma'ayan Fishelson, Anna Tzemach 								 */					 
/*************************************************************************/

#include "fft.h"
#include "gdata.h"
#include "likehood.h"
#include "sl_Auxilary.h"

void  reverseTF(double *inp);
//extern	sl_thetaCompTheta	*thetaVals;
int		power2[31]={1,2 ,4 ,8 ,16 ,32 ,64 ,128 ,256 ,512 ,1024 ,2048 ,4096 ,8192 ,16384 ,32768 ,65536 ,131072 ,262144 ,524288 ,1048576 ,2097152 ,4194304 ,8388608 ,16777216 ,33554432 ,67108864 ,134217728 ,268435456 ,536870912 ,1073741824 };
					

/******************************************************************************
*
*       Function        FFT
*       Parameters		result (a vector used to store the result of the 
*						multiplication), input (the input vector), theta (the
*						recombination value), size (the size of both vectors,
*						the input vector and the result vector).
*       Output          none
*       Description		
*       Comments		the vector 'result' needs to be allocated before the
*						sl_function is called.					
*
******************************************************************************/
void gh_calcFFT(double *result, double *input, double theta, int size)
{
	int		n;

	#if (PRINT_FFT_TIME == 1)
		fprintf(fftFile, "Start writing = %lf\t\t", ((double)clock())/(double)CLOCKS_PER_SEC);	
	#endif

	/* 1.	Calculate n (= the size of the inheritance vector). */
	n = (int)ceil(log(size)/log(2.0));

    if (n < 32) {
		gh_calcComplexFFT(result, input, theta, size, n);
	}
	else {
		gh_calcSimpleFFT(result, input, theta, size, n);
	}

	#if (PRINT_FFT_TIME == 1)
		fprintf(fftFile, "End writing = %lf\n", ((double)clock())/(double)CLOCKS_PER_SEC);	
	#endif
}


/******************************************************************************
*
*       Function        gh_calcComplexFFT
*       Parameters		result (a vector used to store the result of the 
*						multiplication), input (the input vector), theta (the
*						recombination value), size (the size of both vectors,
*						the input vector and the result vector).
*       Output          none
*       Description		
*       Comments		the vector 'result' needs to be allocated before the
*						sl_function is called.					
*
******************************************************************************/
void gh_calcComplexFFT(double *result, double *input, double theta, int size, int n){
	int		i, j;
	double base, *Tvals;

	/* 1.	Copy the input vector onto the result vector. */
	memcpy(result, input, sizeof(double) * size);

	/* 2.	Calculate the Fourier Transform of the input vector. */
	gh_calcComplexFTrans(result, size, n);

	/* 3.	Calculate the Fourier Transform of the transition matrix. */
	base = 1 - 2 * theta;

	/* 4.	Allocate memory for an array that would hold the "n+1" 
			different values in the transition matrix, which correspond
			to the number of bits that are different. */
	sl_myMalloc(Tvals, sizeof(double) * (PedigreeData.NumOfMeioses+1));

	/* 5.	Initialize the array. */
	Tvals[0] = 1;
	for (i = 1; i <= PedigreeData.NumOfMeioses; i++) {
		Tvals[i] = Tvals[i-1] * base;
	}

	/* 7.	Calculate the Fourier Transform of the result vector. */
 	for (j = 0; j < PedigreeData.Possibilities; j++){
 		result[j] *= Tvals[WeightArrayTransform[j]];
	}

	/* 8.	Calculate the Reverse Fourier Transform of the result vector. */
	gh_calcComplexFTrans(result, size, n);
    
	for (j = 0; j < size; j++)
		result[j] /= size;

	/* 9.	Free memory. */
	free(Tvals);
}



/******************************************************************************
*
*       Function        gh_calcComplexFTrans
*       Parameters		input (the input vector), size (the size of both vectors,
*						the input vector and the result vector).
*       Output          none
*       Description		Calculate the Fourier Transform of inp.
*       Comments		none
*
******************************************************************************/

void gh_calcComplexFTrans(double *inp, int size, int n)
{
	int		i, j, jFlipBitI;
	double tmp;
	int *ei;
    for (i = 0, ei = power2; i < n; ei++,i++){
		for (j = 0; j < size; j++){

			if ((j & *ei) == 0) {
				jFlipBitI = j ^ *ei;

				tmp = inp[j];
				inp[j] = tmp + inp[jFlipBitI];
				inp[jFlipBitI] = tmp - inp[jFlipBitI];
			}
		}
	}
}


/******************************************************************************
*
*       Function        gh_calcSimpleFFT
*       Parameters		result (a vector used to store the result of the 
*						multiplication), input (the input vector), theta (the
*						recombination value), size (the size of both vectors,
*						the input vector and the result vector).
*       Output          none
*       Description		
*       Comments		the vector 'result' needs to be allocated before the
*						sl_function is called.					
*
******************************************************************************/
void gh_calcSimpleFFT(double *result, double *input, double theta, int size, int n)
{
	int		i, j, *wArr, weight, *wPtr, *wPtrEnd, *wPtrStart, *incArr;
	double	base, *Tvals;
	long	jumpFactor;

	/* 1.	Copy the input vector onto the result vector. */

	memcpy(result, input, sizeof(double) * size);

	/* 2.	Allocate memory for an n-bit array that would be used to
			iterate over all possible values for the inheritance vector. */

	sl_myMalloc(wArr, sizeof(int) * n);

	/* 3.	Allocate memory for an auxilary array used to store the 
			necessary increment/decrement in the index of the 
			input/result vectors as a result of an increment/decrement
			of 1 in each one of the n bits of the inheritance vector. */
	
	sl_myMalloc(incArr, sizeof(int) * n);

	/* 4.	Initialize the array. */

	/* 4.1.	Initialize the jump factor. */

	jumpFactor = 1;

	/* 4.2.	Iterate over the vector's bits. */

	for (i = n-1; i >= 0; i--) {

		/* 4.2.1.	Initialize the jumFactor of bit i. */

		incArr[i] = jumpFactor;

		/* 4.2.2	Multiply the jumpFactor by the number of values of the bit (2). */

		jumpFactor *= 2;
	}

	/* 5.	Calculate the Fourier Transform of the input vector. */

	gh_calcSimpleFTrans(result, size, wArr, n, incArr);


	/* 6.	Calculate the Fourier Transform of the transition matrix. */

	base = 1 - 2 * theta;

	/* 7.	Allocate memory for an array that would hold the "n+1" 
			different values in the transition matrix, which correspond
			to the number of bits that are different. */

	sl_myMalloc(Tvals, sizeof(double) * (n+1));

	/* 8.	Initialize the array. */

	Tvals[0] = 1;

	for (i = 1; i <= n; i++) {
		Tvals[i] = Tvals[i-1] * base;
	}



	for (i=0; i < n; i++)
		wArr[i] = 0;

	wPtrStart = wArr + n - 1;
	wPtrEnd = wArr;

	

	/* 4.	Calculate the Fourier Transform of the result vector. */

	for (j = 0, weight = 0; j < size; j++){
		result[j] *= Tvals[weight];

		for (wPtr = wPtrStart; wPtr >= wPtrEnd; wPtr--) { 
			if (*wPtr == 0) { 
				*wPtr = 1; 
				weight++;
				break;
			} 
			else {
				*wPtr = 0; 
				weight--;
			}
		}
	}

	/* 5.	Calculate the Reverse Fourier Transform of the result vector. */

	gh_calcSimpleFTrans(result, size, wArr, n, incArr);

	for (j = 0; j < size; j++)
		result[j] /= size;

	/* 6.	Free memory. */

	free(wArr);
	free(incArr);
	free(Tvals);
}



/******************************************************************************
*
*       Function        gh_calcSimpleFTrans
*       Parameters		inp (the input vector), size (the size of both vectors,
*						the input vector and the result vector).
*       Output          none
*       Description		Calculates the Fourier Transform of input.
*       Comments		none
*
******************************************************************************/
void gh_calcSimpleFTrans(double *inp, int size, int *wArr, int n, int *incArr)
{
	int		i, j, *wPtr, *wPtrEnd, *wPtrStart, ind;
	double	operand1, operand2;


	/* 1.	Calculate the Fourier Transform of the input vector. */

	for (i=0; i < n; i++)
		wArr[i] = 0;

	wPtrStart = wArr + n - 1;
	wPtrEnd = wArr;

    for (i = 0; i < n; i++){
		for (j = 0; j < size; j++){

			if (wArr[i] == 0) {
				operand1 = inp[j];
				ind = j + incArr[i];
				operand2 = inp[ind];

                inp[j] = operand1 + operand2;
				inp[ind] = operand1 - operand2;
			}

			for (wPtr = wPtrStart; wPtr >= wPtrEnd; wPtr--) { 
				if (*wPtr == 0) { 
					*wPtr = 1; 
					break;
				} 
				else {
					*wPtr = 0; 
				}
			}
		}
	}
}


