/* $Id: haplotypingAlleles.c,v 1.2 2008/08/03 08:15:47 superlink Exp $ */
/****************************************************************************
 ************** haplotypingAlleles.c ****************************************
 ****************************************************************************
 ************** created by Anisenia Andrei **********************************
 *	given the most probable vectors path through the map,					*
 *	we calculate the exact alleles of each person and store them			*
 ****************************************************************************/

#include <math.h>
#include <string.h>
#include "haplotypingAlleles.h"
#include "queue.h"
#include "restrictionsGraph.h"

/********************************************************/
/*	help functions for printResultHaplotypes function	*/

void initRecombinationsArrays(HaplotypingPrintData* haplotypingPrintData, int numHaplotypes);

void addRecombination(HaplotypingPrintData* haplotypingPrintData, int nonFounder);

void printRecombinations(FILE* haplotypesFile,int numRecombinations,
						 int* recombCountersArray, int numHaplotypes,  double sumRecombValues);

/********************************************************/



/****************************************************************************
*
*       Function		initiateFounders  
*       Parameters		pedigree
*						numF, numN
*						locus
*		Output			none
*       Description		initiates founders alleles in "pedigree"
*       Comments		none
*
*****************************************************************************/
void initiateFounders(ped_node** pedigree,int numF,int numN,int locus)
{
	allelesHunter* AllelesHunter;
	
	AllelesHunter = initiateAllelesHunter(pedigree, numF,numN,locus);
	runAllelesHunter(pedigree,AllelesHunter);
	destroyAllelesHunter(AllelesHunter);
}



/****************************************************************************
*
*       Function		initiateNonFounders  
*       Parameters		pedigree
*						numF, size
*						locus
*		Output			none
*       Description		initiates nonfounders alleles in "pedigree"
*       Comments		none
*
*****************************************************************************/
void initiateNonFounders(ped_node** pedigree,int numF,int size,int locus,Bool* firstChild)
{
	int vector, person;
	ped_node* p;

	vector = haplotypeVectors[locus];
	for(person = numF; person < size; person++)
	{
		p = pedigree[person];
		/*	handling the father of the person	*/
		/****************************************/
		if (firstChild[p->father] == true)
		{
			firstChild[p->father] = false;
			p->myAlleles[locus].allele1 = pedigree[p->father]->myAlleles[locus].allele1;
			p->myAlleles[locus].paternalInheritance = 0;
		}
		else
		{
			if (vector%2 == 0)
			{
				p->myAlleles[locus].allele1 = 
					pedigree[p->father]->myAlleles[locus].allele1;
				p->myAlleles[locus].paternalInheritance = 0;
			}
			else
			{
					p->myAlleles[locus].allele1 = 
							pedigree[p->father]->myAlleles[locus].allele2;
					p->myAlleles[locus].paternalInheritance = 1;
			}
			vector = vector>>1;
		}

		/*	handling the mother of the person	*/
		/****************************************/
		if (firstChild[p->mother] == true)
		{
			firstChild[p->mother] = false;
			p->myAlleles[locus].allele2 = pedigree[p->mother]->myAlleles[locus].allele1;
			p->myAlleles[locus].maternalInheritance = 0;
		}
		else
		{
			if (vector%2 == 0)
			{
				p->myAlleles[locus].allele2 = 
					pedigree[p->mother]->myAlleles[locus].allele1;
				p->myAlleles[locus].maternalInheritance = 0;
			}
			else
			{
					p->myAlleles[locus].allele2 = 
							pedigree[p->mother]->myAlleles[locus].allele2;
					p->myAlleles[locus].maternalInheritance = 1;
			}
			vector = vector>>1;
		}
	}
}


/****************************************************************************
*
*       Function		setLocusAlleles  
*       Parameters	  	locus - number of locus
*						pedigree
*						numF - number of founders
*						numN - number of nonfounders
*       Output			none
*       Description		sets alleles of nonfounders in the locus="locus"
*						using inheritance vector = haplotypeVectors[locus]  
*       Comments		none
*
*****************************************************************************/
void setLocusAlleles(ped_node** pedigree,int numF,int numN,int locus)
{
	int size;
	Bool* firstChild;	/*	for handling first children */

	size = numN+numF; 
	/*	initiating founders alleles at current locus*/
	initiateFounders(pedigree,numF,numN,locus);

	/*  initiating nonfounders alleles at current locus	*/
	firstChild = FirstChildren(numF,size);
	initiateNonFounders(pedigree,numF,size,locus,firstChild);
	free(firstChild);

}



/****************************************************************************
*
*       Function		setHaplotypeAlleles  
*       Parameters		  
*									haplotypingPrintData - the structure with specific data for haplotypes printing
*		Output			none
*       Description		for each person the most probable pair of alleles 
*						at each locus is calculated
*       Comments		the function uses "haplotypeVectors"
*
*****************************************************************************/
void setHaplotypeAlleles(HaplotypingPrintData* haplotypingPrintData) {
	SortedPedigree* sortedPed;
	int curLocus, numIBDalleles;
	int** recMatrix;
	int startLocus;

	/*	intiating pedigree	*/
	sortedPed = InitSortedPedigree(NumInfMarkers);
	numIBDalleles = initiateFoundersIBDalleles(sortedPed);
	/*	setting alleles at each marker for all non-founders */
	for(curLocus = 0; curLocus < NumInfMarkers; curLocus++) {	
		setLocusAlleles_New(sortedPed, curLocus, numIBDalleles);
	}
	
	if (haplotypingPrintData->indicationFirstPrint) {
		//	Print pedigree persons names
		printPedigreePersons(haplotypingPrintData->haplotypesFile, 
				sortedPed->pedigree, sortedPed->numF, sortedPed->numN);
		startLocus = 0;
		//	Initiate log-likelihood and recombinations arrays
		haplotypingPrintData->logLikelihoodHaplo = initiateLogLikelihood(sortedPed->numF, sortedPed->numN);
		initRecombinationsArrays(haplotypingPrintData, 2*(sortedPed->numN));
	}
	else {
		startLocus = 1;
	}
	//	Find the recombinations in current window
	recMatrix = 
		findRecombinations(sortedPed->pedigree,sortedPed->numF,sortedPed->numN);
	//	Print haplotypes for the current window
	printHaplotypes(haplotypingPrintData->haplotypesFile, haplotypingPrintData->indicationLastPrint,
			sortedPed, startLocus, recMatrix);
	//	Count the recombinations and update log-likelihood for the current window
	updateRecombinationsAndLikelihood(haplotypingPrintData, sortedPed, recMatrix, startLocus);
	//	Print the resulted recombinations and log-likelihood in the last print (last window calculation)
	if (haplotypingPrintData->indicationLastPrint) {
		printRecombinationsAndLikelihood(haplotypingPrintData, sortedPed->numN);
	}
	
	destroyRecMatrix(recMatrix,sortedPed->numN);
	destroySortedPedigree(sortedPed);
}



/****************************************************************************
*
*       Function			printRecombinationsAndLikelihood  
*       Parameters		haplotypingPrintData - the structure with specific data for haplotypes printing
*									numN
*		Output				none
*       Description		calculates recombinations and log-likelihood in the end of the calculation (last window)
*       Comments			none
*
*****************************************************************************/
void printRecombinationsAndLikelihood(HaplotypingPrintData* haplotypingPrintData, int numN) {
	double probability;

	fprintf(haplotypingPrintData->haplotypesFile,"\n");
	printRecombinations(haplotypingPrintData->haplotypesFile, haplotypingPrintData->numRecombinations,
		haplotypingPrintData->recombCountersArray, 2*numN, haplotypingPrintData->sumRecombValues);
	probability = (haplotypingPrintData->logLikelihoodHaplo)/log(10);
	fprintf(haplotypingPrintData->haplotypesFile,"\nThe Log-likelihood of this assignment: %lf\n",probability);
}



/****************************************************************************
*
*       Function			updateRecombinationsAndLikelihood  
*       Parameters		haplotypingPrintData - the structure with specific data for haplotypes printing
*									sortedPed, recMatrix, startLocus
*		Output				none
*       Description		calculates recombinations and log-likelihood addition for the current window
*       Comments			none
*
*****************************************************************************/
void updateRecombinationsAndLikelihood(HaplotypingPrintData* haplotypingPrintData,
									   SortedPedigree* sortedPed, int** recMatrix, int foundersStartLocus) {
	int numN, numF;
	ped_node** pedigree;
	
	pedigree = sortedPed->pedigree;
	numF = sortedPed->numF;
	numN = sortedPed->numN;
	
	//	Updating log-likelihood for founder haplotype alleles
	updateLikelihoodUsingFounders(haplotypingPrintData, numF, pedigree, foundersStartLocus);
	
	//	Updating log-likelihood and recombinations for non-founder haplotype alleles
	updateLikeAndRecombUsingNonFounders(haplotypingPrintData, numF, numN, pedigree, recMatrix);
}



/****************************************************************************
*
*       Function			updateLikelihoodUsingFounders  
*       Parameters		haplotypingPrintData - the structure with specific data for haplotypes printing
*									numF, pedigree, foundersStartLocus
*		Output				none
*       Description		calculates log-likelihood addition for the founders in the current window
*       Comments			none
*
*****************************************************************************/
void updateLikelihoodUsingFounders(HaplotypingPrintData* haplotypingPrintData, 
				int numF, ped_node** pedigree, int foundersStartLocus) {
	
	int locus, person, allele;
	for(locus = foundersStartLocus; locus < NumInfMarkers; locus++) {
		for(person = 0; person < numF; person++) {
			//	treating first haplotype of the person	
			allele = pedigree[person]->myAlleles[locus].allele1;
			haplotypingPrintData->logLikelihoodHaplo += 
					log(lociDataArray[MarkerOrder[locus]+1].frequencies[allele-1]);
			//	treating second haplotype of the person	
			allele = pedigree[person]->myAlleles[locus].allele2;
			haplotypingPrintData->logLikelihoodHaplo += 
					log(lociDataArray[MarkerOrder[locus]+1].frequencies[allele-1]);
		}
	}
}



/****************************************************************************
*
*       Function			updateLikeAndRecombUsingNonFounders  
*       Parameters		haplotypingPrintData - the structure with specific data for haplotypes printing
*									numF, numN, pedigree, recMatrix
*		Output				none
*       Description		calculates log-likelihood addition for the non-founders in the current window
*									and updates the recombinations using recMatrix
*       Comments			none
*
*****************************************************************************/
void updateLikeAndRecombUsingNonFounders(HaplotypingPrintData* haplotypingPrintData, 
								   int numF, int numN, ped_node** pedigree, int** recMatrix)  {
	
	int locus, person, allele;
	for(locus = 0; locus < NumInfMarkers - 1; locus++) {
		for(person = numF; person < numF + numN; person++) {
			
			//	treating first haplotype of the person	
			allele = pedigree[person]->myAlleles[locus].allele1;
			if (recMatrix[2*(person-numF)][locus] == 1) {
				//	there was a recombination between ("locus") and ("locus"+1)
				haplotypingPrintData->logLikelihoodHaplo += log(Recomb_values[MarkerOrder[locus]]);
				haplotypingPrintData->numRecombinations++;
				addRecombination(haplotypingPrintData, 2*(person-numF));
			}
			else {
				//	there was no recombination between ("locus") and ("locus"+1)
				haplotypingPrintData->logLikelihoodHaplo += log(1 - Recomb_values[MarkerOrder[locus]]);
			}

			//	treating second haplotype of the person	
			allele = pedigree[person]->myAlleles[locus].allele2;
			if (recMatrix[2*(person-numF)+1][locus] == 1) {
				/*	there was a recombination between ("locus") and ("locus"+1)	*/
				haplotypingPrintData->numRecombinations++;
				addRecombination(haplotypingPrintData, 2*(person-numF)+1);
				haplotypingPrintData->logLikelihoodHaplo += log(Recomb_values[MarkerOrder[locus]]);
			}
			else {
				/*	there was no recombination between ("locus") and ("locus"+1)	*/
				haplotypingPrintData->logLikelihoodHaplo += log(1 - Recomb_values[MarkerOrder[locus]]);
			}
		}
	}

}


/****************************************************************************
*
*       Function		initiateLogLikelihood  
*       Parameters	numF, numN
*		Output			probability
*       Description	initiates the log-likelihood of haplotyping (for each non-founder adds log(0.25))
*       Comments		none
*
*****************************************************************************/
double initiateLogLikelihood(int numF, int numN) {
	int person;
	double probability = 0;

	for(person = numF; person < numF + numN; person++) {
			probability += log(0.25);
	}
	return probability;
}


/****************************************************************************
*
*       Function		setLocusAlleles_New 
*       Parameters	  	pedigree
*						curlocus - number of locus
*						numIBDalleles - number of founders IBD alleles
*       Output			none
*       Description		sets alleles of nonfounders in the locus="locus"
*						using inheritance vector = haplotypeVectors[locus]  
*       Comments		none
*
*****************************************************************************/
void setLocusAlleles_New(SortedPedigree* sortedPed,int curLocus,int numIBDalleles)
{
	int person, numPersons, IBD1, IBD2, allele1, allele2;
	RestrictionsGraph* rGraph;
	Pedigree* Ped;
	ped_node** pedigree;

	Ped = AllPedigreesList->PedigeeGraph;
	pedigree = sortedPed->pedigree;
	/*	1. initiating non-founders alleles due to max vector in current locus*/
	initiateNonFoundersAlleles(sortedPed, haplotypeVectors[curLocus]);

	/*	2. creating empty restrictions graph of founder alleles	*/
	rGraph = createRestrictionsGraph(numIBDalleles, 
		lociDataArray[MarkerOrder[curLocus] + 1].frequencies,
		lociDataArray[MarkerOrder[curLocus] + 1].num_of_freq);

	/*	3. adding restrictions	to founders alleles	*/
	numPersons = sortedPed->numF + sortedPed->numN;
	for(person = 0; person < numPersons; ++person)
	{
		/*	for each person in the pedigree:	*/
		allele1 = Ped[pedigree[person]->PedIndex].myLocus[MarkerOrder[curLocus] + 1].allele1;
		allele2 = Ped[pedigree[person]->PedIndex].myLocus[MarkerOrder[curLocus] + 1].allele2;
		if ( (allele1 != 0) && (allele2 != 0) )
		{
			/*	if both person's alleles are known => get his IBD alleles	*/
			IBD1 = pedigree[person]->allele1;
			IBD2 = pedigree[person]->allele2;
			if (IBD1 == ALLELE_DOESNT_EXIST)
			{
				if (IBD2 == ALLELE_DOESNT_EXIST)
				{
					/*	person has nor paternal neither maternal alleles!	*/
					continue;
				}
				else
				{
					/*	person has no paternal allele	*/
					IBD1 = IBD2;
				}
			}
			else
			{
				if (IBD2 == ALLELE_DOESNT_EXIST)
				{
					/*	person has no maternal allele	*/
					IBD2 = IBD1;
				}
			}
			/*	IBD1 and IBD2 are nodes while allele1 and allele2 are restrictions	*/
			/*	in the restriction graph of founder alleles	(rGraph)	*/
			/*	so we add new restriction (IBD1,IBD2,allele1,allele2) to the graph	*/
			addRestriction(rGraph, IBD1, IBD2, allele1, allele2);
		}
	}

	/*	4. satisfying alleles restrictions in graph of founder alleles	*/
	satisfyRestrictions(rGraph);

	/*	5. copying alleles restrictions to pedigree	*/
	for(person = 0; person < numPersons; ++person)
	{
		pedigree[person]->myAlleles[curLocus].allele1 = 
			getVertexRestriction(rGraph, pedigree[person]->allele1);
		pedigree[person]->myAlleles[curLocus].allele2 = 
			getVertexRestriction(rGraph, pedigree[person]->allele2);
	}

	/*	6.	*/
	destroyRestrictionsGraph(rGraph);
}



/****************************************************************************
*
*       Function		printPedigreePersons  
*       Parameters		haplotypesFile, pedigree, numF, numN	  	
*		Output			none
*       Description		prints names of the persons in the given pedigree
*       Comments		none
*
*****************************************************************************/
void printPedigreePersons(FILE* haplotypesFile,  ped_node** pedigree, int numF, int numN) {
	
	int locusNameMaxLength, i, person;
	locusNameMaxLength = strlen(lociDataArray[NumOfLocuses - 1].locusName);

	for(i = 0; i < locusNameMaxLength; i++) {
		fprintf(haplotypesFile," ");
	}
	fprintf(haplotypesFile,"\t\t");
	for(person = 0; person < numF + numN; person++) {
		fprintf(haplotypesFile,"person%s\t\t\t\t", PedigreeGraph[pedigree[person]->PedIndex].personName);
	}
	fprintf(haplotypesFile,"\n\n");
}


/****************************************************************************
*
*       Function		printHaplotypes  
*       Parameters		haplotypesFile, indicationLastPrint, sortedPed	  	
*									startPrintHaplotypesFrom - the locus to start the haplotype printing from
*									recMatrix - matrix with recombinations
*		Output			none
*       Description		prints resulted exact pedigree haplotypes
*       Comments		none
*
*****************************************************************************/
void printHaplotypes(FILE* haplotypesFile, Bool indicationLastPrint, SortedPedigree* sortedPed, 
					 int startPrintHaplotypesFrom, int** recMatrix) {

	int locusNameMaxLength, locus, numF, numN;
	ped_node** pedigree;

	pedigree = sortedPed->pedigree;
	numF = sortedPed->numF;
	numN = sortedPed->numN;
	locusNameMaxLength = strlen(lociDataArray[NumOfLocuses - 1].locusName);	
	//	printing haplotypes for each locus
	for(locus = 0; locus < NumInfMarkers; locus++) {
		if (locus >= startPrintHaplotypesFrom) {
			printLocusHaplotypes(haplotypesFile, pedigree, numN + numF, locus, locusNameMaxLength);
		}
		//	Print recombinations between 2 loci: (locus) and (locus + 1)
		if (locus < NumInfMarkers - 1) {
			printLineWithRecombinations(haplotypesFile, numF, numN, recMatrix, locus, locusNameMaxLength);
		}
	}
	//	In the end print line without recombinations (locus equals NumInfMarkers, so no recombinations)
	if (indicationLastPrint) {
		printLineWithRecombinations(haplotypesFile, numF, numN, recMatrix, locus, locusNameMaxLength);
	}
}



/****************************************************************************
*
*       Function		printLocusHaplotypes  
*       Parameters		haplotypesFile, pedigree, numPersons  	
*									locus, locusNameMaxLength
*		Output			none
*       Description		prints resulted exact pedigree haplotypes for the given locus
*       Comments		is used by "printHaplotypes" function
*
*****************************************************************************/
void printLocusHaplotypes(FILE* haplotypesFile, ped_node** pedigree, int numPersons, 
						  int locus, int locusNameMaxLength) {
	
	int locusNameLength, i, person, allele1, allele2;

	// Printing locus name
	fprintf(haplotypesFile,"%s", lociDataArray[MarkerOrder[locus]+1].locusName);
	locusNameLength = strlen(lociDataArray[MarkerOrder[locus]+1].locusName);
	for(i = 0; i < locusNameMaxLength - locusNameLength; i++) {
		fprintf(haplotypesFile," ");
	}
	fprintf(haplotypesFile,"\t");

	// For each person print his haplotypes
	for(person = 0; person < numPersons; person++) {
		allele1 = pedigree[person]->myAlleles[locus].allele1;
		allele2 = pedigree[person]->myAlleles[locus].allele2;
		fprintf(haplotypesFile,"\t%d\t|\t%d\t", allele1, allele2);
	}	
	fprintf(haplotypesFile,"\n");
}




/****************************************************************************
*
*       Function		printLineWithRecombinations  
*       Parameters		haplotypesFile, numF, numN	  	
*									recMatrix - matrix with recombinations
*		Output			none
*       Description		prints the line with recombinations signed "X" between 2 adjacent loci  
*       Comments		none
*
*****************************************************************************/
void printLineWithRecombinations(FILE* haplotypesFile,  int numF, int numN, int** recMatrix, int locus,
																		int locusNameMaxLength){
	int i, person;

	for(i = 0; i < locusNameMaxLength; i++) {
		fprintf(haplotypesFile," ");
	}
	fprintf(haplotypesFile,"\t\t");
	for(person = 0; person < numF + numN; person++) {
		//	print line with recombinations
		if (person < numF) {
			//	for founder:
			fprintf(haplotypesFile,"\t|\t\t\t");	
		} 
		else {
			//	for non-founder in paternal locus:
			if ((locus >= 0) && (locus < NumInfMarkers - 1) && (recMatrix[2*(person-numF)][locus] == 1)) {
				//	there was a recombination between ("locus") and ("locus"+1)
				fprintf(haplotypesFile,"X\t|\t");
			}
			else {
				fprintf(haplotypesFile,"\t|\t");
			}
			//	for non-founder in maternal locus:
			if ((locus >= 0) && (locus < NumInfMarkers - 1) && (recMatrix[2*(person-numF)+1][locus] == 1)) {
				//	there was a recombination between ("locus") and ("locus"+1)
				fprintf(haplotypesFile,"X\t\t");
			}
			else {
				fprintf(haplotypesFile,"\t\t");
			}
		}
	}
	fprintf(haplotypesFile,"\n");
}




/****************************************************************************
*
*       Function		printResultHaplotypes  
*       Parameters		haplotypesFile, pedigree, numF, numN	  	
*		Output			none
*       Description		prints resulted exact pedigree haplotypes
*       Comments		none
*
*****************************************************************************/
void printResultHaplotypes(FILE* haplotypesFile, ped_node** pedigree,
						   int numF,int numN, int** recMatrix) {
	int person, locus, numRecombinations, allele, allele1, allele2, locusNameMaxLength, locusNameLength, i;
	int* recombCountersArray, *recombHaplotypesArray;
	double probability;
	char tmp[50];

	//	initializations
	probability = 0;
	numRecombinations = 0;
	sl_myMalloc(recombCountersArray, 5*sizeof(int));
	sl_myMalloc(recombHaplotypesArray, 2*numN*sizeof(int));
	//initRecombinationsArrays(recombCountersArray,recombHaplotypesArray, 2*numN);
	locusNameMaxLength = strlen(lociDataArray[MarkerOrder[NumInfMarkers - 1]+1].locusName);

	//	person numbers
	for(i = 0; i < locusNameMaxLength; i++) {
		fprintf(haplotypesFile," ");
	}
	fprintf(haplotypesFile,"\t\t");
	for(person = 0; person < numF + numN; person++) {
		fprintf(haplotypesFile,"person%d\t\t\t\t", pedigree[person]->PedIndex + 1);
	}
	fprintf(haplotypesFile,"\n\n");
	
	//	printing haplotypes for each locus
	for(locus = 0; locus < NumInfMarkers; locus++) {
		
		fprintf(haplotypesFile,"%s", lociDataArray[MarkerOrder[locus]+1].locusName);
		locusNameLength = strlen(lociDataArray[MarkerOrder[locus]+1].locusName);
		for(i = 0; i < locusNameMaxLength - locusNameLength; i++) {
			fprintf(haplotypesFile," ");
		}
		fprintf(haplotypesFile,"\t");

		//	for each person
		for(person = 0; person < numF + numN; person++) {
			//	print 2 haplotyping alleles of the person
			allele1 = pedigree[person]->myAlleles[locus].allele1;
			allele2 = pedigree[person]->myAlleles[locus].allele2;
			fprintf(haplotypesFile,"\t%d\t|\t%d\t", allele1, allele2);
		}

		fprintf(haplotypesFile,"\n");
		for(i = 0; i < locusNameMaxLength; i++) {
			fprintf(haplotypesFile," ");
		}
		fprintf(haplotypesFile,"\t\t");
		for(person = 0; person < numF + numN; person++) {
			//	print line with recombinations
			if (person < numF) {
				//	for founder:
				fprintf(haplotypesFile,"\t|\t\t\t");	
			} 
			else {
				//	for non-founder in paternal locus:
				if (recMatrix[2*(person-numF)][locus] == 1) {
					//	there was a recombination between ("locus") and ("locus"+1)
					fprintf(haplotypesFile,"X\t|\t");
				}
				else {
					fprintf(haplotypesFile,"\t|\t");
				}
				//	for non-founder in maternal locus:
				if (recMatrix[2*(person-numF)+1][locus] == 1) {
					//	there was a recombination between ("locus") and ("locus"+1)
					fprintf(haplotypesFile,"X\t\t");
				}
				else {
					fprintf(haplotypesFile,"\t\t");
				}
			}
		}

		fprintf(haplotypesFile,"\n");
	}
	
	fprintf(haplotypesFile,"\n");

	
	//	calculating log-likelihood of the assignment
	for(person=0; person<numF+numN; person++) {
		if (person >= numF) {
			probability += log(0.25);
		}
		//	treating first haplotype of the person	
		for(locus=0; locus < NumInfMarkers; locus++) {
			
			allele = pedigree[person]->myAlleles[locus].allele1;
			if (person < numF) {
				//	for founder:
				probability += 
					log(lociDataArray[MarkerOrder[locus]+1].frequencies[allele-1]);
			}
			else if (locus < NumInfMarkers-1) {
				//	for non-founder:	
				if (recMatrix[2*(person-numF)][locus] == 1) {
					//	there was a recombination between ("locus") and ("locus"+1)
					numRecombinations++;
					//addRecombination(recombCountersArray,recombHaplotypesArray,
						//												2*(person-numF));
					probability += log(Recomb_values[locus]);
				}
				else {
					//	there was no recombination between ("locus") and ("locus"+1)
					probability += log(1 - Recomb_values[locus]);
				}
			}
		}
		//	treating second haplotype of the person
		for(locus=0; locus < NumInfMarkers; locus++) {
			
			allele = pedigree[person]->myAlleles[locus].allele2;
			if (person < numF) {
				//	for founder:
				probability += 
					log(lociDataArray[MarkerOrder[locus]+1].frequencies[allele-1]);
			}
			else if (locus < NumInfMarkers-1) {
				//	for non-founder:
				if (recMatrix[2*(person-numF)+1][locus] == 1) {
					/*	there was a recombination between ("locus") and ("locus"+1)	*/
					numRecombinations++;
					//addRecombination(recombCountersArray,recombHaplotypesArray,
					//													2*(person-numF)+1);
					probability += log(Recomb_values[locus]);
				}
				else {
					/*	there was no recombination between ("locus") and ("locus"+1)	*/
					probability += log(1 - Recomb_values[locus]);
				}
			}
		}
	}

	//printRecombinations(haplotypesFile,numRecombinations,recombCountersArray,2*numN);
	probability /= log(10);
	fprintf(haplotypesFile,"\nThe Log-likelihood of this assignment: %lf\n",probability);
	free(recombCountersArray);
	free(recombHaplotypesArray);
}

/********************************************************/
/*	help functions for printResultHaplotypes function	*/

void initRecombinationsArrays(HaplotypingPrintData* haplotypingPrintData, int numHaplotypes) {
	int i;
	
	sl_myMalloc(haplotypingPrintData->recombCountersArray, 5*sizeof(int));
	sl_myMalloc(haplotypingPrintData->recombHaplotypesArray, numHaplotypes*sizeof(int));
	haplotypingPrintData->recombCountersArray[0] = numHaplotypes;
	for(i = 1; i < 5; i++) {
		haplotypingPrintData->recombCountersArray[i] = 0;
	} 
	for(i = 0; i < numHaplotypes; i++) {
		haplotypingPrintData->recombHaplotypesArray[i] = 0;
	}
}


void addRecombination(HaplotypingPrintData* haplotypingPrintData, int haplotype)
{
	if (haplotypingPrintData->recombHaplotypesArray[haplotype] == 4)
	{
		return;
	}
	haplotypingPrintData->recombCountersArray[haplotypingPrintData->recombHaplotypesArray[haplotype]]--;
	haplotypingPrintData->recombHaplotypesArray[haplotype]++;
	haplotypingPrintData->recombCountersArray[haplotypingPrintData->recombHaplotypesArray[haplotype]]++;
}



void printRecombinations(FILE* haplotypesFile,int numRecombinations,
						 int* recombCountersArray, int numHaplotypes, double sumRecombValues) {
	int i;
	double expectedRecombs;

	fprintf(haplotypesFile,"\nNUMBER OF RECOMBINATIONS\t\t");
	fprintf(haplotypesFile,"NUMBER OF HAPLOTYPES WITH r RECOMBINATIONS:\n");
	fprintf(haplotypesFile,"EXPECTED:\tSEEN:\t\t\t\t");
	for(i = 0; i < 5; i++)
	{
		if (i < 4)
		{
			fprintf(haplotypesFile,"r = %d\t\t",i);
		}
		else
		{
			fprintf(haplotypesFile,"r > 3\n");
		}
	}
	fprintf(haplotypesFile,"------------------------------------------------------------");
	fprintf(haplotypesFile,"------------------------------\n");
	expectedRecombs =  sumRecombValues * numHaplotypes;
	fprintf(haplotypesFile,"%.2lf\t\t", expectedRecombs);
	fprintf(haplotypesFile,"%d\t\t\t\t",numRecombinations);
	for(i = 0; i < 5; i++)
	{
		fprintf(haplotypesFile,"%d\t\t",recombCountersArray[i]);
	}
	fprintf(haplotypesFile,"\n\n");
}


 

/********************************************************/


/****************************************************************************
*
*       Function		findRecombinations  
*       Parameters		pedigree, numF, numN	  	
*		Output			matrix with recombinations
*       Description		computes the matrix with recombinations
*       Comments		none
*
*****************************************************************************/
int** findRecombinations(ped_node** pedigree,int numF,int numN)
{
	int** recMatrix, lociPass, nfAllele, nonfounder, *vector1, *vector2, *tmpVector;
	AlleleData* treatedAllele;
	Pedigree* Ped;

	/*	initiation of recMatrix		*/
	sl_myMalloc(recMatrix, 2*numN*sizeof(int*));
	for(nfAllele=0; nfAllele<2*numN; nfAllele++)
	{
		sl_myMalloc(recMatrix[nfAllele], (NumInfMarkers-1)*sizeof(int));
		for(lociPass=0; lociPass<NumInfMarkers-1; lociPass++)
		{
			recMatrix[nfAllele][lociPass] = 0;
		}
	}
	/*	initiation of vectors arrays	*/
	vector1 = getInheritanceVector(haplotypeVectors[0]);
	sl_myMalloc(vector2, (PedigreeData.SpeedUpNumOfMeioses)*sizeof(int));
	Ped = AllPedigreesList->PedigeeGraph;

	/*	computation of recombinations			*/
	/*	for each pass between 2 adjacent loci	*/
	for(lociPass = 0; lociPass < NumInfMarkers - 1; lociPass++)
	{
		setInheritanceVector(vector2, haplotypeVectors[lociPass+1]);
		/*	go through all non-founders and for each allele of each n.-f. do:	*/
		for(nonfounder = numF, nfAllele = 0; nonfounder < (numN + numF);
				nonfounder += nfAllele, nfAllele = 1 - nfAllele)
		{
			/*	choosing allele to treat	*/
			if (nfAllele == 0)
			{
				treatedAllele = Ped[pedigree[nonfounder]->PedIndex].paternalAllele;
			}
			else
			{
				treatedAllele = Ped[pedigree[nonfounder]->PedIndex].maternalAllele;
			}
			if  ( (treatedAllele->bitType == RegularBit) && 
				  (vector1[treatedAllele->number] != vector2[treatedAllele->number]) )
			{
					/*	signing recombination	*/
					recMatrix[2*(nonfounder - numF) + nfAllele][lociPass] = 1;
			}
		}	
		/*	swapping vectors for the next lociPass	*/
		tmpVector = vector1;
		vector1 = vector2;
		vector2 = tmpVector;
	}
	free(vector1);
	free(vector2);

	return recMatrix;
}



/****************************************************************************
*
*       Function		destroyRecMatrix  
*       Parameters		recMatrix,numN	  	
*		Output			none
*       Description		destroys the matrix with recombinations
*       Comments		none
*
*****************************************************************************/
void destroyRecMatrix(int** recMatrix,int numN)
{
	int nfAllele;

	for(nfAllele=0; nfAllele<2*numN; nfAllele++)
	{
		free(recMatrix[nfAllele]);
	}
	free(recMatrix);
}



/****************************************************************************
*
*       Function		printInput  
*       Parameters		pedigree
*						numF, numN
*		Output			none
*       Description		prints initial unordered haplotypes and the most 
*						probable vectors at each locus
*       Comments		none
*
*****************************************************************************/
void printInput(ped_node** pedigree,int numF,int numN)
{	
	int person,locus;
	Pedigree* Ped;
	FILE* haplotypesFile;

	Ped = AllPedigreesList->PedigeeGraph;

	haplotypesFile = fopen(UNORDERED_HAPLOTYPES_FILE_NAME, "w");
	if (haplotypesFile == NULL)
	{
		printf("File %s could not be opened.", UNORDERED_HAPLOTYPES_FILE_NAME);
		exit(1);
	}

	fprintf(haplotypesFile,"\t\t");
	for(locus=0; locus<NumInfMarkers; locus++)
	{
		fprintf(haplotypesFile,"locus%d\t",locus);
	}
	
	for(person=0; person< numF+numN; person++)
	{
		fprintf(haplotypesFile,"\n\n");
		fprintf(haplotypesFile,"person%d\t",person);
		for(locus=0; locus<NumInfMarkers; locus++)
		{
			fprintf(haplotypesFile,"{%d,",
					Ped[pedigree[person]->PedIndex].myLocus[locus+1].allele1);
			fprintf(haplotypesFile,"%d}\t\t",
					Ped[pedigree[person]->PedIndex].myLocus[locus+1].allele2);
		}
	}
	fprintf(haplotypesFile,"\n\n");
	fprintf(haplotypesFile,"path:\t\t");
	for(locus=0; locus<NumInfMarkers; locus++)
	{
		fprintf(haplotypesFile,"[%d]\t\t",haplotypeVectors[locus]);
	}
	fclose(haplotypesFile);

}



