/* $Id: haplotypingAlleles.h,v 1.2 2008/08/03 08:15:47 superlink Exp $ */
/****************************************************************************
 ************** haplotypingAlleles.h ****************************************
 ****************************************************************************
 ************** created by Anisenia Andrei **********************************
 *	given the most probable vectors path through the map,					*
 *	we calculate the exact alleles of each person and store them			*
 ****************************************************************************/

#ifndef HAPLOTYPING_ALLELES
#define HAPLOTYPING_ALLELES

#include "haplotypingSimple.h"
#include "initiatePedigree.h"
#include "initAllelesHunter.h"
#include "runAllelesHunter.h"

#define HAPLOTYPES_FILE_NAME			"haplotypes.txt"
#define UNORDERED_HAPLOTYPES_FILE_NAME	"unorderedHaplotypesNew.txt"

/****************************************************************************
*
*       Function		initiateFounders  
*       Parameters		pedigree
*						numF, numN
*						locus
*		Output			none
*       Description		initiates founders alleles in "pedigree"
*       Comments		none
*
*****************************************************************************/
void initiateFounders(ped_node** pedigree,int numF, int numN, int locus);


/****************************************************************************
*
*       Function		initiateNonFounders  
*       Parameters		pedigree
*						numF, size
*						locus
*		Output			none
*       Description		initiates nonfounders alleles in "pedigree"
*       Comments		none
*
*****************************************************************************/
void initiateNonFounders(ped_node** pedigree,int numF,int size,int locus,Bool* firstChild);


/****************************************************************************
*
*       Function		setLocusAlleles  
*       Parameters	  	locus - number of locus
*						pedigree
*						numF - number of founders
*						numN - number of nonfounders
*       Output			none
*       Description		sets alleles of nonfounders in the locus="locus"
*						using inheritance vector = haplotypeVectors[locus]  
*       Comments		none
*
*****************************************************************************/
void setLocusAlleles(ped_node** pedigree,int numF,int numN,int locus);



/****************************************************************************
*
*       Function		setHaplotypeAlleles  
*       Parameters			  
*									haplotypingPrintData - the structure with specific data for haplotypes printing
*		Output			none
*       Description		for each person the most probable pair of alleles 
*						at each locus is calculated
*       Comments		the function uses "haplotypeVectors"
*
*****************************************************************************/
void setHaplotypeAlleles(HaplotypingPrintData* haplotypingPrintData);



/****************************************************************************
*
*       Function			updateRecombinationsAndLikelihood  
*       Parameters		haplotypingPrintData - the structure with specific data for haplotypes printing
*									sortedPed, recMatrix, startLocus
*		Output				none
*       Description		calculates recombinations and log-likelihood addition for the current window
*       Comments			none
*
*****************************************************************************/
void updateRecombinationsAndLikelihood(HaplotypingPrintData* haplotypingPrintData,
									   SortedPedigree* sortedPed, int** recMatrix, int startLocus);


/****************************************************************************
*
*       Function			updateLikelihoodUsingFounders  
*       Parameters		haplotypingPrintData - the structure with specific data for haplotypes printing
*									numF, pedigree, foundersStartLocus
*		Output				none
*       Description		calculates log-likelihood addition for the founders in the current window
*       Comments			none
*
*****************************************************************************/
void updateLikelihoodUsingFounders(HaplotypingPrintData* haplotypingPrintData, 
				int numF, ped_node** pedigree, int foundersStartLocus);



/****************************************************************************
*
*       Function			updateLikeAndRecombUsingNonFounders  
*       Parameters		haplotypingPrintData - the structure with specific data for haplotypes printing
*									numF, numN, pedigree, recMatrix
*		Output				none
*       Description		calculates log-likelihood addition for the non-founders in the current window
*									and updates the recombinations using recMatrix
*       Comments			none
*
*****************************************************************************/
void updateLikeAndRecombUsingNonFounders(HaplotypingPrintData* haplotypingPrintData, 
								   int numF, int numN, ped_node** pedigree, int** recMatrix);



/****************************************************************************
*
*       Function			printRecombinationsAndLikelihood  
*       Parameters		haplotypingPrintData - the structure with specific data for haplotypes printing
*									numN
*		Output				none
*       Description		calculates recombinations and log-likelihood in the end of the calculation (last window)
*       Comments			none
*
*****************************************************************************/
void printRecombinationsAndLikelihood(HaplotypingPrintData* haplotypingPrintData, int numN);



/****************************************************************************
*
*       Function		initiateLogLikelihood  
*       Parameters	numF, numN
*		Output			probability
*       Description	initiates the log-likelihood of haplotyping (for each non-founder adds log(0.25))
*       Comments		none
*
*****************************************************************************/
double initiateLogLikelihood(int numF, int numN);


/****************************************************************************
*
*       Function		setLocusAlleles_New 
*       Parameters	  	sortedPed
*						curlocus - number of locus
*						numIBDalleles - number of founders IBD alleles
*       Output			none
*       Description		sets alleles of nonfounders in the locus="locus"
*						using inheritance vector = haplotypeVectors[locus]  
*       Comments		none
*
*****************************************************************************/
void setLocusAlleles_New(SortedPedigree* sortedPed,int curLocus,int numIBDalleles);



/****************************************************************************
*
*       Function		printPedigreePersons  
*       Parameters		haplotypesFile, pedigree, numF, numN	  	
*		Output			none
*       Description		prints names of the persons in the given pedigree
*       Comments		none
*
*****************************************************************************/
void printPedigreePersons(FILE* haplotypesFile,  ped_node** pedigree, int numF, int numN) ;



/****************************************************************************
*
*       Function		printHaplotypes  
*       Parameters		haplotypesFile, indicationLastPrint, sortedPed	  	
*									startPrintHaplotypesFrom - the locus to start the haplotype printing from
*									recMatrix - matrix with recombinations
*		Output			none
*       Description		prints resulted exact pedigree haplotypes
*       Comments		none
*
*****************************************************************************/
void printHaplotypes(FILE* haplotypesFile, Bool indicationLastPrint, SortedPedigree* sortedPed, 
					 int startPrintHaplotypesFrom, int** recMatrix) ;



/****************************************************************************
*
*       Function		printLocusHaplotypes  
*       Parameters		haplotypesFile, pedigree, numPersons  	
*									locus, locusNameMaxLength
*		Output			none
*       Description		prints resulted exact pedigree haplotypes for the given locus
*       Comments		is used by "printHaplotypes" function
*
*****************************************************************************/
void printLocusHaplotypes(FILE* haplotypesFile, ped_node** pedigree, int numPersons, 
						  int locus, int locusNameMaxLength);


/****************************************************************************
*
*       Function		printLineWithRecombinations  
*       Parameters		haplotypesFile, numF, numN	  	
*									recMatrix - matrix with recombinations
*		Output			none
*       Description		prints the line with recombinations signed "X" between 2 adjacent loci  
*       Comments		none
*
*****************************************************************************/
void printLineWithRecombinations(FILE* haplotypesFile,  int numF, int numN, int** recMatrix, int locus,
																		int locusNameMaxLength);

/****************************************************************************
*
*       Function		printResultHaplotypes  
*       Parameters		haplotypesFile, pedigree, numF, numN	  	
*		Output			none
*       Description		prints resulted exact pedigree haplotypes
*       Comments		none
*
*****************************************************************************/
void printResultHaplotypes(FILE* haplotypesFile, ped_node** pedigree,
						   int numF,int numN, int** recMatrix);



/****************************************************************************
*
*       Function		findRecombinations  
*       Parameters		pedigree, numF, numN	  	
*		Output			matrix with recombinations
*       Description		computes the matrix with recombinations
*       Comments		none
*
*****************************************************************************/
int** findRecombinations(ped_node** pedigree,int numF,int numN);


/****************************************************************************
*
*       Function		destroyRecMatrix  
*       Parameters		recMatrix,numN	  	
*		Output			none
*       Description		destroys the matrix with recombinations
*       Comments		none
*
*****************************************************************************/
void destroyRecMatrix(int** recMatrix,int numN);



/****************************************************************************
*
*       Function		printInput  
*       Parameters		pedigree
*						numF, numN
*		Output			none
*       Description		prints initial unordered haplotypes and the most 
*						probable vectors at each locus
*       Comments		none
*
*****************************************************************************/
void printInput(ped_node** pedigree,int numF,int numN);







#endif




