/* $Id: haplotypingSimple.c,v 1.2 2008/08/03 08:15:47 superlink Exp $ */
/****************************************************************************
 ************** haplotypingSimple.c *****************************************
 ****************************************************************************
 ************** created by Anisenia Andrei **********************************/

#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "haplotypingSimple.h"
#include "haplotypingAlleles.h"
#include "haplotypingLevelsTree.h"
#include "haplotypingReversedTree.h"
#include "haplotypingTreesAlgorithms.h"



/****************************************************************************
*
*       Function			haplotypeCalculationBackwardPhase
*       Parameters	  	backwardPhaseIterationInput	
*       Output				none 
*       Description		calculates haplotype by using NumOfLocuses as
*									number of HMM nodes, PedigreeData.Possibilities as number of 
*									hidden states (=vectors), ProbabilityLocus as 
*									emissions matrix
*       Comments			implements only the backward phase of Viterbi algorithm
*														
*
*****************************************************************************/
void haplotypeCalculationBackwardPhase(BackwardPhaseIterationInput* backwardPhaseIterationInput) {
	int ** backPointers;
	int i;
	double* probsArray;
	
	#if (PRINT_HAPLO_TIME == 1) 
		fprintf(haploFile, "Start writing = %lf\t\t", ((double)clock())/(double)CLOCKS_PER_SEC);	
	#endif

	/*	if the size of inheritance vectors is non-positive => return*/
	if (PedigreeData.Possibilities <= 1)
	{
		return;
	}
	size = PedigreeData.SpeedUpNumOfMeioses;

	/*	allocating matrix backPointers of size (NumInfMarkers-1)*PedigreeData.Possibilities	*/
	sl_myMalloc(backPointers, ((NumInfMarkers-1)*sizeof(int*)) );
	
	/*	initialization of powersOf2 array*/
	powers2 = initiatePowersOf2(size+1);

	/*	performing backward phase of Viterbi */	
	probsArray = algorithmChooser(backPointers, backwardPhaseIterationInput->haplotypingProbs, NULL);

	//	Storing the results (resulting probabilities vector and back pointers at the first locus)
	//	into the backwardPhaseIterationInput
	if (backwardPhaseIterationInput->haplotypingBackPointers != NULL) {
		free(backwardPhaseIterationInput->haplotypingBackPointers);
	}
	backwardPhaseIterationInput->haplotypingBackPointers = backPointers[0];
	backPointers[0] = NULL;
	if (backwardPhaseIterationInput->haplotypingProbs != NULL) {
		free(backwardPhaseIterationInput->haplotypingProbs);
	}
	backwardPhaseIterationInput->haplotypingProbs = probsArray;
	
	free(powers2);
	deleteBackPointers(backPointers);
	deleteNonZeroMarkers();

	#if (PRINT_HAPLO_TIME == 1)
		fprintf(haploFile, "End writing = %lf\n", ((double)clock())/(double)CLOCKS_PER_SEC);	
	#endif
}




/****************************************************************************
*
*       Function		haplotypeCalculationForwardPhase  
*       Parameters	  	leftLocusBackPointers - back pointers at the first window locus
*									maxVectorLeftLocus - max probability Viterbi vector at the first window locus
*									rightPriorProbs - haplotyping probabilities vector at the end locus
*       Output			the probability of the most probable vectors path    
*       Description		calculates haplotype by using NumInfMarkers as
*						number of HMM nodes, PedigreeData.Possibilities as number of 
*						hidden states (=vectors), ProbabilityLocus as 
*						emissions matrix
*       Comments		implements Viterbi algorithm in O( m*(2n-f)*2^(2n-f) )
*
*****************************************************************************/
double haplotypeCalculationForwardPhase(int* leftLocusBackPointers, 
			int* maxVectorLeftLocus, double* rightPriorProbs)
{
	int ** backPointers;
	int max;
	double* probsArray;
	double probMostProbablePath;

	#if (PRINT_HAPLO_TIME == 1) 
		fprintf(haploFile, "Start writing = %lf\t\t", ((double)clock())/(double)CLOCKS_PER_SEC);	
	#endif

	/*	if the size of inheritance vectors is non-positive => return*/
	if (PedigreeData.Possibilities <= 1)
	{
		return -1;
	}
	size = PedigreeData.SpeedUpNumOfMeioses;
	
	/*	allocating matrix backPointers of size (NumInfMarkers-1)*PedigreeData.Possibilities	*/
	sl_myMalloc(backPointers, ((NumInfMarkers-1)*sizeof(int*)) );
	/*	initialization of powersOf2 array*/
	powers2 = initiatePowersOf2(size+1);
	/*	performing backward phase of Viterbi */	
	probsArray = algorithmChooser(backPointers, rightPriorProbs, leftLocusBackPointers);
	if (leftLocusBackPointers == NULL) {
		max = findMaxVectorInArray(probsArray);
	}
	else {
		max = *maxVectorLeftLocus;
	}
	probMostProbablePath = probsArray[max];
	free(powers2);

	/*	performing forward phase of Viterbi	*/
	findMaxPath(max, backPointers, maxVectorLeftLocus);

	if (leftLocusBackPointers != NULL) {
		backPointers[0] = NULL;
	}
	deleteBackPointers(backPointers);
	free(probsArray);
	deleteNonZeroMarkers();

	#if (PRINT_HAPLO_TIME == 1)
		fprintf(haploFile, "End writing = %lf\n", ((double)clock())/(double)CLOCKS_PER_SEC);	
	#endif

	return probMostProbablePath;
}



/****************************************************************************
*
*       Function			algorithmChooser  
*       Parameters	  	backPointers, haplotypingRightPriorProbs, leftLocusBackPointers
*       Output				probabilities at the first locus    
*       Description		chooses and operates the appropriate algorithm for 
*									calculating backward step from locus = "locus" to 
*									locus = "locus"-1
*       Comments			none
*
*****************************************************************************/
double* algorithmChooser(int** backPointers, double* haplotypingRightPriorProbs, int* leftLocusBackPointers)
{
	VECTOR prevStepProbsArray;
	int i, numVectorsPrevLocus, numVectorsCurrLocus, firstLocus;
	levelsTree* prevStepProbsTree;
	reversedTree* thisStepProbsTree;
	nonZeroMarker* currNonZeroMarker;

	/*	initiating currNonZeroMarker and prevStepProbsArray	*/
	currNonZeroMarker = NonZeroMarkersList; 
	sl_myMalloc(prevStepProbsArray,PedigreeData.Possibilities*sizeof(double));
	if (haplotypingRightPriorProbs != NULL) {
		copyVector(haplotypingRightPriorProbs, prevStepProbsArray);
	}
	else {
		copyVector(ProbabilityLocus[currNonZeroMarker->probabilityLocusIndex],
				prevStepProbsArray);
	}
	if (leftLocusBackPointers == NULL) {
		firstLocus = 1;
	}
	else {
		firstLocus = 2;
		backPointers[0] = leftLocusBackPointers;
	}

	for (i = NumInfMarkers - 1; i >= firstLocus; i--)
	{
		numVectorsPrevLocus = currNonZeroMarker->numNonZeroVectors;
		currNonZeroMarker = currNonZeroMarker->prev;
		numVectorsCurrLocus = currNonZeroMarker->numNonZeroVectors;
		normalizeProbsArray(prevStepProbsArray);

		/*	choosing the algorithm to operate	*/
		if (numVectorsPrevLocus <= PedigreeData.Possibilities/6)
		{
			prevStepProbsTree = levelsTreeConstruction(prevStepProbsArray,
									size,numVectorsPrevLocus);
			if (prevStepProbsTree != NULL)
			{
				copyVector(ProbabilityLocus[currNonZeroMarker->probabilityLocusIndex],
						prevStepProbsArray);
				sl_myMalloc(backPointers[i-1], PedigreeData.Possibilities*sizeof(int));
				if ( (numVectorsPrevLocus <= PedigreeData.Possibilities/8) &&
					(numVectorsCurrLocus <= PedigreeData.Possibilities/8))
				{
					/*	algorithm: transition from tree to tree	*/
					thisStepProbsTree = reversedTreeConstruction(prevStepProbsArray,
									size,numVectorsCurrLocus);
					prevStepProbsArray = treeToTreeCalculation(prevStepProbsArray, 
												thisStepProbsTree,prevStepProbsTree,
												backPointers[i-1], Recomb_values[MarkerOrder[i-1]]);
					destroyReversedTree(thisStepProbsTree);
				}
				else
				{
					/*	algorithm: transition from array to tree	*/
					prevStepProbsArray = arrayToTreeCalculation(prevStepProbsArray,
								prevStepProbsTree,backPointers[i-1], Recomb_values[MarkerOrder[i-1]]);
				}
				destroyLevelsTree(prevStepProbsTree);
				continue;
			}
		}
		if (numVectorsCurrLocus <= PedigreeData.Possibilities/1024)
		{
			/*	algorithm: transition from tree to array	*/
			thisStepProbsTree = reversedTreeConstruction(
							ProbabilityLocus[currNonZeroMarker->probabilityLocusIndex],
							size,numVectorsCurrLocus);
			prevStepProbsArray = treeToArrayCalculation(thisStepProbsTree,
								prevStepProbsArray, backPointers,i-1, Recomb_values[MarkerOrder[i-1]]);
			MultDiagonalMatrix(prevStepProbsArray,
							ProbabilityLocus[currNonZeroMarker->probabilityLocusIndex],
							prevStepProbsArray);
			destroyReversedTree(thisStepProbsTree);
			continue;	
		}
		/*	algorithm: transition from array to array	*/
		prevStepProbsArray = arrayToArrayCalculation(prevStepProbsArray, backPointers,
										i-1, Recomb_values[MarkerOrder[i-1]]);
		MultDiagonalMatrix(prevStepProbsArray,
							ProbabilityLocus[currNonZeroMarker->probabilityLocusIndex],
							prevStepProbsArray);
	}
	
	return prevStepProbsArray;
}




/****************************************************/
/*	help functions for algorithmChooser() function	*/

int findMaxVectorInArray(VECTOR prevStepProbsArray)
{
	int max,i;
	double res;
	double* result = &res;

	*result = *prevStepProbsArray;
	max = 0;
	for(i=0; i < PedigreeData.Possibilities; i++, prevStepProbsArray++)
	{
		if (*result < *prevStepProbsArray)
		{
			*result = *prevStepProbsArray;
			max = i;
		}
	}
	return max;
}



void deleteNonZeroMarkers()
{
	nonZeroMarker* tmp;

	tmp = NonZeroMarkersList;
	while(tmp != NULL)
	{
		NonZeroMarkersList = tmp->prev;
		free(tmp);
		tmp = NonZeroMarkersList;
	}
}

void normalizeProbsArray(VECTOR prevStepProbsArray) {
	int i;
	double divider, *probPtr;

	for(i = 0, divider = 0, probPtr = prevStepProbsArray; i < PedigreeData.Possibilities; i++, probPtr++) {
		if (*probPtr != 0) {
			/*	probability of the current vector is non-zero => should be normalized	*/
			if (divider == 0) {
				/*	updating divider for the first time	*/
				divider = *probPtr;
				/*	normalizing vector's probability	*/
				*probPtr = 1;
			}
			else {
				/*	normalizing vector's probability without updating divider	*/
				*probPtr /= divider;
			}
		}		
	}

}


/****************************************************************************
*
*       Function		findMaxPath  
*       Parameters	  	maxVector - the result of the last step of backward phase
*									backPointers
*									maxVectorRightLocusForUpdate - vector to be updated after finding the max path
*       Output			none  
*       Description		traces back the most probable path using backPointers matrix 
*       Comments		none
*
*****************************************************************************/
void findMaxPath(int maxVector,int** backPointers, int* maxVectorRightLocusForUpdate)
{
	int i;

	/*	tracing back the most probable path using backPointers matrix	*/
	haplotypeVectors[0] = maxVector;
	for(i=0; i <= NumInfMarkers-2 ; i++)
	{
		maxVector = backPointers[i][maxVector];
		haplotypeVectors[i+1] = maxVector;
	}
	*maxVectorRightLocusForUpdate = maxVector;
}



/****************************************************************************
*
*       Function		haplotyping  
*       Parameters	  	none
*       Output			none  
*       Description		main function in haplotyping 
*       Comments		none
*
*****************************************************************************/
void haplotyping()
{
	double haplo;
	FILE* haplotypesFile;
	
	haplo = haplotypeCalculationForwardPhase(NULL, NULL, NULL);
	haplotypesFile = fopen(HAPLOTYPES_FILE_NAME, "w");
	if (haplotypesFile == NULL)
	{
		printf("File %s could not be opened.", HAPLOTYPES_FILE_NAME);
		exit(1);
	}
	if (haplo >= 0)
	{
		if (haplo > 0)
		{
			//setHaplotypeAlleles(haplotypesFile, 0);
		}
		else
		{
			printf("The log probability of the most probable path is: -infinity\n");
			printf("Haplotyping has no considerable result\n");
			fprintf(haplotypesFile,
				"The log probability of the most probable path is: -infinity\n");
			fprintf(haplotypesFile,"Haplotyping has no considerable result\n");
		}
	}
	else
	{
		printf("No information to perform haplotyping\n");
		fprintf(haplotypesFile,"No information to perform haplotyping\n");
	}
	fclose(haplotypesFile);
}



/****************************************************************************
*
*       Function		deleteBackPointers  
*       Parameters	  	backPointers
*       Output			none  
*       Description		deletes backPointers
*       Comments		none
*
*****************************************************************************/
void deleteBackPointers(int** backPointers) {
	int i;

	for(i = 0; i < NumInfMarkers-1; i++) {
		if (backPointers[i] != NULL) {
			free(backPointers[i]);
			backPointers[i] = NULL;
		}
	}
	free(backPointers);
}


