/* $Id: initiatePedigree.c,v 1.2 2008/08/03 08:15:48 superlink Exp $ */
/****************************************************************************/
/*  initiatePedigree.c									   					*/
/*  initiates pedigree structure										 	*/
/*  Author: Andrei Anisenia													*/
/****************************************************************************/

#include "initiatePedigree.h"




/****************************************************************************
*
*       Function		InitSortedPedigree
*       Parameters	  	numLoci - number of locuses
*       Output			initiated sorted pedgree    
*       Description		intiates in each node in "ped" the next fields:
*						"af", "mother" and "father" according to the 
*						information in Pedigree
*       Comments		none
*
*****************************************************************************/
SortedPedigree* InitSortedPedigree(int numLoci)
{
	SortedPedigree* sortedPed;

	sl_myMalloc(sortedPed, sizeof(SortedPedigree));
	sortedPed->pedigree = 
		InitPed( &(sortedPed->numF), &(sortedPed->numN), numLoci);
	return sortedPed;
}


/****************************************************************************
*
*       Function		destroySortedPedigree
*       Parameters	  	sortedPed - pedigree to destroy	
*       Output			none
*       Description		deletes sortedPed
*       Comments		none
*
*****************************************************************************/
void destroySortedPedigree(SortedPedigree* sortedPed)
{
	destroyPedigree(sortedPed->pedigree, sortedPed->numF, sortedPed->numN);
	free(sortedPed);
}



/****************************************************************************
*
*       Function		getRealFounders
*       Parameters	  	f - number of not childless founders	
*       Output			array of indexes of founders in PedigreeGraph   
*       Description		these are indexes of founders which have children
*						(not childless founders)
*       Comments		none
*
*****************************************************************************/
int* getRealFounders(int f)
{
	int* Real, i, ind;
	UnlinkedFounder* unf;

	Real = (int*)malloc(f*sizeof(int));
	ind = 0;
	for(i=0; i<AllPedigreesList->numOfFounders; i++)
	{
		unf = unlinkedFounders;
		while(unf != NULL)
		{
			if (unf->founder == PedigreeData.Founders[i])
				break;
			else unf = unf->next;
		}
		if (unf == NULL)
		{
			Real[ind] = PedigreeData.Founders[i];
			ind++;
		}
	}
	return Real;
}


/****************************************************************************
*
*       Function		InitPed
*       Parameters	  	f - pointer to int (number of founders)
*						n - pointer to int (number of nonfounders)
*						numLoci - number of locuses
*       Output			initiated pedgree    
*       Description		intiates in each node in "ped" the next fields:
*						"af", "mother" and "father" according to the 
*						information in Pedigree
*       Comments		none
*
*****************************************************************************/
ped_node** InitPed(int* found, int* nonfound, int numLoci)
{
	Pedigree* Ped;
	ped_node** pedig;
	int* help, *RealFounders;
	/*	i - index in "ped"; P_index - index in "Ped"						*/
	int i, P_index;
	int founders, nonfounders; 

	Ped = AllPedigreesList->PedigeeGraph;
	founders = AllPedigreesList->numOfFounders - PedigreeData.ChildlessFounders;
	RealFounders = getRealFounders(founders);
	nonfounders = AllPedigreesList->numOfNonFounders;
	help = (int*)malloc((AllPedigreesList->size)*sizeof(int));
	/*	allocating "ped" array												*/
	pedig = (ped_node**)malloc((founders+nonfounders)*sizeof(ped_node*));
	for(i=0; i<founders+nonfounders; i++)
		pedig[i] =(ped_node*)malloc(sizeof(ped_node));
	/*	initiation for founders	and nonfounders								*/
	for(i=0; i<founders+nonfounders; i++)
	{
		if (i<founders)
		{
			P_index = RealFounders[i];
			pedig[i]->father = -1;
			pedig[i]->mother = -1;
		}
		else{
			P_index = PedigreeData.NonFounders[i-founders];
			pedig[i]->father = help[Ped[P_index].father-1];
			pedig[i]->mother = help[Ped[P_index].mother-1];
		}
		pedig[i]->PedIndex = P_index;
		if (numLoci == 0)
		{
			pedig[i]->myAlleles = NULL;
		}
		else
		{
			sl_myMalloc(pedig[i]->myAlleles, (numLoci*sizeof(ped_locus)) );
		}
		/*	if the person is affected										*/
		if (Ped[P_index].myLocus->allele1 == 2)
				pedig[i]->af = true;
		/*	if the person is not affected									*/
		else	pedig[i]->af = false;
		help[P_index] = i;
	}
	*found = founders;
	*nonfound = nonfounders;
	free(help);
	free(RealFounders);
	return pedig;
}




/****************************************************************************
*
*       Function		deleteUnlinkedFounders
*       Parameters	  	none	
*       Output			none
*       Description		deletes list of childless founders
*       Comments		none
*
*****************************************************************************/
void deleteUnlinkedFounders()
{
	UnlinkedFounder* unf = unlinkedFounders;
	while(unf != NULL)
	{
		unf = unf->next;
		unlinkedFounders->next = NULL;
		free(unlinkedFounders);
		unlinkedFounders = unf;
	}
}



/****************************************************************************
*
*       Function		destroyPedigree
*       Parameters	  	ped - pedigree to destroy	
*       Output			none
*       Description		deletes ped
*       Comments		none
*
*****************************************************************************/
void destroyPedigree(ped_node** ped,int numF, int numN)
{
	int i;

	for(i=0; i<numF+numN; i++)
	{
		if (ped[i]->myAlleles != NULL)
		{
			free(ped[i]->myAlleles);
		}
		free(ped[i]);
	}
	free(ped);
}




/****************************************************************************
*
*       Function		FirstChildren  
*       Parameters		firstChild - array to be full
*						numF,size
*		Output			array firstChildren
*       Description		creates and initiates first children 
*       Comments		none
*
*****************************************************************************/
Bool* FirstChildren(int numF, int size)
{
	Bool* firstChild;

	sl_myMalloc(firstChild, size*sizeof(Bool));
	reinitFirstChildren(firstChild,numF,size);
	return firstChild;
}




/****************************************************************************
*
*       Function		reinitFirstChildren  
*       Parameters		firstChild - array to be full
*						numF,size
*		Output			none
*       Description		reinitiates first children 
*       Comments		none
*
*****************************************************************************/
void reinitFirstChildren(Bool* firstChild, int numF, int size)
{
	int person;

	for(person = 0; person < size; person++)
	{
		if (person < numF)
		{
			firstChild[person] = true;
		}
		else
		{
			firstChild[person] = false;
		}
	}
}




/****************************************************************************
*
*       Function		initiatePowersOf2  
*       Parameters	  	size
*       Output			initiated array of size of powers of 2    
*       Description		initiates powersOf2 array
*       Comments		none
*
*****************************************************************************/
int* initiatePowersOf2(int size)
{
	int tmp,i;
	int* powers;

	sl_myMalloc(powers, (size*sizeof(int)) );
	tmp = 1;
	for(i=0; i<size; i++)
	{
		powers[i] = tmp;
		tmp = tmp*2;
	}
	return powers;
}



/****************************************************************************
*
*       Function		initiateFoundersIBDalleles  
*       Parameters	  	ped
*       Output			number of IBD alleles    
*       Description		parses IBD alleles between founders
*       Comments		none
*
*****************************************************************************/
int initiateFoundersIBDalleles(SortedPedigree* sortedPed)
{
	int numIBDalleles, numFounders, i;
	Pedigree* Ped;
	ped_node** pedigree;

	/*	initiations	*/
	numIBDalleles = 0;
	numFounders = sortedPed->numF;
	Ped = AllPedigreesList->PedigeeGraph;
	pedigree = sortedPed->pedigree;

	/*	parsing IBD alleles (starting 0) between founders	*/
	for(i = 0; i < numFounders; ++i)
	{
		/*	for each founder:	*/
		/*	check that his paternal allele is used	*/
		if (Ped[pedigree[i]->PedIndex].paternalAllele->bitType != NotUsedBit)
		{	
			pedigree[i]->allele1 = numIBDalleles;
			numIBDalleles++;
		}
		else
		{
			pedigree[i]->allele1 = ALLELE_DOESNT_EXIST;
		}
		/*	check that his maternal allele is used	*/
		if (Ped[pedigree[i]->PedIndex].maternalAllele->bitType != NotUsedBit)
		{	
			pedigree[i]->allele2 = numIBDalleles;
			numIBDalleles++;
		}
		else
		{
			pedigree[i]->allele2 = ALLELE_DOESNT_EXIST;
		}
	}

	return numIBDalleles;
}



/****************************************************************************
*
*       Function		initiateNonFoundersAlleles  
*       Parameters	  	ped , vectorIndex
*       Output			none    
*       Description		parses founder alleles between non-founders using 
*						given inheritance vector
*       Comments		none
*
*****************************************************************************/
void initiateNonFoundersAlleles(SortedPedigree* sortedPed, int vectorIndex)
{
	int* vector, currNonFounder, numFounders, numNonFounders;
	Pedigree* Ped;
	ped_node** pedigree;

	/*	checking the validity of vector	*/
	if ( (vectorIndex < 0) || (vectorIndex >= PedigreeData.Possibilities) )
	{
		return;
	}
	vector = getInheritanceVector(vectorIndex);
	Ped = AllPedigreesList->PedigeeGraph;
	numFounders = sortedPed->numF;
	numNonFounders = sortedPed->numN;
	pedigree = sortedPed->pedigree;
	
	/*	going through all non-founders and updating alleles 1/2	*/
	for(currNonFounder = numFounders; 
			currNonFounder < (numFounders + numNonFounders); ++currNonFounder)
	{
		/*	treating paternal allele of current non-founder	*/
		pedigree[currNonFounder]->allele1 = 
			getParentalAllele(Ped[pedigree[currNonFounder]->PedIndex].paternalAllele,
							pedigree[pedigree[currNonFounder]->father], vector);
		/*	treating maternal allele of current non-founder	*/
		pedigree[currNonFounder]->allele2 = 
			getParentalAllele(Ped[pedigree[currNonFounder]->PedIndex].maternalAllele,
							pedigree[pedigree[currNonFounder]->mother], vector);
	}

	free(vector);
}


// help function of "initiateNonFoundersAlleles"
int getParentalAllele(AlleleData* parentalBit, ped_node* parent, int* vector)
{
	int allele;

	switch(parentalBit->bitType)
	{
		case RegularBit:	(vector[parentalBit->number] == 0) ? 
								(allele = parent->allele1) : (allele = parent->allele2);
							break;
		case ReductionBit:	allele = parent->allele1;
							break;
		case ConstBit:		(parentalBit->value == 0) ? 
								(allele = parent->allele1) : (allele = parent->allele2);
							break;
		case NotUsedBit:	allele = ALLELE_DOESNT_EXIST;
							break;
		default:			break;
	}

	return allele;
}


/****************************************************************************
*
*       Function		getInheritanceVector  
*       Parameters	  	vectorIndex
*       Output			vector array    
*       Description		getting vector from vector index
*       Comments		none
*
*****************************************************************************/
int* getInheritanceVector(int vectorIndex)
{
	int* vector, i;

	sl_myMalloc(vector, (PedigreeData.SpeedUpNumOfMeioses)*sizeof(int));
	for(i = 0; i < (PedigreeData.SpeedUpNumOfMeioses); ++i)
	{
		vector[i] = vectorIndex % 2;
		vectorIndex  = vectorIndex >> 1;
	}
	return vector;
}


/****************************************************************************
*
*       Function		setInheritanceVector  
*       Parameters	  	vectorPtr - pointer to vector, vectorIndex
*       Output			none    
*       Description		getting vector from vector index and setting it to 
*						vectorPtr
*       Comments		assuming vectorPtr array's size is equal to 
*						"PedigreeData.SpeedUpNumOfMeioses"
*
*****************************************************************************/
void setInheritanceVector(int* vectorPtr, int vectorIndex)
{
	int i;

	for(i = 0; i < (PedigreeData.SpeedUpNumOfMeioses); ++i)
	{
		vectorPtr[i] = vectorIndex % 2;
		vectorIndex  = vectorIndex >> 1;
	}
}


