/* $Id: parseChildrenLists.c,v 1.2 2008/08/03 08:15:48 superlink Exp $ */
/****************************************************************************/
/*  parseChildrenLists.c					   								*/
/*  for each parent in the pedigree (father or mother) sort his chidren	 	*/
/*	in increasing order of indexes and set his first child to be the child	*/
/*  with the minimal index (between all his children)						*/
/*  Author: Anisenia Andrei													*/
/*  Last updated: 09.02.07													*/
/****************************************************************************/
#include "gdata.h"
#include "sl_Auxilary.h"

//	Functions for the external use
void getChidrenForParents(Pedigree* Ped, int pedSize);

void sortChildrenIncreasingIndexes();





/****************************************************************************/
/* internal definitions														*/
#define GET_NEXT_SIBLING(currentSiblingPtr,parentSex)	\
	((parentSex == Male) ? currentSiblingPtr->NeSiPa : currentSiblingPtr->NeSiMa)

#define SET_NEXT_SIBLING(currentSiblingPtr,parentSex,nextSiblingIndex)	\
	((parentSex == Male) ? (currentSiblingPtr->NeSiPa = nextSiblingIndex) :	\
	(currentSiblingPtr->NeSiMa = nextSiblingIndex))



/* internal functions														*/
int sortParentChildren(int firstChild, Gender parentSex, int* arrayIndexes);
void bubbleSort(int* arr, int size); 

int getParentIndex(Pedigree* personPtr, Gender parentSex);
Bool* setParentsIndication(Pedigree* Ped, int pedSize);
void getChidrenForParent(int parentIndex, Pedigree* Ped, int pedSize);
void setNextSibling(Pedigree* personPtr, int nextSiblingIndex, Gender parentSex);
void setChild(Pedigree* personPtr, int childIndex);
void clearChidrenLists(Pedigree* Ped, int pedSize);


/****************************************************************************/
/* Implementation															*/





/****************************************************************************
*
*       Function		getChidrenForParents  
*       Parameters		Ped - pedigree, pedSize - size of the pedigree
*		Output			none
*       Description		sets fields "child", "NeSiPa" and "NeSiMa" using only
*						"index", "father" and "mother" fields of the persons in the pedigree
*       Comments		there may be many errors in pedigree structure in the pedfile,
*						but if the indexes of the persons (field "index") 
*						and links of children to parents ("father" and "mother" fields)
*						are valid then the links of the parent to his first child
*						and sibling links may be computed automatically 
*						without using pedfile information (which may contain errors)
*
*****************************************************************************/
void getChidrenForParents(Pedigree* Ped, int pedSize) {
	int personIndex;
	Bool* isParent;	//	is used to distinct between parents and no-parents in the pedigree

	clearChidrenLists(Ped, pedSize);
	isParent = setParentsIndication(Ped, pedSize);
	//	Go over all people in the pedigree
	for(personIndex = 0; personIndex < pedSize; personIndex++) {
		//	For each parent create children list
		if (isParent[personIndex]) {
			getChidrenForParent(personIndex, Ped, pedSize);
		}
	}
	free(isParent);
}



/****************************************************************************
*
*       Function		getChidrenForParent  
*       Parameters		personIndex - index of the treated parent in the pedigree
*						Ped - pedigree, pedSize - size of the pedigree
*		Output			none
*       Description		sets children list of the parent ("child" field of the parent
						and "NeSiPa"/"NeSiMa" fields of his children)
*       Comments		none
*
*****************************************************************************/
void getChidrenForParent(int parentIndex, Pedigree* Ped, int pedSize) {
	int lastTreatedChildIndex; 	// To keep the last node in the list
	int currPersonIndex;
	Gender parentSex;			//	Sex of the current parent

	//	Initiation 
	lastTreatedChildIndex = -1;
	parentSex = Ped[parentIndex].sex;

	//	Go over all people in the pedigree
	for(currPersonIndex = 0; currPersonIndex < pedSize; currPersonIndex++) {
		//	Find parent's children in the pedigree:
		if (getParentIndex((&Ped[currPersonIndex]), parentSex) == parentIndex) {
			//	Add current person to the children list of the treated parent
			if (lastTreatedChildIndex == -1) {
				//	Add current person as the first child in the parent's children list
				setChild((&Ped[parentIndex]), currPersonIndex);
			}
			else {
				//	Add current person to the end in the parent's children list
				setNextSibling((&Ped[lastTreatedChildIndex]), currPersonIndex, parentSex);
			}
			lastTreatedChildIndex = currPersonIndex;
		}
	}
}


//	Set "child", "NeSiPa" and "NeSiMa" fields to 0 for all persons in the pedigree
void clearChidrenLists(Pedigree* Ped, int pedSize) {
	int i;
	for(i = 0; i < pedSize; i++) {
		Ped[i].child = 0;
		Ped[i].NeSiPa = 0;
		Ped[i].NeSiMa = 0;
	}
}



//	Allocation and initiation of the array isParent
Bool* setParentsIndication(Pedigree* Ped, int pedSize) {
	Bool* isParent;
	int i, fatherIndex, motherIndex;
	
	sl_myMalloc(isParent, pedSize*sizeof(Bool));
	//	set all persons to be "not-parents"
	for(i = 0; i < pedSize; i++) {
		isParent[i] = false;
	}
	//	for each person in the pedigree
	for(i = 0; i < pedSize; i++) {
		//	set the person's parents to be "parent"
		fatherIndex = getParentIndex((&Ped[i]), Male);
		motherIndex = getParentIndex((&Ped[i]), Female);
		if (fatherIndex >= 0) {
			isParent[fatherIndex] = true;
			isParent[motherIndex] = true;
		}
	}
	return isParent;
}


//	Getting the parent's index of the person from the pedigree: 
int getParentIndex(Pedigree* personPtr, Gender parentSex) {
	int parentIndex = (parentSex == Male) ? (personPtr->father - 1) : (personPtr->mother - 1);
	return parentIndex;
}


//	Setting the person's next sibling in the pedigree: 
void setNextSibling(Pedigree* personPtr, int nextSiblingIndex, Gender parentSex) {
	(parentSex == Male) ? 
		(personPtr->NeSiPa = nextSiblingIndex + 1) : (personPtr->NeSiMa = nextSiblingIndex + 1);
}


//	Setting the person's child in the pedigree
void setChild(Pedigree* personPtr, int childIndex) {
	personPtr->child = childIndex + 1;
}




/****************************************************************************
*
*       Function		sortChildrenIncreasingIndexes  
*       Parameters		none 	
*		Output			none
*       Description		for each parent in the pedigree (father or mother) sort his chidren
*						in increasing order of indexes and set his first child to be the child
*						with the minimal index (between all his children)
*       Comments		uses sortParentChildren function
*
*****************************************************************************/
void sortChildrenIncreasingIndexes() {
	int i;
	int *arrayIndexes;
	Pedigree* currPerson, *Ped;

	/*	initiations	*/
	sl_myMalloc(arrayIndexes, (AllPedigreesList->size + 1)*sizeof(int));
	Ped = AllPedigreesList->PedigeeGraph;

	/*	going through all persons in the pedigree	*/
	for(i = 0, currPerson = Ped; i < AllPedigreesList->size; ++i, ++currPerson) {
		/*	if current person is a parent => sort his chidren		*/
		/*	in increasing order of indexes and set his first child	*/
		/*	to be with the minimal index (between his children)		*/
		if (currPerson->child) {
			currPerson->child = 
				sortParentChildren(currPerson->child, currPerson->sex, arrayIndexes);
		}
	}
	free(arrayIndexes);
}



/****************************************************************************
*
*       Function		sortParentChildren  
*       Parameters		firstChildIndex - index of the first child of the parent 	
*						parentSex - Male or Female
*		Output			minimal index between children indexes	
*       Description		sorts list of children (starting firstChildIndex) 
*						in the increasing order of their indexes
*       Comments		is used by sortChildrenIncreasingIndexes function
*
*****************************************************************************/
int sortParentChildren(int firstChildIndex, Gender parentSex, int* arrayIndexes) {
	int *currArrayCell, *thisSibling;
	int childIndex, arraySize, i, nextSiblingIndex;
	Pedigree* currChild, *Ped;

	/*	initiations	*/
	Ped = AllPedigreesList->PedigeeGraph;
	arraySize = 0;

	/*	go through children list starting firstChildIndex	*/
	/*	and saving their indexes in array					*/
	for(childIndex = firstChildIndex, currArrayCell = arrayIndexes;
		childIndex > 0; ++currArrayCell) {
		/*	saving index in array	*/
		*currArrayCell = childIndex;
		arraySize++;
		/*	going to the next child	(sibling of the current child)	*/
		currChild = &(Ped[childIndex-1]);
		childIndex = GET_NEXT_SIBLING(currChild, parentSex);
	}
	arrayIndexes[arraySize] = 0;
	/*	sorting children indexes in array	*/
	bubbleSort(arrayIndexes, arraySize);

	/*	copying sorted indexes to the pedigree graph	*/
	for(thisSibling = &(arrayIndexes[0]), i = 0; i < arraySize; ++i, ++thisSibling) {
		currChild = &(Ped[*thisSibling-1]);
		nextSiblingIndex = *(thisSibling+1);
		SET_NEXT_SIBLING(currChild,parentSex,nextSiblingIndex);
	}
	return arrayIndexes[0];
}



/****************************************************************************
*
*       Function		bubbleSort  
*       Parameters		arr - array of integers to be sorted	
*						size - size of arr
*		Output			none
*       Description		bubble sort of arr (increasing order)
*       Comments		none
*
*****************************************************************************/
void bubbleSort(int* arr, int size) {
	int i,j,tmp;

	if ( (size == 0) || (size == 1) ) {
		return;
	}
	for(i = size - 1; i > 0; --i) {
		for(j = 0; j < i; ++j) {
			if (arr[j+1] < arr[j]) {
				tmp = arr[j+1];
				arr[j+1] = arr[j];
				arr[j] = tmp;
			}
		}
	}
}








