/* $Id: restrictionsGraph.c,v 1.2 2008/08/03 08:15:48 superlink Exp $ */
/********************************************************************************/
/*  restrictionsGraph.c									   						*/
/*  RestrictionsGraph : given const set of vertices and changable (addible)		*/
/*	set of edges when each edge carry a restriction of type (number1, number2)	*/
/*	try to satisfy all number restrictions of edges using vertices,				*/
/*	which means give to each vertex some number such that for each edge			*/
/*	edge = (vertex1,vertex2) the following will hold:							*/	 
/*	{vertex1.number, vertex2.number} = {edge.number1, edge.number2}				*/
/*	Each restriction has some double score, so restriction graph tries			*/
/*	to satisfy all restrictions in the way that the multiplication of scores	*/
/*	of all numbers at vertices will be maximal									*/
/*  Author: Anisenia Andrei														*/
/********************************************************************************/

#include "restrictionsGraph.h"


/****************************************************************************/
/*	TestRestrictionsGraph													*/
void TestRestrictionsGraph()
{
	RestrictionsGraph* graph;
	double* restrictionScores;
	int numVertices, numRestrictions;
	Bool res;

/*	example1:	*
	numVertices = 8;
	numRestrictions = 3;
	sl_myMalloc(restrictionScores,numRestrictions*sizeof(double));
	restrictionScores[0] = 4;
	restrictionScores[1] = 3;
	restrictionScores[2] = 2;
	graph = createRestrictionsGraph(numVertices,restrictionScores,numRestrictions);
	addRestriction(graph, 0,1, 1,3);
	addRestriction(graph, 2,3, 2,3);
	addRestriction(graph, 0,4, 1,3);
	addRestriction(graph, 0,7, 3,2);
	addRestriction(graph, 0,7, 2,3);
	addRestriction(graph, 4,5, 1,2);
*/
	numVertices = 7;
	numRestrictions = 3;
	sl_myMalloc(restrictionScores,numRestrictions*sizeof(double));
	restrictionScores[0] = 4;
	restrictionScores[1] = 3;
	restrictionScores[2] = 2;
	graph = createRestrictionsGraph(numVertices,restrictionScores,numRestrictions);
	addRestriction(graph, 0,0, 3,3);
	addRestriction(graph, 1,2, 2,3);
	addRestriction(graph, 1,3, 1,3);
	addRestriction(graph, 1,5, 3,2);
	addRestriction(graph, 3,3, 1,1);
	addRestriction(graph, 3,4, 1,2);
	//addRestriction(graph, 5,5, 3,3);

	res = satisfyRestrictions(graph);

	destroyRestrictionsGraph(graph);
	free(restrictionScores);
}


/****************************************************************************
*
*       Function		createRestrictionsGraph  
*       Parameters		numVertices, restrictionScores, numRestrictions
*		Output			restrictionGraph
*       Description		creates restriction graph with {0,1,..,(numVertices-1}
*						set of vertices and empty set of restrictions
*       Comments		none
*
*****************************************************************************/
RestrictionsGraph* createRestrictionsGraph(int numVertices, 
								double* restrictionScores, int numRestrictions)
{
	RestrictionsGraph* graph;
	int vertexIndex, restrictionIndex, maxRestrictionIndex;
	RestrictionVertex* currVertex;
	
	/*	initiations	*/
	sl_myMalloc(graph, sizeof(RestrictionsGraph));
	graph->numRestrictions = numRestrictions;
	graph->numVertices = numVertices;
	graph->restrictions = NULL;
	graph->restrictionScores = restrictionScores;
	sl_myMalloc(graph->vertices, numVertices*sizeof(RestrictionVertex));

	/*	initializing vertices	*/
	for(currVertex = graph->vertices, vertexIndex = 0; 
		vertexIndex < numVertices; ++vertexIndex, ++currVertex)
	{
		currVertex->restrictionNumber = -1;
		currVertex->used = false;
		currVertex->vertexRestrictions = NULL;
	}

	/*	finding restrictionWithMaxScore		*/
	maxRestrictionIndex = 0;
	for(restrictionIndex = 1; restrictionIndex < numRestrictions; ++restrictionIndex)
	{
		if (restrictionScores[restrictionIndex] > 
						restrictionScores[maxRestrictionIndex])
		{
			maxRestrictionIndex = restrictionIndex;
		}
	}
	graph->restrictionWithMaxScore = maxRestrictionIndex + 1;
		
	return graph;
}



/****************************************************************************
*
*       Function		addRestriction  
*       Parameters		vertexIndex1,vertexIndex2,restriction1,restriction2
*		Output			false if some of the above numbers is out of scope,
*						else true
*       Description		creates restriction edge and inserts into the graph
*						queue and into 2 queues of 2 above vertices
*       Comments		none
*
*****************************************************************************/
Bool addRestriction(RestrictionsGraph* graph, int vertexIndex1, int vertexIndex2, 
					int restriction1, int restriction2)
{
	RestrictionEdge* newRestriction;

	/*	checking that all numbers are in their scopes	*/	
	if ( isOutOfScope(vertexIndex1, 0, graph->numVertices - 1) ||
		 isOutOfScope(vertexIndex2, 0, graph->numVertices - 1) ||
		 isOutOfScope(restriction1, 1, graph->numRestrictions) ||
		 isOutOfScope(restriction2, 1, graph->numRestrictions) )
	{
		return false;
	}

	/*	creating new restriction edge and adding it	to restriction graph	*/
	sl_myMalloc(newRestriction, sizeof(RestrictionEdge));
	newRestriction->restriction1 = restriction1;
	newRestriction->restriction2 = restriction2;
	newRestriction->vertexIndex1 = vertexIndex1;
	newRestriction->vertexIndex2 = vertexIndex2;
	/*	adding new restriction to graph list	*/
	if (graph->restrictions == NULL)
	{
		graph->restrictions = CreateQueue(newRestriction);
	}
	else
	{
		InsertToTail(graph->restrictions, newRestriction);
	}
	/*	checking if the vertices are the same	*/
	if (vertexIndex1 == vertexIndex2)
	{
		if (restriction1 != restriction2)
		{
			/*	this restriction cannot be satisfied, so we don't add it	*/
			return false;
		}
		else
		{
			/*	adding new restriction to single vertex (v.1 = v.2) list	*/
			addRestrictionToVertex(&(graph->vertices[vertexIndex1]), newRestriction);
		}
	}
	else
	{
		/*	adding new restriction to vertex 1 list	*/
		addRestrictionToVertex(&(graph->vertices[vertexIndex1]), newRestriction);
		/*	adding new restriction to vertex 2 list	*/
		addRestrictionToVertex(&(graph->vertices[vertexIndex2]), newRestriction);
	}

	return true;
}



// help function
void addRestrictionToVertex
		(RestrictionVertex* vertex, RestrictionEdge* newRestriction)
{
	if (vertex->vertexRestrictions == NULL)
	{
		vertex->vertexRestrictions = CreateQueue(newRestriction);
	}
	else
	{
		InsertToTail(vertex->vertexRestrictions, newRestriction);
	}
}



/****************************************************************************
*
*       Function		satisfyRestrictions  
*       Parameters		RestrictionsGraph
*		Output			false if graph restrictions can't be satisfied,
*						else true
*       Description		tries to satisfy restrictions in the "maximal" way,
*						if the restrictions are satisfied, then the
*						"satisfying" numbers are written in vertices
*       Comments		none
*
*****************************************************************************/
Bool satisfyRestrictions(RestrictionsGraph* graph)
{
	int vertexIndex;
	RestrictionVertex* vertex;

	/*	going through all vertices	*/
	vertex = graph->vertices;
	for(vertexIndex = 0; vertexIndex < graph->numVertices; ++vertexIndex, ++vertex)
	{
		if (!vertex->used)
		{
			/*	vertex is not used => treat it's component	*/
			if (vertex->vertexRestrictions == NULL)
			{
				/*	vertex has no restrictions => it's component is empty	*/
				vertex->restrictionNumber = graph->restrictionWithMaxScore;
				vertex->used = true;
			}
			else
			{
				/*	treat vertex + it's component	*/
				if (treatVertex(vertexIndex, graph) == false)
				{
					/*	no satisfaction for current vertex component	*/
					/*	=> no satisfaction for overall graph			*/
					return false;
				}
			}
		}
	}
	return true;
}



/****************************************************************************
*
*       Function		getVertexRestriction  
*       Parameters		vertexIndex
*		Output			restriction number in vertex (-1 if was not initiated
*						or vertex is out of scope)
*       Description		getter
*       Comments		none
*
*****************************************************************************/
int getVertexRestriction(RestrictionsGraph* graph, int vertexIndex)
{
	if ( (vertexIndex < 0) || (vertexIndex >= graph->numVertices) )
	{
		return -1;
	}
	return graph->vertices[vertexIndex].restrictionNumber;
}



/****************************************************************************
*
*       Function		treatVertex  
*       Parameters		vertex index, RestrictionsGraph graph
*		Output			false if vertex component restrictions can't be satisfied,
*						else true
*       Description		tries to satisfy vertex component restrictions 
*						in the "maximal" way,
*						if the restrictions are satisfied, then the
*						"satisfying" numbers are written in component vertices
*       Comments		none
*
*****************************************************************************/
/*
Bool treatVertex(int vertexIndex, RestrictionsGraph* graph)
{
	RestrictionEdge* rEdge;
	double score1, score2;

	/*	using first restriction of current vertex in it's list	*
	rEdge = 
		(RestrictionEdge*)getFirst(graph->vertices[vertexIndex].vertexRestrictions);
	/*	trying first configuration	*
	graph->vertices[vertexIndex].restrictionNumber = rEdge->restriction1;
	graph->vertices[vertexIndex].used = true;
	score1 = treatVertexComponent(vertexIndex, graph);
	/*	checking the result	*
	if (rEdge->restriction1 == rEdge->restriction2)
	{
		/*	only 1 configuration is possible	*
		if (score1 == 0)
		{
			/*	no satisfaction for vertex component	*
			return false;
		}
		else
		{
			/*	maximal satisfaction was saved in vertices array	*
			return true;
		}
	}
	else
	{
		/*	trying second configuration		*
		setComponentUnused(vertexIndex, graph);
		graph->vertices[vertexIndex].restrictionNumber = rEdge->restriction2;
		graph->vertices[vertexIndex].used = true;
		score2 = treatVertexComponent(vertexIndex, graph);
		/*	comparing 2 configurations and choosing maximal		*
		if ( (score2 >= score1) && (score2 != 0) )
		{
			/*	second configuration is maximal and is saved in vertices array	*
			return true;
		}
		else
		{
			/*	first configuration is maximal or there is no satisfaction	*
			if (score1 == 0)
			{
				/*	no satisfaction for vertex component	*
				return false;
			}
			else
			{
				/*	first configuration is maximal	and we need to save it	*
				setComponentUnused(vertexIndex, graph);
				graph->vertices[vertexIndex].restrictionNumber = rEdge->restriction1;
				graph->vertices[vertexIndex].used = true;
				treatVertexComponent(vertexIndex, graph);
				return true;
			}
		}
	}
}
*/



Bool treatVertex(int vertexIndex, RestrictionsGraph* graph)
{
	RestrictionEdge* rEdge;
	double score1, score2;

	/*	using first restriction of current vertex in it's list	*/
	rEdge = 
		(RestrictionEdge*)getFirst(graph->vertices[vertexIndex].vertexRestrictions);
	/*	trying first configuration	*/
	graph->vertices[vertexIndex].restrictionNumber = rEdge->restriction2;
	graph->vertices[vertexIndex].used = true;
	score2 = treatVertexComponent(vertexIndex, graph);
	/*	checking the result	*/
	if (rEdge->restriction1 == rEdge->restriction2)
	{
		/*	only 1 configuration is possible	*/
		if (score2 == 0)
		{
			/*	no satisfaction for vertex component	*/
			return false;
		}
		else
		{
			/*	maximal satisfaction was saved in vertices array	*/
			return true;
		}
	}
	else
	{
		/*	trying second configuration		*/
		setComponentUnused(vertexIndex, graph);
		graph->vertices[vertexIndex].restrictionNumber = rEdge->restriction1;
		graph->vertices[vertexIndex].used = true;
		score1 = treatVertexComponent(vertexIndex, graph);
		/*	comparing 2 configurations and choosing maximal		*/
		if ( (score1 >= score2) && (score1 != 0) )
		{
			/*	first configuration is maximal and is saved in vertices array	*/	
			return true;
		}
		else
		{
			/*	second configuration is maximal or there is no satisfaction	*/
			if (score2 == 0)
			{
				/*	no satisfaction for vertex component	*/
				return false;
			}
			else
			{
				/*	second configuration is maximal	and we need to save it	*/
				setComponentUnused(vertexIndex, graph);
				graph->vertices[vertexIndex].restrictionNumber = rEdge->restriction2;
				graph->vertices[vertexIndex].used = true;
				treatVertexComponent(vertexIndex, graph);
				return true;
			}
		}
	}
}





/****************************************************************************
*
*       Function		treatVertexComponent
*       Parameters		vertex index, RestrictionsGraph graph
*		Output			score for current configuration satisfaction
*       Description		tries to satisfy vertex component restrictions 
*						in the "maximal" way,
*						if the restrictions are satisfied, then the
*						"satisfying" numbers are written in component vertices
*       Comments		assumes some number is assigned to vertex
*
*****************************************************************************/
double treatVertexComponent(int vertexIndex, RestrictionsGraph* graph)
{
	double resultScore;
	RestrictionEdge* rEdge;
	RestrictionVertex* currVertex, *secondVertex;
	int currRestriction, secondRestriction, secondVertexIndex;

	/*	initiations	*/
	currVertex = &(graph->vertices[vertexIndex]);
	currRestriction = currVertex->restrictionNumber;
	resultScore = graph->restrictionScores[currRestriction - 1];
	
	/*	going through all restrictions of currVertex	*/
	restartCurr(currVertex->vertexRestrictions);
	while ( (rEdge = (RestrictionEdge*)getCurr(currVertex->vertexRestrictions)) != NULL)
	{
		/*	find second vertex of current vector in current "rEdge" restriction	*/
		secondVertexIndex = (rEdge->vertexIndex1 == vertexIndex) ? 
								rEdge->vertexIndex2 : rEdge->vertexIndex1;
		secondVertex = &(graph->vertices[secondVertexIndex]);
		/*	check if second vertex is used or not	*/
		if (secondVertex->used)
		{
			/*	second vertex is already used (has restriction number)	*/
			secondRestriction = secondVertex->restrictionNumber;
			/*	check the cosistensy of restriction in current and second vertices	*/
			if ( !( ((currRestriction == rEdge->restriction1) &&
				     (secondRestriction == rEdge->restriction2)) ||
				    ((currRestriction == rEdge->restriction2) &&
				     (secondRestriction == rEdge->restriction1)) 
				   ) 
				)
			{
				/*	restriction is not valid => no satisfaction	*/
				return 0;
			}
		}
		else
		{
			/*	second vertex is not used	*/
			/*	check that current vector number presents in current restriction	*/
			if ( (currRestriction != rEdge->restriction1) &&
				 (currRestriction != rEdge->restriction2) )
			{
				/*	restriction is not valid => no satisfaction	*/
				return 0;
			}
			/*	set second vertex restriction	*/
			secondVertex->restrictionNumber = (currRestriction == rEdge->restriction1) ?
				rEdge->restriction2 : rEdge->restriction1;
			secondVertex->used = true;
			/*	recursive call for second vertex	*/
			resultScore *= treatVertexComponent(secondVertexIndex, graph);
			/*	check if there was no satisfaction	*/
			if (resultScore == 0)
			{
				return 0;
			}
		}
		increaseCurr(currVertex->vertexRestrictions);
	}

	return resultScore;
}



/****************************************************************************
*
*       Function		setComponentUnused
*       Parameters		vertex index, RestrictionsGraph graph
*		Output			none
*       Description		sets all vertices in the component of given vertex unused
*       Comments		none
*
*****************************************************************************/
void setComponentUnused(int vertexIndex, RestrictionsGraph* graph)
{
	RestrictionEdge* rEdge;
	RestrictionVertex* currVertex;
	int secondVertexIndex;

	/*	set current vertex unused	*/
	currVertex = &(graph->vertices[vertexIndex]);
	currVertex->used = false;
	/*	going through all restrictions of currVertex	*/
	restartCurr(currVertex->vertexRestrictions);
	while ( (rEdge = (RestrictionEdge*)getCurr(currVertex->vertexRestrictions)) != NULL)
	{
		/*	find second vertex of current vector in current "rEdge" restriction	*/
		secondVertexIndex = (rEdge->vertexIndex1 == vertexIndex) ? 
								rEdge->vertexIndex2 : rEdge->vertexIndex1;
		if (graph->vertices[secondVertexIndex].used)
		{
			/*	recursive call	*/
			setComponentUnused(secondVertexIndex, graph);
		}
		increaseCurr(currVertex->vertexRestrictions);
	}
}



/****************************************************************************
*
*       Function		destroyRestrictionsGraph  
*       Parameters		RestrictionsGraph
*		Output			none
*       Description		frees all the memory occupied by the RestrictionsGraph
*       Comments		none
*
*****************************************************************************/
void destroyRestrictionsGraph(RestrictionsGraph* graph)
{
	RestrictionEdge* edge;
	RestrictionVertex* vertex;
	int i;

	/*	freeing all restrictions		*/
	if (graph->restrictions != NULL)
	{
		restartCurr(graph->restrictions);
		while( (edge = (RestrictionEdge*)getCurr(graph->restrictions)) != NULL)
		{
			free(edge);
			increaseCurr(graph->restrictions);
		}
		destroyQueue(graph->restrictions);
	}
	for(vertex = graph->vertices, i = 0; i < graph->numVertices; ++i, ++vertex)
	{
		if (vertex->vertexRestrictions != NULL)
		{
			destroyQueue(vertex->vertexRestrictions);
		}
	}
	/*	destroying graph	*/
	free(graph->vertices);
	free(graph);
}


