/* $Id: restrictionsGraph.h,v 1.2 2008/08/03 08:15:48 superlink Exp $ */
/********************************************************************************/
/*  restrictionsGraph.h									   						*/
/*  RestrictionsGraph : given const set of vertices and changable (addible)		*/
/*	set of edges when each edge carry a restriction of type (number1, number2)	*/
/*	try to satisfy all number restrictions of edges using vertices,				*/
/*	which means give to each vertex some number such that for each edge			*/
/*	edge = (vertex1,vertex2) the following will hold:							*/	 
/*	{vertex1.number, vertex2.number} = {edge.number1, edge.number2}				*/
/*	Each restriction has some double score, so restriction graph tries			*/
/*	to satisfy all restrictions in the way that the multiplication of scores	*/
/*	of all numbers at vertices will be maximal									*/
/*  Author: Anisenia Andrei														*/
/********************************************************************************/

#ifndef RESTRICTIONS_GRAPH_H
#define RESTRICTIONS_GRAPH_H

#include "gdata.h"
#include "queue.h"

/********************************************************************************/
/*	Data structure definitions	*/

/*	restriction graph vertices	*/
typedef struct vertex_t
{
	/*	number which will be given to current vertex	*/
	int restrictionNumber;
	Bool used;
	/*	list of edges with restrictions touching current vertex	*/	
	Queue* vertexRestrictions;
} RestrictionVertex;


/*	restriction graph edges		*/
typedef struct edge_t
{
	/*	vertice indexes are from 1 to (numVertices - 1)		*/
	int vertexIndex1;	
	int vertexIndex2;
	/*	restrictions numbers are from 1 to numRestrictions	*/
	int restriction1;
	int restriction2;
}RestrictionEdge;


/*	restriction graph	*/
typedef struct RestrictionGraph_t
{
	/*	array of vertices	*/
	RestrictionVertex* vertices;	
	int numVertices;
	/*	list of all graph restrictions (=edges)		*/
	Queue* restrictions;
	/*	array of scores (score for each restriction)	*/
	double* restrictionScores;
	int numRestrictions;
	int restrictionWithMaxScore;
}RestrictionsGraph;


/********************************************************************************/
/********************************************************************************/
/*	Function declarations	*/


/****************************************************************************/
/*	TestRestrictionsGraph													*/
void TestRestrictionsGraph();

/****************************************************************************
*
*       Function		createRestrictionsGraph  
*       Parameters		numVertices, restrictionScores, numRestrictions
*		Output			restrictionGraph
*       Description		creates restriction graph with {0,1,..,(numVertices-1}
*						set of vertices and empty set of restrictions
*       Comments		none
*
*****************************************************************************/
RestrictionsGraph* createRestrictionsGraph(int numVertices, 
								double* restrictionScores, int numRestrictions);


/****************************************************************************
*
*       Function		addRestriction  
*       Parameters		vertexIndex1,vertexIndex2,restriction1,restriction2
*		Output			false if some of the above numbers is out of scope,
*						else true
*       Description		creates restriction edge and inserts into the graph
*						queue and into 2 queues of 2 above vertices
*       Comments		none
*
*****************************************************************************/
Bool addRestriction(RestrictionsGraph* graph, int vertexIndex1, int vertexIndex2, 
					int restriction1, int restriction2);

// help function
void addRestrictionToVertex
		(RestrictionVertex* vertex, RestrictionEdge* newRestriction);



/****************************************************************************
*
*       Function		satisfyRestrictions  
*       Parameters		RestrictionsGraph
*		Output			false if graph restrictions can't be satisfied,
*						else true
*       Description		tries to satisfy restrictions in the "maximal" way,
*						if the restrictions are satisfied, then the
*						"satisfying" numbers are written in vertices
*       Comments		none
*
*****************************************************************************/
Bool satisfyRestrictions(RestrictionsGraph* graph);


/****************************************************************************
*
*       Function		getVertexRestriction  
*       Parameters		vertexIndex
*		Output			restriction number in vertex (-1 if was not initiated
*						or vertex is out of scope)
*       Description		getter
*       Comments		none
*
*****************************************************************************/
int getVertexRestriction(RestrictionsGraph* graph, int vertexIndex);

/****************************************************************************
*
*       Function		treatVertex  
*       Parameters		vertex index, RestrictionsGraph graph
*		Output			false if vertex component restrictions can't be satisfied,
*						else true
*       Description		tries to satisfy vertex component restrictions 
*						in the "maximal" way,
*						if the restrictions are satisfied, then the
*						"satisfying" numbers are written in component vertices
*       Comments		none
*
*****************************************************************************/
Bool treatVertex(int vertexIndex, RestrictionsGraph* graph);



/****************************************************************************
*
*       Function		treatVertexComponent
*       Parameters		vertex index, RestrictionsGraph graph
*		Output			score for current configuration satisfaction
*       Description		tries to satisfy vertex component restrictions 
*						in the "maximal" way,
*						if the restrictions are satisfied, then the
*						"satisfying" numbers are written in component vertices
*       Comments		assumes some number is assigned to vertex
*
*****************************************************************************/
double treatVertexComponent(int vertexIndex, RestrictionsGraph* graph);



/****************************************************************************
*
*       Function		setComponentUnused
*       Parameters		vertex index, RestrictionsGraph graph
*		Output			none
*       Description		sets all vertices in the component of given vertex unused
*       Comments		none
*
*****************************************************************************/
void setComponentUnused(int vertexIndex, RestrictionsGraph* graph);



/****************************************************************************
*
*       Function		destroyRestrictionsGraph  
*       Parameters		RestrictionsGraph
*		Output			none
*       Description		frees all the memory occupied by the RestrictionsGraph
*       Comments		none
*
*****************************************************************************/
void destroyRestrictionsGraph(RestrictionsGraph* graph);

#endif


