/* $Id: sequence.c,v 1.2 2008/08/03 08:15:48 superlink Exp $ */
/****************************************************************************/
/*  sequence.c											   					*/
/*  computes "gray code" sequence										 	*/
/*  Author: Andrei Anisenia													*/
/****************************************************************************/
#include <stdlib.h>
#include <math.h>
#include "sequence.h"
#include "sl_Auxilary.h"


/****************************************************************************
*
*       Function		power
*       Parameters	  	x,y	
*       Output			x^y  
*       Description		none
*       Comments		none
*
*****************************************************************************/
unsigned long power(int x, int y)
{
	int i;
	unsigned long result;

	result = 1;
	for(i=1; i<=y; i++)
		result *= x;
	return result;
}



/****************************************************************************
*
*       Function		compute_strait_to_array
*       Parameters	  	k - size of the "gray code" to generate
*						size - will be set to 2^k-1 (size of helper array)	
*       Output			helper array (basic piece of "gray code")   
*       Description		fills in the array of size = 2^k-1 with {1,2,1,3,1,2,1,4...}
*       Comments		k<=20, else it will fail
*
*****************************************************************************/
int* compute_strait_to_array(int k, unsigned long* size)
{
	unsigned long limit, i;
	int* arr;
	int times, j;
	
	
	if (k>20)
		return NULL;
	limit = power(2,k);
	arr = (int*)malloc((limit-1)*sizeof(int));
	for (i=0; i<limit-1; i++)
	{
		times = 0;
		j = i+1;
		while (j%2 == 0)
		{
			j = j/2;
			times++;
		}
		arr[i] = times+1;
	}
	*size = limit-1;
	return arr;
}



/****************************************************************************
*
*       Function		compute
*       Parameters	  	ph - helper struct
*						uF - pointer to function wich will get the sequence of "gray code"	
*       Output			none
*       Description		computes "gray code" given the helper struct ph
*						and passes it to uF (number by number)
*       Comments		none
*
*****************************************************************************/
void compute(help_alg* ph, void (*uF)(int) )
{	
	unsigned long j;
	int ind;

	for ((ph->i[0])=0; (ph->i[0])<=ph->size; ph->i[0]++)
	{
		/* iterating basic array of k						*/
		for (j=0; j<ph->size; j++)
			uF(ph->arr[j]-1);		
		
		/* loop for getting the "gray code" sequence		*/
		/* in the iterative manner (not recursive)			*/
		if (ph->m == 1) ph->flag = false;
		for (ind=1; ind<ph->m; ind++)
		{
			if (ph->i[ind] < ph->size)
			{
				ph->lim = ind*(ph->k) + ph->arr[ph->i[ind]];
				if (ph->lim > ph->n)
				{
					ph->flag = false;
					break;
				}
				uF(ph->lim-1);						
				ph->i[ind]++;
				break;
			}
			else
			{
				if (ind == (ph->m)-1)
				{
					ph->flag = false;
					break;
				}
				else
				{
					ph->i[ind] = 0;
					ph->i[0] = -1;
				}
			}
		}
		if (ph->flag == false) break;
	}
}



/****************************************************************************
*
*       Function		compute_algorithm
*       Parameters	  	n - size of the whole "gray code" to generate
*						uF - pointer to function wich will get the sequence of "gray code"	
*       Output			none   
*       Description		prepares helper struct (using function "compute_strait_to_array"
*						and calls the "gray code" generating function "compute"
*       Comments		none
*
*****************************************************************************/
void compute_algorithm(int n, void (*uF)(int) )
{
	help_alg H;
	int ind;
	int k,m;
	
	m = n/15;
	if ((m==0)||(n==15))	/* n<15 or n=15		*/
	{
		k = n;
		m = 1;
	}
	else		/* n>15							*/
	{
		m++;
		k = 15;
	}
	H.k = k;
	H.m = m;
	H.n = n;
	/* allocating arr of k						*/
	H.arr = compute_strait_to_array(k, &(H.size));
	/* allocating helper array i of m indexes	*/
	H.i =(int*)malloc(m*sizeof(int));
	for (ind=0; ind<m; ind++)
		H.i[ind] = 0;
	/* if flag=false => stop and exit			*/
	H.flag = true;
	
	compute(&H, uF);
	free(H.arr);
	free(H.i);
}





/****************************************************************************/
/*	alternative algorithm													*/

int* getGrayCodeArray(int k, unsigned long* size)
{
	unsigned long limit, i;
	int* arr;
	int j;
	
	limit = power(2,k);
	sl_myMalloc(arr, (limit-1)*sizeof(int));
	for (i=0; i<limit-1; i++)
	{
		arr[i] = 0;
		j = i + 1;
		while (j%2 == 0)
		{
			j = j >> 1;
			arr[i]++;
		}
	}
	*size = limit-1;
	return arr;
}



grayCodeGenerator* initiateGrayCodeGenerator(int size)
{
	grayCodeGenerator* generator;
	int i, addedNumber;
	
	/*	initiations	*/
	sl_myMalloc(generator, sizeof(grayCodeGenerator));
	if (size <= LIMIT)
	{
		/*	part of the required gray code in generated straight to array	*/
		generator->basicArray = getGrayCodeArray(size, &(generator->basicSize));
		generator->controlSize = 1;
	}
	else
	{
		generator->basicArray = getGrayCodeArray(LIMIT, &(generator->basicSize));
		generator->controlSize = 2;
	}

	/*	filling in control array	*/
	for(i = 0, addedNumber = 0; i < (generator->controlSize); i++, addedNumber += LIMIT)
	{
		generator->controlArray[i].addedNumber = addedNumber;
		generator->controlArray[i].currentBasicIndex = 0;
		generator->controlArray[i].currentBasicNumberPtr = generator->basicArray;
	}

	generator->currentControlIndex = 0;
	generator->currentControlNodePtr = generator->controlArray;
	
	return generator;
}



int getNextGrayCodeNumber(grayCodeGenerator* generator)
{
	int number, needChangeControlNode;
	controlNode* controlNodePtr;

	/*	finding required gray code number	*/
	controlNodePtr = generator->currentControlNodePtr;
	number = controlNodePtr->addedNumber + (*(controlNodePtr->currentBasicNumberPtr));

	/*	updating gray code generator structure in order to move to the next number	*/
	/*	moving forward in basic array	*/
	controlNodePtr->currentBasicIndex++;
	controlNodePtr->currentBasicNumberPtr++;

	needChangeControlNode = generator->currentControlIndex;
	if (controlNodePtr->currentBasicIndex == generator->basicSize)
	{
		/*	reached the end of the basic array => go to the start of basic array	*/
		controlNodePtr->currentBasicIndex = 0;
		controlNodePtr->currentBasicNumberPtr = generator->basicArray;
		needChangeControlNode++;
	}
	/*	we need to change control node when we reached the end of the basic array	*/
	/*	or when current control index is 1	*/
	if ( (generator->controlSize > 1) && (needChangeControlNode) )
	{
		generator->currentControlIndex = 1 - generator->currentControlIndex;
		generator->currentControlNodePtr = 
					generator->controlArray + generator->currentControlIndex;
	}

	return number;
}



void destroyGrayCodeGenerator(grayCodeGenerator* generator)
{
	free(generator->basicArray);
	free(generator);
}



